{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE FlexibleContexts #-}

-- | Syntax-directed automata constructions based on formulas.
module Formula.Automata (
    mkGTT,
    mkRR1,
    mkRR2,
    mkFormula,
    M,
    nextI,
    tellM,
    evalM,
    fit,
) where

import Formula.Type
import qualified GTT
import qualified RRn
import qualified TA
import qualified Data.Rewriting.Rule as R
import qualified Util.Set as S
import qualified Data.Map as M
import Util.Triple
import Util.Signature
import Control.Monad.RWS.Strict
import Data.List

-- | State of the 'M' monad. Contains the cache, and index of next inference
data Data s f = Data {
    gtts :: !(M.Map (Signature s f, GTT_rel) (GTT.GTT f Int)),
    rr1s :: !(M.Map (Signature s f, RR1_rel) (RRn.RRn s f Int)),
    rr2s :: !(M.Map (Signature s f, RR2_rel) (RRn.RRn s f Int)),
    fms  :: !(M.Map (Signature s f, Formula) (Int, RRn.RRn s f Int, S.Set Int)),
    idx  :: !Int
}

-- | Auxiliary monad for formula manipulation, supporting caching of
-- subformulas and logging.
newtype M s f a = M { runM :: RWS () (Endo [Step]) (Data s f) a }
    deriving (Functor, Applicative, Monad, MonadState (Data s f), MonadWriter (Endo [Step]))

-- | Run in auxiliary monad.
evalM :: M s f a -> (a, [Step])
evalM m = (a, appEndo ss []) where
    (a, _, ss) = runRWS (runM m) () data0

    data0 = Data {
        gtts = M.empty,
        rr1s = M.empty,
        rr2s = M.empty,
        fms  = M.empty,
        idx  = 0
    }

tellM :: Step -> M s f ()
tellM s = tell $ Endo (s :)

nextI :: M s f Int
nextI = do
    i <- gets idx
    modify (\d -> d { idx = i+1 })
    return i

mkTRS :: (Ord f, Ord v) => [[R.Rule f v]] -> [Trs] -> [R.Rule f v]
mkTRS trss ts = S.nubOrd $ ts >>= getTRS
  where
    rev r = R.Rule{ R.lhs = R.rhs r, R.rhs = R.lhs r }
    getTRS (Fwd i) = getTRS' i
    getTRS (Bwd i) = map rev (getTRS' i)
    getTRS' i
        | i < 0 || i >= length trss = error $ "TRS index " ++ show i ++
            " out of bounds (0..<" ++ show (length trss) ++ ")"
        | otherwise = trss !! i

-- | Construct GTT corresponding to description.
mkGTT :: (Ord s, Ord f, Ord v) => Signature s f -> [[R.Rule f v]] -> GTT_rel -> M s f (GTT.GTT f Int)
mkGTT sig trss = go' where
    go' g = do
        gtts' <- gets gtts
        case (sig, g) `M.lookup` gtts' of
            Just g' -> return g'
            Nothing -> do
                g' <- GTT.trim . GTT.unique <$> go g
                modify (\d -> d { gtts = M.insert (sig, g) g' (gtts d) })
                return g'

    go (ARoot ts) = return $ GTT.relabel $ GTT.fromTRSI sig (mkTRS trss ts)
    go (GInv g) = GTT.ginv <$> go' g
    go (AUnion g1 g2) = GTT.relabel <$> (GTT.aunion <$> go' g1 <*> go' g2)
    go (ATrancl g) = GTT.relabel . GTT.atrancl . GTT.anchor <$> go' g
    go (GTrancl g) = GTT.gtrancl <$> go' g
    go (AComp g1 g2) = GTT.relabel <$> (gtt_acomp <$> go' g1 <*> go' g2)
    go (GComp g1 g2) = GTT.relabel <$> (GTT.gcomp <$> go' g1 <*> go' g2)
    -- derived
    go (GSteps ts) = go' (GTrancl (ARoot ts))

    gtt_acomp g1 g2 = GTT.acomp (GTT.anchor g1) (GTT.anchor g2)

-- | Construct /RR1/ automaton corresponding to description.
mkRR1 :: (Ord s, Ord f, Ord v) => Signature s f -> [[R.Rule f v]] -> RR1_rel -> M s f (RRn.RRn s f Int)
mkRR1 sig trss = go' where
    go' s = do
        rr1s' <- gets rr1s
        case (sig, s) `M.lookup` rr1s' of
            Just s' -> return s'
            Nothing -> do
                s' <- simpRRn . RRn.unique <$> go s
                modify (\d -> d { rr1s = M.insert (sig, s) s' (rr1s d) })
                return s'

    go2 = mkRR2 sig trss
    go R1Terms = return $ RRn.fromTA sig (TA.relabel $ TA.fromMSSig sig (sorts sig))
    go (R1NF ts) =
        return $ RRn.fromTA sig (TA.relabel $ TA.normalForm sig (mkTRS trss ts))
    go (R1Inf r) = RRn.relabel . RRn.infinite <$> go2 r
    go (R1Proj i r) = RRn.project [i] <$> go2 r
    go (R1Union s1 s2) = RRn.relabel <$> (RRn.union <$> go s1 <*> go' s2)
    go (R1Inter s1 s2) = RRn.relabel <$> (RRn.intersection <$> go' s1 <*> go' s2)
    go (R1Diff s1 s2) = RRn.relabel <$> (RRn.difference <$> go' s1 <*> go' s2)
    -- derived
    go (R1Fin r) = go' (R1Diff R1Terms (R1Inf r))

-- | Construct /RR2/ automaton corresponding to description.
mkRR2 :: (Ord s, Ord f, Ord v) => Signature s f -> [[R.Rule f v]] -> RR2_rel -> M s f (RRn.RRn s f Int)
mkRR2 sig trss = go' where
    go' r = do
        rr2s' <- gets rr2s
        case (sig, r) `M.lookup` rr2s' of
            Just r' -> return r'
            Nothing -> do
                r' <- simpRRn' . RRn.unique <$> go r
                modify (\d -> d { rr2s = M.insert (sig, r) r' (rr2s d) })
                return r'

    go1 = mkRR1 sig trss
    t1 = go1 R1Terms
    go (R2GTT g w x) = RRn.relabel <$>
       (RRn.intersection <$> (RRn.prod <$> t1 <*> t1)
                         <*> (RRn.parallel w x . RRn.fromGTT sig <$>
                              mkGTT sig trss g))
    go (R2Diag s) = RRn.diag <$> go1 s
    go (R2Prod s1 s2) = RRn.relabel <$> (RRn.prod <$> go1 s1 <*> go1 s2)
    go (R2Inv r) = RRn.unsafeLift (TA.map (\[f, g] -> [g, f]) id) <$> go' r
    go (R2Union r1 r2) = RRn.relabel <$> (RRn.union <$> go' r1 <*> go' r2)
    go (R2Inter r1 r2) = RRn.relabel <$> (RRn.intersection <$> go' r1 <*> go' r2)
    go (R2Diff r1 r2) = RRn.relabel <$> (RRn.difference <$> go' r1 <*> go' r2)
    go (R2Comp r1 r2) = RRn.relabel . RRn.project [0,2] . simpRRn <$>
       (RRn.intersection <$> (simpRRn <$> (RRn.prod <$> go' r1 <*> t1))
                         <*> (simpRRn <$> (RRn.prod <$> t1 <*> go' r2)))
    -- derived
    go R2Eq = go' (R2Diag R1Terms)
    go (R2Reflc r) = go' (R2Union r R2Eq)
    go (R2Step ts) = go' (R2GTT (ARoot ts) PAny ESingle)
    go (R2StepEq ts) = go' (R2Reflc (R2Step ts))
    go (R2Steps ts) = go' (R2GTT (GSteps ts) PAny EStrictParallel)
    go (R2StepsEq ts) = go' (R2GTT (GSteps ts) PAny EParallel)
    go (R2StepsNF ts) = go' (R2Inter (R2StepsEq ts) (R2Prod R1Terms (R1NF ts)))
    go (R2ParStep ts) = go' (R2GTT (ARoot ts) PAny EParallel)
    go (R2RootStep ts) = go' (R2GTT (ARoot ts) PRoot ESingle)
    go (R2RootStepEq ts) = go' (R2Reflc (R2RootStep ts))
    go (R2RootSteps ts) = go' (R2GTT (ATrancl (ARoot ts)) PRoot ESingle)
    go (R2RootStepsEq ts) = go' (R2Reflc (R2RootSteps ts))
    go (R2NonRootStep ts) = go' (R2GTT (ARoot ts) PNonRoot ESingle)
    go (R2NonRootStepEq ts) = go' (R2Reflc (R2NonRootStep ts))
    go (R2NonRootSteps ts) = go' (R2GTT (GSteps ts) PNonRoot EStrictParallel)
    go (R2NonRootStepsEq ts) = go' (R2GTT (GSteps ts) PNonRoot EParallel)
    go (R2Meet ts) = go' (R2GTT (GComp (GInv (GSteps ts)) (GSteps ts)) PAny EParallel)
    go (R2Join ts) = go' (R2GTT (GComp (GSteps ts) (GInv (GSteps ts))) PAny EParallel)

-- | Construct /RRn/ automaton corresponding to formula.
--
-- This is a naive, syntax-driven translation. It returns the set
-- of free variables along with the RRn automaton, which is restricted
-- to those variables.
--
-- If the boolean argument is false, no log entries are added.
mkFormula :: (Ord s, Ord f, Ord v) => Bool -> Signature s f -> [Signature s f]
          -> [[R.Rule f v]] -> Formula -> M s f (Int, RRn.RRn s f Int, S.Set Int)
mkFormula log sig extSig trss = go' sig where
    go' sig f = do
        fms' <- gets fms
        case (sig, f) `M.lookup` fms' of
            Just f' -> return f'
            Nothing -> do
                (inf, r, vs) <- go sig f
                i <- nextI
                let f' = (i, simpRRn' . RRn.unique $ r, vs)
                modify (\d -> d { fms = M.insert (sig, f) f' (fms d) })
                tellM (i, inf, f, mkInfo f')
                return f'

    mkInfo (_, r, _) = [
        Size (S.size (TA.states (RRn.ta r)))
             (length (TA.trans (RRn.ta r)))
             (length (TA.eps (RRn.ta r)))
        | log]

    go sig (FRR1 s x) = do
        rs <- mkRR1 sig trss s
        return (IRR1 s x, rs, S.singleton x)
    go sig (FRR2 r x y) = do
        rr <- mkRR2 sig trss (if x > y then R2Inv r else r)
        return $ case x `compare` y of
            LT -> (IRR2 r x y, rr, S.fromAscList [x, y])
            EQ -> (IRR2 r x y, RRn.subst [0, 0] rr, S.singleton x)
            GT -> (IRR2 r x y, rr, S.fromAscList [y, x])
    go sig (FAnd []) = go sig FTrue
    go sig (FAnd [f1, FNot f2]) = do
        r1@(i1, _, _) <- go' sig f1
        r2 <- go' sig f2
        -- we skip the automaton produced by the negation
        (i3, _, _) <- mkFormula False sig extSig trss (FNot f2)
        let ([r1', r2'], vs) = fit [r1, r2]
        return (IAnd [i1, i3], RRn.relabel (RRn.difference r1' r2'), vs)
    go sig (FAnd fs) = do
        rs <- mapM (go' sig) fs
        let (rs', vs) = fit rs
        return (IAnd (map fst3 rs),
            foldl1 (\r1 r2 -> RRn.relabel (RRn.intersection r1 r2)) rs', vs)
    go sig (FOr []) = go sig FFalse
    go sig (FOr fs) = do
        rs <- mapM (go' sig) fs
        let (rs', vs) = fit rs
        return (IOr (map fst3 rs),
            foldl1 (\r1 r2 -> RRn.relabel (RRn.union r1 r2)) rs', vs)
    go sig (FNot f) = do
        (i, r, vs) <- go' sig f
        return (INot i,
            RRn.relabel (RRn.difference (mkFull (RRn.signature r) (RRn.arity r)) r), vs)
    go sig (FExists f) = do
        (i, r, vs) <- go' sig f
        return $ case S.minView vs of
            Just (0, vs') -> (IExists i,
                              RRn.project [1..RRn.arity r - 1] r,
                              S.mapMonotonic pred vs')
            _             -> (IExists i,
                              r, S.mapMonotonic pred vs)
    go sig (FForall f) = go sig (FNot (FExists (FNot f)))
    go _ FTrue = return (IAnd [], RRn.relabel (RRn.rr0 True), S.empty)
    go _ FFalse = return (IOr [], RRn.relabel (RRn.rr0 False), S.empty)
    go sig (FExtendSig f n) = go (extend sig n) f
    go _ FRestrict{} = error "mkFormula: expected formula without restrictions (FRestrict)"

    extend sig n = concatMap (take n . filter (`notElem` sig)) extSig ++ sig

mkFull :: (Ord s, Ord f) => Signature s f -> Int -> RRn.RRn s f Int
mkFull _ 0 = RRn.relabel (RRn.rr0 True)
mkFull sig 1 = RRn.relabel (RRn.fromTA sig (TA.fromMSSig sig (sorts sig)))
mkFull sig n = simpRRn (RRn.prod (mkFull sig (n-1)) (mkFull sig 1))

-- | Cylindrifies the given relations, such that they fit together.
-- After 'fit' the RRn encoding of the returned tree automata,
-- is compatible to be combined by a binary relation (i.e. intersection, union, difference..)
fit :: (Ord s, Ord f) => [(Int, RRn.RRn s f Int, S.Set Int)] -> ([RRn.RRn s f Int], S.Set Int)
fit rs = (map adj rs, vs') where
    sig = S.nubOrd . concat $ map (RRn.signature . snd3) rs
    vs' = foldl' S.union S.empty (map thd3 rs)
    adj (_, r, vs) = RRn.relabel (simpRRn (RRn.project is (RRn.prod r rf)))
      where
        ws = vs' `S.difference` vs
        rf = mkFull sig (S.size ws)
        is = flip map (S.toAscList vs') $ \j ->
            case j `S.lookupIndex` ws of
                Just i  -> S.size vs + i
                Nothing -> j `S.findIndex` vs

-- | Simplify intermediate RRn automaton
simpRRn :: (Ord s, Ord f, Ord q) => RRn.RRn s f q -> RRn.RRn s f Int
simpRRn = simpRRn' . RRn.relabel

-- | simpRRn spezialized to 'Int' states
simpRRn' :: (Ord s, Ord f) => RRn.RRn s f Int -> RRn.RRn s f Int
simpRRn' = RRn.trim . RRn.epsFree . RRn.unsafeLift (TA.collapseEps)
