{-# LANGUAGE FlexibleContexts #-}

-- | FORT compatible formula parser
module Formula.FORT (
    parseFormula,
    parseCertifiableFormula,
    fortFormula,
) where

import Formula.Type
import qualified Util.OpenTerms as Open

import Control.Monad (msum, void)
import Control.Monad.Reader (MonadReader, asks, local, msum, runReader, void)
import Data.List (elemIndex)
import Text.Parsec (ParseError, ParsecT, (<?>), (<|>))
import qualified Text.Parsec as P
import Util.Parsec


-- | The @ParseContext@ contains the list of idices of TRSs and a signature
-- extension function.
--
-- Note: the default system used for relations and properties is the union
-- of all TRSs in the @trss@ list.
data ParseContext = ParseContext
    { trss :: [Int] -- ^ List of indexes of TRSs
    , sigExtension :: Open.Property -> Int -- ^ Signature extension function
    }

setTrss :: [Int] -> ParseContext -> ParseContext
setTrss ts ctxt = ctxt { trss = ts }

parseFormula ::
    (Open.Property -> Int) -> [Int] -> String -> Either ParseError Formula
parseFormula extend trss inp =
    runReader (P.runParserT fortFormula () "<formula>" inp) ctxt
  where
    ctxt = ParseContext { trss = trss, sigExtension = extend}

parseCertifiableFormula ::
    (Open.Property -> Int) ->
    [Int] ->
    String ->
    Either ParseError (Either (Open.Property, Formula) Formula)
parseCertifiableFormula extend trss inp =
    runReader (P.runParserT fortCertFormula () "<formula>" inp) ctxt
  where
    ctxt =
        ParseContext
            { trss = trss
            , sigExtension = extend
            }

fortCertFormula ::
    (MonadReader ParseContext m) =>
    ParsecT String u m (Either (Open.Property, Formula) Formula)
fortCertFormula =
    P.spaces *> (Left <$> nonGroundProperty <|> Right <$> formula []) <* P.eof

-- | Parser for FORT formula.
fortFormula :: (MonadReader ParseContext m) => ParsecT String u m Formula
fortFormula = P.spaces *> formula [] <* P.eof

toks :: Monad m => [String] -> ParsecT String u m ()
toks ts = void $ msum [P.try (P.string t) <* P.spaces | t <- ts]

name :: Monad m => ParsecT String u m String
name = (:) <$> P.letter <*> P.many P.alphaNum <* P.spaces

names :: Monad m => ParsecT String u m [String]
names = (:) <$> name <*> P.many (toks [","] *> name)

nat :: Monad m => ParsecT String u m Int
nat = read <$> P.many1 P.digit <* P.spaces

ixs :: Monad m => ParsecT String u m [Int]
ixs = P.between (toks ["["]) (toks ["]"]) (P.sepBy1 nat (toks [","]))

-- <var>
var :: Monad m => [String] -> ParsecT String u m Int
var vs = do
    p <- P.getPosition
    v <- name
    case elemIndex v vs of
        Just i -> return i
        Nothing ->
            sysUnexpectedAt p $
                "variable `" ++ v ++ "': closed formula expected"

-- <formula>
formula :: (MonadReader ParseContext m) => [String] -> ParsecT String u m Formula
formula vs = iff
  where
    fAnd [x] = x; fAnd xs = FAnd xs
    fOr [x] = x; fOr xs = FOr xs
    iff = foldr1 fIff <$> ((:) <$> imp <*> P.many (toks ["<=>", "⇔"] *> imp))
    imp = foldr1 fImplies <$> ((:) <$> dis <*> P.many (toks ["=>", "⇒"] *> dis))
    dis = fOr <$> ((:) <$> con <*> P.many (toks ["|", "∨"] *> con))
    con = fAnd <$> ((:) <$> neg <*> P.many (toks ["&", "∧"] *> neg))
    neg = FNot <$> (toks ["~", "¬"] *> neg) <|> bas
    bas = toks ["("] *> iff <* toks [")"]
        <|> property vs
        <|> quantified
        <|> restricted
        <|> flip FExtendSig <$> (toks ["{+"] *> nat <* toks ["}"]) <*> bas
        <|> flip FRR2 <$> var vs <*> relation <*> var vs
    quantified = do
        q <- quantifier
        ws <- names
        toks ["("]
        f <- formula (reverse ws ++ vs)
        toks [")"]
        return $ iterate q f !! length ws
    restricted = do
      ts <- ixs
      local (setTrss [0..length ts - 1]) $ flip FRestrict ts <$> bas

-- <relation>
relation :: MonadReader ParseContext m => ParsecT String u m RR2_rel
relation = asks trss >>= rel
  where
    rel ts
      = R2RootSteps fwd <$ toks ["->e+", "->ε+", "→e+", "→ε+"]
      <|> R2RootStepsEq fwd <$ toks ["->e*", "->ε*", "→e*", "→ε*"]
      <|> R2RootStepEq fwd <$ toks ["->e=", "->ε=", "→e=", "→ε="]
      <|> R2RootStep fwd <$ toks ["->e", "->ε", "→e", "→ε"]
      <|> R2NonRootSteps fwd <$ toks ["->be+", "->⌄ε+", "→be+", "→⌄ε+"]
      <|> R2NonRootStepsEq fwd <$ toks ["->be*", "->⌄ε*", "→be*", "→⌄ε*"]
      <|> R2NonRootStepEq fwd <$ toks ["->be=", "->⌄ε=", "→be=", "→⌄ε="]
      <|> R2NonRootStep fwd <$ toks ["->be", "->⌄ε", "→be", "→⌄ε"]
      <|> R2Steps fwd <$ toks ["->+", "→+"]
      <|> R2StepsEq fwd <$ toks ["->*", "→*"]
      <|> R2StepEq fwd <$ toks ["->=", "→="]
      <|> R2StepsNF fwd <$ toks ["->!", "→!"]
      <|> R2Step fwd <$ toks ["->", "→"]
      <|> R2ParStep fwd <$ toks ["-||->", "⇻"]
      <|> R2StepsEq conv <$ toks ["<->*", "↔*"]
      <|> R2Step conv <$ toks ["<->", "↔"]
      <|> R2RootSteps bwd <$ toks ["+e<-", "+ε<-", "+e←", "+ε←"]
      <|> R2RootStepsEq bwd <$ toks ["*e<-", "*ε<-", "*e←", "*ε←"]
      <|> R2RootStepEq bwd <$ toks ["=e<-", "=ε<-", "=e←", "=ε←"]
      <|> R2RootStep bwd <$ toks ["e<-", "ε<-", "e←", "ε←"]
      <|> R2NonRootSteps bwd <$ toks ["+be<-", "+⌄ε<-", "+be←", "+⌄ε←"]
      <|> R2NonRootStepsEq bwd <$ toks ["*be<-", "*⌄ε<-", "*be←", "*⌄ε←"]
      <|> R2NonRootStepEq bwd <$ toks ["=be<-", "=⌄ε<-", "=be←", "=⌄ε←"]
      <|> R2NonRootStep bwd <$ toks ["be<-", "⌄ε<-", "be←", "⌄ε←"]
      <|> R2Steps bwd <$ toks ["+<-", "+←"]
      <|> R2StepsEq bwd <$ toks ["*<-", "*←"]
      <|> R2StepEq bwd <$ toks ["=<-", "=←"]
      <|> R2StepsNF bwd <$ toks ["!<-", "!←", "<-!", "←!"]
      <|> R2ParStep bwd <$ toks ["<-||-", "⇺"]
      <|> R2Step bwd <$ toks ["<-", "←"]
      <|> R2Join fwd <$ toks ["join", "↓"]
      <|> R2Meet fwd <$ toks ["meet", "↑"]
      <|> R2Eq <$ toks ["="]
      <?> "relation"
      where
        fwd = map Fwd ts
        bwd = map Bwd ts
        conv = fwd ++ bwd

-- <quantifier>
quantifier :: Monad m => ParsecT String u m (Formula -> Formula)
quantifier = FForall <$ toks ["forall", "∀"]
         <|> FExists <$ toks ["exists", "∃"]


-- <property>
property :: (MonadReader ParseContext m) => [String] -> ParsecT String u m Formula
property vs = P.try (property1 vs) <|> groundProperty <|> nonGroundProperty'
  where
    nonGroundProperty' = nonGroundProperty >>= extend
    extend (prop, f) = do
        extF <- asks sigExtension
        return $ FExtendSig f (extF prop)

groundProperty :: MonadReader ParseContext m => ParsecT String u m Formula
groundProperty = asks trss >>= go
  where
    go ts = fWN ts <$ toks ["WN"] -- not ground, but needs no sig-extension
        <|> fSN ts <$ toks ["SN"] -- not ground, but needs no sig-extension
        <|> fGUNC ts <$ toks ["GUNC"]
        <|> fGCR ts <$ toks ["GCR"]
        <|> fGNFP ts <$ toks ["GNFP"]
        <|> fGUN ts <$ toks ["GUNR", "GUN"]
        <|> fGWCR ts <$ toks ["GWCR"]
        <|> fGSCR ts <$ toks ["GSCR"]
        <|> ( fGCOM <$ toks ["GCOM", "GCom"]
          <|> fGLCOM <$ toks ["GLCOM", "GLCom"]
          <|> fGNE <$ toks ["GNE"]
          <|> fGCE <$ toks ["GCE"]
            ) <*> (toks ["("] *> ixs)
              <*> (toks [","] *> ixs <* toks [")"])

nonGroundProperty ::
  MonadReader ParseContext m => ParsecT String u m (Open.Property, Formula)
nonGroundProperty = asks trss >>= go
  where
    go ts = unc <$ toks ["UNC"]
        <|> cr <$ toks ["CR"]
        <|> nfp <$ toks ["NFP"]
        <|> un <$ toks ["UNR", "UN"]
        <|> wcr <$ toks ["WCR"]
        <|> scr <$ toks ["SCR"]
        <|> ( com <$ toks ["COM", "Com"]
            <|> ne <$ toks ["NE"]
            <|> ce <$ toks ["CE"]
            <|> lcom <$ toks ["LCOM", "LCom"]
            ) <*> (toks ["("] *> ixs) <*> (toks [","] *> ixs <* toks [")"])
      where
        cr = (Open.CR, fGCR ts)
        ne rs ss = (Open.NE, fGNE rs ss)
        ce rs ss = (Open.CE, fGCE rs ss)
        wcr = (Open.WCR, fGWCR ts)
        scr = (Open.SCR, fGSCR ts)
        un = (Open.UNR, fGUN ts)
        unc = (Open.UNC, fGUNC ts)
        nfp = (Open.NFP, fGNFP ts)
        lcom is js = (Open.COM, fGLCOM is js)
        com is js = (Open.COM, fGCOM is js)

-- | Properties which take a variable as argument
--
-- Example: forall t (SN(t))
property1 :: MonadReader ParseContext m => [String] -> ParsecT String u m Formula
property1 vs = asks trss >>= go
  where
    go ts = prop <*> (toks ["("] *> var vs <* toks [")"])
        <|> finiteness
             <*> (toks ["("] *> (formula2 <* toks [","] <|> return (R2Step fwd)))
             <*> (var vs <* toks [")"])
      where
        prop = cr <$ toks ["CR"]
             <|> wcr <$ toks ["WCR"]
             <|> wn <$ toks ["WN"]
             <|> nfp <$ toks ["NFP"]
             <|> sn <$ toks ["SN"]
             <|> nf <$ toks ["NF"]
             <|> scr <$ toks ["SCR"]
             <|> un <$ toks ["UNR", "UN"]
        finiteness
          = inf <$ toks ["INF", "Inf"]
          <|> fin <$ toks ["FIN", "Fin"]

        fwd = map Fwd ts
        bwd = map Bwd ts
        conv = fwd ++ bwd

        r = R2Step fwd
        rs = R2StepsEq fwd
        cr v =
          FForall (FForall
                   (fImplies
                    (FAnd [FRR2 rs (v + 2) 0, FRR2 rs (v + 2) 1])
                    (FRR2 (R2Comp rs (R2Inv rs)) 0 1)
                   ))
        wcr v =
          FForall (FForall
                   (fImplies
                    (FAnd [FRR2 r (v + 2) 0, FRR2 r (v + 2) 1])
                    (FRR2 (R2Comp rs (R2Inv rs)) 0 1)
                   ))
        wn v = FExists (FRR2 (R2StepsNF fwd) (v + 1) 0)
        nfp v =
          FForall
            (fImplies
             (FRR2 (R2Inter (R2StepsEq conv) (R2Prod R1Terms (R1NF fwd))) (v + 1) 0)
             (FRR2 (R2StepsEq fwd) (v + 1) 0)
            )
        sn v =
          FAnd [ FForall
                   (fImplies
                    (FRR2 (R2StepsEq fwd) (v + 1) 0)
                    (FNot (FRR2 (R2Steps fwd) 0 0))
                   )
               , FRR1 (R1Fin (R2StepsEq fwd)) v
               ]
        nf v = FRR1 (R1NF fwd) v
        scr v =
          let (r, rs) = (R2Step fwd, R2StepsEq fwd)
           in FForall (FForall
                       (fImplies
                        (FAnd [FRR2 r (v + 2) 0, FRR2 r (v + 2) 1])
                        (FRR2 (R2Comp (R2Reflc r) (R2Inv rs)) 0 1)
                       ))
        un v =
          let rs = R2StepsNF fwd
           in FForall (FForall
                       (fImplies
                        (FAnd [FRR2 rs (v + 2) 0, FRR2 rs (v + 2) 1])
                        (FRR2 R2Eq 0 1)
                       ))
        fin r v = FRR1 (R1Fin r) v
        inf r v = FRR1 (R1Inf r) v

-- <formula2>
formula2 :: MonadReader ParseContext m => ParsecT String u m RR2_rel
formula2 = iff
  where
    iff = foldr1 r2Iff <$> ((:) <$> imp <*> P.many (toks ["<=>", "⇔"] *> imp))
    imp = foldr1 r2Implies <$> ((:) <$> dis <*> P.many (toks ["=>", "⇒"] *> dis))
    dis = foldr1 R2Union <$> ((:) <$> con <*> P.many (toks ["|", "∨"] *> con))
    con = foldr1 R2Inter <$> ((:) <$> neg <*> P.many (toks ["&", "∧"] *> neg))
    neg = r2Neg <$> (toks ["~", "¬"] *> neg) <|> bas
    bas = toks ["("] *> iff <* toks [")"] <|> relation

---------------------------------------------------
-- Macros

fWN :: [Int] -> Formula
fWN ts = FForall (FExists (FAnd [FRR2 (R2StepsEq fwd) 1 0, FRR1 (R1NF fwd) 0]))
  where
    fwd = map Fwd ts

fSN :: [Int] -> Formula
fSN ts =
  FAnd [ FForall (FNot (FRR2 (R2Steps fwd) 0 0))
       , FForall (FNot (FRR1 (R1Inf (R2StepsEq fwd)) 0))
       ]
  where
    fwd = map Fwd ts

-- | Ground-normalization-equivalence
fGNE :: [Int] -> [Int] -> Formula
fGNE rs ss =
    FForall (FForall
                (fIff
                    (FRR2 (R2StepsNF (map Fwd rs)) 0 1)
                    (FRR2 (R2StepsNF (map Fwd ss)) 0 1)
                ))

-- | Ground-conversion-equivalence
fGCE :: (Foldable t1, Foldable t2) => t1 Int -> t2 Int -> Formula
fGCE rs ss =
    FForall (FForall
                (fIff
                    (FRR2 (R2StepsEq (concatMap (\t -> [Fwd t, Bwd t]) rs)) 0 1)
                    (FRR2 (R2StepsEq (concatMap (\t -> [Fwd t, Bwd t]) ss)) 0 1)
                ))

fGLCOM' :: (RR2_rel, RR2_rel) -> (RR2_rel, RR2_rel) -> Formula
fGLCOM' (r, rs) (s, ss) =
  FForall (FForall
           (fImplies
            (FRR2 (R2Comp (R2Inv s) r) 0 1)
            (FRR2 (R2Comp rs (R2Inv ss)) 0 1)
           ))

fGCOM' :: RR2_rel -> RR2_rel -> Formula
fGCOM' rs ss =
  FForall (FForall
           (fImplies
            (FRR2 (R2Comp (R2Inv ss) rs) 0 1)
            (FRR2 (R2Comp rs (R2Inv ss)) 0 1)
           ))

-- | Ground confluence
fGCR :: [Int] -> Formula
fGCR ts =
    FForall ( FForall
              ( fImplies
                (FRR2 (R2Comp (R2Step bwd) (R2StepsEq fwd)) 0 1)
                (FRR2 (R2Comp (R2StepsEq fwd) (R2StepsEq bwd)) 0 1)
              ))
  where
    fwd = map Fwd ts
    bwd = map Bwd ts

-- | Ground local-confluence
fGWCR :: [Int] -> Formula
fGWCR ts = fGLCOM' (r, rs) (r, rs)
  where
    fwd = map Fwd ts
    r = R2Step fwd
    rs = R2StepsEq fwd

-- | Ground strong-confluence
fGSCR :: [Int] -> Formula
fGSCR ts =
    FForall (FForall
                (fImplies
                    (FRR2 (R2Comp (R2Inv r) r) 0 1)
                    (FRR2 (R2Comp (R2Reflc r) (R2Inv rs)) 0 1)
                ))
  where
    fwd = map Fwd ts
    r = R2Step fwd
    rs = R2StepsEq fwd

-- | Ground unique normal forms w.r.t. reduction
fGUN :: [Int] -> Formula
fGUN ts =
    FForall (FForall
                (fImplies
                    (FRR2 (R2Comp (R2Inv rs) rs) 0 1)
                    (FRR2 R2Eq 0 1)
                ))
  where
    fwd = map Fwd ts
    rs = R2StepsNF fwd

-- | Ground unique normal forms w.r.t. conversion
fGUNC :: [Int] -> Formula
fGUNC ts =
    FForall (FForall
                (fImplies
                    (FAnd [ FRR2 (R2Prod (R1NF fwd) (R1NF fwd)) 0 1
                          , FNot (FRR2 R2Eq 0 1)])
                    (FNot (FRR2 (R2StepsEq conv) 0 1))
                ))
  where
    fwd = map Fwd ts
    bwd = map Bwd ts
    conv = fwd ++ bwd

-- | Ground normal form property
fGNFP :: [Int] -> Formula
fGNFP ts =
    FForall (FForall
                (fImplies
                    (FRR2 (R2Comp (R2Step bwd) (R2StepsNF fwd)) 1 0)
                    (FRR2 (R2StepsEq fwd) 1 0)
                ))
  where
    fwd = map Fwd ts
    bwd = map Bwd ts

-- | Ground local commutation
fGLCOM :: [Int] -> [Int] -> Formula
fGLCOM is js =
  fGLCOM'
    (R2Step (map Fwd is), R2StepsEq (map Fwd is))
    (R2Step (map Fwd js), R2StepsEq (map Fwd js))

-- | Ground commutation
fGCOM :: [Int] -> [Int] -> Formula
fGCOM is js = fGCOM' (R2StepsEq (map Fwd is)) (R2StepsEq (map Fwd js))
