{-# LANGUAGE OverloadedStrings #-}

-- | Certificate output.
module Formula.Output (
    Output (..)
) where

import Formula.Type
import Prettyprinter

-- | Type class for certificate output.
class Output a where
    -- | Render a value in a s-expression based machine-readable way.
    output :: a -> Doc ann

sexp :: Doc ann -> [Doc ann] -> Doc ann
sexp n [] = "(" <> n <> ")"
sexp n xs = "(" <> n <+> nest 1 (sep xs <> ")")

slist' :: [Doc ann] -> Doc ann
slist' [] = "(" <> ")"
slist' xs = "(" <> nest 1 (sep xs <> ")")

slist :: Output p => [p] -> Doc ann
slist ps = slist' (map output ps)

instance Output Int where
    output = pretty

instance Output Trs where
    output (Fwd i) = output i
    output (Bwd i) = output i <> "-"

instance Output GTT_rel where
    output (ARoot ts)        = sexp "root-step" [slist ts]
    output (GInv g)          = sexp "inverse" [output g]
    output (AUnion g1 g2)    = sexp "aunion" [output g1, output g2]
    output (ATrancl g)       = sexp "atc" [output g]
    output (GTrancl g)       = sexp "gtc" [output g]
    output (AComp g1 g2)     = sexp "acomp" [output g1, output g2]
    output (GComp g1 g2)     = sexp "gcomp" [output g1, output g2]
    -- derived
    output (GSteps ts)       = sexp "gstep+" [slist ts]

instance Output Pos_step where
    output PRoot             = "e" -- "root"
    output PNonRoot          = ">" -- "non-root"
    output PAny              = ">=" -- "any"

instance Output Ext_step where
    output ESingle           = "1" -- "single"
    output EParallel         = ">=" -- "parallel"
    output EStrictParallel   = ">" -- "strictparallel"

instance Output RR1_rel where
    output R1Terms           = sexp "terms" []
    output (R1NF ts)         = sexp "nf" [slist ts]
    output (R1Inf r)         = sexp "inf" [output r]
    output (R1Proj i r)      = sexp "proj" [output i, output r]
    output (R1Union s1 s2)   = sexp "union" [output s1, output s2]
    output (R1Inter s1 s2)   = sexp "inter" [output s1, output s2]
    output (R1Diff s1 s2)    = sexp "diff" [output s1, output s2]
    -- derived
    output (R1Fin r)         = sexp "fin" [output r]

instance Output RR2_rel where
    output (R2GTT g w x)     = sexp "gtt" [output g, output w, output x]
    output (R2Diag s)        = sexp "id" [output s]
    output (R2Prod s1 s2)    = sexp "product" [output s1, output s2]
    output (R2Inv r)         = sexp "inverse" [output r]
    output (R2Union r1 r2)   = sexp "union" [output r1, output r2]
    output (R2Inter r1 r2)   = sexp "inter" [output r1, output r2]
    output (R2Diff r1 r2)    = sexp "diff" [output r1, output r2]
    output (R2Comp r1 r2)    = sexp "comp" [output r1, output r2]
    -- derived
    output R2Eq              = sexp "equality" []
    output (R2Reflc r)       = sexp "reflcl" [output r]
    output (R2Step ts)       = sexp "step" [slist ts]
    output (R2StepEq ts)     = sexp "step=" [slist ts]
    output (R2ParStep ts)    = sexp "parallel-step" [slist ts]
    output (R2Steps ts)      = sexp "step+" [slist ts]
    output (R2StepsEq ts)    = sexp "step*" [slist ts]
    output (R2StepsNF ts)    = sexp "step!" [slist ts]
    output (R2RootStep ts)   = sexp "root-step" [slist ts]
    output (R2RootStepEq ts) = sexp "root-step=" [slist ts]
    output (R2RootSteps ts)  = sexp "root-step+" [slist ts]
    output (R2RootStepsEq ts) = sexp "root-step*" [slist ts]
    output (R2NonRootStep ts) = sexp "non-root-step" [slist ts]
    output (R2NonRootStepEq ts) = sexp "non-root-step=" [slist ts]
    output (R2NonRootSteps ts) = sexp "non-root-step+" [slist ts]
    output (R2NonRootStepsEq ts) = sexp "non-root-step*" [slist ts]
    output (R2Meet ts)       = sexp "meet" [slist ts]
    output (R2Join ts)       = sexp "join" [slist ts]

instance Output Formula where
    output (FRR1 s i)        = sexp "rr1" [output s, output i]
    output (FRR2 r i j)      = sexp "rr2" [output r, output i, output j]
    output (FAnd fs)         = sexp "and" [slist fs]
    output (FOr fs)          = sexp "or" [slist fs]
    output (FNot f)          = sexp "not" [output f]
    output (FForall f)       = sexp "forall" [output f]
    output (FExists f)       = sexp "exists" [output f]
    output (FExtendSig f n)  = output f -- sexp "extend-sig" [output n, output f]
    -- derived
    output FTrue             = sexp "true" []
    output FFalse            = sexp "false" []
    output (FRestrict f ts)  = sexp "restrict" [output f, slist ts]

instance Output Equivalence where
    output EDistribAndOr     = "distrib-and-or"
    output EDistribOrAnd     = "distrib-or-and"

instance Output Inference where
    output (IRR1 s i)        = sexp "rr1" [output s, output i]
    output (IRR2 r i j)      = sexp "rr2" [output r, output i, output j]
    output (IAnd is)         = sexp "and" [slist is]
    output (IOr is)          = sexp "or" [slist is]
    output (INot i)          = sexp "not" [output i]
    output (IExists i)       = sexp "exists" [output i]
    output (IRename i vs)    = sexp "rename" [output i, slist vs]
    output (INNFPlus i)      = sexp "nnf" [output i]
    output (IRepl i ps e)    = sexp "replace" [output i, slist ps, output e]

instance Output Info where
    output (Size s t e)      = sexp "size" [output s, output t, output e]

instance Output Claim where
    output Empty             = "empty"
    output Nonempty          = "nonempty"

instance Output Certificate where
    output (Certificate is c i) = vsep [
        vsep [slist' (output j : output r : output f : map output es)
             | (j, r, f, es) <- is],
        slist' [output c, output i]]
