-- | Certificate parser.
module Formula.Parse (
    certificate,
    formula,
) where

import Prelude hiding (lex)
import Control.Monad
import Data.Functor (($>))
import Text.Parsec

import Formula.Type as Fm

-- | Parser for the certificate
certificate :: Parsec String () Certificate
certificate = spaces *> certificateP <* eof

-- | Parser for formula
-- Note: this parses formulas used in the certificat which
-- using de Bruijn indices.
-- See Formula.FORT for the formulas used as input to FORT.
formula :: Parsec String () Formula
formula = spaces *> formulaP <* eof

------------------------------------------------------------
-- Parsers

-- <trs>
trsP :: Parsec String () Trs
trsP = do
  i <- read <$> many1 digit
  dir <- lex $ option Fwd (tok "-" $> Bwd)
  return $ dir i

-- <gtt>
gttP :: Parsec String () GTT_rel
gttP = sexp
  [ ("root-step", ARoot <$> slist1 trsP)
  , ("inverse", GInv <$> gttP)
  , ("aunion", AUnion <$> gttP <*> gttP)
  , ("atc", ATrancl <$> gttP)
  , ("gtc", GTrancl <$> gttP)
  , ("acomp", AComp <$> gttP <*> gttP)
  , ("gcomp", GComp <$> gttP <*> gttP)
  -- derived
  , ("gstep+", GSteps <$> slist1 trsP)
  ]

-- <pos>
posStepP :: Parsec String () Pos_step
posStepP = ((tok "e" <|> tok "root") $> PRoot)
       <|> ((tok ">" <|> tok "non-root") $> PNonRoot)
       <|> ((tok ">=" <|> tok "any") $> PAny)
       <?> "Step Position"

-- <num>
extStepP :: Parsec String () Ext_step
extStepP = ((tok "1" <|> tok "single") $> ESingle)
       <|> ((tok ">=" <|> tok "parallel") $> EParallel)
       <|> ((tok ">" <|> tok "strictparallel") $> EStrictParallel)

-- <rr1>
rr1P :: Parsec String () RR1_rel
rr1P = sexp
  [ ("terms", pure R1Terms)
  , ("nf", R1NF <$> slist1 trsP)
  , ("inf", R1Inf <$> rr2P)
  , ("proj", R1Proj <$> projPos <*> rr2P)
  , ("union", R1Union <$> rr1P <*> rr1P)
  , ("inter", R1Inter <$> rr1P <*> rr1P)
  , ("diff", R1Diff <$> rr1P <*> rr1P)
  -- derived
  , ("fin", R1Fin <$> rr2P)
  ] <?> "RR1 relation"
  where
    -- ('1' | '2')
    projPos = (tok "1" $> 1) <|> (tok "2" $> 2)


-- <rr2>
rr2P :: Parsec String () RR2_rel
rr2P = sexp
  [ ("gtt", R2GTT <$> gttP <*> posStepP <*> extStepP)
  , ("id", R2Diag <$> rr1P)
  , ("product", R2Prod <$> rr1P <*> rr1P)
  , ("inverse", R2Inv <$> rr2P)
  , ("union", R2Union <$> rr2P <*> rr2P)
  , ("inter", R2Inter <$> rr2P <*> rr2P)
  , ("diff", R2Diff <$> rr2P <*> rr2P)
  , ("comp", R2Comp <$> rr2P <*> rr2P)
  -- derived
  , ("equality", pure R2Eq)
  , ("reflcl", R2Reflc <$> rr2P)
  , ("step", R2Step <$> slist1 trsP)
  , ("step=", R2StepEq <$> slist1 trsP)
  , ("parallel-step", R2ParStep <$> slist1 trsP)
  , ("step+", R2Steps <$> slist1 trsP)
  , ("step*", R2StepsEq <$> slist1 trsP)
  , ("step!", R2StepsNF <$> slist1 trsP)
  , ("stepsnf", R2StepsNF <$> slist1 trsP)
  , ("root-step", R2RootStep <$> slist1 trsP)
  , ("root-step=", R2RootStepEq <$> slist1 trsP)
  , ("root-step+", R2RootSteps <$> slist1 trsP)
  , ("root-step*", R2RootStepsEq <$> slist1 trsP)
  , ("non-root-step", R2NonRootStep <$> slist1 trsP)
  , ("non-root-step=", R2NonRootStepEq <$> slist1 trsP)
  , ("non-root-step+", R2NonRootSteps <$> slist1 trsP)
  , ("non-root-step*", R2NonRootStepsEq <$> slist1 trsP)
  , ("meet", R2Meet <$> slist1 trsP)
  , ("join", R2Join <$> slist1 trsP)
  ] <?> "RR2 relation"

-- <formula>
formulaP :: Parsec String () Formula
formulaP = sexp tbl <?> "Formula"
  where tbl = [ ("rr1", FRR1 <$> rr1P <*> nat)
              , ("rr2", FRR2 <$> rr2P <*> nat <*> nat)
              , ("and", FAnd <$> slist formulaP)
              , ("or", FOr <$> slist formulaP)
              , ("not", FNot <$> formulaP)
              , ("forall", FForall <$> formulaP)
              , ("exists", FExists <$> formulaP)
              , ("extend-sig", flip FExtendSig <$> nat <*> formulaP)
              -- derived
              , ("true", pure FTrue)
              , ("false", pure FFalse)
              , ("restrict", FRestrict <$> formulaP <*> slist1 nat)
              ]

-- <equivalence>
equivalenceP :: Parsec String () Equivalence
equivalenceP = (tok "distrib-and-or" $> EDistribAndOr)
           <|> (tok "distrib-or-and" $> EDistribOrAnd)
           <?> "Equivalence"

-- <inference>
inferenceP :: Parsec String () Inference
inferenceP = sexp
   [ ("rr1", IRR1 <$> rr1P <*> nat)
   , ("rr2", IRR2 <$> rr2P <*> nat <*> nat)
   , ("and", IAnd <$> slist nat)
   , ("or", IOr <$> slist nat)
   , ("not", INot <$> nat)
   , ("exists", IExists <$> nat)
   , ("rename", IRename <$> nat <*> slist nat)
   , ("nnf", INNFPlus <$> nat)
   , ("replace", IRepl <$> equivalenceP <*> slist1 nat <*> nat)
   ] <?> "Inference"

-- <info>
infoP :: Parsec String () Info
infoP = parens (tok "size" *> sizeP) <?> "Size Info"
  where sizeP = Size <$> nat <*> nat <*> nat

stepP :: Parsec String () Step
stepP = stepP' <?> "Inference Step"
  where stepP' = (,,,) <$> nat <*> inferenceP <*> formulaP <*> many infoP

claimP :: Parsec String () (Claim, Int)
claimP = (,) <$> claimP' <*> nat
  where claimP' = (tok "empty" $> Fm.Empty)
              <|> (tok "nonempty" $> Fm.Nonempty)

-- <proof>
certificateP :: Parsec String () Certificate
certificateP = do
  r <- parens ((Right <$> stepP) <|> Left <$> claimP)
  case r of
    Left (c, i) -> pure $ Certificate [] c i
    Right s -> do
      Certificate ss c i <- certificateP
      pure $ Certificate (s:ss) c i

-- | parses a positive integer
-- <nat> <var> <term> <item> <itemref>
nat :: Parsec String () Int
nat = read <$> lex (many1 digit)

------------------------------------------------------------
-- Utility functions

parens :: Parsec String () a -> Parsec String () a
parens = between (lex $ char '(') (lex $ char ')')

tok :: String -> Parsec String () ()
tok s = void . lex $ try (string s)

lex :: Parsec String () a -> Parsec String () a
lex p = p <* spaces

sexp :: [(String, Parsec String () b)] -> Parsec String () b
sexp tbl = between (lex $ char '(') (lex $ char ')') go
  where
    go = msum [ident s *> p | (s, p) <- tbl]
    ident s = lex $ try (string s *> lookAhead delim)
    delim = space <|> oneOf "()"

slist :: Parsec String () a -> Parsec String () [a]
slist p = parens (many p)

slist1 :: Parsec String () a -> Parsec String () [a]
slist1 p = parens (many1 p)

