{-# LANGUAGE OverloadedStrings #-}

-- | Pretty printing for formulas.
module Formula.Pretty (
    pretty,
) where

import Formula.Type
import Prettyprinter hiding (list)
import qualified Data.Set as S

instance Pretty Trs where
    pretty (Fwd i) = pretty i
    pretty (Bwd i) = pretty i <> "-"

list :: [Doc ann] -> Doc ann
list ds = encloseSep "[" "]" "," ds

list' :: Pretty p => [p] -> Doc ann
list' = list . map pretty

arr :: Doc ann -> Doc ann -> Doc ann -> [Trs] -> Doc ann
arr l b r ts
    | ts' == s [Fwd 0]        = r
    | ts' == s [Fwd 0, Bwd 0] = b
    | ts' == s [Bwd 0]        = l
    | fs' == bs'              = b <> list' (S.toList fs')
    | S.null fs'              = list' (S.toList bs') <> l
    | otherwise               = r <> list' (S.toList ts')
  where
    s :: Ord a => [a] -> S.Set a
    s = S.fromList
    fs' = s [i | Fwd i <- ts]
    bs' = s [i | Bwd i <- ts]
    ts' = s ts

instance Pretty GTT_rel where
    pretty (ARoot ts) = arr "ε<-" "<->ε" "->ε" ts
    pretty (GInv g) = "(" <> pretty g <> ")⁻"
    pretty (AUnion g1 g2) = "(" <> pretty g1 <+> "∪" <+> pretty g2 <> ")"
    pretty (ATrancl g) = "(" <> pretty g <> ")+"
    pretty (GTrancl g) = "(" <> pretty g <> ")+̂ "
    pretty (AComp g1 g2) = "(" <> pretty g1 <+> "o" <+> pretty g2 <> ")"
    pretty (GComp g1 g2) = "(" <> pretty g1 <+> "ô" <+> pretty g2 <> ")"
    pretty (GSteps ts) = arr "+̂<-" "<->+̂" "->+̂" ts

instance Pretty Pos_step where
    pretty PRoot = "ε"
    pretty PNonRoot = ">ε"
    pretty PAny = "⩾ε"

instance Pretty Ext_step where
    pretty ESingle = "1"
    pretty EParallel = "∥"
    pretty EStrictParallel = "∥̇"

instance Pretty RR1_rel where
    pretty R1Terms = "𝒯"
    pretty (R1NF ts) = "NF" <> list' ts
    pretty (R1Inf r) = "Inf(" <> pretty r <> ")"
    pretty (R1Proj i r) = "(" <> pretty r <> if i == 0 then ")₁" else ")₂"
    pretty (R1Union s1 s2) = "(" <> pretty s1 <+> "∪" <+> pretty s2 <> ")"
    pretty (R1Inter s1 s2) = "(" <> pretty s1 <+> "∩" <+> pretty s2 <> ")"
    pretty (R1Diff s1 s2) = "(" <> pretty s1 <+> "-" <+> pretty s2 <> ")"
    pretty (R1Fin r) = "Fin(" <> pretty r <> ")"

instance Pretty RR2_rel where
    pretty (R2GTT r w x) = "(" <> pretty r <> ")" <> pretty w <> "," <> pretty x
    pretty (R2Diag s) = "Δ(" <> pretty s <> ")"
    pretty (R2Prod s1 s2) = "(" <> pretty s1 <+> "×" <+> pretty s2 <> ")"
    pretty (R2Inv r) = "(" <> pretty r <> ")⁻"
    pretty (R2Union r1 r2) = "(" <> pretty r1 <+> "∪" <+> pretty r2 <> ")"
    pretty (R2Inter r1 r2) = "(" <> pretty r1 <+> "∩" <+> pretty r2 <> ")"
    pretty (R2Diff r1 r2) = "(" <> pretty r1 <+> "-" <+> pretty r2 <> ")"
    pretty (R2Comp r1 r2) = "(" <> pretty r1 <+> "o" <+> pretty r2 <> ")"
    pretty R2Eq = "="
    pretty (R2Reflc r) = "(" <> pretty r <> ")="
    pretty (R2Step ts) = arr "<-" "<->" "->" ts
    pretty (R2StepEq ts) = arr "=<-" "<->=" "->=" ts
    pretty (R2Steps ts) = arr "+<-" "<->+" "->+" ts
    pretty (R2StepsEq ts) = arr "*<-" "<->*" "->*" ts
    pretty (R2StepsNF ts) = arr "!<-" "<->!" "->!" ts
    pretty (R2ParStep ts) = arr "<-∥-" "<-∥->" "-∥->" ts
    pretty (R2RootStep ts) = arr "ε<-" "<->ε" "->ε" ts
    pretty (R2RootStepEq ts) = arr "=ε<-" "<->ε=" "->ε=" ts
    pretty (R2RootSteps ts) = arr "+ε<-" "<->ε+" "->ε+" ts
    pretty (R2RootStepsEq ts) = arr "*ε<-" "<->ε*" "->ε*" ts
    pretty (R2NonRootStep ts) = arr "¬ε<-" "<->¬ε" "->¬ε" ts
    pretty (R2NonRootStepEq ts) = arr "=¬ε<-" "<->¬ε=" "->¬ε=" ts
    pretty (R2NonRootSteps ts) = arr "+¬ε<-" "<->¬ε+" "->¬ε+" ts
    pretty (R2NonRootStepsEq ts) = arr "*¬ε<-" "<->¬ε*" "->¬ε*" ts
    pretty (R2Meet ts) = arr "↓" "<->*" "↑" ts
    pretty (R2Join ts) = arr "↑" "<->*" "↓" ts

instance Pretty Formula where
    pretty (FRR1 s x) = pretty x <+> "∈" <+> pretty s
    pretty (FRR2 r x y) = pretty x <+> pretty r <+> pretty y
    pretty (FAnd fs) = encloseSep "(" ")" (flatAlt "∧ " " ∧ ") (map pretty fs)
    pretty (FOr fs) = encloseSep "(" ")" (flatAlt "∨ " " ∨ ") (map pretty fs)
    pretty (FNot f) = "¬" <+> pretty f
    pretty (FExists f) = "∃(" <> pretty f <> ")"
    pretty (FForall f) = "∀(" <> pretty f <> ")"
    pretty FTrue = "⊤"
    pretty FFalse = "⊥"
    pretty (FExtendSig f n) = "Σ+" <> "[" <> pretty n <> "]" <> "(" <> pretty f <> ")"
    pretty (FRestrict f ts) = list' ts <> "(" <> pretty f <> ")"

{-
-- | Equational reasoning.
data Equivalence
    = EDistribAndOr              -- ^ and distributes over or
    | EDistribOrAnd              -- ^ or distributes over and
    deriving Show

-- | Inference rule.
data Inference
    = IRR1 RR1_rel Int           -- ^ introduction by RR1 relation
    | IRR2 RR2_rel Int Int       -- ^ introduction by RR2 relation
    | IAnd [Int]                 -- ^ conjunction
    | IOr [Int]                  -- ^ disjunction
    | INot Int                   -- ^ negation
    | IExists Int                -- ^ existential quantification
    | IRename Int [Int]          -- ^ rename variables
    | INNFPlus Int               -- ^ equivalence by NNF plus ACUI01
    | IRepl Equivalence [Int] Int -- ^ equivalence by equational reasoning
    deriving Show

-- | Extra information.
data Info
    = Size Int Int Int           -- ^ states, transitions, ε-transitions
    deriving Show

-- | Final proof claim.
data Claim
    = Empty                      -- ^ empty = unsatisfiable
    | Nonempty                   -- ^ non-empty = satisfiable
    deriving Show

-- | Certificate type.
data Certificate =
    Certificate [(Int, Inference, Formula, [Info])] Claim Int
    deriving Show
-}
