-- | Formula type definition.
module Formula.Type where

-- | TRS description.
data Trs
    = Fwd Int                    -- ^ TRS in forward direction
    | Bwd Int                    -- ^ TRS in backward direction
    deriving (Eq, Ord, Show)

-- | GTT relations.
data GTT_rel
    = ARoot [Trs]                -- ^ root steps
    | GInv GTT_rel               -- ^ inverse
    | AUnion GTT_rel GTT_rel     -- ^ union
    | ATrancl GTT_rel            -- ^ transitive closure (root)
    | GTrancl GTT_rel            -- ^ transitive closure (parallel)
    | AComp GTT_rel GTT_rel      -- ^ composition (root)
    | GComp GTT_rel GTT_rel      -- ^ composition (parallel)
    -- derived
    | GSteps [Trs]               -- ^ rewrite sequence with a root step
    deriving (Eq, Ord, Show)

-- | Position for lifting root steps.
data Pos_step
    = PRoot                      -- ^ root steps only
    | PNonRoot                   -- ^ non-root steps only
    | PAny                       -- ^ unrestricted steps
    deriving (Eq, Ord, Show)

-- | Multiplicity for lifting root steps.
data Ext_step
    = ESingle                    -- ^ single step
    | EParallel                  -- ^ parallel step
    | EStrictParallel            -- ^ non-empty parallel step
    deriving (Eq, Ord, Show)

-- | /RR1/ relation.
data RR1_rel
    = R1Terms                    -- ^ all terms
    | R1NF [Trs]                 -- ^ normal forms
    | R1Inf RR2_rel              -- ^ infiniteness
    | R1Proj Int RR2_rel         -- ^ projection of /RR2/ relation
    | R1Union RR1_rel RR1_rel    -- ^ union
    | R1Inter RR1_rel RR1_rel    -- ^ intersection
    | R1Diff RR1_rel RR1_rel     -- ^ difference
    -- derived
    | R1Fin RR2_rel              -- ^ finiteness
    deriving (Eq, Ord, Show)

-- | /RR2/ relation.
data RR2_rel
    = R2GTT GTT_rel Pos_step Ext_step -- ^ lift GTT to /RR2/
    | R2Diag RR1_rel             -- ^ identity
    | R2Prod RR1_rel RR1_rel     -- ^ cartesian product
    | R2Inv RR2_rel              -- ^ inverse
    | R2Union RR2_rel RR2_rel    -- ^ union
    | R2Inter RR2_rel RR2_rel    -- ^ intersection
    | R2Diff RR2_rel RR2_rel     -- ^ difference
    | R2Comp RR2_rel RR2_rel     -- ^ composition
    -- derived
    | R2Eq                       -- ^ equality on terms
    | R2Reflc RR2_rel            -- ^ reflexive closure
    | R2Step [Trs]               -- ^ single step
    | R2StepEq [Trs]             -- ^ single or empty step
    | R2Steps [Trs]              -- ^ at least one step
    | R2StepsEq [Trs]            -- ^ many steps
    | R2StepsNF [Trs]            -- ^ rewrite to NF
    | R2ParStep [Trs]            -- ^ parallel step
    | R2RootStep [Trs]           -- ^ root step
    | R2RootStepEq [Trs]         -- ^ root or empty step
    | R2RootSteps [Trs]          -- ^ at least one root step
    | R2RootStepsEq [Trs]        -- ^ many root steps
    | R2NonRootStep [Trs]        -- ^ non-root step
    | R2NonRootStepEq [Trs]      -- ^ non-root or empty step
    | R2NonRootSteps [Trs]       -- ^ at least one non-root step
    | R2NonRootStepsEq [Trs]     -- ^ many non-root steps
    | R2Meet [Trs]               -- ^ meetability
    | R2Join [Trs]               -- ^ joinability
    deriving (Eq, Ord, Show)

-- | Complement of /RR2/ relation.
r2Neg :: RR2_rel -> RR2_rel
r2Neg r = R2Diff (R2Prod R1Terms R1Terms) r

-- | Implication of /RR2/ relations.
r2Implies :: RR2_rel -> RR2_rel -> RR2_rel
r2Implies r1 r2 = R2Union (r2Neg r1) r2

-- | Equivalence of /RR2/ relations.
r2Iff :: RR2_rel -> RR2_rel -> RR2_rel
r2Iff r1 r2 = R2Inter (r2Implies r1 r2) (r2Implies r2 r1)

-- | First-order formula.
data Formula
    = FRR1 RR1_rel Int           -- ^ /RR1/ predicate
    | FRR2 RR2_rel Int Int       -- ^ /RR2/ predicate
    | FAnd [Formula]             -- ^ conjunction
    | FOr [Formula]              -- ^ disjunction
    | FNot Formula               -- ^ negation
    | FExists Formula            -- ^ existential quantification
    | FForall Formula            -- ^ universal quantification
    -- derived
    | FTrue                      -- ^ True
    | FFalse                     -- ^ False
    | FExtendSig Formula Int     -- ^ extend signature by n constants
    | FRestrict Formula [Int]    -- ^ restrict TRSs
    deriving (Eq, Ord, Show)

-- | Implication of formulas.
fImplies :: Formula -> Formula -> Formula
fImplies f g = FOr [FNot f, g]

-- | Equivalence of formulas.
fIff :: Formula -> Formula -> Formula
fIff r1 r2 = FAnd [fImplies r1 r2, fImplies r2 r1]

-- | Equational reasoning.
data Equivalence
    = EDistribAndOr              -- ^ and distributes over or
    | EDistribOrAnd              -- ^ or distributes over and
    deriving (Eq, Ord, Show)

-- | Inference rule.
data Inference
    = IRR1 RR1_rel Int           -- ^ introduction by RR1 relation
    | IRR2 RR2_rel Int Int       -- ^ introduction by RR2 relation
    | IAnd [Int]                 -- ^ conjunction
    | IOr [Int]                  -- ^ disjunction
    | INot Int                   -- ^ negation
    | IExists Int                -- ^ existential quantification
    | IRename Int [Int]          -- ^ rename variables
    | INNFPlus Int               -- ^ equivalence by NNF plus ACUI01
    | IRepl Equivalence [Int] Int -- ^ equivalence by equational reasoning
    deriving (Eq, Ord, Show)

-- | Extra information.
data Info
    = Size Int Int Int           -- ^ states, transitions, ε-transitions
    deriving (Eq, Ord, Show)

-- | Inference step.
type Step = (Int, Inference, Formula, [Info])

-- | Final proof claim.
data Claim
    = Empty                      -- ^ empty = unsatisfiable
    | Nonempty                   -- ^ non-empty = satisfiable
    deriving (Eq, Ord, Show)

-- | Certificate type.
data Certificate =
    Certificate [Step] Claim Int
    deriving (Eq, Ord, Show)
