-- | Basic GTT operations.
module GTT.Basics (
    unique,
    trim,
    relabel,
    anchor,
    ginv,
    aunion,
) where

import qualified TA
import GTT.Type as GTT
import qualified Data.Set as S
import qualified Data.Map as M

import Data.Hashable

-- | De-duplicate states and transitions of a GTT.
unique :: (Ord f, Ord q) => GTT f q -> GTT f q
unique gtt = GTT {
    leftTA  = TA.unique (leftTA  gtt),
    rightTA = TA.unique (rightTA gtt)
}

-- | Trim GTT.
trim :: (Ord q, Hashable q) => GTT f q -> GTT f q
trim gtt = GTT { leftTA = l'', rightTA = r'' }
  where
    ls = TA.reachable (leftTA gtt)
    rs = TA.reachable (rightTA gtt)
    is = ls `S.intersection` rs
    l' = TA.restrict (leftTA  gtt) ls
    r' = TA.restrict (rightTA gtt) rs
    is' = S.toList is
    l'' = TA.restrict l' (TA.productive l'{ TA.finals = is' })
    r'' = TA.restrict r' (TA.productive r'{ TA.finals = is' })

-- | Anchor GTT.
--
-- Produce equivalent GTT with the property that no interface state
-- occurs in any transition's lhs.
anchor :: Ord q => GTT f q -> GTT f (Either q (Either q q))
anchor gtt = GTT { leftTA = l'', rightTA = r'' }
  where
    is = interface gtt
{-
    -- We rename the two automata apart and introduce a fresh state for
    -- each interface state, with epsilon transitions to connect them.
    l' = TA.map id (Right . Left) (leftTA gtt) `TA.merge`
        TA.empty{ TA.eps = [(Right (Left q), Left q) | q <- S.toList is] }
    r' = TA.map id (Right . Right) (rightTA gtt) `TA.merge`
        TA.empty{ TA.eps = [(Right (Right q), Left q) | q <- S.toList is] }
-}
    -- Optimized version that elides epsilon transitions if an interface
    -- state does not occur in the lhss of one of the automata.
    lsL = TA.lStates (leftTA gtt)
    lsR = TA.lStates (rightTA gtt)
    goodL = is `S.difference` lsL; badL = is `S.intersection` lsL
    goodR = is `S.difference` lsR; badR = is `S.intersection` lsR
    renL q = if q `S.member` goodL then Left q else Right (Left q)
    renR q = if q `S.member` goodR then Left q else Right (Right q)
    l'' = TA.map id renL (leftTA gtt) `TA.merge`
        TA.empty{ TA.eps = [(Right (Left q), Left q) | q <- S.toList badL] }
    r'' = TA.map id renR (rightTA gtt) `TA.merge`
        TA.empty{ TA.eps = [(Right (Right q), Left q) | q <- S.toList badR] }

-- | Relabel GTT.
relabel :: Ord q => GTT f q -> GTT f Int
relabel gtt = GTT.map id fq gtt where
    fq = (M.fromAscList (zip (S.toAscList (states gtt)) [0..]) M.!)

-- | Invert GTT.
ginv :: GTT f q -> GTT f q
ginv GTT { leftTA = l, rightTA = r } = GTT { leftTA = r, rightTA = l }

-- | Anchored union of GTTs.
aunion :: GTT f p -> GTT f q -> GTT f (Either p q)
aunion g1 g2 = merge (GTT.map id Left g1) (GTT.map id Right g2)
