-- | GTT composition and transitive closure.
module GTT.Compose (
    gcomp,
    acomp,
    gtrancl,
    atrancl,
) where

import Prelude as P
import qualified TA
import GTT.Type as GTT
import Util.Saturate
import qualified Data.Set as S
import qualified Data.Map as Map
import Data.Maybe
import qualified Util.Set as S
import qualified GTT.Basics as GTT

-- | GTT composition.
--
-- May destroy root-cleanliness.
gcomp :: (Eq f, Ord p, Ord q) => GTT f p -> GTT f q -> GTT f (Either p q)
gcomp g1 g2 = g' `GTT.merge` GTT {
    leftTA  = TA.empty{ TA.trans = getTransl g' epsl },
    rightTA = TA.empty{ TA.trans = getTransr g' epsl }
} where
    (g1', g2') = (GTT.map id Left g1, GTT.map id Right g2)
    g' = GTT.merge g1' g2'
    epsl' = S.toList (saturate sat0 sat1)
    epsl = epsFree2 epsl'
    sat0 = gc_sat0 g1' g2'
    sat1 = gc_sat1 g1' g2'


-- | Anchored GTT composition.
--
-- Assumes and preserves root-cleanliness (see `anchor`).
acomp :: (Eq f, Ord p, Ord q) => GTT f p -> GTT f q -> GTT f (Either p q)
acomp g1 g2 = g' `GTT.merge` GTT {
    leftTA  = TA.empty{ TA.trans = getTransl g' epsl },
    rightTA = TA.empty{ TA.trans = getTransr g' epsl }
} where
    (g1', g2') = (GTT.map id Left g1, GTT.map id Right g2)
    g' = GTT.merge g1' g2'
    ifs1 = (`S.member` interface g1')
    ifs2 = (`S.member` interface g2')
    epsl' = P.filter (\(p, q) -> ifs1 p && ifs2 q) (S.toList (saturate sat0 sat1))
    epsl = epsFree2 epsl'
    sat0 = gc_sat0 g1' g2'
    sat1 = gc_sat1 g1' g2'


-- | GTT transitive closure.
--
-- May destroy root-cleanliness.
gtrancl :: (Ord f, Ord q) => GTT f q -> GTT f q
gtrancl g = GTT.unique $ g' `GTT.merge` GTT {
    leftTA  = TA.empty{ TA.trans = getTransl g' epsl },
    rightTA = TA.empty{ TA.trans = getTransr g' epsl }
} where
    epsl' = S.toList (saturate sat0 sat1)
    epsfree = epsFree epsl' g
    epsl = fst epsfree 
    g' = snd epsfree
    sat0 = gc_sat0 g g
    sat1 e@(p, q) es =
        [(p, q'') | (q', q'') <- S.toList es, q == q'] ++
        [(p'', q) | (p'', p') <- S.toList es, p' == p] ++
        gc_sat1 g g e es
    --epsStep = (Map.fromListWith S.union [(q,S.singleton p) | (q,p) <- epsl] Map.!)
    --epsStepRev = (Map.fromListWith S.union [(q,S.singleton p) | (p,q) <- epsl] Map.!) -- ! not inside
    --replFunM = Map.fromList [ (q,S.findMin $ (S.intersection (epsStep q) (epsStepRev q))) --findMin Empty add q itself
    --        | (q,_) <- epsl]

-- | Anchored GTT transitive closure.
--
-- Requires and preserves root-cleanliness.
atrancl :: (Ord f, Ord q) => GTT f q -> GTT f q
atrancl g = GTT.unique $ g' `GTT.merge` GTT {
    leftTA  = TA.empty{ TA.trans = getTransl g' epsl },
    rightTA = TA.empty{ TA.trans = getTransr g' epsl }
} where
    ifs = (`S.member` interface g)
    epsl' = P.filter (\(p, q) -> ifs p && ifs q) (S.toList (saturate sat0 sat1))
    epsfree = epsFree epsl' g
    epsl = fst epsfree 
    g' = snd epsfree
    sat0 = gc_sat0 g g
    sat1 e@(p, q) es =
        [(p, q'') | (q', q'') <- S.toList es, q == q'] ++
        [(p'', q) | (p'', p') <- S.toList es, p' == p] ++
        gc_sat1 g g e es

-- reduces epsilon transitions
epsFree :: (Ord f, Ord q) => [(q,q)] -> GTT f q -> ([(q,q)], GTT f q)
epsFree epsl' g = (epsl, g')
    where
        epsl = S.nubOrd [ (replFun q, replFun p) | (q,p) <- epsl' ]
        g' = GTT.unique $ 
            GTT { leftTA = TA.map id replFun (leftTA g)
            , rightTA = TA.map id replFun (rightTA g) 
            }
        replFun q = Map.findWithDefault q q replFunM
        intersect q = S.insert q (S.intersection (epsStep q) (epsStepRev q))
        map = Map.fromListWith S.union [(q,S.singleton p) | (q,p) <- epsl']
        mapRev = Map.fromListWith S.union [(q,S.singleton p) | (p,q) <- epsl']
        epsStep q = Map.findWithDefault (S.singleton q) q map
        epsStepRev q = Map.findWithDefault (S.singleton q) q mapRev
        replFunM = Map.fromList [ (q,S.findMin $ intersect q) | (q,_) <- epsl']

epsFree2 :: (Ord q) => [(q,q)] -> [(q,q)]
epsFree2 epsl' = S.nubOrd [ (replFun q, replFun p) | (q,p) <- epsl' ]
    where
        replFun q = Map.findWithDefault q q replFunM
        intersect q = S.insert q (S.intersection (epsStep q) (epsStepRev q))
        map = Map.fromListWith S.union [(q,S.singleton p) | (q,p) <- epsl']
        mapRev = Map.fromListWith S.union [(q,S.singleton p) | (p,q) <- epsl']
        epsStep q = Map.findWithDefault (S.singleton q) q map
        epsStepRev q = Map.findWithDefault (S.singleton q) q mapRev
        replFunM = Map.fromList [ (q,S.findMin $ intersect q) | (q,_) <- epsl']
      

getTransl ::  (Eq f, Ord q) => GTT f q -> [(q, q)] -> [(f, [q], q)]
getTransl GTT{ leftTA = a1 } eps = 
    [(f, ps, q) | (f, ps, p1) <- TA.trans a1, let z = p1, q <- Map.findWithDefault [] z m]
    where
        m = listToMap eps

getTransr ::  (Eq f, Ord q) => GTT f q -> [(q, q)] -> [(f, [q], q)]
getTransr GTT{ rightTA = a1 } eps = 
    [(f, ps, q) | (f, ps, p1) <- TA.trans a1, let z = p1, q <- Map.findWithDefault [] z m]
    where
        m = listToRevMap eps

listToMap :: (Ord a) => [(a, b)] -> Map.Map a [b]
listToMap eps =
  Map.fromListWith (++) . P.map (\(x,y) -> (x,[y])) $ eps

listToRevMap :: (Ord a) => [(a, a)] -> Map.Map a [a]
listToRevMap eps =
  Map.fromListWith (++) . P.map (\(x,y) -> (y,[x])) $ eps

elimMaybe :: Maybe [a] -> [a]
elimMaybe (Just xs) = xs
elimMaybe Nothing = []

-- Auxiliary functions shared between composition and transitive closure.
gc_sat0 :: (Eq f, Ord q) => GTT f q -> GTT f q -> [(q, q)]
gc_sat0 GTT{ rightTA = a1 } GTT{ leftTA = a2 } =
    [(p, q) | (f, [], p) <- TA.trans a1, (f', [], q) <- TA.trans a2, f == f']

gc_sat1 :: (Eq f, Ord q) =>
    GTT f q -> GTT f q -> (q, q) -> S.Set (q, q) -> [(q, q)]
gc_sat1 GTT{ rightTA = a1 } GTT{ leftTA = a2 } (p, q) es =
    [(p'', q) | (p', p'') <- TA.eps a1, p == p'] ++
    [(p, q'') | (q', q'') <- TA.eps a2, q == q'] ++
    [(p', q')
    | (f, ps, p') <- TA.trans a1,
      (f', qs, q') <- TA.trans a2,
      (f, length ps) == (f', length qs),
      (p, q) `elem` zip ps qs,
      all (`S.member` es) (zip ps qs)]