-- | Construct GTT from TRS.
module GTT.TRS (
    fromTRS,
    fromTRSI,
) where

import GTT.Type as GTT
import qualified TA as TA
import qualified Data.Rewriting.Term as T
import qualified Data.Rewriting.Rule as R
import Util.Rewriting (isLVTRS)
import Util.Signature
import Control.Exception (assert)

-- | Construct GTT accepting (root/parallel) steps of given LV-TRS.
fromTRS :: (Ord s, Ord f, Ord v) => Signature s f ->
    [R.Rule f v] -> GTT f (Either (T.Term f s) (T.Term f s))
fromTRS sig trs
    | assert (isLVTRS trs) False = undefined
    | otherwise = GTT {
        leftTA = (TA.merges [TA.map id Left (TA.fromTerm sig (R.lhs r)) | r <- trs]) {
            TA.eps = [(Left l, Right r)
                   | rl <- trs
                   , s <- sortOf rl
                   , let l = applySort (R.lhs rl) s
                   , let r = applySort (R.rhs rl) s
                   ]
        },
        rightTA = TA.merges [TA.map id Right (TA.fromTerm sig (R.rhs r)) | r <- trs]
    }
 where
   -- sortOf :: R.Rule f v -> [s]
   sortOf r = case (R.lhs r, R.rhs r) of
     (T.Fun f _, _) -> [snd $ getSort f]
     (_, T.Fun f _) -> [snd $ getSort f]
     (T.Var _, T.Var _) -> sorts sig

  
   getSort = lookupSig sig
   applySort (T.Var _) s = T.Var s
   applySort (T.Fun f ts) _ = T.Fun f (zipWith applySort ts (fst $ getSort f))

-- | Construct GTT accepting (root/parallel) steps of given LV-TRS,
-- with designated interface state `Left i` for each rule.
fromTRSI :: (Ord s, Ord f, Ord v) => Signature s f ->
    [R.Rule f v] -> GTT f (Either (Int, s) (Either (T.Term f s) (T.Term f s)))
fromTRSI sig trs
    | assert (isLVTRS trs) False = undefined
    | otherwise = GTT {
        leftTA = TA.merges [TA.map id (maybe (Left (i, s)) (Right . Left))
            (TA.fromTermSI sig s (R.lhs r)) | (i, r) <- zip [0..] trs, s <- sortOfR r],
        rightTA = TA.merges [TA.map id (maybe (Left (i, s)) (Right . Right))
            (TA.fromTermSI sig s (R.rhs r)) | (i, r) <- zip [0..] trs, s <- sortOfR r]
    }

    where
      sortOfR (R.Rule (T.Fun f _) _) = [snd $ lookupSig sig f]
      sortOfR (R.Rule (T.Var _) (T.Fun f _ )) = [snd $ lookupSig sig f]
      sortOfR (R.Rule (T.Var _) (T.Var _ )) = sorts sig
