{-# LANGUAGE FlexibleContexts #-}
-- | Parser for problems in the [COPS](http://cops.uibk.ac.at/) Formats
module Problem.Parse
    ( parseProblemFile
      -- * Unsorted TRSs
      -- $unsortedTRSs
    , trsProb
      -- * Many-sorted TRSs
      -- $manysortedTRSs
    , mstrsProb
      -- * Commutation Problem
      -- $commutationProblem
    , commutationProb
    ) where

import Prelude hiding (lex)

import Control.Monad
import Data.Char (isSpace, isDigit)
import Data.List (nub)
import Data.Maybe

import qualified Data.Rewriting.Term as P
import qualified Data.Rewriting.Rule as R
import Text.Parsec

import Problem.Type (Problem)
import qualified Problem.Type as T

-- | Parses a problem File
--
-- @parseProblemFile file.trs@ parses a 'Problem' in the TRS Format
-- of MSTRS Format from file @file.trs@.
parseProblemFile :: FilePath -> IO [Problem String String String]
parseProblemFile file = do
    inp <- readFile file
    case parse problem file inp of
      Left e -> error $ show e
      Right p -> return p


problem :: Parsec String u [Problem String String String]
problem = spaces *> (try commutationProb <|> fmap pure (try mstrsProb <|> trsProb)) <* eof

--------------------
---- Commutation Problem
-- $commutationProblem
--
-- The inlined
-- [Commutation Format](http://project-coco.uibk.ac.at/2019/categories/commutation.php)
-- used during [CoCo](http://project-coco.uibk.ac.at/).
--
-- This is parsed as two 'Problem's with the single sort @"0"@.

-- | Parser for the /inlined Commutation/ TRS Formats.
commutationProb :: Parsec String u [Problem String String String]
commutationProb = do
    _ <- sExpr_ "PROBLEM" (lex . string $ "COMMUTATION")
    _ <- optional comment
    p1 <- trsProb
    p2 <- trsProb
    return [p1,p2]

--------------------
---- Unsorted trss
-- $unsortedTRSs
--
-- The /Basic/ and /Extended/
-- [TRS Formats](http://project-coco.uibk.ac.at/problems/trs.php)
-- used in [COPS](http://cops.uibk.ac.at/) represent term rewrite systems
-- without sort information.
--
-- This is parsed as a 'Problem' with the single sort @"0"@.

-- | Parser for the /Basic\/Extended/ TRS Formats.
trsProb :: Parsec String u (Problem String String String)
trsProb = do
    vs <- option [] (try vars)
    sig <- optionMaybe (try unsortedSig)
    t <- trs ident (guardedIdent (`elem` vs))
    _ <- optional (try comment)
    let sig' = nub . map (\(f, n) -> (f, replicate n "0", "0"))
             $ fromMaybe [ fa
                         | R.Rule l r <- t
                         , fa <- P.funs (P.withArity l) ++ P.funs (P.withArity r)
                         ] sig
    return $ T.Problem { T.signature = sig'
                       , T.rules = t
                       , T.sorts = ["0"]
                       , T.symbols = nub $ t >>= R.funs
                       , T.vars = vs
                       }

unsortedSig :: Parsec String u [(String,Int)]
unsortedSig = sExpr_ "SIG" (many sig')
    where
        sig' = sExpr ident num
        num = lex (read <$> many1 (satisfy isDigit))

---------------------
---- Many-sorted trss
--
-- $manysortedTRSs
--
-- The [MSTRS Formats](http://project-coco.uibk.ac.at/problems/trs.php)
-- used in [COPS](http://cops.uibk.ac.at/) represent many-sorted
-- term rewrite systems.

-- | Parser for the /MSTRS/ Format.
mstrsProb :: Parsec String u (Problem String String String)
mstrsProb = do
    s <- msSig
    let fs = nub [f | (f, _, _) <- s]
    t <- trs ident (guardedIdent (`notElem` fs))
    _ <- optional comment
    return $ T.Problem { T.signature = s
                       , T.rules = t
                       , T.sorts = nub $ concat [s:ss | (_, ss, s) <- s]
                       , T.symbols = fs
                       , T.vars = nub $ t >>= R.vars
                       }

msSig :: Parsec String u [(String, [String], String)]
msSig = sExpr_ "SIG" (many sig')
    where
        sig' = do
            (f, (ss, s)) <- sExpr ident sort
            return (f, ss, s)
        sort = (,) <$> many ident <*> (lex (string "->") *> ident)


--------------------
-- General Parsers

vars :: Parsec String u [String]
vars = sExpr_ "VAR" vars'
    where
        vars' = many ident

trs :: Parsec String u f -> Parsec String u v -> Parsec String u [R.Rule f v]
trs f v = sExpr_ "RULES" (many rule')
    where
        rule' = R.Rule <$> term <*> (lex (string "->") *> term)
        term = P.parse f v

comment :: Parsec String u String
comment = sExpr_ "COMMENT" comment'
    where
        comment' = ((++) <$> withpar <*> comment')
               <|> ((++) <$> many1 (noneOf "()") <*> comment')
               <|> return ""
        withpar = parens comment'

sExpr_ :: (Stream s m Char) => String -> ParsecT s u m a -> ParsecT s u m a
sExpr_ name p = snd <$> sExpr (lex (string name)) p

--------------------
-- Utility functions

sExpr :: (Stream s m Char) => ParsecT s u m h -> ParsecT s u m b -> ParsecT s u m (h, b)
sExpr head body = parens $ (,) <$> head <*> body

parens :: (Stream s m Char) => ParsecT s u m a -> ParsecT s u m a
parens = between (lex $ char '(') (lex $ char ')')

lex :: (Stream s m Char) => ParsecT s u m a -> ParsecT s u m a
lex p = p <* spaces

ident :: (Stream s m Char) => ParsecT s u m String
ident = try . lex $ do
   i <- many1 (satisfy (\c -> not (isSpace c) && c `notElem` "(),"))
   guard (i `notElem` ["->"])
   return i

guardedIdent :: (String -> Bool) -> Parsec String u String
guardedIdent p = do
    i <- ident
    guard $ p i
    return i

