-- | Basic /RRn/ automata functionality.
module RRn.Basics (
    fromTA,
    rr0,
    diag,
    prod,
    -- * Lifted operations from tree automata
    unsafeLift,
    unsafeLift2,
    unique,
    trim,
    minimize,
    epsFree,
    relabel,
    subset,
    union,
    intersection,
    difference,
) where

import RRn.Type
import RRn.Internal
import qualified TA
import Control.Exception (assert)
import qualified Util.Set as S
import qualified Data.IntSet as IS
import Util.Signature

import Data.Hashable

-- | Make /RR1/ relation from tree automaton.
fromTA :: (Ord f) => Signature s f -> TA.TA f q -> RRn s f q
fromTA sig a = RRn {
    arity = 1,
    signature = sig,
    ta = TA.map (\f -> [Just f]) id a
}

-- | Make /RR0/ relation corresponding to `True` or `False`
rr0 :: Bool -> RRn s f ()
rr0 b = RRn {
    arity = 0,
    signature = [],
    ta = TA.empty {
        TA.finals = [() | b],
        TA.trans = [([], [], ()) | b]
    }
}

-- | Make diagonal relation from /RR1/ relation.
diag :: RRn s f q -> RRn s f q
diag r | assert (arity r == 1) False = undefined
diag r = RRn {
    arity = 2,
    signature = signature r,
    ta = TA.map (\[Just f] -> [Just f, Just f]) id (ta r)
}

-- | Cartesian product (≈ cylindrification) of two /RRn/ relations
prod :: (Ord s, Ord f, Ord p, Ord q) => RRn s f p -> RRn s f q -> RRn s f (Maybe p, Maybe q)
prod r1 r2 = RRn {
    arity = arity r1 + arity r2,
    signature = S.nubOrd $ signature r1 ++ signature r2,
    ta = (TA.map mg id (prodAux (ta r1) (ta r2))) {
        TA.finals = [(Just p, Just q) | p <- TA.finals (ta r1), q <- TA.finals (ta r2)]
    }
} where
    mg (Just fs, Just gs) = fs ++ gs
    mg (Nothing, Just gs) = replicate (arity r1) Nothing ++ gs
    mg (Just fs, Nothing) = fs ++ replicate (arity r2) Nothing
    mg (Nothing, Nothing) = error "RRn.Basics.prod.mg: impossible"

-- | Lift operation on automata to /RRn/ relations,
-- assuming the underlying operation respects the arity.
unsafeLift :: (TA.TA [Maybe f] q -> TA.TA [Maybe f] q') ->
    RRn s f q -> RRn s f q'
unsafeLift f r = RRn {
    arity = arity r,
    signature = signature r,
    ta = f (ta r)
}

-- | Lift binary operation on automata to /RRn/ relations of equal arity,
-- assuming the underlying operation respects the arity.
unsafeLift2 :: (Ord s, Ord f) => (TA.TA [Maybe f] p -> TA.TA [Maybe f] q -> TA.TA [Maybe f] r) ->
    RRn s f p -> RRn s f q -> RRn s f r
unsafeLift2 _ r1 r2 | assert (arity r1 == arity r2) False = undefined
unsafeLift2 f r1 r2 = RRn {
    arity = arity r1,
    signature = S.nubOrd $ signature r1 ++ signature r2,
    ta = f (ta r1) (ta r2)
}

-- | De-duplicate states and transitions of underlying automaton.
unique :: (Ord s, Ord f, Ord q) => RRn s f q -> RRn s f q
unique = unsafeLift TA.unique

-- | Trim underlying automaton.
trim :: (Ord s, Ord q, Hashable q) => RRn s f q -> RRn s f q
trim = unsafeLift TA.trim

-- | Remove epsilon-closures from underlying automaton.
epsFree :: (Ord s, Ord f, Ord q, Hashable q) => RRn s f q -> RRn s f q
epsFree = unsafeLift TA.epsFree

-- | Trim underlying automaton.
minimize :: (Ord s, Ord f) => RRn s f Int -> RRn s f Int
minimize = unsafeLift TA.minimize

-- | Relabel underlying automaton.
relabel :: (Ord s, Ord q) => RRn s f q -> RRn s f Int
relabel = unsafeLift TA.relabel

-- | Check inclusion of /RRn/ relations.
subset :: (Ord f, Ord p) => RRn s f p -> RRn s f Int -> Bool
subset r1 r2 | assert (arity r1 == arity r2) False = undefined
subset r1 r2 = TA.subset (ta r1) (ta r2)

-- | Union of two RRn relations.
union :: (Ord s, Ord f, Ord p, Ord q) => RRn s f p -> RRn s f q -> RRn s f (Either p q)
union = unsafeLift2 TA.union

-- | Intersection of two RRn relations.
intersection :: (Ord s, Ord f, Ord p, Ord q) => RRn s f p -> RRn s f q -> RRn s f (p, q)
intersection = unsafeLift2 TA.intersection

-- | Difference of two RRn relations.
difference :: (Ord s, Ord f, Ord p) => RRn s f p -> RRn s f Int -> RRn s f (p, IS.IntSet)
difference = unsafeLift2 TA.difference
