-- | Projection operations on /RRn/ relations.
module RRn.Project (
    project,
    infinite,
    subst,
) where

import Prelude as P
import RRn.Type as RRn
import RRn.Basics
import qualified TA
import qualified Util.Set as S
import qualified Data.Map as M
import Control.Exception (assert)
import Util.Signature

import Data.Hashable

-- | Compute projection of an /RRn/ relation.
--
-- The inputs are a list of indices into the old /RRn/
-- relation, and an /RRn/ automaton.
--
-- Specification: L(project is r) = {map (rs !!) is | rs ∈ L(r)}
--
-- Note: All states of the given /RRn/ automaton should be
-- reachable. This property is preserved by projection.
project :: Ord f => [Int] -> RRn s f q -> RRn s f q
project is rrn | any (\i -> i < 0 || i >= arity rrn) is =
    error $ "RRn.project: index out of range " ++ show (is, arity rrn)
project is rrn =  RRn {
    arity = length is,
    signature = sig,
    ta = adjust (TA.map (\fs -> P.map (fs !!) is) id (ta rrn))
} where
   sig = signature rrn
   adjust ta = ta{ TA.trans = P.map adjustTrans (TA.trans ta) }
   adjustTrans (fs, qs, q) =
       (fs, take (maximum (0 : P.map (maybe 0 farity) fs)) qs, q)
   farity = (M.fromList (arities sig) M.!)

-- | Compute substitution of an /RRn/ relation.
--
-- Specification: L(subst is r) = {rs | map (rs !!) is ∈ L(r)}
--
-- Note: This function assumes that the given list of indices `is`
-- covers a consecutive range `[0..ari-1]` for some arity `ari`;
-- it does not do any cylindrification. Furthermore, the signature
-- is assumed to have a fixed arity for every function symbol.
subst :: Ord f => [Int] -> RRn s f q -> RRn s f q
subst is rrn
    | length is /= arity rrn =
        error $ "RRn.subst: length/arity mismatch " ++ show (is, arity rrn)
    | S.nubOrd is /= [0..ari - 1] =
        error $ "RRn.subst: index range not [0..ari-1]" ++ show (is, ari)
    | otherwise = RRn {
        arity = ari,
        signature = signature rrn,
        ta = (ta rrn) {
            TA.trans = [(fs', qs, q)
                | (fs, qs, q) <- TA.trans (ta rrn),
                  let fs' = P.map ((fs !!) . inv) [0..ari - 1],
                  fs == P.map (fs' !!) is]
        }
    }
  where
    ari = maximum (0 : P.map (+1) is)
    inv = (M.fromList (zip is [0..]) M.!)

-- | Compute infiniteness predicate from /RR2/ relation.
--
-- The inputs is an /RR2/ relation.
--
-- Specification: L(infinite fs r) = Inf(L(r)) where
-- Inf(R) = {[x] | {y | [x,y] ∈ R} is infinite}
infinite :: (Ord s, Ord f, Ord q, Hashable q) => RRn s f q -> RRn s f (Either q q)
infinite rrn | assert (arity rrn /= 2) False = undefined
infinite rrn = project [0] (trim rrn')
  where
    infR = TA.infinite TA.TA{
        TA.finals = undefined,
        TA.trans = [(f, ps, q)
            | ([Nothing, Just f], ps, q) <- TA.trans (ta rrn)],
        TA.eps = TA.eps (ta rrn)
    }
    rrn' = rrn {
        -- use Left/Right to keep track of infR state usage
        ta = TA.TA {
            TA.finals = P.map Right (TA.finals (ta rrn)),
            TA.trans =
                [ (fs, P.map b ps, b q)
                | (fs@[Nothing, Just _], ps, q) <- TA.trans (ta rrn),
                  let b q = if q `S.member` infR then Right q else Left q] ++
                [ (fs, P.map Left ps, Left q)
                | (fs@[Just _, Nothing], ps, q) <- TA.trans (ta rrn)] ++
                [ (fs, ps', q')
                | (fs@[Just _, Just _], ps, q) <- TA.trans (ta rrn),
                  (ps', q') <- pick1 ps q],
            TA.eps =
                [(Right p, Left p) | p <- S.toList (TA.states (ta rrn))] ++
                [(b p, b q) | (p, q) <- TA.eps (ta rrn), b <- [Left, Right]]
        }
    }
    pick1 [] q = [([], Left q)]
    pick1 (p : ps) q = (Right p : P.map Left ps, Right q) :
        [(Left p : ps', q') | (ps', q') <- pick1 ps q]
