-- | Basic tree automata functionality.
module TA.Basics (
    unique,
    merge,
    merges,
    union,
    intersection,
    restrict,
    reachable,
    productive,
    trim,
    epsFree,
    infinite,
    relabel,
    isEmpty,
) where

import TA.Type as TA
import Util.Saturate
import Util.Trancl
import qualified Util.Set as S
import qualified Data.Map as M
import qualified Data.Map.Strict as MS

import qualified Data.HashMap.Strict as HM
import qualified Data.HashSet as HS
import Data.Hashable

-- | De-duplicate states and transitions of an automaton.
unique :: (Ord f, Ord q) => TA f q -> TA f q
unique ta = TA {
    finals = S.nubOrd $ finals ta,
    trans  = S.nubOrd $ trans ta,
    eps    = S.nubOrd $ eps ta
}

-- | Merge two tree automata, taking the union of final states,
-- transitions, and epsilon transitions.
merge :: TA f q -> TA f q -> TA f q
merge ta1 ta2 = TA {
    finals = finals ta1 ++ finals ta2,
    trans  = trans ta1 ++ trans ta2,
    eps    = eps ta1 ++ eps ta2
}

-- | Merge and de-duplicate several tree automata.
merges :: (Ord f, Ord q) => [TA f q] -> TA f q
merges = unique . foldr merge TA.empty

-- | Compute union of two tree automata.
union :: TA f p -> TA f q -> TA f (Either p q)
union ta1 ta2 = TA.map id Left ta1 `merge` TA.map id Right ta2

-- | Compute intersection of two tree automata.
intersection :: (Ord f, Ord p, Ord q) => TA f p -> TA f q -> TA f (p, q)
intersection ta1 ta2 = TA {
    finals = [(p, q) | p <- finals ta1, q <- finals ta2],
    trans = [(f, zip ps qs, (p, q))
        | (f, ps, p) <- trans ta1,
          (qs, q) <- M.findWithDefault [] (f, length ps) transTa2],
    eps = [((p, q), (p', q)) | (p, p') <- eps ta1, q <- S.toList rs] ++
        [((p, q), (p, q')) | p <- S.toList ls, (q, q') <- eps ta2]
} where
    (ls, rs) = (states ta1, states ta2)
    transTa2 = MS.fromListWith (++) [((g, length qs), [(qs, q)])
                                    | (g, qs, q) <- trans ta2]

-- | Restrict states of a tree automaton.
restrict :: Ord q => TA f q -> S.Set q -> TA f q
restrict ta qs = TA {
    finals = filter (`S.member` qs) (finals ta),
    trans  = filter (\(_, ps, p) -> all (`S.member` qs) (p : ps)) (trans ta),
    eps    = filter (\(p, q) -> p `S.member` qs && q `S.member` qs) (eps ta)
}

-- | Compute reachable states of a tree automaton.
reachable :: (Ord q, Hashable q) => TA f q -> S.Set q
reachable ta = saturate sat0 sat1
  where
    trans' = HM.fromListWith (++) [(q, [(ps, p)]) | (_, ps, p) <- trans ta, q <- ps]
    sat0 = [p | (_, [], p) <- trans ta]
    sat1 q qs =
        [p | (q', p) <- eps ta, q == q'] ++
        [p | (ps, p) <- HM.lookupDefault [] q trans', all (`S.member` qs) ps]

-- | Compute productive states of a tree automaton.
productive :: (Ord q, Hashable q) => TA f q -> S.Set q
productive ta = saturate sat0 sat1
  where
    trans' = HM.fromListWith (++) [(p, [ps]) | (_, ps, p) <- trans ta]
    sat0 = finals ta
    sat1 q _ =
        [p | (p, q') <- eps ta, q == q'] ++
        [p | ps <- HM.lookupDefault [] q trans', p <- ps]

-- | Trim tree automaton.
trim :: (Ord q, Hashable q) => TA f q -> TA f q
trim ta = ta''
  where
    ta' = restrict ta (reachable ta)
    ta'' = restrict ta' (productive ta')

-- | Compute epsilon-free version of tree automaton.
epsFree :: (Ord f, Ord q, Hashable q) => TA f q -> TA f q
epsFree ta = TA {
    finals = finals ta,
    trans = S.nubOrd [(f, ps, q) | (f, ps, p) <- trans ta, q <- eps' p],
    eps = []
} where
    eps' = rtrancl' (eps ta)

-- | Compute states accepting an infinite number of terms.
--
-- Assumes that all states are reachable.
infinite :: (Ord q, Hashable q) => TA f q -> S.Set q
infinite ta = S.fromList [q | p <- loopy, q <- rdeps p]
  where
    eps' = rtrancl' (eps ta)
    deps = [(p, q') | (_, ps, q) <- trans ta, q' <- eps' q, p <- ps]
    rdeps = rtrancl' deps
    -- Compute loopy states, i.e., states for which there is a ground
    -- context C such that C[q] ->+ q. Since all states a reachable, we
    -- only need to check the spine of C.
    -- Infinite states are those that are reachable from a loopy state.
    loopy = S.nubOrd [p | (p, p') <- HS.toList (trancl deps), p == p']

-- | Relabel tree automaton.
relabel :: Ord q => TA f q -> TA f Int
relabel ta = TA.map id fq ta where
    fq = (M.fromAscList (zip (S.toAscList (states ta)) [0..]) M.!)

-- | Check emptiness of a tree automaton language.
isEmpty :: (Ord q, Hashable q) => TA f q -> Bool
isEmpty ta = all (not . (`S.member` reachable ta)) (finals ta)
