{-# LANGUAGE FlexibleContexts #-}

module TA.Minimize (
    minimize,
    collapseEps,
) where

import Prelude as P
import TA.Type as TA
import TA.Basics
import qualified Data.Graph as G
import qualified Data.Tree as T
import Control.Monad.ST
import Data.Array.ST
-- import Data.Array.Unboxed ((!))
import Data.Array.Base (unsafeWrite, unsafeAt)
import Control.Monad
import Data.List
import Util.Trancl
import qualified Util.Set as S
import qualified Data.Map as M

newSTUArray_ :: (MArray (STUArray s) e (ST s), Ix i) =>
    (i, i) -> ST s (STUArray s i e)
newSTUArray_ = newArray_

newSTArray :: Ix i => (i, i) -> e -> ST s (STArray s i e)
newSTArray = newArray

collapseEpsRef :: (Ord f, Ord q) => TA f q -> TA f Int
collapseEpsRef ta'
  | null (eps ta') = ta
  | otherwise = unique (TA.map id get ta)
  where
    ta = relabel ta'
    eps' = M.toList . M.fromListWith (++) $ [(p,[q]) | (p,q) <- eps ta]
    sccs = G.stronglyConnComp [(p, p, qs) | (p,qs) <- eps']
    eM = M.fromList [(y, x) | scc@(x:_) <- P.map G.flattenSCC sccs, y <- scc]
    get x = M.findWithDefault x x eM

-- | Collapse components induced by epsilon transitions.
collapseEps :: Ord f => TA f Int -> TA f Int
collapseEps ta
  | null (eps ta) = ta
  | otherwise = unique (TA.map id (unsafeAt epsSCC) ta)
  where
    epsSCC = runSTUArray $ do
        let sz = maximum . (-1 :) $ statesL ta
        res <- newSTUArray_ (0, sz)
        forM_ (zip (G.scc $ G.buildG (0, sz) (eps ta)) [0..]) $ \(t, i) -> do
            unsafeWrite res (T.rootLabel t) i
            forM_ (T.subForest t) $ \t' -> forM_ t' $ \n -> do
                unsafeWrite res n i
        return res

-- | Minimize the given tree automaton.
--
-- Assumes that -1 is not a state of the given automaton.

-- It's not clear how to minimize tree automata properly. In the code
-- below, a word automaton is built from the tree automaton that captures
-- paths through an accepted term. To this end, each transition
--   f(q_1..q_n) -> q
-- is turned into transitions
--   q_i --(f,[q_(i-1)..q_1],[q_(i+1)..q_n])-> q
-- for i = 1..n. Constants are artificially treated as unary function
-- symbols; for any transition c() -> q,
--   -1 --(f,[],[])-> q
-- is added to the word automaton (-1 is a fresh state). The final states
-- of the word automaton are those of the tree automaton; -1 is the intial
-- state.
--
-- `refine` finds a coarsest bisimulation of the word automaton; to this
-- end it takes a partition of the states (initially separating states
-- -1, the finaly states of the input automaton, and the non-final states),
-- and refines it according to observations (two states p ~ q are
-- distinguishable for a partition ~ if for some symbol x,
--   {[p']_~ | p' --x-> p} != {[q']_~ | q' --x-> q}

minimize :: Ord f => TA f Int -> TA f Int
minimize ta0 = unique (TA.map id (\i -> (\(Just x) -> x) $ findIndex (i `S.member`) discr) ta')
  where
    ta' = collapseEps ta0 -- unique ta0
    sz = maximum . (-1 :) $ statesL ta'
    rinits = scanl (flip (:)) []
    feps' = rtrancl' (eps ta')
    reps' = rtrancl' (P.map (\(p, q) -> (q, p)) (eps ta'))
    pseudotrans = S.nubOrd $ [(p, (f, psl, psr), q)
        | (f, ps, q) <- trans ta',
          (psl, p : psr) <- zip (rinits ps) (tails ps)] ++
        [(-1, (f,[],[]), q) | (f, [], q) <- trans ta']
    idx2 = M.fromListWith (++) [(k, [(p, q)]) | (p, k, q) <- pseudotrans]
    discr = let (l, r) = partition (`elem` finals ta') [0..sz] in
        refine $ P.map S.fromList [[-1], l, r]
    refine d = if length d' == length d then d else refine d'
      where
        d' = d >>= \pt ->
            case
                do  (p, q) <- eps ta'
                    guard $ q `elem` pt
                    let [ptp] = filter (p `S.member`) d
                    let t = S.fromList $ do
                            q' <- S.toList pt
                            guard $ any (`S.member` ptp) (reps' q')
                            return q'
                    guard $ S.size t /= S.size pt
                    return [t, S.difference pt t]
                ++
                do  (p, s, q) <- pseudotrans
                    guard $ q `elem` pt
                    let [ptp] = filter (p `S.member`) d
                    let t = S.fromList $ do
                            (p', q') <- idx2 M.! s
                            guard $ any (`S.member` ptp) (reps' p')
                            filter (`S.member` pt) (feps' q')
                    guard $ S.size t /= S.size pt
                    return [t, S.difference pt t]
            of  [] -> [pt]
                pt' : _ -> pt'
