module TA.NormalForm
    ( normalForm
    ) where

import Prelude as P
import Data.List (foldl1')
import qualified Data.Map.Strict as M
import qualified Data.Rewriting.Term as T
import qualified Data.Rewriting.Rule as R

import TA.Type (TA(..))
import Util.Saturate
import qualified Util.Set as S
import Util.Signature

-- | Tree automaton accepting normal forms w.r.t. the given TRS
--
-- A sorted version of the construction by (Comon 2000)
-- https://doi.org/10.1006/inco.1999.2838
--
-- The TRS is assumed to be left linear.
--
-- Resulting Automaton is not trim if some sort
-- has no ground normal form as a member.
normalForm :: (Ord f, Ord s)
           => Signature s f -> [R.Rule f v] -> TA f (T.Term f s)
normalForm sig trs =
    TA { finals = states
       , eps = []
       , trans = [ (f, ps, q) | (f, ss, _) <- sig
                 , ps <- combinations (length ss) states
                 , all (\(p, s) -> s == getSort' p) (zip ps ss)
                 , let trm = T.Fun f ps
                 , not $ isRedex trm
                 , let q = maxState trm
                 ]
       }
    where
        -- left-hand sides where variables are replaced by sorts
        lhss = [ applySort' s (R.lhs r) | r <- trs, s <- sortsR r ]
        sbtrms = S.nubOrd [ sbt | t <- lhss
                          , sbt@(T.Fun _ _) <- T.properSubterms t]

        states = [T.Var s | s <- S.nubOrd [s | (_, ss, s') <- sig , s <- s':ss]]
              ++ S.toList (saturate sat0 sat1)
        sat0 = filter (not . isRedex) sbtrms
        sat1 trm st = [ trm' | Just trm' <- P.map (unify' trm) (S.toList st)
                      , not $ isRedex trm'
                      ]

        -- due to the construction of `states`
        -- the set of states which are at least as general as a term
        -- contain a unique 'least general' term (maxState term).
        maxState trm = foldl1' (\p q -> if q `instOf` p then q else p)
                     $ filter (trm `instOf`) states

        isRedex t = any (t `instOf`) lhss
        instOf   = isInstanceOf getSort'
        unify'   = unify getSort'
        getSort' = getSort sig
        applySort' = applySort sig

        -- sort of the given rule
        -- If both terms are variables then return all sorts.
        sortsR r = case (R.lhs r, R.rhs r) of
          (T.Fun f _, _) -> [snd $ lookupSig sig f]
          (_, T.Fun f _) -> [snd $ lookupSig sig f]
          (T.Var _, T.Var _) -> sorts sig


-- | Construct most general term that is an instance of the two arguments
-- after renaming all variables apart
--
--  unify s t = unify t s = t  where s is the sort of t
--  unify (Fun f fs) (Fun f gs) = Fun f [unify l r | (l,r) <- zip fs gs]
unify :: (Eq f, Eq s)
      => (T.Term f s -> s) -> T.Term f s -> T.Term f s -> Maybe (T.Term f s)
unify getSort' = go
    where
        go t (T.Var s) = if s == getSort' t then Just t else Nothing
        go (T.Var s) t = if s == getSort' t then Just t else Nothing
        go (T.Fun f fs) (T.Fun g gs)
          | g == f    = T.Fun f <$> sequence (zipWith go fs gs)
          | otherwise = Nothing

-- | Checks if one term is an instance of another
-- after renaming all variables apart
--
-- Defined by the following inference rules:
--     s is sort of t => t `isInstanceOf` s
--
--     (forall 0 < i <= n. ti `isInstanceOf` si)
--     => f(t1,...,tn) `isInstanceOf` f(s1,...,sn)
--
-- Defines a partial order (>=) on terms,
-- where the maximal term is the least general.
isInstanceOf :: (Eq f, Eq s)
             => (T.Term f s -> s) -> T.Term f s -> T.Term f s -> Bool
isInstanceOf getSort' = go
    where
        go (T.Var _)    (T.Fun _ _)  = False
        go t            (T.Var s  )  = s == getSort' t
        go (T.Fun f fs) (T.Fun g gs) =
          f == g && all (uncurry go) (zip fs gs)

-- | Get sort of the term in the given signature
getSort :: Ord f => Signature s f -> T.Term f s -> s
getSort sig = go
    where
        go (T.Var s) = s
        go (T.Fun f fs) = case M.lookup (f, length fs) srtMp of
            Just s -> s
            Nothing -> error "Function symbol not in signature"
        srtMp = M.fromList [ ((f, length ss), s)
                           | (f, ss, s) <- sig
                           ]


-- | Replaces variables in term by their sort.
--
-- The second argument must be a non variable term.
applySort :: (Ord f, Ord s) => Signature s f -> s -> T.Term f v -> T.Term f s
applySort sig = go
    where
        sortOf (T.Fun f fs) = srtMp M.! (f, length fs)
        sortOf (T.Var _) =
            error "Cannot determine sort of a variable at a root position."
        srtMp = M.fromList [ ((f,length ss), (ss, s))
                           | (f, ss, s) <- sig
                           ]

        go s (T.Var _)      = T.Var s
        go _ t@(T.Fun f fs) = let (ss, _) = sortOf t
                               in T.Fun f (zipWith go ss fs)


-- Util functions

-- | Produces the list of all lists of length n containing elements form xs
--
--   ls in (combinations n xs) <=> length ls = n and all elements x of ls are in xs
--
combinations :: Int -> [a] -> [[a]]
combinations n xs
  | n <= 0    = [[]]
  | otherwise = [y:ys | y <- xs, ys <- combinations (n-1) xs]
