-- | Subset check and difference construction for tree automata.
module TA.Subset (
    subset,
    difference,
) where

import Prelude as P
import TA.Type
import TA.Det
import Util.Saturate
import qualified Util.Set as S
import qualified Data.IntSet as IS

-- Basic Hopcroft-Karp procedure for language inclusion.
--
-- We enumerate pairs (p, q) :: (p, S.Set q) such that there is a run
-- in `la` reaching state `p`, with the corresponding run to `q` in
-- the determinized version of `ra`.
--
-- `subset` checks whether for each such state, `q` is a final state
-- whenever `p` is.
--
-- `difference` uses the same enumeration to produce the states of the
-- resulting automaton.

-- | Check whether the first automaton is included in the second one.
subset :: (Ord f, Ord p) => TA f p -> TA f Int -> Bool
subset la ra = all check (saturateI sat0 sat1)
  where
    finalr = detFinal ra
    finall = (`S.member` S.fromList (finals la))
    (sat0, sat1) = subsetAux la ra
    check (p, q) = not (finall p) || finalr q

-- | Difference of two regular tree languages.
difference  :: (Ord f, Ord p) => TA f p -> TA f Int -> TA f (p, IS.IntSet)
difference la ra = TA {
    finals = filter (\(p, q) -> finall p && not (finalr q)) (S.toList ns),
    trans = S.toList . S.fromList $ [(f, zip ps qss, (p, transr f qss))
        | (f, ps, p) <- trans la,
          qss <- mapM (S.lookupFstL ns) ps],
    eps = [((p, q), (p', q)) | (p, p') <- eps la, q <- S.lookupFstL ns p]
} where
    (finalr, transr) = (detFinal ra, detTrans ra)
    finall = (`S.member` S.fromList (finals la))
    (sat0, sat1) = subsetAux la ra
    ns = saturate sat0 sat1

-- Enumeration of reachable states (p, q).
subsetAux :: (Ord f, Ord p) => TA f p -> TA f Int ->
   ([(p, IS.IntSet)], (p, IS.IntSet) -> S.Set (p, IS.IntSet) -> [(p, IS.IntSet)])
subsetAux la ra = (sat0, sat1)
  where
    transr = detTrans ra
    sat0 = [(p, transr f []) | (f, [], p) <- trans la]
    sat1 (p, q) rs = [(p', transr f qs)
         | (f, ps', p') <- trans la,
           p `elem` ps',
           let qss = P.map (S.lookupFstL rs) ps',
           all (not . null) qss, -- check that `sequence` produces a result
           qs <- sequence qss,
           q `elem` qs -- is this check worthwhile?
         ] ++
         [(p'', q) | (p', p'') <- eps la, p' == p]
