-- | Term and signature based constructions of tree automata.
module TA.Term (
    fromTerm,
    fromTermI,
    fromSig,
    fromMSSig,
    fromTermSI,
    ) where

import Prelude as P
import TA.Type as TA
import qualified Data.Rewriting.Term as T
import Control.Exception (assert)
import Util.Signature

mkTrans :: Signature s f -> T.Term f s -> [(f, [T.Term f s], T.Term f s)]
mkTrans sig v@(T.Var _) = [(f, P.map T.Var ss, T.Var s) | (f, ss, s) <- sig]
mkTrans _   t@(T.Fun f ts) = [(f, ts, t)]

varToSort :: Ord f => Signature s f -> T.Term f v -> T.Term f s
varToSort _ (T.Var _) =  error "Cannot determine sort of a variable at a root position."
varToSort sig t@(T.Fun f _) = go t (snd $ getSort f)
    where
        getSort = lookupSig sig
        go (T.Var _) s = T.Var s
        go (T.Fun f ts) _ = T.Fun f (zipWith go ts ss)
            where (ss, s) = getSort f

-- | Tree automaton accepting instances of given linear term.
fromTerm :: (Ord s, Ord f, Ord v) => Signature s f -> T.Term f v -> TA f (T.Term f s)
fromTerm _   t | assert (T.isLinear t) False = undefined
fromTerm sig (T.Var _) = TA.map id T.Var (fromMSSig sig (sorts sig))
fromTerm sig t@(T.Fun _ _) = TA {
    finals = [t'],
    trans  = T.subterms t' >>= mkTrans sig,
    eps    = []
} where
    t' = varToSort sig t

-- | Tree automaton accepting instances of given linear term,
-- with a distinguished interface state (`Nothing`).
fromTermI :: (Ord s, Ord f, Ord v) => Signature s f -> T.Term f v -> TA f (Maybe (T.Term f s))
fromTermI _   t | assert (T.isLinear t) False = undefined
fromTermI sig t@(T.Var _) = (TA.map id Just (fromTerm sig t)) {
        finals = [Nothing],
        eps = [(Just (T.Var s), Nothing) | s <- sorts sig]
    }
fromTermI sig t@(T.Fun _ _) = case t' of
    T.Fun f ts' -> ta' {
        finals = [Nothing],
        trans = (f, P.map Just ts', Nothing) : trans ta'
    }
  where
    t' = varToSort sig t
    ta' = (TA.map id Just (TA {
        finals = [],
        trans  = T.properSubterms t' >>= mkTrans sig,
        eps    = []
    })) { finals = [Nothing] }

-- | Tree automaton accepting instances of given linear term, of a given sort
-- with a distinguished interface state (`Nothing`).
fromTermSI :: (Ord s, Ord f, Ord v) => Signature s f -> s -> T.Term f v -> TA f (Maybe (T.Term f s))
fromTermSI _ _  t | assert (T.isLinear t) False = undefined
fromTermSI sig s t@(T.Var _) = (TA.map id Just (fromTerm sig t)) {
        finals = [Nothing],
        eps = [(Just (T.Var s), Nothing)]
    }
fromTermSI sig s t@(T.Fun _ _) = fromTermI sig t

-- | Tree automaton accepting all terms according to a simple signature.
fromSig :: Signature s f -> TA f ()
fromSig fs = fromMSSig (P.map (\(f,a) -> (f, replicate a (), ())) (arities fs)) [()]

-- | Tree automaton accepting all terms of one of the specified types,
-- according to given many-sorted signature.
fromMSSig :: Signature s f -> [s] -> TA f s
fromMSSig fs ts = TA {
    finals = ts,
    trans = fs,
    eps = []
}
