-- | Tree automaton type definition.
module TA.Type (
    TA (..),
    map,
    empty,
    states, statesL,
    lStates, lStatesL,
    rStates, rStatesL,
) where

import qualified Prelude as P
import Prelude hiding (map)
import qualified Data.Set as S

-- | A tree automaton.
data TA f q = TA {
    finals :: [q],            -- ^ final states
    trans  :: [(f, [q], q)],  -- ^ transitions
    eps    :: [(q, q)]        -- ^ epsilon transitions
}

instance (Show f, Show q) => Show (TA f q) where
    show ta = "TA {finals = " ++ show (finals ta) ++ ", trans/eps =\n" ++
        unlines ["  " ++ show f ++ show qs ++ " -> " ++ show q
                | (f, qs, q) <- trans ta] ++
        unlines ["  " ++ show p ++ " -> " ++ show q
                | (p, q) <- eps ta] ++
        "}"

-- | Map function symbols and states.
map :: (f -> f') -> (q -> q') -> TA f q -> TA f' q'
map ff fq ta = TA {
    finals = [fq q | q <- finals ta],
    trans  = [(ff f, P.map fq qs, fq q) | (f, qs, q) <- trans ta],
    eps    = [(fq p, fq q) | (p, q) <- eps ta]
}

-- | An empty TA.
empty :: TA f q
empty = TA { finals = [], trans = [], eps = [] }

-- | Compute states as set.
states :: Ord q => TA f q -> S.Set q
states = S.fromList . statesL

-- | Compute states as list.
statesL :: TA f q -> [q]
statesL ta =
    finals ta ++
    concat [q : qs | (_, qs, q) <- trans ta] ++
    concat [[p, q] | (p, q) <- eps ta]

-- | Compute states occurring in the lhs of a transition as set.
lStates :: Ord q => TA f q -> S.Set q
lStates = S.fromList . lStatesL

-- | Compute states occurring in the lhs of a transition as list.
lStatesL :: TA f q -> [q]
lStatesL ta = concat [qs | (_, qs, _) <- trans ta] ++ [p | (p, _) <- eps ta]

-- | Compute states occurring in the rhs of a transition as set.
rStates :: Ord q => TA f q -> S.Set q
rStates = S.fromList . lStatesL

-- | Compute states occurring in the rhs of a transition as list.
rStatesL :: TA f q -> [q]
rStatesL ta = [q | (_, _, q) <- trans ta] ++ [q | (_, q) <- eps ta]
