-- | Witness generating emptiness/subset check.
module TA.Witness (
    -- * Witness generation
    witness,
    witnessDiff,
    -- * General search algorithm
    prioritySearch
) where

import Prelude as P
import TA.Type
import qualified Data.Rewriting.Term as T
import qualified Data.OrdPSQ as PSQ
import qualified Data.Map.Strict as M
import Data.List
import Util.Ann

import qualified Util.Set as S
import TA.Det

-- | Produce witness of non-emptiness, if any; the witness will be one
-- of minimal weight according to the given weight function.
--
-- Weights must be non-negative.
witness :: Ord q => TA f q -> (f -> Int) -> Maybe (T.Term f ())
witness ta wf = annotation <$> prioritySearch (`elem` finals ta) sat0 sat1
  where
    -- `sat0` and `sat1` are analogous to those in `TA.Basics.reachable`.
    sat0 = [q :-> Ann (wf f) (T.Fun f []) | (f, [], q) <- trans ta]
    sat1 (Ann w wit) q wits =
        [p :-> Ann (wf f + sum (P.map value as)) (T.Fun f (P.map annotation as))
        | (f, ps, p) <- trans ta, q `elem` ps, all (`M.member` wits) ps,
          let as = P.map (wits M.!) ps] ++
        [p :-> Ann w wit | (q', p) <- eps ta, q == q']

-- | Produce witness that the first TA is not subset of the second
-- (i.e. difference is non-empty), if one exists;
-- the witness will be one of minimal weight
-- according to the given weight function.
--
-- Weights must be non-negative.
witnessDiff :: (Ord f, Ord q)
            => TA f q -> TA f Int -> (f -> Int) -> Maybe (T.Term f ())
witnessDiff la ra wf = annotation <$> prioritySearch (not . check) sat0 sat1
  where
    finalr = detFinal ra
    finall = (`S.member` S.fromList (finals la))
    check (p, q) = not (finall p) || finalr q
    transr = detTrans ra
    -- `sat0` and `sat1` are analogous to those in `TA.Subset.subsetAux`.
    sat0 = [ (p, transr f []) :-> Ann (wf f) (T.Fun f [])
           | (f, [], p) <- trans la]
    sat1 (Ann w wit) (p, q) wits =
      [ (p', transr f qs)
      :-> Ann (wf f + sum (P.map value as)) (T.Fun f (P.map annotation as))
      | (f, ps', p') <- trans la, p `elem` ps',
         let qss = P.map (lookupFstL wits) ps',
         all (not . null) qss, -- check that `sequence` produces a result
         qs <- sequence qss, q `elem` qs, -- is this check worthwhile?
         let as = P.map (wits M.!) (zip ps' qs)
      ] ++ [(p'', q) :-> Ann w wit | (p', p'') <- eps la, p' == p]


data Weighted a w = a :-> w

-- | Search similar to Dijkstra's shortest path algorithm.
--
-- By using @'Util.Ann.Ann' weight witness@ to annotate a states weight
-- with a witness (path), the smallest witness (shortest path) can be found.
--
-- See source of `witness` and `witnessDiff` for examples of use.
--
-- Note: Weight of successor states must be
-- greater or equal the weight of the current state.
-- (e.g.: @w' = w + weight a@, where @weight a@ is non-negative)
prioritySearch :: (Ord a, Ord w)
              => (a -> Bool) -- ^ Predicate determining goal state
              -> [Weighted a w] -- ^ Initial states with weights
              -> (w -> a -> M.Map a w -> [Weighted a w])
              -- ^ Function returning successor states given
              -- current weight, state, and weight map
              -> Maybe w
              -- ^ Returns smallest weight of path from
              -- some initial state to some goal state
prioritySearch isGoal sat0 sat1 = go (fromList sat0) M.empty
  where
    -- We use a priority queue to enumerate witnesses in order of increasing
    -- weight, similar to Dijkstra's shortest path algorithm.
    go pq0 wits = case PSQ.minView pq0 of
        Nothing -> Nothing
        Just (a, w, _, pq)
            | a `M.member` wits -> go pq wits
            | isGoal a -> Just w
            | otherwise -> go pq' wits'
          where
            wits' = M.insert a w wits
            pq' = foldl' (\pq (a :-> q) -> insertMin a q pq) pq
                (sat1 w a wits')

    fromList = PSQ.fromList . P.map (\(a :-> w) -> (a, w, ()))

insertMin :: (Ord a, Ord w) => a -> w -> PSQ.OrdPSQ a w () -> PSQ.OrdPSQ a w ()
insertMin key weight psq = case PSQ.insertView key weight () psq of
  (Nothing, psq') -> psq'
  (Just (prevWeight, _), psq')
    | prevWeight <= weight -> psq
    | otherwise -> psq'


-- | Similar to functions from 'Util.Set'
lookupFstL :: (Ord a, Ord b) => M.Map (a, b) c -> a -> [b]
lookupFstL m = M.keys . lookupFst m
    where
        filterFst :: (Ord a, Ord b) => M.Map (a, b) c -> a -> M.Map (a, b) c
        filterFst m a = M.takeWhileAntitone ((== a) . fst) .
            M.dropWhileAntitone ((< a) . fst) $ m

        lookupFst :: (Ord a, Ord b) => M.Map (a, b) c -> a -> M.Map b c
        lookupFst m = M.mapKeysMonotonic snd . filterFst m

