{-# LANGUAGE LambdaCase #-}
-- | This module contains Types and functions related
-- to properties on open (non-ground) terms.
module Util.OpenTerms (
  -- * Properties on open terms

  -- These properties need a signature extension to reduce to ground-properties
  Property (..),

  -- * Signature extension
  neededExtension,
  formalizedNeededExtension,
  -- ** Default results

  -- Use these when parsing a formula without knowing the system.
  defaultLLRGExtension,
  defaultLVExtension,

  -- * Predicates on TRSs

  -- The following predicates may be useful to determine how many additional
  -- constants are required.
  Rules.isGround,
  Rules.isLinear,
  isMonadic,
  isVarSep,
  -- * References
  -- $references
) where

import Data.Rewriting.Rule (Rule (..))
import qualified Data.Rewriting.Rules as Rules
import qualified Data.Rewriting.Term as Term

import qualified Util.Set as S

data Property
  = -- | Commutation
    COM
  | -- | Confluence
    CR
  | -- | Local-confluence
    WCR
  | -- | Strong-confluence
    SCR
  | -- | Unique normal-forms w.r.t. conversion
    UNC
  | -- | Unique normal-forms w.r.t. reduction
    UNR
  | -- | Normal from property
    NFP
  | -- | Conversion Equivalence
    CE
  | -- | Normalisation Equivalence
    NE
  deriving (Eq, Ord, Show)

-- | Given a @'Property'@ and a trs, returns the number of constants the signature
-- needs to extended by to reduce the property to the ground-property.
neededExtension :: Ord v => [Rule f v] -> Property -> Int
neededExtension = formalizedNeededExtension

formalizedNeededExtension :: Ord v => [Rule f v] -> Property -> Int
formalizedNeededExtension trs prop = minimum [i | (ps, i) <- lkup prop, all ($ trs) ps]
  where
    lkup COM = crResults
    lkup CR = crResults
    lkup WCR = crResults
    lkup SCR = crResults
    lkup UNR = crResults
    lkup NFP = crResults
    lkup UNC = uncResults
    lkup NE = neResults
    lkup CE = ceResults


-- | Returns the number of constants needed to reduce a property P to its
-- ground equivalent GP for left-linear right-ground (LLRG) systems.
--
-- Useful when then the system is unknown at the time of parsing the formula.
--
-- >>> defaultLLRGExtension CR
-- 1
-- >>> defaultLLRGExtension UNC
-- 2
defaultLLRGExtension :: Property -> Int
defaultLLRGExtension = \case
  UNC -> 2
  COM -> 1
  CR  -> 1
  WCR -> 1
  SCR -> 1
  UNR -> 1
  NFP -> 1
  CE  -> 1
  NE  -> 1

-- | Returns the number of constants needed to reduce a property P to its
-- ground equivalent GP for linear variable-separated (LV) systems.
--
-- Useful when then the system is unknown at the time of parsing the formula.
--
-- >>> defaultLVExtension CR
-- 2
-- >>> defaultLVExtension CE
-- 1
defaultLVExtension :: Property -> Int
defaultLVExtension = \case
  CE -> 1
  COM -> 2
  CR -> 2
  WCR -> 2
  SCR -> 2
  UNC -> 2
  UNR -> 2
  NFP -> 2
  NE -> 2

-- | Known results for reducing CR to GCR
crResults :: Ord v => [([[Rule f v] -> Bool], Int)]
crResults =
  [ ([Rules.isGround], 0), -- [1], [2]
    ([Rules.isLeftLinear, Rules.isRightGround, isMonadic], 0), -- [1], [2]
    ([Rules.isLeftLinear, Rules.isRightGround], defaultLLRGExtension CR), -- [1], [2]
    ([Rules.isLinear, isVarSep], defaultLVExtension CR), -- [2]
    ([], defaultLVExtension CR) -- default case equivalent to LV
  ]

-- | Known results for reducing UNC to GUNC
uncResults :: Ord v => [([[Rule f v] -> Bool], Int)]
uncResults =
  [ ([Rules.isGround], 0), -- [1], [2]
    ([Rules.isLeftLinear, Rules.isRightGround, isMonadic], 0), -- [1], [2]
    ([Rules.isLeftLinear, Rules.isRightGround], defaultLLRGExtension UNC), -- [1]
    ([Rules.isLinear, isVarSep], defaultLVExtension UNC), -- [2]
    ([], defaultLVExtension UNC) -- default case equivalent to LV
  ]

-- | Results for reducing NE to GNE
neResults :: Ord v => [([[Rule f v] -> Bool], Int)]
neResults =
  [ ([Rules.isGround], 0),
    ([Rules.isLeftLinear, Rules.isRightGround], defaultLLRGExtension NE),
    ([Rules.isLinear, isVarSep, isMonadic], 1),
    ([Rules.isLinear, isVarSep], defaultLVExtension NE),
    ([], defaultLVExtension NE) -- default case equivalent to LV
  ]

-- | Results for reducing CE to GCE
ceResults :: Ord v => [([[Rule f v] -> Bool], Int)]
ceResults =
  [ ([Rules.isGround], 0),
    ([Rules.isLinear, isVarSep, isMonadic], 1),
    ([Rules.isLinear, isVarSep], defaultLVExtension CE),
    ([], defaultLVExtension CE) -- default case equivalent to LV
  ]
-- Predicates

-- | Returns @'True'@ iff the given rules contain no function symbol with arity > 1
isMonadic :: [Rule f v] -> Bool
isMonadic trs = not (any nonMonadic terms)
  where
    terms = Rules.lhss trs ++ Rules.rhss trs
    nonMonadic = any (\(_, a) -> a > 1) . Term.funs . Term.withArity

-- | Returns @'True'@ iff the given rules are variable-separated. That is the set
-- of variables on the left and right hand side are disjoint.
isVarSep :: (Ord v) => [Rule f v] -> Bool
isVarSep = all ruleVarSep
  where
    ruleVarSep (Rule lhs rhs) =
      S.fromList (Term.vars lhs) `S.disjoint` S.fromList (Term.vars rhs)

-- $references
--
-- (1) F. Rapp and A.Middeldorp. Confluence properties on open terms in the
--   first-order theory of rewriting. IWC 2016
-- (2) B. Felgenhauer, A. Middeldorp, T. V. H. Prathamesh, and F. Rapp.
--   A verified ground confluence tool for linear variable-separated rewrite
--   systems in Isabelle/HOL. CPP 2019. doi:10.1145/3293880.3294098
