{-# LANGUAGE ScopedTypeVariables #-}
-- | Generic saturation by positive Horn inference.
--
-- Saturation is specified in terms of two functions `sat0` and `sat1`, where
-- `sat0` makes inferences that don't require any premises, while `sat1` is
-- given a new premise and a list of further premises that have already been
-- inferred, and computes consequences of that. For example, the transitive
-- closure of a relation can be computed as follows.
--
-- > -- | Compute transitive closure of a relation.
-- > trancl :: Ord a => [(a, a)] -> S.Set (a, a)
-- > trancl r = saturate sat0 sat1 where
-- >     sat0 = r
-- >     sat1 (x, y) _ = [(x, z) | (y', z) <- r, y == y']
module Util.Saturate (
    saturate,
    saturateI,
    saturateHash,
) where

import qualified Data.Set as S
import qualified Data.HashSet as HS
import Data.Hashable
import Data.List

-- | Generic saturation procedure.
saturate :: Ord x => [x] -> (x -> S.Set x -> [x]) -> S.Set x
saturate sat0 sat1 = foldl' go S.empty sat0 where
    go xs x
        | x `S.member` xs = xs
        | otherwise = let xs' = S.insert x xs in foldl' go xs' (sat1 x xs')

-- | Generic saturation procedure, incremental (i.e., lazy) version.
saturateI :: forall x. Ord x => [x] -> (x -> S.Set x -> [x]) -> [x]
saturateI sat0 sat1 = foldr go (\_ -> []) sat0 S.empty where
    go :: x -> (S.Set x -> [x]) -> S.Set x -> [x]
    go x c xs
        | x `S.member` xs = c xs
        | otherwise = let xs' = S.insert x xs in
            x : foldr go c (sat1 x xs') xs'

saturateHash :: (Eq x, Hashable x) => [x] -> (x -> HS.HashSet x -> [x]) -> HS.HashSet x
saturateHash sat0 sat1 = foldl' go HS.empty sat0 where
    go xs x
        | x `HS.member` xs = xs
        | otherwise = let xs' = HS.insert x xs in foldl' go xs' (sat1 x xs')
