{-# LANGUAGE OverloadedStrings #-}

module Main (
    main,
) where

import qualified Formula as Fm
import qualified Formula.Automata as Fm
import qualified Formula.Simplify as Fm
import qualified Formula.FORT as Fm
import qualified RRn
import qualified TA
import Util.Pretty
import Util.Rewriting
import Util.Signature

import qualified Problem as P
import qualified Data.Rewriting.Rule as R
import qualified Data.Rewriting.Term as T
import qualified Data.Rewriting.Term.Parse as TP
import qualified ArgumentParser as AP
import qualified Util.Set as S

import Satisfiability
import System.IO
import System.Exit
import System.Environment
import Control.Monad
import Data.List

import Prettyprinter as PP hiding (list)
import Prettyprinter.Render.Text
import Text.Parsec (parse, eof)

import System.Console.GetOpt

import Control.Parallel.Strategies
import Control.Concurrent (setNumCapabilities)
import Util.OpenTerms (defaultLVExtension, defaultLLRGExtension)


data Options = Options {
    optVerbose :: Bool,
    optLV :: Bool,
    optSigstep :: Int,
    optMaxArity :: Int,
    optMaxDepth :: Int,
    optMaxRules :: Int,
    optMinRules :: Int,
    optVars :: Int,
    optSig :: [(String, Int)],
    optNum :: Int
}

opt0 = Options {
    optVerbose = False,
    optLV = False,
    optSigstep = 2,
    optMaxArity = 3,
    optMaxDepth = 3,
    optMaxRules = 3,
    optMinRules = 0,
    optVars = 1,
    optSig = [],
    optNum = 1
}

options :: [OptDescr (Options -> IO Options)]
options = [
    Option "h?"  ["help"]
        (NoArg (\_ -> usage ExitSuccess))
        "print usage information",
    Option ""  ["verbose", "history"]
        (NoArg (\o -> return o{ optVerbose = True }))
        "activate additional history output",
    Option "l" ["lv"]
        (NoArg (\o -> return $ o{ optLV = True }))
        "search for linear variable separated ES",
    Option "n" ["num"]
        (ReqArg (\arg o -> do
            n <- parseNum "num (-n)" arg
            return $ o{ optNum = n }) "NUM")
        ("number of systems to synthesise, default: " ++ show (optNum opt0)),
    Option "S" ["signature"]
        (ReqArg (\arg o -> do
            sig <- parseSignature "signature" arg
            return $ o { optSig = sig }) "STRING")
        "signature, default: creates sig with sigstep (-s)",
    Option "a" ["arity"]
        (ReqArg (\s o -> do
            sig <- parseArities "arity (-a)" s
            return $ o{ optSig = sig }) "STRING")
        "signature with arities, default: creates sig with sigstep (-s)",
    Option "s" ["sigStep"]
        (ReqArg (\arg o -> do
            n <- parseNum "sigStep (-s)" arg
            return $ o{ optSigstep = n }) "NUM")
        ("steps to generate default signature, default: " ++ show (optSigstep opt0)),
    Option "A" ["maxArity"]
        (ReqArg (\arg o -> do
            n <- parseNum "maxArity (-A)" arg
            return $ o{ optMaxArity = n }) "NUM")
        ("maximal arity in generated signature, default: " ++ show (optMaxArity opt0)),
    Option "D" ["maxDepth"]
        (ReqArg (\arg o -> do
            n <- parseNum "maxDepth (-D)" arg
            return $ o{ optMaxDepth = n }) "NUM")
        ("maximum depth, default: " ++ show (optMaxDepth opt0)),
    Option "r" ["minNumRules"]
        (ReqArg (\arg o -> do
            n <- parseNum "minNumRules (-r)" arg
            return $ o{ optMinRules = n
                      , optMaxRules = max (optMaxRules o) n
                      }) "NUM")
        ("minimum number of rules, default: " ++ show (optMinRules opt0)),
    Option "R" ["maxNumRules"]
        (ReqArg (\arg o -> do
            n <- parseNum "maxNumRules (-R)" arg
            return $ o{ optMaxRules = n }) "NUM")
        ("maximum number of rules, default: " ++ show (optMaxRules opt0)),
    Option "v" ["vars"]
        (ReqArg (\arg o -> do
            n <- parseNum "vars (-v)" arg
            return $ o{ optVars = n }) "NUM")
        ("maximum number of variables, default: " ++ show (optVars opt0)),
    Option "j" ["jobs"]
        (ReqArg (\v o -> setNumCapabilities (read v) >> pure o) "NUM")
        "set number of jobs to run in parallel, default: 1"
    ]
  where
    parseNum name inp = parseArg AP.nat name inp
    parseSignature = parseArg AP.signatureParser
    parseArities name inp = generateSig <$> parseArg AP.intListParser name inp

    generateSig as = zip (map (\i -> 'c' : show i) [0 .. ]) as

    parseArg :: AP.ArgParser a -> String -> String -> IO a
    parseArg p name inp = case parse (p <* eof) name inp of
                    Left e -> ioError (userError $ show e)
                    Right f -> return f


usage :: ExitCode -> IO a
usage e = do
    hPutStrLn stderr $
        usageInfo header options
    exitWith e
  where
    header =
      unlines
        [ "Usage:\tfort-s [OPTION]... FORMULA [FILE.trs]..."
        , "Synthesise one or more systems with property FORMULA."
        , ""
        , "OPTIONS:"
        ]

main :: IO()
main = do
    args <- getArgs
    case getOpt Permute options args of
        (_,_ , e:_) -> do
            hPutStrLn stderr $ "Error: " ++ e
            usage (ExitFailure 1)
        (os, fs, []) -> do
            opt <- foldl (>>=) (return opt0) os
            -- putStrLn (show fs)
            main' opt fs
    where
        main' opt [] = do
            hPutStrLn stderr "Error: expected a formula\n"
            usage (ExitFailure 1)
        main' opt (formula : fs) = do
            let sig = optSig opt
                sigExtension =
                    if optLV opt
                        then defaultLVExtension
                        else defaultLLRGExtension
                numSys = optNum opt + length fs
                ixs = [0..numSys-1]
            case Fm.parseFormula sigExtension ixs formula of
                Left e -> do
                    hPutStrLn stderr "Error: formula could not be parsed"
                    return()
                Right f -> do
                    print . vsep $
                      [ hang 4 $ sep ["Property:", pretty formula]
                      , mempty
                      , hang 4 $ sep ["Internal representation:", pretty f]
                      ]
                    ps <- concat <$> mapM P.parseProblemFile fs
                    let trss = map P.rules ps
                        -- signature from files
                        sig' = Util.Signature.arities (S.nubOrd $ ps >>= P.signature)
                    searchTrs opt (sig ++ sig') f trss


-- adds new constant/function to default signature
increaseSig :: Options -> [(String, Int)] -> Int -> [(String,Int)]
increaseSig opt _ 0 = [("s0", 0)]
increaseSig opt oldSig cnt =
    let arity = mod (div cnt (optSigstep opt)) (optMaxArity opt + 1) in
    oldSig ++ [("s" ++ show cnt, arity)]

-- calls searchTrsUntilDepth with parameter
searchTrs :: Options -> [(String, Int)] -> Fm.Formula -> [[R.Rule String String]] -> IO()
searchTrs opt sig formula f
    | null (optSig opt) && f == [] = searchTrsUntilDepth opt (increaseSig opt sig 0) formula numSys 0 1 mempty [] f
    | otherwise = searchTrsUntilDepth opt sig formula numSys 0 1 mempty [] f
    where
      numSys = optNum opt
      toSys (Fm.Fwd i) = i
      toSys (Fm.Bwd i) = i

data GenState = GenSt
    { terms :: [T.Term String String]
    , rules :: [R.Rule String String]
    , trss :: [[R.Rule String String]]
    }

instance Semigroup GenState where
    GenSt ts1 rs1 trss1 <> GenSt ts2 rs2 trss2 = GenSt (ts1 <> ts2) (rs1 <> rs2) (trss1 <> trss2)

instance Monoid GenState where
    mempty = GenSt [] [] []

searchTrsUntilDepth :: Options -> [(String, Int)] -> Fm.Formula -> Int -> Int -> Int -> GenState -> [T.Term String String] -> [[R.Rule String String]] -> IO()
searchTrsUntilDepth opt sig formula numSys dCnt sigCnt genState newTerms file
    -- if max depth is reached: try again with increased sig
    | (dCnt > optMaxDepth opt) && null (optSig opt) =
        searchTrsUntilDepth opt (increaseSig opt sig sigCnt) formula numSys 0 (sigCnt+1) mempty [] file
    -- if max depth is reached: stop
    | dCnt > optMaxDepth opt =
        putStrLn "\nNo TRS found within maximum depth.\n"
    -- max depth is not reached
    | otherwise =
        let oldTerms = terms genState
            oldRules = rules genState
            oldTrss = trss genState
            newTerms' = getNewTerms sig (optVars opt) newTerms oldTerms
            newRules = getNewRules sig newTerms' oldTerms opt
            newTrss = getNewTrss oldRules newRules opt
            newTrss' = getMinDepthTrss dCnt newTrss
            newCand = getNewCand numSys newTrss' oldTrss
            newState =
                GenSt { terms = newTerms'
                      , rules = newRules
                      , trss = newTrss'
                      }
            nextState = genState <> newState
            foundTrss = checkGiven sig formula newCand file
        in
        case foundTrss of
          Nothing -> do
            when (optVerbose opt) $ do
                putStrLn("trss: " ++ show (Fm.pretty newCand))
                putStrLn("\nUsed Signature: " ++ show sig)
            searchTrsUntilDepth opt sig formula numSys (dCnt+1) sigCnt nextState newTerms' file
          Just result -> do
            when (optVerbose opt) $ do
                putStrLn("trs: " ++ show (Fm.pretty newCand))
            print . vsep $
              [ mempty
              , "Successfully found TRSs:"
              , indent 4 $ prettyResult result
              , mempty
              , hang 4 . sep $
                [ "Used Signature:"
                , fillSep (pretty <$> sig)
                ]
              ]
  where
    prettyResult :: [[R.Rule String String]] -> Doc ann
    prettyResult [trs] = vsep (Fm.pretty <$> trs)
    prettyResult trss = vsep $ zipWith go [0..] trss
      where
        go n trs =
          vsep [ "TRS" <+> pretty (n :: Int) <> ":"
               , indent 4 $ vsep (Fm.pretty <$> trs)
               ]

-- return trss that have at least depth min depth
getMinDepthTrss :: Int -> [[R.Rule String String]] -> [[R.Rule String String]]
getMinDepthTrss minDepth trss
    | minDepth == 0 = trss
    | otherwise = filter (not . all (R.both (\t -> depth t < minDepth))) trss
    where
        depth = T.fold (const 0) (\_ args -> if null args then 0 else 1+maximum args)

getNewCand :: Int -> [[R.Rule String String]] -> [[R.Rule String String]] -> [[[R.Rule String String]]]
getNewCand 0 newTrss oldTrss = []
getNewCand n newTrss oldTrss =
    [t : ts | t <- oldTrss, ts <- getNewCand (n - 1) newTrss oldTrss]
    ++ [t : ts | t <- newTrss, ts <- subPermLen (n - 1) allTrss]
  where
    allTrss = oldTrss ++ newTrss

    subPermLen 0 _ = [[]]
    subPermLen n xs = [ y:ys | y <- xs, ys <- subPermLen (n-1) xs ]




getNewTerms :: [(String, Int)] -> Int -> [T.Term String String] -> [T.Term String String] -> [T.Term String String]
-- get constants and variables
getNewTerms sig numVars [] [] = makeConst sig ++ makeVar numVars
-- get deeper nested terms
getNewTerms sig _ newTerms allTerms =
        let funL = getFuns sig in
        getNewFunTerms funL newTerms allTerms

-- get deeper nested terms
getNewFunTerms :: [(String, Int)] -> [T.Term String String] -> [T.Term String String] -> [T.Term String String]
getNewFunTerms [] _ _ = []
getNewFunTerms _ [] _ = []
--getNewFunTerms (f:fs) newTerms allTerms =
getNewFunTerms fs newTerms allTerms =
    [nT | (f, arity) <- fs, perm <- getPermuts arity newTerms allTerms, nT <- createTerm f [perm]]

getPermuts :: Int -> [T.Term String String] -> [T.Term String String] -> [[T.Term String String]]
getPermuts 1 newTerms _ = filter (\x -> checkVarRenaming x) (replicateM 1 newTerms)
--getPermuts arity nT allTerms = testOneNewTerm (filter (\x -> checkVarRenaming x) (replicateM arity allTerms)) nT --check
getPermuts arity _ allTerms = filter (\x -> checkVarRenaming x) (replicateM arity allTerms) --no check

testOneNewTerm :: [[T.Term String String]] -> [T.Term String String] -> [[T.Term String String]]
testOneNewTerm reps newTerms = [ts | nt <- newTerms, let nt' = nt, ts <- filter (elem nt') reps]

checkVarRenaming :: [T.Term String String] -> Bool
checkVarRenaming xs =
    let vars = varList xs in
        makeVar2 vars == makeVar (length vars)

varList :: [T.Term String String] -> [String]
varList xs = concat (map (\x -> T.vars x) xs)

-- get deeper nested trss
getNewTrss :: [R.Rule String String] -> [R.Rule String String] -> Options -> [[R.Rule String String]]
getNewTrss oldRules newRules opt = subseqLens minR maxR allRules
  where
    minR = optMinRules opt
    maxR = optMaxRules opt
    allRules = oldRules ++ newRules

subseqLens :: Int -> Int -> [a] -> [[a]]
subseqLens minN maxN xs = concatMap (`subseqLen` xs) [minN..maxN]

subseqLen 0 _ = [[]]
subseqLen _ [] = []
subseqLen n (x:xs) = map (x:) (subseqLen (n-1) xs) ++ subseqLen n xs

subseqMaxLen :: Int -> [a] -> [[a]]
subseqMaxLen n [] = [[]]
subseqMaxLen 0 _ = [[]]
subseqMaxLen n (x:xs) = map (x:) (subseqMaxLen (n-1) xs) ++ subseqMaxLen n xs

testOneNewRule :: [[R.Rule String String]] -> [R.Rule String String] -> [[R.Rule String String]]
testOneNewRule seq newRules = [rs | nr <- newRules, rs <- filter (elem nr) seq]

-- get deeper nested rules
getNewRules :: [(String, Int)] -> [T.Term String String] -> [T.Term String String] -> Options -> [R.Rule String String]
getNewRules sig newTerms oldTerms opt =
    let newLinearTerms = [i | i <- filter (\x -> T.isLinear x) newTerms]
        oldLinearTerms = [i | i <- filter (\x -> T.isLinear x) oldTerms]
        newRules = generateNewRules newLinearTerms oldLinearTerms opt
    in  makeRules newRules

generateNewRules :: [T.Term String String] -> [T.Term String String] -> Options -> [(T.Term String String, T.Term String String)]
generateNewRules newTerms oldTerms opt =
    let allLinear = (newTerms ++ oldTerms)
        allGround = filter (\x -> T.isGround x) (newTerms ++ oldTerms)
        newGround = filter (\x -> T.isGround x) newTerms
        oldTerms' = filter (\x -> not (T.isVar x)) oldTerms
        newTerms' = filter (\x -> not (T.isVar x)) newTerms
    in
        if optLV opt then --linear variable-separated terms
            [(x, y) | x <- newTerms , y <- allLinear, intersect (T.vars x) (T.vars y) == [] ]
            ++ [(x, y) | x <- oldTerms, y <- newTerms, intersect (T.vars x) (T.vars y) == []]
        else --left-linear right-ground terms
            [(x, y) | x <- newTerms' , y <- allGround] ++ [(x, y) | x <- oldTerms', y <- newGround]

makeRules :: [(T.Term String String, T.Term String String)] -> [R.Rule String String]
makeRules trss = map (\trs -> R.Rule (fst trs) (snd trs)) trss

createTerm :: String -> [[T.Term String String]] -> [T.Term String String]
createTerm fun permuts = map (\i -> T.Fun fun i) permuts

makeVar :: Int -> [T.Term String String]
makeVar num = map (\x -> T.Var x) (map (\i -> "x" ++ show i) [0 .. (num-1)])

makeVar2 :: [String] -> [T.Term String String]
makeVar2 xs = map (\x -> T.Var x) xs

makeConst :: [(String, Int)]  -> [T.Term String String]
makeConst sig = map (\i -> T.Fun i []) (map fst (filter (\x -> snd x == 0) sig))

getFuns :: [(String, Int)] -> [(String, Int)]
getFuns sig = filter (\x -> not (snd x == 0)) sig

toList :: [a] -> [[a]]
toList as = map (\a -> [a]) as

-- satisfies :: (Ord f, Ord v) => [(f, Int)] -> [[R.Rule f v]] -> Fm.Formula -> Bool
checkGiven :: [(String, Int)] -> Fm.Formula -> [[[R.Rule String String]]] -> [[R.Rule String String]] -> Maybe [[R.Rule String String]]
checkGiven sig formula trsss file =
    fmap fst $ find snd $ (map (\cand -> (cand, satisfies sig (cand ++ file) formula)) trsss `using` parBuffer 128 (evalTuple2 r0 rseq)) --par
    --(filter (\trs -> satisfies sig ([trs] ++ file) formula) trss) --nopar

-- pretty-print terms
instance (Pretty f, Pretty v) => Pretty (R.Rule f v) where
    pretty (R.Rule lhs rhs) = pretty lhs <> " -> " <> pretty rhs

-- pretty-print terms
instance (Pretty f, Pretty v) => Pretty (T.Term f v) where
    pretty (T.Var x) = pretty x
    pretty (T.Fun f []) = pretty f
    pretty (T.Fun f ts) = pretty f <> encloseSep' "(" ")" "," (map pretty ts)
