(******************************************************************************
 * Copyright (c) 2012-2014, Toyama&Aoto Laboratory, Tohoku University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the 
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(******************************************************************************
 * file: rwtools/commands/acp.sml
 * description: command interface for confluence checker
 * author: AOTO Takahito
 * 
 ******************************************************************************)

signature AGCP =  
sig
    val runDebug: bool ref
    val runProfile: bool ref
    val main: string * string list -> OS.Process.status
    val mkCommand: unit -> unit
    val install: unit -> unit
(*    val mkCommand2: int -> unit *)
end;

structure Agcp: AGCP  =
struct

val runDebug = ref false : bool ref
fun debug f = if !runDebug then f () else ()

val runProfile = ref false

local
    structure PU = PrintUtil
    val numOfSuccess = ref 0;
    val numOfFailure = ref 0;

    exception Abort;

    datatype option_result = HelpOption | Proceed | Error of string
    fun printOptionError [] = PU.printlnStdErr "option error?"
      | printOptionError ((Error str)::_) =  PU.printlnStdErr ("Error: " ^ str)
      | printOptionError (x::xs) = printOptionError xs

    val usage = "Usage: cmd <file or directory>, for details use --help"

    val useDirectOption = ref true
    val useCommutativeOption = ref true
    val useNonMinimalCommutativeOption = ref false
    val useLayerOption = ref true

    val timeLimit = ref 0 
    val timeLimitForInd = ref 5
    val verboseMode = ref 0
    val version = "0.02"

    val useCertifiableOutputOption = ref false

    fun setDebug () = 
	let val _ = if (!verboseMode) >= 3 then runDebug := true else ()
	    val _ = if (!verboseMode) >= 3 then Cr.runDebug := true else ()
	    val _ = if (!verboseMode) >= 3 then CrSolver.runDebug := true else ()
	    val _ = if (!verboseMode) >= 3 then CrCompletion.runDebug := true else ()
	    val _ = if (!verboseMode) >= 3 then CrDiagram.runDebug := true else ()
	    val _ = if (!verboseMode) >= 3 then NonCr.runDebug := true else ()
	    val _ = if (!verboseMode) >= 4 
		    then  
			print ("Warning:  Verbosity level of > 3 is unofficial. \n"
				 ^ "You need to delete temporary files *manually*.\n")
		    else ()
	    val _ = if (!verboseMode) >= 4 then  Solver.runDebug := true else ()
	    val _ = if (!verboseMode) >= 4 then  DpSolver.runDebug := true else ()
	    val _ = if (!verboseMode) >= 4 then  Dp.runDebug := true else ()
	    val _ = if (!verboseMode) >= 5 then  PolynomialInterpretation.runDebug := true else ()
	    val _ = if (!verboseMode) >= 5 then  MatrixInterpretation.runDebug := true else ()
	    val _ = if (!verboseMode) >= 5 then  PoSolver.runDebug := true else ()
	    val _ = if (!verboseMode) >= 5 then  Order.runDebug := true else ()
	in ()
	end

    fun resetDebug () = 
	let val _ = runDebug := false
	    val _ = Cr.runDebug := false
	    val _ = CrSolver.runDebug := false
	    val _ = CrCompletion.runDebug := false
	    val _ = CrDiagram.runDebug := false
	    val _ =  Solver.runDebug := false
	    val _ =  PoSolver.runDebug := false
	    val _ =  DpSolver.runDebug := false
	    val _ =  Dp.runDebug := false
	    val _ =  Order.runDebug := false
	    val _ = PolynomialInterpretation.runDebug := false
	    val _ = MatrixInterpretation.runDebug := false
	in ()
	end

    fun setInit () = (
        useDirectOption := true;
        useCommutativeOption := true;
        useNonMinimalCommutativeOption := false;
        useLayerOption := true;
        useCertifiableOutputOption := false;
        CrSolver.useHuetToyamaOostrom:= true;
        CrSolver.useHuetStrongClosed:= true;
        CrSolver.useOyamaguchiOhta:= true;
        CrSolver.useToyamaParallel:= true;
        CrSolver.useOkuiSimultaneous:= true;
        CrSolver.useDecreasingDiagrams := true;
        CrSolver.useOhtaOyamaguchiToyama:= true;
        CrSolver.useGomiOyamaguchiOhtaDepth:= true;
        CrSolver.useGomiOyamaguchiOhtaWeight:= true;
        CrSolver.useSuzukiAotoToyama:= true;
        CrSolver.useUchidaAotoToyama:= true;
        CrSolver.useUchidaAotoToyamaStrong:= true;
        CrSolver.useSakaiOgawa:= true;
        CrSolver.useKnuthBendix:= true;
        CrSolver.useHuetModulo:= true;
	NonCr.useNonConfluenceByApproximation := true;
	NonCr.useNonConfluenceByTreeAutomata := true;
	NonCr.useNonConfluenceByInterpretationAndOrder := true;
        CrSolver.useBidirectionalCompletion := true;
        CrSolver.proveGroundConfluence := false;
	resetDebug ();
        verboseMode := 1;
        timeLimit := 0)

    val _ = setInit ();

    fun setUseOption useOption str name =
	case name
            of "0" => (useOption := false; Proceed)
             | "1" => (useOption := true; Proceed)
             | "yes" => (useOption := true; Proceed)
             | "no" => (useOption := false; Proceed)
             | ow => Error ("illegal option: " ^ ow); 

    val tmpDir = ref (OS.FileSys.fullPath ".")

    fun setTmpDir path = 
        let val path' = OS.FileSys.fullPath path
            val _ = debug (fn _ => PU.printStdErr ("tmpDir: " ^ path' ^ "\n"))
        in if OS.FileSys.access (path', [OS.FileSys.A_WRITE])
           then (tmpDir := path'; Proceed)
           else Error ("directory not writable: " ^ path)
        end
        handle OS.SysErr (msg,_) => (Error ("directory not found: " ^ path))

    fun setExecutableFile name path =
        let val _ = debug (fn _ => PU.printStdErr ("[" ^ path ^ "]\n"))
	    val tokens = String.tokens (fn c => c = #" ") path
	    val command = hd tokens
	    val options = tl tokens
	    val path' = OS.FileSys.fullPath command
        in if OS.FileSys.access (path', [OS.FileSys.A_EXEC])
           then (name := String.concatWith " " (path'::options); Proceed)
           else Error ("file not executable: " ^ path)
        end
        handle OS.SysErr (msg,_) => Error ("file not found: " ^ path)

    val minisatPath = ref ""
    fun setMinisatPath path = setExecutableFile minisatPath path 

    val yicesPath = ref ""
    fun setYicesPath path = setExecutableFile yicesPath path 

    val terminationProver = ref ""
    fun setTerminationProver path = setExecutableFile terminationProver path 

    val relativeTerminationProver = ref ""
    fun setRelativeTerminationProver path = setExecutableFile relativeTerminationProver path 

    val certifiableOutputFile = ref ""

    fun setCertifiableOutputFile path =
        let val path' = OS.FileSys.fullPath path
        in if not (OS.FileSys.isDir path') 
	      andalso OS.FileSys.access (path', [OS.FileSys.A_WRITE])
           then (useCertifiableOutputOption := true;
		 debug (fn _ => PU.printStdErr ("certifiable output file: " ^ path' ^ "\n"));
		 certifiableOutputFile := path'; Proceed)
           else Error ("file not writable: " ^ path)
	end
	handle OS.SysErr (msg,_) => 
	       let val dir = OS.Path.getParent path
		   val file = OS.Path.file path
		   val dir' = OS.FileSys.fullPath dir
		   val path' = OS.Path.concat (dir',file)
               in if OS.FileSys.access (dir', [OS.FileSys.A_WRITE])
		  then (useCertifiableOutputOption := true;
			debug (fn _ => PU.printStdErr ("certifiable output file: " ^ path' ^ "\n"));
			certifiableOutputFile := path'; Proceed)
		  else Error ("file not writable: " ^ path)
	       end

    fun setCertifiableOutputOption path = 
	(CertifyUtil.acpVersion := version;
	 setCertifiableOutputFile path)
	
    fun setTimerOption num  =
        case Int.fromString num
         of SOME i => (timeLimit := i; Proceed)
          | NONE => Error ("illegal option: -t" ^ num)

    fun setDirectOption () = 
        (useDirectOption := false;
         useCommutativeOption := false;
	 useNonMinimalCommutativeOption := false;
         useLayerOption := false;
         Proceed)

    fun setVerboseOption num  =
	case Int.fromString num
         of SOME i => (verboseMode := i; setDebug (); Proceed)
          | NONE => Error ("illegal option: -v" ^ num)

    fun setResetOption () = 
		(CrSolver.useHuetToyamaOostrom:= false;
		 CrSolver.useHuetStrongClosed:= false;
		 CrSolver.useOyamaguchiOhta:= false;
		 CrSolver.useToyamaParallel:= false;
		 CrSolver.useOkuiSimultaneous:= false;
		 CrSolver.useDecreasingDiagrams := false;
		 CrSolver.useOhtaOyamaguchiToyama:= false;
		 CrSolver.useGomiOyamaguchiOhtaDepth:= false;
		 CrSolver.useGomiOyamaguchiOhtaWeight:= false;
		 CrSolver.useSuzukiAotoToyama:= false;
		 CrSolver.useUchidaAotoToyama:= false;
		 CrSolver.useUchidaAotoToyamaStrong:= false;
		 CrSolver.useSakaiOgawa:= false;
		 CrSolver.useKnuthBendix:= false;
		 CrSolver.useHuetModulo:= false;
		 NonCr.useNonConfluenceByApproximation := false;
		 NonCr.useNonConfluenceByTreeAutomata := false;
		 NonCr.useNonConfluenceByInterpretationAndOrder := false;
		 CrSolver.useBidirectionalCompletion := false;
                 CrSolver.proveGroundConfluence := false;
		 Proceed)

    val options = [{short = "h",
                    long = ["help"],
                    desc = GetOpt.NoArg (fn () => HelpOption),
                    help = "print this message"},
                   {short = "v",
                    long = ["verbose"],
                    desc = GetOpt.ReqArg (setVerboseOption, "0|1|2|3"),
                    help = ("verbosity level [0:YES/NO/MAYBE,1(default):YES/NO/MAYBE+explanation,2--]")},
                   (* {short = "", *)
                   (*  long = ["ps"], *)
                   (*  desc = GetOpt.ReqArg (setUseOption useDirectOption "ps","0|1|no|yes"), *)
                   (*  help = "use persistent decomposition [default:yes]"}, *)
                   (* {short = "", *)
                   (*  long = ["lp"], *)
                   (*  desc = GetOpt.ReqArg (setUseOption useLayerOption "lp","0|1|no|yes"), *)
                   (*  help = "use layer-preserving decomposition [default:yes]"}, *)
                   (* {short = "", *)
                   (*  long = ["cm"], *)
                   (*  desc = GetOpt.ReqArg (setUseOption useCommutativeOption "cm","0|1|no|yes"), *)
                   (*  help = "use commutative decomposition [default:yes]"}, *)
                   (* {short = "", *)
                   (*  long = ["nm"], *)
                   (*  desc = GetOpt.ReqArg (setUseOption useNonMinimalCommutativeOption  *)
                   (*                                     "nm","0|1|no|yes"), *)
                   (*  help = "with non-minimal commutative decomposition [default:no]"}, *)
                   (* {short = "d", *)
                   (*  long = ["direct"], *)
                   (*  desc = GetOpt.NoArg (fn () => setDirectOption ()), *)
                   (*  help = "equivalent to --ps=0 --lp=0 --cm=0"}, *)
                   {short = "t",
                    long = ["timer"],
                    desc = GetOpt.ReqArg (setTimerOption, "int"),
                    help = "specify the time limit in seconds [default:0 (no-limit)]"},
                   {short = "",
                    long = ["minisat-path"],
                    desc = GetOpt.ReqArg (setMinisatPath, "path"),
                    help = "specify an executable minisat [default:./minisat]"},
                   {short = "",
                    long = ["yices-path"],
                    desc = GetOpt.ReqArg (setYicesPath, "path"),
                    help = "specify an executable yices [default:./yices]"},
                   {short = "",
                    long = ["tmp-dir"],
                    desc = GetOpt.ReqArg (setTmpDir, "path"),
                    help = "specify a directory for putting temporal files [default:.]"}
                   (* {short = "", *)
                   (*  long = ["termination-prover"], *)
                   (*  desc = GetOpt.ReqArg (setTerminationProver, "path"), *)
                   (*  help = "specify an external termination prover [default:(internal)]"}, *)
                   (* {short = "", *)
                   (*  long = ["relative-termination-prover"], *)
                   (*  desc = GetOpt.ReqArg (setRelativeTerminationProver, "path"), *)
                   (*  help = "specify an external relative termination prover [default:(internal)]"}, *)
                   (* {short = "", *)
                   (*  long = ["certifiable-proofs-output"], *)
                   (*  desc = GetOpt.ReqArg (setCertifiableOutputOption, "path"), *)
                   (*  help = "specify output file for certifiable proofs"}, *)
                   (* {short = "", *)
                   (*  long = ["ground-confluence"], *)
                   (*  desc = GetOpt.NoArg (fn () => (CrSolver.proveGroundConfluence := true; Proceed)), *)
                   (*  help = "prove ground confluence"}, *)
                   (* {short = "r", *)
                   (*  long = ["reset"], *)
                   (*  desc = GetOpt.NoArg (fn () => setResetOption ()), *)
                   (*  help = "turn off all criteria by default (use before enable options)"}, *)
                   (* {short = "", *)
                   (*  long = ["enable-KB"], *)
                   (*  desc = GetOpt.NoArg (fn () => (CrSolver.useKnuthBendix := true; Proceed)), *)
                   (*  help = "(Knuth&Bendix,1970)"}, *)
                   (* {short = "", *)
                   (*  (* long = ["enable-strong"] 2014/4/14 overlap with strongly-quasi-linear is harmful *) *)
                   (*  long = ["enable-strongly-closed"], *)
                   (*  desc = GetOpt.NoArg (fn () => (CrSolver.useHuetStrongClosed := true; Proceed)), *)
                   (*  help = "(Huet,1980) strongly closed"}, *)
                   (* {short = "", *)
                   (*  long = ["enable-modulo"], *)
                   (*  desc = GetOpt.NoArg (fn () => (CrSolver.useHuetModulo := true; Proceed)), *)
                   (*  help = "(Huet,1980) Church-Rosser modulo"}, *)
                   (* {short = "", *)
                   (*  long = ["enable-parallel"], *)
                   (*  desc = GetOpt.NoArg (fn () => (CrSolver.useToyamaParallel := true; Proceed)), *)
                   (*  help = "(Toyama,1981) parallel critical pairs"}, *)
                   (* {short = "", *)
                   (*  long = ["enable-simple"], *)
                   (*  desc = GetOpt.NoArg (fn () => (CrSolver.useOhtaOyamaguchiToyama := true; Proceed)), *)
                   (*  help = "(Ohta&Oyamaguchi&Toyama,1995) simple-right-linear"}, *)
                   (* {short = "", *)
                   (*  long = ["enable-development"], *)
                   (*  desc = GetOpt.NoArg (fn () => (CrSolver.useHuetToyamaOostrom := true; Proceed)), *)
                   (*  help = "(Huet,1980)-(Toyama,1988)-(van Oostrom,1997) development closed"}, *)
                   (* {short = "", *)
                   (*  long = ["enable-simultaneous"], *)
                   (*  desc = GetOpt.NoArg (fn () => (CrSolver.useOkuiSimultaneous := true; Proceed)), *)
                   (*  help = "(Okui,1998) simultaneous critical pairs"}, *)
                   (* {short = "", *)
                   (*  long = ["enable-depth"], *)
                   (*  desc = GetOpt.NoArg (fn () => (CrSolver.useGomiOyamaguchiOhtaDepth := true; Proceed)), *)
                   (*  help = "(Gomi&Oyamaguchi&Ohta,1996/1998) strongly depth-preserving"}, *)
                   (* {short = "", *)
                   (*  long = ["enable-weight"], *)
                   (*  desc = GetOpt.NoArg (fn () => (CrSolver.useGomiOyamaguchiOhtaWeight := true; Proceed)), *)
                   (*  help = "(Gomi&Oyamaguchi&Ohta,1998) strongly weight-preserving"}, *)
                   (* {short = "", *)
                   (*  long = ["enable-upside"], *)
                   (*  desc = GetOpt.NoArg (fn () => (CrSolver.useOyamaguchiOhta := true; Proceed)), *)
                   (*  help = "(Oyamaguchi&Ohta,2004) upside-parallel-closed and outside-closed "}, *)
                   (* {short = "", *)
                   (*  long = ["enable-decreasing"], *)
                   (*  desc = GetOpt.NoArg (fn () => (CrSolver.useDecreasingDiagrams := true; Proceed)), *)
                   (*  help = "(van Oostrom,2008)-(Aoto,2010) decreasing diagram with rule-labelling"}, *)
                   (* {short = "", *)
                   (*  long = ["enable-shallow"], *)
                   (*  desc = GetOpt.NoArg (fn () => (CrSolver.useSakaiOgawa := true; Proceed)), *)
                   (*  help = "(Sakai&Ogawa,2010) weakly-non-overlapping non-collapsing and shallow"}, *)
                   (* {short = "", *)
                   (*  long = ["enable-completion"], *)
                   (*  desc = GetOpt.NoArg (fn () => (CrSolver.useBidirectionalCompletion := true; Proceed)), *)
                   (*  help = "(Aoto&Toyama,2012) reduction-preserving completion"}, *)
                   (* {short = "", *)
                   (*  long = ["enable-quasi-left-linear"], *)
                   (*  desc = GetOpt.NoArg (fn () => (CrSolver.useSuzukiAotoToyama := true; Proceed)), *)
                   (*  help = "(Suzuki&Aoto&Toyama,2013) quasi-left-linear"}, *)
                   (* {short = "", *)
                   (*  long = ["enable-quasi-linear"], *)
                   (*  desc = GetOpt.NoArg (fn () => (CrSolver.useUchidaAotoToyama := true; Proceed)), *)
                   (*  help = "(Aoto&Toyama&Uchida,2014) quasi-linear"}, *)
                   (* {short = "", *)
                   (*  long = ["enable-strongly-quasi-linear"], *)
                   (*  desc = GetOpt.NoArg (fn () => (CrSolver.useUchidaAotoToyamaStrong := true; Proceed)), *)
                   (*  help = "(Aoto&Toyama&Uchida,2014) strongly quasi-linear"}, *)
                   (* {short = "", *)
                   (*  long = ["enable-direct-approximation"], *)
                   (*  desc = GetOpt.NoArg (fn () => (NonCr.useNonConfluenceByApproximation := true; Proceed)), *)
                   (*  help = "disproving by direct approximations"}, *)
                   (* {short = "", *)
                   (*  long = ["enable-tree-automata"], *)
                   (*  desc = GetOpt.NoArg (fn () => (NonCr.useNonConfluenceByTreeAutomata := true; Proceed)), *)
                   (*  help = "(Jacquemard,1996)-(Durand&Middeldorp,1997) disproving by tree-automata approximation"}, *)
                   (* {short = "", *)
                   (*  long = ["enable-interpretation-and-order"], *)
                   (*  desc = GetOpt.NoArg (fn () => (NonCr.useNonConfluenceByInterpretationAndOrder := true; Proceed)), *)
                   (*  help = "(Aoto,2013) disproving by interpretation and ordering"} *)

                  (*                {short = "", *)
                  (*                 long = ["debug"], *)
                  (*             desc = GetOpt.NoArg (fn () => (runDebug := true; *)
                  (*                                Solver.runDebug := true; *)
                  (*                                PoSolver.runDebug := true; *)
                  (*                                DpSolver.runDebug := true; *)
                  (*                                Dp.runDebug := true; *)
                  (*                                CrSolver.runDebug := true; *)
                  (*                                Proceed)), *)
                  (*             help = "debug mode"} *)
                  ]

	exception ReadFileError

    fun checkFile (path,filename) =
        let 
	    open PrintUtil
            fun main () = 
                let 
                    val _ = useDirectOption := !useDirectOption
                    val _ = useCommutativeOption := !useCommutativeOption
                    val _ = useNonMinimalCommutativeOption := (!useCommutativeOption andalso 
							       !useNonMinimalCommutativeOption)
                    val _ = useLayerOption := !useLayerOption
                    val (_,rs,_) = case IOMsTpdb.rdFile filename of
				      SOME ans => ans
				    | NONE => raise Abort

                    val opt0 = {satSolver = (!minisatPath),
                                smtSolver = (!yicesPath),
                                terminationProver = (!terminationProver),
                                relativeTerminationProver = (!relativeTerminationProver),
                                tmpDir = (!tmpDir) }
                    val opt1 = DpSolver.defaultOptions
                    val opt2 = PoSolver.defaultOptions
                    val opt3 = {useDirect=(!useDirectOption),
                                useCommutative=(!useCommutativeOption),
                                useNonMinimalCommutative=(!useNonMinimalCommutativeOption),
                                useLayer=(!useLayerOption),
				useCertifiableOutputOption=(!useCertifiableOutputOption)}
		    val _ = CertifyUtil.outputFile := (!certifiableOutputFile)
                    val _ = debug (fn _ => PU.print ("\n"))

                    val _ = debug (fn _ => 
                                    let val mesg = 
                                             "Start Checking " 
                                              ^ (if path = "" 
                                                 then filename 
                                                 else (path ^ "/" ^ filename))
                                    in
                             PU.print (mesg ^ "\n");
                             PU.print (implode (List.tabulate (size mesg, fn x => #"=")));
                             PU.print "\n"
                                      end)

                    val _ = debug (fn _ => PU.print ("Persistent Decomposition: " ^ (if !useDirectOption 
                                                              then "on\n" else "off\n"))) 
                    val _ = debug (fn _ => PU.print ("Commutative Decompositoin: " ^ (if !useCommutativeOption 
                                                             then "on\n" else "off\n"))) 
                    val _ = debug (fn _ => PU.print ("Non-min Commutative Decomposition : " 
						     ^ (if !useNonMinimalCommutativeOption 
                                                             then "on\n" else "off\n"))) 
                    val _ = debug (fn _ => PU.print ("Layer-Preserving Decomposition: " ^ (if !useLayerOption 
                                                               then "on\n" else "off\n"))) 

                    val _ = debug (fn _ => PU.print ("Output Certifierable Proof Trees: " 
						     ^ (if !useCertifiableOutputOption 
                                                        then "on\n" else "off\n")))

                    val result  = CrSolver.crSolver  opt0 opt1 opt2 opt3 rs
				      
                in 
		    (PU.print (if path = "" then filename else (path ^ "/" ^ filename)); 
		     case result of
			 Cr.CR => (numOfSuccess := !numOfSuccess + 1; 
				   PU.println ": Success(CR)";
				   PU.printlnStdOut "YES")
		       | Cr.NotCR => (numOfSuccess := !numOfSuccess + 1; 
				      PU.println ": Success(not CR)";
				      PU.printlnStdOut "NO")
		       | Unknown => (numOfFailure := !numOfFailure + 1; 
				     PU.println ": Failure(unknown)";
				     PU.printlnStdOut "MAYBE"))
                end

            fun main2 () =  (* for Ground Confluence *)
                let 
                    val (decls,rs,_) = case IOMsTpdb.rdFile filename of
					   SOME ans => ans
					 | NONE => (print "Error: reading input\n";  raise Abort)

                    val _ = debug (fn _ => 
                                    let val mesg = 
                                             "Start Checking " 
                                              ^ (if path = "" 
                                                 then filename 
                                                 else (path ^ "/" ^ filename))
                                    in
                             PU.print (mesg ^ "\n");
                             PU.print (implode (List.tabulate (size mesg, fn x => #"=")));
                             PU.print "\n"
                                      end)

		    (* val (cPat,dPat,candidates) = Gcr.partRulesAndIndThm decls rs *)
		    val (cPat,dPat,cPatReducible,candidates) = Gcr.partRulesAndIndThmFSCD decls rs
(***		    val (cPat,dPat,cPatReducible,candidates) = Gcr.partRulesAndIndThmIWC decls rs ***)

		    val indOpt = { useE2 = false,
			     useQB = false,
				 useDC = false,
				 useSDC = false,
				 useSG = false,
				 useMG = false,
				 useESG = false,
				 useDP = false,
				 useTP = false,
				 useRS = false,
				 useLS = false  } 

		    fun try [] = ProverResult.MAYBE
		      | try (cand::rest) = 
			let val (msrs,conj) = cand
(***)
			    val _ = print "Trying with:\n"
			    val _ = print "R:\n"
			    val _ = print (Trs.prRules msrs)
			    val _ = print "E:\n"
			    val _ = print (Trs.prEqs conj)
(***)
			    (* val result = Riprover.gcrProver (!minisatPath) (!tmpDir) indOpt *)
			    (* 				    (decls, cPat,dPat, msrs, conj) *)
			    fun indmain () = Riprover.gcrProverFSCD (!minisatPath) (!tmpDir) indOpt
							    (decls, cPat,dPat, cPatReducible, msrs, conj)
			    val result = TimeUtil.timeLimitRun (IntInf.fromInt (!timeLimitForInd)) indmain () 
					 handle TimeLimit.TimeOut => 
						(if OS.FileSys.access (!Solver.inTmpFileName, [OS.FileSys.A_WRITE])
						 then (OS.FileSys.remove (!Solver.inTmpFileName);
						       Solver.inTmpFileName := "")
						 else ();
						 if OS.FileSys.access (!Solver.outTmpFileName, [OS.FileSys.A_WRITE])    
						 then (OS.FileSys.remove (!Solver.outTmpFileName);
						       Solver.outTmpFileName := "")
						 else ();
						 PU.println "rewriting induction stopped because of timeout.";
						 ProverResult.MAYBE)

			in if result = ProverResult.MAYBE
			   then try rest
			   else result
			end

		    val result = try candidates

                in 
		    (PU.print (if path = "" then filename else (path ^ "/" ^ filename)); 
		     case result of
			 ProverResult.YES => (numOfSuccess := !numOfSuccess + 1; 
				   PU.println ": Success(GCR)";
				   PU.printlnStdOut "YES")
		       | ProverResult.NO => (numOfSuccess := !numOfSuccess + 1; 
				      PU.println ": Success(not GCR)";
				      PU.printlnStdOut "NO")
		       | ProverResult.MAYBE => (numOfFailure := !numOfFailure + 1; 
				     PU.println ": Failure(unknown)";
				     PU.printlnStdOut "MAYBE"))
                end

            val timer = Timer.startCPUTimer ()
            val start = TimeUtil.checkTime timer 
            (*         val result = main ()   *)
	    val _ = CrSolver.proveGroundConfluence := true

            val result = TimeUtil.timeLimitRun (IntInf.fromInt (!timeLimit)) main2 () 

                handle TimeLimit.TimeOut => 
                       (if OS.FileSys.access (!Solver.inTmpFileName, [OS.FileSys.A_WRITE])
                        then (OS.FileSys.remove (!Solver.inTmpFileName);
                              Solver.inTmpFileName := "")
                        else ();
                        if OS.FileSys.access (!Solver.outTmpFileName, [OS.FileSys.A_WRITE])    
                        then (OS.FileSys.remove (!Solver.outTmpFileName);
                              Solver.outTmpFileName := "")
                        else ();
                        PU.print (if path = "" then filename else (path ^ "/" ^ filename));
                        numOfFailure := !numOfFailure + 1; 
                        PU.println ": Failure(timeout)";
			PU.printlnStdOut "MAYBE")
		     | ReadFileError =>
		       (PU.print (if path = "" then filename else (path ^ "/" ^ filename));
			numOfFailure := !numOfFailure + 1; 
			PU.println ": Failure(input file error)";
			PU.printlnStdOut "MAYBE")
            val stop = TimeUtil.checkTime timer 
            val _ = PU.print ("("  ^ (TimeUtil.reportTime (start,stop)) ^ " msec.)\n")
        in
            result
        end


    fun reportTotal timer = 
        let val _ = PU.print "\n=============================\n"
            val _ = PU.print ("Total: " 
                           ^ (Int.toString (!numOfSuccess)) ^ " Success, " 
                           ^ (Int.toString (!numOfFailure)) ^ " Failure: ")
            val _ = PU.print (Int.toString (LargeInt.toInt (TimeUtil.checkTime timer)) 
                           ^ " millisec.\n")
        in () end

in
fun main (name:string, args:string list) =
    let  
        val timer = Timer.startCPUTimer()
        val start = TimeUtil.checkTime timer 
        val commandDir = OS.FileSys.getDir();
	val _ = tmpDir := commandDir
        val _ = setInit ();

        val targets = 
            case (GetOpt.getOpt {argOrder = GetOpt.Permute,
                                 options = options,
                                 errFn = (fn str => (PU.printlnStdErr "option error\n"; 
						     PU.printlnStdErr usage; 
						     raise Abort))}
                                args) of
                (xs,args) =>  if List.exists (fn x => x = HelpOption) xs
                              then (PU.printlnStdErr (GetOpt.usageInfo 
                                                           {header=("agcp v" ^ version),
                                                            options=options});  
                                    raise Abort)
			      else if not (List.all (fn x => x = Proceed) xs)
			      then (printOptionError xs; PU.printlnStdErr usage; raise Abort)
                              else case args of 
                                       [] =>(PU.printlnStdErr "file or directory not specified";
                                            PU.printlnStdErr usage; raise Abort)
                                     | xs => (List.map 
                                                 (fn x => OS.FileSys.realPath x
                                                     handle OS.SysErr (msg,_) => 
                                                            (PU.printlnStdErr ("Error: file or directory not found: " 
                                                                                ^ x);
							     PU.printlnStdErr usage;
                                                             raise Abort))
                                                 xs)
       val _ = debug (fn _ => PU.printStdErr ("commandDir: " ^ commandDir ^ "\n"))
       val _ = debug (fn _ => PU.printStdErr ("tmpDir: " ^ (!tmpDir) ^ "\n"))


                val _ = if ("" = !minisatPath) 
                        then let val path' = OS.FileSys.fullPath "./minisat"
                                     handle OS.SysErr (msg,_) => 
                                            (PU.printlnStdErr ("Error: ./minisat not found; " 
                                                                ^ "use --minisat-path to specify minisat-path");
					     PU.printlnStdErr usage;
                                             raise Abort)
                             in if OS.FileSys.access (path', [OS.FileSys.A_EXEC])
                                then minisatPath := path'
                                else (PU.printlnStdErr ("Error: ./minisat not executable; "
                                                         ^ "use --minisat-path to specify minisat-path");
				      PU.printlnStdErr usage;
                                      raise Abort)
                             end
                        else ()

        val _ = if ("" = !yicesPath) 
                then let val path' = OS.FileSys.fullPath "./yices"
                             handle OS.SysErr (msg,_) => 
                                    (PU.printlnStdErr ("Error: ./yices not found; " 
                                                        ^ "use --yices-path to specify yices-path");
				     PU.printlnStdErr usage;
                                     raise Abort)
                     in if OS.FileSys.access (path', [OS.FileSys.A_EXEC])
                        then yicesPath := path'
                        else (PU.printlnStdErr ("Error: ./yices not executable; "
                                                 ^ "use --yices-path to specify yices-path");
			      PU.printlnStdErr usage;
                              raise Abort)
                     end
                else ()

		    val _ = if (!useCertifiableOutputOption)
			       andalso ((!useDirectOption) 
					orelse (!useCommutativeOption)
					orelse (!useLayerOption))
			    then (PU.printlnStdErr ("Error: currently, certification and decomposition techniques can not combied;");
				  PU.printlnStdErr ("       specify -d and --certifiable-proofs-output together"); 
				  raise Abort)
			    else ()


(**        val _ = if (!CrSolver.useHuetModulo)
                    andalso ("" = !relativeTerminationProver) 
                then 
                    (PU.println ("relative termination prover not specified; "
                                        ^ "disable some options or specify one using --relative-termination-prover");
                     raise Abort)
                else ()
**)

	val logFile = if ((!verboseMode) = 0)
		      then OS.Path.joinDirFile {dir="/dev",file="null"} 
		      else if ((!verboseMode) = 1)
		      then
			  let val tmpName = 
			      #file (OS.Path.splitDirFile (OS.FileSys.tmpName ()))
			  val _ = debug (fn () => print ("logfile: " ^ tmpName ^ "\n"))
			  in OS.Path.joinDirFile {dir=(!tmpDir),file=tmpName}
			  end
		      else ""

	val saveOut = ref (!PU.outs)
	val _ = if (!runDebug) orelse (!verboseMode) >= 2
		then PU.outs := TextIO.stdOut
		else PU.outs := TextIO.openOut logFile

	fun readIns ins = 
	    if TextIO.endOfStream ins 
	    then ()
	    else case TextIO.inputLine ins of
		     SOME line => (PU.printStdOut line; readIns ins)
		   | NONE => readIns ins

	fun finish () =  
	    if (!verboseMode = 1)
	    then let
		    val _ = TextIO.closeOut (!PU.outs)
		    val ins = TextIO.openIn logFile
		    val _ = readIns ins
		    val _ = TextIO.closeIn ins; 
		    val _ = PU.outs := (!saveOut)
		    val _ = if !runDebug then () else OS.FileSys.remove logFile
		in ()
		end
	    else (PU.outs := (!saveOut); ())

    in
        if length targets = 1 
           andalso not (OS.FileSys.isDir (hd targets))
        then 
            (CommandUtil.applyCmdToDirTree checkFile ("", hd targets);
	     finish ();
             setInit ();
             OS.Process.success)
        else
            (List.app (fn item => 
                          CommandUtil.applyCmdToDirTree checkFile ("", item))
                      targets;
             reportTotal timer;
 	        finish ();
             setInit ();
             OS.Process.success)
            handle e  => (OS.FileSys.chDir commandDir; finish (); raise (e))
    end
    handle Abort => OS.Process.failure
end

fun mkCommand () = SMLofNJ.exportFn("../../work/agcp/agcp",main);
fun install () = SMLofNJ.exportFn("work/acp/agcp",main);

(*
fun main2 1  (name:string, args:string list) = let
    val _ = CrCompletion.usePCP := false
    val _ = CrCompletion.useForward := false
    val _ = CrCompletion.useLinear := true
    val _ = CrCompletion.useParallel := false
    val _ = CrCompletion.useRelative := false
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := false
in main (name,args)end

| main2 2  (name:string, args:string list) = let 
    val _ = CrCompletion.usePCP := false
    val _ = CrCompletion.useForward := false
    val _ = CrCompletion.useLinear := true
    val _ = CrCompletion.useParallel := false
    val _ = CrCompletion.useRelative := true
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := false
in main (name,args)end

| main2 3  (name:string, args:string list) = let 
    val _ = CrCompletion.usePCP := false
    val _ = CrCompletion.useForward := false
    val _ = CrCompletion.useLinear := false
    val _ = CrCompletion.useParallel := true
    val _ = CrCompletion.useRelative := false
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := false
in main (name,args)end

| main2 4  (name:string, args:string list) = let 
    val _ = CrCompletion.usePCP := false
    val _ = CrCompletion.useForward := false
    val _ = CrCompletion.useLinear := false
    val _ = CrCompletion.useParallel := true
    val _ = CrCompletion.useRelative := true
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := false
in main (name,args)end

| main2 5  (name:string, args:string list) = let 
    val _ = CrCompletion.usePCP := true
    val _ = CrCompletion.useForward := false
    val _ = CrCompletion.useLinear := false
    val _ = CrCompletion.useParallel := true
    val _ = CrCompletion.useRelative := false
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := false
in main (name,args)end

| main2 6  (name:string, args:string list) = let 
    val _ = CrCompletion.usePCP := true
    val _ = CrCompletion.useForward := false
    val _ = CrCompletion.useLinear := false
    val _ = CrCompletion.useParallel := true
    val _ = CrCompletion.useRelative := true
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := false
in main (name,args)end

| main2 7  (name:string, args:string list) = let 
    val _ = CrCompletion.usePCP := true
    val _ = CrCompletion.useForward := false
    val _ = CrCompletion.useLinear := true
    val _ = CrCompletion.useParallel := true
    val _ = CrCompletion.useRelative := false
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := false
in main (name,args)end

| main2 8  (name:string, args:string list) = let 
    val _ = CrCompletion.usePCP := true
    val _ = CrCompletion.useForward := false
    val _ = CrCompletion.useLinear := true
    val _ = CrCompletion.useParallel := true
    val _ = CrCompletion.useRelative := true
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := false
in main (name,args)end

| main2 9  (name:string, args:string list) = let 
    val _ = CrCompletion.usePCP := false
    val _ = CrCompletion.useForward := false
    val _ = CrCompletion.useLinear := true
    val _ = CrCompletion.useParallel := false
    val _ = CrCompletion.useRelative := false
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := true
in main (name,args)end

| main2 10  (name:string, args:string list) = let 
    val _ = CrCompletion.usePCP := false
    val _ = CrCompletion.useForward := false
    val _ = CrCompletion.useLinear := true
    val _ = CrCompletion.useParallel := false
    val _ = CrCompletion.useRelative := true
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := true
in main (name,args)end

| main2 11  (name:string, args:string list) = let 
    val _ = CrCompletion.usePCP := true
    val _ = CrCompletion.useForward := false
    val _ = CrCompletion.useLinear := false
    val _ = CrCompletion.useParallel := true
    val _ = CrCompletion.useRelative := false
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := true
in main (name,args)end

| main2 12  (name:string, args:string list) = let 
    val _ = CrCompletion.usePCP := true
    val _ = CrCompletion.useForward := false
    val _ = CrCompletion.useLinear := false
    val _ = CrCompletion.useParallel := true
    val _ = CrCompletion.useRelative := true
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := true
in main (name,args)end

| main2 13  (name:string, args:string list) = let 
    val _ = CrCompletion.usePCP := true
    val _ = CrCompletion.useForward := false
    val _ = CrCompletion.useLinear := true
    val _ = CrCompletion.useParallel := true
    val _ = CrCompletion.useRelative := false
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := true
in main (name,args)end

| main2 14  (name:string, args:string list) = let 
    val _ = CrCompletion.usePCP := true
    val _ = CrCompletion.useForward := false
    val _ = CrCompletion.useLinear := true
    val _ = CrCompletion.useParallel := true
    val _ = CrCompletion.useRelative := true
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := true
in main (name,args)end

| main2  15  (name:string, args:string list) = let
    val _ = CrCompletion.usePCP := false
    val _ = CrCompletion.useForward := true
    val _ = CrCompletion.useLinear := true
    val _ = CrCompletion.useParallel := false
    val _ = CrCompletion.useRelative := false
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := true      
in main (name,args)end

| main2  16  (name:string, args:string list) = let
    val _ = CrCompletion.usePCP := false
    val _ = CrCompletion.useForward := true
    val _ = CrCompletion.useLinear := true
    val _ = CrCompletion.useParallel := false
    val _ = CrCompletion.useRelative := true
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := true      
in main (name,args)end

| main2  17  (name:string, args:string list) = let
    val _ = CrCompletion.usePCP := true
    val _ = CrCompletion.useForward := true
    val _ = CrCompletion.useLinear := false
    val _ = CrCompletion.useParallel := true
    val _ = CrCompletion.useRelative := false
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := true      
in main (name,args)end

| main2  18  (name:string, args:string list) = let
    val _ = CrCompletion.usePCP := true
    val _ = CrCompletion.useForward := true
    val _ = CrCompletion.useLinear := false
    val _ = CrCompletion.useParallel := true
    val _ = CrCompletion.useRelative := true
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := true      
in main (name,args)end

| main2  19  (name:string, args:string list) = let
    val _ = CrCompletion.usePCP := true
    val _ = CrCompletion.useForward := true
    val _ = CrCompletion.useLinear := true
    val _ = CrCompletion.useParallel := true
    val _ = CrCompletion.useRelative := false
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := true      
in main (name,args)end

| main2  20  (name:string, args:string list) = let
    val _ = CrCompletion.usePCP := true
    val _ = CrCompletion.useForward := true
    val _ = CrCompletion.useLinear := true
    val _ = CrCompletion.useParallel := true
    val _ = CrCompletion.useRelative := true
    val _ = CrCompletion.useHuet := false
    val _ = CrCompletion.useCompletion := true      
in main (name,args)end

| main2  21 (name:string, args:string list) = let
    val _ = CrCompletion.usePCP := true
    val _ = CrCompletion.useForward := true
    val _ = CrCompletion.useLinear := true
    val _ = CrCompletion.useParallel := true
    val _ = CrCompletion.useRelative := true
    val _ = CrCompletion.useHuet := true
    val _ = CrCompletion.useCompletion := true
in main (name,args)end

| main2  _ (name:string, args:string list) = let
    val _ = CrCompletion.usePCP := false
    val _ = CrCompletion.useForward := false
    val _ = CrCompletion.useLinear := false
    val _ = CrCompletion.useParallel := false
    val _ = CrCompletion.useRelative := false
    val _ = CrCompletion.useHuet := true
    val _ = CrCompletion.useCompletion := false
in main (name,args)end


fun mkCommand2 i  = SMLofNJ.exportFn("../../work/acp/acp",main2 i);

*)
end



