(******************************************************************************
 * file: rwtools/commands/kbcomp.sml
 * description: command interface for various Knuth-Bendix Completion Procedures
 * author: AOTO Takahito
 * 
 ******************************************************************************)

signature KBCOMP =  
    sig
	val runDebug: bool ref
	val runProfile: bool ref
	val main: string * string list -> OS.Process.status
	val mkCommand: unit -> unit
    end;

structure Kbcomp : KBCOMP =
    struct

    exception ReadFileError

    val runDebug = ref false : bool ref
    fun debug f = if !runDebug then f () else ()

    val runProfile = ref false

    local
	val numOfSuccess = ref 0;
	val numOfFailure = ref 0;

	exception Abort;

	datatype option_result = HelpOption | Proceed

	val useCheckerOption = ref false
	val useParallelOption = ref false
	val useMKBOption = ref false

	val usage = "Usage: cmd <file or directory>"
	
	val timeLimit = ref 0
	val timeLimitForSNChecker = ref 0

	fun setUseOption useOption str name =
	    let val _ = case name
			of "0" => useOption := false
			 | "1" => useOption := true
			 | "yes" => useOption := true
			 | "no" => useOption := false
			 | ow => (PrintUtil.println ("illegal option: -" ^ str ^ ow); 
				  PrintUtil.println usage;
				  raise Abort)
	    in Proceed
	    end

	val minisatPath = ref ""
	fun setMinisatPath path = 
	    let val path' = OS.FileSys.fullPath path
		    handle OS.SysErr (msg,_) => 
			   (PrintUtil.println ("file not found: " ^ path);
			    raise Abort)
	    in if OS.FileSys.access (path', [OS.FileSys.A_EXEC])
	       then (minisatPath := path'; Proceed)
	       else (PrintUtil.println ("file not executable: " ^ path);
		     raise Abort)
	    end

	val tmpDir = ref (OS.FileSys.fullPath ".")
	fun setTmpDir path = 
	    let val path' = OS.FileSys.fullPath path
		val _ = debug (fn _ => print ("tmpDir: " ^ path' ^ "\n"))
	    in if OS.FileSys.access (path', [OS.FileSys.A_WRITE])
	       then (tmpDir := path'; Proceed)
	       else (PrintUtil.println ("directory not writable: " ^ path);
		     raise Abort)
	    end

	val externalProver = ref ""
	fun setExternalProver path = 
	    let val path' = OS.FileSys.fullPath path
		    handle OS.SysErr (msg,_) => 
			   (PrintUtil.println ("file not found: " ^ path);
			    raise Abort)
	    in if OS.FileSys.access (path', [OS.FileSys.A_EXEC])
	       then (externalProver := path'; Proceed)
	       else (PrintUtil.println ("file not executable: " ^ path);
		     raise Abort)
	    end

	fun setTimerOption timeVar str num  =
	    if (!timeLimit) <> 0
		andalso !timeLimitForSNChecker <> 0
	    then
		(PrintUtil.println "not capable of multiple time limit"; 
		 PrintUtil.println usage;
		 raise Abort)
	    else
		let val _ = case Int.fromString num
                             of SOME i => timeVar := i
                              | NONE => (PrintUtil.println ("illegal option: -" ^ str ^ num); 
					 PrintUtil.println usage;
					 raise Abort)
		in Proceed
		end


	   val options = [{short = "h",
			   long = ["help"],
			   desc = GetOpt.NoArg (fn () => HelpOption),
			   help = "print this message"},
		       {short = "c",
		        long = ["check"],
			desc = GetOpt.NoArg (fn () => (useCheckerOption := true; Proceed)),
			help = "use a termination checker"},
		       {short = "p",
		        long = ["use-parallel"],
			desc = GetOpt.NoArg (fn () => (useParallelOption := true; Proceed)),
			help = "use parallel completion"},
		       {short = "m",
		        long = ["use-mkb"],
			desc = GetOpt.NoArg (fn () => (useMKBOption := true; Proceed)),
			help = "use parallel completion"},
                       {short = "t",
                        long = ["time-limit"],
                        desc = GetOpt.ReqArg (setTimerOption timeLimit "t", "int"),
                        help = "specify the time limit in seconds"},
                       {short = "T",
                        long = ["termination-time-limit"],
                        desc = GetOpt.ReqArg (setTimerOption timeLimitForSNChecker "T", "int"),
                        help = "specify the time limit for SN checker in seconds"},
                       {short = "",
                        long = ["minisat-path"],
                        desc = GetOpt.ReqArg (setMinisatPath, "path"),
                        help = "specify a executable minisat"},
                       {short = "",
                        long = ["tmp-dir"],
                        desc = GetOpt.ReqArg (setTmpDir, "dir"),
                        help = "specify a directory where temporal files will be placed"},
                       {short = "P",
                        long = ["termination-prover"],
                        desc = GetOpt.ReqArg (setExternalProver, "path"),
                        help = "specify a termination prover command"},
                       {short = "",
                        long = ["debug"],
                        desc = GetOpt.NoArg (fn () => (runDebug := true; 
(*                                                       Solver.runDebug := true;  *)
(*                                                       Ind.runDebug := true;  *)
                                                       Mkbtc.runDebug := true; 
                                                       Proceed)),
                        help = "debug mode"}
		      ]

	fun checkFile (path,filename) =
	    let 
		fun main () = 
		    let
			val opt1 = DpSolver.defaultOptions
			val opt2 = PoSolver.defaultOptions

                        val _ = debug (fn _ => print ("\n"))

                        val _ = debug (fn _ => 
                                          let val mesg = 
                                                  "Start Checking " 
                                                  ^ (if path = "" 
                                                     then filename 
                                                     else (path ^ "/" ^ filename))
                                          in
                                              print (mesg ^ "\n");
                                              print (implode (List.tabulate (size mesg, fn x => #"=")));
                                              print "\n"
                                          end)

			val _ = TestKb.minisatPath := !minisatPath
			val _ = TestKb.tmpDir := !tmpDir

			val _ = debug (fn _ => print ("Checker: " ^ (if !useCheckerOption 
							      then "on\n" else "off\n"))) 

			val _ = debug (fn _ => if !useCheckerOption 
					       then print ("Parallel: " ^ (if !useParallelOption 
									   then "on\n" else "off\n"))
					       else ())


			val result = if !useCheckerOption 
				     then 
					 TestKb.kbtcCheck 
					     (!minisatPath) (!tmpDir) opt1 opt2
					     (!useParallelOption)
					     (!useMKBOption)
					     (!externalProver)
					     (!timeLimitForSNChecker)
					     ((fn (func,rs,eq,conj,prec,poly,kbprec)
						  => (func,eq,conj))
						  (IOSpec.rdFile filename))
					 else 
					     TestKb.kbCheck 
						 ((fn (func,rs,eq,conj,prec,poly,kbprec)
						      => (func,eq,conj,prec,poly,kbprec))
						      (IOSpec.rdFile filename))
		    in 
                        (print (if path = "" then filename else (path ^ "/" ^ filename));
                         if result
			 then (numOfSuccess := !numOfSuccess + 1; 
			       print ": Success")
			 else (numOfFailure := !numOfFailure + 1; 
			       print ": Failure"))
		    end

		val timer = Timer.startCPUTimer ()
 		val start = TimeUtil.checkTime timer 
		val result = if !timeLimit = 0
			     then main () 
			     else TimeUtil.timeLimitRun (IntInf.fromInt (!timeLimit)) main () 
		    handle TimeLimit.TimeOut => 
                           (if OS.FileSys.access (!Solver.inTmpFileName, [OS.FileSys.A_WRITE])
			    then (OS.FileSys.remove (!Solver.inTmpFileName);
				  Solver.inTmpFileName := "")
			    else ();
			    if OS.FileSys.access (!Solver.outTmpFileName, [OS.FileSys.A_WRITE])	
			    then (OS.FileSys.remove (!Solver.outTmpFileName);
				  Solver.outTmpFileName := "")
			    else ();
			    print (if path = "" then filename else (path ^ "/" ^ filename));
                            numOfFailure := !numOfFailure + 1; 
                            print ": Failure(timeout)")
                val stop = TimeUtil.checkTime timer 
                val _ = print (" ("  ^ (TimeUtil.reportTime (start,stop)) ^ " msec.)\n")
	    in
		result
	    end

	fun reportTotal timer = 
            let val _ = print "\n=============================\n"
		val _ = print ("Total: " 
                               ^ (Int.toString (!numOfSuccess)) ^ " Success, " 
                               ^ (Int.toString (!numOfFailure)) ^ " Failure: ")
		val _ = print (Int.toString (LargeInt.toInt (TimeUtil.checkTime timer)) 
                               ^ " millisec.\n")
            in () end

	fun setInit () = ();

    in
    fun main (name:string, args:string list) =
	let  
	    val timer = Timer.startCPUTimer()
 	    val start = TimeUtil.checkTime timer 
	    val commandDir = OS.FileSys.getDir();
	    val _ = setInit ();
	    val targets = 
		case (GetOpt.getOpt {argOrder = GetOpt.Permute,
				     options = options,
				     errFn = (fn str => (PrintUtil.println usage;
							 raise Abort))}
				    args) of
		    (xs,args) =>  if List.exists (fn x => x = HelpOption) xs
				  then (PrintUtil.println (GetOpt.usageInfo 
							  {header="sprint v.0.08",
							   options=options});  
					raise Abort)
				  else case args of 
					   [] =>(PrintUtil.println "file or directory not specified";
						 PrintUtil.println usage; raise Abort)
					 | xs => List.map 
						 (fn x => OS.FileSys.realPath x
 						     handle OS.SysErr (msg,_) => 
							    (PrintUtil.println 
								 ("file or directory not found: " 
								  ^ x);
							     raise Abort))
						 xs

	    val _ = if "" = !minisatPath 
		    then let val path' = OS.FileSys.fullPath "./minisat"
				 handle OS.SysErr (msg,_) => 
					(PrintUtil.println ("./minisat not found; " 
						       ^ "use --minisat-path to specify minisat-path");
					 raise Abort)
			 in if OS.FileSys.access (path', [OS.FileSys.A_EXEC])
			    then minisatPath := path'
			    else (PrintUtil.println ("./minisat not executable; "
						^ "use -m to specify minisat-path");
				  raise Abort)
			 end
		    else ()
				 
	in
	    if length targets = 1 
	       andalso not (OS.FileSys.isDir (hd targets))
	    then 
		(CommandUtil.applyCmdToDirTree checkFile ("", hd targets);
		 setInit ();
		 OS.Process.success)
	    else
		(List.app (fn item => 
			      CommandUtil.applyCmdToDirTree checkFile ("", item))
			  targets;
		 reportTotal timer;
		 setInit ();
		 OS.Process.success)
	    handle e  => (OS.FileSys.chDir commandDir; raise (e))
	end
        handle Abort => OS.Process.failure
    end

    fun mkCommand () = SMLofNJ.exportFn("../../work/kbcomp/kbcomp",main);

end






