(******************************************************************************
 * file: rwtools/commands/rwchecer.sml
 * description: command interface for termination checker
 * author: AOTO Takahito
 * 
 ******************************************************************************)

signature RWCHECKER =  
sig
    val runDebug: bool ref
    val main: string * string list -> OS.Process.status
    val mkCommand: unit -> unit
end;

structure Rwchecker: RWCHECKER =
struct

val runDebug = ref true : bool ref
fun debug f = if !runDebug then f () else ()

local
    val numOfSuccess = ref 0;
    val numOfFailure = ref 0;

    exception Abort;

    datatype option_result = HelpOption | Proceed

    val usage = "Usage: cmd <file or directory>, for details use --help"
				
    val useDpOption = ref false
    val useRpOption = ref true
    val usePiOption = ref true
    val useScOption = ref true
    val useUrOption = ref true
    val useInOption = ref false

    val useQuasiOption = ref true
    val useLexOption = ref false
    val useMulOption = ref false
    val useAfOption = ref true

    val timeLimit = ref 0
    val multipleEncodingMethods = ref false (* $B7PO)=g=x$,(B2$B$D0J>e;XDj$5$l$J$$$3$H$r%A%'%C%/(B *)

    fun setInit () = (

        useDpOption := false;
        useRpOption := true;
        useScOption := true;
        useUrOption := true;
        useInOption := false;

        useQuasiOption := true;
        useLexOption := false;
        useMulOption := false;
        useAfOption := true;

        timeLimit := 0;
        multipleEncodingMethods := false)

    val _ = setInit ();


    fun setEncodingMethod name =
        if (!multipleEncodingMethods = true) then
            (PrintUtil.println "multiple -o option found"; 
             PrintUtil.println usage;
             raise Abort)
        else
            let val _ = multipleEncodingMethods := true;
                val _ = case name
                         of "c" => (useLexOption := false;useMulOption := false)
                          | "l" => (useLexOption := true;useMulOption := false)
                          | "m" => (useLexOption := false;useMulOption := true)
                          | "r" => (useLexOption := true;useMulOption := true)
                          | ow => (PrintUtil.println ("illegal option: -o" ^ ow); 
                                   PrintUtil.println usage;
                                   raise Abort)
            in Proceed
            end

    fun setUseOption useOption str name =
        let val _ = case name
                     of "0" => useOption := false
                      | "1" => useOption := true
                      | "yes" => useOption := true
                      | "no" => useOption := false
                      | ow => (PrintUtil.println ("illegal option: -" ^ str ^ ow); 
                               PrintUtil.println usage;
                               raise Abort)
        in Proceed
        end

    fun setExecutableFile name path =
        let val _ = debug (fn _ => print ("[" ^ path ^ "]\n"))
	    val tokens = String.tokens (fn c => c = #" ") path
	    val command = hd tokens
	    val options = tl tokens
	    val path' = OS.FileSys.fullPath command
                handle OS.SysErr (msg,_) => 
                       (PrintUtil.println ("file not found: " ^ path);
                        raise Abort)
        in if OS.FileSys.access (path', [OS.FileSys.A_EXEC])
           then (name := String.concatWith " " (path'::options); Proceed)
           else (PrintUtil.println ("file not executable: " ^ path);
                 raise Abort)
        end

    val minisatPath = ref ""
    fun setMinisatPath path = setExecutableFile minisatPath path 

    val yicesPath = ref ""
    fun setYicesPath path = setExecutableFile yicesPath path 

    val tmpDir = ref (OS.FileSys.fullPath ".")
    fun setTmpDir path = 
        let val path' = OS.FileSys.fullPath path
            val _ = debug (fn _ => print ("tmpDir: " ^ path' ^ "\n"))
        in if OS.FileSys.access (path', [OS.FileSys.A_WRITE])
           then (tmpDir := path'; Proceed)
           else (PrintUtil.println ("directory not writable: " ^ path);
                 raise Abort)
        end
        
    fun setTimerOption num  =
        let val _ = case Int.fromString num
                     of SOME i => timeLimit := i
                      | NONE => (PrintUtil.println ("illegal option: -t" ^ num); 
                                 PrintUtil.println usage;
                                 raise Abort)
        in Proceed
        end

    val options = [{short = "h",
                    long = ["help"],
                    desc = GetOpt.NoArg (fn () => HelpOption),
                    help = "print this message"},
                   {short = "o",
                    long = ["order"],
                    desc = GetOpt.ReqArg (setEncodingMethod,"name"),
                    help = "specify a path order c=>lpo,l=>lpos,m=>mpo,r=>rpo"},
                   {short = "",
                    long = ["qp"],
                    desc = GetOpt.ReqArg (setUseOption useQuasiOption "qp","0|1|no|yes"),
                    help = "use a quasi-order precedence"},
                   {short = "",
                    long = ["af"],
                    desc = GetOpt.ReqArg (setUseOption useAfOption "af","0|1|no|yes"),
                    help = "use argument filtering"},
                   {short = "d",
                    long = ["dp"],
                    desc = GetOpt.ReqArg (setUseOption useDpOption "d","0|1|no|yes"),
                    help = "use dependency pair method"},
                   {short = "",
                    long = ["sc"],
                    desc = GetOpt.ReqArg (setUseOption useScOption "sc","0|1|no|yes"),
                    help = "use subterm criteria"},
                   {short = "",
                    long = ["ur"],
                    desc = GetOpt.ReqArg (setUseOption useUrOption "ur","0|1|no|yes"),
                    help = "use usable rules criteria"},
                   {short = "",
                    long = ["rp"],
                    desc = GetOpt.ReqArg (setUseOption useRpOption "rp","0|1|no|yes"),
                    help = "use reduction pair"},
                   {short = "",
                    long = ["pi"],
                    desc = GetOpt.ReqArg (setUseOption usePiOption "rp","0|1|no|yes"),
                    help = "use polynomial interpretation"},
                   {short = "",
                    long = ["in"],
                    desc = GetOpt.ReqArg (setUseOption useInOption "in","0|1|no|yes"),
                    help = "use innermost termination check"},
                   {short = "t",
                    long = ["timer"],
                    desc = GetOpt.ReqArg (setTimerOption, "int"),
                    help = "specify the time limit in seconds"},
                   {short = "",
                    long = ["minisat-path"],
                    desc = GetOpt.ReqArg (setMinisatPath, "path"),
                    help = "specify a executable minisat"},
                   {short = "",
                    long = ["yices-path"],
                    desc = GetOpt.ReqArg (setYicesPath, "path"),
                    help = "specify an executable yices [default:./yices]"},
                   {short = "",
                    long = ["tmp-dir"],
                    desc = GetOpt.ReqArg (setTmpDir, "dir"),
                    help = "specify a directory where temporal files will be placed"},
                   {short = "",
                    long = ["debug"],
                    desc = GetOpt.NoArg (fn () => (runDebug := true; 
                                                   Solver.runDebug := true; 
                                                   PoSolver.runDebug := true; 
                                                   DpSolver.runDebug := true; 
                                                   Dp.runDebug := true; 
                                                   Proceed)),
                    help = "debug mode"}
                  ]

	exception ReadFileError
    fun checkFile (path,filename) =
        let 
            fun main () = 
                let 
                    val _ = useRpOption := ((!useDpOption) andalso (!useRpOption))
                    val _ = usePiOption := ((!useDpOption) andalso (!usePiOption))
                    val _ = useScOption := ((!useDpOption) andalso (!useScOption))
                    val _ = useUrOption := ((!useDpOption) andalso (!useUrOption))
                    val _ = useInOption := ((!useDpOption) andalso (!useInOption))
                    val _ = useAfOption := ((!useDpOption) andalso (!useAfOption))
                    val (rs,ss) = case IOTpdb.rdFile filename of
				 SOME (rs,ss) => (rs,ss)
			       | NONE=>raise ReadFileError
                    val opt1 = {useDp=(!useDpOption),
                                useRp=(!useRpOption),
                                usePi=(!usePiOption),
                                useSc=(!useScOption),
                                useUr=(!useUrOption),
                                useIn=(!useInOption)}
                    val opt2 = {useQuasi=(!useQuasiOption),
                                useAf=(!useAfOption),
                                useLex=(!useLexOption),
                                useMul=(!useMulOption)}

                    val _ = debug (fn _ => print ("\n"))

                    val _ = debug (fn _ => 
                                      let val mesg = 
                                              "Start Checking " 
                                              ^ (if path = "" 
                                                 then filename 
                                                 else (path ^ "/" ^ filename))
                                      in
                                          print (mesg ^ "\n");
                                          print (implode (List.tabulate (size mesg, fn x => #"=")));
                                          print "\n"
                                      end)

                    val _ = debug (fn _ => print ("DP: " ^ (if !useDpOption 
                                                            then "on\n" else "off\n"))) 
                    val _ = debug (fn _ => print ("SC: " ^ (if !useScOption 
                                                            then "on\n" else "off\n"))) 
                    val _ = debug (fn _ => print ("RP: " ^ (if !useRpOption 
                                                            then "on\n" else "off\n"))) 
                    val _ = debug (fn _ => print ("PI: " ^ (if !usePiOption 
                                                            then "on\n" else "off\n"))) 
                    val _ = debug (fn _ => print ("UR: " ^ (if !useUrOption 
                                                            then "on\n" else "off\n")))
                    val _ = debug (fn _ => print ("In: " ^ (if !useInOption 
                                                            then "on\n" else "off\n"))) 
                    val _ = debug (fn _ => print ("Quasi: " ^ (if !useQuasiOption 
                                                               then "on\n" else "off\n"))) 
                    val _ = debug (fn _ => print ("Lex: " ^ (if !useLexOption 
                                                             then "on\n" else "off\n"))) 
                    val _ = debug (fn _ => print ("Mul: " ^ (if !useMulOption 
                                                             then "on\n" else "off\n"))) 
                    val _ = debug (fn _ => print ("AF: " ^ (if !useAfOption 
                                                            then "on\n" else "off\n"))) 
                    val opt0 = {satSolver = (!minisatPath),
                                smtSolver = (!yicesPath),
                                terminationProver = "",
                                relativeTerminationProver = "",
                                tmpDir = (!tmpDir) }

                    val result  = 
			if (null ss)
			then if !useDpOption 
                             then DpSolver.dpSolver (!minisatPath) (!tmpDir) opt1 opt2 rs
                             else PoSolver.poSolver (!minisatPath) (!tmpDir) opt2 (rs,[])
			else MatrixInterpretation.relsnSolver opt0 opt1 opt2 (rs,ss)
			      
                in 
                    (print (if path = "" then filename else (path ^ "/" ^ filename));
                     if result
                     then (numOfSuccess := !numOfSuccess + 1; 
                           print ": Success")
                     else (numOfFailure := !numOfFailure + 1; 
                           print ": Failure"))
                end
            val timer = Timer.startCPUTimer ()
            val start = TimeUtil.checkTime timer 
			(*                 val result = main ()   *)
            val result = TimeUtil.timeLimitRun (IntInf.fromInt (!timeLimit)) main () 
                handle TimeLimit.TimeOut => 
                       (if OS.FileSys.access (!Solver.inTmpFileName, [OS.FileSys.A_WRITE])
                        then (OS.FileSys.remove (!Solver.inTmpFileName);
                              Solver.inTmpFileName := "")
                        else ();
                        if OS.FileSys.access (!Solver.outTmpFileName, [OS.FileSys.A_WRITE])        
                        then (OS.FileSys.remove (!Solver.outTmpFileName);
                              Solver.outTmpFileName := "")
                        else ();
                        print (if path = "" then filename else (path ^ "/" ^ filename));
                        numOfFailure := !numOfFailure + 1; 
                        print ": Failure(timeout)")
					 | ReadFileError =>
							(print (if path = "" then filename else (path ^ "/" ^ filename));
							 numOfFailure := !numOfFailure + 1; 
							 print ": Failure(input file error)")

            val stop = TimeUtil.checkTime timer 
            val _ = print (" ("  ^ (TimeUtil.reportTime (start,stop)) ^ " msec.)\n")
        in
            result
        end


    fun reportTotal timer = 
        let val _ = print "\n=============================\n"
            val _ = print ("Total: " 
                           ^ (Int.toString (!numOfSuccess)) ^ " Success, " 
                           ^ (Int.toString (!numOfFailure)) ^ " Failure: ")
            val _ = print (Int.toString (LargeInt.toInt (TimeUtil.checkTime timer)) 
                           ^ " millisec.\n")
        in () end

in
fun main (name:string, args:string list) =
    let  
        val timer = Timer.startCPUTimer()
        val start = TimeUtil.checkTime timer 
        val commandDir = OS.FileSys.getDir();
        val _ = setInit ();
        val targets = 
            case (GetOpt.getOpt {argOrder = GetOpt.Permute,
                                 options = options,
                                 errFn = (fn str => (PrintUtil.println usage;
                                                     raise Abort))}
                                args) of
                (xs,args) =>  if List.exists (fn x => x = HelpOption) xs
                              then (PrintUtil.println (GetOpt.usageInfo 
                                                           {header="rewriting checker v.0.04",
															options=options});  
                                    raise Abort)
                              else case args of 
                                       [] =>(PrintUtil.println "file or directory not specified";
                                             PrintUtil.println usage; raise Abort)
                                     | xs => List.map 
                                                 (fn x => OS.FileSys.realPath x
                                                     handle OS.SysErr (msg,_) => 
                                                            (PrintUtil.println ("file or directory not found: " 
																				^ x);
                                                             raise Abort))
                                                 xs

        val _ = if "" = !minisatPath 
                then let val path' = OS.FileSys.fullPath "./minisat"
                             handle OS.SysErr (msg,_) => 
                                    (PrintUtil.println ("./minisat not found; " 
														^ "use --minisat-path to specify minisat-path");
                                     raise Abort)
                     in if OS.FileSys.access (path', [OS.FileSys.A_EXEC])
                        then minisatPath := path'
                        else (PrintUtil.println ("./minisat not executable; "
                                                 ^ "use --minisat-path to specify minisat-path");
                              raise Abort)
                     end
                else ()

        val _ = if ("" = !yicesPath)
                then let val path' = OS.FileSys.fullPath "./yices"
                             handle OS.SysErr (msg,_) => 
                              (PrintUtil.println ("./yices not found; " 
						  ^ "use --yices-path to specify yices-path");
                                     raise Abort)
                     in if OS.FileSys.access (path', [OS.FileSys.A_EXEC])
                        then minisatPath := path'
                        else (PrintUtil.println ("./yices not executable; "
                                                 ^ "use --yices-path to specify yices-path");
                              raise Abort)
                     end
                else ()
    in
        if length targets = 1 
           andalso not (OS.FileSys.isDir (hd targets))
        then 
            (CommandUtil.applyCmdToDirTree checkFile ("", hd targets);
             setInit ();
             OS.Process.success)
        else
            (List.app (fn item => 
                          CommandUtil.applyCmdToDirTree checkFile ("", item))
                      targets;
             reportTotal timer;
             setInit ();
             OS.Process.success)
            handle e  => (OS.FileSys.chDir commandDir; raise (e))
    end
    handle Abort => OS.Process.failure
end

fun mkCommand () = SMLofNJ.exportFn("../../work/rwchecker/rwchecker",main);

end



