(******************************************************************************
 * file: rwtools/commands/sprint.sml
 * description: command interface for Rewriting Induction Prover
 * author: AOTO Takahito
 * 
 ******************************************************************************)

signature SPRINT =  
    sig
	val runDebug: bool ref
	val runProfile: bool ref
	val main: string * string list -> OS.Process.status
	val mkCommand: unit -> unit
    end;

structure Sprint : SPRINT =
    struct

    exception ReadFileError

    val runDebug = ref false : bool ref
    fun debug f = if !runDebug then f () else ()

    val runProfile = ref false

    local
	val numOfSuccess = ref 0;
	val numOfFailure = ref 0;

	exception Abort;

	datatype option_result = HelpOption | Proceed

	val usage = "Usage: cmd <file or directory>"
	
	val useE2Option = ref false  (* use expan2 *)
	val useQbOption = ref false  (* use quasi-basic subterm *)
	val useDcOption = ref false  (* use (original) divergence critic *)
	val useSdcOption = ref false (* use sound divergence critic *)
	val useSgOption = ref false  (* use sound generalization *)
	val useMgOption = ref false  (* use maximal generalization *)
	val useEsgOption = ref false (* use extended sound generalization *)
	val useDpOption = ref false  (* use disproof *)
	val useTpOption = ref false  (* use term partition *)
	val useRsOption = ref false  (* use reset *)
	val useLsOption = ref false  (* use lemma suggestion *)

	fun setAllOptions b = 
	    (useE2Option := b;
	     useQbOption := b; 
	     useDcOption := b;
	     useSdcOption := b;
	     useSgOption := b;
	     useMgOption := b;
	     useEsgOption := b;
	     useDpOption := b;
	     useTpOption := b;
	     useRsOption := b;
	     useLsOption := b)

	val timeLimit = ref 0

	fun setUseOption str useOption name =
            let val _ = case name
                         of "0" => useOption := false
                          | "1" => useOption := true
                          | "yes" => useOption := true
                          | "no" => useOption := false
                          | ow => (PrintUtil.println ("illegal option: -" ^ str ^ ow); 
                                   PrintUtil.println usage;
                                   raise Abort)
            in Proceed
            end

	val minisatPath = ref ""
	fun setMinisatPath path = 
	    let val path' = OS.FileSys.fullPath path
		    handle OS.SysErr (msg,_) => 
			   (PrintUtil.println ("file not found: " ^ path);
			    raise Abort)
	    in if OS.FileSys.access (path', [OS.FileSys.A_EXEC])
	       then (minisatPath := path'; Proceed)
	       else (PrintUtil.println ("file not executable: " ^ path);
		     raise Abort)
	    end

	val tmpDir = ref (OS.FileSys.fullPath ".")
	fun setTmpDir path = 
	    let val path' = OS.FileSys.fullPath path
		val _ = debug (fn _ => print ("tmpDir: " ^ path' ^ "\n"))
	    in if OS.FileSys.access (path', [OS.FileSys.A_WRITE])
	       then (tmpDir := path'; Proceed)
	       else (PrintUtil.println ("directory not writable: " ^ path);
		     raise Abort)
	    end

       fun setTimerOption num  =
            let val _ = case Int.fromString num
                        of SOME i => timeLimit := i
                         | NONE => (PrintUtil.println ("illegal option: -t" ^ num); 
                                    PrintUtil.println usage;
                                    raise Abort)
            in Proceed
            end


	val options = [{short = "h",
		        long = ["help"],
			desc = GetOpt.NoArg (fn () => HelpOption),
			help = "print this message"},
                       {short = "a",
                        long = ["all"],
                        desc = GetOpt.NoArg (fn () => (setAllOptions true; Proceed)),
                        help = "use all"},
		       {short = "e",
                        long = ["use-expand2"],
                        desc = GetOpt.ReqArg (setUseOption "e" useE2Option, "0|1|no|yes"),
                        help = "use expand2"},
                       {short = "q",
                        long = ["use-quasi-basic"],
                        desc = GetOpt.ReqArg (setUseOption "q" useQbOption, "0|1|no|yes"),
                        help = "use quasi-basic subterms"},
					   {short = "c",
                        long = ["use-divergence-critic"],
                        desc = GetOpt.ReqArg (setUseOption "c" useDcOption, "0|1|no|yes"),
                        help = "use original divergence critic"},
		       {short = "s",
                        long = ["use-sound-divergence-critic"],
                        desc = GetOpt.ReqArg (setUseOption "s" useSdcOption, "0|1|no|yes"),
                        help = "use sound divergence critic"},
                       {short = "g",
                        long = ["use-sound-generalization"],
                        desc = GetOpt.ReqArg (setUseOption "g" useSgOption, "0|1|no|yes"),
                        help = "use sound generalization"},
                       {short = "m",
                        long = ["use-maximal-generalization"],
                        desc = GetOpt.ReqArg (setUseOption "m" useMgOption, "0|1|no|yes"),
                        help = "use maximal generalization"},
(*                       {short = "",
                        long = ["use-extended-sound-generalization"],
                        desc = GetOpt.ReqArg (setUseOption "" useEsgOption, "0|1|no|yes"),
                        help = "use extended sound generalization"}, *)
                       {short = "d",
                        long = ["use-disproof"],
                        desc = GetOpt.ReqArg (setUseOption "d" useDpOption, "0|1|no|yes"),
                        help = "use disproof and decomposition"},
                       {short = "p",
                        long = ["use-term-partition"],
                        desc = GetOpt.ReqArg (setUseOption "p" useTpOption, "0|1|no|yes"),
                        help = "use term partition"},
                       {short = "r",
                        long = ["use-reset"],
                        desc = GetOpt.ReqArg (setUseOption "r" useRsOption, "0|1|no|yes"),
                        help = "use rest"},
                       {short = "l",
                        long = ["lemma-suggestion"],
                        desc = GetOpt.ReqArg (setUseOption "l" useLsOption, "0|1|no|yes"),
                        help = "use lemma suggestion"},
                       {short = "t",
                        long = ["timer"],
                        desc = GetOpt.ReqArg (setTimerOption, "int"),
                        help = "specify the time limit in seconds"},
                       {short = "",
                        long = ["minisat-path"],
                        desc = GetOpt.ReqArg (setMinisatPath, "path"),
                        help = "specify a executable minisat"},
                       {short = "",
                        long = ["tmp-dir"],
                        desc = GetOpt.ReqArg (setTmpDir, "dir"),
(*                        help = "specify a directory where temporal files will be placed"}, *)
                        help = "specify a directory for temporal files"},
                       {short = "",
                        long = ["debug"],
                        desc = GetOpt.NoArg (fn () => (runDebug := true; 
                                                       Solver.runDebug := true; 
                                                       Ind.runDebug := true; 
                                                       Proceed)),
                        help = "debug mode"}
		      ]

	fun checkFile (path,filename) =
	    let 
		fun main () = 
		    let
                        val _ = debug (fn _ => print ("\n"))

                        val _ = debug (fn _ => 
                                          let val mesg = 
                                                  "Start Checking " 
                                                  ^ (if path = "" 
                                                     then filename 
                                                     else (path ^ "/" ^ filename))
                                          in
                                              print (mesg ^ "\n");
                                              print (implode (List.tabulate (size mesg, fn x => #"=")));
                                              print "\n"
                                          end)

                        val _ = debug (fn _ => print ("QB: " ^ (if !useQbOption 
                                                              then "on\n" else "off\n"))) 
                        val _ = debug (fn _ => print ("DC: " ^ (if !useDcOption 
                                                              then "on\n" else "off\n"))) 
                        val _ = debug (fn _ => print ("SG: " ^ (if !useSgOption 
                                                               then "on\n" else "off\n"))) 
                        val _ = debug (fn _ => print ("DP: " ^ (if !useDpOption 
                                                               then "on\n" else "off\n"))) 
                        val _ = debug (fn _ => print ("TP: " ^ (if !useTpOption 
                                                               then "on\n" else "off\n"))) 

			val result = Riprover.riprover (!minisatPath) (!tmpDir)
				{ useE2 = (!useE2Option),
				 useQB = (!useQbOption),
				 useDC = (!useDcOption),
				 useSDC = (!useSdcOption),
				 useSG = (!useSgOption),
				 useMG = (!useMgOption),
				 useESG = (!useEsgOption),
				 useDP = (!useDpOption),
				 useTP = (!useTpOption),
				 useRS = (!useRsOption),
				 useLS = (!useLsOption) } 
				((fn (func,rs,eq,conj,prec,poly,kbprec)
				     => (func,rs,conj,prec))
				     (IOSpec.rdFile filename))
		    in 
                (print "Result: ";
                 print (if path = "" then filename else (path ^ "/" ^ filename));
                 if (result = ProverResult.YES orelse result = ProverResult.NO)
				 then (numOfSuccess := !numOfSuccess + 1; 
					   print ": Success")
				 else (numOfFailure := !numOfFailure + 1; 
					   print ": Failure"))
		    end

		val _ = print ("Filename: "
			       ^ (if path = "" then filename else (path ^ "/" ^ filename))
			       ^ "\n")

		val timer = Timer.startCPUTimer ()
 		val start = TimeUtil.checkTime timer 
		val result = TimeUtil.timeLimitRun (IntInf.fromInt (!timeLimit)) main () 
                    handle TimeLimit.TimeOut => 
                           (if OS.FileSys.access (!Solver.inTmpFileName, [OS.FileSys.A_WRITE])
			    then (OS.FileSys.remove (!Solver.inTmpFileName);
				  Solver.inTmpFileName := "")
			    else ();
			    if OS.FileSys.access (!Solver.outTmpFileName, [OS.FileSys.A_WRITE])	
			    then (OS.FileSys.remove (!Solver.outTmpFileName);
				  Solver.outTmpFileName := "")
			    else ();
                print "\nResult: ";
			    print (if path = "" then filename else (path ^ "/" ^ filename));
                            numOfFailure := !numOfFailure + 1; 
                            print ": Failure(timeout)")
                val stop = TimeUtil.checkTime timer 
                val _ = print (" ("  ^ (TimeUtil.reportTime (start,stop)) ^ " msec.)\n")
	    in
		result
	    end

	fun reportTotal timer = 
            let val _ = print "\n=============================\n"
		val _ = print ("Total: " 
                               ^ (Int.toString (!numOfSuccess)) ^ " Success, " 
                               ^ (Int.toString (!numOfFailure)) ^ " Failure: ")
		val _ = print (Int.toString (LargeInt.toInt (TimeUtil.checkTime timer)) 
                               ^ " millisec.\n")
            in () end

	fun setInit () = setAllOptions false

    in
    fun main (name:string, args:string list) =
	let  
	    val timer = Timer.startCPUTimer()
 	    val start = TimeUtil.checkTime timer 
	    val commandDir = OS.FileSys.getDir();
	    val _ = setInit ();
	    val targets = 
		case (GetOpt.getOpt {argOrder = GetOpt.Permute,
				     options = options,
				     errFn = (fn str => (PrintUtil.println usage;
							 raise Abort))}
				    args) of
		    (xs,args) =>  if List.exists (fn x => x = HelpOption) xs
				  then (PrintUtil.println (GetOpt.usageInfo 
							  {header="sprint v.0.10",
							   options=options});  
					raise Abort)
				  else case args of 
					   [] =>(PrintUtil.println "file or directory not specified";
						 PrintUtil.println usage; raise Abort)
					 | xs => List.map 
						 (fn x => OS.FileSys.realPath x
 						     handle OS.SysErr (msg,_) => 
							    (PrintUtil.println 
								 ("file or directory not found: " 
								  ^ x);
							     raise Abort))
						 xs

	    val _ = if "" = !minisatPath 
		    then let val path' = OS.FileSys.fullPath "./minisat"
				 handle OS.SysErr (msg,_) => 
					(PrintUtil.println ("./minisat not found; " 
						       ^ "use --minisat-path to specify minisat-path");
					 raise Abort)
			 in if OS.FileSys.access (path', [OS.FileSys.A_EXEC])
			    then minisatPath := path'
			    else (PrintUtil.println ("./minisat not executable; "
						^ "use -m to specify minisat-path");
				  raise Abort)
			 end
		    else ()
				 
	in
	    if length targets = 1 
	       andalso not (OS.FileSys.isDir (hd targets))
	    then 
		(CommandUtil.applyCmdToDirTree checkFile ("", hd targets);
		 setInit ();
		 OS.Process.success)
	    else
		(List.app (fn item => 
			      CommandUtil.applyCmdToDirTree checkFile ("", item))
			  targets;
		 reportTotal timer;
		 setInit ();
		 OS.Process.success)
	    handle e  => (OS.FileSys.chDir commandDir; raise (e))
	end
        handle Abort => OS.Process.failure
    end

    fun mkCommand () = SMLofNJ.exportFn("../../work/sprint/sprint",main);

end






