(******************************************************************************
 * file: rwtools/kbcomp/kbtc.sml
 * description: Knuth-Bendix Completion Procedure With Termination Checker
 * author: AOTO Takahito
 * 
 * Copyright (c) 2008, Toyama&Aoto Laboratory, Tohoku University
 ******************************************************************************)

signature KBTC = 
sig
    (* $BDd;_@-%A%'%C%+!<$rMQ$$$?40Hw2=<jB3$-(B *)
    val kbtc : ((Term.term * Term.term) list -> bool) -> Trs.equations -> bool

    (* $BDd;_@-%A%'%C%+!<$rMQ$$$?JBNs40Hw2=<jB3$-(B *)
    val pkbtc : ((Term.term * Term.term) list -> bool) -> Trs.equations -> bool
end;

structure Kbtc : KBTC = 
   struct

   local
       structure L = List
   in

   exception KbtcError

   (*   RuleSet$B$+$i%5%$%:$,:G$b>.$5$$5,B'(Br$B$rA*$S!$AH(B( r, RuleSet - {r} ) $B$rJV$94X?t(B  *)
   fun orientation isTerminating (cs, rs, []) = (print "no equation to expand\n";  raise KbtcError)
    |  orientation isTerminating (cs, rs, es) =
	let

            (* $B9`%5%$%:(B($BD9(B,$BC;(B)$B$HBP1~$9$k9`$r4^$s$@%j%9%H$r%=!<%H$7JV$9(B *)
            fun makeVarNumlist rs =
		let 
                    fun withSize (l,r) = let val llen = Term.termSize l
					     val rlen = Term.termSize r
					 in if rlen > llen
					    then ((rlen,llen),(r,l))
					    else ((llen,rlen),(l,r))
					 end
		in
                    ListMergeSort.sort 
			(fn (((n1,m1),_),((n2,m2),_)) => n1 > n2 orelse 
							 (n1 = n2 andalso m1 > m2))
			(List.map (fn lr => withSize lr) rs)
		end

	   (* $B$D$$$G$K!$%i%U$K=EJ#$r=|5n(B *)
	    local fun removesub [] ans = (rev ans)
		    | removesub (x::xs) [] = removesub xs [x]
		    | removesub ((ls,s)::xs) ((lt,t)::ys) = 
		      if ls = lt (* $B:8JU$H1&JU%5%$%:$,F1$8$J$i(B *)
			 andalso (Subst.identicalModuloRenamingRule s t)
		      then removesub xs ((lt,t)::ys)
		      else removesub xs ((ls,s)::(lt,t)::ys)
	    in
	    fun removeDuplication xs =  removesub xs []
	    end

            (*  $B9`%5%$%:(B($B:8JU$H1&JU$N(Bmax)$B$N>.$5$$=g$K$_$F!$:G=i$K=g=x$KIU$/$NEy<0$H$=$l0J30$KJ,3d$9$k(B *)
            fun select isTerminating [] done = (print "No orientable equations\n"; raise KbtcError)
              | select isTerminating ((l,r)::es) done = 
		if isTerminating ((l,r)::cs) then ((l,r), (rev done) @ es )
		else if isTerminating ((r,l)::cs) then  ((r,l), (rev done) @ es)
		else select isTerminating es ((l,r)::done)
                     
            val sorted = List.map (#2) (removeDuplication (makeVarNumlist es))

            val ((a,b),rest) = select isTerminating sorted []
	in
            ((a,b)::cs, (a,b), rs , rest)
	end                   


   (* rs $B$N1&JU$r(B r::rs $B$rMQ$$$F@55,2=$9$k(B *)
   fun composition (cs, r, rs, es) =
       let
           fun comp (a,b) = (a, Rewrite.linf (r::rs) b)
       in
           (cs, r, List.map comp rs, es)
       end


   (* $B=q$-49$(5,B'(B r $B$H(B $B=q$-49$(5,B'=89g(B r::rs $B$N4V$N4m81BP$rEy<0$KDI2C$9$k(B *)
   fun deduction (cs, r, rs, es) = 
       (cs, r, rs, List.@(Cr.criticalPairsKB r rs, es)) 

   (* rs $B$+$i!$:8JU$,(B r $B$G=q$-49$(2DG=$J$b$N$r<h$j=|$/(B *)
   fun collapse (cs, r, rs, es) = 
       let
           fun collaplist (l,_) rs =
               let
                   fun subcol l (Term.Var _) = false
                     | subcol l (m as Term.Fun( _ , ts , _) ) =
                       case Subst.match l m of
                           SOME _ => true
			 | NONE   => subcollist l ts
                   and subcollist _    []   = false
                     | subcollist l (t::ts) = (subcol l t) orelse (subcollist l ts) 
               in
                   List.filter (fn (m,_) => not (subcol l m)) rs
               end
       in
           (cs, r, collaplist r rs, es)
       end

   (* $BEy<0=89g$NN>JU$r@55,2=$9$k(B *)
   fun simplification (cs, rs, es) = 
       let
           fun simpl (l,r) = (Rewrite.linf rs l, Rewrite.linf rs r)
       in
           (cs, rs, List.map simpl es)
       end

   (* $BEy<0=89g$+$i!$N>JU$,F1$8Ey<0$r:o=|$9$k(B *)
   fun deletion (cs, rs, es) = (cs, rs, List.filter (not o Term.equal) es)


   (* $B40Hw2=$N5"G<%9%F%C%W(B *)
   fun kbstep isTerminating (cs, rs, es) =
       (deletion
	o simplification
	o (fn (cs, r, rs, es) => (cs, r::rs, es))
	o collapse
	o deduction
	o composition 
	o orientation isTerminating) (cs, rs, es)

   (* $B40Hw2=%a%$%s%k!<%A%s(B *)
   fun kbtc isTerminating es =
       let 
	   val _ = print "Start Knuth-Bendix Completion Procedure with a Termination Checker...\n"
	   val _ = print "Initial set of equations:\n"
	   val _ = print (Trs.prEqs es)
	   fun kbsub n isTerminating (cs, rs, es)  = 
	       let val _ = print ("Step " ^ (Int.toString n) ^ "\n")
		   val (cs', rs', es') = kbstep isTerminating (cs, rs, es)
		   val _ = print (Trs.prRules rs')
		   val _ = print (Trs.prEqs es')
	       in
		   if null es'
		   then (print ("Succeed in " ^ (Int.toString n) ^ " steps \n");
			 print (Trs.prRules rs'); 
			 true)
		   else  kbsub (n+1) isTerminating (cs',rs',es')
	       end
       in
	   kbsub 1 isTerminating ([], [], es)
	   handle KbtcError => (print "Failed:\n"; false)
       end


   (****************************)
   (********* $BJB9T40Hw2=(B *********)
   (****************************)

   (*  RuleSet$B$+$i(B SN({l->r} \cup cs) $B$^$?$O(B  SN({r->l} \cup cs) 
       $B$H$J$k%5%$%:$,:G$b>.$5$$5,B'$rA*$S!$(B
       $B2DG=$JAH$N%j%9%H(B $B"<(B  [ ((l,r) \cup cs, l->r, rs, $B;D$j$N(Bes),
                          ((r,l) \cup cs, l->r, rs, $B;D$j$N(Bes) ]
       $B$rJV$9(B  *)
   fun parallelOrientation isTerminating (cs, rs, []) = []
    |  parallelOrientation isTerminating (cs, rs, es) =
	let
            (* $B9`%5%$%:(B($BD9(B,$BC;(B)$B$HBP1~$9$k9`$r4^$s$@%j%9%H$r%=!<%H$7JV$9(B *)
            fun makeVarNumlist rs =
		let 
                    fun withSize (l,r) = let val llen = Term.termSize l
					     val rlen = Term.termSize r
					 in if rlen > llen
					    then ((rlen,llen),(r,l))
					    else ((llen,rlen),(l,r))
					 end
		in
                    ListMergeSort.sort 
			(fn (((n1,m1),_),((n2,m2),_)) => n1 > n2 orelse 
							 (n1 = n2 andalso m1 > m2))
			(List.map (fn lr => withSize lr) rs)
		end

	   (* $B$D$$$G$K!$%i%U$K=EJ#$r=|5n(B *)
	    local fun removesub [] ans = (rev ans)
		    | removesub (x::xs) [] = removesub xs [x]
		    | removesub ((ls,s)::xs) ((lt,t)::ys) = 
		      if ls = lt (* $B:8JU$H1&JU%5%$%:$,F1$8$J$i(B *)
			 andalso (Subst.identicalModuloRenamingRule s t)
		      then removesub xs ((lt,t)::ys)
		      else removesub xs ((ls,s)::(lt,t)::ys)
	    in
	    fun removeDuplication xs =  removesub xs []
	    end

            (*  $B9`%5%$%:(B($B:8JU$H1&JU$N(Bmax)$B$N>.$5$$=g$K$_$F!$(B
                $B=q$-49$(5,B'$K0\$;$k:G=i$NEy<0(B l=r $B$rA*$S!$(B
                $B2DG=$J(B process $B$N%j%9%H(B $B"<(B  { l->r$B$N>l9g!$(Br->l$B$N>l9g(B } $B$rA4$FJV$9(B *)

            fun select isTerminating [] done = []
              | select isTerminating ((l,r)::es) done = 
		case (isTerminating ((l,r)::cs),
		      isTerminating ((r,l)::cs)) of
		    (true,true) => [((l,r)::cs, (l,r), rs, (rev done) @ es),  
				    ((r,l)::cs, (r,l), rs, (rev done) @ es)]
		  | (true,false) => [((l,r)::cs, (l,r), rs, (rev done) @ es)]
		  | (false,true) => [((r,l)::cs, (r,l), rs, (rev done) @ es)]
		  | (false,false) => select isTerminating es ((l,r)::done)
                     
            val sorted = List.map (#2) (removeDuplication (makeVarNumlist es))
	in
            select isTerminating sorted []
	end                   


   (* $BJB9T40Hw2=$N5"G<%9%F%C%W(B *)
   fun pkbstep isTerminating ps =
       let 
	   val _ = print ("Num. of Procs: " ^ (Int.toString (length ps)) ^ "\n")

         (* $B@)Ls%5%$%:$NAmOB$,>.$5$$$b$N$rA*$V(B *)
	   fun select [] = (print "No process to select\n"; raise KbtcError)
	     | select (x::xs) = 
	       let
		   fun getSize (_,(cs,_,es)) = 
		       List.foldl
			   (fn ((l,r),m) =>  m + (Term.termSize l) + (Term.termSize r))
			   0
			   (es @ cs)
		   fun selectsub (n,x) [] rest = (x,rest)
		     | selectsub (n,x) (y::ys) rest = 
		       let val m = getSize y
		       in if n <= m 
			  then selectsub (n,x) ys (y::rest)
			  else selectsub (m,y) ys (x::rest)
		       end
	       in
		   selectsub (getSize x, x) xs []
	       end

	   val ((id,(cs,rs,es)),rest) = select ps

	   val _ = print ("Selected Process: "
			  ^ "ID " ^ (if id = "" then "e" else id) ^ "\n")

	   (* $B40Hw2=$NE,MQ(B *)
	   fun applyKb (cs,rs,es) = 
	       List.map 
		   (deletion 
		    o simplification
		    o (fn (cs, r, rs, es) => (cs, r::rs, es))
		    o collapse
		    o deduction
		    o composition)
		   (parallelOrientation isTerminating (cs,rs,es))

	   (* $B?7$7$$(BID $B$O(B 0 ^ $B5l(BID, 1 ^ $B5l(BID, ... $B$H$9$k(B *)
	   fun makeNodes ps = 
	       let val names = List.tabulate (length ps, fn x => (Int.toString x) ^ id)
	       in
		   ListPair.zip (names, ps)
	       end

       in
	   (makeNodes (applyKb (cs,rs,es)), rest)
       end

   (* $BJB9T40Hw2=$N%a%$%s%k!<%A%s(B *)
   fun pkbtc isTerminating es =
       let 
	   val _ = print "Start Parallel Knuth-Bendix Completion Procedure with a Termination Checker...\n"
	   val _ = print "Initial set of equations:\n"
	   val _ = print (Trs.prEqs es)
	   fun pkbsub n isTerminating ps  = 
	       let val _ = print ("STEP " ^ (Int.toString n) ^ "\n")
		   val (obtained,rest)  =  pkbstep isTerminating ps
		   val _ = print ("Num. of obtained Procs: " 
				  ^ (Int.toString (length obtained)) ^ "\n")
		   val _ = List.app (fn (id,(_,rs,es)) => 
					(print ("ID " ^ id ^ " ("
							^ "Step " ^ (Int.toString ((String.size id) - 1))
							^ ")\n");
					 print (Trs.prRules rs); 
					 print (Trs.prEqs es)))
				    obtained
	       in 
		   case List.find (fn (_,(_,_,es)) => null es) obtained of
		       SOME (id',(_,rs',_)) => (print ("Succeed in " ^ (Int.toString n) ^ " steps \n");
						print ("ID " ^ id' ^ " ("
						       ^ "Step " ^ (Int.toString ((String.size id') - 1))
						       ^ ")\n");
						print (Trs.prRules rs'); true)
		     | NONE => pkbsub (n+1) isTerminating (obtained @ rest)
	       end
       in
	   pkbsub 1 isTerminating [("e",([],[],es))]
	   handle KbtcError => (print "Failed\n"; false)
       end

   end (* of local *)

   end (* of struct *)




