(* Unfailing Knuth-Bendix Completion Prover *)
(* 2009.Feb.03 Pham Tuong Van *)
signature SigMkb = 
sig
	exception ERROR

	val mkbstep : (atom * (term * term -> bool)) list
		-> ((term * term) * (set * set * set)) list * ((term * term) * (set * set * set)) list
		-> ((term * term) * (set * set * set)) list	* ((term * term) * (set * set * set)) list
	val mkbsub  : int
		-> (atom * (term * term -> bool)) list
		-> ((term * term) * (set * set * set)) list	* ((term * term) * (set * set * set)) list
		-> bool
	(* Manual order, MKB, input: Order list, terms, where Order: Function lex order list *)
	val mmkb  : string list list -> (term * term) list -> bool	
	(* Auto all order MKB, input: funtion key list *)
	val amkb  : string list -> (term * term) list -> bool	
    val debug : bool   

end;

structure Mkb : SigMkb = struct
	
    open Term
	open Trs
	open Rewrite
	open Subst
	open ECr
	structure A = Atom
	structure AS = AtomSet
	structure L = List
	fun mapAppend f xs = List.foldr (fn (x,ys) => List.@(f x, ys)) [] xs
		
    exception ERROR;	
	
	(* Input: Function key list ex. ["D","Plus","+"] Output: [(1,Order1),(2,Order2),... *)
	fun generateOrders FList =
		let
			fun n! 0 = 1 | n! x = x * n! (x-1);
			(* choose m th Atom *)
			fun AChoose _ _ [] = NONE |
				AChoose 0 xs (y::ys) = SOME (y,xs@ys) |
				AChoose m xs (y::ys) = AChoose (m-1) (xs@[y]) ys
			(* Convert order number to natural num order list *)
			fun convertToNList 0 _ = [0] | 
				convertToNList m a = (a div (n! m))::(convertToNList (m-1) (a mod (n! m))); 
			(* Convert a nartural num order list to Atom order list *)
			fun convertToAList [] _ = [] |
				convertToAList (x::xs) ls =
				case AChoose x [] ls of
					NONE => convertToAList xs ls
				  | SOME (at,nls) => at::(convertToAList xs nls)		
			val len = L.length FList - 1 
			val AList = L.map (fn x=> A.atom x) FList 
			fun convertToOrder num = 
				Order.qlposCheck (Order.comparePrec (convertToAList (convertToNList len num) AList)) 
				(fn _ => (fn xs => xs)) Order.GT
			val orderNum = n! (len+1) 
			fun generate 0 = [] |
				generate n = (A.atom (Int.toString n),(convertToOrder (n-1)))::(generate (n-1));
		in rev (generate orderNum)
		end;
	(* Input: Order function key list ex. [["D","Plus","+"],[...],...] Output: [(1,Order1),(2,Order2),... *)
	fun manualOrder OList =
		let 
			fun toOrder ls = 
				Order.qlposCheck (Order.comparePrec (L.map (fn x => A.atom x) ls)) 
				(fn _ => (fn xs => xs)) Order.GT
			fun output _ [] = [] |
				output n (od::ods) = (A.atom (Int.toString n),(toOrder od))::(output (n+1) ods)
		in output 1 OList
		end;	

	(* -------- Base on mkbtc.sml by Aoto ---------*)
	fun switchNode ((l,r),(L1,L2,L3)) = ((r,l),(L2,L1,L3))
		
	fun r_Proj id ps = mapAppend (fn ((l,r),(L1,L2,_)) => 
								  (if AS.member (L1,id) then [(l,r)] else [])
									  @ (if AS.member (L2,id) then [(r,l)] else [])) ps
				   
	fun e_Proj id ps = L.mapPartial (fn (e,(_,_,L3)) => 
									   if AS.member (L3,id) then SOME e else NONE) 
		ps
		
	fun mapAppend f xs = List.foldr (fn (x,ys) => List.@(f x, ys)) [] xs
		
	fun isSucceeded maxId (oNodes,cNodes) = 
		let 
			fun generateId 0 = []
			  | generateId n = (Atom.atom (Int.toString n))::(generateId (n-1))

			val idls = generateId maxId
		in
			L.find (fn id => null (e_Proj id oNodes)
					andalso null (e_Proj id cNodes)
					andalso null (r_Proj id oNodes))
			idls
		end

	fun printNode ((s,t),(R1,R2,E)) =
		let val _ = print ("[<" ^ (Term.toString s) ^ " :+:  "
						   ^ (Term.toString t) ^ ">, ")
			val _ = print ("L1:" ^ (PrintUtil.prSetInOneLine
									A.toString (AS.listItems R1)) ^ ", ")
			val _ = print ("L2:" ^ (PrintUtil.prSetInOneLine
									A.toString (AS.listItems R2)) ^ ", ")
			val _ = print ("L3:" ^ (PrintUtil.prSetInOneLine
								   A.toString (AS.listItems E)) ^ ", ")
		in ()
		end

	(* ------ End of Modification ------- *)

	(* Choose from open Node a Node with smallest size of s:t*)
	fun chooseNode [] = (print "No equation to expand\n";  raise ERROR)
	  | chooseNode (onod::oNodes) =
		let
			fun nodeSize ((l,r),(_,_,_)) = termSize l + termSize r;
			
			fun findLeast rest anod min [] = (anod,rest)
			  | findLeast rest anod min (nod::nods) = 
				let 
					val thisSize = nodeSize nod
				in 
					if thisSize < min then findLeast (anod::rest) nod thisSize nods
					else findLeast (nod::rest) anod min nods
				end
		in 
			findLeast [] onod (nodeSize onod) oNodes
		end   

	(* Delete *)
	fun delete nodes = L.filter (fn ((l,r),(_,_,_)) => not (Term.equal (l,r))) nodes

	(* Garbage Collect *)
	fun gc nodes = L.filter 
		(fn ((l,r),(L1,L2,L3)) =>
		 not (AS.isEmpty L1)
		 orelse not (AS.isEmpty L2)
		 orelse not (AS.isEmpty L3))
		nodes
	
	(* -------- Rewrite -------*)
	local 
		(* tryRewrite (by rule) (in term) *)
		fun tryRewrite ((l,r),(L,_,_)) ((s,t),(L1,L2,L3)) =
		 if not (AS.isEmpty L)
		 then (* rewrite1 *)
		   if Subst.isVariant t l
		   then case Rewrite.rootRewrite [(l,r)] t of
		       SOME u => let val LcapL1 = AS.intersection (L,L1)
							 val LcapL3 = AS.intersection (L,L3)
						 in if not (AS.isEmpty LcapL1)
							 orelse not (AS.isEmpty LcapL3)
							then SOME (((s,t),(AS.difference (L1,LcapL1),L2,
											   AS.difference (L3,LcapL3))),
									   ((s,u),(LcapL1,AS.empty,LcapL3)))
							else NONE 
						 end
		     | NONE => NONE
		   else (* rewrite2 *)
		       case Rewrite.linfPartial [(l,r)] t of
				   SOME u => let val LcapL1 = AS.intersection (L,L1)
								 val LcapL2 = AS.intersection (L,L2)
								 val LcapL3 = AS.intersection (L,L3)
							 in if not (AS.isEmpty LcapL1)
								 orelse not (AS.isEmpty LcapL2)
								 orelse not (AS.isEmpty LcapL3)
								then SOME (((s,t),(AS.difference (L1,LcapL1),
												   AS.difference (L2,LcapL2),
												   AS.difference (L3,LcapL3))),
										   ((s,u),(LcapL1,AS.empty,
												   AS.union (LcapL2,LcapL3))))
								else NONE
							 end
				 | NONE => NONE
		 else NONE

		(* rewriteNode (by RULES) (in TERM) *)
		fun rewriteNode [] _ = NONE
		  | rewriteNode (((l,r),(L,L',E))::ns) ((s,t),(L1,L2,L3)) = 
			case tryRewrite ((l,r),(L,L',E)) ((s,t),(L1,L2,L3)) of
				SOME (n,m) => SOME (n,m)
			  | NONE => case tryRewrite ((r,l),(L',L,E)) ((s,t),(L1,L2,L3)) of
					SOME (n,m) => SOME (n,m)
				  | NONE => case tryRewrite ((l,r),(L,L',E)) ((t,s),(L2,L1,L3)) of
						SOME (n,m) => SOME (switchNode n, switchNode m)
					  | NONE => case tryRewrite ((r,l),(L',L,E)) ((s,t),(L2,L1,L3)) of
						    SOME (n,m) => SOME (switchNode n, switchNode m)
						  | NONE => rewriteNode ns ((s,t),(L1,L2,L3))

		(* divideNode byRULES inTERMS : rewrite terms by rules and divide to 
		 Nothing happen, Updated, Created *)
		fun divideNodes rls ts =
			let
				fun merge (nod,(occur,os,us,cs)) = case rewriteNode rls nod of
					NONE => (occur,nod::os,us,cs)
				  | SOME (n,m) => (true,os,n::us,m::cs)
			in
				L.foldr merge (false,[],[],[]) ts
			end

		(* rewriteStep by_RULES in_TERMS Updated Created -> (Updated, Created Node) *)
		fun rewriteStep origin rls = 
			let
				val (occur1,ORG,u1,c1) = divideNodes rls origin
				fun findUpdate nds uds crs = 
					case divideNodes rls nds of
						(false,_,_,_) => (nds@uds,crs)
					  | (true,a,b,c) => findUpdate b (a@uds) (c@crs)
				val (UPD,c2) = findUpdate u1 [] []
				fun findCreate nds crs = case divideNodes rls nds of
					(false,_,_,_) => nds@crs
				  | (true,a,b,c) => findCreate (b@c) (a@crs)
				val CRT = findCreate (c1@c2) []
			in (ORG,UPD,CRT)
			end
	in
		   fun rewrite (terms,rules) =  rewriteStep terms rules

	end

	(* - End of Rewrite ([Nc,{n}]) - *)   

	(* Orient *)
	fun orient order ((lt,rt),(aL1,aL2,aL3)) =
		let
			fun findOrder key =
				case L.find (fn (id,_) => A.sameAtom(key,id)) order of
					SOME (_,odr) => odr
				  | NONE => (fn _ =>false)
			
			fun tryOrient (L1,L2,L3) [] = (L1,L2,L3)
			  | tryOrient (L1,L2,L3) (l::ls) = 
				let
					val thisOrder = findOrder l
				in
					if thisOrder (lt,rt) 
						then tryOrient (AS.add(L1,l),L2,L3) ls
					else if thisOrder (rt,lt) 
							 then tryOrient (L1,AS.add(L2,l),L3) ls
						 else tryOrient (L1,L2,AS.add(L3,l)) ls
				end
		in ((lt,rt),tryOrient (aL1,aL2,AS.empty) (AS.listItems aL3))
		end

	(* Deduction *)

	fun deduce n0 nodes =
		let 
			fun addcpstep0 ((l,r),(L1,L2,L3)) =
				if AS.isEmpty L1 then [] else
				List.map
				(fn (u,v) => ((u,v),(AS.empty,AS.empty,L1)))
				(Cr.insideCriticalPairs [(l,r)])
			
			fun addcpstep ((l,r),(R1,R2,E)) ((x,y),(X1,X2,E')) = 
				let 
					val R1capX1 = AS.intersection (R1,X1)
				in if not (AS.isEmpty R1capX1)
					   then
						   List.map
						   (fn (u,v) => ((u,v),(AS.empty,AS.empty,R1capX1)))
						   (L.@(Cr.criticalPairs2 ([(l,r)],[(x,y)]),
								Cr.insideCriticalPairs2 ([(x,y)],[(l,r)])))
				   else []
				end

			fun addcp0 ((l,r),(R1,R2,E)) =
				(addcpstep0 ((l,r),(R1,R2,E)))
				@ (addcpstep0 ((r,l),(R2,R1,E)))
				
			fun addcp ((l,r),(R1,R2,E)) ((x,y),(X1,X2,E')) =
				(addcpstep ((l,r),(R1,R2,E)) ((x,y),(X1,X2,E')))
				@ (addcpstep ((l,r),(R1,R2,E)) ((y,x),(X2,X1,E')))
				@ (addcpstep ((r,l),(R2,R1,E)) ((x,y),(X1,X2,E')))
				@ (addcpstep ((r,l),(R2,R1,E)) ((y,x),(X2,X1,E')))
		in
			(addcp0 n0) @ (mapAppend (fn n1 => addcp n0 n1) nodes)
		end

	(* -----End of Deduction -----*)

	(* Union and subsume *)
	fun addNodes ([], stable) = stable
	  | addNodes (n0::ns, stable) =
		let 
			fun insert node [] ans = node::ans
			  | insert ((x,y),(X1,X2,Y)) (((l,r),(R1,R2,E))::ns) ans =
				if Subst.identicalModuloRenamingRule (x,y) (l,r)
				then let val R1' = AS.union (X1,R1)
						 val R2' = AS.union (X2,R2)
						 val E'  = AS.difference 
							 (AS.union (Y,E),
							  AS.union (R1',R2')) 			

					 in ((l,r),(R1',R2',E'))::(ns @ ans)
					 end
				else 
					if Subst.identicalModuloRenamingRule (x,y) (r,l)
					then let 
							 val R1' = AS.union (X1,R2)
							 val R2' = AS.union (X2,R1)
							 val  E'  = AS.difference 
								 (AS.union (Y,E),
								  AS.union (R1',R2')) 			
								 
						 in ((r,l),(R1',R2',E'))::(ns @ ans)
						 end
					else
						insert ((x,y),(X1,X2,Y)) ns (((l,r),(R1,R2,E))::ans)

		in
			addNodes (ns, insert n0 stable [])
		end

	(* --- End of Union and subsume --- *)
	fun mkbstep Orders (oNodes,cNodes) = 
		let 
			val (chosen, restoNodes) = chooseNode oNodes
			   
			val _ = (print ("Selected Node:\n"); printNode chosen; print "\n")
				
			val (t1,t2,t3) = rewrite ([chosen], cNodes)
			val chosen1 = if null t2 then chosen else (fn (a::b) => a) t2

			val oNodes1 = addNodes(delete(t3), restoNodes)
			
			val an = orient Orders chosen1

			val (oNodes2,cNodes2) = (fn (a,b,c) => (c,a@b)) (rewrite (cNodes, [an]))
			val oNodes3 = addNodes(oNodes1, delete(oNodes2))
			val cNodes3 = rev(gc(cNodes2))

			val oNodes4 = addNodes(oNodes3, delete(deduce an cNodes3))
			val (t4,t5,t6) = rewrite (oNodes4, cNodes3)
			val oNodes5 = t4@t5
			val oNodes6 = addNodes(t6, oNodes5)
			val oNodes7 = rev (gc(delete oNodes6))

			val cNodes4 = addNodes([an],cNodes3)

		in 
			(oNodes7,cNodes4)
		end

	fun mkbsub n Orders (oNodes,cNodes) =
		let 
			val _ = print ("Step: " ^ (Int.toString n) ^ "\n")
			val _ = (Int.toString (length (oNodes))) ^ " open Nodes | " ^
				    (Int.toString (length (cNodes))) ^ " close Nodes \n"
			val (oNodes', cNodes') = mkbstep Orders (oNodes,cNodes)
			val Odlen = L.length Orders
		in
			case isSucceeded Odlen (oNodes',cNodes') of
				SOME id => 
					let
						val rs' = r_Proj id cNodes'
					in 
						(print ("\n Succeed in " ^ (Int.toString n) ^ " steps \n");
						 print (Trs.prRules rs'); true)
					end
			  | NONE => mkbsub (n+1) Orders (oNodes',cNodes')
		end
	
	(* Manual order mbk *)
	fun mmkb orderStrings es =
		let
 			val orders = manualOrder orderStrings
				
			val _ = "Manual order MKB Completion ...\n"
			
			fun generateInitialId 0 IID = IID 
			  | generateInitialId n IID = 
				generateInitialId (n-1) (AS.add(IID, A.atom (Int.toString n)))
			
			val initialID = generateInitialId (L.length orders) AS.empty
			
			val initialNodes = L.map (fn e => (e,(AS.empty,AS.empty,initialID))) es
		in mkbsub 1 orders (initialNodes,[])
		
		handle ERROR => (print "Failed\n"; false) 
		
		end

	fun amkb functionString es =
		let
 			val orders = generateOrders functionString
				
			val _ = "Automatic all order MKB Completion ...\n"
			
			fun generateInitialId 0 IID = IID 
			  | generateInitialId n IID = 
				generateInitialId (n-1) (AS.add(IID, A.atom (Int.toString n)))
			
			val initialID = generateInitialId (L.length orders) AS.empty
			
			val initialNodes = L.map (fn e => (e,(AS.empty,AS.empty,initialID))) es
		in mkbsub 1 orders (initialNodes,[])
		
		handle ERROR => (print "Failed\n"; false) 
		
		end

	val debug =
		let
		in false
		end;

end;

