(******************************************************************************
 * file: rwtools/kbcomp/kbtc.sml
 * description: Multi Knuth-Bendix Completion Procedure With Termination Checker
 * author: AOTO Takahito
 * 
 * Copyright (c) 2008, Toyama&Aoto Laboratory, Tohoku University
 ******************************************************************************)

signature MKBTC = 
sig
    val runDebug: bool ref
    val minisatPath: string ref (* not used *)
    val tmpDir: string ref (* not used *)

    val mkbtc : ((Term.term * Term.term) list -> bool) 
		-> Trs.equations  
		-> bool
		    
end;


structure Mkbtc : MKBTC = 
   struct

   local 
       structure L = List
       structure A = Atom
       structure VS = VarSet
       structure AS = AtomSet
       fun mapAppend f xs = List.foldr (fn (x,ys) => List.@(f x, ys)) [] xs
   in

   exception MkbtcError;
   val runDebug = ref false
   fun debug f = if (!runDebug) then f () else ()
    val minisatPath = ref "" (* not used *)
    val tmpDir = ref ""  (* not used *)

   fun printNode ((s,t),(R1,R2,E,C1,C2)) =
       let val _ = print ("[<" ^ (Term.toString s) ^ " :+:  "
		   ^ (Term.toString t) ^ ">, ")
	   val _ = print ("R1:" ^ (PrintUtil.prSetInOneLine
		       A.toString (AS.listItems R1)) ^ ", ")
	   val _ = print ("R2:" ^ (PrintUtil.prSetInOneLine
		       A.toString (AS.listItems R2)) ^ ", ")
	   val _ = print ("E:" ^ (PrintUtil.prSetInOneLine
		       A.toString (AS.listItems E)) ^ ", ")
	   val _ = print ("C1:" ^ (PrintUtil.prSetInOneLine
		       A.toString (AS.listItems C1)) ^ ", ")
	   val _ = print ("C2:" ^ (PrintUtil.prSetInOneLine
		       A.toString (AS.listItems C2)) ^ "]\n")
       in ()
       end

   fun switchNode ((l,r),(R1,R2,E,C1,C2)) = ((r,l),(R2,R1,E,C2,C1))

   fun rPart id ps = mapAppend (fn ((l,r),(r1,r2,_,_,_)) => 
				   (if AS.member (r1,id) then [(l,r)] else [])
				   @ (if AS.member (r2,id) then [(r,l)] else []))
			       ps

   fun ePart id ps = List.mapPartial (fn (e,(_,_,es,_,_)) => 
				       if AS.member (es,id) then SOME e else NONE) 
				   ps

   fun cPart id ps = mapAppend (fn ((l,r),(_,_,_,c1,c2)) => 
				   (if AS.member (c1,id) then [(l,r)] else [])
				   @ (if AS.member (c2,id) then [(r,l)] else []))
			       ps


   (* Ƥ ID Υꥹ *)
   val idlist = ref ([]:(Atom.atom list))
   fun member x ys = List.exists (fn y => AS.equal (x,y)) ys
   fun delete x [] = []
     | delete x (y::ys) = if A.compare (x,y) = EQUAL then ys else y::(delete x ys)
   fun removeId id = idlist := delete id (!idlist)
   fun addId id = idlist := (id::(!idlist))
   fun prId id = Atom.toString id

   fun isSucceeded (oNodes,cNodes) = 
	   L.find (fn id => null (ePart id oNodes)
			    andalso null (ePart id cNodes)
			    andalso null (rPart id oNodes))
		  (!idlist)

   fun deleteIdsFromNode idSet (lr,(r1,r2,e,c1,c2)) = 
	   (lr, (AS.difference(r1,idSet),
			 AS.difference(r2,idSet),
			 AS.difference(e,idSet),
			 AS.difference(c1,idSet),
			 AS.difference(c2,idSet)))


   fun choose isTerminating (oNodes,cNodes) =
       let
           fun sizeOfEqs es = L.foldl 
				    (fn ((l,r),i) => (Term.termSize l) + (Term.termSize r) + i) 
(*				    (fn ((l,r),i) => Int.max (Term.termSize l, Term.termSize r) + i) *)
				    0 
				    es

           fun sizeOfRules rs = L.foldl 
(* 				    (fn ((l,r),i) => (Term.termSize l)*(Term.termSize l) *)
(* 						     + (Term.termSize r)*(Term.termSize r) + i) *)
				    (fn ((l,r),i) => (Term.termSize l)*(Term.termSize l)
						     + (Term.termSize r)*(Term.termSize r) + i)
				    0 
				    (L.filter (fn (x,y) =>  not (Term.equal (Rewrite.linf rs x,
									     Rewrite.linf rs y)))
					      (Cr.criticalPairs rs))

	   (* id Υ E_id Υ + R_id Υ *)
           fun idWithSize id = (id, sizeOfEqs (ePart id oNodes) 
				    + sizeOfEqs (ePart id cNodes) 
 				   (*  + sizeOfRules (rPart id cNodes) *)
 				   (*  + sizeOfRules (rPart id oNodes) *)
 				    + sizeOfEqs (cPart id cNodes)
 				    + sizeOfEqs (cPart id oNodes))

(*           fun idWithSize id = (id, sizeOfEqs (ePart id oNodes)) *)

           val idlistWithSize = L.map (fn id => idWithSize id) (!idlist)


           val _ = debug (fn _ => print "Proc score:")
           val _ = debug (fn _ => 
			     L.app (fn (id,plen) => print ((prId id) ^ "(" ^ (Int.toString plen) ^ ") ")) 
				   idlistWithSize)
           val _ = debug (fn _ => print "\n")


 	   fun sizeOfId id = 
	       case L.find (fn (id2,i) => A.compare (id,id2) = EQUAL) idlistWithSize of
		   SOME (_,i) => i
		 | NONE => 0

	   (* id ν E Υϡޤޤ id Υ min *)
	   fun sizeOfSet X = 
	       case AS.find (fn _=>true) X of
		   SOME id0 => AS.foldl (fn (id,i)=> Int.min (sizeOfId id, i)) 
					(sizeOfId id0) X
		 | NONE => 0 (* E = \emptyset *)

           fun makeVarNumlist ps =
		let
                    fun withSize ((l,r),(R1,R2,E,C1,C2))=
			let val llen = Term.termSize l 
			    val rlen = Term.termSize r 
			    val maxlen = Int.max (llen, rlen)
			    val sumlen = llen + rlen
			    val ldep = Term.termDepth l
			    val rdep = Term.termDepth r
			    val plen = sizeOfSet E
			    val elen = AS.numItems E
			in if rlen > llen
(* Reasonable but not Good  *)
			   then ((plen,sumlen,sumlen,elen),(r,l),(R2,R1,E,C2,C1))
			   else ((plen,sumlen,sumlen,elen),(l,r),(R1,R2,E,C1,C2))
(* Odd but Good : cge2=> 50, endo=>44, groups=>31, equiv_proofs=>36 *)
(* 			   then ((plen,llen,ldep+rdep),(r,l),(R2,R1,E,C2,C1)) *)
(* 			   else ((plen,ldep,llen),(l,r),(R1,R2,E,C1,C2)) *)
(* test: cge2=>46, endo=>38, groups=>31, equiv_proofs=>40 *)
(*			   then ((plen,llen,rdep),(r,l),(R2,R1,E,C2,C1)) *)
(*			   else ((plen,ldep,rlen),(l,r),(R1,R2,E,C1,C2)) *)
(* test : cge2=>43, endo=>43, groups=>31, equiv_proofs=>40*)
(*  			   then ((plen,llen,ldep+rdep),(r,l),(R2,R1,E,C2,C1)) *)
(*  			   else ((plen,ldep,rlen+llen),(l,r),(R1,R2,E,C1,C2)) *)
(* test : cge2=>45, endo=>38, groups=>31, equiv_proofs=>40 *)
(*  			   then ((plen,llen,ldep+rdep),(r,l),(R2,R1,E,C2,C1))  *)
(*  			   else ((plen,ldep,rlen),(l,r),(R1,R2,E,C1,C2)) *)
(* test: cge2=>42, endo=>39,groups=>22,equiv_proofs=>40  *)
(*  			   then ((plen,llen,ldep+rdep),(r,l),(R2,R1,E,C2,C1)) *)
(*  			   else if (length ps) mod 3 = 0 *)
(* 			   then ((plen,ldep,llen),(l,r),(R1,R2,E,C1,C2)) *)
(* 			   else ((plen,ldep,rlen),(l,r),(R1,R2,E,C1,C2)) *)
(* Test: cge2=>45, cge3=>68, endo=>38,groups=>31,equiv_proofs=>38  *)
(*  			   then ((plen,llen,ldep+rdep),(r,l),(R2,R1,E,C2,C1)) *)
(*  			   else if (length ps) mod 3 = 0 *)
(* 			   then ((plen,ldep,rdep),(l,r),(R1,R2,E,C1,C2)) *)
(* 			   else ((plen,ldep,rlen),(l,r),(R1,R2,E,C1,C2)) *)
(* test: cge2=>46, endo=>53,groups=>28,equiv_proofs=>37  *)
(*  			   then ((plen,llen,ldep+rdep),(r,l),(R2,R1,E,C2,C1))  *)
(*  			   else if (length ps) mod 7 = 0 *)
(* 			   then ((plen,rdep,rlen),(l,r),(R1,R2,E,C1,C2)) *)
(* 			   else ((plen,ldep,rlen),(l,r),(R1,R2,E,C1,C2)) *)
(* test: cge2=>60, endo=>41,groups=>31,equiv_proofs=>40  *)
(*  			   then ((plen,llen,ldep+rdep),(r,l),(R2,R1,E,C2,C1))  *)
(*  			   else if (length ps) mod 3 = 0 *)
(* 			   then ((plen,rdep,llen),(l,r),(R1,R2,E,C1,C2)) *)
(* 			   else ((plen,ldep,rlen),(l,r),(R1,R2,E,C1,C2)) *)
			end
		in
                    ListMergeSort.sort
			(fn (((p1,n1,m1,e1),_,_),((p2,n2,m2,e2),_,_))
			    => p1 = 0 andalso p2 > 0
			       orelse
			       p1 > p2 andalso p2 > 0
			       orelse
			       (p1 = p2 andalso p2 > 0 andalso n1 > n2)
			       orelse
			       (p1 = p2 andalso p2 > 0 andalso n1 = n2 andalso m1 > m2)
			       orelse
			       (p1 = p2 andalso p2 > 0 andalso n1 = n2 andalso m1 = m2 andalso e1 < e2))
			(List.map (fn (lr,labels) => withSize (lr,labels)) ps)
		end

            fun select isTerminating [] _ = (print "No orientable equations\n"; raise MkbtcError)
              | select isTerminating ((p as ((l,r),(R1,R2,E,C1,C2)))::ps) rest = 
(* 		if  (\*  ҥ塼ꥹƥl -> r  r -> l ޤץȤϡ*\) *)
(*             (\*  l = r ŸϤ  *\) *)
(* 			not (AS.isEmpty R1) *)
(* 			(\*  orelse not (AS.isEmpty R2) ... needed for some examples *\) *)
(* 			andalso not (AS.isEmpty R2) *)
(* 		then select isTerminating ps (p::rest) *)
(* 		else  *)
		let 
		    val (E1,E2,D) = 
			AS.foldl (fn (id,(X,Y,Z)) => 
				     let val cs = cPart id cNodes 
					 val _ = debug (fn _ => print ("Check Termination: (ID " 
								       ^ (Atom.toString id) ^ ")\n"))
					 val _ = debug (fn _ =>print ("Cpart: \n" ^ (Trs.prRules cs)))
					 val _ = debug (fn _ => print ("(l: " 
								       ^ (Term.toString l) ^ ", r: "
								       ^ (Term.toString r) ^ ")\n"))
				     (*	val _ = debug (fn _ => L.app printNode (rest @ ps)) *)
				     in case (isTerminating ((l,r)::cs),
					      isTerminating ((r,l)::cs)) of
					    (true,true) => (debug (fn _ => print ": TT\n");
								(AS.add (X,id),AS.add (Y,id),Z))
					      | (true,false) => (debug (fn _ => print ": TF\n");
								 (AS.add (X,id), Y, Z))
					      | (false,true) => (debug (fn _=> print ": FT\n");
								 (X, AS.add (Y,id),Z))
					      | (false,false) => (debug (fn _=> print ": FF\n");
								  (X,Y,Z))
					 end)
				 (AS.empty,AS.empty,AS.empty)
				 E
				 
		    val size = sizeOfSet E
			       

		in  

(* 			if (\* ҥ塼ꥹƥE ǤʤȤϡ˾ΤE1E2˴ޤޤƤ褦ˤȤ *\) *)
(* 			    false andalso  (\* disable *\) *)
(* 				not (AS.isEmpty E) *)
(* 				andalso sizeOfSet E1 <> size  *)
(* 				andalso sizeOfSet E2 <> size *)
(* 			then select isTerminating ps (p::rest) *)
(* 			(\*** else (p, E1, E2, rest @ ps)  ***\) *)
(* 			else *)
(*   		      (\***  ҥ塼ꥹƥ򤷤 orient ˼ԤץϽƤޤ ***\) *)
(* 			    (idlist := L.filter (fn id => not (AS.member (!toBeDeleted,id))) (!idlist); *)
(* 			     (p,  *)
(* 			      AS.difference(E1,!toBeDeleted),  *)
(* 			      AS.difference(E2,!toBeDeleted),  *)
(* 			      L.map (deleteIdsFromNode (!toBeDeleted)) (rest @ ps),  *)
(* 			      L.map (deleteIdsFromNode (!toBeDeleted)) cNodes)) *)

		    if (AS.isEmpty E1) andalso (AS.isEmpty E2)
		    then select isTerminating ps (p::rest)
		    else (p, E1, E2, ps, rest)
		end
			
	    val (oNodesPos,oNodesNeg) = L.partition (fn (_,(_,_,E,_,_)) => AS.isEmpty E) oNodes

	    val sorted0 = makeVarNumlist oNodesNeg

            val sorted = List.map (fn p => (#2 p, #3 p)) sorted0

            val _ = debug (fn _ => print "Open nodes:(sorted)\n")
 	    val _ = debug (fn _ => (L.app (fn ((p1,n1,m1,e1),p,q) =>
					      (print ((Int.toString p1)
						     ^ "-" ^ (Int.toString n1)
						     ^ "-" ^ (Int.toString m1)
						     ^ "-" ^ (Int.toString e1));
					       (printNode (p,q))))
 					  sorted0))

       in
	   select isTerminating sorted (cNodes @ oNodesPos)
       end


   fun orient ((l,r),(R1,R2,E,C1,C2)) (E1,E2) (oNodes,cNodes) =
       let
	   val E1cupE2 = AS.union (E1,E2)
	   val _ = AS.app (fn id => (removeId id)) E1cupE2  (* delete from idList *)
	   val E' = AS.difference (E,E1cupE2)
	   val R1' = AS.map (fn id=> let val id0 = Atom.atom ("0" ^ (prId id))
					 val _ = addId id0
				     in id0
				     end)
			    E1
	   val R2' = AS.map (fn id=> let val id1 = Atom.atom ("1" ^ (prId id))
					 val _ = addId id1
				     in id1
				     end)
			    E2

	   fun replace X = AS.foldl
			       (fn (id,set) => if AS.member (E2,id)
					       then AS.add 
							(if AS.member (E1,id)
							 then AS.add (set,Atom.atom ("0" ^ (prId id)))
							 else set,
							 Atom.atom ("1" ^ (prId id)))
					       else if AS.member (E1,id)
					       then AS.add (set,Atom.atom ("0" ^ (prId id)))
					       else AS.add (set,id))
			       AS.empty
			       X

	   fun update nodes = List.map
			      (fn ((x,y),(X1,X2,Y,Z1,Z2))
				  => ((x,y), (replace X1,replace X2,replace Y,replace Z1,replace Z2)))
			      nodes

	   val cNodes' = update cNodes
	   val oNodes' = update oNodes
	   val p' = ((l,r),(AS.union (R1,R1'),AS.union (R2,R2'), E',AS.union (C1,R1'), AS.union (C2,R2')))
       in
	   (p', oNodes', cNodes')
       end


   fun deduce n0 nodes =
       let fun addcpstep0 ((l,r),(R1,R2,E,C1,C2)) =
	       List.map
		   (fn (u,v) => ((u,v),(AS.empty,AS.empty,R1,AS.empty,AS.empty)))
		   (Cr.insideCriticalPairs [(l,r)])

	   fun addcpstep ((l,r),(R1,R2,E,C1,C2)) ((x,y),(X1,X2,Y,Z1,Z2)) = 
	       let val R1capX1 = AS.intersection (R1,X1)
	       in if not (AS.isEmpty R1capX1)
		  then
		      List.map
			  (fn (u,v) => ((u,v),(AS.empty,AS.empty,R1capX1,AS.empty,AS.empty)))
			  (L.@(Cr.criticalPairs2 ([(l,r)],[(x,y)]),
			       Cr.insideCriticalPairs2 ([(x,y)],[(l,r)])))
		  else []
	       end

	   fun addcp0 ((l,r),(R1,R2,E,C1,C2)) =
	       (addcpstep0 ((l,r),(R1,R2,E,C1,C2)))
	       @ (addcpstep0 ((r,l),(R2,R1,E,C2,C1)))

	   fun addcp ((l,r),(R1,R2,E,C1,C2)) ((x,y),(X1,X2,Y,Z1,Z2)) =
	       (addcpstep ((l,r),(R1,R2,E,C1,C2)) ((x,y),(X1,X2,Y,Z1,Z2)))
	       @ (addcpstep ((l,r),(R1,R2,E,C1,C2)) ((y,x),(X2,X1,Y,Z2,Z1)))
	       @ (addcpstep ((r,l),(R2,R1,E,C2,C1)) ((x,y),(X1,X2,Y,Z1,Z2)))
	       @ (addcpstep ((r,l),(R2,R1,E,C2,C1)) ((y,x),(X2,X1,Y,Z2,Z1)))
       in
	   (addcp0 n0) @ (mapAppend (fn n1 => addcp n0 n1) nodes)
       end


       local fun tryRewrite ((l,r),(R1,R2,E,C1,C2)) ((x,y),(X1,X2,Y,Z1,Z2)) =
	       if not (AS.isEmpty R1)
		  andalso (not (AS.isEmpty X1) 
			   orelse not (AS.isEmpty X2) 
			   orelse not (AS.isEmpty Y))
	       then (* rewrite1 *)
		   if Subst.isVariant l y
		   then if (not (AS.isEmpty X1) orelse not (AS.isEmpty Y))
			then case Rewrite.rootRewrite [(l,r)] y of
				 SOME y' => let val R1capX1 = AS.intersection (R1,X1)
				      val R1capY = AS.intersection (R1,Y)
(*				      val _ = debug (fn _ =>print ("root rewrite " ^ (Term.toString y) 
						     ^ " => " ^ (Term.toString y') 
						     ^ " by rule " ^ (Term.toString l) 
						     ^ " -> " ^ (Term.toString r) 
						     ^ (PrintUtil.prSetInOneLine 
							    prId 
							    (AS.listItems (AS.union (R1capX1,R1capY))))
						     ^ "\n")) *)
				  in if not (AS.isEmpty R1capX1)
					orelse not (AS.isEmpty R1capY)
				     then SOME (((x,y),(AS.difference (X1,R1capX1),X2,
							AS.difference (Y,R1capY),Z1,Z2)),
(**	09/07/29			((x,y'),(R1capX1,AS.empty,R1capY,Z1,Z2)))  **)
						((x,y'),(R1capX1,AS.empty,R1capY,AS.empty,AS.empty)))  
				     else NONE 
				  end
			       | NONE => NONE
			else NONE
		   else (* rewrite2 *)
		       let val R1capX1 = AS.intersection (R1,X1)
			   val R1capX2 = AS.intersection (R1,X2)
			   val R1capY = AS.intersection (R1,Y)
		       in
			   if  not (AS.isEmpty R1capX1)
			       orelse not (AS.isEmpty R1capX2)
			       orelse not (AS.isEmpty R1capY)
			   then case Rewrite.linfPartial [(l,r)] y of
				    SOME y' => let 
(*					val _ = debug (fn _ => print ("linf rewrite " ^ (Term.toString y) 
						  ^ " => " ^ (Term.toString y') 
						  ^ " by rule " ^  (Term.toString l) 
						  ^ " -> " ^ (Term.toString r) 
						  ^ (PrintUtil.prSetInOneLine 
							 prId 
							 (AS.listItems (AS.union (R1capX1,AS.union (R1capX2,R1capY)))))
						  ^ "\n"))
*)
				    in 
					SOME (((x,y),(AS.difference (X1,R1capX1),
						      AS.difference (X2,R1capX2),
						      AS.difference (Y,R1capY),Z1,Z2)),
					      ((x,y'),(R1capX1,AS.empty,
						       AS.union (R1capX2,R1capY),
						       AS.empty,AS.empty)))
				    end
				  | NONE => NONE
			   else NONE
		       end
	       else NONE

	   (* rewriteNode rules-nodes term-node *)
           fun rewriteNode [] _ = NONE
	     | rewriteNode (((l,r),(R1,R2,E,C1,C2))::ns)
			   ((x,y),(X1,X2,Y,Z1,Z2)) = 
(* 	       (debug (fn _ =>  *)
(* 			  print ("term: " ^ (Term.toString x) ^ " and " ^ (Term.toString y) ^ "\n")); *)
(* 		debug (fn _=> print ("rule: " ^ (Term.toString l)  ^" -> " ^ (Term.toString r) ^ "\n")); *)

	       case tryRewrite ((l,r),(R1,R2,E,C1,C2)) 
			       ((x,y),(X1,X2,Y,Z1,Z2)) of
		   SOME (n,m) => SOME (n,m)
		 | NONE => case tryRewrite ((r,l),(R2,R1,E,C2,C1))
	 			           ((x,y),(X1,X2,Y,Z1,Z2)) of
                               SOME (n,m) => SOME (n,m)
			     | NONE => case tryRewrite ((l,r),(R1,R2,E,C1,C2))
						       ((y,x),(X2,X1,Y,Z2,Z1)) of
					   SOME (n,m) => SOME (switchNode n, switchNode m)
					 | NONE => case tryRewrite ((r,l),(R2,R1,E,C2,C1))
								   ((y,x),(X2,X1,Y,Z2,Z1)) of
						       SOME (n,m) => SOME (switchNode n, switchNode m)
						     | NONE => rewriteNode ns
									    ((x,y),(X1,X2,Y,Z1,Z2)) (* ) *)
      
	 (* initial call: rewriteStep terms rules ([],[]) *)
	   fun rewriteStep [] ys (new,stable) = 
	       if null new then (rev stable) else rewriteStep (rev new) ys ([], stable)
	     | rewriteStep (x::xs) ys (new,stable) =
	       case rewriteNode ys x of 
		       SOME (n,m) => rewriteStep xs ys (n::m::new, stable) 
		     | NONE => rewriteStep xs ys (new,x::stable)

	   fun init xs ys = 
	       L.foldl (fn (x,(os,ns,ms)) => 
					case rewriteNode ys x of 
						SOME (n,m) => (os,n::ns,m::ms)
					  | NONE => (x::os,ns,ms))
		   ([],[],[])
		   xs

	   fun properRewriteStep (stablec,[],[]) rs (new1,new2,stableo) = 
	       if (null new1) andalso (null new2)
	       then (rev stablec,stableo) (*  stay closed  and go open *)
	       else properRewriteStep (stablec,new1,new2) rs ([],[],stableo)
	     | properRewriteStep (stablec,x::ns,ms) rs (new1,new2,stableo) = 
	       (case rewriteNode rs x of 
		       SOME (n,m) => properRewriteStep (stablec,ns,ms) rs (n::new1,m::new2,stableo)
		     | NONE => properRewriteStep (x::stablec,ns,ms) rs (new1,new2,stableo))
	     | properRewriteStep (stablec,[],y::ms) rs (new1,new2,stableo) = 
	       (case rewriteNode rs y of 
		       SOME (n,m) => properRewriteStep (stablec,[],ms) rs (new1,n::m::new2,stableo)
		     | NONE => properRewriteStep (stablec,[],ms) rs (new1,new2,y::stableo))

       in
       fun rewrite (terms,rules) =  rewriteStep terms rules ([],[])
       fun properRewrite (terms,rules) =  
			properRewriteStep (init terms rules) rules ([],[],[])
       end


   fun delete nodes = L.filter
		       (fn ((l,r),(R1,R2,E,C1,C2)) => not (Term.equal (l,r)))
		       nodes
		       
   fun gc nodes = L.filter
		       (fn ((l,r),(R1,R2,E,C1,C2)) => 
			   not (AS.isEmpty R1)
			   orelse not (AS.isEmpty R2)
			   orelse not (AS.isEmpty E)
			   orelse not (AS.isEmpty C1)
			   orelse not (AS.isEmpty C2))
		       nodes

   fun add ([], stable) = stable
     | add (n0::ns, stable) =
       let fun insert node [] ans = rev (node::ans)
	     | insert ((x,y),(X1,X2,Y,Z1,Z2)) (((l,r),(R1,R2,E,C1,C2))::ns) ans =
	       if Subst.identicalModuloRenamingRule (x,y) (l,r)
	       then let val R1' = AS.union (X1,R1)
			val R2' = AS.union (X2,R2)
			val C1' = AS.union (Z1,C1)
			val C2' = AS.union (Z2,C2)
			val E' = AS.union (AS.difference (E,
							  AS.union (AS.union (X1,X2),
								    AS.union (Z1,Z2))),
					   AS.difference (Y,
							  AS.union (AS.union (R1,R2),
								    AS.union (C1,C2))))
		    in L.revAppend (ans, ((l,r),(R1',R2',E',C1',C2'))::ns)
		    end
	       else if Subst.identicalModuloRenamingRule (x,y) (r,l)
	       then let val R1' = AS.union (X1,R2)
			val R2' = AS.union (X2,R1)
			val C1' = AS.union (Z1,C2)
			val C2' = AS.union (Z2,C1)
			val E' = AS.union (AS.difference (E,
							  AS.union (AS.union (X1,X2),
								    AS.union (Z1,Z2))),
					   AS.difference (Y,
							  AS.union (AS.union (R2,R1),
								    AS.union (C2,C1))))
		    in L.revAppend (ans, ((r,l),(R1',R2',E',C1',C2'))::ns)
		    end
	       else
		   insert ((x,y),(X1,X2,Y,Z1,Z2)) ns (((l,r),(R1,R2,E,C1,C2))::ans)

       in
	   add (ns, insert n0 stable [])
       end

   fun elim nodes = 
       let fun similar (_,(R1,R2,E,C1,C2)) (x,y) = 
	       let val xR1 = AS.member (R1,x)
		   val yR1 = AS.member (R1,y)
	       in (xR1 andalso yR1) orelse (not xR1 andalso not yR1) end
	       andalso
	       let val xR2 = AS.member (R2,x)
		   val yR2 = AS.member (R2,y)
	       in (xR2 andalso yR2) orelse (not xR2 andalso not yR2) end
	       andalso
	       let val xE = AS.member (E,x)
		   val yE = AS.member (E,y)
	       in (xE andalso yE) orelse (not xE andalso not yE) end
	       andalso
	       let val xC1 = AS.member (C1,x)
		   val yC1 = AS.member (C1,y)
	       in (xC1 andalso yC1) orelse (not xC1 andalso not yC1) end
	       andalso
	       let val xC2 = AS.member (C2,x)
		   val yC2 = AS.member (C2,y)
	       in (xC2 andalso yC2) orelse (not xC2 andalso not yC2) end
	   fun elimSub [] ys = ys
	     | elimSub (x::xs) [] = elimSub xs [x]
	     | elimSub (x::xs) ys = 
	       if L.all (fn y => not (L.all (fn n => similar n (x,y)) nodes)) ys
	       then elimSub xs (x::ys)
	       else elimSub xs ys
       in
	   elimSub (!idlist) []
       end


   (* ¿ŴεǼƥå *)
   fun mkbstep isTerminating (oNodes,cNodes) = 
       let 
	   val (p, E1, E2, oNodes1,cNodes1) = choose isTerminating (oNodes,cNodes)
					      
	   val _ = debug (fn _ =>print ("Selected Node:\n"))
	   val _ = debug (fn _ => printNode p)

(*	   val _ = debug (fn _ =>print ("oNodes1:\n"))
	   val _ = debug (fn _ => L.app printNode oNodes1) *)

	   val (p2,oNodes2,cNodes2) = orient p (E1,E2) (oNodes1,cNodes1)

	   val oNodes4 = add (delete (deduce p2 cNodes2),add (oNodes2,[]))

	   val _ = debug (fn _ =>print ("oNodes4:\n"))
	   val _ = debug (fn _ => L.app printNode oNodes4)

	   val (cNodes4,oNodes4b) = properRewrite (cNodes2,[p2])

	   val _ = debug (fn _ =>print ("oNodes4b:\n"))
	   val _ = debug (fn _ =>L.app printNode oNodes4b)

	   val oNodes5 = gc (add (delete oNodes4b, oNodes4))
	   val cNodes5 = rev (gc (add (rev cNodes4,[p2])))

	   (*		   val _ = L.app printNode cNodes5 *)

	   (*		   val oNodes6 = gc (delete (rewrite (oNodes5,cNodes5))) *)
	   val _ = debug (fn _ =>print ("cNodes5:\n"))
	   val _ = debug (fn _ =>L.app printNode cNodes5)

	   val cNodes6 = ListMergeSort.sort
				     (fn (((l1,r1),_),((l2,r2),_))
					 => (Term.termSize l1) + Term.termSize (r1) 
					    > (Term.termSize l2) + Term.termSize (r2))
				     cNodes5


	   val oNodes6 = gc (add (delete (rewrite (oNodes5,cNodes6)), []))

	   val _ = debug (fn _ =>print ("oNodes6:\n"))
	   val _ = debug (fn _ =>L.app printNode oNodes6)


       (*		   val _ = idlist :=  elim (oNodes6 @ cNodes6) *)

       in 
	   (oNodes6,cNodes6)
       end

   fun mkbstep2 isTerminating (oNodes,cNodes) = 
       let 
		   val (p, E1, E2, oNodes1,cNodes1) = choose isTerminating (oNodes,cNodes)
			   
		   val _ = debug (fn _ =>print ("Selected Node:\n"))
		   val _ = debug (fn _ => printNode p)

		   val (p2,oNodes2,cNodes2) = orient p (E1,E2) (oNodes1,cNodes1)

		   val (cNodes3,oNodes3b) = properRewrite (cNodes2,[p2])

		   val oNodes5 = gc (add (delete (deduce p2 cNodes3), 
					  add (delete oNodes3b, add (delete oNodes2,[]))))

		   val cNodes5 = gc (add (cNodes3,[p2]))

		   val cNodes6 = ListMergeSort.sort
				     (fn (((l1,r1),_),((l2,r2),_))
					 => (Term.termSize l1) + Term.termSize (r1) 
					    > (Term.termSize l2) + Term.termSize (r2))
				     cNodes5

(*		   val oNodes6 = gc (delete (rewrite (oNodes5,cNodes5))) *)
		   val oNodes6 = gc (add (delete (rewrite (oNodes5,cNodes5)), []))

		   val _ = idlist :=  elim (oNodes6 @ cNodes6)

	   in 
		   (oNodes6,cNodes6)
	   end

					       
   (* ¿ŴΥᥤ롼 *)
   fun mkbtc isTerminating es =
       let val _ = print "Start Multi Knuth-Bendix Completion Procedure with a Termination Checker...\n"
	   val _ = print "Initial set of equations:\n"
	   val _ = print (Trs.prEqs es)
	   fun mkbsub n isTerminating (oNodes,cNodes)  = 
	       let val interval = 10
		   val _ = if n mod interval = 0
			   then print ("STEP " ^ (Int.toString n) ^ "\n")
			   else debug (fn _ => print ("STEP " ^ (Int.toString n) ^ "\n"))

(* 		   val _ = if n =  50 *)
(* 			   then (runDebug := true; *)
(* 				 DpSolver.runDebug := true) *)
(* 			   else () *)

		   val _ = debug (fn _ => print ("Num. of Nodes: Open " 
				  ^ (Int.toString (length (oNodes))) ^ ", Closed "
				  ^ (Int.toString (length (cNodes))) ^ "\n"))
		   val _ = if n mod interval = 0
			   then print ("Num. of Proc.: " 
				       ^ (Int.toString (length (!idlist))) ^ "\n")
			   else debug (fn _ => print ("Num. of Proc.: " 
						      ^ (Int.toString (length (!idlist))) ^ "\n"))

		   val _ = debug (fn _ => print "Open nodes:\n")
 		   val _ = debug (fn _ => L.app printNode oNodes)
		   val _ = debug (fn _ => print "Closed nodes:\n")
 		   val _ = debug (fn _ => L.app printNode cNodes)

(*		   val (oNodes',cNodes') = mkbstep isTerminating (oNodes,cNodes) *)
 		   val (oNodes',cNodes') = mkbstep2 isTerminating (oNodes,cNodes) 

		   val _ = debug (fn _ =>
			       List.app 
			       (fn id =>
				   let val idstr = prId id
				   in
				   (print ("ID " ^ idstr ^ " ("  ^ "Step " 
					   ^ (Int.toString ((String.size idstr) - 1)) ^ ")\n");
				    print ("O Rules:\n" ^ (Trs.prRules (rPart id oNodes')));
				    print ("O Equations:\n" ^ (Trs.prEqs (ePart id oNodes')));
				    print ("O Constraints:\n" ^ (Trs.prRules (cPart id oNodes')));
				    print ("C Rules:\n" ^ (Trs.prRules (rPart id cNodes'))));
				    print ("C Equations:\n" ^ (Trs.prEqs (ePart id cNodes')));
				    print ("C Constraints:\n" ^ (Trs.prRules (cPart id cNodes')))
				   end)
			       (!idlist))
	       in
		   case isSucceeded (oNodes',cNodes') of
		       SOME id => let val rs' = (rPart id cNodes') @ (rPart id oNodes')
				      val idstr = prId id
				  in (print ("Succeed in " ^ (Int.toString n) ^ " steps \n");
				      print ("ID " ^ idstr ^ " ("  ^ "Step " 
					     ^ (Int.toString ((String.size idstr)-1)) ^ ")\n");
				      print (Trs.prRules rs'); true)
				  end
		     | NONE => mkbsub (n+1) isTerminating (oNodes',cNodes')
	       end

	   val initialId = Atom.atom "e" 
	   val _ = addId initialId
	   val initialNodes = List.map 
				  (fn e => (e,(AS.empty,AS.empty,
					       AS.singleton initialId,
					       AS.empty,AS.empty)))
				  es
       in
	   mkbsub 1 isTerminating (initialNodes,[])
	   handle MkbtcError => (print "Failed\n"; false)
       end

   end (* of local *)

   end (* struct *)
