(* Unfailing Knuth-Bendix Completion Prover *)
(* 2009.Feb.03 Pham Tuong Van *)
signature SigMuc = 
sig
	exception ERROR

	val mucstep : (atom * (term * term -> bool)) list
		-> ((term * term) * (set * set * set)) list * ((term * term) * (set * set * set)) list
		-> ((term * term) * (set * set * set)) list	* ((term * term) * (set * set * set)) list
	val mucsub  : int
		-> (atom * (term * term -> bool)) list
		-> ((term * term) * (set * set * set)) list	* ((term * term) * (set * set * set)) list
		-> bool
	(* Manual order, MUC, input: Order list, terms, where Order: Function lex order list *)
	val mmuc  : string list list -> (term * term) list -> bool	
	(* Auto all order MUC, input: funtion key list *)
	val amuc  : string list -> (term * term) list -> bool	
    val debug : bool   

end;

structure Muc : SigMuc = struct
	
    open Term
	open Trs
	open Rewrite
	open Subst
	open ECr
	structure A = Atom
	structure AS = AtomSet
	structure L = List
	fun mapAppend f xs = List.foldr (fn (x,ys) => List.@(f x, ys)) [] xs
		
    exception ERROR;	
	
	(* Input: Function key list ex. ["D","Plus","+"] Output: [(1,Order1),(2,Order2),... *)
	fun generateOrders FList =
		let
			fun n! 0 = 1 | n! x = x * n! (x-1);
			(* choose m th Atom *)
			fun AChoose _ _ [] = NONE |
				AChoose 0 xs (y::ys) = SOME (y,xs@ys) |
				AChoose m xs (y::ys) = AChoose (m-1) (xs@[y]) ys
			(* Convert order number to natural num order list *)
			fun convertToNList 0 _ = [0] | 
				convertToNList m a = (a div (n! m))::(convertToNList (m-1) (a mod (n! m))); 
			(* Convert a nartural num order list to Atom order list *)
			fun convertToAList [] _ = [] |
				convertToAList (x::xs) ls =
				case AChoose x [] ls of
					NONE => convertToAList xs ls
				  | SOME (at,nls) => at::(convertToAList xs nls)		
			val len = L.length FList - 1 
			val AList = L.map (fn x=> A.atom x) FList 
			fun convertToOrder num = 
				Order.qlposCheck (Order.comparePrec (convertToAList (convertToNList len num) AList)) 
				(fn _ => (fn xs => xs)) Order.GT
			val orderNum = n! (len+1) 
			fun generate 0 = [] |
				generate n = (A.atom (Int.toString n),(convertToOrder (n-1)))::(generate (n-1));
		in rev (generate orderNum)
		end;
	(* Input: Order function key list ex. [["D","Plus","+"],[...],...] Output: [(1,Order1),(2,Order2),... *)
	fun manualOrder OList =
		let 
			fun toOrder ls = 
				Order.qlposCheck (Order.comparePrec (L.map (fn x => A.atom x) ls)) 
				(fn _ => (fn xs => xs)) Order.GT
			fun output _ [] = [] |
				output n (od::ods) = (A.atom (Int.toString n),(toOrder od))::(output (n+1) ods)
		in output 1 OList
		end;	

	(* -------- Base on mkbtc.sml by Aoto ---------*)
	fun switchNode ((l,r),(L1,L2,L3)) = ((r,l),(L2,L1,L3))
		
	fun r_Proj id ps = mapAppend (fn ((l,r),(L1,L2,_)) => 
								  (if AS.member (L1,id) then [(l,r)] else [])
									  @ (if AS.member (L2,id) then [(r,l)] else [])) ps
				   
	fun e_Proj id ps = L.mapPartial (fn (e,(_,_,L3)) => 
									   if AS.member (L3,id) then SOME e else NONE) 
		ps
		
	fun mapAppend f xs = List.foldr (fn (x,ys) => List.@(f x, ys)) [] xs
		
	fun isSucceeded maxId (oNodes,cNodes) = 
		let 
			fun generateId 0 = []
			  | generateId n = (Atom.atom (Int.toString n))::(generateId (n-1))

			val idls = generateId maxId
		in if null oNodes then SOME (A.atom "1") else
			L.find (fn id => null (e_Proj id oNodes)
					andalso null (e_Proj id cNodes)
					andalso null (r_Proj id oNodes))
			idls
		end

	fun printNode ((s,t),(R1,R2,E)) =
		let val _ = print ("[<" ^ (Term.toString s) ^ " :+:  "
						   ^ (Term.toString t) ^ ">, ")
			val _ = print ("L1:" ^ (PrintUtil.prSetInOneLine
									A.toString (AS.listItems R1)) ^ ", ")
			val _ = print ("L2:" ^ (PrintUtil.prSetInOneLine
									A.toString (AS.listItems R2)) ^ ", ")
			val _ = print ("L3:" ^ (PrintUtil.prSetInOneLine
								   A.toString (AS.listItems E)) ^ ", ")
		in ()
		end

	fun prns es = L.map (fn nd => (printNode nd; print"\n")) es

	fun prPeak (a,b,c) = (print (IOFotrs.prTerm b); print " <- "; print (IOFotrs.prTerm a); print " -> "; print (IOFotrs.prTerm c); print "\n")
	fun prPeaks es = L.map (fn p => prPeak p) es

	(* ------ End of Modification ------- *)

	(* Choose from open Node a Node with smallest size of s:t*)
	fun chooseNode [] = (print "No equation to expand\n";  raise ERROR)
	  | chooseNode (onod::oNodes) =
		let
			(* Choose the node with least size and appear in much Label (Process) *)
			fun nodeSize ((l,r),(_,_,L3)) = termSize l + termSize r + 1000 - (AS.numItems L3);
			
			fun findLeast rest anod min [] = (anod,rev rest)
			  | findLeast rest anod min (nod::nods) = 
				let 
					val thisSize = nodeSize nod
				in 
					if thisSize < min then findLeast (anod::rest) nod thisSize nods
					else findLeast (nod::rest) anod min nods
				end
		in 
			findLeast [] onod (nodeSize onod) oNodes
		end   

	(* Delete *)
	fun delete nodes = L.filter (fn ((l,r),(_,_,_)) => not (Term.equal (l,r))) nodes

	(* Garbage Collect *)
	fun gc nodes = L.filter 
		(fn ((l,r),(L1,L2,L3)) =>
		 not (AS.isEmpty L1)
		 orelse not (AS.isEmpty L2)
		 orelse not (AS.isEmpty L3))
		nodes

	fun findOrder order key =
		case L.find (fn (id,_) => A.sameAtom(key,id)) order of
			SOME (_,odr) => odr
		  | NONE => (fn _ =>false)

	fun rootORewrite ord [] term = NONE
	  | rootORewrite ord ((l,r)::rs) term = 
		case match l term
			of SOME sigma => 
				let 
					val rsigma = applySubst sigma r
				in if ord (term, rsigma) then SOME (applySubst sigma r)
				   else rootORewrite ord rs term
				end
		  | NONE => rootORewrite ord rs term
				 
   fun linfO ord R term =
       case term of 
		   Var _ => term
		 | Fun (f,ts,sort) => 
			   let
				   val ss = L.map (linfO ord R) ts
			   in case rootORewrite ord R (Fun (f,ss,sort)) of
				   SOME t => linfO ord R t
				 | NONE => Fun (f,ss,sort)
			   end

   fun statusLinfO ord R term =
       case term 
		   of Var _ => (false, term)
		 | Fun (f,ts,sort) => 
			   let val statusXss = L.map (statusLinfO ord R) ts
				   val ss = L.map #2 statusXss
			   in case rootORewrite ord R (Fun (f,ss,sort))
				   of SOME t => (true, linfO ord R t)
				 | NONE => (L.exists (fn x => #1 x) statusXss, Fun (f,ss,sort))
			   end

   fun linfOPartial ord R term =
       let val (status, nf) = statusLinfO ord R term
       in if status then SOME nf else NONE
       end
	
	(* -------- Rewrite -------*)
	local 

		(* divide L in L1 U L2, which for all i L1, so >i to, otherwise i in L2 *)
		(*fun divideL ord so to aL =
			AS.foldr (fn (key, (aL1,aL2)) =>
					  if (findOrder ord key) (so,to) then 
						  (AS.add(aL1,key),aL2)
					  else (aL1, AS.add (aL2,key)))
			(AS.empty, AS.empty) aL *)

		(* Fill all item in aL that so >_i to *)
		fun findLabel ord so to aL =
			AS.filter (fn key => (findOrder ord key) (so,to)) aL

		(* tryRewrite (by rule) (in term) *)
		fun tryRewrite order ((l,r),(L1',L2',L3')) ((s,t),(L1,L2,L3)) =
		 if not (AS.isEmpty L3')
			 orelse not (AS.isEmpty L1')
		 then (* rewrite1 *)
			 if Subst.isVariant t l
			 then case Rewrite.rootRewrite [(l,r)] t of
				 SOME u => let 							
							   val L = findLabel order t u L3'
							   val L' = AS.union (L1',L)
							   val L1capL' = AS.intersection (L',L1)
							   val L3capL1' = AS.intersection (L1',L3)
						   in if not (AS.isEmpty L1capL')
							   orelse not (AS.isEmpty L3capL1')
							  then SOME (((s,t),(AS.difference (L1,L1capL'),L2,
												 AS.difference (L3,L3capL1'))),
										 ((s,u),(L1capL',AS.empty,L3capL1')))
							  else NONE 
						   end
			   | NONE => NONE
			 else (* rewrite2 *)
				 (* Rule: l -> r is really exist *)
				 if not (AS.isEmpty L1') then
					 case Rewrite.linfPartial [(l,r)] t of
						 SOME u => let 
								 val L = findLabel order t u L3'
								 val L' = AS.union (L1',L)
								 val L1capL' = AS.intersection (L',L1)
								 val L2capL' = AS.intersection (L',L2)
								 val L3capL' = AS.intersection (L',L3)
							 in if not (AS.isEmpty L1capL')
								 orelse not (AS.isEmpty L2capL')
								 orelse not (AS.isEmpty L3capL')
								then SOME (((s,t),(AS.difference (L1,L1capL'),
												   AS.difference (L2,L2capL'),
												   AS.difference (L3,L3capL'))),
										   ((s,u),(L1capL',AS.empty,
												   AS.union (L2capL',L3capL'))))
								else NONE
							 end
					   | NONE => NONE
				 (* Rule l->r exist or not, this may lead to unfinite loop *)
				 else 
					 let
						 val L' = AS.filter 
							 (fn key =>((fn (a,b) => a)
										(statusLinfO (findOrder order key) [(l,r)] t )
										))
							 L3'
						 val LL = AS.find
							 (fn key =>((fn (a,b) => a)
										(statusLinfO (findOrder order key) [(l,r)] t )
										))
							 L3'	   
							 
						 val su = case LL of 
							 SOME key => (s, linfO (findOrder order key) [(l,r)] t)
						   | NONE => (s,s)

						 val L1capL' = AS.intersection (L',L1)
						 val L2capL' = AS.intersection (L',L2)
						 val L3capL' = AS.intersection (L',L3)						 
					 in if not (AS.isEmpty L1capL')
						 orelse not (AS.isEmpty L2capL')
						 orelse not (AS.isEmpty L3capL')
						then SOME (((s,t),(AS.difference (L1,L1capL'),
										   AS.difference (L2,L2capL'),
										   AS.difference (L3,L3capL'))),
								   (su,(L1capL',AS.empty,
										   AS.union (L2capL',L3capL'))))
						else NONE
					 end
		 else NONE

		(* rewriteNode (by RULES) (in TERM) *)
		fun rewriteNode order [] _ = NONE
		  | rewriteNode order (((l,r),(L,L',E))::ns) ((s,t),(L1,L2,L3)) = 
			case tryRewrite order ((l,r),(L,L',E)) ((s,t),(L1,L2,L3)) of
				SOME (n,m) => SOME (n,m)
			  | NONE => case tryRewrite order ((r,l),(L',L,E)) ((s,t),(L1,L2,L3)) of
					SOME (n,m) => SOME (n,m)
				  | NONE => case tryRewrite order ((l,r),(L,L',E)) ((t,s),(L2,L1,L3)) of
						SOME (n,m) => SOME (switchNode n, switchNode m)
					  | NONE => case tryRewrite order ((r,l),(L',L,E)) ((t,s),(L2,L1,L3)) of
						    SOME (n,m) => SOME (switchNode n, switchNode m)
						  | NONE => rewriteNode order ns ((s,t),(L1,L2,L3))

		(* divideNode byRULES inTERMS : rewrite terms by rules and divide to 
		 Nothing happen, Updated, Created *)
		fun divideNodes order rls ts =
			let
				fun merge (nod,(occur,os,us,cs)) = case rewriteNode order rls nod of
					NONE => (occur,nod::os,us,cs)
				  | SOME (n,m) => (true,os,n::us,m::cs)
			in
				L.foldr merge (false,[],[],[]) ts
			end

		(* rewriteStep by_RULES in_TERMS Updated Created -> (Updated, Created Node) *)
		fun rewriteStep order origin rls = 
			let
				val (occur1,ORG,u1,c1) = divideNodes order rls origin
				fun findUpdate nds uds crs = 
					case divideNodes order rls nds of
						(false,_,_,_) => (nds@uds,crs)
					  | (true,a,b,c) => findUpdate b (a@uds) (c@crs)
				val (UPD,c2) = findUpdate u1 [] []
				fun findCreate nds crs = case divideNodes order rls nds of
					(false,_,_,_) => nds@crs
				  | (true,a,b,c) => findCreate (b@c) (a@crs)
				val CRT = findCreate (c1@c2) []
			in (ORG,UPD,CRT)
			end
	in
		   fun rewrite order (terms,rules) =  rewriteStep order terms rules

	end

	(* - End of Rewrite ([Nc,{n}]) - *)   
				
	(* Orient *)
	fun orient order ((lt,rt),(aL1,aL2,aL3)) =
		let
			fun tryOrient (L1,L2,L3) [] = (L1,L2,L3)
			  | tryOrient (L1,L2,L3) (l::ls) = 
				let
					val thisOrder = findOrder order l
				in
					if thisOrder (lt,rt) 
						then tryOrient (AS.add(L1,l),L2,L3) ls
					else if thisOrder (rt,lt) 
							 then tryOrient (L1,AS.add(L2,l),L3) ls
						 else tryOrient (L1,L2,AS.add(L3,l)) ls
				end
		in ((lt,rt),tryOrient (aL1,aL2,AS.empty) (AS.listItems aL3))
		end

	(* Deduction *)
	
	fun deduce Order n0 nodes =
		let 
			fun isLegalCriticalPair key (a,u,v) = 
				let
					val ord = findOrder Order key 
					val isLegal = (not (ord (u,a))) andalso (not (ord(v,a)))
					val isNotVarial = not (Subst.isVariant u v)
				in
					isLegal andalso isNotVarial
				end
			
			fun filterLabel (a,u,v) aL =
				AS.filter (fn key=> isLegalCriticalPair key (a,u,v)) aL
			
			fun addcpstep0 ((l,r),(L1,L2,L3)) =	
				let
					val L1UL3 = AS.union (L1,L3)
				in if not (AS.isEmpty L1UL3) then
					L.foldr
					(fn ((a,u,v),nls) => 				 
					 let
						 val nL = filterLabel (a,u,v) L1UL3
						 (* Disable this will turn of print out Crpeaks *)
						 (*val _ = if not (AS.isEmpty nL) then prPeak (a,u,v) else () *)
					     (* --- *)	 
					 in						 
						 if AS.isEmpty nL then nls 
						 else (((u,v),(AS.empty,AS.empty,nL))::nls)
					 end)
					[] (Cr.insideCriticalPeaks [(l,r)])
				   else []
				end
			
			fun addcpstep ((l,r),(L1',L2',L3')) ((s,t),(L1,L2,L3)) = 
				let 
					val L1UL3 = AS.union (L1,L3)
					val L1UL3' = AS.union (L1',L3')
					val aL = AS.intersection (L1UL3,L1UL3')
				in if not (AS.isEmpty aL)
					   then
						   L.foldr
						   (fn ((a,u,v),nls) =>
							let
								val nL = filterLabel (a,u,v) aL
								(* Disable this will turn of print out Crpeaks *)
								(* val _ = if not (AS.isEmpty nL) then prPeak (a,u,v) else () *)
							    (* --- *)
							in						 
								if AS.isEmpty nL then nls 
								else (((u,v),(AS.empty,AS.empty,nL))::nls)
							end)
							[]
							(L.@(Cr.criticalPeaks2 ([(l,r)],[(s,t)]),
								 Cr.insideCriticalPeaks2 ([(s,t)],[(l,r)])))
				   else []
				end
			
			fun addcp0 ((l,r),(R1,R2,E)) =
				(addcpstep0 ((l,r),(R1,R2,E)))
				@ (addcpstep0 ((r,l),(R2,R1,E)))
				
			fun addcp ((l,r),(R1,R2,E)) ((x,y),(X1,X2,E')) =
				(addcpstep ((l,r),(R1,R2,E)) ((x,y),(X1,X2,E')))
				@ (addcpstep ((l,r),(R1,R2,E)) ((y,x),(X2,X1,E')))
				@ (addcpstep ((r,l),(R2,R1,E)) ((x,y),(X1,X2,E')))
				@ (addcpstep ((r,l),(R2,R1,E)) ((y,x),(X2,X1,E')))
		in
			(addcp0 n0) @ (mapAppend (fn n1 => addcp n0 n1) nodes)
		end

	(* -----End of Deduction -----*)

	(* Union and subsume *)
	fun addNodes ([], stable) = stable
	  | addNodes (n0::ns, stable) =
		let 
			fun insert node [] ans = node::ans
			  | insert ((x,y),(X1,X2,Y)) (((l,r),(R1,R2,E))::ns) ans =
				if Subst.identicalModuloRenamingRule (x,y) (l,r)
				then let val R1' = AS.union (X1,R1)
						 val R2' = AS.union (X2,R2)
						 val E'  = AS.difference 
							 (AS.union (Y,E),
							  AS.union (R1',R2')) 			

					 in ((l,r),(R1',R2',E'))::(ns @ ans)
					 end
				else 
					if Subst.identicalModuloRenamingRule (x,y) (r,l)
					then let 
							 val R1' = AS.union (X1,R2)
							 val R2' = AS.union (X2,R1)
							 val  E'  = AS.difference 
								 (AS.union (Y,E),
								  AS.union (R1',R2')) 			
								 
						 in ((r,l),(R1',R2',E'))::(ns @ ans)
						 end
					else
						insert ((x,y),(X1,X2,Y)) ns (((l,r),(R1,R2,E))::ans)

		in
			addNodes (ns, insert n0 stable [])
		end

	(* --- End of Union and subsume --- *)
	fun mucstep Orders (oNodes,cNodes) = 
		let 
			fun isNonsense (_,(a,b,c)) = (AS.isEmpty a) andalso (AS.isEmpty b) andalso (AS.isEmpty c)
			fun isNonRule (_, (a,b,c)) = (AS.isEmpty a) andalso (AS.isEmpty b)
			
			(* if a node is nonrule and have L3 is subset of another node in another Closed node *) 
			fun isMeaningless _ [] = false
			  | isMeaningless ((a,a'),(b,c,d)) (((e,e'),(f,g,h))::xs) =
				if (AS.isSubset (d,AS.union(h,AS.union(f,g)))) andalso
					((Subst.identicalModuloRenamingRule (a,a') (e,e'))
					 orelse (Subst.identicalModuloRenamingRule (a,a') (e',e)))
				then true else isMeaningless ((a,a'),(b,c,d)) xs

			fun chooseOrientable chosed [] = chooseNode (rev (chosed))
			  | chooseOrientable chosed nodes =
				let
					val (ok,rest) = chooseNode nodes
					fun orientable nod =
						case orient Orders nod of
							(x,(a,b,c)) => not ((AS.isEmpty a) andalso (AS.isEmpty b))
				in if isNonRule ok then
					  if isMeaningless ok cNodes then chooseOrientable chosed rest 
					  else if orientable ok then (ok, chosed @ rest)
						   else chooseOrientable (ok::chosed) rest
				   else (ok,chosed @ rest)
				end

			val (chosen, restoNodes) = chooseOrientable [] oNodes
			   
			val _ = (print ("-----Selected Node: \n"); printNode chosen; print "\n")
			(*val _ = (print ("-----Open Nodes: \n"); prns restoNodes)
			val _ = (print ("-----Close Nodes: \n"); prns cNodes)*)
				
			val (t1,t2,t3) = rewrite Orders ([chosen], cNodes)
			val chosen1 = if null t2 then chosen else (fn (a::b) => a) t2

			val oNodes1 = addNodes(delete(t3), restoNodes)
			val an = orient Orders chosen1

			val (oNodes2,cNodes2) = (fn (a,b,c) => (c,a@b)) (rewrite Orders (cNodes, [an]))

			val oNodes3 = addNodes(delete(oNodes2), oNodes1)

			val cNodes3 = rev(gc(cNodes2))

			val deduceNodes = delete(deduce Orders an cNodes3)
			(*val _ = (print ("-----Deduce Nodes: \n"); prns deduceNodes)*)
			val oNodes4 = addNodes(deduceNodes, oNodes3)
			val (t4,t5,t6) = rewrite Orders (oNodes4, cNodes3)
			val oNodes5 = addNodes(t5,t4)
			val oNodes6 = addNodes(t6, oNodes5)
			val oNodes7 = rev (gc(delete oNodes6))

			val cNodes4 = addNodes([an],cNodes3)

		in 
			
			(oNodes7,cNodes4)
		end

	fun mucsub n Orders (oNodes,cNodes) =
		let 
			val _ = print ("Step: " ^ (Int.toString n) ^ "\n")
			val _ = (Int.toString (length (oNodes))) ^ " open Nodes | " ^
				    (Int.toString (length (cNodes))) ^ " close Nodes \n"
			val (oNodes', cNodes') = mucstep Orders (oNodes,cNodes)
			val Odlen = L.length Orders
		in case isSucceeded Odlen (oNodes',cNodes') of
				SOME id => 
					let
						val rs' = r_Proj id cNodes'
						val es' = e_Proj id cNodes'
					in 
						(print ("\n Succeed in " ^ (Int.toString n) ^ " steps \n");
						 print ("--------Rules:--------\n"); print (Trs.prRules rs'); 
						 print ("-------Equation:------\n"); print (Trs.prEqs es');
						 true)
					end
			  | NONE => mucsub (n+1) Orders (oNodes',cNodes')
		end
	
	(* Manual order mbk *)
	fun mmuc orderStrings es =
		let
 			val orders = manualOrder orderStrings
				
			val _ = "Manual order MUC Completion ...\n"
			
			fun generateInitialId 0 IID = IID 
			  | generateInitialId n IID = 
				generateInitialId (n-1) (AS.add(IID, A.atom (Int.toString n)))
			
			val initialID = generateInitialId (L.length orders) AS.empty
			
			val initialNodes = L.map (fn e => (e,(AS.empty,AS.empty,initialID))) es
		in mucsub 1 orders (initialNodes,[])
		
		handle ERROR => (print "Failed\n"; false) 
		
		end

	fun amuc functionString es =
		let
 			val orders = generateOrders functionString
				
			val _ = "Automatic all order MUC Completion ...\n"
			
			fun generateInitialId 0 IID = IID 
			  | generateInitialId n IID = 
				generateInitialId (n-1) (AS.add(IID, A.atom (Int.toString n)))
			
			val initialID = generateInitialId (L.length orders) AS.empty
			
			val initialNodes = L.map (fn e => (e,(AS.empty,AS.empty,initialID))) es
		in mucsub 1 orders (initialNodes,[])
		
		handle ERROR => (print "Failed\n"; false) 
		
		end

	val debug = true;

end;

(*
fun try data = (fn (a,b) => (print "Open: \n"; pns a; print "Closed: \n"; pns b; (a,b))) (Muc.mucstep ODS data); *)