(******************************************************************************
 * file: rwtools/kbcomp/test_kb.sml
 * description: tests routine for various Knuth-Bendix Completion Procedures
 * author: AOTO Takahito
 * 
 * Copyright (c) 2008, Toyama&Aoto Laboratory, Tohoku University
 ******************************************************************************)

signature TEST_KB=  
  sig
      val minisatPath: string ref
      val tmpDir: string ref

      val kbCheck:   Term.decl list
                   * (Term.term * Term.term) list
                   * (Term.term * Term.term) list
                   * (int * Fun.ord_key * Order.rpo_status * int list) list
                   * (Fun.ord_key * Poly.poly) list
		   * (int * (int * int * Fun.ord_key * int list) list)
                   -> bool

      val kbtcCheck: string 
		     -> string
		     -> DpSolver.options
		     -> PoSolver.options
		     -> bool  (* $BJB9T$G$d$k$+$I$&$+(B *)
		     -> bool  (* $BB?=E$G$d$k$+$I$&$+(B *)
		     -> string (* $BDd;_@-8!>Z4o$N%3%^%s%I(B *)
		     -> int   (* $BDd;_@-8!>Z4o$N%?%$%`%j%_%C%H(B($BIC(B) *)
		     -> Term.decl list
			* (Term.term * Term.term) list
			* (Term.term * Term.term) list
                     -> bool

      val kbTest: int -> bool
      val kbTest2: int -> bool

  end;

structure TestKb : TEST_KB =
    struct

    local 
       open Trs
       structure IO = IOMstrs
       structure FS = FunSet
       structure SS = SortSet
       structure FM = FunMap
       structure L = List
       structure PU = PrintUtil
   in

    exception TestKbError

    val minisatPath = ref "../../work/tools/bin/minisat"   (* may over-rewritten *)
    val tmpDir = ref "." (* may over-rewritten *)

    fun time f arg=
        let
            val timer = Timer.startCPUTimer ()
            val start = TimeUtil.checkTime timer 
            val result = f arg
            val stop = TimeUtil.checkTime timer 
            val _ = print (" ("  
                           ^ (TimeUtil.reportTime (start,stop)) 
                           ^ " msec.)\n")        
        in
            result
        end

    fun compilePrec (sorts,decls:Term.decl list ,prec) = 
        let 
           val sortSet = List.foldr (fn (s,set) => SS.add (set,s)) SS.empty sorts
           val fs = map (fn x => #sym x) decls
           val faMap = let fun arity (Sort.Base _) = 0
                             | arity (Sort.Proc (xs,_)) = length xs
                       in
                           L.foldr (fn (x,fm) => FM.insert (fm,#sym x,arity (#sort x)))
                                   FM.empty decls
                       end

           val prec' = List.map 
                           (fn (i,f,st,ns) => 
                                   (i,f,st,
                                          case st of
                                              Order.LEX => let val arity = valOf (FM.find (faMap,f))
                                                           in
                                                               if (length ns) = arity
                                                               then ns
                                                               else if (null ns)
                                                               then List.tabulate (arity,fn x=>x+1)
                                                               else (PU.error "invalid prec spec";
                                                                     raise TestKbError)
                                                           end
                                            | Order.MUL => if not (null ns)
                                                           then (PU.error "invalid prec spec";
                                                                  raise TestKbError)
                                                           else ns)
                               )
                           prec
           val stMap = List.foldr (fn ((_,f,st,_),fm) => FM.insert (fm,f,st)) FM.empty prec'
           val lexMap = L.foldr (fn ((_,f,_,ns),fm) => FM.insert (fm,f,ns)) FM.empty prec'
           val funqprec = List.foldr (fn ((i,f,_,_),m) => FM.insert (m,f,i))  FM.empty prec'
           fun isMul f = valOf (FM.find (stMap,f)) = Order.MUL
           fun lex f ts = let val ns = valOf (FM.find (lexMap,f))
                              val ns2 = ListPair.zip (ns,ts)
                              val sorted = ListMergeSort.sort
                                               (fn (x,y) => Int.> (#1 x,#1 y))
                                               ns2
                          in List.map #2 sorted
                          end
        in
            (funqprec,isMul,lex)
        end

    fun kbCheckByPrec (sorts:Sort.ord_key list,decls,seqs,sconj,prec) = 
        let val (funqprec,isMul,lex)  = compilePrec (sorts,decls,prec)
            fun greaterWithMul (l,r) =
                   let 
                           val symCount = ref 1
                           val prop = #1 (Order.encodeRpoCheck (Order.compareQprec funqprec) 
                                                          isMul lex symCount Order.GT (l,r))
                   in
                           if prop = Prop.True
                                   then true
                           else if prop = Prop.False
 
                           then false
                           else (* print ("---check--- " ^ (Prop.printProp prop) ^ "\n");
                                 print ("(l,r) =     " ^ (Trs.prRule (l,r)) ^ "\n");
                                 Solver.runDebug := true; *)
                               #1 (Solver.propSolver (!minisatPath) (!tmpDir) (prop,!symCount))
                   end

           fun greaterWithMul2 (l,r) =
	       Order.qrpoCheck (Order.compareQprec funqprec) 
                                  isMul lex Order.GT (l,r)

           fun greaterWithoutMul (l,r) = 
                   Order.qlposCheck (Order.compareQprec funqprec) lex Order.GT (l,r)

           val containMul = L.exists (isMul o (fn x => #sym x)) decls

            val greater = if containMul 
(*			  then greaterWithMul2  ... test yet to be done *)
			  then greaterWithMul
			  else greaterWithoutMul 

(*           val greater = greaterWithMul *)

        in
	    if null sconj
	    then Kb.kb greater seqs
	    else KbProver.kbProver greater seqs (hd sconj)
        end


    fun compileKbPrec (sorts,decls:Term.decl list,(_,prec)) = 
        let 
           val sortSet = List.foldr (fn (s,set) => SS.add (set,s)) SS.empty sorts
           val fs = map (fn x => #sym x) decls
           val faMap = let fun arity (Sort.Base _) = 0
                             | arity (Sort.Proc (xs,_)) = length xs
                       in
                           L.foldr (fn (x,fm) => FM.insert (fm,#sym x,arity (#sort x)))
                                   FM.empty decls
                       end

           fun getNs f ns = 
	       let val arity = valOf (FM.find (faMap,f))
               in
                   if (length ns) = arity
                   then ns
                   else if (null ns)
                   then List.tabulate (arity,fn x=>x+1)
                   else (PU.error "invalid prec spec";
                         raise TestKbError)
               end

           val lexMap = L.foldr (fn ((_,_,f,ns),fm) => FM.insert (fm,f,getNs f ns)) FM.empty prec
           val funqprec = List.foldr (fn ((i,_,f,_),m) => FM.insert (m,f,i))  FM.empty prec
           val weight = List.foldr (fn ((_,j,f,_),m) => FM.insert (m,f,j))  FM.empty prec

           fun lex f ts = let val ns = valOf (FM.find (lexMap,f))
                              val ns2 = ListPair.zip (ns,ts)
                              val sorted = ListMergeSort.sort
                                               (fn (x,y) => Int.> (#1 x,#1 y))
                                               ns2
                          in List.map #2 sorted
                          end
        in
            (funqprec,weight,lex)
        end


    fun kbCheckByKbprec (sorts:Sort.ord_key list,decls,seqs,sconj,kbprec) = 
        let val (funqprec,weight,lex)  = compileKbPrec (sorts,decls,kbprec)
	    val varWeight = #1 kbprec
            fun greater (l,r) = 
                   Kbo.kboCheck varWeight weight (Order.compareQprec funqprec) lex (l,r)
        in
	    if Kbo.isValidKboSpec varWeight weight (Order.compareQprec funqprec) decls
	    then if null sconj
		 then Kb.kb greater seqs
		 else KbProver.kbProver greater seqs (hd sconj)
	    else (PU.error "Invalid KBO specification (or maybe no order is specified?) \n";
                  raise TestKbError)
        end


    fun kbCheckByPoly (sorts:Sort.ord_key list,decls,seqs,sconj,poly) = 
	let fun greater (l,r) = Poly.poly_strict poly (l,r)
	in
	    if null sconj
	    then Kb.kb greater seqs
	    else KbProver.kbProver greater seqs (hd sconj)
        end

    fun kbCheck (decls,eqs,conj,prec,poly,kbprec) = 
	let val seqs = let val res = Trs.attachSortToRules decls eqs
                       in if (isSome res)
                          then valOf res
                          else (PU.error "not well-sorted equations";
                                raise TestKbError)
                       end
	    val sorts = SortSet.listItems (Term.sortSetInDecls decls)

            val sconj = let val res = Trs.attachSortToRules decls conj
                        in if (isSome res)
                           then valOf res
                           else (PU.error "not well-sorted equations";
                                 raise TestKbError)
                        end
	in
	    if (null prec) 
            then if (null poly) 
		 then kbCheckByKbprec (sorts,decls,seqs,sconj,kbprec) 
		 else kbCheckByPoly (sorts,decls,seqs,sconj,poly) 
	    else kbCheckByPrec (sorts,decls,seqs,sconj,prec) 
	end

    fun kbtcCheck minisatPath tmpDir 
		  (opt1:DpSolver.options) (opt2:PoSolver.options) 
		  isParallel isMulti
		  snChecker timeLimit (decls,eqs,conj) = 
	let val seqs = let val res = Trs.attachSortToRules decls eqs
                       in if (isSome res)
                          then valOf res
                          else (PU.error "not well-sorted equations";
                                raise TestKbError)
                       end
            val sconj = let val res = Trs.attachSortToRules decls conj
                        in if (isSome res)
                           then valOf res
                           else (PU.error "not well-sorted equations";
                                 raise TestKbError)
                        end

            fun isTerminating rs = 
		let
		    fun main () = 
			if snChecker = ""
			then
			    if (#useDp opt1)
			    then (print "."; DpSolver.dpSolver minisatPath tmpDir opt1 opt2 rs)
			    else PoSolver.poSolver minisatPath tmpDir opt2 (rs,[])
			else
			    Solver.snSolver snChecker tmpDir rs
			    
 		    val result = 
			if timeLimit = 0
			then main () 
			else TimeUtil.timeLimitRun (IntInf.fromInt timeLimit) main () 
			     handle TimeLimit.TimeOut => 
				    (if OS.FileSys.access (!Solver.inTmpFileName, [OS.FileSys.A_WRITE])
				     then (OS.FileSys.remove (!Solver.inTmpFileName);
					   Solver.inTmpFileName := "")
				     else ();
				     if OS.FileSys.access (!Solver.outTmpFileName, [OS.FileSys.A_WRITE])	
				     then (OS.FileSys.remove (!Solver.outTmpFileName);
					   Solver.outTmpFileName := "")
				     else ();
				     false)
		in
		    result
		end
	in
	    if isMulti
	    then (Mkbtc.minisatPath := minisatPath;
		  Mkbtc.tmpDir := tmpDir;
		  Mkbtc.mkbtc isTerminating seqs)
	    else
		if isParallel
		then Kbtc.pkbtc isTerminating seqs
		else Kbtc.kbtc isTerminating seqs
	end

    fun kbTest 1 = 
        time kbCheck 
             ( (* (IOMstrs.rdSorts ["Any"]), *)
              (IOMstrs.rdDecls ["D : Any*Any=>Any",
                                "I : Any=>Any",
                                "E : Any"]),
              (IOMstrs.rdEqs ["D(E, ?x:Any) = ?x:Any",
                              "D(I(?x:Any), ?x:Any) = E",
                              "D(D(?x:Any, ?y:Any), ?z:Any) = D(?x:Any, D(?y:Any, ?z:Any))"]),
	      [],
              ([(3, Fun.fromString "I", Order.rdRpoStatus "Lex", [1]),
                (2, Fun.fromString "D", Order.rdRpoStatus "Lex", [1,2]),
                (1, Fun.fromString "E", Order.rdRpoStatus "Lex", [])]),
	      [],
              (0,[]))
        
      | kbTest _ = false ;

    end

    fun kbTest2 1 = 
        time (Kb.kb 
                  (Order.qlposCheck
                       (Order.comparePrec [Atom.atom "I", Atom.atom "D", Atom.atom "E"])
                       (fn _ => (fn xs => xs))
                       Order.GT))
             (IOFotrs.rdEqs
                  ["D(E, ?x) = ?x",
                   "D(I(?x), ?x) = E",
                   "D(D(?x, ?y), ?z) = D(?x, D(?y, ?z))"])

      | kbTest2 _ = false ;


end
