structure Str_Literal =
struct

fun map f [] = []
  | map f (x :: xs) = f x :: map f xs; (* deliberate clone not relying on List._ module *)

fun check_ascii (k : IntInf.int) =
  if 0 <= k andalso k < 128
  then k
  else raise Fail "Non-ASCII character in literal";

val char_of_ascii = Char.chr o IntInf.toInt o check_ascii;

val ascii_of_char = check_ascii o IntInf.fromInt o Char.ord;

val literal_of_asciis = String.implode o map char_of_ascii;

val asciis_of_literal = map ascii_of_char o String.explode;

end;

structure Pat_Complete : sig
  type nat
  type 'a equal
  type num
  datatype ('a, 'b) term = Var of 'b | Fun of 'a * ('a, 'b) term list
  type char
  type ('a, 'b) sum
  val pat_complete_alg :
    'a equal -> 'b equal -> 'c equal ->
      (('a * 'b list) * 'b) list ->
        (('a * 'b list) * 'b) list -> ('a, 'c) term list -> bool
  val create_agcp_input :
    (string -> 'a) ->
      IntInf.int ->
        IntInf.int ->
          (IntInf.int list * IntInf.int list) list ->
            ('a list) list * ('a list) list
  val pat_complete_alg_test :
    IntInf.int -> IntInf.int -> (IntInf.int list * IntInf.int list) list -> bool
  val show_pat_complete_test :
    IntInf.int ->
      IntInf.int -> (IntInf.int list * IntInf.int list) list -> string
  val strong_quasi_reducible_alg :
    'a equal -> 'b equal -> 'c equal ->
      (('a * 'b list) * 'b) list ->
        (('a * 'b list) * 'b) list -> ('a, 'c) term list -> bool
end = struct

datatype nat = Nat of IntInf.int;

fun integer_of_nat (Nat x) = x;

fun equal_nata m n = (((integer_of_nat m) : IntInf.int) = (integer_of_nat n));

type 'a equal = {equal : 'a -> 'a -> bool};
val equal = #equal : 'a equal -> 'a -> 'a -> bool;

val equal_nat = {equal = equal_nata} : nat equal;

fun times_nata m n = Nat (IntInf.* (integer_of_nat m, integer_of_nat n));

type 'a times = {times : 'a -> 'a -> 'a};
val times = #times : 'a times -> 'a -> 'a -> 'a;

type 'a dvd = {times_dvd : 'a times};
val times_dvd = #times_dvd : 'a dvd -> 'a times;

val times_nat = {times = times_nata} : nat times;

val dvd_nat = {times_dvd = times_nat} : nat dvd;

datatype num = One | Bit0 of num | Bit1 of num;

val one_nata : nat = Nat (1 : IntInf.int);

type 'a one = {one : 'a};
val one = #one : 'a one -> 'a;

val one_nat = {one = one_nata} : nat one;

fun plus_nata m n = Nat (IntInf.+ (integer_of_nat m, integer_of_nat n));

type 'a plus = {plus : 'a -> 'a -> 'a};
val plus = #plus : 'a plus -> 'a -> 'a -> 'a;

val plus_nat = {plus = plus_nata} : nat plus;

val zero_nata : nat = Nat (0 : IntInf.int);

type 'a zero = {zero : 'a};
val zero = #zero : 'a zero -> 'a;

val zero_nat = {zero = zero_nata} : nat zero;

type 'a semigroup_add = {plus_semigroup_add : 'a plus};
val plus_semigroup_add = #plus_semigroup_add : 'a semigroup_add -> 'a plus;

type 'a numeral =
  {one_numeral : 'a one, semigroup_add_numeral : 'a semigroup_add};
val one_numeral = #one_numeral : 'a numeral -> 'a one;
val semigroup_add_numeral = #semigroup_add_numeral :
  'a numeral -> 'a semigroup_add;

val semigroup_add_nat = {plus_semigroup_add = plus_nat} : nat semigroup_add;

val numeral_nat =
  {one_numeral = one_nat, semigroup_add_numeral = semigroup_add_nat} :
  nat numeral;

type 'a power = {one_power : 'a one, times_power : 'a times};
val one_power = #one_power : 'a power -> 'a one;
val times_power = #times_power : 'a power -> 'a times;

val power_nat = {one_power = one_nat, times_power = times_nat} : nat power;

type 'a ord = {less_eq : 'a -> 'a -> bool, less : 'a -> 'a -> bool};
val less_eq = #less_eq : 'a ord -> 'a -> 'a -> bool;
val less = #less : 'a ord -> 'a -> 'a -> bool;

fun max A_ a b = (if less_eq A_ a b then b else a);

val ord_integer =
  {less_eq = (fn a => fn b => IntInf.<= (a, b)),
    less = (fn a => fn b => IntInf.< (a, b))}
  : IntInf.int ord;

fun minus_nata m n =
  Nat (max ord_integer (0 : IntInf.int)
        (IntInf.- (integer_of_nat m, integer_of_nat n)));

type 'a minus = {minus : 'a -> 'a -> 'a};
val minus = #minus : 'a minus -> 'a -> 'a -> 'a;

val minus_nat = {minus = minus_nata} : nat minus;

fun apsnd f (x, y) = (x, f y);

fun divmod_integer k l =
  (if ((k : IntInf.int) = (0 : IntInf.int))
    then ((0 : IntInf.int), (0 : IntInf.int))
    else (if IntInf.< ((0 : IntInf.int), l)
           then (if IntInf.< ((0 : IntInf.int), k)
                  then IntInf.divMod (IntInf.abs k, IntInf.abs l)
                  else let
                         val (r, s) =
                           IntInf.divMod (IntInf.abs k, IntInf.abs l);
                       in
                         (if ((s : IntInf.int) = (0 : IntInf.int))
                           then (IntInf.~ r, (0 : IntInf.int))
                           else (IntInf.- (IntInf.~ r, (1 : IntInf.int)),
                                  IntInf.- (l, s)))
                       end)
           else (if ((l : IntInf.int) = (0 : IntInf.int))
                  then ((0 : IntInf.int), k)
                  else apsnd IntInf.~
                         (if IntInf.< (k, (0 : IntInf.int))
                           then IntInf.divMod (IntInf.abs k, IntInf.abs l)
                           else let
                                  val (r, s) =
                                    IntInf.divMod (IntInf.abs k, IntInf.abs l);
                                in
                                  (if ((s : IntInf.int) = (0 : IntInf.int))
                                    then (IntInf.~ r, (0 : IntInf.int))
                                    else (IntInf.- (IntInf.~
              r, (1 : IntInf.int)),
   IntInf.- (IntInf.~ l, s)))
                                end))));

fun fst (x1, x2) = x1;

fun divide_integer k l = fst (divmod_integer k l);

fun divide_nata m n =
  Nat (divide_integer (integer_of_nat m) (integer_of_nat n));

type 'a divide = {divide : 'a -> 'a -> 'a};
val divide = #divide : 'a divide -> 'a -> 'a -> 'a;

val divide_nat = {divide = divide_nata} : nat divide;

fun snd (x1, x2) = x2;

fun modulo_integer k l = snd (divmod_integer k l);

fun modulo_nata m n =
  Nat (modulo_integer (integer_of_nat m) (integer_of_nat n));

type 'a modulo =
  {divide_modulo : 'a divide, dvd_modulo : 'a dvd, modulo : 'a -> 'a -> 'a};
val divide_modulo = #divide_modulo : 'a modulo -> 'a divide;
val dvd_modulo = #dvd_modulo : 'a modulo -> 'a dvd;
val modulo = #modulo : 'a modulo -> 'a -> 'a -> 'a;

val modulo_nat =
  {divide_modulo = divide_nat, dvd_modulo = dvd_nat, modulo = modulo_nata} :
  nat modulo;

fun less_eq_nat m n = IntInf.<= (integer_of_nat m, integer_of_nat n);

fun less_nat m n = IntInf.< (integer_of_nat m, integer_of_nat n);

val ord_nat = {less_eq = less_eq_nat, less = less_nat} : nat ord;

type 'a ab_semigroup_add = {semigroup_add_ab_semigroup_add : 'a semigroup_add};
val semigroup_add_ab_semigroup_add = #semigroup_add_ab_semigroup_add :
  'a ab_semigroup_add -> 'a semigroup_add;

type 'a monoid_add =
  {semigroup_add_monoid_add : 'a semigroup_add, zero_monoid_add : 'a zero};
val semigroup_add_monoid_add = #semigroup_add_monoid_add :
  'a monoid_add -> 'a semigroup_add;
val zero_monoid_add = #zero_monoid_add : 'a monoid_add -> 'a zero;

type 'a comm_monoid_add =
  {ab_semigroup_add_comm_monoid_add : 'a ab_semigroup_add,
    monoid_add_comm_monoid_add : 'a monoid_add};
val ab_semigroup_add_comm_monoid_add = #ab_semigroup_add_comm_monoid_add :
  'a comm_monoid_add -> 'a ab_semigroup_add;
val monoid_add_comm_monoid_add = #monoid_add_comm_monoid_add :
  'a comm_monoid_add -> 'a monoid_add;

type 'a mult_zero = {times_mult_zero : 'a times, zero_mult_zero : 'a zero};
val times_mult_zero = #times_mult_zero : 'a mult_zero -> 'a times;
val zero_mult_zero = #zero_mult_zero : 'a mult_zero -> 'a zero;

type 'a semigroup_mult = {times_semigroup_mult : 'a times};
val times_semigroup_mult = #times_semigroup_mult :
  'a semigroup_mult -> 'a times;

type 'a semiring =
  {ab_semigroup_add_semiring : 'a ab_semigroup_add,
    semigroup_mult_semiring : 'a semigroup_mult};
val ab_semigroup_add_semiring = #ab_semigroup_add_semiring :
  'a semiring -> 'a ab_semigroup_add;
val semigroup_mult_semiring = #semigroup_mult_semiring :
  'a semiring -> 'a semigroup_mult;

type 'a semiring_0 =
  {comm_monoid_add_semiring_0 : 'a comm_monoid_add,
    mult_zero_semiring_0 : 'a mult_zero, semiring_semiring_0 : 'a semiring};
val comm_monoid_add_semiring_0 = #comm_monoid_add_semiring_0 :
  'a semiring_0 -> 'a comm_monoid_add;
val mult_zero_semiring_0 = #mult_zero_semiring_0 :
  'a semiring_0 -> 'a mult_zero;
val semiring_semiring_0 = #semiring_semiring_0 : 'a semiring_0 -> 'a semiring;

type 'a semiring_no_zero_divisors =
  {semiring_0_semiring_no_zero_divisors : 'a semiring_0};
val semiring_0_semiring_no_zero_divisors = #semiring_0_semiring_no_zero_divisors
  : 'a semiring_no_zero_divisors -> 'a semiring_0;

type 'a monoid_mult =
  {semigroup_mult_monoid_mult : 'a semigroup_mult,
    power_monoid_mult : 'a power};
val semigroup_mult_monoid_mult = #semigroup_mult_monoid_mult :
  'a monoid_mult -> 'a semigroup_mult;
val power_monoid_mult = #power_monoid_mult : 'a monoid_mult -> 'a power;

type 'a semiring_numeral =
  {monoid_mult_semiring_numeral : 'a monoid_mult,
    numeral_semiring_numeral : 'a numeral,
    semiring_semiring_numeral : 'a semiring};
val monoid_mult_semiring_numeral = #monoid_mult_semiring_numeral :
  'a semiring_numeral -> 'a monoid_mult;
val numeral_semiring_numeral = #numeral_semiring_numeral :
  'a semiring_numeral -> 'a numeral;
val semiring_semiring_numeral = #semiring_semiring_numeral :
  'a semiring_numeral -> 'a semiring;

type 'a zero_neq_one = {one_zero_neq_one : 'a one, zero_zero_neq_one : 'a zero};
val one_zero_neq_one = #one_zero_neq_one : 'a zero_neq_one -> 'a one;
val zero_zero_neq_one = #zero_zero_neq_one : 'a zero_neq_one -> 'a zero;

type 'a semiring_1 =
  {semiring_numeral_semiring_1 : 'a semiring_numeral,
    semiring_0_semiring_1 : 'a semiring_0,
    zero_neq_one_semiring_1 : 'a zero_neq_one};
val semiring_numeral_semiring_1 = #semiring_numeral_semiring_1 :
  'a semiring_1 -> 'a semiring_numeral;
val semiring_0_semiring_1 = #semiring_0_semiring_1 :
  'a semiring_1 -> 'a semiring_0;
val zero_neq_one_semiring_1 = #zero_neq_one_semiring_1 :
  'a semiring_1 -> 'a zero_neq_one;

type 'a semiring_1_no_zero_divisors =
  {semiring_1_semiring_1_no_zero_divisors : 'a semiring_1,
    semiring_no_zero_divisors_semiring_1_no_zero_divisors :
      'a semiring_no_zero_divisors};
val semiring_1_semiring_1_no_zero_divisors =
  #semiring_1_semiring_1_no_zero_divisors :
  'a semiring_1_no_zero_divisors -> 'a semiring_1;
val semiring_no_zero_divisors_semiring_1_no_zero_divisors =
  #semiring_no_zero_divisors_semiring_1_no_zero_divisors :
  'a semiring_1_no_zero_divisors -> 'a semiring_no_zero_divisors;

type 'a cancel_semigroup_add =
  {semigroup_add_cancel_semigroup_add : 'a semigroup_add};
val semigroup_add_cancel_semigroup_add = #semigroup_add_cancel_semigroup_add :
  'a cancel_semigroup_add -> 'a semigroup_add;

type 'a cancel_ab_semigroup_add =
  {ab_semigroup_add_cancel_ab_semigroup_add : 'a ab_semigroup_add,
    cancel_semigroup_add_cancel_ab_semigroup_add : 'a cancel_semigroup_add,
    minus_cancel_ab_semigroup_add : 'a minus};
val ab_semigroup_add_cancel_ab_semigroup_add =
  #ab_semigroup_add_cancel_ab_semigroup_add :
  'a cancel_ab_semigroup_add -> 'a ab_semigroup_add;
val cancel_semigroup_add_cancel_ab_semigroup_add =
  #cancel_semigroup_add_cancel_ab_semigroup_add :
  'a cancel_ab_semigroup_add -> 'a cancel_semigroup_add;
val minus_cancel_ab_semigroup_add = #minus_cancel_ab_semigroup_add :
  'a cancel_ab_semigroup_add -> 'a minus;

type 'a cancel_comm_monoid_add =
  {cancel_ab_semigroup_add_cancel_comm_monoid_add : 'a cancel_ab_semigroup_add,
    comm_monoid_add_cancel_comm_monoid_add : 'a comm_monoid_add};
val cancel_ab_semigroup_add_cancel_comm_monoid_add =
  #cancel_ab_semigroup_add_cancel_comm_monoid_add :
  'a cancel_comm_monoid_add -> 'a cancel_ab_semigroup_add;
val comm_monoid_add_cancel_comm_monoid_add =
  #comm_monoid_add_cancel_comm_monoid_add :
  'a cancel_comm_monoid_add -> 'a comm_monoid_add;

type 'a semiring_0_cancel =
  {cancel_comm_monoid_add_semiring_0_cancel : 'a cancel_comm_monoid_add,
    semiring_0_semiring_0_cancel : 'a semiring_0};
val cancel_comm_monoid_add_semiring_0_cancel =
  #cancel_comm_monoid_add_semiring_0_cancel :
  'a semiring_0_cancel -> 'a cancel_comm_monoid_add;
val semiring_0_semiring_0_cancel = #semiring_0_semiring_0_cancel :
  'a semiring_0_cancel -> 'a semiring_0;

type 'a ab_semigroup_mult =
  {semigroup_mult_ab_semigroup_mult : 'a semigroup_mult};
val semigroup_mult_ab_semigroup_mult = #semigroup_mult_ab_semigroup_mult :
  'a ab_semigroup_mult -> 'a semigroup_mult;

type 'a comm_semiring =
  {ab_semigroup_mult_comm_semiring : 'a ab_semigroup_mult,
    semiring_comm_semiring : 'a semiring};
val ab_semigroup_mult_comm_semiring = #ab_semigroup_mult_comm_semiring :
  'a comm_semiring -> 'a ab_semigroup_mult;
val semiring_comm_semiring = #semiring_comm_semiring :
  'a comm_semiring -> 'a semiring;

type 'a comm_semiring_0 =
  {comm_semiring_comm_semiring_0 : 'a comm_semiring,
    semiring_0_comm_semiring_0 : 'a semiring_0};
val comm_semiring_comm_semiring_0 = #comm_semiring_comm_semiring_0 :
  'a comm_semiring_0 -> 'a comm_semiring;
val semiring_0_comm_semiring_0 = #semiring_0_comm_semiring_0 :
  'a comm_semiring_0 -> 'a semiring_0;

type 'a comm_semiring_0_cancel =
  {comm_semiring_0_comm_semiring_0_cancel : 'a comm_semiring_0,
    semiring_0_cancel_comm_semiring_0_cancel : 'a semiring_0_cancel};
val comm_semiring_0_comm_semiring_0_cancel =
  #comm_semiring_0_comm_semiring_0_cancel :
  'a comm_semiring_0_cancel -> 'a comm_semiring_0;
val semiring_0_cancel_comm_semiring_0_cancel =
  #semiring_0_cancel_comm_semiring_0_cancel :
  'a comm_semiring_0_cancel -> 'a semiring_0_cancel;

type 'a semiring_1_cancel =
  {semiring_0_cancel_semiring_1_cancel : 'a semiring_0_cancel,
    semiring_1_semiring_1_cancel : 'a semiring_1};
val semiring_0_cancel_semiring_1_cancel = #semiring_0_cancel_semiring_1_cancel :
  'a semiring_1_cancel -> 'a semiring_0_cancel;
val semiring_1_semiring_1_cancel = #semiring_1_semiring_1_cancel :
  'a semiring_1_cancel -> 'a semiring_1;

type 'a comm_monoid_mult =
  {ab_semigroup_mult_comm_monoid_mult : 'a ab_semigroup_mult,
    monoid_mult_comm_monoid_mult : 'a monoid_mult,
    dvd_comm_monoid_mult : 'a dvd};
val ab_semigroup_mult_comm_monoid_mult = #ab_semigroup_mult_comm_monoid_mult :
  'a comm_monoid_mult -> 'a ab_semigroup_mult;
val monoid_mult_comm_monoid_mult = #monoid_mult_comm_monoid_mult :
  'a comm_monoid_mult -> 'a monoid_mult;
val dvd_comm_monoid_mult = #dvd_comm_monoid_mult :
  'a comm_monoid_mult -> 'a dvd;

type 'a comm_semiring_1 =
  {comm_monoid_mult_comm_semiring_1 : 'a comm_monoid_mult,
    comm_semiring_0_comm_semiring_1 : 'a comm_semiring_0,
    semiring_1_comm_semiring_1 : 'a semiring_1};
val comm_monoid_mult_comm_semiring_1 = #comm_monoid_mult_comm_semiring_1 :
  'a comm_semiring_1 -> 'a comm_monoid_mult;
val comm_semiring_0_comm_semiring_1 = #comm_semiring_0_comm_semiring_1 :
  'a comm_semiring_1 -> 'a comm_semiring_0;
val semiring_1_comm_semiring_1 = #semiring_1_comm_semiring_1 :
  'a comm_semiring_1 -> 'a semiring_1;

type 'a comm_semiring_1_cancel =
  {comm_semiring_0_cancel_comm_semiring_1_cancel : 'a comm_semiring_0_cancel,
    comm_semiring_1_comm_semiring_1_cancel : 'a comm_semiring_1,
    semiring_1_cancel_comm_semiring_1_cancel : 'a semiring_1_cancel};
val comm_semiring_0_cancel_comm_semiring_1_cancel =
  #comm_semiring_0_cancel_comm_semiring_1_cancel :
  'a comm_semiring_1_cancel -> 'a comm_semiring_0_cancel;
val comm_semiring_1_comm_semiring_1_cancel =
  #comm_semiring_1_comm_semiring_1_cancel :
  'a comm_semiring_1_cancel -> 'a comm_semiring_1;
val semiring_1_cancel_comm_semiring_1_cancel =
  #semiring_1_cancel_comm_semiring_1_cancel :
  'a comm_semiring_1_cancel -> 'a semiring_1_cancel;

type 'a semidom =
  {comm_semiring_1_cancel_semidom : 'a comm_semiring_1_cancel,
    semiring_1_no_zero_divisors_semidom : 'a semiring_1_no_zero_divisors};
val comm_semiring_1_cancel_semidom = #comm_semiring_1_cancel_semidom :
  'a semidom -> 'a comm_semiring_1_cancel;
val semiring_1_no_zero_divisors_semidom = #semiring_1_no_zero_divisors_semidom :
  'a semidom -> 'a semiring_1_no_zero_divisors;

val ab_semigroup_add_nat = {semigroup_add_ab_semigroup_add = semigroup_add_nat}
  : nat ab_semigroup_add;

val monoid_add_nat =
  {semigroup_add_monoid_add = semigroup_add_nat, zero_monoid_add = zero_nat} :
  nat monoid_add;

val comm_monoid_add_nat =
  {ab_semigroup_add_comm_monoid_add = ab_semigroup_add_nat,
    monoid_add_comm_monoid_add = monoid_add_nat}
  : nat comm_monoid_add;

val mult_zero_nat = {times_mult_zero = times_nat, zero_mult_zero = zero_nat} :
  nat mult_zero;

val semigroup_mult_nat = {times_semigroup_mult = times_nat} :
  nat semigroup_mult;

val semiring_nat =
  {ab_semigroup_add_semiring = ab_semigroup_add_nat,
    semigroup_mult_semiring = semigroup_mult_nat}
  : nat semiring;

val semiring_0_nat =
  {comm_monoid_add_semiring_0 = comm_monoid_add_nat,
    mult_zero_semiring_0 = mult_zero_nat, semiring_semiring_0 = semiring_nat}
  : nat semiring_0;

val semiring_no_zero_divisors_nat =
  {semiring_0_semiring_no_zero_divisors = semiring_0_nat} :
  nat semiring_no_zero_divisors;

val monoid_mult_nat =
  {semigroup_mult_monoid_mult = semigroup_mult_nat,
    power_monoid_mult = power_nat}
  : nat monoid_mult;

val semiring_numeral_nat =
  {monoid_mult_semiring_numeral = monoid_mult_nat,
    numeral_semiring_numeral = numeral_nat,
    semiring_semiring_numeral = semiring_nat}
  : nat semiring_numeral;

val zero_neq_one_nat =
  {one_zero_neq_one = one_nat, zero_zero_neq_one = zero_nat} : nat zero_neq_one;

val semiring_1_nat =
  {semiring_numeral_semiring_1 = semiring_numeral_nat,
    semiring_0_semiring_1 = semiring_0_nat,
    zero_neq_one_semiring_1 = zero_neq_one_nat}
  : nat semiring_1;

val semiring_1_no_zero_divisors_nat =
  {semiring_1_semiring_1_no_zero_divisors = semiring_1_nat,
    semiring_no_zero_divisors_semiring_1_no_zero_divisors =
      semiring_no_zero_divisors_nat}
  : nat semiring_1_no_zero_divisors;

val cancel_semigroup_add_nat =
  {semigroup_add_cancel_semigroup_add = semigroup_add_nat} :
  nat cancel_semigroup_add;

val cancel_ab_semigroup_add_nat =
  {ab_semigroup_add_cancel_ab_semigroup_add = ab_semigroup_add_nat,
    cancel_semigroup_add_cancel_ab_semigroup_add = cancel_semigroup_add_nat,
    minus_cancel_ab_semigroup_add = minus_nat}
  : nat cancel_ab_semigroup_add;

val cancel_comm_monoid_add_nat =
  {cancel_ab_semigroup_add_cancel_comm_monoid_add = cancel_ab_semigroup_add_nat,
    comm_monoid_add_cancel_comm_monoid_add = comm_monoid_add_nat}
  : nat cancel_comm_monoid_add;

val semiring_0_cancel_nat =
  {cancel_comm_monoid_add_semiring_0_cancel = cancel_comm_monoid_add_nat,
    semiring_0_semiring_0_cancel = semiring_0_nat}
  : nat semiring_0_cancel;

val ab_semigroup_mult_nat =
  {semigroup_mult_ab_semigroup_mult = semigroup_mult_nat} :
  nat ab_semigroup_mult;

val comm_semiring_nat =
  {ab_semigroup_mult_comm_semiring = ab_semigroup_mult_nat,
    semiring_comm_semiring = semiring_nat}
  : nat comm_semiring;

val comm_semiring_0_nat =
  {comm_semiring_comm_semiring_0 = comm_semiring_nat,
    semiring_0_comm_semiring_0 = semiring_0_nat}
  : nat comm_semiring_0;

val comm_semiring_0_cancel_nat =
  {comm_semiring_0_comm_semiring_0_cancel = comm_semiring_0_nat,
    semiring_0_cancel_comm_semiring_0_cancel = semiring_0_cancel_nat}
  : nat comm_semiring_0_cancel;

val semiring_1_cancel_nat =
  {semiring_0_cancel_semiring_1_cancel = semiring_0_cancel_nat,
    semiring_1_semiring_1_cancel = semiring_1_nat}
  : nat semiring_1_cancel;

val comm_monoid_mult_nat =
  {ab_semigroup_mult_comm_monoid_mult = ab_semigroup_mult_nat,
    monoid_mult_comm_monoid_mult = monoid_mult_nat,
    dvd_comm_monoid_mult = dvd_nat}
  : nat comm_monoid_mult;

val comm_semiring_1_nat =
  {comm_monoid_mult_comm_semiring_1 = comm_monoid_mult_nat,
    comm_semiring_0_comm_semiring_1 = comm_semiring_0_nat,
    semiring_1_comm_semiring_1 = semiring_1_nat}
  : nat comm_semiring_1;

val comm_semiring_1_cancel_nat =
  {comm_semiring_0_cancel_comm_semiring_1_cancel = comm_semiring_0_cancel_nat,
    comm_semiring_1_comm_semiring_1_cancel = comm_semiring_1_nat,
    semiring_1_cancel_comm_semiring_1_cancel = semiring_1_cancel_nat}
  : nat comm_semiring_1_cancel;

val semidom_nat =
  {comm_semiring_1_cancel_semidom = comm_semiring_1_cancel_nat,
    semiring_1_no_zero_divisors_semidom = semiring_1_no_zero_divisors_nat}
  : nat semidom;

type 'a semiring_no_zero_divisors_cancel =
  {semiring_no_zero_divisors_semiring_no_zero_divisors_cancel :
     'a semiring_no_zero_divisors};
val semiring_no_zero_divisors_semiring_no_zero_divisors_cancel =
  #semiring_no_zero_divisors_semiring_no_zero_divisors_cancel :
  'a semiring_no_zero_divisors_cancel -> 'a semiring_no_zero_divisors;

type 'a semidom_divide =
  {divide_semidom_divide : 'a divide, semidom_semidom_divide : 'a semidom,
    semiring_no_zero_divisors_cancel_semidom_divide :
      'a semiring_no_zero_divisors_cancel};
val divide_semidom_divide = #divide_semidom_divide :
  'a semidom_divide -> 'a divide;
val semidom_semidom_divide = #semidom_semidom_divide :
  'a semidom_divide -> 'a semidom;
val semiring_no_zero_divisors_cancel_semidom_divide =
  #semiring_no_zero_divisors_cancel_semidom_divide :
  'a semidom_divide -> 'a semiring_no_zero_divisors_cancel;

val semiring_no_zero_divisors_cancel_nat =
  {semiring_no_zero_divisors_semiring_no_zero_divisors_cancel =
     semiring_no_zero_divisors_nat}
  : nat semiring_no_zero_divisors_cancel;

val semidom_divide_nat =
  {divide_semidom_divide = divide_nat, semidom_semidom_divide = semidom_nat,
    semiring_no_zero_divisors_cancel_semidom_divide =
      semiring_no_zero_divisors_cancel_nat}
  : nat semidom_divide;

type 'a algebraic_semidom =
  {semidom_divide_algebraic_semidom : 'a semidom_divide};
val semidom_divide_algebraic_semidom = #semidom_divide_algebraic_semidom :
  'a algebraic_semidom -> 'a semidom_divide;

type 'a semiring_modulo =
  {comm_semiring_1_cancel_semiring_modulo : 'a comm_semiring_1_cancel,
    modulo_semiring_modulo : 'a modulo};
val comm_semiring_1_cancel_semiring_modulo =
  #comm_semiring_1_cancel_semiring_modulo :
  'a semiring_modulo -> 'a comm_semiring_1_cancel;
val modulo_semiring_modulo = #modulo_semiring_modulo :
  'a semiring_modulo -> 'a modulo;

type 'a semidom_modulo =
  {algebraic_semidom_semidom_modulo : 'a algebraic_semidom,
    semiring_modulo_semidom_modulo : 'a semiring_modulo};
val algebraic_semidom_semidom_modulo = #algebraic_semidom_semidom_modulo :
  'a semidom_modulo -> 'a algebraic_semidom;
val semiring_modulo_semidom_modulo = #semiring_modulo_semidom_modulo :
  'a semidom_modulo -> 'a semiring_modulo;

val algebraic_semidom_nat =
  {semidom_divide_algebraic_semidom = semidom_divide_nat} :
  nat algebraic_semidom;

val semiring_modulo_nat =
  {comm_semiring_1_cancel_semiring_modulo = comm_semiring_1_cancel_nat,
    modulo_semiring_modulo = modulo_nat}
  : nat semiring_modulo;

val semidom_modulo_nat =
  {algebraic_semidom_semidom_modulo = algebraic_semidom_nat,
    semiring_modulo_semidom_modulo = semiring_modulo_nat}
  : nat semidom_modulo;

fun eq A_ a b = equal A_ a b;

fun equal_lista A_ [] (x21 :: x22) = false
  | equal_lista A_ (x21 :: x22) [] = false
  | equal_lista A_ (x21 :: x22) (y21 :: y22) =
    eq A_ x21 y21 andalso equal_lista A_ x22 y22
  | equal_lista A_ [] [] = true;

fun equal_list A_ = {equal = equal_lista A_} : ('a list) equal;

type 'a showl =
  {showsl : 'a -> string -> string, showsl_list : 'a list -> string -> string};
val showsl = #showsl : 'a showl -> 'a -> string -> string;
val showsl_list = #showsl_list : 'a showl -> 'a list -> string -> string;

fun showsl_lista A_ xs = showsl_list A_ xs;

fun showsl_lit x = (fn a => x ^ a);

fun showsl_sep s sep [] = showsl_lit ""
  | showsl_sep s sep [x] = s x
  | showsl_sep s sep (x :: v :: va) = s x o sep o showsl_sep s sep (v :: va);

fun null [] = true
  | null (x :: xs) = false;

fun showsl_list_gen showslx e l s r xs =
  (if null xs then showsl_lit e
    else showsl_lit l o showsl_sep showslx (showsl_lit s) xs o showsl_lit r);

fun default_showsl_list sl = showsl_list_gen sl "[]" "[" ", " "]";

fun showsl_list_list A_ xs = default_showsl_list (showsl_lista A_) xs;

fun showl_list A_ =
  {showsl = showsl_lista A_, showsl_list = showsl_list_list A_} :
  ('a list) showl;

datatype ('a, 'b) term = Var of 'b | Fun of 'a * ('a, 'b) term list;

fun map f [] = []
  | map f (x21 :: x22) = f x21 :: map f x22;

fun id x = (fn xa => xa) x;

fun showsl_term funa var (Var x) = var x
  | showsl_term funa var (Fun (f, ts)) =
    funa f o showsl_list_gen id "" "(" ", " ")" (map (showsl_term funa var) ts);

fun showsl_terma A_ B_ t = showsl_term (showsl A_) (showsl B_) t;

fun showsl_list_term A_ B_ xs = default_showsl_list (showsl_terma A_ B_) xs;

fun showl_term A_ B_ =
  {showsl = showsl_terma A_ B_, showsl_list = showsl_list_term A_ B_} :
  ('a, 'b) term showl;

datatype char = Chr of IntInf.int;

fun integer_of_char (Chr x) = x;

fun equal_chara c d =
  (((integer_of_char c) : IntInf.int) = (integer_of_char d));

val equal_char = {equal = equal_chara} : char equal;

fun implode cs = Str_Literal.literal_of_asciis (map integer_of_char cs);

fun showsl_list_char cs s = showsl_lit (implode cs) s;

fun showsl_char c = showsl_lit (implode [c]);

val showl_char = {showsl = showsl_char, showsl_list = showsl_list_char} :
  char showl;

fun equal_proda A_ B_ (x1, x2) (y1, y2) = eq A_ x1 y1 andalso eq B_ x2 y2;

fun equal_prod A_ B_ = {equal = equal_proda A_ B_} : ('a * 'b) equal;

datatype 'a set = Set of 'a list | Coset of 'a list;

datatype ('a, 'b) sum = Inl of 'a | Inr of 'b;

fun suc n = plus_nata n one_nata;

fun nth (x :: xs) n =
  (if equal_nata n zero_nata then x else nth xs (minus_nata n one_nata));

fun fold f (x :: xs) s = fold f xs (f x s)
  | fold f [] s = s;

fun rev xs = fold (fn a => fn b => a :: b) xs [];

fun upt i j = (if less_nat i j then i :: upt (suc i) j else []);

fun zip (x :: xs) (y :: ys) = (x, y) :: zip xs ys
  | zip xs [] = []
  | zip [] ys = [];

fun find uu [] = NONE
  | find p (x :: xs) = (if p x then SOME x else find p xs);

fun maps f [] = []
  | maps f (x :: xs) = f x @ maps f xs;

fun foldr f [] = id
  | foldr f (x :: xs) = f x o foldr f xs;

fun map_option f NONE = NONE
  | map_option f (SOME x2) = SOME (f x2);

fun those [] = SOME []
  | those (x :: xs) =
    (case x of NONE => NONE | SOME y => map_option (fn a => y :: a) (those xs));

fun map_of A_ ((l, v) :: ps) k = (if eq A_ l k then SOME v else map_of A_ ps k)
  | map_of A_ [] k = NONE;

fun removeAll A_ x [] = []
  | removeAll A_ x (y :: xs) =
    (if eq A_ x y then removeAll A_ x xs else y :: removeAll A_ x xs);

fun membera A_ [] y = false
  | membera A_ (x :: xs) y = eq A_ x y orelse membera A_ xs y;

fun inserta A_ x xs = (if membera A_ xs x then xs else x :: xs);

fun insert A_ x (Coset xs) = Coset (removeAll A_ x xs)
  | insert A_ x (Set xs) = Set (inserta A_ x xs);

fun member A_ x (Coset xs) = not (membera A_ xs x)
  | member A_ x (Set xs) = membera A_ xs x;

fun remove A_ x (Coset xs) = Coset (inserta A_ x xs)
  | remove A_ x (Set xs) = Set (removeAll A_ x xs);

fun fun_upd A_ f a b = (fn x => (if eq A_ x a then b else f x));

fun subst A_ x t = fun_upd A_ Var x t;

fun concat xss = foldr (fn a => fn b => a @ b) xss [];

fun filter p [] = []
  | filter p (x :: xs) = (if p x then x :: filter p xs else filter p xs);

fun update A_ k v [] = [(k, v)]
  | update A_ k v (p :: ps) =
    (if eq A_ (fst p) k then (k, v) :: ps else p :: update A_ k v ps);

fun hd (x21 :: x22) = x21;

fun tl [] = []
  | tl (x21 :: x22) = x22;

fun list_ex p [] = false
  | list_ex p (x :: xs) = p x orelse list_ex p xs;

fun remdups A_ [] = []
  | remdups A_ (x :: xs) =
    (if membera A_ xs x then remdups A_ xs else x :: remdups A_ xs);

fun distinct A_ [] = true
  | distinct A_ (x :: xs) = not (membera A_ xs x) andalso distinct A_ xs;

fun partition p [] = ([], [])
  | partition p (x :: xs) = let
                              val (yes, no) = partition p xs;
                            in
                              (if p x then (x :: yes, no) else (yes, x :: no))
                            end;

fun replicate n x =
  (if equal_nata n zero_nata then []
    else x :: replicate (minus_nata n one_nata) x);

fun eval_term i (Var x) alpha = alpha x
  | eval_term i (Fun (f, ss)) alpha =
    i f (map (fn s => eval_term i s alpha) ss);

fun gen_length n (x :: xs) = gen_length (suc n) xs
  | gen_length n [] = n;

fun map_filter f [] = []
  | map_filter f (x :: xs) =
    (case f x of NONE => map_filter f xs | SOME y => y :: map_filter f xs);

fun bind m f = (case m of Inl a => Inl a | Inr a => f a);

fun max_list [] = zero_nata
  | max_list (x :: xs) = max ord_nat x (max_list xs);

fun check b e = (if b then Inr () else Inl e);

fun the (SOME x2) = x2;

fun shows_string x = (fn a => x @ a);

fun list_all p [] = true
  | list_all p (x :: xs) = p x andalso list_all p xs;

fun catch_error m f = (case m of Inl a => f a | Inr a => Inr a);

fun forallM f [] = Inr ()
  | forallM f (x :: xs) =
    bind (catch_error (f x) (fn xa => Inl (x, xa))) (fn _ => forallM f xs);

fun dvd (A1_, A2_) a b =
  eq A1_
    (modulo ((modulo_semiring_modulo o semiring_modulo_semidom_modulo) A2_) b a)
    (zero ((zero_mult_zero o mult_zero_semiring_0 o semiring_0_semiring_1 o
             semiring_1_comm_semiring_1 o
             comm_semiring_1_comm_semiring_1_cancel o
             comm_semiring_1_cancel_semidom o semidom_semidom_divide o
             semidom_divide_algebraic_semidom o
             algebraic_semidom_semidom_modulo)
            A2_));

fun map_prod f g (a, b) = (f a, g b);

fun add_vars_term (Var x) xs = x :: xs
  | add_vars_term (Fun (uu, ts)) xs = foldr add_vars_term ts xs;

val bot_set : 'a set = Set [];

fun sup_set A_ (Coset xs) a = Coset (filter (fn x => not (member A_ x a)) xs)
  | sup_set A_ (Set xs) a = fold (insert A_) xs a;

fun vars_term_list t = add_vars_term t [];

val char_0x39 : char = Chr (57 : IntInf.int);

val char_0x38 : char = Chr (56 : IntInf.int);

val char_0x37 : char = Chr (55 : IntInf.int);

val char_0x36 : char = Chr (54 : IntInf.int);

val char_0x35 : char = Chr (53 : IntInf.int);

val char_0x34 : char = Chr (52 : IntInf.int);

val char_0x33 : char = Chr (51 : IntInf.int);

val char_0x32 : char = Chr (50 : IntInf.int);

val char_0x31 : char = Chr (49 : IntInf.int);

val char_0x30 : char = Chr (48 : IntInf.int);

fun nat_of_integer k = Nat (max ord_integer (0 : IntInf.int) k);

fun string_of_digit n =
  (if equal_nata n zero_nata then [char_0x30]
    else (if equal_nata n one_nata then [char_0x31]
           else (if equal_nata n (nat_of_integer (2 : IntInf.int))
                  then [char_0x32]
                  else (if equal_nata n (nat_of_integer (3 : IntInf.int))
                         then [char_0x33]
                         else (if equal_nata n (nat_of_integer (4 : IntInf.int))
                                then [char_0x34]
                                else (if equal_nata n
   (nat_of_integer (5 : IntInf.int))
                                       then [char_0x35]
                                       else (if equal_nata n
          (nat_of_integer (6 : IntInf.int))
      then [char_0x36]
      else (if equal_nata n (nat_of_integer (7 : IntInf.int)) then [char_0x37]
             else (if equal_nata n (nat_of_integer (8 : IntInf.int))
                    then [char_0x38] else [char_0x39])))))))));

fun showsp_nat p n =
  (if less_nat n (nat_of_integer (10 : IntInf.int))
    then shows_string (string_of_digit n)
    else showsp_nat p (divide_nata n (nat_of_integer (10 : IntInf.int))) o
           shows_string
             (string_of_digit
               (modulo_nata n (nat_of_integer (10 : IntInf.int)))));

fun showsl_lines A_ desc_empty =
  showsl_list_gen (showsl A_) desc_empty "" "\n" "";

val char_0x3A : char = Chr (58 : IntInf.int);

val char_0x3F : char = Chr (63 : IntInf.int);

val char_0x42 : char = Chr (66 : IntInf.int);

val char_0x62 : char = Chr (98 : IntInf.int);

val char_0x66 : char = Chr (102 : IntInf.int);

val char_0x67 : char = Chr (103 : IntInf.int);

val char_0x74 : char = Chr (116 : IntInf.int);

val char_0x78 : char = Chr (120 : IntInf.int);

val char_0x79 : char = Chr (121 : IntInf.int);

fun minus_set A_ a (Coset xs) = Set (filter (fn x => member A_ x a) xs)
  | minus_set A_ a (Set xs) = fold (remove A_) xs a;

fun less_eq_set A_ (Coset []) (Set []) = false
  | less_eq_set A_ a (Coset ys) = list_all (fn y => not (member A_ y a)) ys
  | less_eq_set A_ (Set xs) b = list_all (fn x => member A_ x b) xs;

fun size_list x = gen_length zero_nata x;

fun shows_prec_nat x = showsp_nat x;

fun testproblem c n =
  let
    val _ = implode;
    val s = id;
    val c1 =
      dvd (equal_nat, semidom_modulo_nat) (nat_of_integer (2 : IntInf.int)) c;
    val c2 =
      dvd (equal_nat, semidom_modulo_nat) (nat_of_integer (2 : IntInf.int))
        (divide_nata c (nat_of_integer (2 : IntInf.int)));
    val c3 =
      dvd (equal_nat, semidom_modulo_nat) (nat_of_integer (2 : IntInf.int))
        (divide_nata c (nat_of_integer (4 : IntInf.int)));
    val c4 =
      dvd (equal_nat, semidom_modulo_nat) (nat_of_integer (2 : IntInf.int))
        (divide_nata c (nat_of_integer (8 : IntInf.int)));
    val revo = (if c4 then id else rev);
    val nn = upt zero_nata n;
    val rnn = (if c4 then id nn else rev nn);
    val b = s [char_0x62];
    val t = s [char_0x74, char_0x74];
    val f = s [char_0x66, char_0x66];
    val g = s [char_0x67];
    val gg = (fn ts => Fun (g, revo ts));
    val ff = Fun (f, []);
    val tt = Fun (t, []);
    val ca = [((t, []), b), ((f, []), b)];
    val d =
      [((g, replicate (times_nata (nat_of_integer (2 : IntInf.int)) n) b), b)];
    val x = (fn i => Var (s ([char_0x78] @ shows_prec_nat zero_nata i [])));
    val y = (fn i => Var (s ([char_0x79] @ shows_prec_nat zero_nata i [])));
    val lhsF =
      gg (if c1 then maps (fn i => [ff, y i]) rnn
           else replicate n ff @ map y rnn);
    val lhsT =
      (fn ba => fn j =>
        gg (if c1
             then maps (fn i =>
                         (if equal_nata i j then [tt, ba] else [x i, y i]))
                    rnn
             else map (fn i => (if equal_nata i j then tt else x i)) rnn @
                    map (fn i => (if equal_nata i j then ba else y i)) rnn));
    val lhssT =
      (if c2 then maps (fn i => [lhsT tt i, lhsT ff i]) nn
        else maps (fn ba => map (lhsT ba) nn) [tt, ff]);
    val lhss = (if c3 then [lhsF] @ lhssT else lhssT @ [lhsF]);
  in
    (ca, (d, lhss))
  end;

fun test_problem c n perms =
  (if less_nat c (nat_of_integer (16 : IntInf.int)) then testproblem c n
    else let
           val (ca, (d, lhss)) = testproblem zero_nata n;
           val (permRow, permCol) =
             nth perms (minus_nata c (nat_of_integer (16 : IntInf.int)));
           val permRows = map (nth lhss) permRow;
           val pCol = (fn Fun (g, ts) => Fun (g, map (nth ts) permCol));
         in
           (ca, (d, map pCol permRows))
         end);

fun tau_c A_ n x c =
  let
    val (f, ss) = c;
  in
    subst (equal_prod equal_nat A_) x
      (Fun (f, map Var (zip (upt n (plus_nata n (size_list ss))) ss)))
  end;

fun term_and_args f ts = Fun (f, ts) :: ts;

fun empty_lr mp = let
                    val (lx, (rx, _)) = mp;
                  in
                    null lx andalso null rx
                  end;

fun conflicts A_ B_ (Var x) (Var y) =
  (if eq B_ x y then SOME [] else SOME [x, y])
  | conflicts A_ B_ (Var x) (Fun (uu, uv)) = SOME [x]
  | conflicts A_ B_ (Fun (uw, ux)) (Var x) = SOME [x]
  | conflicts A_ B_ (Fun (f, ss)) (Fun (g, ts)) =
    (if equal_proda A_ equal_nat (f, size_list ss) (g, size_list ts)
      then map_option concat
             (those (map (fn (a, b) => conflicts A_ B_ a b) (zip ss ts)))
      else NONE);

fun find_var A_ B_ p =
  (case maps (fn (lx, _) => lx) p
    of [] => let
               val (_, (rx, _)) = hd (filter (not o snd o snd) p);
               val (_, s :: t :: _) = hd rx;
             in
               hd (the (conflicts B_ (equal_prod equal_nat A_) s t))
             end
    | (x, _) :: _ => x);

fun compute_nonempty_main A_ ne ls =
  let
    val rem_ls = filter (fn f => not (member A_ (snd f) ne)) ls;
    val (new, rem) =
      partition (fn (a, b) => let
                                val (_, args) = a;
                              in
                                (fn _ => less_eq_set A_ (Set args) ne)
                              end
                                b)
        rem_ls;
  in
    (if null new then ne
      else compute_nonempty_main A_ (sup_set A_ ne (Set (map snd new))) rem)
  end;

fun compute_nonempty_sorts B_ cs = compute_nonempty_main B_ bot_set cs;

fun decide_nonempty_sorts A_ tau_s cs =
  let
    val ne = compute_nonempty_sorts A_ cs;
  in
    find (fn tau => not (member A_ tau ne)) tau_s
  end;

fun tau_s_list A_ cl n x = map (tau_c A_ n x) (cl (snd x));

fun insert_rx A_ B_ C_ inf_sort t x rxb =
  let
    val (rx, b) = rxb;
  in
    (case map_of C_ rx x of NONE => SOME ((x, [t]) :: rx, b)
      | SOME ts =>
        (case those (map (conflicts B_ (equal_prod equal_nat A_) t) ts)
          of NONE => NONE
          | SOME cs =>
            (if membera (equal_list (equal_prod equal_nat A_)) cs []
              then SOME rxb
              else SOME (update C_ x (t :: ts) rx,
                          b orelse
                            list_ex (fn y => inf_sort (snd y)) (concat cs)))))
  end;

fun decomp_impl A_ B_ C_ inf_sort ((t, Var y) :: mp) =
  (case decomp_impl A_ B_ C_ inf_sort mp of NONE => NONE
    | SOME (lx, rx) =>
      (case insert_rx A_ B_ C_ inf_sort t y rx of NONE => NONE
        | SOME rxa => SOME (lx, rxa)))
  | decomp_impl A_ B_ C_ inf_sort ((Var x, Fun (g, ls)) :: mp) =
    (case decomp_impl A_ B_ C_ inf_sort mp of NONE => NONE
      | SOME (lx, rx) => SOME ((x, Fun (g, ls)) :: lx, rx))
  | decomp_impl A_ B_ C_ inf_sort ((Fun (f, ts), Fun (g, ls)) :: mp) =
    (if equal_proda B_ equal_nat (f, size_list ts) (g, size_list ls)
      then decomp_impl A_ B_ C_ inf_sort (zip ts ls @ mp) else NONE)
  | decomp_impl A_ B_ C_ inf_sort [] = SOME ([], ([], false));

fun match_var_impl C_ mp =
  let
    val (xl, (rx, b)) = mp;
    val xs = remdups C_ (maps (vars_term_list o snd) xl);
  in
    (xl, (filter (fn (x, ts) => not (null (tl ts)) orelse membera C_ xs x) rx,
           b))
  end;

fun match_steps_impl A_ B_ C_ inf_sort mp =
  map_option (match_var_impl C_) (decomp_impl A_ B_ C_ inf_sort mp);

fun pat_inner_impl A_ B_ C_ inf_sort (mp :: p) pd =
  (case match_steps_impl A_ B_ C_ inf_sort mp
    of NONE => pat_inner_impl A_ B_ C_ inf_sort p pd
    | SOME mpa =>
      (if empty_lr mpa then NONE
        else pat_inner_impl A_ B_ C_ inf_sort p (mpa :: pd)))
  | pat_inner_impl A_ B_ C_ inf_sort [] pd = SOME pd;

fun subst_left tau =
  map_prod (fn t => eval_term (fn a => fn b => Fun (a, b)) t tau) id;

fun subst_match_problem_list tau = map (subst_left tau);

fun subst_pat_problem_list tau = map (subst_match_problem_list tau);

fun mp_rx_list mp = maps (fn (x, a) => map (fn t => (t, Var x)) a) (fst mp);

fun mp_lr_list pair = let
                        val (lx, rx) = pair;
                      in
                        map (map_prod Var id) lx @ mp_rx_list rx
                      end;

fun pat_impl A_ B_ C_ cl inf_sort n p =
  (case pat_inner_impl A_ B_ C_ inf_sort p [] of NONE => SOME []
    | SOME pa =>
      (if list_all (fn mp => snd (snd mp)) pa then NONE
        else let
               val p_l = map mp_lr_list pa;
               val x = find_var A_ B_ pa;
             in
               SOME (map (fn tau => subst_pat_problem_list tau p_l)
                      (tau_s_list A_ cl n x))
             end));

fun pats_impl A_ B_ C_ m cl inf_sort n ps =
  (case ps of [] => true
    | p :: ps1 =>
      (case pat_impl A_ B_ C_ cl inf_sort n p of NONE => false
        | SOME ps2 =>
          pats_impl A_ B_ C_ m cl inf_sort (plus_nata n m) (ps2 @ ps1)));

fun pat_complete_impl A_ B_ C_ m cl inf_sort ps =
  let
    val n =
      suc (max_list
            (maps (map fst o vars_term_list o fst) (concat (concat ps))));
  in
    pats_impl A_ B_ C_ m cl inf_sort n ps
  end;

fun compute_inf_main A_ m_inf ls =
  let
    val (fin, lsa) =
      partition
        (fn (_, fs) =>
          list_all (list_all (fn tau => not (member A_ tau m_inf)))
            (map snd fs))
        ls;
  in
    (if null fin then m_inf
      else compute_inf_main A_ (minus_set A_ m_inf (Set (map fst fin))) lsa)
  end;

fun compute_inf_sorts B_ cs =
  let
    val csa =
      map (fn tau =>
            (tau, map_filter
                    (fn x => (if eq B_ (snd x) tau then SOME (fst x) else NONE))
                    cs))
        (remdups B_ (map snd cs));
  in
    compute_inf_main B_ (Set (map fst csa)) csa
  end;

fun sorts_of_ssig_list B_ cs =
  remdups B_ (maps (fn (a, b) => let
                                   val (_, ss) = a;
                                 in
                                   (fn s => s :: ss)
                                 end
                                   b)
               cs);

fun decide_pat_complete A_ B_ C_ cs p =
  let
    val _ = sorts_of_ssig_list B_ cs;
    val m = max_list (map (size_list o snd o fst) cs);
    val cl =
      (fn s =>
        map_filter (fn x => (if (eq B_ s o snd) x then SOME (fst x) else NONE))
          cs);
    val is = compute_inf_sorts B_ cs;
  in
    pat_complete_impl B_ A_ C_ m cl (fn s => member B_ s is)
  end
    p;

fun decide_pat_complete_lhss A_ B_ C_ c d lhss =
  bind (check (distinct (equal_prod A_ (equal_list B_)) (map fst c))
         (showsl_lit "constructor information contains duplicate"))
    (fn _ =>
      bind (check (distinct (equal_prod A_ (equal_list B_)) (map fst d))
             (showsl_lit "defined symbol information contains duplicate"))
        (fn _ =>
          let
            val s = sorts_of_ssig_list B_ c;
          in
            bind (catch_error
                   (forallM
                     (fn (a, b) =>
                       let
                         val (_, ss) = a;
                       in
                         (fn _ =>
                           catch_error
                             (forallM
                               (fn sa =>
                                 check (membera B_ s sa)
                                   (showsl_lit
                                     "a defined symbol has argument sort that is not known in constructors"))
                               ss)
                             (fn x => Inl (snd x)))
                       end
                         b)
                     d)
                   (fn x => Inl (snd x)))
              (fn _ =>
                bind (case decide_nonempty_sorts B_ s c of NONE => Inr ()
                       | SOME _ => Inl (showsl_lit "some sort is empty"))
                  (fn _ =>
                    let
                      val pats =
                        map (fn (a, b) =>
                              let
                                val (f, ss) = a;
                              in
                                (fn _ =>
                                  Fun (f,
map Var (zip (upt zero_nata (size_list ss)) ss)))
                              end
                                b)
                          d;
                      val p =
                        map (fn pat => map (fn lhs => [(pat, lhs)]) lhss) pats;
                    in
                      Inr (decide_pat_complete A_ B_ C_ c p)
                    end))
          end));

fun pat_complete_alg A_ B_ C_ c d lhss =
  (case decide_pat_complete_lhss A_ B_ C_ c d lhss
    of Inl err => (raise Fail (err "")) (fn _ => true) | Inr res => res);

fun test_problem_integer c n perms =
  test_problem (nat_of_integer c) (nat_of_integer n)
    (map (map_prod (map nat_of_integer) (map nat_of_integer)) perms);

fun create_agcp_input term c n perms =
  let
    val na = nat_of_integer n;
    val _ = nat_of_integer c;
    val lhss = (snd o snd) (test_problem_integer c n perms);
    val tt =
      (fn a =>
        (case a
          of Var x => term (implode ([char_0x3F] @ x @ [char_0x3A, char_0x42]))
          | Fun (f, []) => term (implode f)));
    val pslist =
      map (fn i => tt (Var ([char_0x78] @ shows_prec_nat zero_nata i [])))
        (upt zero_nata (times_nata (nat_of_integer (2 : IntInf.int)) na));
    val a = map (fn Fun (_, a) => map tt a) lhss;
  in
    ([pslist], a)
  end;

fun pat_complete_alg_test c n perms =
  let
    val (ca, (a, b)) = test_problem_integer c n perms;
  in
    pat_complete_alg (equal_list equal_char) (equal_list equal_char)
      (equal_list equal_char) ca a b
  end;

fun show_pat_complete_test c n perms =
  let
    val (_, (_, lhss)) = test_problem_integer c n perms;
  in
    showsl_lines (showl_term (showl_list showl_char) (showl_list showl_char))
      "empty" lhss ""
  end;

fun decide_strong_quasi_reducible A_ B_ C_ c d lhss =
  bind (check (distinct (equal_prod A_ (equal_list B_)) (map fst c))
         (showsl_lit "constructor information contains duplicate"))
    (fn _ =>
      bind (check (distinct (equal_prod A_ (equal_list B_)) (map fst d))
             (showsl_lit "defined symbol information contains duplicate"))
        (fn _ =>
          let
            val s = sorts_of_ssig_list B_ c;
          in
            bind (catch_error
                   (forallM
                     (fn (a, b) =>
                       let
                         val (_, ss) = a;
                       in
                         (fn _ =>
                           catch_error
                             (forallM
                               (fn sa =>
                                 check (membera B_ s sa)
                                   (showsl_lit
                                     "defined symbol f has argument sort s that is not known in constructors"))
                               ss)
                             (fn x => Inl (snd x)))
                       end
                         b)
                     d)
                   (fn x => Inl (snd x)))
              (fn _ =>
                bind (case decide_nonempty_sorts B_ s c of NONE => Inr ()
                       | SOME _ => Inl (showsl_lit "sort s is empty"))
                  (fn _ =>
                    let
                      val pats =
                        map (fn (a, b) =>
                              let
                                val (f, ss) = a;
                              in
                                (fn _ =>
                                  term_and_args f
                                    (map Var
                                      (zip (upt zero_nata (size_list ss)) ss)))
                              end
                                b)
                          d;
                      val p =
                        map (maps (fn pat => map (fn lhs => [(pat, lhs)]) lhss))
                          pats;
                    in
                      Inr (decide_pat_complete A_ B_ C_ c p)
                    end))
          end));

fun strong_quasi_reducible_alg A_ B_ C_ c d lhss =
  (case decide_strong_quasi_reducible A_ B_ C_ c d lhss
    of Inl err => (raise Fail (err "")) (fn _ => true) | Inr res => res);

end; (*struct Pat_Complete*)
