(******************************************************************************
 * Copyright (c) 2012-2015, Toyama&Aoto Laboratory, Tohoku University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the 
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(******************************************************************************
 * file: rwtools/rwchecker/cr_solver.sml
 * description: confluence checker
 * author: YOSHIDA Junichi
 * author: AOTO Takahito
 * 
 ******************************************************************************)

signature CR_SOLVER = 
sig
    val runProfile: bool ref
    val runDebug: bool ref
    val runCertification: bool ref

    type options = {
         useDirect:bool,   (*  $BD>OBJ,2r$r;H$&$+(B *)
         useCommutative:bool,   (*  $B2D49J,2r$r;H$&$+(B *)
         useNonMinimalCommutative:bool,   (* $BHs6K>.2D49J,2r$r;H$&$+(B *)
         useLayer:bool,    (*  $BAXJ]B8J,2r$r;H$&$+(B *)
         useCertifiableOutputOption:bool    (*  $B8!>Z=PNO$r=P$9$+(B *)
    }

    val crSolver: Solver.options
                  -> DpSolver.options
                  -> PoSolver.options
                  -> options
                  -> (Term.term * Term.term) list 
                  ->  Cr.ConfluenceResult

    val useHuetToyamaOostrom: bool ref;
    val useHuetStrongClosed: bool ref;
    val useOyamaguchiOhta: bool ref;
    val useToyamaParallel: bool ref;
    val useOkuiSimultaneous: bool ref;
    val useDecreasingDiagrams: bool ref;
    val useOhtaOyamaguchiToyama: bool ref;
    val useGomiOyamaguchiOhtaDepth: bool ref;
    val useGomiOyamaguchiOhtaWeight: bool ref;
    val useSuzukiAotoToyama: bool ref;
    val useUchidaAotoToyama: bool ref;
    val useUchidaAotoToyamaStrong: bool ref;
    val useSakaiOgawa: bool ref;
    val useKnuthBendix: bool ref;
    val useHuetModulo: bool ref;
    val useBidirectionalCompletion: bool ref;
    val proveGroundConfluence: bool ref;

end

structure CrSolver : CR_SOLVER =
struct
    local
	open Prop
	open Cr
	structure A = Atom
	structure CU = CertifyUtil
	structure L = List
	structure LU = ListUtil
	structure FS = FunSet
	structure FIS = FunIntSet
	structure ILM = IntListMap2
	structure TS = TermSet
	structure VS = VarSet
	structure YI = YicesInput
	open PrintUtil
    in
    
    exception CrSolverError
          
    val runProfile = ref false : bool ref
    val runDebug = ref false : bool ref
    val runCertification = ref false : bool ref

    fun debug f = if !runDebug then f () else ()
    fun cpout f = if !runCertification then f () else ()

    type options = {
         useDirect:bool,   (*  $BD>OBJ,2r$r;H$&$+(B *)
         useCommutative:bool,   (*  $B2D49J,2r$r;H$&$+(B *)
         useNonMinimalCommutative:bool,   (* $BHs6K>.2D49J,2r$r;H$&$+(B *)
         useLayer:bool,    (*  $BAXJ]B8J,2r$r;H$&$+(B *)
         useCertifiableOutputOption:bool    (*  $B8!>Z=PNO$r=P$9$+(B *)
    }

    val useHuetToyamaOostrom = ref true
    val useHuetStrongClosed = ref true
    val useOyamaguchiOhta = ref true
    val useToyamaParallel = ref true
    val useOkuiSimultaneous = ref true
    val useDecreasingDiagrams = ref true
    val useOhtaOyamaguchiToyama = ref true
    val useGomiOyamaguchiOhtaDepth = ref true
    val useGomiOyamaguchiOhtaWeight = ref true
    val useSuzukiAotoToyama = ref true
    val useUchidaAotoToyama = ref true
    val useUchidaAotoToyamaStrong = ref true
    val useSakaiOgawa = ref true
    val useKnuthBendix = ref true
    val useHuetModulo = ref false (* included in BidirectionalCompletion *)
    val useBidirectionalCompletion = ref true
    val proveGroundConfluence = ref false

    val nameHuetToyamaOostrom = "Development Closed"
    val nameHuetStrongClosed = "Strongly Closed"
    val nameOyamaguchiOhta = "Upside-Parallel-Closed/Outside-Closed"
    val nameToyamaParallel = "Toyama (Parallel CPs)"
    val nameOkuiSimultaneous = "Okui (Simultaneous CPs)"
    val nameHuetModulo = "Huet (modulo AC)"
    val nameDecreasingDiagrams = "Diagram Decreasing"
    val nameOhtaOyamaguchiToyama = "Simple-Right-Linear"
    val nameGomiOyamaguchiOhtaDepth = "Strongly Depth-Preserving & Root-E-Closed/Non-E-Overlapping"
    val nameGomiOyamaguchiOhtaWeight = "Strongly Weight-Preserving & Root-E-Closed/Non-E-Overlapping"
    val nameSuzukiAotoToyama = "Quasi-Left-Linear & Parallel-Closed"
    val nameUchidaAotoToyama = "Quasi-Linear & Linearized-Decreasing"
    val nameUchidaAotoToyamaStrong = "Strongly Quasi-Linear & Hierarchically Decreasing"
    val nameSakaiOgawa = "Weakly-Non-Overlapping & Non-Collapsing & Shallow"
    val nameNonConfluence = "Non-Confluence"
    val nameKnuthBendix = "Knuth & Bendix"

    fun unknown s = "unknown " ^ s

    local 
	open Term
	open Cr
	open PrintUtil
  (* $B9gN.@-$N4pK\H=Dj(B *)
    in
   fun checkConfluenceConditions2 satSolver smtSolver tmpDir 
				  isTerminating isRelativeTerminating rs = 
       let 
	   val _ = print "Apply Direct Methods...\n"


	   val inCpeaksWithIndexAndPosFs = Cr.insideCriticalPeaksWithIndexAndPosFs rs

	   val inCpsPos = L.map (fn (mn,(ps,fs,x,y,z)) => (ps,y,z)) inCpeaksWithIndexAndPosFs
	   val inCpeaks = L.map (fn (mn,(ps,fs,x,y,z)) => (x,y,z)) inCpeaksWithIndexAndPosFs
	   val outCpeaks = outsideCriticalPeaksInOneside rs

	   val inCps = L.map (fn (_,y,z) => (y,z)) inCpeaks
	   val outCps = L.map (fn (_,y,z) => (y,z)) outCpeaks

	   val _ = print ("Inner CPs:\n" ^ (LU.toStringCommaLnSquare (IOFotrs.prEq) inCps))
	   val _ = print ("Outer CPs:\n" ^ (LU.toStringCommaLnSquare (IOFotrs.prEq) outCps))

	   val (_,dSymSet,cSymSet,_) = Trs.fdcSetInRules rs

	   val maxRewriteLenForNonConfluenceCandidates = ref 2

	   val developSetTable = ref TermMap.empty

	   fun insertDevelopSet term = 
	       let val tset = Rewrite.developOneStepReductSet rs term
		   val _ = developSetTable := TermMap.insert (!developSetTable,term,tset)
	       in tset
	       end

	   fun lookupDevelopSet term =
	       case TermMap.find (!developSetTable,term) of
		   SOME tset => tset
		 | NONE => insertDevelopSet term

	   val reductSetTable = ref TermMap.empty

	   fun insertReductSet term = 
	       let val maxLen = !maxRewriteLenForNonConfluenceCandidates
		   val tset = Rewrite.manyStepsReductSet rs maxLen term
		   val _ = reductSetTable := TermMap.insert (!reductSetTable,term,tset)
	       in tset
	       end

	   fun lookupReductSet term =
	       case TermMap.find (!reductSetTable,term) of
		   SOME tset => tset
		 | NONE => insertReductSet term

	   fun checkHuetToyamaOostrom _ = 
	       (!useHuetToyamaOostrom)
	       andalso
	       let fun checkIn (l,r) = Term.equal (l,r) 
				       orelse let val dl = lookupDevelopSet l
						  val _ = debug (fn _ => println ("check inner CP " ^ (Trs.prEq (l,r))))
						  val _ = debug (fn _ => println ("{ s | l -o-> s } = " ^
								 (prSetInOneLine Term.toString (TS.listItems dl))))
					      in TS.member (dl,r)
					      end
						  
		   fun checkOut (l,r) = 
		       Term.equal (l,r) 
		       orelse 
		       let val l1 = lookupDevelopSet l
			   val r1 = lookupDevelopSet r
			   val _ = debug (fn _ => println ("check outer CP " ^ (Trs.prEq (l,r))))
			   val _ = debug (fn _ => println ("X1 = { s | l -o-> s } = " ^ 
						  (prSetInOneLine Term.toString (TS.listItems l1))))
			   val _ = debug (fn _ => println ("Y1 = { t | r -o-> t } = " ^
						  (prSetInOneLine Term.toString (TS.listItems r1))))
			   val _ = debug (fn _ => println ("check X1 cap Y1 /= {} "))
		       in not (TS.isEmpty (TS.intersection (l1,r1)))
			  orelse let val l2 = lookupReductSet l
				     val r2 = lookupReductSet r
				     val _ = debug (fn _ => println ("X2 = { s | l -n-> s } = " ^
						    (prSetInOneLine Term.toString (TS.listItems l2))))
				     val _ = debug (fn _ => println ("Y2 = { s | r -n-> s } = " ^
								   (prSetInOneLine Term.toString (TS.listItems r2))))
				     val _ = debug (fn _ => println ("check X1 cap Y2 /= {} and X2 cap Y1 /= {}"))
				 in not (TS.isEmpty (TS.intersection (l1,r2)))
				    andalso not (TS.isEmpty (TS.intersection (l2,r1)))
				 end
		       end
	       in
		   if L.all checkIn inCps
		      andalso L.all checkOut outCps
		   then (println nameHuetToyamaOostrom; true)
		   else (println (unknown nameHuetToyamaOostrom); false)
	       end

	   fun checkHuetStrongClosed _ = 
	       (!useHuetStrongClosed)
	       andalso
	       let val certificate = ref 0
		   fun strongClosed (l,r) = 
		       Term.equal (l,r) 
		       orelse 		    
		       let 
			   val _ = debug (fn _ => println ("check CP " ^ (Trs.prEq (l,r))))
			   val r1 = lookupDevelopSet r
			   val l1 = TS.add (Rewrite.oneStepReductSet rs l, l)
			   val _ = debug (fn _ => println ("X = { s | l -=-> s } = " ^
							   (prSetInOneLine Term.toString (TS.listItems l1))))
			   val _ = debug (fn _ => println ("Y1 = { s | r -o-> s } = " ^
							   (prSetInOneLine Term.toString (TS.listItems r1))))
			   val _ = debug (fn _ => println ("check X cap Y1 /= {} "))
		       in not (TS.isEmpty (TS.intersection (r1,l1)))
			      andalso (if !runCertification 
				       then certificate := Int.max(!certificate, TS.numItems r1)
				       else (); true)
			  orelse
			  let val r2 = lookupReductSet r
			   val _ = debug (fn _ => println ("Y2 = { s | r -n-> s } = " ^
							   (prSetInOneLine Term.toString (TS.listItems r2))))
			   val _ = debug (fn _ => println ("check X cap Y2 /= {} "))
			  in not (TS.isEmpty (TS.intersection (r2,l1)))
			     andalso (if !runCertification 
				      then certificate := Int.max(!certificate, TS.numItems r2)
				      else (); true)
			  end
		       end
	       in
		   if L.all strongClosed outCps
		      andalso L.all strongClosed (L.map (fn (x,y) => (y,x)) outCps)
		      andalso L.all strongClosed inCps
		      andalso L.all strongClosed (L.map (fn (x,y) => (y,x)) inCps)
		   then (println nameHuetStrongClosed; 
			 cpout (fn _ => CU.output (CU.beginProof (CU.beginCrProof
                               (fn _ => CU.encloseProofLeafBy "stronglyClosed" (Int.toString (!certificate))))));
			 true)
		   else (println (unknown nameHuetStrongClosed);false)
	       end


	   fun checkOyamaguchiOhta _ = 
	       (!useOyamaguchiOhta)
	       andalso
	       let val depthRestrectedClosedCps = isDepthRestrictedClosedCps rs inCpsPos outCps
		   val _ = if depthRestrectedClosedCps
			   then println nameOyamaguchiOhta
			   else println ("unknown " ^ nameOyamaguchiOhta)
	       in depthRestrectedClosedCps
	       end


	   fun checkToyamaParallel _ = 
	       (!useToyamaParallel)
	       andalso
	       let val parallelCriticalPairClosed = 
		       isParallelCriticalPairClosed rs (inCps @ outCps) 
		   val _ = if parallelCriticalPairClosed
			   then println nameToyamaParallel
			   else println (unknown nameToyamaParallel)
	       in parallelCriticalPairClosed
	       end

	   fun checkOkuiSimultaneous _ = 
	       (!useOkuiSimultaneous)
	       andalso
	       let val okuiClosedCps = isOkuiClosedCps rs
		   val _ = if okuiClosedCps
			   then println nameOkuiSimultaneous
			   else println (unknown nameOkuiSimultaneous)
	       in okuiClosedCps
	       end

	   fun checkHuetModulo _ = 
	       (!useHuetModulo)
	       andalso
	       if Cr.isHuetModulo isRelativeTerminating rs
	       then (println nameHuetModulo; true)
	       else (println (unknown nameHuetModulo);false)
	       
	   fun checkDecreasingDiagrams _ = 
	       (!useDecreasingDiagrams)
	       andalso
	       let val _ = print "check Locally Decreasing Diagrams by Rule Labelling...\n"
(*		   val decreasingCps = CrDiagram2.hasDecreasingCps0 smtSolver tmpDir rs *)
(*		   val decreasingCps = CrDiagram2.hasDecreasingCps smtSolver tmpDir rs  *)
		   val decreasingCps = CrDiagram.hasDecreasingCps smtSolver tmpDir rs  
		   val _ = if decreasingCps
			   then println nameDecreasingDiagrams
			   else println (unknown nameDecreasingDiagrams)
	       in decreasingCps
	       end

	   fun checkOhtaOyamaguchiToyama _ =
	       (!useOhtaOyamaguchiToyama)
	       andalso
	       let fun checkVar (l,r) = 
		       VS.isEmpty
			   (VS.intersection (		      
			    VS.difference (Term.varSetInTerm l, Term.linearVarSetInTerm l),
			    Term.varSetInTerm r))
		   val _ = debug (fn _ => println "check all x in V(r) is linear in l")
	       in if L.all checkVar rs
		     andalso 
		     let val cp = Cr.criticalPairs (L.map (fn (l,r) => (Subst.linearize l,r)) rs)
			 val _ = debug (fn _ => print ("CPs after linearization of lhs:\n"))
			 val _ = debug (fn _ => print (LU.toStringCommaLnSquare (IOFotrs.prEq) cp))
		     in  null cp
		     end
		  then (println nameOhtaOyamaguchiToyama; true)
		  else (println (unknown nameOhtaOyamaguchiToyama); false)
	       end

	   fun transformationForWightPreserving smtSolver rs =
	       let val faSet = Trs.funAritySetInRules rs
		   val funlist = L.map (fn (f,_) => f) (FIS.listItems faSet)

		   fun varWithFuncOnPath (Var (x,_)) =  [(x,[])]
		     | varWithFuncOnPath (Fun (f,ts,_)) =  
		       L.map (fn (x,fs) => (x,f::fs)) (LU.mapAppend varWithFuncOnPath ts)

		   fun prVarAndFuncs (x,fs) = "<" ^ (Var.toString x) ^ ","
					      ^ (LU.toStringCommaSquare Fun.toString fs) ^ ">"

		   fun prVarAndFuncsInRule (l,r) =
		       let (* val _ = print ((Term.toString l) ^ " -> " ^ (Term.toString r) ^ "\n") *)
			   val listl = varWithFuncOnPath l
			   (* val _ = print " lhs:" *)
			   (* val _ = print (LU.toStringSemicolonCurly prVarAndFuncs listl) *)
			   (* val _ = print "\n" *)
			   val listr = varWithFuncOnPath r
			   (* val _ = print " rhs:" *)
			   (* val _ = print (LU.toStringSemicolonCurly prVarAndFuncs listr) *)
			   (* val _ = print "\n" *)
		       in () end

		   (* val _ = L.app prVarAndFuncsInRule rs *)

		   fun prConstraint pr (l,r) =
		       let (* val _ = print ((Term.toString l) ^ " -> " ^ (Term.toString r) ^ "\n") *)
		   	   val listl = varWithFuncOnPath l
		   	   val listr = varWithFuncOnPath r
		   	   val vs = VS.listItems (VS.intersection (Term.varSetInTerm l, Term.varSetInTerm r))
		   	   fun prConstraint x =
			       let fun select xs = L.mapPartial 
						       (fn (y,fs) => if Var.equal (x,y) then SOME fs else NONE) 
						       xs
				   fun mkpoly fs = YI.prPlus (L.map Fun.toString fs)
				   val rstrings = L.map mkpoly (select listr)
				   val lstrings = L.map mkpoly (select listl)
			       in YI.prAnd (ListXProd.mapX YI.prLe (rstrings,lstrings))
			       end
		       in L.app (fn x => pr (YI.prAssert (prConstraint x))) vs
		       end

		   (*	  val tmpName = #file (OS.Path.splitDirFile (OS.FileSys.tmpName ())) *)
		   (*	  val _ = debug (fn _=> print ("[" ^ tmpName ^ "]\n")) *)
		   (*	  val inTmpFile = OS.Path.joinDirFile {dir=".",file=tmpName} *)
		   (*	  val fouts = TextIO.openOut inTmpFile *)
		   (*     val cmd = smtSolver ^ " " ^ inTmpFile ^ " 2>&1\n"  *)
 		   val cmd = smtSolver ^ " 2>/dev/null\n"

 		   val proverCmd = ("/bin/sh", ["-c",cmd])
		   val proc = Unix.execute proverCmd
		   val (ins,outs) = Unix.streamsOf proc

		   (* fun yprint str = (TextIO.output (fouts, str); TextIO.flushOut fouts) *)
		   fun yprint str = (TextIO.output (outs, str); TextIO.flushOut outs)
		   val _ = yprint "(set-evidence! true)\n"

		   val _ = L.app (fn (f,ar) => 
				     let val fname = Fun.toString f
					 val _ = yprint (YI.prDefInt fname)
					 val _ = yprint (YI.prAssert (YI.prAnd [YI.prLe ("1",fname), 
									      YI.prLe (fname,"4")]))
				     in () end)
				 (FIS.listItems faSet)
		   val _ = L.app (prConstraint yprint) rs
		   val _ = yprint "(check)\n"
 

		   val answer = TextIO.inputLine ins

		   (* val _ = if (isSome answer)  *)
		   (* 	   then debug (fn _=> print ("output:" ^ (valOf answer))) *)
		   (* 	   else () *)

		   val result =  (isSome answer) andalso (valOf answer = "sat\n")

		   (* $B%W%m%;%9=*N;=hM}(B*)
		   fun finish () = 
		       let (* val _ = debug (fn _=> print "finish the yices process\n")  *)
			   val _ = TextIO.closeOut outs
			   val _ = TextIO.closeIn ins
			   val _ = Unix.reap proc  
			   (* val _ = OS.FileSys.remove inTmpFile *)
		       in () end

	       in if not result 
		  then let val _ = debug (fn _=> print "No suitable weight found ...\n")
		       in (finish ();NONE)
		       end
		  else let 
			  val assign = YI.readAssigns ins
			  (* val _ = print "assignment:\n" *)
			  (* val _ = StringMap.appi (fn (k,str) => print (k ^ ": " ^ str ^ "\n")) assign *)
			  (* val _ = print "end of assignment.\n" *)
			  val funNumlist = L.mapPartial
					       (fn (f,str) => let val i = valOf (Int.fromString str)
							      in if i > 1 then SOME (Fun.fromString f, i) else NONE
							      end)
					       (StringMap.listItemsi assign)
			  
			  fun mknewfun f (init,n) = 
			      let val newfs = L.tabulate (n, fn i => Fun.fromString (
								     (Fun.toString f) ^ "_" 
								     ^ (Int.toString (i+init))))
				  fun mkterm [] ts = (print "mkterm\n"; raise CrSolverError)
				    | mkterm [f] ts = Fun (f,ts,Sort.null)
				    | mkterm (f::fs) ts = Fun (f,[mkterm fs ts],Sort.null)
			      in if LU.disjoint' Fun.equal (newfs,funlist)
				 then (f,fn args => mkterm newfs args)
				 else mknewfun f (init+n,n)
			      end
			  val replacelist = L.map (fn (f,n)=> mknewfun f (1,n)) funNumlist

			  fun replace (t as (Var _)) = t
			    | replace (t as (Fun (f,args,ty))) = 
			      case L.find (fn (g,mk) => Fun.equal (f,g)) replacelist of
				  SOME (g,mk) => mk args | NONE => Fun (f,L.map replace args,ty)
			  val ans = L.map (fn (l,r) => (replace l, replace r)) rs
		      in (finish (); SOME ans)
		      end
	       end
       
	   fun checkGomiOyamaguchiOhta _ = 
	       ((!useGomiOyamaguchiOhtaDepth) orelse (!useGomiOyamaguchiOhtaWeight))
	       andalso
	       if Cr.confluenceByDepthPreserving rs
	       then (println nameGomiOyamaguchiOhtaDepth; true)
	       else if (!useGomiOyamaguchiOhtaWeight)
	       then
		   let val _ = println (unknown nameGomiOyamaguchiOhtaDepth)
		       val _ = debug (fn _ => print "try to apply CR-preserving transformation\n")
		   in case transformationForWightPreserving smtSolver rs of
			  NONE => (println (unknown nameGomiOyamaguchiOhtaWeight); false)
			| SOME rs' => 
			  let val _ =  debug (fn _ => print "new rules obtained:\n")
			      val _ =  debug (fn _ => print (Trs.prRules rs'))
			  in if Cr.confluenceByDepthPreserving rs'
			     then (println nameGomiOyamaguchiOhtaWeight; true)
			     else (println (unknown nameGomiOyamaguchiOhtaWeight); false)
			  end
		   end
	       else (println (unknown nameGomiOyamaguchiOhtaDepth); false)


	   fun checkSuzukiAotoToyama _ = 
	       (!useSuzukiAotoToyama)
	       andalso
	       if CrDirect.isConfluentWeakLeftLinearSystem isTerminating (inCps,outCps) rs
	       then (println nameSuzukiAotoToyama; true)
	       else (println (unknown nameSuzukiAotoToyama); false)

	   fun checkUchidaAotoToyama _ = 
	       (!useUchidaAotoToyama)
	       andalso
	       if CrDiagram.isConfluentQuasiLinearSystem isTerminating smtSolver tmpDir rs
	       then (println nameUchidaAotoToyama; true)
	       else (println (unknown nameUchidaAotoToyama); false)

	   fun checkUchidaAotoToyamaStrong _ = 
	       (!useUchidaAotoToyamaStrong)
	       andalso
	       if CrDiagram.isConfluentStronglyQuasiLinearSystem isTerminating smtSolver tmpDir rs
	       then (println nameUchidaAotoToyamaStrong; true)
	       else (println (unknown nameUchidaAotoToyamaStrong); false)

	   fun checkSakaiOgawa _ = 
	       (!useSakaiOgawa)
	       andalso
	       if Trs.areShallowRules rs
		  andalso L.all Term.equal inCps
		  andalso L.all Term.equal outCps
		  andalso not (Trs.isCollapsingTrs rs)
	       then (println nameSakaiOgawa; true)
	       else (println (unknown nameSakaiOgawa); false)

	   fun checkNonConfluence _ = 
	       ((!NonCr.useNonConfluenceByApproximation)
		orelse
		(!NonCr.useNonConfluenceByTreeAutomata)
		orelse
		(!NonCr.useNonConfluenceByInterpretationAndOrder))
	       andalso
	       let val _ = print "check Non-Confluence...\n"
	       in if NonCr.checkNonConfluence satSolver smtSolver tmpDir rs
		  then true
		  else (println (unknown nameNonConfluence); false)
	       end

	   fun checkBidirectionalCompletion _ = 
	       (!useBidirectionalCompletion)
	       andalso
	       let val _ = print "check by Reduction-Preserving Completion...\n"
		   val result = CrCompletion.bidirectionalCompletion 
				    (isTerminating false)
				    isRelativeTerminating
				    dSymSet rs
		   val _ = if result
			   then print "Reduction-Preserving Completion\n"
			   else print "unknown Reduction-Preserving Completion\n"
	       in result
	       end

	   fun checkConfluenceConditionsForLinearSystem _ = 
	       if checkHuetToyamaOostrom ()
		  orelse
		  checkHuetStrongClosed ()
		  orelse 
		  checkSakaiOgawa ()
		  orelse
		  checkOyamaguchiOhta ()
		  orelse
		  checkToyamaParallel ()
		  orelse
		  checkOkuiSimultaneous ()
		  orelse
		  checkGomiOyamaguchiOhta ()
		  orelse
		  checkDecreasingDiagrams ()
	       then (print "Direct Methods"; report CR)
	       else if checkNonConfluence ()
	       then (print "Direct Methods"; report NotCR)
	       else if
		       checkHuetModulo ()
		       orelse
		       checkBidirectionalCompletion ()
		   then (print "Direct Methods"; report CR)
		   else (print "Direct Methods"; report Unknown)

	   fun checkConfluenceConditionsForLeftLinearSystem _ = 
	       if checkHuetToyamaOostrom ()
		  orelse 
		  checkSakaiOgawa ()
		  orelse
		  checkOyamaguchiOhta ()
		  orelse
		  checkToyamaParallel ()
		  orelse
		  checkOkuiSimultaneous ()
		  orelse
		  checkGomiOyamaguchiOhta ()
		  orelse
		  checkDecreasingDiagrams ()
		  orelse 
		  checkUchidaAotoToyama ()
		  orelse 
		  checkUchidaAotoToyamaStrong ()
	       then (print "Direct Methods"; report CR)
	       else if checkNonConfluence ()
	       then (print "Direct Methods"; report NotCR)
	       else if checkHuetModulo ()
		       orelse
		       checkBidirectionalCompletion ()
	       then (print "Direct Methods"; report CR)
	       else (print "Direct Methods"; report Unknown)

	   fun checkConfluenceConditionsForRightLinearSystem _ = 
	       if checkOhtaOyamaguchiToyama ()
		  orelse 
		  checkSakaiOgawa ()
		  orelse
		  checkGomiOyamaguchiOhta ()
		  orelse 
		  checkSuzukiAotoToyama ()
		  orelse 
		  checkUchidaAotoToyama ()
		  orelse 
		  checkUchidaAotoToyamaStrong ()
	       then (print "Direct Methods"; report CR)
	       else if checkNonConfluence ()
	       then (print "Direct Methods"; report NotCR)
	       else if checkBidirectionalCompletion ()
	       then (print "Direct Methods"; report CR)
	       else (print "Direct Methods"; report Unknown)

	   fun checkConfluenceConditionsForNonLeftRightLinearSystem _ = 
	       if checkSakaiOgawa ()
		  orelse
		  checkGomiOyamaguchiOhta ()
		  orelse 
		  checkSuzukiAotoToyama ()
		  orelse 
		  checkUchidaAotoToyama ()
		  orelse 
		  checkUchidaAotoToyamaStrong ()
	       then (print "Direct Methods"; report CR)
	       else if checkNonConfluence ()
	       then (print "Direct Methods"; report NotCR)
	       else if checkBidirectionalCompletion ()
	       then (print "Direct Methods"; report CR)
	       else (print "Direct Methods"; report Unknown)

	   fun checkConfluenceConditionsForNonTerminatingSystem _ = 
	       case (Trs.areLeftLinearRules rs,
		     Trs.areRightLinearRules rs) of
		   (true,true) => (print "Linear\n";
				   checkConfluenceConditionsForLinearSystem ())
		 | (true,false) => (print "Left-Linear, not Right-Linear\n";
				    checkConfluenceConditionsForLeftLinearSystem ())
		 | (false,true) => (print "not Left-Linear, Right-Linear\n";
				    checkConfluenceConditionsForRightLinearSystem ())
		 | (false,false) => (print "not Left-Linear, not Right-Linear\n";
				     checkConfluenceConditionsForNonLeftRightLinearSystem ())

       in
	   if (!useKnuthBendix)
	   then
	       if null inCps  (* overlay *)
	       then (print "Overlay, check Innermost Termination...\n";
		     if isTerminating true rs
		    then let val _ = print "Innermost Terminating (hence Terminating)"
			     val joinableCps = isInnerJoinableCps rs outCps
			     val _ = if joinableCps
				     then println ", WCR"
				     else println ", not WCR"
			     val _ = println nameKnuthBendix
			 in
			     if joinableCps then 
				 (print "Direct Methods"; report CR)
			     else 
				 (print "Direct Methods"; report NotCR)
			 end
		    else 
			(print "unknown Innermost Terminating\n";
			 println (unknown nameKnuthBendix);
			 checkConfluenceConditionsForNonTerminatingSystem rs))
	       else (print "not Overlay, check Termination...\n";
		     if isTerminating false rs
		     then let val _ = print "Terminating"
			      val joinableCps = isJoinableCps rs (inCps @ outCps)
			      val _ = if joinableCps
				      then println ", WCR"
				      else println ", not WCR"
			      val _ = println nameKnuthBendix
			  in
			     if joinableCps then 
				 (print "Direct Methods"; report CR)
			     else 
				 (print "Direct Methods"; report NotCR)
			  end
		     else
			 (println  "unknown/not Terminating";
			  println (unknown nameKnuthBendix);
			  checkConfluenceConditionsForNonTerminatingSystem rs))
	   else
	       checkConfluenceConditionsForNonTerminatingSystem rs
       end
    end


   fun applyDirectCheck directCheck rs (n,fs) = 
       if null rs
       then report CR
       else let val _ = print "Rewrite Rules:\n"
		val _ = print (Trs.prRules rs)
		val checkCC = directCheck rs
	    in case checkCC of
		   CR => CR
		 | NotCR => NotCR
		 | Unknown => applyDecomposition directCheck rs (n,fs)
	    end
   and applyDecomposition directCheck rs (n,fs) = 
       if (null fs)
       then 
	   Unknown
       else if not (null fs)
	  andalso n = length fs
       then (print "No further decomposition possible\n\n";
	     Unknown)
       else 
	   let 
	       val fs' = (tl fs) @ [hd fs]
	       fun direct rs = applyDirectCheck directCheck rs (1,fs')
	       fun decomp rs = applyDecomposition directCheck rs (n+1,fs')
	   in case (hd fs) of
(*		0 => CrDirect.tryDirectSumDecomposition rs (direct,decomp) *)
		0 => CrDirect.tryPersistentDecomposition rs (direct,decomp)
	      | 1 => CrLayer.tryLayerPreservingDecomposition rs (direct,decomp)
	      | 2 => CrCom.tryCommutativeDecomposition rs (direct,decomp)
	      | _ => (print "Error at applyDecomposition\n";
		      raise CrSolverError)
	   end

   exception Failure



  (* $B9gN.@-%=%k%P(B *)
  fun crSolver (opt0:Solver.options) (opt1:DpSolver.options) 
	       (opt2:PoSolver.options) (opt3:options) rs =
      let
	   val trs = Trs.rulesToTrs rs
	   val _ = debug (fn _ => print "Rewrite Rules:\n") 
	   val _ = debug (fn _ => print (Trs.prRules (#Rules trs))) 

           val _ = runCertification := (#useCertifiableOutputOption opt3)
           val _ = NonCr.runCertification := (!runCertification)
           val _ = PoSolver.runCertification := (!runCertification)

	   val _ = cpout (fn _ => CU.start ())
	   val _ = cpout (fn _ => CU.output (CU.beginInput (CU.inputTrs (Trs.prTrsInProofTree rs))))
	   val _ = cpout (fn _ => CU.output (CU.version "2.1"))

           val satSolver = (#satSolver opt0)
           val smtSolver = (#smtSolver opt0)
           val snProver = (#terminationProver opt0)
           val relSnProver = (#relativeTerminationProver opt0)

           val tmpDir = (#tmpDir opt0)

	   val _ = if  (#VarCond trs) then ()
		   else (print "Variable condition is not satisfied\n";
			 raise Failure)

          fun isTerminating isOverlay rs = 
	      if snProver = ""
	      then 
		  if (#useDp opt1)
		  then if isOverlay 
		       then DpSolver.dpSolverInner satSolver tmpDir opt1 opt2 rs
		       else DpSolver.dpSolver satSolver tmpDir opt1 opt2 rs
		  else PoSolver.poSolver satSolver tmpDir opt2 (rs,[])
	      else 
		  if isOverlay 
		  then Solver.sinSolver snProver tmpDir rs
		  else Solver.snSolver snProver tmpDir rs

	  fun isRelativeTerminating (rs,es) =
	      if relSnProver = ""
	      then 
		  MatrixInterpretation.relsnSolver 
		      (opt0:Solver.options) 
		      (opt1:DpSolver.options) 
		      (opt2:PoSolver.options) (rs,es)
	      else
		  (* print "relative termination proof\n";
		   print "rs:\n";
		   print (Trs.prRules rs);
		   print "es:\n";
		   print (Trs.prRules es);
		   print relSnProver;
		   print "\n"; *)
		  Solver.relsnSolver relSnProver tmpDir (rs,es)

	  fun indexOf (l,r) = 
	      let fun indexOfSub [] _ = (print "indexOf failed\n";
					 raise CrSolverError)
		    | indexOfSub ((l',r')::rest) i
		      = if Term.equal (l,l')
			   andalso Term.equal (r,r')
			then i
			else indexOfSub rest (i+1)
	      in indexOfSub rs 0 
	      end
	   fun getRules ns = L.map (fn i => L.nth (rs,i)) ns
	   val rem = ref ILM.empty

           (* $B<+A3?t%j%9%H$GF~NO(BTRS$B$NItJ,=89g$rBP1~$5$;$k(B *)
	   fun lookup xs  = 
	       case ILM.find (!rem,xs) of
		   SOME ans=> (print "(cached)"; ans)
		 | NONE => let val ans = checkConfluenceConditions2 
					     satSolver smtSolver tmpDir
					     isTerminating 
					     isRelativeTerminating 
					     (getRules xs) 
			       val _ = ILM.insert (!rem, xs, ans)
			   in ans
			   end
	   fun directCheck rs = lookup (L.map indexOf rs)

	   val _ = if #useNonMinimalCommutative opt3
		   then CrCom.useNonMinimal := true
		   else CrCom.useNonMinimal := false

(* (*temporal *)      val _ = NonCr.satSolver := (#satSolver opt0) *)
(* (*temporal *)      val _ = NonCr.smtSolver := (#smtSolver opt0) *)
(* (*temporal *)      val _ = NonCr.snProver := (#terminationProver opt0) *)
(* (*temporal *)      val _ = NonCr.relSnProver := (#relativeTerminationProver opt0) *)
(* (*temporal *)      val _ = NonCr.tmpDir := (#tmpDir opt0) *)

(* (*temporal *)	  val crResult = if NonCr.checkNonConfluence rs *)
(* 			 then NotCR *)
(* 			 else Unknown *)

	  val crResult = 
          case (#useDirect opt3, #useCommutative opt3, #useLayer opt3) of
               (false,false,false)
               => applyDirectCheck directCheck rs (0,[]) 
             | (true,false,false)
               => applyDirectCheck directCheck rs (0,[0]) 
             | (false,true,false)
               => applyDirectCheck directCheck rs (0,[2]) 
             | (false,false,true)
               => applyDirectCheck directCheck rs (0,[1]) 
             | (true,true,false)
               => applyDirectCheck directCheck rs (0,[0,2]) 
             | (true,false,true)
               => applyDirectCheck directCheck rs (0,[0,1]) 
             | (false,true,true)
               => applyDirectCheck directCheck rs (0,[1,2]) 
             | (true,true,true)
               => applyDirectCheck directCheck rs (0,[0,1,2]) 

	  val _ = cpout (fn _ => CU.finish ())
      in
	  case crResult of 
           CR => (print "\nFinal result: CR\n"; CR)
         | NotCR => (print "\nFinal result: not CR\n"; NotCR)
         | Unknown => (print "\nFinal result: Can't judge\n"; Unknown)
      end
      handle Failure => (print "\nFinal result: Can't judge\n"; Unknown)

    end (* of local *)

end (* of structre *)

