(******************************************************************************
 * Copyright (c) 2015, Toyama&Aoto Laboratory, Tohoku University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the 
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(******************************************************************************
 * file: rwtools/rwchecker/gcr.sml
 * description: ingredients for checking ground Chuch-Rosser Property
 * author: AOTO Takahito
 * 
 ******************************************************************************)

signature GCR = 
sig
    val runDebug: bool ref
    val complementOfTerm : Term.term list -> Term.term -> Term.term list
    val complementOfSubst : Term.term list -> Subst.subst -> Subst.subst list 
    val gcrSolver : Term.decl list -> (Term.term * Term.term) list -> Cr.ConfluenceResult
    (* val declsToPattern: Term.decl list -> Term.term list *)
    (* val getConstructor: (Term.term list *  Term.term list * Term.term list) -> Term.term list *)
    (* val getDefined: Term.term list -> Term.term list -> Term.term list *)
    val patCovered: Term.term list -> (Term.term list list * Term.term list list) -> bool
    val coveringPatternSequence: Term.term list -> Term.term list list -> bool

    val subtractPatternSequence: Term.term list -> (Term.term list * Term.term list) ->  Term.term list list option

   val patSubtraction: Term.term list -> (Term.term list list * Term.term list list) ->  Term.term list list
 
    val partRulesAndIndThm: Term.decl list -> (Term.term * Term.term) list 
			    -> (Term.term list 
				* Term.term list 
				* ((Term.term * Term.term) list * (Term.term * Term.term) list) list)
    val partRulesAndIndThmFSCD: Term.decl list -> (Term.term * Term.term) list 
			    -> (Term.term list 
				* Term.term list 
				* Term.term list 
				* ((Term.term * Term.term) list * (Term.term * Term.term) list) list)

    val partRulesAndIndThmIWC: Term.decl list -> (Term.term * Term.term) list 
			    -> (Term.term list 
				* Term.term list 
				* Term.term list 
				* ((Term.term * Term.term) list * (Term.term * Term.term) list) list)
end;

structure Gcr : GCR = 
   struct
   local 
       open Term
       open Trs
       open Rewrite
       open Subst
       structure VS = VarSet
       structure VM = VarMap
       structure FS = FunSet
       structure FM = FunMap
       structure SS = SortSet
       structure FIS = FunIntSet
       structure L = List
       structure LP = ListPair
       structure LU = ListUtil
       structure PU = PrintUtil
       structure TP = TermPair
       structure TPS = TermPairSet
       structure TS = TermSet
       open PrintUtil
   in

   val runDebug = ref false : bool ref
   fun debug f = if !runDebug then f () else ()
   exception GcrError


   fun sameSubst (sigma, rho) = 
       let val kvlist0 = VM.listItemsi sigma
	   val klist0 = L.map (fn (k,v) => k) kvlist0
	   val kvlist1 = VM.listItemsi rho
	   val klist1 = L.map (fn (k,v) => k) kvlist1
       in  LU.setEqual' Var.equal (klist0,klist1)
	   andalso L.all (fn k => Term.equal (VM.lookup (sigma,k), VM.lookup (rho,k))) klist0
       end

   type single_pattern = term (* should be linear term *)
   type pattern = term list (* should be linear *)
   (* W: Υѥ    *)
   (* P: ѥ   *)
   (* pat \in P: signle pattern   *)

   fun selectPatternBySort P ty = L.filter (fn pat => Sort.equal (ty, sortOfTerm pat)) P

  (* pat \subseteq P*)
   fun isNonEmptyTerm W pat = 
       L.exists (fn p =>  isSome (Subst.match p pat)) W

   fun isNonEmptyPattern W P = 
       L.exists (fn pat => isNonEmptyTerm W pat) P

   fun isEmptyPattern W P = not (isNonEmptyPattern W P)

   (* pos is supposed to be a variable position of pat *)
   fun extendTermAtPos W pat pos = 
       let  fun extendPatSub W pat pos = 
		case pos of [] =>  (* root position *)
			    (case pat of Var (x,ty) => selectPatternBySort W ty
				       | Fun (f,ts,ty) => (print "extendTermAtPos: not a variable position\n"; raise GcrError))
			  | i::ps => (* non-root position *)
			    (case pat of Var (x,ty) => (print "extendTermAtPos: not a variable position\n"; raise GcrError)
				       | Fun (f,ts,ty) => let val pati = extendPatSub  W (L.nth (ts,i)) ps
							  in L.map (fn pi => Fun (f, LU.replaceNth (ts,i,pi), ty)) pati 
							  end)
       in L.map Term.linearize (extendPatSub W pat pos)
       end

   fun intersectionOfSingePatterns (pat1,pat2) =
       let val [pat1',pat2'] = Subst.renameTerms [pat1,pat2]
       in case Subst.unify pat1' pat2' of
   	      NONE => []
   	    | SOME sigma => [Subst.applySubst sigma pat1']
       end

  (* pat1  pat2  subsume 뤫 pat2  pat1 üʾ礫 *)
   fun subsumeSingePattern pat1 pat2 = isSome (Subst.match pat1 pat2)

   fun isEmptyTerm W pat = L.all (fn p=> null (intersectionOfSingePatterns (p,pat))) W
   fun isEmptyPattern W P = L.all (fn p => isEmptyTerm W p) P

   fun isNonEmptyTerm W pat = L.exists (fn p=> not (null (intersectionOfSingePatterns (p,pat)))) W
   fun isNonEmptyPattern W P = L.exists (fn p => isNonEmptyTerm W p) P

   fun deleteAuxFromTerm P pat = 
       let val (pat'::P')  = Subst.renameTerms (pat::P)
       in  LU.eliminateDuplication' Term.equal 
				    (LU.mapAppend (fn p => intersectionOfSingePatterns (p,pat')) P')
       end

   fun applySublist sublist t = L.map (fn sigma => (Subst.applySubst sigma t)) sublist

   fun intersectionOfPatterns (P,Q) = 
       LU.mapAppend (fn p => deleteAuxFromTerm P p) Q

   local
       fun listToSubst ks vs = LP.foldl (fn (k,v,map) => VM.insert (map,k,v)) VM.empty (ks,vs)
   in
   (* աsort 㤦Τ complement ʤ *)
   (*  W Ǥϡ f(..variable only ...) η *)
   (* fun complementOfTerm W t = *)
   (*     let fun compOfTerm [] t ans = L.rev ans *)
   (* 	     | compOfTerm (p::ps) t ans = *)
   (* 	       if not (Term.haveSameSort (p,t)) *)
   (* 	       then compOfTerm ps t ans *)
   (* 	       else case Subst.match p t of *)
   (* 			NONE => compOfTerm ps t (p::ans) *)
   (* 		      | SOME sigma => *)
   (* 			let (* val _ = PrintUtil.println ("\nP: " ^ (LU.toStringCommaSquare Term.toString W)) *) *)
   (* 			    (* val _ = PrintUtil.println ("pattern: " ^ (Term.toString p)) *) *)
   (* 			    (* val _ = PrintUtil.println ("term: " ^ (Term.toString t)) *) *)
   (* 			    (* val _ = PrintUtil.println ("sub: " ^ (Subst.toString sigma)) *) *)
   (* 			    (* val _ = PrintUtil.println ("compute complement sub of " ^ (Subst.toString sigma) ^ "...") *) *)
   (* 			    val ss = complementOfSubst W sigma *)
   (* 			    (* val _ = PrintUtil.println ("...complement sub of " ^ (Subst.toString sigma) *) *)
   (* 			    (* 			       ^ " := " ^ (LU.toStringCommaSquare Subst.toString ss)) *) *)
   (* 			    val pcomp = L.map Term.linearize (applySublist ss p) *)
   (* 			in compOfTerm ps t (pcomp @ ans) *)
   (* 			end *)
   (*     in compOfTerm W t [] *)
   (*     end *)
   fun complementOfTerm W t =
       let fun compOfTermList [] _ = []
   	     | compOfTermList ps [] = ps
   	     | compOfTermList ps (t::ts) =
   	       let val compt = compOfTerm ps t
   	       in compOfTermList compt ts
   	       end
   	   and compOfTerm [] _ = []
   	     | compOfTerm (p::ps) t =
   	       let val [p',t'] = Subst.renameTerms [p,t]
   	       in if isSome (Subst.match t' p')
   		  then compOfTerm ps t
   		  else case Subst.unify p' t' of
   			   NONE => p::(compOfTerm ps t)
   			 | SOME sigma =>
   			let (* val _ = PrintUtil.println ("\nP: " ^ (LU.toStringCommaSquare Term.toString W)) *)
   			    (* val _ = PrintUtil.println ("pattern: " ^ (Term.toString p)) *)
   			    (* val _ = PrintUtil.println ("term: " ^ (Term.toString t)) *)
   			    (* val _ = PrintUtil.println ("sub: " ^ (Subst.toString sigma)) *)
   			    (* val _ = PrintUtil.println ("compute complement sub of " ^ (Subst.toString sigma) ^ "...") *)
   			    val ss = complementOfSubst W sigma
   			    (* val _ = PrintUtil.println ("...complement sub of " ^ (Subst.toString sigma) *)
   			    (* 			       ^ " := " ^ (LU.toStringCommaSquare Subst.toString ss)) *)
   			    val pcomp = L.map Term.linearize (applySublist ss p')
   			    val tcomp = L.map Term.linearize (applySublist ss t')
   			    val pcomp' = LU.eliminateDuplication' (fn (x,y) => Subst.isVariant x y) (pcomp @ ps)
   			in compOfTermList pcomp' tcomp
   			end
   	       end
   	   val (W1,W2) = L.partition (fn p => Term.haveSameSort (p,t)) W
       in W2 @ (compOfTerm W1 t)
       end
   and complementOfSubst W sigma = 
       let fun compOfSubst sigma = 
	       let val sigma' = VM.filteri (fn (k,v) => (not (isVar v))) sigma
	       in if VM.isEmpty sigma'
		  then []
		  else let val kvlist = VM.listItemsi sigma'
			   val klist = L.map (fn (k,v) => k) kvlist
			   val vlist = L.map (fn (k,v) => v) kvlist
			   (* val _ = PrintUtil.println ("take complements of terms: " ^ (LU.toStringComma Term.toString vlist)) *)
		           val entlist =  L.map (fn t => (if isNonEmptyPattern W [t] then [t] else [])
							 @  (L.filter (fn p => Term.haveSameSort (p,t)) (complementOfTerm W t)))
						vlist
			   val sublist0 = LU.allCombinations entlist
			   val result = LU.deleteOne' sameSubst sigma' (L.map (listToSubst klist) sublist0)
		       in result 
		       end
	       end
       in compOfSubst sigma
       end
   end (* local *)

   fun complementOfPattern W [] =  W
     | complementOfPattern W (p::ps) = 
       intersectionOfPatterns (complementOfTerm W p, complementOfPattern W ps)

   fun subtractPattern W (P,Q) = 
       intersectionOfPatterns (P, complementOfPattern W Q)

   fun equalPatterns W (P,Q) = 
       isEmptyPattern W (subtractPattern W (P,Q))
       andalso isEmptyPattern W (subtractPattern W (Q,P))

  (* P  Τ򥫥СƤ뤫 *)
   fun coveringPattern W P = 
       isEmptyPattern W (subtractPattern W (W,P))
       (* let fun coverSub [] _ = true *)
       (* 	     | coverSub (w::ws) qs  = *)
       (* 	       let val (w'::qs')  = Subst.renameTerms (w::qs) *)
       (* 		   val cand = L.mapPartial (fn q => case Subst.unify w' q of  *)
       (* 							NONE => NONE *)
       (* 						      | SOME sigma => SOME (q,sigma)) qs' *)
       (* 	       in if null cand  *)
       (* 		  then false *)
       (* 		  else let val (q,sigma) = hd cand *)
       (* 		       in if isSome (Subst.match q w') *)
       (* 			  then coverSub ws qs *)
       (* 			  else let val qs2 = LU.deleteOne' Term.equal q qs' *)
       (* 				   val sublist = complementOfSubst W sigma *)
       (* 				   val wcomp = applySublist sublist w' *)
       (* 				   val qcomp = applySublist sublist q *)
       (* 			       in coverSub (LU.eliminateDuplication' (fn (x,y) => Subst.isVariant x y) (wcomp @ ws)) *)
       (* 					   (LU.eliminateDuplication' (fn (x,y) => Subst.isVariant x y) (qcomp @ qs')) *)
       (* 			       end *)
       (* 		       end *)
       (* 	       end *)
       (* in coverSub W P *)
       (* end *)


   local
       exception Found of Term.term * Term.term * Term.term * Term.term * Subst.subst
   in
   fun findUnif (ps,ls) =
       ((L.app (fn l => L.app (fn p => let val [p',l'] = Subst.renameTerms [p,l]
				       in case Subst.unify p' l' of
					      SOME sigma => raise Found (p,l,p',l',sigma)
					    | NONE => ()
				       end) ps) ls); NONE)
       handle (Found (p,l,p',l',sigma)) => SOME (p,l,p',l',sigma)
   end	       

   fun matchSubterm hs t  = 
       let val subs = Term.properSubterms t
       in L.exists (fn h => L.exists (fn s => isSome (Subst.match h s)) subs) hs
       end

   fun getConstructor (ps,ls,hs) =
       let  fun getConstructorStep k (ps,ls,hs) =
		if k > 5
		then (print "Failure: step exceeds the limit\n"; ps)
		else case findUnif (ps,ls) of
			 NONE => ps
		       | SOME (p,l,p',l',sigma) => 
			 let val t = Subst.applySubst sigma l'
			     (* val _ = PU.println ("\nt = " ^ (Term.toString t)) *)
			     val hs2 =  t::hs
			     val sigma1 = Subst.restrictDomain (Term.varSetInTerm l') sigma
			     (* val _ = PU.println ("sigma1 = " ^ (Subst.toString sigma1)) *)
			     val sigma2 = Subst.restrictDomain (Term.varSetInTerm p') sigma
			     (* val _ = PU.println ("sigma2 = " ^ (Subst.toString sigma2)) *)
			     val L = LU.deleteOne' Term.equal l ls
			     val P = LU.deleteOne' Term.equal p ps
			     (* val _ = PU.println ("L = " ^ (LU.toStringCommaSquare Term.toString L)) *)
			     (* val _ = PU.println ("P = " ^ (LU.toStringCommaSquare Term.toString P)) *)
			     val comp1 = complementOfSubst P sigma1
			     (* val _ = PU.println ("comp1 = " ^ (LU.toStringCommaSquare Subst.toString comp1)) *)
			     val comp2 = complementOfSubst P sigma2
			     (* val _ = PU.println ("comp2 = " ^ (LU.toStringCommaSquare Subst.toString comp2)) *)
			     val ls' = applySublist comp1 l'
			     (* val _ = PU.println ("ls' = " ^ (LU.toStringCommaSquare Term.toString ls')) *)
			     val ps' = applySublist comp2 p'
			     (* val _ = PU.println ("ps' = " ^ (LU.toStringCommaSquare Term.toString ps')) *)
			     val ls2 = L.filter (fn t => not (matchSubterm hs2 t)) (L @ ls')
			     val ps2 = L.filter (fn t => not (matchSubterm hs2 t)) (P @ ps')
			     (* val _ = PU.println ("ls2 = " ^ (LU.toStringCommaSquare Term.toString ls2)) *)
			     (* val _ = PU.println ("ps2 = " ^ (LU.toStringCommaSquare Term.toString ps2)) *)
			 in getConstructorStep (k+1) (ps2,ls2,hs2)
			 end
       in getConstructorStep 0 (ps,ls,hs)
       end

  (* lhs 꿼 depth ޤǴäơΤnf ˻Ȥؿ롥 *)
  (* ls --- աps --- ѥ *)
   fun getConstructor2 (ps,ls,hs) = 
       let fun takeMaxDepth xs = hd (LU.takeMaximals (fn (x,y) => SOME (Int.compare (x,y))) (L.map Term.termDepth xs))
	   fun takeMinDepth xs = hd (LU.takeMinimals (fn (x,y) => SOME (Int.compare (x,y))) (L.map Term.termDepth xs))
	   val depth = takeMaxDepth ls
	   val _ = PU.println (Int.toString depth) 
	   fun mkGroundFromTerm t = let (* val vars = VarSet.listItems (Term.varSetInTerm t)  *)
					(* val len = L.length vars *)
					(* val pslist = LU.allCombinations (L.tabulate (len, fn x=>ps)) *)
					(* fun makesub ps = LP.foldl (fn (x,p,fm) => VM.insert (fm,x,p)) VM.empty (vars,ps) *)
					(* val sublist = L.map makesub pslist *)
	                                val vars = SortedVarSet.listItems (Term.sortedVarSetInTerm t) 
					val len = L.length vars
					val pslist = LU.allCombinations (L.tabulate (len, fn x=>ps))
					exception SortInconsistence
					fun makesub ps = let val ans = LP.foldl (fn ((x,ty),p,fm) => 
										    (if Sort.equal (ty, Term.sortOfTerm p)
										     then VM.insert (fm,x,p)
										     else raise SortInconsistence))
										VM.empty (vars,ps)
							 in SOME ans
							 end handle SortInconsistence => NONE
					val sublist = L.mapPartial makesub pslist
				    in L.map Subst.linearize (applySublist sublist t)
				    end
	   fun mkstep yet done = if takeMinDepth yet < depth + 1
				 then let val (done2,yet2) = L.partition Term.isGroundTerm (LU.mapAppend  mkGroundFromTerm yet)
				      in mkstep yet2 (done@done2)
				      end
				 else done
	   val grounds = mkstep ps []
	   val _ = PU.println (Int.toString (L.length grounds))
	   val _ = PU.println (LU.toStringCommaSquare Term.toString grounds)
	   val dummyRules = L.map (fn l => (l, Term.Fun (Fun.fromString "a",[],Sort.null))) ls
	   val gnfs = L.filter (fn t => Rewrite.isNormalForm dummyRules t) grounds
	   val _ = PU.println (LU.toStringCommaSquare Term.toString gnfs)
	   val cSet = Term.funSetInTerms gnfs
       in L.filter (fn pat => FS.member (cSet,valOf (funRootOfTerm pat))) ps
       end

  (* lhs 꿼 depth ޤ äơnf ˻Ȥؿ롥 *)
  (* ls --- աps --- ѥ *)
  (* efficient version *)
   fun getConstructor3 (patlist,lhslist,hs) = 
       let fun takeMaxDepth xs = hd (LU.takeMaximals (fn (x,y) => SOME (Int.compare (x,y))) (L.map Term.termDepth xs))
	   fun takeMinDepth xs = hd (LU.takeMinimals (fn (x,y) => SOME (Int.compare (x,y))) (L.map Term.termDepth xs))
	   (* val _ = PU.println (LU.toStringCommaCurly Term.toString lhslist) *)
	   val depth = takeMaxDepth lhslist
	   (* val _ = PU.println (Int.toString depth)  *)
	   val patlist2 = L.map Term.linearize patlist
	   (* val _ = PU.println (LU.toStringCommaCurly Term.toString patlist2) *)
	   val dummyRules = L.map (fn l => (l, Term.Fun (Fun.fromString "a",[],Sort.null))) lhslist
	   val patlist3 = L.filter (Rewrite.isNormalForm dummyRules) patlist2
	   (* val _ = PU.println (LU.toStringCommaCurly Term.toString patlist3) *)

	   fun varLen (Var _) = 1
	     | varLen (Fun (f,ts,ty)) = varLenList ts
	   and varLenList [] = 0
	     | varLenList (t::ts) = (varLen t) + (varLenList ts)

	   exception SortInconsistence
	   fun repVars t ps = let val (ans,_) = repVarsSub (t,ps)
			      in SOME ans end
			      handle SortInconsistence => NONE
	   and repVarsSub (t,[]) = (t,[])
	     | repVarsSub (Var (v,ty), p::ps) = if Sort.equal (ty, Term.sortOfTerm p)
						then (p, ps) 
						else raise SortInconsistence
	     | repVarsSub (Fun (f,ts,ty), ps) = let val (ts2,ps2)  = repVarsSubList (ts,ps)
						   in (Fun (f,ts2,ty),   ps2)
						   end
	   and repVarsSubList ([], ps) = ([],ps)
	     | repVarsSubList (t::ts, ps) = let val (t2,ps2) = repVarsSub (t, ps)
	   				       val (ts2,ps3) = repVarsSubList (ts,ps2)
	   				   in (t2::ts2, ps3) 
					   end

	   (* val _ = L.app (fn u=> PU.println (Int.toString (varLen u))) lhslist *)

	   fun mkGroundFromTerm t = let val len = varLen t
					val pslist = LU.allCombinations (L.tabulate (len, fn x=>patlist3))
				    in L.mapPartial (fn ps => repVars t ps) pslist
				    end

	   (* val _ = PU.println (LU.toStringCommaCurly Term.toString (LU.mapAppend mkGroundFromTerm lhslist)) *)

	   fun mkstep n yet done = if n < depth
				   then let (* val _ = PU.println ("CURRENT DEPTH: " ^ (Int.toString n)) *)
					    val next = LU.mapAppend mkGroundFromTerm yet
					    (* val _ = print "NEXT: " *)
					    (* val _ = PU.println (LU.toStringCommaCurly Term.toString next) *)
					    val next2 = L.filter (Rewrite.isNormalForm dummyRules) next
					    (* val _ = print "NEXT2: " *)
					    (* val _ = PU.println (LU.toStringCommaCurly Term.toString next2) *)
					    val (done2,yet2) = L.partition Term.isGroundTerm next2
					    (* val _ = print "DONE2: " *)
					    (* val _ = PU.println (LU.toStringCommaCurly Term.toString done2) *)
					    (* val _ = print "YET2: " *)
					    (* val _ = PU.println (LU.toStringCommaCurly Term.toString yet2) *)
					in mkstep (n+1) yet2 (done@done2)
					end
				 else done
	   val grounds = mkstep 1 patlist3 []

	   (* val _ = PU.println (Int.toString (L.length grounds)) *)
	   (* val _ = PU.println (LU.toStringCommaSquare Term.toString grounds) *)
	   (* val gnfs = L.filter (fn t => Rewrite.isNormalForm dummyRules t) grounds *)
	   (* val _ = PU.println (LU.toStringCommaSquare Term.toString gnfs) *)
	   val cSet = Term.funSetInTerms grounds
       in L.filter (fn pat => FS.member (cSet,valOf (funRootOfTerm pat))) patlist 
       end

(*   val env = IOMstrs.rdDecls ["plus:Nat*Nat=> Nat", "s:Nat=> Nat", "0:Nat", "nil:List", "cons:Nat*List=>List" ] *)
   val env = IOMstrs.rdDecls ["plus:Nat*Nat=> Nat", "s:Nat=> Nat", "0:Nat", "nil:List", "cons:Nat*List=>List", "times:Coef*Nat=>Nat", "1:Coef" ]

   fun rdSortedTerm str = case Term.attachSortToTerm env (IOMstrs.rdTerm str) of
			      SOME t => t
			    | NONE => (PU.println ("Error: non-sorted term " ^ str); raise GcrError)

   val envInt = IOMstrs.rdDecls ["s:Int=>Int", "p:Int=>Int", "0:Int" ]
   fun rdIntTerm str = case Term.attachSortToTerm envInt (IOMstrs.rdTerm str) of
			      SOME t => t
			    | NONE => (PU.println ("Error: non-sorted term " ^ str); raise GcrError)

   val patPlus = rdSortedTerm "plus(?x:Nat,?y:Nat)"
   val patPlus0Y = rdSortedTerm "plus(0,?y:Nat)"
   val patS = rdSortedTerm "s(?x:Nat)"
   val patSS = rdSortedTerm "s(s(?x:Nat))"
   val patS0 = rdSortedTerm "s(0)"
   val pat0 = rdSortedTerm "0"
   val patCons = rdSortedTerm "cons(?x:Nat,?ys:List)"
   val patNil = rdSortedTerm "nil" 
   val patX = rdSortedTerm "?x:Nat" 
   val PNat = [patPlus, patS, pat0]
   val PInt = [rdIntTerm "s(?x:Int)", rdIntTerm "p(?x:Int)", rdIntTerm "0"]
   val PList = [patCons, patNil]
   val PNatList = PNat @ PList




   val _ = PU.println "extendTermAtPos: [+(0,y), +(s(x),y)]"
   val patPlusPlusY = extendTermAtPos [pat0,patS] patPlus [0]
   val _ = PU.println (LU.toStringCommaSquare Term.toString patPlusPlusY)

   val _ = PU.println "delete auxiliary pattern: [+(0,y)]"
   val _ = PU.println (LU.toStringCommaSquare Term.toStringWithSort (deleteAuxFromTerm [patPlus] patPlus0Y))

   val _ = PU.println "delete auxiliary pattern: []"
   val _ = PU.println (LU.toStringCommaSquare Term.toStringWithSort (deleteAuxFromTerm [patS] patPlus0Y))

   val _ = PU.println "isNonEmpty Pattern: true" 
   val _ = PU.println (if isNonEmptyPattern PNatList [pat0] then "true" else "false")

   val ps = complementOfTerm PNat (rdSortedTerm "plus(0,?y:Nat)")
   val _ = PU.println (LU.toStringCommaSquare Term.toStringWithSort ps)
   val _ = PU.println (LU.toStringCommaSquare Term.toString ps)
   val _ = PU.println "complement of term: [+(s(x),y), +(+(x,y),z), s(x), 0]"

   val ps = complementOfPattern PNatList [rdSortedTerm "cons(0,?xs:List)"]
   val _ = PU.println (LU.toStringCommaSquare Term.toStringWithSort ps)
   val _ = PU.println (LU.toStringCommaSquare Term.toString ps)
   val _ = PU.println "complement of pattern: [s(x)::xs, +(x,y)::xs, nil]"

   val ps = complementOfPattern PNatList [rdSortedTerm "cons(0,?ys:List)", rdSortedTerm "cons(s(?x:Nat),nil)"]
   val _ = PU.println (LU.toStringCommaSquare Term.toStringWithSort ps)
   val _ = PU.println (LU.toStringCommaSquare Term.toString ps)
   val _ = PU.println "complement of pattern: [s(x)::y::xs, +(x,y)::xs, nil]"
 (**************  
      =========> ºݤˤϡ
      [cons(plus(?x_1,?y_1),cons(?x_2,?ys_1)),cons(plus(?x_1,?y_1),nil),cons(s(?x_1),cons(?x_2,?ys_1)),nil]
                  ֤äƤ롥ѥ xs  nil  x::xs  ŸƤޤΤϡޤʤ
  **************)

   val P1 = [rdSortedTerm "cons(0,nil)", rdSortedTerm "cons(0,cons(?x:Nat,?ys:List))"]
   val P2 = [rdSortedTerm "cons(0,?ys:List)"]
   val _ = if equalPatterns PNatList (P1,P2) then "true" else "false"
   val _ = PU.println "equalPatterns: true"

   val _ = print "\n------------------------\n"

   val _ = if coveringPattern [patS,pat0] [patSS,patS0,pat0]
   	   then print "[s(x),0] covers [s(s(x)),s(0),0]\n"
   	   else print "[s(x),0] not covers [s(s(x)),s(0),0]\n"

   val ps = complementOfTerm [patSS,patS0,pat0] patS
   val _ = PU.println (LU.toStringCommaSquare Term.toString ps)

   val _ = if coveringPattern [patSS,patS0,pat0] [patS,pat0]
   	   then print "[s(s(x)),s(0),0] covers [s(x),0]\n"
   	   else print "[s(s(x)),s(0),0] not covers [s(x),0]\n"

   val declNat = IOMstrs.rdDecls ["plus:Nat*Nat=> Nat", "s:Nat=> Nat", "0:Nat"]
   val declNatList = IOMstrs.rdDecls ["plus:Nat*Nat=> Nat", "s:Nat=> Nat", "0:Nat", "nil:List", "cons:Nat*List=>List" ]
   val declInt = IOMstrs.rdDecls ["s:Int=>Int","p:Int=>Int", "0:Int"] 

   fun declToTerm (decl:Term.decl) =
       let val f = #sym decl
	   val ty = #sort decl
	   val count = ref 0
	   fun newVar () = (count := (!count) + 1; Var.fromStringAndInt ("x",!count))
	   fun newSortedVar ty = Term.Var (newVar (), ty)
       in if Sort.isBaseType ty
	  then Fun (f, [], ty)
	  else Fun (f, L.map newSortedVar (Sort.args ty), Sort.return ty)
       end

   fun declsToPattern decls = L.map declToTerm decls

   fun rdSortedRules env ss
       = let val rs = IOMstrs.rdRules ss
	     fun attach t = case Term.attachSortToTerm env t of
				SOME u => u
			      | NONE => (PU.println ("Error: non-sorted term " ^ (Term.toString t)); raise GcrError)
	 in L.map  (fn (l,r) => (attach l, attach r)) rs
	 end
				 
   fun getLHS rs = L.map (fn (l,r) => l) rs

   val R = rdSortedRules declNat ["plus(0,?y:Nat) ->?y:Nat",
				  "plus(s(?x:Nat),?y:Nat) ->s(plus(?x:Nat,?y:Nat))"]
   val _ = print (Trs.prRules R)
   val cpat = getConstructor3 (PNat, getLHS R, [])
   val _ = PU.println ("Constructor3 pattern: " ^ (LU.toStringCommaSquare Term.toString cpat))


   val R = rdSortedRules declNat ["plus(0,?y:Nat) ->?y:Nat",
				  "plus(s(?x:Nat),?y:Nat) ->s(plus(?x:Nat,?y:Nat))"]
   val _ = print (Trs.prRules R)
   val cpat = getConstructor (PNat, getLHS R, [])
   val _ = PU.println ("Constructor pattern: " ^ (LU.toStringCommaSquare Term.toString cpat))

   val R = rdSortedRules declNat ["plus(0,?y:Nat) -> ?y:Nat",
				  "plus(s(?x:Nat),?y:Nat) -> s(plus(?x:Nat,?y:Nat))",
				  "s(s(?x:Nat)) -> ?x:Nat"]
   val _ = print (Trs.prRules R)
   val cpat = getConstructor (PNat, getLHS R, [])
   val _ = PU.println ("Constructor pattern: " ^ (LU.toStringCommaSquare Term.toString cpat))

   val R = rdSortedRules declNat ["plus(0,?y:Nat) -> ?y:Nat",
				  "plus(s(0),?y:Nat) -> s(?y:Nat)",
				  "s(s(?x:Nat)) -> ?x:Nat"]
   val _ = print (Trs.prRules R)
   val cpat = getConstructor (PNat, getLHS R, [])
   val _ = PU.println ("Constructor pattern: " ^ (LU.toStringCommaSquare Term.toString cpat))

   val R = rdSortedRules declInt ["s(p(?x:Int)) -> ?x:Int",
   				  "p(s(?x:Int)) -> ?x:Int"]
   val _ = print (Trs.prRules R)
   val cpat = getConstructor (PInt, getLHS R, [])
   val _ = PU.println ("Constructor pattern: " ^ (LU.toStringCommaSquare Term.toString cpat))

   val R = rdSortedRules declInt ["s(p(?x:Int)) -> ?x:Int"]
   val _ = print (Trs.prRules R)
   val cpat = getConstructor (PInt, getLHS R, [])
   val _ = PU.println ("Constructor pattern: " ^ (LU.toStringCommaSquare Term.toString cpat))

				  
   val R = rdSortedRules declNat ["plus(0,?y:Nat) -> ?y:Nat",
				  "plus(s(?x:Nat),?y:Nat) -> s(plus(?x:Nat,?y:Nat))",
				  "plus(plus(?x:Nat,?y:Nat),?z:Nat) -> plus(?x:Nat,plus(?y:Nat,?z:Nat))"]
   val _ = print (Trs.prRules R)
   val cpat = getConstructor (PNat, getLHS R, [])
   val _ = PU.println ("Constructor pattern: " ^ (LU.toStringCommaSquare Term.toString cpat))

(* added test *)
   val declCoef = IOMstrs.rdDecls ["times:Coef*Nat=> Nat", "s:Nat=> Nat", "0:Nat", "1:Coef" ]
   val R = rdSortedRules declCoef ["times(1,0) -> 0",
                                   "times(1,s(?y:Nat)) ->s(times(1,?y:Nat))"]
   val _ = print (Trs.prRules R)
   val pat1 = rdSortedTerm "1"
   val patTimes = rdSortedTerm "times(?x:Coef,?y:Nat)"
   val PCoef = [patTimes, pat1, patS, pat0]
   val cpat = getConstructor3 (PCoef, getLHS R, [])
   val _ = PU.println ("Constructor2 pattern: " ^ (LU.toStringCommaSquare Term.toString cpat))


   fun subtractPatternSequence cPat (ps,qs) = 
       if LP.all (fn (p,q) => isSome (Subst.match q p)) (ps,qs)
       then SOME []  (* qs is more general than ps *)
       else let val len = L.length ps
		val terms = Subst.renameTerms (ps @ qs)
		val ps' = L.take (terms, len)
		val qs' = L.drop (terms, len)
		val unifs = LP.map (fn (p,q) => Subst.unify p q) (ps',qs')
		fun complement sigma = if VM.isEmpty sigma
					  orelse L.all Term.isVar (VarMap.listItems sigma)
				       then NONE
				       else SOME (complementOfSubst cPat sigma)
		fun apply sublistlist ps = let val ls = L.map (fn n => 
								  LU.eliminateDuplication' (fn (x,y) => Subst.isVariant x y)
											   (applySublist (L.nth (sublistlist,n))  
													 (L.nth (ps,n))))
							      (L.tabulate (L.length ps, fn n => n))
					   in LU.allCombinations ls end
	    in if L.exists (not o isSome) unifs
	       then NONE (* no common part *)
	       else let val sublistlist = L.map (fn sigmaop => case complement (valOf sigmaop) of
								   SOME sublist => ((valOf sigmaop)::sublist)
								 | NONE => [valOf sigmaop]) unifs 
			val inst = LP.map (fn (sigmaop,p) => Subst.applySubst (valOf sigmaop) p) (unifs,ps')
			val ans = LU.deleteOne' (fn (ps,qs) => LP.all (fn (p,q) => Subst.isVariant p q) (ps,qs))
						inst
						(apply sublistlist ps')
		    in SOME ans
		    end
	    end

   val _ = print "coveringPattern\n"
   val PNat0 = [patS, pat0]
   val pslist = valOf (subtractPatternSequence PNat0 ([patX,patX], [pat0,patS]))
(*   val pslist = valOf (subtractPatternSequence PNat0 ([patX], [pat0])) *)
   val _ = PU.println ("Constructor patterns: " ^
    		 (LU.toStringSemicolonSquare  
		     (fn ps => LU.toStringCommaCurly Term.toString ps) pslist))

  (* cPat = [0, s(x)] 
     patlist = [[0,y] [s(x),y]] 
     ps = [0,const]
   *)


  (* cPat = [0, s(x)] 
     pslist=[x,y], patlist=[[0,y] [s(x),y]]     => true
     pslist=[0,y], patlist = [[0,0],[x,s(y)]]   => true
     pslist=[0,y], patlist = [[0,0],[0,s(y)]]   => true
   *)

   (* pslist  patlist ǥСƤ뤫 *)
   fun patCovered cPat (pslist,patlist)=
       let fun subsumeseq (ps,qs) = LP.all (fn (p,q) => isSome (Subst.match p q)) (ps,qs)
   	   fun elimdup [] ans = L.rev ans
   	     | elimdup (ps::rest) ans = (L.filter (fn qs => not (subsumeseq (ps,qs))) rest)
   					@ (ps::(L.filter (fn qs => not (subsumeseq (ps,qs))) ans))
   	   fun covers [] qslist = true
   	     | covers pslist [] = false
   	     | covers (ps::pslist) qslist =
	       if L.exists (fn qs => subsumeseq (qs,ps)) qslist
	       then covers pslist qslist
	       else case L.find (fn qs => let val len = L.length ps
					      val terms = Subst.renameTerms (ps @ qs)
					      val ps' = L.take (terms, len)
					      val qs' = L.drop (terms, len)
					  in  LP.all (fn (p,q) => isSome (Subst.unify p q)) (ps',qs') 
					  end) 
				qslist of
			NONE => false
		      | SOME qs => case subtractPatternSequence cPat (ps,qs) of
				       NONE => (print "Error: no common part found ???\n"; raise GcrError)
				     | SOME comp => covers (elimdup (comp @ pslist) []) qslist
          in covers pslist patlist
       end

   val _ = if patCovered PNat0 ([[patX,patX]], [[pat0,patX], [patS,patX]])
   	   then print "[[x,y]] is covered by [[0,y],[s(x),y]].\n"
   	   else print "[[x,y]] is not covered by [[0,y],[s(x),y]].???\n"

   val _ = if patCovered PNat0 ([[pat0,patX]], [[pat0,pat0], [pat0,patS]])
   	   then print "[[0,y]] is covered by [[0,0],[0,s(y)]].\n"
   	   else print "[[0,y]] is not covered by [[0,0],[0,s(y)]].???\n"

   val _ = if patCovered PNat0 ([[pat0,patX]], [[pat0,pat0], [patS,patX]])
   	   then print "[[0,y]] is covered by [[0,0],[s(x),y]].???\n"
   	   else print "[[0,y]] is not covered by [[0,0],[s(x),y]].\n"

   (* pslist - patlist  *)
   fun patSubtraction cPat (pslist,patlist)=
       let fun subsumeseq (ps,qs) = LP.all (fn (p,q) => isSome (Subst.match p q)) (ps,qs)
   	   fun elimdup [] ans = L.rev ans
   	     | elimdup (ps::rest) ans = (L.filter (fn qs => not (subsumeseq (ps,qs))) rest)
   					@ (ps::(L.filter (fn qs => not (subsumeseq (ps,qs))) ans))
   	   fun subtract [] qslist = []
   	     | subtract pslist [] = pslist
   	     | subtract (ps::pslist) qslist =
	       if L.exists (fn qs => subsumeseq (qs,ps)) qslist
	       then subtract pslist qslist
	       else case L.find (fn qs => let val len = L.length ps
					      val terms = Subst.renameTerms (ps @ qs)
					      val ps' = L.take (terms, len)
					      val qs' = L.drop (terms, len)
					  in  LP.all (fn (p,q) => isSome (Subst.unify p q)) (ps',qs') 
					  end) 
				qslist of
			NONE => ps::(subtract pslist qslist)
		      | SOME qs => case subtractPatternSequence cPat (ps,qs) of
				       NONE => (print "Error: no common part found ???\n"; raise GcrError)
				     | SOME comp => subtract (elimdup (comp @ pslist) []) qslist
          in subtract pslist patlist
       end


  (* cPat = [0, s(x)] 
     patlist = [[0,y] [s(x),y]]
     patlist = [[0,0] [s(x),0],[x,s(y)]]
     patlist = [[0,y] [s(0),y],[s(s(x)),y]]
   *)

   fun coveringPatternSequence cPat [] = null cPat
     | coveringPatternSequence cPat patlist =
       let val hdpat = hd patlist
   	   val count = ref 0
   	   fun newVar () = (count := (!count) + 1; Var.fromStringAndInt ("x",!count))
   	   fun newSortedVar ty = Term.Var (newVar (), ty)
   	   val basepat = L.map (fn t => newSortedVar (Term.sortOfTerm t)) hdpat
       in patCovered cPat ([basepat],patlist)
       end


   val _ = if coveringPatternSequence PNat0 [[pat0,patX], [patS,patX]]
   	   then print "[[0,y],[s(x),y]]covering Nat over 0, s.\n"
   	   else print "[[0,y],[s(x),y]] not covering Nat over 0, s.???\n"

   val _ = if coveringPatternSequence PNat0 [[pat0,pat0], [patS,pat0], [patX,patS]]
   	   then print "[[0,0],[s(x),0],[x,s(y)]] covering Nat over 0, s.\n"
   	   else print "[[0,0],[s(x),0],[x,s(y)]] not covering Nat over 0, s.???\n"

   val _ = if coveringPatternSequence PNat0 [[pat0,patX], [patS0,patX], [patSS, patX]]
   	   then print "[[0,y],[s(0),y],[s(s(x)),y]] covering Nat over 0, s.\n"
   	   else print "[[0,y],[s(0),y],[s(s(x)),y]] not covering Nat over 0, s.???\n"


   (* dpat 򤼤 reduce Ǥ롼뽸 *)
   fun selectRulesForDpat dpat cPat cPatReducible rs =
       let val target = L.mapPartial (fn i => let val (l,r) = L.nth (rs, i)
					      in case Subst.match dpat l of
						     SOME sigma => SOME (i,l,r,sigma)
						   | NONE => NONE
					      end)
				     (L.tabulate (L.length rs, fn x=>x))
	   val rs' = L.map (fn (_,l,r,_) => (l,r)) target
	   val num = L.map (fn (i,_,_,_) => i) target
	   val patlist = L.map (fn (_,_,_,sigma) => VM.listItems sigma) target
(******
	   val _ = print "(... "
	   val _ = print (LU.toStringCommaCurly Term.toString cPat)
	   val _ = print "...)\n"
	   val _ = print "[:"
	   val _ = print (LU.toStringSemicolonSquare
	   		      (fn pat => LU.toStringCommaCurly Term.toString pat)
	   		      patlist)
	   val _ = print ":]\n"
******)
   	   val count = ref 0
   	   fun newVar () = (count := (!count) + 1; Var.fromStringAndInt ("x",!count))
   	   fun newSortedVar ty = Term.Var (newVar (), ty)
       in if null patlist
	  then NONE
	  else let val len = L.length (hd patlist) 
		   val varps = L.map (fn p => newSortedVar (Term.sortOfTerm p)) (hd patlist) 
		   val psFromReducible = LU.mapAppend (fn n => let val ty = Term.sortOfTerm (L.nth (varps, n))
								   val cPatForTy = L.filter (fn p => Sort.equal (ty, Term.sortOfTerm p))
											    cPatReducible
							       in L.map (fn p=> LU.replaceNth (varps,n,p)) cPatForTy
							       end)
                                                      (L.tabulate (len, fn x => x))
(*****
		   val _ = print "[..."
		   val _ = print (LU.toStringSemicolonSquare
		   		      (fn pat => LU.toStringCommaCurly Term.toString pat)
		   		      psFromReducible)
		   val _ = print "...]\n"
*****)
		   val cands = LU.powerlist (L.tabulate (L.length patlist, fn x=>x))
		   (* val cPatLen = LU.allCombinations (L.tabulate (len,fn x=>cPat)) *)
		   fun check ns = let val ps = L.map (fn l => L.nth (patlist,l)) ns
				      (* val _ = print "["  *)
				      (* val _ = print (LU.toStringSemicolonSquare  *)
				      (* 			 (fn pat => LU.toStringCommaCurly Term.toString pat)  *)
				      (* 			 ps)  *)
				      (* val _ = print "]\n"  *)
				  in coveringPatternSequence cPat (ps @ psFromReducible)
				  end
		   val allAns = L.filter check cands
(*
		  val _ = print (LU.toStringSemicolonSquare (fn ns  => LU.toStringCommaCurly Int.toString ns) cands) 
		  val _ = print (LU.toStringSemicolonSquare (fn ns  => LU.toStringCommaCurly Int.toString ns) allAns) 
*)
		   val minAns = LU.takeMinimals (fn (xs,ys)=>  if LU.subseteq (xs,ys) 
							       then if LU.subseteq (ys,xs) then SOME EQUAL
								    else SOME LESS
							       else if LU.subseteq (ys,xs) then SOME GREATER
							       else NONE)
						allAns
	       in (*  SOME (L.map (fn xs => L.map (fn x=>L.nth (num,x)) xs) minAns) *)
		   if null minAns  (*** corrected(?)  2016/0629 ****)
		  then NONE
		  else  SOME (L.map (fn xs => L.map (fn x=>L.nth (num,x)) xs) minAns)
	       end
       end


   fun  getDefined W P = let val comp = complementOfPattern W P
			     fun matchsub p t = L.exists (fn s => isSome (Subst.match p s)) (Term.properSubterms t)
			     fun check [] ans = ans
			       | check (p::ps) ans =
				 check (L.filter (fn t => not (matchsub p t)) ps) (p::(L.filter (fn t => not (matchsub p t)) ans))
			 in check comp []
			 end

   val R = let val _ = rdSortedRules declNat ["plus(0,?y:Nat) -> ?y:Nat",
					      "plus(s(?x:Nat),?y:Nat) -> s(plus(?x:Nat,?y:Nat))",
					      "plus(?x:Nat,0) -> ?x:Nat",
					      "plus(?x:Nat,s(?y:Nat)) -> s(plus(?x:Nat,?y:Nat))"]
	       val R = rdSortedRules declNat ["plus(0,?y:Nat) -> ?y:Nat",
					      "plus(s(0),?y:Nat) -> s(plus(?x:Nat,?y:Nat))",
					      "plus(?x:Nat,0) -> ?x:Nat",
					      "s(s(?x:Nat)) -> ?x:Nat",
					      "plus(?x:Nat,s(?y:Nat)) -> s(plus(?x:Nat,?y:Nat))"]
	   in R end

   (* val W = declsToPattern declNat *)
   (* val _ = PU.println ("Nat pattern: " ^ (LU.toStringCommaSquare Term.toStringWithSort W)) *)
   (* val _ = PU.println ("Nat pattern: " ^ (LU.toStringCommaSquare Term.toStringWithSort PNat)) *)
   (* val cPat = getConstructor (W, getLHS R, [])  *)
   (* val _ = PU.println ("Constructor pattern: " ^ (LU.toStringCommaSquare Term.toString cPat)) *)


   fun partRulesAndIndThm (decls:Term.decl list) R = 
       let (* val _ = print (Trs.prRules R) *)
	   val (linR,nlinR) = L.partition (fn (l,_) => Term.isLinearTerm l) R
	   val W = declsToPattern decls 
	   val cPat = getConstructor2 (W, getLHS linR, []) 
	   val cSet = Term.funSetInTerms cPat
	   val cfs = FunSet.listItems cSet
	   val (clinR,dlinR) = L.partition (fn (l,_) => FS.isSubset (Term.funSetInTerm l, cSet))
					   linR
	   val cPatReducible = L.map (fn (l,_) => l) clinR
	   val _ = PU.println ("Constructor pattern: " ^ (LU.toStringCommaSquare Term.toString cPat))
	   val dPat = getDefined W cPat
	   val _ = PU.println ("Defined pattern: " ^ (LU.toStringCommaSquare Term.toString dPat))
	   val mins =  L.map (fn dpat => 
				 let (* val _ = PU.println ("dpattern: " ^ (Term.toString dpat)) *)
				     (* val _ = PU.println ("min sets of rules:") *)
				 in case selectRulesForDpat dpat cPat cPatReducible dlinR of
					NONE => []
				      | SOME nums => let fun prns ns = print (Trs.prRules (L.map (fn n=> L.nth (dlinR,n)) ns))
							 val _ = L.app prns nums
						     in nums end
				 end) dPat
	   val candidates = 
	       if L.exists null mins
	       then (PU.println "Failure: not quasi-reducible"; [])
	       else let val answers = L.map (fn ns => ListMergeSort.sort Int.> (LU.mapAppend (fn xs=>xs) ns)) (LU.allCombinations mins)
			val nums = L.tabulate (L.length dlinR, fn x=>x)
			fun getConj ns = LU.differenceByAll (nums, ns)
			fun prPartition (rs,conj) = let val _ = PU.println "Rule part & Conj Part:"
							val _ = print (Trs.prRules rs)
							val _ = print (Trs.prRules conj)
						    in () end
		    in L.map (fn ns => let val rs = L.map (fn n => L.nth (dlinR,n)) ns
					   val ms = getConj ns
					   val conj = L.map (fn n => L.nth (dlinR,n)) ms
					   val _ = prPartition (rs,conj)
				       in (clinR @ rs, conj @ nlinR) end) answers
		    end

       in (cPat,dPat,candidates)
       end


   val _  =  partRulesAndIndThm declNat R

   fun gcrSolver decls R = 
       let val _ = print (Trs.prRules R)
	   val W = declsToPattern decls 
	   val cPat = getConstructor (W, getLHS R, []) 
	   val _ = PU.println ("Constructor pattern: " ^ (LU.toStringCommaSquare Term.toString cPat))
	   val dPat = getDefined W cPat
	   val _ = PU.println ("Defined pattern: " ^ (LU.toStringCommaSquare Term.toString dPat))
	   val cands = partRulesAndIndThm decls R
       in Cr.Unknown
       end

   fun partRulesAndIndThmFSCD (decls:Term.decl list) R = 
       let val _ = print "*** Computating Strongly Quasi-Reducible Parts ***\n"
	   val _ = PU.println ("TRS:")
	   val _ = print (Trs.prRules R)

           (* non-left-linear rules  conjecture part  *)
	   val (rules00,conj00) = L.partition (fn (l,r) => (Term.isLinearTerm l
							    andalso not (Subst.matchSubterm l r))
							   orelse Term.isGroundTerm l) R

           (* conj0 κᤷ롼κդ˽ФƤʤؿ椬 conjecture Ǥʤ *)
	   val fset00 = Term.funSetInTerms (L.map (fn (l,r)=>l) rules00)
	   val (conj0a, conj0b) = L.partition (fn (l,r) => FS.isSubset (Term.funSetInTerms [l,r], fset00)) conj00
	   val (rules0,conj0) = (rules00 @conj0b, conj0a)

	   (* val _ = PU.println ("conj0:") *)
	   (* val _ = print (Trs.prRules conj0) *)

	   val W = declsToPattern decls 

	   val cPat0 = let val tops = L.mapPartial (fn (l,_) => Term.funRootOfTerm l) rules0
		       in L.filter (fn pat => not (LU.member'
						       Fun.equal
						       (valOf (Term.funRootOfTerm pat))
						       tops))
				   W
		       end
	   val cPat = let val cPat1 = TimeUtil.timeLimitRun (IntInf.fromInt 1) getConstructor3 (W, getLHS rules0, [])
			  (* val _ = print (LU.toStringCommaCurly Term.toString cPat1) *)
		      in LU.union' Term.equal (cPat0,cPat1)
		      end handle TimeLimit.TimeOut => let
			      val _ = PU.println "constructor detection timeout; switch to an approximation."
			  in cPat0
			  end

	   val cSet = Term.funSetInTerms cPat
	   val cfs = FunSet.listItems cSet
	   val _ = PU.println ("Constructors: " ^ (LU.toStringCommaCurly Fun.toString cfs))
			      
	   val dPat = getDefined W cPat
	   val dSet = Term.funSetInTerms dPat
	   val dfs = FunSet.listItems dSet
	   val _ = PU.println ("Defined function symbols: " ^ (LU.toStringCommaCurly Fun.toString dfs))

 	   val (Crules,Drules) = L.partition (fn (l,r) => FS.isSubset (Term.funSetInTerm l, cSet)
						  andalso FS.isSubset (Term.funSetInTerm r, cSet))
				     rules0

	   val _ = PU.println ("Constructor subsystem:")
	   val _ = print (Trs.prRules Crules)
	   val cPatReducible = L.filter Term.isLinearTerm (L.map (fn (l,_) => l) Crules)

            (* R = rules0 U conj0,  rules0 = Crules0 U Drules, conj0 = conj0a U conj0b, 
               Crules = Crules0 U conj0a, Drules = CS U nonCS *)
	   val (CS,nonCS) = L.partition (fn (l,_) =>  Term.isFun l 
						      andalso Term.isLinearTerm l
						      andalso FunSet.isSubset (Term.funSetInTerms (Term.argsOfTerm l), cSet))
					Drules  

	   (* val _ = PU.println ("nonCS:") *)
	   (* val _ = print (Trs.prRules nonCS) *)

	   val mins =  L.map (fn dpat => 
				 let (* val _ = PU.println ("dpattern: " ^ (Term.toString dpat))  *)
				 (*     val _ = PU.println ("min sets of rules:") *)
				 in case selectRulesForDpat dpat cPat cPatReducible CS of
					NONE => []
				      | SOME nums => let fun prns ns = print (Trs.prRules (L.map (fn n=> L.nth (CS,n)) ns))
							(* val _ = L.app prns nums *)
						     in nums end
				 end) dPat

	   val candidates = 
	       if L.exists null mins
	       then (PU.println "no strongly quasi-reducible parts found"; [])
	       else let val answers = L.map (fn ns => ListMergeSort.sort Int.> (LU.mapAppend (fn xs=>xs) ns)) (LU.allCombinations mins)
			val nums = L.tabulate (L.length CS, fn x=>x)
			fun getConj ns = LU.differenceByAll (nums, ns)
			fun prPartition (rpart,cpart) = let val _ = PU.println "Rule part & Conj Part:"
							val _ = print (Trs.prRules rpart)
							val _ = print (Trs.prRules cpart)
						    in () end
		    in L.map (fn ns => let val rs = L.map (fn n => L.nth (CS,n)) ns
					   val ms = getConj ns
					   val conj = L.map (fn n => L.nth (CS,n)) ms
					   val _ = prPartition (Crules @ rs, conj @ nonCS @ conj0) 
				       in (Crules @ rs, conj @ nonCS @ conj0) 
				       end) 
			     answers
		    end

       in (cPat,dPat,cPatReducible,candidates)
       end


   fun partRulesAndIndThmIWC (decls:Term.decl list) R = 
       let val _ = PU.print "*** Computating Strongly Quasi-Reducible Parts ***\n"
	   val _ = PU.println ("TRS:")
	   val _ = PU.print (Trs.prRules R)

           (* non-left-linear rules  conjecture part  *)
	   val (rules00,conj00) = L.partition (fn (l,r) => (Term.isLinearTerm l
							    andalso not (Subst.matchSubterm l r))
							   orelse Term.isGroundTerm l) R

           (* conj0 κᤷ롼κդ˽ФƤʤؿ椬 conjecture Ǥʤ *)
	   val fset00 = Term.funSetInTerms (L.map (fn (l,r)=>l) rules00)
	   val (conj0a, conj0b) = L.partition (fn (l,r) => FS.isSubset (Term.funSetInTerms [l,r], fset00)) conj00
	   val (rules0,conj0) = (rules00 @conj0b, conj0a)

	   (* val _ = PU.println ("conj0:") *)
	   (* val _ = print (Trs.prRules conj0) *)

	   val W = declsToPattern decls 

	   val cPat0 = let val tops = L.mapPartial (fn (l,_) => Term.funRootOfTerm l) rules0
		       in L.filter (fn pat => not (LU.member'
						       Fun.equal
						       (valOf (Term.funRootOfTerm pat))
						       tops))
				   W
		       end

(*	   val cPat = let val cPat1 = TimeUtil.timeLimitRun (IntInf.fromInt 1) getConstructor3 (W, getLHS rules0, []) *)
	   val cPat = let val cPat1 = TimeUtil.timeLimitRun (IntInf.fromInt 1) getConstructor3 (W, getLHS R,  [])
	   		  (* val _ = print (LU.toStringCommaCurly Term.toString cPat1) *)
	   	      in LU.union' Term.equal (cPat0,cPat1)
	   	      end handle TimeLimit.TimeOut => let
	   		      val _ = PU.println "constructor detection timeout; switch to an approximation."
	   		  in cPat0
	   		  end

(*	   val cPat = cPat0 *)


	   val cSet = Term.funSetInTerms cPat
	   val cfs = FunSet.listItems cSet
	   val _ = PU.println ("Constructors: " ^ (LU.toStringCommaCurly Fun.toString cfs))
			      
	   val dPat = getDefined W cPat
	   val dSet = Term.funSetInTerms dPat
	   val dfs = FunSet.listItems dSet
	   val _ = PU.println ("Defined function symbols: " ^ (LU.toStringCommaCurly Fun.toString dfs))

 	   val (Crules,Drules) = L.partition (fn (l,r) => FS.isSubset (Term.funSetInTerm l, cSet)
						  andalso FS.isSubset (Term.funSetInTerm r, cSet))
				     rules0

	   val _ = PU.println ("Constructor subsystem:")
	   val _ = PU.print (Trs.prRules Crules)
	   val cPatReducible = L.filter Term.isLinearTerm (L.map (fn (l,_) => l) Crules)

            (* R = rules0 U conj0,  rules0 = Crules0 U Drules, conj0 = conj0a U conj0b, 
               Crules = Crules0 U conj0a, Drules = CS U nonCS *)
	   val (CS,nonCS) = L.partition (fn (l,_) =>  Term.isFun l 
						      andalso Term.isLinearTerm l
						      andalso FunSet.isSubset (Term.funSetInTerms (Term.argsOfTerm l), cSet))
					Drules  

	   (* val _ = PU.println ("nonCS:") *)
	   (* val _ = print (Trs.prRules nonCS) *)

	   val minisatPath = "./minisat"
	   val tmpDir = "."
	   fun findSupplements dpat = 
	       let  val patNeeded = Term.argsOfTerm dpat
		    val rulesForDpat = L.mapPartial (fn (l,r) => if isSome (Subst.match dpat l) then SOME (l,r) else NONE)  CS
		    val lhsForDpat = L.map (fn (l,r)=>l) rulesForDpat
		    val patlist = L.map  Term.argsOfTerm lhsForDpat
		    val  subtracted = patSubtraction cPat ([patNeeded],patlist)
		    val f = case Term.funRootOfTerm dpat of 
				SOME f => f | NONE => (print "Error: no fun root???\n"; raise GcrError)
		    val ty = Term.sortOfTerm  dpat
		    val shortOf = L.map (fn ts =>  Term.Fun (f, ts,ty))  subtracted 

		    val _ = PU.println  ("Required Patterns: "^ (LU.toStringCommaCurly Term.toString shortOf))

		    val numOfSteps = 1
		    fun getCand l = L.mapPartial (fn r => if isSome (Subst.match l r) then NONE else SOME (l, r))
						 (TermSet.listItems (Rewrite.manyStepsReductSet R numOfSteps l))
		    val candidates = LU.allCombinations (L.map  getCand shortOf)
		    val _ = PU.println ("Candidates for "^ (Term.toString dpat))
		    val _ = List.app (fn rs => PU.print (Trs.prRules rs)) candidates
		    val opt2 = { useQuasi=false, useLex = false, useMul = true, useAf = false }
	       in L.find (fn cand2 => isSome (PoSolver.poSolverForGcr minisatPath tmpDir opt2 (CS @ cand2,[])))
		   	 (L.map (fn cand => rulesForDpat@cand) candidates)
	       end

	   val mins0 =  L.map (fn dpat => 
				 let val _ = PU.println ("dpattern: " ^ (Term.toString dpat))  (* *)
				     val _ = PU.println ("min sets of rules:")  (* *)
				 in case selectRulesForDpat dpat cPat cPatReducible CS of
					NONE => ([], findSupplements dpat)
				      | SOME nums => let fun prns ns = PU.print (Trs.prRules (L.map (fn n=> L.nth (CS,n)) ns))
							 val _ = PU.println "aaa"
							 val _ = L.app prns nums 
						     in (nums, NONE) end
				 end) dPat

	   val candidates = 
	       if L.exists (fn (ns,rs) => null ns andalso not (isSome rs)) mins0
	       then (PU.println "no strongly quasi-reducible parts found"; [])
	       else let val mins = L.mapPartial (fn (u,v) => if null u then NONE else SOME u) mins0
			val additions = LU.mapAppend (fn x=>x) (L.mapPartial (fn (u,v) => v) mins0)
			val _ = PU.println "added rules:"
			val _ = PU.print (Trs.prRules additions)
			val answers = L.map (fn ns => ListMergeSort.sort Int.> (LU.mapAppend (fn xs=>xs) ns)) (LU.allCombinations mins)
			val _ = PU.println "bbb"
			val nums = L.tabulate (L.length CS, fn x=>x)
			val _ = PU.println "ccc"
			fun getConj ns = LU.differenceByAll (nums, ns)
			val _ = PU.println "ddd"
			fun prPartition (rpart,cpart) = let val _ = PU.println "Rule part & Conj Part:"
							val _ = PU.print (Trs.prRules rpart)
							val _ = PU.print (Trs.prRules cpart)
						    in () end
		    in L.map (fn ns => let val rs = L.map (fn n => L.nth (CS,n)) ns
					   val ms = getConj ns
					   val conj = L.map (fn n => L.nth (CS,n)) ms
					   val _ = prPartition (Crules @ rs, conj @ nonCS @ conj0) 
				       in (Crules @ rs @ additions, conj @ nonCS @ conj0) 
				       end) 
			     answers
		    end

       in (cPat,dPat,cPatReducible,candidates)
       end




   end (* of local *)
   
   end; (* of structure Gcr *)
