(******************************************************************************
 * file: rwtools/sprint/ind.sml
 * description: rewriting induction
 * author: AOTO Takahito
 * 
 * Copyright (c) 2007-2015, Toyama&Aoto Laboratory, Tohoku University
 ******************************************************************************)

signature IND = 
sig
    type opt = {  useE2:bool (* use expand2 *) 
		, useQB:bool (* use quasi-basic subterm *)
		, useDC:bool (* use divergence critic *)
		, useSDC:bool (* use sound divergence critic *)
		, useSG:bool (* use sound generalization *)
		, useMG:bool (* use maximal generalization *)
		, useESG:bool (* use extended sound generalization *)
		, useDP:bool (* use decompose and disproof *)
		, useTP:bool (* use term partition *)
		, useRS:bool (* use reset *)
		, useLS:bool (* use lemma suggestion *)
	       }

    val indsimple: (Term.term * Term.term -> bool)
		   ->  (Term.term * Term.term -> bool)
		   -> opt
		   -> (Term.decl list * Term.term list * Term.term list)
		   -> Trs.trs
		   -> (Term.term * Term.term) list 
		   -> ProverResult.prover_result

    val indsimpleFSCD: (Term.term * Term.term -> bool)
		   ->  (Term.term * Term.term -> bool)
		   -> opt
		   -> (Term.decl list * Term.term list * Term.term list * Term.term list)
		   -> Trs.trs
		   -> (Term.term * Term.term) list 
		   -> ProverResult.prover_result

    val ind: (int -> Term.term * Term.term -> bool)
	     -> (int -> Term.term * Term.term -> bool)
	     -> int list
	     -> (Term.term * Term.term -> bool)
	     -> opt
	     -> Term.decl list
	     -> Trs.trs
	     -> Sg.argTables 
	     -> (Term.term * Term.term) list 
	     -> ProverResult.prover_result

    val runDebug: bool ref

end;

structure Ind : IND = 
struct

    type opt = {  useE2:bool (* use expand2 *) 
	    , useQB:bool (* use quasi-basic subterm *)
	    , useDC:bool (* use divergence critic *)
	    , useSDC:bool (* use sound divergence critic *)
	    , useSG:bool (* use sound generalization *)
	    , useMG:bool (* use maximal generalization *)
	    , useESG:bool (* use extended sound generalization *)
	    , useDP:bool (* use decompose and disproof *)
	    , useTP:bool (* use term partition *)
	    , useRS:bool (* use reset *)
	    , useLS:bool (* use lemma suggestion *)
	   }


    local
	open Term
	open Trs
	structure F = Fun
	structure V = Var
	structure S = Sort
	structure A = Atom
	structure L = List
	structure LP = ListPair
	structure LU = ListUtil
	structure Ar = Array
	structure AS = AtomSet
	structure FS = FunSet
	structure VS = VarSet
	structure VM = VarMap
	structure O = Order
	structure TS = TermSet
	fun arrayToList ar = List.tabulate (Array.length ar, fn n  => Array.sub (ar,n))
	fun mapAppend f xs = List.foldr (fn (x,ys) => List.@(f x, ys)) [] xs
	fun mapAppendTermSet f xset = TS.foldr (fn (x,yset) => TS.union (f x, yset)) TS.empty xset
	fun member x ys = isSome (List.find (fn y => x = y) ys)
	fun notMember x ys = not (member x ys)
	fun listMinus (xs,ys) =  L.filter (fn x => notMember x ys) xs
	fun replaceNth (xs,i,y) = (List.take (xs,i)) @ [y] @ (List.drop (xs,i+1))
	fun exceptNth n [] = []
	  | exceptNth n (x::xs) = 
            if n = 0 
            then xs
            else (x::(exceptNth (n-1) xs))
	fun changeSide xs = L.map (fn (l,r) => (r,l)) xs
	open PrintUtil
    in

    exception IndError of string
    exception IndSuccess
    exception IndFailure

    val trial = false
    val useSubProof = false

    datatype partialOrder = GT | LT | EQ | IC
					   
    val runDebug = ref false : bool ref
    fun debug f = if !runDebug then f () else ()

    (* $BEy<0$NA*Br(B: size $B$N>.$5$$$b$N$rA*Br$9$k(B *)
    fun selectAnEquationToExpand [] = (print "eq set empty"; raise IndError "selectAnEquationToExpand")
      | selectAnEquationToExpand (e::es) = 
	let
	    fun size (l,r) = (termSize l) + (termSize r)

	    fun select (i,e) [] acc = (e,acc)
	      | select (i,e) (d::ds) acc = 
		let val j = size d
		in if j < i 
		   then select (j,d) ds (e::acc)
		   else select (i,e) ds (d::acc)
		end
	in
	    select (size e,e) es []
	end

    (* $B4pK\ItJ,9`(B *)

    fun isBasicTerm (trs:trs) (Var _) = false
      | isBasicTerm (trs:trs) (Fun (f,ts,_)) = 
	if not (FS.member (#DSymSet trs, f))
	then false
	else FS.isSubset(funSetInTerms ts, #CSymSet trs)

    fun isBasicTermByPattern (cPat,dPat) (Var _) = false
      | isBasicTermByPattern (cPat,dPat) (t as (Fun (f,ts,_))) = 
	L.exists (fn dpat =>
		     case Subst.match dpat t of
			 SOME sigma => L.all (fn (_,v) => L.all (fn c => not (isSome (Subst.match c v))) dPat)
					      (VM.listItemsi sigma)
		       | NONE => false) dPat

    fun basicSubtermOCs (trs:trs) term = 
	let 
	    val ans = Term.collectSubtermOCs (isBasicTerm trs) term
	    val _ = debug (fn _ => print ("Basic Ocs of " ^ (Term.toString term) ^ "\n"));
	    val _ = debug (fn _ => print (List.foldr (fn (p,str) => (Term.prPosition p) ^ ",") "\n" ans));
	in
	    ans
	end

    fun basicSubtermOCsByPattern (cPat,dPat) term = 
	let 
	    val ans = Term.collectSubtermOCs (isBasicTermByPattern (cPat,dPat)) term
	    val _ = debug (fn _ => print ("Basic Ocs of " ^ (Term.toString term) ^ "\n"));
	    val _ = debug (fn _ => print (List.foldr (fn (p,str) => (Term.prPosition p) ^ ",") "\n" ans));
	in
	    ans
	end

(*     fun isQuasiBasicTerm (trs:trs) (Var _) = false *)
(*       | isQuasiBasicTerm (trs:trs) (term as Fun (f,ts,_)) =  *)
(* 	if not (FS.member (#DSymSet trs, f)) *)
(* 	then false *)
(* 	else *)
(* 	    let val rs = #Rules trs *)
(* 		val cSymSet = #CSymSet trs *)
(* 		val dSymSet = #DSymSet trs *)
(* 		val capVsetOp = Term.basicCap (cSymSet,dSymSet) term *)
(* 	    in *)
(* 		if not (isSome capVsetOp) *)
(* 		then false *)
(* 		else *)
(* 		    let val (cap,vset) = valOf capVsetOp *)
(* 			val mgus = List.mapPartial (fn (l,r) => Subst.disjointUnify cap l) rs *)
(* 		    in  *)
(* 			if null mgus *)
(* 			then false *)
(* 			else List.all  *)
(* 				 (fn (sigma,_) =>  *)
(* 				     not (VS.exists *)
(* 					      (fn v => VM.inDomain (sigma, v) *)
(* 						       andalso (isFun (valOf (VM.find (sigma,v))))) *)
(* 					      vset)) *)
(* 				 mgus *)
(* 		    end *)
(* 	    end *)

    fun isQuasiBasicTerm (trs:trs) (Var _) = false
      | isQuasiBasicTerm (trs:trs) (term as Fun (f,ts,_)) = 
	if not (FS.member (#DSymSet trs, f))
	then false
	else
	    let val rs = #Rules trs
		val fRules = L.filter (fn (l,_) => case funRootOfTerm l of
						    SOME g => Fun.equal (f,g)
						  | NONE => false)
				      rs
		val cSymSet = #CSymSet trs
		val dSymSet = #DSymSet trs
		fun allMaximalPathToDSym _ (Var _) = []
		  | allMaximalPathToDSym p (Fun (f,ts,_)) 
		    = if FS.member (dSymSet, f)
		      then [rev p]
		      else mapAppend 
			   (fn i => allMaximalPathToDSym (i+1::p) (L.nth (ts,i)))
			   (L.tabulate (length ts, fn x => x))
		val allPaths = mapAppend 
				   (fn i => allMaximalPathToDSym [i+1] (L.nth (ts,i)))
				   (L.tabulate (length ts, fn x => x))
		fun checkPathCond _ (Var _) = true
		  | checkPathCond [] (Fun (f,ts,_)) = false
		  | checkPathCond (i::p) (Fun (f,ts,_)) = 
		    if i-1 < length ts
		    then checkPathCond p (L.nth (ts, i-1))
		    else false
	    in
		L.all 
		    (fn p => L.all (fn (l,_) => checkPathCond p l) fRules)
		    allPaths
	    end

   fun isQuasiBasicTermNonFree (cPat,dPat,cPatReducible) rules (Var _) = false
     | isQuasiBasicTermNonFree (cPat,dPat,cPatReducible) rules (Fun (f,ts,ty)) = 
       let 
	   val cPatReducible2 = L.filter Term.isLinearTerm cPatReducible
           (* dummy $B$NDj?t$N@8@.(B *)
	   val c_counter = ref 0
	   fun dummyConstOfType ty = let val _ = c_counter := (1 + !c_counter)
					 val f = Fun.fromString ("_dummyConst_" ^ (Int.toString (!c_counter)))
				     in Term.mkFunTerm (f,[],ty) end

	   fun pcap u  =  if Term.isVar u 
			  then u
			  else if L.exists (fn pat=> isSome (Subst.match pat u)) (dPat@cPatReducible2)
			  then dummyConstOfType (Term.sortOfTerm u)
			  else Term.mkFunTerm (valOf (Term.funRootOfTerm u), L.map pcap (Term.argsOfTerm u), Term.sortOfTerm u)

           (* $B0z?t$N(B pcat $B$N%j%9%H(B *)
	   val ps = L.map pcap ts
	   val cap = Fun (f, ps, ty)

           (* cap $B$KC10l2=$G$-$k:8JU$N%j%9%H(B *)
	   val patlist = L.mapPartial (fn (l,_) => if Subst.disjointUnify (l,cap)
						   then SOME (Term.argsOfTerm l) else NONE)
				      rules

	   val len = L.length ps
   	   val count = ref 0
   	   fun newVar () = (count := (!count) + 1; Var.fromStringAndInt ("x",!count))
   	   fun newSortedVar ty = Term.Var (newVar (), ty)
	   val varps = L.map (fn p => newSortedVar (Term.sortOfTerm p)) ps
	   val psFromReducible = LU.mapAppend (fn n => let val ty = Term.sortOfTerm (L.nth (varps, n))
							   val cPatForTy = L.filter (fn p => Sort.equal (ty, Term.sortOfTerm p))
										    cPatReducible2
						       in L.map (fn p=> LU.replaceNth (varps,n,p)) cPatForTy
						       end)
                                              (L.tabulate (len, fn x => x))
       in Gcr.patCovered cPat ([ps],patlist @ psFromReducible)
       end

    fun quasiBasicSubtermOCsNonFree (cPat,dPat,cPatReducible) rules term = 
	let 
	    val ans = Term.collectSubtermOCs 
			  (isQuasiBasicTermNonFree (cPat,dPat,cPatReducible) rules) term
	    val _ = debug (fn _ => print ("Quasi Basic Ocs of " ^ (Term.toString term) ^ "\n"))
	    val _ = debug (fn _ => print (List.foldr (fn (p,str) => (Term.prPosition p) ^ ",") "\n" ans))
	in
	    ans
	end

    fun quasiBasicSubtermOCs (trs:trs) term = 
	let 
	    val ans = Term.collectSubtermOCs (isQuasiBasicTerm trs) term
	    val _ = debug (fn _ => print ("Quasi Basic Ocs of " ^ (Term.toString term) ^ "\n"))
	    val _ = debug (fn _ => print (List.foldr (fn (p,str) => (Term.prPosition p) ^ ",") "\n" ans))
	in
	    ans
	end

    (* Expand $B5,B'(B: (l,r) $B$N:8JU$N(B pos $B$K=P8=$9$kItJ,9`$r(B expand $B$9$k(B *)
    fun expand rs (l,r) pos =
	let val ((l',r')::rs') = Trs.renameRules ((l,r)::rs)
	in L.mapPartial
	       (fn (x,y) =>
		   case (Subst.unify (valOf (subterm pos l')) x)
		    of SOME sigma => 
		       SOME (Subst.applySubst sigma (valOf (replaceSubterm l' pos y)), 
			     Subst.applySubst sigma r')
		     | NONE => NONE)
	       rs'
	end

    (* Expand $B5,B'(B: (l,r) $B$N:8JU$N(B pos $B$K=P8=$9$kItJ,9`$r(B expand $B$9$k(B *)
   (*  non-free constructor version *)
    fun expandNonFree rules cPat (lhs,rhs) pos = 
	let val clist = L.map (fn t => valOf (Term.funRootOfTerm t)) cPat
	    val qred = valOf (Term.subterm pos lhs)
	    val (root,args,ty) = (valOf (Term.funRootOfTerm qred), Term.argsOfTerm qred, Term.sortOfTerm qred)

	    val len = L.length args
	    val (crules,drules) = L.partition (fn (l2,r2) => 
						  let val X = FS.listItems (Term.funSetInTerm l2)
						  in L.all (fn f => LU.member' Fun.equal f clist) X
						  end) rules
	    val varlist  = L.tabulate (len, fn n => Var.fromStringAndInt ("x",n))
	    val varTerms = LP.map Term.mkVarTerm (varlist, L.map Term.sortOfTerm args)
	    val origin = Term.mkFunTerm (root,varTerms,ty)

	    val ((origin',_)::crules') = Trs.renameRules ((origin,origin)::crules)
	    val varTerms' = Term.argsOfTerm origin'

	    val extraRules = LU.mapAppend (fn (l2,r2) => L.tabulate (len, 
								     fn n => 
									(Term.mkFunTerm (root, LU.replaceNth (varTerms',n,l2), ty),
									 Term.mkFunTerm (root, LU.replaceNth (varTerms',n,r2), ty))))
					  crules'

	    val ((lhs',rhs')::generatedRules) = Trs.renameRules ((lhs,rhs)::rules @ extraRules)
	    val rules' = L.take (generatedRules,L.length rules)
	    val extraRules' = L.drop (generatedRules,L.length rules)

	    val qred' = valOf (Term.subterm pos lhs')


	    val rules2' = L.filter (fn (l,_) => isSome (Subst.unify qred' l)) rules'
	    val extraRules2' = L.filter (fn (l,_) => isSome (Subst.unify qred' l)) extraRules'

	    val patlistR = L.map (fn (l,_) => Term.argsOfTerm l) rules2'


	    fun dropExtra cand = 
		let val canDrop = L.filter (fn x => let val candex = LU.exceptNth (cand,x)
							val patlistX = L.map (fn (l,_) => Term.argsOfTerm l) candex
						    in Gcr.coveringPatternSequence cPat (patlistR@patlistX)
						    end) 
					   (L.tabulate (L.length cand, fn x=>x))
		in if null canDrop
		   then cand
		   else dropExtra (LU.exceptNth (cand, hd canDrop))
		end

	    val result = L.mapPartial
			      (fn (x,y) =>
				  case (Subst.unify qred' x)
				   of SOME sigma => 
				      SOME (Subst.applySubst sigma (valOf (replaceSubterm lhs' pos y)), 
					    Subst.applySubst sigma rhs')
				    | NONE => NONE)
			      (rules' @ (dropExtra extraRules'))
	in 
	    result
	end

    (* $BE83+0LCV$NA*Br(B: $BJQ?t(Bindex$B$N(Bmaximum$B$,0lHV>.$5$$0LCV(B(1$BHV:G8e$KE83+$5$l$?0LCV(B)$B$rA*Br$9$k(B *)
    fun selectAPositionToExpand [] _ = 
	(print "positions to expand empty"; raise IndError "selectAPositionToExpand")
      | selectAPositionToExpand (p::ps) t = 
	let
	    fun year p = case (Term.subterm p t) of
			     NONE => (print "not a subterm pos"; 
				      raise IndError "selectAPositionToExpand")
			   | SOME u => let val vset = varSetInTerm u 
				       in VS.foldr (fn ((_,j),i) => Int.max (i,j)) 0 vset
				       end

	    fun select (i,p) [] = p
	      | select (i,p) (q::qs) = 
		let val j = year q
		in if j < i 
		   then select (j,q) qs
		   else select (i,p) qs
		end
	in
	    select (year p,p) ps
	end


    fun infExpand trs (opt:opt) (l,r) =
	let val ps = if (#useQB opt)
		     then quasiBasicSubtermOCs trs l
		     else basicSubtermOCs trs l
	    val rules = #Rules trs
	in if null ps 
	   then NONE
	   else let val p = selectAPositionToExpand ps l
		in SOME (expand rules (l,r) p)
		end
	end

    fun infExpandByPattern (cPat,dPat,rules) (opt:opt) (l,r) =
	let val ps = (* if (#useQB opt)
		     then quasiBasicSubtermOCs (cPat,dPat) l
		     else *) basicSubtermOCsByPattern (cPat,dPat) l
	in if null ps 
	   then NONE
	   else let val p = selectAPositionToExpand ps l
(*		in SOME (expand rules (l,r) p) *)
		in SOME (expandNonFree rules cPat (l,r) p)
		end
	end

    fun infExpandByPatternNonFree (cPat,dPat,cPatReducible) rules (l,r) = 
	let val ps = quasiBasicSubtermOCsNonFree (cPat,dPat,cPatReducible) rules l
	in if null ps 
	   then NONE
	   else let val p = selectAPositionToExpand ps l
		in SOME (expand rules (l,r) p)
		end
	end

    fun infExpandByPatternNonFree2 (cPat,dPat,cPatReducible) rules (l,r) = 
	let val ps = quasiBasicSubtermOCsNonFree (cPat,dPat,cPatReducible) rules l
	in if null ps 
	   then NONE
	   else let val p = selectAPositionToExpand ps l
		in SOME (expandNonFree rules cPat (l,r) p)
		end
	end


    fun infExpand2 trs (opt:opt) (l,r) = 
	let val ps1 = if (#useQB opt)
		      then quasiBasicSubtermOCs trs l
		      else basicSubtermOCs trs l
	    val ps2 = if (#useQB opt)
		      then quasiBasicSubtermOCs trs r
		      else basicSubtermOCs trs r
	    val rules = #Rules trs
	    fun isSameEq (l,r) (l',r') = 
		Subst.identicalModuloRenamingRule (l,r) (l',r')
		orelse Subst.identicalModuloRenamingRule (l,r) (r',l')
	    fun elimDup es = 
		let  fun elimDupSub [] ans = rev ans
		       | elimDupSub ((l,r)::es) ans = 
			 elimDupSub (L.filter (not o (isSameEq (l,r))) es) 
				    ((l,r)::ans)
		in elimDupSub es []
		end

	in if (null ps1) orelse (null ps2)
	   then NONE
	   else let val es = expand rules (l,r) (hd ps1)
		in
		    SOME (elimDup 
			      (L.map (fn (x,y) => (y,x)) 
				     (mapAppend (fn (x,y) => expand rules (y,x) (hd ps2)) es)))
		end
	end

    fun infExpand2ByPattern (cPat,dPat,rules) (opt:opt) (l,r) =
	let val ps1 = (* if (#useQB opt)
		      then quasiBasicSubtermOCs trs l
		      else *) basicSubtermOCsByPattern (cPat,dPat) l
	    val ps2 = (* if (#useQB opt)
		      then quasiBasicSubtermOCs trs r
		      else *) basicSubtermOCsByPattern (cPat,dPat) r
	    fun isSameEq (l,r) (l',r') = 
		Subst.identicalModuloRenamingRule (l,r) (l',r')
		orelse Subst.identicalModuloRenamingRule (l,r) (r',l')
	    fun elimDup es = 
		let  fun elimDupSub [] ans = rev ans
		       | elimDupSub ((l,r)::es) ans = 
			 elimDupSub (L.filter (not o (isSameEq (l,r))) es) 
				    ((l,r)::ans)
		in elimDupSub es []
		end

	in if (null ps1) orelse (null ps2)
	   then NONE
	   else let val es = expand rules (l,r) (hd ps1)
		in
		    SOME (elimDup 
			      (L.map (fn (x,y) => (y,x)) 
				     (mapAppend (fn (x,y) => expand rules (y,x) (hd ps2)) es)))
		end
	end


    fun nonVarProperSubtermSet t  = 
	L.foldr (fn (x,xs) => TS.add (xs,x)) TS.empty (Term.nonVarProperSubterms t)

    (* Delete $B5,B'!':o=|$K@.8y$9$l$P(B NONE$B!$<:GT$7$?$i(B SOME $B85$NEy<0(B *)
    (*  K^\approx $B$NJQ7A$O$^$@F~$l$F$J$$!%!%!%(B*)
    fun delete (G,K,E) (x,y) = 
	let
	    (* val _ = print "Delete\n" *)
	    val xx = Rewrite.reductSet G x
	    fun choose [] ss = raise IndError "delete"
	      | choose ts [] = raise IndError "delete"
	      | choose ts ss = 
		let val ts' = if (length ts) <= 3 then ts else (L.take (ts,3))
		    val ss' = if (length ss) <= 3 then ss else (L.take (ss,3))
	            val cand = 
			ListXProd.mapX 
			    (fn (s,t) =>  
				let val ssub = nonVarProperSubtermSet s
				    val tsub = nonVarProperSubtermSet t
				    val len = L.length (TS.listItems (TS.intersection (ssub,tsub)))
				in (len,(s,t))
				end)
			    (ts',ss')
		    fun select [] _ (k,e) = e
		      | select ((k,e)::xs) l accum =
			if k > l
			then select xs k (k,e)
			else select xs l accum
		in 
		    select (tl cand) 0 (hd cand)
		end
	in
	    if TS.member (xx,y)
	    then NONE
	    else 
		let
		    (* val xxx = mapAppendTermSet (Rewrite.oneStepReductSet K) xx *)
		    val xxx = mapAppendTermSet (Rewrite.parallelOneStepReductSet K) xx
		    val xxx' = mapAppendTermSet (Rewrite.parallelOneStepReductSet E) xx
		    val yy = Rewrite.reductSet G y
		in
		    if not (TS.isEmpty (TS.intersection (xxx, yy)))
		       orelse not (TS.isEmpty (TS.intersection (xxx', yy)))
		    then NONE
		    else
			let 
			    (* 	   val yyy = mapAppendTermSet (Rewrite.oneStepReductSet K) yy *)
			    val yyy = mapAppendTermSet (Rewrite.parallelOneStepReductSet K) yy
			    val yyy' = mapAppendTermSet (Rewrite.parallelOneStepReductSet E) yy
			in
			    if not (TS.isEmpty (TS.intersection (xx, yyy)))
			       orelse not (TS.isEmpty (TS.intersection (xx, yyy')))
			    then NONE
			    else SOME (x,y) 
                            (* $B6&DL$NItJ,9`$,B?$$BeI=85$rJV$9(B, x,y $B$r%G%U%)%k%H$K$7$?$$$N$G@hF,$KDI2C(B *)
(* 			    else SOME (choose (x::(TS.listItems xx)) (y::(TS.listItems yy))) *)
			end
		end
	end

    (* Decompose $B5,B'(B *)
    fun decompose cSymSet (x,y) = 
	case (x,y) of 
	    (Fun (f,ts,_), Fun (g,ss,_)) 
	    => if Fun.equal (f,g) andalso FS.member (cSymSet,f)
	       then mapAppend (decompose cSymSet) (ListPair.zip (ts,ss))
	       else [(x,y)]
	  | (_,_) => [(x,y)]
		     
    (* Simplify $B5,B'(B *)
    fun simplify isOrdered isEquivalent (H,G,K) (l,r) = 
	let 
(* 	    val _ = print "s" *)
(* 	    val _ = print (Trs.prRules H) *)
(* 	    val _ = print (Trs.prEqs G) *)
(* 	    val _ = print (Trs.prEqs K) *)
	    val (st1,nf1) = Rewrite.statusModuloClinf isOrdered isEquivalent H G K l
(*	    val _ = print "lhs" *)
            val (st2,nf2) = Rewrite.statusModuloClinf isOrdered isEquivalent H G K r
(*	    val _ = print "rhs" *)
	in (st1 orelse st2, (nf1,nf2))
	end

    (* kSimpfily: K-instance $B$b;H$C$F(B1step reduction $B$9$k(B *)
    fun kSimplify isOrdered isEquivalent (H,G,K) (l,r) = 
	let 
	    val Kord = 
		L.filter (fn (l,r)=> VS.isSubset (Term.varSetInTerm r, Term.varSetInTerm l)) K
	    val Keq = 
		L.filter (fn (l,r)=> VS.equal (Term.varSetInTerm l, Term.varSetInTerm r)) K

	    (* val _ = print "k1" *)

	    fun clos x = TermSet.listItems (Rewrite.constrainedReductSet
						isEquivalent G Keq x)

	    fun check [] = NONE
	      | check (x::xs) = case Rewrite.oneStepConstrainedRewritePartial
					 isOrdered H Kord x
				 of SOME y => SOME y
				  | NONE => check xs

	    val closL = clos l
	    val closR = clos r

	    (* val _ = print "k2" *)

	    fun check2 [] z = NONE
	      | check2 (x::ys) z  = let val xs = TS.listItems (Rewrite.reductSet K x)
				    in 
					case L.find (fn x' => isOrdered (z,x')) xs  of
					    SOME x' => SOME x'
					  | NONE => check2 ys z
				    end
	in
	    case (check closL, check closR) of
		(SOME newl, SOME newr) => SOME (newl,newr)
	      | (SOME newl, NONE) => SOME (newl,r)
	      | (NONE, SOME newr) => SOME (l,newr)
	      | (NONE, NONE) => NONE
	(**************** 
	     Simplify-C $B$r;H$C$?(B Simplification$B!'5U$NJU$h$j>.$5$/$J$l$P4JLs$7$F$b$h$$(B 
			--- $B;~4V$,$+$+$k$N$G$d$i$J$$!%!%!%(B
	 if isOrdered (l,r) orelse isOrdered (r,l)
	 then NONE
	 else case check2 closL r of
		  SOME newl => (print "check2a!!"; SOME (newl,r))
		| NONE => case check2 closR l of
			      SOME newr => (print "check2b!!"; SOME (l,newr))
			    | NONE => NONE
	 ***************)
	end

    (* Simplify2 $B5,B'(B : K $B$r(B1$BEY$@$1$7$+;H$o$J$$(B *)
    fun simplify2 isOrdered isEquivalent (H,G,K) (l,r) = 
	let val (st,e1) = simplify isOrdered isEquivalent (H,G,[]) (l,r)
	    (* val _ = print "k0" *)
	in case kSimplify isOrdered isEquivalent (H,G,K) e1
	    (*	   of SOME e2 => simplify isOrdered isEquivalent (H,G,K) e2 *)
 	    of SOME e2 => simplify isOrdered isEquivalent (H,G,[]) e2
	     | NONE => (st,e1)
	end


    fun infSimplify isOrdered isEquivalent (trs:Trs.trs) (opt:opt) (es,hs,gs,ks) = 
	let val rules = #Rules trs
	    val cSymSet = #CSymSet trs
	    val H = rules @ hs
	    val G = gs @ (changeSide gs)
	    val K = ks @ (changeSide ks)

	    (* val _ = print (Trs.prRules es) *)
	    (* val _ = print (Trs.prRules H) *)
	    (* val _ = print (Trs.prEqs G) *)
	    (* val _ = print (Trs.prEqs K) *)

	    fun infSimplifySub [] acc = acc
	      | infSimplifySub (e::es) acc = 
		let val E = es @ (changeSide es) 
			    @ acc @ (changeSide acc)

		    (* val _ = print ("e:" ^ (Trs.prEq e) ^ "\n") *)
		    (* val _ = print ("E:" ^ (Trs.prEqs E)) *)

		    val (st,e') =  simplify2 isOrdered isEquivalent (H,G,K) e  

		    (* val _ = print ("e':" ^ (Trs.prEq e') ^ "\n") *)

		    val ee = L.mapPartial (delete (G,K,E))
					  (if (#useDP opt)
					   then (decompose cSymSet e')
					   else [e'])
		    (* val _ = print ("ee:" ^ (Trs.prEqs ee)) *)
		in
		    infSimplifySub es (ee @ acc)
		end
	in
	    infSimplifySub es []
	end




    (* $B7rA40lHL2=K!(B **)
    fun infSG2 (argTables:Sg.argTables) eqs (l,r)  =
	let fun again e = case Sg.soundGeneralization argTables e of
			      NONE => e
			    | SOME e' => again e'
	in
	    case (Sg.soundGeneralization argTables (l,r)) of 
		NONE => NONE
	      | SOME e' => SOME (again e')
	end       

    fun infSG (argTables:Sg.argTables) eqs (l,r)  =
	let fun try (u,v) = 
		case Sg.soundGeneralization argTables (u,v) of
		    SOME e' => SOME e'
		  | NONE => 
		    if null eqs 
		    then NONE
		    else
			let
			    val variants = ListXProd.mapX
					       (fn (x,y) => (x,y))
					       (TS.listItems (Rewrite.developOneStepReductSet eqs u),
						TS.listItems (Rewrite.developOneStepReductSet eqs v))
			    fun tryAll [] = NONE
			      | tryAll (x::xs) = 
				case Sg.soundGeneralization argTables x of
				    SOME x' => SOME x'
				  | NONE => tryAll xs
			in tryAll variants
			end
	    fun again e = case try e of 
			      NONE => e 
			    | SOME e' => again e'
	in
	    case try (l,r) of
		NONE => NONE
	      | SOME e' => SOME (again e')
	end       


    (* Disproof $B5,B'(B *)
    fun infDisproof (trs:Trs.trs) (l,r) = 
	let val rules = #Rules trs
	    val cSymSet = #CSymSet trs
	    val vsetl = varSetInTerm l
	    val vsetr = varSetInTerm r
	in
	    case (l,r) of
		(Var (x,_), Var (y,_))=>  if Var.equal (x,y) 
					  then NONE
					  else SOME (l,r)
	      | (Var (x,_), Fun (g,_,_))=> if FS.member (cSymSet, g)
					      orelse not (VS.member (vsetr, x))
					   then SOME (l,r)
					   else NONE
	      | (Fun (f,_,_), Var (y,_))=> if FS.member (cSymSet, f)
					      orelse not (VS.member (vsetl, y))
					   then SOME (l,r)
					   else NONE
	      | (Fun (f,_,_), Fun (g,_,_)) => if FS.member (cSymSet,f)
						 andalso FS.member (cSymSet,g)
						 andalso not (Fun.equal (f,g))
					      then SOME (l,r)
					      else NONE
	end

    (* $B9`J,3dK!(B **)
    fun infTermPartition (trs:Trs.trs) (argTables:Sg.argTables) (l,r) = 
	let fun infTP (l,r) = 
	    let val sigMap = Trs.signatureInRules (#Rules trs)
		(* val _ = print ("Try TP to " ^ (prEq (l,r)) ^ "\n") *)
			
 		val seqL = Sg.termPartition sigMap trs argTables l 
		(* val _ =  L.map (fn (_,s,t) => print ( "tp of L...<" ^ (Term.toString s) *)
		(* 				      ^  " ---- " ^ (Term.toString t)  ^ ">\n")) *)
		(* 	       seqL *)
			       
 		val seqR = Sg.termPartition sigMap trs argTables r 
		(* val _ =  L.map (fn (_,s,t) => print ( "tp of R...<" ^ (Term.toString s) *)
		(* 				      ^  " --- " ^ (Term.toString t)  ^ ">\n")) *)
		(* 	       seqR *)
		val cand = ListXProd.mapX
			       (fn ((_,s1,s2),(_,t1,t2)) => ((s1,t1),(s2,t2)))
			       (seqL,seqR)

		fun process e [] = SOME e
		  | process e (((s1,t1),(s2,t2))::xs) =
		    let val eq1 = Term.equal (s1,t1) 
			val eq2 = Term.equal (s2,t2) 
		    in
			if eq1 andalso eq2
			then NONE  (* nothing to prove remained *)
			else if eq1
			then
			    infTP (s2,t2)
			else if eq2
			then
			    infTP (s1,t1)
			else process e xs
		end
	    in
		process (l,r) cand
	    end
	in
	    case infTP (l,r) of 
		SOME (l2,r2) => if Term.equal (l,l2) andalso  Term.equal (r,r2) 
				then NONE
				else SOME [(l2,r2)]
	      | NONE => SOME []  
	end

    (* $B7rA4H/;64UDjK!(B *)
    (* $BF1$8JdBj$,(B2$B2s@8@.$5$l$k%A%'%C%/$7$F$$$k(B *)
    fun infSoundDivergenceCritic _ [] = NONE
      | infSoundDivergenceCritic _ (lr::[]) = NONE
      | infSoundDivergenceCritic sg (lr::rs) = 
	let exception Found of Term.term * Term.term
		fun sdc skel rule  = 
			case Dc.speculateSoundLemma skel rule of
				SOME e2 => sg e2  | NONE => NONE
	    fun speculate e  = 
		case sdc e lr of
		    SOME lem1 => 
		    (List.app  (fn e2 => 
				   case Dc.speculateLemma e2 e of
				       SOME lem2 => 
				       if Subst.identicalModuloRenamingRule lem2 lem1
				       then raise Found lem1
				       else ()
				     | NONE => ())
			       rs)
		  | NONE => ()
	in
	    (L.app speculate rs; NONE)
	    handle (Found ans) => (print ("Add Lemma (SDC): " ^ (IOMstrs.prEq ans) ^ "\n");
				   SOME ans)
	end

    (* $BH/;64UDjK!(B **)
    (* $BF1$8JdBj$,(B2$B2s@8@.$5$l$k%A%'%C%/$7$F$$$k(B => $B%A%'%C%/$7$J$$(B *)
    fun infDivergenceCritic [] = NONE
      | infDivergenceCritic (lr::[]) = NONE
      | infDivergenceCritic (lr::rs) = 
	let exception Found of (Term.term * Term.term)
	    fun speculate e  = 
		case Dc.speculateLemma e lr of
 		    SOME lem1 => raise Found lem1
(* 		    SOME lem1 => *)
(* 		    (List.app  *)
(* 			 (fn e2 =>  *)
(* 			     case Dc.speculateLemma e2 e of *)
(* 				 SOME lem2 =>  *)
(* 				 if Subst.identicalModuloRenamingRule lem2 lem1 *)
(* 				 then raise Found lem1 *)
(* 				 else () *)
(* 			       | NONE => ()) *)
(* 			 rs)  *)
		  | NONE => ()
	in 
	    (L.app speculate rs; NONE)
	    handle (Found ans) => (print ("Add Lemma (DC): " ^ (IOMstrs.prEq ans) ^ "\n");
				   SOME ans)
	end


    (* $B0lHL2=(B *)
    fun checkGeneralizationBySubproof (trs:Trs.trs) (l,r) = 
	let val dSymSet = #DSymSet trs
	    fun rootDefined (Fun (f,_,_)) = FS.member (dSymSet,f)
	      | rootDefined (Var (x,_)) = false

            (* $B:,5-9f$,Dj5A5-9f$+$DHsJQ?t$JItJ,9`$G!$N>JU$K4^$^$l$F$$$k$b$N$r0lHL2=$9$k(B *)
	    val lset = L.foldr (fn (t,xset) => if rootDefined t 
					       then TS.add (xset,t)
					       else xset)
			       TS.empty (Term.nonVarSubterms l)
	    val rset = L.foldr (fn (t,xset) => if rootDefined t 
					       then TS.add (xset,t)
					       else xset)
			       TS.empty (Term.nonVarSubterms r)
		       
            (* $B$3$3$G!$(Bk-equal $B$J$b$N$r0lHL2=$G$-$k$H$h$$!%!%!%(B*)
	    val both = TS.intersection (lset,rset)

	    fun replaceTerm (u,x) (t as (Var _)) = 
		if Term.equal (u,t) then x else t
	      | replaceTerm (u,x) (t as (Fun (f,ts,a))) = 
		if Term.equal (u,t)
		then x
		else Fun (f,L.map (replaceTerm (u,x)) ts,a)

	in
	    if (TS.isEmpty both)
	    then NONE
	    else 
		let
		    val sorted = ListMergeSort.sort
				     (fn (x,y) => isASubterm x y)
				     (TS.listItems both)

		    val _ = debug (fn _ => print (PrintUtil.prList Term.toString sorted))

		    val m  = Term.maxVarIndexInTerms [l,r]

		    val vs = ListPair.map 
				 (fn (i,u) => Var ((A.atom "x", m + i), sortOfTerm u))
				 (L.tabulate (length sorted, fn i=>i+1), sorted)


		    val l' = ListPair.foldl
				 (fn (u,x,t) => replaceTerm (u,x) t)
				 l
				 (sorted,vs)

		    val r' = ListPair.foldl
				 (fn (u,x,t) => replaceTerm (u,x) t)
				 r
				 (sorted,vs)

		    val _ = print ("***Enter Subproof of " ^ (prEq (l',r')) ^ "\n")

		in
		    SOME (l',r')
		end
	end


    fun equalEq (e1,e2) = Subst.identicalModuloRenamingRule e1 e2

    fun equalEq2 (e1,(l,r)) = Subst.identicalModuloRenamingRule e1 (l,r)
			     orelse
			     Subst.identicalModuloRenamingRule e1 (r,l)

    fun equalEqs (es1,es2) = 
	(length es1 = length es2)
	andalso
	null (LU.differenceByOne' equalEq (es1,es2))

    (***************************************)
    (******* $B=q$-49$(5"G<K!$N5"G<E*<jB3$-(B [Aoto,SCSS] ********)
    (***************************************)

    local
	exception Disproved
	exception Proved
	exception Stopped
	val sizeLimit = 40
	val countLimit = 30
	val lastTheorems = ref []
    in
    fun indsimple isOrdered isEquivalent (opt:opt) (decls:decl list,cPat,dPat) (trs:trs) conjectures  = 
	let 
	    val rules = #Rules trs

	    fun indStep (es,hs,gs,ks) count = 
		(* es: equations to prove *)
		(* hs: rewrite rules + orientable hypothesis *)
		(* gs: hypothesis with equivalent sides *)
		(* ks: hypothesis with incomparable sides *)
		let 
		    (* $B>uBVI=<((B *)
		    val _ = print ("STEP " ^ (Int.toString (count)) ^ "\n")
		    val _ = print ("ES:\n" ^ (Trs.prEqs es))
		    val _ = print ("HS:\n" ^ (Trs.prRules hs))
		    val _ = if (#useE2 opt) then print ("GS:\n" ^ (Trs.prEqs gs)) else ()
		    val _ = if (#useE2 opt) then print ("KS:\n" ^ (Trs.prEqs ks))  else ()

		    (* Simplification $B%9%F%C%W(B *)
		    val (es0,hs0,gs0,ks0) = applyInfSimplify (es,hs,gs,ks)
		    (* val _ = print ("...(end of applyInfSimplify)\n") *)

		    (* $B=*N;%A%'%C%/(B *)
		    val _ = if null es0 then raise Proved
			    else if count = countLimit then 
				(print "Stopped: the number of steps reaches to the limit.\n"; raise Stopped)
			    else ()

		    (* Expand $B%9%F%C%W(B *)
		    val (extra, (es1,hs1,gs1,ks1)) = applyInfExpand (es0,hs0,gs0,ks0) [] 

		    (* Simplification $B%9%F%C%W(B *)
		    val (es2,hs2,gs2,ks2) = applyInfSimplify  (es1,hs1,gs1,ks1)

		    val (new1,es3,hs3,gs3,ks3) = 
			if (#useDC opt) then
			    (* Divergence Critic *)
			    case infDivergenceCritic (List.nth ([es2,hs2,gs2,ks2],extra)) of
				SOME e => ([e],es2,hs2,gs2,ks2)
			      | NONE => ([],es2,hs2,gs2,ks2)
			else
			    ([],es2,hs2,gs2,ks2)

		in indStep (es3,hs3,gs3,ks3) (count + 1)
		end

	    and applyInfExpand (es0,hs,gs,ks) es1 = 	   
		(*  es1 $B$O$9$G$KE83+$r;n$7$?$b$N$N!$E83+$K<:GT$7$?Ey<0$N%j%9%H(B *)
		let 
		    (* val _ = print ("Expand\n") *)

		    (* $BE83+$G$-$kEy<0$,$J$$>l9g$O=*N;(B *)
		    val _ = if null es0 
			    then  (print ("No equation to expand\n"); raise Stopped)
			    else ()

		    val ((l,r),es2) = selectAnEquationToExpand es0
				      
		    (* val _ = print ("eq to expand: " ^ (Trs.prEq (l,r)) ^ "\n") *)

		    (* val _ = if isOrdered (l,r) *)
		    (* 	    then print "ordered\n" *)
		    (* 	    else print "not ordered\n" *)

		    val result = 
			if isOrdered (l,r)
			then case infExpandByPattern (cPat,dPat,rules) opt (l,r) of
				 SOME ee => SOME (1, (ee,[(l,r)],[],[]))
			       | NONE => NONE
			else if isOrdered (r,l)
			then case infExpandByPattern (cPat,dPat,rules) opt (r,l) of
				 SOME ee => SOME (1, (ee,[(r,l)],[],[]))
			       | NONE => NONE
			else if (#useE2 opt)
			then case infExpand2ByPattern (cPat,dPat,rules) opt (l,r) of
				 SOME ee => 
				 if isEquivalent (l,r)
				 then SOME (2, (ee,[],[(l,r)],[]))
				 else SOME (3, (ee,[],[],[(l,r)]))
			       | NONE => NONE
			else NONE

		    (* $B7P2aI=<((B *)
		    val _ = if isSome result
			    then (print ("Expand " ^ (Trs.prEq (l,r)) ^ "\n");
				  print (Trs.prEqs (#1 (#2 (valOf result)))))
			    else ()
				 
		in case result of
		       SOME (ex, (ee,eh,eg,ek)) => (ex,  (* $B$I$N>l=j$K(B I.H. $B$,F~$C$?$+(B *)
						    (ee @ es2 @ es1, 
						     eh @ hs, 
						     eg @ gs, 
						     ek @ ks))
		     | NONE => applyInfExpand (es2,hs,gs,ks) ((l,r)::es1) 
		(* $BE83+<:GT$N>l9g$OB>$NEy<0$NE83+$r;n$_$k(B *)
		end

	    and applyInfMG gs es = 
		if (#useMG opt)
		then  let
		    (* generalization *)
			val dSymSet = #DSymSet trs
			fun rootDefined (Fun (f,_,_)) = FS.member (dSymSet,f)
			  | rootDefined (Var (x,_)) = false

			fun equivalent t = (t,TS.addList (TS.empty,
							  t::Rewrite.rootRewriteAll gs t))

		    fun getGeneralization (u,v) = 
			let val us = L.filter (not o Term.isConstant) (Term.nonVarSubterms u)
			    val vs = L.filter (not o Term.isConstant) (Term.nonVarSubterms v)
			    (* val _ = print (LU.toStringCommaSquare Term.toString us) *)
			    (* val _ = print "\n" *)
			    (* val _ = print (LU.toStringCommaSquare Term.toString vs) *)
			    (* val _ = print "\n" *)

			    val uss = L.map equivalent us
			    val vss = L.map equivalent vs

			    fun checkEqual ((_,xset),(_,yset)) = 
				not (TS.isEmpty (TS.intersection (xset,yset)))

			    val vcoms = List.filter
					   (fn x => LU.member' checkEqual x uss)
					   vss

			    val ucoms = List.filter
					   (fn x => LU.member' checkEqual x vss)
					   uss

			    val coms = ListPair.map (fn ((x,_),(y,_)) => (x,y)) (ucoms,vcoms)
					   
			    val coms2 = List.filter
					    (fn (x,y) => 
						L.all (fn (x2,y2) => 
							  (not (Term.isAProperSubterm x x2)
							   andalso
							   not (Term.isAProperSubterm y y2)))
							   coms)
					    coms
			    (* val _ = print ("common:\n" ^ (LU.toStringCommaSquare  
							      (fn (x,y) => 
								  "<" ^ (Term.toString x)
								  ^ "," ^ (Term.toString y) ^ ">")
							      coms2)) *)

			    val max = Term.maxVarIndexInTerms [u,v]
			    val newVs = L.tabulate (L.length coms2, 
						 fn i => Var.fromStringAndInt ("x",max+i+1))

			    fun replace t [] [] = NONE
			      | replace t (u::us) (v::vs) =
				if Term.equal (t,u)
				then SOME (Var (v,sortOfTerm u))
				else replace t us vs

			    fun replaceTerm1 (Var x) = (Var x)
			      |	replaceTerm1 (t as (Fun (f,ts,ty))) =
				case replace t (L.map (fn (x,y) => x) coms2) newVs of
				    SOME v => v
				  | NONE => Fun (f,L.map replaceTerm1 ts, ty)

			    fun replaceTerm2 (Var x) = (Var x)
			      |	replaceTerm2 (t as (Fun (f,ts,ty))) =
				case replace t (L.map (fn (x,y) => y) coms2) newVs of
				    SOME v => v
				  | NONE => Fun (f,L.map replaceTerm2 ts, ty)

			    val genEq = (replaceTerm1 u,replaceTerm2 v)
			in 
			   if null coms2 
			   then NONE
			   else SOME genEq 
			end
		    in
			initialSimplification (L.mapPartial getGeneralization es)
		    end
		else []

            (* gs : variation $B$r:n$kF1CM<0(B
               es : generalization $B$NBP>]<0(B *)
	    and applyInfSG gs es = es

	    and applyInfSimplify (es,hs,gs,ks) = 
		let 
		    (* val _ = print ("Simplify" ^ "\n") *)

		    (* es $B$rC1=c2=(B *)
		    val es1 = infSimplify isOrdered isEquivalent trs opt (es,hs,gs,ks)

		    (* $B7rA40lHL2=K!$G!$Ey<0$r(B($B2DG=$J$i$P(B)$B0lHL2=(B *)
		    val es2 = applyInfSG gs es1
				   
		    (* $BHsDjM}$,4^$^$l$F$$$J$$$+%A%'%C%/(B *)
		    val es3 = if (#useDP opt)
                              then L.map (fn e => 
					     case infDisproof trs e
                                              of SOME e => (print ("Disproof: " ^ (prEq e) ^ "\n");
                                                            raise Disproved;
                                                            e)
                                               | NONE => e)
					 es2
                              else es2

		in 
		    (es3, hs, gs, ks)
		end

	    and initialSimplification es =
		let
		    (* $B7rA40lHL2=(B---$BC1=c2=$r$d$kA0$G$J$$$H0lHL2=$G$-$J$$>l9g$b$"$k$N$G(B $B!%!%!%(B *)
		    val es1 = applyInfSG [] es

		    (* $BC1=c2=(B + $B7rA40lHL2=(B + $BH?G}>ZL@(B *)
		    val (es2,_,_,_) = applyInfSimplify (es1,[],[],[])
		in
		    es2
		end 
	in  
	    indStep (conjectures,[],[],[]) 0
	end
	handle Disproved => ProverResult.NO
	     | Proved => ProverResult.YES
 	     | Stopped => ProverResult.MAYBE

    end

    (***************************************)
    (******* $B=q$-49$(5"G<K!$N5"G<E*<jB3$-(B [FSCD 2016] ********)
    (***************************************)

    local
	exception Disproved
	exception Proved
	exception Stopped
	val sizeLimit = 40
	val countLimit = 30
	val lastTheorems = ref []
	fun invRules rs = L.mapPartial (fn (l,r) => 
					   if VarSet.isSubset (Term.varSetInTerm l, Term.varSetInTerm r)
					   then SOME (r,l) else NONE) rs

    in
    fun indsimpleFSCD isOrdered isEquivalent (opt:opt) (decls:decl list,cPat,dPat,cPatReducible) (trs:trs) conjectures  = 
	let 
	    val rules = #Rules trs
	    val _ = print ("Rules:\n" ^ (Trs.prRules rules))
	    val _ = print ("Conjectures:\n" ^ (Trs.prEqs conjectures))

	    fun indStep (es,hs) count = 
		(* es: equations to prove *)
		(* hs: rewrite rules + hypothesis, ordered rewriting should be used  *)
		let 
		    (* $B>uBVI=<((B *)
		    val _ = print ("STEP " ^ (Int.toString (count)) ^ "\n")
		    val _ = print ("ES:\n" ^ (Trs.prEqs es))
		    val _ = print ("HS:\n" ^ (Trs.prRules hs))

		    (* Simplification $B%9%F%C%W(B *)
		    val (es0,hs0) = applyInfSimplify (es,hs) false
		    (* val _ = print ("...(end of applyInfSimplify)\n") *)
		    val _ = print ("ES0:\n" ^ (Trs.prEqs es0))
		    val _ = print ("HS0:\n" ^ (Trs.prRules hs0))

		    (* Delete $B%9%F%C%W(B *)
		    val (es1,hs1) = applyInfDelete (es0,hs0)
		    val _ = print ("ES1:\n" ^ (Trs.prEqs es1))
		    val _ = print ("HS1:\n" ^ (Trs.prRules hs1))

		    (* val _ = print ("...(end of applyInfDelete)\n") *)

		    (* $B=*N;%A%'%C%/(B *)
		    val _ = if null es1 then raise Proved
			    (* we use time limit instead *)
			    (* else if count = countLimit 
			    then (print "Stopped: the number of steps reaches to the limit.\n"; raise Stopped) *)
			    else ()

		    (* Expand $B%9%F%C%W(B *)
		    val (es2,hs2) = applyInfExpand (es1,hs1) [] 
		    val _ = print ("ES2:\n" ^ (Trs.prEqs es2))
		    val _ = print ("HS2:\n" ^ (Trs.prRules hs2))

		in indStep (es2,hs2) (count + 1)
		end

	    and applyInfExpand (es0,hs) es1 = 	   
		(*  es1 $B$O$9$G$KE83+$r;n$7$?$b$N$N!$E83+$K<:GT$7$?Ey<0$N%j%9%H(B *)
		let 
		    (* val _ = print ("Expand\n") *)

		    (* $BE83+$G$-$kEy<0$,$J$$>l9g$O=*N;(B *)
		    val _ = if null es0 
			    then  (print ("No equation to expand\n"); raise Stopped)
			    else ()

		    val ((l,r),es2) = selectAnEquationToExpand es0
				      
		    (* val _ = print ("eq to expand: " ^ (Trs.prEq (l,r)) ^ "\n") *)

		    (* val _ = if isOrdered (l,r) *)
		    (* 	    then print "ordered\n" *)
		    (* 	    else print "not ordered\n" *)

		    val result = 
			if isOrdered (l,r)
			then case infExpandByPattern (cPat,dPat,rules) opt (l,r) of (* FSCD 2016 *)
			    (* case infExpandByPatternNonFree2 (cPat,dPat,cPatReducible) rules (l,r) of *)
				 SOME ee => SOME (ee,(l,r))
			       | NONE => NONE
			else if isOrdered (r,l)
			then case infExpandByPattern (cPat,dPat,rules) opt (r,l) of (* FSCD 2016 *)
			    (* case infExpandByPatternNonFree2 (cPat,dPat,cPatReducible) rules (r,l) of *)
				 SOME ee => SOME (ee,(r,l))
			       | NONE => NONE
			else case infExpandByPattern (cPat,dPat,rules) opt (l,r) of  (* FSCD 2016 *)
			    (* case infExpandByPatternNonFree2 (cPat,dPat,cPatReducible) rules (l,r) of FSCD 2016 *)
				 SOME ee => SOME (ee,(l,r))
			       | NONE => (case infExpandByPattern (cPat,dPat,rules) opt (r,l) of (* FSCD 2016 *)
				   (* case infExpandByPatternNonFree2 (cPat,dPat,cPatReducible) rules (r,l) of *)
					     SOME ee => SOME (ee,(r,l))
					   | NONE => NONE)

		    (* $B7P2aI=<((B *)
		    val _ = case result of
				SOME (ee,(l1,r1)) => (print ("Expand " ^ (Trs.prEq (l1,r1)) ^ "\n");
						      print (Trs.prEqs ee))
			      | NONE => ()
				 
		in case result of
		       SOME (es3,(l1,r1)) => let val (es4,_) = applyInfSimplify (es3,hs) true
					     in (es4 @ es2 @ es1, (l1,r1)::hs)
					     end
		     | NONE => applyInfExpand (es2,hs) ((l,r)::es1) 
		(* $BE83+<:GT$N>l9g$OB>$NEy<0$NE83+$r;n$_$k(B *)
		end

	    and applyInfSimplify (es,hs) afterExpand = 
		(* afterExpand = true $B$J$i(B es $B$N:8JU$@$1$r(B $B%b%8%e%mIU$-$G4JLs2=(B *)
		let 
                    fun orderedRewriteOneStepStatus rs t =
			let val ts = L.filter (fn s=> isOrdered (t,s)) (Rewrite.oneStepReducts rs t)
			in if null ts then (false,t) else (true, hd ts)
			end

                    fun orderedRewriteStepsStatus rs t =
			let fun rewriteAsMany s = case orderedRewriteOneStepStatus rs s of
						      (false,s1) => s1
						    | (true,s1) => rewriteAsMany s1
			in case orderedRewriteOneStepStatus rs t of
			       (false,t1) => (false,t1)
			     | (true,t1) => (true,rewriteAsMany t1)
			end

                   (* $BIaDL$N=q$-49$((B(rs)$B$H=g=x=q$-49$((B(ks)$B$NN>J}(B *)
		    fun simpleAndOrderedRewriteStepsStatus ks rs t = 
			let fun rewriteAsMany s = let val s1 = Rewrite.linf rs s
						  in case orderedRewriteStepsStatus ks s1 of
							 (true,s2) => rewriteAsMany s2
						       | (false,s2) => s2
						  end
			    val t1 = Rewrite.linf rs t
			in case orderedRewriteStepsStatus ks t1 of
			       (true,t2) => (true, rewriteAsMany t2)
			     | (false,t2) => (false, t2)
			end

                   (* isEquivalence $B$N%b%8%e%m(B(gs)$B$N$b$H$G$N!$IaDL$N=q$-49$((B(rs)$B$H=g=x=q$-49$((B(ks)$B$NN>J}(B *)
                    fun moduloRewriteStepsStatus gs ks rs t =
			let val cands = (* $B$H$j$"$($:!$(Bmodulo $B$O(B1step $B$N$_(B *)
				L.filter (fn s=> isEquivalent (t,s)) (Rewrite.oneStepReducts gs t)
			    fun tryall [] = NONE
			      | tryall (u::us) = case simpleAndOrderedRewriteStepsStatus ks rs u of
						     (true,u1) => SOME u1
						   | (false,_) => tryall us
			in case tryall cands of
			       SOME ans => (true,ans)
			     | NONE => (false,t)
			end

                    (* fun rewriteTermWOmodulo t = *)
		    (* 	case simpleAndOrderedRewriteStepsStatus hs rules t of *)
		    (* 	    (true,t1) => rewriteTermWOmodulo t1 *)
		    (* 	  | (false,t2) => t2 *)

                    fun rewriteTermWithModulo t =
			case simpleAndOrderedRewriteStepsStatus hs rules t of
			    (true,t1) => rewriteTermWithModulo t1
			  | (false,t2) => let val hsBoth = hs @ (invRules hs)
					 in case moduloRewriteStepsStatus hsBoth hs rules t2 of
						   (true,t3) => rewriteTermWithModulo t3
						 | (false,t3) => t3
					 end

                    (* $B0lEY(BR0$B$+(BH$B$G=q$-$+$($?$i!$(Bmodulo $B$r;H$($k(B *)
                    fun rewriteTerm t =
			case simpleAndOrderedRewriteStepsStatus hs rules t of
			    (true,t1) => rewriteTermWithModulo t1
			  | (false,t2) => t2

		    (* val _ = print ("Simplify" ^ "\n") *)
		    (* fun isOriginalConjecture (l,r) =  *)
		    (* 	LU.member' (fn ((l1,r1),(l2,r2))=> Term.equal (l1,l2) andalso Term.equal (r1,r2)) *)
		    (* 		   (l,r) *)
		    (* 		   conjectures *)

		    (* val es1 = L.map (fn (l,r) =>  *)
                    (*                     (* $BM?$($i$l$?(B conjecture $B$N$H$-$O!$(Bmodulo $B$O;H$($J$$(B *) *)
		    (* 			if isOriginalConjecture (l,r) *)
		    (* 			then (rewriteTermWOmodulo l, rewriteTermWOmodulo r) *)
		    (* 			else (rewriteTerm l, rewriteTerm r)) *)
		    (* 		    es *)

		    val es1 = if afterExpand
			      then L.map (fn (l,r) => (rewriteTermWithModulo l, r)) es
			      else L.map (fn (l,r) => (rewriteTerm l, rewriteTerm r)) es

		in  (es1, hs)
		end

	    and applyInfDelete (es,hs) = 
		let val es1 = L.filter (fn (l,r) => not (Term.equal (l,r))) es
		    val es2 = L.filter (fn (l,r) => 
					   not (LU.member' Term.equal r (Rewrite.oneStepReducts hs l))
					   andalso not (LU.member' Term.equal l (Rewrite.oneStepReducts hs r))) 
				       es1
		in (es2,hs)
		end


	in  
	    indStep (conjectures,[]) 0
	end
	handle Disproved => ProverResult.NO
	     | Proved => ProverResult.YES
 	     | Stopped => ProverResult.MAYBE

    end







    (***************************************)
    (******* $B=q$-49$(5"G<%U%l!<%`%o!<%/$K4p$E$/5"G<E*<jB3$-(B ********)
    (***************************************)

    local
	exception Disproved
	exception Proved
	exception Stopped
	val sizeLimit = 40
	val lastTheorems = ref []

    in
    fun ind isOrdered isEquivalent orderParams isMonomorphicEq
	    (opt:opt) (decls:decl list) (trs:trs) (argTables:Sg.argTables)
	    conjectures  = 
	let 
            (* stat = 0 ==> lex , stat = 1 ==> mul for binary *)
	    fun checkOrder stat (l,r) = if isOrdered stat (l,r)
					then ([(l,r)],[],[])
					else if isOrdered stat (r,l)
					then ([(r,l)],[],[])
					else if isEquivalent stat (l,r)
					then ([],[(l,r)],[])
					else ([],[],[(l,r)])

	    fun checkOrderEqs stat [] = ([],[],[])
	      | checkOrderEqs stat ((l,r)::es) =
		let val (xs,ys,zs) = checkOrderEqs stat es
		in if isOrdered stat (l,r)
		   then ((l,r)::xs,ys,zs)
		   else if isOrdered stat (r,l)
		   then ((r,l)::xs,ys,zs)
		   else if isEquivalent stat (l,r)
		   then (xs,(l,r)::ys,zs)
		   else (xs,ys,(l,r)::zs)
		end
		
	    datatype goal = GOAL | NON_GOAL

	    fun isGoalProc (_,_,g,_,_,_) = (GOAL = g)

	    fun sizeOfEs [] = 0
	      | sizeOfEs ((l0,r0)::es) =
		List.foldl
		    (fn ((l,r),w) => Int.min (Term.termSize l + Term.termSize r, w))
		    (Term.termSize l0 + Term.termSize r0)
		    es

	    exception SkipIndStep

	    fun indStep (size,conj,isGoal,stat,count,(es,hs,gs,ks)) countLimit = 
		
		(* es: equations to prove *)
		(* hs: rewrite rules + orientable hypothesis *)
		(* gs: hypothesis with equivalent sides *)
		(* ks: hypothesis with incomparable sides *)

		let 
		    (* $B>uBVI=<((B *)
		    val _ = print ("STEP " ^ (Int.toString (count)))

		    val _  = print (if isGoal = GOAL
				    then " (Goal) "
				    else " (Non-Goal) ")

		    val _  = print (if stat = 0
				    then " (Specified)\n"
				    else if stat = 1
				    then " (Lex)\n"
				    else " (Mul)\n")

		    val _ = print ("ES:\n" ^ (Trs.prEqs es))
		    val _ = print ("HS:\n" ^ (Trs.prRules hs))
		    val _ = if (#useE2 opt) then print ("GS:\n" ^ (Trs.prEqs gs))
			    else ()
		    val _ = if (#useE2 opt) then print ("KS:\n" ^ (Trs.prEqs ks))
			    else ()

		    (* Simplification $B%9%F%C%W(B *)
		    val (es0,hs0,gs0,ks0) = applyInfSimplify stat (es,hs,gs,ks)


		    (* $B=*N;%A%'%C%/(B *)
		    val _ = if null es0 then raise Proved
			    else if count = countLimit then raise Stopped
			    else ()

		    (* simplify $B$,(B proper $B$K$"$C$?$i!$(Bexpand $B$r%9%-%C%W(B *)
		    val skip = not (equalEqs (es0,es))

		    val _  = if skip
			     then print "Expand skipped\n"
			     else ()

		    (* Expand $B%9%F%C%W(B *)
		    val (extra, (es1,hs1,gs1,ks1)) = 
			if skip
			then (0,(es0,hs0,gs0,ks0))
			else applyInfExpand stat (es0,hs0,gs0,ks0) [] 

		    (* val _ = print "aaa" *)

		    (* Simplification $B%9%F%C%W(B *)
		    val (es2,hs2,gs2,ks2) = 
			if skip
			then (es1,hs1,gs1,ks1)
			else applyInfSimplify stat (es1,hs1,gs1,ks1)

		    (* val _ = print "bbb" *)
		    (* val _ = print ("ES2:\n" ^ (Trs.prEqs es2)) *)

		    val (new1,es3,hs3,gs3,ks3) = 
			if skip
			then ([],es2,hs2,gs2,ks2)
			else 
			if (#useDC opt) then
			    (* Divergence Critic *)
			    case infDivergenceCritic (List.nth ([es2,hs2,gs2,ks2],extra)) of
				SOME e => ([e],es2,hs2,gs2,ks2)
			      | NONE => 
				if (#useSDC opt) then
				    (* Sound Divergence Critic *)
				    case infSoundDivergenceCritic 
					     (infSG argTables gs2)
					     (* (Sg.soundGeneralization argTables) *)
					     (List.nth ([es2,hs2,gs2,ks2],extra)) of
					SOME e => ([],e::es2,hs2,gs2,ks2)
				      | NONE => ([],es2,hs2,gs2,ks2)
				else 
				    ([],es2,hs2,gs2,ks2)
			else
			    if (#useSDC opt) then
				(* Sound Divergence Critic *)
				case infSoundDivergenceCritic 
					     (infSG argTables gs2)
					 (* (Sg.soundGeneralization argTables) *)
					 (List.nth ([es2,hs2,gs2,ks2],extra)) of
				    SOME e => ([],e::es2,hs2,gs2,ks2)
				  | NONE => ([],es2,hs2,gs2,ks2)
			    else
				([],es2,hs2,gs2,ks2)


		   val _ = print "ccc"
		   val _ = print ("new2:\n" ^ (Trs.prEqs new1))

		   val new2 = applyInfMG gs2 es3

		   val _ = print "cccc"
		   val _ = print ("new2:\n" ^ (Trs.prEqs new2))

		   val _ = print ("ES3:\n" ^ (Trs.prEqs es3))

		   val new = new1 @ new2

		   val count' = if skip 
				then count
				else (count + 1)
		in
		    (0, SOME (new,(sizeOfEs es3,conj,isGoal,stat,count',(es3,hs3,gs3,ks3))))
		end

	    and applyInfExpand stat (es0,hs,gs,ks) es1 = 	   
		(*  es1 $B$O$9$G$KE83+$r;n$7$?$b$N$N!$E83+$K<:GT$7$?Ey<0$N%j%9%H(B *)
		let 
		    (* $BE83+$G$-$kEy<0$,$J$$>l9g$O=*N;(B *)
		    val _ = if null es0 then raise Stopped
			    else ()

		    val ((l,r),es2) = selectAnEquationToExpand es0
				      
		    val result = 
			if isOrdered stat (l,r)
			then case infExpand trs opt (l,r) of
				 SOME ee => SOME (1, (ee,[(l,r)],[],[]))
			       | NONE => NONE
			else if isOrdered stat (r,l)
			then case infExpand trs opt (r,l) of
				 SOME ee => SOME (1, (ee,[(r,l)],[],[]))
			       | NONE => NONE
			else if (#useE2 opt)
			then case infExpand2 trs opt (l,r) of
				 SOME ee => 
				 if isEquivalent stat (l,r)
				 then SOME (2, (ee,[],[(l,r)],[]))
				 else SOME (3, (ee,[],[],[(l,r)]))
			       | NONE => NONE
			else NONE


		    (* $B7P2aI=<((B *)
		    val _ = if isSome result
			    then (print ("Expand " ^ (Trs.prEq (l,r)) ^ "\n");
				  print (Trs.prEqs (#1 (#2 (valOf result)))))
			    else ()
				 
		in case result of
		       SOME (ex, (ee,eh,eg,ek)) => (ex,  (* $B$I$N>l=j$K(B I.H. $B$,F~$C$?$+(B *)
						    (ee @ es2 @ es1, 
						     eh @ hs, 
						     eg @ gs, 
						     ek @ ks))
		     | NONE => applyInfExpand stat (es2,hs,gs,ks) ((l,r)::es1) 
		(* $BE83+<:GT$N>l9g$OB>$NEy<0$NE83+$r;n$_$k(B *)
		end

	    and applyInfMG gs es = 
		if (#useMG opt)
		then  let
		    (* generalization *)
			val dSymSet = #DSymSet trs
			fun rootDefined (Fun (f,_,_)) = FS.member (dSymSet,f)
			  | rootDefined (Var (x,_)) = false

			fun equivalent t = (t,TS.addList (TS.empty,
							  t::Rewrite.rootRewriteAll gs t))

		    fun getGeneralization (u,v) = 
			let val us = L.filter (not o Term.isConstant) (Term.nonVarSubterms u)
			    val vs = L.filter (not o Term.isConstant) (Term.nonVarSubterms v)
			    (* val _ = print (LU.toStringCommaSquare Term.toString us) *)
			    (* val _ = print "\n" *)
			    (* val _ = print (LU.toStringCommaSquare Term.toString vs) *)
			    (* val _ = print "\n" *)

			    val uss = L.map equivalent us
			    val vss = L.map equivalent vs

			    fun checkEqual ((_,xset),(_,yset)) = 
				not (TS.isEmpty (TS.intersection (xset,yset)))

			    val vcoms = List.filter
					   (fn x => LU.member' checkEqual x uss)
					   vss

			    val ucoms = List.filter
					   (fn x => LU.member' checkEqual x vss)
					   uss

			    val coms = ListPair.map (fn ((x,_),(y,_)) => (x,y)) (ucoms,vcoms)
					   
			    val coms2 = List.filter
					    (fn (x,y) => 
						L.all (fn (x2,y2) => 
							  (not (Term.isAProperSubterm x x2)
							   andalso
							   not (Term.isAProperSubterm y y2)))
							   coms)
					    coms
			    (* val _ = print ("common:\n" ^ (LU.toStringCommaSquare  
							      (fn (x,y) => 
								  "<" ^ (Term.toString x)
								  ^ "," ^ (Term.toString y) ^ ">")
							      coms2)) *)

			    val max = Term.maxVarIndexInTerms [u,v]
			    val newVs = L.tabulate (L.length coms2, 
						 fn i => Var.fromStringAndInt ("x",max+i+1))

			    fun replace t [] [] = NONE
			      | replace t (u::us) (v::vs) =
				if Term.equal (t,u)
				then SOME (Var (v,sortOfTerm u))
				else replace t us vs

			    fun replaceTerm1 (Var x) = (Var x)
			      |	replaceTerm1 (t as (Fun (f,ts,ty))) =
				case replace t (L.map (fn (x,y) => x) coms2) newVs of
				    SOME v => v
				  | NONE => Fun (f,L.map replaceTerm1 ts, ty)

			    fun replaceTerm2 (Var x) = (Var x)
			      |	replaceTerm2 (t as (Fun (f,ts,ty))) =
				case replace t (L.map (fn (x,y) => y) coms2) newVs of
				    SOME v => v
				  | NONE => Fun (f,L.map replaceTerm2 ts, ty)

			    val genEq = (replaceTerm1 u,replaceTerm2 v)
			in 
			   if null coms2 
			   then NONE
			   else SOME genEq 
			end
		    in
			initialSimplification (L.mapPartial getGeneralization es)
		    end
		else []

            (* gs : variation $B$r:n$kF1CM<0(B
               es : generalization $B$NBP>]<0(B *)
	    and applyInfSG gs es = 
		if (#useSG opt)
		then L.map 
			 (fn e => 
			     if isMonomorphicEq e
			     then (case (infSG argTables gs e)
				    of SOME e' => (print ("SG: " ^ (prEq e') ^ "\n");
						   print ("(from) " ^ (prEq e) ^ "\n");
						   e')
				     | NONE => e)
			     else e)
			 es
		else es


	    and applyInfSimplify stat (es,hs,gs,ks) = 
		let 
		    (* es $B$rC1=c2=(B *)
		    val es1 = infSimplify (isOrdered stat) (isEquivalent stat) trs opt (es,hs,gs,ks)
			      
		    (* val _ = print "a1" *)
		    val rules = #Rules trs
				
		    (* val _ = print "a2" *)
		    (* $B7rA40lHL2=K!$G!$Ey<0$r(B($B2DG=$J$i$P(B)$B0lHL2=(B *)
		    val es2 = applyInfSG gs es1
				   
		    (* val _ = print "a3"*)

		    (* $BHsDjM}$,4^$^$l$F$$$J$$$+%A%'%C%/(B *)
		    val es3 = if (#useDP opt)
                              then L.map (fn e => 
					     case infDisproof trs e
                                              of SOME e => (print ("Disproof: " ^ (prEq e) ^ "\n");
                                                            raise Disproved;
                                                            e)
                                               | NONE => e)
					 es2
                              else es2

		    (* val _ = print "a4" *)


		in 
		    (es3, hs, gs, ks)
		end

	    and initialSimplification es =
		let
		    (* $B7rA40lHL2=(B---$BC1=c2=$r$d$kA0$G$J$$$H0lHL2=$G$-$J$$>l9g$b$"$k$N$G(B $B!%!%!%(B *)
		    val es1 = applyInfSG [] es

		    (* $BC1=c2=(B + $B7rA40lHL2=(B + $BH?G}>ZL@(B *)
		    val (es2,_,_,_) = applyInfSimplify 0 (es1,[],[],[])
		in
		    es2
		end 


	    fun selectProcess [] = (print "selectProcess\n"; raise IndError "selectProcess")
	      | selectProcess (x::xs) = 
		let fun select (y as (size,_,_,_,_,_)) ys zs =
			if null ys
			then (y,zs)
			else let val (y2 as (size2,_,_,_,_,_)) = hd ys
			     in if size <= size2
				then select y (tl ys) (y2::zs)
				else select y2 (tl ys) (y::zs)
			     end
		in select x xs []
		end

	    fun isSameConj (es1,es2) =
		(List.length es1) = (List.length es2)
		andalso
		(ListPair.allEq
		     (fn ((l1,r1),(l2,r2)) => Term.equal (l1,l2) andalso Term.equal (r1,r2))
		     (es1,es2))

	    fun makeTPproc es = if (#useTP opt) 
				then LU.mapAppend
				     (fn eq => 
					     if isMonomorphicEq eq
					     then
					     case infTermPartition trs argTables eq of
						 SOME es' => 
						 (print ("term partition for " ^ (Trs.prEq eq) ^
							 ": " ^ (Trs.prEqs es') ^ "\n");
						  L.map (fn i=>
							    (sizeOfEs es',eq::es',NON_GOAL,0,1,(es',[],[],[])))
							orderParams)
					       | NONE => 
						 (print ("no term partition for " ^ (Trs.prEq eq) ^ "\n");
						  [])
					     else [])
 				     es
				else []


	    fun applyIndStep proc = indStep proc 0
		handle Disproved => (1,NONE)
		     | Proved => (2,NONE)
		     | Stopped => (3,NONE)

	    fun mergeTheoremsToProc thms (size,conj,isGoal,stat,count,(es,hs,gs,ks)) =
		let val (xs,ys,zs) = checkOrderEqs stat thms
		in (size,conj,isGoal,stat,count,
		    (es,
		     LU.union' equalEq (xs,hs),
		     LU.union' equalEq (ys,gs),
		     LU.union' equalEq2 (zs,ks)))
		end

	    fun mergeTheoremsToProcs thms procs = L.map (mergeTheoremsToProc thms) procs

	    fun containedEqs eqs (size,conj,isGoal,stat,count,(es,hs,gs,ks)) =
		L.all (fn eq => LU.member' equalEq eq eqs) conj

	    fun deleteProvedProcs es procs = L.filter (not o (containedEqs es)) procs

	    fun containEqs eqs (size,conj,isGoal,stat,count,(es,hs,gs,ks)) =
		L.all (fn eq => LU.member' equalEq eq conj) eqs

	    fun deleteDisprovedProcs es procs = L.filter (not o (containEqs es)) procs

	    fun resetProcess (theorems,examming,procs) =
		if equalEqs (theorems,!lastTheorems)
		then (print "Give up\n"; ProverResult.MAYBE)
		else let val _ = lastTheorems := theorems
			 val _ = print "Reset Process...\n"
			 val newProcs = LU.mapAppend
					    (fn (size,conj,isGoal,_,_,_) => 
						if isGoal = NON_GOAL
						then []
						else
						    L.map (fn i =>
							      (sizeOfEs conj,conj,
							       GOAL,i,1,(conj,[],[],[])))
							  orderParams)
					    procs
			 val newExamming = L.mapPartial
					   (fn (size,conj,isGoal,_,_,_) => 
					       if isGoal = NON_GOAL
					       then NONE
					       else SOME conj)
					   procs

		     in proceedOneProcess (theorems,newExamming,mergeTheoremsToProcs theorems newProcs)
		     end

	    and  proceedOneProcess (theorems,examming,procs) = 
		if List.all (not o isGoalProc) procs
		then ProverResult.YES
		else 
		    let val _ = print "start\n"
			val _ = print ("Num. of Goal procs: "
				       ^ Int.toString
				       (L.length (L.filter (fn (_,_,isGoal,_,_,_) => isGoal = GOAL) procs)))

			val _ = print (" / Num. of Non Goal procs: "
				       ^ Int.toString
				       (L.length (L.filter (fn (_,_,isGoal,_,_,_) => isGoal = NON_GOAL) procs))
				       ^ "\n")

			val (proc,rest) = selectProcess procs
			val (size,conj,isGoal,_,_,_) = proc
(*
			val _ = print ("number of processes: " 
				       ^ (Int.toString (List.length procs))
				       ^ "\n")
*)
			val _ = print ("size of the selected process: " 
				       ^ (Int.toString size)
				       ^ "\n")
		    in 
			if (#useRS opt) 
			   andalso size > sizeLimit
			then resetProcess (theorems,examming,procs)
			else 
			    case applyIndStep proc of
				(0,SOME (new,proc2)) => (let 
							    val unknowns = LU.differenceByAll' equalEq 
											       (new,theorems @ (changeSide theorems))
							    val reallyNew = L.filter
										(fn eq => not (LU.member' equalEqs [eq] examming))
										unknowns
							(* val _ = L.map (fn es => print (Trs.prEqs es)) examming *)
							in if null reallyNew
							   then proceedOneProcess (theorems,examming,proc2::rest)
							   else
							       let 
								   val _ = print ("New Lemmas:\n" ^ (Trs.prEqs reallyNew))
								   val newProcs = 
								       (LU.mapAppend (fn eq => 
											 L.map (fn i =>
												   (sizeOfEs [eq],[eq],NON_GOAL,i,1,([eq],[],[],[])))
											       orderParams)
										     reallyNew)
								       @ (makeTPproc reallyNew)
							       in proceedOneProcess (theorems,
										     examming @
										     (L.map (fn (_,x,_,_,_,_) => x) newProcs),
										     (mergeTheoremsToProcs theorems (proc2::newProcs))
										     @ rest)
							       end
							end)
			      | (1,_) => handleDisproved (conj,isGoal) 
							 (theorems,examming,rest)
			      | (2,_) => handleProved (conj,isGoal) 
						      (theorems,examming,rest)
			      | (3,_) => handleStopped (conj,isGoal) 
						       (theorems,examming,rest)
			      | _ => raise IndError "processOneProcess"
		    end

	    and handleProved (conj,isGoal) (theorems,examming,procs) =
		(print "proved!\n";
		if List.all (not o isGoalProc) procs
		then ProverResult.YES
		else let 
			val proper = LU.differenceByAll' equalEq (conj,theorems)
			 fun addEqs (size,conj,isGoal,stat,count,(es,hs,gs,ks)) =
			     let val (xs,ys,zs) = checkOrderEqs stat proper
			     in
				 (size,conj,isGoal,stat,count,(es,hs@xs,gs@ys,ks@zs))
			     end
			 val _ = print "New Theorems:\n"
			 val _ = print (Trs.prEqs proper)
			 val _ = print (Int.toString (L.length procs));
			 val _ = print ":"
			 val _ = print (Int.toString (L.length (deleteProvedProcs conj procs)))
		    in
			 proceedOneProcess (proper @ theorems,
					    LU.deleteAll' equalEqs proper examming,
					    L.map addEqs 
						  (deleteProvedProcs conj procs))
		     end)

	    and handleDisproved (conj,isGoal) (theorems,examming,procs) =
		(print "disproved!\n";
		if isGoal = GOAL
		then ProverResult.NO
		else (print (Trs.prEqs conj);
		      print (Int.toString (L.length procs));
		      print ":";
		      print (Int.toString (L.length (deleteDisprovedProcs conj procs)));
		      proceedOneProcess (theorems,examming,deleteDisprovedProcs conj procs)))

	    and handleStopped (conj,isGoal) (theorems,examming,procs) =
		(print "stopped!\n";
		if isGoal = GOAL
		then ProverResult.MAYBE
		else proceedOneProcess (theorems,examming, (* to be added *)
					procs))

	    val genConjs = initialSimplification conjectures
		handle Disproved => []
		     | Proved => []
		     | Stopped => []

	    fun makeInitProcs g conjs = 
		L.map (fn i=> (sizeOfEs conjs,conjs,g,i,1,(conjs,[],[],[]))) orderParams

	    val suggestLemma =
		if not (#useLS opt) 
		then []
		else
		let val dsymSet = (#DSymSet trs)
		    val dsyms = FS.listItems dsymSet
		    val csymSet = (#CSymSet trs)
		    val csyms = FS.listItems csymSet
		    val bins = L.mapPartial 
				   (fn decl => if FS.member (dsymSet,#sym decl)
					       then case (#sort decl) of
							Sort.Proc (tys,ty) =>
							if (L.length tys) = 2
							then if Sort.equal (L.nth (tys,0), L.nth (tys,1))
								andalso Sort.equal (L.nth (tys,0), ty)
							     then SOME (#sym decl,ty)
							     else NONE
							else NONE
						      | Sort.Base _ => NONE
					       else NONE)
				   decls
		    val consts = L.mapPartial 
				   (fn decl => if FS.member (csymSet,#sym decl)
					       then case (#sort decl) of
							Sort.Proc _ => NONE
						      | Sort.Base _ => SOME (#sym decl, #sort decl)
					       else NONE)
				   decls
		    val comConjs = L.map (fn (f,ty) => let val x = Var.fromStringAndInt ("x",0)
							   val y = Var.fromStringAndInt ("y",0)
						       in
							   (Fun (f, [Var (x,ty), Var (y,ty)],ty),
							    Fun (f, [Var (y,ty), Var (x,ty)],ty))
						       end)
					 bins
		    val assocConjs = 
			L.map (fn (f,ty) => let val x = Var.fromStringAndInt ("x",0)
						val y = Var.fromStringAndInt ("y",0)
						val z = Var.fromStringAndInt ("z",0)
					    in
						(Fun (f, [Fun (f, [Var (x,ty), Var (y,ty)],ty),
							  Var (z,ty)], ty),
						 Fun (f, [Var (x,ty), 
							  Fun (f, [Var (y,ty), Var (z,ty)],ty)],
						      ty))
					    end)
					 bins

		    val unitConjs = 
			ListXProd.foldX
			       (fn ((f,ty),(c,ty1),xs) =>
				   if Sort.equal (ty,ty1)
				   then 
				       let val x = Var.fromStringAndInt ("x",0)
				       in
					   (Fun (f, [Var (x,ty), Fun (c,[],ty)], ty),
					     Var (x,ty))::
					   (Fun (f, [Fun (c,[],ty), Var (x,ty)], ty),
					    Var (x,ty))::xs
				       end
				   else xs)
			       (bins,consts)
			       []

		    val lemmas = comConjs @ assocConjs @ unitConjs
		in
		    LU.mapAppend (fn e => makeInitProcs NON_GOAL [e]) lemmas
		end

	in  
	    proceedOneProcess ([], (* theorems *)
			       [genConjs], (* list of conjectures *)
			       (makeInitProcs GOAL genConjs) @
			       suggestLemma @ 
			       (makeTPproc genConjs))
	end
    end


    end (* of local *)

end
