(******************************************************************************
 * file: rwtools/sprint/riprover.sml
 * description: inductive theorem prover based on rewriting induction
 * author: AOTO Takahito
 * author: SEKO Hiroyuki
 * 
 * Copyright (c) 2007-2015, Toyama&Aoto Laboratory, Tohoku University
 ******************************************************************************)

signature RIPROVER =  
   sig
       val riprover: string 
		     -> string 
		     -> Ind.opt
                     -> Term.decl list
			* (Term.term * Term.term) list
			* (Term.term * Term.term) list
		 	* (int * Atom.atom * Order.rpo_status * int list) list
(*            * (Fun.ord_key * Poly.poly) list  *)
(*	         -> bool *)
		     -> ProverResult.prover_result

       val gcrProver: string 
		     -> string 
		     -> Ind.opt
                     -> Term.decl list
			* Term.term list
			* Term.term list
			* (Term.term * Term.term) list
			* (Term.term * Term.term) list
		     -> ProverResult.prover_result

       (* FSCD version *)
       val gcrProverFSCD: string 
		     -> string 
		     -> Ind.opt
                     -> Term.decl list
			* Term.term list
			* Term.term list
			* Term.term list
			* (Term.term * Term.term) list
			* (Term.term * Term.term) list
		     -> ProverResult.prover_result

       val printSpikeSpec: string 
		     -> string 
		     -> Ind.opt
                     -> Term.decl list
			* (Term.term * Term.term) list
			* (Term.term * Term.term) list
		 	* (int * Atom.atom * Order.rpo_status * int list) list
	         -> bool
   end ; 


structure Riprover : RIPROVER =
   struct

   exception RiproverError of string

   local 
       open Trs
       structure IO = IOMstrs
       structure FS = FunSet
       structure FM = FunMap
       structure L = List
       structure PU = PrintUtil
       structure LU = ListUtil
       fun printErrorMsg msg = TextIO.output (TextIO.stdOut, msg ^ "\n")
       fun error e = printErrorMsg (exnMessage e)
       open PrintUtil
   in

   fun riprover minisatPath tmpDir (opt:Ind.opt) 
(* 		(sorts,decls,rules,conj,funpoly) = *)
		(decls,rules,conj,prec) =
       let val _ = print "Specification:\n"

	   (*** check SORTS section ***)
(*	   val _ = print (PU.prList Sort.toString sorts) *)

	   (*** check FUNCTIONS section ***)
	   val fs = map (fn x => #sym x) decls
	   val sorts = SortSet.listItems (Term.sortSetInDecls decls)

	   val faMap = let fun arity (Sort.Base _) = 0
			     | arity (Sort.Proc (xs,_)) = length xs
		       in
			   L.foldr (fn (x,fm) => FM.insert (fm,#sym x,arity (#sort x)))
				   FM.empty decls
		       end
	   val _ = print (Term.prDecls decls)

	   (*** check RULES section ***)
	   val srules = let val res = Trs.attachSortToRules decls rules
			in if (isSome res)
			   then valOf res
			   else (PU.error "Error: not well-sorted?";
				 raise RiproverError "riprover")
			end


	   val _ = print (Trs.prRules srules)
	   fun indent _ = print ("    ")

	   val trs = let val (fsymSet, dsymSet,csymSet,varCond) = Trs.fdcSetInRules srules
			 val csymSet2 = FS.union (csymSet,
						  FS.difference (FS.addList (FS.empty,fs),fsymSet))
		     in 
			 indent ();
			 print "functions: ";
			 PU.println (LU.toStringCommaCurly Atom.toString (FS.listItems fsymSet));
			 indent ();
			 print "defined functions: ";
			 PU.println (LU.toStringCommaCurly Atom.toString (FS.listItems dsymSet));
			 indent ();
			 print "constructors: ";
			 PU.println (LU.toStringCommaCurly Atom.toString (FS.listItems csymSet2));
			 {FSymSet=fsymSet,DSymSet=dsymSet,CSymSet=csymSet2,
			  VarCond=varCond,Rules=srules}
		     end

	   (*** check CONJECTURES section ***)
	   val sconj = valOf (Trs.attachSortToRules decls conj)
	   val _ = print (Trs.prEqs sconj)

	   (*** check PRECEDENCE section ***)
	   val _ = indent ()

	   val _ = if (length prec) = 0
		   then ()
		   else
		       let val count = ref 0
			   val status = ref Order.MUL
			   val _ = count := (fn (i,_,_,_) => i) (hd prec)
			   val _ = status := (fn (_,_,st,_) => st) (hd prec)
			   val _ = print ("prec: " ^ (Atom.toString ((fn (_,f,_,_) => f) (hd prec))))
			   val _ = L.app (fn (i,f,st,_) => 
					     if i = (!count)
					     then if Order.eqRpoStatus (st,!status)
						  then print (" = " ^ (Atom.toString f))
						  else (PU.error "invalid prec spec";
							raise RiproverError "riprover")
					     else (count := i;
						   status := st;
						   print (" > " ^ (Atom.toString f))))
					 (tl prec)
			   val _ = print  "\n"
		       in ()
		       end

	   val prec' = L.map 
		       (fn (i,f,st,ns) => (i,f,st,
					  case st of
					      Order.LEX => let val arity = valOf (FM.find (faMap,f))
							   in
							       if (length ns) = arity
							       then ns
							       else if (null ns)
							       then L.tabulate (arity,fn x=>x+1)
							       else (PU.error "invalid prec spec";
								     raise RiproverError "riprover")
							   end
					    | Order.MUL => if not (null ns)
							   then (PU.error "invalid prec spec";
								 raise RiproverError "riprover")
							   else ns))
		       prec
	   val _ = (indent (); print "status: ")
	   val _ = print (LU.toStringComma
						  (fn (_,id,st,ns) => 
							  (Atom.toString id) ^ ":" ^ (Order.prRpoStatus st)
							  ^ (LU.toStringCommaSquare Int.toString ns))
				     prec');
	   val _ = print "\n"

	   (*** check Termination of TRS ***)
	   val stMap = L.foldr (fn ((_,f,st,_),fm) => FM.insert (fm,f,st)) FM.empty prec'
	   val lexMap = L.foldr (fn ((_,f,_,ns),fm) => FM.insert (fm,f,ns)) FM.empty prec'
	   fun lex f ts = let val ns = valOf (FM.find (lexMap,f))
			      val ns2 = ListPair.zip (ns,ts)
			      val sorted = ListMergeSort.sort
					       (fn (x,y) => Int.> (#1 x,#1 y))
					       ns2
			  in L.map #2 sorted
			  end
	   fun isMul f = valOf (FM.find (stMap,f)) = Order.MUL
	   val funqprec = List.foldr (fn ((i,f,_,_),m) => FM.insert (m,f,i))  FM.empty prec'

	   val isTerminating = 
	       let val symCount = ref 1
		   val prop = 
		       Prop.Conj (L.map (fn (l,r) =>
					    #1 (Order.encodeRpoCheck (Order.compareQprec funqprec) 
								     isMul lex symCount Order.GT (l,r)))
					srules)
(* 		   val _ = Solver.runDebug := true *)
		   val (result,_) = Solver.propSolver minisatPath tmpDir (prop,!symCount)
	       in
		   if result 
		   then print "Orientable rules.\n"
		   else (print "Rules not orientable."; raise RiproverError "riprover")
	       end

	   (*** check CONFLUENCE of TRS ***)
	   val isConfluent = 
		   if Cr.isLocalConfluent srules
		   then (print "Confluent trs.\n"; true)
		   else (print "Non-Confluent trs.\n"; false)

	   (*** check CONSISTENCY of TRS ***)
	   val isConsistent = 
	       if FS.numItems (#CSymSet trs) > 1
	       then (print "Consistent trs.\n"; true)
	       else (print "Non-Consisten trs.\n"; false)

	   (*** check LEFT-LINEARITY of TRS ***)
(*         NOT NEEDED ANY MORE *)
(* 	   val isLeftLinear =  *)
(* 		   if Trs.areLeftLinearRules srules *)
(* 		   then (print "Left-Linear trs.\n"; true) *)
(* 		   else (print "Non-Left-Linear trs.\n"; false) *)

	   (*** check MONOMORPHIC signature ***)
	   val (monomorphicSortSet,argTables) =
	       if (#useSDC opt) orelse (#useSG opt) orelse (#useTP opt)
	       then Sg.getMonomorphicSortSet decls trs
	       else (SortSet.empty,Sg.nullTables)
		    
	   fun isMonomorphicEq (s,t) = SortSet.exists (fn ty => Sort.equal (ty,Term.sortOfTerm s))
						      monomorphicSortSet
	   val isMonomorphicRules = isMonomorphicEq 

	  (*
	   val addlem = ref []
	   val _ = if isConfluent
		   then addlem := Sg.upPosLemmas argTables trs
		   else () 
	  *)

	   (*** set options ***)
	   val opt' = {useE2 = #useE2 opt,
		       useQB = #useQB opt,
		       useDC = #useDC opt,
		       useSDC = #useSDC opt,
		       useSG = #useSG opt,
		       useMG = #useMG opt,
		       useESG = false,
		       useDP = if isConfluent andalso isConsistent
			       then #useDP opt
			       else false,
		       useTP = #useTP opt,
		       useRS = #useRS opt,
		       useLS = #useLS opt
		      }


	   val _ = print "Options: \n"
	   val _ = print (" Expand2: " ^ (if #useE2 opt' then "on\n" else "off\n"))
	   val _ = print (" Quasi-Basic: " ^ (if #useQB opt' then "on\n" else "off\n"))
	   val _ = print (" Divergence Critic: " ^ (if #useDC opt' then "on\n" else "off\n"))
	   val _ = print (" Sound Divergence Critic: " ^ (if #useSDC opt' then "on\n" else "off\n"))
	   val _ = print (" Sound Generalization: " ^ (if #useSG opt' then "on\n" else "off\n"))
	   val _ = print (" Maximal Generalization: " ^ (if #useMG opt' then "on\n" else "off\n"))
(*	   val _ = print (" Extended Sound Generalization: " ^ (if #useESG opt' then "on\n" else "off\n")) *)
	   val _ = print (" Decompose & Disproof: " ^ (if #useDP opt' then "on\n" else "off\n"))
	   val _ = print (" Term Partition: " ^ (if #useTP opt' then "on\n" else "off\n"))
	   val _ = print (" Reset: " ^ (if #useRS opt' then "on\n" else "off\n"))
	   val _ = print (" Lemma Suggestion: " ^ (if #useLS opt' then "on\n" else "off\n"))


	   fun isMul 0 f = valOf (FM.find (stMap,f)) = Order.MUL
	     | isMul 1 _ =  false
	     | isMul _ f =  let val arity = valOf (FM.find (faMap,f))
			    in arity = 2 
			    end

	   (*** preparation of order checking tools ***)
	   fun isOrdered i = Order.qrpoCheck (Order.compareQprec funqprec) (isMul i) lex Order.GT

	   fun isEquivalent i = Order.qrpoCheck (Order.compareQprec funqprec) (isMul i) lex Order.EQ

	   val orderParams = 
	       L.filter
		   (fn i => L.all
				(Order.qrpoCheck (Order.compareQprec funqprec) (isMul i) lex Order.GT)
				srules)
		   [0]
		   (* [0,1,2] *)

	   fun main () = 
	       if null orderParams
	       then (print "Failure (ordering)\n"; ProverResult.MAYBE)
	       else
	       Ind.ind  isOrdered isEquivalent orderParams isMonomorphicEq
				  (*  (Poly.poly_strict funpoly) (Poly.poly_eq funpoly) *)
				 opt' decls trs argTables sconj

       in 
           main ()
       end
	   handle RiproverError msg => ProverResult.MAYBE

 (* for GCR not using prec *)
   fun riprover2 minisatPath tmpDir (opt:Ind.opt) (decls,cPat,dPat,rules,conj) =
       let (* val _ = print "Specification:\n"*)
	   (*** check SORTS section ***)
(*	   val _ = print (PU.prList Sort.toString sorts) *)

	   (*** check FUNCTIONS section ***)
	   val fs = map (fn x => #sym x) decls
	   val sorts = SortSet.listItems (Term.sortSetInDecls decls)

	   val faMap = let fun arity (Sort.Base _) = 0
			     | arity (Sort.Proc (xs,_)) = length xs
		       in
			   L.foldr (fn (x,fm) => FM.insert (fm,#sym x,arity (#sort x)))
				   FM.empty decls
		       end
(*	   val _ = print (Term.prDecls decls) *)

	   (*** check RULES section ***)
	   val srules = rules
	   (* val srules = let val res = Trs.attachSortToRules decls rules *)
	   (* 		in if (isSome res) *)
	   (* 		   then valOf res *)
	   (* 		   else (PU.error "Error: not well-sorted?"; *)
	   (* 			 raise RiproverError "riprover") *)
	   (* 		end *)


	   (* val _ = print (Trs.prRules srules) *)
	   (* fun indent _ = print ("    ") *)

	   (*** check Free Constructor section ***)
	   val cSet = Term.funSetInTerms cPat
	   val isFreeConstructor = 
	       if L.all (fn (l,_) => not (FunSet.isSubset (Term.funSetInTerm l, cSet))) rules
	       then (print "Constructors are free.\n"; true)
	       else (print "Constructors are not free.\n"; false)

	   val trs = let val (fsymSet, dsymSet,csymSet,varCond) = Trs.fdcSetInRules srules
			 val csymSet2 = FS.union (csymSet,
						  FS.difference (FS.addList (FS.empty,fs),fsymSet))
		     in 
			 (* indent (); *)
			 (* print "functions: "; *)
			 (* PU.println (LU.toStringCommaCurly Atom.toString (FS.listItems fsymSet)); *)
			 (* indent (); *)
			 (* print "defined functions: "; *)
			 (* PU.println (LU.toStringCommaCurly Atom.toString (FS.listItems dsymSet)); *)
			 (* indent (); *)
			 (* print "constructors: "; *)
			 (* PU.println (LU.toStringCommaCurly Atom.toString (FS.listItems csymSet2)); *)
			 {FSymSet=fsymSet,DSymSet=dsymSet,CSymSet=csymSet2,
			  VarCond=varCond,Rules=srules}
		     end

	   (*** check CONJECTURES section ***)
	   val sconj = conj
	   (* val sconj = valOf (Trs.attachSortToRules decls conj) *)
	   (* val _ = print (Trs.prEqs sconj) *)

	   (*** check PRECEDENCE section ***)
	   (* val _ = indent () *)
	   val prec' = [] 


	   (*** get PRECEDENCE section ***)
	   val opt0 = { useQuasi=true, useLex = true, useMul = true, useAf = false }
	   val info0 = PoSolver.poSolverForGcr minisatPath tmpDir opt0 (srules,[])

	   val opt1 = { useQuasi=true, useLex = false, useMul = true, useAf = false }
	   val info1 = PoSolver.poSolverForGcr minisatPath tmpDir opt1 (srules,[])

	   val opt2 = { useQuasi=true, useLex = true, useMul = false, useAf = false }
	   val info2 = PoSolver.poSolverForGcr minisatPath tmpDir opt2 (srules,[])

	   val orderParams = L.mapPartial (fn x=>x)
					  [ if isSome info0 then SOME 0 else NONE,
					    if isSome info1 then SOME 1 else NONE,
					    if isSome info2 then SOME 2 else NONE ]

	   val _ = if null (orderParams)
		   then (print "Termination proof failed.\n"; raise RiproverError "")
		   else ()
				       
	   val (isOrdered0, isEquivalent0) = 
	       if isSome info0
	       then let val (funqprec,mul,lex) = valOf info0
			val prec = Order.compareQprec funqprec
			fun isMul f = LU.member' Fun.equal f mul
			fun order f ts = case L.find (fn (g,_) => Fun.equal (f,g)) lex of
					     NONE => (print "Lex status not found.\n";  [];
						      raise RiproverError "")  (* should not happen *)
					   | SOME (_,ns) => L.map (fn n => L.nth (ts,n)) ns
		    in (Order.qrpoCheck prec isMul order Order.GT,
			Order.qrpoCheck prec isMul order Order.EQ)
		    end 
	       else (fn _ => false, fn _=> false)

	   val (isOrdered1, isEquivalent1) = 
	       if isSome info1
	       then let val (funqprec,mul,lex) = valOf info1
			val prec = Order.compareQprec funqprec
		    in (Order.qmpoCheck prec Order.GT,
			Order.qmpoCheck prec Order.EQ)
		    end 
	       else (fn _ => false, fn _=> false)

	   val (isOrdered2, isEquivalent2) = 
	       if isSome info2
	       then let val (funqprec,mul,lex) = valOf info2
			val prec = Order.compareQprec funqprec
			fun order f ts = case L.find (fn (g,_) => Fun.equal (f,g)) lex of
					     NONE => (print "Lex status not found.\n"; 
						      raise RiproverError "")  (* should not happen *)
					   | SOME (_,ns) => L.map (fn n => L.nth (ts,n)) ns
		    in (Order.qlposCheck prec order Order.GT,
			Order.qlposCheck prec order Order.EQ)
		    end 
	       else (fn _ => false, fn _=> false)

	   fun isOrdered 0 = isOrdered0
	     | isOrdered 1 = isOrdered1
	     | isOrdered _ = isOrdered2 (* used for 2 *)

	   fun isEquivalent 0 = isEquivalent0
	     | isEquivalent 1 = isEquivalent1
	     | isEquivalent _ = isEquivalent2  (* used for 2 *)

	   (*** check CONFLUENCE of TRS ***)
	   val isConfluent = 
		   if Cr.isLocalConfluent srules
		   then (print "Confluent trs.\n"; true)
		   else (print "Non-Confluent trs.\n"; false)

	   (*** check CONSISTENCY of TRS ***)
	   val isConsistent = 
	       let val consts = L.filter Term.isConstant cPat
	       in if (L.length consts) > 1
		     orelse (not (null consts)
			     andalso (L.exists (not o Term.isConstant) cPat))
		  then (print "Consistent trs.\n"; true)
		  else (print "Non-Consisten trs.\n"; false)
	       end
		   
	   (*** set options ***)
	   val opt' = {useE2 = true,
		       useQB = false,
		       useDC = true,
		       useSDC = true,
		       useSG = isFreeConstructor,
		       useMG = true,
		       useESG = isFreeConstructor,
		       useDP = isConfluent andalso isConsistent,
		       useTP = isFreeConstructor,
		       useRS = true,
		       useLS = isFreeConstructor
		      }

	   (*** check MONOMORPHIC signature ***)
	   val (monomorphicSortSet,argTables) =
	       if isFreeConstructor
		  andalso 
		  ((#useSDC opt) orelse (#useSG opt) orelse (#useTP opt))
	       then Sg.getMonomorphicSortSet decls trs
	       else (SortSet.empty,Sg.nullTables)
		    
	   fun isMonomorphicEq (s,t) = SortSet.exists (fn ty => Sort.equal (ty,Term.sortOfTerm s))
						      monomorphicSortSet
	   val isMonomorphicRules = isMonomorphicEq 

	  (*
	   val addlem = ref []
	   val _ = if isConfluent
		   then addlem := Sg.upPosLemmas argTables trs
		   else () 
	  *)

	   val _ = print "Options: \n"
	   val _ = print (" Expand2: " ^ (if #useE2 opt' then "on\n" else "off\n"))
	   val _ = print (" Quasi-Basic: " ^ (if #useQB opt' then "on\n" else "off\n"))
	   val _ = print (" Divergence Critic: " ^ (if #useDC opt' then "on\n" else "off\n"))
	   val _ = print (" Sound Divergence Critic: " ^ (if #useSDC opt' then "on\n" else "off\n"))
	   val _ = print (" Sound Generalization: " ^ (if #useSG opt' then "on\n" else "off\n"))
	   val _ = print (" Maximal Generalization: " ^ (if #useMG opt' then "on\n" else "off\n"))
(*	   val _ = print (" Extended Sound Generalization: " ^ (if #useESG opt' then "on\n" else "off\n")) *)
	   val _ = print (" Decompose & Disproof: " ^ (if #useDP opt' then "on\n" else "off\n"))
	   val _ = print (" Term Partition: " ^ (if #useTP opt' then "on\n" else "off\n"))
	   val _ = print (" Reset: " ^ (if #useRS opt' then "on\n" else "off\n"))
	   val _ = print (" Lemma Suggestion: " ^ (if #useLS opt' then "on\n" else "off\n"))


       in 
	   Ind.ind  isOrdered isEquivalent orderParams isMonomorphicEq
		    opt' decls trs argTables sconj
       end
	   handle RiproverError msg => ProverResult.MAYBE



   fun gcrProver minisatPath tmpDir (opt:Ind.opt) (decls,cPat,dPat,rules,conj) =
       let (* val _ = print "Specification:\n" *)

	   (*** check SORTS section ***)
	   val sortSet = SortSet.union (Trs.sortSetInRules rules,
				      Term.sortSetInDecls decls)
	   val sorts = SortSet.listItems sortSet
	   val _ = print (PU.prList Sort.toString sorts) 

	   (*** check FUNCTIONS section ***)
	   val _ = print (Term.prDecls decls)
	   val fs = map (fn x => #sym x) decls
	   val faMap = let fun arity (Sort.Base _) = 0
			     | arity (Sort.Proc (xs,_)) = length xs
		       in
			   L.foldr (fn (x,fm) => FM.insert (fm,#sym x,arity (#sort x)))
				   FM.empty decls
		       end

	   (*** check RULES section ***)
	   val srules = rules
	   val _ = print (Trs.prRules srules)

	   (*** make TRS section ***)
	   val trs = { FSymSet = FunSet.empty,
		       DSymSet = FunSet.empty,  (* temporaly use empty set *)
		       CSymSet = FunSet.empty,  (* temporaly use empty set *)
		       VarCond = Trs.checkVarCond srules,
		       Rules = rules }:Trs.trs

	   (*** check CONJECTURES section ***)
	   (* val sconj = valOf (Trs.attachSortToRules decls conj) *)
	   val sconj = conj
	   val _ = print (Trs.prEqs sconj)

	   (*** get PRECEDENCE section ***)
	   val opt2 = { useQuasi=true, useLex = false, useMul = true, useAf = false }
	   val funqprec = case PoSolver.poSolverForGcr minisatPath tmpDir opt2 (srules,[]) of
			      SOME (prec,_,_) => prec
			    | NONE => (print "Termination proof failed.\n";
				       raise RiproverError "")

	   val _ = print "Precedence:\n"
	   val _ = print (LU.toStringCommaCurly (fn (f,i) => "(" ^ (Fun.toString f) ^ "," ^ (Int.toString i) ^ ")")
						     (FM.listItemsi funqprec))
	   val _ = print "\n"

	   (*** check CONFLUENCE of TRS ***)
	   val cps = Cr.criticalPairs srules
       
	   val cps2 = L.filter (fn (l,r) => not (Term.equal (l,r)))
			       (List.map (fn (u,v) => (Rewrite.linf srules u,  Rewrite.linf srules v)) cps)

	   val isConfluent = 
		   if null cps2
		   then (print "Confluent trs.\n"; true)
		   else (print "Non-Confluent trs.\n"; false)

	   (*** check CONSISTENCY of TRS ***)
	   (* val isConsistent =  *)
	   (*     if FS.numItems (#CSymSet trs) > 1 *)
	   (*     then (print "Consistent trs.\n"; true) *)
	   (*     else (print "Non-Consisten trs.\n"; false) *)

	   (*** check MONOMORPHIC signature ***)
	   (* val (monomorphicSortSet,argTables) = *)
	   (*     if (#useSDC opt) orelse (#useSG opt) orelse (#useTP opt) *)
	   (*     then Sg.getMonomorphicSortSet decls (trs:Trs.trs) *)
	   (*     else (SortSet.empty,Sg.nullTables) *)
		    
	   (* fun isMonomorphicEq (s,t) = SortSet.exists (fn ty => Sort.equal (ty,Term.sortOfTerm s)) *)
	   (* 					      monomorphicSortSet *)
	   (* val isMonomorphicRules = isMonomorphicEq  *)

	   (*** set options ***)
	   val opt' = {useE2 = true,
		       useQB = false,
		       useDC = false,
		       useSDC = false,
		       useSG =  false,
		       useMG =  false,
		       useESG = false,
		       useDP =  false,
		       useTP =  false,
		       useRS =  false,
		       useLS =  false
		      }

(* 	   val _ = print "Options: \n" *)
(* 	   val _ = print (" Expand2: " ^ (if #useE2 opt' then "on\n" else "off\n")) *)
(* 	   val _ = print (" Quasi-Basic: " ^ (if #useQB opt' then "on\n" else "off\n")) *)
(* 	   val _ = print (" Divergence Critic: " ^ (if #useDC opt' then "on\n" else "off\n")) *)
(* 	   val _ = print (" Sound Divergence Critic: " ^ (if #useSDC opt' then "on\n" else "off\n")) *)
(* 	   val _ = print (" Sound Generalization: " ^ (if #useSG opt' then "on\n" else "off\n")) *)
(* 	   val _ = print (" Maximal Generalization: " ^ (if #useMG opt' then "on\n" else "off\n")) *)
(* (*	   val _ = print (" Extended Sound Generalization: " ^ (if #useESG opt' then "on\n" else "off\n")) *) *)
(* 	   val _ = print (" Decompose & Disproof: " ^ (if #useDP opt' then "on\n" else "off\n")) *)
(* 	   val _ = print (" Term Partition: " ^ (if #useTP opt' then "on\n" else "off\n")) *)
(* 	   val _ = print (" Reset: " ^ (if #useRS opt' then "on\n" else "off\n")) *)
(* 	   val _ = print (" Lemma Suggestion: " ^ (if #useLS opt' then "on\n" else "off\n")) *)

	   fun lex f ts = []

	   (*** preparation of order checking tools ***)
	   (* val isOrdered = Order.qrpoCheck (Order.compareQprec funqprec) (fn _ => true) lex Order.GT *)
	   (* val isEquivalent = Order.qrpoCheck (Order.compareQprec funqprec) (fn _ => true) lex Order.EQ *)
	   val isOrdered = Order.qmpoCheck (Order.compareQprec funqprec) Order.GT
	   val isEquivalent = Order.qmpoCheck (Order.compareQprec funqprec) Order.EQ

	   val gcrOfR = if isConfluent
			then ProverResult.YES
			else let val _ = print "Check ground confluence of R by basic RI.\n"
			     in Ind.indsimple isOrdered isEquivalent opt' (decls,cPat,dPat) trs cps2
			     end

       in if gcrOfR = ProverResult.YES
          then if null conj
	       then (print "Conjuctures parts are empty.\n";
		     ProverResult.YES)
	       else (print "Check conjuctures parts are inductive theorem of R.\n";
		     riprover2 minisatPath tmpDir (opt:Ind.opt) (decls,cPat,dPat,rules,conj))
	  else (print "Failed to prove ground confluece R.\n";
		ProverResult.MAYBE)
       end
	   handle RiproverError msg => ProverResult.MAYBE


   fun gcrProverFSCD minisatPath tmpDir (opt:Ind.opt) (decls,cPat,dPat,cPatReducible,rules,conj) =
       let val _ = print "*** Ground Confluence Check by Rewriting Induction ***\n"

	   (*** check SORTS section ***)
	   val sortSet = SortSet.union (Trs.sortSetInRules rules,
				      Term.sortSetInDecls decls)
	   val sorts = SortSet.listItems sortSet
	   val _ = PU.println ("Sort: " ^ (LU.toStringCommaCurly Sort.toString sorts))

	   (*** check FUNCTIONS section ***)
	   val _ = PU.println ("Signature:")
	   val _ = print (Term.prDecls decls)

	   val fs = map (fn x => #sym x) decls
	   val faMap = let fun arity (Sort.Base _) = 0
			     | arity (Sort.Proc (xs,_)) = length xs
		       in
			   L.foldr (fn (x,fm) => FM.insert (fm,#sym x,arity (#sort x)))
				   FM.empty decls
		       end

	   (*** check RULES section ***)
	   val srules = rules
	   val _ = PU.println ("Rule Part:")
	   val _ = print (Trs.prRules srules)

	   (*** make TRS section ***)
	   val trs = { FSymSet = FunSet.empty,
		       DSymSet = FunSet.empty,  (* temporaly use empty set *)
		       CSymSet = FunSet.empty,  (* temporaly use empty set *)
		       VarCond = true,  (* we don't care---conj part may contain e-rules *)
		       Rules = rules }:Trs.trs

	   (*** check CONJECTURES section ***)
	   (* val sconj = valOf (Trs.attachSortToRules decls conj) *)
	   val _ = PU.println ("Conjecture Part:")
	   val sconj = conj
	   val _ = print (Trs.prEqs sconj)

	   (*** get PRECEDENCE section ***)
	  (* val opt2 = { useQuasi=true, useLex = false, useMul = true, useAf = false } *)
	   val opt2 = { useQuasi=false, useLex = false, useMul = true, useAf = false }
	   val funqprec = case PoSolver.poSolverForGcr minisatPath tmpDir opt2 (srules,[]) of
			      SOME (prec,_,_) => prec
			    | NONE => (print "Termination proof failed.\n";
				       raise RiproverError "")

	   val _ = print "Precedence (by weight): "
	   val _ = PU.println (LU.toStringCommaCurly (fn (f,i) => "(" ^ (Fun.toString f) ^ "," ^ (Int.toString i) ^ ")")
						     (FM.listItemsi funqprec))

	   (*** check CONFLUENCE of TRS ***)
	   val cps = Cr.criticalPairs srules
       
	   val cps2 = L.filter (fn (l,r) => not (Term.equal (l,r)))
			       (List.map (fn (u,v) => (Rewrite.linf srules u,  Rewrite.linf srules v)) cps)

	   val isConfluent = 
		   if null cps2
		   then (print "Rule part is confluent.\n"; true)
		   else (print "Rule part is not confluent.\n"; false)

	   (*** set options (ȤʤΤǴطʤ) ***)
	   val opt' = {useE2 = false,
		       useQB = false,
		       useDC = false,
		       useSDC = false,
		       useSG =  false,
		       useMG =  false,
		       useESG = false,
		       useDP =  false,
		       useTP =  false,
		       useRS =  false,
		       useLS =  false
		      }

	   fun lex f ts = []

	   (*** preparation of order checking tools ***)
	   (* val isOrdered = Order.qrpoCheck (Order.compareQprec funqprec) (fn _ => true) lex Order.GT *)
	   (* val isEquivalent = Order.qrpoCheck (Order.compareQprec funqprec) (fn _ => true) lex Order.EQ *)
	   val isOrdered = Order.qmpoCheck (Order.compareQprec funqprec) Order.GT
	   val isEquivalent = Order.qmpoCheck (Order.compareQprec funqprec) Order.EQ

	   val gcrOfR0 = if isConfluent
			then ProverResult.YES
			else let val _ = print "Check ground confluence of Rule part (R0) by basic RI.\n"
			     in Ind.indsimpleFSCD isOrdered isEquivalent opt' (decls,cPat,dPat,cPatReducible) trs cps2
			     end

       in if gcrOfR0 = ProverResult.YES
          then (print "R0 is ground confluent.\n";
		if null conj
		then (print "Conjucture part is empty.\n";
		      ProverResult.YES)
		else (print "Check conj part consists of inductive theorems of R0.\n";
		      (*		     riprover2 minisatPath tmpDir (opt:Ind.opt) (decls,cPat,dPat,rules,conj) *)
		      let val gcrOfRall = Ind.indsimpleFSCD isOrdered isEquivalent opt' (decls,cPat,dPat,cPatReducible) trs conj
		      in if gcrOfRall = ProverResult.YES
			 then (print "Conj part consisits of inductive theorems of R0.\n";ProverResult.YES)
			 else (print "Failed to prove conj part consists of inductive theorems of R0.\n";
			       ProverResult.MAYBE)
		      end))
	  else (print "Failed to prove ground confluece R0.\n";
		ProverResult.MAYBE)
       end
	   handle RiproverError msg => ProverResult.MAYBE







   fun printSpikeSpec minisatPath tmpDir (opt:Ind.opt) 
		(decls,rules,conj,prec) =
       let 
	   (*** check FUNCTIONS section ***)
	   val fs = map (fn x => #sym x) decls
	   val sorts = SortSet.listItems (Term.sortSetInDecls decls)

	   val faMap = let fun arity (Sort.Base _) = 0
			     | arity (Sort.Proc (xs,_)) = length xs
		       in
			   L.foldr (fn (x,fm) => FM.insert (fm,#sym x,arity (#sort x)))
				   FM.empty decls
		       end

	   (*** check RULES section ***)
	   val srules = let val res = Trs.attachSortToRules decls rules
			in if (isSome res)
			   then valOf res
			   else (PU.error "Error: not well-sorted?";
				 raise RiproverError "riprover")
			end

	   val trs = let val (fsymSet, dsymSet,csymSet,varCond) = Trs.fdcSetInRules srules
			 val csymSet2 = FS.union (csymSet,
						  FS.difference (FS.addList (FS.empty,fs),fsymSet))
		     in 
			 {FSymSet=fsymSet,DSymSet=dsymSet,CSymSet=csymSet2,
			  VarCond=varCond,Rules=srules}
		     end

	   (*** check CONJECTURES section ***)
	   val sconj = valOf (Trs.attachSortToRules decls conj)

	   (*** check PRECEDENCE section ***)
	   val _ = if (length prec) = 0
		   then ()
		   else
		       let val count = ref 0
			   val status = ref Order.MUL
			   val _ = count := (fn (i,_,_,_) => i) (hd prec)
			   val _ = status := (fn (_,_,st,_) => st) (hd prec)
			   val _ = L.app (fn (i,f,st,_) => 
					     if i = (!count)
					     then if Order.eqRpoStatus (st,!status)
						  then ()
						  else (PU.error "invalid prec spec";
							raise RiproverError "riprover")
					     else (count := i;
						   status := st))
					 (tl prec)
		       in ()
		       end

	   val prec' = L.map 
		       (fn (i,f,st,ns) => (i,f,st,
					  case st of
					      Order.LEX => let val arity = valOf (FM.find (faMap,f))
							   in
							       if (length ns) = arity
							       then ns
							       else if (null ns)
							       then L.tabulate (arity,fn x=>x+1)
							       else (PU.error "invalid prec spec";
								     raise RiproverError "riprover")
							   end
					    | Order.MUL => if not (null ns)
							   then (PU.error "invalid prec spec";
								 raise RiproverError "riprover")
							   else ns))
		       prec

	   (*** check Termination of TRS ***)
	   val stMap = L.foldr (fn ((_,f,st,_),fm) => FM.insert (fm,f,st)) FM.empty prec'
	   val lexMap = L.foldr (fn ((_,f,_,ns),fm) => FM.insert (fm,f,ns)) FM.empty prec'
	   fun lex f ts = let val ns = valOf (FM.find (lexMap,f))
			      val ns2 = ListPair.zip (ns,ts)
			      val sorted = ListMergeSort.sort
					       (fn (x,y) => Int.> (#1 x,#1 y))
					       ns2
			  in L.map #2 sorted
			  end
	   fun isMul f = valOf (FM.find (stMap,f)) = Order.MUL
	   val funqprec = List.foldr (fn ((i,f,_,_),m) => FM.insert (m,f,i))  FM.empty prec'
	   val isTerminating = 
	       let val symCount = ref 1
		   val prop = 
		       Prop.Conj (L.map (fn (l,r) =>
					    #1 (Order.encodeRpoCheck (Order.compareQprec funqprec) 
								     isMul lex symCount Order.GT (l,r)))
					srules)
(* 		   val _ = Solver.runDebug := true *)
		   val (result,_) = Solver.propSolver minisatPath tmpDir (prop,!symCount)
	       in
		   if result 
		   then ()
		   else (print "Rules not orientable."; raise RiproverError "riprover")
	       end



	   fun toSpike () =
	       let (* val _ = print "specification : test\n\n" *)
		   val _ = print "sorts:"
		   val _ = List.app
			       (fn ty => print (" " ^ (Sort.toString ty)))
			       sorts
		   val _ = print " ;\n\n"

		   val sigMap = L.foldr (fn (x,fm) => 
					    FM.insert (fm,#sym x,
						       case (#sort x) of
							   Sort.Base a => ([],Sort.Base a)
							 | Sort.Proc (tys,ty) => (tys,ty)))
					FM.empty decls

		   val _ = print "constructors:\n\n"
		   val _ = List.app
			   (fn (f,(tys,ty)) =>
			       let val _ = print (Fun.toString f)
				   val _ = L.app (fn _ => print "_") tys
				   val _ = print " :"
				   val _ = L.app (fn s => print (" " ^ (Sort.toString s))) tys
				   val _ = print " ->"
			    	   val _ = print (" " ^ (Sort.toString ty))
				   val _ = print ";\n"
			       in () end)
			   (FM.listItemsi
				(FM.filteri (fn (f,_) => FS.member (#CSymSet trs, f))
					   sigMap))

		   val _ = print "\ndefined functions:\n\n"
		   val _ = List.app
			   (fn (f,(tys,ty)) =>
			       let val _ = print (Fun.toString f)
				   val _ = L.app (fn _ => print "_") tys
				   val _ = print " :"
				   val _ = L.app (fn s => print (" " ^ (Sort.toString s))) tys
				   val _ = print " ->"
			    	   val _ = print (" " ^ (Sort.toString ty))
				   val _ = print ";\n"
			       in () end)
			   (FM.listItemsi
				(FM.filteri (fn (f,ty) => FS.member (#DSymSet trs, f))
					   sigMap))

		   fun VtoString (x,0) = (Atom.toString x) 
		     | VtoString (x,i) = (Atom.toString x) ^ (Int.toString i) 
		   open Term;
		   fun TtoString (Var (x,_)) = VtoString x
		     | TtoString (Fun (f,[],_)) = Fun.toString f
		     | TtoString (Fun (f,ts,_)) = (Fun.toString f) ^ "(" ^ (toStringList ts) ^ ")" 
		   and toStringList ts = PrintUtil.prSeq TtoString ts

		   val _ = print "\naxioms:\n\n"
		   val _ = L.app
			   (fn (l,r) => print ((TtoString l) ^ " = " ^ (TtoString r) ^ ";\n" ))
			   srules

		   val _ = print "\ngreater:\n\n"
		   val _ = L.app (fn (_,f,_,_) => print ((Atom.toString f) ^ " ")) prec
		   val _ = print ";\n\n"

		   val _ = print "status :\n\n"
		   val _ = L.app (fn (_,f,st,_) => 
				     if Order.eqRpoStatus (st,Order.MUL)
				     then print ((Atom.toString f) ^ " ")
				     else ())
				 prec
		   val _ = print "ms;\n\n"

		   val _ = print "properties:\n\n"
		   val _ = print "system_is_sufficiently_complete ;\n"
		   val _ = print "system_is_strongly_sufficiently_complete ;\n"
		   val _ = print "system_is_ground_convergent ;\n"

		   val _ = print "\n\nStrategy:\n\n"

		   val _ = print "% instances of Delete\n"
		   val _ = print "tautology_rule                   = delete(id, [tautology]);\n"
		   val _ = print "subsumption_rule                 = delete(id, [subsumption (L|C)]);\n"
		   val _ = print "\n"

		   val _ = print "% instances of Simplify\n"
		   val _ = print "decomposition_rule = simplify(id, [negative_decomposition]);\n"
		   val _ = print "conditional_rewriting_rule = simplify(id, [conditional_rewriting(normalize, L|R|C, *)]);\n"
		   val _ = print "\n"

		   val _ = print "% instances of AddPremise\n"
		   val _ = print "induction = add_premise(generate,[id]);\n"
		   val _ = print "induction_eq = add_premise(generate_eq,[id]);"
		   val _ = print "\n\n"

		   val _ = print "stra = repeat (try (\n"
		   val _ = print "                    tautology_rule,\n"
		   val _ = print "                    decomposition_rule,\n"
		   val _ = print "                    conditional_rewriting_rule,\n"
		   val _ = print "                    subsumption_rule,\n"
		   val _ = print "                    print_goals\n"
		   val _ = print "));\n"
		   val _ = print "\n"

		   val _ = print "fullind = (repeat(stra, induction_eq), print_goals_with_history);\n"
		   val _ = print "\nstart_with: fullind\n"
		   val _ = print "\n"

		   val _ = print "conjectures:\n\n"
		   val _ = L.app
			   (fn (l,r) => print ((TtoString l) ^ " = " ^ (TtoString r) ^ ";\n" ))
			   sconj
	       in
		   true
	       end

       in 
		   toSpike ()
       end


   end (* of local *)
   end;


