(******************************************************************************
 * file: rwtools/stsrs/io_stsrs.sml
 * description: Input/Output utilities for STSRS
 * author: AOTO Takahito
 * 
 * Copyright (c) 2008, Toyama&Aoto Laboratory, Tohoku University
 ******************************************************************************)

structure IOStsrs : sig

   include IO_TRS

   val rdSort: string -> Sort.sort
   val rdSorts: string list -> Sort.sort list
   val rdDecl: string -> Term.decl
   val rdDecls: string list -> Term.decl list

   val rdTerm': string -> Term.term
   val rdRule': string -> Term.term * Term.term
   val rdEq': string -> Term.term * Term.term
   val rdTerms': string list -> Term.term list
   val rdRules': string list -> (Term.term * Term.term) list
   val rdEqs': string list -> (Term.term * Term.term) list

   val prTerm': Term.term -> string
   val prRule': Term.term * Term.term -> string
   val prEq': Term.term * Term.term -> string
   val prTerms': Term.term list -> string list
   val prRules': (Term.term * Term.term) list -> string list
   val prEqs': (Term.term * Term.term) list -> string list

end  =

  struct

  structure StsrsKey =
      struct 
      val alphas = [ ]
      and symbols = ["(", ")", ",", ":", "->", ">", "=", "*"]
      end;

   structure StsrsLex = Lexical (StsrsKey);
   structure StsrsParsing = Parsing (StsrsLex);

   local 
       open Term
       structure A = Atom
       structure S = Sort
       structure L = List
   in

   local
       infix 6 $--
       infix 5 --
       infix 3 >>
       infix 0 ||
       fun makeFun ts = Fun (A.atom "@", ts, Sort.null)
       fun makeConst id = Fun (A.atom id, nil, Sort.null)
       fun makeVar (id, ty)  = Var ((A.atom id, 0), ty)
       fun makeTermList (t, ts) = t::ts
       fun makeTermList1 t = t::nil
       fun makeNil _ = nil
       fun makeSort id = S.Base (A.atom id)
       fun makeBra (ty,_) = ty
       fun makeProc (args,ty) = S.Proc (args,ty)
       fun makeProc1 ty = S.Proc (nil,ty)
       fun makeArgs (ty,args) = ty :: args
       fun makeArgs1 ty = ty :: nil
       fun makeDecl (id,s) = {sym=A.atom id, sort=s}:decl
       fun makeFun' (ts,ty) = Fun (A.atom "@", ts, ty)
       fun makeConst' (id, ty) = Fun (A.atom id, nil, ty)

       open StsrsParsing

       fun rule toks =
           ( term -- "->" $-- term ) toks
       and equation toks =
           ( term -- "=" $-- term ) toks
       and term toks =
           ( "(" $-- termlist >> makeFun || atom ) toks
       and termlist toks =
           ( $ ")" >> makeNil || termseq -- $ ")" >> #1 ) toks
       and termseq toks =
           ( term --  termseq >> makeTermList || term >> makeTermList1 ) toks
       and atom toks  =
           ( "?" $-- id -- ":" $-- proc >> makeVar || id >> makeConst ) toks
       and rule' toks =
           ( term' -- "->" $-- term' ) toks
       and equation' toks =
           ( term' -- "=" $-- term' ) toks
       and term' toks =
           ( "(" $-- termlist' -- ":" $-- proc >> makeFun' || atom' ) toks
       and termlist' toks =
           ( $ ")" >> makeNil || termseq' -- $ ")" >> #1 ) toks
       and termseq' toks =
           ( term' --  termseq' >> makeTermList || term' >> makeTermList1 ) toks
       and atom' toks  =
           ( "?" $-- id -- ":" $-- proc >> makeVar || id -- ":" $-- proc >> makeConst' ) toks
       and sort toks  =
           ( id >> makeSort) toks
       and decl toks  =
           ( id -- ":" $-- proc >> makeDecl) toks
       and bra toks  =
           ( "(" $-- proc -- $ ")" >> makeBra || sort ) toks        
       and proc toks  =
           ( args -- "=" $-- (">" $-- bra ) >> makeProc || sort ) toks        
       and args toks  =
           ( bra -- "*" $-- args >> makeArgs || bra >> makeArgs1 ) toks        
   in 
   val rdTerm = reader term
   val rdRule = reader rule
   val rdEq = reader equation
   val rdDecl = reader decl
   val rdSort = reader bra
   val rdTerm' = reader term'
   val rdRule' = reader rule'
   val rdEq' = reader equation'
   end

   val rdTerms = L.map rdTerm
   val rdRules = L.map rdRule
   val rdEqs = L.map rdEq
   val rdSorts = L.map rdSort
   val rdDecls = L.map rdDecl
   val rdTerms' = L.map rdTerm'
   val rdRules' = L.map rdRule'
   val rdEqs' = L.map rdEq'

   fun prTerm (Var (x,ty)) = (Var.toString x) ^ ":" ^ (S.toString ty)
     | prTerm (Fun (f,[],ty)) = (Fun.toString f) 
     | prTerm (Fun (f,ts,ty)) = "(" ^ (prTermList ts) ^ ")"
   and prTermList ts = PrintUtil.prSpace prTerm ts

   fun prRule (l,r) = (prTerm l) ^ " -> " ^ (prTerm r)  
   fun prEq (l,r) = (prTerm l) ^ " = " ^ (prTerm r)  

   val prTerms = L.map prTerm
   val prRules = L.map prRule
   val prEqs = L.map prEq


   fun prTerm' (Var (x,ty)) = (Var.toString x) ^ ":" ^ (S.toString ty)
     | prTerm' (Fun (f,[],ty)) = (Fun.toString f) ^ ":" ^ (S.toString ty)
     | prTerm' (Fun (f,ts,ty)) = "(" ^ (prTermList' ts) ^ ")" ^ ":" ^ (S.toString ty)
   and prTermList' ts = PrintUtil.prSpace prTerm' ts

   fun prRule' (l,r) = (prTerm' l) ^ " -> " ^ (prTerm' r)  
   fun prEq' (l,r) = (prTerm' l) ^ " = " ^ (prTerm' r)  

   val prTerms' = L.map prTerm'
   val prRules' = L.map prRule'
   val prEqs' = L.map prEq'


   end (* of local *)

   end; (* of structure IOStsrs *)

