(******************************************************************************
 * file: rwtools/stsrs/st_dp.sml
 * description: Dependency Pairs for Simply Typed SRS
 * author: AOTO Takahito
 * 
 * Copyright (c) 2008-2009, Toyama&Aoto Laboratory, Tohoku University
 ******************************************************************************)

signature ST_DP = 
   sig
   val markSymbol: Atom.atom -> Atom.atom
   val isMarkedSymbol: Atom.atom -> bool
   val unmarkSymbol: Atom.atom -> Atom.atom

   val markHead: Term.term -> Term.term
   val isMarkedHead: Term.term -> bool
   val unmarkHead: Term.term -> Term.term

   val dependencyPairs: FunSet.set 
			-> (Term.term * Term.term) list * (Term.term * Term.term) list 
			-> (Term.term * Term.term) list 

   val dependencyGraph: int FunMap.map 			
			-> (Term.term * Term.term) list 
			-> int list array

   val headInstantiation: (Term.term * Term.term) list 
			  -> (Term.term * Term.term) list 

   val getUsableHeadMap: (Term.term * Term.term) list 
			-> TermSet.set TermMap.map

   val getUsableHeadPairMap: FunSet.set -> TermIntSet.set 
			     -> (Term.term * Term.term) list 
			     -> TermIntSet.set TermIntMap.map

   val getUsableRules: TermSet.set TermMap.map 
		       -> (Term.term * Term.term) list 
                       -> (Term.term * Term.term) list 
		       -> int list 

   val applyAfToTerm: (int list) TermIntMap.map
		      -> TermIntSet.set
		      -> Term.term 
		      -> Term.term 

end;


structure STDp : ST_DP = 
   struct

   local 
       open Term
       open STTerm
       structure P = Prop
       structure S = Sort
       structure FS = FunSet
       structure VS = VarSet
       structure VM = VarMap
       structure TS = TermSet
       structure FM = FunMap
       structure TM = TermMap
       structure FIS = FunIntSet
       structure FIM = FunIntMap
       structure TIS = TermIntSet
       structure TIM = TermIntMap
       structure L = List
       structure LP = ListPair
       fun mapAppend f xs = List.foldr (fn (x,ys) => List.@(f x, ys)) [] xs
   in

   exception STDpError;

   val markingLetter = ref #"#"

   fun markSymbol a = 
       Atom.atom (concat [(Atom.toString a),(str (!markingLetter))])

   fun isMarkedSymbol a =
       let val s = (Atom.toString a)
	   val f = String.sub (s, (String.size s) - 1)
       in f = (!markingLetter)
       end

   fun unmarkSymbol a = 
       if isMarkedSymbol a
       then
	   let val s = (Atom.toString a)
	       val removed = String.substring (s, 0, (String.size s) - 1)
	   in Atom.atom removed
	   end
       else a

   fun markHead (Var ((v,i),a)) = Var ((markSymbol v,i),a)
     | markHead (Fun (f,[],a)) = Fun (markSymbol f,[],a)
     | markHead (Fun (f,t0::ts,a)) = 
       let val t0' = markHead t0
       in (Fun (f,t0'::ts,a))
       end

   fun isMarkedHead (Var ((v,_),_)) = isMarkedSymbol v
     | isMarkedHead (Fun (f,[],_))= isMarkedSymbol f
     | isMarkedHead (Fun (_,t0::_,_))= isMarkedHead t0

   fun unmarkHead (Var ((v,i),a)) = Var ((unmarkSymbol v,i),a)
     | unmarkHead (Fun (f,[],a)) =  Fun (unmarkSymbol f,[],a)
     | unmarkHead (Fun (f,t0::ts,a)) = 
       let val t0' = unmarkHead t0
       in (Fun (f,t0'::ts,a))
       end

   fun dependencyPairs dSymSet (rs,exrs) =
       let
	   fun headNonConstructorSubtermSet t = 
	       let  (* (headDefined?,  set of subterms) *)
		   fun dSub (t as Var _) accum = (true, TS.add (accum,t))
		     | dSub (t as Fun (f,[],_)) accum =
		       if FS.member (dSymSet,f) 
		       then (true,  TS.add (accum, t))
		       else (false, accum)
		     | dSub (t as Fun (f,t0::ts,_)) accum =
		       let val (b,accum2) = dSub t0 accum
		       in if b then (b, TS.add (dSubList ts accum2, t))
			  else (b, dSubList ts accum2)
		       end
		   and dSubList [] accum = accum
		     | dSubList (ti::ts) accum = 
		       let val (_,accum2) = dSub ti accum
		       in dSubList ts accum2
		       end
	       in
		   #2 (dSub t TS.empty)
	       end

       
	   fun mkDps (l,r) = 
	       let val ts = TS.listItems (headNonConstructorSubtermSet r)
(* 		   val _ = print (PrintUtil.prList toString ts) *)
		   val ts' = List.filter (fn t => not (Term.isAProperSubterm t l)) ts
		   val l' = markHead l
	       in L.map (fn s => (l', markHead s)) ts'
	       end

	   fun mkDpsEx (l,r) = 
	       case headOfTerm r of
		   Var _ => SOME (markHead l, markHead r)
		 | Fun (f,ts,a) => if FS.member (dSymSet,f)
				   then SOME (markHead l, markHead r)
				   else NONE
					
       in
	  L.@(mapAppend mkDps rs, L.mapPartial mkDpsEx exrs)
       end

  (* REN o CAP(t)  $B$H(B s $B$,(B unifiable $B$+(B *) 
  (* --- marking $B$,$5$l$F$$$k$H2>Dj(B *)
   fun isConnectable _ (Var (_,ty1), Var (_,ty2)) = Sort.equal (ty1,ty2)
     | isConnectable _ (Var (_,ty1), Fun (_,_,ty2)) = Sort.equal (ty1,ty2)
     | isConnectable _ (Fun (_,_,ty1), Var (_,ty2)) = Sort.equal (ty1,ty2)
     | isConnectable hdMap (t as (Fun (f,[],ty1)), Fun (g,ss,ty2)) = 
       Fun.equal (f,g) andalso (null ss)
       orelse (Sort.equal (ty1,ty2)
	       andalso case FM.find (hdMap, f) of
			   SOME d2 => d2 = 0
			 | NONE => false)
     | isConnectable hdMap (t as (Fun (_,t0::ts,ty1)), Fun (g,ss,ty2)) =
       Sort.equal (ty1,ty2)
       andalso 
       let val (h,d) = headAndDepthOfTerm t
       in case h of 
	      Var _ => if isMarkedHead h
		       then length (t0::ts) = length ss
			    andalso LP.all (isConnectable hdMap) (t0::ts,ss)
		       else true
	    | Fun (f,_,_) => case FM.find (hdMap, f) of
				 SOME d2 => d2 <= d 
					    orelse (length (t0::ts) = length ss
						    andalso LP.all (isConnectable hdMap) (t0::ts,ss))
			       | NONE => length (t0::ts) = length ss
					 andalso LP.all (isConnectable hdMap) (t0::ts,ss)
       end
	   
   fun dependencyGraph hdMap dps = 
       let val dg = Array.array (length dps, [])
	   val count = ref 0
	   fun updateDg r = 
	       let fun addEdge _ [] accum = accum
		     | addEdge n ((l,_)::ts) accum= 
		       if isConnectable hdMap (r,l)
		       then addEdge (n+1) ts (n::accum)
		       else addEdge (n+1) ts accum
		   val edges = rev (addEdge 0 dps [])
	       in Array.update (dg, !count, edges);
		  count := !count + 1
	       end
	   val _ = L.app (fn (_,r) => updateDg r) dps
       in
	   dg
       end

   fun headInstantiateRule (l,r) ts =
       let val h = headOfTerm r
	   fun getConnection (p as Var (v,ty1), t) =
	       if Sort.equal (ty1, sortOfTerm t)
	       then SOME (VM.singleton (v, extractPattern t))
	       else NONE
	     | getConnection (Fun (_,_::_,_), Fun (_,[],_)) = NONE
	     | getConnection (Fun (_,t0::_,ty1), Fun (_,s0::_,ty2)) =
	       if Sort.equal (ty1,ty2)
	       then getConnection (t0, s0)
	       else NONE
	     | getConnection (_, _) = 
	       (print "headInstantiation"; raise STDpError)
       in case h of 
	      Fun _ => [(l,r)]
	    | Var _ => let val r1 = unmarkHead r
			   val ts1 = L.map unmarkHead ts 
			   val dummy = Atom.atom ""
			   val (Fun (_,[l',r'],_) :: Fun (_,ts',_) :: nil) =
			       Subst.renameTerms [ Fun (dummy,[l,r1],Sort.null),
						   Fun (dummy,ts1,Sort.null) ]
		       in L.mapPartial
			      (fn u => case getConnection (r',u) of
					   SOME sigma => SOME (Subst.applySubst sigma l',
							       markHead (Subst.applySubst sigma r'))
					 | NONE => NONE)
			      ts'
		       end
       end

   fun headInstantiation dps =
       let val lhss = TS.listItems (L.foldl (fn ((l,_),set)=> TS.add (set,l)) TS.empty dps)
       in
	   mapAppend (fn (l,r) => headInstantiateRule (l,r) lhss) dps
       end


   fun getUsableHeadMap rules =
       let
	   val definedHeads = L.map (fn (l,_) => headOfTerm l) rules
	   val definedHeadSet = TS.addList (TS.empty, definedHeads)

	   fun derivative (l,r) =
	       let val lHead = headOfTerm l 
		   val rHeadSet = TS.intersection (definedHeadSet,
						TS.addList (TS.empty, headOfSubterms r))
		   val pVarSet = L.foldl
				 (fn (v,set) => TS.add (set,STTerm.anonymizeTerm v))
				 TS.empty
				 (STTerm.primaryVariablesOfTerm r)
	       in (lHead, rHeadSet, pVarSet)
	       end

	   val exprules = rules @ (mapAppend (fn lr => lr::STSrs.etaExpandRule lr) rules)
	   val infoList = List.map derivative exprules
	   val primaryVarSet = List.foldl
				   (fn ((_,_,pVarSet),set) => TS.union (set,pVarSet))
				   TS.empty
				   infoList

           val items = definedHeads @ (TS.listItems primaryVarSet)
           fun findIndex s = 
	       let fun findIndexSub _ _ [] = NONE
		     | findIndexSub s n (t::ts) = if Term.equal (s,t)
						  then SOME n 
						  else findIndexSub s (n+1) ts
	       in
		   findIndexSub s 0 items
	       end

           val termMap0 = 
	       let fun makeTableSub [] tm = tm
                     | makeTableSub ((h,fset,vset)::xs) tm =
		       let val set = case (TM.find (tm,h)) of
					       SOME gset => gset | NONE => TS.empty
		       in
                           makeTableSub xs (TM.insert (tm, h, 
						       TS.union (vset, TS.union (fset,set))))
		       end
	       in
                   makeTableSub infoList TM.empty
               end

           val termMap1 = 
	       TS.foldl       
	       (fn (v,tm) => let val set = case (TM.find (tm,v)) of
					       SOME gset => gset | NONE => TS.empty
			     in
				 TM.insert (tm, v, 
					    TS.addList (set,
							L.filter
							    (fn h => Sort.isASubsort
									 (Term.sortOfTerm v)
									 (Term.sortOfTerm h))
							    definedHeads))
			     end)
	       termMap0
	       primaryVarSet

           fun getClosure tm = 
               let val ar = Array.array (length items,[])
                   val _ = TM.appi (fn (t,set) => 
                                       Array.update (ar,
                                                     valOf (findIndex t),
                                                     L.map (fn u => valOf (findIndex u))
							   (TS.listItems set)))
                                  tm
                   val graph2 = Graph.reflexiveTransitiveClosure ar
               in
                  Array.foldri
                      (fn (i,ns,tm) => 
			  TM.insert (tm,
				     L.nth (items,i),
				     TS.addList (TS.empty, L.map (fn j => L.nth (items,j)) ns)))
                      TM.empty
                      graph2
               end
       in
	   getClosure termMap1
       end
       
  (* varUPairSet $B$O(B rules $B$H(B hdps $B$+$i(B $B!%!%!%(B *)
   fun getUsableHeadPairMap dSymSet varUpairSet rules =
       let
        (* $B:8JU$N(B hpair(l)$B!$1&JU$N(Busable pairs$B$N=89g(B *)
	   fun derivative (l,r) =
	       let val hpair = STTerm.headAndDepthOfTerm l
		   val definedUpairs = L.filter (fn (h,_) => STTerm.isDefinedHead dSymSet h)
						(STTerm.headAndDepthOfSubterms r)
		   val varUpairs = L.filter (fn (h,_) => STTerm.isVarHead h)
					    (STTerm.headAndDepthOfApplicativeSubterms' r)
		   val definedUpairSet = L.foldl
					 (fn (headDep,set) => TIS.add (set,headDep))
					 TIS.empty definedUpairs
		   val varUpairSet = L.foldl
					 (fn (headDep,set) => TIS.add (set,headDep))
					 TIS.empty varUpairs
				 
	       in (hpair, definedUpairSet, varUpairSet)
	       end

	   val exprules = rules @ (mapAppend (fn lr => lr::STSrs.etaExpandRule lr) rules)
	   val infoList = List.map derivative exprules

	   val dUpairSet =  List.foldl (fn ((hi,dup,_),set) => TIS.add (TIS.union (dup,set), hi))
				       TIS.empty infoList

           val items = (TIS.listItems dUpairSet) @ (TIS.listItems varUpairSet)

(* 	   val _ = print "index of items\n" *)
(* 	   val _ = L.app (fn (s,i) => print ((Term.toString s) ^ "," ^  (Int.toString i) ^ "\n")) *)
(* 			 items *)
(* 	   val _ = print "end index\n" *)

           fun findIndex (s,i) = 
	       let fun findIndexSub _ [] = (print ("Error: findIndex, the index of <"
						   ^ (Term.toString s) ^ "," ^  (Int.toString i) 
						   ^ "> is not found\n"); NONE)
		     | findIndexSub n ((t,j)::ts) = 
		       if Term.equal (s,t) andalso i = j
		       then SOME n 
		       else findIndexSub (n+1) ts
	       in
		   findIndexSub 0 items
	       end

         (* defined usable pairs  => usable pairs in rhs *)
           val termMap0 = 
	       let fun makeTableSub [] tim = tim
                     | makeTableSub ((hp,fset,vset)::xs) tim =
		       let val set = case (TIM.find (tim,hp)) of
					       SOME gset => gset | NONE => TIS.empty
		       in
                           makeTableSub xs (TIM.insert (tim, hp, 
						       TIS.union (vset, TIS.union (fset,set))))
		       end
	       in
                   makeTableSub infoList TIM.empty
               end


         (* applicative var => defined usable pairs *)
           val termMap1 = 
	       TIS.foldl       
	       (fn ((v,i),tim) => let val set = case (TIM.find (tim,(v,i))) of
					       SOME gset => gset | NONE => TIS.empty
                                  (* $B$3$3$O6u$N$O$:!$ITMW$G$O(B ? *)
			     in
				 TIM.insert (tim, (v,i), 
					    TIS.union (set,
						       TIS.filter
							   (fn (h,j) => case
								   STTerm.sortOfHeadAndDepth (h,j-i) of
								   NONE => false
								 | SOME ty => Sort.compare
										  (Term.sortOfTerm v,ty)
									      = EQUAL)
							   dUpairSet))
			     end)
	       termMap0
	       varUpairSet



           fun getClosure tim = 
               let val ar = Array.array (length items,[])
                   val _ = TIM.appi (fn ((t,i),set) => 
					 Array.update (ar,
                                                     valOf (findIndex (t,i)),
                                                     L.map (fn u => valOf (findIndex u))
							   (TIS.listItems set)))
                                  tim

                   val graph2 = Graph.reflexiveTransitiveClosure ar
               in
                  Array.foldri
                      (fn (i,ns,tim) => 
			  TIM.insert (tim,
				     L.nth (items,i),
				     TIS.addList (TIS.empty, L.map (fn j => L.nth (items,j)) ns)))
                      TIM.empty
                      graph2
               end
       in
	   getClosure termMap1
       end




   fun getUsableRules usableHM rules dps = 
       let
	   fun usableTermSet tm ts =
	       L.foldl (fn (h,ac) => case (TM.find (tm,h)) of
					  SOME set => TS.union (set,ac)
					| NONE => ac)
			TS.empty
			ts
			
           val usableTS = 
	       usableTermSet usableHM 
		             (mapAppend 
				  (fn (_,r) => STTerm.headOfNonVariableSubterms' r)
				  dps)
(****
	   val _ = print "###############\n"
	   val _ = L.app (fn t => print (" " ^ (Term.toString t)))
			  (mapAppend 
			       (fn (_,r) => STTerm.headOfNonVariableSubterms' r)
			       dps)
	   val _ = print "\n###############\n"
	   val _ = TM.appi (fn (k,tset) => (print ((Term.toString k) ^ " := ");
					    TS.app (fn t => print (", " ^ (Term.toString t)))
						   tset;
					    print "\n"))
			   usableHM
	   val _ = print "\n###############\n"
	   val _ = TS.app (fn t => print (" " ^ (Term.toString t)))
			  usableTS
	   val _ = print "\n###############\n"
******)

       in
           L.filter (fn i => let val (l,_) = L.nth (rules,i)
			     in TS.member (usableTS, STTerm.headOfTerm l)
			     end)
		    (L.tabulate (L.length rules, fn i =>i))
       end



   fun applyAfToTerm piMap colSet (t as (Var _)) = t
     | applyAfToTerm piMap colSet (t as (Fun (_,[],_))) = t
     | applyAfToTerm piMap colSet (t as (Fun (f,t0::ts,ty))) =
       let val (h,d) = STTerm.headAndDepthOfTerm' t0
(*	   val _ = print ("(" ^ (STTerm.toString h)
			  ^ ","  ^ (Int.toString d) ^ ")") *)
       in
	   case TIM.find (piMap, (h,d)) of 
	       SOME ns => if TIS.member (colSet, (h,d))
			  then applyAfToTerm piMap colSet (L.nth (t0::ts, hd ns))
			  else Fun (f,
				    L.map (fn i => applyAfToTerm piMap colSet
								 (L.nth (t0::ts,i))) ns,
				    Sort.null)
	     | _=> (print "applyAfToTerm: (";
		    print (STTerm.toString h);
		    print ",";
		    print (Int.toString d);
		    print ")\n";
		    raise STDpError)
       end

   end (* of local *)



end (* of structure STDp *)

