(******************************************************************************
 * file: rwtools/stsrs/st_dpsolver.sml
 * description: Termination Solver for STSRS by Dependency Pair Framework
 * author: AOTO Takahito
 * 
 * Copyright (c) 2008-2009, Toyama&Aoto Laboratory, Tohoku University
 ******************************************************************************)

signature ST_DPSOLVER =  
   sig
      val runDebug: bool ref

      type options = {
	   useRp:bool         (* reduction pair $B$r;H$&$+$I$&$+(B *)
	 , useSc:bool         (* subterm criteria $B$r;H$&$+$I$&$+(B *)
	 , useUr:bool         (* usable rules criteria $B$r;H$&$+$I$&$+(B *)
	 , useIn: bool       (* inntermost $B$X$N%9%$%C%A$r;H$&$+$I$&$+(B[$BL$<BAu(B]*)
      }

      val dpSolver: string 
		    -> string
		    -> options
		    -> STPoSolver.options
		    -> (Atom.atom list * Term.decl list * (Term.term * Term.term) list)
		    -> bool

   end;

structure STDpSolver : ST_DPSOLVER =
   struct

   local 
       open Term
       structure A = Atom
       structure AT = AtomTable
(*       structure CP = Compiler.Profile *)
       structure FM = FunMap
       structure FS = FunSet
       structure IM = IntMap
       structure IS = IntSet
       structure L = List
       structure LP = ListPair
       structure FIS = FunIntSet
       structure FIT = FunIntTable
       structure FIIT = FunIntIntTable
       structure FPT = FunPairTable
       structure SS = SortSet
       structure TS = TermSet
       structure TM = TermMap
       structure TPM = TermPairMap
       structure TIS = TermIntSet
       structure TIM = TermIntMap
       fun mapAppend f xs = List.foldr (fn (x,ys) => List.@(f x, ys)) [] xs
       fun member x ys = isSome (L.find (fn y => x = y) ys) 
       fun selectIndexes sub all = 
	   let fun select [] _ _ ans = SOME (rev ans)
		 | select (x::xs) [] _ _ = NONE
		 | select (x::xs) (y::ys) i ans = 
		   if x = y 
		   then select xs ys (i+1) (i::ans)
		   else select (x::xs) ys (i+1) ans
	   in  
	       select sub all 0 []
	   end
       local
	   fun logMain n k ans = if n <= k then ans
				 else logMain n (k * 2) (ans + 1)
       in
       fun log n = logMain n 2 1
      end
   in 

   val runDebug = ref false : bool ref
   fun debug f = if !runDebug then f () else ()

   val inTmpFileName = ref ""
   val outTmpFileName = ref ""

   val minisatPath = ref ""
   val tmpDir = ref ""

   datatype strategy = SN | SIN  (* termination or inntermost termination *)
			    
   type dpproblem  = (Term.term * Term.term) list    (* dependency pairs *)
		      * Trs.trs * strategy

   type options = {
	useRp:bool         (* reduction pair $B$r;H$&$+$I$&$+(B *)
      ,	useSc:bool         (* subterm criteria $B$r;H$&$+$I$&$+(B *)
      ,	useUr:bool         (* usable rules criteria $B$r;H$&$+$I$&$+(B *)
      , useIn:bool       (* inntermost $B$X$N%9%$%C%A$r;H$&$+$I$&$+(B[$BL$<BAu(B]*)
   }


   exception STDpSolverError
   exception Success
   exception Failure

   fun dependencyGraphProcessor (hdps,stsrs:STSrs.stsrs,strat) idxes =
       if null idxes
       then []
       else
	   let 
	       val _ = debug (fn _ => print "\nDependency Graph Processor...\n")

	       val _ = debug (fn _ => print "Rewrite Rules:\n")
	       val _ = debug (fn _ => print (STSrs.prRules (#Rules stsrs)))

	       val _ = debug (fn _ => print "Dependency Pairs:\n")
	       val current = L.map (fn i => L.nth (hdps,i)) idxes
	       val _ = debug (fn _ => print (STSrs.prRules current))

	       val dependencyGraph = STDp.dependencyGraph (#HdMap stsrs) current

	       val _ = debug (fn _ => if strat = DpSolver.SN
				      then print "Dependency Graph:\n"
				      else print "Innermost Dependency Graph:\n")

	       val _ = debug (fn _ => print (Graph.toString dependencyGraph))

	       val scc = Graph.scc dependencyGraph (* list of int set *)
	       val _ = debug (fn _ => print "scc: ")
	       val _ = debug (fn _ => L.app (print o IntSet.toString) scc)
	       val _ = debug (fn _ => print "\n")

	   in
	       L.map (fn ns => L.map (fn n=> L.nth (idxes,n)) (IntSet.listItems ns)) scc
	   end

   fun subtermCriteriaProcessor path dir (isFlatted,dps,stsrs:STSrs.stsrs,strat) idxes =
       if null idxes
       then []
       else
	   let 
	   val _ = debug (fn _ => print "\nSubterm Criteria Processor...\n")

	   val fdps = if isFlatted
		      then dps
		      else STSrs.flattingRules dps

	   val current = L.map (fn i => L.nth (fdps,i)) idxes
			 
	   val _ = debug (fn _ => print "Flatted Dependency Pairs:\n")
	   val _ = debug (fn _ => print (Trs.prRules current))

	   val (result,remIdxes) = Dp.applySubtermCriteria path dir current
	   val _ = debug (fn _ => print ("=>" ^ (if result then "T" else "F") ^ "\n"))
	   val _ = debug (fn _ => if result 
				  then print ("Remained Flatted Dps:\n"
					      ^ (Trs.prRules 
						     (L.map (fn i=> L.nth (current,i)) remIdxes)))
				  else ())
						      
	   val rems = L.map (fn n=> L.nth (idxes,n)) remIdxes

	   in
	       if result 
	       then if null rems
		    then []
		    else mapAppend (subtermCriteriaProcessor path dir (true,fdps,stsrs,strat))
				   (dependencyGraphProcessor (dps,stsrs,strat) rems)
					     
	       else [ rems ]
	   end


   fun dpSolver path dir (opt:options) (opt2:STPoSolver.options) (sorts,decls,srules) = 
       let 
	   val _ = minisatPath := path
	   val _ = tmpDir := dir

           (* compute eta expansion *)
	   val _ = debug (fn _=> print "Expansions:\n")
	   val exprules =  STSrs.etaExpandRules srules
	   val _ = debug (fn _=> print (PrintUtil.prList IOStsrs.prRule exprules))

           (* check defined constants and constructor constants *)
	   val stsrs =  STSrs.rulesToStsrs srules 
	   val _ = debug (fn _=> print "Constants:\n")
	   val _ = debug (fn _=> print (PrintUtil.prList Fun.toString (FS.listItems (#FSymSet stsrs))))
	   val _ = debug (fn _=> print "Defined symbols:\n")
	   val _ = debug (fn _=> print (PrintUtil.prList Fun.toString (FS.listItems (#DSymSet stsrs))))
	   val _ = debug (fn _=> print "Constructors:\n")
	   val _ = debug (fn _=> print (PrintUtil.prList Fun.toString (FS.listItems (#CSymSet stsrs))))
	   val _ = debug (fn _=> print "Head and Depth:\n")
	   val _ = debug (fn _=> print (PrintUtil.prList 
			      (fn (f,i) => "(" ^ (Fun.toString f) 
					   ^ "," ^ (Int.toString i) ^ ")")
			      (FM.listItemsi (#HdMap stsrs))))

           (* compuate dependency pairs *)
	   val dependencyPairs =  STDp.dependencyPairs (#DSymSet stsrs) (srules,exprules)
	   val _ = debug (fn _=> print "Dependency Pairs:\n")
	   val _ = debug (fn _=> print (PrintUtil.prList IOStsrs.prRule dependencyPairs))

           (* compuate head instantiation *)
	   val hdps =  STDp.headInstantiation dependencyPairs
	   val _ = debug (fn _=> print "Head-Instatiated Dependency Pairs:\n")
	   val _ = debug (fn _=> print (PrintUtil.prList IOStsrs.prRule hdps))

           (* switch to innermost termination (yet not implmemented) *)
	   val strat = DpSolver.SN

           (* apply the standard processer: SC and DPG Processor *)
	   val problems = 
	       let
		   val idxes = L.tabulate(L.length hdps,fn i=>i)
	       in
		   if #useSc opt
		   then
 		       mapAppend (subtermCriteriaProcessor path dir (false,hdps,stsrs,strat))
	       			 (dependencyGraphProcessor (hdps,stsrs,strat) idxes)
		   
		   else
	       	       dependencyGraphProcessor (hdps,stsrs,strat) idxes
	       end

	   val _ = if null problems 
		   then raise Success
		   else if not (#useRp opt)
		   then raise Failure
		   else ()

	   val hdpsIndexesAll = IS.listItems (L.foldl (fn (ns,set)=>IS.addList (set,ns)) IS.empty problems)
	   val idxesList = L.map (fn sub => case selectIndexes sub hdpsIndexesAll of
						SOME ns => ns
					      | NONE => (print "dpSolver:selectIndexes\n";
							 raise STDpSolverError))
				 problems
	   val hdplist = L.map (fn i => L.nth (hdps, i)) hdpsIndexesAll

	   val trs =  Trs.rulesToTrs srules 
	   val rulesAndHdps =  srules @ hdplist

	   val arules =  STSrs.annotateRules srules 
	   val aexprules =  STSrs.annotateRules exprules 
	   val adplist =  STSrs.annotateRules hdplist
	   val arulesAndHdps = arules @ adplist

	   val faMap  = Trs.funArityMapInRules arulesAndHdps
	   val faSet  = FM.foldri (fn (k,v,xs)=> FIS.add (xs,(k,v))) FIS.empty faMap
	   val funSet  = FM.foldri (fn (k,v,xs)=> FS.add (xs,k)) FS.empty faMap
	   val faList  = FM.listItemsi faMap
	   val fs  = L.map #1 faList
	   val lenFs  = L.length fs
	   val rowLen  = log lenFs


	   (****************************************)
	   (*** Prepare Reduction Pair Processor ***)
	   (****************************************)

	   val _ = debug (fn _ => print "\nPreparing for Reduction Pair Processor...\n")

	   val symCount = ref 0

	   val encoding = STPoSolver.mkEncodingInfo symCount (fs,lenFs,rowLen,faList,faSet,faMap)

(* 	   val _ = debug (fn _ => print "List of Annotated Functions with Arity\n") *)
(* 	   val _ = debug (fn _ => L.app (fn (f,i) => (print " "; *)
(* 				       print (Atom.toString f); *)
(* 				       print " (arity: "; *)
(* 				       print (Int.toString i); *)
(* 				       print ")\n")) *)
(* 			 faList) *)

	   (* for argument filtering *)
	   val applicativeHeadAndDepthSet = 
	       L.foldl (fn ((l,r),set) => 
			   TIS.addList (TIS.addList (set,
						      STTerm.headAndDepthOfApplicativeSubterms' l),
					STTerm.headAndDepthOfApplicativeSubterms' r))
			TIS.empty
			rulesAndHdps

(* 	   val _ = debug (fn _ => print "List of (applicative Heads)x(Depth) pairs with Type\n") *)
(* 	   val _ = debug (fn _ => TIS.app (fn (h,d) => (print " <"; *)
(* 				       print (Term.toString h); *)
(* 				       print ", "; *)
(* 				       print (Int.toString d); *)
(* 				       print ">: "; *)
(* 				       print (Sort.toString (valOf (STTerm.sortOfHeadAndDepth (h,d)))); *)
(* 				       print "\n")) *)
(* 			 applicativeHeadAndDepthSet) *)

	   (* types that need to be syncronized *)
	   fun stabilityTypes ds rs = 
	       let 
                   (* vhSubterm: a subterm u s.t. hd(u) \in V *)
		   val sortSetOfVhSubterms =
		       L.foldl (fn ((l,r),set) => 
				   SS.addList(SS.addList (set,
							  STTerm.sortsOfVhSubterms l),
					      STTerm.sortsOfVhSubterms r))
			       SS.empty
			       (ds @ rs)

		   (* add all supersorts of rules of non basic types *)
		   val sortsOfNonBasicRules = 
		       L.filter 
			   (not o Sort.isBaseType)
			   (mapAppend Sort.supersorts
				      (L.map (fn (l,_) => sortOfTerm l) rs))
	       in
		   SS.addList (SS.filter (not o Sort.isBaseType) sortSetOfVhSubterms, 
			       sortsOfNonBasicRules)
	       end

	   val stypes = SS.listItems (stabilityTypes hdps srules)

(* 	   val _ = debug (fn _ => print "Stability Types\n") *)
(* 	   val _ = debug (fn _ => print (PrintUtil.prList Sort.toString stypes)) *)

	   (* syncronization conditions for each stability types *)
	   val stabilityConds = 
	       let 
		   val applicativeUnmarkedHeadAndDepthSet = 
		       TIS.filter
			   (fn (h,d) => not (Sort.isBaseType (valOf (STTerm.sortOfHeadAndDepth (h,d))))
					andalso not (STDp.isMarkedHead h)) 
			   applicativeHeadAndDepthSet
			   
		   val hdList = TIS.foldl
				     (fn ((h,d),xs) => (h,d,valOf (STTerm.sortOfHeadAndDepth (h,d)))::xs)
				     []
				     applicativeUnmarkedHeadAndDepthSet
	       in
		   L.map (fn s =>
			      case s of
				  Sort.Base _ => (print "unexpected error";
						  raise STDpSolverError)
				| Sort.Proc (ts,t0) =>
				  STPoSolver.mkEqAfCond 
				      encoding
				      (length ts + 1,
			     	       L.mapPartial
					   (fn (h,d,ty) => if Sort.equal (s,ty)
							   then SOME (STTerm.annotateApp (h,d,ty))
							   else NONE)
					   hdList))
			 stypes
	       end

	   val strictPrecCond = STPoSolver.mkStrictPrecCond opt2 encoding
	   val lexCond = STPoSolver.mkLexCond opt2 encoding
	   val quasiRpoCond = STPoSolver.mkQuasiRpoCond opt2 encoding
	   val afCond = STPoSolver.mkAfCond opt2 encoding

          (* DP $B$d(B rules $B$KBP1~$9$k(B index $B$rMQ0U(B *)
	   val lenrules = length arules
	   val lendplist = length adplist
	   val lenstcondlist = length stypes
	   val count = !symCount + 1
	   val dplistForGt = LP.zip (L.tabulate (lendplist,fn x => count + x), adplist)
	   val dplistForEq = LP.zip (L.tabulate (lendplist,fn x => count + lendplist + x), adplist)
	   val rslistForGe = LP.zip (L.tabulate (lenrules,fn x => count + 2*lendplist + x), arules)
	   val stcondlist = LP.zip (L.tabulate (lenrules,fn x => count + 2*lendplist + lenrules + x), 
				    stypes)
 	   val _ = symCount := count + lendplist*2  + lenrules + lenstcondlist
           (* +1 for condition prop *)

          (* index for each hpair, used for usable rules *)

	   val hpairsCountMap =
	       if #useRp opt
	       then let 
		       val hpairSet = 
			   L.foldl (fn ((l,_),set) => TIS.add (set, STTerm.headAndDepthOfTerm' l))
				   TIS.empty (srules @ exprules)
		   in
		       TIS.foldl
			(fn ((t,i),map) => TIM.insert(map,(t,i),
						      (symCount := (!symCount) + 1;!symCount)))
			TIM.empty hpairSet
		   end
	       else TIM.empty

           (* $B@)LsO@M}<0$KBP1~$7$?L?BjJQ?t$rMQ0U$7$F!$(BIff ($BL?BjJQ?t!$@)LsO@M}<0(B) $B$rEPO?(B  *)
	   (* $B<B:]$N(B encoding $B$O8zN($N$?$a!$8e$GI,MW$K$J$C$?$H$-$K$d$k(B *)
	   (* constraints u > v for  u -> v \in DP(R) *)
	   val propMap0 =
	       List.foldr
		   (fn ((i,lr),mp) =>
		       IM.insert (mp,i,
			       fn ()
				  => Prop.eqCnf' symCount
						 (Prop.simplifyProp
						      (Prop.Iff (Prop.Atom i,
								 STPoSolver.encodeStOrderConstraint
								     Order.GT symCount encoding
								     opt2 lr)))))
		   IM.empty
		   dplistForGt
		   
	   (* adding constraints u \simeq v for  u -> v \in DP(R) to propMap0 *)
	   val propMap1 =
	       List.foldr
	           (fn ((i,lr),mp) =>
		       IM.insert (mp,i,
			       fn ()
				  => Prop.eqCnf' symCount
						 (Prop.simplifyProp
						      (Prop.Iff (Prop.Atom i,
								 STPoSolver.encodeStOrderConstraint
								     Order.EQ symCount encoding
								     opt2 lr)))))
		   propMap0
		   dplistForEq

	   (* adding constraints u \leq v for  l -> r \in R  to propMap1 *)
	   val propMap2 =
	       List.foldr
	           (fn ((i,lr),mp) =>
		       IM.insert(mp,i,
			      fn () =>
				 Prop.eqCnf' symCount
					     (Prop.simplifyProp
						  (Prop.Iff (Prop.Atom i,
							     STPoSolver.encodeStOrderConstraint
								 Order.GE symCount encoding
								 opt2 lr)))))
		   propMap1
		   rslistForGe

	   (* adding encoding conditions to propMap3 *)
	   val propMap3 =
	       let val stCond = (* stability condition *)
		       Prop.Conj (L.tabulate (lenstcondlist,
					   fn i => Prop.Iff (Prop.Atom (count + 2*lendplist 
									+ lenrules + i),
							     L.nth (stabilityConds, i))))
	       in
		   IM.insert (propMap2,
			      count + lendplist*2 + lenrules,
			   fn () => Prop.eqCnf' symCount
						(Prop.simplifyProp 
						     (Prop.Conj
							  [strictPrecCond,
							   lexCond,
							   stCond,
							   afCond,
							   quasiRpoCond ])))
	       end


          (* Iff ($BL?BjJQ?t!$@)LsO@M}<0(B) $B$N(B encoding $B$O(B lookup $B$,=i$a$F$"$C$?;~$K$9$k(B *)
          (* $B:FMxMQ$9$k$?$a$K(B propMapDone $B$K(B encoding $B$,:Q$s$@$b$N$rJ];}(B *)
	   val propMapDone = ref IM.empty
	   fun lookupProp i =
	       case IM.find (!propMapDone,i) of
		   SOME p => p
		 | NONE => (case IM.find (propMap3, i) of
				SOME f => let val q = f ()
					  in (propMapDone := IM.insert (!propMapDone,i,q); q)
					  end
			      | NONE => (print ("Failing lookupProp: " ^ (Int.toString i)
						^ "(base " ^ (Int.toString count) ^ ")\n");
					 raise STDpSolverError))

	   (* $BMxMQ2DG=5,B'$r7W;;$9$k$?$a$N%F!<%V%k(B *)
 	   val usableHM = if #useUr opt
			  then STDp.getUsableHeadMap srules
			  else TermMap.empty

	   (* $B0z?t%U%#%k%?%j%s%0$r$7$?8e$KMxMQ2DG=5,B'$N@)Ls$,@.N)$9$k$3$H$r<($9L?Bj$r@8@.(B *)
	   (* ts: dp $B$N1&JU$r=8$a$?%j%9%H(B *)
(* 	   fun mkUsableRulesProp ts rsIdxs = *)
(* 	       let *)
(* 		   open Term; *)
(* 		   fun funRulesIdx hSet =  *)
(* 		       List.mapPartial *)
(* 			   (fn i => if TS.member (hSet,  *)
(* 						  STTerm.headOfTerm (#1 (L.nth (arules,i)))) *)
(* 				    then SOME (Prop.Atom (count + 2*lendplist + i)) *)
(* 				    else NONE) *)
(* 			   rsIdxs *)
(* 		   fun usables (Var _)  = Prop.True *)
(* 		     | usables (t as (Fun (f,ts,_))) = *)
(* 		       let *)
(* 			   val ps = L.tabulate *)
(* 					(length ts, *)
(* 				      fn i => if #useAf opt2 *)
(* 					      then Prop.Imp (STPoSolver.encodePiStatus encoding (f,i+1), *)
(* 							     usables (L.nth (ts,i))) *)
(* 					      else usables (L.nth (ts,i))) *)

(* 			   val h = STTerm.anonymizeTerm (STTerm.headOfTerm t) *)

(* 		       in *)
(* 			   case TM.find (usableHM,h) of *)
(* 			       SOME tset => Prop.Conj (L.@(funRulesIdx tset, ps)) *)
(* 			     | NONE => Prop.Conj ps *)
(* 		       end *)
(* 	       in *)
(* 		   Prop.eqCnf' symCount (Prop.simplifyProp (Prop.Conj (L.map usables ts))) *)
(* 	       end *)

	   

(*  	   val usableHPM = if #useUr opt *)
(* 			  then let val varUpairSet = TIS.filter (fn (h,_) => STTerm.isVarHead h)  *)
(* 								applicativeHeadAndDepthSet *)
(* 			       in *)
(* 				   STDp.getUsableHeadPairMap (#DSymSet stsrs) varUpairSet srules *)
(* 			       end *)
(* 			   else TermIntMap.empty *)

(* 	   val _ = debug (fn () => print "Usable Pair Map\n") *)
(* 	   val _ = debug (fn () => List.app *)
(* 		   (fn (((s,i),set)) =>  *)
(* 		       (print ("<" ^ (Term.toString s) ^ ","   *)
(* 			       ^ (Int.toString i) ^ ">: "); *)
(* 			print (PrintUtil.prSetInOneLine (fn (t,j) => "<" ^ (Term.toString t) ^ ","   *)
(* 							 ^ (Int.toString j) ^ ">") *)
(* 							(TIS.listItems set)); *)
(* 			print "\n")) *)
(* 		   (TermIntMap.listItemsi usableHPM)) *)


	   (* $B0z?t%U%#%k%?%j%s%0$r$7$?8e$KMxMQ2DG=5,B'$N@)Ls$,@.N)$9$k$3$H$r<($9L?Bj$r@8@.(B *)
	   (* TO DO: rule $B$K$D$$$F$N%U%#%k%?%j%s%0$,$J$$!%!%!%(B *)
	   (* ts: dp $B$N1&JU$r=8$a$?%j%9%H(B *)
(* 	   fun mkUsableRulesProp2 ts rsIdxs = *)
(* 	       let *)
(* 		   open Term; *)
(* 		   fun funRulesIdx hiSet = *)
(* 		       List.mapPartial *)
(* 			   (fn i => if TIS.member (hiSet, *)
(* 						  STTerm.headAndDepthOfTerm' (#1 (L.nth (srules,i)))) *)
(* 				    then SOME (Prop.Atom (count + 2*lendplist + i)) *)
(* 				    else NONE) *)
(* 			   rsIdxs *)
(* 		   fun usables (Var _)  = Prop.True *)
(* 		     | usables (t as (Fun (f,ts,_))) = *)
(* 		       let *)
(* 			   val ps = L.tabulate *)
(* 					(length ts, *)
(* 				      fn i => if #useAf opt2 *)
(* 					      then Prop.Imp (STPoSolver.encodePiStatus encoding (f,i+1), *)
(* 							     usables (L.nth (ts,i))) *)
(* 					      else usables (L.nth (ts,i))) *)
(* 			   val (h,i) = STTerm.headAndDepthOfTerm' t *)
(*  			   val his = L.tabulate (i+1, fn x => (h,x))   *)
(* 			   val tiset = L.foldl *)
(* 					   (fn (set1,set2) => TIS.union (set1,set2)) *)
(* 					   TIS.empty *)
(* 					   (L.mapPartial (fn hi => TIM.find (usableHPM,hi)) his) *)
(* 		       in *)
(* 			   Prop.Conj (L.@(funRulesIdx tiset, ps)) *)
(* 		       end *)
(* 	       in *)
(* 		   Prop.eqCnf' symCount (Prop.simplifyProp (Prop.Conj (L.map usables ts))) *)
(* 	       end *)


	   fun mkUsableRulesProp3 ts rsIdxs =
	       let
		   open Prop
		   open Term

		   fun useFunArg (f,i) = STPoSolver.encodePiStatus encoding (f,i+1)

		   fun usableTerm s = 
		       case s of
			   Var _ => True
			 | Fun (f,ts,_) => (* applicative *)
			   let val (h,d) = STTerm.headAndDepthOfTerm' s
			       val headProps = 
				   case h of
				       Var (x,ty0) => 
					   L.mapPartial
					       (fn ((h1,d1),j) => 
						   case STTerm.sortOfHeadAndDepth (h1,d1 - d)
						    of SOME ty1 => if Sort.compare (ty0,ty1)
								      = EQUAL
								   then SOME (Atom j)
								   else NONE
						     | NONE => NONE)
					       (TIM.listItemsi hpairsCountMap)
				     | Fun _ => L.mapPartial 
						    (fn i => case TIM.find (hpairsCountMap,(h,i)) of
								 SOME j => SOME (Atom j) | NONE => NONE)
						    (L.tabulate (d+1, fn x => x))

			       val argProps = 
				   L.tabulate (length ts, 
					    fn i => if #useAf opt2 
						    then Imp (useFunArg (f,i), usableTerm (L.nth (ts,i)))
						    else usableTerm (L.nth (ts,i)))
			   in Conj (argProps @ headProps)
			   end

		   val usableRule = 
		       Conj (L.mapPartial 
				 (fn i => 
				     let val (l,r) = L.nth (arules,i)
					 val (h,d) = STTerm.headAndDepthOfTerm' l
				     in case TIM.find (hpairsCountMap,(h,d)) of
						SOME j => SOME (Imp (Atom j,
								     Conj [Atom (count + 2*lendplist + i),
									   usableTerm r]))
    					      | NONE => NONE
				     end)
				 rsIdxs)

		   val expUsableRule = 
		       Conj (L.mapPartial 
				 (fn (l,r) => 
				     let val (h,d) = STTerm.headAndDepthOfTerm' l
				     in case TIM.find (hpairsCountMap,(h,d)) of
					    SOME j => SOME (Imp (Atom j,usableTerm r))
    					  | NONE => NONE
				     end)
				 aexprules)

		   val uprop = Prop.Conj (L.@ (L.map usableTerm ts, [usableRule,expUsableRule]))

(* 		   val _ = print (Prop.printProp uprop)  *)

	       in 
		   Prop.eqCnf' symCount (Prop.simplifyProp uprop)
	       end



	   (* $B0z?t%U%#%k%?%j%s%0$r$7$?8e$K(B Stability Condition $B$N@)Ls$,@.N)$9$k$3$H$r<($9L?Bj$r@8@.(B *)
	   (* ts: $BMxMQ2DG=$J=q$-49$(5,B'$NN>JU(B + dp $B$NN>JU(B *)
	   (* To Do: usable $B5,B'$HM;9g$G$-$k!)!)!)(B *)
	   fun mkStabilityCondProp dps usableRS = 
	       let
		   val ts = L.foldl (fn ((l,r),ss) => l::r::ss) [] (dps @ usableRS)

		   val supersortSetOfNonBasicRules = 
		       L.foldl
			   (fn ((l,_),set) => (SS.addList (set, Sort.supersorts (sortOfTerm l))))
			   SS.empty
			   usableRS

 	          (* $B4X?tJQ?t$N%"%W%j%1!<%7%g%s$+$i@8$8$k@)Ls(B *)
		   fun usables (Var _)  = Prop.True
		     | usables (t as (Fun (f,ts,ty))) =
		       let
			   val ps = L.tabulate
					(length ts,
				      fn i => if #useAf opt2
					      then Prop.Imp (STPoSolver.encodePiStatus encoding (f,i+1),
							     usables (L.nth (ts,i)))
					      else  usables (L.nth (ts,i)))
				    
			   val h = STTerm.headOfTerm t
			   val hy = Term.sortOfTerm h
		       in
			   Prop.Conj 
			       (if Term.isVar h 
				then
				    (L.mapPartial
					 (fn i=> if Sort.isASubsort (L.nth (stypes,i)) hy
						 then SOME (Prop.Atom (count + lendplist*2 + 	
								       lenrules + i))
						 else NONE)
					 (L.tabulate (L.length stypes, fn i=>i)))
				    @ ps
				else ps)
		       end


 	          (* $B4X?t7?$N=q$-49$(5,B'$+$i@8$8$k@)Ls(B *)
		   val frometa = L.mapPartial
				     (fn i=> if SS.member (supersortSetOfNonBasicRules,
							   L.nth (stypes,i))
					     then SOME (Prop.Atom (count + lendplist*2 + 	
								   lenrules + i))
					     else NONE)
				     (L.tabulate (L.length stypes, fn i=>i))
				 
	       in
			   Prop.eqCnf' symCount 
			   (Prop.simplifyProp 
				(Prop.Conj ((L.map usables ts) @ frometa)))
	       end

	   (********************************)
	   (*** Reduction Pair Processor ***)
	   (********************************)

	   fun reductionPairProcessor idxes =
	       if null idxes
	       then SOME []
	       else
		   let
		       val _ = debug (fn _ => print "\nReduction Pair Processor...\n")

		       (* $BBP>]$H$9$k0MB8BP(B *)
		       val dps = L.map (fn i=> L.nth (hdplist,i)) idxes
		       val adps = L.map (fn i=> L.nth (adplist,i)) idxes
		       val _ = debug (fn _ => print "Dependency Pairs:\n")
		       val _ = debug (fn _ => print (STSrs.prRules dps))

		       (* AF$B$r9MN8$7$J$$4J0WHG$N7W;;$GMxMQ2DG=5,B'$r@)8B$7$F$*$/(B *)
		       val rsIdxs = if #useUr opt
				    then STDp.getUsableRules usableHM arules dps 
				    else
					L.tabulate (lenrules,fn x => x)
		       val usableRS = L.map (fn i => L.nth (arules,i)) rsIdxs
		       val _ = debug (fn _ => print "Rules:\n")
		       val _ = debug (fn _ => print (STSrs.prRules usableRS))

		       (* $B@)LsO@M}<0(B *)
		       val gtPropsDP = L.map (fn i => lookupProp (count + i)) idxes
		       val eqPropsDP = L.map (fn i => lookupProp (count + lendplist + i)) idxes
		       val gePropsRS = L.map (fn i => lookupProp (count + lendplist*2 + i)) rsIdxs
		       val condProp = lookupProp (count + lendplist*2 + lenrules)

                       (*  AF $B$r9MN8$7$?(B Usable Rule $B$NL?Bj(B *)
		       val usableRulesProp = 
				   if (#useUr opt)
				   (* then mkUsableRulesProp (L.map (fn (l,r)=>r) adps) rsIdxs *)
				   then mkUsableRulesProp3 (L.map (fn (l,r)=>r) adps) rsIdxs 
				   else 
				       Prop.Conj 
					   (L.map (fn i => Prop.Atom (count + lendplist*2 + i)) rsIdxs)


			   (*  AF $B$r9MN8$7$?(B Stability Condition $B$NL?Bj(B *)
		       val stCondProp = if (#useUr opt)
					then mkStabilityCondProp adps usableRS
					else 
					    Prop.Conj (L.tabulate (L.length stypes,
								fn i=> Prop.Atom (count + lendplist*2 
										  + lenrules + i)))
					    
			   (*  $B:G=*E*$K2r$/$Y$-@)Ls!'MWAG$O$9$G$K(B CNF $B$K$7$F$"$k(B *)
		       val props =
			   Prop.appendCnf
			       (mapAppend (fn x => x)
					  [ gtPropsDP,
					    eqPropsDP,
					    gePropsRS,
					    [condProp],
					    [usableRulesProp],
					    [stCondProp],
 					    [Prop.Disj
						 (L.map (fn i => Prop.Atom (count + i)) idxes)],
 					    L.map (fn i => Prop.Disj
							       [Prop.Atom (count + i),
								Prop.Atom (count + lendplist + i)])
						  idxes ])

		       val (res,resAr) = Solver.cnfSolver (!minisatPath) (!tmpDir)
							  (Prop.Conj props,length props,!symCount)

		       val _ = debug (fn _ => if res
					      then STPoSolver.printInfo resAr encoding opt2
					      else ())

(* 		       val _ = debug (fn _=> if res then *)
(* 						 STPoSolver.printPiMap resAr encoding *)
(* 					     else ()) *)
(* 		       val _ = debug (fn _=> if res then *)
(* 						 STPoSolver.printColMap resAr encoding *)
(* 					     else ()) *)
(* 		       val _ = debug (fn _=> if res then *)
(* 						 STPoSolver.printWeight resAr encoding *)
(* 					     else ()) *)



		       (* $B%G%P%C%0MQ(B *)
		       fun printVal resAr i  = if i < Array.length resAr then
						   print ("value: " ^ (Int.toString i) ^ "=>"
							  ^ (if STPoSolver.isAssignedByTrue resAr i
							     then "T\n" else "F\n"))
					       else ()

                       (* stability $B$N%l%]!<%H(B *)
		       val _ = debug (fn _=> if res
			       then
				   (print "Stability:\n";
				   L.app
				   (fn i => if STPoSolver.isAssignedByTrue 
						   resAr (count + lendplist*2 + lenrules + i)
					    then print ("  " ^ (Sort.toString (L.nth (stypes,i)))
							^ ": stable\n")
					    else print ("  " ^ (Sort.toString (L.nth (stypes,i)))
							^ ": not stable\n"))
				   (L.tabulate (lenstcondlist,fn i=>i)))
			       else ())


(*		       val _ = L.tabulate (1, fn i => printVal resAr (888 + i)) *)
(* 		       val _ = let val p = STPoSolver.lookupGtPrecTable encoding (Fun.fromString "s", *)
(* 								       Fun.fromString "double") *)
(* 			       in if STPoSolver.evalProp resAre ncoding p  *)
(* 				  then print "s > double holds\n" *)
(* 				  else print "s > double does not holds\n" *)
(* 			       end *)


                       (*  strict order $B$,$D$$$F$$$J$$(B dp $B$rH4$-$@$9(B *)
		       val idxes2 = if res
				     then
					 L.filter (fn i => not (STPoSolver.isAssignedByTrue 
								    resAr (count + i)))
						  idxes
				     else []

                       (*  ge $B$H$J$C$F$$$J$$(B rules $B$rH4$-$@$9(B *)
		       val rsIdxs2 = if res
				     then
					 L.filter (fn i => not (STPoSolver.isAssignedByTrue
								    resAr (count + 2*lendplist + i)))
						  rsIdxs
				     else []


		       val _ = debug (fn _ =>
					 if res
					 then
					     let val _ = print "Obtained:\n"
						 val colSet = STPoSolver.getColSet resAr encoding
										   applicativeHeadAndDepthSet
						 val piMap = STPoSolver.getPiMap resAr encoding
										 applicativeHeadAndDepthSet
						 fun af t  = STDp.applyAfToTerm piMap colSet t
					     in
						 (L.app (fn i =>
							    let val (l,r) = L.nth (adplist,i)
							    in
								if member i idxes2
								then
								    print (" "
									   ^ (STTerm.toString l)
									   ^ " :(eq): "
									   ^ (STTerm.toString r) 
									   ^ (if #useAf opt2 
											  then "\n [after AF] "
												  ^ (STTerm.toString (af l))
												  ^ " :(eq): "
												  ^ (STTerm.toString (af r))
										  else "")
									   ^ "\n")
								else
								    print (" "
									   ^ (STTerm.toString l) 
									   ^ " :(gt): "
									   ^ (STTerm.toString r) 
									   ^ (if #useAf opt2 
											  then "\n [after AF] "
											  ^ (STTerm.toString (af l)) 
											  ^ " :(gt): "
											  ^ (STTerm.toString (af r))
										  else "")
										   ^ "\n")
							    end)
							idxes;
						  L.app (fn i =>
							    let val (l,r) = L.nth (arules,i)
							    in
								if member i rsIdxs2
								then
								    print (" "
									   ^ (STTerm.toString l)
									   ^ " :(not ge): "
									   ^ (STTerm.toString r) ^ "\n")
								else
								    print (" "
									   ^ (STTerm.toString l)
									   ^ " :(ge): "
									   ^ (STTerm.toString r) 
									   ^ (if #useAf opt2 
											  then "\n [after AF] "
												  ^ (STTerm.toString (af l)) 
												  ^ " :(ge): "
												  ^ (STTerm.toString (af r))
										  else "")
									   ^ "\n")
							    end)
							rsIdxs)
					     end
					 else ())
			       
		   in
		       if res
		       then SOME idxes2
		       else NONE
		   end


          (* $B0MB8BPDd;_@-%A%'%+!<$N%a%$%s%k!<%A%s(B *)
	   fun step [] = raise Success
	     | step (p::ps) =
	       case reductionPairProcessor p of
		   SOME ys => if null ys
			      then step ps
			      else step ((if #useSc opt
					 then
					     mapAppend (subtermCriteriaProcessor 
							    path dir
							    (false,hdplist,stsrs,strat))
						       (dependencyGraphProcessor
							    (hdplist,stsrs,strat) ys)
					 else
					     dependencyGraphProcessor
						 (hdplist,stsrs,strat) ys)
					 @ ps)
		 | NONE => raise Failure

       in
	   step idxesList
       end
       handle Success => true
	    | Failure => false

   end (* of local *)

end (* of STSolver *)


