(******************************************************************************
 * file: rwtools/stsrs/st_order.sml
 * description: Encoding of Recursive Path Ordering for STSRS
 * author: AOTO Takahito
 * 
 * Copyright (c) 2008, Toyama&Aoto Laboratory, Tohoku University
 ******************************************************************************)

signature ST_ORDER = 
sig
    val runDebug: bool ref

(* SAT $B$X$N%(%s%3!<%I(B *)
    (* lex on strict-precedence *)
    val lpoEncoding: (Term.fun_key * Term.fun_key -> Prop.prop)
		     -> Order.encoding_type
		     -> Term.term * Term.term
		     -> Prop.prop

    (* lex on quasi-precedence *)
    val qlpoEncoding: (Term.fun_key * Term.fun_key -> Prop.prop)
		      -> (Term.fun_key * Term.fun_key -> Prop.prop)
		      -> Order.encoding_type
		      -> Term.term * Term.term
		      -> Prop.prop

    (* lex on quasi-precedence with argument filtering *)
    val afQlpoEncoding: (Term.fun_key * Term.fun_key -> Prop.prop)
		       -> (Term.fun_key * Term.fun_key -> Prop.prop)
		       -> (Term.fun_key -> Prop.prop)
		       -> (Term.fun_key * int -> Prop.prop)
			-> int ref
		       -> Order.encoding_type
		       -> Term.term * Term.term
		       -> Prop.prop


    exception STOrderError of string
end;

structure STOrder : ST_ORDER =
  struct

  local
      
  open Term
  open Trs
  open Prop

  structure VS = VarSet
  structure VM = VarMap
  structure A = Atom
  structure FS = FunSet
  structure AT = AtomTable
  structure FPS = FunPairSet
  structure FPT = FunPairTable
  structure L = List
  structure LP = ListPair
  structure FIS = FunIntSet
  structure FIT = FunIntTable
  structure IPM = IntPairMap
  structure FIIT = FunIntIntTable
  structure TP = TermPair
  structure TPM = TermPairMap
  structure ILM = IntListMap2
  structure ILPM = IntListPairMap
(*  structure CP = Compiler.Profile *)

  exception STOrderError of string

  fun member x ys = isSome (L.find (fn y => x = y) ys) 

  fun containVar [] _ = false
    | containVar ((Var (x,_))::ts) y =
      (Var.equal (x,y)) orelse (containVar ts y)
    | containVar ((Fun (_,ss,_))::ts) y =
      (containVar ss y) orelse (containVar ts y)

  in 

  val runDebug = ref false

  fun debug f = if !runDebug then f () else ()

  exception STOrderError of string;

(************************************************************************************)
(* lpoEncoding:  forall l -> r \in R. l >lpo r $B$H(B $BF1Ey$J(B satisfiablity $B$NO@M}<0$KJQ49(B  *)
(************************************************************************************)

  fun lpoEncoding precEncoding encodingType (l,r)  = 
      let
	  (*** >lpo $B$H$J$k$?$a$N@)Ls$rJV$9(B ***)
	  fun lpoPropGt (Var _) _ = False
	    | lpoPropGt (Fun (_,ss,_)) (Var (y,_)) =
	      if (containVar ss y) then True else False
	    | lpoPropGt (s as Fun (f,ss,_)) (t as Fun (g,ts,_)) =
	      if equal (s,t)
		 orelse not (VS.isSubset (varSetInTerm t, varSetInTerm s))
	      then False
	      else case (STTerm.isAppSym f, STTerm.isAppSym g) of
		       (true,true) => lexPropGt s t ss ts
		     | (true,false) => alphaProp ss t
		     | (false,true) => majoProp s ts
		     | (false,false) => precEncoding (f,g)

	  (*** \exists si \in ss.  si >=lpo t $B$H$J$k$?$a$N@)Ls$rJV$9(B ***)
	  and alphaProp ss t =
 	      if List.exists (fn si => Term.equal (si,t)) ss 
 	      then True 
 	      else Disj (L.map (fn si => lpoPropGt si t) ss)
	      
	  (*** \forall t \in ts. s >lpo t $B$H$J$k$?$a$N@)Ls$rJV$9(B ***)
	  and majoProp s ts =
	      Conj (L.map (fn ti => lpoPropGt s ti) ts)

	  (*** ss >lpo^lex tt and \forall ti \in tt. s >lpo ti  ***)
	  (***      or  \exists si. si >= t  $B$H$J$k$?$a$N@)Ls(B ***)
	  and lexPropGt _ _ [] [] = False
	    | lexPropGt _ _ (si::ss) [] = True
	    | lexPropGt _ _ [] (ti::ts) = False
	    | lexPropGt s t (si::ss) (ti::ts) =
	      if Term.equal (si,ti)
	      then lexPropGt s t ss ts
	      else let val p = lpoPropGt si ti
		   in
		       if p = False
		       then alphaProp ss t
		       else Disj [ Conj [p, majoProp s ts],
				   alphaProp (si::ss) t  ]
		   end
      in
	  case encodingType of
	      Order.GT => lpoPropGt l r
	    | Order.GE => if Term.equal (l,r) 
		    then True
		    else lpoPropGt l r
	    | Order.EQ => if Term.equal (l,r) 
		    then True
		    else False
      end



(************************************************************************************)
(* qlpoEncoding:  forall l -> r \in R. l >lpo r $B$H(B $BF1Ey$J(B satisfiablity $B$NO@M}<0$KJQ49(B  *)
(* base order $B$O(B quasi-order *)
(************************************************************************************)

  fun qlpoEncoding precEncoding precEqEncoding encodingType (l,r)  = 
      let
	  (*** >lpo $B$H$J$k$?$a$N@)Ls$rJV$9(B ***)
	  fun qlpoPropGt (Var _) _ = False
	    | qlpoPropGt (Fun (_,ss,_)) (Var (y,_)) =
	      if (containVar ss y) then True else False
	    | qlpoPropGt (s as Fun (f,ss,_)) (t as Fun (g,ts,_)) =
	      if equal (s,t)
		 orelse not (VS.isSubset (varSetInTerm t, varSetInTerm s))
	      then False
	      else case (STTerm.isAppSym f, STTerm.isAppSym g) of
		       (true,true) => lexPropGt s t ss ts
		     | (true,false) => alphaProp ss t
		     | (false,true) => majoProp s ts
		     | (false,false) => precEncoding (f,g)

	  (*** s =qlpo t $B$K$J$k@)Ls$rJV$9(B ***)
	  and qlpoPropEq (Var (x,_)) (Var (y,_)) =
	      if Var.equal (x,y) then True else False
	    | qlpoPropEq (s as (Var _)) (t as (Fun _)) = False
	    | qlpoPropEq (Fun (_,ts,_)) (Var (y,_)) = False
	    | qlpoPropEq (s as Fun (f,ss,_)) (t as Fun (g,ts,_)) =
	      if not (length ss = length ts)
		 orelse not (VS.equal (varSetInTerm s,varSetInTerm t))
	      then False
	      else if Term.equal (s,t)
	      then True
	      else case (STTerm.isAppSym f, STTerm.isAppSym g) of
		       (true,true) => Conj (LP.map (fn (x,y) => qlpoPropEq  x y) (ss,ts))
		     | (true,false) => False
		     | (false,true) => False
		     | (false,false) => precEqEncoding (f,g)

	  (*** \exists si \in ss.  si >=qlpo t $B$H$J$k$?$a$N@)Ls$rJV$9(B ***)
	  and alphaProp [] t = False
	    | alphaProp (si::ss) t = 
	      let val p = qlpoPropEq si t
	      in if p = True
		 then True
		 else let val q = qlpoPropGt si t
		      in
			  if q = True
			  then True
			  else Disj [p,q,alphaProp ss t]
		      end
	      end
	      
	  (*** \forall t \in ts. s >qlpo t $B$H$J$k$?$a$N@)Ls$rJV$9(B ***)
	  and majoProp s ts =
	      Conj (L.map (fn ti => qlpoPropGt s ti) ts)

	  (*** ss >qlpo^lex tt and \forall ti \in tt. s >qlpo ti  ***)
	  (***      or  \exists si. si >= t  $B$H$J$k$?$a$N@)Ls(B ***)
	  and lexPropGt _ _ [] [] = False
	    | lexPropGt _ t  (si::ss) [] = alphaProp (si::ss) t
	    | lexPropGt _ _ [] (ti::ts) = False
	    | lexPropGt s t (si::ss) (ti::ts) =
	      let val p = qlpoPropEq si ti
	      in if p = True
		 then lexPropGt s t ss ts
		 else if p = False
		 then let val q = qlpoPropGt si ti
		      in
			  if q = False
			  then alphaProp ss t
			  else Disj [Conj [q, majoProp s ts], 
				     alphaProp (si::ss) t ]
		      end
		 else let val q = qlpoPropGt si ti
		      in
			  if q = False
			  then Disj [Conj [p, lexPropGt s t ss ts],
				     alphaProp ss t]
			  else Disj [Conj [p, lexPropGt s t ss ts],
				     Conj [q, (majoProp s ts)],
				     alphaProp (si::ss) t ]
		      end
	      end

      in
	  case encodingType of
	      Order.GT => qlpoPropGt l r
	    | Order.GE => Disj [qlpoPropGt l r, qlpoPropEq l r]
	    | Order.EQ => qlpoPropEq l r
      end


(************************************************************************************)
(* afQlpoEncoding:                                                                   *)
(*         forall l -> r \in R. pi(l) >lpo pi(r) $B$H(B $BF1Ey$J(B satisfiablity $B$NO@M}<0$KJQ49(B  *)
(* permutation $B$J$7(B                                                                  *)
(************************************************************************************)

  fun afQlpoEncoding precEncoding precEqEncoding isCollapse pi 
		    symCount encodingType (l,r) =
      let 
	  val gtMap= ref TermPairMap.empty
	  val eqMap= ref TermPairMap.empty
	  val iffProps = ref []

	  fun lookupGt s t = 
	      case TPM.find (!gtMap, (s,t)) of 
		  SOME p => (debug (fn _ => print ( "encoding cach used (" 
						    ^ (Term.toString s) ^ " :gt: "
						    ^ (Term.toString t) ^ ")\n" ));
			     p)
		| NONE => let val _ =  debug (fn _ => print ( "try to encode (" 
							      ^ (Term.toString s) ^ " :gt: "
							      ^ (Term.toString t) ^ ")\n" ))
			      val p = lpoPropGt s t
			  in 
			      if p = False orelse p = True
			      then
				  (gtMap := TPM.insert (!gtMap, (s,t), p); p)
			      else
				  let
				      val _ = symCount  := (!symCount) + 1;
				      val q = Prop.Atom (!symCount)
				      val _ = iffProps := (Prop.Iff (q,p))::(!iffProps)
				      val _ =  debug (fn _ => print ( "end to encode (" 
								      ^ (Term.toString s) ^ " :gt: "
								      ^ (Term.toString t) ^ ")\n" ))
				  in (gtMap := TPM.insert (!gtMap, (s,t), q); q)
				  end
			  end

	  and lookupEq s t = 
	      case TPM.find (!eqMap, (s,t)) of 
		  SOME p => (debug (fn _=> print ( "encoding cach used (" 
						   ^ (Term.toString s) ^ " :eq: "
						   ^ (Term.toString t) ^ ")\n" ));
			     p)
		| NONE => let val _ = debug (fn _=> print ( "try to encode (" 
							    ^ (Term.toString s) ^ " :eq: "
							    ^ (Term.toString t) ^ ")\n" ))
			      val p = lpoPropEq s t
			  in
			      if p = False orelse p = True
			      then
				  (eqMap := TPM.insert (!eqMap, (s,t), p); p)
			      else
				  let 
				      val _ = symCount  := (!symCount) + 1;
				      val q = Prop.Atom (!symCount)
				      val _ = iffProps := (Prop.Iff (q,p))::(!iffProps)
				      val _ =  debug (fn _ => print ( "end to encode (" 
								      ^ (Term.toString s) ^ " :eq: "
								      ^ (Term.toString t) ^ ")\n" ))
				  in (eqMap := TPM.insert (!eqMap, (s,t), q); q)
				  end
			  end

	  (*** >lpo $B$H$J$k$?$a$N@)Ls$rJV$9(B ***)
	  and lpoPropGt (Var _) _ = False
	    | lpoPropGt (s as Fun (f,ss,_)) (t as Var (y,_)) = 
	      if (containVar ss y)  (* Thus ss <> [] and f is application symbol *)
	      then
		  let val ps1 = L.tabulate (length ss,fn i => lookupGt (L.nth (ss,i)) t)
		      val ps2 = L.tabulate (length ss,fn i => lookupEq (L.nth (ss,i)) t)
		      val ps = LP.map (fn (p,q) => Disj [p,q]) (ps1,ps2)
		      val ns = L.tabulate (length ss,fn x=>x)
		  in
		    IfThenElse (isCollapse f,
				Disj (LP.map (fn (i,p) => Conj [pi (f,i+1),p]) (ns,ps1)),
				Disj (LP.map (fn (i,p) => Conj [pi (f,i+1),p]) (ns,ps)))
		  end
	      else False
	    | lpoPropGt (s as Fun (f,ss,_)) (t as Fun (g,ts,_)) =
	      if equal (s,t)
	      then False
	      else case (STTerm.isAppSym f, STTerm.isAppSym g) of
		       (true,false) => 
		       IfThenElse (isCollapse f,
				   Disj (L.tabulate (length ss,
						  fn i => Conj [pi (f,i+1), 
								lookupGt (L.nth (ss,i)) 
									  t])),
				   alphaProp f ss t)
		     | (false,true) => majoProp g 0 s ts
		     | (false,false) => precEncoding (f,g)
		     | (true,true) => 
		       if Fun.equal (f,g)
		       then IfThenElse (isCollapse f,
					Disj (L.tabulate (length ss,
						       fn i => Conj [pi (f,i+1), 
								     lookupGt (L.nth (ss,i)) 
									       (L.nth (ts,i)) ])),
					Disj [alphaProp f ss t,
					      lexPropGt f g s t ss ts])
			   else IfThenElse 
				(isCollapse f,
				 Disj (L.tabulate (length ss,
						fn i => Conj [pi (f,i+1), 
							      lookupGt (L.nth (ss,i)) t])),
				 Disj [alphaProp f ss t,
				       IfThenElse
					   (isCollapse g,
					    Disj (L.tabulate (length ts,
							   fn j => Conj [pi (g,j+1), 
									 lookupGt s (L.nth (ts,j))])),
					    
					    lexPropGt f g s t ss ts)])

	  (*** s = t $B$K$J$k@)Ls$rJV$9(B ***)
	  and lpoPropEq (Var (x,_)) (Var (y,_)) =
	      if Var.equal (x,y) then True else False
	    | lpoPropEq (s as (Var (x,_))) (t as (Fun (g,ts,_))) =
	      if containVar ts x (* hence ts <> [] *)
	      then
		  Conj [isCollapse g,
			Disj (L.tabulate (length ts,
				       fn j => Conj [pi (g,j+1),
						     lookupEq s (L.nth (ts,j))]))]
      	      else False
	    | lpoPropEq (s as Fun (f,ss,_)) (t as Var (y,_)) = lookupEq t s 
	    | lpoPropEq (s as Fun (f,ss,_)) (t as Fun (g,ts,_)) =
	      if Term.equal (s,t)
	      then True
	      else case (STTerm.isAppSym f, STTerm.isAppSym g) of
		       (true,false) => 
		       Conj (isCollapse f:: 
			     (L.tabulate (length ss,
				       fn j => Imp (pi (f,j+1),
						    lookupEq (L.nth (ss,j)) t))))
		     | (false,true) => 
		       Conj (isCollapse g:: 
			     (L.tabulate (length ts,
				       fn j => Imp (pi (g,j+1),
						    lookupEq s (L.nth (ts,j))))))
		     | (false,false) => precEqEncoding (f,g)
		     | (true,true) => 
		       if Fun.equal (f,g) 
		       then
 			   Conj (L.tabulate (length ss,
					  fn j => Imp (pi (f,j+1),
 						       lookupEq (L.nth (ss,j)) (L.nth (ts,j)))))
		       else
 			   IfThenElse
			       (isCollapse f,
				Conj (L.tabulate (length ss,
					 fn j => Imp (pi (f,j+1),
						      lookupEq (L.nth (ss,j)) t))),
 				IfThenElse
				    (isCollapse g,
				     Conj (L.tabulate (length ts,
					      fn j => Imp (pi (g,j+1),
							   lookupEq s (L.nth (ts,j))))),
 				     lexPropEq f g ss ts))
		       

	  (*** s >= t $B$K$J$k@)Ls$rJV$9(B ***)
 	  and lpoPropGe s t = Disj [lookupEq s t,lookupGt s t]

	  (*** \exists si \in ss.  si >=lpo t $B$H$J$k$?$a$N@)Ls$rJV$9(B ***)
	  and alphaProp f ss t = (* f is assumed not collapsing *)
	      Disj (L.tabulate (length ss,
			     fn i => Conj [pi (f,i+1), lpoPropGe (L.nth (ss,i)) t]))

	  (*** \forall t \in ts. s >lpo t $B$H$J$k$?$a$N@)Ls$rJV$9(B ***)
	  and majoProp f j s ts =
	      Conj (L.tabulate (length ts, 
			     fn i => Imp (pi (f,j+i+1), lookupGt s (L.nth (ts,i)))))
			  

	  (*** ss >qlpo^lex tt and \forall ti \in tt. s >qlpo ti  ***)
	  (*** $B$H$J$k$?$a$N@)Ls(B ***)
          (*  f,g $B$O(B not collapsing $B$H2>Dj(B *)
	  and lexPropGt f g s t ss ts = 
	      let 
 		  val _ = debug (fn _=> 
				    print ("(" ^ (Fun.toString f) ^ "," ^ (Fun.toString g) ^ "): "
					   ^ "[" ^ (PrintUtil.prSeq Term.toString  ss) ^ "] >> [" 
					   ^ (PrintUtil.prSeq Term.toString  ts) ^ "]\n"))
		  val lexGtMap = ref IPM.empty
		  val lenss  = length ss
		  val lents  = length ts
		  fun lookupLexGt i j  = 
		      if i < lenss andalso j = lents
		      then True
		      else if i = lenss orelse j = lents
		      then False
		      else case IPM.find (!lexGtMap, (i,j)) of 
			       SOME p => p
			     | NONE => let 
 				   val _ = debug (fn _=> 
						     print ("lookupLexGt (" ^ (Int.toString i) ^ "," 
							    ^ (Int.toString j) ^ ")\n"))
				   val p = lexGtSub i j
 				   val _ = debug (fn _=> 
						     print ("-lookupLexGt (" ^ (Int.toString i) ^ "," 
							    ^ (Int.toString j) ^ ")\n"))
				   val _ = symCount  := (!symCount) + 1
				   val q = Prop.Atom (!symCount)
				   val _ = iffProps := (Prop.Iff (q,p))::(!iffProps)
			       in (lexGtMap := IPM.insert (!lexGtMap, (i,j), q); q)
				       end
		  and lexGtSub i j = 
		      let val si = L.nth (ss,i)
			  val tj = L.nth (ts,j)
			  val ts0 = L.drop (ts,j+1)  (* CHANGE 2008/12/27 *)
 			  val _ = debug (fn _=> 
					    print ("lexGtSub (" ^ (Int.toString i) ^ "," 
						   ^ (Int.toString j) ^ ")\n"))
		      in
			  Disj [ Conj [pi (f,i+1), pi (g,j+1),  
				       Disj [Conj [ lookupLexGt (i+1) (j+1),
						    lookupEq si tj ],
					     Conj [ majoProp g (j+1) s ts0,
						    lookupGt si tj ]]],
				 Conj  [pi (f,i+1), Neg (pi (g,j+1)), 
					lookupLexGt i (j+1) ],
				 Conj  [Neg (pi (f,i+1)), pi (g,j+1), 
					lookupLexGt (i+1) j ],
				 Conj  [Neg (pi (f,i+1)), Neg (pi (g,j+1)),
					lookupLexGt (i+1) (j+1)] ]
		      end
	      in
		  lookupLexGt 0 0
	      end

	  (*** ss ==^lex tt $B$H$J$k$?$a$N@)Ls(B ***)
	  and lexPropEq f g ss ts = 
	      let val _ = debug (fn _=> 
				    print ("(" ^ (Fun.toString f) ^ "," ^ (Fun.toString g) ^ "): "
					   ^ "[" ^ (PrintUtil.prSeq Term.toString  ss) ^ "] == [" 
					   ^ (PrintUtil.prSeq Term.toString  ts) ^ "]\n"))
		  val lexEqMap = ref IPM.empty
		  val lenss  = length ss
		  val lents  = length ts
		  fun lookupLexEq i j  = 
		      if i = lenss andalso j = lents
		      then True
		      else if i >= lenss orelse j >= lents
		      then False
		      else case IPM.find (!lexEqMap, (i,j)) of 
			       SOME p => p
			     | NONE => let val p = lexEqsub i j
					   val _ = symCount  := (!symCount) + 1;
					   val q = Prop.Atom (!symCount)
					   val _ = iffProps := (Prop.Iff (q,p))::(!iffProps)
				       in (lexEqMap := IPM.insert (!lexEqMap, (i,j), q); q)
				       end
		  and lexEqsub i j = 
		      let val si = L.nth (ss,i)
			  val tj = L.nth (ts,j)
		      in
			  Disj [ Conj [pi (f,i+1), pi (g,j+1),  
				       lookupLexEq (i+1) (j+1),
				       lookupEq si tj ],
				 Conj  [pi (f,i+1), Neg (pi (g,j+1)), 
					lookupLexEq i (j+1) ],
				 Conj  [Neg (pi (f,i+1)), pi (g,j+1), 
					lookupLexEq (i+1) j ],
				 Conj  [Neg (pi (f,i+1)), Neg (pi (g,j+1)),
					lookupLexEq (i+1) (j+1)] ]
		      end
	      in
		  lookupLexEq 0 0
	      end


      in
	  case encodingType of 
	      Order.GT => let val p = lookupGt l r in Conj (p::(!iffProps)) end
	    | Order.GE => let val p = lpoPropGe l r in Conj (p::(!iffProps)) end 
	    | Order.EQ => let val p = lookupEq l r in Conj (p::(!iffProps)) end

      end (* of let *)



  end (* of local *)
				     
  end (* of structre *)


