(******************************************************************************
 * file: rwtools/stsrs/st_posolver.sml
 * description: Termination Solver for STSRS by Path Orderings
 * author: AOTO Takahito
 * 
 * Copyright (c) 2008-2009, Toyama&Aoto Laboratory, Tohoku University
 ******************************************************************************)

signature ST_POSOLVER = 
sig
    val runProfile: bool ref
    val runDebug: bool ref

    type options = {
	   useQuasi:bool,     (* quasi precedence $B$r;H$&$+$I$&$+(B *)
	   useLex:bool,       (* lexicographic status $B$r;H$&$+$I$&$+(B *)
	   useMul:bool,       (* multiset status $B$r;H$&$+$I$&$+(B *)
	   useAf:bool         (* argument filtering $B$r;H$&$+$I$&$+(B *)
    }

    val poSolver: string 
		    -> string
		    -> options
		    -> (Atom.atom list * Term.decl list * (Term.term * Term.term) list)
		    -> bool

    type encoding_info = { Fs : Fun.ord_key list,
			   LenFs : int,
			   RowLen : int,
			   FaList : (Fun.ord_key * int) list,
			   FaSet : FunIntSet.set,
			   FaMap : int FunMap.map,
			   PrecMap : int FunIntMap.map,
                           MulMap : int FunMap.map,
                           LexMap : int FunIntIntMap.map,
                           ColMap : int FunMap.map,
                           PiMap : int FunIntMap.map,
			   GtPrecTable : Prop.prop FunPairTable.hash_table,
			   EqPrecTable : Prop.prop FunPairTable.hash_table,
			   LexTable : Prop.prop FunIntIntTable.hash_table,
			   SameNthTable : Prop.prop FunIntPairTable.hash_table
			 }

    val mkSymMap: int ref -> Fun.ord_key list -> int FunMap.map
    val mkPrecMap: int ref -> (Fun.ord_key list * int) -> int FunIntMap.map
    val mkMulMap: int ref -> Fun.ord_key list -> int FunMap.map
    val mkLexMap: int ref -> FunIntSet.set -> int FunIntIntMap.map
    val mkColMap: int ref -> Fun.ord_key list -> int FunMap.map
    val mkPiMap: int ref -> FunIntSet.set -> int FunIntMap.map

    val mkEncodingInfo: int ref 
			-> Fun.ord_key list * int * int * ('a * int) list * FunIntSet.set
			   * 'b
             -> {ColMap:int FunMap.map, EqPrecTable:'c FunPairTable.hash_table,
                 FaList:('a * int) list, FaMap:'b, FaSet:FunIntSet.set,
                 Fs:Fun.ord_key list, GtPrecTable:'d FunPairTable.hash_table,
                 LenFs:int, LexMap:int FunIntIntMap.map,
                 LexTable:'e FunIntIntTable.hash_table, MulMap:int FunMap.map,
                 PiMap:int FunIntMap.map, PrecMap:int FunIntMap.map,
                 RowLen:int, SameNthTable:'f FunIntPairTable.hash_table}


    val mkLexCond: options -> encoding_info -> Prop.prop
    val mkStrictPrecCond: options -> encoding_info -> Prop.prop
    val mkAfCond: options -> encoding_info -> Prop.prop
    val mkQuasiRpoCond: options -> encoding_info -> Prop.prop
    val encodeMulStatus: encoding_info -> Fun.ord_key -> Prop.prop
    val encodeLexStatus: encoding_info -> Fun.ord_key * int * int -> Prop.prop
    val encodeSameNthStatus: encoding_info -> (Fun.ord_key * int) * (Fun.ord_key * int) -> Prop.prop
    val encodeColStatus: encoding_info -> Fun.ord_key -> Prop.prop
    val encodePiStatus: encoding_info -> Fun.ord_key * int -> Prop.prop
    val mkEqAfCond: encoding_info -> int * (Fun.ord_key list) -> Prop.prop

    val encodeGtPrec: encoding_info -> Fun.ord_key * Fun.ord_key -> Prop.prop
    val encodeEqPrec: encoding_info -> Fun.ord_key * Fun.ord_key -> Prop.prop
    val lookupFaMap: (int FunMap.map) * Fun.ord_key -> int
    val lookupGtPrecTable: encoding_info -> Fun.ord_key * Fun.ord_key -> Prop.prop
    val lookupEqPrecTable: encoding_info -> Fun.ord_key * Fun.ord_key -> Prop.prop
    val lookupLexTable: encoding_info -> Fun.ord_key * int * int -> Prop.prop
    val lookupSameNthTable: encoding_info -> (Fun.ord_key * int) * (Fun.ord_key * int) -> Prop.prop

    val isAssignedByTrue: int array -> int -> bool
    val evalProp: int array -> encoding_info -> Prop.prop -> bool
    val printWeight: int array -> encoding_info -> unit
    val printPiMap : int array -> encoding_info -> unit
    val getPiMap: int array -> encoding_info -> TermIntSet.set -> (int list) TermIntMap.map
    val printColMap : int array -> encoding_info -> unit
    val getColSet: int array -> encoding_info -> TermIntSet.set -> TermIntSet.set

    val mkPrecList: int array -> encoding_info -> (Fun.ord_key list) list
    val prArgStatusAF: int array -> encoding_info -> options -> (Fun.ord_key * int) -> string
    val prArgStatus: int array -> encoding_info -> options -> (Fun.ord_key * int) -> string
    val printInfo: int array -> encoding_info -> options -> unit

    val encodeStOrderConstraint: Order.encoding_type 
			       -> int ref -> encoding_info
			       -> options 
			       -> Term.term * Term.term -> Prop.prop
			       
end

structure STPoSolver : ST_POSOLVER =
struct
local
    open Prop
    structure A = Atom
    structure L = List
    structure LP = ListPair
    structure FS = FunSet
    structure FPT = FunPairTable
    structure FIT = FunIntTable
    structure FIPT = FunIntPairTable
    structure FM = FunMap
    structure FIS = FunIntSet
    structure FIM = FunIntMap
    structure FII = FunIntInt
    structure FIIM = FunIntIntMap
    structure FIIT = FunIntIntTable
    structure TPM = TermPairMap
    structure TIS = TermIntSet
    structure TIM = TermIntMap
    fun mapAppend f xs = List.foldr (fn (x,ys) => List.@(f x, ys)) [] xs
    local
	fun logMain n k ans = if n <= k then ans
			      else logMain n (k * 2) (ans + 1)
    in
    fun log n = logMain n 2 1
    end
in


    exception PoSolverError
	  
    val runProfile = ref false : bool ref
    val runDebug = ref false : bool ref
    fun debug f = if !runDebug then f () else ()
    val minisatPath = ref ""
    val tmpDir = ref ""

    type options = {
	   useQuasi:bool,     (* quasi precedence $B$r;H$&$+$I$&$+(B *)
	   useLex:bool,       (* lexicographic status $B$r;H$&$+$I$&$+(B *)
	   useMul:bool,       (* multiset status $B$r;H$&$+$I$&$+(B *)
	   useAf:bool         (* argument filtering $B$r;H$&$+$I$&$+(B *)
    }

    fun lookupFaMap (faMap,f) = 
	case FM.find (faMap,f) of
	    SOME i => i
	  | NONE => (print ("looking up faMap fails: (" ^ (Fun.toString f)  ^ "\n");
		     raise PoSolverError)

    (****************)
    (** Symbol Map **)
    (****************)
    (* +-+------+ *)
    (* |f|!c    | *)
    (* |g|!c + 1| *)
    (* |h|!c + 2| *)
    (* |s|!c + 3| *)
    (* +-+------+ *)
    (*  lenFs = 4      *)

    fun mkSymMap counterRef fs =
	L.foldr 
	    (fn (f,en) => (counterRef := 1 + !counterRef;
			   FM.insert (en,f,!counterRef)))
	    FM.empty 
	    fs

    fun lookupSymMap (symMap,f) = 
	case FM.find (symMap,f) of
	    SOME i => i
	  | NONE => (print ("looking up symMap fails: "  ^ (Fun.toString f) ^ "\n");
		     raise PoSolverError)


    (************************)
    (*** ENCODING SECTION ***)
    (************************)

    type encoding_info = { Fs : Fun.ord_key list,
			   LenFs : int,
			   RowLen : int,
			   FaList : (Fun.ord_key * int) list,
			   FaSet : FIS.set,
			   FaMap : int FM.map,
			   PrecMap : int FIM.map,
                           MulMap : int FM.map,
                           LexMap : int FIIM.map,
                           ColMap : int FM.map,
                           PiMap : int FIM.map,
			   GtPrecTable : Prop.prop FPT.hash_table,
			   EqPrecTable : Prop.prop FPT.hash_table,
			   LexTable : Prop.prop FIIT.hash_table,
			   SameNthTable : Prop.prop FIPT.hash_table
			 }


    (******************)
    (** Codish Table **)
    (******************)
    (* --+---------------------    *)
    (*row    1      2      3       *)
    (* --+---------------------    *)
    (* f | init   init+1  init+2   *)
    (* g | init+3 init+4  init+5   *)
    (* h | init+6 init+7  init+8   *)
    (* s | init+9 init+10 init+11  *)
    (* --+---------------------    *)
    (*  lowLen = 3, lenFs = 4      *)

    fun mkPrecMap counterRef (fs,rowLen) = 
	let
	    val rows = List.tabulate (rowLen, fn n => n + 1)
	    val products = ListXProd.mapX (fn (f,i) => (f,i)) (fs,rows)
	in
	    L.foldr 
		(fn (fi,en) => (counterRef := 1 + !counterRef;
				FIM.insert (en,fi,!counterRef)))
		FIM.empty 
		products
	end

    fun lookupPrecMap (precMap,(f,i)) = 
	case FIM.find (precMap,(f,i)) of
	    SOME n => n
	  | NONE => (print ("looking up precMap fails: (" 
			    ^ (Fun.toString f)  ^ "," ^ (Int.toString i)  ^ ")\n");
		     raise PoSolverError)


    (* f > g $B$G$"$k$3$H$rI=$o$9O@M}<0(B *)
    fun encodeGtPrec (encoding:encoding_info) (f,g) =
	let val rowLen = #RowLen encoding
	    val precMap = #PrecMap encoding
	in
	    Disj (L.tabulate (rowLen,
		    fn i => Conj
				(Atom (lookupPrecMap (precMap, (f,i+1)))::
				 Atom (~(lookupPrecMap (precMap, (g,i+1))))::
				 (L.tabulate (i,
					   fn j => (Iff (Atom (lookupPrecMap (precMap, (f,j+1))),
							 Atom (lookupPrecMap (precMap, (g,j+1))))))))))
	end	    

    (* $B%a%b(B & Look Up *)
    fun mkGtPrecTable lenFs = FPT.mkTable (lenFs * lenFs, PoSolverError)

    fun lookupGtPrecTable encoding (f,g) =
	let val gtPrecTable = #GtPrecTable encoding 
	in
	    case (FPT.find gtPrecTable (f,g))
	     of SOME p => p
	      | NONE => let val p = encodeGtPrec encoding (f,g)
			in (FPT.insert gtPrecTable ((f,g),p); p)
			end
	end

    (* f $B$H(B g $B$,(B $BF1$8(B precedence $B$G$"$k$3$H$rI=$o$9O@M}<0(B *)
    fun encodeEqPrec (encoding:encoding_info) (f,g) =
	let val rowLen = #RowLen encoding
	    val precMap = #PrecMap encoding
	in
	    Conj (L.tabulate (rowLen,
			   fn i => (Iff (Atom (lookupPrecMap (precMap, (f,i+1))),
					 Atom (lookupPrecMap (precMap, (g,i+1)))))))
	end

    (* $B%a%b(B & Look Up *)
    fun mkEqPrecTable lenFs = FPT.mkTable (lenFs * lenFs, PoSolverError)

    fun lookupEqPrecTable encoding (f,g) =
	let val eqPrecTable = #EqPrecTable encoding
	in
	   case (FPT.find eqPrecTable (f,g))
	    of SOME p => p
	       | NONE => let val p = encodeEqPrec encoding (f,g)
			 in (FPT.insert eqPrecTable ((f,g),p); p)
			 end
	end

    (* f $B$H(B g $B$,(B $BF1$8(B precedence $B$G$J$$$3$H$rI=$o$9O@M}<0(B *)
    fun encodeNeqPrec (encoding:encoding_info) (f,g) = 
	Neg (lookupEqPrecTable encoding (f,g))

    (* strict precedence $B$K$9$k$?$a$N@)Ls(B *)
    fun mkStrictPrecCond (opt:options) (encoding:encoding_info) =
	if #useQuasi opt
	then Prop.True
	else let  
		val eqPrecTable = #EqPrecTable encoding
		fun prod [] = []
		  | prod (x::ys) = L.@ (L.map (fn y => (x,y)) ys, prod ys)
		val consts = L.mapPartial (fn (f,i) => if i = 0 then SOME f else NONE)
					  (#FaList encoding)
	    in
		Conj (List.map (encodeNeqPrec encoding) (prod consts))
	    end
      

    (******************)
    (** Status Table **)
    (******************)
    (* +-+------+ *)
    (* |f|init  | *)
    (* |g|init+1| *)
    (* |h|init+2| *)
    (* |s|init+3| *)
    (* +-+------+ *)

    fun mkMulMap counterRef fs = mkSymMap counterRef fs  

    fun lookupMulMap (mulMap,f) = 
	case FM.find (mulMap,f) of
	    SOME i => i
	  | NONE => (print ("looking up mulMap fails: " ^ (Fun.toString f)  ^ "\n");
		     raise PoSolverError)

    (* f $B$,(B Staus MUL $B$G$"$k$3$H$rI=$o$9O@M}<0(B *)
    fun encodeMulStatus (encoding:encoding_info) f = 
	Atom (lookupMulMap (#MulMap encoding, f))

    fun mkQuasiRpoCond (opt:options) (encoding:encoding_info) =
	if (#useLex opt) 
	   andalso  (#useMul opt) 
	   andalso  (#useQuasi opt) 
	then let fun mkCond [] = []
		   | mkCond (f::xs) = 
		     L.@ (L.map (fn g => 
				    Imp (lookupEqPrecTable encoding (f,g), 
					 Iff (encodeMulStatus encoding f, 
					      encodeMulStatus encoding g)))
				xs,
			  mkCond xs)
	     in
		 Conj (mkCond (#Fs encoding))
	     end
	else Prop.True


    (************************************)
    (** $B<-=q<0%9%F!<%?%9$N%(%s%3!<%G%#%s%0(B *****)
    (*****************  CODISH ENCODING ****************)
    (*  f      1 2    prec     prop vars               *)
    (*      +-+---+          --+------------           *)
    (*      |1|1 1| 2+1=3    f1| x31 x32 x33           *)
    (*      +-+---+            |                       *)
    (*      |2|1 0| 2+0=2    f2| x34 x35 x36           *)
    (*      +-+---+            |                       *)
    (*      |3|0 1| 0+1=1    f3| x37 x38 x39           *)
    (*      +-+---+            |                       *)
    (*      |4|0 0| 0+0=0    f4| x40 x41 x42           *)
    (*      +-+---+                                    *)
    (*  arity(f) = 4                                   *)
    (*  arity(f) = 1 $B$N$H$-$O$J$7(B                         *)
    (***************************************************)

    fun mkLexMap counterRef faSet = 
	FIS.foldr (fn ((f,ar),map) => 
		      if ar <= 1 then map
		      else
			  ListXProd.foldX
			      (fn (i,j,m) => (counterRef := 1 + (!counterRef);
					      FIIM.insert (m,(f,i,j),!counterRef)))
			      (L.tabulate (ar,fn i=>i+1),
			       L.tabulate (log ar,fn i=>i+1))
			      map)
		  FIIM.empty
		  faSet

    fun lookupLexMap (lexMap, (f,i,j)) = 
	if i > 0 andalso j > 0
	then
	    case FIIM.find (lexMap,(f,i,j)) of
		SOME i => i
	      | NONE => (print ("looking up lexMap fails: (" ^ (Fun.toString f)  ^ ","
				^ (Int.toString i)  ^ "," ^ (Int.toString j)  ^ ")\n");
			 raise PoSolverError)
	else  (print ("looking up lexMap fails: (" ^ (Fun.toString f)  ^ ","
		      ^ (Int.toString i)  ^ "," ^ (Int.toString j)  ^ ")\n");
	       raise PoSolverError)

    (* $BBh(Bi$B0z?t$OBh(Bj$B0z?t$h$j@h$KHf3S$9$k$3$H$rI=$o$9O@M}<0(B ($B$?$@$7!$(Bi <> j) *)
    fun encodeLexStatus (encoding:encoding_info) (f,i,j) =  
	let val faMap = #FaMap encoding
	    val lexMap = #LexMap encoding
	in
	    if i <> j 
	    then
		Disj (L.tabulate (log (lookupFaMap (faMap, f)),
			       fn k => Conj
					   (Atom (lookupLexMap (lexMap, (f,i+1,k+1)))::
					    Atom (~(lookupLexMap (lexMap, (f,j+1,k+1))))::
					    (L.tabulate (k,
						      fn l => 
							 (Iff (Atom (lookupLexMap (lexMap,(f,i+1,l+1))),
							       Atom (lookupLexMap (lexMap,(f,j+1,l+1))))))))))
	    else (print ("encodeLexStatus Error: (" ^ (Int.toString i) ^ ","
			 ^ (Int.toString j) ^ ")\n");
		  raise PoSolverError)
	end
      
    (* $B%a%b(B & Look Up *)
    fun mkLexTable faList = 
	let val len = L.foldr (fn ((_,ar),m) => ar * (log ar) + m) 0 faList
	in FIIT.mkTable (len, PoSolverError)
	end

    fun lookupLexTable encoding fij =
	let val lexTable = #LexTable encoding
	in
	   case (FIIT.find lexTable fij)
	    of SOME p => p
	     | NONE => let val p = encodeLexStatus encoding fij
		       in (FIIT.insert lexTable (fij,p); p)
		       end
	end

    fun encodeSameNthStatus (encoding:encoding_info) ((f,i),(g,j)) =  
	let 
	    val faMap = #FaMap encoding
	    val lexMap = #LexMap encoding
	    val arF = lookupFaMap (faMap, f)
	    val arG = lookupFaMap (faMap, g)
	    val logArF = log arF
	    val logArG = log arG
	    fun sameNthStatusSub ((f,i),(g,j)) =  
		Conj [Conj (L.tabulate (logArF,
				     fn k => Iff (Atom (lookupLexMap (lexMap, (f,i+1,k+1))), 
						  Atom (lookupLexMap (lexMap, (g,j+1,logArF-logArG+k+1)))))),
		      Conj (L.tabulate (logArF-logArG,
				     fn k => Iff (False, Atom (lookupLexMap (lexMap,(g,j+1,k+1))))))]
	in
	    if arF <= arG
	    then sameNthStatusSub ((f,i),(g,j))
	    else sameNthStatusSub ((g,j),(f,i))
	end

    (* $B%a%b(B & Look Up *)
    fun mkSameNthTable faList  = 
	let val len = L.foldr (fn ((_,ar),m) => ar + m) 0 faList
	in FIPT.mkTable (len*len, PoSolverError)
	end

    fun lookupSameNthTable encoding (fi,gj) =
	let val sameNthTable = #SameNthTable encoding
	in
	    case (FIPT.find sameNthTable (fi,gj))
	     of SOME p => p
	      | NONE => let val p = encodeSameNthStatus encoding (fi,gj)
			in (FIPT.insert sameNthTable ((fi,gj),p); p)
			end
	end

    fun mkLexCond (opt:options) (encoding:encoding_info) =
	if (#useLex opt) 
	then let val faList = #FaList encoding 
		 val sameNthTable = #SameNthTable encoding
	     in
		 Conj (L.map
			   (fn (f,ar) =>
			       let val xs = L.tabulate (ar,fn x=>x)
			       in
				   Conj (ListXProd.mapX
					     (fn (i,j) 
						 => if i < j
						    then Neg (lookupSameNthTable
								  encoding ((f,i),(f,j)))
						    else True)
					     (xs,xs))
			       end)
			   faList)
	     end
	else Prop.True

    (*********************************************)
    (** $B0z?t%U%#%k%?%j%s%0$N%9%F!<%?%9$N%(%s%3!<%G%#%s%0(B ****)
    (*********************************************)
    (* encodeColStatus f                              *)
    (*  \pi(f) = i => true                       *)
    (*  \pi(f) = [...] => false                  *)
    (*********************************************)

    fun mkColMap counterRef fs = 
	mkSymMap counterRef fs

    fun lookupColMap (colMap,f) = 
	case FM.find (colMap,f) of
	    SOME i => i
	  | NONE => (print ("looking up colMap fails: (" ^ (Fun.toString f)  ^ ")\n");
		     raise PoSolverError)

    fun encodeColStatus (encoding:encoding_info) f = 
	let val colMap = #ColMap encoding
	in Atom (lookupColMap (colMap,f))
	end

    (**************************************************)
    (****** $B0z?t%U%#%k%?%j%s%0$N%9%F!<%?%9$N%(%s%3!<%G%#%s%0(B *****)
    (**************************************************)
    (**** CHANGE  CHANGE CHANGE                   ***)
    (* pi (f,i) = true iff  pi(f) = i            ***)
    (*                      or i  \in pi(f)       ***)
    (* E.g. (plus,0), (plus,1) --> (plus,1), (plus,2) *)
		    
    fun mkPiMap counterRef faSet =
	FIS.foldr (fn ((f,ar),pm) => 
		      List.foldr
			  (fn (i,m) => (counterRef := 1 + !counterRef;
					FIM.insert (m,(f,i),!counterRef)))
			  pm
 			  (L.tabulate (ar,fn i=>i + 1)))
		  FIM.empty
		  faSet

    fun lookupPiMap (piMap,(f,i)) = 
	case FIM.find (piMap,(f,i))
	 of SOME n => n
	  | NONE =>  (print ("looking up piMap fails: (" ^ (Atom.toString f) ^ "," 
			     ^ (Int.toString i) ^ ")\n");
		      raise PoSolverError)

    fun encodePiStatus (encoding:encoding_info) (f,i) = 
	let val piMap = #PiMap encoding
	in Atom (lookupPiMap (piMap,(f,i)))
	end

    (****** encodeColStatus f = true $B$J$i(B pi(f)$B$,(Btrue$B$H$J$k$N$O(B1$B$D$@$1(B*****)
    fun mkAfCond (opt:options) (encoding:encoding_info) =
	if (#useAf opt) 
	then let val faList = #FaList encoding
	     in Conj (L.map (fn (f,ar) =>
				Imp (encodeColStatus encoding f,
				     one (L.tabulate (ar,fn j => encodePiStatus encoding (f,j+1)))))
			    faList)
	     end
	else Prop.True

    (****** pi(f,i)=true iff pi(g,i) = true for (f,g) \in fs and i = 0,..,ar-1 *****)
    fun mkEqAfCond (encoding:encoding_info) (ar,[]) = True
      | mkEqAfCond (encoding:encoding_info) (ar,f::fs) =
	let val piMap = #PiMap encoding
	    val colMap = #ColMap encoding
	in Conj (mapAppend
		     (fn g =>
			 (Iff (Atom (lookupColMap (colMap,f)),
				  Atom (lookupColMap (colMap,g))))::
			 (L.tabulate (ar,
				       fn i=>
					  Iff (Atom (lookupPiMap (piMap,(f,i+1))),
					       Atom (lookupPiMap (piMap,(g,i+1)))))))
		     fs)
	end


    (***************************************)
    (*** $BA4It$N(B encoding map & table $B$N:n@.(B ***)
    (***************************************)

   (* couterRef: $B$9$G$K;H$C$F$$$k%+%&%s%?HV9f$N:GBgCM(B *)
    fun mkEncodingInfo counterRef (fs,lenFs,rowLen,faList,faSet,faMap) =
	let val precMap = mkPrecMap counterRef (fs,rowLen)
	    val mulMap = mkMulMap counterRef fs
	    val lexMap = mkLexMap counterRef faSet
	    val colMap = mkColMap counterRef fs
	    val piMap = mkPiMap counterRef faSet
	    val gtPrecTable = mkGtPrecTable lenFs
	    val eqPrecTable = mkEqPrecTable lenFs
	    val lexTable = mkLexTable faList
	    val sameNthTable = mkSameNthTable faList
	in
	    { Fs = fs,
	      LenFs = lenFs,
	      RowLen = rowLen,
	      FaList = faList,
	      FaSet = faSet,
	      FaMap = faMap,
	      PrecMap = precMap,
              MulMap = mulMap,
              LexMap = lexMap,
              ColMap = colMap,
              PiMap = piMap,
	      GtPrecTable = gtPrecTable,
	      EqPrecTable = eqPrecTable,
	      LexTable = lexTable,
	      SameNthTable = sameNthTable }
	end


    (*******************)
    (*** $B=PNO7k2L$N2r@O(B ***)
    (*******************)

    (* $BBh(Bi$BHVL\$NL?BjJQ?t$,??$+(B *)
    fun isAssignedByTrue resultAr i = 
	if i <= Array.length resultAr
	   andalso 0 < i
	then Array.sub (resultAr,i-1) >  0
	else (print ("isAssingedByTrue: index out of bouds: " 
		     ^ (Int.toString (i-1))
		     ^ "\n"); raise PoSolverError)

    fun evalProp resultAr (encoding:encoding_info) p = 
	case p of
	    Prop.Atom i => if i > 0 then isAssignedByTrue resultAr i 
		      else if i < 0 then not (isAssignedByTrue resultAr (~i))
		      else (print "zero is not used as var index.\n"; raise PoSolverError)
	  | Prop.Conj ps => L.all (evalProp resultAr encoding) ps
	  | Prop.Disj ps => L.exists (evalProp resultAr encoding) ps
	  | Prop.Neg p => not (evalProp resultAr encoding p)
	  | Prop.Imp (p,q)  => if evalProp resultAr encoding p
			       then evalProp resultAr encoding q
			       else true
	  | Prop.Iff (p,q) => (evalProp resultAr encoding p)
			      = (evalProp resultAr encoding q)
	  | Prop.IfThenElse (c,p,q) => 
	    if evalProp resultAr encoding c
	    then evalProp resultAr encoding p
	    else evalProp resultAr encoding q


    (*** $B4X?t5-9f$N=E$_(B ***)
    fun weightOfFun resultAr (encoding:encoding_info) f =
	let val rowLen = #RowLen encoding
	    val precMap = #PrecMap encoding
	    val ns = L.map (fn i => lookupPrecMap (precMap, (f,i+1)))
			   (L.tabulate (rowLen, fn x => x))
	in
	    List.foldl (fn (k,sum) 
			   => 2 * sum + (if isAssignedByTrue resultAr k
					 then 1 else 0)) 
		       0 ns
	end

    fun printWeight resultAr (encoding:encoding_info) = 
	let val fs = #Fs encoding
	    val _ = debug (fn _ => print "weight:\n")
	in
	    L.app (fn f =>
		      (print (" " ^ (Fun.toString f) ^ ": "
			      ^  (Int.toString (weightOfFun resultAr encoding f))
			      ^ " ")))
		  fs;
	    print "\n"
	end


    (*** $B4X?t5-9f$N=g0L%j%9%H$N:n@.(B ***)	  
    fun mkPrecList resultAr (encoding:encoding_info) = 
	let val fs = #Fs encoding
	    val mp = List.foldr 
			 (fn (f,map) 
			     => FM.insert (map,f,weightOfFun resultAr encoding f))
			 FM.empty
			 fs
	    val sorted = 
		ListMergeSort.sort  
		    (fn (f,g) => Int.< (valOf (FM.find (mp,f)), valOf (FM.find (mp,g))))
		    fs
	    (* [1,2,2,3,3,3] => [[1],[2,2],[3,3,3]] *)
	    fun divideToEqClass equal [] = []
	      | divideToEqClass equal (a::sorted) =
		let
		    fun divide a [] = [[a]]
		      | divide a (b::xs) = 
			let val pre = divide b xs
			in if null pre
			   then (print "divideToEqClass\n"; raise PoSolverError)
			   else if equal (a, b)
			   then  (a::hd pre)::(tl pre)
			   else  ([a]::pre)
			end
		in
		    divide a sorted
		end
	    fun equal (f,g) = (valOf (FM.find (mp,f))) = (valOf (FM.find (mp,g)))
	in
	    divideToEqClass equal sorted
	end


    (*** Lex status $B$N2r@O(B ***)
    fun greaterLex resultAr (encoding:encoding_info) f (i,j) =  (* i <> j, i,j = 1,.. *)
	let val faMap = #FaMap encoding
	    val lexMap = #LexMap encoding
	    val ar = lookupFaMap (faMap, f)
	    val lnAr = log ar
	    val w_i = List.foldl (fn (k,sum) 
				     => 2 * sum 
					+ (if isAssignedByTrue resultAr k 
					   then 0 else 1)) 
				 0 
				 (L.tabulate (lnAr, fn k => lookupLexMap (lexMap,(f,i,k+1))))
	    val w_j = List.foldl (fn (k,sum) 
				     => 2 * sum + (if isAssignedByTrue resultAr k 
						   then 0 else 1)) 
				 0 
				 (L.tabulate (lnAr, fn k => lookupLexMap (lexMap, (f,j,k+1))))
	in
	    w_i > w_j
	end


    fun printPiMap resultAr (encoding:encoding_info) = 
	let val piMap = #PiMap encoding
	    val faList = #FaList encoding
	    val _ = debug (fn _ => print "piMap:\n")
	in
	    L.map (fn (f,ar) =>
		      (print (" " ^ (Fun.toString f) ^ ": ");
		       L.app (fn i =>
				 if isAssignedByTrue resultAr (lookupPiMap (piMap,(f,i+1)))
				 then print ((Int.toString (i+1)) ^ " ")
				 else ())
			     (L.tabulate (ar, fn i=>i))))
		  faList;
	    print "\n"
	end

    fun getPiMap resultAr (encoding:encoding_info) hdSet = 
	let val piMap = #PiMap encoding
	    val faList = #FaList encoding
	in
	    TIS.foldl (fn ((h,d),map) =>
			  let val ty = valOf (STTerm.sortOfHeadAndDepth (h,d))
			      val f = STTerm.annotateApp (h,d,ty)
			      val ar = Sort.arityOfSort ty + 1
			  in 
			      TIM.insert (map,(h,d),
					  L.filter (fn i =>
						       isAssignedByTrue resultAr 
									(lookupPiMap (piMap,(f,i+1))))
						   (L.tabulate (ar, fn i=>i)))
			  end)
		    TIM.empty
		    hdSet
	end


    fun printColMap resultAr (encoding:encoding_info) = 
	let val colMap = #ColMap encoding
	    val fs = #Fs encoding
	    val _ = debug (fn _ => print "colMap:\n")
	in
	    L.map (fn f =>
		      (print (" " ^ (Fun.toString f) ^ ": ");
		       if isAssignedByTrue resultAr (lookupColMap (colMap,f))
		       then print "Y "
		       else print "N "))
		  fs;
	    print "\n"
	end

    fun getColSet resultAr (encoding:encoding_info) hdSet = 
	let val colMap = #ColMap encoding
	    val fs = #Fs encoding
	in
	    TIS.foldl (fn ((h,d),set) =>
			  let val ty = valOf (STTerm.sortOfHeadAndDepth (h,d))
			      val f = STTerm.annotateApp (h,d,ty)
			  in if isAssignedByTrue resultAr (lookupColMap (colMap,f))
			     then TIS.add (set,(h,d))
			     else set
			  end)
		      TIS.empty
		      hdSet
	end


    (*** $B4X?t5-9f$N0z?tHf3S=g=x%j%9%H$N:n@.(B[AF$B$J$7$N>l9g(B] ***)	  
    fun mkLexList resultAr (encoding:encoding_info) (f,ar) = 
	ListMergeSort.sort 
	    (fn (i,j) => greaterLex resultAr encoding f (i+1,j+1))
	    (L.tabulate (ar, fn i=>i))

    (*** $B4X?t5-9f$N0z?tHf3S=g=x%j%9%H$N:n@.(B[AF$B$"$j$N>l9g(B] ***)	  
    fun mkLexListAF resultAr (encoding:encoding_info) (f,ar) = 
	let
	    val piMap = #PiMap encoding
	    val ns = List.filter 
			 (fn i => isAssignedByTrue resultAr (lookupPiMap (piMap,(f,i+1))))
			 (L.tabulate (ar, fn i=>i))
	in
	    ListMergeSort.sort 
		(fn (i,j) => greaterLex resultAr encoding f (i+1,j+1))
		ns
	end

    (*** Lex/Mul$BIU$-$N0z?tHf3S=g=x%j%9%H=PNO(B[AF$B$J$7$N>l9g(B] ***)
    fun prArgStatus resultAr (encoding:encoding_info) (opt:options) (f,ar)  = 
	let val mulMap = #MulMap encoding
	    val ns = mkLexList resultAr encoding (f,ar)
	in
	    if (#useMul opt) andalso (#useLex opt)
	    then
		if isAssignedByTrue resultAr (lookupMulMap (mulMap, f))
  		then "Mul" 
		else "Lex [" ^ (PrintUtil.prSeq Int.toString ns) ^ "]"
	    else if (#useMul opt) 
	    then "Mul" 
	    else 
		"Lex [" ^  (PrintUtil.prSeq Int.toString ns) ^ "]"
	end

    (*** Lex/Mul$BIU$-$N0z?tHf3S=g=x%j%9%H=PNO(B[AF$B$"$j$N>l9g(B] ***)
    fun prArgStatusAF resultAr (encoding:encoding_info) (opt:options) (f,ar)  = 
	let val ns = mkLexListAF resultAr encoding (f,ar)
	    val colMap = #ColMap encoding
	    val mulMap = #MulMap encoding
	    val isCol = isAssignedByTrue resultAr (lookupColMap (colMap,f))
	    val ms = ListMergeSort.sort Int.> ns
	in
	    if isCol 
	    then if length ns <> 1 
		 then (print "forget AF condition?\n"; raise PoSolverError)
		 else "!" ^ (Int.toString (hd ns))
	    else (* non collapsing *)
		if (#useMul opt) andalso (#useLex opt)
		then
		    if isAssignedByTrue resultAr (lookupMulMap (mulMap,f))
  		    then "Mul [" ^ (PrintUtil.prSeq Int.toString ms) ^ "]"
		    else "Lex [" ^ (PrintUtil.prSeq Int.toString ns) ^ "]"
		else if (#useMul opt) 
		then "Mul [" ^ (PrintUtil.prSeq Int.toString ms) ^ "]"
		else if (#useLex opt) 
		then "Lex [" ^ (PrintUtil.prSeq Int.toString ns) ^ "]"
		else "Lex [" ^  (PrintUtil.prSeq Int.toString ms) ^ "]" 
	end

    (*** $B4X?t5-9f$NM%@h=g0L!$(BLex/Mul$BIU$-$N0z?tHf3S=g=x%j%9%H$N=PNO(B ***)
    fun printInfo resultAr (encoding:encoding_info) (opt:options) = 
	let
	    val faMap = #FaMap encoding
	    val prec = List.filter (not o null)
		       (List.map
			    (fn fs => (List.filter (not o STTerm.isAppSym) fs))
			    (mkPrecList resultAr encoding))

	    val appSyms = List.filter STTerm.isAppSym (#Fs encoding)

	    fun prWithArg f = " " ^ (Fun.toString f)

	    fun prWithArgAf f = let 
				  val ar = lookupFaMap (faMap, f)
			      in
				  " " ^ (STTerm.prAnnotatedAppSymbol f) ^ " : " 
				  ^ prArgStatusAF resultAr encoding
						  (opt:options) (f,ar)
			      end
	in
	    print "Precedence:\n";
	    print (PrintUtil.prSeqWith 
		       (fn xs => PrintUtil.prSeqWith 
				     prWithArg ", " xs)

		       ";\n" prec);
            print ";\n";
	    (if #useAf opt
	     then
		 (print "Argument filtering:\n";
		  print (PrintUtil.prSeqWith
			     prWithArgAf ";\n" appSyms);
		  print ";\n")
	     else
		 ())
	end




    fun encodeStOrderConstraint encodingType counterRef encoding 
			      (opt:options) rule =
	let 
	    val faSet = #FaSet encoding
	    val funSet = FIS.foldr (fn ((f,i),set)=> FS.add (set,f)) FS.empty faSet

	    fun prec (f,g) = lookupGtPrecTable encoding (f,g)
	    fun preceq (f,g) = lookupEqPrecTable encoding (f,g)
	    fun mul f = encodeMulStatus encoding f
	    fun lex (f,i,j) = lookupLexTable encoding (f,i,j)
	    fun same (fi,gj) = lookupSameNthTable encoding (fi,gj)
	    fun col f = encodeColStatus encoding f
	    fun pi (f,i) = encodePiStatus encoding (f,i)

	    val _ = debug (fn _ => let val (l,r) = rule
				   in print ("do encoding: " ^ (STTerm.toString l)
					     ^ (case encodingType of
						    Order.GT => " :(gt): "
						  | Order.GE => " :(ge): "
						  | Order.EQ => " :(eq): ")
					     ^ (STTerm.toString r) ^ "\n")
				   end)
	in case (#useAf opt,#useQuasi opt,#useLex opt,#useMul opt) of
	       (false,false,false,false)
	       => STOrder.lpoEncoding prec encodingType rule
	     | (false,true,false,false)
	       => STOrder.qlpoEncoding prec preceq encodingType rule
	     | (true,true,false,false)
	       => STOrder.afQlpoEncoding prec preceq col pi counterRef
				       encodingType rule
	     | _ => (print "Option not supported\n"; Prop.False)
	end


   fun poSolver path dir (opt:options) (sorts,decls,srules) = 
       let
	   val _ = minisatPath := path
	   val _ = tmpDir := dir
	   val arules =  STSrs.annotateRules srules 
	   val trs =  Trs.rulesToTrs arules 

	   val faMap  = Trs.funArityMapInRules arules
	   val faSet  = FM.foldri (fn (k,v,xs)=> FIS.add (xs,(k,v))) FIS.empty faMap
	   val funSet  = FM.foldri (fn (k,v,xs)=> FS.add (xs,k)) FS.empty faMap
	   val faList  = FM.listItemsi faMap
	   val fs  = L.map #1 faList
	   val lenFs  = L.length fs
	   val rowLen  = log lenFs

	   val symCount = ref 0
	   val encoding = mkEncodingInfo symCount (fs,lenFs,rowLen,faList,faSet,faMap)

	   val strictPrecCond = mkStrictPrecCond opt encoding 
	   val lexCond = mkLexCond opt encoding 
	   val quasiRpoCond = mkQuasiRpoCond opt encoding 

 	   val gtProps = L.map (Prop.simplifyProp 
				o (encodeStOrderConstraint Order.GT symCount encoding opt))
			       arules

   	   val condProp = Prop.simplifyProp (Prop.Conj [strictPrecCond, lexCond, quasiRpoCond])

	   val prop = Prop.Conj [ condProp,
				  Prop.Conj gtProps ]

	   val (result,resultArray) = Solver.propSolver (!minisatPath) (!tmpDir) (prop,!symCount)
	   val _ = debug (fn _ => if result
				  then printInfo resultArray encoding opt
				  else ())

       in
	   result
       end






  end (* of local *)

  end (* of structre *)

