(******************************************************************************
 * file: rwtools/stsrs/st_srs.sml
 * description: Simply Typed S-Expression Rewriting Systems
 * author: AOTO Takahito
 * 
 * Copyright (c) 2008-2009, Toyama&Aoto Laboratory, Tohoku University
 ******************************************************************************)

signature ST_SRS = 
   sig
   type stsrs = { FSymSet:FunSet.set, 
		  DSymSet:FunSet.set,
		  CSymSet:FunSet.set,
		  VarCond:bool,    (* Does satisfy l \notin V, V(r) \subseteq (l) ? *)
		  Rules: (Term.term * Term.term) list,
		  HdMap: int FunMap.map  (* depth of head occurrence of lhs *)
		}
   val prRule: Term.term * Term.term -> string
   val prEq: Term.term * Term.term -> string
   val prRules: (Term.term * Term.term) list -> string
   val prEqs: (Term.term * Term.term) list -> string

   val attachTypeToRules: Term.decl list 
			  -> (Term.term * Term.term) list 
			  -> (Term.term * Term.term) list option
   val rulesToStsrs: (Term.term * Term.term) list 
		     -> stsrs
   val etaExpandRule: Term.term * Term.term 
		      -> (Term.term * Term.term) list
   val etaExpandRules: (Term.term * Term.term) list 
		       -> (Term.term * Term.term) list
   val flattingRules: (Term.term * Term.term) list 
		      -> (Term.term * Term.term) list 
   val annotateRules: (Term.term * Term.term) list 
		      -> (Term.term * Term.term) list 
   val unannotateRules: (Term.term * Term.term) list 
			-> (Term.term * Term.term) list 

end;

structure STSrs : ST_SRS = 
   struct

   local 
       open Term
       open STTerm
       structure P = Prop
       structure S = Sort
       structure FS = FunSet
       structure VS = VarSet
       structure VM = VarMap
       structure TS = TermSet
       structure FM = FunMap
       structure TM = TermMap
       structure FIS = FunIntSet
       structure FIM = FunIntMap
       structure L = List
       structure LP = ListPair
       fun mapAppend f xs = List.foldr (fn (x,ys) => List.@(f x, ys)) [] xs
   in

   exception STSrsError

   type stsrs = { FSymSet:FunSet.set, 
		  DSymSet:FunSet.set,
		  CSymSet:FunSet.set,
		  VarCond:bool,    (* Does satisfy l \notin V, V(r) \subseteq (l) ? *)
		  Rules: (Term.term * Term.term) list,
		  HdMap: int FunMap.map  (* depth of head occurrence of lhs *)
		}

   fun prRule (l,r) = (STTerm.toString l) ^ " -> " ^ (STTerm.toString r)  
   fun prEq (l,r) = (STTerm.toString l) ^ " = " ^ (STTerm.toString r) 
   val prRules = PrintUtil.prList prRule
   val prEqs = PrintUtil.prList prEq

   (* well-typed $B$J$i(B SOME typedRules $B$rJV$7!$$=$&$G$J$1$l$P(B NONE $B$rJV$9(B *)
   fun attachTypeToRules (decls:decl list) rules =
       let exception NotWellSorted
	   val srules = 
	       L.map (fn (l,r) => 
			 case (attachTypeToTerm decls l,
			       attachTypeToTerm decls r) of
			 (SOME sl, SOME sr) => 
			 if Sort.equal (sortOfTerm sl, sortOfTerm sr)
			 then (sl,sr)
			 else (print ("Lhs&Rhs sorts not equal: " 
				      ^ (IOStsrs.prRule (l,r)) ^ "\n");
			       raise NotWellSorted)
		       | _ => (print ("Non-well-sorted term found: " 
				      ^ (IOStsrs.prRule (l,r)) ^ "\n");
			       raise NotWellSorted))
		     rules
       in
	   SOME srules
	   handle NotWellSorted => NONE
       end

   fun fdcSetInRules rs = 
       let fun headFun (Var _) = NONE
             | headFun (Fun (f,[],_)) = SOME f
             | headFun (Fun (f,t::ts,_)) = headFun t
	   val heads = L.map (fn (l,_) => headFun l) rs
	   val vars = L.map (fn (l,r) => (varSetInTerm l, varSetInTerm r)) rs
	   val varCond = (L.all isSome heads)
			 andalso (L.all (fn (vl,vr) => VS.isSubset (vr,vl)) vars)
	   val _ = if varCond then ()
		   else if (L.all isSome heads)
		   then (print "The following rules have a extra variable in rhs.\n";
			 L.app (fn (l,r) => if not (VS.isSubset (varSetInTerm r, varSetInTerm l))
					    then print (IOStsrs.prRule (l,r) ^ "\n")
					    else ())
			       rs)
		   else (print "The following rules have a variable lhs:\n";
			 L.app (fn (l,r) => if (isVar l) 
					    then print (IOStsrs.prRule (l,r) ^ "\n")
					    else ())
			       rs)
	   val dSyms = L.foldr 
			   (fn (x,xset) => if isSome x 
					   then FS.add (xset, valOf x) 
					   else xset)
			   FS.empty 
			   heads
 	   val fSyms = FS.difference (Trs.funSetInRules rs, FS.singleton (Fun.fromString "@"))
	   val cSyms = FS.difference (fSyms, dSyms)
       in 
	   (fSyms,dSyms,cSyms,varCond)
       end

   fun mkHdMap rs = 
       foldl (fn ((l,_),m) => let val (h,d) = headAndDepthOfTerm l
			      in case h of
				     Var _ => m
				   | Fun (f,_,_) => 
				     case FM.find (m,f) of  (* modified 2009.6.15 *)
					 NONE => FM.insert (m,f,d)
				       | SOME d' => if d' <= d then m
						    else FM.insert (m,f,d)
			      end)
	     FM.empty rs

   fun rulesToStsrs rs =
       let val (fSyms,dSyms,cSyms,varCond) = fdcSetInRules rs
       in { FSymSet = fSyms,
	    DSymSet = dSyms,
	    CSymSet = cSyms,
	    VarCond = varCond,
	    Rules = rs,
	    HdMap = mkHdMap rs }
       end

  (* argument expansion $B$7$?$b$N$rA4$FJV$9(B *)
   fun etaExpandRule (l,r) = 
       case (sortOfTerm l) of
	   S.Base _ => []
	 | S.Proc (ss,s) => 
	   let val m = Term.maxVarIndexInTerms [l,r]
	       val name = Atom.atom "x"
	       val vars = L.map (fn ty => Var ((name,m+1),ty)) ss
	       val (l',r') = case (appendArgs l vars,
				   appendArgs r vars) of 
				 (SOME l', SOME r')  => (l',r')
			       | (_, _) => (print "etaExpand fails\n";
					    raise STSrsError)
	   in
	       (l',r') :: (etaExpandRule (l',r'))
	   end
   
   fun etaExpandRules rules = mapAppend etaExpandRule rules

   fun flattingRules rs = L.map (fn (l,r) => (flattingTerm l,flattingTerm r)) rs

   fun annotateRules rs = L.map (fn (l,r) => (annotateTerm l,annotateTerm r)) rs

   fun unannotateRules rs = L.map (fn (l,r) => (unannotateTerm l,unannotateTerm r)) rs


   end (* of local *)

end (* of structure STSrs *)

