(******************************************************************************
 * file: rwtools/stsrs/st_term.sml
 * description: definition and utility functions for Simply Typed S-Expression
 * author: AOTO Takahito
 * 
 * Copyright (c) 2008-2009, Toyama&Aoto Laboratory, Tohoku University
 ******************************************************************************)

signature ST_TERM = 
   sig
   val isAppSym: Fun.ord_key -> bool
   val isDefinedHead: FunSet.set -> Term.term -> bool
   val isVarHead: Term.term -> bool
   val isDefinedOrVarHead: FunSet.set -> Term.term -> bool

   val headOfTerm: Term.term -> Term.term
   val headOfSubterms: Term.term -> Term.term list
   val headOfNonVariableSubterms: Term.term -> Term.term list

   val anonymizeTerm: Term.term -> Term.term
   val headOfTerm': Term.term -> Term.term
   val headOfSubterms': Term.term -> Term.term list
   val headOfNonVariableSubterms': Term.term -> Term.term list
   val headAndDepthOfTerm: Term.term -> Term.term * int
   val headAndDepthOfSubterms: Term.term -> (Term.term * int) list
   val headAndDepthOfApplicativeSubterms: Term.term -> (Term.term * int) list
   val headAndDepthOfTerm': Term.term -> Term.term * int
   val headAndDepthOfSubterms': Term.term -> (Term.term * int) list
   val headAndDepthOfApplicativeSubterms': Term.term -> (Term.term * int) list
   val sortOfHeadAndDepth: Term.term * int -> Sort.ord_key option
   val sortsOfVhSubterms: Term.term -> Sort.ord_key list
   val primaryVariablesOfTerm: Term.term -> Term.term list 
   val toString: Term.term -> string
   val appendArgs: Term.term -> Term.term list -> Term.term option
   val extractPattern: Term.term -> Term.term
   val attachTypeToTerm: Term.decl list -> Term.term -> Term.term option
   val flattingTerm: Term.term -> Term.term
   val annotateApp: Term.term * int * Sort.ord_key -> Fun.ord_key 
   val prAnnotatedAppSymbol: Fun.ord_key -> string
   val annotateTerm: Term.term -> Term.term
   val unannotateTerm: Term.term -> Term.term
end;

structure STTerm : ST_TERM = 
   struct

   local 
       open Term
       structure P = Prop
       structure S = Sort
       structure FS = FunSet
       structure VS = VarSet
       structure VM = VarMap
       structure TS = TermSet
       structure FM = FunMap
       structure TM = TermMap
       structure FIS = FunIntSet
       structure FIM = FunIntMap
       structure L = List
       structure LP = ListPair
       fun mapAppend f xs = List.foldr (fn (x,ys) => List.@(f x, ys)) [] xs
   in

   exception STTermError;

   val appendLetter = ref #"@"
   val appSymbol  = Fun.fromString (str (!appendLetter))

   fun isAppSym f = (String.sub (Fun.toString f,0)) = (!appendLetter)

   fun isDefinedHead dSymSet (Var _) = false
     | isDefinedHead dSymSet (Fun (f,_,_)) = FS.member (dSymSet,f)

   fun isDefinedOrVarHead dSymSet (Var _) = true
     | isDefinedOrVarHead dSymSet (Fun (f,_,_)) = FS.member (dSymSet,f)

   fun isVarHead (Var _) = true
     | isVarHead (Fun _) = false

   fun headOfTerm (t as (Var _)) = t
     | headOfTerm (t as (Fun (f,[],_))) = t 
     | headOfTerm (Fun (f,t::ts,_)) = headOfTerm t

   fun headOfSubterms (t as (Var _)) = [t]
     | headOfSubterms (t as (Fun (f,[],_))) = [t]
     | headOfSubterms (Fun (f,t::ts,_)) = mapAppend headOfSubterms (t::ts)

   fun headOfNonVariableSubterms (t as (Var _)) = []
     | headOfNonVariableSubterms (t as (Fun (f,[],_))) = [t]
     | headOfNonVariableSubterms (t as (Fun (f,t0::ts,_))) = 
       (headOfTerm t)::(mapAppend headOfNonVariableSubterms (t0::ts))

   (* $BJQ?t$r(B anonymous $B$K$9$k(B *)
   fun anonymizeTerm (Var (_,ty)) = 
       let (* $BL>A0$K7?$rF~$l$F$*$+$J$$$HB>$N7?$NJQ?t$H(B identify $B$5$l$F$7$^$&$+$b(B *)
           val name = Atom.atom ("var(" ^ (Sort.toString ty) ^ ")")
       in Var ((name,0),ty)
       end 
     | anonymizeTerm t = t

   (* $BJQ?t$,(B anonymous $B$K$J$k%P!<%8%g%s(B *)
   fun headOfTerm' (t as (Var _)) = anonymizeTerm t
     | headOfTerm' (t as (Fun (f,[],_))) = t 
     | headOfTerm' (Fun (f,t::ts,_)) = headOfTerm' t

   fun headOfSubterms' (t as (Var _)) = [anonymizeTerm t]
     | headOfSubterms' (t as (Fun (f,[],_))) = [t]
     | headOfSubterms' (Fun (f,t::ts,_)) = mapAppend headOfSubterms' (t::ts)

   fun headOfNonVariableSubterms' (t as (Var _)) = []
     | headOfNonVariableSubterms' (t as (Fun (f,[],_))) = [t]
     | headOfNonVariableSubterms' (t as (Fun (f,t0::ts,_))) = 
       (anonymizeTerm (headOfTerm t))::(mapAppend headOfNonVariableSubterms' (t0::ts))

   (* The depth defined like this: 
      d(t) = 0 for t \in \Sigma \cup V; d((t0 ... tN)) = d(t0) + 1 *)
   fun headAndDepthOfTerm (t as Var (_,ty)) = (t,0)
     | headAndDepthOfTerm (t as Fun (f,[],ty)) = (t,0)
     | headAndDepthOfTerm (t as Fun (_,t0::ts,_)) =
       let val (h,d) = headAndDepthOfTerm t0
       in (h, d+1)
       end

   (* $BJQ?t$,(B anonymous $B$K$J$k%P!<%8%g%s(B *)
   fun headAndDepthOfTerm' (t as Var (_,ty)) = (anonymizeTerm t,0)
     | headAndDepthOfTerm' (t as Fun (f,[],ty)) = (t,0)
     | headAndDepthOfTerm' (t as Fun (_,t0::ts,_)) =
       let val (h,d) = headAndDepthOfTerm t0
       in (anonymizeTerm h, d+1)
       end

   fun sortOfHeadAndDepth (t,n) = Sort.plus (Term.sortOfTerm t) n

   fun headAndDepthOfSubterms (t as Var (_,ty)) = [(t,0)]
     | headAndDepthOfSubterms (t as Fun (f,[],ty)) = [(t,0)]
     | headAndDepthOfSubterms (t as Fun (_,t0::ts,_)) =
       let val (h,d) = headAndDepthOfTerm t0
       in (h, d+1)::(mapAppend headAndDepthOfSubterms (t0::ts))
       end

   (* $BJQ?t$,(B anonymous $B$K$J$k%P!<%8%g%s(B *)
   fun headAndDepthOfSubterms' (t as Var (_,ty)) = [(anonymizeTerm t,0)]
     | headAndDepthOfSubterms' (t as Fun (f,[],ty)) = [(t,0)]
     | headAndDepthOfSubterms' (t as Fun (_,t0::ts,_)) =
       let val (h,d) = headAndDepthOfTerm' t0
       in (anonymizeTerm h, d+1)::(mapAppend headAndDepthOfSubterms' (t0::ts))
       end

   (* $BDj?t$dJQ?t$N$N(B(h,d)$B$OF~$l$J$$!$(Bapplication $B$5$l$k(B (h,d)$B$N$_(B *)
   fun headAndDepthOfApplicativeSubterms (t as Var (_,ty)) = []
     | headAndDepthOfApplicativeSubterms (t as Fun (f,[],ty)) = []
     | headAndDepthOfApplicativeSubterms (t as Fun (_,t0::ts,_)) =
       let val (h,d) = headAndDepthOfTerm t0
       in (h, d)::(mapAppend headAndDepthOfApplicativeSubterms (t0::ts))
       end

   fun headAndDepthOfApplicativeSubterms' (t as Var (_,ty)) = []
     | headAndDepthOfApplicativeSubterms' (t as Fun (f,[],ty)) = []
     | headAndDepthOfApplicativeSubterms' (t as Fun (_,t0::ts,_)) =
       let val (h,d) = headAndDepthOfTerm' t0
       in (anonymizeTerm h, d)::(mapAppend headAndDepthOfApplicativeSubterms' (t0::ts))
       end

   (* VhSubterms: subterm u s.t. hd(u) \in V *)
   fun sortsOfVhSubterms (Var (_,ty)) = [ty]
     | sortsOfVhSubterms (Fun (f,[],_)) = []
     | sortsOfVhSubterms (t as Fun (_,t0::ts,ty)) = 
       if Term.isVar (headOfTerm t0)
       then ty::(mapAppend sortsOfVhSubterms (t0::ts))
       else (mapAppend sortsOfVhSubterms (t0::ts))

   fun primaryVariablesOfTerm (t as Var (_,ty)) = []
     | primaryVariablesOfTerm (t as Fun (f,[],ty)) = []
     | primaryVariablesOfTerm (t as Fun (_,t0::ts,_)) =
       if Term.isVar t0
       then t0::(mapAppend primaryVariablesOfTerm (t0::ts))
       else (mapAppend primaryVariablesOfTerm (t0::ts))


   (* appendArgs t [s0,...,sn] = (t s0 ... tn) *)
   fun appendArgs t ts  =  
       let val sortOfArgs = L.map sortOfTerm ts
       in
	   case sortOfTerm t of
	       S.Proc (ss,ty) => if L.all 
				      (fn (a,b) => Sort.equal (a,b)) 
				      (LP.zip (ss,sortOfArgs))
			       then SOME (Fun (Atom.atom "@", t::ts, ty))
			       else NONE
	     | S.Base _ => NONE
       end

   fun prTerm (Var (x,ty)) = (Var.toString x) ^ ":" ^ (S.toString ty)
     | prTerm (Fun (f,ts,ty)) = 
       if isAppSym f 
       then "(" ^ (prTermList ts) ^ ")"
       else (Fun.toString f) 
   and prTermList ts = PrintUtil.prSpace prTerm ts

   val toString = prTerm

   (* pattern:  ((hd(t) x1 .. xN) y1 ...yM) ...)  *)
   fun extractPattern (t as Var _)  = t
     | extractPattern (t as Fun (f,[],_))  = t
     | extractPattern (Fun (f,t0::ts,ty))  =
       let val p0 = extractPattern t0
	   val mvi = maxVarIndexInTerm p0
	   val vname = Atom.atom "z" 
	   val vs =  L.tabulate (length ts,
				 fn i => Var ((vname,mvi+i+1),sortOfTerm (L.nth (ts,i))))
       in Fun (f,p0::vs,ty)
       end

   (* well-typed $B$J$i(B SOME typedTerm $B$rJV$7!$$=$&$G$J$1$l$P(B NONE $B$rJV$9(B *)
   fun attachTypeToTerm (decls:decl list) term =
       case term of
	   Var _ => SOME term
	 | Fun (f,[],_) =>
	   let
	       val spec = L.find (fn x => (Fun.equal (#sym x, f))) decls
	   in
	       if (isSome spec) 
	       then SOME (Fun (f,[],#sort (valOf spec)))
	       else NONE
	   end
	 | Fun (f,t::ts,_) =>
	   let
	       val s = attachTypeToTerm decls t
	       val ss = map (attachTypeToTerm decls) ts
	   in
	       if (isSome s) andalso (L.all isSome ss) 
	       then 
		   let 
		       val h = valOf s
		       val args = L.map valOf ss
		       val hsort = sortOfTerm h
		       val sorts = L.map sortOfTerm args
		   in case sortOfTerm h of
			  S.Base _ => NONE
			| S.Proc (types,ty) =>
			  if ListPair.all (fn (x,y) => Sort.equal (x, y)) (sorts,types)
			  then SOME (Fun (f,h::args,ty))
			  else NONE
		   end
	       else
		   NONE
	   end


   (* flattening:
      (((f s1 ... sl) t1 ... tm) u1 ... un)  
          =>  f(s1, ... ,sl,t1, ... ,tm,u1, ... ,un)
    *) 

   fun flattingTerm (t as (Var (v,_))) =  (print "Error: flattingTerm\n";
					   raise STTermError)
     | flattingTerm (t as (Fun (c,[],_))) = Fun (c,[],Sort.null)
     | flattingTerm (Fun (f,t0::ts,ty)) = 
       let val s0 = flattingTerm t0
       in  case s0 of
	       Var _ =>  (print "Error: flattingTerm\n";
					   raise STTermError)
	     | Fun (c,us,_) =>  Fun (c, us@ts, Sort.null)
       end


   (* annotating:
      ( @(t0,t1,...,tn)
          =  @_{<hd(t0),depth(t0)>}(t0,...,tn)  when hd(t0) \in \Sigma
          =  @_{<\tau,depth(t0)>}(t0,...,tn)  when hd(t0) \in V
    *) 

   fun prAnnotatedAppSymbol s = 
       let val s' = String.extract (Fun.toString s,2,NONE)
	   val ss = String.tokens (fn (#",":char)=>true | _ => false) s'
	   val s1 = hd ss
	   val s2 = hd (tl ss)
	   val s3 = hd (tl (tl ss))
       in s1 ^ "," ^  s3 ^ " (type: " ^ s2 ^ ")"
       end

   fun annotateApp (Var (v,_),d,ty) = 
       Fun.fromString ("@_<var(" ^  (Sort.toString ty) ^ "),"  
		       ^(Sort.toString ty) ^ "," ^ (Int.toString d) ^ ">")
     | annotateApp (Fun (c,_,_),d,ty) = 
       Fun.fromString ("@_<" ^  (Atom.toString c) ^ ","  
		       ^  (Sort.toString ty) ^ "," ^ (Int.toString d) ^ ">")

   fun annotateTerm (t as (Var _)) =  t
     | annotateTerm (t as (Fun (_,[],_))) = t
     | annotateTerm (Fun (f,t0::ts,ty)) = 
       let val (h,d) = headAndDepthOfTerm t0 (* annotateApp $B$G(B annonymous $B2=$7$F$$$k(B *)
	   val ty' = case sortOfHeadAndDepth (h,d) of
			NONE => (print "Error: annotateTerm failed\n"; raise STTermError)
		      | SOME ty => ty
	   val f' = annotateApp (h,d,ty')
	   val ts' = L.map annotateTerm (t0::ts)
       in  Fun (f',ts',ty)
       end

   val appSymbol = Fun.fromString (str (!appendLetter))

   fun unannotateTerm (t as (Var _)) =  t
     | unannotateTerm (t as (Fun (_,[],_))) = t
     | unannotateTerm (Fun (f,t0::ts,ty)) = Fun (appSymbol,L.map annotateTerm (t0::ts),ty)

   end (* of local *)

   end; (* of structure STTerm *)

