(******************************************************************************
 * file: rwtools/stsrs/stchecker.sml
 * description: Termination Checker for STSRS
 * author: AOTO Takahito
 * 
 * Copyright (c) 2008, Toyama&Aoto Laboratory, Tohoku University
 ******************************************************************************)

signature STCHECKER =  
  sig
      val runDebug: bool ref
      val main: string * string list -> OS.Process.status
      val mkCommand: unit -> unit
  end;

structure Stchecker: STCHECKER =
    struct

    val runDebug = ref false : bool ref
    fun debug f = if !runDebug then f () else ()

    exception StcheckerError

    local
	val numOfSuccess = ref 0;
	val numOfFailure = ref 0;

	exception Abort;

	datatype option_result = HelpOption | Proceed

	val usage = "Usage: cmd <file or directory>"
	
	val useDpOption = ref false
	val useRpOption = ref true
	val useScOption = ref true
	val useUrOption = ref true
	val useInOption = ref false

	val useQuasiOption = ref true
	val useLexOption = ref false
	val useMulOption = ref false
	val useAfOption = ref true

	val timeLimit = ref 0
	val multipleEncodingMethods = ref false (* $B7PO)=g=x$,(B2$B$D0J>e;XDj$5$l$J$$$3$H$r%A%'%C%/(B *)

	fun setInit () = (
			  useDpOption := false;
			  useRpOption := true;
			  useScOption := true;
			  useUrOption := true;
			  useInOption := false;

			  useQuasiOption := true;
			  useLexOption := false;
			  useMulOption := false;
			  useAfOption := true;

			  timeLimit := 0;
			  multipleEncodingMethods := false)

	val _ = setInit ();


	fun setEncodingMethod name =
	    if (!multipleEncodingMethods = true) then
		(PrintUtil.println "multiple -o option found"; 
		 PrintUtil.println usage;
		 raise Abort)
	    else
		let val _ = multipleEncodingMethods := true;
		    val _ = case name
			     of "c" => (useLexOption := false;useMulOption := false)
  			      | "l" => (useLexOption := true;useMulOption := false)
  			      | "m" => (useLexOption := false;useMulOption := true)
  			      | "r" => (useLexOption := true;useMulOption := true)
			      | ow => (PrintUtil.println ("illegal option: -o" ^ ow); 
				       PrintUtil.println usage;
				       raise Abort)
		in Proceed
		end

	fun setUseOption useOption str name =
	    let val _ = case name
			of "0" => useOption := false
			 | "1" => useOption := true
			 | "yes" => useOption := true
			 | "no" => useOption := false
			 | ow => (PrintUtil.println ("illegal option: -" ^ str ^ ow); 
				  PrintUtil.println usage;
				  raise Abort)
	    in Proceed
	    end

	val minisatPath = ref ""
	fun setMinisatPath path = 
	    let val path' = OS.FileSys.fullPath path
		    handle OS.SysErr (msg,_) => 
			   (PrintUtil.println ("file not found: " ^ path);
			    raise Abort)
	    in if OS.FileSys.access (path', [OS.FileSys.A_EXEC])
	       then (minisatPath := path'; Proceed)
	       else (PrintUtil.println ("file not executable: " ^ path);
		     raise Abort)
	    end

	val tmpDir = ref (OS.FileSys.fullPath ".")
	fun setTmpDir path = 
	    let val path' = OS.FileSys.fullPath path
		val _ = debug (fn _ => print ("tmpDir: " ^ path' ^ "\n"))
	    in if OS.FileSys.access (path', [OS.FileSys.A_WRITE])
	       then (tmpDir := path'; Proceed)
	       else (PrintUtil.println ("directory not writable: " ^ path);
		     raise Abort)
	    end

	fun setTimerOption num  =
	    let val _ = case Int.fromString num
			of SOME i => timeLimit := i
			 | NONE => (PrintUtil.println ("illegal option: -t" ^ num); 
				    PrintUtil.println usage;
				    raise Abort)
	    in Proceed
	    end

	val options = [{short = "h",
		        long = ["help"],
			desc = GetOpt.NoArg (fn () => HelpOption),
			help = "print this message"},
		       {short = "o",
		        long = ["order"],
			desc = GetOpt.ReqArg (setEncodingMethod,"name"),
			help = "specify a path order c=>lpo,l=>lpos,m=>mpo,r=>rpo"},
		       {short = "",
		        long = ["qo", "use-quasi-order"],
			desc = GetOpt.ReqArg (setUseOption useQuasiOption "qo","0|1|no|yes"),
			help = "use a quasi base order"},
		       {short = "",
		        long = ["af", "use-argument-filtering"],
			desc = GetOpt.ReqArg (setUseOption useAfOption "af","0|1|no|yes"),
			help = "use a argument filtering"},
		       {short = "",
		        long = ["sc", "use-subterm-criteria"],
			desc = GetOpt.ReqArg (setUseOption useScOption "sc","0|1|no|yes"),
			help = "use subterm criteria"},
		       {short = "",
		        long = ["rp", "use-reduction-pair"],
			desc = GetOpt.ReqArg (setUseOption useRpOption "rp","0|1|no|yes"),
			help = "use reduction pair"},
		       {short = "",
		        long = ["ur", "use-usable-rules"],
			desc = GetOpt.ReqArg (setUseOption useUrOption "ur","0|1|no|yes"),
			help = "use reduction pair"},
		       {short = "d",
		        long = ["dp", "use-dp"],
			desc = GetOpt.ReqArg (setUseOption useDpOption "d","0|1|no|yes"),
			help = "use dependency pair method"},
		       {short = "t",
		        long = ["timer"],
			desc = GetOpt.ReqArg (setTimerOption, "int"),
			help = "specify the time limit in seconds"},
		       {short = "",
		        long = ["minisat-path"],
			desc = GetOpt.ReqArg (setMinisatPath, "path"),
			help = "specify a executable minisat"},
		       {short = "",
		        long = ["tmp-dir"],
			desc = GetOpt.ReqArg (setTmpDir, "dir"),
			help = "specify a directory where temporal files will be placed"},
		       {short = "",
		        long = ["debug"],
			desc = GetOpt.NoArg (fn () => (runDebug := true; 
						       STDpSolver.runDebug := true; 
						       STPoSolver.runDebug := true; 
						       DpSolver.runDebug := true; 
						       Dp.runDebug := true; 
						       Proceed)),
			help = "debug mode"}
		      ]

	fun checkFile (path,filename) =
	    let 
		fun main () = 
		    let 
			val _ = useAfOption := ((!useDpOption) andalso (!useAfOption))
			val _ = useScOption := ((!useDpOption) andalso (!useScOption))
			val _ = useRpOption := ((!useDpOption) andalso (!useRpOption))
			val _ = useInOption := ((!useDpOption) andalso (!useInOption))
			val _ = useUrOption := ((!useDpOption) andalso (!useUrOption))
			val opt1 = {useSc=(!useScOption),
				    useRp=(!useRpOption),
				    useUr=(!useUrOption),
				    useIn= false}
			val opt2 = {useQuasi=(!useQuasiOption),
				    useAf=(!useAfOption),
				    useLex=(!useLexOption),
				    useMul=(!useMulOption)}
			val (sorts,tupdecl,rules) = IOStspec.rdFile filename
			val (decls : Term.decl list)
			  = map (fn x => {sym = #1 x,sort = #2 x}) tupdecl 
			val _ = debug (fn _=> print "Sorts:\n")
			val _ = debug (fn _=> print (PrintUtil.prList Atom.toString sorts))
				
			val _ = debug (fn _=> print "Declaration:\n")
			val _ = debug (fn _=> print (Term.prDecls decls))

			val _ = debug (fn _=> print "Rules:\n")
			val _ = debug (fn _=> print (PrintUtil.prList IOStsrs.prRule rules))

			val srules =  case STSrs.attachTypeToRules decls rules of 
					   SOME srules => srules
					 | NONE => (print "not well-sorted rules found\n";
						    raise StcheckerError) 

			val _ = debug (fn _=> print "Sorted rules in detail:\n")
			val _ = debug (fn _=> print (PrintUtil.prList IOStsrs.prRule' srules))

			val _ = debug (fn _ => print ("\n"))
			val _ = debug (fn _ => 
					  let val mesg = 
						  "Start Checking " 
						  ^ (if path = "" 
						     then filename 
						     else (path ^ "/" ^ filename))
					  in
					      print (mesg ^ "\n");
					      print (implode (List.tabulate (size mesg, fn x => #"=")));
					      print "\n"
					  end)

			val result  = if !useDpOption
				      then STDpSolver.dpSolver (!minisatPath) (!tmpDir)
							     opt1 opt2 (sorts,decls,srules)
				      else STPoSolver.poSolver (!minisatPath) (!tmpDir)
							     opt2 (sorts,decls,srules)


		    in 
			(print (if path = "" then filename else (path ^ "/" ^ filename));
			 if result
			 then (numOfSuccess := !numOfSuccess + 1; 
			       print ": Success")
			 else (numOfFailure := !numOfFailure + 1; 
			       print ": Failure"))
		    end
		val timer = Timer.startCPUTimer ()
 		val start = TimeUtil.checkTime timer 
 		val result = TimeUtil.timeLimitRun (IntInf.fromInt (!timeLimit)) main () 
		    handle TimeLimit.TimeOut => 
                           (if OS.FileSys.access (!Solver.inTmpFileName, [OS.FileSys.A_WRITE])
			    then (OS.FileSys.remove (!Solver.inTmpFileName);
				  Solver.inTmpFileName := "")
			    else ();
			    if OS.FileSys.access (!Solver.outTmpFileName, [OS.FileSys.A_WRITE])	
			    then (OS.FileSys.remove (!Solver.outTmpFileName);
				  Solver.outTmpFileName := "")
			    else ();
			    print (if path = "" then filename else (path ^ "/" ^ filename));
                            numOfFailure := !numOfFailure + 1; 
                            print ": Failure(timeout)")
 		val stop = TimeUtil.checkTime timer 
		val _ = print (" ("  ^ (TimeUtil.reportTime (start,stop)) ^ " msec.)\n")
	    in
		result
	    end

	fun reportTotal timer = 
 	   let val _ = print "\n=============================\n"
	       val _ = print ("Total: " 
			      ^ (Int.toString (!numOfSuccess)) ^ " Success, " 
			      ^ (Int.toString (!numOfFailure)) ^ " Failure: ")
	       val _ = print (Int.toString (LargeInt.toInt (TimeUtil.checkTime timer)) 
			      ^ " millisec.\n")
	   in () end


    in
    fun main (name:string, args:string list) =
	let  
	    val timer = Timer.startCPUTimer()
 	    val start = TimeUtil.checkTime timer 
	    val commandDir = OS.FileSys.getDir();
	    val _ = setInit ();
	    val targets = 
		case (GetOpt.getOpt {argOrder = GetOpt.Permute,
				     options = options,
				     errFn = (fn str => (PrintUtil.println usage;
							 raise Abort))}
				    args) of
		    (xs,args) =>  if List.exists (fn x => x = HelpOption) xs
				  then (PrintUtil.println (GetOpt.usageInfo 
							  {header="rewriting checker v.0.04",
							   options=options});  
					raise Abort)
				  else case args of 
					   [] =>(PrintUtil.println "file or directory not specified";
						 PrintUtil.println usage; raise Abort)
					 | xs => List.map 
						 (fn x => OS.FileSys.realPath x
 						     handle OS.SysErr (msg,_) => 
							    (PrintUtil.println ("file or directory not found: " 
									   ^ x);
							     raise Abort))
						 xs

	    val _ = if "" = !minisatPath 
		    then let val path' = OS.FileSys.fullPath "./minisat"
				 handle OS.SysErr (msg,_) => 
					(PrintUtil.println ("./minisat not found; " 
						       ^ "use -m to specify minisat-path");
					 raise Abort)
			 in if OS.FileSys.access (path', [OS.FileSys.A_EXEC])
			    then minisatPath := path'
			    else (PrintUtil.println ("./minisat not executable; "
						^ "use -m to specify minisat-path");
				  raise Abort)
			 end
		    else ()
	in
	    if length targets = 1 
	       andalso not (OS.FileSys.isDir (hd targets))
	    then 
		(CommandUtil.applyCmdToDirTree checkFile ("", hd targets);
		 setInit ();
		 OS.Process.success)
	    else
		(List.app (fn item => 
			      CommandUtil.applyCmdToDirTree checkFile ("", item))
			  targets;
		 reportTotal timer;
		 setInit ();
		 OS.Process.success)
	    handle e  => (OS.FileSys.chDir commandDir; raise (e))
	end
        handle Abort => OS.Process.failure
    end

    fun mkCommand () = SMLofNJ.exportFn("../../work/stchecker/stchecker",main);

end



