(******************************************************************************
 * Copyright (c) 2012-2015, Toyama&Aoto Laboratory, Tohoku University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the 
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(******************************************************************************
 * file: rwtools/term_rewriting/subst.sml
 * description: substitution, matching, unification
 * author: AOTO Takahito
 * 
 ******************************************************************************)

signature SUBST = 
   sig
   type subst = Term.term VarMap.map
   val empty: subst
   val applySubst: subst -> Term.term -> Term.term
   val inDomain: Term.var_key -> subst -> bool
   val restrictDomain: VarSet.set -> subst -> subst
   val compose: subst -> subst -> subst
   val merge: subst * subst -> subst option
   val mergeList: subst list -> subst option
   val match: Term.term -> Term.term -> subst option
   val matchSubterm: Term.term -> Term.term -> bool
   val isVariant: Term.term -> Term.term -> bool
   val identicalModuloRenamingRule: (Term.term * Term.term) -> (Term.term * Term.term) -> bool
   val renameTermToStd: Term.term -> Term.term
   val renameTerms: Term.term list -> Term.term list
   val renameTermDisjointFrom: Var.ord_key list -> Term.term -> Term.term
   val unify: Term.term -> Term.term -> subst option 
   val disjointUnify: Term.term * Term.term -> bool
   val mergeUnify: subst * subst -> subst option
   val mergeUnifyList: subst list -> subst option
   val linearize: Term.term -> Term.term
   val toString: subst -> string
   val toStringWithVarSort: subst -> string 
   val fromString: string -> subst option

   end;

structure Subst : SUBST = 
   struct

   local 
       open Term;
       structure VS = VarSet
       structure VM = VarMap
       structure L = List
       structure LP = ListPair
   in

   type subst = term VM.map

   val empty = VM.empty

   (* applySubst sigma term :  sigma  term Ŭ  *)

   fun applySubst sigma term = 
       case term 
	of Var (x,_) =>  (case (VM.find (sigma, x))
			   of SOME t => t | NONE => term)
	 | Fun (f,ts,sort) => Fun (f, L.map (applySubst sigma) ts, sort)

   fun inDomain x sigma =  VM.inDomain (sigma, x)

   (* f = compose g h defined by f(x) = g(h(x)) *)
   fun compose sub1 sub2 = 
       let val sub3 = 
	       VM.foldli (fn (v,t,map) => VM.insert (map,v,applySubst sub1 t)) VM.empty sub2
       in
	   VM.foldli (fn (v,t,map) => VM.insert (map,v,t)) sub1 sub3
       end

   (*** Example 
   - val sub1 = valOf (Subst.fromString "{?x:=f(?y),?y:=a}");
   - val sub2 = valOf (Subst.fromString "{?y:=g(?z),?z:=h(?y)}");
   - Subst.toString (Subst.compose sub1 sub2);
     val it = "{?x:=f(?y),?y:=g(?z),?z:=h(a)}" : string
   - Subst.toString (Subst.compose sub2 sub1);
     val it = "{?x:=f(g(?z)),?y:=a,?z:=h(?y)}" : string
    ***)

   fun restrictDomain vSet sigma =  VM.filteri (fn (v,t)=> VS.member (vSet,v)) sigma

   fun merge (sigma1,sigma2) =
       let val common = VM.intersectWith (fn (u,v)=> if Term.equal (u,v)
						     then SOME u
						     else NONE) (sigma1,sigma2)
       in if List.all isSome (VM.listItems common)
	  then SOME (VM.unionWith (fn (u,v) => u) (sigma1,sigma2))
	  else NONE
       end

   fun mergeList [] = SOME VM.empty
     | mergeList [sigma] = SOME sigma
     | mergeList (sigma1::sigma2::rest)  = 
       case merge (sigma1,sigma2) of
	   SOME sigma => mergeList (sigma::rest)
	 | NONE => NONE


   (* match pattern term : term  pattern ˾ȹ礹֤ *)
   local
       exception UnMatched
       fun matchMain [] sigma = SOME sigma
	 | matchMain ((s0,t0)::rest) sigma = 
	   case s0 
	    of Var (x,_) => (case (VM.find (sigma,x))
			      of SOME t1 => (case compare (t0,t1) 
					      of EQUAL => matchMain rest sigma 
					       | _ => raise UnMatched)
			       | NONE => matchMain rest (VM.insert (sigma, x,t0)))
	     | Fun (f,ts,_) => case t0 
				of Var _ => raise UnMatched
				 | Fun (g,ss,_) => case Fun.compare (f,g)
						    of EQUAL => matchMain (L.@(LP.zip (ts,ss), rest)) sigma
						     | _ => raise UnMatched
   in 
   fun match s t = matchMain [(s,t)] VM.empty
       handle UnMatched => NONE
   end

   fun matchSubterm pat (t as (Fun (f,ts,_))) =  isSome (match pat t) orelse
						 L.exists (fn ti => matchSubterm pat ti) ts
     | matchSubterm pat t = isSome (match pat t)




(*    fun identicalModuloRenamingRule (l,r) (l',r') = *)
(*        let val null = Atom.atom "" *)
(* 	   val rule1 = Fun (null,[l,r],Sort.null) *)
(* 	   val rule2 = Fun (null,[l',r'],Sort.null) *)
(*        in *)
(* 	   isSome (match rule1 rule2) *)
(* 	   andalso isSome (match rule2 rule1) *)
(*        end *)


   local
       exception NonRenaming

       fun isMono ren = 
	   let fun checkMono [] ys = true
		 | checkMono (x::xs) ys = 
		   if L.exists (fn y => Var.equal (x,y)) ys 
		   then false
		   else checkMono xs (x::ys)
	   in checkMono (L.map (fn (k,v) => v) (VM.listItemsi ren)) []
	   end

       fun isRenaming (Var _) (Fun _) ren = raise NonRenaming
	 | isRenaming (Fun _) (Var _) ren = raise NonRenaming
	 | isRenaming (Var (x,_)) (Var (y,_)) ren = 
	   (case VM.find (ren, x) of
		SOME z => if Var.equal (y,z) then ren 
			  else raise NonRenaming
	      | NONE => VM.insert (ren,x,y))
	 | isRenaming (Fun (f,ts,_)) (Fun (g,ss,_)) ren = 
	   if Fun.equal (f,g)
	   then LP.foldl 
		    (fn (ti,si,ans) => isRenaming ti si ans)
		    ren
		    (ts,ss)
	   else raise NonRenaming
   in
   fun isVariant s t = 
       let val sigma = isRenaming s t VM.empty
       in isMono sigma
       end
       handle NonRenaming => false
   fun identicalModuloRenamingRule (l1,r1) (l2,r2) = 
       let val sigma = isRenaming r1 r2 (isRenaming l1 l2 VM.empty)
       in isMono sigma
       end
       handle NonRenaming => false
   end


   fun renameTermToStd term =
       let val count = ref 0
	   val ren = ref VM.empty
	   fun newvar () = (count := (!count+1); 
			    Var.fromStringAndInt ("x",!count))
	   fun renaming (Var (x,ty)) = 
	       (case VM.find (!ren,x) of 
		    SOME y => Var (y,ty)
		  | NONE => let val y = newvar ()
				val _ = ren := VM.insert (!ren,x,y)
			    in Var (y,ty)
			    end)
	     | renaming (Fun (f,ts,ty)) =
	       Fun (f, List.map renaming ts,ty)
       in
	   renaming term
       end

   (* renameTerms [t0,t1,...] : t0,t1,...ѿ dijoint ˤʤ褦 term Υǥå
						  ѹΥꥹȤ֤ *)

   local
       fun varIndexes (Var (x,_)) = [#2 x]
	 | varIndexes (Fun (_,ts,_)) = L.concat (L.map varIndexes ts)
       fun renameTermsMain [] _ accum = (L.rev accum)
	 | renameTermsMain (t0::ts) n accum =
	   let val idx = (varIndexes t0)
	   in if null idx then renameTermsMain ts n (t0::accum)
	      else let val (i, j) = (L.foldr Int.min (L.hd idx) (L.tl idx), 
				     L.foldr Int.max (L.hd idx) (L.tl idx))
		   in
		       renameTermsMain ts (n-i+j+1) ((increaseVarIndexBy (n-i) t0)::accum)
		   end
	   end
   in
   fun renameTerms ts = renameTermsMain ts 0 []
   fun renameTermDisjointFrom vs t = 
       if null vs then t
       else let val (n::ns) = L.map (fn (x,y) => y) vs
		val m = L.foldl Int.max n ns
       		val i = L.foldl Int.min 0 (varIndexes t)
	    in if m < i then t
	       else increaseVarIndexBy (m-i+1) t
	    end
   end


   (* unify term1 term2 : term1  term2 ñ첽ǽʤ mgu ֤ *)
   (* ɬפʤ unify  renaming Τ˺ʤ *)
   local
       exception UnUnified
       fun applySubst1 (x,t) term = 
	   case term 
	    of Var (y,_) =>  (case Var.compare (x,y) of EQUAL => t | _ => term)
	     | Fun (f,ts,sort) => Fun (f, map (applySubst1 (x,t)) ts, sort)
       fun unifyMain [] sigma = SOME sigma
	 | unifyMain ((s0,t0)::rest) sigma = 
	   case (s0,t0)
	    of (Var (x,_), Var (y,_)) => 
	       (case Var.compare (x,y) 
		 of EQUAL => unifyMain rest sigma
		  | _ => eliminate x t0 rest sigma)
	     | (Fun (f,ts,_), Fun (g,ss,_)) => 
	       (case Fun.compare (f,g)
		 of EQUAL => unifyMain ((LP.zip (ts,ss))@rest) sigma
		  | _ => raise UnUnified)
	     | (Var (x,_), Fun _) => 
	       if VS.member (varSetInTerm t0, x) 
	       then raise UnUnified
	       else eliminate x t0 rest sigma
	     | (Fun _, Var (y,_)) => 
	       if VS.member (varSetInTerm s0, y) 
	       then raise UnUnified
	       else eliminate y s0 rest sigma
       and eliminate x t0 rest sigma =
	   let
	       val rest' = (map (fn (s,t) => (applySubst1 (x,t0) s, applySubst1 (x,t0) t)) rest)
	       val sigma' = VM.map (fn s => applySubst1 (x,t0) s) sigma
	   in 
               unifyMain rest' (VM.insert (sigma',x,t0))
	   end
   in 
   fun unify s t = unifyMain [(s,t)] VM.empty
       handle UnUnified => NONE
   end (* of local *)


   (* disjointUnify term1 term2 : term1  term2 ñ첽ǽʤ true ֤ *)
   (* term1  term2 ѿ disjoint ȸ *)
   fun disjointUnify (term1,term2) = 
       let val [s,t] = renameTerms [term1,term2]
       in isSome (unify s t)
       end

(*
   fun disjointUnify term1 term2 = 
       let val count = (maxVarIndexInTerm term1) + 1
	   val term2' = increaseVarIndexBy count term2
	   val varset1 = varSetInTerm term1
	   val varset2 = varSetInTerm term2'
	   fun incIndexesBy n (Var ((x,i),sort)) = if VS.member (varset2,(x,i)) then Var ((x,i+n), sort)
						   else Var ((x,i), sort)
	     | incIndexesBy n (Fun (f,ts,sort)) = Fun (f, L.map (incIndexesBy n) ts, sort)


	   val makeSigma1 = VM.foldli (fn (v,t,map) => if VS.member (varset1,v)
						       then VM.insert (map,v,incIndexesBy (~count) t)
						       else map)
				      VM.empty 
	   val makeSigma2 = VM.foldli (fn (v,t,map) => if VS.member (varset2,v)
						       then VM.insert (map,Var.increaseIndexBy (~count) v,
								       incIndexesBy (~count) t)
						       else map)
				      VM.empty
	   val sigmaOp = unify term1 term2'
       in
	   case sigmaOp of 
	       SOME sigma => SOME (makeSigma1 sigma, makeSigma2 sigma)
	     | NONE => NONE
       end
Bug?: +(x,y)  +((y,z),x)  disjointUnify  
      (x := y+z, z:=y) Ȥʤꡤ   +(+(y,z),y) ȤʤäƤޤ

*)
   (*** Example
       - Subst.disjointUnify (IOFotrs.rdTerm "ap(ap(L,?x),?y)") (IOFotrs.rdTerm "ap(?y0,?y0)");
       - print (let val (sub1,sub2) = valOf it in (Subst.toString sub1) ^ (Subst.toString sub2) end);
         {?y:=ap(L,?x)}{?y0:=ap(L,?x)}val it = () : unit
    ******)


   fun mergeUnify (sigma1,sigma2) =
       let val common = VM.intersectWith (fn (u,v)=> unify u v) (sigma1,sigma2)
       in if List.all isSome (VM.listItems common)
	  then SOME (VM.unionWithi (fn (k,u,v) => let val sub = valOf (valOf (VM.find (common,k)))
						  in applySubst sub u
						  end)
				   (sigma1,sigma2))
	  else NONE
       end

   fun mergeUnifyList [] = SOME VM.empty
     | mergeUnifyList [sigma] = SOME sigma
     | mergeUnifyList (sigma1::sigma2::rest)  = 
       case mergeUnify (sigma1,sigma2) of
	   SOME sigma => mergeUnifyList (sigma::rest)
	 | NONE => NONE



   (* 2ܤиѿ̾ؤơ *)
   fun linearize term = 
       let val count = ref (maxVarIndexInTerm term)
	   fun newvar () = (count := (!count+1); 
			    Var.fromStringAndInt ("x",!count))
	   fun rename vset (Var (x,ty)) = if VS.member (vset,x)
					  then (Var (newvar (),ty), vset)
					  else (Var (x,ty), VS.add (vset,x))
	     | rename vset (Fun (f,ts,ty)) = let val (ts',vset') = renameList vset ts
					     in (Fun (f,ts',ty),vset') 
					     end
	   and renameList vset []  = ([],vset)
	     | renameList vset (t::ts) = 
	       let val (t',vset1) = rename vset t
		   val (ts',vset2) = renameList vset1 ts
	       in (t'::ts', vset2)
	       end
	   val (new,_) = rename VS.empty term
       in new
       end

   fun toString sub = let fun prPair (x,t) = (Var.toString x) ^ ":=" ^ (Term.toString t)
		      in ListUtil.toStringCommaCurly prPair (VM.listItemsi sub)
		      end

   fun toStringWithVarSort sub = 
       let fun prPair (x,t) = (Var.toString x) ^ ":=" ^ (Term.toStringWithVarSort t)
       in ListUtil.toStringCommaCurly prPair (VM.listItemsi sub)
       end


   fun fromString str = let val ts = IOFotrs.rdAssignmentSet str
			in
			    if L.all (fn (x,_) => isVar x) ts
			    then 
				SOME (L.foldl (fn ((v,t),vmap) => VM.insert (vmap,valOf (varRootOfTerm v),t))
					      VM.empty
					      ts)
			    else NONE
			end
				   

   end (* of local *)

   end;

