(******************************************************************************
 * file: rwtools/sprint/dec.sml
 * description: decision procedures
 * author: AOTO Takahito
 * 
 * $Id: dec.sml,v 1.7 2011/09/23 03:58:29 aoto Exp $
 * Copyright (c) 2014, Toyama&Aoto Laboratory, Tohoku University
 ******************************************************************************)

signature DEC = 
sig
    val indNatTimes: (Fun.ord_key * Fun.ord_key * Fun.ord_key * Fun.ord_key)
			    -> (Term.term * Term.term) -> bool
    val indNatPlus: (Fun.ord_key * Fun.ord_key * Fun.ord_key) 
			    -> (Term.term * Term.term) -> bool
    val indListLen: (Fun.ord_key * Fun.ord_key * Fun.ord_key * Fun.ord_key  * Fun.ord_key
		     * Fun.ord_key * Fun.ord_key * Fun.ord_key * Fun.ord_key)
		    -> (Sort.ord_key * Sort.ord_key)
		    -> (Term.term * Term.term) -> bool
    val indNatMaxMin: (Fun.ord_key * Fun.ord_key * Fun.ord_key) -> Sort.ord_key
			    -> (Term.term * Term.term) -> bool
    val indNatMax: (Fun.ord_key * Fun.ord_key * Fun.ord_key)
			    -> (Term.term * Term.term) -> bool
    val indNatMin: (Fun.ord_key * Fun.ord_key * Fun.ord_key)
			    -> (Term.term * Term.term) -> bool
end;

structure Dec : DEC = 
struct
local
    structure F = Fun
    structure FS = FunSet
    structure S = Sort
    structure SS = SortSet
    structure L = List
    structure LP = ListPair
    structure LU = ListUtil
    structure VM = VarMap
    structure VS = VarSet
    structure YI = YicesInput
in
val runDebug = ref false : bool ref
fun debug f = if !runDebug then f () else ()

exception DecError of string

fun toString xs = LU.toStringCommaCurly
		  (fn (x,N) => "<" ^ (Int.toString x) ^ "," ^ (LU.toStringCommaSquare Var.toString N) ^ ">")
		  xs

local (* for indNatTimes, indNatPlus, indListLen *)
    fun add (x,N) [] = [(x,N)]
      | add (x,N) ((y,M)::ys) = if LU.multisetEqual' Var.equal (N,M)
				then (x+y,N)::ys
				else (y,M)::(add (x,N) ys)
    fun union (xs,ys) = L.foldl (fn ((x,N),ys') => add (x,N) ys') ys xs
    fun join (xs,ys) = ListXProd.foldX (fn ((x,N),(y,M),zs) => add (x*y, N @ M) zs) (xs,ys) [] 

    val singleton = Fun.fromString "_singleton"
in
fun indNatTimes (times,plus,succ,zero) (term1,term2) = 
    let fun trans (s as (Term.Var (x,_))) = [(1,[x])]
	  | trans (Term.Fun (f,ts,ty)) = 
	    if Fun.equal (f,times) then join (trans (hd ts), trans (hd (tl ts)))
	    else if Fun.equal (f,plus) then union (trans (hd ts), trans (hd (tl ts)))
	    else if Fun.equal (f,succ)  then add (1,[]) (trans (hd ts))
	    else if Fun.equal (f,zero) then []
	    else (print "indNatTimes"; raise DecError "indNatTimes")
	val ts1 = trans term1
	val _ = print ("tranlated lhs: " ^ (toString ts1) ^ "\n")
	val ts2 = trans term2
	val _ = print ("tranlated rhs: " ^ (toString ts2) ^ "\n")
    in LU.multisetEqual' (fn ((x,N),(y,M)) => x = y andalso (LU.multisetEqual' Var.equal (N,M)))
			    (ts1, ts2)
    end

fun indNatPlus (plus,succ,zero) (term1,term2) = 
    indNatTimes (Fun.fromString "", plus,succ,zero) (term1,term2)

fun indListLen (len,rev,app,times,plus,cons,nilX,succ,zero) (Lsort,Nsort) (term1,term2) = 
    let val count = ref (Term.maxVarIndexInTerms [term1, term2])
	fun newVar () = (count:= (!count)+1; Var.fromStringAndInt ("x",!count))

	val table = ref VM.empty
	fun rename x = case VM.find (!table, x) of
			   SOME y => y
			 | NONE => let val y = newVar () 
				   in (table := VM.insert (!table,x,y); y) end

	(* cons => singleton *)
	fun trans1 (s as (Term.Var _)) = s
	  | trans1 (Term.Fun (f,ts,ty)) =
	    if Fun.equal (f,cons) 
	    then Term.Fun (app, [Term.Fun (singleton,[trans1 (hd ts)],ty), trans1 (hd (tl ts))], ty)
	    else Term.Fun (f,L.map trans1 ts,ty)

	(* eliminate len *)
	fun trans2 (s as (Term.Var _)) = s
	  | trans2 (s as Term.Fun (f,ts,ty)) =
	    if Fun.equal (f,len) then trans2sub (hd ts)
	    else Term.Fun (f,L.map trans2 ts,ty)
	and trans2sub (Term.Var (y,ty)) = Term.Var (rename y, Nsort)
	  | trans2sub (Term.Fun (f,ts,ty)) = 
	    if Fun.equal (f,nilX) 
	    then Term.Fun (zero,[],Nsort)
	    else if Fun.equal (f,singleton)
	    then Term.Fun (succ,[Term.Fun (zero,[],Nsort)],Nsort)
	    else if Fun.equal (f,rev) 
	    then trans2sub (hd ts)
	    else if Fun.equal (f,app) 
	    then Term.Fun (plus,[trans2sub (hd ts), trans2sub (hd (tl ts))], Nsort)
	    else (print "indListLen"; raise DecError "indListLen")

	(* distribute rev *)
	fun trans3 s =
	    let fun even n = (Int.mod (n,2) = 0)
		fun trans3sub n (t as (Term.Var (_,ty))) = 
		    if even n then t else Term.Fun (rev,[t],ty)
		  | trans3sub n (Term.Fun (f,ts,ty)) = 
		    if Fun.equal (f,rev) 
		    then trans3sub (n+1) (hd ts)
		    else if Fun.equal (f,singleton) 
		    then Term.Fun (f, [trans3sub 0 (hd ts)],ty)
		    else if Fun.equal (f,app)
		    then if even n 
			 then Term.Fun (f, [trans3sub n (hd ts), trans3sub n (hd (tl ts))], ty)
			 else Term.Fun (f, [trans3sub n (hd (tl ts)), trans3sub n (hd ts)], ty)
		    else Term.Fun (f,L.map (trans3sub n) ts,ty)
	    in trans3sub 0 s end

	fun decomp (s as (Term.Var _)) = [s]
	  | decomp (s as (Term.Fun (f,ts,ty))) = 
	    if Fun.equal (f,app)
	    then decomp (hd ts) @ decomp (hd (tl ts))
	    else [s]

	fun chkelm (Term.Var (x1,ty1), Term.Var(x2,ty2)) = Var.equal (x1,x2)
	  | chkelm (Term.Fun _, Term.Var _) = false
	  | chkelm (Term.Var _, Term.Fun _) = false
	  | chkelm (Term.Fun (f1,ts1,ty1), Term.Fun (f2,ts2,ty2)) =
	    (Fun.equal (f1, rev) andalso Fun.equal (f2, rev) 
	     andalso Term.isVar (hd ts1) andalso Term.equal (hd ts1, hd ts2))
	    orelse 
	    (Fun.equal (f1, singleton) andalso Fun.equal (f2, singleton) 
	     andalso indNatTimes (times,plus,succ,zero) (hd ts1,hd ts2))

	(* val _ = print "test...1\n" *)

    in if not (Sort.equal (Term.sortOfTerm term1, Term.sortOfTerm term2))
       then (print "indListLen"; raise DecError "indListLen")
       else if Sort.equal (Term.sortOfTerm term1, Nsort)
       then let val t1 = (trans3 o trans2 o trans1) term1
		val t2 = (trans3 o trans2 o trans1) term2
	    in indNatTimes (times,plus,succ,zero) (t1,t2)
	    end
       else if Sort.equal (Term.sortOfTerm term1, Lsort)
       then let (* val _ = print "test...2\n" *)
		(* val _ = print ((Term.toString term1) ^ "\n") *)
		(* val a1 = trans1 term1 *)
		(* val _ = print "test...2a\n" *)
		(* val _ = print ((Term.toString a1) ^ "\n") *)
		(* val a2 = trans2 a1 *)
		(* val _ = print "test...2b\n" *)
		(* val _ = print ((Term.toString a2) ^ "\n") *)
		(* val a3 = trans3 a2 *)
		(* val _ = print "test...2c\n" *)
		(* val _ = print ((Term.toString a3) ^ "\n") *)
		(* val ts1 = decomp a3 *)
		val ts1 = (decomp o trans3 o trans2 o trans1) term1
		val _ = print ("translated lhs: " ^ (LU.toStringCommaSquare Term.toString ts1) ^ "\n")
		(* val _ = print (LU.toStringCommaSquare Term.toString ts1) *)
		(* val _ = print "test...3\n" *)
		val ts2 = (decomp o trans3 o trans2 o trans1) term2 
		val _ = print ("translated rhs: " ^ (LU.toStringCommaSquare Term.toString ts2) ^ "\n")
		(* val _ = print (LU.toStringCommaSquare Term.toString ts2) *)
		(* val _ = print "test...4\n" *)
	    in L.length ts1 = L.length ts2
		  andalso L.all chkelm (LP.zip (ts1,ts2))
	    end
       else (print "indListLen"; raise DecError "indListLen")
    end
end (* of local *)

structure InfVarSet: sig
    include ORD_KEY2 
datatype elem = Vars of VarSet.set | Top
datatype quasi_order = LESS | EQUAL | GREATER | INCOMPARABLE
val compareAsSet: ord_key * ord_key -> quasi_order
val takeMins: ord_key list -> ord_key list
end = struct 
datatype elem = Vars of VarSet.set | Top
type ord_key = elem
fun compare (Top, Top) = EQUAL
  | compare (Vars _, Top) = LESS
  | compare (Top, Vars _) = GREATER
  | compare (Vars vs1, Vars vs2) = VarSet.compare (vs1,vs2)
fun equal (x,y) = (compare (x,y)) = EQUAL
datatype quasi_order = LESS | EQUAL | GREATER | INCOMPARABLE
fun compareAsSet (Top, Top) = EQUAL
  | compareAsSet (Vars _, Top) = LESS
  | compareAsSet (Top, Vars _) = GREATER
  | compareAsSet (Vars vs1, Vars vs2) = 
    case (VarSet.isSubset (vs1,vs2), VarSet.isSubset (vs2,vs1)) of
	(true,true) => EQUAL
      | (true,false) => LESS
      | (false,true) => GREATER
      | (false,false) => INCOMPARABLE
fun toString Top = "top"
  | toString (Vars vs) = ListUtil.toStringCommaCurly Var.toString (VarSet.listItems vs)

local (* takeMinsSub *)
    exception FoundSmaller
in
fun takeMins [] = []
  | takeMins (x::ys) = case (takeMinsSub x ys) of
			   SOME ys' => x::(takeMins ys')
			 | NONE => takeMins ys
and takeMinsSub x ys = 
    let val ys' = L.mapPartial (fn y => case compareAsSet (x,y) of 
					    GREATER => raise FoundSmaller
					  | LESS => NONE
					  | EQUAL => NONE
					  | INCOMPARABLE => SOME y)  ys
    in SOME ys' 
    end handle FoundSmaller => NONE
end (* of local *)
end (* of struct *)

structure InfVarSetSet = RedBlackSetFn (InfVarSet) : ORD_SET

local (* for indNatMaxMin *)
fun mkMaxMinRules (max,min) sortN = 
    let val varX = Var.fromStringAndInt ("x",0)
	val varY = Var.fromStringAndInt ("y",0)
	val varZ = Var.fromStringAndInt ("z",0)
	val x = Term.mkVarTerm (varX, sortN)
	val y = Term.mkVarTerm (varY, sortN)
	val z = Term.mkVarTerm (varZ, sortN)
	val maxXY = Term.mkFunTerm (max,[x,y],sortN)
	val maxYZ = Term.mkFunTerm (max,[y,z],sortN)
	val minXZ = Term.mkFunTerm (min,[x,z],sortN)
	val minYZ = Term.mkFunTerm (min,[y,z],sortN)
	val minXY = Term.mkFunTerm (min,[x,y],sortN)
	val rule1 = Trs.mkRule (Term.mkFunTerm (min,[maxXY,z],sortN),
				Term.mkFunTerm (max,[minXZ,minYZ],sortN))
	val rule2 = Trs.mkRule (Term.mkFunTerm (min,[x,maxYZ],sortN),
				Term.mkFunTerm (max,[minXY,minXZ],sortN))
    in [rule1,rule2]
    end

structure IVS = InfVarSet
structure IVSS = InfVarSetSet

fun interpretMinTerm (min,zero) s = 
    if FS.member (Term.funSetInTerm s,zero) 
    then IVS.Top
    else IVS.Vars (Term.varSetInTerm s)

fun interpretMaxTerm (max,min,zero) s = 
    if Term.isFun s
       andalso Fun.equal (max,valOf (Term.funRootOfTerm s))
    then let val [vss1,vss2] = L.map (interpretMaxTerm (max,min,zero)) (Term.argsOfTerm s)
	 in IVSS.union (vss1,vss2)
	 end
    else IVSS.singleton (interpretMinTerm (min,zero) s)

in
fun indNatMaxMin (max,min,zero) sortN (term1,term2) = 
    let val rulesMaxMin = mkMaxMinRules (max,min) sortN
	val nf1 = Rewrite.linf rulesMaxMin term1
	val nf2 = Rewrite.linf rulesMaxMin term2
	val vss1 = IVSS.listItems (interpretMaxTerm (max,min,zero) nf1)
	val vss2 = IVSS.listItems (interpretMaxTerm (max,min,zero) nf2)
	(* val _ = print ((LU.toStringSemicolonCurly IVS.toString vss1) ^ "\n") *)
	(* val _ = print ((LU.toStringSemicolonCurly IVS.toString vss2) ^ "\n") *)
	val mins1 = IVS.takeMins vss1
	val mins2 = IVS.takeMins vss2

	val _ = print ((LU.toStringSemicolonCurly IVS.toString mins1) ^ "\n")
	val _ = print ((LU.toStringSemicolonCurly IVS.toString mins2) ^ "\n")
    in LU.setEqual' (fn (x,y) => IVS.compareAsSet (x,y) = IVS.EQUAL) (mins1,mins2)
    end
end (* of local *)


local (* for indNatMax, indNatMin *)
fun compare ((n1,s1), (n2,s2)) =
    if Term.equal (s1,s2)
    then if n1 > n2 then SOME GREATER
	 else if n1 < n2 then SOME LESS
	 else SOME EQUAL
    else if Term.isFun s1 andalso n1 <= n2
    then SOME LESS
    else if Term.isFun s2 andalso n1 >= n2
    then SOME GREATER
    else NONE

(* maxOrMin \in { max, min } *)
fun trans (maxOrMin,succ,zero) n s = 
    if Term.isVar s then [(n,s)]
    else let val (f,ts,ty) = valOf (Term.decFunTerm s)
	 in if Fun.equal (f,maxOrMin) then LU.mapAppend (trans (maxOrMin,succ,zero) n) ts
	    else if Fun.equal (f,succ) then trans (maxOrMin,succ,zero) (n+1) (hd ts)
	    else if Fun.equal (f,zero) then [(n,s)]
	    else (print "indNatMax/Min"; raise DecError "indNatMax/Min")
	 end
in
fun indNatMax (max,succ,zero) (term1,term2) = 
    let val ts1 = trans (max,succ,zero) 0 term1
	val ts2 = trans (max,succ,zero) 0 term2
	(* val _ = print ((LU.toStringCommaCurly *)
	(* 		    (fn (n,s) => "<" ^ (Int.toString n) ^ "," ^ (Term.toString s) ^ ">") ts1) ^ "\n") *)
	(* val _ = print ((LU.toStringCommaCurly *)
	(* 		    (fn (n,s) => "<" ^ (Int.toString n) ^ "," ^ (Term.toString s) ^ ">") ts2) ^ "\n") *)
	val maxs1 = LU.takeMaximals compare ts1
	val maxs2 = LU.takeMaximals compare ts2
	val _ = print ("translated lhs: " ^ 
		       (LU.toStringCommaCurly
	 		    (fn (n,s) => "<" ^ (Int.toString n) ^ "," ^ (Term.toString s) ^ ">") maxs1)
		       ^ "\n")
	val _ = print ("translated rhs: " ^ 
		       (LU.toStringCommaCurly
	 		    (fn (n,s) => "<" ^ (Int.toString n) ^ "," ^ (Term.toString s) ^ ">") maxs2)
		       ^ "\n")
	(* val _ = print ((LU.toStringCommaCurly *)
	(* 		    (fn (n,s) => "<" ^ (Int.toString n) ^ "," ^ (Term.toString s) ^ ">") maxs2) ^ "\n") *)
    in LU.setEqual' (fn (x,y)=> compare (x,y) = SOME EQUAL) (maxs1,maxs2)
    end

fun indNatMin (min,succ,zero) (term1,term2) = 
    let val ts1 = trans (min,succ,zero) 0 term1
	val ts2 = trans (min,succ,zero) 0 term2
	(* val _ = print ((LU.toStringCommaCurly  *)
	(* 		    (fn (n,s) => "<" ^ (Int.toString n) ^ "," ^ (Term.toString s) ^ ">") ts1) ^ "\n") *)
	(* val _ = print ((LU.toStringCommaCurly  *)
	(* 		    (fn (n,s) => "<" ^ (Int.toString n) ^ "," ^ (Term.toString s) ^ ">") ts2) ^ "\n") *)
	val mins1 = LU.takeMinimals compare ts1
	val mins2 = LU.takeMinimals compare ts2
	val _ = print ("translated lhs: " ^ 
		       (LU.toStringCommaCurly 
			    (fn (n,s) => "<" ^ (Int.toString n) ^ "," ^ (Term.toString s) ^ ">") mins1)
		       ^ "\n")
	val _ = print ("translated rhs: " ^ 
		       (LU.toStringCommaCurly 
			    (fn (n,s) => "<" ^ (Int.toString n) ^ "," ^ (Term.toString s) ^ ">") mins2)
		       ^ "\n")

	(* val _ = print ((LU.toStringCommaCurly  *)
	(* 		    (fn (n,s) => "<" ^ (Int.toString n) ^ "," ^ (Term.toString s) ^ ">") mins2) ^ "\n") *)
    in LU.setEqual' (fn (x,y)=> compare (x,y) = SOME EQUAL) (mins1,mins2)
    end

end (* of local *)		 

fun encodeToFormula (max,min,plus,succ,zero) (term1,term2) =
    let val count = ref (Term.maxVarIndexInTerms [term1, term2])
	fun newVar () = (count:= (!count)+1; Var.fromStringAndInt ("x",!count))
	fun trans s =
	    if Term.isVar s
	    then ([],s)
	    else let val (f,ts,ty) = valOf (Term.decFunTerm s)
		 in if Fun.equal (f,max) orelse Fun.equal (f,min)
		    then let val (ts1,t1) =  trans (hd ts)
			     val (ts2,t2) =  trans (hd (tl ts))
			     val x = newVar ()
			     val varX = Term.mkVarTerm (x,ty)
			 in (((varX,Term.mkFunTerm (f,[t1,t2],ty))::ts1@ts2),varX)
			 end
		    else if Fun.equal (f,plus)
		    then let val (ts1,t1) =  trans (hd ts)
			     val (ts2,t2) =  trans (hd (tl ts))
			 in (ts1@ts2,Term.mkFunTerm (f,[t1,t2],ty))
			 end
		    else if Fun.equal (f,succ)
		    then let val (ts1,t1) =  trans (hd ts)
			 in (ts1,Term.mkFunTerm (f,[t1],ty))
			 end
		    else if Fun.equal (f,zero)
		    then ([],s)
		    else (print "indNatMaxMinPlus"; raise DecError "indNatMaxMinPlus")
		 end
	val (ts1,t1) = trans term1
	val (ts2,t2) = trans term2
	(* val (xs,ys) = LP.unzip (ts1 @ ts2) *)
	(* val vset = Term.varSetInTerms (t1::t2::xs@ys) *)
	(* val _ = print "forall " *)
	(* val _ = print (LU.toStringComma Var.toString (VS.listItems vset)) *)
	(* val _ = print ". " *)
	(* val _ = print (LU.toStringComma (fn (l,r) => (Term.toString l) ^ "=" ^ (Term.toString r)) (ts1@ts2)) *)
	(* val _ = print " => " *)
	(* val _ = print ((Term.toString t1) ^ "=" ^ (Term.toString t2) ^ "\n") *)
    in (ts1@ts2, t1, t2)
    end

fun indNatMaxMinPlus smtSolver (max,min,plus,succ,zero) (term1,term2) =
    let val (ts,t1,t2) = encodeToFormula (max,min,plus,succ,zero) (term1,term2)
	val (xs,ys) = LP.unzip ts
	val vset = VS.listItems (Term.varSetInTerms (t1::t2::xs@ys))
	
	val cmd = smtSolver ^ " 2>/dev/null\n"
 	val proverCmd = ("/bin/sh", ["-c",cmd])
	val proc = Unix.execute proverCmd
	val (ins,outs) = Unix.streamsOf proc

	(* val tmpName = #file (OS.Path.splitDirFile (OS.FileSys.tmpName ())) *)
	(* val inTmpFile = OS.Path.joinDirFile {dir=".",file=tmpName} *)
	(* val outs = TextIO.openOut inTmpFile *)
        (* val _ = print ("tmpfile: " ^ tmpName ^ "\n") *)

	val _ = L.app (fn x => TextIO.output (outs, YI.prDefInt (Var.toString x))) vset
	val _ = L.app (fn x => TextIO.output (outs, YI.prAssert (YI.prGe (Var.toString x, YI.prZero)))) vset

	fun encodePlusTerm t = 
	    if Term.isVar t
	    then Term.toString t
	    else let val (f,ts,ty) = valOf (Term.decFunTerm t)
		 in if Fun.equal (f,plus)
		    then YI.prPlus [encodePlusTerm (hd ts), encodePlusTerm (hd (tl ts))]
		    else if Fun.equal (f,succ)
		    then YI.prPlus [encodePlusTerm (hd ts), Int.toString 1]
		    else if Fun.equal (f,zero)
		    then YI.prZero
		    else (print "indNatMaxMinPlus"; raise DecError "indNatMaxMinPlus")
		 end

        (* l: var term, r = min (u,v) or r = max (u,v) for u,v: T({+,s,0},V) *)
	fun encodeMinMax (l,r) = 
	    let val (f,ts,ty) = valOf (Term.decFunTerm r)
		val strL = encodePlusTerm l
		val strU = encodePlusTerm (hd ts)
		val strV = encodePlusTerm (hd (tl ts))
	    in if Fun.equal (f,min)
	       then YI.prIfThenElse (YI.prLe (strU,strV)) (YI.prEq (strL,strU)) (YI.prEq (strL,strV))
	       else if Fun.equal (f,max)
	       then YI.prIfThenElse (YI.prLe (strU,strV)) (YI.prEq (strL,strV)) (YI.prEq (strL,strU))
	       else (print "indNatMaxMinPlus"; raise DecError "indNatMaxMinPlus")
	    end

	val _ = L.app (fn (l,r) => TextIO.output (outs, YI.prAssert (encodeMinMax (l,r)))) ts

	val _ = TextIO.output (outs, YI.prAssert (YI.prNeq (encodePlusTerm t1,encodePlusTerm t2)))

	val _ = TextIO.output (outs, "(set-evidence! true)\n")
	val _ = TextIO.output (outs, "(check)\n")
	val _ = TextIO.flushOut outs

(*	val answer = SOME "unsat\n" *)
	val answer = TextIO.inputLine ins
	val _ = if (isSome answer)
                then debug (fn _=> print ("output:" ^ (valOf answer)))
                else ()

	val result =  (isSome answer) andalso (valOf answer = "unsat\n")

(*	val assign = YI.readAssigns ins *)
(*	val _ = print "assignment:\n" *)
(*	val _ = print (LU.toStringCommaSquare 
				    (fn i => Int.toString (lookupInt assign i)) 
				    [1,2,3,4,5,6,7,8,9,10,11,12,13]) *)
(*		 val _ = print "\n" *)

	val _ = debug (fn _=> print "finish the yices process\n")
	val _ = TextIO.closeOut outs
	val _ = TextIO.closeIn ins
	val _ = Unix.reap proc
    in result 
    end

val funNatTimes = (Fun.fromString "times", Fun.fromString "plus", 
		       Fun.fromString "s", Fun.fromString "0")
val funNatPlus = (Fun.fromString "plus", Fun.fromString "s", Fun.fromString "0")
val funListLen = (Fun.fromString "len", Fun.fromString "rev", Fun.fromString "app", 
		   Fun.fromString "times", Fun.fromString "plus", 
		   Fun.fromString "cons", Fun.fromString "nil", 
		   Fun.fromString "s", Fun.fromString "0")
val funNatMaxMinPlus = (Fun.fromString "max",Fun.fromString "min",Fun.fromString "plus",
			 Fun.fromString "s",Fun.fromString "0")
val funNatMaxMin = (Fun.fromString "max", Fun.fromString "min",Fun.fromString "0")
val funNatMax = (Fun.fromString "max", Fun.fromString "s",Fun.fromString "0")
val funNatMin = (Fun.fromString "min", Fun.fromString "s",Fun.fromString "0")

val sortList = (Sort.fromString "List", Sort.fromString "Nat")
val sortNat = Sort.fromString "Nat"

val declNatTimes = IOMstrs.rdDecls [ "times:Nat*Nat=>Nat","plus:Nat*Nat=>Nat", "s:Nat=>Nat", "0: Nat" ]
val declNatPlus = IOMstrs.rdDecls [ "plus:Nat*Nat=>Nat", "s:Nat=>Nat", "0: Nat" ]
val declListLen = IOMstrs.rdDecls [ "len:List=>Nat", "rev:List=>List", "app:List*List=>List", 
				     "times:Nat*Nat=>Nat","plus:Nat*Nat=>Nat", 
				     "cons:Nat*List=>List", "nil:List", "s:Nat=>Nat", "0: Nat" ]
val declNatMaxMinPlus = IOMstrs.rdDecls [ "max:Nat*Nat=>Nat", "min:Nat*Nat=>Nat", 
					"plus:Nat*Nat=>Nat", "s:Nat=>Nat", "0:Nat" ]
val declNatMaxMin = IOMstrs.rdDecls [ "max:Nat*Nat=>Nat", "min:Nat*Nat=>Nat", "0:Nat" ]
val declNatMax = IOMstrs.rdDecls [ "max:Nat*Nat=>Nat", "s:Nat=>Nat", "0:Nat" ]
val declNatMin = IOMstrs.rdDecls [ "min:Nat*Nat=>Nat", "s:Nat=>Nat", "0:Nat"]

val smtSolver = "./yices"

fun runInd 1 = indNatTimes funNatTimes
  | runInd 2 = indNatPlus funNatPlus
  | runInd 3 = indListLen funListLen sortList
  | runInd 4 = indNatMaxMin funNatMaxMin sortNat
  | runInd 5 = indNatMax funNatMax
  | runInd 6 = indNatMin funNatMin
  | runInd 7 = indNatMaxMinPlus smtSolver funNatMaxMinPlus

fun checkDecls 1 = declNatTimes
  | checkDecls 2 = declNatPlus
  | checkDecls 3 = declListLen
  | checkDecls 4 = declNatMaxMin
  | checkDecls 5 = declNatMax
  | checkDecls 6 = declNatMin
  | checkDecls 7 = declNatMaxMinPlus

val count = ref 1


fun check num [x,y] ans = 
    let val n = num mod 10
	val [u,v] = IOMstrs.rdTerms [x,y]
	val s = case Term.attachSortToTerm (checkDecls n) u of 
		    NONE => (print ("not well-sorted term: " ^ Term.toString u); raise DecError "check")
		  | SOME u' => u'
	val t = case Term.attachSortToTerm (checkDecls n) v of 
		    NONE => (print ("not well-sorted term: " ^ Term.toString v); raise DecError "check")
		  | SOME v' => v'
	val result = runInd n (s,t)
	val _ = print ((Int.toString (!count))
		       ^ ". " ^ (Term.toString s) 
		       ^ " = " ^ (Term.toString t) ^ ": "
		       ^ (if result then "yes" else "no")
		       ^ "(" ^ ans ^ ")\n")
	val _ = count:= (!count) + 1;
    in result end

(* val _ = check 1 ["?x:Nat", "?x:Nat"] "yes" *)
(* val _ = check 1 ["plus(?x:Nat,0)", "?x:Nat"] "yes" *)
(* val _ = check 1 ["plus(?x:Nat,0)", "plus(0,?x:Nat)"] "yes" *)
(* val _ = check 1 ["plus(?x:Nat,s(0))", "s(?x:Nat)"] "yes" *)
(* val _ = check 1 ["plus(s(0),?x:Nat)", "s(?x:Nat)"] "yes" *)
(* val _ = check 1 ["plus(s(0),?x:Nat)", "s(s(?x:Nat))"] "no" *)
(* val _ = check 1 ["times(s(0),?x:Nat)", "s(s(?x:Nat))"] "no" *)
(* val _ = check 1 ["times(s(0),?x:Nat)", "?x:Nat"] "yes" *)
(* val _ = check 1 ["times(0,s(?x:Nat))", "0"] "yes" *)
(* val _ = check 1 ["times(0,s(?x:Nat))", "?x:Nat"] "no" *)
(* val _ = check 1 ["times(plus(?x:Nat,?y:Nat),plus(?y:Nat,?x:Nat))",  *)
(* 	       "plus(times(?x:Nat,?x:Nat),times(s(s(0)),?x:Nat))"] "no" *)
(* val _ = check 1 ["times(plus(?x:Nat,?y:Nat),s(s(0)))",  *)
(* 	       "plus(plus(?x:Nat,?x:Nat),plus(?y:Nat,?y:Nat))"] "yes" *)
(* val _ = check 1 ["times(plus(?x:Nat,?y:Nat),?z)",  *)
(* 	       "plus(times(?z,?x:Nat),times(?y:Nat,?z))"] "yes" *)
(* val _ = check 1 ["times(plus(?x:Nat,s(0)),?x:Nat)",  *)
(* 	       "plus(times(?x:Nat,?x:Nat),?x:Nat)"] "yes" *)
(* val _ = check 1 ["times(plus(?x:Nat,s(0)),plus(s(0),?x:Nat))",  *)
(* 	       "plus(times(?x:Nat,?x:Nat),s(times(s(s(0)),?x:Nat)))"] "yes" *)
(* val _ = check 1  ["times(plus(?x:Nat,?y:Nat),plus(?y:Nat,?x:Nat))",  *)
(* 		"plus(times(?y:Nat,?y:Nat),plus(times(?x:Nat,?x:Nat),times(s(s(0)),times(?x:Nat,?y:Nat))))"] "yes" *)

(* val _ = check 2 ["plus(?x:Nat,0)", "?x:Nat"] "yes" *)
(* val _ = check 2 ["plus(?x:Nat,0)", "plus(0,?x:Nat)"] "yes" *)
(* val _ = check 2 ["plus(?x:Nat,s(0))", "s(?x:Nat)"] "yes" *)
(* val _ = check 2 ["plus(?y:Nat,s(?x:Nat))", "plus(?x:Nat,s(?y:Nat))"] "yes" *)
(* val _ = check 2 ["plus(s(0),?x:Nat)", "s(s(?x:Nat))"] "no" *)

(* val _ = check 3 ["?x:Nat", "?x:Nat"] "yes" *)
(* val _ = check 3 ["rev(?xs:List)", "?xs:List"] "no" *)
(* val _ = check 3 ["rev(rev(?xs:List))", "?xs:List"] "yes" *)
(* val _ = check 3 ["rev(app(?xs:List,?ys:List))", "app(rev(?ys:List),rev(?xs:List))"] "yes" *)
(* val _ = check 3 ["plus(len(?xs:List),len(?ys:List))", "len(app(?ys:List,?xs:List))"] "yes" *)
(* val _ = check 3 ["times(len(?xs:List),s(s(0)))", "len(app(?xs:List,?xs:List))"] "yes" *)
(* val _ = check 3 ["s(times(len(?xs:List),s(s(0))))", "len(cons(?x:Nat,app(?xs:List,?xs:List)))"] "yes" *)

(*val _ = check 3 ["cons(len(cons(0,?xs:List)),?xs:List)", 
		 "cons(plus(s(?z:Nat),plus(?y:Nat,?x:Nat)),cons(s(0),app(?ys:List,?zs:List)))"] "?" *)
(* val _ = check 3 ["rev(cons(len(?ys:List),cons(?y:Nat,?ys:List)))",
		 "cons(times(len(?ys:List),times(?x:Nat,0)),cons(plus(?z:Nat,?y:Nat),app(?xs:List,?zs:List)))"] "?" *)

 (* val _ = check 3 ["rev(cons(len(?ys:List),cons(?y:Nat,?ys:List)))", *)
 (* 		    "cons(times(len(?ys:List),times(?x:Nat,0)),cons(plus(?z:Nat,?y:Nat),app(?xs:List,?zs:List)))"] "?" *)

 (* val _ = check 3 ["rev(cons(len(?xs:List),?xs:List))", *)
 (* 		  "rev(cons(0,?xs:List))"] "?" *)


(* val _ = check 4 ["?x:Nat", "max(?x:Nat,?x:Nat)"] "yes" *)
(* val _ = check 4 ["?x:Nat", "min(?x:Nat,?x:Nat)"] "yes" *)
(* val _ = check 4 ["?x:Nat", "min(?x:Nat,?y:Nat)"] "no" *)
(* val _ = check 4 ["?x:Nat", "max(?x:Nat,?y:Nat)"] "no" *)
(* val _ = check 4 ["min(?x:Nat,?x:Nat)", "max(?x:Nat,?x:Nat)"] "yes" *)
(* val _ = check 4 ["min(?x:Nat,?y:Nat)", "max(?x:Nat,?y:Nat)"] "no" *)
(* val _ = check 4 ["min(?x:Nat,?y:Nat)", "min(?y:Nat,?x:Nat)"] "yes" *)
(* val _ = check 4 ["min(?x:Nat,max(?x:Nat,?y:Nat))", "?x:Nat"] "yes" *)
(* val _ = check 4 ["max(?x:Nat,min(?x:Nat,?y:Nat))", "?x:Nat"] "yes" *)
(* val _ = check 4 ["?x:Nat", "max(?x:Nat,0)"] "yes" *)
(* val _ = check 4 ["?x:Nat", "min(?x:Nat,0)"] "no" *)
(* val _ = check 4 ["0", "max(?x:Nat,0)"] "no" *)
(* val _ = check 4 ["0", "min(?x:Nat,0)"] "yes" *)
(* val _ = check 4 ["max(min(?y:Nat,0),max(min(?y:Nat,?z:Nat),?x:Nat))",  *)
(* 		"max(?x:Nat,min(max(?x:Nat,?y:Nat),?z:Nat))"] "yes" *)

(* val _ = check 5 ["s(?x:Nat)", "max(s(?x:Nat),?x:Nat)"] "yes" *)
(* val _ = check 5 ["?x:Nat", "max(s(0),?x:Nat)"] "no" *)
(* val _ = check 5 ["s(?x:Nat)", "max(s(0),s(?x:Nat))"] "yes" *)
(* val _ = check 5 ["s(max(?x:Nat,?y:Nat))", "max(s(?x:Nat),s(?y:Nat))"] "yes" *)

(* val _ = check 5 ["s(max(?x:Nat,?y:Nat))", "max(s(?x:Nat),s(?y:Nat))"] "yes" *)

(* val _ = check 5 ["max(?x:Nat,max(?z:Nat,0))", "max(?x:Nat,?z:Nat)"] "yes" *)


(* val _ = check 6 ["s(?x:Nat)", "min(s(?x:Nat),?x:Nat)"] "no"  *)
(* val _ = check 6 ["?x:Nat", "min(s(?x:Nat),?x:Nat)"] "yes" *)
(* val _ = check 6 ["?x:Nat", "min(s(0),?x:Nat)"] "no" *)
(* val _ = check 6 ["s(?x:Nat)", "min(s(0),s(?x:Nat))"] "no" *)
(* val _ = check 6 ["s(min(?x:Nat,?y:Nat))", "min(s(?x:Nat),s(?y:Nat))"] "yes" *)
(* val _ = check 6 ["min(min(?x:Nat,?y:Nat),?z:Nat)", "min(?x:Nat,min(?y:Nat,?z:Nat))"] "yes" *)
(* val _ = check 6 ["min(min(?x:Nat,s(?z:Nat)),s(s(?z:Nat)))", "min(s(s(?x:Nat)),min(s(?z:Nat),?x:Nat))"] "yes" *)

(* val _ = check 7 ["?x:Nat", "?x:Nat"] "yes" *)
(* val _ = check 7 ["s(min(?x:Nat,?y:Nat))", "min(s(?x:Nat),s(?y:Nat))"] "yes" *)
(* val _ = check 7 ["min(min(?x:Nat,s(?z:Nat)),s(s(?z:Nat)))", "min(s(s(?x:Nat)),min(s(?z:Nat),?x:Nat))"] "yes" *)
(* val _ = check 7 ["min(plus(?x:Nat,?z:Nat),plus(?x:Nat,?y:Nat))", "plus(?x:Nat,min(?y:Nat,?z:Nat))"] "yes" *)
(* val _ = check 7 ["s(?x:Nat)", "max(s(0),s(?x:Nat))"] "yes" *)
(* val _ = check 7 ["s(max(?x:Nat,?y:Nat))", "max(s(?x:Nat),s(?y:Nat))"] "yes" *)
(* val _ = check 7 ["?x:Nat", "max(?x:Nat,?x:Nat)"] "yes" *)
(* val _ = check 7 ["?x:Nat", "min(?x:Nat,?x:Nat)"] "yes" *)
(* val _ = check 7 ["?x:Nat", "min(?x:Nat,?y:Nat)"] "no" *)
(* val _ = check 7 ["?x:Nat", "max(?x:Nat,?y:Nat)"] "no" *)
(* val _ = check 7 ["min(?x:Nat,?x:Nat)", "max(?x:Nat,?x:Nat)"] "yes" *)
(* val _ = check 7 ["min(?x:Nat,?y:Nat)", "max(?x:Nat,?y:Nat)"] "no" *)
(* val _ = check 7 ["min(?x:Nat,?y:Nat)", "min(?y:Nat,?x:Nat)"] "yes" *)
(* val _ = check 7 ["min(?x:Nat,max(?x:Nat,?y:Nat))", "?x:Nat"] "yes" *)
(* val _ = check 7 ["max(?x:Nat,min(?x:Nat,?y:Nat))", "?x:Nat"] "yes" *)
(* val _ = check 7 ["?x:Nat", "max(?x:Nat,0)"] "yes" *)
(* val _ = check 7 ["?x:Nat", "min(?x:Nat,0)"] "no" *)
(* val _ = check 7 ["0", "max(?x:Nat,0)"] "no" *)
(* val _ = check 7 ["0", "min(?x:Nat,0)"] "yes" *)
(* val _ = check 7 ["max(min(?y:Nat,0),max(min(?y:Nat,?z:Nat),?x:Nat))", *)
(* 		"max(?x:Nat,min(max(?x:Nat,?y:Nat),?z:Nat))"] "yes" *)

(* val randKey = ref (1,2)  (* (0,1) *) *)
(* fun updateRandKey () = let val (x,y) = !randKye *)
(* 		       in randKey := (y,x+y) *)
(* 		       end *)


(* fun randomInt seed n = Random.randRange (0,n-1) seed *)

fun mkRandTerm randomInt symbols depth sort  =
    let val len = L.length symbols
	val i = randomInt len
	val (isVar,name,argSort,retSort) = L.nth (symbols, i)
    in if not (String.compare (retSort,sort) = EQUAL)
	      orelse (depth >= 5 andalso not (null argSort))
	      orelse (depth = 0 andalso (null argSort))
       then mkRandTerm randomInt symbols depth sort  (* run again *)
       else if isVar
       then Term.mkVarTerm (Var.fromStringAndInt (name,0),Sort.fromString retSort)
       else let val args = L.map (mkRandTerm randomInt symbols (depth+1)) argSort
	    in Term.mkFunTerm (Fun.fromString name, args, Sort.fromString retSort)
	    end
    end

fun mkRandRule randomInt symbols sort = (mkRandTerm randomInt symbols 0 sort, mkRandTerm randomInt symbols 0 sort)
fun mkConjectures randomInt symbols sort n = L.tabulate (n, fn _ => mkRandRule randomInt symbols sort)

structure TermPair2 : ORD_KEY = 
struct 
type ord_key = Term.term * Term.term
fun compare ((x0,y0),(x1,y1)) =
	  let val sizeX0 = Term.termSize x0
	      val sizeY0 = Term.termSize y0
	      val sizeX1 = Term.termSize x1
	      val sizeY1 = Term.termSize y1
	      val sizeZ0 = sizeX0 + sizeY0
	      val sizeZ1 = sizeX1 + sizeY1
	  in if sizeZ0 > sizeZ1 
		orelse (sizeZ0 = sizeZ1 andalso sizeX0 > sizeX1)
		orelse (sizeZ0 = sizeZ1 andalso sizeX0 = sizeX1
			andalso Term.compare (x0,x1) = GREATER)
		orelse (sizeZ0 = sizeZ1 andalso sizeX0 = sizeX1
			andalso Term.compare (x0,x1) = EQUAL
			andalso Term.compare (y0,y1) = GREATER)
	     then GREATER
	     else if Term.compare (x0,x1) = EQUAL
		     andalso Term.compare (y0,y1) = EQUAL
	     then EQUAL
	     else LESS
	  end
      fun equal (ss,tt) = compare (ss,tt) = EQUAL
      fun toString (s,t) = ("(" ^ (Term.toString s) ^ ", " ^ (Term.toString t) ^ ")")
      end 

structure TermPairSet2 = RedBlackSetFn (TermPair2) : ORD_SET
structure TPS2 = TermPairSet2

fun deleteIdenticalEqs eqs =
    let fun delete [] ans = rev ans
	  | delete ((l,r)::rs) ans = 
	    if L.exists (fn lr2 => (Subst.identicalModuloRenamingRule (l,r) lr2
				    orelse Subst.identicalModuloRenamingRule (r,l) lr2)) ans
	    then delete rs ans
	    else delete rs ((l,r)::ans)
    in delete eqs []
    end

fun deleteIdenticalEqsFrom eqs eqs' =
    LU.differenceByAll' (fn ((l,r),lr) => (Subst.identicalModuloRenamingRule (l,r) lr
					   orelse Subst.identicalModuloRenamingRule (r,l) lr))
			(eqs,eqs')

fun symbols 1 = [(false,"times",["Nat","Nat"],"Nat"),
		 (false,"plus",["Nat","Nat"],"Nat"),
		 (false,"plus",["Nat","Nat"],"Nat"),
		 (false,"s",["Nat"],"Nat"),
		 (true,"x",[],"Nat"),
		 (true,"y",[],"Nat"),
		 (false,"times",["Nat","Nat"],"Nat"),
		 (false,"s",["Nat"],"Nat"),
		 (true,"z",[],"Nat") ]
(* 1 with zero *)
  | symbols 11 = [(false,"times",["Nat","Nat"],"Nat"),
		 (false,"plus",["Nat","Nat"],"Nat"),
		 (false,"s",["Nat"],"Nat"),
		 (true,"x",[],"Nat"),
		 (true,"x",[],"Nat"),
		 (true,"y",[],"Nat"),
		 (false,"times",["Nat","Nat"],"Nat"),
		 (false,"s",["Nat"],"Nat"),
		 (true,"z",[],"Nat"),
		 (false,"plus",["Nat","Nat"],"Nat"),
		 (false,"0",[],"Nat") ] 
  | symbols 2 = [(false,"plus",["Nat","Nat"],"Nat"),
		 (false,"plus",["Nat","Nat"],"Nat"),
		 (false,"s",["Nat"],"Nat"),
		 (true,"x",[],"Nat"),
		 (true,"x",[],"Nat"),
		 (true,"x",[],"Nat"),
		 (true,"y",[],"Nat"),
		 (true,"y",[],"Nat"),
		 (false,"s",["Nat"],"Nat"),
		 (false,"s",["Nat"],"Nat"),
		 (true,"z",[],"Nat"),
		 (true,"z",[],"Nat"),
		 (false,"plus",["Nat","Nat"],"Nat"),
		 (false,"plus",["Nat","Nat"],"Nat"),
		 (true,"x",[],"Nat"),
		 (true,"y",[],"Nat") ]
(* 2 with zero *)
  | symbols 12 = [(false,"plus",["Nat","Nat"],"Nat"),
		 (false,"plus",["Nat","Nat"],"Nat"),
		 (false,"s",["Nat"],"Nat"),
		 (true,"x",[],"Nat"),
		 (true,"x",[],"Nat"),
		 (true,"x",[],"Nat"),
		 (true,"y",[],"Nat"),
		 (true,"y",[],"Nat"),
		 (false,"s",["Nat"],"Nat"),
		 (false,"s",["Nat"],"Nat"),
		 (true,"z",[],"Nat"),
		 (true,"z",[],"Nat"),
		 (false,"plus",["Nat","Nat"],"Nat"),
		 (false,"plus",["Nat","Nat"],"Nat"),
		 (true,"x",[],"Nat"),
		 (true,"y",[],"Nat"),
		 (false,"0",[],"Nat")]
  | symbols 3 = [(false,"len",["List"],"Nat"), 
		 (true,"xs",[],"List"),
		 (false,"rev",["List"],"List"),
		 (false,"app",["List","List"],"List"),
		 (true,"xs",[],"List"),
		 (true,"x",[],"Nat"),
		 (true,"ys",[],"List"),
		 (false,"app",["List","List"],"List"),
		 (false,"plus",["Nat","Nat"],"Nat"),
		 (false,"app",["List","List"],"List"),
		 (true,"x",[],"Nat"),
		 (true,"zs",[],"List"),
		 (false,"times",["Nat","Nat"],"Nat"),
		 (true,"xs",[],"List"),
		 (true,"ys",[],"List"),
		 (true,"zs",[],"List"),
		 (false,"app",["List","List"],"List"),
		 (false,"s",["Nat"],"Nat"),
		 (false,"cons",["Nat","List"],"List"),
		 (true,"y",[],"Nat"),
		 (false,"0",[],"Nat")
	      ]
  | symbols 4 = [(false,"min",["Nat","Nat"],"Nat"),
		 (true,"x",[],"Nat"),
		 (false,"min",["Nat","Nat"],"Nat"),
		 (true,"x",[],"Nat"),
		 (false,"max",["Nat","Nat"],"Nat"),
		 (true,"y",[],"Nat"),
		 (false,"max",["Nat","Nat"],"Nat"),
		 (true,"y",[],"Nat"),
		 (true,"z",[],"Nat"),
		 (false,"0",[],"Nat")
		]
  | symbols 5 = [
		 (false,"s",["Nat"],"Nat"),
		 (true,"x",[],"Nat"),
		 (false,"s",["Nat"],"Nat"),
		 (false,"max",["Nat","Nat"],"Nat"),
		 (true,"x",[],"Nat"),
		 (true,"y",[],"Nat"),
		 (false,"max",["Nat","Nat"],"Nat"),
		 (true,"y",[],"Nat"),
		 (true,"z",[],"Nat"),
		 (false,"0",[],"Nat")
		]
  | symbols 6 = [
		 (false,"s",["Nat"],"Nat"),
		 (true,"x",[],"Nat"),
		 (false,"s",["Nat"],"Nat"),
		 (false,"min",["Nat","Nat"],"Nat"),
		 (true,"x",[],"Nat"),
		 (true,"y",[],"Nat"),
		 (false,"min",["Nat","Nat"],"Nat"),
		 (true,"y",[],"Nat"),
		 (true,"z",[],"Nat"),
		 (false,"0",[],"Nat")
		]
  | symbols 7 = [(false,"plus",["Nat","Nat"],"Nat"),
		 (false,"s",["Nat"],"Nat"),
		 (false,"max",["Nat","Nat"],"Nat"),
		 (false,"min",["Nat","Nat"],"Nat"),
		 (true,"x",[],"Nat"),
		 (false,"s",["Nat"],"Nat"),
		 (false,"min",["Nat","Nat"],"Nat"),
		 (true,"x",[],"Nat"),
		 (false,"max",["Nat","Nat"],"Nat"),
		 (false,"plus",["Nat","Nat"],"Nat"),
		 (true,"y",[],"Nat"),
		 (false,"s",["Nat"],"Nat"),
		 (true,"z",[],"Nat"),
		 (false,"plus",["Nat","Nat"],"Nat"),
		 (false,"0",[],"Nat") ] 


fun topSort 3 = "List"
  | topSort n = "Nat"

exception DecGiveUpGeneration of (Term.term * Term.term) list

val correctAns = "yes"

fun scond (l,r) = not (Term.isFun l andalso Term.isFun r
		   andalso LU.member' Fun.equal (valOf (Term.funRootOfTerm l)) [Fun.fromString "s"]
		   andalso Fun.equal (valOf (Term.funRootOfTerm l), valOf (Term.funRootOfTerm r)))
fun scond2 (l,r) = LU.member' Fun.equal (Fun.fromString "s") (FS.listItems (Term.funSetInTerms [l,r]))
fun zcond (l,r) = LU.member' Fun.equal (Fun.fromString "0") (FS.listItems (Term.funSetInTerms [l,r]))
fun tcond (l,r) = LU.member' Fun.equal (Fun.fromString "times") (FS.listItems (Term.funSetInTerms [l,r]))
fun mmcond (l,r) = LU.member' Fun.equal (Fun.fromString "min") (FS.listItems (Term.funSetInTerms [l,r]))
		   andalso
		   LU.member' Fun.equal (Fun.fromString "max") (FS.listItems (Term.funSetInTerms [l,r]))
fun mmcond2 (l,r) = LU.member' Fun.equal (Fun.fromString "min") (FS.listItems (Term.funSetInTerms [l,r]))
		   orelse
		   LU.member' Fun.equal (Fun.fromString "max") (FS.listItems (Term.funSetInTerms [l,r]))
fun pcond (l,r) = LU.member' Fun.equal (Fun.fromString "plus") (FS.listItems (Term.funSetInTerms [l,r]))

fun tzcond (l,r) = (* not subterms of the form  times(0,...)  times(...,0) *) 
    let val natSort = Sort.fromString "Nat"
	val subterms = (Term.subterms l) @ (Term.subterms r)
	val zterm = Term.mkFunTerm (Fun.fromString "0", [], natSort)
    in L.all (fn u => not (Term.isFun u
			   andalso let val (f,ts,ty) = valOf (Term.decFunTerm  u)
				   in Fun.equal (f, Fun.fromString "times")
				      andalso LU.member' Term.equal zterm ts
				   end)) subterms
    end

fun condition 1 _ (l,r) = (scond (l,r)) andalso (tcond (l,r))  (* contain times *)
  | condition 11 yesno (l,r) = if yesno
			       then (scond (l,r)) andalso (tcond (l,r))
				    andalso (zcond (l,r) andalso (tzcond (l,r)) )
			       else 
				   (scond (l,r)) andalso (tcond (l,r)) andalso (zcond (l,r))
  | condition 2 _ (l,r) = (scond (l,r)) (* not s(...) = s(...) *)
  | condition 12 _ (l,r) = (scond (l,r)) andalso (zcond (l,r)) (* contain 0 *)
  | condition 4 _ (l,r) = (scond (l,r)) andalso (mmcond (l,r)) (* not s(...) = s(...) *)  (* contain both min/max *)
  | condition 5 _ (l,r) = (scond (l,r)) (* not s(...) = s(...) *) 
  | condition 6 _ (l,r) = (scond (l,r)) (* not s(...) = s(...) *) 
  | condition 7 _ (l,r) = (mmcond2 (l,r)) andalso (pcond (l,r))
  | condition _ _ _ = true

fun generate randomInt n yesno known 
  = let val _ = print (Int.toString (L.length known) ^ "\n")
(*	val conj = mkConjectures (symbols n) (topSort n) 10000 *)
	val conj = mkConjectures randomInt (symbols n) (topSort n) 2000

	val conj2 = deleteIdenticalEqsFrom conj known
	(* val _ = print (Trs.prEqs conj2)  *)
	(* val _ = print (Trs.prEqs [L.nth (conj2,46),L.nth (conj2,47),L.nth (conj2,48),L.nth (conj2,49)]) *)

	val thm0 = if yesno
		   then L.filter (fn (s,t)=> check n [IOMstrs.prTerm  s, IOMstrs.prTerm t] "?") conj2
		   else L.filter (fn (s,t)=> not (check n [IOMstrs.prTerm  s, IOMstrs.prTerm t] "?")) conj2

	val thm1 = L.filter (not o Term.equal) thm0
	val thm = L.filter (condition n yesno) thm1

	val _ = if null thm
		then (print "!!! GIVE UP !!!\n"; raise DecGiveUpGeneration known)
		else ()

	val total = deleteIdenticalEqs (TPS2.listItems (TPS2.addList (TPS2.empty, thm@known)))

    in if L.length total >= 100
       then (true, total)
       else generate randomInt n yesno total
    end
    handle DecGiveUpGeneration known => (false, known)

fun test pnum yesno n known 
  = if n >= 30 
    then known
    else let fun sum d =  if d = 0 then (1,2)  
 			  else let val (x,y) = sum (d-1) in (y,x+y) end
	     val seed = Random.rand (sum (n mod 10))
 	     fun randomInt len = Random.randRange (0,len-1) seed
	     val _ = L.tabulate (n, fn i => randomInt 3)
	     val (ans,known') =  generate randomInt pnum yesno known 
	 in if ans 
	    then known'
	    else test pnum yesno (n+1) known'
	 end
	 
fun start pnum yesno = test pnum yesno 0 []

(* val problemsNatTimesWithoutZeroNo = start 1 false *)
(* val problemsNatTimesWithZeroNo = start 11 false *)
(* val problemsListLenNo = start 3 false *)
(* val problemsTimesWithoutZeroYes = start 1 true *)
(* val problemsTimesWithZeroYes = start 11 true *)
(* val problemsPlusWithoutZeroYes = start 2 true *)
(* val problemsPlusWithZeroYes = start 21 true *)
(* val problemsListLenYes = start 3 true *)

(* val problemsNatPlusWithoutZeroNo = start 2 false *)
(* val _ = print "......problemsNatPlusWithoutZeroNo.........\n" *)
(* val _ = print (LU.toStringCommaSquare *)
(*   		   (fn (l,r) => "\n\"" ^ (IOMstrs.prTerm l) ^ " = " ^ (IOMstrs.prTerm r) ^ "\"") *)
(*   		   problemsNatPlusWithoutZeroNo) *)

(* val problemsNatPlusWithZeroNo = start 12 false *)
(* val _ = print "......problemsNatPlusWithZeroNo.........\n" *)
(* val _ = print (LU.toStringCommaSquare *)
(*   		   (fn (l,r) => "\n\"" ^ (IOMstrs.prTerm l) ^ " = " ^ (IOMstrs.prTerm r) ^ "\"") *)
(*   		   problemsNatPlusWithZeroNo) *)

(* val problemsListLenNo = start 3 false *)
(* val _ = print "......problemsListLenNo.........\n" *)
(* val _ = print (LU.toStringCommaSquare *)
(*   		   (fn (l,r) => "\n\"" ^ (IOMstrs.prTerm l) ^ " = " ^ (IOMstrs.prTerm r) ^ "\"") *)
(*   		   problemsListLenNo) *)


(* val problemsNatMaxMinYes = start 4 true *)
(* val _ = print "......problemsNatMaxMinYes.........\n" *)
(* val _ = print (LU.toStringCommaSquare *)
(*   		   (fn (l,r) => "\n\"" ^ (IOMstrs.prTerm l) ^ " = " ^ (IOMstrs.prTerm r) ^ "\"") *)
(*   		   problemsNatMaxMinYes) *)

(* val problemsNatMaxYes = start 5 true *)
(* val _ = print "......problemsNatMaxYes.........\n" *)
(* val _ = print (LU.toStringCommaSquare *)
(*   		   (fn (l,r) => "\n\"" ^ (IOMstrs.prTerm l) ^ " = " ^ (IOMstrs.prTerm r) ^ "\"") *)
(*   		   problemsNatMaxYes) *)

(* val problemsNatMinYes = start 6 true *)
(* val _ = print "......problemsNatMinYes.........\n" *)
(* val _ = print (LU.toStringCommaSquare *)
(*   		   (fn (l,r) => "\n\"" ^ (IOMstrs.prTerm l) ^ " = " ^ (IOMstrs.prTerm r) ^ "\"") *)
(*   		   problemsNatMinYes) *)

(* val problemsNatMaxMinNo = start 4 false *)
(* val _ = print "......problemsNatMaxMinNo.........\n" *)
(* val _ = print (LU.toStringCommaSquare  *)
(*   		   (fn (l,r) => "\n\"" ^ (IOMstrs.prTerm l) ^ " = " ^ (IOMstrs.prTerm r) ^ "\"")  *)
(*   		   problemsNatMaxMinNo)  *)

(* val problemsNatMaxNo = start 5 false *)
(* val _ = print "......problemsNatMaxNo.........\n" *)
(* val _ = print (LU.toStringCommaSquare *)
(*   		   (fn (l,r) => "\n\"" ^ (IOMstrs.prTerm l) ^ " = " ^ (IOMstrs.prTerm r) ^ "\"") *)
(*   		   problemsNatMaxNo) *)

(* val problemsNatMinNo = start 6 false *)
(* val _ = print "......problemsNatMinNo.........\n" *)
(* val _ = print (LU.toStringCommaSquare *)
(*   		   (fn (l,r) => "\n\"" ^ (IOMstrs.prTerm l) ^ " = " ^ (IOMstrs.prTerm r) ^ "\"") *)
(*   		   problemsNatMinNo) *)

(* val problemsMaxMinPlusNo = start 7 false *)
(* val _ = print "......problemsMaxMinPlusNo.........\n" *)
(* val _ = print (LU.toStringCommaSquare *)
(*   		   (fn (l,r) => "\n\"" ^ (IOMstrs.prTerm l) ^ " = " ^ (IOMstrs.prTerm r) ^ "\"") *)
(*   		   problemsMaxMinPlusNo) *)


fun toStringSpike t =
    if Term.isVar t
    then Var.name (valOf (Term.varRootOfTerm t))
    else let val (f,ts,ty) = valOf (Term.decFunTerm t)
	 in if null ts
	    then Fun.toString f
	    else if Fun.equal (Fun.fromString "plus",f)
	    then "(" ^ toStringSpike (hd ts) ^ " + " ^ toStringSpike (hd (tl ts)) ^  ")"
	    else if Fun.equal (Fun.fromString "times",f)
	    then "(" ^ toStringSpike (hd ts) ^ " * " ^ toStringSpike (hd (tl ts)) ^  ")"
	    else (Fun.toString f) ^ (ListUtil.toStringCommaRound toStringSpike ts)
	 end

fun toStringSail t =
    if Term.isVar t
    then String.implode (List.map Char.toUpper (String.explode (Var.name (valOf (Term.varRootOfTerm t)))))
    else let val (f,ts,ty) = valOf (Term.decFunTerm t)
	 in if Fun.equal (Fun.fromString "0",f)
	    then "zero" 
	    else (Fun.toString f) ^ (LU.toStringCommaRound toStringSail ts)
	 end

fun toStringSpikeOrg t =
    if Term.isVar t
    then Var.name (valOf (Term.varRootOfTerm t))
    else let val (f,ts,ty) = valOf (Term.decFunTerm t)
	 in if null ts
	    then Fun.toString f
	    else if Fun.equal (Fun.fromString "s",f)
	    then "S" ^ (LU.toStringCommaRound toStringSpikeOrg ts)
	    else (Fun.toString f) ^ (ListUtil.toStringCommaRound toStringSpikeOrg ts)
	 end

val problemsPlusWithoutZeroYes =
IOMstrs.rdEqs [
"plus(?x:Nat,?z:Nat) = plus(?z:Nat,?x:Nat)",
"s(plus(?x:Nat,?y:Nat)) = plus(s(?x:Nat),?y:Nat)",
"s(plus(?x:Nat,?z:Nat)) = s(plus(?z:Nat,?x:Nat))",
"s(plus(?y:Nat,?x:Nat)) = plus(s(?x:Nat),?y:Nat)",
"plus(?x:Nat,s(?x:Nat)) = s(plus(?x:Nat,?x:Nat))",
"plus(?x:Nat,s(?y:Nat)) = plus(s(?x:Nat),?y:Nat)",
"plus(?x:Nat,s(?z:Nat)) = s(plus(?z:Nat,?x:Nat))",
"plus(?x:Nat,s(?z:Nat)) = plus(?z:Nat,s(?x:Nat))",
"plus(?y:Nat,s(?y:Nat)) = plus(s(?y:Nat),?y:Nat)",
"plus(?z:Nat,s(?x:Nat)) = s(plus(?z:Nat,?x:Nat))",
"plus(s(?x:Nat),?y:Nat) = plus(?y:Nat,s(?x:Nat))",
"plus(s(?x:Nat),?y:Nat) = plus(s(?y:Nat),?x:Nat)",
"plus(s(?y:Nat),?y:Nat) = s(plus(?y:Nat,?y:Nat))",
"s(s(plus(?x:Nat,?x:Nat))) = s(plus(s(?x:Nat),?x:Nat))",
"s(s(plus(?x:Nat,?z:Nat))) = s(s(plus(?z:Nat,?x:Nat)))",
"s(s(plus(?y:Nat,?x:Nat))) = s(plus(s(?y:Nat),?x:Nat))",
"s(s(plus(?y:Nat,?x:Nat))) = plus(s(s(?x:Nat)),?y:Nat)",
"s(s(plus(?y:Nat,?y:Nat))) = plus(?y:Nat,s(s(?y:Nat)))",
"s(plus(?x:Nat,s(?z:Nat))) = plus(?x:Nat,s(s(?z:Nat)))",
"s(plus(?y:Nat,s(?x:Nat))) = s(s(plus(?y:Nat,?x:Nat)))",
"s(plus(?y:Nat,s(?x:Nat))) = plus(?x:Nat,s(s(?y:Nat)))",
"s(plus(?y:Nat,s(?x:Nat))) = plus(s(s(?y:Nat)),?x:Nat)",
"s(plus(?y:Nat,s(?y:Nat))) = s(s(plus(?y:Nat,?y:Nat)))",
"s(plus(s(?x:Nat),?x:Nat)) = plus(?x:Nat,s(s(?x:Nat)))",
"s(plus(s(?x:Nat),?y:Nat)) = plus(s(?x:Nat),s(?y:Nat))",
"s(plus(s(?y:Nat),?z:Nat)) = s(s(plus(?z:Nat,?y:Nat)))",
"s(plus(s(?y:Nat),?z:Nat)) = s(plus(?z:Nat,s(?y:Nat)))",
"s(plus(s(?z:Nat),?y:Nat)) = s(plus(?z:Nat,s(?y:Nat)))",
"plus(?y:Nat,s(s(?x:Nat))) = s(s(plus(?x:Nat,?y:Nat)))",
"plus(?y:Nat,s(s(?x:Nat))) = plus(?x:Nat,s(s(?y:Nat)))",
"plus(?y:Nat,plus(?x:Nat,?z:Nat)) = plus(?z:Nat,plus(?x:Nat,?y:Nat))",
"plus(?y:Nat,plus(?z:Nat,?z:Nat)) = plus(?z:Nat,plus(?y:Nat,?z:Nat))",
"plus(?z:Nat,s(s(?x:Nat))) = s(s(plus(?z:Nat,?x:Nat)))",
"plus(?z:Nat,plus(?x:Nat,?y:Nat)) = plus(plus(?y:Nat,?x:Nat),?z:Nat)",
"plus(s(?y:Nat),s(?z:Nat)) = plus(s(?z:Nat),s(?y:Nat))",
"plus(s(?z:Nat),s(?x:Nat)) = plus(?z:Nat,s(s(?x:Nat)))",
"plus(s(s(?x:Nat)),?z:Nat) = s(plus(?z:Nat,s(?x:Nat)))",
"plus(s(s(?y:Nat)),?z:Nat) = s(plus(s(?z:Nat),?y:Nat))",
"plus(plus(?x:Nat,?z:Nat),?x:Nat) = plus(?x:Nat,plus(?x:Nat,?z:Nat))",
"plus(plus(?y:Nat,?x:Nat),?x:Nat) = plus(?y:Nat,plus(?x:Nat,?x:Nat))",
"plus(plus(?y:Nat,?z:Nat),?x:Nat) = plus(plus(?z:Nat,?x:Nat),?y:Nat)",
"s(plus(?x:Nat,plus(?y:Nat,?z:Nat))) = plus(?y:Nat,plus(s(?z:Nat),?x:Nat))",
"s(plus(?y:Nat,plus(?x:Nat,?z:Nat))) = plus(s(plus(?x:Nat,?y:Nat)),?z:Nat)",
"s(plus(?y:Nat,plus(?z:Nat,?x:Nat))) = plus(?z:Nat,plus(s(?y:Nat),?x:Nat))",
"s(plus(?z:Nat,plus(?y:Nat,?z:Nat))) = plus(?y:Nat,plus(s(?z:Nat),?z:Nat))",
"s(plus(plus(?x:Nat,?x:Nat),?y:Nat)) = s(plus(?x:Nat,plus(?x:Nat,?y:Nat)))",
"s(plus(plus(?x:Nat,?y:Nat),?x:Nat)) = plus(?x:Nat,s(plus(?y:Nat,?x:Nat)))",
"s(plus(plus(?x:Nat,?y:Nat),?z:Nat)) = plus(s(plus(?y:Nat,?z:Nat)),?x:Nat)",
"s(plus(plus(?x:Nat,?z:Nat),?x:Nat)) = plus(s(?x:Nat),plus(?z:Nat,?x:Nat))",
"s(plus(plus(?x:Nat,?z:Nat),?y:Nat)) = plus(?z:Nat,plus(?x:Nat,s(?y:Nat)))",
"s(plus(plus(?y:Nat,?x:Nat),?x:Nat)) = plus(s(plus(?x:Nat,?y:Nat)),?x:Nat)",
"s(plus(plus(?y:Nat,?z:Nat),?z:Nat)) = s(plus(?z:Nat,plus(?y:Nat,?z:Nat)))",
"s(plus(plus(?z:Nat,?y:Nat),?x:Nat)) = plus(?z:Nat,plus(s(?x:Nat),?y:Nat))",
"plus(?x:Nat,s(plus(?x:Nat,?y:Nat))) = s(plus(?y:Nat,plus(?x:Nat,?x:Nat)))",
"plus(?x:Nat,s(plus(?x:Nat,?y:Nat))) = plus(s(?x:Nat),plus(?y:Nat,?x:Nat))",
"plus(?x:Nat,plus(?x:Nat,s(?z:Nat))) = plus(?x:Nat,s(plus(?z:Nat,?x:Nat)))",
"plus(?x:Nat,plus(s(?y:Nat),?x:Nat)) = s(plus(?y:Nat,plus(?x:Nat,?x:Nat)))",
"plus(?x:Nat,plus(s(?y:Nat),?y:Nat)) = plus(plus(?y:Nat,?y:Nat),s(?x:Nat))",
"plus(?y:Nat,plus(s(?y:Nat),?x:Nat)) = plus(s(?y:Nat),plus(?y:Nat,?x:Nat))",
"plus(?y:Nat,plus(s(?z:Nat),?y:Nat)) = plus(?z:Nat,plus(s(?y:Nat),?y:Nat))",
"plus(?z:Nat,s(plus(?x:Nat,?z:Nat))) = plus(s(plus(?x:Nat,?z:Nat)),?z:Nat)",
"plus(?z:Nat,plus(s(?y:Nat),?y:Nat)) = plus(s(plus(?z:Nat,?y:Nat)),?y:Nat)",
"plus(s(?x:Nat),plus(?x:Nat,?x:Nat)) = plus(?x:Nat,s(plus(?x:Nat,?x:Nat)))",
"plus(s(?z:Nat),plus(?y:Nat,?x:Nat)) = plus(s(plus(?y:Nat,?z:Nat)),?x:Nat)",
"plus(s(s(?y:Nat)),s(?y:Nat)) = plus(s(?y:Nat),s(s(?y:Nat)))",
"plus(s(plus(?x:Nat,?y:Nat)),?y:Nat) = plus(plus(s(?x:Nat),?y:Nat),?y:Nat)",
"plus(s(plus(?y:Nat,?x:Nat)),?x:Nat) = plus(s(plus(?x:Nat,?y:Nat)),?x:Nat)",
"plus(s(plus(?y:Nat,?x:Nat)),?z:Nat) = plus(?z:Nat,plus(s(?y:Nat),?x:Nat))",
"plus(s(plus(?y:Nat,?y:Nat)),?x:Nat) = plus(s(?x:Nat),plus(?y:Nat,?y:Nat))",
"plus(s(plus(?z:Nat,?x:Nat)),?y:Nat) = plus(?z:Nat,plus(?x:Nat,s(?y:Nat)))",
"plus(s(plus(?z:Nat,?z:Nat)),?x:Nat) = plus(?z:Nat,plus(?x:Nat,s(?z:Nat)))",
"plus(plus(?x:Nat,?x:Nat),s(?y:Nat)) = plus(?y:Nat,s(plus(?x:Nat,?x:Nat)))",
"plus(plus(?x:Nat,?y:Nat),s(?y:Nat)) = plus(plus(s(?y:Nat),?x:Nat),?y:Nat)",
"plus(plus(?x:Nat,?y:Nat),s(?z:Nat)) = plus(s(plus(?x:Nat,?y:Nat)),?z:Nat)",
"plus(plus(?x:Nat,s(?y:Nat)),?x:Nat) = plus(s(plus(?x:Nat,?x:Nat)),?y:Nat)",
"plus(plus(?y:Nat,s(?x:Nat)),?x:Nat) = plus(?x:Nat,plus(?x:Nat,s(?y:Nat)))",
"plus(plus(?y:Nat,s(?x:Nat)),?y:Nat) = plus(?x:Nat,plus(s(?y:Nat),?y:Nat))",
"plus(plus(s(?x:Nat),?x:Nat),?y:Nat) = plus(?y:Nat,plus(s(?x:Nat),?x:Nat))",
"plus(plus(s(?x:Nat),?y:Nat),?y:Nat) = plus(?y:Nat,plus(?x:Nat,s(?y:Nat)))",
"plus(plus(s(?x:Nat),?y:Nat),?z:Nat) = s(plus(plus(?x:Nat,?y:Nat),?z:Nat))",
"plus(plus(s(?y:Nat),?z:Nat),?y:Nat) = plus(s(?y:Nat),plus(?y:Nat,?z:Nat))",
"s(plus(plus(?x:Nat,?x:Nat),s(?x:Nat))) = plus(plus(s(?x:Nat),s(?x:Nat)),?x:Nat)",
"s(plus(plus(?x:Nat,?x:Nat),s(?z:Nat))) = s(plus(s(?z:Nat),plus(?x:Nat,?x:Nat)))",
"s(plus(plus(?x:Nat,?y:Nat),s(?z:Nat))) = s(plus(s(?x:Nat),plus(?y:Nat,?z:Nat)))",
"plus(?x:Nat,plus(?x:Nat,plus(?y:Nat,?z:Nat))) = plus(?y:Nat,plus(plus(?z:Nat,?x:Nat),?x:Nat))",
"plus(?x:Nat,plus(s(?x:Nat),s(?x:Nat))) = plus(s(plus(?x:Nat,?x:Nat)),s(?x:Nat))",
"plus(?x:Nat,plus(plus(?z:Nat,?x:Nat),?z:Nat)) = plus(plus(?z:Nat,plus(?x:Nat,?x:Nat)),?z:Nat)",
"plus(s(?x:Nat),plus(s(?x:Nat),?y:Nat)) = s(plus(plus(?x:Nat,?y:Nat),s(?x:Nat)))",
"plus(s(?z:Nat),s(plus(?x:Nat,?x:Nat))) = plus(?z:Nat,plus(s(?x:Nat),s(?x:Nat)))",
"plus(plus(?x:Nat,plus(?y:Nat,?x:Nat)),?y:Nat) = plus(?x:Nat,plus(?x:Nat,plus(?y:Nat,?y:Nat)))",
"plus(plus(?x:Nat,plus(?y:Nat,?x:Nat)),?z:Nat) = plus(?x:Nat,plus(?x:Nat,plus(?z:Nat,?y:Nat)))",
"plus(plus(plus(?y:Nat,?x:Nat),?x:Nat),?z:Nat) = plus(plus(plus(?x:Nat,?y:Nat),?z:Nat),?x:Nat)",
"s(plus(plus(?x:Nat,?x:Nat),plus(?y:Nat,?x:Nat))) = plus(plus(?x:Nat,?x:Nat),s(plus(?y:Nat,?x:Nat)))",
"s(plus(plus(?z:Nat,?z:Nat),plus(?z:Nat,?x:Nat))) = plus(plus(plus(?z:Nat,?z:Nat),s(?x:Nat)),?z:Nat)",
"plus(?y:Nat,plus(s(?z:Nat),plus(?x:Nat,?x:Nat))) = plus(plus(?z:Nat,plus(?x:Nat,?x:Nat)),s(?y:Nat))",
"plus(s(?x:Nat),plus(plus(?y:Nat,?x:Nat),?z:Nat)) = plus(s(?y:Nat),plus(plus(?x:Nat,?x:Nat),?z:Nat))",
"plus(plus(?x:Nat,?x:Nat),plus(?y:Nat,s(?y:Nat))) = plus(plus(s(?x:Nat),plus(?x:Nat,?y:Nat)),?y:Nat)",
"plus(plus(?x:Nat,?y:Nat),s(plus(?x:Nat,?x:Nat))) = plus(plus(s(?y:Nat),plus(?x:Nat,?x:Nat)),?x:Nat)",
"plus(plus(s(?y:Nat),?x:Nat),plus(?y:Nat,?x:Nat)) = plus(plus(?x:Nat,plus(?y:Nat,?x:Nat)),s(?y:Nat))",
"plus(plus(plus(?y:Nat,?z:Nat),s(?x:Nat)),?x:Nat) = plus(s(plus(?y:Nat,?x:Nat)),plus(?x:Nat,?z:Nat))",
"plus(plus(plus(?y:Nat,?x:Nat),?z:Nat),plus(?x:Nat,?y:Nat)) = plus(plus(plus(?x:Nat,?y:Nat),plus(?z:Nat,?y:Nat)),?x:Nat)"
]

val problemsTimesWithZeroYes =
IOMstrs.rdEqs [
"times(?x:Nat,0) = times(0,?x:Nat)",
"times(?x:Nat,0) = times(0,0)",
"times(?y:Nat,0) = times(?z:Nat,0)",
"times(?y:Nat,0) = times(0,?x:Nat)",
"times(0,?y:Nat) = times(0,?z:Nat)",
"times(0,0) = times(0,?y:Nat)",
"s(?x:Nat) = times(s(0),s(?x:Nat))",
"s(?y:Nat) = times(s(?y:Nat),s(0))",
"plus(?x:Nat,0) = times(?x:Nat,s(0))",
"times(?x:Nat,s(0)) = plus(0,?x:Nat)",
"times(s(?x:Nat),0) = times(0,?y:Nat)",
"times(s(0),?y:Nat) = plus(0,?y:Nat)",
"s(?x:Nat) = plus(?x:Nat,s(times(0,?y:Nat)))",
"s(?x:Nat) = plus(s(?x:Nat),times(0,?y:Nat))",
"s(?x:Nat) = plus(times(0,?z:Nat),s(?x:Nat))",
"s(?y:Nat) = plus(s(times(?y:Nat,0)),?y:Nat)",
"s(?z:Nat) = plus(s(times(?x:Nat,0)),?z:Nat)",
"plus(0,0) = times(s(s(?x:Nat)),0)",
"times(?x:Nat,?x:Nat) = times(plus(0,?x:Nat),?x:Nat)",
"times(?x:Nat,?y:Nat) = times(plus(0,?y:Nat),?x:Nat)",
"times(?x:Nat,0) = times(0,plus(0,?z:Nat))",
"times(?y:Nat,?x:Nat) = times(?x:Nat,plus(?y:Nat,0))",
"times(?z:Nat,?x:Nat) = plus(0,times(?x:Nat,?z:Nat))",
"times(?z:Nat,?y:Nat) = times(?z:Nat,plus(0,?y:Nat))",
"plus(?y:Nat,times(0,0)) = plus(0,?y:Nat)",
"plus(0,times(?y:Nat,?z:Nat)) = times(?y:Nat,?z:Nat)",
"plus(times(?x:Nat,?y:Nat),0) = times(?x:Nat,?y:Nat)",
"plus(times(?y:Nat,?x:Nat),0) = times(?x:Nat,?y:Nat)",
"times(?x:Nat,plus(?x:Nat,0)) = times(?x:Nat,?x:Nat)",
"times(?y:Nat,times(?x:Nat,0)) = times(0,?x:Nat)",
"times(0,plus(?x:Nat,0)) = times(?x:Nat,0)",
"times(0,times(?y:Nat,?x:Nat)) = times(?x:Nat,0)",
"times(0,times(?z:Nat,0)) = times(?x:Nat,0)",
"times(s(s(?y:Nat)),0) = times(?z:Nat,0)",
"times(plus(?x:Nat,0),?x:Nat) = times(?x:Nat,?x:Nat)",
"times(plus(?x:Nat,0),?y:Nat) = times(?x:Nat,?y:Nat)",
"times(times(?x:Nat,0),?x:Nat) = times(0,0)",
"plus(?x:Nat,s(times(0,?x:Nat))) = s(?x:Nat)",
"plus(?y:Nat,s(times(?z:Nat,0))) = s(?y:Nat)",
"plus(s(?y:Nat),times(?x:Nat,0)) = s(?y:Nat)",
"plus(s(times(0,?x:Nat)),0) = s(0)",
"s(?x:Nat) = plus(0,times(s(?x:Nat),s(0)))",
"plus(?x:Nat,0) = plus(times(s(?x:Nat),0),?x:Nat)",
"times(?x:Nat,?z:Nat) = times(s(0),times(?z:Nat,?x:Nat))",
"times(?x:Nat,0) = times(plus(s(?x:Nat),0),0)",
"times(0,?x:Nat) = times(0,plus(0,s(?z:Nat)))",
"times(0,?x:Nat) = times(times(s(?z:Nat),0),?x:Nat)",
"s(times(0,s(?z:Nat))) = plus(0,s(0))",
"times(plus(?x:Nat,?x:Nat),0) = times(s(?x:Nat),0)",
"plus(0,times(0,s(?x:Nat))) = times(?x:Nat,0)",
"plus(times(?x:Nat,s(0)),0) = plus(0,?x:Nat)",
"plus(times(s(0),?x:Nat),?y:Nat) = plus(?y:Nat,?x:Nat)",
"times(?x:Nat,times(s(?y:Nat),0)) = times(?z:Nat,0)",
"times(0,s(plus(0,?x:Nat))) = times(?x:Nat,0)",
"times(0,plus(?y:Nat,s(?x:Nat))) = times(?x:Nat,0)",
"times(plus(?x:Nat,s(?x:Nat)),0) = times(0,?y:Nat)",
"times(plus(?x:Nat,s(0)),0) = times(0,?x:Nat)",
"times(plus(s(?z:Nat),?z:Nat),0) = times(0,0)",
"times(times(0,s(?z:Nat)),0) = times(0,?y:Nat)",
"times(times(s(?y:Nat),0),?y:Nat) = times(?y:Nat,0)",
"plus(s(times(s(0),?x:Nat)),0) = s(?x:Nat)",
"plus(times(s(0),s(?y:Nat)),0) = s(?y:Nat)",
"s(?x:Nat) = plus(times(?z:Nat,0),plus(s(?x:Nat),0))",
"s(?y:Nat) = plus(s(?y:Nat),times(0,plus(?x:Nat,?x:Nat)))",
"s(?y:Nat) = plus(plus(s(0),times(0,?x:Nat)),?y:Nat)",
"plus(0,?x:Nat) = plus(times(plus(?x:Nat,?z:Nat),0),?x:Nat)",
"plus(0,0) = times(plus(0,0),plus(?y:Nat,?z:Nat))",
"times(?y:Nat,?x:Nat) = times(plus(?y:Nat,times(?y:Nat,0)),?x:Nat)",
"times(?y:Nat,0) = times(s(?x:Nat),times(0,s(0)))",
"times(?y:Nat,0) = times(plus(plus(?y:Nat,?z:Nat),0),0)",
"times(0,?x:Nat) = times(plus(plus(?z:Nat,?y:Nat),?x:Nat),0)",
"plus(0,s(?x:Nat)) = plus(s(times(0,?y:Nat)),?x:Nat)",
"plus(s(?x:Nat),0) = plus(s(times(0,?y:Nat)),?x:Nat)",
"times(?x:Nat,s(?x:Nat)) = times(?x:Nat,plus(s(0),?x:Nat))",
"times(s(0),0) = times(s(times(?y:Nat,?z:Nat)),0)",
"times(0,plus(?y:Nat,0)) = times(?y:Nat,times(0,?x:Nat))",
"times(s(s(?x:Nat)),0) = times(times(?x:Nat,0),?x:Nat)",
"times(plus(?x:Nat,?x:Nat),0) = times(0,times(?y:Nat,?x:Nat))",
"times(times(?x:Nat,?x:Nat),0) = times(times(?z:Nat,?y:Nat),0)",
"times(times(?y:Nat,?x:Nat),0) = times(plus(0,0),0)",
"times(0,plus(s(?x:Nat),0)) = times(s(?x:Nat),0)",
"times(s(?x:Nat),plus(?x:Nat,0)) = times(s(?x:Nat),?x:Nat)",
"plus(?x:Nat,plus(?x:Nat,times(0,?x:Nat))) = plus(?x:Nat,?x:Nat)",
"plus(?x:Nat,plus(times(0,?z:Nat),?y:Nat)) = plus(?y:Nat,?x:Nat)",
"plus(?z:Nat,times(times(?x:Nat,?y:Nat),0)) = plus(?z:Nat,0)",
"plus(plus(times(?z:Nat,0),?x:Nat),?y:Nat) = plus(?x:Nat,?y:Nat)",
"plus(plus(times(0,?x:Nat),?z:Nat),?x:Nat) = plus(?x:Nat,?z:Nat)",
"plus(times(plus(?z:Nat,?z:Nat),0),0) = times(0,?x:Nat)",
"times(?x:Nat,plus(?y:Nat,times(?y:Nat,0))) = times(?y:Nat,?x:Nat)",
"times(?x:Nat,times(times(?x:Nat,?x:Nat),0)) = times(?z:Nat,0)",
"times(0,times(?y:Nat,times(?y:Nat,?z:Nat))) = times(?z:Nat,0)",
"times(0,times(0,plus(?x:Nat,?z:Nat))) = times(0,?x:Nat)",
"times(0,times(0,times(?y:Nat,?y:Nat))) = times(?x:Nat,0)",
"times(plus(?x:Nat,plus(0,?x:Nat)),0) = plus(0,0)",
"times(plus(times(?y:Nat,?x:Nat),0),0) = times(0,?z:Nat)",
"plus(times(0,plus(?x:Nat,?x:Nat)),s(?x:Nat)) = s(?x:Nat)",
"plus(times(plus(?y:Nat,?y:Nat),0),s(?x:Nat)) = s(?x:Nat)",
"s(?z:Nat) = plus(s(?z:Nat),times(times(?x:Nat,0),s(?x:Nat)))",
"plus(0,?y:Nat) = times(plus(0,plus(0,s(0))),?y:Nat)",
"times(0,?x:Nat) = times(s(?z:Nat),plus(times(0,?x:Nat),0))",
"times(0,?y:Nat) = times(times(?y:Nat,plus(0,0)),s(?z:Nat))",
"times(0,0) = times(0,plus(plus(?y:Nat,s(?y:Nat)),?x:Nat))",
"times(?y:Nat,s(?x:Nat)) = plus(times(?x:Nat,plus(0,?y:Nat)),?y:Nat)",
"plus(?x:Nat,times(?z:Nat,?x:Nat)) = times(s(plus(0,?z:Nat)),?x:Nat)",
"times(plus(?y:Nat,?y:Nat),0) = times(0,s(plus(?y:Nat,?z:Nat)))",
"times(0,plus(0,s(?x:Nat))) = times(?x:Nat,times(0,?z:Nat))",
"times(s(0),times(?x:Nat,?x:Nat)) = times(plus(?x:Nat,0),?x:Nat)",
"times(plus(?y:Nat,s(0)),0) = times(0,s(s(?y:Nat)))",
"plus(?x:Nat,s(times(s(0),?z:Nat))) = plus(?x:Nat,s(?z:Nat))",
"plus(times(?z:Nat,?x:Nat),times(s(?y:Nat),0)) = times(?x:Nat,?z:Nat)",
"times(s(times(s(s(?z:Nat)),0)),0) = times(0,?y:Nat)",
"times(times(s(?x:Nat),?x:Nat),times(0,?z:Nat)) = times(0,?z:Nat)",
"plus(s(times(times(0,?x:Nat),s(?x:Nat))),?x:Nat) = s(?x:Nat)",
"s(s(?x:Nat)) = plus(s(s(?x:Nat)),times(times(0,?z:Nat),?y:Nat))",
"plus(?y:Nat,?x:Nat) = plus(plus(0,times(0,?y:Nat)),plus(?y:Nat,?x:Nat))",
"times(?z:Nat,0) = times(times(times(?z:Nat,?y:Nat),times(?x:Nat,?x:Nat)),0)",
"times(0,?x:Nat) = times(plus(s(s(times(0,0))),?x:Nat),0)",
"times(0,?z:Nat) = times(times(0,s(s(plus(?z:Nat,0)))),?x:Nat)",
"s(plus(?y:Nat,?x:Nat)) = plus(plus(times(0,?x:Nat),?x:Nat),s(?y:Nat))",
"times(?x:Nat,s(?x:Nat)) = plus(?x:Nat,times(times(?x:Nat,s(0)),?x:Nat))",
"times(s(0),s(?x:Nat)) = s(plus(times(s(0),?x:Nat),0))",
"times(times(0,0),?x:Nat) = times(plus(?y:Nat,times(?x:Nat,0)),0)",
"plus(?y:Nat,times(plus(?x:Nat,?x:Nat),0)) = plus(times(0,?z:Nat),?y:Nat)",
"plus(0,times(times(?x:Nat,?y:Nat),?y:Nat)) = times(?x:Nat,times(?y:Nat,?y:Nat))",
"times(0,plus(plus(?z:Nat,?x:Nat),?x:Nat)) = times(0,plus(?y:Nat,?x:Nat))",
"plus(plus(?x:Nat,s(times(?z:Nat,0))),?x:Nat) = plus(s(?x:Nat),?x:Nat)",
"plus(?z:Nat,s(s(times(0,times(?x:Nat,?x:Nat))))) = s(s(?z:Nat))",
"plus(0,times(?y:Nat,plus(0,times(0,?y:Nat)))) = times(?y:Nat,0)",
"times(?x:Nat,s(s(times(0,times(?x:Nat,?x:Nat))))) = plus(?x:Nat,?x:Nat)",
"times(?y:Nat,times(0,times(s(s(?x:Nat)),?y:Nat))) = times(?x:Nat,0)",
"times(s(plus(?y:Nat,plus(0,s(?z:Nat)))),0) = times(0,?y:Nat)",
"times(plus(plus(0,?z:Nat),times(0,?y:Nat)),?x:Nat) = times(?z:Nat,?x:Nat)",
"times(times(0,s(times(0,s(0)))),?x:Nat) = times(?x:Nat,0)",
"plus(plus(s(0),?y:Nat),times(plus(?x:Nat,?y:Nat),0)) = s(?y:Nat)",
"s(0) = plus(0,s(times(s(plus(0,?x:Nat)),plus(0,0))))",
"plus(?x:Nat,?y:Nat) = times(plus(plus(?x:Nat,0),?y:Nat),s(times(?x:Nat,0)))",
"times(0,?x:Nat) = times(times(?z:Nat,plus(plus(s(?x:Nat),?z:Nat),?x:Nat)),0)",
"times(0,0) = times(?z:Nat,times(0,plus(plus(?z:Nat,0),s(?y:Nat))))",
"s(plus(0,?x:Nat)) = times(s(times(0,?y:Nat)),plus(s(0),?x:Nat))",
"times(s(?x:Nat),0) = times(plus(plus(?x:Nat,?z:Nat),times(?x:Nat,?y:Nat)),0)",
"times(s(0),0) = plus(0,times(?x:Nat,times(times(?z:Nat,0),?x:Nat)))",
"times(0,times(?x:Nat,?z:Nat)) = times(times(?y:Nat,0),times(0,s(?y:Nat)))",
"times(0,s(plus(?x:Nat,?y:Nat))) = times(plus(?x:Nat,plus(?y:Nat,?x:Nat)),0)",
"times(s(plus(?y:Nat,?z:Nat)),0) = times(times(0,?z:Nat),plus(?x:Nat,?x:Nat))",
"times(plus(?x:Nat,s(?x:Nat)),0) = times(plus(plus(?x:Nat,?x:Nat),?x:Nat),0)",
"plus(times(plus(?x:Nat,?y:Nat),s(0)),?x:Nat) = plus(?x:Nat,plus(?x:Nat,?y:Nat))",
"times(s(?y:Nat),times(times(?x:Nat,0),?x:Nat)) = times(0,times(?z:Nat,?y:Nat))",
"times(0,times(times(s(s(?x:Nat)),0),?x:Nat)) = times(0,s(?x:Nat))",
"plus(plus(s(times(s(0),plus(0,?x:Nat))),0),0) = s(?x:Nat)",
"s(?z:Nat) = plus(times(0,times(?y:Nat,?z:Nat)),plus(s(0),plus(0,?z:Nat)))",
"times(?x:Nat,0) = times(plus(plus(?x:Nat,?z:Nat),plus(?y:Nat,?x:Nat)),times(?x:Nat,0))",
"times(?x:Nat,0) = times(times(?y:Nat,?x:Nat),times(s(plus(?x:Nat,s(0))),0))",
"times(0,?y:Nat) = times(times(plus(?x:Nat,?x:Nat),0),plus(?z:Nat,plus(0,0)))",
"times(0,s(0)) = times(0,plus(times(plus(s(?x:Nat),0),?z:Nat),?z:Nat))",
"times(0,plus(?x:Nat,0)) = times(times(plus(?z:Nat,?y:Nat),times(?y:Nat,0)),?y:Nat)",
"times(?z:Nat,times(0,plus(0,0))) = times(times(0,plus(?z:Nat,?x:Nat)),?x:Nat)",
"times(plus(0,0),s(s(?x:Nat))) = times(0,plus(?y:Nat,plus(?x:Nat,?x:Nat)))",
"times(0,plus(s(?y:Nat),plus(?x:Nat,?z:Nat))) = times(s(plus(?z:Nat,0)),0)",
"plus(plus(?x:Nat,times(?x:Nat,0)),plus(?z:Nat,?y:Nat)) = plus(plus(?y:Nat,?z:Nat),?x:Nat)",
"times(times(times(0,times(0,0)),?z:Nat),?y:Nat) = times(?y:Nat,plus(0,0))",
"plus(s(plus(?y:Nat,?y:Nat)),plus(0,times(?x:Nat,0))) = plus(s(?y:Nat),?y:Nat)",
"times(0,plus(?z:Nat,plus(?y:Nat,s(plus(0,?y:Nat))))) = times(s(?y:Nat),0)",
"times(times(0,times(?x:Nat,?y:Nat)),plus(?z:Nat,s(?x:Nat))) = times(0,s(?z:Nat))",
"times(s(plus(?z:Nat,s(times(?z:Nat,?x:Nat)))),times(?x:Nat,0)) = times(0,?z:Nat)",
"times(?x:Nat,s(?y:Nat)) = plus(plus(plus(0,?x:Nat),0),plus(times(?y:Nat,?x:Nat),0))",
"times(0,s(plus(0,?x:Nat))) = times(plus(times(?y:Nat,?y:Nat),plus(?x:Nat,?z:Nat)),0)",
"times(0,s(times(0,?x:Nat))) = times(plus(plus(?x:Nat,?y:Nat),times(?z:Nat,?x:Nat)),0)",
"times(plus(0,s(?x:Nat)),0) = times(0,times(plus(?x:Nat,0),plus(?x:Nat,?x:Nat)))",
"times(0,plus(times(?x:Nat,?y:Nat),s(?z:Nat))) = times(0,plus(plus(?x:Nat,?y:Nat),0))",
"times(times(?x:Nat,0),plus(?y:Nat,plus(?z:Nat,?z:Nat))) = times(0,plus(s(?x:Nat),0))",
"plus(plus(s(?y:Nat),s(?x:Nat)),times(s(?y:Nat),0)) = plus(s(?y:Nat),s(?x:Nat))",
"times(times(plus(?y:Nat,times(?y:Nat,0)),0),s(?x:Nat)) = times(?x:Nat,times(?y:Nat,0))",
"times(0,times(s(times(plus(?z:Nat,?x:Nat),?x:Nat)),s(s(?x:Nat)))) = times(0,?x:Nat)",
"plus(0,0) = times(times(times(?x:Nat,?y:Nat),times(?x:Nat,?x:Nat)),times(0,times(?y:Nat,0)))",
"times(?x:Nat,0) = times(plus(?x:Nat,?y:Nat),times(times(?x:Nat,times(times(0,?y:Nat),?x:Nat)),?x:Nat))",
"times(s(?z:Nat),0) = times(times(times(?x:Nat,0),plus(?x:Nat,?z:Nat)),s(times(?x:Nat,?x:Nat)))",
"times(plus(?y:Nat,?x:Nat),0) = times(0,times(plus(?x:Nat,times(?x:Nat,times(?z:Nat,?x:Nat))),?x:Nat))",
"plus(plus(?y:Nat,?y:Nat),s(?x:Nat)) = plus(plus(?x:Nat,?y:Nat),plus(times(?x:Nat,0),s(?y:Nat)))",
"times(0,s(plus(?y:Nat,?z:Nat))) = times(plus(?y:Nat,s(?x:Nat)),times(plus(?z:Nat,?y:Nat),0))",
"times(plus(?z:Nat,?z:Nat),times(times(0,?x:Nat),?x:Nat)) = times(times(plus(?z:Nat,?y:Nat),?y:Nat),0)",
"times(plus(?y:Nat,plus(?z:Nat,times(s(?y:Nat),?y:Nat))),0) = times(s(times(?x:Nat,?y:Nat)),0)",
"times(times(plus(s(?x:Nat),s(times(?x:Nat,?x:Nat))),s(?y:Nat)),0) = times(0,s(?y:Nat))",
"times(plus(plus(?y:Nat,?y:Nat),times(?x:Nat,?x:Nat)),times(0,plus(?x:Nat,?x:Nat))) = times(0,?y:Nat)",
"s(?x:Nat) = plus(?x:Nat,plus(times(times(?x:Nat,0),0),s(times(times(0,?z:Nat),s(?x:Nat)))))",
"times(?z:Nat,0) = times(plus(s(plus(0,plus(?y:Nat,?x:Nat))),?x:Nat),times(0,times(?y:Nat,0)))",
"times(0,plus(times(?y:Nat,?x:Nat),?z:Nat)) = plus(times(0,s(0)),times(0,plus(?x:Nat,?z:Nat)))",
"times(times(?x:Nat,0),times(s(?z:Nat),times(?y:Nat,0))) = times(?x:Nat,times(times(?y:Nat,0),?x:Nat))",
"times(?x:Nat,0) = times(?y:Nat,times(s(s(times(0,0))),times(0,times(plus(0,?z:Nat),?y:Nat))))",
"times(?y:Nat,0) = times(times(times(times(plus(?x:Nat,?y:Nat),?x:Nat),0),plus(0,times(?z:Nat,?z:Nat))),?y:Nat)",
"times(0,times(?z:Nat,?x:Nat)) = plus(times(?x:Nat,times(times(?x:Nat,times(0,?x:Nat)),?x:Nat)),times(0,0))",
"times(0,times(plus(?z:Nat,s(s(?z:Nat))),?x:Nat)) = times(0,times(plus(?z:Nat,?z:Nat),plus(?x:Nat,0)))",
"plus(plus(times(0,?x:Nat),?z:Nat),plus(?y:Nat,plus(?x:Nat,?z:Nat))) = plus(plus(plus(?z:Nat,?y:Nat),?z:Nat),?x:Nat)",
"times(times(times(plus(s(?x:Nat),s(0)),?y:Nat),0),plus(?x:Nat,s(?x:Nat))) = times(s(0),0)",
"times(0,times(?z:Nat,times(times(plus(?x:Nat,?z:Nat),s(?z:Nat)),plus(?x:Nat,s(?z:Nat))))) = plus(0,0)",
"times(times(times(?z:Nat,plus(0,0)),?x:Nat),times(?x:Nat,s(times(s(?z:Nat),?y:Nat)))) = plus(0,0)",
"plus(times(times(?y:Nat,times(times(?y:Nat,?y:Nat),times(0,?z:Nat))),times(?z:Nat,?z:Nat)),s(?x:Nat)) = s(?x:Nat)",
"times(times(0,?x:Nat),0) = times(times(?y:Nat,times(times(0,times(?x:Nat,?z:Nat)),times(?x:Nat,s(?x:Nat)))),?z:Nat)",
"times(plus(s(?y:Nat),?x:Nat),0) = times(times(0,plus(?y:Nat,?x:Nat)),plus(times(?y:Nat,?z:Nat),plus(?x:Nat,?y:Nat)))",
"times(plus(plus(?x:Nat,s(?x:Nat)),plus(times(times(?y:Nat,?x:Nat),?z:Nat),0)),0) = times(?x:Nat,times(0,?x:Nat))",
"plus(0,?x:Nat) = plus(?x:Nat,times(s(s(?y:Nat)),times(plus(times(0,?z:Nat),0),times(?z:Nat,plus(?z:Nat,?x:Nat)))))",
"times(?y:Nat,0) = times(times(times(s(times(?x:Nat,?z:Nat)),?x:Nat),plus(times(?x:Nat,?y:Nat),?z:Nat)),times(0,s(?y:Nat)))",
"times(times(0,?x:Nat),s(s(?z:Nat))) = times(plus(?x:Nat,plus(?y:Nat,?y:Nat)),times(plus(?x:Nat,?z:Nat),times(0,?y:Nat)))",
"plus(times(plus(?y:Nat,?x:Nat),0),times(?y:Nat,s(0))) = times(s(times(?z:Nat,times(times(?y:Nat,0),?x:Nat))),?y:Nat)",
"times(times(0,times(s(0),plus(times(0,?x:Nat),times(?x:Nat,0)))),0) = times(s(?y:Nat),times(?y:Nat,0))",
"plus(?x:Nat,plus(times(times(0,times(?y:Nat,?x:Nat)),s(?y:Nat)),times(s(times(0,?x:Nat)),0))) = plus(0,?x:Nat)",
"times(times(times(times(plus(?y:Nat,?x:Nat),plus(?x:Nat,0)),plus(?z:Nat,plus(?y:Nat,?x:Nat))),?x:Nat),0) = times(?x:Nat,0)",
"times(?y:Nat,0) = times(plus(?y:Nat,?x:Nat),times(times(s(plus(?x:Nat,?z:Nat)),plus(s(?z:Nat),s(?z:Nat))),times(0,?x:Nat)))",
"times(0,s(times(?x:Nat,s(times(?x:Nat,?z:Nat))))) = times(plus(plus(?x:Nat,s(plus(?y:Nat,0))),?y:Nat),times(0,?x:Nat))",
"times(plus(0,s(0)),times(0,plus(?y:Nat,?x:Nat))) = times(times(?y:Nat,times(?x:Nat,0)),times(plus(0,?x:Nat),?y:Nat))",
"times(times(0,0),times(times(?z:Nat,plus(?z:Nat,?x:Nat)),s(0))) = times(0,times(?x:Nat,times(times(?x:Nat,?y:Nat),?x:Nat)))",
"times(s(plus(plus(times(0,?z:Nat),times(0,?x:Nat)),times(0,plus(?z:Nat,0)))),?x:Nat) = plus(times(0,?x:Nat),?x:Nat)",
"times(plus(?z:Nat,times(times(?x:Nat,times(?z:Nat,?x:Nat)),times(s(?y:Nat),times(?x:Nat,?x:Nat)))),0) = times(times(0,?y:Nat),0)",
"times(s(?y:Nat),0) = times(times(0,s(0)),plus(times(?y:Nat,plus(times(?z:Nat,?x:Nat),plus(?x:Nat,?z:Nat))),times(?y:Nat,?y:Nat)))",
"times(s(plus(?z:Nat,s(plus(?x:Nat,?x:Nat)))),0) = times(plus(plus(plus(plus(?x:Nat,?z:Nat),plus(?x:Nat,?x:Nat)),?z:Nat),?x:Nat),0)",
"times(plus(times(plus(?z:Nat,?x:Nat),plus(times(?x:Nat,?z:Nat),?y:Nat)),s(?z:Nat)),plus(0,0)) = times(0,plus(?x:Nat,s(?z:Nat)))",
"times(0,?x:Nat) = times(times(0,?z:Nat),times(s(?x:Nat),times(plus(?x:Nat,times(?x:Nat,?z:Nat)),plus(plus(?x:Nat,?x:Nat),plus(?z:Nat,?x:Nat)))))",
"times(times(?x:Nat,s(s(?y:Nat))),0) = times(times(s(plus(?z:Nat,?y:Nat)),0),plus(?x:Nat,plus(?z:Nat,times(times(?z:Nat,?z:Nat),?x:Nat))))",
"times(times(0,plus(?x:Nat,plus(times(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat)))),?z:Nat) = times(0,s(times(?z:Nat,times(s(0),s(?x:Nat)))))",
"plus(s(s(0)),times(plus(plus(plus(?x:Nat,0),s(?z:Nat)),0),s(0))) = s(s(s(plus(plus(?z:Nat,0),?x:Nat))))",
"times(plus(times(s(?z:Nat),plus(plus(?x:Nat,?z:Nat),?x:Nat)),plus(?x:Nat,plus(?z:Nat,times(?x:Nat,?y:Nat)))),0) = times(0,times(?x:Nat,?x:Nat))",
"times(0,times(times(times(times(?z:Nat,?x:Nat),s(?z:Nat)),plus(?z:Nat,s(?z:Nat))),s(plus(times(?z:Nat,0),?z:Nat)))) = times(?x:Nat,0)",
"times(0,s(?z:Nat)) = times(s(s(plus(?x:Nat,times(0,?x:Nat)))),times(0,times(s(times(?y:Nat,?x:Nat)),plus(?x:Nat,times(0,?x:Nat)))))",
"plus(plus(s(?x:Nat),times(?x:Nat,?z:Nat)),plus(plus(?y:Nat,0),?z:Nat)) = plus(s(plus(0,plus(plus(?x:Nat,?y:Nat),times(?z:Nat,?x:Nat)))),?z:Nat)",
"times(times(?y:Nat,0),times(times(s(0),plus(times(?z:Nat,0),times(?x:Nat,?y:Nat))),?x:Nat)) = times(times(times(plus(?x:Nat,?x:Nat),?y:Nat),0),?y:Nat)",
"times(plus(0,times(0,plus(s(?x:Nat),plus(0,?z:Nat)))),s(plus(?x:Nat,plus(s(?x:Nat),times(0,?z:Nat))))) = times(?x:Nat,plus(0,0))",
"times(times(?y:Nat,0),times(?x:Nat,plus(times(times(?y:Nat,?z:Nat),times(?y:Nat,?x:Nat)),times(times(?z:Nat,?y:Nat),s(?x:Nat))))) = times(plus(?x:Nat,?y:Nat),0)",
"times(times(plus(s(0),?z:Nat),plus(?x:Nat,?x:Nat)),0) = times(0,s(times(times(plus(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),plus(?x:Nat,plus(?y:Nat,0)))))",
"times(times(?x:Nat,0),times(plus(?x:Nat,?x:Nat),plus(?y:Nat,?z:Nat))) = times(times(times(s(s(?x:Nat)),?x:Nat),s(plus(times(?x:Nat,?y:Nat),s(?y:Nat)))),0)",
"times(plus(s(0),plus(times(s(?x:Nat),s(?x:Nat)),?y:Nat)),plus(times(0,?x:Nat),0)) = times(0,plus(?z:Nat,s(plus(?x:Nat,plus(?y:Nat,?z:Nat)))))",
"times(0,times(times(?z:Nat,plus(plus(?z:Nat,?y:Nat),?x:Nat)),times(?x:Nat,plus(times(?x:Nat,0),?x:Nat)))) = times(0,times(plus(times(?x:Nat,?z:Nat),0),?x:Nat))",
"times(plus(times(?x:Nat,times(?x:Nat,?z:Nat)),s(times(plus(?y:Nat,?x:Nat),?x:Nat))),times(plus(?z:Nat,times(?x:Nat,plus(?z:Nat,?y:Nat))),0)) = times(s(s(s(0))),0)",
"times(times(0,0),s(0)) = times(s(times(plus(plus(0,?x:Nat),plus(?x:Nat,?z:Nat)),times(times(?y:Nat,?z:Nat),?z:Nat))),times(s(plus(plus(?x:Nat,0),0)),0))",
"plus(times(0,times(?y:Nat,plus(?y:Nat,times(?x:Nat,?y:Nat)))),times(0,?x:Nat)) = times(times(?y:Nat,times(times(plus(0,0),?x:Nat),s(?x:Nat))),plus(plus(?x:Nat,?y:Nat),?x:Nat))",
"times(times(times(?x:Nat,times(0,plus(?y:Nat,0))),0),s(plus(?y:Nat,s(?x:Nat)))) = times(?y:Nat,times(plus(plus(0,times(?x:Nat,?y:Nat)),s(times(?x:Nat,?z:Nat))),0))",
"times(times(s(plus(times(?z:Nat,?x:Nat),plus(?z:Nat,?z:Nat))),times(0,plus(plus(0,?y:Nat),?z:Nat))),?z:Nat) = times(plus(times(times(?x:Nat,plus(?x:Nat,?z:Nat)),?y:Nat),?x:Nat),0)",
"times(s(plus(s(times(?y:Nat,?y:Nat)),s(plus(?x:Nat,?y:Nat)))),times(times(times(?z:Nat,0),times(times(?x:Nat,?y:Nat),0)),?x:Nat)) = times(times(times(?y:Nat,?z:Nat),0),?x:Nat)",
"times(s(plus(s(s(?x:Nat)),?x:Nat)),0) = times(0,times(plus(plus(times(?y:Nat,?z:Nat),s(0)),plus(times(?x:Nat,?x:Nat),?z:Nat)),s(times(s(?z:Nat),plus(?x:Nat,?y:Nat)))))",
"times(times(times(0,0),times(0,plus(plus(?x:Nat,?x:Nat),0))),?y:Nat) = times(times(plus(s(plus(?z:Nat,?z:Nat)),?z:Nat),0),times(times(?y:Nat,?z:Nat),s(times(?x:Nat,?y:Nat))))",
"times(?x:Nat,times(s(0),times(plus(s(?y:Nat),?z:Nat),plus(0,times(0,?z:Nat))))) = times(s(?x:Nat),times(times(s(plus(?z:Nat,?x:Nat)),plus(times(?y:Nat,?y:Nat),0)),0))",
"times(times(0,times(times(times(?x:Nat,?x:Nat),plus(?x:Nat,0)),?x:Nat)),plus(?x:Nat,times(0,plus(times(?z:Nat,?z:Nat),?z:Nat)))) = times(times(?z:Nat,plus(?x:Nat,?z:Nat)),times(0,?y:Nat))",
"times(times(times(times(s(?x:Nat),0),?x:Nat),0),?x:Nat) = times(times(0,plus(s(times(?x:Nat,?y:Nat)),?x:Nat)),times(0,times(s(plus(?y:Nat,?y:Nat)),times(times(?y:Nat,?x:Nat),?x:Nat))))",
"times(0,times(s(plus(plus(?x:Nat,?x:Nat),times(?y:Nat,?x:Nat))),plus(?x:Nat,?x:Nat))) = times(0,s(times(plus(plus(?z:Nat,?z:Nat),times(?x:Nat,?x:Nat)),plus(times(?x:Nat,?z:Nat),plus(?x:Nat,?x:Nat)))))",
"times(0,times(times(times(plus(?x:Nat,?y:Nat),times(?y:Nat,?x:Nat)),?y:Nat),times(?x:Nat,s(plus(?z:Nat,?y:Nat))))) = times(?x:Nat,times(times(?y:Nat,0),times(plus(plus(0,?x:Nat),plus(0,?x:Nat)),?y:Nat)))",
"times(0,0) = times(times(times(times(times(?y:Nat,0),?x:Nat),s(times(?y:Nat,?x:Nat))),?y:Nat),plus(times(plus(times(?y:Nat,0),0),s(plus(?y:Nat,?z:Nat))),s(plus(plus(?y:Nat,0),plus(?z:Nat,?x:Nat)))))"
]

val problemsTimesWithZeroNo =
IOMstrs.rdEqs [
"s(?x:Nat) = times(?x:Nat,0)",
"s(?x:Nat) = times(0,?x:Nat)",
"s(?y:Nat) = times(0,?z:Nat)",
"s(?y:Nat) = times(0,0)",
"s(0) = times(?x:Nat,?x:Nat)",
"s(0) = times(?x:Nat,0)",
"s(0) = times(?y:Nat,?x:Nat)",
"s(?y:Nat) = times(s(?y:Nat),0)",
"s(0) = times(0,s(?x:Nat))",
"s(0) = times(s(?x:Nat),0)",
"s(s(?y:Nat)) = times(0,?x:Nat)",
"s(s(0)) = times(0,?z:Nat)",
"plus(?x:Nat,?x:Nat) = times(?y:Nat,0)",
"plus(?z:Nat,0) = times(0,?x:Nat)",
"plus(0,?y:Nat) = times(?x:Nat,?y:Nat)",
"plus(0,?z:Nat) = times(?x:Nat,0)",
"times(?x:Nat,?x:Nat) = times(0,0)",
"times(?x:Nat,?y:Nat) = plus(0,?z:Nat)",
"times(?x:Nat,?z:Nat) = plus(?z:Nat,0)",
"times(?y:Nat,?z:Nat) = plus(?y:Nat,0)",
"times(?z:Nat,0) = plus(?y:Nat,0)",
"times(0,?z:Nat) = times(?z:Nat,?z:Nat)",
"times(0,0) = s(s(?z:Nat))",
"times(0,0) = times(?z:Nat,?y:Nat)",
"times(?x:Nat,s(0)) = s(?z:Nat)",
"times(0,s(?y:Nat)) = s(?y:Nat)",
"times(s(?x:Nat),?x:Nat) = s(0)",
"s(?x:Nat) = plus(?x:Nat,times(?y:Nat,0))",
"s(?x:Nat) = times(?x:Nat,times(0,?x:Nat))",
"s(?x:Nat) = times(times(?x:Nat,?x:Nat),0)",
"s(?y:Nat) = plus(times(?x:Nat,0),?x:Nat)",
"s(?y:Nat) = times(s(s(0)),?y:Nat)",
"s(?y:Nat) = times(plus(0,?x:Nat),?x:Nat)",
"s(?y:Nat) = times(times(0,?x:Nat),?y:Nat)",
"s(0) = plus(?y:Nat,times(0,?x:Nat))",
"s(0) = plus(0,times(?y:Nat,?x:Nat))",
"s(0) = times(?x:Nat,s(s(?x:Nat)))",
"s(0) = times(?y:Nat,plus(?x:Nat,0))",
"plus(?x:Nat,0) = s(times(?x:Nat,?x:Nat))",
"times(?x:Nat,0) = plus(s(?x:Nat),0)",
"times(?z:Nat,?z:Nat) = s(times(0,0))",
"times(?z:Nat,0) = s(times(?y:Nat,?y:Nat))",
"times(0,0) = plus(s(?y:Nat),?x:Nat)",
"times(0,0) = times(?x:Nat,s(?x:Nat))",
"s(plus(?z:Nat,?y:Nat)) = times(0,?y:Nat)",
"s(times(?y:Nat,?x:Nat)) = times(0,0)",
"times(?x:Nat,s(0)) = plus(?y:Nat,?x:Nat)",
"times(?y:Nat,s(0)) = s(s(?z:Nat))",
"times(0,s(?x:Nat)) = plus(?y:Nat,?z:Nat)",
"times(s(?y:Nat),0) = plus(?y:Nat,0)",
"times(s(0),?x:Nat) = times(?y:Nat,?x:Nat)",
"times(s(0),?z:Nat) = plus(?z:Nat,?x:Nat)",
"times(0,plus(?x:Nat,?x:Nat)) = s(?z:Nat)",
"times(0,times(?y:Nat,?x:Nat)) = s(?y:Nat)",
"times(0,times(0,?x:Nat)) = s(?x:Nat)",
"times(plus(?x:Nat,0),?x:Nat) = s(?x:Nat)",
"times(plus(?x:Nat,0),0) = s(?x:Nat)",
"times(times(?x:Nat,0),?y:Nat) = s(?x:Nat)",
"s(s(?y:Nat)) = plus(?x:Nat,times(0,?y:Nat))",
"s(s(0)) = plus(times(?y:Nat,0),?y:Nat)",
"s(s(0)) = plus(times(0,0),0)",
"plus(?z:Nat,0) = times(0,times(?z:Nat,?z:Nat))",
"times(?x:Nat,?y:Nat) = times(?x:Nat,times(?y:Nat,0))",
"times(?y:Nat,?x:Nat) = s(times(0,s(?y:Nat)))",
"times(?y:Nat,0) = s(plus(s(?x:Nat),?z:Nat))",
"times(?y:Nat,0) = times(plus(0,?x:Nat),?y:Nat)",
"times(0,?y:Nat) = s(times(s(?x:Nat),?z:Nat))",
"s(times(0,0)) = plus(0,s(?x:Nat))",
"times(0,s(?y:Nat)) = times(s(?x:Nat),?y:Nat)",
"s(times(s(0),0)) = plus(0,?x:Nat)",
"plus(0,plus(?z:Nat,?z:Nat)) = times(?y:Nat,?x:Nat)",
"plus(times(?x:Nat,0),?x:Nat) = plus(?x:Nat,?y:Nat)",
"times(?x:Nat,times(?z:Nat,0)) = plus(?y:Nat,?x:Nat)",
"times(?y:Nat,plus(?z:Nat,?x:Nat)) = s(s(0))",
"times(s(?y:Nat),s(0)) = times(?x:Nat,?z:Nat)",
"times(times(?x:Nat,?x:Nat),?x:Nat) = times(0,?x:Nat)",
"times(times(?y:Nat,?x:Nat),?x:Nat) = plus(0,?x:Nat)",
"times(?x:Nat,times(s(?x:Nat),?z:Nat)) = s(0)",
"times(?z:Nat,s(times(?x:Nat,?x:Nat))) = s(0)",
"times(s(0),plus(?y:Nat,?z:Nat)) = s(?x:Nat)",
"times(times(?y:Nat,0),s(?y:Nat)) = s(?x:Nat)",
"s(?x:Nat) = plus(0,times(?y:Nat,times(?x:Nat,?y:Nat)))",
"s(?y:Nat) = plus(plus(times(?y:Nat,?z:Nat),0),?z:Nat)",
"s(?y:Nat) = times(0,plus(?x:Nat,times(?x:Nat,?x:Nat)))",
"s(?y:Nat) = times(times(?x:Nat,?y:Nat),times(?x:Nat,0))",
"s(?z:Nat) = plus(?z:Nat,times(s(s(?y:Nat)),0))",
"s(0) = plus(s(times(s(?z:Nat),?x:Nat)),?z:Nat)",
"s(s(?x:Nat)) = plus(s(times(?y:Nat,0)),?x:Nat)",
"s(s(?x:Nat)) = times(s(?y:Nat),plus(0,0))",
"s(s(?y:Nat)) = plus(s(times(0,?x:Nat)),0)",
"s(s(?z:Nat)) = times(plus(0,?x:Nat),s(?x:Nat))",
"plus(?z:Nat,?x:Nat) = times(s(?x:Nat),times(?z:Nat,0))",
"plus(0,?x:Nat) = times(times(?z:Nat,?z:Nat),s(?z:Nat))",
"plus(0,?z:Nat) = s(s(s(times(0,?z:Nat))))",
"times(?x:Nat,0) = s(s(plus(0,s(?y:Nat))))",
"times(?x:Nat,0) = times(?x:Nat,s(s(s(?z:Nat))))",
"times(?y:Nat,?x:Nat) = times(s(plus(0,0)),?y:Nat)",
"times(?z:Nat,?y:Nat) = times(plus(?y:Nat,s(0)),0)",
"times(?z:Nat,0) = s(plus(?y:Nat,plus(?x:Nat,?x:Nat)))",
"times(0,?x:Nat) = s(s(s(times(?x:Nat,0))))",
"times(0,?x:Nat) = times(s(?z:Nat),plus(0,?x:Nat))",
"times(0,?y:Nat) = times(s(times(0,?z:Nat)),?z:Nat)",
"s(plus(?x:Nat,?x:Nat)) = times(?x:Nat,times(?x:Nat,0))",
"s(plus(?x:Nat,0)) = times(plus(?x:Nat,0),?y:Nat)",
"s(plus(?z:Nat,0)) = times(?x:Nat,times(?x:Nat,?x:Nat))",
"s(plus(?z:Nat,0)) = times(times(?y:Nat,?x:Nat),?x:Nat)",
"s(times(?z:Nat,0)) = plus(s(s(0)),?y:Nat)",
"s(times(0,?z:Nat)) = plus(?y:Nat,times(?x:Nat,?x:Nat))",
"plus(s(?x:Nat),0) = s(times(s(?x:Nat),?x:Nat))",
"times(?y:Nat,s(?z:Nat)) = times(plus(0,?z:Nat),?x:Nat)",
"times(0,s(?x:Nat)) = times(plus(?z:Nat,?y:Nat),?x:Nat)",
"times(s(0),?x:Nat) = times(?x:Nat,times(?x:Nat,?z:Nat))",
"plus(0,plus(?x:Nat,?y:Nat)) = times(?z:Nat,s(?z:Nat))",
"plus(plus(?x:Nat,?x:Nat),0) = times(s(?x:Nat),?z:Nat)",
"plus(times(0,?x:Nat),?x:Nat) = s(plus(?z:Nat,?y:Nat))",
"times(?x:Nat,s(s(0))) = plus(s(0),?z:Nat)",
"times(?x:Nat,plus(0,?x:Nat)) = times(s(?x:Nat),?y:Nat)",
"times(?z:Nat,s(s(?x:Nat))) = times(0,s(?y:Nat))",
"times(plus(?x:Nat,?x:Nat),?z:Nat) = plus(s(0),?y:Nat)",
"times(times(?y:Nat,0),?y:Nat) = s(plus(?x:Nat,?y:Nat))",
"times(times(0,?z:Nat),?x:Nat) = plus(s(?x:Nat),0)",
"s(s(s(times(?y:Nat,?y:Nat)))) = times(?x:Nat,0)",
"s(s(times(s(0),?z:Nat))) = times(?x:Nat,?x:Nat)",
"s(times(plus(0,?y:Nat),?x:Nat)) = plus(0,?y:Nat)",
"plus(?z:Nat,plus(?x:Nat,s(?x:Nat))) = times(?x:Nat,0)",
"plus(0,s(plus(?z:Nat,?y:Nat))) = times(?x:Nat,?x:Nat)",
"plus(s(s(s(?x:Nat))),0) = times(?x:Nat,0)",
"plus(times(s(?z:Nat),?y:Nat),?z:Nat) = plus(0,?x:Nat)",
"times(?x:Nat,plus(?x:Nat,s(?y:Nat))) = plus(?x:Nat,0)",
"times(?x:Nat,plus(?x:Nat,s(0))) = times(?x:Nat,?x:Nat)",
"times(times(?x:Nat,s(?y:Nat)),0) = plus(?y:Nat,?x:Nat)",
"times(times(s(0),?x:Nat),?x:Nat) = s(s(?x:Nat))",
"plus(?x:Nat,times(plus(?y:Nat,0),?x:Nat)) = s(?x:Nat)",
"times(?y:Nat,plus(s(s(?x:Nat)),0)) = s(?x:Nat)",
"times(?z:Nat,s(s(plus(?x:Nat,?x:Nat)))) = s(0)",
"times(?z:Nat,s(s(plus(?y:Nat,0)))) = s(?y:Nat)",
"times(0,s(s(times(0,?x:Nat)))) = s(?x:Nat)",
"times(0,s(times(?x:Nat,s(?x:Nat)))) = s(?z:Nat)",
"times(s(?x:Nat),times(0,s(0))) = s(?x:Nat)",
"times(s(s(plus(?y:Nat,0))),?x:Nat) = s(?z:Nat)",
"times(times(0,plus(0,?x:Nat)),?x:Nat) = s(?x:Nat)",
"times(times(plus(?z:Nat,0),?z:Nat),?z:Nat) = s(?x:Nat)",
"s(?x:Nat) = plus(?y:Nat,times(s(?y:Nat),plus(?z:Nat,0)))",
"s(?x:Nat) = plus(times(s(plus(?x:Nat,0)),0),?z:Nat)",
"s(?x:Nat) = times(s(?z:Nat),s(times(s(0),?x:Nat)))",
"s(?y:Nat) = plus(times(0,plus(?z:Nat,s(?x:Nat))),?x:Nat)",
"s(?z:Nat) = plus(?z:Nat,plus(s(times(0,?x:Nat)),?y:Nat))",
"s(?z:Nat) = times(0,s(times(s(?y:Nat),s(0))))",
"s(0) = plus(?y:Nat,times(s(times(0,?x:Nat)),0))",
"s(0) = plus(?z:Nat,times(?z:Nat,plus(0,s(?z:Nat))))",
"s(0) = times(times(?y:Nat,times(?y:Nat,s(?z:Nat))),?x:Nat)",
"s(0) = times(times(0,?y:Nat),plus(?x:Nat,s(0)))",
"plus(?y:Nat,?x:Nat) = plus(0,plus(?x:Nat,times(?x:Nat,?y:Nat)))",
"plus(0,?x:Nat) = times(plus(?y:Nat,times(?y:Nat,?x:Nat)),?x:Nat)",
"plus(0,?z:Nat) = s(s(times(?y:Nat,plus(?x:Nat,0))))",
"times(?y:Nat,?y:Nat) = plus(plus(0,0),s(s(?y:Nat)))",
"times(?y:Nat,?z:Nat) = s(plus(0,times(s(?x:Nat),?z:Nat)))",
"times(?z:Nat,?y:Nat) = s(plus(s(times(?x:Nat,?x:Nat)),0))",
"times(0,0) = plus(times(?x:Nat,0),times(?y:Nat,?x:Nat))",
"plus(s(?x:Nat),?x:Nat) = times(0,plus(s(?x:Nat),0))",
"times(?z:Nat,s(?x:Nat)) = plus(s(0),times(?x:Nat,?x:Nat))",
"plus(?x:Nat,times(?y:Nat,?z:Nat)) = plus(s(0),s(?z:Nat))",
"times(0,plus(?z:Nat,?z:Nat)) = plus(plus(?x:Nat,?x:Nat),?x:Nat)",
"times(plus(?x:Nat,?z:Nat),?z:Nat) = times(plus(0,0),?z:Nat)",
"times(plus(?x:Nat,0),?x:Nat) = plus(?y:Nat,plus(?y:Nat,?x:Nat))",
"times(times(?x:Nat,?y:Nat),0) = plus(times(?y:Nat,?y:Nat),?y:Nat)",
"s(times(plus(?x:Nat,?x:Nat),?x:Nat)) = times(s(?y:Nat),0)",
"s(times(times(?z:Nat,0),?z:Nat)) = times(?y:Nat,s(0))",
"plus(s(?x:Nat),times(0,?y:Nat)) = s(plus(?x:Nat,?x:Nat))",
"plus(times(s(?z:Nat),?x:Nat),0) = s(plus(?z:Nat,0))",
"times(?x:Nat,s(plus(0,?x:Nat))) = times(?z:Nat,s(0))",
"times(s(s(s(0))),?x:Nat) = times(?x:Nat,s(?z:Nat))",
"times(plus(?y:Nat,?x:Nat),s(?x:Nat)) = plus(s(?z:Nat),0)",
"s(plus(0,s(plus(0,?z:Nat)))) = times(?x:Nat,?x:Nat)",
"s(plus(plus(s(0),?y:Nat),?x:Nat)) = times(?x:Nat,?y:Nat)",
"s(times(?x:Nat,s(times(?y:Nat,?z:Nat)))) = times(0,0)",
"s(times(?x:Nat,plus(0,s(?z:Nat)))) = plus(0,0)",
"s(times(s(?z:Nat),plus(0,?y:Nat))) = times(?x:Nat,?x:Nat)",
"s(times(s(plus(?z:Nat,?z:Nat)),?x:Nat)) = plus(?y:Nat,0)",
"plus(?y:Nat,plus(plus(?x:Nat,?x:Nat),0)) = times(?z:Nat,?y:Nat)",
"plus(?z:Nat,s(s(times(0,?z:Nat)))) = times(0,?x:Nat)",
"plus(s(0),times(s(?x:Nat),?z:Nat)) = times(?x:Nat,?y:Nat)",
"plus(s(times(?x:Nat,0)),s(0)) = s(s(?y:Nat))",
"plus(times(0,times(?y:Nat,?x:Nat)),?y:Nat) = plus(?z:Nat,?x:Nat)",
"plus(times(s(?y:Nat),s(0)),?x:Nat) = times(0,?y:Nat)",
"times(?x:Nat,times(?x:Nat,s(s(?z:Nat)))) = s(s(0))",
"times(0,s(s(times(?x:Nat,?y:Nat)))) = times(?y:Nat,?y:Nat)",
"times(0,s(plus(0,s(?x:Nat)))) = s(s(0))",
"times(s(s(0)),plus(?x:Nat,?x:Nat)) = times(?y:Nat,0)",
"times(s(plus(0,?x:Nat)),s(0)) = plus(?z:Nat,?x:Nat)",
"times(plus(s(s(0)),?y:Nat),?y:Nat) = s(s(0))",
"plus(0,times(?y:Nat,s(times(?x:Nat,?x:Nat)))) = s(0)",
"plus(s(plus(times(?x:Nat,?z:Nat),?x:Nat)),0) = s(?y:Nat)",
"plus(s(times(s(?z:Nat),s(?z:Nat))),0) = s(?z:Nat)",
"times(?y:Nat,s(plus(times(?y:Nat,?x:Nat),0))) = s(?x:Nat)",
"times(?z:Nat,plus(s(plus(0,0)),?x:Nat)) = s(?z:Nat)",
"times(s(plus(?y:Nat,plus(?x:Nat,?y:Nat))),0) = s(?x:Nat)",
"times(s(plus(s(?y:Nat),s(?y:Nat))),?x:Nat) = s(0)",
"s(?x:Nat) = plus(s(?y:Nat),plus(?x:Nat,s(times(0,?x:Nat))))",
"s(?x:Nat) = times(?x:Nat,plus(0,plus(?y:Nat,plus(?x:Nat,?x:Nat))))",
"s(?y:Nat) = plus(times(s(?x:Nat),plus(?x:Nat,s(0))),0)",
"s(?y:Nat) = times(s(?x:Nat),plus(?x:Nat,times(?z:Nat,s(0))))",
"s(?z:Nat) = plus(?x:Nat,times(plus(?x:Nat,0),times(0,0)))",
"s(?z:Nat) = plus(0,times(plus(plus(?x:Nat,?y:Nat),?z:Nat),?y:Nat))",
"s(?z:Nat) = plus(s(?x:Nat),s(plus(times(?y:Nat,?x:Nat),0)))",
"s(?z:Nat) = plus(times(?x:Nat,plus(plus(?x:Nat,?z:Nat),0)),0)",
"s(?z:Nat) = times(?y:Nat,plus(plus(?z:Nat,plus(?x:Nat,?x:Nat)),0))",
"s(?z:Nat) = times(plus(?x:Nat,plus(0,times(?x:Nat,?x:Nat))),0)",
"s(0) = plus(s(plus(?x:Nat,times(?y:Nat,?y:Nat))),s(?y:Nat))",
"s(0) = plus(s(times(times(?x:Nat,?z:Nat),?x:Nat)),s(?z:Nat))",
"s(0) = plus(times(times(plus(?x:Nat,?x:Nat),0),?x:Nat),?x:Nat)",
"s(0) = times(times(?y:Nat,times(plus(?z:Nat,?y:Nat),?y:Nat)),?z:Nat)",
"s(0) = times(times(s(plus(?x:Nat,?x:Nat)),0),s(?x:Nat))",
"s(s(?z:Nat)) = times(plus(times(s(?x:Nat),?z:Nat),0),?z:Nat)",
"plus(?x:Nat,?y:Nat) = times(s(times(?x:Nat,plus(0,?y:Nat))),0)",
"plus(?x:Nat,?y:Nat) = times(times(s(plus(?z:Nat,0)),?y:Nat),?z:Nat)",
"plus(?x:Nat,?z:Nat) = times(s(0),plus(?z:Nat,s(s(?z:Nat))))",
"plus(?x:Nat,?z:Nat) = times(times(s(times(?x:Nat,?x:Nat)),?x:Nat),0)",
"plus(?x:Nat,0) = times(plus(s(s(?x:Nat)),s(?y:Nat)),?y:Nat)",
"plus(?z:Nat,?y:Nat) = s(plus(times(?z:Nat,times(0,?x:Nat)),?x:Nat))",
"plus(0,?x:Nat) = plus(plus(s(times(0,?y:Nat)),?x:Nat),?x:Nat)",
"times(?x:Nat,?x:Nat) = plus(times(s(plus(0,?y:Nat)),0),?x:Nat)",
"times(?x:Nat,?x:Nat) = times(times(s(0),times(?x:Nat,?x:Nat)),?x:Nat)",
"times(?y:Nat,0) = s(times(?z:Nat,times(0,times(0,?x:Nat))))",
"times(?z:Nat,?x:Nat) = times(?y:Nat,times(s(times(0,0)),0))",
"times(?z:Nat,0) = plus(?x:Nat,times(plus(?y:Nat,?x:Nat),s(?x:Nat)))",
"times(0,?x:Nat) = plus(times(s(?x:Nat),?x:Nat),times(0,?x:Nat))",
"times(0,?x:Nat) = times(plus(?x:Nat,s(plus(?x:Nat,?x:Nat))),?y:Nat)",
"s(plus(?y:Nat,?x:Nat)) = plus(plus(times(?y:Nat,0),0),?x:Nat)",
"plus(s(?z:Nat),?z:Nat) = plus(s(s(times(?z:Nat,0))),?y:Nat)",
"plus(s(0),?x:Nat) = times(?x:Nat,times(?x:Nat,times(?x:Nat,?z:Nat)))",
"plus(s(0),0) = s(times(s(?z:Nat),plus(?z:Nat,?x:Nat)))",
"times(0,s(?y:Nat)) = plus(0,s(s(plus(?x:Nat,?y:Nat))))",
"times(0,s(?z:Nat)) = times(s(s(plus(?x:Nat,0))),?x:Nat)",
"plus(times(0,?x:Nat),?z:Nat) = s(times(plus(?x:Nat,?y:Nat),?x:Nat))",
"times(0,plus(?x:Nat,?x:Nat)) = times(?x:Nat,times(s(?x:Nat),?x:Nat))",
"times(plus(?x:Nat,0),0) = times(plus(s(?y:Nat),0),?y:Nat)",
"times(plus(?z:Nat,?x:Nat),0) = plus(?z:Nat,plus(?x:Nat,s(?x:Nat)))",
"times(times(?x:Nat,?x:Nat),0) = plus(?x:Nat,plus(s(?x:Nat),0))",
"times(times(?x:Nat,?y:Nat),0) = plus(times(?z:Nat,s(?z:Nat)),?x:Nat)",
"s(plus(?z:Nat,times(?x:Nat,0))) = times(?x:Nat,times(?z:Nat,?y:Nat))",
"plus(s(s(?x:Nat)),s(?z:Nat)) = times(?z:Nat,times(0,0))",
"plus(?z:Nat,times(?y:Nat,plus(?x:Nat,?z:Nat))) = s(plus(0,0))",
"times(?x:Nat,s(times(s(?x:Nat),?z:Nat))) = s(plus(0,?y:Nat))",
"times(0,s(times(?y:Nat,s(?x:Nat)))) = times(?x:Nat,s(?x:Nat))",
"times(s(?x:Nat),times(?z:Nat,s(0))) = times(s(?x:Nat),?x:Nat)",
"times(plus(?y:Nat,plus(0,?x:Nat)),0) = times(?y:Nat,s(?z:Nat))",
"times(times(?y:Nat,0),s(s(0))) = plus(s(?y:Nat),?x:Nat)",
"s(plus(?x:Nat,times(?z:Nat,plus(?y:Nat,0)))) = times(?y:Nat,?y:Nat)",
"s(plus(0,times(plus(0,?y:Nat),0))) = plus(?x:Nat,?y:Nat)",
"s(plus(times(?x:Nat,plus(?y:Nat,?y:Nat)),?x:Nat)) = times(0,?x:Nat)",
"plus(s(?x:Nat),s(s(plus(?z:Nat,?y:Nat)))) = times(0,?y:Nat)",
"plus(s(?x:Nat),plus(plus(?x:Nat,?y:Nat),?z:Nat)) = times(0,?z:Nat)",
"plus(plus(0,plus(?y:Nat,s(?x:Nat))),?x:Nat) = times(?x:Nat,?z:Nat)",
"plus(times(0,s(times(?y:Nat,?z:Nat))),0) = s(s(?y:Nat))",
"plus(times(s(?z:Nat),s(s(?x:Nat))),0) = times(?x:Nat,?z:Nat)",
"times(?y:Nat,times(?z:Nat,times(?y:Nat,s(?x:Nat)))) = times(0,?x:Nat)",
"times(?y:Nat,times(0,times(s(?x:Nat),?z:Nat))) = plus(0,?x:Nat)",
"times(s(s(s(?x:Nat))),times(?x:Nat,?x:Nat)) = plus(?y:Nat,0)",
"times(s(plus(times(0,?z:Nat),?z:Nat)),0) = plus(0,?x:Nat)",
"times(plus(s(?x:Nat),?z:Nat),times(?x:Nat,?x:Nat)) = times(?y:Nat,0)",
"times(plus(plus(?x:Nat,s(?y:Nat)),0),?y:Nat) = times(0,?x:Nat)",
"times(times(s(s(?z:Nat)),0),s(?y:Nat)) = s(s(0))",
"plus(0,s(plus(s(?x:Nat),times(?y:Nat,?y:Nat)))) = s(?z:Nat)",
"plus(0,times(plus(s(0),s(?x:Nat)),?x:Nat)) = s(?x:Nat)",
"plus(s(?y:Nat),s(times(times(?z:Nat,?y:Nat),0))) = s(?y:Nat)",
"plus(plus(?z:Nat,?x:Nat),s(s(times(0,?x:Nat)))) = s(?y:Nat)",
"times(0,plus(times(plus(?y:Nat,?y:Nat),0),?z:Nat)) = s(?y:Nat)",
"times(0,times(plus(?x:Nat,times(?x:Nat,?z:Nat)),?x:Nat)) = s(?x:Nat)",
"times(plus(times(?y:Nat,times(?z:Nat,?x:Nat)),0),?z:Nat) = s(?z:Nat)",
"times(plus(times(plus(?x:Nat,?x:Nat),0),?z:Nat),?x:Nat) = s(?x:Nat)",
"times(times(plus(s(?y:Nat),s(?z:Nat)),?x:Nat),?y:Nat) = s(0)",
"s(?x:Nat) = plus(s(times(?x:Nat,times(?z:Nat,0))),times(?x:Nat,?z:Nat))",
"s(?x:Nat) = plus(plus(s(?y:Nat),times(plus(?z:Nat,0),?z:Nat)),?z:Nat)",
"s(?x:Nat) = plus(times(?y:Nat,times(plus(0,0),s(?y:Nat))),0)",
"s(?x:Nat) = plus(times(?z:Nat,0),s(plus(times(?x:Nat,?x:Nat),?x:Nat)))",
"s(?x:Nat) = times(times(?y:Nat,plus(?x:Nat,?x:Nat)),plus(0,s(0)))",
"s(?y:Nat) = times(plus(plus(?x:Nat,?y:Nat),?y:Nat),plus(s(0),?z:Nat))",
"s(?z:Nat) = times(plus(times(times(0,?x:Nat),s(0)),?z:Nat),?x:Nat)",
"s(0) = times(?y:Nat,plus(plus(?x:Nat,?y:Nat),plus(0,s(?x:Nat))))",
"s(0) = times(plus(plus(?x:Nat,s(?x:Nat)),?x:Nat),plus(?x:Nat,0))",
"plus(?x:Nat,?x:Nat) = plus(?z:Nat,times(plus(?x:Nat,plus(?y:Nat,0)),?x:Nat))",
"plus(?x:Nat,?y:Nat) = plus(0,plus(?z:Nat,times(times(?x:Nat,0),0)))",
"plus(?x:Nat,?z:Nat) = times(?z:Nat,times(plus(?y:Nat,times(?x:Nat,?x:Nat)),0))",
"plus(?y:Nat,?x:Nat) = s(plus(s(s(0)),s(times(?x:Nat,?z:Nat))))",
"plus(?y:Nat,?y:Nat) = times(times(s(s(?z:Nat)),s(?x:Nat)),s(0))",
"plus(?y:Nat,0) = s(plus(s(?y:Nat),times(?x:Nat,plus(?z:Nat,?x:Nat))))",
"plus(?y:Nat,0) = plus(?x:Nat,times(?x:Nat,times(s(?x:Nat),s(?z:Nat))))",
"plus(?z:Nat,0) = s(s(times(times(?z:Nat,?x:Nat),times(?y:Nat,?z:Nat))))",
"plus(0,?z:Nat) = plus(times(plus(?x:Nat,?x:Nat),0),times(?z:Nat,?z:Nat))",
"plus(0,?z:Nat) = times(times(plus(times(?x:Nat,?z:Nat),?x:Nat),?x:Nat),?z:Nat)",
"times(?x:Nat,?x:Nat) = times(0,plus(s(times(?z:Nat,?x:Nat)),s(?y:Nat)))",
"times(?x:Nat,?z:Nat) = times(?z:Nat,plus(plus(s(0),s(0)),?x:Nat))",
"times(?x:Nat,0) = plus(times(s(times(?z:Nat,?x:Nat)),s(?x:Nat)),?y:Nat)",
"times(?y:Nat,?y:Nat) = times(plus(?y:Nat,plus(plus(?x:Nat,?x:Nat),0)),0)",
"times(?z:Nat,?z:Nat) = plus(?x:Nat,times(0,times(s(0),s(?x:Nat))))",
"times(?z:Nat,0) = s(times(plus(s(?y:Nat),?x:Nat),s(s(0))))",
"times(?z:Nat,0) = times(?x:Nat,plus(plus(times(?x:Nat,?x:Nat),?x:Nat),?z:Nat))",
"times(0,?x:Nat) = times(times(s(0),plus(?y:Nat,s(0))),?y:Nat)",
"s(times(?x:Nat,?x:Nat)) = times(s(times(?z:Nat,plus(?y:Nat,?x:Nat))),0)",
"s(times(?z:Nat,?x:Nat)) = plus(s(plus(times(?z:Nat,?y:Nat),0)),0)",
"s(times(0,?y:Nat)) = times(?z:Nat,s(times(?z:Nat,times(?y:Nat,0))))",
"plus(s(?x:Nat),?x:Nat) = times(s(times(plus(?z:Nat,?x:Nat),0)),?z:Nat)",
"times(0,s(?z:Nat)) = s(times(times(?x:Nat,times(?y:Nat,?y:Nat)),?y:Nat))",
"times(s(?x:Nat),?x:Nat) = s(times(0,times(?y:Nat,plus(?y:Nat,?y:Nat))))",
"times(s(?x:Nat),0) = plus(plus(s(times(?z:Nat,?x:Nat)),?x:Nat),?z:Nat)",
"plus(?x:Nat,plus(?y:Nat,?x:Nat)) = s(s(plus(times(?z:Nat,0),?y:Nat)))",
"plus(?x:Nat,plus(0,0)) = plus(plus(?x:Nat,?x:Nat),times(?x:Nat,0))",
"plus(s(?x:Nat),s(?z:Nat)) = s(times(s(times(?x:Nat,0)),?x:Nat))",
"plus(s(s(?y:Nat)),?z:Nat) = plus(s(?x:Nat),s(times(?x:Nat,0)))",
"plus(plus(?x:Nat,?z:Nat),?z:Nat) = plus(?x:Nat,s(times(s(?y:Nat),0)))",
"times(?x:Nat,plus(?x:Nat,0)) = plus(?y:Nat,times(s(s(?y:Nat)),?y:Nat))",
"times(0,times(?z:Nat,?x:Nat)) = plus(times(?x:Nat,?x:Nat),times(?x:Nat,?x:Nat))",
"times(s(?x:Nat),s(?x:Nat)) = times(s(times(s(?z:Nat),0)),0)",
"times(s(?z:Nat),s(?z:Nat)) = s(plus(times(0,s(?x:Nat)),?z:Nat))",
"times(plus(?x:Nat,?x:Nat),?z:Nat) = times(?z:Nat,plus(times(?x:Nat,?x:Nat),0))",
"s(s(plus(0,s(?x:Nat)))) = times(?x:Nat,plus(?y:Nat,s(?z:Nat)))",
"s(plus(s(s(?x:Nat)),0)) = times(?x:Nat,s(s(s(?x:Nat))))",
"s(times(?x:Nat,times(?x:Nat,?z:Nat))) = plus(?x:Nat,s(plus(?x:Nat,0)))",
"times(?y:Nat,s(s(s(?z:Nat)))) = s(s(s(plus(0,?z:Nat))))",
"times(plus(?y:Nat,s(?y:Nat)),?z:Nat) = plus(?x:Nat,plus(0,s(?x:Nat)))",
"times(times(s(?y:Nat),?z:Nat),0) = plus(times(0,0),s(?y:Nat))",
"s(plus(plus(?x:Nat,0),s(?y:Nat))) = times(s(?x:Nat),s(0))",
"plus(s(s(times(?x:Nat,?x:Nat))),?x:Nat) = times(?y:Nat,times(0,?z:Nat))",
"times(plus(0,times(?x:Nat,?x:Nat)),?x:Nat) = s(plus(s(0),?x:Nat))",
"s(s(plus(times(?x:Nat,?y:Nat),s(?y:Nat)))) = times(s(?x:Nat),0)",
"plus(?x:Nat,times(?z:Nat,s(times(?y:Nat,?x:Nat)))) = s(plus(0,?x:Nat))",
"plus(?z:Nat,times(s(plus(?x:Nat,?z:Nat)),?y:Nat)) = times(?y:Nat,s(0))",
"plus(s(?y:Nat),plus(?y:Nat,plus(?x:Nat,?z:Nat))) = times(s(?y:Nat),0)",
"plus(plus(times(?x:Nat,s(0)),?x:Nat),?y:Nat) = s(times(?y:Nat,?x:Nat))",
"times(0,times(?z:Nat,s(times(?x:Nat,?x:Nat)))) = s(s(s(?x:Nat)))",
"times(times(0,times(s(?z:Nat),?z:Nat)),0) = s(plus(0,?x:Nat))",
"s(plus(0,times(s(?z:Nat),times(?y:Nat,?z:Nat)))) = times(?x:Nat,0)",
"plus(?z:Nat,s(times(plus(?y:Nat,?x:Nat),s(?y:Nat)))) = plus(0,0)",
"plus(0,plus(s(?y:Nat),times(s(0),0))) = plus(?y:Nat,0)",
"plus(s(times(s(?y:Nat),?x:Nat)),plus(?y:Nat,?x:Nat)) = times(0,0)",
"plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),times(?y:Nat,?y:Nat)) = plus(0,?y:Nat)",
"plus(times(?z:Nat,plus(plus(0,?x:Nat),?x:Nat)),?z:Nat) = plus(?x:Nat,?y:Nat)",
"plus(times(s(0),s(times(?z:Nat,?x:Nat))),?x:Nat) = plus(?x:Nat,0)",
"plus(times(times(plus(?x:Nat,?x:Nat),?x:Nat),0),0) = times(?z:Nat,?x:Nat)",
"times(?z:Nat,plus(s(plus(0,?y:Nat)),s(?y:Nat))) = plus(?z:Nat,?x:Nat)",
"times(?z:Nat,plus(times(times(?x:Nat,?z:Nat),?x:Nat),0)) = plus(?x:Nat,?x:Nat)",
"times(?z:Nat,times(?x:Nat,plus(s(?x:Nat),s(?x:Nat)))) = plus(?x:Nat,0)",
"times(0,times(?x:Nat,times(times(?x:Nat,?y:Nat),?x:Nat))) = times(?x:Nat,?x:Nat)",
"times(0,times(plus(times(?z:Nat,?x:Nat),?x:Nat),?z:Nat)) = times(?x:Nat,?y:Nat)",
"times(plus(?y:Nat,plus(times(?x:Nat,0),?x:Nat)),?z:Nat) = plus(?y:Nat,?x:Nat)",
"times(plus(s(?z:Nat),times(?z:Nat,s(?z:Nat))),?x:Nat) = plus(?x:Nat,0)",
"times(0,s(times(plus(?z:Nat,?x:Nat),times(?y:Nat,?x:Nat)))) = s(?x:Nat)",
"times(plus(0,?z:Nat),plus(times(s(?y:Nat),?z:Nat),?z:Nat)) = s(?x:Nat)",
"times(plus(times(plus(0,?y:Nat),s(?y:Nat)),?x:Nat),?x:Nat) = s(0)",
"s(?x:Nat) = times(plus(times(times(?x:Nat,?x:Nat),0),plus(?x:Nat,?y:Nat)),?x:Nat)",
"s(?x:Nat) = times(times(plus(?x:Nat,s(0)),plus(s(?y:Nat),?x:Nat)),?z:Nat)",
"s(?x:Nat) = times(times(times(plus(?x:Nat,?x:Nat),0),0),s(s(?z:Nat)))",
"s(?y:Nat) = plus(times(times(s(?x:Nat),s(0)),?x:Nat),times(?x:Nat,?x:Nat))",
"s(?z:Nat) = plus(?x:Nat,times(times(s(?z:Nat),0),times(s(?x:Nat),?z:Nat)))",
"s(?z:Nat) = plus(?y:Nat,plus(times(plus(?x:Nat,?z:Nat),times(0,0)),?y:Nat))",
"s(?z:Nat) = times(0,times(times(times(?x:Nat,?y:Nat),times(?y:Nat,?x:Nat)),?y:Nat))",
"s(0) = plus(?x:Nat,plus(times(s(0),?y:Nat),plus(s(?y:Nat),?x:Nat)))",
"s(0) = plus(times(s(s(?y:Nat)),times(plus(?z:Nat,?x:Nat),?x:Nat)),?x:Nat)",
"s(s(?x:Nat)) = plus(0,times(s(s(?x:Nat)),plus(s(?x:Nat),?y:Nat)))",
"s(s(?y:Nat)) = plus(times(plus(?x:Nat,times(0,?x:Nat)),s(?x:Nat)),0)",
"plus(?x:Nat,?z:Nat) = times(times(times(s(0),?x:Nat),times(?x:Nat,?y:Nat)),?z:Nat)",
"plus(?y:Nat,?x:Nat) = times(times(plus(times(?y:Nat,0),s(?z:Nat)),0),?y:Nat)",
"plus(?z:Nat,?y:Nat) = plus(?x:Nat,plus(s(?y:Nat),plus(0,times(?y:Nat,?z:Nat))))",
"plus(?z:Nat,?z:Nat) = plus(times(times(?z:Nat,?x:Nat),s(times(?x:Nat,0))),?z:Nat)",
"plus(0,?x:Nat) = times(0,times(?x:Nat,plus(times(0,?z:Nat),s(?z:Nat))))",
"times(?x:Nat,?y:Nat) = s(plus(?x:Nat,times(times(0,?y:Nat),times(0,?y:Nat))))",
"times(?x:Nat,?y:Nat) = plus(?y:Nat,times(s(plus(?z:Nat,?z:Nat)),times(?z:Nat,0)))",
"times(?x:Nat,0) = times(plus(?y:Nat,s(s(?y:Nat))),times(?y:Nat,s(?z:Nat)))",
"times(?z:Nat,0) = plus(s(?x:Nat),s(times(times(?z:Nat,?y:Nat),s(?x:Nat))))",
"times(0,?y:Nat) = times(times(?z:Nat,times(s(0),s(?z:Nat))),s(?y:Nat))",
"times(0,0) = plus(plus(?z:Nat,s(plus(?z:Nat,?x:Nat))),times(?z:Nat,0))",
"s(s(s(?x:Nat))) = plus(?x:Nat,times(?x:Nat,times(plus(?z:Nat,0),?y:Nat)))",
"s(s(s(?x:Nat))) = times(plus(times(0,0),s(s(?x:Nat))),?y:Nat)",
"s(plus(?x:Nat,0)) = times(s(times(times(?y:Nat,?x:Nat),s(?y:Nat))),0)",
"s(times(0,0)) = plus(times(times(?x:Nat,times(?y:Nat,0)),?x:Nat),?z:Nat)",
"plus(?y:Nat,s(0)) = s(plus(?x:Nat,times(plus(?x:Nat,0),s(?x:Nat))))",
"plus(s(?x:Nat),?x:Nat) = plus(times(plus(0,s(?x:Nat)),s(?y:Nat)),?x:Nat)",
"plus(s(0),0) = plus(0,times(plus(times(?y:Nat,?x:Nat),?x:Nat),?x:Nat))",
"plus(s(0),0) = plus(s(?x:Nat),plus(?y:Nat,s(times(?x:Nat,?x:Nat))))",
"times(s(?z:Nat),?x:Nat) = plus(times(?z:Nat,?z:Nat),times(times(?x:Nat,0),?x:Nat))",
"plus(0,plus(?x:Nat,?x:Nat)) = s(s(plus(s(?x:Nat),times(?x:Nat,?x:Nat))))",
"plus(s(s(0)),0) = times(?x:Nat,plus(0,times(?y:Nat,s(?x:Nat))))",
"plus(plus(?x:Nat,?x:Nat),0) = plus(?y:Nat,s(times(s(?x:Nat),s(0))))",
"times(plus(?z:Nat,?x:Nat),?z:Nat) = plus(plus(s(plus(?x:Nat,?y:Nat)),?x:Nat),0)",
"s(s(plus(?z:Nat,s(?x:Nat)))) = times(0,s(times(0,s(?z:Nat))))",
"s(s(times(?y:Nat,s(0)))) = times(?x:Nat,s(s(times(?x:Nat,?y:Nat))))",
"plus(?x:Nat,plus(0,s(?y:Nat))) = s(times(s(s(0)),s(?x:Nat)))",
"plus(0,plus(s(?y:Nat),0)) = times(0,times(plus(?y:Nat,?x:Nat),?z:Nat))",
"plus(plus(?x:Nat,s(?x:Nat)),?y:Nat) = plus(times(0,?y:Nat),plus(0,?x:Nat))",
"plus(times(s(0),?x:Nat),?y:Nat) = plus(0,times(plus(?x:Nat,?z:Nat),?x:Nat))",
"times(?y:Nat,times(times(0,0),0)) = s(s(s(plus(?x:Nat,0))))",
"times(0,times(?z:Nat,plus(0,?x:Nat))) = times(s(?z:Nat),s(s(?z:Nat)))",
"plus(?x:Nat,s(plus(plus(?y:Nat,?x:Nat),0))) = times(plus(?y:Nat,?x:Nat),0)",
"plus(?z:Nat,plus(?y:Nat,times(0,s(?z:Nat)))) = times(?y:Nat,plus(?x:Nat,?x:Nat))",
"plus(0,times(?x:Nat,s(times(?x:Nat,?x:Nat)))) = times(plus(?y:Nat,?z:Nat),?x:Nat)",
"s(s(plus(plus(?x:Nat,0),times(?x:Nat,?y:Nat)))) = times(s(?x:Nat),?x:Nat)",
"plus(plus(?z:Nat,plus(plus(?y:Nat,?x:Nat),?y:Nat)),?x:Nat) = times(?z:Nat,s(0))",
"plus(plus(times(plus(0,?y:Nat),?x:Nat),?z:Nat),0) = plus(?y:Nat,s(0))",
"plus(times(s(plus(0,?x:Nat)),?z:Nat),s(0)) = s(plus(?x:Nat,0))",
"times(s(0),times(s(times(0,?x:Nat)),0)) = plus(?z:Nat,s(?x:Nat))",
"times(s(times(s(0),times(?x:Nat,?x:Nat))),?y:Nat) = times(?y:Nat,s(?x:Nat))",
"times(plus(?z:Nat,times(0,times(?z:Nat,?x:Nat))),?z:Nat) = plus(s(?z:Nat),?x:Nat)",
"times(times(s(?y:Nat),times(s(0),?y:Nat)),?z:Nat) = s(plus(?y:Nat,0))",
"s(plus(?z:Nat,times(plus(?z:Nat,0),plus(?x:Nat,0)))) = times(?x:Nat,?z:Nat)",
"s(plus(times(plus(?z:Nat,?x:Nat),plus(?y:Nat,0)),?z:Nat)) = times(?y:Nat,0)",
"s(plus(times(times(?z:Nat,0),?z:Nat),times(?x:Nat,0))) = times(?z:Nat,0)",
"s(times(s(s(?x:Nat)),plus(plus(0,?x:Nat),?y:Nat))) = times(0,0)",
"plus(s(plus(plus(?y:Nat,?x:Nat),plus(?y:Nat,0))),?x:Nat) = times(?x:Nat,0)",
"plus(plus(plus(s(0),plus(?x:Nat,0)),?x:Nat),?y:Nat) = times(?y:Nat,?x:Nat)",
"plus(times(?z:Nat,plus(times(?y:Nat,?x:Nat),s(0))),?x:Nat) = times(?x:Nat,0)",
"plus(times(s(?x:Nat),times(?x:Nat,plus(?y:Nat,?z:Nat))),?y:Nat) = plus(0,?y:Nat)",
"times(?x:Nat,times(s(0),times(times(0,?x:Nat),?x:Nat))) = times(?x:Nat,?x:Nat)",
"times(?z:Nat,plus(s(?z:Nat),plus(times(?x:Nat,?x:Nat),?z:Nat))) = times(0,?z:Nat)",
"times(?z:Nat,plus(s(plus(?z:Nat,?x:Nat)),s(s(0)))) = times(?x:Nat,?y:Nat)",
"times(s(?y:Nat),times(s(?x:Nat),s(plus(?y:Nat,0)))) = times(?x:Nat,?z:Nat)",
"times(s(0),plus(times(s(?x:Nat),s(?x:Nat)),?z:Nat)) = s(s(0))",
"times(s(plus(?x:Nat,times(0,?x:Nat))),plus(?x:Nat,?z:Nat)) = times(?x:Nat,?z:Nat)",
"times(s(plus(0,plus(?x:Nat,?y:Nat))),times(?z:Nat,?z:Nat)) = plus(?x:Nat,?x:Nat)",
"times(plus(plus(times(?y:Nat,?z:Nat),?z:Nat),s(?x:Nat)),0) = plus(?z:Nat,?x:Nat)",
"times(plus(plus(times(0,?y:Nat),?y:Nat),s(0)),?z:Nat) = s(s(?x:Nat))",
"plus(?x:Nat,plus(times(times(?y:Nat,?x:Nat),times(0,?y:Nat)),?x:Nat)) = s(0)",
"plus(s(?x:Nat),s(plus(times(?x:Nat,?x:Nat),plus(0,?x:Nat)))) = s(0)",
"plus(s(times(?x:Nat,0)),times(?x:Nat,plus(s(?x:Nat),?y:Nat))) = s(?y:Nat)",
"plus(plus(times(plus(0,?z:Nat),?x:Nat),?z:Nat),plus(?x:Nat,?x:Nat)) = s(?x:Nat)",
"plus(times(?x:Nat,plus(times(?x:Nat,?x:Nat),plus(?z:Nat,?x:Nat))),?z:Nat) = s(0)",
"plus(times(?z:Nat,times(plus(?x:Nat,?y:Nat),0)),plus(?y:Nat,?y:Nat)) = s(?y:Nat)",
"plus(times(plus(0,plus(?z:Nat,0)),s(?y:Nat)),s(0)) = s(?x:Nat)",
"plus(times(plus(plus(?y:Nat,?x:Nat),plus(0,?y:Nat)),?x:Nat),0) = s(?z:Nat)",
"times(?y:Nat,times(times(?y:Nat,times(?x:Nat,?z:Nat)),times(?x:Nat,?x:Nat))) = s(0)",
"times(?y:Nat,times(times(plus(?x:Nat,?x:Nat),times(?x:Nat,?x:Nat)),?z:Nat)) = s(0)",
"times(s(plus(?x:Nat,?x:Nat)),s(times(?z:Nat,times(0,?x:Nat)))) = s(?z:Nat)",
"times(plus(?x:Nat,?x:Nat),plus(times(plus(0,?x:Nat),0),?y:Nat)) = s(?y:Nat)",
"times(plus(plus(plus(0,?x:Nat),plus(?x:Nat,?x:Nat)),?y:Nat),?x:Nat) = s(?y:Nat)",
"s(?x:Nat) = times(?z:Nat,times(plus(?x:Nat,s(?x:Nat)),times(?x:Nat,plus(0,?z:Nat))))",
"s(?y:Nat) = plus(plus(?y:Nat,times(s(0),s(?z:Nat))),s(plus(?z:Nat,?y:Nat)))",
"s(?y:Nat) = plus(times(s(s(?z:Nat)),plus(?z:Nat,plus(?x:Nat,?x:Nat))),s(0))",
"s(?y:Nat) = times(s(times(?x:Nat,times(0,?x:Nat))),s(plus(s(?x:Nat),?x:Nat)))",
"s(?z:Nat) = plus(s(s(plus(?y:Nat,?y:Nat))),times(0,s(plus(?x:Nat,?x:Nat))))",
"s(0) = times(?z:Nat,plus(s(plus(0,?y:Nat)),times(times(?z:Nat,?x:Nat),?x:Nat)))",
"s(0) = times(times(?x:Nat,?x:Nat),plus(plus(times(0,?x:Nat),?y:Nat),s(0)))",
"s(s(?y:Nat)) = plus(times(s(times(?y:Nat,?x:Nat)),s(?x:Nat)),times(0,?x:Nat))",
"plus(?x:Nat,?x:Nat) = times(?z:Nat,plus(times(?x:Nat,?z:Nat),plus(s(0),s(?y:Nat))))",
"plus(?x:Nat,?y:Nat) = plus(times(?x:Nat,times(times(?z:Nat,?x:Nat),times(?x:Nat,?y:Nat))),0)",
"plus(?y:Nat,?x:Nat) = times(?x:Nat,plus(?y:Nat,plus(plus(0,?z:Nat),times(?x:Nat,?y:Nat))))",
"plus(?z:Nat,?z:Nat) = times(s(s(plus(0,?x:Nat))),plus(0,s(s(?x:Nat))))",
"plus(0,?x:Nat) = plus(plus(s(?x:Nat),?z:Nat),s(times(plus(?x:Nat,?z:Nat),?x:Nat)))",
"times(?x:Nat,?x:Nat) = plus(plus(times(plus(?y:Nat,0),times(?z:Nat,?x:Nat)),?x:Nat),0)",
"times(?x:Nat,0) = times(times(?x:Nat,plus(plus(?y:Nat,?x:Nat),times(0,0))),?x:Nat)",
"times(?y:Nat,?x:Nat) = plus(?x:Nat,plus(?y:Nat,times(plus(?y:Nat,?y:Nat),plus(?z:Nat,0))))",
"times(?y:Nat,?x:Nat) = plus(times(?y:Nat,plus(plus(?y:Nat,0),plus(?x:Nat,0))),?x:Nat)",
"times(?y:Nat,?y:Nat) = s(times(plus(?z:Nat,times(0,?y:Nat)),plus(s(?y:Nat),0)))",
"times(?z:Nat,0) = s(times(s(?z:Nat),times(times(?z:Nat,?x:Nat),times(?z:Nat,?y:Nat))))",
"times(0,0) = times(plus(times(plus(?x:Nat,?x:Nat),0),?y:Nat),plus(?y:Nat,?x:Nat))",
"s(plus(?y:Nat,?x:Nat)) = times(plus(?x:Nat,times(plus(0,0),?z:Nat)),s(?y:Nat))",
"s(times(?x:Nat,?x:Nat)) = times(times(times(s(?x:Nat),plus(0,?y:Nat)),?z:Nat),?y:Nat)",
"plus(s(?x:Nat),?x:Nat) = times(s(?x:Nat),plus(0,times(0,times(?x:Nat,?x:Nat))))",
"times(s(?z:Nat),?x:Nat) = s(plus(times(times(0,?z:Nat),?y:Nat),times(?z:Nat,?x:Nat)))",
"times(s(?z:Nat),?z:Nat) = plus(?z:Nat,times(times(s(?x:Nat),times(?y:Nat,0)),?x:Nat))",
"s(s(s(s(?x:Nat)))) = times(0,plus(plus(s(0),s(?x:Nat)),0))",
"s(times(?x:Nat,s(?z:Nat))) = times(?z:Nat,s(times(s(0),plus(?x:Nat,0))))",
"plus(?x:Nat,plus(?x:Nat,?z:Nat)) = s(plus(times(s(0),times(0,?x:Nat)),?x:Nat))",
"times(?x:Nat,plus(?x:Nat,?x:Nat)) = s(s(plus(plus(0,0),times(?y:Nat,?x:Nat))))",
"times(?y:Nat,plus(?z:Nat,0)) = plus(?x:Nat,plus(plus(times(?x:Nat,?x:Nat),?x:Nat),0))",
"times(0,times(?z:Nat,0)) = times(times(plus(?x:Nat,?x:Nat),s(?y:Nat)),s(?z:Nat))",
"times(s(0),s(?y:Nat)) = s(times(plus(plus(?z:Nat,0),s(?x:Nat)),?y:Nat))",
"s(times(?y:Nat,times(?x:Nat,?x:Nat))) = plus(0,plus(s(s(?x:Nat)),s(?y:Nat)))",
"plus(?y:Nat,s(s(s(?x:Nat)))) = s(s(times(s(0),plus(?z:Nat,?x:Nat))))",
"plus(s(times(?y:Nat,0)),?y:Nat) = plus(plus(s(plus(?y:Nat,?x:Nat)),?y:Nat),?y:Nat)",
"plus(times(s(?x:Nat),?z:Nat),0) = plus(?x:Nat,plus(plus(s(?x:Nat),?x:Nat),0))",
"plus(?x:Nat,s(s(times(0,?x:Nat)))) = times(s(?x:Nat),s(times(?y:Nat,?y:Nat)))",
"times(?z:Nat,times(s(?x:Nat),s(?x:Nat))) = times(plus(0,s(?y:Nat)),s(?x:Nat))",
"times(plus(?y:Nat,s(s(?z:Nat))),0) = plus(plus(?y:Nat,s(s(0))),?z:Nat)",
"times(plus(s(?z:Nat),s(?x:Nat)),?x:Nat) = s(s(times(0,plus(?z:Nat,?x:Nat))))",
"s(plus(?x:Nat,plus(times(?x:Nat,?x:Nat),0))) = times(?x:Nat,times(s(?y:Nat),?x:Nat))",
"s(plus(0,times(s(?x:Nat),times(?x:Nat,?x:Nat)))) = times(?x:Nat,times(?x:Nat,0))",
"s(times(plus(times(?x:Nat,0),?y:Nat),s(?x:Nat))) = times(times(?y:Nat,0),0)",
"plus(s(?z:Nat),s(plus(s(?z:Nat),s(?x:Nat)))) = times(plus(?z:Nat,0),?z:Nat)",
"plus(times(0,plus(?x:Nat,?x:Nat)),plus(?x:Nat,?x:Nat)) = plus(?y:Nat,plus(?x:Nat,?y:Nat))",
"plus(times(times(plus(?z:Nat,?z:Nat),?x:Nat),0),?x:Nat) = plus(s(s(0)),0)",
"times(s(?x:Nat),plus(0,plus(s(?x:Nat),?x:Nat))) = s(times(s(0),0))",
"times(plus(?x:Nat,times(0,plus(?x:Nat,?x:Nat))),?y:Nat) = times(s(?x:Nat),s(?x:Nat))",
"times(times(s(?y:Nat),s(0)),plus(?x:Nat,?z:Nat)) = plus(times(?x:Nat,?y:Nat),?x:Nat)",
"plus(plus(times(s(?z:Nat),times(?x:Nat,?x:Nat)),?x:Nat),0) = s(s(s(?x:Nat)))",
"plus(times(plus(times(?x:Nat,0),?x:Nat),?z:Nat),s(?y:Nat)) = s(plus(0,?x:Nat))",
"plus(times(plus(times(?z:Nat,?x:Nat),?y:Nat),s(?y:Nat)),?z:Nat) = plus(0,s(?y:Nat))",
"times(?y:Nat,plus(s(plus(?x:Nat,?x:Nat)),s(s(?z:Nat)))) = s(times(?x:Nat,0))",
"times(plus(plus(times(?x:Nat,?x:Nat),?x:Nat),0),s(?x:Nat)) = s(times(0,?z:Nat))",
"times(plus(times(s(?z:Nat),0),plus(?x:Nat,?x:Nat)),?z:Nat) = plus(?y:Nat,s(?x:Nat))",
"s(times(plus(?x:Nat,plus(0,?x:Nat)),s(plus(?y:Nat,?y:Nat)))) = times(?y:Nat,?x:Nat)",
"plus(?x:Nat,times(plus(s(?x:Nat),s(0)),plus(?x:Nat,0))) = s(s(0))",
"plus(?y:Nat,plus(plus(0,?x:Nat),times(?x:Nat,times(?y:Nat,?x:Nat)))) = times(?x:Nat,?x:Nat)",
"plus(0,plus(s(times(0,?x:Nat)),plus(s(?x:Nat),?x:Nat))) = times(?x:Nat,0)",
"plus(s(?y:Nat),times(plus(?x:Nat,plus(?z:Nat,0)),s(0))) = plus(?z:Nat,?x:Nat)",
"plus(s(0),plus(s(s(?y:Nat)),s(times(?y:Nat,?x:Nat)))) = times(?z:Nat,?z:Nat)",
"plus(s(times(plus(?x:Nat,0),0)),plus(s(?x:Nat),?z:Nat)) = times(0,?y:Nat)",
"plus(plus(?x:Nat,0),times(plus(times(0,?z:Nat),?z:Nat),?z:Nat)) = s(s(0))",
"plus(plus(times(times(0,?z:Nat),plus(?z:Nat,?x:Nat)),?z:Nat),?y:Nat) = plus(?y:Nat,?y:Nat)",
"plus(times(0,plus(plus(0,?z:Nat),plus(?z:Nat,?z:Nat))),?x:Nat) = times(?x:Nat,?x:Nat)",
"times(?x:Nat,times(s(s(?z:Nat)),plus(times(?x:Nat,?x:Nat),?x:Nat))) = plus(0,?y:Nat)",
"times(?y:Nat,plus(times(times(?z:Nat,?x:Nat),times(?y:Nat,?x:Nat)),?y:Nat)) = plus(0,?z:Nat)",
"times(s(?z:Nat),plus(times(plus(?y:Nat,?x:Nat),?z:Nat),s(?x:Nat))) = times(?y:Nat,0)",
"times(s(s(plus(?x:Nat,?z:Nat))),s(s(plus(?x:Nat,0)))) = times(?x:Nat,?z:Nat)",
"times(s(plus(?z:Nat,s(?z:Nat))),plus(?x:Nat,times(?x:Nat,0))) = times(0,?z:Nat)",
"times(plus(?z:Nat,s(times(?y:Nat,0))),s(s(s(?z:Nat)))) = times(?y:Nat,?y:Nat)",
"times(plus(plus(s(0),?x:Nat),plus(0,?z:Nat)),s(?x:Nat)) = s(s(?x:Nat))",
"times(plus(plus(times(0,0),times(?z:Nat,?x:Nat)),?z:Nat),?x:Nat) = times(?z:Nat,?x:Nat)",
"times(times(s(s(?y:Nat)),?z:Nat),times(s(?z:Nat),s(?x:Nat))) = times(0,?x:Nat)",
"times(times(plus(plus(?x:Nat,0),?x:Nat),?z:Nat),plus(?y:Nat,?z:Nat)) = times(?z:Nat,?y:Nat)",
"plus(s(plus(plus(?z:Nat,?x:Nat),?x:Nat)),s(times(?x:Nat,s(?x:Nat)))) = s(0)",
"plus(plus(times(plus(?z:Nat,0),?x:Nat),s(s(?z:Nat))),s(?x:Nat)) = s(?x:Nat)",
"plus(times(0,s(?z:Nat)),times(0,times(?x:Nat,plus(?z:Nat,?x:Nat)))) = s(?y:Nat)",
"plus(times(s(times(?x:Nat,0)),plus(?x:Nat,times(?x:Nat,?x:Nat))),?x:Nat) = s(0)",
"plus(times(plus(?x:Nat,?y:Nat),plus(s(?x:Nat),times(?x:Nat,0))),?y:Nat) = s(?y:Nat)",
"times(?x:Nat,plus(plus(s(0),times(?y:Nat,?x:Nat)),plus(?y:Nat,?x:Nat))) = s(0)",
"times(plus(plus(times(?x:Nat,?y:Nat),plus(?z:Nat,0)),s(0)),?z:Nat) = s(?x:Nat)",
"times(times(plus(0,?y:Nat),0),times(0,times(s(0),0))) = s(?z:Nat)",
"times(times(plus(times(?x:Nat,0),s(0)),plus(0,?x:Nat)),?z:Nat) = s(0)",
"s(?x:Nat) = plus(plus(s(times(?x:Nat,?x:Nat)),times(plus(?x:Nat,?x:Nat),s(?x:Nat))),0)",
"s(?x:Nat) = times(0,times(times(times(?z:Nat,?x:Nat),s(?z:Nat)),plus(?z:Nat,s(?z:Nat))))",
"s(?x:Nat) = times(times(plus(?z:Nat,?y:Nat),s(0)),times(?z:Nat,s(times(0,?y:Nat))))",
"s(?z:Nat) = times(times(s(times(?x:Nat,?y:Nat)),0),times(s(?x:Nat),plus(?z:Nat,0)))",
"s(s(0)) = times(times(?z:Nat,0),times(s(?z:Nat),plus(plus(0,?x:Nat),?x:Nat)))",
"plus(?x:Nat,?x:Nat) = s(plus(plus(s(?x:Nat),plus(0,?x:Nat)),s(times(?x:Nat,0))))",
"plus(?x:Nat,?x:Nat) = times(times(plus(s(0),times(?x:Nat,0)),plus(?y:Nat,?z:Nat)),0)",
"plus(?x:Nat,?z:Nat) = s(times(s(times(?x:Nat,?x:Nat)),times(s(?y:Nat),times(0,?y:Nat))))",
"plus(?y:Nat,?x:Nat) = plus(plus(plus(times(?z:Nat,?x:Nat),?y:Nat),s(0)),plus(?x:Nat,?x:Nat))",
"plus(?y:Nat,?z:Nat) = plus(plus(plus(times(0,?z:Nat),?x:Nat),times(0,s(?x:Nat))),?z:Nat)",
"plus(?z:Nat,?x:Nat) = times(plus(?z:Nat,s(s(0))),times(s(plus(?y:Nat,?x:Nat)),?z:Nat))",
"plus(?z:Nat,?y:Nat) = plus(s(plus(s(?z:Nat),?x:Nat)),times(s(plus(0,?z:Nat)),?y:Nat))",
"plus(?z:Nat,?z:Nat) = plus(?x:Nat,plus(s(?x:Nat),times(times(?x:Nat,?z:Nat),times(0,?z:Nat))))",
"plus(0,?y:Nat) = plus(times(times(s(?y:Nat),0),times(plus(?x:Nat,?z:Nat),?y:Nat)),?x:Nat)",
"plus(0,?z:Nat) = plus(s(plus(?x:Nat,?z:Nat)),times(?y:Nat,plus(times(?z:Nat,?y:Nat),0)))",
"times(?x:Nat,?z:Nat) = plus(0,plus(s(0),plus(plus(?z:Nat,?z:Nat),times(?x:Nat,?y:Nat))))",
"plus(?x:Nat,s(?x:Nat)) = times(times(times(times(?x:Nat,0),times(?x:Nat,?y:Nat)),0),?z:Nat)",
"plus(s(?x:Nat),0) = plus(times(?y:Nat,plus(?x:Nat,?x:Nat)),s(plus(0,s(?x:Nat))))",
"plus(s(?y:Nat),?x:Nat) = s(times(s(times(?y:Nat,?z:Nat)),times(0,plus(?y:Nat,?x:Nat))))",
"plus(s(?z:Nat),?x:Nat) = times(s(?x:Nat),plus(?z:Nat,times(s(0),plus(?x:Nat,?z:Nat))))",
"times(?x:Nat,s(?x:Nat)) = plus(times(times(?y:Nat,?z:Nat),?y:Nat),s(s(times(?x:Nat,0))))",
"times(?z:Nat,s(?x:Nat)) = plus(?x:Nat,times(plus(times(0,?x:Nat),times(0,0)),?x:Nat))",
"times(?z:Nat,s(0)) = plus(plus(?y:Nat,s(?y:Nat)),s(times(0,plus(?x:Nat,?z:Nat))))",
"s(s(plus(0,?x:Nat))) = times(?x:Nat,s(plus(plus(?x:Nat,?z:Nat),times(?y:Nat,?y:Nat))))",
"s(times(s(?x:Nat),?y:Nat)) = times(s(times(times(?x:Nat,0),?x:Nat)),plus(?x:Nat,0))",
"plus(0,s(s(0))) = times(s(?x:Nat),plus(plus(s(?x:Nat),s(0)),?x:Nat))",
"plus(s(?x:Nat),times(0,0)) = s(times(s(s(?z:Nat)),plus(s(0),?x:Nat)))",
"plus(times(s(?z:Nat),0),?z:Nat) = s(plus(times(s(?z:Nat),plus(?x:Nat,0)),0))",
"times(plus(s(0),?y:Nat),?z:Nat) = s(s(plus(times(?y:Nat,?y:Nat),plus(?x:Nat,0))))",
"times(times(?x:Nat,s(?x:Nat)),?x:Nat) = times(?z:Nat,times(?x:Nat,plus(0,plus(0,?z:Nat))))",
"s(s(times(?y:Nat,times(?z:Nat,?z:Nat)))) = plus(times(s(times(?x:Nat,?z:Nat)),?x:Nat),0)",
"s(plus(s(times(?x:Nat,?z:Nat)),0)) = plus(times(?y:Nat,s(plus(?z:Nat,0))),?z:Nat)",
"s(times(s(plus(0,0)),0)) = plus(?x:Nat,s(plus(plus(?z:Nat,?x:Nat),?x:Nat)))",
"plus(?y:Nat,plus(times(0,0),?x:Nat)) = plus(?x:Nat,times(times(?z:Nat,?z:Nat),s(?y:Nat)))",
"plus(?z:Nat,times(?x:Nat,s(s(0)))) = times(0,s(times(?z:Nat,plus(?z:Nat,?z:Nat))))",
"plus(times(?z:Nat,0),plus(?x:Nat,?y:Nat)) = s(times(plus(0,plus(?y:Nat,0)),?x:Nat))",
"times(?x:Nat,plus(?x:Nat,plus(0,0))) = times(plus(s(times(?y:Nat,?y:Nat)),?x:Nat),?x:Nat)",
"s(plus(0,times(?y:Nat,plus(?z:Nat,?y:Nat)))) = times(plus(s(s(0)),0),0)",
"s(times(?z:Nat,times(?z:Nat,plus(?x:Nat,0)))) = times(times(times(?y:Nat,0),?z:Nat),?x:Nat)",
"plus(?x:Nat,times(0,s(times(0,?z:Nat)))) = s(plus(0,times(?x:Nat,s(?z:Nat))))",
"plus(?y:Nat,times(s(times(?x:Nat,?y:Nat)),0)) = plus(s(?y:Nat),s(s(s(?x:Nat))))",
"plus(plus(s(plus(?z:Nat,?x:Nat)),?x:Nat),?z:Nat) = times(plus(?x:Nat,?x:Nat),plus(?x:Nat,0))",
"times(?x:Nat,s(plus(s(?x:Nat),s(?z:Nat)))) = plus(0,s(s(plus(?x:Nat,?y:Nat))))",
"times(?x:Nat,s(plus(plus(0,?x:Nat),?x:Nat))) = times(s(times(s(?y:Nat),?y:Nat)),?y:Nat)",
"times(?x:Nat,plus(?z:Nat,s(plus(?x:Nat,0)))) = s(plus(?z:Nat,times(s(?z:Nat),0)))",
"times(?x:Nat,times(s(plus(?x:Nat,?y:Nat)),?y:Nat)) = times(s(times(s(?x:Nat),0)),0)",
"times(0,times(0,times(s(?x:Nat),?x:Nat))) = s(plus(s(?x:Nat),s(s(?z:Nat))))",
"s(times(plus(times(?z:Nat,?x:Nat),s(?y:Nat)),?x:Nat)) = plus(s(plus(?y:Nat,0)),0)",
"plus(plus(times(0,times(?z:Nat,0)),?z:Nat),?y:Nat) = s(s(times(s(0),?z:Nat)))",
"times(?x:Nat,plus(times(0,times(?x:Nat,?x:Nat)),?x:Nat)) = plus(s(0),plus(0,?x:Nat))",
"plus(s(?y:Nat),plus(?z:Nat,times(?x:Nat,plus(?x:Nat,?x:Nat)))) = plus(times(0,?x:Nat),0)",
"plus(s(plus(?x:Nat,?z:Nat)),s(times(?y:Nat,s(0)))) = s(plus(s(?z:Nat),?z:Nat))",
"plus(plus(?y:Nat,times(s(?x:Nat),?z:Nat)),times(?x:Nat,0)) = times(plus(?x:Nat,0),?z:Nat)",
"times(?x:Nat,plus(plus(?x:Nat,times(?y:Nat,?y:Nat)),s(?y:Nat))) = plus(times(0,?z:Nat),?x:Nat)",
"times(s(0),times(?x:Nat,times(times(?x:Nat,?x:Nat),?z:Nat))) = times(s(?x:Nat),s(?x:Nat))",
"times(s(plus(plus(0,?x:Nat),plus(?x:Nat,?x:Nat))),?x:Nat) = s(times(s(?y:Nat),?x:Nat))",
"times(plus(0,?x:Nat),times(?x:Nat,s(plus(?y:Nat,?x:Nat)))) = times(plus(?x:Nat,?y:Nat),?z:Nat)",
"times(plus(0,times(s(?z:Nat),?x:Nat)),s(s(?z:Nat))) = times(?x:Nat,times(?x:Nat,?x:Nat))",
"times(times(?y:Nat,times(times(0,0),?x:Nat)),s(?x:Nat)) = plus(plus(?x:Nat,?y:Nat),?x:Nat)",
"plus(?y:Nat,times(times(times(?x:Nat,0),plus(?z:Nat,?z:Nat)),?z:Nat)) = s(times(0,?x:Nat))",
"plus(plus(plus(times(0,?y:Nat),plus(0,0)),?x:Nat),?y:Nat) = times(s(?x:Nat),?z:Nat)",
"plus(times(?x:Nat,plus(times(0,?y:Nat),s(?y:Nat))),s(?x:Nat)) = plus(?x:Nat,s(?x:Nat))",
"times(plus(?x:Nat,?z:Nat),times(plus(times(0,0),?x:Nat),?x:Nat)) = plus(s(?x:Nat),?x:Nat)",
"times(plus(s(s(0)),?y:Nat),plus(times(?x:Nat,?z:Nat),0)) = plus(?x:Nat,s(0))",
"times(plus(s(times(?y:Nat,?x:Nat)),0),times(s(0),?y:Nat)) = s(plus(?x:Nat,?x:Nat))",
"plus(s(s(s(0))),times(?x:Nat,times(s(?x:Nat),s(?x:Nat)))) = plus(?x:Nat,?y:Nat)",
"plus(s(plus(?x:Nat,s(0))),s(plus(0,times(?x:Nat,?y:Nat)))) = times(?y:Nat,0)",
"plus(s(times(s(?x:Nat),0)),times(plus(s(?x:Nat),?x:Nat),?z:Nat)) = times(?x:Nat,?y:Nat)",
"plus(plus(plus(?x:Nat,plus(?z:Nat,?x:Nat)),times(s(?z:Nat),?y:Nat)),?z:Nat) = plus(?z:Nat,0)",
"plus(plus(plus(?z:Nat,s(0)),?x:Nat),plus(?x:Nat,plus(?y:Nat,0))) = times(0,?y:Nat)",
"plus(times(s(times(0,?z:Nat)),s(?x:Nat)),times(s(?y:Nat),?z:Nat)) = plus(?y:Nat,0)",
"plus(times(times(times(0,?z:Nat),?x:Nat),plus(?y:Nat,s(?x:Nat))),0) = times(?z:Nat,?x:Nat)",
"times(times(0,times(s(?z:Nat),times(?y:Nat,?z:Nat))),times(?x:Nat,?x:Nat)) = plus(?x:Nat,?z:Nat)",
"plus(?x:Nat,plus(times(?z:Nat,times(?y:Nat,?y:Nat)),plus(plus(?y:Nat,?y:Nat),?x:Nat))) = s(0)",
"plus(0,plus(times(s(0),0),times(s(?x:Nat),times(0,?y:Nat)))) = s(?x:Nat)",
"plus(s(?y:Nat),plus(s(plus(0,?z:Nat)),times(times(?y:Nat,?y:Nat),?x:Nat))) = s(0)",
"plus(s(times(plus(?x:Nat,?x:Nat),0)),plus(0,plus(s(?y:Nat),?x:Nat))) = s(?x:Nat)",
"plus(times(times(?x:Nat,plus(0,?y:Nat)),times(s(?x:Nat),?z:Nat)),s(?x:Nat)) = s(?x:Nat)",
"times(?y:Nat,times(plus(plus(?x:Nat,?x:Nat),?x:Nat),times(times(0,0),?y:Nat))) = s(?y:Nat)",
"times(s(s(times(?z:Nat,?y:Nat))),times(plus(?x:Nat,plus(0,?x:Nat)),?y:Nat)) = s(0)",
"times(s(times(times(?x:Nat,?x:Nat),s(?z:Nat))),plus(s(s(0)),?x:Nat)) = s(?x:Nat)",
"times(plus(?y:Nat,times(plus(?z:Nat,?x:Nat),0)),s(plus(?y:Nat,s(?x:Nat)))) = s(?z:Nat)",
"times(plus(times(?x:Nat,0),times(times(?x:Nat,?x:Nat),plus(?x:Nat,?z:Nat))),0) = s(?y:Nat)",
"s(?x:Nat) = plus(times(times(?z:Nat,plus(?x:Nat,?x:Nat)),?y:Nat),times(s(times(?x:Nat,?y:Nat)),0))",
"s(?x:Nat) = times(s(times(?x:Nat,0)),plus(plus(0,times(?z:Nat,?z:Nat)),plus(?x:Nat,0)))",
"s(?x:Nat) = times(plus(?x:Nat,s(plus(?x:Nat,0))),plus(plus(times(?x:Nat,?y:Nat),0),0))",
"s(?x:Nat) = times(times(s(plus(0,?y:Nat)),times(plus(?z:Nat,?x:Nat),times(?x:Nat,?x:Nat))),?x:Nat)",
"s(?y:Nat) = plus(times(plus(0,?z:Nat),plus(plus(?y:Nat,0),s(0))),plus(?y:Nat,?z:Nat))",
"s(?z:Nat) = plus(plus(s(plus(?y:Nat,?x:Nat)),0),plus(times(s(?y:Nat),s(?y:Nat)),?x:Nat))",
"s(?z:Nat) = times(plus(times(plus(?x:Nat,0),?x:Nat),s(plus(?x:Nat,?x:Nat))),plus(?z:Nat,?x:Nat))",
"s(0) = plus(?x:Nat,plus(times(times(0,?x:Nat),times(?z:Nat,?z:Nat)),s(plus(?z:Nat,?z:Nat))))",
"s(0) = plus(times(times(plus(?z:Nat,0),?z:Nat),plus(s(?x:Nat),?y:Nat)),s(s(?x:Nat)))",
"s(s(?y:Nat)) = times(plus(s(0),?x:Nat),s(times(plus(?y:Nat,?x:Nat),plus(?x:Nat,?z:Nat))))",
"plus(?x:Nat,?x:Nat) = s(times(times(s(?x:Nat),plus(?z:Nat,?x:Nat)),times(s(0),s(?z:Nat))))",
"plus(?y:Nat,?x:Nat) = times(times(s(?y:Nat),plus(?z:Nat,s(?z:Nat))),plus(?x:Nat,times(0,?z:Nat)))",
"times(?y:Nat,?y:Nat) = times(plus(?z:Nat,0),plus(s(plus(?x:Nat,?z:Nat)),s(plus(0,0))))",
"times(?z:Nat,?y:Nat) = plus(times(?z:Nat,s(?y:Nat)),s(times(plus(?x:Nat,?x:Nat),plus(?x:Nat,0))))",
"times(?z:Nat,?z:Nat) = plus(plus(?y:Nat,s(times(?x:Nat,?x:Nat))),times(times(s(0),?x:Nat),?x:Nat))",
"s(plus(?x:Nat,?x:Nat)) = plus(times(?x:Nat,?x:Nat),times(?z:Nat,plus(s(?z:Nat),plus(?x:Nat,0))))",
"plus(s(?z:Nat),?y:Nat) = plus(plus(times(times(?x:Nat,?z:Nat),0),times(?x:Nat,0)),s(?z:Nat))",
"times(s(?x:Nat),0) = plus(times(plus(?y:Nat,?z:Nat),times(s(?x:Nat),?x:Nat)),times(?x:Nat,?x:Nat))",
"times(s(0),?z:Nat) = plus(plus(s(plus(?z:Nat,?z:Nat)),times(plus(?y:Nat,0),?x:Nat)),0)",
"plus(?x:Nat,times(?z:Nat,?z:Nat)) = plus(?x:Nat,times(plus(plus(?x:Nat,?x:Nat),s(?y:Nat)),s(0)))",
"plus(?z:Nat,plus(0,?x:Nat)) = plus(plus(times(plus(?z:Nat,0),0),times(?y:Nat,?x:Nat)),0)",
"s(plus(plus(?y:Nat,?x:Nat),?z:Nat)) = plus(?x:Nat,plus(?z:Nat,plus(s(?x:Nat),times(?z:Nat,0))))",
"plus(s(0),s(s(?y:Nat))) = plus(?z:Nat,times(plus(s(0),times(?x:Nat,?x:Nat)),?x:Nat))",
"plus(plus(?x:Nat,?x:Nat),s(0)) = times(0,times(s(times(?y:Nat,?x:Nat)),times(0,?x:Nat)))",
"plus(plus(s(?x:Nat),?y:Nat),?y:Nat) = times(times(0,times(?x:Nat,?y:Nat)),plus(?z:Nat,s(?x:Nat)))",
"times(plus(?y:Nat,s(0)),?x:Nat) = s(times(?x:Nat,plus(times(?x:Nat,?x:Nat),times(?y:Nat,?y:Nat))))",
"plus(?y:Nat,s(plus(?z:Nat,s(0)))) = s(s(times(plus(?x:Nat,?x:Nat),times(0,0))))",
"times(?x:Nat,s(plus(0,s(?z:Nat)))) = s(plus(0,times(plus(?x:Nat,?y:Nat),s(?z:Nat))))",
"times(?x:Nat,plus(?x:Nat,s(s(?x:Nat)))) = plus(plus(?x:Nat,times(?z:Nat,plus(0,?x:Nat))),?y:Nat)",
"times(?z:Nat,times(times(?x:Nat,?x:Nat),0)) = plus(plus(?x:Nat,times(?x:Nat,?y:Nat)),s(s(0)))",
"times(s(plus(?x:Nat,s(?x:Nat))),0) = plus(times(plus(?x:Nat,plus(0,?y:Nat)),?z:Nat),?x:Nat)",
"s(plus(plus(?y:Nat,times(?x:Nat,0)),?x:Nat)) = times(?x:Nat,plus(s(plus(?x:Nat,?x:Nat)),?x:Nat))",
"s(times(plus(?z:Nat,s(?z:Nat)),s(?z:Nat))) = times(?y:Nat,s(plus(plus(?z:Nat,0),?z:Nat)))",
"plus(s(?x:Nat),s(plus(?x:Nat,s(?z:Nat)))) = plus(times(?x:Nat,s(plus(0,?z:Nat))),?z:Nat)",
"plus(s(0),plus(times(?y:Nat,?y:Nat),0)) = plus(plus(times(?x:Nat,s(?x:Nat)),?x:Nat),?x:Nat)",
"plus(times(s(0),?x:Nat),times(?x:Nat,?z:Nat)) = s(plus(s(s(?z:Nat)),times(?z:Nat,?x:Nat)))",
"plus(times(s(0),plus(?y:Nat,?z:Nat)),?x:Nat) = times(?x:Nat,times(s(times(0,?x:Nat)),?y:Nat))",
"plus(times(s(plus(?y:Nat,?y:Nat)),?y:Nat),0) = times(plus(0,plus(s(0),?y:Nat)),?z:Nat)",
"times(?y:Nat,plus(plus(?x:Nat,s(0)),?y:Nat)) = plus(s(times(?y:Nat,plus(?y:Nat,?z:Nat))),?x:Nat)",
"times(0,s(plus(plus(?y:Nat,?x:Nat),?x:Nat))) = s(times(?x:Nat,times(?x:Nat,times(?y:Nat,?z:Nat))))",
"times(0,s(plus(plus(?z:Nat,?z:Nat),?y:Nat))) = s(plus(0,plus(?y:Nat,times(?x:Nat,0))))",
"times(s(plus(plus(?z:Nat,?y:Nat),0)),0) = plus(plus(times(?y:Nat,s(?z:Nat)),?y:Nat),?z:Nat)",
"times(plus(?z:Nat,s(times(?x:Nat,?x:Nat))),0) = plus(plus(?x:Nat,s(times(?y:Nat,0))),?x:Nat)",
"times(plus(?z:Nat,times(s(?y:Nat),?x:Nat)),0) = plus(?y:Nat,plus(?x:Nat,times(s(0),?y:Nat)))",
"times(times(times(?y:Nat,s(?z:Nat)),?z:Nat),?x:Nat) = times(s(plus(plus(?x:Nat,0),?x:Nat)),?y:Nat)",
"plus(?x:Nat,plus(?x:Nat,plus(?x:Nat,plus(?y:Nat,?x:Nat)))) = times(?x:Nat,plus(0,times(0,?z:Nat)))",
"plus(0,plus(plus(s(?y:Nat),?y:Nat),s(?x:Nat))) = times(?x:Nat,times(plus(?z:Nat,0),?x:Nat))",
"plus(times(?x:Nat,times(?z:Nat,plus(0,?x:Nat))),?z:Nat) = s(plus(plus(?z:Nat,s(?y:Nat)),?z:Nat))",
"times(?x:Nat,s(plus(plus(?x:Nat,?z:Nat),s(?x:Nat)))) = times(0,times(plus(?z:Nat,?x:Nat),?z:Nat))",
"times(?y:Nat,plus(times(plus(0,0),?x:Nat),?x:Nat)) = times(?z:Nat,s(s(times(?x:Nat,?z:Nat))))",
"times(?z:Nat,plus(0,plus(?z:Nat,plus(?y:Nat,0)))) = plus(?x:Nat,s(plus(s(?z:Nat),?z:Nat)))",
"times(times(s(?x:Nat),s(?y:Nat)),s(s(?z:Nat))) = plus(0,s(s(plus(0,?x:Nat))))",
"plus(s(?y:Nat),plus(0,plus(?y:Nat,times(?x:Nat,0)))) = s(plus(?z:Nat,times(?x:Nat,?x:Nat)))",
"plus(plus(0,plus(?z:Nat,?x:Nat)),plus(s(?x:Nat),?y:Nat)) = times(?x:Nat,plus(?x:Nat,s(0)))",
"plus(plus(0,plus(0,s(0))),s(s(?y:Nat))) = plus(times(s(?z:Nat),0),?z:Nat)",
"plus(plus(times(0,times(?x:Nat,?z:Nat)),s(?x:Nat)),?y:Nat) = times(?z:Nat,plus(s(0),?x:Nat))",
"plus(times(?z:Nat,times(s(?y:Nat),plus(?z:Nat,?y:Nat))),?y:Nat) = times(plus(s(?x:Nat),?x:Nat),0)",
"plus(times(plus(times(?y:Nat,?x:Nat),s(?y:Nat)),?x:Nat),0) = times(s(plus(0,0)),?z:Nat)",
"times(?x:Nat,times(times(times(0,?x:Nat),s(?x:Nat)),?z:Nat)) = times(?x:Nat,plus(s(0),?y:Nat))",
"times(times(times(times(0,?x:Nat),?x:Nat),?x:Nat),s(?z:Nat)) = times(plus(?z:Nat,0),s(0))",
"plus(s(?x:Nat),times(times(s(?x:Nat),times(?y:Nat,?x:Nat)),?x:Nat)) = times(0,plus(0,?y:Nat))",
"plus(times(times(?z:Nat,?x:Nat),?x:Nat),plus(0,s(s(?z:Nat)))) = plus(plus(?z:Nat,?x:Nat),0)",
"times(s(?y:Nat),s(plus(plus(0,?y:Nat),times(?x:Nat,?z:Nat)))) = times(?z:Nat,plus(0,?y:Nat))",
"times(plus(?x:Nat,?x:Nat),times(plus(?y:Nat,times(0,0)),?y:Nat)) = plus(0,plus(?x:Nat,?y:Nat))",
"times(plus(s(?x:Nat),0),times(plus(?x:Nat,s(0)),?z:Nat)) = s(plus(0,s(?z:Nat)))",
"plus(s(plus(?z:Nat,s(0))),times(plus(0,s(?x:Nat)),0)) = times(?z:Nat,s(0))",
"plus(plus(0,s(s(?z:Nat))),times(0,s(times(?x:Nat,0)))) = times(s(?x:Nat),?z:Nat)",
"plus(plus(times(times(?x:Nat,?y:Nat),?x:Nat),times(s(?y:Nat),?x:Nat)),?x:Nat) = s(plus(?x:Nat,0))",
"times(?z:Nat,plus(plus(s(?z:Nat),plus(?z:Nat,?x:Nat)),plus(?y:Nat,0))) = plus(s(?y:Nat),?x:Nat)",
"times(s(?y:Nat),plus(plus(?z:Nat,0),times(?x:Nat,times(0,?z:Nat)))) = s(times(?x:Nat,?x:Nat))",
"times(times(0,plus(?x:Nat,?x:Nat)),s(plus(plus(?x:Nat,?x:Nat),?x:Nat))) = s(plus(0,?x:Nat))",
"times(times(s(?y:Nat),0),s(times(times(?x:Nat,?x:Nat),s(?y:Nat)))) = s(times(0,?x:Nat))",
"times(times(times(plus(?z:Nat,0),times(?x:Nat,?x:Nat)),?y:Nat),s(0)) = times(?z:Nat,s(?x:Nat))",
"s(plus(plus(?z:Nat,plus(?z:Nat,0)),times(s(?x:Nat),plus(?y:Nat,?x:Nat)))) = plus(0,?x:Nat)",
"s(plus(times(?y:Nat,plus(?z:Nat,?z:Nat)),plus(plus(?x:Nat,?y:Nat),s(?y:Nat)))) = times(0,?x:Nat)",
"times(?y:Nat,plus(plus(?x:Nat,plus(0,?x:Nat)),times(plus(?x:Nat,?x:Nat),?x:Nat))) = plus(?z:Nat,?x:Nat)",
"times(s(?x:Nat),times(times(?x:Nat,s(?x:Nat)),plus(?z:Nat,times(0,?z:Nat)))) = plus(?z:Nat,?x:Nat)",
"times(plus(?x:Nat,?y:Nat),times(?y:Nat,times(plus(?x:Nat,?y:Nat),plus(?x:Nat,0)))) = s(s(?x:Nat))",
"times(plus(s(times(?x:Nat,?x:Nat)),times(s(?z:Nat),plus(?y:Nat,?x:Nat))),?x:Nat) = plus(?x:Nat,0)",
"times(plus(times(plus(0,?x:Nat),plus(0,?z:Nat)),?z:Nat),times(?x:Nat,?x:Nat)) = plus(?y:Nat,?z:Nat)",
"plus(plus(?y:Nat,s(?x:Nat)),plus(times(plus(?z:Nat,?x:Nat),times(?x:Nat,0)),?x:Nat)) = s(?y:Nat)",
"plus(plus(plus(s(?z:Nat),?y:Nat),?x:Nat),times(plus(?z:Nat,0),s(s(0)))) = s(?x:Nat)",
"plus(times(?y:Nat,plus(times(?x:Nat,?y:Nat),plus(0,?x:Nat))),times(s(?y:Nat),?x:Nat)) = s(?y:Nat)",
"plus(times(plus(0,?x:Nat),plus(?x:Nat,plus(?x:Nat,0))),s(times(?x:Nat,?y:Nat))) = s(?x:Nat)",
"s(?x:Nat) = plus(s(times(?x:Nat,times(0,?x:Nat))),plus(s(?z:Nat),plus(plus(?x:Nat,?z:Nat),?z:Nat)))",
"s(?y:Nat) = times(plus(?y:Nat,times(plus(?x:Nat,?x:Nat),s(?z:Nat))),times(?x:Nat,times(s(0),?z:Nat)))",
"s(?y:Nat) = times(times(s(times(?x:Nat,?z:Nat)),times(times(?x:Nat,0),?x:Nat)),s(s(s(?x:Nat))))",
"s(?z:Nat) = times(times(times(s(?x:Nat),?y:Nat),0),s(times(times(?z:Nat,?y:Nat),plus(?y:Nat,0))))",
"s(0) = plus(plus(0,?z:Nat),plus(plus(s(?z:Nat),s(?x:Nat)),times(?x:Nat,times(?y:Nat,?y:Nat))))",
"plus(?z:Nat,?x:Nat) = plus(plus(times(plus(?y:Nat,?x:Nat),0),?y:Nat),s(plus(?y:Nat,plus(?x:Nat,?x:Nat))))",
"times(?y:Nat,0) = plus(s(?x:Nat),plus(s(plus(0,?y:Nat)),plus(s(?x:Nat),times(?x:Nat,?z:Nat))))",
"s(plus(0,0)) = plus(times(times(plus(?x:Nat,?y:Nat),times(?x:Nat,?x:Nat)),?y:Nat),plus(?z:Nat,0))",
"plus(0,s(?x:Nat)) = times(times(s(times(?x:Nat,?x:Nat)),plus(s(?z:Nat),times(?y:Nat,?y:Nat))),?z:Nat)",
"times(s(?x:Nat),?z:Nat) = times(plus(plus(s(?y:Nat),plus(?x:Nat,?x:Nat)),s(?x:Nat)),times(0,0))",
"s(times(s(?z:Nat),?x:Nat)) = plus(times(?y:Nat,0),s(times(plus(?z:Nat,?z:Nat),plus(?z:Nat,?x:Nat))))",
"plus(s(s(0)),0) = plus(times(times(?x:Nat,?x:Nat),times(?x:Nat,plus(?x:Nat,0))),s(?x:Nat))",
"times(?x:Nat,s(s(?x:Nat))) = times(s(plus(s(0),?y:Nat)),s(times(s(?x:Nat),s(?y:Nat))))",
"times(?x:Nat,plus(0,0)) = plus(0,plus(times(s(?x:Nat),?x:Nat),times(plus(?x:Nat,?x:Nat),?z:Nat)))",
"times(0,plus(?x:Nat,?y:Nat)) = plus(plus(s(?x:Nat),s(0)),s(times(?y:Nat,plus(?z:Nat,?x:Nat))))",
"times(plus(0,?x:Nat),?x:Nat) = plus(0,plus(plus(?z:Nat,times(?x:Nat,?z:Nat)),s(times(?x:Nat,0))))",
"s(plus(?z:Nat,plus(?z:Nat,?y:Nat))) = plus(times(?y:Nat,s(?z:Nat)),times(s(0),times(?z:Nat,0)))",
"times(?x:Nat,times(?z:Nat,s(0))) = times(?x:Nat,plus(times(plus(0,?x:Nat),plus(?x:Nat,?y:Nat)),?x:Nat))",
"s(s(times(plus(?z:Nat,?x:Nat),?z:Nat))) = times(?z:Nat,plus(0,times(plus(?x:Nat,?z:Nat),s(?z:Nat))))",
"s(times(?z:Nat,times(0,s(?y:Nat)))) = plus(times(?x:Nat,s(?x:Nat)),s(s(plus(?z:Nat,0))))",
"times(?x:Nat,s(s(plus(?x:Nat,?y:Nat)))) = s(times(0,plus(times(0,?y:Nat),plus(?x:Nat,?y:Nat))))",
"times(s(s(plus(?x:Nat,?y:Nat))),?x:Nat) = times(?y:Nat,plus(0,plus(times(?z:Nat,?z:Nat),s(?z:Nat))))",
"s(s(times(s(?x:Nat),plus(0,?x:Nat)))) = plus(?y:Nat,s(times(s(0),plus(?y:Nat,?x:Nat))))",
"plus(s(?x:Nat),plus(s(s(0)),?x:Nat)) = times(plus(s(s(?x:Nat)),times(?y:Nat,?x:Nat)),?z:Nat)",
"plus(s(times(plus(?z:Nat,0),?y:Nat)),?x:Nat) = plus(times(plus(times(?x:Nat,?z:Nat),?z:Nat),?z:Nat),?x:Nat)",
"times(0,s(plus(0,plus(0,?y:Nat)))) = s(times(plus(plus(?x:Nat,0),s(?x:Nat)),?y:Nat))",
"times(plus(0,s(times(?z:Nat,?y:Nat))),?x:Nat) = plus(times(plus(?y:Nat,plus(0,0)),0),?y:Nat)",
"s(plus(s(plus(0,?x:Nat)),times(0,?x:Nat))) = times(s(0),plus(times(?x:Nat,?x:Nat),0))",
"s(plus(plus(plus(0,0),s(?x:Nat)),?z:Nat)) = plus(s(s(?x:Nat)),times(s(?x:Nat),?x:Nat))",
"s(times(times(?z:Nat,times(?x:Nat,?x:Nat)),s(0))) = plus(?y:Nat,s(times(?x:Nat,plus(?x:Nat,0))))",
"plus(?x:Nat,plus(times(?x:Nat,?x:Nat),plus(?y:Nat,0))) = plus(times(?x:Nat,times(?x:Nat,s(0))),?z:Nat)",
"times(?x:Nat,plus(s(s(?z:Nat)),times(?z:Nat,?y:Nat))) = times(?z:Nat,times(?x:Nat,times(0,s(?x:Nat))))",
"times(s(plus(times(?z:Nat,?z:Nat),0)),s(0)) = s(plus(times(0,times(?y:Nat,0)),?z:Nat))",
"s(times(s(plus(?z:Nat,?z:Nat)),s(times(?x:Nat,?y:Nat)))) = times(0,times(plus(?y:Nat,?x:Nat),?x:Nat))",
"plus(plus(s(s(?z:Nat)),s(0)),times(?x:Nat,?y:Nat)) = plus(?x:Nat,times(s(?y:Nat),s(0)))",
"plus(plus(times(?x:Nat,plus(?x:Nat,?x:Nat)),s(?x:Nat)),?x:Nat) = s(plus(0,plus(?x:Nat,s(?y:Nat))))",
"plus(plus(times(s(?z:Nat),s(?y:Nat)),s(?x:Nat)),?x:Nat) = s(times(s(times(?x:Nat,0)),?x:Nat))",
"times(plus(times(?y:Nat,?z:Nat),plus(?z:Nat,s(?z:Nat))),?x:Nat) = times(0,s(s(times(?y:Nat,?x:Nat))))",
"times(times(?z:Nat,s(?y:Nat)),times(?z:Nat,s(s(?y:Nat)))) = times(times(?y:Nat,plus(0,?x:Nat)),?x:Nat)",
"plus(plus(0,plus(plus(?y:Nat,0),0)),times(?x:Nat,?z:Nat)) = s(plus(s(?y:Nat),s(0)))",
"plus(plus(times(0,s(?x:Nat)),plus(s(?z:Nat),?z:Nat)),?z:Nat) = times(?y:Nat,plus(s(?x:Nat),?z:Nat))",
"plus(times(0,plus(?y:Nat,times(0,?y:Nat))),times(?z:Nat,?x:Nat)) = s(times(?y:Nat,s(s(?x:Nat))))",
"plus(times(s(plus(?x:Nat,?x:Nat)),s(times(?z:Nat,?z:Nat))),?x:Nat) = s(s(s(times(0,?z:Nat))))",
"plus(times(times(plus(?x:Nat,?x:Nat),?y:Nat),0),times(?x:Nat,?z:Nat)) = plus(times(?z:Nat,?y:Nat),s(?y:Nat))",
"times(?x:Nat,times(?x:Nat,times(times(?y:Nat,?z:Nat),plus(?x:Nat,?x:Nat)))) = plus(s(plus(0,?z:Nat)),?x:Nat)",
"times(s(times(s(0),times(?x:Nat,?x:Nat))),plus(0,?x:Nat)) = plus(?y:Nat,s(plus(?x:Nat,0)))",
"times(plus(plus(?x:Nat,times(?x:Nat,?z:Nat)),?x:Nat),s(s(0))) = plus(0,plus(?y:Nat,s(0)))",
"times(plus(plus(?y:Nat,?z:Nat),times(0,?y:Nat)),plus(0,?x:Nat)) = plus(s(?x:Nat),times(?x:Nat,?x:Nat))",
"times(plus(times(?z:Nat,?x:Nat),plus(times(?x:Nat,?x:Nat),?y:Nat)),?x:Nat) = s(plus(0,times(?z:Nat,0)))",
"times(times(?y:Nat,times(?x:Nat,0)),times(s(s(0)),?x:Nat)) = s(s(times(?x:Nat,s(?x:Nat))))",
"s(plus(times(?z:Nat,times(?y:Nat,?x:Nat)),times(plus(?y:Nat,?x:Nat),0))) = times(?x:Nat,plus(0,0))",
"plus(plus(plus(0,?z:Nat),plus(?z:Nat,times(?z:Nat,?z:Nat))),s(?x:Nat)) = s(plus(s(?z:Nat),0))",
"plus(times(?x:Nat,s(?y:Nat)),plus(?x:Nat,plus(plus(?y:Nat,?x:Nat),?x:Nat))) = s(s(plus(?y:Nat,0)))",
"plus(times(plus(s(?x:Nat),plus(?x:Nat,?z:Nat)),s(s(0))),?x:Nat) = plus(plus(?x:Nat,?y:Nat),0)",
"times(plus(?y:Nat,?x:Nat),plus(?x:Nat,plus(plus(0,?y:Nat),s(0)))) = plus(0,plus(?z:Nat,?x:Nat))",
"times(plus(s(?y:Nat),times(plus(?x:Nat,?y:Nat),times(?x:Nat,?y:Nat))),0) = times(?y:Nat,plus(?x:Nat,?y:Nat))",
"times(plus(times(times(?x:Nat,0),times(?z:Nat,?y:Nat)),?y:Nat),s(0)) = plus(times(?x:Nat,?x:Nat),?x:Nat)",
"plus(plus(?y:Nat,0),times(times(times(0,?y:Nat),times(?y:Nat,?y:Nat)),?y:Nat)) = s(plus(?z:Nat,?x:Nat))",
"plus(times(plus(times(?x:Nat,?y:Nat),?z:Nat),times(plus(?z:Nat,?x:Nat),?x:Nat)),?x:Nat) = times(s(0),?z:Nat)",
"times(s(s(times(0,?x:Nat))),s(plus(times(?y:Nat,?y:Nat),s(?x:Nat)))) = s(times(?x:Nat,?x:Nat))",
"times(plus(s(0),times(?y:Nat,?x:Nat)),times(?z:Nat,plus(0,s(?z:Nat)))) = s(times(?y:Nat,?x:Nat))",
"times(plus(times(plus(?x:Nat,?x:Nat),?z:Nat),plus(?x:Nat,?z:Nat)),plus(?x:Nat,?x:Nat)) = s(plus(0,0))",
"s(times(plus(s(?x:Nat),plus(?z:Nat,?y:Nat)),times(times(?y:Nat,?x:Nat),s(?x:Nat)))) = times(?z:Nat,0)",
"plus(times(?x:Nat,s(plus(0,?y:Nat))),plus(s(?y:Nat),plus(s(0),0))) = times(?z:Nat,?x:Nat)",
"plus(times(s(times(0,0)),?z:Nat),s(times(s(0),times(?y:Nat,0)))) = s(s(?x:Nat))",
"times(s(plus(?x:Nat,?x:Nat)),times(?z:Nat,times(plus(0,?x:Nat),plus(0,?z:Nat)))) = plus(?x:Nat,?z:Nat)",
"times(plus(s(?y:Nat),times(s(?y:Nat),0)),times(?x:Nat,s(times(?x:Nat,?x:Nat)))) = plus(?y:Nat,?x:Nat)",
"times(times(0,s(0)),plus(times(?y:Nat,plus(?z:Nat,?x:Nat)),plus(?x:Nat,?z:Nat))) = times(?y:Nat,?y:Nat)",
"times(times(s(?y:Nat),?z:Nat),times(s(plus(?x:Nat,?x:Nat)),s(plus(0,0)))) = s(s(?y:Nat))",
"times(plus(plus(s(?y:Nat),?x:Nat),0),plus(s(?x:Nat),plus(s(?x:Nat),s(0)))) = s(?x:Nat)",
"times(times(times(?y:Nat,0),?z:Nat),times(?x:Nat,times(times(?z:Nat,?z:Nat),times(?z:Nat,?y:Nat)))) = s(?x:Nat)",
"s(?x:Nat) = plus(s(plus(?z:Nat,plus(?x:Nat,?x:Nat))),plus(plus(s(?y:Nat),?x:Nat),s(times(?x:Nat,0))))",
"s(?y:Nat) = plus(?y:Nat,times(plus(times(0,?x:Nat),s(?x:Nat)),plus(plus(?y:Nat,?z:Nat),plus(?x:Nat,?x:Nat))))",
"s(0) = plus(plus(0,s(times(?x:Nat,?y:Nat))),plus(?x:Nat,plus(plus(?y:Nat,?x:Nat),times(0,?x:Nat))))",
"plus(?x:Nat,?x:Nat) = plus(s(times(0,s(0))),times(plus(?x:Nat,plus(?x:Nat,?z:Nat)),s(s(0))))",
"plus(?x:Nat,?x:Nat) = times(plus(times(plus(?z:Nat,?x:Nat),plus(?x:Nat,?z:Nat)),times(0,s(0))),s(0))",
"plus(?y:Nat,?z:Nat) = plus(times(s(s(?x:Nat)),times(?y:Nat,?x:Nat)),s(plus(s(?x:Nat),times(?x:Nat,0))))",
"times(?x:Nat,?y:Nat) = plus(?y:Nat,plus(times(plus(?x:Nat,?y:Nat),times(?x:Nat,?y:Nat)),times(times(0,?x:Nat),0)))",
"times(?x:Nat,s(0)) = times(?z:Nat,times(times(s(?x:Nat),0),plus(times(?z:Nat,?x:Nat),times(?z:Nat,?x:Nat))))",
"times(s(?x:Nat),?x:Nat) = plus(plus(times(s(0),plus(?x:Nat,?y:Nat)),times(0,?y:Nat)),s(s(0)))",
"times(s(?x:Nat),?x:Nat) = times(plus(s(?y:Nat),plus(times(0,?x:Nat),times(?x:Nat,?y:Nat))),plus(?z:Nat,?x:Nat))",
"s(times(s(0),0)) = plus(?x:Nat,times(times(?y:Nat,0),times(plus(?x:Nat,?y:Nat),plus(?x:Nat,?y:Nat))))",
"plus(s(?y:Nat),s(?x:Nat)) = plus(times(?x:Nat,times(?y:Nat,s(?y:Nat))),s(times(times(0,0),?x:Nat)))",
"plus(plus(0,?y:Nat),?y:Nat) = plus(times(?x:Nat,s(s(0))),times(plus(?x:Nat,plus(?z:Nat,?y:Nat)),?z:Nat))",
"times(?x:Nat,times(?x:Nat,?x:Nat)) = s(times(times(times(?y:Nat,0),s(?x:Nat)),plus(times(?y:Nat,?x:Nat),0)))",
"s(times(s(s(0)),?x:Nat)) = times(plus(?y:Nat,0),plus(s(?x:Nat),times(plus(?y:Nat,?z:Nat),?y:Nat)))",
"plus(plus(0,s(?y:Nat)),?z:Nat) = times(plus(times(plus(?z:Nat,?x:Nat),?x:Nat),?x:Nat),plus(s(?x:Nat),?z:Nat))",
"times(s(s(?x:Nat)),s(0)) = plus(plus(s(?x:Nat),plus(plus(0,?x:Nat),plus(?x:Nat,?x:Nat))),?x:Nat)",
"times(plus(s(0),?x:Nat),?y:Nat) = times(times(times(s(?z:Nat),plus(?x:Nat,0)),?x:Nat),times(0,?x:Nat))",
"times(times(s(?x:Nat),?y:Nat),?z:Nat) = plus(s(plus(?y:Nat,times(0,?y:Nat))),s(times(?x:Nat,s(?z:Nat))))",
"plus(0,s(s(plus(0,?x:Nat)))) = plus(plus(s(?z:Nat),times(plus(?x:Nat,?y:Nat),s(?x:Nat))),?x:Nat)",
"plus(s(plus(?x:Nat,s(?x:Nat))),?y:Nat) = plus(times(s(0),?y:Nat),s(plus(?z:Nat,times(?z:Nat,?y:Nat))))",
"plus(?z:Nat,s(times(times(?x:Nat,?x:Nat),0))) = plus(s(0),s(times(times(?x:Nat,?y:Nat),s(?x:Nat))))",
"plus(times(?x:Nat,s(?x:Nat)),times(?x:Nat,0)) = s(plus(plus(?x:Nat,s(?z:Nat)),s(times(0,?x:Nat))))",
"times(?x:Nat,s(plus(plus(?x:Nat,?x:Nat),?z:Nat))) = times(plus(plus(0,?y:Nat),?x:Nat),plus(s(0),?x:Nat))",
"times(plus(times(?x:Nat,s(?x:Nat)),?z:Nat),0) = times(?y:Nat,s(times(times(0,?x:Nat),plus(?y:Nat,?y:Nat))))",
"times(plus(times(s(0),?x:Nat),?z:Nat),?x:Nat) = s(plus(plus(s(?x:Nat),?z:Nat),times(s(?z:Nat),?x:Nat)))",
"s(s(times(plus(0,?x:Nat),plus(0,?z:Nat)))) = plus(times(0,plus(?y:Nat,times(?x:Nat,?y:Nat))),?z:Nat)",
"s(plus(s(?y:Nat),plus(?y:Nat,plus(?z:Nat,0)))) = times(plus(?z:Nat,0),times(plus(0,?x:Nat),?x:Nat))",
"plus(?x:Nat,plus(s(?y:Nat),times(s(?x:Nat),0))) = s(s(times(plus(?x:Nat,?x:Nat),plus(?y:Nat,?x:Nat))))",
"plus(0,s(plus(s(?x:Nat),plus(?y:Nat,?x:Nat)))) = plus(0,s(plus(times(?z:Nat,?x:Nat),s(?x:Nat))))",
"plus(s(times(plus(?x:Nat,0),s(?z:Nat))),0) = plus(plus(plus(?z:Nat,s(?x:Nat)),?y:Nat),s(?x:Nat))",
"plus(plus(s(plus(?z:Nat,?x:Nat)),?x:Nat),s(?x:Nat)) = times(times(0,?y:Nat),times(s(s(?y:Nat)),0))",
"times(?y:Nat,times(plus(?x:Nat,?x:Nat),times(?z:Nat,?x:Nat))) = plus(s(plus(s(?z:Nat),0)),plus(?x:Nat,?z:Nat))",
"times(0,times(0,times(?x:Nat,times(?y:Nat,?x:Nat)))) = plus(?z:Nat,plus(times(?x:Nat,0),s(s(?z:Nat))))",
"times(s(plus(0,s(?x:Nat))),plus(0,?x:Nat)) = plus(?y:Nat,times(s(plus(0,?x:Nat)),s(?z:Nat)))",
"times(s(times(?x:Nat,s(0))),times(?x:Nat,?z:Nat)) = times(s(times(times(?y:Nat,?y:Nat),?y:Nat)),s(0))",
"times(times(0,times(?y:Nat,times(?y:Nat,?x:Nat))),?z:Nat) = plus(?x:Nat,s(times(times(?x:Nat,0),s(?z:Nat))))",
"s(plus(plus(s(?x:Nat),times(?y:Nat,?y:Nat)),s(0))) = times(?y:Nat,times(?y:Nat,s(times(0,?y:Nat))))",
"plus(times(plus(?y:Nat,?x:Nat),0),times(?y:Nat,s(0))) = times(s(times(?z:Nat,times(?y:Nat,0))),?x:Nat)",
"plus(times(plus(0,plus(?x:Nat,?z:Nat)),0),s(?z:Nat)) = plus(times(plus(?z:Nat,?y:Nat),?z:Nat),s(?x:Nat))",
"times(?x:Nat,plus(0,times(s(?x:Nat),times(?x:Nat,?y:Nat)))) = s(plus(plus(times(?x:Nat,?z:Nat),?x:Nat),?y:Nat))",
"times(?x:Nat,plus(plus(plus(?x:Nat,?x:Nat),s(?y:Nat)),?y:Nat)) = times(plus(?y:Nat,s(?y:Nat)),times(0,?z:Nat))",
"times(0,s(times(plus(?x:Nat,?y:Nat),times(?x:Nat,?x:Nat)))) = s(plus(times(s(?y:Nat),s(?x:Nat)),?x:Nat))",
"times(0,plus(s(times(?z:Nat,?x:Nat)),plus(?y:Nat,?z:Nat))) = plus(times(s(?y:Nat),plus(?z:Nat,?y:Nat)),?x:Nat)",
"times(s(s(?z:Nat)),times(?z:Nat,s(times(?x:Nat,?x:Nat)))) = plus(times(?y:Nat,s(plus(0,?x:Nat))),?x:Nat)",
"times(s(times(plus(?x:Nat,?y:Nat),0)),plus(0,?y:Nat)) = times(plus(plus(0,s(?y:Nat)),?x:Nat),0)",
"times(times(?x:Nat,times(0,plus(0,?z:Nat))),s(?y:Nat)) = s(times(times(0,plus(?x:Nat,?y:Nat)),?y:Nat))",
"s(plus(s(?y:Nat),plus(plus(?x:Nat,0),times(?y:Nat,?x:Nat)))) = plus(?x:Nat,s(plus(0,s(?z:Nat))))",
"plus(?x:Nat,plus(?z:Nat,times(times(0,?x:Nat),times(?z:Nat,?z:Nat)))) = s(s(plus(0,times(?x:Nat,?y:Nat))))",
"plus(0,plus(times(times(?z:Nat,?x:Nat),times(?x:Nat,?z:Nat)),?y:Nat)) = times(?y:Nat,times(s(?x:Nat),s(?x:Nat)))",
"plus(s(s(0)),plus(?x:Nat,times(?z:Nat,times(?x:Nat,0)))) = times(s(times(?x:Nat,s(?z:Nat))),?x:Nat)",
"plus(s(times(?y:Nat,s(?x:Nat))),times(s(s(?y:Nat)),?y:Nat)) = plus(s(?x:Nat),s(times(?y:Nat,0)))",
"plus(times(0,times(times(0,?z:Nat),0)),times(?y:Nat,?x:Nat)) = plus(?x:Nat,plus(plus(?x:Nat,?x:Nat),?z:Nat))",
"times(plus(times(s(?x:Nat),0),times(?x:Nat,s(?z:Nat))),0) = plus(s(0),s(s(s(0))))",
"times(times(?x:Nat,0),times(times(plus(?z:Nat,0),?x:Nat),?y:Nat)) = times(?x:Nat,times(s(s(?z:Nat)),?z:Nat))",
"s(times(times(times(?x:Nat,?x:Nat),?x:Nat),times(plus(?y:Nat,?x:Nat),?x:Nat))) = times(0,s(times(?z:Nat,0)))",
"plus(0,times(plus(?x:Nat,s(?y:Nat)),times(times(?y:Nat,?y:Nat),?x:Nat))) = plus(?x:Nat,plus(s(?z:Nat),?y:Nat))",
"times(0,times(times(?x:Nat,times(?x:Nat,?x:Nat)),plus(?z:Nat,s(?y:Nat)))) = s(s(s(times(?z:Nat,?x:Nat))))",
"s(times(s(plus(?x:Nat,?x:Nat)),times(times(?x:Nat,?z:Nat),plus(?y:Nat,?y:Nat)))) = plus(s(?x:Nat),s(0))",
"plus(?z:Nat,plus(times(?x:Nat,plus(0,?z:Nat)),times(?y:Nat,times(?y:Nat,?x:Nat)))) = plus(?y:Nat,s(s(?y:Nat)))",
"plus(s(times(0,s(0))),plus(s(times(?x:Nat,?z:Nat)),s(?z:Nat))) = plus(plus(?x:Nat,0),?x:Nat)",
"plus(plus(?y:Nat,times(plus(?z:Nat,?y:Nat),times(?x:Nat,?x:Nat))),plus(?y:Nat,?z:Nat)) = times(?y:Nat,times(0,?x:Nat))",
"times(times(times(?z:Nat,s(0)),0),times(s(plus(?x:Nat,?z:Nat)),?z:Nat)) = s(plus(?z:Nat,s(?x:Nat)))",
"times(times(times(0,plus(0,0)),?x:Nat),plus(?x:Nat,plus(?x:Nat,?y:Nat))) = s(s(plus(?x:Nat,?y:Nat)))",
"plus(?x:Nat,plus(times(times(0,?x:Nat),s(?x:Nat)),times(?x:Nat,plus(0,0)))) = s(times(?x:Nat,?y:Nat))",
"plus(?x:Nat,times(plus(plus(0,?x:Nat),?x:Nat),plus(plus(?z:Nat,?x:Nat),s(0)))) = times(s(?x:Nat),?x:Nat)",
"plus(plus(?y:Nat,?x:Nat),times(plus(plus(?y:Nat,?z:Nat),?y:Nat),s(plus(0,?x:Nat)))) = s(plus(?x:Nat,?x:Nat))",
"plus(times(plus(?z:Nat,times(?y:Nat,?x:Nat)),?y:Nat),times(?x:Nat,s(times(0,?z:Nat)))) = s(plus(0,?y:Nat))",
"plus(times(plus(s(?y:Nat),plus(?z:Nat,0)),times(times(0,?x:Nat),0)),?y:Nat) = times(s(?x:Nat),?x:Nat)",
"times(s(times(?x:Nat,plus(?y:Nat,?x:Nat))),times(plus(s(?y:Nat),s(?z:Nat)),0)) = plus(?x:Nat,s(0))",
"times(times(?z:Nat,plus(?z:Nat,?x:Nat)),plus(?y:Nat,plus(s(0),plus(?z:Nat,?x:Nat)))) = plus(0,s(?z:Nat))",
"plus(plus(s(?x:Nat),?z:Nat),times(times(?z:Nat,times(?x:Nat,?x:Nat)),times(?x:Nat,s(0)))) = plus(?z:Nat,?z:Nat)",
"plus(plus(times(?y:Nat,times(?x:Nat,0)),s(?y:Nat)),times(s(times(?x:Nat,?x:Nat)),?x:Nat)) = times(?x:Nat,?x:Nat)",
"plus(times(s(?y:Nat),?y:Nat),times(times(?y:Nat,?y:Nat),plus(times(0,?z:Nat),s(?x:Nat)))) = times(?y:Nat,?z:Nat)",
"plus(times(times(times(?y:Nat,0),s(0)),plus(s(0),s(0))),s(?x:Nat)) = times(0,?x:Nat)",
"times(s(plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat))),s(times(?x:Nat,times(?y:Nat,?z:Nat)))) = times(0,?x:Nat)",
"times(plus(s(0),plus(?z:Nat,plus(?z:Nat,?x:Nat))),plus(s(times(?y:Nat,?z:Nat)),?x:Nat)) = s(s(?x:Nat))",
"times(times(plus(?z:Nat,0),?x:Nat),times(plus(plus(?y:Nat,0),times(?y:Nat,?y:Nat)),0)) = s(s(0))",
"plus(?y:Nat,times(plus(times(?z:Nat,0),times(0,?z:Nat)),times(times(?y:Nat,?x:Nat),s(?z:Nat)))) = s(?x:Nat)",
"plus(plus(plus(plus(?z:Nat,?x:Nat),?x:Nat),?x:Nat),plus(?z:Nat,times(s(0),plus(?z:Nat,?z:Nat)))) = s(?x:Nat)",
"plus(plus(plus(times(?z:Nat,?y:Nat),times(0,0)),times(?x:Nat,times(0,?z:Nat))),s(0)) = s(?y:Nat)",
"plus(times(plus(times(0,?x:Nat),plus(?x:Nat,?z:Nat)),plus(0,?x:Nat)),s(s(s(0)))) = s(?z:Nat)",
"times(plus(times(plus(?x:Nat,?x:Nat),plus(?z:Nat,0)),times(s(?z:Nat),times(?z:Nat,?x:Nat))),?x:Nat) = s(?x:Nat)",
"s(s(?x:Nat)) = times(times(?x:Nat,plus(?x:Nat,s(?x:Nat))),plus(plus(times(?x:Nat,?z:Nat),times(0,?z:Nat)),?x:Nat))",
"plus(?y:Nat,?y:Nat) = plus(plus(s(plus(?x:Nat,?z:Nat)),0),times(times(?x:Nat,s(?x:Nat)),s(plus(?x:Nat,?z:Nat))))",
"plus(0,?x:Nat) = times(plus(times(plus(?y:Nat,?y:Nat),?y:Nat),times(s(?z:Nat),times(?x:Nat,?y:Nat))),s(s(?x:Nat)))",
"plus(0,0) = plus(times(?x:Nat,times(times(?y:Nat,?x:Nat),times(?y:Nat,?y:Nat))),s(times(times(?z:Nat,0),?y:Nat)))",
"times(?x:Nat,?y:Nat) = plus(plus(times(?x:Nat,?x:Nat),plus(s(0),plus(?y:Nat,?x:Nat))),s(s(times(?x:Nat,?x:Nat))))",
"times(?x:Nat,0) = times(s(plus(0,times(?y:Nat,0))),plus(times(plus(?x:Nat,0),times(?x:Nat,0)),?y:Nat))",
"times(?y:Nat,?x:Nat) = times(plus(times(?x:Nat,?z:Nat),plus(times(0,?z:Nat),s(?z:Nat))),s(plus(?x:Nat,s(0))))",
"s(s(s(0))) = times(plus(?x:Nat,times(plus(?x:Nat,0),plus(?y:Nat,?x:Nat))),times(?x:Nat,times(0,?x:Nat)))",
"plus(s(?y:Nat),?x:Nat) = plus(s(times(times(?x:Nat,?x:Nat),?z:Nat)),plus(?x:Nat,times(s(?x:Nat),times(0,?x:Nat))))",
"plus(s(0),0) = plus(plus(?y:Nat,plus(?x:Nat,?x:Nat)),plus(plus(?z:Nat,times(?y:Nat,?z:Nat)),s(s(?z:Nat))))",
"plus(?y:Nat,plus(?x:Nat,?x:Nat)) = times(0,plus(times(?x:Nat,plus(?y:Nat,?z:Nat)),times(plus(?z:Nat,0),s(0))))",
"plus(times(?x:Nat,?y:Nat),0) = times(s(?y:Nat),times(s(plus(?x:Nat,?x:Nat)),times(times(?y:Nat,?x:Nat),s(?z:Nat))))",
"times(?x:Nat,plus(?x:Nat,0)) = plus(s(0),times(times(s(?z:Nat),times(?y:Nat,?x:Nat)),s(plus(?y:Nat,?x:Nat))))",
"times(?y:Nat,plus(?x:Nat,?x:Nat)) = times(s(plus(plus(?y:Nat,0),times(?x:Nat,0))),s(times(?z:Nat,s(?x:Nat))))",
"s(times(?z:Nat,plus(?x:Nat,0))) = times(plus(s(s(?x:Nat)),?x:Nat),plus(?x:Nat,times(0,times(?x:Nat,?z:Nat))))",
"plus(s(?x:Nat),plus(?x:Nat,?x:Nat)) = plus(s(s(0)),times(plus(s(?z:Nat),?x:Nat),plus(?z:Nat,s(?x:Nat))))",
"plus(0,s(times(?y:Nat,s(?x:Nat)))) = times(s(times(times(?z:Nat,?x:Nat),?x:Nat)),plus(s(s(0)),?x:Nat))",
"plus(0,plus(?x:Nat,plus(?x:Nat,?x:Nat))) = times(plus(plus(plus(?x:Nat,?x:Nat),s(?y:Nat)),?x:Nat),times(?z:Nat,?x:Nat))",
"plus(s(s(times(?x:Nat,0))),?x:Nat) = plus(plus(plus(times(0,?x:Nat),?x:Nat),?z:Nat),s(times(0,?y:Nat)))",
"plus(times(?x:Nat,times(?x:Nat,?x:Nat)),?y:Nat) = times(s(0),times(plus(times(?y:Nat,0),plus(?x:Nat,?x:Nat)),?y:Nat))",
"times(?x:Nat,s(s(times(?x:Nat,?x:Nat)))) = plus(?x:Nat,plus(s(plus(?x:Nat,?x:Nat)),times(?x:Nat,times(0,?y:Nat))))",
"times(s(plus(?y:Nat,s(?x:Nat))),0) = plus(plus(?z:Nat,s(times(?x:Nat,0))),plus(s(0),s(?y:Nat)))",
"times(times(s(s(?x:Nat)),?x:Nat),?y:Nat) = plus(plus(times(?y:Nat,s(?y:Nat)),times(?y:Nat,times(?z:Nat,0))),?z:Nat)",
"s(plus(0,plus(plus(?z:Nat,?z:Nat),?x:Nat))) = plus(times(s(plus(?z:Nat,?x:Nat)),?y:Nat),times(0,s(?x:Nat)))",
"s(times(?x:Nat,plus(plus(?x:Nat,?z:Nat),?y:Nat))) = times(times(?z:Nat,times(plus(0,?z:Nat),plus(?x:Nat,?x:Nat))),?x:Nat)",
"plus(0,s(plus(times(?x:Nat,?x:Nat),?x:Nat))) = times(times(?z:Nat,s(?y:Nat)),times(plus(s(?x:Nat),?x:Nat),?x:Nat))",
"plus(0,s(times(times(?z:Nat,?x:Nat),?x:Nat))) = times(s(?x:Nat),s(times(times(0,?z:Nat),times(?x:Nat,?x:Nat))))",
"plus(s(times(plus(?x:Nat,0),?x:Nat)),0) = plus(s(?z:Nat),plus(plus(?x:Nat,?z:Nat),plus(s(?x:Nat),?z:Nat)))",
"plus(times(0,?x:Nat),s(plus(?y:Nat,?x:Nat))) = s(plus(times(s(?x:Nat),s(0)),s(times(0,?z:Nat))))",
"times(?x:Nat,s(plus(?z:Nat,plus(?x:Nat,?z:Nat)))) = plus(plus(?y:Nat,s(0)),s(times(0,plus(0,?x:Nat))))",
"times(times(plus(?y:Nat,?z:Nat),s(0)),?x:Nat) = plus(?x:Nat,plus(0,plus(times(?z:Nat,?x:Nat),times(?z:Nat,0))))",
"plus(?z:Nat,times(times(?x:Nat,s(?y:Nat)),s(?x:Nat))) = plus(s(0),s(times(plus(?x:Nat,?x:Nat),s(?x:Nat))))",
"plus(plus(?z:Nat,plus(?x:Nat,times(?y:Nat,0))),?x:Nat) = s(times(plus(0,plus(0,?z:Nat)),s(s(?y:Nat))))",
"plus(times(?x:Nat,plus(?z:Nat,plus(?x:Nat,?z:Nat))),?z:Nat) = plus(s(times(times(0,?x:Nat),?x:Nat)),plus(?z:Nat,0))",
"plus(times(?z:Nat,times(s(?x:Nat),?z:Nat)),s(?z:Nat)) = plus(plus(s(times(?z:Nat,?y:Nat)),0),times(?y:Nat,?z:Nat))",
"plus(times(0,s(times(?z:Nat,?z:Nat))),s(?x:Nat)) = s(times(?y:Nat,plus(times(?x:Nat,?x:Nat),plus(?x:Nat,?z:Nat))))",
"times(?z:Nat,times(?z:Nat,times(0,plus(?y:Nat,?z:Nat)))) = times(?x:Nat,plus(0,times(plus(?z:Nat,?x:Nat),s(?x:Nat))))",
"s(times(plus(?y:Nat,times(?z:Nat,?z:Nat)),plus(0,?y:Nat))) = times(?x:Nat,plus(times(?x:Nat,?x:Nat),s(s(?x:Nat))))",
"plus(0,plus(s(plus(0,?y:Nat)),s(s(0)))) = s(times(s(?x:Nat),times(?x:Nat,times(?x:Nat,?y:Nat))))",
"plus(0,times(times(?y:Nat,s(?x:Nat)),s(s(?x:Nat)))) = s(times(plus(s(0),times(?y:Nat,?x:Nat)),0))",
"plus(s(?x:Nat),plus(plus(times(?z:Nat,?y:Nat),0),0)) = s(plus(s(0),plus(times(0,?y:Nat),?y:Nat)))",
"plus(s(plus(s(?z:Nat),s(?x:Nat))),times(0,?x:Nat)) = s(plus(s(s(?x:Nat)),s(plus(?y:Nat,?y:Nat))))",
"plus(times(times(times(0,?x:Nat),s(?z:Nat)),0),?x:Nat) = s(times(?x:Nat,times(plus(?y:Nat,?x:Nat),s(?z:Nat))))",
"plus(?x:Nat,plus(times(s(?y:Nat),times(0,?x:Nat)),s(?x:Nat))) = times(s(?x:Nat),plus(?x:Nat,times(?z:Nat,?z:Nat)))",
"times(plus(0,s(plus(?x:Nat,?x:Nat))),times(s(?x:Nat),?x:Nat)) = times(?z:Nat,times(?x:Nat,times(?z:Nat,s(?y:Nat))))",
"times(times(plus(?z:Nat,0),times(s(?x:Nat),?x:Nat)),s(?y:Nat)) = times(s(times(plus(?x:Nat,0),0)),?y:Nat)",
"times(?x:Nat,plus(plus(plus(0,?y:Nat),s(?z:Nat)),s(s(?x:Nat)))) = plus(s(s(plus(?z:Nat,?x:Nat))),?x:Nat)",
"times(0,times(s(?y:Nat),times(times(?z:Nat,?y:Nat),plus(?y:Nat,?z:Nat)))) = times(plus(?x:Nat,?x:Nat),plus(0,?y:Nat))",
"times(plus(plus(0,times(?x:Nat,?y:Nat)),times(0,?y:Nat)),s(?x:Nat)) = s(plus(?x:Nat,plus(s(?y:Nat),?x:Nat)))",
"times(times(times(plus(0,0),?x:Nat),?y:Nat),s(plus(?x:Nat,?y:Nat))) = plus(?z:Nat,times(?x:Nat,plus(?x:Nat,?x:Nat)))",
"plus(times(?x:Nat,plus(?y:Nat,?x:Nat)),plus(?z:Nat,times(0,times(?x:Nat,?x:Nat)))) = plus(?x:Nat,plus(?z:Nat,s(?x:Nat)))",
"plus(times(0,0),times(s(s(?x:Nat)),plus(?x:Nat,plus(0,?x:Nat)))) = s(plus(s(s(?z:Nat)),?y:Nat))",
"times(plus(?z:Nat,plus(?x:Nat,?z:Nat)),times(s(?y:Nat),times(0,s(?x:Nat)))) = times(s(times(?x:Nat,0)),?x:Nat)",
"times(times(s(s(?x:Nat)),times(plus(0,?y:Nat),?y:Nat)),times(?x:Nat,0)) = s(s(s(plus(?x:Nat,?x:Nat))))",
"plus(times(?x:Nat,plus(?z:Nat,?y:Nat)),times(?x:Nat,plus(s(?x:Nat),plus(0,?x:Nat)))) = times(0,plus(?x:Nat,?y:Nat))",
"plus(times(?x:Nat,plus(plus(0,?z:Nat),s(?x:Nat))),plus(plus(0,?y:Nat),0)) = plus(plus(0,?x:Nat),?x:Nat)",
"plus(times(times(?x:Nat,times(?x:Nat,?x:Nat)),times(?z:Nat,plus(0,?x:Nat))),s(?x:Nat)) = plus(s(s(?x:Nat)),0)",
"plus(plus(?z:Nat,plus(?y:Nat,?x:Nat)),plus(plus(times(?x:Nat,0),?y:Nat),s(s(?x:Nat)))) = times(s(?y:Nat),?z:Nat)",
"plus(plus(s(s(?z:Nat)),times(?z:Nat,?z:Nat)),plus(?z:Nat,plus(?x:Nat,times(0,?y:Nat)))) = times(s(?y:Nat),0)",
"times(s(s(plus(?x:Nat,0))),plus(times(0,plus(?x:Nat,?y:Nat)),s(s(?z:Nat)))) = times(0,s(?z:Nat))",
"times(plus(plus(times(0,?x:Nat),s(?x:Nat)),0),s(plus(s(?x:Nat),s(?x:Nat)))) = plus(s(?z:Nat),?z:Nat)",
"times(s(plus(?y:Nat,times(?y:Nat,0))),times(?y:Nat,times(plus(?z:Nat,?y:Nat),times(?y:Nat,?x:Nat)))) = times(0,?y:Nat)",
"plus(s(s(plus(?z:Nat,0))),times(plus(0,?x:Nat),plus(times(?x:Nat,?x:Nat),times(0,0)))) = s(?y:Nat)",
"plus(s(s(plus(0,0))),plus(plus(s(?y:Nat),s(?y:Nat)),times(times(?z:Nat,?x:Nat),?x:Nat))) = s(?z:Nat)",
"plus(plus(?x:Nat,plus(times(0,0),times(?y:Nat,?x:Nat))),times(?x:Nat,times(?z:Nat,times(0,?x:Nat)))) = s(0)",
"plus(plus(plus(?y:Nat,plus(0,?z:Nat)),s(plus(0,?x:Nat))),plus(?z:Nat,times(?z:Nat,s(?x:Nat)))) = s(?x:Nat)",
"times(times(plus(plus(?y:Nat,?z:Nat),?x:Nat),s(?z:Nat)),plus(plus(?y:Nat,0),times(s(?x:Nat),?x:Nat))) = s(?z:Nat)",
"plus(0,?y:Nat) = plus(plus(times(s(?x:Nat),times(?x:Nat,?y:Nat)),?x:Nat),plus(?y:Nat,plus(s(?z:Nat),plus(?y:Nat,?x:Nat))))",
"times(?x:Nat,?z:Nat) = times(plus(times(s(0),?z:Nat),plus(?z:Nat,?y:Nat)),times(s(?x:Nat),plus(times(?x:Nat,?z:Nat),?x:Nat)))",
"times(?z:Nat,0) = plus(times(times(times(?y:Nat,0),s(?z:Nat)),s(plus(0,?y:Nat))),times(?y:Nat,plus(?x:Nat,0)))",
"times(0,s(0)) = s(times(times(times(?x:Nat,0),plus(?z:Nat,?x:Nat)),plus(times(0,?y:Nat),plus(?z:Nat,?y:Nat))))",
"s(s(plus(?z:Nat,?x:Nat))) = times(plus(times(times(?y:Nat,?y:Nat),?y:Nat),s(plus(?z:Nat,0))),plus(?z:Nat,s(?x:Nat)))",
"plus(?y:Nat,times(?y:Nat,?y:Nat)) = times(plus(plus(plus(?x:Nat,?x:Nat),?x:Nat),0),times(?x:Nat,plus(times(0,?x:Nat),?y:Nat)))",
"times(?x:Nat,plus(0,?x:Nat)) = plus(plus(times(?x:Nat,s(?y:Nat)),times(plus(0,0),?x:Nat)),plus(s(?x:Nat),?x:Nat))",
"times(plus(?x:Nat,?y:Nat),?z:Nat) = plus(s(times(times(0,?x:Nat),s(0))),times(?z:Nat,plus(?z:Nat,plus(?y:Nat,0))))",
"plus(?x:Nat,s(s(s(0)))) = plus(s(times(times(?z:Nat,?x:Nat),0)),times(times(?y:Nat,plus(?x:Nat,0)),?x:Nat))",
"plus(plus(?y:Nat,?z:Nat),s(?x:Nat)) = plus(plus(times(plus(?z:Nat,0),0),plus(times(?x:Nat,?x:Nat),s(0))),?z:Nat)",
"plus(times(s(s(?x:Nat)),?z:Nat),0) = plus(times(?y:Nat,?x:Nat),times(s(?x:Nat),times(s(?z:Nat),times(?y:Nat,?x:Nat))))",
"s(times(?z:Nat,times(?z:Nat,times(?y:Nat,0)))) = plus(?z:Nat,plus(s(?x:Nat),plus(times(?x:Nat,0),plus(?z:Nat,0))))",
"plus(?x:Nat,times(s(times(?x:Nat,?x:Nat)),?x:Nat)) = plus(0,times(plus(s(?z:Nat),plus(0,?x:Nat)),plus(?x:Nat,?x:Nat)))",
"plus(s(times(?x:Nat,plus(?x:Nat,?z:Nat))),?y:Nat) = times(plus(plus(?x:Nat,?y:Nat),?y:Nat),s(plus(0,plus(?y:Nat,?z:Nat))))",
"times(?x:Nat,s(plus(times(?x:Nat,?x:Nat),0))) = times(?y:Nat,plus(s(?x:Nat),times(plus(?x:Nat,?x:Nat),plus(0,?x:Nat))))",
"times(?x:Nat,s(times(plus(?x:Nat,?z:Nat),?x:Nat))) = times(times(s(?x:Nat),0),times(?y:Nat,times(times(?y:Nat,0),?x:Nat)))",
"s(s(times(plus(?z:Nat,?z:Nat),plus(?y:Nat,?z:Nat)))) = times(times(?x:Nat,plus(s(0),?z:Nat)),times(?x:Nat,s(?x:Nat)))",
"plus(plus(plus(0,?y:Nat),plus(?x:Nat,?x:Nat)),?y:Nat) = times(times(times(times(?x:Nat,?y:Nat),times(0,?y:Nat)),?z:Nat),?x:Nat)",
"times(?x:Nat,times(plus(?x:Nat,plus(?x:Nat,0)),0)) = s(times(s(plus(?y:Nat,?x:Nat)),times(times(?x:Nat,?x:Nat),?z:Nat)))",
"s(plus(times(times(?z:Nat,?x:Nat),plus(0,?y:Nat)),?x:Nat)) = plus(?x:Nat,times(plus(plus(?x:Nat,?z:Nat),s(?z:Nat)),0))",
"plus(?z:Nat,s(plus(plus(?z:Nat,?z:Nat),plus(?y:Nat,0)))) = plus(plus(plus(?y:Nat,times(0,?z:Nat)),?x:Nat),s(?x:Nat))",
"plus(s(s(?z:Nat)),plus(times(?y:Nat,s(?z:Nat)),?y:Nat)) = plus(0,times(plus(plus(0,?x:Nat),s(?y:Nat)),?x:Nat))",
"plus(plus(s(times(?x:Nat,?x:Nat)),plus(?y:Nat,?x:Nat)),?x:Nat) = times(s(times(plus(?y:Nat,0),plus(?x:Nat,0))),?x:Nat)",
"plus(plus(times(s(?x:Nat),times(?x:Nat,?x:Nat)),0),?y:Nat) = times(plus(0,s(plus(?z:Nat,?z:Nat))),s(s(0)))",
"plus(times(0,s(?x:Nat)),s(s(plus(?x:Nat,?x:Nat)))) = times(plus(?x:Nat,?x:Nat),plus(s(plus(0,?x:Nat)),?x:Nat))",
"plus(?y:Nat,plus(s(?x:Nat),times(plus(?x:Nat,?x:Nat),s(?y:Nat)))) = plus(plus(?z:Nat,plus(s(0),s(?x:Nat))),?z:Nat)",
"plus(s(plus(?x:Nat,?x:Nat)),s(times(?y:Nat,times(0,?y:Nat)))) = times(?x:Nat,times(plus(times(?x:Nat,?x:Nat),?z:Nat),?x:Nat))",
"plus(times(?z:Nat,times(plus(?x:Nat,?x:Nat),plus(0,?x:Nat))),?y:Nat) = plus(times(plus(s(0),?z:Nat),?x:Nat),s(?y:Nat))",
"times(?x:Nat,plus(plus(times(?x:Nat,?x:Nat),plus(?y:Nat,0)),?x:Nat)) = s(s(times(plus(0,?x:Nat),plus(0,?y:Nat))))",
"times(times(times(times(?y:Nat,?x:Nat),s(?x:Nat)),s(0)),0) = s(plus(?y:Nat,times(times(?x:Nat,?z:Nat),s(?y:Nat))))",
"s(plus(plus(times(?x:Nat,?y:Nat),plus(?x:Nat,0)),plus(?y:Nat,?z:Nat))) = times(?z:Nat,times(0,times(s(?y:Nat),?y:Nat)))",
"s(times(times(?y:Nat,s(?x:Nat)),plus(times(?x:Nat,?x:Nat),s(0)))) = times(plus(0,?x:Nat),s(plus(?x:Nat,?x:Nat)))",
"plus(s(times(times(?x:Nat,0),s(?x:Nat))),s(s(s(0)))) = s(plus(times(?z:Nat,times(?x:Nat,?x:Nat)),?z:Nat))",
"plus(times(0,times(s(?y:Nat),s(?x:Nat))),plus(s(0),?y:Nat)) = times(times(?z:Nat,s(plus(?y:Nat,0))),0)",
"times(0,times(s(times(?y:Nat,?x:Nat)),plus(?x:Nat,times(0,?x:Nat)))) = plus(times(times(?y:Nat,s(0)),?z:Nat),?y:Nat)",
"plus(?x:Nat,plus(times(s(0),plus(?x:Nat,?x:Nat)),s(plus(?z:Nat,?x:Nat)))) = times(times(?x:Nat,?y:Nat),times(?y:Nat,?x:Nat))",
"times(s(times(?x:Nat,s(?z:Nat))),plus(times(0,s(?y:Nat)),s(?x:Nat))) = s(plus(plus(?x:Nat,s(?x:Nat)),?x:Nat))",
"s(times(times(times(?z:Nat,?x:Nat),times(?x:Nat,?x:Nat)),times(plus(0,?x:Nat),?x:Nat))) = times(?z:Nat,plus(s(?z:Nat),?z:Nat))",
"plus(s(?y:Nat),plus(plus(?x:Nat,times(?x:Nat,?x:Nat)),times(times(0,?y:Nat),?x:Nat))) = plus(s(?x:Nat),plus(?x:Nat,?z:Nat))",
"times(plus(plus(?y:Nat,times(0,?x:Nat)),plus(plus(?z:Nat,?z:Nat),s(?x:Nat))),?y:Nat) = plus(?x:Nat,plus(s(?z:Nat),0))",
"plus(times(?x:Nat,s(times(?z:Nat,?y:Nat))),plus(s(s(0)),plus(?z:Nat,s(?z:Nat)))) = plus(?x:Nat,plus(0,?z:Nat))",
"plus(times(times(?z:Nat,0),s(?x:Nat)),times(plus(?y:Nat,times(?y:Nat,?z:Nat)),s(?z:Nat))) = times(times(?x:Nat,0),?x:Nat)",
"times(s(s(?y:Nat)),plus(plus(?z:Nat,plus(?y:Nat,0)),times(?y:Nat,times(?y:Nat,?z:Nat)))) = times(times(?x:Nat,?z:Nat),?x:Nat)",
"times(plus(?x:Nat,s(?y:Nat)),times(times(?z:Nat,plus(?x:Nat,0)),s(times(0,?x:Nat)))) = times(s(?z:Nat),s(?x:Nat))",
"times(times(?x:Nat,s(s(?x:Nat))),plus(0,plus(plus(?y:Nat,?y:Nat),plus(?x:Nat,?z:Nat)))) = plus(?x:Nat,times(?x:Nat,?x:Nat))",
"plus(plus(plus(plus(0,?y:Nat),s(0)),0),plus(plus(times(0,?y:Nat),?z:Nat),0)) = plus(?x:Nat,s(?x:Nat))",
"times(s(?y:Nat),plus(plus(plus(?x:Nat,0),?x:Nat),plus(times(?y:Nat,?x:Nat),plus(?x:Nat,?y:Nat)))) = times(?x:Nat,s(?x:Nat))",
"times(plus(times(s(?y:Nat),?y:Nat),plus(?y:Nat,?z:Nat)),times(times(s(?y:Nat),?x:Nat),times(?x:Nat,0))) = plus(?y:Nat,?z:Nat)",
"plus(times(?x:Nat,times(times(?y:Nat,0),?x:Nat)),plus(plus(?x:Nat,times(0,?z:Nat)),plus(?x:Nat,s(0)))) = s(?y:Nat)",
"plus(times(?y:Nat,times(s(0),times(?x:Nat,0))),times(times(times(?y:Nat,?y:Nat),times(0,?x:Nat)),0)) = s(0)",
"times(times(s(s(0)),?x:Nat),plus(times(0,s(?z:Nat)),times(plus(?z:Nat,0),times(?y:Nat,?z:Nat)))) = s(?x:Nat)",
"s(?x:Nat) = times(s(times(times(?x:Nat,0),times(?x:Nat,0))),times(times(?x:Nat,s(?z:Nat)),plus(0,plus(0,?x:Nat))))",
"s(?y:Nat) = times(times(times(s(0),0),?x:Nat),times(times(times(?y:Nat,?x:Nat),plus(0,?x:Nat)),plus(?x:Nat,s(?z:Nat))))",
"s(s(?y:Nat)) = plus(times(plus(?x:Nat,?z:Nat),s(plus(?y:Nat,?y:Nat))),plus(plus(times(?x:Nat,?y:Nat),?y:Nat),times(?z:Nat,0)))",
"s(plus(?x:Nat,?x:Nat)) = plus(s(plus(plus(?y:Nat,?x:Nat),times(?x:Nat,0))),plus(plus(0,?x:Nat),times(?x:Nat,s(?x:Nat))))",
"plus(?z:Nat,times(?z:Nat,?y:Nat)) = times(times(plus(?z:Nat,plus(?x:Nat,0)),times(times(?x:Nat,?y:Nat),?x:Nat)),plus(?x:Nat,s(0)))",
"times(s(s(plus(?x:Nat,?z:Nat))),?x:Nat) = plus(plus(times(0,plus(?x:Nat,?x:Nat)),times(s(?x:Nat),0)),plus(?y:Nat,?x:Nat))",
"s(plus(s(?z:Nat),s(times(?x:Nat,0)))) = plus(plus(plus(times(?x:Nat,?x:Nat),?x:Nat),times(times(?y:Nat,0),0)),?y:Nat)",
"s(plus(plus(?z:Nat,?x:Nat),times(?z:Nat,?y:Nat))) = plus(times(times(?z:Nat,plus(?x:Nat,?y:Nat)),times(?x:Nat,plus(0,?x:Nat))),?y:Nat)",
"plus(s(times(s(0),s(?x:Nat))),?x:Nat) = times(?x:Nat,times(times(s(?y:Nat),plus(0,?x:Nat)),s(plus(0,?x:Nat))))",
"times(?x:Nat,s(plus(?z:Nat,plus(?y:Nat,?x:Nat)))) = times(?x:Nat,times(times(?y:Nat,0),times(plus(0,?x:Nat),plus(0,?x:Nat))))",
"times(?x:Nat,times(?x:Nat,s(times(?x:Nat,?x:Nat)))) = plus(s(times(?z:Nat,0)),s(times(plus(?y:Nat,?z:Nat),plus(?x:Nat,?x:Nat))))",
"times(?x:Nat,times(times(s(0),?x:Nat),?x:Nat)) = times(s(times(s(?x:Nat),0)),s(plus(times(?x:Nat,0),s(0))))",
"times(?y:Nat,s(times(?z:Nat,plus(?x:Nat,?z:Nat)))) = times(plus(?x:Nat,times(0,plus(?y:Nat,?x:Nat))),times(times(?z:Nat,?y:Nat),?x:Nat))",
"s(s(plus(times(?x:Nat,0),times(0,?z:Nat)))) = times(plus(times(times(?y:Nat,?z:Nat),s(?z:Nat)),s(s(?y:Nat))),0)",
"s(plus(plus(s(?z:Nat),plus(?z:Nat,?y:Nat)),?x:Nat)) = times(plus(times(?y:Nat,times(?x:Nat,0)),s(times(?x:Nat,?x:Nat))),?x:Nat)",
"plus(0,plus(plus(times(?x:Nat,?y:Nat),0),?x:Nat)) = plus(times(plus(plus(?x:Nat,?x:Nat),times(?z:Nat,0)),s(?x:Nat)),?x:Nat)",
"plus(0,times(times(plus(?x:Nat,?x:Nat),?z:Nat),?x:Nat)) = times(s(plus(?z:Nat,?y:Nat)),s(plus(times(?x:Nat,?x:Nat),s(0))))",
"times(s(s(plus(?y:Nat,?x:Nat))),plus(?z:Nat,?y:Nat)) = plus(?x:Nat,times(times(times(?x:Nat,?z:Nat),0),plus(?x:Nat,s(?z:Nat))))",
"times(s(s(times(?x:Nat,?z:Nat))),times(0,?y:Nat)) = plus(plus(s(times(?z:Nat,?z:Nat)),0),times(?z:Nat,plus(0,?y:Nat)))",
"plus(s(?z:Nat),times(s(plus(0,?z:Nat)),s(?z:Nat))) = plus(times(times(?x:Nat,s(?y:Nat)),s(s(?y:Nat))),s(?x:Nat))",
"plus(times(?x:Nat,plus(plus(?x:Nat,?y:Nat),s(?x:Nat))),0) = plus(0,plus(plus(plus(0,?x:Nat),s(?z:Nat)),s(?z:Nat)))",
"plus(times(?z:Nat,plus(times(?y:Nat,?z:Nat),s(?y:Nat))),?y:Nat) = plus(plus(0,?y:Nat),plus(?y:Nat,times(plus(0,?x:Nat),?x:Nat)))",
"times(?z:Nat,plus(plus(s(0),plus(?x:Nat,?x:Nat)),?x:Nat)) = s(plus(plus(?y:Nat,?x:Nat),times(s(?x:Nat),times(?x:Nat,?x:Nat))))",
"times(s(?z:Nat),plus(plus(times(?x:Nat,?z:Nat),?x:Nat),?x:Nat)) = times(plus(plus(?x:Nat,times(?x:Nat,?y:Nat)),plus(?z:Nat,0)),?x:Nat)",
"times(times(plus(?z:Nat,?x:Nat),plus(0,?x:Nat)),s(?z:Nat)) = times(plus(s(?x:Nat),plus(s(?z:Nat),times(?x:Nat,?y:Nat))),0)",
"plus(?z:Nat,times(?x:Nat,plus(plus(?x:Nat,?x:Nat),plus(0,0)))) = plus(s(times(s(?y:Nat),s(?y:Nat))),times(?z:Nat,0))",
"plus(0,times(?x:Nat,plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?z:Nat)))) = times(s(times(times(0,?x:Nat),plus(0,?y:Nat))),?x:Nat)",
"plus(s(?x:Nat),plus(?z:Nat,plus(times(0,?x:Nat),s(?z:Nat)))) = times(plus(0,?y:Nat),s(times(plus(?y:Nat,?x:Nat),?y:Nat)))",
"times(?x:Nat,times(times(plus(?x:Nat,?z:Nat),plus(0,?y:Nat)),?z:Nat)) = plus(times(0,times(s(?y:Nat),s(?x:Nat))),s(?z:Nat))",
"s(plus(plus(plus(?x:Nat,0),0),plus(?x:Nat,times(?y:Nat,?z:Nat)))) = plus(plus(s(?z:Nat),s(?x:Nat)),times(?x:Nat,?x:Nat))",
"times(plus(0,?x:Nat),s(times(plus(0,?z:Nat),plus(?x:Nat,?y:Nat)))) = s(s(times(times(0,0),times(?y:Nat,?y:Nat))))",
"times(times(s(?x:Nat),?y:Nat),plus(plus(?x:Nat,plus(?z:Nat,?x:Nat)),?z:Nat)) = plus(0,times(s(times(?x:Nat,?x:Nat)),s(?z:Nat)))",
"times(times(s(s(?x:Nat)),times(?z:Nat,s(0))),times(?y:Nat,0)) = times(s(times(plus(?y:Nat,?x:Nat),s(?z:Nat))),?x:Nat)",
"plus(plus(plus(times(?x:Nat,?y:Nat),?z:Nat),s(s(?z:Nat))),plus(?x:Nat,?z:Nat)) = plus(times(?x:Nat,plus(?y:Nat,s(?x:Nat))),0)",
"times(plus(0,s(times(0,?x:Nat))),plus(s(times(?z:Nat,?x:Nat)),?y:Nat)) = plus(plus(?x:Nat,s(?x:Nat)),plus(?z:Nat,?y:Nat))",
"times(times(times(times(?z:Nat,0),times(?y:Nat,?x:Nat)),plus(?x:Nat,s(?x:Nat))),?y:Nat) = plus(0,s(times(s(?y:Nat),?x:Nat)))",
"plus(s(?x:Nat),times(plus(times(?y:Nat,?y:Nat),times(?z:Nat,0)),s(plus(?x:Nat,0)))) = s(times(?x:Nat,plus(?x:Nat,?z:Nat)))",
"plus(s(times(s(?x:Nat),times(0,?z:Nat))),s(plus(plus(?x:Nat,0),s(?y:Nat)))) = plus(times(0,?z:Nat),s(?x:Nat))",
"times(plus(?z:Nat,times(times(?x:Nat,?z:Nat),?x:Nat)),times(s(s(?y:Nat)),times(s(?x:Nat),?x:Nat))) = times(times(0,?y:Nat),0)",
"plus(times(times(s(?x:Nat),times(?z:Nat,?z:Nat)),s(s(0))),s(times(s(?y:Nat),s(?y:Nat)))) = s(s(s(?z:Nat)))",
"times(s(times(times(?x:Nat,0),0)),times(plus(times(?z:Nat,0),0),times(?z:Nat,plus(?z:Nat,?x:Nat)))) = times(?x:Nat,?y:Nat)",
"plus(times(plus(times(?y:Nat,?z:Nat),plus(?z:Nat,?y:Nat)),plus(times(?x:Nat,?y:Nat),times(?y:Nat,?y:Nat))),plus(?z:Nat,0)) = s(?x:Nat)",
"times(times(plus(plus(?z:Nat,?x:Nat),0),plus(?x:Nat,?x:Nat)),times(times(0,s(?x:Nat)),plus(?x:Nat,s(0)))) = s(?y:Nat)",
"s(times(0,?z:Nat)) = times(plus(plus(0,plus(?z:Nat,0)),?z:Nat),plus(s(times(?y:Nat,?z:Nat)),times(s(?z:Nat),s(?z:Nat))))",
"s(s(plus(?y:Nat,0))) = plus(times(times(times(0,?x:Nat),?z:Nat),times(times(?z:Nat,?z:Nat),plus(?y:Nat,?x:Nat))),plus(?x:Nat,?x:Nat))",
"s(plus(0,s(?x:Nat))) = times(plus(?x:Nat,times(plus(?x:Nat,?x:Nat),?z:Nat)),times(?x:Nat,times(plus(0,?y:Nat),times(?y:Nat,?x:Nat))))",
"s(times(s(?x:Nat),0)) = plus(plus(plus(plus(?y:Nat,?x:Nat),times(?x:Nat,?x:Nat)),times(plus(?z:Nat,?y:Nat),times(?x:Nat,?z:Nat))),0)",
"times(?x:Nat,plus(s(?z:Nat),?x:Nat)) = times(times(plus(plus(?y:Nat,0),?z:Nat),times(times(?y:Nat,?z:Nat),?x:Nat)),times(s(?y:Nat),?z:Nat))",
"times(times(s(?z:Nat),?x:Nat),?z:Nat) = plus(plus(times(?x:Nat,times(?x:Nat,0)),s(s(?x:Nat))),times(0,times(s(0),0)))",
"plus(s(s(0)),s(s(?y:Nat))) = plus(times(times(times(?x:Nat,?z:Nat),?x:Nat),plus(?y:Nat,?y:Nat)),times(s(?x:Nat),s(?x:Nat)))",
"times(times(s(s(?y:Nat)),0),0) = plus(plus(s(plus(?z:Nat,?y:Nat)),s(times(?y:Nat,?x:Nat))),times(times(?y:Nat,?y:Nat),?z:Nat))",
"s(times(plus(times(0,0),?z:Nat),?x:Nat)) = plus(times(plus(plus(0,?x:Nat),s(?y:Nat)),plus(?x:Nat,plus(0,?x:Nat))),?x:Nat)",
"s(times(times(times(?y:Nat,0),?y:Nat),?x:Nat)) = times(times(?x:Nat,plus(?z:Nat,?x:Nat)),s(plus(times(?y:Nat,?x:Nat),plus(?z:Nat,?x:Nat))))",
"s(plus(?x:Nat,plus(times(?z:Nat,?z:Nat),s(?z:Nat)))) = plus(plus(s(s(?x:Nat)),times(plus(?z:Nat,0),plus(?y:Nat,?x:Nat))),0)",
"s(plus(?y:Nat,times(s(?y:Nat),plus(?x:Nat,?z:Nat)))) = times(plus(plus(times(0,?x:Nat),?x:Nat),?y:Nat),times(?y:Nat,s(s(?y:Nat))))",
"s(plus(s(s(0)),s(plus(?z:Nat,?x:Nat)))) = plus(plus(times(?y:Nat,plus(?z:Nat,?x:Nat)),times(?x:Nat,plus(?y:Nat,?y:Nat))),?x:Nat)",
"plus(plus(?y:Nat,s(plus(?x:Nat,0))),s(?x:Nat)) = plus(?x:Nat,times(times(plus(?x:Nat,?y:Nat),?x:Nat),plus(0,times(?x:Nat,?x:Nat))))",
"times(?x:Nat,plus(plus(?x:Nat,times(0,?y:Nat)),?x:Nat)) = times(0,times(plus(plus(?z:Nat,?x:Nat),?x:Nat),plus(times(?y:Nat,?x:Nat),?z:Nat)))",
"s(plus(s(s(?y:Nat)),times(?x:Nat,times(?y:Nat,?x:Nat)))) = times(s(?x:Nat),times(times(0,0),times(0,plus(?y:Nat,0))))",
"plus(plus(plus(plus(?z:Nat,?y:Nat),s(?z:Nat)),?x:Nat),?y:Nat) = times(plus(plus(plus(0,?y:Nat),?z:Nat),s(plus(?x:Nat,?x:Nat))),?y:Nat)",
"plus(times(?y:Nat,s(plus(?x:Nat,?x:Nat))),plus(0,?x:Nat)) = times(times(0,0),times(times(?z:Nat,plus(?z:Nat,?x:Nat)),s(0)))",
"times(plus(plus(plus(?z:Nat,?y:Nat),0),plus(?y:Nat,?y:Nat)),?y:Nat) = plus(?y:Nat,times(plus(plus(?y:Nat,?x:Nat),?x:Nat),plus(?y:Nat,0)))",
"times(plus(times(s(?z:Nat),0),times(?x:Nat,s(?y:Nat))),0) = plus(?x:Nat,plus(s(times(?x:Nat,0)),s(plus(?z:Nat,0))))",
"times(times(times(?z:Nat,?x:Nat),plus(times(?z:Nat,?z:Nat),?x:Nat)),?y:Nat) = s(times(s(plus(0,?x:Nat)),times(times(?x:Nat,?y:Nat),?x:Nat)))",
"plus(s(0),times(plus(s(?x:Nat),times(?y:Nat,?x:Nat)),s(?z:Nat))) = times(times(times(s(?z:Nat),s(?x:Nat)),?x:Nat),s(?x:Nat))",
"times(?y:Nat,plus(plus(plus(0,?x:Nat),?x:Nat),plus(?y:Nat,s(0)))) = times(s(?x:Nat),times(?x:Nat,plus(?x:Nat,times(0,?x:Nat))))",
"times(s(times(times(?x:Nat,0),?x:Nat)),plus(times(?x:Nat,?x:Nat),?x:Nat)) = s(plus(plus(?z:Nat,times(?z:Nat,?x:Nat)),plus(0,0)))",
"times(times(?z:Nat,s(?x:Nat)),plus(plus(times(0,?y:Nat),?x:Nat),0)) = times(times(times(?y:Nat,times(?x:Nat,?x:Nat)),?x:Nat),s(?y:Nat))",
"plus(plus(?x:Nat,times(plus(?y:Nat,?z:Nat),0)),plus(s(s(?x:Nat)),?x:Nat)) = plus(?y:Nat,plus(plus(plus(0,?x:Nat),?x:Nat),?y:Nat))",
"plus(plus(plus(s(?x:Nat),?x:Nat),?z:Nat),s(times(s(?y:Nat),s(?x:Nat)))) = times(?x:Nat,s(plus(times(?z:Nat,?x:Nat),s(0))))",
"times(?x:Nat,times(times(plus(?x:Nat,0),?x:Nat),times(?x:Nat,plus(?x:Nat,?x:Nat)))) = times(?z:Nat,plus(?z:Nat,plus(?x:Nat,plus(?z:Nat,0))))",
"times(s(?y:Nat),plus(plus(times(?x:Nat,?x:Nat),plus(0,?y:Nat)),times(0,?x:Nat))) = times(s(plus(?z:Nat,times(?x:Nat,?x:Nat))),?y:Nat)",
"times(plus(s(?y:Nat),times(0,?y:Nat)),times(?z:Nat,plus(s(?z:Nat),plus(?z:Nat,?x:Nat)))) = times(?y:Nat,s(times(s(0),?x:Nat)))",
"plus(times(times(0,times(?x:Nat,?y:Nat)),s(s(?x:Nat))),s(times(times(?z:Nat,?y:Nat),?z:Nat))) = times(0,s(times(0,?x:Nat)))",
"times(plus(plus(0,?y:Nat),plus(0,0)),plus(s(?x:Nat),times(?z:Nat,plus(?z:Nat,0)))) = plus(s(?y:Nat),plus(0,?x:Nat))",
"plus(s(?x:Nat),times(times(plus(?x:Nat,?x:Nat),times(?y:Nat,0)),times(s(?y:Nat),times(?x:Nat,?y:Nat)))) = plus(?y:Nat,plus(?z:Nat,0))",
"plus(times(s(s(?y:Nat)),s(plus(0,?z:Nat))),plus(0,times(s(?x:Nat),plus(?x:Nat,?z:Nat)))) = s(s(s(s(?x:Nat))))",
"times(plus(times(plus(?z:Nat,0),times(?x:Nat,?y:Nat)),plus(?x:Nat,times(?y:Nat,?x:Nat))),times(?z:Nat,plus(?y:Nat,0))) = plus(?x:Nat,?y:Nat)",
"plus(s(plus(times(0,?x:Nat),times(?y:Nat,?x:Nat))),plus(times(plus(0,?y:Nat),?x:Nat),times(times(?z:Nat,?x:Nat),0))) = s(0)",
"times(s(plus(times(?y:Nat,0),?z:Nat)),plus(plus(0,times(0,?z:Nat)),times(times(?y:Nat,?x:Nat),times(?x:Nat,?x:Nat)))) = s(?y:Nat)",
"s(?z:Nat) = times(plus(times(times(0,?y:Nat),s(?y:Nat)),plus(s(?z:Nat),times(?y:Nat,?z:Nat))),plus(?x:Nat,times(times(?z:Nat,?y:Nat),?z:Nat)))",
"plus(?z:Nat,s(0)) = plus(plus(times(plus(?x:Nat,?z:Nat),?x:Nat),?x:Nat),plus(times(?y:Nat,?x:Nat),plus(plus(0,0),plus(?x:Nat,?x:Nat))))",
"times(?x:Nat,s(s(0))) = times(plus(s(?y:Nat),plus(s(?x:Nat),plus(?z:Nat,?x:Nat))),s(times(times(?x:Nat,0),plus(?y:Nat,?x:Nat))))",
"times(0,plus(?z:Nat,?z:Nat)) = times(plus(?x:Nat,plus(plus(?x:Nat,?z:Nat),s(?z:Nat))),plus(plus(?z:Nat,?x:Nat),plus(?z:Nat,times(?x:Nat,?x:Nat))))",
"s(plus(?z:Nat,s(times(?z:Nat,?y:Nat)))) = times(plus(plus(s(0),times(?x:Nat,?x:Nat)),s(?z:Nat)),times(times(?z:Nat,?x:Nat),s(?y:Nat)))",
"plus(?x:Nat,plus(plus(?x:Nat,?x:Nat),?x:Nat)) = plus(times(s(?z:Nat),plus(?y:Nat,s(?x:Nat))),plus(s(?x:Nat),plus(s(0),s(?x:Nat))))",
"plus(s(s(plus(?z:Nat,0))),0) = plus(times(plus(s(?z:Nat),s(?x:Nat)),plus(s(?y:Nat),s(?z:Nat))),plus(?x:Nat,s(0)))",
"plus(plus(s(?y:Nat),s(?z:Nat)),?x:Nat) = times(0,times(times(s(?x:Nat),times(0,?y:Nat)),times(times(?y:Nat,?x:Nat),times(?x:Nat,?x:Nat))))",
"s(plus(s(0),s(times(?z:Nat,?y:Nat)))) = plus(s(times(times(?x:Nat,?y:Nat),times(?z:Nat,?z:Nat))),plus(s(times(?x:Nat,?y:Nat)),0))",
"s(times(?y:Nat,times(times(?x:Nat,?x:Nat),times(?x:Nat,?z:Nat)))) = plus(times(?y:Nat,?z:Nat),plus(times(times(?x:Nat,?y:Nat),0),plus(0,?x:Nat)))",
"s(times(0,plus(plus(?z:Nat,?x:Nat),plus(?y:Nat,?x:Nat)))) = times(s(times(?x:Nat,plus(?y:Nat,?z:Nat))),s(times(times(?z:Nat,?y:Nat),0)))",
"plus(?x:Nat,times(times(?y:Nat,0),plus(0,s(?y:Nat)))) = plus(times(0,s(times(?z:Nat,?y:Nat))),s(times(times(0,?z:Nat),?y:Nat)))",
"plus(plus(0,times(times(?x:Nat,?z:Nat),s(?x:Nat))),?y:Nat) = times(plus(?y:Nat,times(0,0)),times(0,times(plus(?z:Nat,?z:Nat),?x:Nat)))",
"plus(times(plus(?x:Nat,0),s(times(?x:Nat,?y:Nat))),?x:Nat) = times(plus(times(?x:Nat,s(0)),times(0,times(?x:Nat,?x:Nat))),s(0))",
"plus(0,plus(times(s(?x:Nat),times(?x:Nat,0)),s(?z:Nat))) = s(times(s(plus(?x:Nat,?x:Nat)),plus(plus(?y:Nat,0),s(?z:Nat))))",
"plus(s(plus(?x:Nat,?z:Nat)),times(s(times(0,?y:Nat)),0)) = plus(s(plus(s(?x:Nat),?y:Nat)),plus(s(plus(?z:Nat,?x:Nat)),?x:Nat))",
"plus(plus(plus(?z:Nat,times(0,?x:Nat)),times(0,?z:Nat)),?x:Nat) = times(?y:Nat,times(s(?x:Nat),times(times(?z:Nat,?z:Nat),times(?x:Nat,?z:Nat))))",
"times(plus(?x:Nat,times(?y:Nat,plus(?y:Nat,?z:Nat))),s(s(?z:Nat))) = times(plus(plus(?y:Nat,?x:Nat),times(times(?y:Nat,?z:Nat),s(?z:Nat))),0)",
"times(s(plus(?x:Nat,?y:Nat)),times(?y:Nat,plus(times(?x:Nat,?z:Nat),?x:Nat))) = times(0,plus(plus(times(0,?x:Nat),times(?x:Nat,0)),?x:Nat))",
"times(s(plus(s(?x:Nat),plus(0,0))),s(s(s(?x:Nat)))) = plus(?x:Nat,times(0,plus(times(?y:Nat,?x:Nat),times(?y:Nat,?x:Nat))))",
"s(plus(times(s(0),s(?x:Nat)),times(s(0),times(?x:Nat,0)))) = times(plus(?x:Nat,s(times(?y:Nat,0))),plus(?x:Nat,?x:Nat))",
"plus(?x:Nat,times(plus(s(?x:Nat),plus(?x:Nat,?x:Nat)),s(times(?z:Nat,0)))) = plus(times(?x:Nat,?z:Nat),times(s(plus(?z:Nat,?x:Nat)),0))",
"times(?z:Nat,times(times(times(?z:Nat,?y:Nat),plus(?x:Nat,0)),s(s(?x:Nat)))) = times(s(?z:Nat),times(?x:Nat,times(times(0,?x:Nat),?y:Nat)))",
"times(plus(s(s(?x:Nat)),0),times(0,plus(?y:Nat,times(?x:Nat,?y:Nat)))) = times(plus(times(plus(0,?x:Nat),?y:Nat),s(?x:Nat)),?x:Nat)",
"times(plus(plus(plus(?z:Nat,?z:Nat),?x:Nat),times(s(?y:Nat),?y:Nat)),s(?y:Nat)) = s(times(times(s(?x:Nat),?z:Nat),plus(s(0),0)))",
"times(times(?x:Nat,s(s(?x:Nat))),times(?x:Nat,times(plus(?y:Nat,?x:Nat),?x:Nat))) = s(plus(0,times(plus(?x:Nat,0),times(?y:Nat,?x:Nat))))",
"times(times(?z:Nat,s(?x:Nat)),times(plus(plus(?z:Nat,?x:Nat),s(?x:Nat)),?z:Nat)) = plus(?x:Nat,plus(plus(?y:Nat,s(?z:Nat)),s(s(0))))",
"plus(s(plus(s(?z:Nat),times(?y:Nat,?x:Nat))),plus(?z:Nat,s(plus(?x:Nat,?x:Nat)))) = plus(times(?x:Nat,0),plus(?z:Nat,plus(?x:Nat,0)))",
"plus(plus(plus(0,s(?x:Nat)),plus(times(?x:Nat,0),plus(?x:Nat,?y:Nat))),0) = times(plus(?z:Nat,plus(?y:Nat,times(?x:Nat,0))),?z:Nat)",
"plus(times(?x:Nat,times(?x:Nat,s(?z:Nat))),s(times(times(?y:Nat,?x:Nat),plus(?z:Nat,?x:Nat)))) = s(times(?z:Nat,times(plus(0,0),?x:Nat)))",
"times(times(s(?x:Nat),plus(s(0),plus(?x:Nat,?z:Nat))),s(plus(s(?x:Nat),?x:Nat))) = plus(plus(?z:Nat,s(times(?z:Nat,?z:Nat))),?y:Nat)",
"plus(times(times(times(?x:Nat,0),times(0,?x:Nat)),plus(0,?x:Nat)),times(?y:Nat,s(plus(?y:Nat,?z:Nat)))) = s(s(plus(?x:Nat,?x:Nat)))",
"times(plus(s(?z:Nat),times(plus(0,?y:Nat),s(0))),times(plus(s(0),plus(?x:Nat,0)),?y:Nat)) = times(plus(?x:Nat,?x:Nat),?z:Nat)",
"times(plus(plus(times(?y:Nat,?z:Nat),plus(?x:Nat,0)),?y:Nat),plus(times(?x:Nat,?z:Nat),times(times(?y:Nat,?x:Nat),?x:Nat))) = times(s(?x:Nat),?x:Nat)",
"times(times(plus(times(?x:Nat,?x:Nat),plus(0,?x:Nat)),plus(plus(?y:Nat,?z:Nat),0)),s(plus(0,times(?x:Nat,?y:Nat)))) = times(?y:Nat,?x:Nat)",
"plus(?x:Nat,?x:Nat) = plus(plus(0,times(s(?x:Nat),plus(?x:Nat,?z:Nat))),times(times(plus(0,0),times(?y:Nat,0)),s(times(?y:Nat,?x:Nat))))",
"times(s(?x:Nat),0) = plus(plus(plus(?y:Nat,?x:Nat),times(times(?y:Nat,?y:Nat),?x:Nat)),times(plus(times(?x:Nat,?z:Nat),plus(?x:Nat,?z:Nat)),s(?x:Nat)))",
"plus(0,times(?z:Nat,?y:Nat)) = times(plus(times(0,0),times(plus(?x:Nat,0),plus(0,?z:Nat))),times(?x:Nat,times(?x:Nat,times(?z:Nat,?x:Nat))))",
"s(plus(?z:Nat,plus(plus(?x:Nat,?y:Nat),times(?x:Nat,?x:Nat)))) = plus(times(plus(times(0,?z:Nat),s(0)),?x:Nat),plus(0,s(s(?x:Nat))))",
"plus(?x:Nat,plus(0,plus(plus(0,?z:Nat),times(?z:Nat,?z:Nat)))) = s(plus(s(times(?z:Nat,?z:Nat)),times(times(?z:Nat,?x:Nat),times(0,?z:Nat))))",
"times(times(?x:Nat,?x:Nat),times(plus(s(?x:Nat),?y:Nat),s(0))) = plus(?z:Nat,plus(plus(plus(?x:Nat,?z:Nat),s(?z:Nat)),s(plus(?x:Nat,?x:Nat))))",
"s(times(times(times(?x:Nat,0),times(0,?x:Nat)),times(0,?x:Nat))) = plus(plus(times(s(?z:Nat),times(?x:Nat,0)),?x:Nat),times(?x:Nat,0))",
"s(plus(plus(plus(?y:Nat,?x:Nat),?x:Nat),times(s(0),times(?y:Nat,?z:Nat)))) = plus(plus(?x:Nat,0),plus(?z:Nat,plus(?y:Nat,times(?z:Nat,?x:Nat))))",
"plus(?x:Nat,times(plus(s(?y:Nat),s(?x:Nat)),plus(0,plus(0,?x:Nat)))) = plus(?z:Nat,plus(plus(plus(?z:Nat,?z:Nat),plus(?x:Nat,?y:Nat)),?y:Nat))",
"plus(plus(s(times(?x:Nat,?z:Nat)),?x:Nat),s(times(plus(?z:Nat,?y:Nat),?y:Nat))) = plus(plus(0,times(plus(?y:Nat,0),plus(?z:Nat,?x:Nat))),?y:Nat)",
"plus(times(s(times(0,?x:Nat)),?z:Nat),times(?y:Nat,plus(s(?y:Nat),?x:Nat))) = times(?y:Nat,plus(times(plus(?x:Nat,?y:Nat),plus(0,?x:Nat)),?x:Nat))",
"plus(times(plus(plus(?x:Nat,?z:Nat),?x:Nat),?z:Nat),plus(?z:Nat,s(s(?y:Nat)))) = plus(0,times(?y:Nat,times(times(?x:Nat,?x:Nat),plus(?y:Nat,?x:Nat))))",
"plus(times(times(s(?z:Nat),?y:Nat),?x:Nat),times(s(plus(?z:Nat,?y:Nat)),?z:Nat)) = times(0,plus(s(plus(0,?z:Nat)),times(?x:Nat,s(?y:Nat))))",
"times(s(s(times(?y:Nat,?x:Nat))),times(times(plus(?z:Nat,0),?z:Nat),?z:Nat)) = plus(plus(s(plus(0,?z:Nat)),s(?x:Nat)),times(?z:Nat,0))",
"times(times(s(0),s(?y:Nat)),s(times(s(?x:Nat),times(?x:Nat,?x:Nat)))) = times(?z:Nat,times(times(times(0,?y:Nat),times(?z:Nat,?y:Nat)),?z:Nat))",
"s(times(times(?y:Nat,times(?x:Nat,?x:Nat)),plus(times(?z:Nat,?x:Nat),plus(?z:Nat,?z:Nat)))) = plus(0,plus(0,times(s(0),plus(0,0))))",
"times(plus(plus(times(?x:Nat,?x:Nat),plus(0,?x:Nat)),s(plus(?x:Nat,0))),?x:Nat) = times(times(?x:Nat,s(s(?x:Nat))),plus(s(?y:Nat),0))",
"plus(plus(?x:Nat,times(plus(?z:Nat,?z:Nat),times(?z:Nat,0))),times(s(?x:Nat),s(?z:Nat))) = s(plus(times(0,s(?y:Nat)),s(s(?x:Nat))))",
"times(plus(s(s(0)),times(s(0),?z:Nat)),s(plus(plus(?x:Nat,?z:Nat),?x:Nat))) = plus(s(?z:Nat),s(times(times(?x:Nat,?x:Nat),0)))",
"times(times(s(plus(?x:Nat,?z:Nat)),?y:Nat),plus(?x:Nat,times(s(0),plus(?z:Nat,?x:Nat)))) = s(times(?x:Nat,times(s(?x:Nat),plus(?y:Nat,?z:Nat))))",
"plus(plus(?x:Nat,s(s(?y:Nat))),times(times(?x:Nat,?z:Nat),plus(plus(?x:Nat,?x:Nat),s(?x:Nat)))) = times(?y:Nat,plus(s(times(?x:Nat,?y:Nat)),0))",
"plus(plus(plus(plus(0,0),plus(?x:Nat,?x:Nat)),0),s(times(times(0,?y:Nat),?x:Nat))) = plus(plus(plus(s(?y:Nat),?z:Nat),?x:Nat),?y:Nat)",
"times(times(plus(plus(?z:Nat,?y:Nat),times(?z:Nat,?z:Nat)),0),s(plus(0,plus(?z:Nat,0)))) = times(times(?x:Nat,?x:Nat),s(plus(?y:Nat,?x:Nat)))",
"times(plus(times(times(?x:Nat,?z:Nat),?z:Nat),s(s(?x:Nat))),times(?x:Nat,times(plus(?z:Nat,?x:Nat),?x:Nat))) = s(plus(?x:Nat,s(times(0,?z:Nat))))",
"times(times(plus(0,0),0),plus(times(?x:Nat,s(?y:Nat)),times(plus(?x:Nat,?z:Nat),s(?z:Nat)))) = times(s(?x:Nat),plus(s(?x:Nat),?z:Nat))",
"plus(plus(?x:Nat,plus(times(?x:Nat,0),s(?x:Nat))),times(plus(s(?y:Nat),?x:Nat),s(times(?x:Nat,?x:Nat)))) = s(s(s(times(0,?x:Nat))))",
"times(plus(plus(0,plus(?x:Nat,?x:Nat)),plus(s(?y:Nat),?z:Nat)),times(times(plus(?x:Nat,?z:Nat),?x:Nat),times(times(?y:Nat,?y:Nat),?z:Nat))) = s(0)",
"times(?x:Nat,0) = plus(times(plus(s(?z:Nat),plus(?y:Nat,?x:Nat)),times(?z:Nat,0)),plus(plus(plus(?z:Nat,?y:Nat),plus(?z:Nat,0)),plus(?z:Nat,?z:Nat)))",
"plus(?y:Nat,s(s(0))) = times(plus(s(plus(0,?x:Nat)),plus(?y:Nat,plus(?x:Nat,?y:Nat))),times(plus(plus(?y:Nat,?z:Nat),times(?y:Nat,?y:Nat)),?z:Nat))",
"times(0,plus(?x:Nat,?z:Nat)) = plus(plus(plus(?x:Nat,plus(?y:Nat,0)),plus(plus(?y:Nat,?x:Nat),?y:Nat)),plus(plus(plus(?x:Nat,?x:Nat),s(?x:Nat)),?z:Nat))",
"s(plus(times(?z:Nat,?x:Nat),s(?x:Nat))) = times(s(plus(?x:Nat,times(?z:Nat,?x:Nat))),times(times(plus(?y:Nat,0),s(?x:Nat)),s(times(?x:Nat,0))))",
"s(times(s(plus(0,?x:Nat)),?x:Nat)) = times(times(?x:Nat,?x:Nat),times(times(times(?z:Nat,?z:Nat),times(?z:Nat,?y:Nat)),times(times(0,0),s(0))))",
"s(times(plus(s(?z:Nat),?x:Nat),0)) = times(plus(times(plus(?z:Nat,?x:Nat),s(?x:Nat)),?x:Nat),plus(0,plus(times(0,?y:Nat),times(0,?x:Nat))))",
"times(?x:Nat,s(plus(s(?z:Nat),s(?x:Nat)))) = plus(plus(plus(plus(?z:Nat,?y:Nat),plus(?x:Nat,?y:Nat)),times(times(?x:Nat,?z:Nat),times(?y:Nat,?x:Nat))),0)",
"s(plus(plus(times(?x:Nat,?x:Nat),?y:Nat),plus(?y:Nat,?y:Nat))) = plus(times(0,times(?y:Nat,plus(?y:Nat,?x:Nat))),plus(times(?y:Nat,times(0,?x:Nat)),?x:Nat))",
"plus(s(0),plus(?x:Nat,times(times(?x:Nat,?x:Nat),?z:Nat))) = times(times(0,times(plus(?z:Nat,?x:Nat),?z:Nat)),plus(?x:Nat,plus(plus(?x:Nat,?x:Nat),?z:Nat)))",
"times(plus(plus(plus(0,?x:Nat),?y:Nat),?y:Nat),s(?z:Nat)) = times(times(?x:Nat,?x:Nat),times(plus(times(?x:Nat,?z:Nat),?z:Nat),plus(plus(?z:Nat,0),0)))",
"times(0,times(?x:Nat,plus(times(?z:Nat,?z:Nat),times(?x:Nat,?z:Nat)))) = plus(times(?x:Nat,?x:Nat),times(times(plus(?z:Nat,0),s(0)),times(?z:Nat,?y:Nat)))",
"times(plus(times(?x:Nat,s(?x:Nat)),?x:Nat),times(s(?x:Nat),0)) = plus(?z:Nat,plus(times(plus(?z:Nat,?y:Nat),s(?x:Nat)),times(?x:Nat,times(?x:Nat,?z:Nat))))",
"plus(plus(s(times(?z:Nat,?y:Nat)),times(times(?x:Nat,?z:Nat),?x:Nat)),?x:Nat) = plus(plus(plus(times(?x:Nat,?y:Nat),s(?z:Nat)),times(?y:Nat,s(0))),?y:Nat)",
"times(plus(s(plus(?z:Nat,?x:Nat)),0),s(s(times(?z:Nat,0)))) = plus(?y:Nat,times(s(plus(0,?z:Nat)),times(times(0,?z:Nat),s(?y:Nat))))",
"plus(plus(times(s(?z:Nat),?x:Nat),?z:Nat),times(?x:Nat,s(plus(?y:Nat,0)))) = plus(plus(plus(times(?y:Nat,?x:Nat),?y:Nat),times(s(?x:Nat),?x:Nat)),?y:Nat)",
"times(times(?x:Nat,s(0)),s(plus(plus(?z:Nat,?x:Nat),times(0,?x:Nat)))) = plus(0,times(s(?y:Nat),times(times(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat))))",
"plus(s(plus(s(?z:Nat),times(?x:Nat,?x:Nat))),times(times(s(?y:Nat),0),s(?z:Nat))) = s(times(times(s(?x:Nat),times(0,?z:Nat)),s(?x:Nat)))",
"plus(plus(times(s(?y:Nat),times(?x:Nat,0)),plus(times(?x:Nat,?z:Nat),s(?x:Nat))),?x:Nat) = s(times(times(times(?x:Nat,?y:Nat),?x:Nat),times(0,?z:Nat)))",
"plus(times(?y:Nat,?x:Nat),plus(plus(s(0),s(?y:Nat)),plus(?z:Nat,times(?y:Nat,?y:Nat)))) = times(?x:Nat,times(s(?x:Nat),plus(plus(?y:Nat,?x:Nat),?y:Nat)))",
"plus(times(s(s(?x:Nat)),s(0)),s(plus(plus(?z:Nat,?y:Nat),plus(?x:Nat,?x:Nat)))) = s(times(s(s(?y:Nat)),plus(?x:Nat,times(0,?z:Nat))))",
"times(plus(times(times(0,?x:Nat),?x:Nat),plus(times(0,?x:Nat),?y:Nat)),times(?z:Nat,?x:Nat)) = plus(s(0),times(times(?y:Nat,times(?x:Nat,?y:Nat)),?z:Nat))",
"plus(s(plus(s(?x:Nat),?y:Nat)),plus(plus(plus(?y:Nat,0),s(?y:Nat)),plus(?z:Nat,0))) = s(s(times(times(?z:Nat,?z:Nat),times(?z:Nat,?y:Nat))))",
"plus(times(?z:Nat,?y:Nat),times(times(s(?x:Nat),?x:Nat),times(plus(?z:Nat,0),times(0,?x:Nat)))) = times(?y:Nat,plus(plus(?y:Nat,plus(?y:Nat,?x:Nat)),0))",
"plus(times(s(s(?z:Nat)),?z:Nat),plus(plus(times(0,?z:Nat),?x:Nat),times(0,s(0)))) = s(plus(plus(plus(?x:Nat,?x:Nat),s(0)),?x:Nat))",
"plus(times(times(?z:Nat,plus(?x:Nat,?x:Nat)),0),times(plus(s(0),times(?y:Nat,?x:Nat)),?x:Nat)) = plus(times(0,plus(?y:Nat,plus(?x:Nat,?x:Nat))),?y:Nat)",
"times(plus(times(s(?y:Nat),times(?y:Nat,?y:Nat)),?x:Nat),plus(plus(times(?x:Nat,?x:Nat),?z:Nat),?z:Nat)) = plus(plus(plus(?x:Nat,?y:Nat),?y:Nat),plus(0,0))",
"times(times(plus(plus(?z:Nat,?x:Nat),?z:Nat),?x:Nat),times(times(plus(?x:Nat,?x:Nat),?z:Nat),s(?y:Nat))) = s(plus(times(0,?y:Nat),times(?y:Nat,s(?z:Nat))))",
"plus(times(times(?x:Nat,plus(0,0)),times(times(0,?z:Nat),plus(?x:Nat,?y:Nat))),plus(?x:Nat,?x:Nat)) = times(plus(times(s(0),?x:Nat),?x:Nat),0)",
"plus(times(plus(times(0,?y:Nat),plus(?x:Nat,?x:Nat)),?x:Nat),times(times(plus(0,0),0),s(times(?z:Nat,0)))) = plus(?x:Nat,plus(0,?x:Nat))",
"times(0,s(?y:Nat)) = times(plus(plus(s(?y:Nat),times(?x:Nat,?x:Nat)),?y:Nat),plus(times(plus(?x:Nat,?y:Nat),s(?y:Nat)),times(times(?x:Nat,?z:Nat),s(?x:Nat))))",
"plus(plus(?x:Nat,?x:Nat),?x:Nat) = times(times(s(times(?x:Nat,?x:Nat)),plus(s(?z:Nat),0)),plus(plus(s(?z:Nat),?y:Nat),times(times(?x:Nat,?x:Nat),s(?z:Nat))))",
"times(?z:Nat,s(plus(times(0,?x:Nat),?z:Nat))) = plus(times(times(?y:Nat,times(0,0)),times(times(0,?x:Nat),s(0))),plus(?x:Nat,plus(?z:Nat,?y:Nat)))",
"s(times(?x:Nat,plus(times(?z:Nat,?z:Nat),s(0)))) = plus(plus(plus(plus(?x:Nat,?y:Nat),?x:Nat),times(?x:Nat,plus(?z:Nat,?y:Nat))),s(s(plus(?y:Nat,?y:Nat))))",
"times(plus(?x:Nat,?z:Nat),times(times(?x:Nat,?z:Nat),?y:Nat)) = plus(times(times(s(?y:Nat),times(?x:Nat,0)),?x:Nat),plus(?y:Nat,times(plus(?z:Nat,?x:Nat),s(?z:Nat))))",
"plus(?y:Nat,plus(times(s(?x:Nat),plus(?x:Nat,?x:Nat)),?z:Nat)) = times(s(?x:Nat),plus(times(times(?y:Nat,?x:Nat),s(?x:Nat)),times(times(?z:Nat,0),s(0))))",
"plus(s(s(plus(?x:Nat,?x:Nat))),times(?y:Nat,s(plus(?x:Nat,?x:Nat)))) = plus(times(times(plus(?x:Nat,?y:Nat),times(0,?x:Nat)),s(0)),s(s(?x:Nat)))",
"times(?z:Nat,plus(s(plus(?z:Nat,?x:Nat)),plus(0,plus(?x:Nat,0)))) = times(times(?x:Nat,?x:Nat),plus(s(s(?y:Nat)),plus(s(?y:Nat),plus(?y:Nat,?x:Nat))))",
"plus(s(times(?x:Nat,plus(?z:Nat,?y:Nat))),plus(plus(s(?z:Nat),?x:Nat),?x:Nat)) = plus(0,plus(s(times(0,?x:Nat)),times(s(?z:Nat),times(?x:Nat,?z:Nat))))",
"plus(?x:Nat,times(s(plus(?z:Nat,?y:Nat)),times(plus(?x:Nat,?y:Nat),plus(?x:Nat,?x:Nat)))) = plus(s(?y:Nat),plus(times(plus(?z:Nat,0),plus(?x:Nat,?x:Nat)),?x:Nat))",
"times(plus(?y:Nat,?x:Nat),plus(plus(s(0),s(?z:Nat)),plus(s(0),0))) = times(s(plus(times(?x:Nat,?x:Nat),?y:Nat)),s(plus(s(?z:Nat),?z:Nat)))",
"times(plus(plus(?x:Nat,plus(?x:Nat,0)),s(?z:Nat)),plus(plus(?x:Nat,?y:Nat),times(?x:Nat,times(?x:Nat,0)))) = times(plus(0,plus(s(?z:Nat),?x:Nat)),?y:Nat)",
"times(plus(plus(s(?y:Nat),times(?z:Nat,0)),s(?z:Nat)),times(s(times(?z:Nat,?x:Nat)),s(s(?x:Nat)))) = times(s(?z:Nat),plus(plus(0,?z:Nat),?x:Nat))",
"plus(?x:Nat,plus(?y:Nat,s(plus(0,?y:Nat)))) = times(plus(?x:Nat,times(?x:Nat,?z:Nat)),times(times(times(?z:Nat,0),plus(?z:Nat,?z:Nat)),plus(?z:Nat,plus(?y:Nat,?x:Nat))))",
"plus(s(?y:Nat),times(0,times(times(?x:Nat,?x:Nat),0))) = times(plus(times(s(?z:Nat),?z:Nat),times(times(?x:Nat,?y:Nat),plus(?x:Nat,?x:Nat))),times(?y:Nat,s(?x:Nat)))",
"plus(times(?x:Nat,0),times(?y:Nat,s(times(?x:Nat,?x:Nat)))) = plus(times(times(times(?y:Nat,?x:Nat),times(?x:Nat,?z:Nat)),?y:Nat),s(times(s(?z:Nat),plus(?z:Nat,?z:Nat))))",
"plus(?z:Nat,times(?z:Nat,plus(plus(?z:Nat,0),times(?z:Nat,?y:Nat)))) = times(plus(?x:Nat,times(times(?x:Nat,?x:Nat),plus(?x:Nat,?z:Nat))),plus(times(?y:Nat,s(?x:Nat)),?x:Nat))",
"times(times(?x:Nat,?z:Nat),times(?x:Nat,times(times(?x:Nat,?y:Nat),0))) = plus(s(times(plus(?y:Nat,?x:Nat),plus(?y:Nat,0))),plus(plus(0,plus(0,0)),?x:Nat))",
"s(plus(plus(plus(?x:Nat,?x:Nat),0),plus(plus(?x:Nat,?y:Nat),?x:Nat))) = plus(plus(?x:Nat,times(times(?x:Nat,?y:Nat),?x:Nat)),plus(?x:Nat,times(s(?y:Nat),s(?x:Nat))))",
"times(plus(plus(?x:Nat,?x:Nat),times(times(0,?x:Nat),?z:Nat)),times(?y:Nat,0)) = times(times(plus(?z:Nat,times(?z:Nat,?x:Nat)),?x:Nat),plus(plus(s(?z:Nat),?y:Nat),?x:Nat))",
"times(?z:Nat,plus(times(plus(0,?x:Nat),plus(?y:Nat,0)),plus(times(?y:Nat,?x:Nat),?x:Nat))) = plus(plus(0,?y:Nat),plus(plus(plus(?y:Nat,?x:Nat),s(0)),?x:Nat))",
"plus(0,plus(plus(s(0),times(?z:Nat,?x:Nat)),plus(times(?x:Nat,?x:Nat),times(?x:Nat,?x:Nat)))) = times(?y:Nat,plus(times(?x:Nat,?x:Nat),plus(times(?x:Nat,?x:Nat),?x:Nat)))",
"times(plus(0,s(plus(?x:Nat,?z:Nat))),times(s(s(?y:Nat)),times(times(?z:Nat,?x:Nat),plus(?x:Nat,?x:Nat)))) = times(times(?x:Nat,times(?y:Nat,plus(?x:Nat,?z:Nat))),0)",
"times(plus(plus(plus(?x:Nat,?y:Nat),times(?x:Nat,?z:Nat)),s(0)),plus(?y:Nat,times(?z:Nat,plus(?x:Nat,?y:Nat)))) = s(times(plus(times(?x:Nat,0),s(?x:Nat)),0))",
"times(s(plus(?x:Nat,times(0,?z:Nat))),plus(plus(times(?x:Nat,?x:Nat),plus(?y:Nat,?z:Nat)),plus(plus(?x:Nat,?z:Nat),?x:Nat))) = times(times(0,?x:Nat),times(?x:Nat,?z:Nat))",
"times(plus(plus(plus(?z:Nat,?x:Nat),times(0,0)),plus(plus(0,?x:Nat),plus(?z:Nat,?x:Nat))),plus(?x:Nat,s(?x:Nat))) = plus(plus(?y:Nat,?y:Nat),times(?y:Nat,?x:Nat))",
"s(?x:Nat) = times(times(plus(plus(?x:Nat,?y:Nat),?x:Nat),plus(plus(?y:Nat,?x:Nat),plus(?y:Nat,?x:Nat))),plus(s(times(0,?x:Nat)),plus(plus(?y:Nat,?y:Nat),plus(?z:Nat,?y:Nat))))",
"plus(plus(times(?z:Nat,times(0,0)),?x:Nat),s(?z:Nat)) = times(plus(?x:Nat,plus(s(?x:Nat),s(0))),plus(s(times(?z:Nat,?z:Nat)),times(times(?x:Nat,?z:Nat),?z:Nat)))",
"times(?x:Nat,plus(times(0,times(?x:Nat,?y:Nat)),s(?x:Nat))) = plus(plus(plus(times(?x:Nat,0),?x:Nat),times(?y:Nat,?y:Nat)),times(?x:Nat,plus(s(?x:Nat),times(?x:Nat,?x:Nat))))",
"times(plus(?y:Nat,times(?y:Nat,?z:Nat)),times(?y:Nat,s(plus(?x:Nat,?x:Nat)))) = times(?z:Nat,plus(plus(times(?z:Nat,0),times(?x:Nat,0)),times(s(?x:Nat),plus(?x:Nat,?x:Nat))))",
"plus(?z:Nat,times(times(times(0,?y:Nat),s(?z:Nat)),plus(s(?x:Nat),?z:Nat))) = plus(times(times(s(?x:Nat),plus(?z:Nat,?z:Nat)),?z:Nat),s(times(?y:Nat,times(?x:Nat,?x:Nat))))",
"times(times(?y:Nat,s(s(?x:Nat))),times(times(plus(?z:Nat,?z:Nat),s(0)),?x:Nat)) = plus(times(times(s(?x:Nat),s(?x:Nat)),?y:Nat),s(plus(?y:Nat,plus(?z:Nat,?y:Nat))))",
"plus(plus(0,times(times(?y:Nat,?x:Nat),s(?x:Nat))),times(?x:Nat,plus(s(?z:Nat),s(?x:Nat)))) = times(plus(times(plus(0,?z:Nat),times(?x:Nat,?y:Nat)),s(?x:Nat)),?y:Nat)",
"times(times(0,plus(?z:Nat,0)),times(plus(?x:Nat,times(?x:Nat,?y:Nat)),times(s(?x:Nat),times(?x:Nat,?x:Nat)))) = plus(?z:Nat,plus(0,plus(s(?y:Nat),times(?x:Nat,0))))",
"plus(times(plus(times(?y:Nat,0),0),s(plus(?y:Nat,?z:Nat))),s(plus(plus(?y:Nat,0),plus(?z:Nat,?x:Nat)))) = times(0,s(plus(s(?y:Nat),times(?x:Nat,?y:Nat))))",
"times(times(times(?x:Nat,times(?x:Nat,?x:Nat)),times(?z:Nat,times(?y:Nat,?x:Nat))),plus(s(times(0,?y:Nat)),plus(?y:Nat,times(0,?y:Nat)))) = times(0,plus(?x:Nat,s(?x:Nat)))",
"times(plus(0,?x:Nat),?z:Nat) = plus(plus(times(?z:Nat,?x:Nat),times(times(?y:Nat,?z:Nat),plus(0,?z:Nat))),times(times(plus(0,?x:Nat),plus(?y:Nat,?z:Nat)),s(times(?x:Nat,?z:Nat))))",
"times(plus(s(s(?z:Nat)),s(?x:Nat)),0) = times(s(plus(times(?z:Nat,?y:Nat),times(?y:Nat,?x:Nat))),times(s(times(?x:Nat,0)),times(plus(?y:Nat,?x:Nat),times(?z:Nat,?z:Nat))))",
"s(plus(times(s(?z:Nat),plus(?x:Nat,?x:Nat)),?y:Nat)) = plus(times(plus(?x:Nat,s(?x:Nat)),plus(times(?x:Nat,?y:Nat),0)),plus(?x:Nat,plus(times(0,?y:Nat),times(?x:Nat,?y:Nat))))",
"times(plus(?x:Nat,?x:Nat),s(times(times(?x:Nat,?y:Nat),plus(?x:Nat,?z:Nat)))) = plus(times(plus(?x:Nat,s(0)),s(s(?x:Nat))),times(?x:Nat,plus(s(?x:Nat),times(?x:Nat,?z:Nat))))",
"s(times(times(plus(?y:Nat,?y:Nat),times(?z:Nat,?z:Nat)),times(?x:Nat,times(?z:Nat,?y:Nat)))) = plus(plus(plus(s(?z:Nat),s(?x:Nat)),plus(plus(0,?y:Nat),plus(?x:Nat,?z:Nat))),0)",
"plus(plus(?x:Nat,times(times(?x:Nat,?y:Nat),?x:Nat)),times(times(s(?z:Nat),plus(0,?x:Nat)),?y:Nat)) = times(plus(plus(0,times(?x:Nat,?x:Nat)),times(times(?y:Nat,?x:Nat),?y:Nat)),?z:Nat)",
"plus(times(s(plus(0,?x:Nat)),plus(?y:Nat,plus(?x:Nat,?x:Nat))),times(?x:Nat,plus(s(?x:Nat),s(?y:Nat)))) = plus(times(plus(?y:Nat,?x:Nat),plus(?z:Nat,plus(?x:Nat,?y:Nat))),?x:Nat)",
"times(plus(s(s(?x:Nat)),?x:Nat),times(plus(0,plus(?y:Nat,?x:Nat)),times(s(?x:Nat),times(0,?x:Nat)))) = plus(?y:Nat,plus(times(times(0,0),plus(?x:Nat,0)),?x:Nat))",
"times(times(times(times(?z:Nat,?x:Nat),?y:Nat),plus(s(?z:Nat),s(?y:Nat))),s(plus(s(?x:Nat),s(?y:Nat)))) = s(plus(s(times(?y:Nat,?x:Nat)),plus(times(0,?x:Nat),?x:Nat)))",
"times(times(times(times(?x:Nat,?y:Nat),times(?z:Nat,?x:Nat)),times(times(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat))),times(?z:Nat,0)) = times(s(plus(plus(?y:Nat,?y:Nat),times(0,0))),?z:Nat)",
"times(s(?x:Nat),plus(s(s(?x:Nat)),s(?z:Nat))) = plus(times(plus(plus(?z:Nat,?x:Nat),?y:Nat),0),times(times(plus(?x:Nat,0),0),times(plus(?y:Nat,?x:Nat),times(0,?y:Nat))))",
"times(times(0,plus(plus(?x:Nat,0),?y:Nat)),times(?z:Nat,?x:Nat)) = times(plus(times(plus(?x:Nat,?z:Nat),plus(?x:Nat,?x:Nat)),times(?x:Nat,?y:Nat)),plus(times(s(?z:Nat),?x:Nat),s(?x:Nat)))",
"times(times(s(plus(0,?z:Nat)),?x:Nat),times(times(?z:Nat,?z:Nat),s(?x:Nat))) = times(plus(?x:Nat,times(?z:Nat,plus(?x:Nat,?y:Nat))),plus(?y:Nat,times(times(?x:Nat,?x:Nat),times(?x:Nat,0))))",
"plus(plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),?y:Nat),plus(?y:Nat,plus(?y:Nat,s(?y:Nat)))) = plus(times(?y:Nat,plus(s(0),s(?y:Nat))),plus(times(times(?z:Nat,?x:Nat),s(?z:Nat)),?z:Nat))",
"s(plus(plus(plus(0,?z:Nat),plus(?x:Nat,?z:Nat)),plus(s(?x:Nat),times(?z:Nat,?z:Nat)))) = plus(?y:Nat,plus(times(?z:Nat,times(?z:Nat,?x:Nat)),times(times(?y:Nat,?x:Nat),times(?x:Nat,0))))",
"plus(times(0,?y:Nat),plus(times(times(?z:Nat,0),plus(?z:Nat,?y:Nat)),s(times(0,?y:Nat)))) = plus(s(plus(s(?y:Nat),times(?z:Nat,?x:Nat))),times(?x:Nat,s(plus(?x:Nat,0))))",
"times(times(times(?x:Nat,0),plus(s(?y:Nat),plus(?z:Nat,?z:Nat))),times(?z:Nat,plus(?x:Nat,?z:Nat))) = plus(times(s(plus(?x:Nat,?y:Nat)),plus(s(?z:Nat),s(?x:Nat))),plus(?z:Nat,?y:Nat))",
"s(plus(?x:Nat,plus(times(?z:Nat,?x:Nat),plus(?z:Nat,?x:Nat)))) = plus(plus(s(times(?z:Nat,?y:Nat)),0),times(plus(times(?x:Nat,?x:Nat),plus(?y:Nat,?y:Nat)),plus(s(?z:Nat),plus(?z:Nat,0))))",
"times(?z:Nat,times(s(times(?x:Nat,?x:Nat)),s(s(?x:Nat)))) = times(times(times(?x:Nat,times(0,0)),plus(plus(?x:Nat,?x:Nat),0)),times(s(?y:Nat),times(s(0),plus(?x:Nat,?x:Nat))))",
"times(times(?z:Nat,times(plus(?x:Nat,?x:Nat),plus(?x:Nat,?z:Nat))),?x:Nat) = times(s(plus(times(?z:Nat,?z:Nat),times(?x:Nat,0))),times(times(plus(?z:Nat,?z:Nat),s(?y:Nat)),s(plus(?y:Nat,0))))",
"plus(s(plus(plus(?x:Nat,?x:Nat),s(?y:Nat))),plus(?y:Nat,plus(s(0),0))) = times(s(times(?x:Nat,?z:Nat)),times(plus(times(?x:Nat,0),?x:Nat),plus(s(?x:Nat),plus(0,?y:Nat))))",
"times(s(plus(?x:Nat,s(?z:Nat))),times(plus(times(0,?y:Nat),0),times(?x:Nat,?y:Nat))) = plus(times(times(s(?x:Nat),?y:Nat),?x:Nat),plus(s(plus(?z:Nat,?x:Nat)),times(s(?x:Nat),0)))",
"times(s(times(plus(?x:Nat,?x:Nat),plus(?x:Nat,0))),times(?y:Nat,plus(s(?x:Nat),?y:Nat))) = plus(times(?x:Nat,times(s(?z:Nat),?x:Nat)),plus(s(times(?x:Nat,0)),s(plus(?y:Nat,?y:Nat))))",
"times(times(?x:Nat,plus(times(?y:Nat,?x:Nat),?x:Nat)),plus(times(times(?y:Nat,?x:Nat),plus(?y:Nat,?x:Nat)),?x:Nat)) = times(times(0,s(times(0,?y:Nat))),s(times(plus(?z:Nat,?y:Nat),s(?x:Nat))))",
"plus(plus(0,times(0,?x:Nat)),plus(plus(plus(?x:Nat,0),plus(?y:Nat,?z:Nat)),plus(plus(?x:Nat,0),?x:Nat))) = times(s(times(?x:Nat,?x:Nat)),times(plus(?y:Nat,plus(?z:Nat,?x:Nat)),?x:Nat))",
"times(times(plus(s(?x:Nat),s(?z:Nat)),times(?x:Nat,?x:Nat)),plus(plus(0,s(0)),s(times(0,?y:Nat)))) = times(plus(?x:Nat,s(plus(?z:Nat,0))),times(s(s(0)),?x:Nat))",
"plus(s(s(s(?x:Nat))),plus(plus(s(?x:Nat),s(?z:Nat)),plus(plus(0,?x:Nat),?x:Nat))) = plus(times(plus(times(0,?z:Nat),?x:Nat),?x:Nat),s(plus(plus(?y:Nat,?x:Nat),plus(?z:Nat,?x:Nat))))",
"plus(times(?x:Nat,s(times(?x:Nat,?y:Nat))),plus(plus(?z:Nat,s(?y:Nat)),times(times(?x:Nat,0),?x:Nat))) = plus(s(times(s(?x:Nat),times(?x:Nat,0))),plus(s(?x:Nat),s(times(?z:Nat,?x:Nat))))",
"times(plus(times(times(?z:Nat,?x:Nat),?x:Nat),plus(plus(?z:Nat,?y:Nat),?z:Nat)),s(plus(?x:Nat,times(0,?x:Nat)))) = plus(plus(times(plus(?x:Nat,?x:Nat),?x:Nat),times(plus(?x:Nat,?z:Nat),s(?x:Nat))),?x:Nat)",
"times(times(times(?y:Nat,s(0)),plus(times(?x:Nat,?z:Nat),?y:Nat)),plus(times(plus(?x:Nat,?y:Nat),?y:Nat),s(?x:Nat))) = plus(0,times(plus(0,s(?z:Nat)),plus(plus(?y:Nat,?z:Nat),s(0))))",
"plus(times(plus(times(0,0),times(?x:Nat,?x:Nat)),times(times(?x:Nat,0),s(?z:Nat))),times(times(plus(?y:Nat,?x:Nat),times(?y:Nat,?x:Nat)),times(s(?z:Nat),?z:Nat))) = plus(plus(?y:Nat,?y:Nat),0)",
"times(?y:Nat,times(times(s(0),s(?z:Nat)),plus(?x:Nat,times(0,0)))) = plus(times(times(?z:Nat,plus(?x:Nat,?x:Nat)),plus(s(?z:Nat),0)),plus(s(times(?z:Nat,?x:Nat)),plus(?x:Nat,s(?z:Nat))))",
"plus(plus(s(times(?x:Nat,?x:Nat)),times(?x:Nat,s(?x:Nat))),times(0,s(times(?y:Nat,0)))) = times(plus(s(?x:Nat),plus(s(?z:Nat),s(?x:Nat))),times(plus(s(?y:Nat),plus(?y:Nat,?x:Nat)),?x:Nat))",
"times(times(times(times(0,?z:Nat),times(?z:Nat,0)),?z:Nat),plus(?y:Nat,times(?y:Nat,times(?x:Nat,0)))) = times(times(?z:Nat,plus(times(?z:Nat,?z:Nat),?x:Nat)),plus(0,times(s(?y:Nat),times(?x:Nat,?x:Nat))))",
"times(plus(times(?y:Nat,times(0,?z:Nat)),s(0)),times(plus(s(?x:Nat),0),plus(plus(?y:Nat,?x:Nat),times(?x:Nat,?x:Nat)))) = plus(plus(?x:Nat,times(?z:Nat,s(0))),times(?y:Nat,s(s(?x:Nat))))",
"times(times(s(0),plus(?z:Nat,?y:Nat)),plus(times(plus(?x:Nat,0),plus(0,?y:Nat)),plus(plus(0,?x:Nat),plus(0,?z:Nat)))) = plus(times(s(0),times(?z:Nat,?x:Nat)),plus(s(?x:Nat),?z:Nat))",
"plus(plus(plus(times(?x:Nat,?y:Nat),s(?z:Nat)),times(s(?x:Nat),0)),times(plus(s(?z:Nat),s(?y:Nat)),?x:Nat)) = plus(s(0),plus(times(times(?y:Nat,?y:Nat),?x:Nat),times(plus(?x:Nat,?x:Nat),times(?x:Nat,?x:Nat))))",
"plus(times(times(s(0),plus(?x:Nat,?x:Nat)),plus(times(?x:Nat,?y:Nat),0)),times(?x:Nat,times(times(?x:Nat,?x:Nat),plus(?z:Nat,0)))) = plus(plus(times(0,?y:Nat),times(?x:Nat,times(?z:Nat,?x:Nat))),plus(?x:Nat,?y:Nat))",
"plus(plus(times(s(?z:Nat),plus(?y:Nat,?x:Nat)),s(?z:Nat)),times(times(plus(?x:Nat,?x:Nat),s(0)),times(s(?z:Nat),plus(?z:Nat,?x:Nat)))) = plus(times(plus(times(?x:Nat,?x:Nat),s(0)),plus(s(0),?x:Nat)),s(?z:Nat))",
"plus(times(plus(plus(?y:Nat,?x:Nat),?y:Nat),plus(?x:Nat,s(?x:Nat))),s(plus(plus(?y:Nat,?x:Nat),s(?x:Nat)))) = plus(times(?y:Nat,times(plus(0,0),?x:Nat)),times(times(times(?x:Nat,?x:Nat),times(?z:Nat,?y:Nat)),plus(?x:Nat,s(?x:Nat))))"
]

val problemsTimesWithoutZeroYes =
IOMstrs.rdEqs [
"times(?x:Nat,?y:Nat) = times(?y:Nat,?x:Nat)",
"times(s(?y:Nat),?x:Nat) = times(?x:Nat,s(?y:Nat))",
"times(s(?y:Nat),?y:Nat) = times(?y:Nat,s(?y:Nat))",
"times(?x:Nat,s(?x:Nat)) = plus(?x:Nat,times(?x:Nat,?x:Nat))",
"times(?x:Nat,s(?x:Nat)) = plus(times(?x:Nat,?x:Nat),?x:Nat)",
"times(?x:Nat,s(?y:Nat)) = plus(times(?x:Nat,?y:Nat),?x:Nat)",
"times(?z:Nat,s(?y:Nat)) = plus(times(?y:Nat,?z:Nat),?z:Nat)",
"times(s(?x:Nat),?x:Nat) = plus(?x:Nat,times(?x:Nat,?x:Nat))",
"times(s(?x:Nat),?x:Nat) = plus(times(?x:Nat,?x:Nat),?x:Nat)",
"times(s(?y:Nat),?x:Nat) = plus(?x:Nat,times(?x:Nat,?y:Nat))",
"plus(?x:Nat,times(?x:Nat,?y:Nat)) = times(?x:Nat,s(?y:Nat))",
"plus(?x:Nat,times(?y:Nat,?x:Nat)) = times(?x:Nat,s(?y:Nat))",
"plus(times(?x:Nat,?y:Nat),?x:Nat) = times(s(?y:Nat),?x:Nat)",
"plus(?x:Nat,times(?x:Nat,?x:Nat)) = plus(times(?x:Nat,?x:Nat),?x:Nat)",
"plus(?x:Nat,times(?x:Nat,?y:Nat)) = plus(times(?y:Nat,?x:Nat),?x:Nat)",
"plus(?x:Nat,times(?y:Nat,?x:Nat)) = plus(times(?x:Nat,?y:Nat),?x:Nat)",
"plus(times(?y:Nat,?y:Nat),?x:Nat) = plus(?x:Nat,times(?y:Nat,?y:Nat))",
"plus(times(?z:Nat,?x:Nat),?x:Nat) = plus(?x:Nat,times(?z:Nat,?x:Nat))",
"times(?x:Nat,s(s(?x:Nat))) = times(s(s(?x:Nat)),?x:Nat)",
"times(?x:Nat,plus(?x:Nat,?y:Nat)) = times(?x:Nat,plus(?y:Nat,?x:Nat))",
"times(?x:Nat,plus(?z:Nat,?z:Nat)) = times(plus(?x:Nat,?x:Nat),?z:Nat)",
"times(?x:Nat,times(?x:Nat,?y:Nat)) = times(times(?x:Nat,?x:Nat),?y:Nat)",
"times(?x:Nat,times(?z:Nat,?z:Nat)) = times(times(?z:Nat,?x:Nat),?z:Nat)",
"times(?y:Nat,times(?x:Nat,?z:Nat)) = times(?y:Nat,times(?z:Nat,?x:Nat))",
"times(plus(?x:Nat,?y:Nat),?y:Nat) = times(?y:Nat,plus(?x:Nat,?y:Nat))",
"times(plus(?y:Nat,?x:Nat),?x:Nat) = times(?x:Nat,plus(?x:Nat,?y:Nat))",
"times(plus(?y:Nat,?y:Nat),?x:Nat) = times(?x:Nat,plus(?y:Nat,?y:Nat))",
"times(times(?x:Nat,?y:Nat),?x:Nat) = times(times(?x:Nat,?x:Nat),?y:Nat)",
"times(times(?x:Nat,?y:Nat),?y:Nat) = times(?x:Nat,times(?y:Nat,?y:Nat))",
"times(times(?y:Nat,?x:Nat),?x:Nat) = times(?x:Nat,times(?x:Nat,?y:Nat))",
"times(times(?y:Nat,?x:Nat),?y:Nat) = times(?y:Nat,times(?y:Nat,?x:Nat))",
"times(times(?z:Nat,?y:Nat),?y:Nat) = times(times(?y:Nat,?y:Nat),?z:Nat)",
"plus(times(?x:Nat,s(?x:Nat)),?x:Nat) = times(s(s(?x:Nat)),?x:Nat)",
"times(s(?x:Nat),s(?y:Nat)) = plus(times(s(?x:Nat),?y:Nat),s(?x:Nat))",
"plus(s(?x:Nat),times(?y:Nat,?x:Nat)) = plus(?x:Nat,s(times(?x:Nat,?y:Nat)))",
"times(times(?x:Nat,s(?x:Nat)),?x:Nat) = times(?x:Nat,times(s(?x:Nat),?x:Nat))",
"times(s(?z:Nat),s(?x:Nat)) = plus(s(times(?z:Nat,?x:Nat)),plus(?x:Nat,?z:Nat))",
"plus(?x:Nat,times(s(?y:Nat),?x:Nat)) = plus(?x:Nat,plus(times(?x:Nat,?y:Nat),?x:Nat))",
"plus(times(s(?z:Nat),?y:Nat),?x:Nat) = plus(?x:Nat,plus(?y:Nat,times(?z:Nat,?y:Nat)))",
"plus(times(?y:Nat,times(?x:Nat,?z:Nat)),?z:Nat) = times(s(times(?y:Nat,?x:Nat)),?z:Nat)",
"plus(times(plus(?x:Nat,?x:Nat),?x:Nat),?x:Nat) = times(plus(?x:Nat,s(?x:Nat)),?x:Nat)",
"times(s(plus(?x:Nat,?y:Nat)),?y:Nat) = plus(times(?y:Nat,s(?x:Nat)),times(?y:Nat,?y:Nat))",
"plus(times(?y:Nat,?x:Nat),plus(?x:Nat,?y:Nat)) = plus(plus(?x:Nat,times(?y:Nat,?x:Nat)),?y:Nat)",
"times(plus(?x:Nat,?x:Nat),times(?x:Nat,?x:Nat)) = times(?x:Nat,times(?x:Nat,plus(?x:Nat,?x:Nat)))",
"times(plus(?x:Nat,plus(?x:Nat,?y:Nat)),?x:Nat) = times(plus(?y:Nat,plus(?x:Nat,?x:Nat)),?x:Nat)",
"times(plus(plus(?y:Nat,?x:Nat),?x:Nat),?x:Nat) = times(plus(?y:Nat,plus(?x:Nat,?x:Nat)),?x:Nat)",
"plus(times(?x:Nat,s(?x:Nat)),s(?x:Nat)) = plus(times(?x:Nat,?x:Nat),plus(?x:Nat,s(?x:Nat)))",
"plus(plus(?x:Nat,?x:Nat),times(?x:Nat,plus(?x:Nat,?x:Nat))) = times(plus(?x:Nat,?x:Nat),s(?x:Nat))",
"plus(s(plus(times(?x:Nat,?y:Nat),?x:Nat)),?y:Nat) = s(plus(plus(?x:Nat,?y:Nat),times(?x:Nat,?y:Nat)))",
"plus(plus(s(times(?z:Nat,?x:Nat)),?y:Nat),?z:Nat) = plus(plus(?z:Nat,s(times(?z:Nat,?x:Nat))),?y:Nat)",
"plus(times(s(plus(?y:Nat,?x:Nat)),?y:Nat),?y:Nat) = plus(times(?y:Nat,s(plus(?y:Nat,?x:Nat))),?y:Nat)",
"times(?x:Nat,plus(?z:Nat,s(plus(?x:Nat,?y:Nat)))) = times(plus(?x:Nat,s(plus(?y:Nat,?z:Nat))),?x:Nat)",
"plus(?y:Nat,plus(times(times(?x:Nat,?y:Nat),?x:Nat),?z:Nat)) = plus(?z:Nat,times(?y:Nat,s(times(?x:Nat,?x:Nat))))",
"plus(s(times(times(?x:Nat,?y:Nat),s(?y:Nat))),?x:Nat) = s(times(s(times(?y:Nat,s(?y:Nat))),?x:Nat))",
"times(s(plus(s(?x:Nat),s(?y:Nat))),?x:Nat) = plus(times(s(plus(?y:Nat,?x:Nat)),?x:Nat),plus(?x:Nat,?x:Nat))",
"s(s(plus(?x:Nat,plus(?x:Nat,times(?x:Nat,?x:Nat))))) = plus(s(plus(?x:Nat,times(?x:Nat,?x:Nat))),s(?x:Nat))",
"plus(?y:Nat,times(?x:Nat,plus(?x:Nat,s(?y:Nat)))) = plus(?x:Nat,plus(?y:Nat,plus(times(?x:Nat,?x:Nat),times(?x:Nat,?y:Nat))))",
"times(s(s(plus(times(?x:Nat,?y:Nat),?y:Nat))),?x:Nat) = plus(plus(times(plus(?y:Nat,times(?y:Nat,?x:Nat)),?x:Nat),?x:Nat),?x:Nat)",
"s(s(plus(plus(s(?x:Nat),times(?x:Nat,?x:Nat)),?x:Nat))) = plus(s(?x:Nat),s(s(plus(?x:Nat,times(?x:Nat,?x:Nat)))))",
"s(plus(times(plus(s(?x:Nat),plus(?x:Nat,?x:Nat)),?x:Nat),plus(plus(?x:Nat,s(?x:Nat)),?x:Nat))) = plus(times(s(plus(plus(?x:Nat,?x:Nat),?x:Nat)),?x:Nat),s(plus(?x:Nat,plus(?x:Nat,s(?x:Nat)))))"]

val problemsTimesWithoutZeroNo =
IOMstrs.rdEqs [
"s(?y:Nat) = times(?x:Nat,?y:Nat)",
"s(?y:Nat) = times(?z:Nat,?x:Nat)",
"s(?y:Nat) = times(?z:Nat,?z:Nat)",
"s(?z:Nat) = times(?z:Nat,?z:Nat)",
"times(?x:Nat,?y:Nat) = s(?x:Nat)",
"s(?x:Nat) = times(?y:Nat,s(?z:Nat))",
"s(?x:Nat) = times(s(?y:Nat),?x:Nat)",
"s(?y:Nat) = times(?x:Nat,s(?x:Nat))",
"s(s(?z:Nat)) = times(?x:Nat,?z:Nat)",
"s(s(?z:Nat)) = times(?z:Nat,?x:Nat)",
"plus(?z:Nat,?x:Nat) = times(?x:Nat,?z:Nat)",
"plus(?z:Nat,?x:Nat) = times(?z:Nat,?y:Nat)",
"plus(?z:Nat,?z:Nat) = times(?y:Nat,?x:Nat)",
"times(?x:Nat,?x:Nat) = times(?x:Nat,?y:Nat)",
"times(?x:Nat,?y:Nat) = s(s(?z:Nat))",
"times(?y:Nat,?z:Nat) = plus(?y:Nat,?y:Nat)",
"times(?z:Nat,?y:Nat) = times(?x:Nat,?z:Nat)",
"times(?z:Nat,?z:Nat) = plus(?x:Nat,?x:Nat)",
"times(?z:Nat,?z:Nat) = plus(?z:Nat,?x:Nat)",
"times(s(?x:Nat),?y:Nat) = s(?z:Nat)",
"s(?x:Nat) = times(plus(?x:Nat,?y:Nat),?y:Nat)",
"s(?y:Nat) = plus(times(?x:Nat,?z:Nat),?y:Nat)",
"s(?y:Nat) = times(?x:Nat,plus(?y:Nat,?x:Nat))",
"s(?y:Nat) = times(plus(?z:Nat,?z:Nat),?z:Nat)",
"s(?z:Nat) = times(?x:Nat,s(s(?z:Nat)))",
"s(?z:Nat) = times(?x:Nat,times(?y:Nat,?x:Nat))",
"s(?z:Nat) = times(s(?z:Nat),s(?y:Nat))",
"s(s(?x:Nat)) = times(?y:Nat,s(?y:Nat))",
"plus(?z:Nat,?z:Nat) = times(s(?x:Nat),?y:Nat)",
"times(?y:Nat,?z:Nat) = s(times(?x:Nat,?y:Nat))",
"times(?y:Nat,?z:Nat) = times(s(?x:Nat),?y:Nat)",
"times(?z:Nat,?x:Nat) = times(s(?z:Nat),?x:Nat)",
"s(times(?z:Nat,?x:Nat)) = plus(?y:Nat,?x:Nat)",
"plus(?z:Nat,s(?x:Nat)) = times(?y:Nat,?z:Nat)",
"plus(s(?x:Nat),?x:Nat) = times(?z:Nat,?x:Nat)",
"times(?x:Nat,s(?x:Nat)) = plus(?y:Nat,?y:Nat)",
"times(?z:Nat,s(?x:Nat)) = plus(?x:Nat,?z:Nat)",
"times(?z:Nat,s(?x:Nat)) = plus(?y:Nat,?x:Nat)",
"plus(?y:Nat,times(?x:Nat,?z:Nat)) = s(?y:Nat)",
"plus(?z:Nat,times(?x:Nat,?z:Nat)) = s(?y:Nat)",
"times(?y:Nat,times(?x:Nat,?y:Nat)) = s(?y:Nat)",
"times(s(?z:Nat),s(?x:Nat)) = s(?y:Nat)",
"times(s(s(?x:Nat)),?y:Nat) = s(?z:Nat)",
"times(plus(?z:Nat,?x:Nat),?x:Nat) = s(?x:Nat)",
"s(?y:Nat) = plus(?y:Nat,s(times(?x:Nat,?z:Nat)))",
"s(?y:Nat) = times(?z:Nat,s(s(s(?x:Nat))))",
"s(?y:Nat) = times(times(?y:Nat,?y:Nat),s(?y:Nat))",
"s(?z:Nat) = plus(s(?z:Nat),times(?x:Nat,?z:Nat))",
"s(?z:Nat) = times(?x:Nat,s(s(s(?z:Nat))))",
"s(?z:Nat) = times(s(?x:Nat),s(s(?x:Nat)))",
"s(s(?x:Nat)) = times(?x:Nat,times(?y:Nat,?y:Nat))",
"s(s(?y:Nat)) = times(?x:Nat,times(?y:Nat,?x:Nat))",
"plus(?y:Nat,?x:Nat) = times(?y:Nat,times(?z:Nat,?x:Nat))",
"plus(?z:Nat,?z:Nat) = times(?y:Nat,plus(?y:Nat,?y:Nat))",
"times(?x:Nat,?x:Nat) = s(s(s(s(?y:Nat))))",
"times(?y:Nat,?x:Nat) = s(times(?x:Nat,s(?z:Nat)))",
"times(?y:Nat,?x:Nat) = plus(s(s(?x:Nat)),?x:Nat)",
"times(?y:Nat,?y:Nat) = times(times(?z:Nat,?z:Nat),?x:Nat)",
"s(s(s(?x:Nat))) = times(s(?y:Nat),?z:Nat)",
"s(s(s(?z:Nat))) = times(s(?z:Nat),?y:Nat)",
"plus(s(?z:Nat),?x:Nat) = s(times(?y:Nat,?x:Nat))",
"plus(s(?z:Nat),?z:Nat) = s(times(?y:Nat,?z:Nat))",
"times(?y:Nat,s(?z:Nat)) = plus(?z:Nat,s(?y:Nat))",
"times(?z:Nat,s(?y:Nat)) = times(?x:Nat,s(?y:Nat))",
"times(s(?y:Nat),?z:Nat) = times(?y:Nat,s(?z:Nat))",
"times(s(?z:Nat),?x:Nat) = times(s(?y:Nat),?x:Nat)",
"times(s(?z:Nat),?z:Nat) = times(s(?x:Nat),?x:Nat)",
"plus(?y:Nat,s(s(?x:Nat))) = times(?z:Nat,?z:Nat)",
"plus(?y:Nat,times(?x:Nat,?z:Nat)) = times(?z:Nat,?y:Nat)",
"plus(times(?x:Nat,?x:Nat),?x:Nat) = plus(?y:Nat,?x:Nat)",
"times(?x:Nat,times(?x:Nat,?y:Nat)) = plus(?z:Nat,?x:Nat)",
"times(s(s(?y:Nat)),?x:Nat) = plus(?y:Nat,?y:Nat)",
"times(plus(?z:Nat,?y:Nat),?x:Nat) = s(s(?z:Nat))",
"plus(s(?z:Nat),times(?z:Nat,?x:Nat)) = s(?x:Nat)",
"times(?x:Nat,plus(?y:Nat,s(?y:Nat))) = s(?z:Nat)",
"times(?x:Nat,times(?y:Nat,s(?z:Nat))) = s(?x:Nat)",
"times(s(?y:Nat),times(?x:Nat,?x:Nat)) = s(?y:Nat)",
"times(s(s(s(?z:Nat))),?y:Nat) = s(?x:Nat)",
"times(times(?z:Nat,s(?z:Nat)),?x:Nat) = s(?z:Nat)",
"s(?x:Nat) = times(?x:Nat,s(times(?z:Nat,s(?y:Nat))))",
"s(?x:Nat) = times(times(s(s(?z:Nat)),?x:Nat),?y:Nat)",
"s(?y:Nat) = plus(s(s(times(?x:Nat,?z:Nat))),?z:Nat)",
"s(?y:Nat) = plus(s(s(times(?y:Nat,?y:Nat))),?z:Nat)",
"s(?y:Nat) = plus(plus(times(?z:Nat,?z:Nat),?y:Nat),?z:Nat)",
"s(?y:Nat) = plus(times(?x:Nat,s(?y:Nat)),s(?z:Nat))",
"s(?y:Nat) = plus(times(plus(?y:Nat,?x:Nat),?x:Nat),?x:Nat)",
"s(?y:Nat) = times(?x:Nat,s(s(plus(?x:Nat,?z:Nat))))",
"s(?z:Nat) = plus(?z:Nat,times(?y:Nat,s(s(?z:Nat))))",
"s(?z:Nat) = times(?x:Nat,times(?y:Nat,plus(?x:Nat,?z:Nat)))",
"s(?z:Nat) = times(times(s(s(?x:Nat)),?x:Nat),?x:Nat)",
"s(s(?y:Nat)) = times(plus(?x:Nat,s(?x:Nat)),?y:Nat)",
"s(s(?z:Nat)) = times(?x:Nat,plus(?y:Nat,s(?y:Nat)))",
"plus(?x:Nat,?z:Nat) = plus(times(?x:Nat,?y:Nat),s(?x:Nat))",
"plus(?y:Nat,?y:Nat) = s(s(times(s(?z:Nat),?z:Nat)))",
"plus(?y:Nat,?y:Nat) = plus(?x:Nat,times(?z:Nat,s(?z:Nat)))",
"times(?x:Nat,?x:Nat) = s(s(plus(?z:Nat,s(?x:Nat))))",
"times(?x:Nat,?x:Nat) = plus(times(?y:Nat,?x:Nat),s(?y:Nat))",
"times(?x:Nat,?y:Nat) = s(s(plus(?z:Nat,s(?x:Nat))))",
"times(?y:Nat,?x:Nat) = times(plus(?z:Nat,s(?z:Nat)),?x:Nat)",
"s(s(s(?z:Nat))) = plus(times(?x:Nat,?y:Nat),?z:Nat)",
"s(plus(?y:Nat,?z:Nat)) = plus(?x:Nat,times(?z:Nat,?x:Nat))",
"s(times(?z:Nat,?z:Nat)) = times(?y:Nat,s(s(?x:Nat)))",
"plus(?z:Nat,s(?x:Nat)) = times(times(?x:Nat,?z:Nat),?y:Nat)",
"times(s(?z:Nat),s(?x:Nat)) = times(s(?z:Nat),?z:Nat)",
"s(s(s(plus(?z:Nat,?x:Nat)))) = times(?y:Nat,?x:Nat)",
"s(s(s(times(?z:Nat,?x:Nat)))) = plus(?z:Nat,?x:Nat)",
"s(s(s(times(?z:Nat,?z:Nat)))) = plus(?y:Nat,?y:Nat)",
"s(plus(?x:Nat,plus(?z:Nat,?x:Nat))) = times(?x:Nat,?y:Nat)",
"plus(?y:Nat,times(?x:Nat,s(?x:Nat))) = times(?y:Nat,?y:Nat)",
"plus(s(s(s(?x:Nat))),?y:Nat) = times(?x:Nat,?y:Nat)",
"plus(plus(?y:Nat,s(?x:Nat)),?z:Nat) = times(?y:Nat,?x:Nat)",
"plus(times(?z:Nat,?x:Nat),s(?y:Nat)) = times(?z:Nat,?z:Nat)",
"times(?y:Nat,s(times(?z:Nat,?y:Nat))) = plus(?x:Nat,?y:Nat)",
"times(s(?x:Nat),plus(?y:Nat,?z:Nat)) = plus(?x:Nat,?y:Nat)",
"times(times(?y:Nat,?y:Nat),s(?y:Nat)) = times(?y:Nat,?z:Nat)",
"plus(?y:Nat,s(times(s(?y:Nat),?z:Nat))) = s(?z:Nat)",
"plus(?z:Nat,times(?y:Nat,times(?x:Nat,?x:Nat))) = s(?y:Nat)",
"times(?x:Nat,plus(times(?y:Nat,?z:Nat),?y:Nat)) = s(?y:Nat)",
"times(?z:Nat,s(s(plus(?x:Nat,?y:Nat)))) = s(?y:Nat)",
"times(?z:Nat,times(times(?x:Nat,?x:Nat),?z:Nat)) = s(?y:Nat)",
"times(s(?x:Nat),s(s(s(?x:Nat)))) = s(?x:Nat)",
"times(s(s(plus(?y:Nat,?z:Nat))),?z:Nat) = s(?x:Nat)",
"times(plus(?z:Nat,plus(?z:Nat,?z:Nat)),?y:Nat) = s(?y:Nat)",
"times(times(s(s(?z:Nat)),?y:Nat),?z:Nat) = s(?y:Nat)",
"s(?x:Nat) = plus(s(?x:Nat),times(?z:Nat,s(s(?z:Nat))))",
"s(?x:Nat) = plus(times(?x:Nat,s(plus(?z:Nat,?x:Nat))),?z:Nat)",
"s(?y:Nat) = plus(?x:Nat,s(times(s(?z:Nat),s(?x:Nat))))",
"s(?y:Nat) = plus(?y:Nat,plus(s(times(?y:Nat,?y:Nat)),?x:Nat))",
"s(?z:Nat) = times(?x:Nat,s(times(?y:Nat,times(?z:Nat,?z:Nat))))",
"s(?z:Nat) = times(?y:Nat,plus(?z:Nat,times(?x:Nat,s(?y:Nat))))",
"s(s(?y:Nat)) = times(s(s(s(?x:Nat))),s(?z:Nat))",
"s(s(?y:Nat)) = times(s(s(times(?y:Nat,?x:Nat))),?y:Nat)",
"s(s(?z:Nat)) = times(?z:Nat,plus(?x:Nat,s(s(?z:Nat))))",
"plus(?x:Nat,?x:Nat) = times(?y:Nat,plus(s(?z:Nat),s(?y:Nat)))",
"plus(?y:Nat,?y:Nat) = times(s(?x:Nat),s(s(s(?y:Nat))))",
"plus(?y:Nat,?z:Nat) = plus(?x:Nat,s(s(times(?x:Nat,?y:Nat))))",
"plus(?z:Nat,?x:Nat) = s(times(?y:Nat,s(plus(?y:Nat,?y:Nat))))",
"times(?x:Nat,?y:Nat) = times(?x:Nat,s(s(times(?z:Nat,?x:Nat))))",
"times(?x:Nat,?y:Nat) = times(s(s(plus(?y:Nat,?x:Nat))),?z:Nat)",
"times(?x:Nat,?z:Nat) = plus(?z:Nat,s(s(times(?y:Nat,?y:Nat))))",
"times(?y:Nat,?x:Nat) = times(s(plus(?x:Nat,s(?z:Nat))),?x:Nat)",
"times(?z:Nat,?x:Nat) = s(plus(s(times(?z:Nat,?z:Nat)),?z:Nat))",
"times(?z:Nat,?y:Nat) = plus(times(?z:Nat,s(s(?x:Nat))),?x:Nat)",
"times(?z:Nat,?y:Nat) = times(?x:Nat,plus(?y:Nat,s(s(?x:Nat))))",
"plus(?x:Nat,s(?x:Nat)) = s(times(s(s(?x:Nat)),?y:Nat))",
"times(?x:Nat,s(?x:Nat)) = s(s(plus(s(?x:Nat),?y:Nat)))",
"times(?z:Nat,s(?z:Nat)) = plus(?x:Nat,s(plus(?y:Nat,?y:Nat)))",
"times(s(?y:Nat),?x:Nat) = times(times(s(?y:Nat),?y:Nat),?y:Nat)",
"s(times(s(?z:Nat),?x:Nat)) = plus(s(?x:Nat),s(?x:Nat))",
"s(times(s(s(?x:Nat)),?y:Nat)) = times(s(?x:Nat),?y:Nat)",
"plus(s(times(?z:Nat,?z:Nat)),?x:Nat) = plus(?y:Nat,s(?y:Nat))",
"times(s(?z:Nat),times(?y:Nat,?x:Nat)) = plus(?y:Nat,s(?x:Nat))",
"s(s(plus(times(?z:Nat,?x:Nat),?x:Nat))) = plus(?z:Nat,?z:Nat)",
"s(s(times(?x:Nat,times(?x:Nat,?y:Nat)))) = times(?z:Nat,?y:Nat)",
"s(plus(?y:Nat,s(times(?z:Nat,?x:Nat)))) = plus(?z:Nat,?x:Nat)",
"s(times(plus(?x:Nat,?x:Nat),s(?y:Nat))) = times(?y:Nat,?y:Nat)",
"plus(?x:Nat,plus(?x:Nat,times(?y:Nat,?z:Nat))) = s(s(?y:Nat))",
"plus(s(plus(?x:Nat,s(?x:Nat))),?z:Nat) = times(?z:Nat,?z:Nat)",
"plus(s(times(?y:Nat,?y:Nat)),s(?y:Nat)) = times(?z:Nat,?z:Nat)",
"plus(times(?x:Nat,?z:Nat),s(s(?z:Nat))) = plus(?z:Nat,?y:Nat)",
"times(?x:Nat,s(times(s(?x:Nat),?z:Nat))) = s(s(?z:Nat))",
"plus(?z:Nat,plus(?x:Nat,s(times(?z:Nat,?y:Nat)))) = s(?x:Nat)",
"plus(plus(?x:Nat,times(s(?x:Nat),?x:Nat)),?y:Nat) = s(?z:Nat)",
"plus(times(s(?z:Nat),?y:Nat),times(?x:Nat,?y:Nat)) = s(?z:Nat)",
"plus(times(s(times(?x:Nat,?y:Nat)),?x:Nat),?y:Nat) = s(?y:Nat)",
"times(?y:Nat,s(plus(?z:Nat,plus(?z:Nat,?y:Nat)))) = s(?z:Nat)",
"times(?y:Nat,s(plus(plus(?y:Nat,?y:Nat),?z:Nat))) = s(?z:Nat)",
"times(?y:Nat,times(s(times(?x:Nat,?y:Nat)),?z:Nat)) = s(?z:Nat)",
"times(s(times(plus(?x:Nat,?y:Nat),?x:Nat)),?z:Nat) = s(?z:Nat)",
"times(plus(?y:Nat,s(times(?z:Nat,?z:Nat))),?x:Nat) = s(?z:Nat)",
"s(?x:Nat) = plus(s(?y:Nat),s(times(times(?x:Nat,?x:Nat),?y:Nat)))",
"s(?x:Nat) = times(?x:Nat,plus(?x:Nat,plus(plus(?y:Nat,?z:Nat),?x:Nat)))",
"s(?x:Nat) = times(s(?x:Nat),s(plus(?y:Nat,plus(?x:Nat,?z:Nat))))",
"s(?y:Nat) = plus(?x:Nat,times(times(?z:Nat,plus(?y:Nat,?z:Nat)),?y:Nat))",
"s(?y:Nat) = plus(?z:Nat,plus(times(?x:Nat,s(?x:Nat)),s(?x:Nat)))",
"s(?y:Nat) = plus(s(s(plus(?y:Nat,?y:Nat))),times(?z:Nat,?x:Nat))",
"s(?y:Nat) = plus(plus(?x:Nat,plus(?y:Nat,times(?z:Nat,?y:Nat))),?z:Nat)",
"s(?z:Nat) = plus(?y:Nat,s(plus(s(?z:Nat),times(?y:Nat,?x:Nat))))",
"s(?z:Nat) = plus(times(?x:Nat,s(?x:Nat)),plus(s(?y:Nat),?x:Nat))",
"s(?z:Nat) = times(times(s(times(?y:Nat,?y:Nat)),?z:Nat),s(?x:Nat))",
"s(s(?x:Nat)) = plus(?z:Nat,times(s(plus(?z:Nat,?z:Nat)),?x:Nat))",
"s(s(?x:Nat)) = times(s(plus(s(?x:Nat),s(?y:Nat))),?x:Nat)",
"s(s(?y:Nat)) = times(?y:Nat,s(plus(plus(?z:Nat,?y:Nat),?y:Nat)))",
"plus(?x:Nat,?x:Nat) = s(times(?y:Nat,plus(?x:Nat,times(?z:Nat,?x:Nat))))",
"plus(?x:Nat,?y:Nat) = times(times(s(times(?x:Nat,?y:Nat)),?z:Nat),?y:Nat)",
"plus(?z:Nat,?x:Nat) = s(times(?x:Nat,plus(times(?x:Nat,?x:Nat),?x:Nat)))",
"plus(?z:Nat,?x:Nat) = times(times(plus(?x:Nat,s(?x:Nat)),?x:Nat),?y:Nat)",
"plus(?z:Nat,?y:Nat) = plus(?y:Nat,s(plus(?z:Nat,times(?y:Nat,?x:Nat))))",
"plus(?z:Nat,?y:Nat) = times(plus(s(?y:Nat),plus(?x:Nat,?x:Nat)),?y:Nat)",
"plus(?z:Nat,?z:Nat) = times(?z:Nat,times(times(s(?x:Nat),?x:Nat),?z:Nat))",
"times(?x:Nat,?y:Nat) = times(s(times(?x:Nat,?y:Nat)),plus(?x:Nat,?y:Nat))",
"times(?y:Nat,?x:Nat) = plus(?y:Nat,times(s(plus(?z:Nat,?x:Nat)),?z:Nat))",
"times(?y:Nat,?z:Nat) = s(times(?y:Nat,plus(s(?z:Nat),s(?z:Nat))))",
"times(?z:Nat,?x:Nat) = plus(?y:Nat,plus(s(plus(?y:Nat,?z:Nat)),?x:Nat))",
"times(?z:Nat,?y:Nat) = times(?z:Nat,s(plus(?z:Nat,plus(?y:Nat,?y:Nat))))",
"times(?z:Nat,?z:Nat) = s(times(times(?x:Nat,plus(?x:Nat,?y:Nat)),?z:Nat))",
"times(?z:Nat,?z:Nat) = plus(?y:Nat,plus(s(?y:Nat),plus(?y:Nat,?y:Nat)))",
"times(?z:Nat,?z:Nat) = times(s(times(s(?z:Nat),s(?x:Nat))),?z:Nat)",
"s(s(s(?y:Nat))) = times(?z:Nat,s(times(?z:Nat,s(?x:Nat))))",
"plus(?y:Nat,s(?z:Nat)) = s(s(times(times(?z:Nat,?x:Nat),?y:Nat)))",
"plus(s(?y:Nat),?y:Nat) = plus(times(?y:Nat,?y:Nat),s(s(?z:Nat)))",
"plus(s(?y:Nat),?y:Nat) = times(s(s(s(?x:Nat))),s(?x:Nat))",
"times(s(?y:Nat),?z:Nat) = s(s(times(plus(?x:Nat,?x:Nat),?x:Nat)))",
"s(s(times(?y:Nat,?y:Nat))) = times(s(s(s(?z:Nat))),?z:Nat)",
"times(?x:Nat,plus(?y:Nat,?z:Nat)) = s(plus(?x:Nat,s(s(?z:Nat))))",
"times(?y:Nat,s(s(?x:Nat))) = s(plus(s(s(?x:Nat)),?z:Nat))",
"times(s(?z:Nat),s(?x:Nat)) = times(s(?z:Nat),plus(?x:Nat,?x:Nat))",
"s(times(?z:Nat,s(s(?y:Nat)))) = plus(?x:Nat,plus(?z:Nat,?x:Nat))",
"times(?y:Nat,s(s(s(?z:Nat)))) = plus(s(?z:Nat),s(?y:Nat))",
"s(s(times(plus(?x:Nat,?y:Nat),?x:Nat))) = times(?x:Nat,s(?x:Nat))",
"plus(?y:Nat,s(s(plus(?y:Nat,?z:Nat)))) = s(times(?y:Nat,?x:Nat))",
"plus(times(times(?y:Nat,?y:Nat),?y:Nat),?z:Nat) = s(times(?z:Nat,?y:Nat))",
"times(times(s(?z:Nat),?y:Nat),s(?y:Nat)) = plus(?y:Nat,s(?z:Nat))",
"s(plus(s(times(?z:Nat,?x:Nat)),s(?y:Nat))) = times(?x:Nat,?z:Nat)",
"s(plus(times(?y:Nat,plus(?y:Nat,?z:Nat)),?z:Nat)) = times(?x:Nat,?x:Nat)",
"s(times(s(?z:Nat),plus(s(?x:Nat),?y:Nat))) = times(?z:Nat,?z:Nat)",
"plus(?z:Nat,s(plus(plus(?x:Nat,?z:Nat),?y:Nat))) = times(?z:Nat,?z:Nat)",
"times(?y:Nat,plus(plus(?x:Nat,s(?z:Nat)),?x:Nat)) = plus(?z:Nat,?z:Nat)",
"times(?z:Nat,times(times(?y:Nat,s(?z:Nat)),?y:Nat)) = plus(?x:Nat,?z:Nat)",
"times(s(times(?x:Nat,s(?y:Nat))),s(?y:Nat)) = s(s(?x:Nat))",
"plus(?x:Nat,plus(plus(times(?x:Nat,?z:Nat),?x:Nat),?z:Nat)) = s(?y:Nat)",
"plus(?y:Nat,plus(times(?z:Nat,times(?x:Nat,?x:Nat)),?x:Nat)) = s(?z:Nat)",
"plus(s(plus(times(?x:Nat,?z:Nat),s(?x:Nat))),?x:Nat) = s(?x:Nat)",
"plus(s(times(s(?z:Nat),times(?z:Nat,?y:Nat))),?y:Nat) = s(?x:Nat)",
"plus(plus(?x:Nat,times(s(?z:Nat),s(?z:Nat))),?x:Nat) = s(?y:Nat)",
"plus(plus(?y:Nat,s(s(?z:Nat))),times(?x:Nat,?z:Nat)) = s(?y:Nat)",
"plus(plus(?z:Nat,plus(?y:Nat,times(?x:Nat,?y:Nat))),?z:Nat) = s(?z:Nat)",
"times(?y:Nat,plus(?z:Nat,times(times(?x:Nat,?x:Nat),?z:Nat))) = s(?z:Nat)",
"times(?y:Nat,plus(?z:Nat,times(times(?y:Nat,?z:Nat),?x:Nat))) = s(?y:Nat)",
"times(s(plus(s(?x:Nat),s(?z:Nat))),s(?x:Nat)) = s(?x:Nat)",
"times(s(plus(plus(?z:Nat,?z:Nat),?y:Nat)),s(?y:Nat)) = s(?z:Nat)",
"times(s(plus(plus(?z:Nat,?z:Nat),s(?y:Nat))),?z:Nat) = s(?y:Nat)",
"times(s(times(times(?y:Nat,?x:Nat),s(?x:Nat))),?y:Nat) = s(?y:Nat)",
"times(plus(?z:Nat,s(s(?x:Nat))),plus(?y:Nat,?y:Nat)) = s(?z:Nat)",
"times(times(?x:Nat,plus(?y:Nat,plus(?z:Nat,?x:Nat))),?x:Nat) = s(?z:Nat)",
"times(times(?y:Nat,plus(?x:Nat,times(?x:Nat,?x:Nat))),?z:Nat) = s(?y:Nat)",
"times(times(times(times(?y:Nat,?x:Nat),?x:Nat),?z:Nat),?x:Nat) = s(?y:Nat)",
"s(?x:Nat) = plus(?z:Nat,plus(times(times(?y:Nat,?x:Nat),s(?y:Nat)),?z:Nat))",
"s(?x:Nat) = plus(s(plus(times(?x:Nat,?y:Nat),times(?z:Nat,?x:Nat))),?y:Nat)",
"s(?x:Nat) = plus(times(?x:Nat,times(times(?x:Nat,?x:Nat),?z:Nat)),s(?z:Nat))",
"s(?x:Nat) = plus(times(?z:Nat,s(times(?z:Nat,?z:Nat))),s(s(?z:Nat)))",
"s(?x:Nat) = plus(times(s(plus(?y:Nat,?x:Nat)),?z:Nat),plus(?z:Nat,?z:Nat))",
"s(?x:Nat) = times(plus(s(plus(?y:Nat,?z:Nat)),s(s(?y:Nat))),?z:Nat)",
"s(?z:Nat) = plus(s(times(times(?y:Nat,?x:Nat),times(?y:Nat,?x:Nat))),?y:Nat)",
"s(?z:Nat) = times(plus(?y:Nat,plus(times(?x:Nat,?y:Nat),s(?z:Nat))),?y:Nat)",
"s(s(?x:Nat)) = plus(?x:Nat,plus(times(?x:Nat,times(?z:Nat,?x:Nat)),?y:Nat))",
"s(s(?x:Nat)) = times(?z:Nat,times(?x:Nat,plus(?z:Nat,plus(?y:Nat,?x:Nat))))",
"s(s(?y:Nat)) = plus(s(times(s(?z:Nat),times(?y:Nat,?y:Nat))),?z:Nat)",
"s(s(?z:Nat)) = times(s(s(plus(?z:Nat,?y:Nat))),times(?x:Nat,?y:Nat))",
"s(s(?z:Nat)) = times(plus(?y:Nat,times(times(?y:Nat,?y:Nat),?y:Nat)),?z:Nat)",
"s(s(?z:Nat)) = times(times(?y:Nat,times(?y:Nat,times(?x:Nat,?z:Nat))),?y:Nat)",
"plus(?y:Nat,?z:Nat) = s(times(s(?y:Nat),times(?x:Nat,plus(?z:Nat,?z:Nat))))",
"plus(?z:Nat,?x:Nat) = s(plus(s(?x:Nat),times(?z:Nat,plus(?x:Nat,?z:Nat))))",
"plus(?z:Nat,?z:Nat) = plus(?x:Nat,times(plus(?z:Nat,plus(?z:Nat,?x:Nat)),?z:Nat))",
"times(?y:Nat,?z:Nat) = times(?x:Nat,times(?x:Nat,plus(?y:Nat,plus(?y:Nat,?x:Nat))))",
"times(?z:Nat,?x:Nat) = s(times(s(?y:Nat),times(?y:Nat,plus(?x:Nat,?z:Nat))))",
"times(?z:Nat,?y:Nat) = plus(?z:Nat,s(plus(s(?y:Nat),plus(?y:Nat,?z:Nat))))",
"times(?z:Nat,?z:Nat) = s(plus(plus(?y:Nat,s(?z:Nat)),s(s(?y:Nat))))",
"s(plus(?x:Nat,?z:Nat)) = plus(?z:Nat,times(?x:Nat,plus(?y:Nat,s(?y:Nat))))",
"s(plus(?x:Nat,?z:Nat)) = times(s(times(?z:Nat,s(?z:Nat))),s(?x:Nat))",
"s(plus(?y:Nat,?x:Nat)) = plus(s(?z:Nat),times(times(?z:Nat,?x:Nat),?z:Nat))",
"plus(?y:Nat,s(?x:Nat)) = times(s(times(times(?x:Nat,?y:Nat),?x:Nat)),?y:Nat)",
"plus(s(?y:Nat),?z:Nat) = s(s(times(s(?x:Nat),plus(?y:Nat,?y:Nat))))",
"plus(s(?y:Nat),?z:Nat) = s(s(times(times(?y:Nat,?z:Nat),s(?y:Nat))))",
"times(?x:Nat,s(?y:Nat)) = times(times(s(plus(?x:Nat,?z:Nat)),?x:Nat),?y:Nat)",
"times(s(?x:Nat),?x:Nat) = s(plus(s(?y:Nat),plus(s(?y:Nat),?z:Nat)))",
"times(s(?y:Nat),?y:Nat) = times(?x:Nat,times(?y:Nat,times(s(?z:Nat),?z:Nat)))",
"times(s(?y:Nat),?y:Nat) = times(s(plus(s(?y:Nat),s(?y:Nat))),?y:Nat)",
"plus(?x:Nat,plus(?y:Nat,?y:Nat)) = plus(?x:Nat,times(plus(?y:Nat,?y:Nat),?y:Nat))",
"plus(s(?z:Nat),s(?y:Nat)) = times(times(?y:Nat,?x:Nat),plus(?x:Nat,?z:Nat))",
"times(s(?z:Nat),s(?z:Nat)) = s(plus(s(times(?y:Nat,?y:Nat)),?y:Nat))",
"times(s(s(?y:Nat)),?z:Nat) = s(plus(plus(s(?x:Nat),?z:Nat),?z:Nat))",
"s(times(s(?z:Nat),s(?x:Nat))) = times(?x:Nat,s(times(?x:Nat,?x:Nat)))",
"s(times(?z:Nat,s(times(?z:Nat,?z:Nat)))) = plus(?x:Nat,s(s(?z:Nat)))",
"times(times(times(?y:Nat,?z:Nat),?x:Nat),?y:Nat) = times(s(?y:Nat),s(?z:Nat))",
"plus(?y:Nat,plus(times(?z:Nat,?x:Nat),s(?y:Nat))) = s(times(?z:Nat,?z:Nat))",
"plus(s(times(?z:Nat,times(?z:Nat,?y:Nat))),?z:Nat) = plus(s(?x:Nat),?y:Nat)",
"plus(plus(s(times(?z:Nat,?x:Nat)),?z:Nat),?y:Nat) = times(?z:Nat,s(?x:Nat))",
"plus(times(s(plus(?x:Nat,?x:Nat)),?y:Nat),?x:Nat) = s(times(?z:Nat,?x:Nat))",
"times(?y:Nat,plus(?z:Nat,s(plus(?z:Nat,?z:Nat)))) = times(s(?z:Nat),?y:Nat)",
"times(plus(plus(s(?z:Nat),?z:Nat),?x:Nat),?x:Nat) = times(?z:Nat,s(?z:Nat))",
"times(times(?y:Nat,?x:Nat),s(s(s(?z:Nat)))) = times(?x:Nat,s(?z:Nat))",
"s(plus(s(s(?x:Nat)),plus(?x:Nat,s(?y:Nat)))) = times(?y:Nat,?z:Nat)",
"s(plus(times(times(?z:Nat,?x:Nat),s(?z:Nat)),?z:Nat)) = plus(?y:Nat,?y:Nat)",
"s(times(?y:Nat,times(plus(?z:Nat,?y:Nat),s(?x:Nat)))) = times(?x:Nat,?y:Nat)",
"plus(?x:Nat,plus(plus(?x:Nat,?y:Nat),times(?x:Nat,?z:Nat))) = times(?x:Nat,?z:Nat)",
"plus(?x:Nat,times(s(s(?z:Nat)),s(s(?x:Nat)))) = plus(?y:Nat,?y:Nat)",
"plus(s(?y:Nat),times(plus(?y:Nat,s(?y:Nat)),?x:Nat)) = times(?z:Nat,?z:Nat)",
"plus(s(times(plus(?y:Nat,?y:Nat),s(?z:Nat))),?z:Nat) = times(?x:Nat,?y:Nat)",
"plus(plus(?y:Nat,?z:Nat),s(s(times(?z:Nat,?x:Nat)))) = s(s(?x:Nat))",
"plus(plus(s(?x:Nat),s(times(?z:Nat,?x:Nat))),?y:Nat) = plus(?y:Nat,?z:Nat)",
"plus(times(?x:Nat,s(times(?z:Nat,?x:Nat))),s(?y:Nat)) = times(?x:Nat,?z:Nat)",
"plus(times(s(s(?y:Nat)),s(s(?x:Nat))),?x:Nat) = s(s(?x:Nat))",
"plus(times(times(times(?z:Nat,?z:Nat),?x:Nat),?x:Nat),?z:Nat) = times(?x:Nat,?y:Nat)",
"times(?x:Nat,s(times(s(?x:Nat),times(?z:Nat,?y:Nat)))) = s(s(?z:Nat))",
"times(?x:Nat,plus(s(plus(?x:Nat,?z:Nat)),s(?x:Nat))) = times(?y:Nat,?y:Nat)",
"times(?y:Nat,s(times(s(?z:Nat),plus(?y:Nat,?z:Nat)))) = times(?x:Nat,?z:Nat)",
"times(s(plus(s(?x:Nat),plus(?x:Nat,?y:Nat))),?x:Nat) = plus(?y:Nat,?x:Nat)",
"times(s(plus(s(?z:Nat),times(?z:Nat,?y:Nat))),?x:Nat) = plus(?x:Nat,?x:Nat)",
"times(s(times(?z:Nat,?y:Nat)),plus(?z:Nat,s(?z:Nat))) = times(?z:Nat,?z:Nat)",
"times(plus(s(plus(?z:Nat,?z:Nat)),?y:Nat),s(?y:Nat)) = s(s(?z:Nat))",
"times(plus(plus(s(?y:Nat),s(?x:Nat)),?z:Nat),?y:Nat) = times(?y:Nat,?y:Nat)",
"times(times(s(s(?z:Nat)),times(?x:Nat,?z:Nat)),?x:Nat) = s(s(?x:Nat))",
"plus(?z:Nat,times(plus(times(?y:Nat,?y:Nat),s(?z:Nat)),?y:Nat)) = s(?z:Nat)",
"plus(s(plus(?x:Nat,plus(?z:Nat,?x:Nat))),times(?x:Nat,?z:Nat)) = s(?y:Nat)",
"plus(plus(plus(times(?z:Nat,?y:Nat),s(?y:Nat)),?y:Nat),?y:Nat) = s(?z:Nat)",
"plus(times(plus(plus(?x:Nat,?z:Nat),?y:Nat),s(?y:Nat)),?x:Nat) = s(?y:Nat)",
"times(?x:Nat,plus(?y:Nat,times(s(?y:Nat),plus(?z:Nat,?y:Nat)))) = s(?z:Nat)",
"times(?y:Nat,times(plus(times(?z:Nat,?y:Nat),?x:Nat),s(?y:Nat))) = s(?y:Nat)",
"times(s(?y:Nat),times(times(s(?x:Nat),s(?x:Nat)),?y:Nat)) = s(?x:Nat)",
"times(s(plus(s(?x:Nat),s(?x:Nat))),plus(?z:Nat,?z:Nat)) = s(?x:Nat)",
"times(s(times(times(?x:Nat,?y:Nat),times(?y:Nat,?y:Nat))),?y:Nat) = s(?y:Nat)",
"times(plus(?y:Nat,s(s(?y:Nat))),s(s(s(?z:Nat)))) = s(?x:Nat)",
"times(plus(times(plus(?x:Nat,?y:Nat),s(?y:Nat)),?x:Nat),?z:Nat) = s(?z:Nat)",
"times(times(?y:Nat,?x:Nat),s(times(times(?z:Nat,?x:Nat),?y:Nat))) = s(?x:Nat)",
"times(times(times(?y:Nat,s(?z:Nat)),?x:Nat),plus(?y:Nat,?x:Nat)) = s(?z:Nat)",
"s(?x:Nat) = times(?z:Nat,plus(plus(s(?z:Nat),s(?z:Nat)),times(?y:Nat,?x:Nat)))",
"s(?x:Nat) = times(s(?y:Nat),s(plus(plus(?z:Nat,?z:Nat),plus(?y:Nat,?x:Nat))))",
"s(?x:Nat) = times(times(plus(s(?x:Nat),?z:Nat),s(times(?y:Nat,?y:Nat))),?z:Nat)",
"s(?y:Nat) = plus(?x:Nat,plus(?z:Nat,times(times(?z:Nat,?y:Nat),plus(?x:Nat,?y:Nat))))",
"s(?y:Nat) = plus(s(?z:Nat),times(?x:Nat,plus(plus(?y:Nat,?y:Nat),s(?z:Nat))))",
"s(?y:Nat) = plus(plus(?x:Nat,times(times(?z:Nat,?x:Nat),times(?z:Nat,?x:Nat))),?z:Nat)",
"s(?y:Nat) = plus(times(?y:Nat,s(?x:Nat)),plus(?x:Nat,s(times(?y:Nat,?x:Nat))))",
"s(?z:Nat) = plus(plus(times(s(?z:Nat),?z:Nat),s(times(?y:Nat,?y:Nat))),?y:Nat)",
"s(?z:Nat) = plus(times(times(plus(?y:Nat,?z:Nat),plus(?z:Nat,?y:Nat)),?z:Nat),?x:Nat)",
"s(?z:Nat) = times(?z:Nat,plus(s(?y:Nat),plus(times(?x:Nat,?x:Nat),s(?z:Nat))))",
"s(?z:Nat) = times(times(plus(?x:Nat,times(?z:Nat,?z:Nat)),times(?y:Nat,?y:Nat)),?x:Nat)",
"s(s(?z:Nat)) = plus(?z:Nat,plus(times(times(?x:Nat,?y:Nat),?y:Nat),s(?x:Nat)))",
"plus(?x:Nat,?y:Nat) = times(times(times(s(?x:Nat),?y:Nat),s(?z:Nat)),s(?y:Nat))",
"plus(?x:Nat,?z:Nat) = plus(?z:Nat,times(times(times(?x:Nat,?x:Nat),s(?x:Nat)),?x:Nat))",
"plus(?x:Nat,?z:Nat) = plus(times(times(plus(?z:Nat,?y:Nat),s(?y:Nat)),?y:Nat),?y:Nat)",
"plus(?z:Nat,?x:Nat) = times(?y:Nat,plus(times(s(?z:Nat),plus(?z:Nat,?z:Nat)),?y:Nat))",
"plus(?z:Nat,?x:Nat) = times(times(s(?x:Nat),plus(?z:Nat,s(?y:Nat))),s(?y:Nat))",
"plus(?z:Nat,?y:Nat) = s(times(?y:Nat,plus(times(?y:Nat,?z:Nat),plus(?z:Nat,?y:Nat))))",
"plus(?z:Nat,?y:Nat) = times(times(?y:Nat,?x:Nat),plus(times(?x:Nat,s(?x:Nat)),?z:Nat))",
"plus(?z:Nat,?z:Nat) = times(s(times(s(?z:Nat),times(?x:Nat,?y:Nat))),s(?x:Nat))",
"times(?x:Nat,?z:Nat) = times(plus(?y:Nat,plus(plus(?x:Nat,?z:Nat),s(?z:Nat))),?y:Nat)",
"times(?y:Nat,?x:Nat) = plus(?y:Nat,plus(times(s(?z:Nat),plus(?x:Nat,?z:Nat)),?x:Nat))",
"times(?y:Nat,?z:Nat) = plus(plus(?z:Nat,?x:Nat),s(times(?y:Nat,plus(?z:Nat,?y:Nat))))",
"times(?z:Nat,?y:Nat) = plus(times(?y:Nat,plus(?z:Nat,?x:Nat)),plus(?z:Nat,s(?y:Nat)))",
"s(times(?x:Nat,?y:Nat)) = times(?y:Nat,plus(plus(times(?x:Nat,?y:Nat),?x:Nat),?x:Nat))",
"plus(?x:Nat,s(?y:Nat)) = times(?x:Nat,s(times(s(?x:Nat),plus(?z:Nat,?x:Nat))))",
"plus(s(?x:Nat),?x:Nat) = plus(s(times(times(?z:Nat,?z:Nat),s(?x:Nat))),?y:Nat)",
"times(s(?x:Nat),?y:Nat) = times(s(times(s(?x:Nat),times(?x:Nat,?z:Nat))),?y:Nat)",
"times(s(?y:Nat),?z:Nat) = s(times(?z:Nat,plus(times(?z:Nat,?x:Nat),s(?y:Nat))))",
"s(s(s(s(?z:Nat)))) = times(?y:Nat,s(times(?y:Nat,plus(?x:Nat,?y:Nat))))",
"s(s(times(?x:Nat,?y:Nat))) = times(s(plus(?z:Nat,plus(?y:Nat,?y:Nat))),?z:Nat)",
"plus(s(s(?x:Nat)),?x:Nat) = times(plus(s(plus(?z:Nat,?x:Nat)),?y:Nat),?y:Nat)",
"times(plus(?y:Nat,?y:Nat),?y:Nat) = s(times(?x:Nat,plus(s(?y:Nat),s(?x:Nat))))",
"s(s(s(plus(?y:Nat,?y:Nat)))) = plus(s(s(?z:Nat)),times(?x:Nat,?z:Nat))",
"s(times(?y:Nat,s(s(?z:Nat)))) = times(?x:Nat,times(plus(?z:Nat,?z:Nat),?y:Nat))",
"plus(?z:Nat,times(?z:Nat,s(?x:Nat))) = plus(s(s(plus(?z:Nat,?y:Nat))),?x:Nat)",
"plus(s(times(?x:Nat,?z:Nat)),?x:Nat) = plus(s(s(times(?y:Nat,?y:Nat))),?z:Nat)",
"plus(plus(s(?z:Nat),?x:Nat),?z:Nat) = times(?y:Nat,plus(?z:Nat,s(s(?z:Nat))))",
"times(?y:Nat,s(plus(?x:Nat,?z:Nat))) = times(plus(?x:Nat,?x:Nat),plus(?y:Nat,?z:Nat))",
"times(s(?x:Nat),times(?z:Nat,?z:Nat)) = s(times(?z:Nat,s(plus(?x:Nat,?x:Nat))))",
"times(s(s(s(?x:Nat))),?z:Nat) = plus(s(?z:Nat),s(s(s(?z:Nat))))",
"s(s(times(?x:Nat,plus(?x:Nat,?y:Nat)))) = times(plus(?x:Nat,s(?x:Nat)),?x:Nat)",
"s(plus(s(plus(?z:Nat,?x:Nat)),?z:Nat)) = times(?y:Nat,s(s(s(?x:Nat))))",
"s(times(s(?z:Nat),plus(?z:Nat,?x:Nat))) = plus(s(s(?x:Nat)),s(?y:Nat))",
"times(s(s(times(?y:Nat,?y:Nat))),?x:Nat) = s(s(s(plus(?y:Nat,?z:Nat))))",
"s(plus(?y:Nat,times(?z:Nat,times(?y:Nat,?x:Nat)))) = times(s(s(?z:Nat)),?y:Nat)",
"plus(?y:Nat,times(?y:Nat,plus(s(?y:Nat),?z:Nat))) = s(times(?x:Nat,s(?y:Nat)))",
"plus(s(times(s(?x:Nat),s(?x:Nat))),?z:Nat) = s(s(s(s(?z:Nat))))",
"times(s(s(?z:Nat)),s(s(s(?x:Nat)))) = plus(times(?z:Nat,?x:Nat),?y:Nat)",
"plus(?y:Nat,s(times(s(?z:Nat),plus(?y:Nat,?x:Nat)))) = s(plus(?y:Nat,?y:Nat))",
"plus(s(times(s(?z:Nat),times(?z:Nat,?z:Nat))),?x:Nat) = s(plus(?y:Nat,?y:Nat))",
"plus(times(s(?x:Nat),?y:Nat),plus(?x:Nat,s(?x:Nat))) = plus(s(?x:Nat),?z:Nat)",
"times(?y:Nat,plus(s(plus(?x:Nat,?x:Nat)),s(?x:Nat))) = s(plus(?z:Nat,?x:Nat))",
"times(?z:Nat,plus(s(?z:Nat),s(times(?x:Nat,?y:Nat)))) = times(s(?z:Nat),?y:Nat)",
"times(plus(s(plus(?y:Nat,?y:Nat)),?z:Nat),s(?z:Nat)) = s(times(?x:Nat,?y:Nat))",
"s(plus(?z:Nat,times(plus(?x:Nat,?y:Nat),times(?z:Nat,?z:Nat)))) = times(?y:Nat,?y:Nat)",
"s(times(?x:Nat,times(times(?z:Nat,?z:Nat),times(?x:Nat,?x:Nat)))) = plus(?y:Nat,?x:Nat)",
"s(times(plus(plus(?z:Nat,?y:Nat),plus(?x:Nat,?x:Nat)),?x:Nat)) = times(?y:Nat,?z:Nat)",
"plus(?x:Nat,plus(plus(times(?z:Nat,?x:Nat),s(?y:Nat)),?z:Nat)) = plus(?x:Nat,?x:Nat)",
"plus(?y:Nat,s(times(times(?x:Nat,?z:Nat),times(?x:Nat,?y:Nat)))) = times(?y:Nat,?y:Nat)",
"plus(?z:Nat,s(times(times(?y:Nat,?x:Nat),times(?x:Nat,?x:Nat)))) = times(?x:Nat,?z:Nat)",
"plus(s(?x:Nat),plus(?x:Nat,times(?y:Nat,times(?z:Nat,?y:Nat)))) = plus(?z:Nat,?x:Nat)",
"plus(plus(times(?y:Nat,?z:Nat),s(s(?z:Nat))),s(?x:Nat)) = times(?y:Nat,?z:Nat)",
"plus(times(?z:Nat,?z:Nat),s(times(plus(?x:Nat,?y:Nat),?z:Nat))) = plus(?y:Nat,?y:Nat)",
"times(plus(times(plus(?y:Nat,?z:Nat),s(?z:Nat)),?y:Nat),?x:Nat) = times(?y:Nat,?y:Nat)",
"times(times(?y:Nat,plus(plus(?z:Nat,?x:Nat),s(?y:Nat))),?y:Nat) = times(?x:Nat,?x:Nat)",
"times(times(plus(s(?y:Nat),times(?x:Nat,?z:Nat)),?x:Nat),?y:Nat) = s(s(?y:Nat))",
"plus(s(s(times(?z:Nat,?y:Nat))),s(times(?z:Nat,s(?x:Nat)))) = s(?z:Nat)",
"plus(times(plus(times(?y:Nat,?z:Nat),?y:Nat),?y:Nat),s(s(?z:Nat))) = s(?y:Nat)",
"times(?y:Nat,plus(times(times(?x:Nat,?z:Nat),plus(?x:Nat,?z:Nat)),?y:Nat)) = s(?x:Nat)",
"times(s(s(times(?z:Nat,?z:Nat))),s(s(times(?y:Nat,?z:Nat)))) = s(?z:Nat)",
"times(s(plus(?x:Nat,times(?z:Nat,?z:Nat))),s(s(s(?x:Nat)))) = s(?z:Nat)",
"times(times(?x:Nat,?x:Nat),plus(times(?z:Nat,plus(?z:Nat,?y:Nat)),?x:Nat)) = s(?x:Nat)",
"times(times(s(times(?z:Nat,?x:Nat)),s(plus(?x:Nat,?x:Nat))),?z:Nat) = s(?y:Nat)",
"s(?x:Nat) = plus(times(s(s(?y:Nat)),?z:Nat),s(plus(times(?x:Nat,?y:Nat),?y:Nat)))",
"s(?x:Nat) = times(?y:Nat,plus(s(plus(?z:Nat,?x:Nat)),plus(?y:Nat,plus(?z:Nat,?z:Nat))))",
"s(?x:Nat) = times(times(plus(?z:Nat,times(?y:Nat,?z:Nat)),s(plus(?y:Nat,?x:Nat))),?x:Nat)",
"s(?y:Nat) = times(plus(s(s(?y:Nat)),plus(s(?y:Nat),s(?z:Nat))),s(?z:Nat))",
"s(?y:Nat) = times(plus(plus(?x:Nat,?z:Nat),?y:Nat),s(times(plus(?y:Nat,?x:Nat),?x:Nat)))",
"s(?y:Nat) = times(times(plus(times(?y:Nat,?x:Nat),?y:Nat),s(times(?y:Nat,?x:Nat))),?x:Nat)",
"s(?z:Nat) = plus(?z:Nat,times(times(?y:Nat,times(?y:Nat,?y:Nat)),s(times(?z:Nat,?x:Nat))))",
"s(?z:Nat) = plus(plus(?y:Nat,s(plus(?z:Nat,?z:Nat))),s(s(times(?z:Nat,?y:Nat))))",
"s(?z:Nat) = times(plus(?z:Nat,s(?z:Nat)),plus(times(times(?x:Nat,?y:Nat),?x:Nat),?x:Nat))",
"s(s(?y:Nat)) = times(plus(s(s(?y:Nat)),?y:Nat),s(s(times(?x:Nat,?x:Nat))))",
"s(s(?y:Nat)) = times(times(?z:Nat,plus(plus(?y:Nat,?z:Nat),times(?y:Nat,?y:Nat))),?z:Nat)",
"s(s(?z:Nat)) = times(s(plus(?y:Nat,?x:Nat)),plus(?y:Nat,s(plus(?y:Nat,?x:Nat))))",
"plus(?z:Nat,?x:Nat) = plus(plus(s(times(?z:Nat,?x:Nat)),plus(s(?y:Nat),?x:Nat)),?x:Nat)",
"plus(?z:Nat,?x:Nat) = times(s(?x:Nat),times(plus(s(?x:Nat),plus(?z:Nat,?x:Nat)),?z:Nat))",
"times(?y:Nat,?x:Nat) = times(times(?y:Nat,s(?z:Nat)),plus(?x:Nat,s(plus(?x:Nat,?z:Nat))))",
"times(?y:Nat,?y:Nat) = plus(times(times(?x:Nat,?y:Nat),s(times(?z:Nat,?x:Nat))),s(?x:Nat))",
"s(plus(?x:Nat,?y:Nat)) = plus(?x:Nat,times(?z:Nat,plus(s(?z:Nat),plus(?y:Nat,?y:Nat))))",
"s(plus(?x:Nat,?z:Nat)) = plus(s(plus(times(?x:Nat,?y:Nat),s(?z:Nat))),s(?y:Nat))",
"s(plus(?x:Nat,?z:Nat)) = times(plus(times(s(?z:Nat),times(?z:Nat,?z:Nat)),?y:Nat),?x:Nat)",
"s(times(?y:Nat,?x:Nat)) = plus(times(plus(s(?y:Nat),plus(?z:Nat,?y:Nat)),?x:Nat),?z:Nat)",
"plus(s(?y:Nat),?z:Nat) = times(?z:Nat,times(plus(times(?y:Nat,?x:Nat),s(?y:Nat)),?z:Nat))",
"times(s(?x:Nat),?z:Nat) = times(?z:Nat,s(times(plus(?x:Nat,?y:Nat),plus(?y:Nat,?z:Nat))))",
"s(s(s(s(?z:Nat)))) = plus(plus(?y:Nat,times(plus(?y:Nat,?z:Nat),?z:Nat)),?z:Nat)",
"s(s(plus(?y:Nat,?z:Nat))) = times(?x:Nat,times(plus(s(?y:Nat),s(?z:Nat)),?x:Nat))",
"plus(times(?x:Nat,?z:Nat),?z:Nat) = times(?y:Nat,times(s(times(?z:Nat,?x:Nat)),s(?x:Nat)))",
"plus(times(?y:Nat,?y:Nat),?y:Nat) = s(times(plus(?z:Nat,?z:Nat),s(times(?y:Nat,?y:Nat))))",
"plus(times(?z:Nat,?x:Nat),?y:Nat) = plus(plus(?y:Nat,plus(s(?y:Nat),?x:Nat)),s(?y:Nat))",
"times(?y:Nat,plus(?y:Nat,?z:Nat)) = times(?y:Nat,plus(plus(?z:Nat,times(?y:Nat,?z:Nat)),?x:Nat))",
"times(s(?z:Nat),s(?z:Nat)) = s(plus(times(times(?z:Nat,?x:Nat),s(?z:Nat)),?y:Nat))",
"times(times(?z:Nat,?x:Nat),?x:Nat) = plus(?z:Nat,times(times(times(?y:Nat,?z:Nat),?y:Nat),?z:Nat))",
"s(s(s(plus(?x:Nat,?x:Nat)))) = times(s(times(times(?x:Nat,?z:Nat),?z:Nat)),?z:Nat)",
"s(s(s(plus(?x:Nat,?z:Nat)))) = times(s(plus(plus(?z:Nat,?y:Nat),?z:Nat)),?y:Nat)",
"s(s(s(plus(?y:Nat,?y:Nat)))) = times(?x:Nat,times(?y:Nat,times(s(?y:Nat),?y:Nat)))",
"s(s(times(?z:Nat,s(?x:Nat)))) = plus(?y:Nat,times(?z:Nat,times(s(?y:Nat),?z:Nat)))",
"s(plus(s(s(?y:Nat)),?x:Nat)) = times(times(s(times(?z:Nat,?z:Nat)),?y:Nat),?z:Nat)",
"s(times(?z:Nat,plus(?z:Nat,?z:Nat))) = plus(s(plus(?z:Nat,plus(?z:Nat,?x:Nat))),?z:Nat)",
"plus(times(?x:Nat,s(?x:Nat)),?x:Nat) = plus(?x:Nat,s(plus(?z:Nat,times(?y:Nat,?x:Nat))))",
"plus(times(?x:Nat,s(?z:Nat)),?x:Nat) = plus(plus(?x:Nat,plus(?y:Nat,s(?z:Nat))),?x:Nat)",
"times(?z:Nat,times(?x:Nat,s(?z:Nat))) = s(times(plus(?y:Nat,times(?y:Nat,?x:Nat)),?x:Nat))",
"times(s(?y:Nat),plus(?y:Nat,?z:Nat)) = plus(plus(?x:Nat,?z:Nat),plus(?z:Nat,s(?z:Nat)))",
"s(s(times(s(?x:Nat),s(?z:Nat)))) = plus(?z:Nat,s(plus(s(?y:Nat),?z:Nat)))",
"s(plus(times(s(?y:Nat),?y:Nat),?z:Nat)) = plus(times(plus(?x:Nat,?x:Nat),?x:Nat),?y:Nat)",
"plus(?y:Nat,plus(times(?z:Nat,?x:Nat),?z:Nat)) = plus(s(s(times(?y:Nat,?z:Nat))),?z:Nat)",
"plus(s(s(plus(?y:Nat,?y:Nat))),?z:Nat) = times(s(s(?z:Nat)),plus(?z:Nat,?x:Nat))",
"times(?z:Nat,plus(plus(?x:Nat,?y:Nat),?y:Nat)) = plus(?x:Nat,plus(?z:Nat,s(s(?z:Nat))))",
"times(s(s(plus(?z:Nat,?z:Nat))),?y:Nat) = plus(s(times(s(?z:Nat),?x:Nat)),?y:Nat)",
"times(s(s(times(?x:Nat,?z:Nat))),?z:Nat) = s(times(?x:Nat,s(plus(?z:Nat,?x:Nat))))",
"plus(s(times(?y:Nat,plus(?y:Nat,?x:Nat))),?x:Nat) = s(times(plus(?z:Nat,?y:Nat),?y:Nat))",
"plus(times(s(?y:Nat),?z:Nat),plus(?y:Nat,?x:Nat)) = plus(plus(s(?z:Nat),?x:Nat),?z:Nat)",
"plus(times(plus(s(?x:Nat),?z:Nat),?z:Nat),?x:Nat) = s(s(s(plus(?x:Nat,?y:Nat))))",
"times(plus(?y:Nat,s(plus(?y:Nat,?z:Nat))),?y:Nat) = times(?z:Nat,times(?y:Nat,s(?y:Nat)))",
"times(times(?y:Nat,times(s(?x:Nat),?z:Nat)),?y:Nat) = s(times(s(s(?z:Nat)),?y:Nat))",
"s(plus(s(?x:Nat),times(plus(?y:Nat,?y:Nat),?z:Nat))) = times(?x:Nat,plus(?x:Nat,?y:Nat))",
"s(plus(plus(s(?x:Nat),?y:Nat),s(s(?z:Nat)))) = times(?x:Nat,plus(?z:Nat,?y:Nat))",
"plus(?y:Nat,times(plus(?z:Nat,?x:Nat),times(?z:Nat,?z:Nat))) = times(plus(?z:Nat,?y:Nat),?y:Nat)",
"plus(plus(?x:Nat,plus(times(?x:Nat,?x:Nat),?z:Nat)),?z:Nat) = plus(s(?y:Nat),s(?z:Nat))",
"times(?y:Nat,plus(s(s(?y:Nat)),s(s(?x:Nat)))) = s(times(s(?x:Nat),?z:Nat))",
"times(?y:Nat,plus(plus(?x:Nat,times(?x:Nat,?z:Nat)),?x:Nat)) = s(s(plus(?z:Nat,?y:Nat)))",
"times(plus(s(times(?z:Nat,?y:Nat)),s(?x:Nat)),?z:Nat) = times(times(?y:Nat,?y:Nat),?y:Nat)",
"times(times(?y:Nat,?z:Nat),s(plus(?y:Nat,s(?z:Nat)))) = plus(plus(?x:Nat,?y:Nat),?z:Nat)",
"times(times(plus(times(?z:Nat,?y:Nat),?y:Nat),?z:Nat),?x:Nat) = plus(times(?x:Nat,?z:Nat),?y:Nat)",
"times(times(times(plus(?z:Nat,?x:Nat),?z:Nat),?z:Nat),?x:Nat) = times(?y:Nat,s(s(?y:Nat)))",
"plus(?y:Nat,times(s(?y:Nat),plus(?x:Nat,plus(?x:Nat,?y:Nat)))) = plus(s(?z:Nat),?x:Nat)",
"plus(times(plus(times(?z:Nat,?x:Nat),?y:Nat),s(?y:Nat)),?z:Nat) = times(s(?x:Nat),?x:Nat)",
"times(?x:Nat,times(?z:Nat,times(s(?z:Nat),plus(?z:Nat,?y:Nat)))) = times(?x:Nat,s(?x:Nat))",
"times(times(?x:Nat,?y:Nat),plus(s(times(?z:Nat,?z:Nat)),?y:Nat)) = times(?x:Nat,s(?z:Nat))",
"plus(times(?y:Nat,times(plus(?y:Nat,?y:Nat),plus(?x:Nat,?x:Nat))),?y:Nat) = times(?y:Nat,?z:Nat)",
"times(?z:Nat,times(?z:Nat,times(times(?y:Nat,?z:Nat),plus(?x:Nat,?z:Nat)))) = plus(?y:Nat,?z:Nat)",
"times(s(?y:Nat),s(plus(plus(?y:Nat,?y:Nat),times(?z:Nat,?x:Nat)))) = times(?z:Nat,?x:Nat)",
"times(s(s(s(?x:Nat))),times(plus(?z:Nat,?x:Nat),s(?z:Nat))) = times(?z:Nat,?y:Nat)",
"times(plus(s(s(?y:Nat)),?x:Nat),s(s(times(?x:Nat,?x:Nat)))) = times(?x:Nat,?x:Nat)",
"times(plus(s(s(?z:Nat)),times(s(?x:Nat),s(?z:Nat))),?z:Nat) = s(s(?x:Nat))",
"times(plus(s(times(?y:Nat,?x:Nat)),s(times(?x:Nat,?z:Nat))),?y:Nat) = plus(?x:Nat,?y:Nat)",
"times(times(?x:Nat,?x:Nat),times(?x:Nat,plus(?x:Nat,plus(?z:Nat,?x:Nat)))) = plus(?x:Nat,?z:Nat)",
"times(times(times(?z:Nat,?x:Nat),plus(?y:Nat,?x:Nat)),s(s(?x:Nat))) = plus(?y:Nat,?x:Nat)",
"plus(?x:Nat,times(plus(plus(?z:Nat,?z:Nat),s(?x:Nat)),plus(?x:Nat,?x:Nat))) = s(?x:Nat)",
"plus(?z:Nat,times(times(s(?x:Nat),?z:Nat),plus(?z:Nat,plus(?y:Nat,?y:Nat)))) = s(?y:Nat)",
"plus(s(?z:Nat),times(s(s(?x:Nat)),times(plus(?y:Nat,?z:Nat),?z:Nat))) = s(?y:Nat)",
"plus(s(plus(?y:Nat,times(?x:Nat,?y:Nat))),times(?z:Nat,s(s(?y:Nat)))) = s(?y:Nat)",
"plus(plus(?x:Nat,s(plus(?x:Nat,?y:Nat))),s(s(times(?z:Nat,?z:Nat)))) = s(?x:Nat)",
"plus(times(plus(s(?x:Nat),times(?x:Nat,?x:Nat)),times(?x:Nat,?x:Nat)),?x:Nat) = s(?z:Nat)",
"times(?z:Nat,times(plus(plus(?z:Nat,?x:Nat),plus(?y:Nat,?x:Nat)),s(?x:Nat))) = s(?x:Nat)",
"times(?z:Nat,times(plus(times(?y:Nat,?y:Nat),plus(?x:Nat,?z:Nat)),s(?x:Nat))) = s(?x:Nat)",
"times(s(?z:Nat),plus(?z:Nat,times(times(?y:Nat,?y:Nat),plus(?y:Nat,?z:Nat)))) = s(?x:Nat)",
"times(s(s(?x:Nat)),plus(s(?x:Nat),times(?y:Nat,times(?z:Nat,?y:Nat)))) = s(?y:Nat)",
"times(s(s(times(?y:Nat,?x:Nat))),s(plus(plus(?z:Nat,?x:Nat),?y:Nat))) = s(?y:Nat)",
"times(times(s(times(?z:Nat,?z:Nat)),times(?z:Nat,times(?x:Nat,?z:Nat))),?z:Nat) = s(?x:Nat)",
"times(times(times(?y:Nat,s(?z:Nat)),plus(times(?z:Nat,?z:Nat),?z:Nat)),?y:Nat) = s(?x:Nat)",
"s(?y:Nat) = plus(s(plus(?z:Nat,?z:Nat)),plus(times(?x:Nat,?y:Nat),times(?y:Nat,s(?x:Nat))))",
"s(?y:Nat) = plus(plus(s(?y:Nat),?z:Nat),times(?z:Nat,times(s(?z:Nat),times(?x:Nat,?y:Nat))))",
"s(?y:Nat) = times(times(?y:Nat,plus(s(?x:Nat),?y:Nat)),s(plus(plus(?x:Nat,?z:Nat),?x:Nat)))",
"s(?z:Nat) = plus(plus(times(plus(?y:Nat,?z:Nat),s(?x:Nat)),s(?y:Nat)),plus(?x:Nat,?y:Nat))",
"s(?z:Nat) = times(plus(times(times(?x:Nat,?z:Nat),?y:Nat),s(times(?z:Nat,?x:Nat))),s(?x:Nat))",
"s(s(?x:Nat)) = plus(?z:Nat,times(plus(times(?z:Nat,?x:Nat),plus(?y:Nat,?y:Nat)),s(?y:Nat)))",
"s(s(?z:Nat)) = plus(?x:Nat,times(s(?y:Nat),times(plus(?x:Nat,?y:Nat),plus(?x:Nat,?x:Nat))))",
"s(s(?z:Nat)) = times(?z:Nat,plus(times(?x:Nat,plus(?y:Nat,?y:Nat)),s(times(?y:Nat,?y:Nat))))",
"plus(?x:Nat,?z:Nat) = plus(plus(times(plus(?z:Nat,?y:Nat),?z:Nat),s(times(?y:Nat,?x:Nat))),?y:Nat)",
"plus(?y:Nat,?x:Nat) = s(plus(plus(times(?x:Nat,?x:Nat),s(?x:Nat)),s(plus(?y:Nat,?z:Nat))))",
"plus(?y:Nat,?z:Nat) = s(times(times(?x:Nat,plus(?y:Nat,?y:Nat)),plus(times(?z:Nat,?z:Nat),?x:Nat)))",
"times(?x:Nat,?x:Nat) = s(plus(times(plus(?x:Nat,?y:Nat),?z:Nat),times(?x:Nat,plus(?y:Nat,?x:Nat))))",
"times(?x:Nat,?x:Nat) = times(s(times(?x:Nat,plus(?x:Nat,?y:Nat))),s(s(times(?z:Nat,?y:Nat))))",
"times(?x:Nat,?x:Nat) = times(times(plus(s(?y:Nat),?x:Nat),times(s(?z:Nat),?z:Nat)),s(?x:Nat))",
"times(?y:Nat,?x:Nat) = plus(?z:Nat,plus(plus(plus(?z:Nat,?x:Nat),plus(?y:Nat,?z:Nat)),s(?y:Nat)))",
"times(?y:Nat,?z:Nat) = times(plus(s(?y:Nat),times(times(?z:Nat,?x:Nat),times(?y:Nat,?x:Nat))),?y:Nat)",
"times(?z:Nat,?y:Nat) = times(?z:Nat,times(plus(s(?x:Nat),plus(?x:Nat,?z:Nat)),times(?x:Nat,?y:Nat)))",
"times(?z:Nat,?z:Nat) = times(plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),plus(?y:Nat,s(?y:Nat))),?y:Nat)",
"s(plus(?x:Nat,?z:Nat)) = times(times(s(plus(?x:Nat,?z:Nat)),?z:Nat),s(plus(?y:Nat,?x:Nat)))",
"s(plus(?y:Nat,?z:Nat)) = plus(?x:Nat,plus(times(plus(?x:Nat,?z:Nat),plus(?y:Nat,?z:Nat)),?x:Nat))",
"plus(?x:Nat,s(?x:Nat)) = times(s(times(times(?x:Nat,?y:Nat),plus(?x:Nat,?y:Nat))),s(?y:Nat))",
"plus(?x:Nat,s(?y:Nat)) = plus(?z:Nat,plus(times(times(?y:Nat,?y:Nat),plus(?x:Nat,?z:Nat)),?y:Nat))",
"s(times(?y:Nat,s(?z:Nat))) = times(s(plus(plus(?y:Nat,?z:Nat),?y:Nat)),s(s(?y:Nat)))",
"plus(?y:Nat,s(s(?y:Nat))) = plus(?z:Nat,s(plus(plus(?x:Nat,?z:Nat),times(?x:Nat,?x:Nat))))",
"plus(s(s(?y:Nat)),?z:Nat) = plus(plus(plus(s(?z:Nat),times(?z:Nat,?y:Nat)),?z:Nat),?x:Nat)",
"plus(plus(?y:Nat,?y:Nat),?y:Nat) = plus(?y:Nat,times(times(times(?z:Nat,?x:Nat),s(?x:Nat)),?z:Nat))",
"times(s(?x:Nat),s(?z:Nat)) = plus(?x:Nat,s(plus(times(?x:Nat,?y:Nat),times(?z:Nat,?z:Nat))))",
"times(plus(?z:Nat,?x:Nat),?z:Nat) = times(?x:Nat,s(plus(plus(?z:Nat,?z:Nat),times(?y:Nat,?z:Nat))))",
"s(s(s(plus(?y:Nat,?z:Nat)))) = times(?y:Nat,times(plus(?z:Nat,s(?x:Nat)),s(?y:Nat)))",
"s(times(?x:Nat,times(?x:Nat,?z:Nat))) = plus(times(times(?x:Nat,times(?z:Nat,?z:Nat)),?x:Nat),?y:Nat)",
"times(?x:Nat,s(s(s(?x:Nat)))) = plus(times(s(plus(?z:Nat,?x:Nat)),?x:Nat),s(?y:Nat))",
"times(?z:Nat,times(?z:Nat,s(?y:Nat))) = s(plus(times(plus(?z:Nat,?z:Nat),?y:Nat),s(?x:Nat)))",
"times(plus(?y:Nat,s(?x:Nat)),?x:Nat) = plus(?y:Nat,times(plus(s(?y:Nat),?x:Nat),s(?z:Nat)))",
"s(s(plus(?z:Nat,plus(?z:Nat,?y:Nat)))) = plus(s(times(times(?y:Nat,?x:Nat),?z:Nat)),?y:Nat)",
"plus(?z:Nat,plus(?y:Nat,s(s(?y:Nat)))) = s(times(?x:Nat,times(s(?x:Nat),s(?y:Nat))))",
"plus(s(s(times(?z:Nat,?z:Nat))),?x:Nat) = times(s(times(times(?z:Nat,?x:Nat),?x:Nat)),?z:Nat)",
"plus(s(plus(?z:Nat,?y:Nat)),s(?x:Nat)) = times(s(?y:Nat),times(?x:Nat,times(?z:Nat,?z:Nat)))",
"times(?x:Nat,plus(?x:Nat,times(?y:Nat,?z:Nat))) = plus(?z:Nat,plus(s(times(?y:Nat,?x:Nat)),?x:Nat))",
"times(?y:Nat,s(s(times(?y:Nat,?x:Nat)))) = plus(plus(?y:Nat,?x:Nat),times(?x:Nat,s(?z:Nat)))",
"times(plus(times(?y:Nat,?y:Nat),?x:Nat),?x:Nat) = s(s(times(times(?y:Nat,?z:Nat),s(?x:Nat))))",
"s(s(plus(s(?y:Nat),times(?y:Nat,?z:Nat)))) = times(s(s(times(?y:Nat,?x:Nat))),?x:Nat)",
"s(s(times(s(?y:Nat),times(?x:Nat,?x:Nat)))) = times(?x:Nat,times(?y:Nat,s(s(?x:Nat))))",
"plus(?x:Nat,times(?z:Nat,times(s(?y:Nat),?x:Nat))) = s(plus(s(times(?x:Nat,?y:Nat)),?y:Nat))",
"plus(?z:Nat,plus(plus(?z:Nat,s(?y:Nat)),?y:Nat)) = plus(times(?z:Nat,s(?z:Nat)),s(?x:Nat))",
"times(plus(?y:Nat,plus(s(?y:Nat),?z:Nat)),?y:Nat) = s(times(s(plus(?x:Nat,?x:Nat)),?x:Nat))",
"s(plus(?x:Nat,times(s(?x:Nat),plus(?y:Nat,?x:Nat)))) = plus(s(?x:Nat),times(?z:Nat,?x:Nat))",
"plus(?y:Nat,plus(plus(times(?z:Nat,?x:Nat),?x:Nat),?x:Nat)) = times(?z:Nat,plus(s(?z:Nat),?z:Nat))",
"plus(plus(s(s(?z:Nat)),?z:Nat),plus(?x:Nat,?x:Nat)) = times(times(?z:Nat,s(?z:Nat)),?z:Nat)",
"plus(times(times(times(?y:Nat,?x:Nat),?y:Nat),?x:Nat),?y:Nat) = times(times(s(?z:Nat),?z:Nat),?z:Nat)",
"plus(?y:Nat,plus(s(plus(?y:Nat,?y:Nat)),times(?x:Nat,?z:Nat))) = s(s(s(s(?z:Nat))))",
"plus(?z:Nat,plus(s(?z:Nat),times(plus(?y:Nat,?y:Nat),?x:Nat))) = plus(s(?y:Nat),s(?x:Nat))",
"plus(times(?x:Nat,plus(s(?z:Nat),plus(?y:Nat,?z:Nat))),?x:Nat) = times(times(?x:Nat,?z:Nat),?z:Nat)",
"plus(times(plus(times(?x:Nat,?x:Nat),s(?y:Nat)),?z:Nat),?z:Nat) = times(times(?y:Nat,?x:Nat),?z:Nat)",
"times(s(s(plus(?y:Nat,?y:Nat))),s(s(s(?x:Nat)))) = s(s(times(?z:Nat,?z:Nat)))",
"plus(?x:Nat,plus(plus(plus(?z:Nat,?x:Nat),plus(?z:Nat,?y:Nat)),?z:Nat)) = times(?z:Nat,s(?z:Nat))",
"plus(plus(?y:Nat,times(times(?y:Nat,?z:Nat),plus(?z:Nat,?x:Nat))),?x:Nat) = plus(?x:Nat,s(?y:Nat))",
"plus(plus(s(s(?y:Nat)),plus(?y:Nat,times(?z:Nat,?y:Nat))),?y:Nat) = plus(s(?z:Nat),?y:Nat)",
"plus(times(times(times(?x:Nat,?y:Nat),?z:Nat),?x:Nat),s(s(?x:Nat))) = times(s(?y:Nat),?y:Nat)",
"times(?x:Nat,plus(plus(s(?x:Nat),plus(?x:Nat,?y:Nat)),s(?z:Nat))) = plus(?y:Nat,s(?x:Nat))",
"times(?z:Nat,times(s(?x:Nat),plus(times(?y:Nat,?x:Nat),s(?x:Nat)))) = plus(?x:Nat,s(?x:Nat))",
"plus(?x:Nat,times(s(?z:Nat),times(times(?z:Nat,?x:Nat),plus(?z:Nat,?z:Nat)))) = s(s(?x:Nat))",
"plus(s(times(?z:Nat,s(?y:Nat))),times(s(times(?x:Nat,?x:Nat)),?x:Nat)) = times(?z:Nat,?x:Nat)",
"plus(plus(times(times(?x:Nat,?x:Nat),?x:Nat),?x:Nat),s(s(s(?z:Nat)))) = times(?x:Nat,?x:Nat)",
"plus(times(s(?z:Nat),times(times(?y:Nat,?x:Nat),plus(?y:Nat,?z:Nat))),?x:Nat) = times(?z:Nat,?y:Nat)",
"plus(times(plus(plus(?z:Nat,?x:Nat),plus(?z:Nat,?y:Nat)),?x:Nat),s(?x:Nat)) = s(s(?z:Nat))",
"times(?y:Nat,plus(s(times(?z:Nat,?x:Nat)),times(?x:Nat,times(?y:Nat,?y:Nat)))) = plus(?x:Nat,?z:Nat)",
"times(s(plus(s(?z:Nat),s(?x:Nat))),times(times(?x:Nat,?y:Nat),?y:Nat)) = s(s(?z:Nat))",
"times(plus(s(times(?x:Nat,?y:Nat)),times(?y:Nat,plus(?y:Nat,?x:Nat))),?x:Nat) = s(s(?z:Nat))",
"times(plus(times(s(?x:Nat),times(?y:Nat,?x:Nat)),?x:Nat),plus(?z:Nat,?y:Nat)) = plus(?y:Nat,?y:Nat)",
"times(times(?y:Nat,?z:Nat),times(plus(s(?y:Nat),times(?x:Nat,?y:Nat)),?x:Nat)) = plus(?y:Nat,?y:Nat)",
"plus(?z:Nat,times(s(s(?y:Nat)),plus(times(?x:Nat,?x:Nat),plus(?x:Nat,?y:Nat)))) = s(?x:Nat)",
"plus(s(plus(s(?y:Nat),s(?y:Nat))),times(?y:Nat,s(times(?x:Nat,?x:Nat)))) = s(?y:Nat)",
"plus(times(?y:Nat,s(s(?y:Nat))),times(?y:Nat,plus(?x:Nat,plus(?y:Nat,?z:Nat)))) = s(?x:Nat)",
"times(plus(s(times(?z:Nat,?z:Nat)),times(s(?y:Nat),times(?x:Nat,?x:Nat))),?x:Nat) = s(?x:Nat)",
"times(times(?z:Nat,s(?z:Nat)),times(?x:Nat,times(s(?x:Nat),plus(?x:Nat,?z:Nat)))) = s(?y:Nat)",
"s(?x:Nat) = plus(plus(times(plus(?y:Nat,?y:Nat),?z:Nat),times(plus(?x:Nat,?y:Nat),s(?x:Nat))),?z:Nat)",
"s(?x:Nat) = times(plus(plus(?z:Nat,times(?y:Nat,?z:Nat)),plus(s(?x:Nat),plus(?z:Nat,?y:Nat))),?y:Nat)",
"s(?y:Nat) = plus(s(times(s(?y:Nat),times(?x:Nat,?x:Nat))),s(times(times(?x:Nat,?x:Nat),?z:Nat)))",
"s(?y:Nat) = plus(plus(?y:Nat,plus(plus(?x:Nat,?x:Nat),times(?z:Nat,?y:Nat))),plus(s(?y:Nat),?x:Nat))",
"s(?z:Nat) = plus(s(s(times(?x:Nat,?y:Nat))),times(?z:Nat,plus(s(?y:Nat),plus(?x:Nat,?y:Nat))))",
"s(?z:Nat) = plus(plus(s(?z:Nat),times(?y:Nat,plus(?x:Nat,?x:Nat))),s(times(?y:Nat,s(?z:Nat))))",
"s(?z:Nat) = times(plus(times(?z:Nat,?y:Nat),times(plus(?y:Nat,?y:Nat),s(?z:Nat))),plus(?x:Nat,?z:Nat))",
"s(?z:Nat) = times(times(times(?z:Nat,plus(?x:Nat,?y:Nat)),plus(times(?x:Nat,?x:Nat),s(?z:Nat))),?x:Nat)",
"s(s(?z:Nat)) = times(times(s(?y:Nat),plus(s(?z:Nat),s(?y:Nat))),s(s(s(?x:Nat))))",
"plus(?x:Nat,?x:Nat) = plus(?y:Nat,times(times(times(?z:Nat,?z:Nat),s(?x:Nat)),s(plus(?x:Nat,?z:Nat))))",
"plus(?x:Nat,?z:Nat) = times(times(?y:Nat,s(plus(?y:Nat,?y:Nat))),times(s(plus(?x:Nat,?z:Nat)),?z:Nat))",
"plus(?y:Nat,?y:Nat) = plus(?x:Nat,times(plus(s(?z:Nat),plus(?y:Nat,?z:Nat)),s(times(?y:Nat,?y:Nat))))",
"plus(?y:Nat,?y:Nat) = times(plus(times(times(?y:Nat,?x:Nat),times(?x:Nat,?x:Nat)),?x:Nat),times(?x:Nat,?z:Nat))",
"plus(?y:Nat,?z:Nat) = plus(plus(s(plus(?x:Nat,?y:Nat)),?z:Nat),s(times(times(?y:Nat,?y:Nat),?x:Nat)))",
"plus(?y:Nat,?z:Nat) = times(plus(?z:Nat,?x:Nat),times(plus(plus(?x:Nat,?y:Nat),times(?z:Nat,?y:Nat)),?z:Nat))",
"times(?x:Nat,?x:Nat) = plus(s(plus(plus(?y:Nat,?z:Nat),?y:Nat)),plus(s(times(?z:Nat,?x:Nat)),?z:Nat))",
"times(?x:Nat,?y:Nat) = times(times(s(plus(?y:Nat,?y:Nat)),plus(s(?y:Nat),plus(?x:Nat,?x:Nat))),?y:Nat)",
"times(?x:Nat,?z:Nat) = times(plus(plus(?x:Nat,s(?y:Nat)),?z:Nat),plus(?y:Nat,plus(s(?y:Nat),?z:Nat)))",
"times(?y:Nat,?x:Nat) = plus(?y:Nat,times(s(times(?z:Nat,?x:Nat)),times(s(?z:Nat),plus(?y:Nat,?x:Nat))))",
"times(?z:Nat,?y:Nat) = times(plus(times(?z:Nat,plus(?z:Nat,?z:Nat)),s(s(?y:Nat))),times(?x:Nat,?x:Nat))",
"s(times(?z:Nat,?y:Nat)) = plus(s(plus(s(?y:Nat),?y:Nat)),times(s(plus(?y:Nat,?y:Nat)),?z:Nat))",
"plus(?z:Nat,s(?y:Nat)) = plus(plus(times(?z:Nat,times(?y:Nat,?z:Nat)),s(times(?x:Nat,?y:Nat))),?x:Nat)",
"plus(?z:Nat,s(?z:Nat)) = s(times(s(plus(?x:Nat,?z:Nat)),times(s(?z:Nat),times(?x:Nat,?x:Nat))))",
"plus(s(?y:Nat),?x:Nat) = plus(s(?x:Nat),times(plus(times(?y:Nat,?x:Nat),plus(?y:Nat,?y:Nat)),?y:Nat))",
"times(s(?x:Nat),?z:Nat) = times(plus(s(plus(?y:Nat,?x:Nat)),?y:Nat),plus(?x:Nat,times(?y:Nat,?x:Nat)))",
"s(s(s(s(?y:Nat)))) = times(?x:Nat,times(s(plus(?z:Nat,?z:Nat)),s(times(?z:Nat,?x:Nat))))",
"s(s(plus(?x:Nat,?x:Nat))) = times(?y:Nat,plus(times(plus(?z:Nat,?y:Nat),times(?z:Nat,?z:Nat)),?x:Nat))",
"s(s(plus(?y:Nat,?x:Nat))) = times(plus(times(times(?y:Nat,?y:Nat),plus(?x:Nat,?x:Nat)),?x:Nat),?y:Nat)",
"s(s(plus(?z:Nat,?z:Nat))) = plus(?z:Nat,times(?y:Nat,times(times(?y:Nat,?y:Nat),times(?z:Nat,?y:Nat))))",
"plus(?y:Nat,times(?z:Nat,?y:Nat)) = plus(times(plus(times(?z:Nat,?z:Nat),times(?z:Nat,?y:Nat)),?y:Nat),?y:Nat)",
"times(?x:Nat,plus(?z:Nat,?x:Nat)) = plus(?y:Nat,plus(?z:Nat,plus(times(?x:Nat,?y:Nat),times(?z:Nat,?x:Nat))))",
"s(s(s(times(?z:Nat,?y:Nat)))) = plus(s(?y:Nat),times(plus(s(?y:Nat),s(?z:Nat)),?y:Nat))",
"s(s(times(s(?y:Nat),?y:Nat))) = plus(?y:Nat,plus(times(s(?y:Nat),?y:Nat),plus(?y:Nat,?x:Nat)))",
"s(s(times(plus(?y:Nat,?z:Nat),?y:Nat))) = times(?x:Nat,plus(s(?z:Nat),plus(?y:Nat,s(?y:Nat))))",
"s(plus(?y:Nat,s(plus(?x:Nat,?x:Nat)))) = plus(times(?x:Nat,plus(?x:Nat,s(?z:Nat))),s(?y:Nat))",
"s(times(?x:Nat,s(times(?x:Nat,?y:Nat)))) = times(?z:Nat,plus(?y:Nat,times(?z:Nat,plus(?z:Nat,?y:Nat))))",
"s(times(?z:Nat,s(times(?z:Nat,?x:Nat)))) = plus(s(s(plus(?x:Nat,?x:Nat))),plus(?x:Nat,?y:Nat))",
"s(times(times(s(?z:Nat),?y:Nat),?x:Nat)) = times(plus(times(s(?y:Nat),?z:Nat),s(?z:Nat)),?x:Nat)",
"plus(?x:Nat,s(times(?z:Nat,s(?x:Nat)))) = times(s(plus(?y:Nat,times(?y:Nat,?y:Nat))),s(?x:Nat))",
"plus(?y:Nat,plus(?z:Nat,plus(?z:Nat,?z:Nat))) = s(times(times(?x:Nat,?y:Nat),times(s(?z:Nat),?z:Nat)))",
"plus(?z:Nat,s(times(?y:Nat,s(?z:Nat)))) = plus(s(?y:Nat),s(times(times(?x:Nat,?z:Nat),?x:Nat)))",
"plus(times(s(?x:Nat),?y:Nat),s(?y:Nat)) = plus(?z:Nat,times(plus(?x:Nat,plus(?z:Nat,?y:Nat)),?x:Nat))",
"times(?y:Nat,times(?z:Nat,s(s(?y:Nat)))) = times(?y:Nat,s(plus(s(?z:Nat),times(?x:Nat,?y:Nat))))",
"times(plus(?y:Nat,?y:Nat),times(?z:Nat,?y:Nat)) = plus(s(plus(s(?z:Nat),plus(?y:Nat,?x:Nat))),?x:Nat)",
"s(s(times(s(?x:Nat),times(?x:Nat,?y:Nat)))) = plus(s(s(?x:Nat)),times(s(?x:Nat),?x:Nat))",
"s(plus(s(?x:Nat),s(times(?z:Nat,?y:Nat)))) = times(times(?x:Nat,plus(s(?z:Nat),?x:Nat)),?y:Nat)",
"s(plus(plus(?z:Nat,?z:Nat),plus(?y:Nat,?x:Nat))) = times(plus(s(s(?x:Nat)),s(?z:Nat)),?y:Nat)",
"s(plus(times(times(?y:Nat,?x:Nat),?y:Nat),?y:Nat)) = plus(?x:Nat,s(times(?x:Nat,times(?z:Nat,?z:Nat))))",
"plus(?y:Nat,plus(?z:Nat,times(s(?x:Nat),?z:Nat))) = s(s(times(times(?x:Nat,?y:Nat),s(?x:Nat))))",
"plus(s(plus(?x:Nat,times(?x:Nat,?z:Nat))),?x:Nat) = times(?z:Nat,times(?y:Nat,s(times(?x:Nat,?y:Nat))))",
"plus(s(plus(times(?x:Nat,?z:Nat),?x:Nat)),?y:Nat) = times(?y:Nat,s(plus(?z:Nat,plus(?y:Nat,?y:Nat))))",
"plus(s(times(plus(?x:Nat,?x:Nat),?z:Nat)),?z:Nat) = times(s(times(times(?x:Nat,?x:Nat),?y:Nat)),?z:Nat)",
"plus(plus(times(?z:Nat,?z:Nat),?z:Nat),s(?z:Nat)) = times(s(times(s(?y:Nat),s(?y:Nat))),?z:Nat)",
"times(?x:Nat,plus(s(s(?y:Nat)),s(?z:Nat))) = plus(?x:Nat,plus(?y:Nat,plus(s(?z:Nat),?z:Nat)))",
"times(s(plus(?x:Nat,times(?z:Nat,?x:Nat))),?x:Nat) = times(?x:Nat,s(plus(plus(?y:Nat,?y:Nat),?x:Nat)))",
"times(s(times(?y:Nat,times(?x:Nat,?z:Nat))),?y:Nat) = times(?y:Nat,times(times(s(?z:Nat),?y:Nat),?y:Nat))",
"times(times(?z:Nat,times(?y:Nat,s(?x:Nat))),?y:Nat) = plus(times(?z:Nat,s(s(?z:Nat))),s(?x:Nat))",
"times(times(s(?x:Nat),s(s(?x:Nat))),?z:Nat) = s(plus(plus(times(?x:Nat,?y:Nat),?z:Nat),?x:Nat))",
"s(plus(plus(?z:Nat,?y:Nat),plus(?y:Nat,s(?x:Nat)))) = times(s(s(plus(?z:Nat,?z:Nat))),?y:Nat)",
"plus(?x:Nat,s(times(plus(?y:Nat,?y:Nat),s(?x:Nat)))) = s(times(?x:Nat,s(times(?y:Nat,?y:Nat))))",
"plus(?x:Nat,times(s(?y:Nat),s(plus(?y:Nat,?z:Nat)))) = s(times(?x:Nat,s(times(?x:Nat,?y:Nat))))",
"plus(s(?y:Nat),times(?x:Nat,s(times(?x:Nat,?z:Nat)))) = plus(s(plus(?z:Nat,s(?x:Nat))),?x:Nat)",
"plus(s(times(?y:Nat,plus(?z:Nat,?z:Nat))),s(?x:Nat)) = times(times(s(s(?y:Nat)),?x:Nat),?z:Nat)",
"times(?x:Nat,times(?z:Nat,times(plus(?z:Nat,?x:Nat),?x:Nat))) = times(?z:Nat,s(s(plus(?z:Nat,?z:Nat))))",
"times(?y:Nat,plus(times(s(?x:Nat),s(?z:Nat)),?x:Nat)) = plus(plus(?z:Nat,?z:Nat),s(s(?y:Nat)))",
"times(plus(?y:Nat,?y:Nat),plus(s(s(?z:Nat)),?x:Nat)) = s(plus(times(?y:Nat,s(?x:Nat)),?z:Nat))",
"plus(?x:Nat,times(times(?y:Nat,times(?y:Nat,?z:Nat)),s(?y:Nat))) = plus(s(?x:Nat),s(s(?x:Nat)))",
"plus(?y:Nat,s(times(times(?y:Nat,?y:Nat),times(?y:Nat,?y:Nat)))) = times(s(s(s(?z:Nat))),?x:Nat)",
"plus(?z:Nat,s(plus(times(?x:Nat,?y:Nat),times(?y:Nat,?z:Nat)))) = s(plus(?z:Nat,times(?y:Nat,?z:Nat)))",
"times(?z:Nat,plus(s(?z:Nat),times(s(?z:Nat),s(?x:Nat)))) = s(s(s(times(?y:Nat,?z:Nat))))",
"times(s(s(?x:Nat)),plus(s(plus(?x:Nat,?z:Nat)),?y:Nat)) = s(s(s(times(?z:Nat,?y:Nat))))",
"times(plus(times(s(?x:Nat),?z:Nat),?y:Nat),s(s(?x:Nat))) = plus(plus(?x:Nat,?y:Nat),s(?x:Nat))",
"times(times(?y:Nat,?y:Nat),s(times(plus(?x:Nat,?y:Nat),?z:Nat))) = times(times(?y:Nat,s(?x:Nat)),?z:Nat)",
"plus(times(?y:Nat,plus(times(?x:Nat,?z:Nat),times(?y:Nat,?y:Nat))),?z:Nat) = plus(times(?y:Nat,?x:Nat),?z:Nat)",
"plus(times(plus(s(?x:Nat),?z:Nat),s(plus(?y:Nat,?x:Nat))),?x:Nat) = s(times(s(?x:Nat),?x:Nat))",
"times(times(plus(plus(?z:Nat,?z:Nat),s(?x:Nat)),s(?x:Nat)),?x:Nat) = times(?y:Nat,plus(?x:Nat,?z:Nat))",
"s(plus(plus(?x:Nat,times(?y:Nat,?y:Nat)),times(?y:Nat,plus(?y:Nat,?x:Nat)))) = plus(?y:Nat,s(?z:Nat))",
"s(times(plus(?y:Nat,plus(?z:Nat,?x:Nat)),times(plus(?z:Nat,?z:Nat),?x:Nat))) = plus(s(?z:Nat),?y:Nat)",
"plus(?x:Nat,plus(times(plus(?y:Nat,?z:Nat),?y:Nat),times(?y:Nat,s(?x:Nat)))) = s(plus(?x:Nat,?x:Nat))",
"plus(?y:Nat,plus(times(plus(?z:Nat,?y:Nat),s(?x:Nat)),s(s(?x:Nat)))) = s(times(?z:Nat,?z:Nat))",
"plus(s(times(plus(?y:Nat,?z:Nat),times(?x:Nat,?y:Nat))),times(?y:Nat,?z:Nat)) = plus(?x:Nat,s(?z:Nat))",
"plus(plus(s(?y:Nat),s(s(?y:Nat))),s(s(times(?y:Nat,?x:Nat)))) = plus(?y:Nat,s(?z:Nat))",
"plus(times(times(?z:Nat,times(?y:Nat,?x:Nat)),s(plus(?x:Nat,?y:Nat))),?z:Nat) = times(?x:Nat,s(?z:Nat))",
"times(times(s(times(?y:Nat,?y:Nat)),s(s(?x:Nat))),plus(?x:Nat,?x:Nat)) = plus(?y:Nat,s(?z:Nat))",
"s(plus(times(?x:Nat,times(?z:Nat,?y:Nat)),plus(plus(?x:Nat,?x:Nat),s(?x:Nat)))) = times(?y:Nat,?x:Nat)",
"plus(s(times(?x:Nat,?x:Nat)),plus(plus(?z:Nat,s(?z:Nat)),times(?z:Nat,?x:Nat))) = plus(?x:Nat,?y:Nat)",
"plus(times(plus(times(?x:Nat,?z:Nat),?x:Nat),s(?z:Nat)),s(s(s(?y:Nat)))) = s(s(?z:Nat))",
"plus(plus(s(?y:Nat),times(times(?y:Nat,?z:Nat),plus(?z:Nat,?z:Nat))),plus(?x:Nat,?z:Nat)) = s(?x:Nat)",
"plus(times(s(times(?x:Nat,?y:Nat)),s(?y:Nat)),s(plus(s(?x:Nat),s(?y:Nat)))) = s(?x:Nat)",
"times(s(?y:Nat),times(times(plus(?x:Nat,?y:Nat),plus(?z:Nat,?x:Nat)),s(s(?x:Nat)))) = s(?x:Nat)",
"times(s(plus(?z:Nat,times(?y:Nat,?z:Nat))),plus(s(?x:Nat),s(plus(?z:Nat,?y:Nat)))) = s(?x:Nat)",
"times(plus(?y:Nat,times(?x:Nat,times(?x:Nat,?x:Nat))),plus(times(s(?x:Nat),?z:Nat),?z:Nat)) = s(?x:Nat)",
"times(plus(s(s(?z:Nat)),plus(times(?x:Nat,?y:Nat),plus(?y:Nat,?y:Nat))),s(?z:Nat)) = s(?z:Nat)",
"s(?x:Nat) = times(s(times(?x:Nat,?z:Nat)),times(s(times(?z:Nat,?z:Nat)),plus(?y:Nat,times(?z:Nat,?y:Nat))))",
"s(?x:Nat) = times(times(plus(?y:Nat,times(?z:Nat,?y:Nat)),times(times(?y:Nat,?x:Nat),s(?z:Nat))),s(?x:Nat))",
"s(?y:Nat) = plus(times(?x:Nat,plus(plus(?y:Nat,?z:Nat),?z:Nat)),s(times(plus(?x:Nat,?x:Nat),s(?z:Nat))))",
"s(?y:Nat) = plus(times(plus(s(?y:Nat),?x:Nat),times(times(?y:Nat,?y:Nat),times(?z:Nat,?x:Nat))),s(?z:Nat))",
"s(?y:Nat) = times(times(s(s(?x:Nat)),s(?y:Nat)),s(plus(plus(?z:Nat,?x:Nat),times(?y:Nat,?x:Nat))))",
"s(?y:Nat) = times(times(s(plus(?z:Nat,?z:Nat)),plus(s(?y:Nat),plus(?z:Nat,?z:Nat))),plus(?x:Nat,?y:Nat))",
"s(?z:Nat) = times(plus(?y:Nat,plus(s(?y:Nat),s(?x:Nat))),s(plus(plus(?x:Nat,?y:Nat),s(?z:Nat))))",
"s(?z:Nat) = times(plus(times(plus(?z:Nat,?y:Nat),plus(?x:Nat,?x:Nat)),plus(plus(?z:Nat,?y:Nat),?z:Nat)),?z:Nat)",
"s(s(?y:Nat)) = plus(?z:Nat,times(times(times(?y:Nat,?z:Nat),times(?x:Nat,?y:Nat)),s(times(?y:Nat,?x:Nat))))",
"s(s(?y:Nat)) = times(s(s(?z:Nat)),plus(times(?z:Nat,times(?x:Nat,?z:Nat)),s(plus(?z:Nat,?x:Nat))))",
"plus(?z:Nat,?y:Nat) = plus(plus(?z:Nat,plus(times(?x:Nat,?z:Nat),s(?z:Nat))),s(plus(s(?x:Nat),?x:Nat)))",
"plus(?z:Nat,?y:Nat) = times(plus(plus(?x:Nat,s(?y:Nat)),times(?x:Nat,plus(?z:Nat,?z:Nat))),plus(?x:Nat,?y:Nat))",
"times(?y:Nat,?y:Nat) = plus(s(plus(times(?x:Nat,?x:Nat),plus(?y:Nat,?z:Nat))),times(?x:Nat,plus(?y:Nat,?z:Nat)))",
"times(?z:Nat,?y:Nat) = plus(times(s(times(?x:Nat,?z:Nat)),plus(times(?z:Nat,?x:Nat),times(?y:Nat,?x:Nat))),?x:Nat)",
"s(s(s(?y:Nat))) = times(?x:Nat,times(times(?z:Nat,times(?y:Nat,?y:Nat)),plus(plus(?y:Nat,?x:Nat),?y:Nat)))",
"s(plus(?x:Nat,?x:Nat)) = times(s(plus(?x:Nat,?x:Nat)),plus(s(?x:Nat),plus(?z:Nat,plus(?z:Nat,?z:Nat))))",
"plus(s(?y:Nat),?x:Nat) = plus(plus(times(s(?x:Nat),?y:Nat),?y:Nat),times(times(s(?y:Nat),?x:Nat),?y:Nat))",
"plus(times(?y:Nat,?y:Nat),?z:Nat) = s(plus(s(plus(?x:Nat,?x:Nat)),times(times(?z:Nat,?z:Nat),s(?x:Nat))))",
"times(s(?y:Nat),s(?y:Nat)) = plus(s(s(times(?z:Nat,?y:Nat))),times(s(times(?y:Nat,?y:Nat)),?y:Nat))",
"times(times(?z:Nat,?z:Nat),?x:Nat) = plus(?y:Nat,times(times(s(?z:Nat),times(?x:Nat,?z:Nat)),s(s(?y:Nat))))",
"s(s(s(plus(?x:Nat,?y:Nat)))) = plus(?x:Nat,plus(plus(times(?y:Nat,?z:Nat),?x:Nat),plus(?x:Nat,?y:Nat)))",
"s(s(s(times(?x:Nat,?x:Nat)))) = plus(s(s(s(?y:Nat))),times(?z:Nat,s(plus(?y:Nat,?x:Nat))))",
"s(times(plus(?z:Nat,?y:Nat),?x:Nat)) = plus(plus(plus(s(?z:Nat),times(?y:Nat,?z:Nat)),s(?x:Nat)),?x:Nat)",
"plus(s(?x:Nat),plus(?z:Nat,?z:Nat)) = plus(times(s(s(?x:Nat)),plus(?z:Nat,times(?y:Nat,?y:Nat))),?z:Nat)",
"plus(s(plus(?x:Nat,?x:Nat)),?y:Nat) = times(times(?y:Nat,times(times(?x:Nat,?z:Nat),plus(?x:Nat,?x:Nat))),?x:Nat)",
"times(s(?x:Nat),times(?x:Nat,?z:Nat)) = plus(plus(s(plus(?x:Nat,?z:Nat)),times(?x:Nat,s(?z:Nat))),?x:Nat)",
"times(s(s(s(?y:Nat))),?x:Nat) = plus(plus(?y:Nat,times(?x:Nat,plus(?x:Nat,?x:Nat))),times(?y:Nat,?x:Nat))",
"times(plus(?y:Nat,?x:Nat),s(?x:Nat)) = plus(plus(s(?x:Nat),plus(s(?z:Nat),plus(?z:Nat,?x:Nat))),?y:Nat)",
"times(times(?x:Nat,?x:Nat),s(?z:Nat)) = times(?y:Nat,plus(?z:Nat,times(times(?x:Nat,?x:Nat),times(?z:Nat,?x:Nat))))",
"s(plus(s(s(?z:Nat)),s(?y:Nat))) = times(times(?y:Nat,?y:Nat),s(plus(?z:Nat,times(?y:Nat,?y:Nat))))",
"s(times(?x:Nat,s(plus(?x:Nat,?x:Nat)))) = times(times(?y:Nat,?z:Nat),plus(?z:Nat,times(?z:Nat,s(?z:Nat))))",
"s(times(?y:Nat,times(?z:Nat,s(?y:Nat)))) = times(?y:Nat,times(?x:Nat,times(plus(?z:Nat,?x:Nat),s(?z:Nat))))",
"plus(plus(?x:Nat,s(?x:Nat)),s(?x:Nat)) = s(plus(times(times(?x:Nat,?x:Nat),times(?z:Nat,?x:Nat)),?y:Nat))",
"times(?x:Nat,s(s(plus(?z:Nat,?x:Nat)))) = plus(s(times(?y:Nat,plus(?x:Nat,?z:Nat))),times(?z:Nat,?z:Nat))",
"times(?z:Nat,s(plus(s(?z:Nat),?z:Nat))) = s(plus(plus(plus(?y:Nat,?z:Nat),plus(?x:Nat,?y:Nat)),?x:Nat))",
"times(?z:Nat,times(?x:Nat,s(s(?x:Nat)))) = s(times(?y:Nat,times(times(?x:Nat,?y:Nat),plus(?x:Nat,?z:Nat))))",
"times(s(times(?x:Nat,s(?z:Nat))),?x:Nat) = plus(times(plus(s(?y:Nat),s(?x:Nat)),?x:Nat),s(?y:Nat))",
"times(plus(?y:Nat,s(s(?x:Nat))),?y:Nat) = times(times(?y:Nat,times(?x:Nat,times(?y:Nat,?z:Nat))),s(?z:Nat))",
"s(times(times(?x:Nat,times(?x:Nat,?y:Nat)),?z:Nat)) = plus(s(times(times(?x:Nat,?y:Nat),s(?y:Nat))),?z:Nat)",
"plus(s(?x:Nat),s(times(s(?x:Nat),?z:Nat))) = times(s(?x:Nat),s(plus(?y:Nat,plus(?x:Nat,?x:Nat))))",
"plus(plus(?x:Nat,s(?x:Nat)),plus(?z:Nat,?y:Nat)) = times(?x:Nat,plus(times(times(?y:Nat,?y:Nat),?z:Nat),?x:Nat))",
"plus(times(s(times(?y:Nat,?z:Nat)),?x:Nat),?y:Nat) = s(times(times(plus(?x:Nat,?z:Nat),s(?z:Nat)),?x:Nat))",
"times(?x:Nat,s(times(times(?z:Nat,?x:Nat),?z:Nat))) = times(?x:Nat,times(plus(plus(?z:Nat,?x:Nat),?y:Nat),?x:Nat))",
"times(times(times(?z:Nat,s(?x:Nat)),?y:Nat),?y:Nat) = s(s(times(times(?z:Nat,?y:Nat),plus(?z:Nat,?y:Nat))))",
"s(s(plus(times(?x:Nat,?x:Nat),plus(?y:Nat,?x:Nat)))) = times(?x:Nat,s(times(plus(?z:Nat,?z:Nat),?z:Nat)))",
"s(plus(times(?z:Nat,s(?y:Nat)),times(?y:Nat,?y:Nat))) = times(s(plus(plus(?z:Nat,?y:Nat),?y:Nat)),?z:Nat)",
"s(times(?z:Nat,times(times(?z:Nat,?x:Nat),s(?x:Nat)))) = times(plus(plus(s(?z:Nat),?y:Nat),?y:Nat),?z:Nat)",
"plus(s(plus(plus(?z:Nat,?y:Nat),s(?x:Nat))),?z:Nat) = times(?z:Nat,s(plus(s(?z:Nat),s(?y:Nat))))",
"plus(times(plus(?z:Nat,plus(?z:Nat,?z:Nat)),?z:Nat),?y:Nat) = times(s(plus(?x:Nat,times(?x:Nat,?x:Nat))),?z:Nat)",
"times(?x:Nat,s(times(times(?x:Nat,?x:Nat),s(?y:Nat)))) = times(times(?x:Nat,s(plus(?y:Nat,?z:Nat))),?x:Nat)",
"times(s(times(s(?x:Nat),times(?y:Nat,?z:Nat))),?x:Nat) = times(?y:Nat,times(s(times(?y:Nat,?x:Nat)),?x:Nat))",
"times(plus(times(plus(?z:Nat,?x:Nat),?y:Nat),?y:Nat),?y:Nat) = s(s(times(times(?z:Nat,?z:Nat),s(?x:Nat))))",
"s(times(?z:Nat,plus(times(?y:Nat,?y:Nat),times(?z:Nat,?y:Nat)))) = times(plus(times(?x:Nat,?y:Nat),?y:Nat),?z:Nat)",
"s(times(s(s(?z:Nat)),times(plus(?x:Nat,?x:Nat),?x:Nat))) = plus(?x:Nat,s(s(plus(?z:Nat,?z:Nat))))",
"plus(?x:Nat,plus(plus(?z:Nat,s(?x:Nat)),plus(?z:Nat,?z:Nat))) = s(times(times(?x:Nat,?x:Nat),s(?y:Nat)))",
"plus(?x:Nat,plus(times(times(?z:Nat,?x:Nat),s(?z:Nat)),?z:Nat)) = s(times(s(times(?z:Nat,?y:Nat)),?y:Nat))",
"plus(?z:Nat,times(plus(times(?x:Nat,?x:Nat),s(?y:Nat)),?y:Nat)) = s(plus(s(plus(?y:Nat,?z:Nat)),?x:Nat))",
"plus(s(times(plus(?x:Nat,?y:Nat),times(?y:Nat,?y:Nat))),?x:Nat) = s(plus(?x:Nat,s(plus(?y:Nat,?x:Nat))))",
"times(?x:Nat,s(times(plus(?z:Nat,?z:Nat),times(?z:Nat,?z:Nat)))) = plus(?z:Nat,plus(s(s(?y:Nat)),?y:Nat))",
"times(s(times(?y:Nat,?z:Nat)),s(s(plus(?y:Nat,?x:Nat)))) = plus(plus(?z:Nat,?z:Nat),times(?z:Nat,?y:Nat))",
"times(plus(?z:Nat,s(times(?y:Nat,?z:Nat))),plus(?x:Nat,?x:Nat)) = times(?y:Nat,times(?x:Nat,s(s(?y:Nat))))",
"times(plus(?z:Nat,times(plus(?z:Nat,?y:Nat),s(?z:Nat))),?z:Nat) = times(times(?z:Nat,times(?z:Nat,?y:Nat)),?y:Nat)",
"times(plus(s(plus(?x:Nat,?y:Nat)),s(?z:Nat)),s(?y:Nat)) = plus(plus(times(?y:Nat,?z:Nat),?x:Nat),?y:Nat)",
"times(plus(s(plus(?z:Nat,?z:Nat)),?z:Nat),plus(?x:Nat,?z:Nat)) = s(times(?z:Nat,s(plus(?y:Nat,?z:Nat))))",
"plus(?x:Nat,times(?y:Nat,plus(times(?y:Nat,?y:Nat),plus(?y:Nat,?x:Nat)))) = plus(?z:Nat,plus(s(?y:Nat),?y:Nat))",
"plus(?z:Nat,plus(?x:Nat,times(plus(?z:Nat,?x:Nat),plus(?z:Nat,?x:Nat)))) = s(s(s(plus(?y:Nat,?y:Nat))))",
"plus(times(s(s(?y:Nat)),times(plus(?x:Nat,?z:Nat),?x:Nat)),?x:Nat) = s(times(times(?y:Nat,?y:Nat),?x:Nat))",
"times(?y:Nat,times(s(s(?z:Nat)),times(?x:Nat,plus(?x:Nat,?x:Nat)))) = s(s(s(times(?z:Nat,?z:Nat))))",
"times(s(s(?z:Nat)),times(times(times(?x:Nat,?z:Nat),?z:Nat),?x:Nat)) = times(plus(?y:Nat,s(?y:Nat)),?z:Nat)",
"times(plus(times(times(?y:Nat,?x:Nat),?y:Nat),times(?x:Nat,?y:Nat)),?z:Nat) = times(s(s(s(?x:Nat))),?z:Nat)",
"times(times(s(?x:Nat),plus(times(?x:Nat,?x:Nat),s(?y:Nat))),?y:Nat) = s(s(times(?z:Nat,s(?z:Nat))))",
"times(times(s(plus(?x:Nat,?z:Nat)),s(plus(?x:Nat,?x:Nat))),?y:Nat) = s(s(s(plus(?z:Nat,?x:Nat))))",
"s(times(times(plus(?x:Nat,?z:Nat),times(?x:Nat,?z:Nat)),times(?x:Nat,?x:Nat))) = times(plus(?z:Nat,?z:Nat),?y:Nat)",
"plus(plus(plus(times(?y:Nat,?y:Nat),times(?y:Nat,?x:Nat)),?y:Nat),s(?y:Nat)) = times(times(?z:Nat,?z:Nat),?z:Nat)",
"plus(times(plus(s(?y:Nat),s(?x:Nat)),plus(?z:Nat,s(?y:Nat))),?y:Nat) = s(s(s(s(?x:Nat))))",
"times(times(times(?z:Nat,?y:Nat),times(?z:Nat,times(?z:Nat,?z:Nat))),s(?x:Nat)) = times(s(s(?z:Nat)),?y:Nat)",
"plus(s(?z:Nat),times(times(?y:Nat,times(?y:Nat,?y:Nat)),times(s(?y:Nat),?z:Nat))) = s(times(?x:Nat,?x:Nat))",
"plus(s(plus(times(?y:Nat,?y:Nat),times(?z:Nat,?y:Nat))),plus(s(?z:Nat),?y:Nat)) = s(plus(?x:Nat,?y:Nat))",
"plus(times(s(s(?y:Nat)),plus(s(?x:Nat),times(?x:Nat,?x:Nat))),s(?x:Nat)) = plus(?x:Nat,s(?x:Nat))",
"times(s(times(s(?z:Nat),times(?z:Nat,?y:Nat))),s(s(plus(?z:Nat,?z:Nat)))) = times(s(?x:Nat),?y:Nat)",
"times(plus(s(s(?y:Nat)),times(plus(?z:Nat,?x:Nat),plus(?z:Nat,?x:Nat))),?z:Nat) = s(s(s(?x:Nat)))",
"plus(plus(times(plus(?y:Nat,?x:Nat),s(?y:Nat)),times(plus(?x:Nat,?z:Nat),?y:Nat)),?z:Nat) = s(s(?x:Nat))",
"plus(times(s(s(?z:Nat)),plus(?z:Nat,times(?z:Nat,?z:Nat))),s(s(s(?z:Nat)))) = times(?y:Nat,?z:Nat)",
"plus(times(plus(times(?y:Nat,?x:Nat),?x:Nat),times(?x:Nat,plus(?z:Nat,?y:Nat))),s(?y:Nat)) = s(s(?y:Nat))",
"times(?x:Nat,plus(plus(s(?x:Nat),plus(?x:Nat,?z:Nat)),times(s(?y:Nat),s(?x:Nat)))) = plus(?x:Nat,?z:Nat)",
"times(?z:Nat,plus(plus(s(?x:Nat),?z:Nat),plus(times(?z:Nat,?x:Nat),times(?z:Nat,?x:Nat)))) = plus(?z:Nat,?x:Nat)",
"times(plus(?x:Nat,plus(s(?y:Nat),?x:Nat)),times(?y:Nat,plus(times(?z:Nat,?z:Nat),?z:Nat))) = s(s(?y:Nat))",
"plus(plus(s(?x:Nat),?z:Nat),plus(plus(plus(?y:Nat,?x:Nat),times(?z:Nat,?x:Nat)),s(?z:Nat))) = s(?x:Nat)",
"plus(times(?y:Nat,times(plus(?x:Nat,?x:Nat),s(?y:Nat))),times(s(times(?y:Nat,?x:Nat)),?z:Nat)) = s(?z:Nat)",
"plus(times(s(plus(?x:Nat,?z:Nat)),?z:Nat),times(s(plus(?x:Nat,?z:Nat)),times(?y:Nat,?z:Nat))) = s(?x:Nat)",
"times(s(plus(?x:Nat,plus(?y:Nat,?y:Nat))),s(times(plus(?z:Nat,?x:Nat),times(?x:Nat,?x:Nat)))) = s(?y:Nat)",
"times(plus(?x:Nat,times(times(?z:Nat,?x:Nat),plus(?y:Nat,?z:Nat))),times(s(s(?y:Nat)),?y:Nat)) = s(?z:Nat)",
"times(plus(s(s(?x:Nat)),plus(plus(?y:Nat,?y:Nat),times(?y:Nat,?y:Nat))),plus(?y:Nat,?z:Nat)) = s(?z:Nat)",
"times(plus(s(times(?z:Nat,?y:Nat)),s(?y:Nat)),s(plus(times(?z:Nat,?x:Nat),s(?y:Nat)))) = s(?x:Nat)",
"times(plus(plus(times(?y:Nat,?x:Nat),s(?z:Nat)),plus(s(?y:Nat),times(?z:Nat,?y:Nat))),?x:Nat) = s(?y:Nat)",
"times(plus(plus(times(?y:Nat,?x:Nat),plus(?x:Nat,?y:Nat)),plus(s(?z:Nat),s(?x:Nat))),?x:Nat) = s(?x:Nat)",
"s(?x:Nat) = plus(s(plus(times(?z:Nat,?y:Nat),times(?x:Nat,?y:Nat))),plus(times(?x:Nat,?x:Nat),s(s(?x:Nat))))",
"s(?x:Nat) = plus(plus(plus(s(?y:Nat),plus(?y:Nat,?z:Nat)),times(times(?z:Nat,?y:Nat),times(?y:Nat,?y:Nat))),?x:Nat)",
"s(?x:Nat) = times(plus(plus(plus(?y:Nat,?z:Nat),s(?x:Nat)),plus(plus(?y:Nat,?y:Nat),times(?y:Nat,?z:Nat))),?x:Nat)",
"s(?z:Nat) = plus(plus(?z:Nat,?z:Nat),plus(times(plus(?z:Nat,?z:Nat),plus(?y:Nat,?y:Nat)),times(?z:Nat,s(?x:Nat))))",
"s(?z:Nat) = plus(plus(plus(?y:Nat,plus(?y:Nat,?y:Nat)),?y:Nat),times(times(?y:Nat,?y:Nat),s(times(?z:Nat,?y:Nat))))",
"s(?z:Nat) = plus(times(s(plus(?y:Nat,?z:Nat)),times(?x:Nat,times(?x:Nat,?z:Nat))),s(plus(?z:Nat,s(?z:Nat))))",
"plus(?x:Nat,?y:Nat) = plus(times(s(plus(?z:Nat,?x:Nat)),times(s(?y:Nat),times(?z:Nat,?x:Nat))),plus(?z:Nat,?y:Nat))",
"plus(?x:Nat,?z:Nat) = times(plus(s(s(?y:Nat)),?z:Nat),plus(plus(?x:Nat,s(?y:Nat)),s(plus(?z:Nat,?z:Nat))))",
"plus(?y:Nat,?z:Nat) = s(plus(plus(plus(?y:Nat,?x:Nat),times(?y:Nat,?y:Nat)),plus(s(?z:Nat),plus(?x:Nat,?z:Nat))))",
"plus(?y:Nat,?z:Nat) = plus(plus(s(times(?x:Nat,?y:Nat)),plus(?x:Nat,s(?z:Nat))),s(times(s(?x:Nat),?y:Nat)))",
"plus(?y:Nat,?z:Nat) = plus(times(times(times(?y:Nat,?y:Nat),s(?y:Nat)),plus(times(?x:Nat,?x:Nat),s(?z:Nat))),?x:Nat)",
"times(?y:Nat,?x:Nat) = plus(plus(s(times(?x:Nat,?y:Nat)),?x:Nat),plus(times(?y:Nat,plus(?z:Nat,?y:Nat)),s(?z:Nat)))",
"s(plus(?x:Nat,?y:Nat)) = times(s(?y:Nat),plus(plus(?x:Nat,?x:Nat),times(plus(?x:Nat,?x:Nat),plus(?z:Nat,?y:Nat))))",
"s(times(?y:Nat,?y:Nat)) = plus(times(plus(plus(?x:Nat,?x:Nat),s(?x:Nat)),?z:Nat),times(s(?y:Nat),s(?x:Nat)))",
"plus(s(?y:Nat),?x:Nat) = plus(?y:Nat,times(s(plus(?z:Nat,?x:Nat)),plus(times(?y:Nat,?x:Nat),plus(?z:Nat,?y:Nat))))",
"times(s(?z:Nat),?y:Nat) = plus(?x:Nat,plus(times(?x:Nat,plus(?x:Nat,?y:Nat)),plus(s(?y:Nat),plus(?x:Nat,?z:Nat))))",
"s(s(plus(?y:Nat,?z:Nat))) = times(plus(plus(?y:Nat,s(?x:Nat)),plus(plus(?y:Nat,?z:Nat),?z:Nat)),s(?y:Nat))",
"s(plus(s(?x:Nat),?x:Nat)) = plus(plus(?z:Nat,?y:Nat),times(?x:Nat,times(times(?x:Nat,?z:Nat),times(?y:Nat,?z:Nat))))",
"times(?x:Nat,times(?y:Nat,?z:Nat)) = plus(times(s(?y:Nat),plus(?z:Nat,s(?y:Nat))),plus(?x:Nat,s(s(?x:Nat))))",
"s(s(s(times(?z:Nat,?y:Nat)))) = times(plus(s(times(?z:Nat,?x:Nat)),plus(?x:Nat,times(?z:Nat,?x:Nat))),?y:Nat)",
"s(s(s(times(?z:Nat,?z:Nat)))) = plus(?y:Nat,times(times(plus(?z:Nat,?y:Nat),plus(?y:Nat,?z:Nat)),s(?x:Nat)))",
"s(s(plus(s(?x:Nat),?x:Nat))) = plus(s(s(?z:Nat)),times(?z:Nat,times(s(?z:Nat),plus(?z:Nat,?z:Nat))))",
"s(times(?z:Nat,s(s(?y:Nat)))) = plus(times(times(?z:Nat,s(?y:Nat)),?x:Nat),s(times(?x:Nat,s(?z:Nat))))",
"plus(s(s(s(?y:Nat))),?z:Nat) = times(times(times(?y:Nat,?z:Nat),times(plus(?x:Nat,?y:Nat),s(?y:Nat))),?y:Nat)",
"plus(plus(?x:Nat,s(?x:Nat)),?z:Nat) = s(plus(times(plus(?z:Nat,?z:Nat),?x:Nat),times(times(?z:Nat,?x:Nat),?x:Nat)))",
"times(s(s(?x:Nat)),s(?y:Nat)) = plus(?x:Nat,plus(plus(?x:Nat,s(?z:Nat)),plus(?z:Nat,plus(?x:Nat,?x:Nat))))",
"times(s(plus(?z:Nat,?y:Nat)),?x:Nat) = plus(s(s(s(?z:Nat))),times(s(times(?x:Nat,?y:Nat)),s(?x:Nat)))",
"s(plus(s(?x:Nat),s(s(?y:Nat)))) = times(s(?z:Nat),times(s(?y:Nat),plus(?x:Nat,times(?x:Nat,?x:Nat))))",
"s(plus(s(times(?z:Nat,?x:Nat)),?z:Nat)) = times(plus(s(s(?y:Nat)),times(?y:Nat,times(?y:Nat,?x:Nat))),?x:Nat)",
"s(times(?z:Nat,plus(?x:Nat,s(?y:Nat)))) = plus(plus(?x:Nat,times(plus(?x:Nat,?y:Nat),times(?x:Nat,?y:Nat))),?y:Nat)",
"plus(?z:Nat,plus(s(s(?z:Nat)),?x:Nat)) = plus(plus(?x:Nat,s(times(?x:Nat,?y:Nat))),s(times(?z:Nat,?y:Nat)))",
"plus(times(s(s(?y:Nat)),?y:Nat),?x:Nat) = plus(?x:Nat,times(plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?y:Nat)),?y:Nat))",
"times(?x:Nat,times(?z:Nat,s(s(?y:Nat)))) = times(plus(?x:Nat,?y:Nat),plus(?x:Nat,plus(?z:Nat,times(?y:Nat,?z:Nat))))",
"times(times(?x:Nat,?x:Nat),plus(?z:Nat,?y:Nat)) = s(plus(s(?x:Nat),plus(plus(?x:Nat,?x:Nat),times(?z:Nat,?y:Nat))))",
"times(times(s(s(?y:Nat)),?z:Nat),?z:Nat) = times(plus(times(s(?y:Nat),times(?z:Nat,?z:Nat)),s(?x:Nat)),?z:Nat)",
"s(s(plus(s(?y:Nat),times(?x:Nat,?x:Nat)))) = times(s(plus(plus(?z:Nat,?z:Nat),s(?z:Nat))),s(?z:Nat))",
"s(plus(?y:Nat,plus(s(?x:Nat),s(?z:Nat)))) = plus(?x:Nat,s(plus(times(?z:Nat,?z:Nat),plus(?x:Nat,?y:Nat))))",
"plus(?x:Nat,s(times(?y:Nat,plus(?y:Nat,?x:Nat)))) = times(times(times(times(?y:Nat,?z:Nat),s(?y:Nat)),?x:Nat),?y:Nat)",
"plus(?y:Nat,s(times(plus(?x:Nat,?z:Nat),?y:Nat))) = s(times(s(plus(?x:Nat,?x:Nat)),plus(s(?z:Nat),?x:Nat)))",
"plus(?y:Nat,plus(s(plus(?y:Nat,?x:Nat)),?y:Nat)) = plus(s(plus(times(?x:Nat,?z:Nat),?z:Nat)),plus(?z:Nat,?y:Nat))",
"plus(plus(?y:Nat,s(times(?z:Nat,?y:Nat))),?z:Nat) = plus(?x:Nat,s(times(plus(?y:Nat,?x:Nat),times(?y:Nat,?y:Nat))))",
"plus(plus(s(times(?y:Nat,?x:Nat)),?z:Nat),?x:Nat) = s(plus(?y:Nat,times(times(?x:Nat,?z:Nat),times(?x:Nat,?x:Nat))))",
"plus(times(s(plus(?x:Nat,?z:Nat)),?y:Nat),?y:Nat) = times(?z:Nat,plus(?x:Nat,plus(plus(?z:Nat,?z:Nat),s(?z:Nat))))",
"times(?y:Nat,times(?z:Nat,s(times(?x:Nat,?y:Nat)))) = times(s(?x:Nat),s(times(s(?x:Nat),plus(?z:Nat,?z:Nat))))",
"times(s(s(?y:Nat)),plus(?x:Nat,s(?z:Nat))) = times(s(plus(plus(?y:Nat,?x:Nat),s(?y:Nat))),s(?y:Nat))",
"s(plus(?y:Nat,plus(s(?y:Nat),plus(?x:Nat,?x:Nat)))) = plus(times(s(?x:Nat),s(plus(?x:Nat,?z:Nat))),?x:Nat)",
"plus(?x:Nat,plus(?x:Nat,times(?z:Nat,plus(?y:Nat,?z:Nat)))) = plus(s(?x:Nat),times(?y:Nat,times(?z:Nat,s(?x:Nat))))",
"plus(?y:Nat,times(plus(?x:Nat,times(?x:Nat,?x:Nat)),?z:Nat)) = plus(?y:Nat,s(plus(s(?z:Nat),plus(?x:Nat,?x:Nat))))",
"plus(plus(plus(s(?z:Nat),?z:Nat),s(?x:Nat)),?y:Nat) = plus(plus(times(?z:Nat,times(?y:Nat,?y:Nat)),?y:Nat),?y:Nat)",
"plus(times(?y:Nat,s(plus(?z:Nat,?z:Nat))),s(?x:Nat)) = plus(s(plus(?y:Nat,times(?x:Nat,?x:Nat))),s(?z:Nat))",
"times(?x:Nat,times(s(s(?x:Nat)),s(s(?x:Nat)))) = plus(?y:Nat,s(times(s(?z:Nat),times(?z:Nat,?x:Nat))))",
"times(?z:Nat,plus(times(times(?x:Nat,?x:Nat),?x:Nat),?y:Nat)) = times(times(s(s(?x:Nat)),plus(?z:Nat,?z:Nat)),?x:Nat)",
"times(s(plus(?x:Nat,?y:Nat)),times(?x:Nat,s(?y:Nat))) = s(times(s(s(?y:Nat)),plus(?z:Nat,s(?x:Nat))))",
"times(s(times(plus(?x:Nat,?y:Nat),s(?y:Nat))),?x:Nat) = plus(plus(s(plus(?x:Nat,?x:Nat)),s(?z:Nat)),?y:Nat)",
"times(s(times(plus(?y:Nat,?x:Nat),s(?y:Nat))),?x:Nat) = plus(s(plus(?x:Nat,s(?z:Nat))),times(?y:Nat,?z:Nat))",
"times(plus(?z:Nat,times(plus(?x:Nat,?z:Nat),?z:Nat)),?z:Nat) = s(times(?y:Nat,times(s(?z:Nat),times(?z:Nat,?y:Nat))))",
"times(times(?x:Nat,?x:Nat),s(s(times(?z:Nat,?z:Nat)))) = s(times(?x:Nat,times(times(?z:Nat,?z:Nat),s(?z:Nat))))",
"s(plus(plus(plus(?x:Nat,?y:Nat),plus(?x:Nat,?y:Nat)),?x:Nat)) = times(s(times(plus(?x:Nat,?y:Nat),?x:Nat)),?z:Nat)",
"s(plus(times(times(?y:Nat,?x:Nat),times(?x:Nat,?z:Nat)),?z:Nat)) = times(s(?y:Nat),times(s(s(?y:Nat)),?x:Nat))",
"s(times(plus(s(?z:Nat),s(?x:Nat)),s(s(?x:Nat)))) = times(s(plus(times(?z:Nat,?x:Nat),?y:Nat)),?x:Nat)",
"s(times(times(?z:Nat,s(?x:Nat)),s(times(?z:Nat,?x:Nat)))) = times(times(plus(s(?z:Nat),?y:Nat),?x:Nat),?x:Nat)",
"plus(?x:Nat,times(s(plus(?y:Nat,?z:Nat)),times(?y:Nat,?z:Nat))) = plus(plus(?x:Nat,s(plus(?x:Nat,?z:Nat))),?z:Nat)",
"plus(?z:Nat,plus(?y:Nat,plus(times(?z:Nat,?x:Nat),s(?z:Nat)))) = s(times(s(times(?z:Nat,?x:Nat)),s(?x:Nat)))",
"plus(plus(?x:Nat,s(plus(?x:Nat,?x:Nat))),s(s(?x:Nat))) = plus(?x:Nat,s(times(?z:Nat,plus(?y:Nat,?x:Nat))))",
"plus(times(?x:Nat,plus(s(?y:Nat),times(?y:Nat,?z:Nat))),?y:Nat) = times(?z:Nat,plus(plus(s(?y:Nat),?y:Nat),?z:Nat))",
"plus(times(?y:Nat,?y:Nat),plus(plus(s(?z:Nat),?y:Nat),?z:Nat)) = plus(plus(?z:Nat,s(times(?x:Nat,?z:Nat))),?x:Nat)",
"plus(times(s(?y:Nat),times(plus(?x:Nat,?y:Nat),?y:Nat)),?z:Nat) = plus(times(?x:Nat,s(plus(?x:Nat,?y:Nat))),?z:Nat)",
"times(?x:Nat,s(plus(times(?y:Nat,?z:Nat),times(?x:Nat,?z:Nat)))) = times(s(times(times(?z:Nat,?z:Nat),?y:Nat)),?x:Nat)",
"times(s(plus(?x:Nat,times(?x:Nat,?z:Nat))),plus(?y:Nat,?x:Nat)) = s(times(plus(plus(?y:Nat,?z:Nat),?y:Nat),?z:Nat))",
"times(s(plus(plus(?y:Nat,?y:Nat),?z:Nat)),times(?x:Nat,?y:Nat)) = times(plus(?z:Nat,s(times(?y:Nat,?z:Nat))),?y:Nat)",
"times(s(times(plus(?y:Nat,?y:Nat),plus(?x:Nat,?x:Nat))),?x:Nat) = plus(s(plus(times(?y:Nat,?z:Nat),?y:Nat)),?x:Nat)",
"times(plus(times(?x:Nat,?y:Nat),s(plus(?x:Nat,?x:Nat))),?y:Nat) = s(times(s(?z:Nat),s(times(?z:Nat,?y:Nat))))",
"s(plus(plus(?z:Nat,times(?z:Nat,?x:Nat)),times(s(?x:Nat),?x:Nat))) = times(?x:Nat,plus(s(s(?y:Nat)),?y:Nat))",
"s(times(times(times(?y:Nat,?z:Nat),?z:Nat),s(times(?x:Nat,?x:Nat)))) = plus(?y:Nat,s(s(times(?y:Nat,?x:Nat))))",
"plus(?y:Nat,times(plus(?x:Nat,plus(?z:Nat,?z:Nat)),times(?x:Nat,?y:Nat))) = s(s(times(s(?x:Nat),s(?y:Nat))))",
"times(times(?x:Nat,s(?x:Nat)),plus(s(plus(?z:Nat,?z:Nat)),?x:Nat)) = plus(?x:Nat,s(s(times(?x:Nat,?y:Nat))))",
"times(times(s(?x:Nat),?z:Nat),times(s(plus(?y:Nat,?z:Nat)),?x:Nat)) = s(times(times(?z:Nat,s(?z:Nat)),?y:Nat))",
"s(plus(plus(s(?y:Nat),times(?x:Nat,?x:Nat)),s(plus(?y:Nat,?z:Nat)))) = plus(times(?x:Nat,?y:Nat),s(?z:Nat))",
"plus(plus(?y:Nat,plus(?z:Nat,?x:Nat)),s(times(?z:Nat,times(?y:Nat,?z:Nat)))) = s(s(s(plus(?x:Nat,?z:Nat))))",
"times(?x:Nat,plus(times(times(?y:Nat,?x:Nat),s(?z:Nat)),s(s(?z:Nat)))) = plus(plus(?y:Nat,s(?x:Nat)),?z:Nat)",
"times(times(s(plus(?z:Nat,?x:Nat)),times(plus(?y:Nat,?y:Nat),?y:Nat)),?y:Nat) = times(?y:Nat,times(?z:Nat,s(?y:Nat)))",
"plus(plus(times(?y:Nat,times(?x:Nat,?z:Nat)),?z:Nat),s(s(times(?z:Nat,?x:Nat)))) = times(plus(?x:Nat,?y:Nat),?y:Nat)",
"plus(times(plus(plus(?z:Nat,?z:Nat),plus(?x:Nat,?z:Nat)),s(s(?x:Nat))),?y:Nat) = s(plus(?x:Nat,s(?z:Nat)))",
"times(s(times(plus(?x:Nat,?y:Nat),times(?x:Nat,?z:Nat))),plus(s(?y:Nat),?z:Nat)) = times(times(?z:Nat,?y:Nat),?z:Nat)",
"times(times(?z:Nat,s(s(?y:Nat))),plus(s(?y:Nat),s(times(?z:Nat,?y:Nat)))) = plus(times(?x:Nat,?y:Nat),?x:Nat)",
"times(times(times(s(?x:Nat),?z:Nat),plus(plus(?y:Nat,?z:Nat),s(?z:Nat))),?y:Nat) = plus(plus(?x:Nat,?y:Nat),?z:Nat)",
"plus(?x:Nat,times(plus(?z:Nat,times(?x:Nat,?z:Nat)),times(s(?z:Nat),times(?z:Nat,?x:Nat)))) = s(times(?x:Nat,?y:Nat))",
"times(times(times(times(?z:Nat,?z:Nat),s(?z:Nat)),s(plus(?x:Nat,?y:Nat))),s(?x:Nat)) = plus(?z:Nat,s(?x:Nat))",
"plus(times(times(?x:Nat,?z:Nat),plus(?z:Nat,plus(?z:Nat,?y:Nat))),times(?y:Nat,plus(?x:Nat,?z:Nat))) = times(?z:Nat,?y:Nat)",
"plus(times(times(times(?y:Nat,?y:Nat),?z:Nat),s(s(?z:Nat))),s(s(plus(?x:Nat,?x:Nat)))) = plus(?x:Nat,?z:Nat)",
"times(plus(times(times(?z:Nat,?z:Nat),plus(?y:Nat,?x:Nat)),plus(?z:Nat,times(?z:Nat,?x:Nat))),?y:Nat) = plus(?y:Nat,?z:Nat)",
"plus(s(times(plus(?x:Nat,?x:Nat),?y:Nat)),times(?x:Nat,plus(plus(?x:Nat,?y:Nat),times(?y:Nat,?y:Nat)))) = s(?x:Nat)",
"plus(plus(s(?x:Nat),times(?z:Nat,plus(?z:Nat,?y:Nat))),times(plus(?x:Nat,s(?z:Nat)),s(?x:Nat))) = s(?z:Nat)",
"plus(times(?y:Nat,?z:Nat),times(plus(plus(?z:Nat,?x:Nat),s(?x:Nat)),plus(s(?x:Nat),s(?x:Nat)))) = s(?x:Nat)",
"times(times(plus(times(?x:Nat,?y:Nat),?y:Nat),?y:Nat),s(times(times(?y:Nat,?y:Nat),times(?x:Nat,?z:Nat)))) = s(?z:Nat)",
"s(?x:Nat) = plus(plus(plus(?x:Nat,s(?y:Nat)),s(times(?y:Nat,?z:Nat))),plus(times(plus(?z:Nat,?y:Nat),?y:Nat),?y:Nat))",
"s(?z:Nat) = plus(plus(s(?y:Nat),s(times(?z:Nat,?y:Nat))),plus(times(?x:Nat,s(?z:Nat)),s(plus(?z:Nat,?z:Nat))))",
"s(s(?x:Nat)) = plus(s(plus(s(?y:Nat),plus(?z:Nat,?y:Nat))),s(times(times(?x:Nat,?x:Nat),plus(?z:Nat,?x:Nat))))",
"s(s(?x:Nat)) = plus(plus(?z:Nat,times(?x:Nat,?z:Nat)),times(plus(plus(?y:Nat,?x:Nat),?x:Nat),times(?z:Nat,s(?y:Nat))))",
"plus(?y:Nat,?x:Nat) = plus(plus(?z:Nat,times(s(?y:Nat),?y:Nat)),plus(?z:Nat,plus(times(?y:Nat,?y:Nat),times(?z:Nat,?y:Nat))))",
"times(?x:Nat,?z:Nat) = times(plus(s(s(?y:Nat)),plus(?z:Nat,plus(?y:Nat,?z:Nat))),plus(plus(?z:Nat,s(?y:Nat)),?y:Nat))",
"times(?z:Nat,?x:Nat) = plus(plus(times(?z:Nat,?z:Nat),?z:Nat),times(plus(s(?x:Nat),plus(?z:Nat,?z:Nat)),s(s(?z:Nat))))",
"times(?z:Nat,?z:Nat) = times(s(plus(?x:Nat,s(?y:Nat))),times(s(s(?x:Nat)),plus(s(?x:Nat),times(?y:Nat,?z:Nat))))",
"s(plus(?y:Nat,?y:Nat)) = times(times(s(s(?x:Nat)),s(?z:Nat)),s(times(plus(?y:Nat,?y:Nat),plus(?x:Nat,?y:Nat))))",
"plus(?x:Nat,s(?y:Nat)) = times(?z:Nat,plus(plus(s(?z:Nat),s(?x:Nat)),times(times(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat))))",
"plus(s(?y:Nat),?y:Nat) = plus(times(?x:Nat,?y:Nat),times(times(plus(?x:Nat,?y:Nat),s(?z:Nat)),plus(s(?y:Nat),?z:Nat)))",
"times(?y:Nat,s(?z:Nat)) = plus(plus(?x:Nat,plus(s(?x:Nat),times(?z:Nat,?y:Nat))),s(plus(times(?z:Nat,?z:Nat),?x:Nat)))",
"plus(s(s(?x:Nat)),s(?z:Nat)) = plus(plus(times(s(?z:Nat),?y:Nat),times(times(?x:Nat,?z:Nat),s(?y:Nat))),?x:Nat)",
"times(?x:Nat,s(plus(?y:Nat,?x:Nat))) = s(times(times(?y:Nat,plus(?x:Nat,?x:Nat)),plus(times(?y:Nat,?x:Nat),s(?x:Nat))))",
"times(s(s(?x:Nat)),s(?x:Nat)) = times(s(times(s(?y:Nat),?z:Nat)),s(plus(s(?y:Nat),times(?z:Nat,?x:Nat))))",
"times(s(s(s(?x:Nat))),?y:Nat) = plus(times(times(s(?x:Nat),times(?z:Nat,?z:Nat)),?y:Nat),plus(?y:Nat,s(?z:Nat)))",
"times(s(plus(?y:Nat,?x:Nat)),?z:Nat) = times(s(plus(times(?z:Nat,?y:Nat),s(?x:Nat))),s(plus(?z:Nat,s(?z:Nat))))",
"s(s(plus(times(?y:Nat,?z:Nat),?y:Nat))) = plus(s(s(s(?z:Nat))),times(s(?z:Nat),s(plus(?z:Nat,?x:Nat))))",
"s(plus(s(plus(?y:Nat,?z:Nat)),?x:Nat)) = plus(times(times(s(?z:Nat),s(?x:Nat)),plus(?y:Nat,?z:Nat)),s(?x:Nat))",
"s(plus(plus(?y:Nat,s(?y:Nat)),?y:Nat)) = plus(?y:Nat,plus(times(?z:Nat,s(?y:Nat)),plus(?y:Nat,plus(?z:Nat,?y:Nat))))",
"s(times(?y:Nat,s(times(?x:Nat,?y:Nat)))) = plus(s(plus(times(?z:Nat,?y:Nat),?x:Nat)),times(?y:Nat,s(s(?y:Nat))))",
"s(times(?z:Nat,s(times(?x:Nat,?z:Nat)))) = times(plus(?z:Nat,?y:Nat),times(s(s(?x:Nat)),s(times(?x:Nat,?x:Nat))))",
"s(times(times(?y:Nat,s(?z:Nat)),?y:Nat)) = times(plus(s(s(?y:Nat)),s(?z:Nat)),plus(s(s(?y:Nat)),?z:Nat))",
"plus(s(?x:Nat),times(?y:Nat,s(?x:Nat))) = s(plus(s(s(?y:Nat)),times(plus(?x:Nat,?y:Nat),times(?y:Nat,?x:Nat))))",
"times(?y:Nat,s(s(times(?x:Nat,?x:Nat)))) = s(times(plus(?y:Nat,plus(?x:Nat,?x:Nat)),times(?x:Nat,plus(?z:Nat,?x:Nat))))",
"times(plus(plus(?y:Nat,?x:Nat),?x:Nat),?x:Nat) = plus(?x:Nat,plus(times(?z:Nat,plus(?z:Nat,?x:Nat)),times(s(?x:Nat),?z:Nat)))",
"s(times(?x:Nat,times(times(?x:Nat,?y:Nat),?y:Nat))) = plus(times(plus(?x:Nat,?z:Nat),plus(s(?x:Nat),s(?x:Nat))),?y:Nat)",
"plus(?z:Nat,s(times(times(?y:Nat,?x:Nat),?x:Nat))) = s(times(s(?y:Nat),plus(plus(?z:Nat,?x:Nat),times(?x:Nat,?z:Nat))))",
"plus(plus(?x:Nat,s(times(?z:Nat,?z:Nat))),?z:Nat) = s(plus(times(?y:Nat,s(?z:Nat)),plus(?x:Nat,plus(?x:Nat,?z:Nat))))",
"plus(times(s(?y:Nat),?x:Nat),times(?z:Nat,?y:Nat)) = times(times(?z:Nat,?y:Nat),plus(plus(s(?z:Nat),?z:Nat),s(?y:Nat)))",
"s(plus(times(times(?z:Nat,?z:Nat),s(?y:Nat)),?z:Nat)) = plus(times(?x:Nat,?x:Nat),s(times(times(?z:Nat,?y:Nat),?z:Nat)))",
"plus(?z:Nat,times(?x:Nat,plus(s(?x:Nat),s(?z:Nat)))) = plus(times(s(times(?x:Nat,?x:Nat)),s(s(?y:Nat))),?x:Nat)",
"plus(s(times(s(?y:Nat),?y:Nat)),s(s(?y:Nat))) = times(times(s(?y:Nat),?y:Nat),s(s(plus(?y:Nat,?z:Nat))))",
"plus(plus(?z:Nat,?z:Nat),s(s(times(?x:Nat,?z:Nat)))) = s(plus(plus(?y:Nat,?x:Nat),plus(plus(?z:Nat,?x:Nat),?y:Nat)))",
"plus(plus(plus(plus(?y:Nat,?z:Nat),?x:Nat),?y:Nat),?y:Nat) = plus(plus(?y:Nat,s(?y:Nat)),times(times(?z:Nat,?x:Nat),?x:Nat))",
"plus(times(s(times(?x:Nat,?z:Nat)),?x:Nat),s(?y:Nat)) = plus(?x:Nat,s(plus(plus(?y:Nat,?y:Nat),plus(?x:Nat,?z:Nat))))",
"plus(times(times(times(?y:Nat,?x:Nat),?y:Nat),?x:Nat),?x:Nat) = times(plus(s(plus(?y:Nat,?y:Nat)),plus(?x:Nat,?z:Nat)),?z:Nat)",
"times(s(plus(s(?x:Nat),times(?y:Nat,?y:Nat))),?z:Nat) = plus(?x:Nat,times(times(plus(?z:Nat,?x:Nat),s(?y:Nat)),?x:Nat))",
"times(plus(s(plus(?y:Nat,?z:Nat)),?z:Nat),s(?z:Nat)) = plus(?x:Nat,plus(?x:Nat,times(times(?y:Nat,?z:Nat),s(?y:Nat))))",
"times(times(?y:Nat,plus(plus(?x:Nat,?y:Nat),?x:Nat)),?y:Nat) = times(?z:Nat,s(plus(plus(?z:Nat,?y:Nat),plus(?y:Nat,?z:Nat))))",
"s(plus(times(?z:Nat,?z:Nat),plus(s(?y:Nat),s(?z:Nat)))) = times(?x:Nat,s(times(plus(?y:Nat,?x:Nat),s(?y:Nat))))",
"plus(?z:Nat,plus(?z:Nat,plus(plus(?y:Nat,?x:Nat),s(?z:Nat)))) = plus(?x:Nat,times(?y:Nat,times(times(?y:Nat,?x:Nat),?x:Nat)))",
"plus(s(times(plus(?y:Nat,?x:Nat),times(?z:Nat,?x:Nat))),?z:Nat) = times(s(plus(times(?z:Nat,?x:Nat),s(?z:Nat))),?y:Nat)",
"plus(times(?z:Nat,times(s(?y:Nat),plus(?z:Nat,?z:Nat))),?z:Nat) = times(s(times(times(?x:Nat,?x:Nat),s(?z:Nat))),?z:Nat)",
"times(s(?y:Nat),s(times(times(?x:Nat,?y:Nat),s(?y:Nat)))) = plus(s(plus(plus(?z:Nat,?x:Nat),s(?x:Nat))),?x:Nat)",
"times(s(plus(?x:Nat,times(?x:Nat,?x:Nat))),s(s(?x:Nat))) = times(plus(?y:Nat,plus(?y:Nat,times(?z:Nat,?y:Nat))),?y:Nat)",
"times(s(plus(plus(?z:Nat,?y:Nat),times(?z:Nat,?y:Nat))),?y:Nat) = times(plus(plus(s(?z:Nat),?x:Nat),s(?y:Nat)),?z:Nat)",
"times(times(?y:Nat,plus(times(?y:Nat,?z:Nat),s(?z:Nat))),?x:Nat) = s(s(times(plus(?y:Nat,?y:Nat),plus(?y:Nat,?x:Nat))))",
"times(times(?z:Nat,times(times(?z:Nat,?x:Nat),s(?z:Nat))),?y:Nat) = plus(?x:Nat,s(plus(s(?z:Nat),times(?y:Nat,?y:Nat))))",
"times(times(times(?x:Nat,plus(?y:Nat,?y:Nat)),s(?y:Nat)),?z:Nat) = s(s(plus(plus(?z:Nat,?x:Nat),times(?y:Nat,?x:Nat))))",
"s(times(plus(?z:Nat,s(?z:Nat)),times(s(?y:Nat),s(?z:Nat)))) = times(plus(?y:Nat,plus(s(?x:Nat),?x:Nat)),?y:Nat)",
"s(times(times(times(?x:Nat,?z:Nat),?z:Nat),s(plus(?x:Nat,?x:Nat)))) = plus(?y:Nat,plus(plus(?y:Nat,s(?z:Nat)),?x:Nat))",
"plus(?x:Nat,plus(?y:Nat,times(times(?z:Nat,?y:Nat),times(?z:Nat,?z:Nat)))) = times(times(?y:Nat,?x:Nat),s(times(?y:Nat,?y:Nat)))",
"plus(plus(plus(?x:Nat,?x:Nat),?z:Nat),plus(?y:Nat,times(?y:Nat,?y:Nat))) = s(plus(plus(plus(?y:Nat,?y:Nat),?z:Nat),?x:Nat))",
"times(?z:Nat,plus(?x:Nat,times(times(?y:Nat,?z:Nat),times(?y:Nat,?x:Nat)))) = times(s(plus(?z:Nat,s(?z:Nat))),s(?y:Nat))",
"times(plus(?y:Nat,plus(plus(?z:Nat,?z:Nat),plus(?x:Nat,?y:Nat))),?z:Nat) = times(s(plus(times(?x:Nat,?x:Nat),?x:Nat)),?z:Nat)",
"times(plus(plus(plus(?x:Nat,?z:Nat),times(?x:Nat,?z:Nat)),?y:Nat),?z:Nat) = times(s(plus(plus(?x:Nat,?x:Nat),?x:Nat)),?y:Nat)",
"times(times(plus(?x:Nat,?x:Nat),times(?y:Nat,plus(?y:Nat,?z:Nat))),?x:Nat) = plus(plus(?z:Nat,?y:Nat),plus(?z:Nat,s(?y:Nat)))",
"plus(s(s(times(?z:Nat,?y:Nat))),plus(s(times(?y:Nat,?y:Nat)),?x:Nat)) = plus(?x:Nat,s(times(s(?y:Nat),?x:Nat)))",
"plus(s(plus(times(?z:Nat,?z:Nat),times(?x:Nat,?y:Nat))),s(s(?z:Nat))) = s(plus(?y:Nat,s(plus(?y:Nat,?y:Nat))))",
"times(?z:Nat,times(times(?y:Nat,?z:Nat),plus(times(?x:Nat,?y:Nat),s(?y:Nat)))) = times(?z:Nat,s(plus(?z:Nat,s(?x:Nat))))",
"times(plus(?x:Nat,s(plus(?z:Nat,?z:Nat))),s(s(plus(?z:Nat,?z:Nat)))) = s(s(plus(?z:Nat,times(?z:Nat,?z:Nat))))",
"times(plus(s(plus(?y:Nat,?x:Nat)),plus(?x:Nat,?y:Nat)),times(?z:Nat,?x:Nat)) = times(s(times(?x:Nat,s(?x:Nat))),?x:Nat)",
"times(times(times(plus(?z:Nat,?z:Nat),s(?z:Nat)),?y:Nat),times(?z:Nat,?x:Nat)) = plus(s(?x:Nat),s(plus(?z:Nat,?y:Nat)))",
"plus(s(times(?x:Nat,?x:Nat)),times(s(times(?x:Nat,?y:Nat)),s(s(?x:Nat)))) = s(s(plus(s(?y:Nat),?y:Nat)))",
"plus(times(?x:Nat,s(times(?y:Nat,?y:Nat))),s(times(?x:Nat,plus(?y:Nat,?x:Nat)))) = times(s(s(s(?y:Nat))),?y:Nat)",
"times(s(times(?y:Nat,?y:Nat)),plus(s(times(?y:Nat,?z:Nat)),s(s(?z:Nat)))) = times(plus(s(?z:Nat),?z:Nat),?y:Nat)",
"times(times(s(plus(?y:Nat,?x:Nat)),plus(?y:Nat,plus(?z:Nat,?y:Nat))),s(?z:Nat)) = s(times(s(s(?z:Nat)),?y:Nat))",
"s(plus(plus(?y:Nat,plus(?z:Nat,?y:Nat)),times(times(?y:Nat,?x:Nat),plus(?y:Nat,?z:Nat)))) = times(plus(?x:Nat,?x:Nat),?y:Nat)",
"plus(plus(?x:Nat,s(times(?y:Nat,?x:Nat))),plus(plus(plus(?y:Nat,?y:Nat),?x:Nat),?y:Nat)) = times(?x:Nat,plus(?z:Nat,?z:Nat))",
"times(s(plus(?y:Nat,s(?x:Nat))),times(?x:Nat,plus(s(?y:Nat),times(?z:Nat,?x:Nat)))) = times(s(s(?x:Nat)),?x:Nat)",
"times(times(s(s(?z:Nat)),times(plus(?y:Nat,?x:Nat),s(?y:Nat))),times(?z:Nat,?y:Nat)) = times(times(?y:Nat,?x:Nat),?y:Nat)",
"times(?z:Nat,times(times(?z:Nat,plus(?x:Nat,?z:Nat)),times(plus(?x:Nat,?y:Nat),times(?z:Nat,?z:Nat)))) = times(?y:Nat,s(?z:Nat))",
"s(plus(plus(plus(?y:Nat,?x:Nat),plus(?y:Nat,?z:Nat)),times(times(?z:Nat,?z:Nat),times(?z:Nat,?y:Nat)))) = times(?y:Nat,?z:Nat)",
"s(times(plus(times(?z:Nat,?x:Nat),times(?y:Nat,?x:Nat)),times(plus(?z:Nat,?z:Nat),times(?y:Nat,?z:Nat)))) = times(?y:Nat,?z:Nat)",
"plus(plus(s(?x:Nat),?y:Nat),times(s(s(?y:Nat)),plus(times(?y:Nat,?y:Nat),times(?z:Nat,?y:Nat)))) = plus(?z:Nat,?x:Nat)",
"plus(plus(plus(plus(?y:Nat,?y:Nat),plus(?y:Nat,?y:Nat)),s(plus(?y:Nat,?x:Nat))),times(?x:Nat,?y:Nat)) = s(s(?x:Nat))",
"plus(plus(times(plus(?x:Nat,?z:Nat),plus(?y:Nat,?y:Nat)),plus(s(?y:Nat),plus(?z:Nat,?z:Nat))),?x:Nat) = times(?x:Nat,?z:Nat)",
"times(plus(?x:Nat,times(?z:Nat,times(?y:Nat,?y:Nat))),plus(plus(times(?y:Nat,?x:Nat),s(?x:Nat)),?x:Nat)) = plus(?z:Nat,?y:Nat)",
"times(plus(?y:Nat,s(times(?z:Nat,?z:Nat))),times(?x:Nat,times(plus(?x:Nat,?y:Nat),plus(?y:Nat,?z:Nat)))) = times(?y:Nat,?y:Nat)",
"times(plus(plus(plus(?z:Nat,?z:Nat),s(?y:Nat)),?z:Nat),times(?y:Nat,times(times(?x:Nat,?z:Nat),?z:Nat))) = times(?y:Nat,?y:Nat)",
"times(plus(times(times(?y:Nat,?x:Nat),times(?z:Nat,?x:Nat)),times(times(?z:Nat,?z:Nat),?y:Nat)),s(?z:Nat)) = plus(?x:Nat,?z:Nat)",
"plus(times(times(?z:Nat,plus(?y:Nat,?x:Nat)),?x:Nat),times(s(times(?x:Nat,?y:Nat)),s(times(?y:Nat,?x:Nat)))) = s(?y:Nat)",
"plus(times(times(plus(?x:Nat,?y:Nat),times(?z:Nat,?x:Nat)),s(s(?z:Nat))),s(s(times(?x:Nat,?y:Nat)))) = s(?x:Nat)",
"times(s(times(plus(?y:Nat,?y:Nat),times(?x:Nat,?x:Nat))),times(?z:Nat,times(s(?x:Nat),plus(?x:Nat,?z:Nat)))) = s(?x:Nat)",
"times(plus(times(plus(?x:Nat,?z:Nat),plus(?x:Nat,?z:Nat)),plus(?x:Nat,?y:Nat)),plus(?x:Nat,times(?y:Nat,?x:Nat))) = s(?x:Nat)",
"times(times(s(s(?z:Nat)),?z:Nat),times(times(plus(?z:Nat,?y:Nat),?x:Nat),plus(?x:Nat,plus(?y:Nat,?x:Nat)))) = s(?z:Nat)",
"s(?x:Nat) = plus(times(?x:Nat,plus(?z:Nat,s(?x:Nat))),times(plus(s(?z:Nat),times(?x:Nat,?y:Nat)),s(plus(?z:Nat,?y:Nat))))",
"s(?y:Nat) = times(plus(plus(plus(?y:Nat,?x:Nat),s(?y:Nat)),plus(times(?z:Nat,?x:Nat),times(?z:Nat,?z:Nat))),times(?x:Nat,?y:Nat))",
"plus(?x:Nat,?x:Nat) = times(times(plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?y:Nat)),?y:Nat),plus(s(?y:Nat),s(times(?y:Nat,?z:Nat))))",
"times(?x:Nat,?x:Nat) = times(s(plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,?y:Nat))),plus(times(times(?x:Nat,?z:Nat),?y:Nat),s(?z:Nat)))",
"times(?x:Nat,?x:Nat) = times(s(plus(plus(?z:Nat,?z:Nat),s(?x:Nat))),times(?x:Nat,plus(times(?x:Nat,?z:Nat),times(?x:Nat,?y:Nat))))",
"times(?y:Nat,?z:Nat) = times(times(s(s(?z:Nat)),?x:Nat),plus(plus(?z:Nat,times(?y:Nat,?x:Nat)),times(?x:Nat,plus(?x:Nat,?x:Nat))))",
"times(?z:Nat,?z:Nat) = plus(plus(times(times(?x:Nat,?y:Nat),plus(?z:Nat,?y:Nat)),?x:Nat),plus(times(?y:Nat,plus(?x:Nat,?z:Nat)),?z:Nat))",
"s(times(?x:Nat,?y:Nat)) = plus(times(s(times(?x:Nat,?z:Nat)),s(s(?z:Nat))),times(s(?x:Nat),plus(s(?y:Nat),?x:Nat)))",
"s(plus(?z:Nat,s(?y:Nat))) = times(plus(s(times(?x:Nat,?z:Nat)),times(s(?y:Nat),?y:Nat)),times(plus(?y:Nat,?z:Nat),?z:Nat))",
"times(plus(?z:Nat,?y:Nat),?z:Nat) = plus(times(s(s(?y:Nat)),?z:Nat),plus(s(plus(?z:Nat,?z:Nat)),s(plus(?z:Nat,?z:Nat))))",
"s(times(plus(?x:Nat,?y:Nat),?y:Nat)) = plus(s(s(plus(?y:Nat,?y:Nat))),plus(plus(plus(?y:Nat,?x:Nat),s(?x:Nat)),?y:Nat))",
"plus(s(times(?x:Nat,?y:Nat)),?y:Nat) = plus(times(s(?x:Nat),plus(plus(?z:Nat,?x:Nat),times(?x:Nat,?x:Nat))),plus(?z:Nat,?x:Nat))",
"times(s(s(s(?y:Nat))),?z:Nat) = plus(plus(times(times(?z:Nat,?z:Nat),s(?y:Nat)),plus(plus(?z:Nat,?y:Nat),?y:Nat)),?y:Nat)",
"s(s(plus(plus(?y:Nat,?x:Nat),?y:Nat))) = plus(?x:Nat,plus(times(s(?x:Nat),plus(?z:Nat,?y:Nat)),s(plus(?x:Nat,?y:Nat))))",
"s(plus(s(plus(?x:Nat,?x:Nat)),?z:Nat)) = plus(times(times(?y:Nat,?y:Nat),times(?y:Nat,times(?x:Nat,?y:Nat))),plus(?x:Nat,?x:Nat))",
"s(plus(s(times(?z:Nat,?x:Nat)),?x:Nat)) = times(plus(plus(plus(?z:Nat,?y:Nat),?y:Nat),s(times(?y:Nat,?x:Nat))),s(?z:Nat))",
"plus(?z:Nat,s(times(?z:Nat,s(?z:Nat)))) = times(plus(plus(?x:Nat,s(?x:Nat)),?y:Nat),s(times(?y:Nat,plus(?z:Nat,?z:Nat))))",
"times(?y:Nat,s(s(times(?y:Nat,?y:Nat)))) = plus(s(s(s(?x:Nat))),plus(plus(s(?y:Nat),times(?x:Nat,?z:Nat)),?x:Nat))",
"s(s(times(s(?x:Nat),plus(?y:Nat,?x:Nat)))) = times(times(?z:Nat,?z:Nat),s(times(times(?y:Nat,?y:Nat),times(?x:Nat,?x:Nat))))",
"s(times(times(s(?y:Nat),?z:Nat),s(?x:Nat))) = times(?y:Nat,times(s(s(?y:Nat)),times(s(?x:Nat),times(?y:Nat,?x:Nat))))",
"plus(?y:Nat,s(plus(?x:Nat,plus(?y:Nat,?y:Nat)))) = plus(plus(?y:Nat,?z:Nat),times(times(?x:Nat,s(?y:Nat)),s(s(?x:Nat))))",
"plus(?y:Nat,plus(s(times(?z:Nat,?y:Nat)),?z:Nat)) = times(times(plus(plus(?y:Nat,?y:Nat),s(?z:Nat)),plus(?z:Nat,?x:Nat)),?x:Nat)",
"plus(?z:Nat,s(times(?z:Nat,times(?y:Nat,?y:Nat)))) = times(times(times(s(?z:Nat),plus(?y:Nat,?x:Nat)),s(?z:Nat)),s(?z:Nat))",
"plus(s(?z:Nat),s(s(plus(?x:Nat,?z:Nat)))) = times(plus(?z:Nat,?y:Nat),s(plus(plus(?z:Nat,?z:Nat),times(?y:Nat,?y:Nat))))",
"plus(plus(plus(s(?x:Nat),?x:Nat),?x:Nat),?x:Nat) = plus(s(plus(?z:Nat,s(?x:Nat))),times(?z:Nat,s(times(?z:Nat,?x:Nat))))",
"s(s(plus(plus(?y:Nat,?z:Nat),times(?x:Nat,?y:Nat)))) = plus(times(?z:Nat,?x:Nat),times(plus(plus(?z:Nat,?x:Nat),?x:Nat),?y:Nat))",
"s(plus(times(?z:Nat,plus(?x:Nat,?y:Nat)),s(?y:Nat))) = times(s(times(?x:Nat,?x:Nat)),plus(?z:Nat,times(s(?x:Nat),?z:Nat)))",
"s(times(times(?z:Nat,?z:Nat),plus(s(?y:Nat),?y:Nat))) = times(times(?y:Nat,s(times(?x:Nat,?y:Nat))),times(?y:Nat,s(?x:Nat)))",
"plus(s(plus(?y:Nat,plus(?x:Nat,?y:Nat))),s(?x:Nat)) = s(times(s(s(?y:Nat)),times(times(?z:Nat,?z:Nat),s(?z:Nat))))",
"plus(times(s(?z:Nat),s(s(?y:Nat))),s(?x:Nat)) = s(plus(times(times(?y:Nat,?x:Nat),?y:Nat),plus(s(?x:Nat),?z:Nat)))",
"plus(times(times(?z:Nat,times(?y:Nat,?z:Nat)),?x:Nat),?z:Nat) = plus(?x:Nat,times(?z:Nat,plus(plus(?x:Nat,?y:Nat),times(?y:Nat,?z:Nat))))",
"s(plus(?x:Nat,times(times(?x:Nat,?x:Nat),times(?x:Nat,?z:Nat)))) = times(times(?x:Nat,times(times(?z:Nat,?z:Nat),s(?x:Nat))),?z:Nat)",
"s(times(?x:Nat,plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,?z:Nat)))) = plus(?z:Nat,s(times(plus(?x:Nat,?z:Nat),times(?y:Nat,?x:Nat))))",
"s(times(plus(s(?z:Nat),plus(?x:Nat,?y:Nat)),s(?x:Nat))) = times(s(?y:Nat),times(s(times(?z:Nat,?z:Nat)),s(?z:Nat)))",
"plus(?x:Nat,plus(times(?z:Nat,?z:Nat),s(plus(?y:Nat,?z:Nat)))) = times(?y:Nat,times(?z:Nat,plus(s(?y:Nat),plus(?z:Nat,?y:Nat))))",
"plus(s(?y:Nat),times(?z:Nat,times(plus(?y:Nat,?y:Nat),?y:Nat))) = plus(?y:Nat,s(times(plus(?x:Nat,?y:Nat),times(?z:Nat,?z:Nat))))",
"plus(plus(?x:Nat,?y:Nat),plus(?y:Nat,times(?y:Nat,s(?z:Nat)))) = times(s(?z:Nat),times(?y:Nat,times(?y:Nat,times(?y:Nat,?z:Nat))))",
"plus(plus(?y:Nat,s(plus(?y:Nat,?y:Nat))),s(s(?x:Nat))) = plus(?z:Nat,s(times(times(?x:Nat,?z:Nat),times(?y:Nat,?x:Nat))))",
"plus(times(?z:Nat,?z:Nat),plus(s(plus(?x:Nat,?y:Nat)),?z:Nat)) = plus(times(s(s(?x:Nat)),s(plus(?x:Nat,?y:Nat))),?x:Nat)",
"times(?y:Nat,s(plus(plus(?x:Nat,?x:Nat),times(?y:Nat,?x:Nat)))) = s(times(s(times(?x:Nat,?y:Nat)),s(plus(?y:Nat,?z:Nat))))",
"times(?y:Nat,times(s(?z:Nat),plus(times(?x:Nat,?x:Nat),?z:Nat))) = times(?z:Nat,times(s(s(?x:Nat)),plus(?z:Nat,s(?z:Nat))))",
"times(s(plus(plus(?x:Nat,?y:Nat),times(?y:Nat,?y:Nat))),?y:Nat) = s(plus(plus(times(?y:Nat,?x:Nat),times(?x:Nat,?x:Nat)),?z:Nat))",
"times(s(plus(times(?y:Nat,?y:Nat),?y:Nat)),s(s(?z:Nat))) = times(?y:Nat,times(times(?y:Nat,?x:Nat),s(plus(?y:Nat,?x:Nat))))",
"times(s(times(?y:Nat,times(?z:Nat,?x:Nat))),plus(?x:Nat,?x:Nat)) = plus(?z:Nat,times(?y:Nat,plus(plus(?z:Nat,?x:Nat),s(?x:Nat))))",
"times(times(?y:Nat,?x:Nat),times(s(plus(?y:Nat,?x:Nat)),?z:Nat)) = plus(times(?x:Nat,s(plus(?y:Nat,?y:Nat))),s(s(?x:Nat)))",
"times(times(times(?y:Nat,?x:Nat),s(times(?z:Nat,?y:Nat))),?z:Nat) = plus(s(times(?y:Nat,plus(?z:Nat,?z:Nat))),times(?z:Nat,?z:Nat))",
"s(plus(plus(plus(?y:Nat,?y:Nat),times(?y:Nat,?y:Nat)),s(?z:Nat))) = times(?z:Nat,plus(?z:Nat,plus(?x:Nat,plus(?y:Nat,?y:Nat))))",
"s(plus(plus(times(?z:Nat,?z:Nat),s(?x:Nat)),plus(?z:Nat,?x:Nat))) = plus(times(?y:Nat,plus(times(?z:Nat,?x:Nat),?z:Nat)),?x:Nat)",
"plus(?z:Nat,times(?x:Nat,times(times(?z:Nat,?z:Nat),times(?y:Nat,?y:Nat)))) = times(?z:Nat,plus(?x:Nat,times(times(?y:Nat,?x:Nat),?x:Nat)))",
"plus(?z:Nat,times(?z:Nat,plus(plus(?y:Nat,?z:Nat),times(?z:Nat,?z:Nat)))) = plus(times(s(?y:Nat),s(plus(?y:Nat,?x:Nat))),?x:Nat)",
"plus(plus(?x:Nat,?z:Nat),s(times(s(?y:Nat),times(?z:Nat,?z:Nat)))) = s(times(?x:Nat,times(s(?z:Nat),plus(?x:Nat,?y:Nat))))",
"times(?x:Nat,plus(?y:Nat,plus(plus(?x:Nat,?y:Nat),times(?y:Nat,?x:Nat)))) = times(s(plus(plus(?y:Nat,?y:Nat),s(?y:Nat))),?y:Nat)",
"times(?y:Nat,times(s(?x:Nat),plus(plus(?x:Nat,?z:Nat),s(?x:Nat)))) = plus(times(?z:Nat,times(plus(?z:Nat,?y:Nat),?z:Nat)),?x:Nat)",
"times(s(plus(?y:Nat,s(?x:Nat))),plus(s(s(?y:Nat)),?x:Nat)) = plus(s(times(s(?x:Nat),times(?y:Nat,?z:Nat))),?z:Nat)",
"times(plus(?x:Nat,s(?y:Nat)),s(times(?x:Nat,plus(?y:Nat,?y:Nat)))) = s(times(s(plus(?x:Nat,?x:Nat)),plus(?x:Nat,?z:Nat)))",
"times(plus(?x:Nat,times(s(?z:Nat),times(?x:Nat,?z:Nat))),s(?y:Nat)) = s(plus(?x:Nat,times(s(?y:Nat),times(?y:Nat,?z:Nat))))",
"times(times(s(plus(?z:Nat,?z:Nat)),s(plus(?z:Nat,?y:Nat))),?z:Nat) = plus(?x:Nat,plus(s(plus(?y:Nat,?y:Nat)),s(?z:Nat)))",
"plus(?x:Nat,plus(s(?z:Nat),times(plus(?z:Nat,?x:Nat),times(?z:Nat,?y:Nat)))) = s(plus(times(plus(?y:Nat,?y:Nat),?z:Nat),?x:Nat))",
"plus(plus(?x:Nat,?z:Nat),s(plus(times(?z:Nat,?x:Nat),times(?x:Nat,?y:Nat)))) = times(plus(?x:Nat,plus(?z:Nat,?y:Nat)),s(?y:Nat))",
"plus(plus(plus(plus(?y:Nat,?z:Nat),s(?z:Nat)),?z:Nat),plus(?z:Nat,?z:Nat)) = times(?y:Nat,s(plus(?x:Nat,plus(?x:Nat,?z:Nat))))",
"plus(times(?y:Nat,s(plus(?x:Nat,?y:Nat))),times(s(s(?z:Nat)),?x:Nat)) = times(?z:Nat,times(?x:Nat,s(plus(?y:Nat,?y:Nat))))",
"times(s(?x:Nat),plus(?y:Nat,times(times(?y:Nat,?z:Nat),plus(?z:Nat,?x:Nat)))) = s(times(s(?y:Nat),s(plus(?y:Nat,?y:Nat))))",
"s(times(plus(times(?x:Nat,?z:Nat),plus(?z:Nat,?x:Nat)),s(plus(?x:Nat,?y:Nat)))) = plus(times(?x:Nat,s(s(?x:Nat))),?y:Nat)",
"plus(?y:Nat,plus(plus(s(?z:Nat),plus(?x:Nat,?z:Nat)),s(times(?z:Nat,?z:Nat)))) = times(s(s(plus(?z:Nat,?x:Nat))),?y:Nat)",
"plus(s(?z:Nat),plus(times(s(?y:Nat),s(?x:Nat)),plus(?x:Nat,s(?z:Nat)))) = s(s(plus(?x:Nat,times(?x:Nat,?z:Nat))))",
"plus(plus(s(?x:Nat),s(s(?x:Nat))),times(s(plus(?x:Nat,?x:Nat)),?z:Nat)) = s(s(times(?x:Nat,times(?y:Nat,?y:Nat))))",
"times(s(?x:Nat),times(plus(s(?x:Nat),s(?z:Nat)),s(plus(?y:Nat,?x:Nat)))) = s(s(plus(?y:Nat,times(?x:Nat,?x:Nat))))",
"times(plus(plus(s(?z:Nat),times(?x:Nat,?y:Nat)),times(?y:Nat,s(?x:Nat))),?x:Nat) = plus(?z:Nat,s(s(plus(?z:Nat,?x:Nat))))",
"s(plus(times(s(?y:Nat),times(?x:Nat,?x:Nat)),times(times(?y:Nat,?x:Nat),s(?y:Nat)))) = times(times(s(?z:Nat),?x:Nat),?z:Nat)",
"times(s(times(times(?x:Nat,?y:Nat),?y:Nat)),s(times(times(?x:Nat,?z:Nat),s(?z:Nat)))) = s(times(times(?x:Nat,?x:Nat),?x:Nat))",
"times(plus(plus(s(?x:Nat),plus(?z:Nat,?x:Nat)),?y:Nat),times(times(?y:Nat,?z:Nat),?z:Nat)) = s(s(times(s(?y:Nat),?x:Nat)))",
"s(plus(plus(plus(?y:Nat,?y:Nat),plus(?y:Nat,?x:Nat)),times(s(?x:Nat),plus(?x:Nat,?y:Nat)))) = times(s(?x:Nat),s(?x:Nat))",
"plus(times(plus(plus(?x:Nat,?x:Nat),times(?y:Nat,?y:Nat)),plus(s(?x:Nat),?z:Nat)),s(?y:Nat)) = s(plus(?z:Nat,s(?x:Nat)))",
"times(s(plus(s(?z:Nat),?y:Nat)),plus(times(plus(?y:Nat,?z:Nat),times(?z:Nat,?x:Nat)),?x:Nat)) = plus(?x:Nat,plus(?y:Nat,?z:Nat))",
"times(times(plus(plus(?y:Nat,?z:Nat),times(?y:Nat,?y:Nat)),times(plus(?z:Nat,?x:Nat),?z:Nat)),?z:Nat) = plus(?z:Nat,s(s(?z:Nat)))",
"plus(s(times(?y:Nat,times(?y:Nat,?z:Nat))),plus(s(times(?x:Nat,?x:Nat)),plus(?y:Nat,s(?z:Nat)))) = times(?x:Nat,s(?y:Nat))",
"plus(s(times(plus(?x:Nat,?z:Nat),times(?z:Nat,?x:Nat))),times(times(times(?z:Nat,?y:Nat),?x:Nat),?y:Nat)) = plus(s(?z:Nat),?z:Nat)",
"times(s(s(plus(?y:Nat,?x:Nat))),times(plus(plus(?y:Nat,?y:Nat),?y:Nat),s(plus(?y:Nat,?x:Nat)))) = s(times(?y:Nat,?y:Nat))",
"times(s(times(times(?y:Nat,?z:Nat),?y:Nat)),plus(?z:Nat,plus(plus(?z:Nat,?x:Nat),plus(?z:Nat,?y:Nat)))) = plus(?y:Nat,s(?y:Nat))",
"times(plus(times(plus(?z:Nat,?x:Nat),times(?y:Nat,?z:Nat)),plus(s(?y:Nat),plus(?z:Nat,?x:Nat))),?x:Nat) = times(?x:Nat,s(?y:Nat))",
"times(times(?y:Nat,plus(plus(?z:Nat,?y:Nat),times(?y:Nat,?y:Nat))),s(times(s(?y:Nat),s(?x:Nat)))) = s(s(s(?x:Nat)))",
"plus(times(?z:Nat,plus(?x:Nat,s(?z:Nat))),times(plus(s(?z:Nat),plus(?y:Nat,?x:Nat)),times(s(?y:Nat),?z:Nat))) = s(?x:Nat)",
"plus(times(s(plus(?x:Nat,?x:Nat)),s(plus(?z:Nat,?z:Nat))),s(plus(plus(?z:Nat,?z:Nat),plus(?x:Nat,?x:Nat)))) = s(?z:Nat)",
"times(plus(plus(?x:Nat,s(?x:Nat)),s(s(?y:Nat))),plus(s(?z:Nat),plus(plus(?z:Nat,?x:Nat),s(?y:Nat)))) = s(?x:Nat)",
"times(plus(times(plus(?x:Nat,?z:Nat),?y:Nat),plus(s(?y:Nat),?x:Nat)),plus(plus(times(?y:Nat,?x:Nat),?z:Nat),?z:Nat)) = s(?y:Nat)",
"times(times(plus(s(?y:Nat),?x:Nat),times(plus(?x:Nat,?z:Nat),s(?y:Nat))),s(times(?z:Nat,plus(?x:Nat,?z:Nat)))) = s(?x:Nat)",
"s(?x:Nat) = plus(s(s(times(?z:Nat,?x:Nat))),plus(plus(times(?y:Nat,?z:Nat),?y:Nat),plus(plus(?y:Nat,?z:Nat),plus(?z:Nat,?y:Nat))))",
"s(?x:Nat) = plus(times(?z:Nat,times(plus(?x:Nat,?y:Nat),plus(?y:Nat,?z:Nat))),times(plus(times(?z:Nat,?y:Nat),times(?z:Nat,?x:Nat)),?x:Nat))",
"s(?x:Nat) = times(times(times(plus(?y:Nat,?y:Nat),s(?y:Nat)),?y:Nat),plus(times(plus(?y:Nat,?y:Nat),times(?x:Nat,?y:Nat)),s(?x:Nat)))",
"s(?z:Nat) = plus(s(plus(plus(?x:Nat,?y:Nat),?y:Nat)),plus(s(plus(?x:Nat,?y:Nat)),plus(times(?x:Nat,?x:Nat),times(?z:Nat,?z:Nat))))",
"s(?z:Nat) = times(plus(?z:Nat,plus(times(?z:Nat,?x:Nat),times(?y:Nat,?y:Nat))),plus(?z:Nat,plus(plus(?x:Nat,?x:Nat),plus(?z:Nat,?y:Nat))))",
"s(?z:Nat) = times(plus(s(times(?y:Nat,?y:Nat)),times(s(?y:Nat),plus(?z:Nat,?x:Nat))),plus(?y:Nat,times(plus(?y:Nat,?x:Nat),?z:Nat)))",
"s(s(?z:Nat)) = times(times(plus(plus(?z:Nat,?y:Nat),?y:Nat),s(plus(?y:Nat,?x:Nat))),s(times(s(?z:Nat),times(?z:Nat,?x:Nat))))",
"plus(?z:Nat,?z:Nat) = times(plus(plus(plus(?z:Nat,?y:Nat),s(?y:Nat)),s(times(?z:Nat,?y:Nat))),plus(?y:Nat,plus(s(?x:Nat),?z:Nat)))",
"times(?x:Nat,?z:Nat) = times(s(times(?y:Nat,times(?x:Nat,?y:Nat))),plus(times(times(?z:Nat,?x:Nat),?y:Nat),times(?z:Nat,plus(?x:Nat,?z:Nat))))",
"times(?y:Nat,?x:Nat) = times(times(s(s(?y:Nat)),plus(?z:Nat,?y:Nat)),plus(s(times(?x:Nat,?z:Nat)),plus(?y:Nat,plus(?x:Nat,?y:Nat))))",
"s(plus(?x:Nat,?z:Nat)) = times(plus(plus(s(?y:Nat),?y:Nat),plus(plus(?z:Nat,?y:Nat),times(?x:Nat,?y:Nat))),times(s(?y:Nat),?y:Nat))",
"s(times(?x:Nat,s(?y:Nat))) = plus(s(times(times(?z:Nat,?z:Nat),?x:Nat)),times(plus(plus(?y:Nat,?z:Nat),s(?z:Nat)),s(?z:Nat)))",
"plus(?y:Nat,plus(?y:Nat,?x:Nat)) = plus(plus(s(times(?x:Nat,?y:Nat)),plus(times(?x:Nat,?y:Nat),?z:Nat)),s(s(times(?x:Nat,?z:Nat))))",
"plus(s(?x:Nat),s(?y:Nat)) = times(plus(?x:Nat,times(?x:Nat,plus(?z:Nat,?z:Nat))),s(times(plus(?y:Nat,?y:Nat),times(?y:Nat,?y:Nat))))",
"plus(plus(?y:Nat,?x:Nat),?x:Nat) = times(plus(?x:Nat,?x:Nat),times(times(plus(?x:Nat,?x:Nat),times(?x:Nat,?z:Nat)),s(plus(?x:Nat,?y:Nat))))",
"times(?x:Nat,plus(?z:Nat,?x:Nat)) = plus(plus(times(s(?y:Nat),plus(?x:Nat,?z:Nat)),times(times(?z:Nat,?x:Nat),times(?z:Nat,?x:Nat))),?y:Nat)",
"s(s(times(s(?x:Nat),?y:Nat))) = times(s(times(s(?z:Nat),plus(?x:Nat,?z:Nat))),plus(?x:Nat,times(plus(?y:Nat,?z:Nat),?x:Nat)))",
"s(s(times(times(?x:Nat,?z:Nat),?y:Nat))) = times(?y:Nat,plus(times(s(?x:Nat),plus(?z:Nat,?x:Nat)),times(times(?z:Nat,?z:Nat),?y:Nat)))",
"s(plus(?y:Nat,times(?x:Nat,s(?y:Nat)))) = plus(plus(?z:Nat,times(s(?z:Nat),?y:Nat)),plus(plus(plus(?z:Nat,?y:Nat),?x:Nat),?y:Nat))",
"s(plus(s(times(?x:Nat,?z:Nat)),?x:Nat)) = plus(s(plus(?y:Nat,times(?x:Nat,?y:Nat))),plus(?z:Nat,times(?y:Nat,times(?x:Nat,?y:Nat))))",
"s(times(?z:Nat,s(times(?x:Nat,?z:Nat)))) = times(s(plus(s(?y:Nat),?z:Nat)),plus(s(?x:Nat),plus(?y:Nat,plus(?z:Nat,?z:Nat))))",
"s(s(times(s(?z:Nat),plus(?y:Nat,?x:Nat)))) = times(plus(plus(?x:Nat,s(?z:Nat)),s(times(?x:Nat,?z:Nat))),plus(?x:Nat,?z:Nat))",
"s(plus(?x:Nat,plus(?z:Nat,times(?y:Nat,?z:Nat)))) = times(plus(?z:Nat,s(times(?z:Nat,?x:Nat))),times(s(plus(?y:Nat,?y:Nat)),?z:Nat))",
"s(plus(?y:Nat,plus(?x:Nat,plus(?x:Nat,?x:Nat)))) = plus(plus(times(?x:Nat,?y:Nat),plus(s(?z:Nat),times(?y:Nat,?x:Nat))),s(?z:Nat))",
"plus(?y:Nat,s(plus(s(?y:Nat),s(?x:Nat)))) = plus(?x:Nat,times(plus(times(?z:Nat,?y:Nat),?y:Nat),times(?z:Nat,plus(?z:Nat,?y:Nat))))",
"plus(s(plus(s(?y:Nat),s(?y:Nat))),?z:Nat) = plus(plus(?z:Nat,s(plus(?y:Nat,?z:Nat))),s(plus(times(?z:Nat,?y:Nat),?x:Nat)))",
"plus(plus(s(plus(?z:Nat,?x:Nat)),?z:Nat),?y:Nat) = plus(times(?y:Nat,?y:Nat),times(plus(plus(?z:Nat,?z:Nat),times(?y:Nat,?z:Nat)),?x:Nat))",
"times(?x:Nat,plus(s(plus(?y:Nat,?x:Nat)),?x:Nat)) = s(times(plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,?y:Nat)),s(plus(?z:Nat,?x:Nat))))",
"times(?z:Nat,plus(?z:Nat,s(times(?y:Nat,?y:Nat)))) = times(?y:Nat,times(s(s(?y:Nat)),plus(plus(?y:Nat,?x:Nat),times(?z:Nat,?z:Nat))))",
"times(?z:Nat,times(?y:Nat,s(plus(?y:Nat,?y:Nat)))) = times(s(times(?y:Nat,?z:Nat)),plus(s(?y:Nat),times(times(?x:Nat,?x:Nat),?z:Nat)))",
"times(plus(?y:Nat,?z:Nat),s(s(s(?x:Nat)))) = plus(plus(s(?x:Nat),plus(?z:Nat,times(?x:Nat,?y:Nat))),s(plus(?z:Nat,?y:Nat)))",
"s(plus(?z:Nat,plus(s(?z:Nat),times(?z:Nat,?y:Nat)))) = times(times(s(s(?x:Nat)),times(times(?z:Nat,?y:Nat),?y:Nat)),s(?y:Nat))",
"s(times(?y:Nat,times(s(?x:Nat),plus(?x:Nat,?z:Nat)))) = times(plus(times(s(?z:Nat),times(?y:Nat,?x:Nat)),?z:Nat),s(s(?y:Nat)))",
"plus(?y:Nat,plus(?z:Nat,plus(?z:Nat,times(?z:Nat,?x:Nat)))) = times(s(plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,?x:Nat))),s(s(?z:Nat)))",
"plus(plus(?z:Nat,plus(plus(?z:Nat,?y:Nat),?x:Nat)),?x:Nat) = plus(times(?y:Nat,times(s(?z:Nat),s(?x:Nat))),s(s(s(?y:Nat))))",
"times(?z:Nat,times(plus(plus(?x:Nat,?x:Nat),?y:Nat),?x:Nat)) = times(?x:Nat,plus(times(?y:Nat,plus(?y:Nat,?x:Nat)),times(?x:Nat,s(?y:Nat))))",
"times(s(?x:Nat),s(times(?z:Nat,plus(?x:Nat,?y:Nat)))) = plus(plus(plus(?y:Nat,?z:Nat),times(s(?x:Nat),plus(?y:Nat,?x:Nat))),?z:Nat)",
"times(s(?z:Nat),s(times(?y:Nat,plus(?z:Nat,?z:Nat)))) = plus(plus(plus(plus(?y:Nat,?z:Nat),?x:Nat),plus(?z:Nat,s(?y:Nat))),?z:Nat)",
"s(plus(?y:Nat,times(times(?z:Nat,?y:Nat),plus(?y:Nat,?x:Nat)))) = plus(times(?y:Nat,?y:Nat),plus(plus(plus(?y:Nat,?x:Nat),?z:Nat),?z:Nat))",
"plus(?z:Nat,times(s(?y:Nat),times(?y:Nat,times(?z:Nat,?z:Nat)))) = plus(s(plus(?z:Nat,s(?z:Nat))),plus(plus(?z:Nat,?z:Nat),?z:Nat))",
"plus(plus(?y:Nat,plus(s(?x:Nat),plus(?y:Nat,?y:Nat))),?y:Nat) = plus(plus(?z:Nat,?z:Nat),plus(times(?z:Nat,?z:Nat),plus(?x:Nat,?y:Nat)))",
"plus(times(?x:Nat,times(plus(?z:Nat,?x:Nat),s(?x:Nat))),?x:Nat) = plus(s(?z:Nat),times(times(?x:Nat,plus(?x:Nat,?y:Nat)),s(?z:Nat)))",
"plus(times(times(s(?y:Nat),?y:Nat),s(s(?z:Nat))),?x:Nat) = plus(times(s(?x:Nat),plus(?x:Nat,?x:Nat)),times(s(?y:Nat),?y:Nat))",
"plus(times(times(times(?y:Nat,?y:Nat),?x:Nat),?y:Nat),s(?y:Nat)) = plus(?x:Nat,plus(plus(times(?y:Nat,?z:Nat),plus(?y:Nat,?y:Nat)),?y:Nat))",
"times(?x:Nat,plus(s(?z:Nat),times(?y:Nat,times(?y:Nat,?x:Nat)))) = plus(plus(?y:Nat,s(?x:Nat)),plus(s(plus(?x:Nat,?z:Nat)),?y:Nat))",
"times(s(?z:Nat),plus(s(s(?z:Nat)),plus(?x:Nat,?x:Nat))) = times(?z:Nat,plus(s(?z:Nat),times(times(?z:Nat,?y:Nat),s(?x:Nat))))",
"times(s(plus(times(?y:Nat,?y:Nat),?z:Nat)),times(?z:Nat,?x:Nat)) = plus(?y:Nat,plus(times(plus(?y:Nat,?z:Nat),s(?z:Nat)),s(?x:Nat)))",
"times(s(times(plus(?x:Nat,?z:Nat),times(?x:Nat,?z:Nat))),?x:Nat) = plus(?z:Nat,plus(plus(s(?x:Nat),times(?y:Nat,?x:Nat)),s(?z:Nat)))",
"times(s(times(times(?x:Nat,?x:Nat),times(?x:Nat,?x:Nat))),?x:Nat) = s(times(s(plus(?z:Nat,?x:Nat)),plus(?x:Nat,plus(?y:Nat,?y:Nat))))",
"s(plus(plus(?x:Nat,s(?y:Nat)),times(?y:Nat,plus(?y:Nat,?z:Nat)))) = times(times(?y:Nat,?z:Nat),s(times(?x:Nat,plus(?z:Nat,?z:Nat))))",
"s(plus(times(?z:Nat,plus(?x:Nat,?z:Nat)),plus(s(?z:Nat),?z:Nat))) = plus(s(plus(times(?z:Nat,?x:Nat),plus(?y:Nat,?x:Nat))),?y:Nat)",
"s(plus(times(times(?y:Nat,?x:Nat),times(?y:Nat,?x:Nat)),s(?z:Nat))) = plus(?x:Nat,times(s(?x:Nat),times(?y:Nat,times(?y:Nat,?x:Nat))))",
"s(times(s(?y:Nat),times(times(?x:Nat,?y:Nat),plus(?z:Nat,?z:Nat)))) = times(times(?y:Nat,?x:Nat),s(plus(plus(?x:Nat,?x:Nat),?z:Nat)))",
"plus(plus(times(s(?y:Nat),?z:Nat),s(plus(?y:Nat,?x:Nat))),?z:Nat) = plus(times(plus(?y:Nat,?x:Nat),times(?y:Nat,s(?y:Nat))),?z:Nat)",
"times(?y:Nat,plus(plus(times(?z:Nat,?z:Nat),times(?y:Nat,?z:Nat)),?y:Nat)) = plus(plus(?z:Nat,?y:Nat),s(plus(plus(?x:Nat,?z:Nat),?y:Nat)))",
"plus(?x:Nat,plus(plus(?z:Nat,?x:Nat),plus(times(?z:Nat,?y:Nat),s(?y:Nat)))) = s(times(?y:Nat,plus(s(?z:Nat),times(?x:Nat,?z:Nat))))",
"plus(s(plus(s(?x:Nat),?x:Nat)),s(times(?z:Nat,plus(?y:Nat,?z:Nat)))) = s(plus(?z:Nat,times(times(?x:Nat,?y:Nat),s(?x:Nat))))",
"plus(plus(times(plus(?x:Nat,?x:Nat),?y:Nat),?z:Nat),plus(s(?z:Nat),?z:Nat)) = times(times(?y:Nat,?z:Nat),s(plus(?z:Nat,s(?y:Nat))))",
"plus(times(times(times(?y:Nat,?z:Nat),?x:Nat),s(s(?y:Nat))),s(?x:Nat)) = plus(?x:Nat,times(?x:Nat,plus(?z:Nat,times(?y:Nat,?x:Nat))))",
"times(s(plus(?y:Nat,?y:Nat)),times(times(plus(?z:Nat,?y:Nat),?x:Nat),?y:Nat)) = plus(times(times(?x:Nat,plus(?z:Nat,?y:Nat)),?x:Nat),?x:Nat)",
"times(plus(?y:Nat,plus(s(?x:Nat),?y:Nat)),plus(?y:Nat,s(s(?x:Nat)))) = times(plus(?y:Nat,plus(?x:Nat,times(?y:Nat,?x:Nat))),?x:Nat)",
"times(plus(times(times(?z:Nat,?y:Nat),s(?x:Nat)),s(s(?z:Nat))),?x:Nat) = times(?z:Nat,s(plus(times(?y:Nat,?y:Nat),s(?y:Nat))))",
"plus(s(?z:Nat),plus(plus(?z:Nat,plus(?x:Nat,?z:Nat)),times(s(?x:Nat),?y:Nat))) = plus(?y:Nat,s(times(?z:Nat,times(?y:Nat,?z:Nat))))",
"plus(plus(s(s(?x:Nat)),plus(?y:Nat,plus(?z:Nat,?y:Nat))),times(?y:Nat,?z:Nat)) = s(plus(?x:Nat,times(?x:Nat,plus(?z:Nat,?z:Nat))))",
"plus(plus(s(s(?x:Nat)),times(plus(?x:Nat,?y:Nat),?z:Nat)),times(?x:Nat,?x:Nat)) = plus(s(s(times(?y:Nat,?y:Nat))),s(?z:Nat))",
"plus(times(plus(?y:Nat,?y:Nat),?y:Nat),plus(?z:Nat,plus(s(?x:Nat),s(?x:Nat)))) = s(times(?x:Nat,times(?y:Nat,times(?z:Nat,?x:Nat))))",
"plus(times(plus(s(?z:Nat),?z:Nat),plus(plus(?z:Nat,?y:Nat),?x:Nat)),s(?x:Nat)) = s(plus(times(times(?z:Nat,?z:Nat),?z:Nat),?y:Nat))",
"plus(times(plus(plus(?x:Nat,?y:Nat),times(?x:Nat,?z:Nat)),plus(?z:Nat,?z:Nat)),?y:Nat) = times(plus(?y:Nat,s(plus(?z:Nat,?y:Nat))),?z:Nat)",
"times(plus(s(plus(?x:Nat,?z:Nat)),?y:Nat),s(plus(?x:Nat,plus(?y:Nat,?z:Nat)))) = plus(?y:Nat,s(times(plus(?y:Nat,?z:Nat),?z:Nat)))",
"times(plus(times(s(?y:Nat),plus(?z:Nat,?x:Nat)),plus(s(?z:Nat),?x:Nat)),?y:Nat) = times(plus(s(times(?z:Nat,?y:Nat)),?z:Nat),?y:Nat)",
"s(times(times(times(?y:Nat,?y:Nat),plus(?x:Nat,?y:Nat)),plus(s(?x:Nat),s(?z:Nat)))) = times(s(s(times(?z:Nat,?x:Nat))),?z:Nat)",
"plus(?z:Nat,plus(plus(plus(?z:Nat,?x:Nat),times(?z:Nat,?y:Nat)),plus(?x:Nat,s(?y:Nat)))) = s(s(times(times(?z:Nat,?y:Nat),?y:Nat)))",
"plus(s(plus(?y:Nat,times(?y:Nat,?x:Nat))),times(times(?x:Nat,plus(?z:Nat,?z:Nat)),?z:Nat)) = plus(s(s(?y:Nat)),s(s(?x:Nat)))",
"plus(plus(plus(?y:Nat,?z:Nat),plus(times(?y:Nat,?z:Nat),plus(?x:Nat,?y:Nat))),s(?z:Nat)) = times(?z:Nat,plus(?y:Nat,s(s(?y:Nat))))",
"times(plus(times(?x:Nat,times(?y:Nat,?x:Nat)),times(s(?y:Nat),plus(?x:Nat,?x:Nat))),?y:Nat) = times(plus(?y:Nat,?y:Nat),plus(?y:Nat,?x:Nat))",
"times(times(s(?z:Nat),s(?y:Nat)),times(s(?z:Nat),plus(times(?x:Nat,?y:Nat),?z:Nat))) = plus(s(times(?y:Nat,?z:Nat)),s(?x:Nat))",
"plus(s(times(s(?y:Nat),?z:Nat)),plus(s(plus(?z:Nat,?y:Nat)),s(times(?y:Nat,?z:Nat)))) = plus(s(times(?y:Nat,?x:Nat)),?x:Nat)",
"plus(times(plus(s(?x:Nat),plus(?x:Nat,?x:Nat)),?x:Nat),s(times(times(?y:Nat,?y:Nat),?z:Nat))) = plus(times(?y:Nat,?x:Nat),s(?x:Nat))",
"times(s(times(times(?y:Nat,?y:Nat),?x:Nat)),plus(plus(plus(?y:Nat,?y:Nat),times(?z:Nat,?z:Nat)),?x:Nat)) = s(times(?x:Nat,s(?x:Nat)))",
"times(plus(?z:Nat,plus(?z:Nat,plus(?x:Nat,?x:Nat))),times(?x:Nat,plus(times(?y:Nat,?x:Nat),s(?y:Nat)))) = s(plus(?y:Nat,s(?y:Nat)))",
"times(times(times(times(?x:Nat,?x:Nat),plus(?z:Nat,?x:Nat)),times(s(?x:Nat),plus(?x:Nat,?z:Nat))),?x:Nat) = times(s(?x:Nat),s(?x:Nat))",
"plus(plus(plus(times(?z:Nat,?z:Nat),s(?y:Nat)),s(s(?y:Nat))),s(times(?x:Nat,plus(?z:Nat,?z:Nat)))) = times(?x:Nat,s(?x:Nat))",
"times(plus(s(times(?z:Nat,?x:Nat)),s(times(?x:Nat,?x:Nat))),plus(plus(?y:Nat,times(?z:Nat,?x:Nat)),?y:Nat)) = times(s(?x:Nat),?y:Nat)",
"times(times(?z:Nat,times(times(?y:Nat,?z:Nat),plus(?x:Nat,?y:Nat))),plus(?z:Nat,plus(times(?y:Nat,?z:Nat),?z:Nat))) = plus(?z:Nat,s(?x:Nat))",
"times(times(plus(?z:Nat,?x:Nat),plus(?z:Nat,plus(?z:Nat,?y:Nat))),plus(?y:Nat,times(?z:Nat,times(?y:Nat,?z:Nat)))) = s(s(s(?y:Nat)))",
"times(times(plus(plus(?z:Nat,?z:Nat),s(?z:Nat)),s(?z:Nat)),times(plus(s(?z:Nat),s(?y:Nat)),?z:Nat)) = s(plus(?y:Nat,?z:Nat))",
"plus(s(plus(times(?x:Nat,?y:Nat),times(?y:Nat,?z:Nat))),times(s(plus(?x:Nat,?y:Nat)),s(times(?y:Nat,?x:Nat)))) = s(s(?z:Nat))",
"plus(times(s(s(?y:Nat)),times(?x:Nat,s(?y:Nat))),plus(plus(times(?x:Nat,?x:Nat),times(?y:Nat,?z:Nat)),?x:Nat)) = plus(?y:Nat,?x:Nat)",
"plus(times(plus(times(?x:Nat,?z:Nat),plus(?y:Nat,?z:Nat)),?z:Nat),times(times(plus(?x:Nat,?y:Nat),s(?y:Nat)),?y:Nat)) = times(?x:Nat,?z:Nat)",
"times(times(plus(?z:Nat,s(?y:Nat)),times(s(?z:Nat),?x:Nat)),times(?z:Nat,plus(s(?y:Nat),times(?z:Nat,?x:Nat)))) = s(s(?y:Nat))",
"times(times(times(plus(?z:Nat,?y:Nat),?y:Nat),plus(?y:Nat,times(?y:Nat,?y:Nat))),times(s(plus(?y:Nat,?y:Nat)),?z:Nat)) = times(?x:Nat,?y:Nat)",
"plus(times(times(?z:Nat,times(?z:Nat,?z:Nat)),plus(plus(?z:Nat,?y:Nat),?x:Nat)),s(plus(s(?x:Nat),plus(?z:Nat,?x:Nat)))) = s(?z:Nat)",
"times(s(times(plus(?y:Nat,?z:Nat),times(?y:Nat,?y:Nat))),plus(s(?x:Nat),times(plus(?x:Nat,?z:Nat),plus(?z:Nat,?z:Nat)))) = s(?y:Nat)",
"s(?x:Nat) = plus(times(times(?y:Nat,times(?z:Nat,?y:Nat)),s(?z:Nat)),plus(times(plus(?z:Nat,?x:Nat),s(?x:Nat)),s(plus(?x:Nat,?x:Nat))))",
"s(?x:Nat) = times(plus(times(?y:Nat,plus(?z:Nat,?y:Nat)),s(times(?x:Nat,?z:Nat))),times(?y:Nat,plus(plus(?y:Nat,?z:Nat),plus(?y:Nat,?z:Nat))))",
"s(?x:Nat) = times(times(s(times(?y:Nat,?y:Nat)),s(plus(?y:Nat,?y:Nat))),plus(plus(?z:Nat,times(?y:Nat,?x:Nat)),plus(s(?x:Nat),?z:Nat)))",
"s(?y:Nat) = plus(times(s(times(?y:Nat,?y:Nat)),times(plus(?x:Nat,?z:Nat),times(?x:Nat,?y:Nat))),plus(times(?x:Nat,s(?y:Nat)),s(?z:Nat)))",
"s(?y:Nat) = times(s(times(times(?x:Nat,?x:Nat),plus(?z:Nat,?y:Nat))),plus(s(plus(?z:Nat,?z:Nat)),times(times(?x:Nat,?z:Nat),s(?y:Nat))))",
"plus(?x:Nat,?z:Nat) = plus(plus(times(?z:Nat,?y:Nat),?x:Nat),times(plus(plus(?x:Nat,?z:Nat),?x:Nat),plus(plus(?y:Nat,?x:Nat),times(?z:Nat,?x:Nat))))",
"plus(?y:Nat,?x:Nat) = plus(plus(times(times(?x:Nat,?x:Nat),plus(?x:Nat,?z:Nat)),s(?x:Nat)),plus(plus(?z:Nat,times(?z:Nat,?x:Nat)),s(?z:Nat)))",
"s(plus(?z:Nat,?y:Nat)) = times(s(plus(s(?y:Nat),times(?y:Nat,?z:Nat))),plus(s(times(?y:Nat,?z:Nat)),plus(?z:Nat,times(?x:Nat,?x:Nat))))",
"plus(?z:Nat,s(?x:Nat)) = plus(times(plus(plus(?z:Nat,?y:Nat),?y:Nat),times(?z:Nat,?x:Nat)),s(plus(plus(?z:Nat,?x:Nat),plus(?y:Nat,?z:Nat))))",
"s(s(s(s(?x:Nat)))) = times(times(times(times(?x:Nat,?y:Nat),plus(?y:Nat,?x:Nat)),plus(times(?x:Nat,?x:Nat),times(?x:Nat,?z:Nat))),?x:Nat)",
"plus(times(?y:Nat,?z:Nat),?y:Nat) = plus(times(plus(s(?y:Nat),times(?z:Nat,?x:Nat)),plus(s(?y:Nat),plus(?z:Nat,?y:Nat))),plus(?z:Nat,?z:Nat))",
"times(times(?z:Nat,?y:Nat),?z:Nat) = plus(times(plus(?y:Nat,?z:Nat),plus(s(?x:Nat),plus(?z:Nat,?y:Nat))),s(times(times(?z:Nat,?y:Nat),?x:Nat)))",
"s(times(?z:Nat,s(s(?z:Nat)))) = times(s(plus(times(?z:Nat,?y:Nat),plus(?z:Nat,?x:Nat))),plus(times(s(?z:Nat),s(?x:Nat)),?x:Nat))",
"plus(?x:Nat,times(s(?x:Nat),?x:Nat)) = plus(?y:Nat,plus(times(plus(?x:Nat,?x:Nat),times(?z:Nat,?z:Nat)),times(s(?x:Nat),plus(?x:Nat,?y:Nat))))",
"plus(?y:Nat,s(plus(?x:Nat,?x:Nat))) = times(?y:Nat,times(times(plus(?y:Nat,?x:Nat),plus(?z:Nat,?z:Nat)),times(times(?z:Nat,?y:Nat),s(?x:Nat))))",
"plus(s(plus(?y:Nat,?x:Nat)),?z:Nat) = plus(plus(plus(times(?z:Nat,?z:Nat),?z:Nat),?y:Nat),times(times(times(?y:Nat,?y:Nat),s(?z:Nat)),?y:Nat))",
"times(times(?y:Nat,s(?z:Nat)),?z:Nat) = times(times(?z:Nat,plus(s(?x:Nat),times(?x:Nat,?z:Nat))),plus(plus(?z:Nat,plus(?x:Nat,?y:Nat)),?z:Nat))",
"s(times(?y:Nat,times(s(?x:Nat),?z:Nat))) = plus(s(plus(s(?x:Nat),s(?y:Nat))),times(s(s(?y:Nat)),plus(s(?x:Nat),?z:Nat)))",
"plus(times(?y:Nat,times(?z:Nat,?y:Nat)),?y:Nat) = s(plus(times(times(?z:Nat,?y:Nat),times(?x:Nat,?x:Nat)),times(plus(?x:Nat,?y:Nat),s(?x:Nat))))",
"times(s(times(s(?z:Nat),?y:Nat)),?y:Nat) = times(plus(plus(plus(?y:Nat,?x:Nat),s(?z:Nat)),s(?z:Nat)),s(s(times(?x:Nat,?x:Nat))))",
"times(plus(?z:Nat,?z:Nat),s(s(?z:Nat))) = plus(s(times(?z:Nat,plus(?y:Nat,?z:Nat))),s(plus(times(?z:Nat,?y:Nat),times(?y:Nat,?z:Nat))))",
"s(s(plus(s(?y:Nat),plus(?y:Nat,?x:Nat)))) = times(times(?y:Nat,?x:Nat),times(times(times(?z:Nat,?x:Nat),times(?x:Nat,?x:Nat)),s(?z:Nat)))",
"s(s(plus(plus(?y:Nat,?y:Nat),s(?y:Nat)))) = times(times(times(plus(?x:Nat,?x:Nat),plus(?y:Nat,?z:Nat)),s(times(?x:Nat,?y:Nat))),?x:Nat)",
"times(?y:Nat,plus(s(?x:Nat),times(?x:Nat,?x:Nat))) = plus(?z:Nat,times(times(times(?y:Nat,?y:Nat),times(?y:Nat,?x:Nat)),times(?y:Nat,s(?z:Nat))))",
"times(s(?z:Nat),s(s(plus(?z:Nat,?y:Nat)))) = s(times(times(plus(?x:Nat,?y:Nat),s(?z:Nat)),times(s(?z:Nat),times(?z:Nat,?x:Nat))))",
"times(plus(?y:Nat,?y:Nat),s(s(s(?x:Nat)))) = times(?z:Nat,times(plus(times(?z:Nat,?z:Nat),s(?y:Nat)),plus(plus(?z:Nat,?z:Nat),?z:Nat)))",
"s(s(times(times(?z:Nat,?z:Nat),times(?x:Nat,?x:Nat)))) = plus(s(times(?y:Nat,?x:Nat)),times(?y:Nat,times(s(?z:Nat),times(?x:Nat,?y:Nat))))",
"s(plus(?z:Nat,times(s(?x:Nat),plus(?z:Nat,?z:Nat)))) = times(plus(times(?y:Nat,s(?y:Nat)),times(s(?y:Nat),?x:Nat)),s(s(?x:Nat)))",
"s(times(s(s(?z:Nat)),times(?x:Nat,s(?x:Nat)))) = plus(plus(times(times(?z:Nat,?x:Nat),?x:Nat),times(?x:Nat,plus(?x:Nat,?y:Nat))),?y:Nat)",
"plus(?z:Nat,plus(times(?y:Nat,?x:Nat),plus(?x:Nat,?z:Nat))) = plus(s(times(?z:Nat,plus(?x:Nat,?x:Nat))),s(times(?y:Nat,times(?z:Nat,?z:Nat))))",
"plus(plus(?y:Nat,times(plus(?y:Nat,?z:Nat),?y:Nat)),?x:Nat) = times(s(?z:Nat),plus(s(?x:Nat),times(times(?y:Nat,?z:Nat),times(?z:Nat,?y:Nat))))",
"plus(times(plus(?y:Nat,plus(?x:Nat,?y:Nat)),?x:Nat),?x:Nat) = times(s(plus(?z:Nat,s(?y:Nat))),times(plus(?z:Nat,s(?z:Nat)),s(?x:Nat)))",
"plus(times(plus(?z:Nat,times(?y:Nat,?x:Nat)),?z:Nat),?y:Nat) = plus(s(s(times(?z:Nat,?x:Nat))),plus(plus(?x:Nat,plus(?y:Nat,?x:Nat)),?z:Nat))",
"times(?y:Nat,times(?x:Nat,times(times(?x:Nat,?x:Nat),?y:Nat))) = times(times(times(?z:Nat,plus(?x:Nat,?z:Nat)),plus(?y:Nat,times(?z:Nat,?x:Nat))),?x:Nat)",
"times(s(times(s(?z:Nat),times(?y:Nat,?x:Nat))),?z:Nat) = times(plus(?y:Nat,times(times(?x:Nat,?y:Nat),?y:Nat)),plus(s(s(?y:Nat)),?y:Nat))",
"plus(?x:Nat,plus(?z:Nat,times(plus(?y:Nat,?y:Nat),s(?x:Nat)))) = plus(times(?x:Nat,?x:Nat),times(?x:Nat,plus(s(?y:Nat),times(?y:Nat,?z:Nat))))",
"plus(s(plus(times(?z:Nat,?x:Nat),times(?z:Nat,?z:Nat))),?z:Nat) = plus(times(s(?x:Nat),times(plus(?z:Nat,?y:Nat),times(?z:Nat,?z:Nat))),?y:Nat)",
"plus(s(plus(times(?z:Nat,?z:Nat),times(?x:Nat,?y:Nat))),?x:Nat) = times(plus(plus(s(?x:Nat),s(?x:Nat)),?y:Nat),plus(s(?y:Nat),?z:Nat))",
"times(s(times(?z:Nat,times(?z:Nat,?z:Nat))),times(?z:Nat,?x:Nat)) = plus(plus(times(s(?y:Nat),times(?y:Nat,?y:Nat)),?y:Nat),s(s(?z:Nat)))",
"times(plus(?y:Nat,?z:Nat),s(times(plus(?y:Nat,?x:Nat),?x:Nat))) = times(s(times(?x:Nat,times(?x:Nat,?z:Nat))),s(s(plus(?z:Nat,?x:Nat))))",
"times(plus(times(?z:Nat,?z:Nat),?x:Nat),times(s(?z:Nat),?z:Nat)) = times(?z:Nat,times(times(s(?x:Nat),times(?y:Nat,?x:Nat)),plus(?x:Nat,?y:Nat)))",
"s(times(times(times(?z:Nat,?z:Nat),s(?z:Nat)),s(s(?x:Nat)))) = plus(times(?x:Nat,times(plus(?y:Nat,?y:Nat),s(?x:Nat))),s(?z:Nat))",
"plus(?y:Nat,plus(s(?z:Nat),plus(s(?y:Nat),times(?z:Nat,?y:Nat)))) = plus(times(plus(s(?x:Nat),plus(?y:Nat,?z:Nat)),s(?x:Nat)),?y:Nat)",
"times(plus(?z:Nat,plus(times(?x:Nat,?y:Nat),plus(?z:Nat,?y:Nat))),?z:Nat) = s(plus(plus(s(?x:Nat),s(?z:Nat)),times(s(?z:Nat),?z:Nat)))",
"times(plus(s(s(?y:Nat)),times(?y:Nat,times(?z:Nat,?x:Nat))),?y:Nat) = s(times(times(s(?z:Nat),times(?y:Nat,?x:Nat)),times(?y:Nat,?x:Nat)))",
"times(times(?x:Nat,plus(?y:Nat,plus(?z:Nat,?y:Nat))),plus(?y:Nat,?y:Nat)) = times(times(s(plus(?y:Nat,?y:Nat)),plus(?z:Nat,s(?x:Nat))),?x:Nat)",
"s(times(s(s(?y:Nat)),times(times(?y:Nat,?x:Nat),plus(?z:Nat,?x:Nat)))) = plus(plus(plus(plus(?x:Nat,?z:Nat),?x:Nat),?y:Nat),s(?z:Nat))",
"plus(plus(?z:Nat,?y:Nat),plus(plus(?z:Nat,s(?x:Nat)),s(s(?y:Nat)))) = times(plus(?y:Nat,plus(s(?z:Nat),times(?z:Nat,?y:Nat))),?x:Nat)",
"times(?y:Nat,times(s(plus(?x:Nat,?x:Nat)),plus(times(?x:Nat,?x:Nat),?y:Nat))) = s(plus(times(times(?z:Nat,?x:Nat),?x:Nat),s(s(?y:Nat))))",
"times(s(?z:Nat),times(s(times(?x:Nat,?y:Nat)),s(times(?y:Nat,?y:Nat)))) = times(?x:Nat,plus(s(times(?z:Nat,?x:Nat)),times(?x:Nat,?z:Nat)))",
"times(plus(?y:Nat,plus(times(?y:Nat,?x:Nat),?y:Nat)),plus(s(?x:Nat),?x:Nat)) = plus(plus(plus(?x:Nat,plus(?z:Nat,?z:Nat)),s(?y:Nat)),?z:Nat)",
"times(plus(s(plus(?y:Nat,?y:Nat)),plus(?z:Nat,?z:Nat)),plus(?y:Nat,?z:Nat)) = plus(?x:Nat,times(plus(?x:Nat,plus(?y:Nat,?x:Nat)),s(?z:Nat)))",
"times(plus(times(?z:Nat,plus(?x:Nat,?y:Nat)),plus(?x:Nat,?z:Nat)),s(?z:Nat)) = times(plus(s(?z:Nat),s(s(?x:Nat))),plus(?z:Nat,?z:Nat))",
"s(times(times(?z:Nat,plus(?y:Nat,?x:Nat)),plus(s(?z:Nat),times(?y:Nat,?z:Nat)))) = plus(?y:Nat,plus(s(?x:Nat),s(plus(?x:Nat,?y:Nat))))",
"s(times(times(s(?z:Nat),plus(?y:Nat,?x:Nat)),plus(plus(?y:Nat,?x:Nat),?y:Nat))) = times(s(?y:Nat),plus(?z:Nat,times(?y:Nat,s(?z:Nat))))",
"plus(s(s(?y:Nat)),plus(s(s(?x:Nat)),times(times(?x:Nat,?y:Nat),?x:Nat))) = s(times(plus(?x:Nat,?x:Nat),s(plus(?y:Nat,?y:Nat))))",
"plus(plus(s(?z:Nat),plus(s(?x:Nat),?z:Nat)),s(s(times(?x:Nat,?y:Nat)))) = plus(s(times(times(?x:Nat,?y:Nat),s(?y:Nat))),?x:Nat)",
"plus(plus(times(plus(?x:Nat,?y:Nat),s(?x:Nat)),?y:Nat),s(s(s(?y:Nat)))) = plus(times(?x:Nat,s(s(?y:Nat))),plus(?z:Nat,?z:Nat))",
"times(?z:Nat,times(times(?y:Nat,?y:Nat),times(plus(?y:Nat,?y:Nat),plus(?y:Nat,?z:Nat)))) = times(plus(s(?x:Nat),s(times(?z:Nat,?z:Nat))),?y:Nat)",
"times(s(times(?x:Nat,?y:Nat)),times(times(s(?z:Nat),times(?x:Nat,?y:Nat)),?x:Nat)) = times(s(?x:Nat),plus(?y:Nat,s(times(?x:Nat,?x:Nat))))",
"times(plus(plus(plus(?y:Nat,?z:Nat),?y:Nat),?x:Nat),plus(s(s(?z:Nat)),?z:Nat)) = plus(times(?x:Nat,times(s(?z:Nat),s(?y:Nat))),?y:Nat)",
"times(times(?y:Nat,times(?y:Nat,times(?x:Nat,?x:Nat))),s(s(times(?x:Nat,?y:Nat)))) = s(s(times(times(?x:Nat,?z:Nat),times(?y:Nat,?x:Nat))))",
"times(times(times(times(?x:Nat,?z:Nat),?x:Nat),plus(times(?z:Nat,?x:Nat),?z:Nat)),?y:Nat) = plus(s(?y:Nat),s(plus(plus(?z:Nat,?z:Nat),?x:Nat)))",
"plus(s(s(times(?y:Nat,?x:Nat))),plus(plus(?z:Nat,?z:Nat),s(times(?z:Nat,?z:Nat)))) = plus(s(times(plus(?z:Nat,?y:Nat),?x:Nat)),?y:Nat)",
"plus(plus(s(times(?y:Nat,?y:Nat)),times(plus(?y:Nat,?x:Nat),times(?y:Nat,?x:Nat))),?y:Nat) = times(s(?y:Nat),plus(s(s(?z:Nat)),?x:Nat))",
"times(?x:Nat,plus(plus(plus(?z:Nat,?y:Nat),s(?y:Nat)),plus(times(?y:Nat,?z:Nat),?z:Nat))) = plus(?x:Nat,plus(?y:Nat,plus(s(?y:Nat),?x:Nat)))",
"times(?x:Nat,plus(times(s(?x:Nat),times(?y:Nat,?z:Nat)),times(?x:Nat,plus(?x:Nat,?y:Nat)))) = plus(plus(plus(?x:Nat,?x:Nat),s(?z:Nat)),?z:Nat)",
"times(?x:Nat,times(plus(plus(?x:Nat,?y:Nat),plus(?z:Nat,?z:Nat)),s(plus(?y:Nat,?y:Nat)))) = times(s(plus(s(?x:Nat),?y:Nat)),s(?x:Nat))",
"times(plus(times(plus(?z:Nat,?z:Nat),?z:Nat),plus(s(?y:Nat),times(?z:Nat,?x:Nat))),?z:Nat) = s(times(times(s(?x:Nat),s(?z:Nat)),?z:Nat))",
"times(times(?y:Nat,times(?y:Nat,?z:Nat)),plus(?y:Nat,plus(plus(?x:Nat,?z:Nat),s(?x:Nat)))) = s(times(plus(?y:Nat,times(?x:Nat,?y:Nat)),?y:Nat))",
"plus(s(?x:Nat),plus(times(s(?x:Nat),times(?y:Nat,?x:Nat)),times(?y:Nat,times(?y:Nat,?y:Nat)))) = s(s(plus(s(?x:Nat),s(?z:Nat))))",
"times(plus(times(s(?x:Nat),?x:Nat),?x:Nat),times(plus(plus(?x:Nat,?z:Nat),s(?z:Nat)),?x:Nat)) = s(s(times(plus(?z:Nat,?y:Nat),?x:Nat)))",
"plus(times(?y:Nat,plus(?z:Nat,plus(?z:Nat,?x:Nat))),plus(plus(plus(?z:Nat,?z:Nat),plus(?y:Nat,?z:Nat)),?y:Nat)) = s(times(s(?x:Nat),?x:Nat))",
"times(times(?y:Nat,?y:Nat),times(plus(plus(?z:Nat,?z:Nat),s(?y:Nat)),plus(times(?z:Nat,?x:Nat),s(?y:Nat)))) = times(plus(?y:Nat,?z:Nat),?z:Nat)",
"times(s(plus(?x:Nat,?x:Nat)),plus(times(s(?y:Nat),plus(?y:Nat,?x:Nat)),plus(s(?z:Nat),plus(?z:Nat,?z:Nat)))) = times(?x:Nat,s(?z:Nat))",
"times(s(times(?x:Nat,plus(?x:Nat,?x:Nat))),times(plus(times(?x:Nat,?z:Nat),times(?y:Nat,?y:Nat)),plus(?z:Nat,?z:Nat))) = s(plus(?z:Nat,?y:Nat))",
"times(plus(?y:Nat,times(s(?z:Nat),?x:Nat)),times(s(times(?z:Nat,?x:Nat)),plus(times(?y:Nat,?y:Nat),s(?x:Nat)))) = plus(?x:Nat,s(?x:Nat))",
"times(plus(times(?x:Nat,times(?x:Nat,?y:Nat)),?x:Nat),plus(times(s(?x:Nat),times(?y:Nat,?y:Nat)),s(s(?x:Nat)))) = s(s(s(?y:Nat)))",
"times(times(s(times(?x:Nat,?x:Nat)),?x:Nat),times(times(s(?x:Nat),s(?z:Nat)),plus(plus(?x:Nat,?z:Nat),?y:Nat))) = times(?x:Nat,s(?x:Nat))",
"times(s(plus(times(?y:Nat,?z:Nat),plus(?y:Nat,?z:Nat))),times(times(?y:Nat,plus(?x:Nat,?x:Nat)),s(plus(?z:Nat,?x:Nat)))) = plus(?z:Nat,?z:Nat)",
"times(plus(?z:Nat,plus(s(?z:Nat),?x:Nat)),plus(plus(times(?z:Nat,?y:Nat),plus(?x:Nat,?x:Nat)),s(times(?z:Nat,?z:Nat)))) = times(?y:Nat,?z:Nat)",
"times(times(times(?x:Nat,plus(?x:Nat,?y:Nat)),times(s(?z:Nat),times(?z:Nat,?y:Nat))),plus(?z:Nat,plus(s(?y:Nat),?x:Nat))) = plus(?z:Nat,?z:Nat)",
"plus(s(times(plus(?z:Nat,?z:Nat),times(?z:Nat,?x:Nat))),times(times(?y:Nat,s(?x:Nat)),plus(plus(?y:Nat,?y:Nat),s(?y:Nat)))) = s(?z:Nat)",
"plus(plus(plus(s(?z:Nat),times(?z:Nat,?y:Nat)),times(?y:Nat,plus(?y:Nat,?z:Nat))),s(plus(s(?z:Nat),plus(?z:Nat,?z:Nat)))) = s(?z:Nat)",
"s(?y:Nat) = times(times(plus(s(?x:Nat),plus(?y:Nat,?z:Nat)),s(times(?z:Nat,?x:Nat))),plus(s(times(?y:Nat,?y:Nat)),s(times(?x:Nat,?x:Nat))))",
"s(plus(?x:Nat,?y:Nat)) = times(plus(times(times(?y:Nat,?x:Nat),?y:Nat),plus(plus(?z:Nat,?x:Nat),s(?x:Nat))),plus(times(?x:Nat,s(?y:Nat)),?y:Nat))",
"s(plus(?y:Nat,?y:Nat)) = plus(plus(times(?z:Nat,times(?z:Nat,?y:Nat)),?x:Nat),plus(times(plus(?x:Nat,?x:Nat),?y:Nat),plus(plus(?y:Nat,?x:Nat),?z:Nat)))",
"times(s(s(?x:Nat)),s(?x:Nat)) = times(times(times(?y:Nat,s(?x:Nat)),plus(?x:Nat,plus(?y:Nat,?y:Nat))),s(plus(?y:Nat,times(?z:Nat,?z:Nat))))",
"s(s(times(times(?y:Nat,?y:Nat),?x:Nat))) = times(?z:Nat,plus(plus(plus(?x:Nat,?z:Nat),plus(?x:Nat,?x:Nat)),times(s(?z:Nat),plus(?y:Nat,?z:Nat))))",
"times(times(s(s(?y:Nat)),?y:Nat),?z:Nat) = times(times(plus(?z:Nat,plus(?y:Nat,?z:Nat)),s(?y:Nat)),times(s(?z:Nat),s(plus(?x:Nat,?z:Nat))))",
"s(times(?z:Nat,times(?y:Nat,plus(?x:Nat,?y:Nat)))) = plus(plus(plus(?x:Nat,?z:Nat),times(plus(?x:Nat,?y:Nat),s(?x:Nat))),s(s(s(?x:Nat))))",
"plus(?z:Nat,s(plus(?z:Nat,plus(?z:Nat,?x:Nat)))) = times(?x:Nat,times(plus(s(?x:Nat),plus(?x:Nat,?x:Nat)),plus(s(?x:Nat),times(?z:Nat,?z:Nat))))",
"plus(s(s(plus(?x:Nat,?z:Nat))),s(?x:Nat)) = times(plus(?x:Nat,plus(plus(?x:Nat,?x:Nat),s(?y:Nat))),s(times(?x:Nat,times(?z:Nat,?z:Nat))))",
"plus(s(times(times(?y:Nat,?z:Nat),s(?y:Nat))),?z:Nat) = times(plus(?y:Nat,?x:Nat),times(plus(s(?y:Nat),s(?y:Nat)),s(plus(?z:Nat,?z:Nat))))",
"times(times(?x:Nat,s(plus(?x:Nat,?y:Nat))),s(?z:Nat)) = times(times(s(times(?z:Nat,?y:Nat)),?x:Nat),s(plus(s(?y:Nat),plus(?x:Nat,?y:Nat))))",
"s(times(?y:Nat,times(times(?z:Nat,?x:Nat),times(?x:Nat,?y:Nat)))) = times(s(times(plus(?x:Nat,?z:Nat),?x:Nat)),plus(s(times(?z:Nat,?z:Nat)),?y:Nat))",
"plus(times(?x:Nat,times(s(?z:Nat),plus(?x:Nat,?x:Nat))),?y:Nat) = plus(s(times(s(?y:Nat),?z:Nat)),plus(times(plus(?x:Nat,?y:Nat),?z:Nat),?x:Nat))",
"plus(times(s(times(?z:Nat,?y:Nat)),times(?x:Nat,?z:Nat)),?x:Nat) = plus(plus(?z:Nat,s(times(?x:Nat,?z:Nat))),s(plus(?z:Nat,times(?y:Nat,?y:Nat))))",
"times(?y:Nat,s(plus(times(?z:Nat,?x:Nat),plus(?z:Nat,?z:Nat)))) = times(?y:Nat,times(plus(plus(?x:Nat,?x:Nat),times(?y:Nat,?y:Nat)),plus(?x:Nat,?z:Nat)))",
"times(?y:Nat,plus(times(s(?z:Nat),times(?z:Nat,?x:Nat)),?y:Nat)) = times(s(plus(?z:Nat,plus(?y:Nat,?z:Nat))),times(?x:Nat,s(times(?z:Nat,?y:Nat))))",
"times(plus(?z:Nat,plus(times(?x:Nat,?z:Nat),s(?y:Nat))),?y:Nat) = plus(times(s(?z:Nat),?y:Nat),times(times(times(?x:Nat,?x:Nat),s(?x:Nat)),?z:Nat))",
"times(times(plus(s(?z:Nat),?y:Nat),times(?y:Nat,?z:Nat)),?y:Nat) = plus(s(times(times(?x:Nat,?y:Nat),?y:Nat)),times(?x:Nat,s(plus(?y:Nat,?z:Nat))))",
"s(plus(s(times(?y:Nat,?x:Nat)),times(s(?z:Nat),s(?z:Nat)))) = plus(?y:Nat,plus(s(plus(?z:Nat,?x:Nat)),times(?y:Nat,plus(?x:Nat,?y:Nat))))",
"s(plus(s(times(?y:Nat,?y:Nat)),times(?z:Nat,plus(?x:Nat,?y:Nat)))) = plus(s(times(times(?z:Nat,?z:Nat),?z:Nat)),s(times(s(?x:Nat),?z:Nat)))",
"s(times(times(times(?z:Nat,?z:Nat),?z:Nat),plus(?y:Nat,s(?z:Nat)))) = times(times(s(?x:Nat),times(plus(?y:Nat,?y:Nat),plus(?x:Nat,?y:Nat))),?z:Nat)",
"plus(?x:Nat,times(times(plus(?y:Nat,?z:Nat),s(?y:Nat)),s(?x:Nat))) = s(plus(s(times(?x:Nat,?z:Nat)),plus(plus(?z:Nat,?z:Nat),s(?z:Nat))))",
"plus(s(s(times(?y:Nat,?x:Nat))),s(s(times(?y:Nat,?x:Nat)))) = times(times(plus(?y:Nat,?x:Nat),plus(s(?y:Nat),s(?z:Nat))),s(?z:Nat))",
"plus(plus(plus(?x:Nat,?x:Nat),s(times(?y:Nat,?y:Nat))),s(?y:Nat)) = s(plus(plus(s(?z:Nat),times(?x:Nat,?x:Nat)),s(times(?x:Nat,?z:Nat))))",
"s(times(times(times(?z:Nat,?z:Nat),s(?z:Nat)),s(times(?z:Nat,?z:Nat)))) = plus(plus(?z:Nat,plus(plus(?z:Nat,?x:Nat),plus(?z:Nat,?x:Nat))),?y:Nat)",
"plus(times(plus(?x:Nat,times(?x:Nat,?y:Nat)),s(plus(?x:Nat,?z:Nat))),?x:Nat) = times(plus(s(times(?x:Nat,?x:Nat)),s(times(?z:Nat,?z:Nat))),?y:Nat)",
"times(s(?z:Nat),plus(s(?x:Nat),plus(s(?x:Nat),times(?y:Nat,?y:Nat)))) = s(plus(plus(times(?y:Nat,?z:Nat),?y:Nat),s(plus(?y:Nat,?z:Nat))))",
"times(plus(s(times(?z:Nat,?x:Nat)),times(?x:Nat,times(?y:Nat,?z:Nat))),?z:Nat) = plus(s(?y:Nat),plus(times(times(?y:Nat,?z:Nat),s(?z:Nat)),?x:Nat))",
"plus(s(plus(times(?x:Nat,?z:Nat),?x:Nat)),times(?z:Nat,s(plus(?x:Nat,?x:Nat)))) = s(plus(times(plus(?z:Nat,?y:Nat),?x:Nat),times(?x:Nat,?x:Nat)))",
"plus(times(?x:Nat,s(?y:Nat)),times(s(plus(?x:Nat,?x:Nat)),s(s(?z:Nat)))) = s(times(?z:Nat,plus(plus(?z:Nat,?y:Nat),times(?x:Nat,?z:Nat))))",
"plus(times(s(?y:Nat),times(plus(?y:Nat,?x:Nat),s(?y:Nat))),times(?y:Nat,?y:Nat)) = times(times(?x:Nat,?z:Nat),times(?x:Nat,s(plus(?y:Nat,?z:Nat))))",
"times(s(?y:Nat),times(s(plus(?y:Nat,?y:Nat)),plus(times(?z:Nat,?y:Nat),?y:Nat))) = times(plus(?z:Nat,?z:Nat),plus(s(?x:Nat),times(?y:Nat,?y:Nat)))",
"times(s(s(s(?z:Nat))),plus(s(plus(?x:Nat,?z:Nat)),s(s(?x:Nat)))) = s(times(s(times(?z:Nat,?y:Nat)),times(s(?z:Nat),?z:Nat)))",
"times(plus(plus(times(?z:Nat,?x:Nat),times(?y:Nat,?y:Nat)),?z:Nat),times(?x:Nat,?y:Nat)) = s(plus(times(plus(?y:Nat,?y:Nat),plus(?z:Nat,?x:Nat)),?z:Nat))",
"times(times(?x:Nat,times(?x:Nat,plus(?x:Nat,?y:Nat))),s(plus(s(?x:Nat),?z:Nat))) = s(times(times(times(?y:Nat,?z:Nat),?x:Nat),s(s(?x:Nat))))",
"plus(times(s(?x:Nat),times(?y:Nat,s(?x:Nat))),times(s(?x:Nat),times(?z:Nat,?z:Nat))) = plus(times(plus(plus(?z:Nat,?x:Nat),?y:Nat),?z:Nat),?z:Nat)",
"plus(times(s(times(?y:Nat,?z:Nat)),plus(times(?z:Nat,?z:Nat),times(?y:Nat,?z:Nat))),?y:Nat) = times(plus(?x:Nat,?x:Nat),times(s(?y:Nat),s(?z:Nat)))",
"times(s(plus(s(?y:Nat),plus(?x:Nat,?z:Nat))),plus(?x:Nat,times(?y:Nat,times(?x:Nat,?x:Nat)))) = s(plus(s(?x:Nat),plus(s(?y:Nat),?y:Nat)))",
"times(s(times(s(?x:Nat),?z:Nat)),times(plus(times(?y:Nat,?y:Nat),plus(?x:Nat,?y:Nat)),?x:Nat)) = s(plus(plus(times(?z:Nat,?y:Nat),?z:Nat),?y:Nat))",
"times(times(plus(plus(?z:Nat,?x:Nat),s(?y:Nat)),s(?y:Nat)),times(times(?y:Nat,?z:Nat),?y:Nat)) = s(times(?z:Nat,times(plus(?z:Nat,?x:Nat),?z:Nat)))",
"times(plus(plus(plus(?x:Nat,?y:Nat),plus(?y:Nat,?z:Nat)),times(plus(?x:Nat,?y:Nat),?y:Nat)),s(?z:Nat)) = s(s(plus(times(?z:Nat,?x:Nat),?y:Nat)))",
"times(times(s(?y:Nat),s(plus(?x:Nat,?y:Nat))),plus(s(s(?x:Nat)),times(s(?z:Nat),?x:Nat))) = s(s(plus(?z:Nat,times(?z:Nat,?y:Nat))))",
"plus(plus(s(?z:Nat),plus(?z:Nat,plus(?z:Nat,?x:Nat))),times(plus(?y:Nat,?y:Nat),times(plus(?z:Nat,?y:Nat),s(?y:Nat)))) = plus(s(?z:Nat),?y:Nat)",
"plus(plus(s(times(?z:Nat,?y:Nat)),times(times(?y:Nat,?y:Nat),times(?z:Nat,?x:Nat))),s(plus(plus(?x:Nat,?z:Nat),s(?y:Nat)))) = plus(?x:Nat,?z:Nat)",
"plus(plus(times(s(?x:Nat),?y:Nat),times(plus(?y:Nat,?z:Nat),s(?x:Nat))),plus(times(plus(?y:Nat,?y:Nat),times(?x:Nat,?x:Nat)),?z:Nat)) = s(?y:Nat)",
"plus(plus(times(plus(?z:Nat,?y:Nat),plus(?y:Nat,?y:Nat)),plus(times(?z:Nat,?y:Nat),?z:Nat)),plus(?y:Nat,times(times(?z:Nat,?x:Nat),?y:Nat))) = s(?z:Nat)",
"plus(times(s(s(?x:Nat)),times(times(?z:Nat,?z:Nat),plus(?y:Nat,?x:Nat))),times(?z:Nat,plus(plus(?x:Nat,?y:Nat),plus(?z:Nat,?z:Nat)))) = s(?y:Nat)",
"times(times(times(s(?z:Nat),?y:Nat),?y:Nat),plus(times(times(?x:Nat,?y:Nat),plus(?x:Nat,?x:Nat)),times(times(?z:Nat,?y:Nat),s(?x:Nat)))) = s(?z:Nat)",
"s(?y:Nat) = plus(plus(times(plus(?y:Nat,?x:Nat),?y:Nat),s(times(?y:Nat,?y:Nat))),plus(times(plus(?y:Nat,?z:Nat),times(?y:Nat,?x:Nat)),s(s(?y:Nat))))",
"s(?y:Nat) = times(times(times(times(?x:Nat,?z:Nat),?y:Nat),times(times(?y:Nat,?x:Nat),plus(?x:Nat,?y:Nat))),times(?x:Nat,times(s(?z:Nat),plus(?z:Nat,?x:Nat))))",
"s(?z:Nat) = times(times(plus(plus(?z:Nat,?y:Nat),s(?y:Nat)),s(times(?z:Nat,?z:Nat))),plus(s(times(?x:Nat,?z:Nat)),plus(plus(?x:Nat,?y:Nat),?y:Nat)))",
"plus(?y:Nat,?x:Nat) = times(times(times(plus(?z:Nat,?z:Nat),plus(?y:Nat,?y:Nat)),s(s(?y:Nat))),times(plus(plus(?z:Nat,?z:Nat),plus(?x:Nat,?y:Nat)),?x:Nat))",
"plus(?z:Nat,?x:Nat) = plus(s(plus(times(?x:Nat,?x:Nat),times(?x:Nat,?x:Nat))),plus(times(times(?x:Nat,?y:Nat),times(?z:Nat,?y:Nat)),s(times(?x:Nat,?z:Nat))))",
"times(?y:Nat,?y:Nat) = plus(plus(?x:Nat,s(times(?z:Nat,?x:Nat))),times(plus(s(?x:Nat),plus(?y:Nat,?x:Nat)),times(plus(?y:Nat,?y:Nat),plus(?z:Nat,?z:Nat))))",
"s(times(?y:Nat,?x:Nat)) = times(times(plus(plus(?z:Nat,?y:Nat),plus(?z:Nat,?x:Nat)),plus(s(?x:Nat),?x:Nat)),times(s(?y:Nat),s(plus(?y:Nat,?z:Nat))))",
"plus(?y:Nat,s(?y:Nat)) = plus(times(s(s(?y:Nat)),s(times(?z:Nat,?x:Nat))),plus(times(s(?y:Nat),times(?z:Nat,?x:Nat)),s(times(?y:Nat,?z:Nat))))",
"times(?x:Nat,plus(?z:Nat,?x:Nat)) = plus(times(plus(s(?y:Nat),times(?y:Nat,?x:Nat)),s(s(?x:Nat))),times(?x:Nat,plus(s(?y:Nat),times(?y:Nat,?x:Nat))))",
"s(times(times(?z:Nat,?z:Nat),?x:Nat)) = plus(times(?x:Nat,s(times(?y:Nat,?y:Nat))),times(times(?x:Nat,plus(?z:Nat,?z:Nat)),times(s(?z:Nat),s(?y:Nat))))",
"plus(?z:Nat,s(s(times(?y:Nat,?y:Nat)))) = times(times(times(s(?z:Nat),?x:Nat),times(times(?x:Nat,?z:Nat),?y:Nat)),plus(?x:Nat,s(times(?x:Nat,?z:Nat))))",
"times(?y:Nat,s(s(times(?x:Nat,?x:Nat)))) = plus(s(times(?z:Nat,times(?x:Nat,?y:Nat))),plus(s(times(?z:Nat,?z:Nat)),plus(?y:Nat,plus(?y:Nat,?x:Nat))))",
"times(times(s(s(?y:Nat)),?z:Nat),?x:Nat) = times(plus(times(?z:Nat,s(?x:Nat)),?z:Nat),times(plus(plus(?y:Nat,?z:Nat),s(?x:Nat)),s(s(?x:Nat))))",
"s(s(times(plus(?y:Nat,?y:Nat),s(?z:Nat)))) = plus(times(?x:Nat,plus(times(?x:Nat,?y:Nat),s(?y:Nat))),plus(?x:Nat,times(times(?x:Nat,?x:Nat),?y:Nat)))",
"plus(times(s(plus(?x:Nat,?y:Nat)),?x:Nat),?y:Nat) = plus(times(plus(s(?z:Nat),?x:Nat),?y:Nat),plus(times(plus(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),?y:Nat))",
"times(plus(?z:Nat,s(times(?z:Nat,?x:Nat))),?y:Nat) = times(s(s(s(?z:Nat))),plus(times(?x:Nat,plus(?x:Nat,?z:Nat)),plus(plus(?y:Nat,?z:Nat),?x:Nat)))",
"plus(?z:Nat,plus(plus(?y:Nat,?z:Nat),s(s(?x:Nat)))) = s(times(plus(plus(?z:Nat,?y:Nat),plus(?x:Nat,?x:Nat)),times(times(?z:Nat,?y:Nat),s(?y:Nat))))",
"plus(times(s(?z:Nat),s(times(?y:Nat,?y:Nat))),?x:Nat) = plus(times(plus(?y:Nat,s(?x:Nat)),plus(s(?y:Nat),plus(?y:Nat,?y:Nat))),times(?z:Nat,?x:Nat))",
"plus(times(s(plus(?x:Nat,?x:Nat)),?y:Nat),s(?y:Nat)) = times(s(s(s(?y:Nat))),times(s(?y:Nat),times(plus(?x:Nat,?y:Nat),times(?x:Nat,?x:Nat))))",
"plus(times(plus(plus(?y:Nat,?y:Nat),?y:Nat),?x:Nat),?x:Nat) = plus(plus(?x:Nat,s(plus(?y:Nat,?y:Nat))),plus(s(s(?y:Nat)),times(?x:Nat,s(?x:Nat))))",
"times(s(?y:Nat),s(times(s(?z:Nat),s(?y:Nat)))) = times(times(s(plus(?z:Nat,?x:Nat)),times(s(?x:Nat),s(?z:Nat))),s(plus(?x:Nat,?y:Nat)))",
"times(s(?y:Nat),times(s(?x:Nat),s(s(?y:Nat)))) = plus(plus(plus(s(?y:Nat),plus(?x:Nat,?y:Nat)),times(times(?y:Nat,?x:Nat),s(?z:Nat))),?z:Nat)",
"s(plus(times(plus(?z:Nat,?x:Nat),times(?y:Nat,?y:Nat)),?z:Nat)) = plus(s(times(plus(?z:Nat,?y:Nat),?z:Nat)),times(plus(?z:Nat,times(?z:Nat,?z:Nat)),?y:Nat))",
"s(times(times(plus(?x:Nat,?z:Nat),times(?x:Nat,?y:Nat)),?y:Nat)) = plus(s(?y:Nat),times(times(times(?y:Nat,?z:Nat),s(?z:Nat)),times(?y:Nat,s(?y:Nat))))",
"plus(times(?y:Nat,times(times(?y:Nat,?z:Nat),s(?x:Nat))),?z:Nat) = plus(s(s(s(?y:Nat))),times(s(s(?y:Nat)),plus(?y:Nat,times(?y:Nat,?z:Nat))))",
"times(?z:Nat,plus(s(times(?z:Nat,?x:Nat)),times(?z:Nat,?z:Nat))) = times(s(s(?z:Nat)),times(plus(times(?z:Nat,?z:Nat),?z:Nat),s(times(?z:Nat,?x:Nat))))",
"times(s(times(times(?z:Nat,?z:Nat),s(?x:Nat))),s(?x:Nat)) = plus(s(times(s(?z:Nat),s(?z:Nat))),s(plus(plus(?y:Nat,?y:Nat),s(?y:Nat))))",
"times(plus(?z:Nat,?x:Nat),times(s(?x:Nat),s(s(?z:Nat)))) = s(plus(plus(?y:Nat,times(?z:Nat,?x:Nat)),times(plus(?z:Nat,?x:Nat),times(?z:Nat,?y:Nat))))",
"times(plus(s(?z:Nat),?y:Nat),s(plus(s(?y:Nat),?y:Nat))) = times(times(s(times(?y:Nat,?x:Nat)),times(?y:Nat,plus(?x:Nat,?y:Nat))),plus(?x:Nat,?x:Nat))",
"times(plus(plus(s(?x:Nat),plus(?x:Nat,?x:Nat)),?x:Nat),?z:Nat) = plus(?y:Nat,plus(times(plus(?x:Nat,?x:Nat),s(?x:Nat)),times(?x:Nat,plus(?x:Nat,?y:Nat))))",
"times(plus(plus(plus(?z:Nat,?y:Nat),?y:Nat),s(?x:Nat)),?y:Nat) = plus(plus(?y:Nat,times(s(?x:Nat),times(?z:Nat,?y:Nat))),times(?z:Nat,plus(?z:Nat,?x:Nat)))",
"times(times(s(?x:Nat),?y:Nat),s(s(times(?z:Nat,?z:Nat)))) = plus(?y:Nat,times(plus(plus(?y:Nat,?x:Nat),times(?z:Nat,?x:Nat)),times(s(?z:Nat),?z:Nat)))",
"s(times(plus(?z:Nat,plus(?y:Nat,?x:Nat)),s(plus(?z:Nat,?x:Nat)))) = times(times(?y:Nat,s(s(?x:Nat))),times(times(?x:Nat,?x:Nat),s(s(?z:Nat))))",
"plus(plus(s(?y:Nat),s(?x:Nat)),times(?y:Nat,s(s(?y:Nat)))) = plus(times(plus(?y:Nat,?y:Nat),?y:Nat),plus(?z:Nat,times(s(?x:Nat),s(?z:Nat))))",
"plus(times(s(?x:Nat),plus(?z:Nat,?x:Nat)),s(times(?z:Nat,?x:Nat))) = times(s(times(times(?y:Nat,?x:Nat),?x:Nat)),times(s(times(?z:Nat,?z:Nat)),?y:Nat))",
"plus(times(times(times(?x:Nat,?z:Nat),plus(?z:Nat,?y:Nat)),?y:Nat),?z:Nat) = plus(plus(s(plus(?x:Nat,?x:Nat)),?y:Nat),s(plus(times(?z:Nat,?z:Nat),?x:Nat)))",
"times(plus(times(?z:Nat,s(?y:Nat)),s(times(?z:Nat,?x:Nat))),?y:Nat) = plus(?x:Nat,plus(plus(?z:Nat,s(?y:Nat)),times(s(?z:Nat),plus(?z:Nat,?z:Nat))))",
"times(times(s(times(?z:Nat,?y:Nat)),s(s(?y:Nat))),s(?y:Nat)) = times(plus(plus(?z:Nat,s(?x:Nat)),?x:Nat),times(s(times(?y:Nat,?z:Nat)),?z:Nat))",
"plus(plus(plus(times(?y:Nat,?y:Nat),?x:Nat),s(plus(?z:Nat,?x:Nat))),?y:Nat) = plus(s(s(times(?y:Nat,?x:Nat))),s(times(?y:Nat,plus(?x:Nat,?y:Nat))))",
"times(?y:Nat,times(plus(?x:Nat,s(?x:Nat)),plus(plus(?z:Nat,?x:Nat),?z:Nat))) = s(times(plus(plus(?y:Nat,?y:Nat),s(?y:Nat)),times(s(?x:Nat),?z:Nat)))",
"times(s(?y:Nat),plus(times(times(?z:Nat,?z:Nat),plus(?z:Nat,?z:Nat)),?z:Nat)) = s(times(times(s(?z:Nat),?x:Nat),plus(times(?x:Nat,?z:Nat),s(?x:Nat))))",
"times(plus(plus(?x:Nat,times(?z:Nat,?z:Nat)),times(?y:Nat,?z:Nat)),s(?x:Nat)) = plus(plus(times(times(?y:Nat,?x:Nat),plus(?x:Nat,?z:Nat)),?z:Nat),s(?y:Nat))",
"times(times(plus(?y:Nat,times(?z:Nat,?z:Nat)),s(times(?x:Nat,?x:Nat))),?z:Nat) = times(plus(times(times(?z:Nat,?x:Nat),?z:Nat),s(plus(?y:Nat,?y:Nat))),?x:Nat)",
"times(times(plus(s(?y:Nat),s(?y:Nat)),s(times(?y:Nat,?y:Nat))),?z:Nat) = times(s(?x:Nat),plus(?y:Nat,times(times(?x:Nat,?z:Nat),plus(?y:Nat,?y:Nat))))",
"plus(plus(plus(?z:Nat,plus(?x:Nat,?z:Nat)),?z:Nat),s(s(times(?x:Nat,?z:Nat)))) = times(plus(?z:Nat,?y:Nat),s(plus(times(?x:Nat,?x:Nat),s(?y:Nat))))",
"plus(plus(times(plus(?x:Nat,?y:Nat),times(?x:Nat,?x:Nat)),s(s(?z:Nat))),?z:Nat) = s(plus(s(s(?x:Nat)),times(plus(?y:Nat,?y:Nat),s(?y:Nat))))",
"plus(times(?z:Nat,plus(s(?x:Nat),times(?y:Nat,?y:Nat))),times(s(?z:Nat),?z:Nat)) = plus(s(?x:Nat),times(s(s(?x:Nat)),s(times(?x:Nat,?z:Nat))))",
"plus(times(plus(times(?y:Nat,?y:Nat),?z:Nat),plus(?y:Nat,plus(?y:Nat,?z:Nat))),?x:Nat) = plus(plus(?y:Nat,plus(times(?y:Nat,?y:Nat),plus(?z:Nat,?x:Nat))),?y:Nat)",
"times(plus(times(times(?z:Nat,?y:Nat),?z:Nat),plus(times(?z:Nat,?x:Nat),?y:Nat)),?z:Nat) = s(plus(s(times(?z:Nat,?y:Nat)),plus(plus(?y:Nat,?z:Nat),?x:Nat)))",
"times(times(s(times(?x:Nat,?y:Nat)),plus(s(?x:Nat),plus(?y:Nat,?x:Nat))),?z:Nat) = times(s(plus(plus(?y:Nat,?z:Nat),?z:Nat)),plus(?z:Nat,s(?x:Nat)))",
"times(times(plus(s(?z:Nat),times(?x:Nat,?z:Nat)),times(s(?x:Nat),?z:Nat)),?x:Nat) = plus(times(?z:Nat,?x:Nat),times(?z:Nat,plus(?y:Nat,times(?x:Nat,?y:Nat))))",
"plus(?x:Nat,plus(times(s(?z:Nat),plus(?x:Nat,?z:Nat)),times(times(?y:Nat,?x:Nat),?y:Nat))) = plus(plus(s(?x:Nat),?x:Nat),plus(plus(?z:Nat,?z:Nat),?x:Nat))",
"plus(plus(?x:Nat,plus(times(?y:Nat,?y:Nat),plus(?x:Nat,?y:Nat))),times(s(?y:Nat),?y:Nat)) = s(plus(times(times(?y:Nat,?y:Nat),times(?x:Nat,?z:Nat)),?x:Nat))",
"plus(times(?z:Nat,times(times(?y:Nat,?x:Nat),?y:Nat)),plus(times(s(?z:Nat),?z:Nat),?x:Nat)) = plus(s(times(times(?y:Nat,?y:Nat),times(?y:Nat,?z:Nat))),?y:Nat)",
"times(plus(s(times(?x:Nat,?z:Nat)),plus(plus(?x:Nat,?z:Nat),s(?z:Nat))),s(?x:Nat)) = plus(s(?z:Nat),s(times(times(?y:Nat,?y:Nat),s(?y:Nat))))",
"s(times(plus(s(?y:Nat),times(?y:Nat,?y:Nat)),times(times(?y:Nat,?x:Nat),plus(?z:Nat,?y:Nat)))) = plus(?x:Nat,s(times(times(?z:Nat,?x:Nat),s(?y:Nat))))",
"plus(plus(?y:Nat,?y:Nat),times(times(?y:Nat,times(?z:Nat,?z:Nat)),plus(?x:Nat,plus(?x:Nat,?z:Nat)))) = plus(plus(?x:Nat,?z:Nat),s(plus(?z:Nat,s(?y:Nat))))",
"plus(plus(s(times(?y:Nat,?z:Nat)),times(s(?x:Nat),?z:Nat)),s(s(times(?x:Nat,?y:Nat)))) = plus(s(?y:Nat),s(times(?x:Nat,times(?x:Nat,?z:Nat))))",
"plus(plus(plus(?z:Nat,plus(?x:Nat,?y:Nat)),plus(plus(?y:Nat,?y:Nat),plus(?y:Nat,?z:Nat))),?z:Nat) = times(?y:Nat,plus(times(?y:Nat,times(?z:Nat,?z:Nat)),?z:Nat))",
"plus(plus(plus(s(?x:Nat),?x:Nat),plus(plus(?x:Nat,?z:Nat),?z:Nat)),s(times(?z:Nat,?x:Nat))) = s(plus(times(times(?x:Nat,?y:Nat),?y:Nat),s(?y:Nat)))",
"plus(times(times(times(?z:Nat,?y:Nat),?y:Nat),plus(times(?x:Nat,?y:Nat),times(?x:Nat,?y:Nat))),?x:Nat) = times(s(plus(s(?y:Nat),plus(?z:Nat,?z:Nat))),?z:Nat)",
"s(plus(plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?y:Nat)),times(times(?y:Nat,?z:Nat),times(?y:Nat,?z:Nat)))) = times(?y:Nat,plus(?y:Nat,times(s(?z:Nat),?x:Nat)))",
"times(?z:Nat,times(plus(times(?z:Nat,?x:Nat),plus(?y:Nat,?y:Nat)),plus(plus(?z:Nat,?x:Nat),s(?x:Nat)))) = s(s(plus(s(?y:Nat),times(?x:Nat,?x:Nat))))",
"plus(?x:Nat,plus(plus(times(?y:Nat,?y:Nat),plus(?y:Nat,?y:Nat)),plus(plus(?z:Nat,?y:Nat),plus(?y:Nat,?z:Nat)))) = times(?z:Nat,s(times(?y:Nat,s(?x:Nat))))",
"plus(plus(?z:Nat,?z:Nat),plus(times(plus(?z:Nat,?y:Nat),times(?x:Nat,?x:Nat)),times(times(?y:Nat,?y:Nat),?z:Nat))) = times(plus(plus(?y:Nat,?z:Nat),?x:Nat),?z:Nat)",
"plus(plus(plus(times(?x:Nat,?x:Nat),?z:Nat),s(s(?x:Nat))),times(s(plus(?x:Nat,?z:Nat)),s(s(?x:Nat)))) = s(s(s(times(?z:Nat,?z:Nat))))",
"plus(times(times(times(?y:Nat,?x:Nat),plus(?z:Nat,?z:Nat)),?z:Nat),times(times(s(?x:Nat),?y:Nat),s(s(?y:Nat)))) = s(s(plus(?z:Nat,s(?z:Nat))))",
"plus(times(?y:Nat,times(plus(?z:Nat,?x:Nat),times(?y:Nat,?y:Nat))),plus(s(s(?y:Nat)),times(?z:Nat,plus(?z:Nat,?z:Nat)))) = times(times(?x:Nat,?z:Nat),?z:Nat)",
"times(times(plus(s(?y:Nat),s(?z:Nat)),plus(?x:Nat,?z:Nat)),plus(s(plus(?y:Nat,?x:Nat)),s(plus(?y:Nat,?z:Nat)))) = times(?y:Nat,s(s(?z:Nat)))",
"plus(s(plus(times(?y:Nat,?y:Nat),plus(?y:Nat,?z:Nat))),plus(plus(?x:Nat,?z:Nat),plus(plus(?x:Nat,?y:Nat),times(?y:Nat,?x:Nat)))) = plus(?y:Nat,s(?y:Nat))",
"plus(times(?z:Nat,times(times(?z:Nat,?x:Nat),?z:Nat)),plus(times(?x:Nat,times(?z:Nat,?z:Nat)),plus(s(?z:Nat),plus(?y:Nat,?x:Nat)))) = plus(s(?y:Nat),?y:Nat)",
"times(plus(plus(times(?x:Nat,?y:Nat),plus(?x:Nat,?z:Nat)),?x:Nat),times(plus(?x:Nat,?y:Nat),times(plus(?y:Nat,?y:Nat),s(?y:Nat)))) = s(s(s(?z:Nat)))",
"plus(plus(plus(plus(?y:Nat,?x:Nat),?y:Nat),s(times(?z:Nat,?x:Nat))),plus(times(s(?z:Nat),times(?z:Nat,?y:Nat)),times(?y:Nat,?y:Nat))) = plus(?y:Nat,?z:Nat)",
"plus(times(times(s(?y:Nat),times(?x:Nat,?x:Nat)),plus(plus(?z:Nat,?x:Nat),plus(?z:Nat,?z:Nat))),s(times(s(?z:Nat),s(?y:Nat)))) = times(?y:Nat,?z:Nat)",
"times(plus(s(times(?x:Nat,?y:Nat)),plus(plus(?y:Nat,?x:Nat),plus(?z:Nat,?z:Nat))),s(plus(times(?x:Nat,?y:Nat),plus(?y:Nat,?x:Nat)))) = plus(?y:Nat,?y:Nat)",
"times(plus(plus(times(?x:Nat,?x:Nat),?x:Nat),?x:Nat),times(plus(times(?x:Nat,?x:Nat),s(?x:Nat)),plus(s(?z:Nat),times(?y:Nat,?z:Nat)))) = s(s(?z:Nat))",
"times(plus(times(s(?x:Nat),plus(?x:Nat,?x:Nat)),plus(plus(?x:Nat,?x:Nat),times(?z:Nat,?z:Nat))),s(plus(times(?x:Nat,?y:Nat),?z:Nat))) = s(s(?y:Nat))",
"times(plus(times(plus(?z:Nat,?z:Nat),times(?x:Nat,?z:Nat)),times(?y:Nat,plus(?z:Nat,?z:Nat))),times(?z:Nat,times(plus(?z:Nat,?z:Nat),?z:Nat))) = times(?y:Nat,?y:Nat)",
"times(plus(times(plus(?x:Nat,?x:Nat),plus(?x:Nat,?y:Nat)),plus(plus(?x:Nat,?z:Nat),s(?z:Nat))),s(plus(s(?x:Nat),times(?z:Nat,?y:Nat)))) = s(?z:Nat)",
"plus(plus(?y:Nat,?z:Nat),?x:Nat) = plus(plus(times(plus(?z:Nat,?z:Nat),s(?y:Nat)),times(?x:Nat,plus(?z:Nat,?x:Nat))),s(times(s(?y:Nat),times(?x:Nat,?x:Nat))))",
"s(s(s(times(?x:Nat,?z:Nat)))) = plus(plus(?y:Nat,s(times(?x:Nat,?y:Nat))),plus(plus(plus(?z:Nat,?y:Nat),s(?y:Nat)),plus(times(?x:Nat,?x:Nat),?x:Nat)))",
"plus(?y:Nat,s(s(s(?z:Nat)))) = plus(times(plus(s(?x:Nat),times(?x:Nat,?x:Nat)),plus(s(?z:Nat),?y:Nat)),s(plus(s(?y:Nat),times(?y:Nat,?x:Nat))))",
"times(s(?y:Nat),times(?y:Nat,?z:Nat)) = plus(times(plus(s(?y:Nat),plus(?z:Nat,?z:Nat)),plus(times(?z:Nat,?z:Nat),?x:Nat)),s(times(?x:Nat,times(?x:Nat,?x:Nat))))",
"times(s(s(s(?z:Nat))),?y:Nat) = plus(s(times(plus(?x:Nat,?x:Nat),plus(?z:Nat,?x:Nat))),times(s(s(?x:Nat)),plus(plus(?y:Nat,?y:Nat),s(?y:Nat))))",
"s(s(plus(s(?x:Nat),s(?x:Nat)))) = plus(s(times(?y:Nat,plus(?y:Nat,?z:Nat))),times(plus(plus(?x:Nat,?y:Nat),?x:Nat),times(?x:Nat,plus(?x:Nat,?y:Nat))))",
"s(plus(?x:Nat,s(times(?x:Nat,?z:Nat)))) = plus(times(?z:Nat,times(times(?z:Nat,?y:Nat),s(?y:Nat))),times(plus(?z:Nat,s(?y:Nat)),times(s(?x:Nat),?y:Nat)))",
"s(s(times(times(?x:Nat,?x:Nat),times(?y:Nat,?z:Nat)))) = times(plus(times(plus(?z:Nat,?z:Nat),s(?x:Nat)),times(times(?z:Nat,?y:Nat),plus(?z:Nat,?x:Nat))),?z:Nat)",
"plus(?z:Nat,times(s(?x:Nat),plus(s(?y:Nat),?x:Nat))) = plus(times(plus(plus(?x:Nat,?x:Nat),times(?x:Nat,?y:Nat)),?y:Nat),s(plus(plus(?y:Nat,?x:Nat),?x:Nat)))",
"times(?z:Nat,plus(?y:Nat,plus(s(?z:Nat),s(?z:Nat)))) = times(s(times(times(?x:Nat,?y:Nat),plus(?y:Nat,?y:Nat))),times(?z:Nat,times(plus(?x:Nat,?x:Nat),?x:Nat)))",
"times(plus(times(?y:Nat,times(?x:Nat,?y:Nat)),?z:Nat),?x:Nat) = times(times(plus(times(?y:Nat,?z:Nat),plus(?x:Nat,?y:Nat)),s(times(?x:Nat,?y:Nat))),times(?y:Nat,?y:Nat))",
"s(plus(times(times(?z:Nat,?x:Nat),?z:Nat),times(?y:Nat,?z:Nat))) = times(plus(plus(?x:Nat,times(?x:Nat,?x:Nat)),?x:Nat),s(plus(times(?x:Nat,?y:Nat),s(?x:Nat))))",
"s(times(?y:Nat,plus(plus(?x:Nat,?y:Nat),times(?z:Nat,?x:Nat)))) = plus(s(plus(?x:Nat,times(?y:Nat,?y:Nat))),plus(times(plus(?z:Nat,?x:Nat),s(?y:Nat)),?y:Nat))",
"plus(times(?x:Nat,plus(?y:Nat,s(?y:Nat))),times(?y:Nat,?z:Nat)) = times(plus(plus(?x:Nat,?x:Nat),?x:Nat),plus(plus(plus(?z:Nat,?y:Nat),times(?x:Nat,?z:Nat)),?z:Nat))",
"times(?z:Nat,plus(times(times(?z:Nat,?z:Nat),s(?y:Nat)),?z:Nat)) = s(times(plus(times(?z:Nat,?x:Nat),s(?z:Nat)),times(plus(?y:Nat,?y:Nat),plus(?y:Nat,?x:Nat))))",
"times(times(times(times(?z:Nat,?x:Nat),s(?x:Nat)),?x:Nat),?y:Nat) = times(plus(?z:Nat,s(plus(?y:Nat,?z:Nat))),times(times(?z:Nat,s(?z:Nat)),plus(?x:Nat,?z:Nat)))",
"s(plus(plus(times(?z:Nat,?y:Nat),?y:Nat),s(times(?y:Nat,?x:Nat)))) = times(s(plus(s(?x:Nat),times(?x:Nat,?y:Nat))),times(?x:Nat,plus(?x:Nat,s(?z:Nat))))",
"s(plus(times(?y:Nat,times(?y:Nat,?z:Nat)),s(times(?y:Nat,?x:Nat)))) = plus(plus(?y:Nat,times(s(?y:Nat),times(?x:Nat,?y:Nat))),s(plus(s(?y:Nat),?z:Nat)))",
"plus(plus(?x:Nat,times(plus(?y:Nat,?x:Nat),times(?y:Nat,?x:Nat))),?x:Nat) = times(?y:Nat,times(plus(?z:Nat,times(?y:Nat,?z:Nat)),times(s(?x:Nat),times(?x:Nat,?y:Nat))))",
"plus(plus(s(?x:Nat),times(?y:Nat,times(?x:Nat,?z:Nat))),s(?y:Nat)) = plus(plus(s(plus(?z:Nat,?x:Nat)),plus(times(?y:Nat,?x:Nat),times(?x:Nat,?y:Nat))),?z:Nat)",
"plus(plus(s(times(?x:Nat,?x:Nat)),s(plus(?x:Nat,?y:Nat))),?z:Nat) = plus(times(?y:Nat,times(?x:Nat,s(?z:Nat))),plus(?z:Nat,times(?x:Nat,times(?z:Nat,?x:Nat))))",
"plus(plus(s(times(?y:Nat,?z:Nat)),plus(?y:Nat,s(?x:Nat))),?z:Nat) = times(plus(times(plus(?x:Nat,?x:Nat),?z:Nat),times(plus(?z:Nat,?z:Nat),s(?z:Nat))),?z:Nat)",
"plus(times(?z:Nat,?y:Nat),s(times(s(?z:Nat),times(?z:Nat,?x:Nat)))) = s(times(plus(s(?z:Nat),plus(?x:Nat,?x:Nat)),plus(times(?z:Nat,?y:Nat),s(?z:Nat))))",
"plus(times(plus(times(?z:Nat,?y:Nat),times(?z:Nat,?x:Nat)),?z:Nat),?x:Nat) = plus(plus(plus(times(?z:Nat,?y:Nat),plus(?z:Nat,?z:Nat)),?x:Nat),plus(s(?z:Nat),?y:Nat))",
"times(plus(plus(plus(?y:Nat,?x:Nat),?y:Nat),plus(?y:Nat,?y:Nat)),?y:Nat) = times(times(plus(plus(?y:Nat,?x:Nat),?x:Nat),s(s(?y:Nat))),times(?y:Nat,s(?z:Nat)))",
"plus(times(?z:Nat,?y:Nat),s(plus(times(?z:Nat,?x:Nat),times(?x:Nat,?y:Nat)))) = times(times(times(plus(?x:Nat,?x:Nat),?z:Nat),plus(plus(?y:Nat,?z:Nat),?y:Nat)),?z:Nat)",
"plus(times(plus(?x:Nat,plus(?y:Nat,?y:Nat)),plus(s(?z:Nat),?y:Nat)),?z:Nat) = plus(times(?y:Nat,times(s(?x:Nat),s(?z:Nat))),s(s(times(?x:Nat,?y:Nat))))",
"times(plus(s(s(?x:Nat)),?z:Nat),plus(s(plus(?y:Nat,?y:Nat)),?z:Nat)) = times(s(times(times(?x:Nat,?y:Nat),?x:Nat)),times(?x:Nat,s(times(?y:Nat,?z:Nat))))",
"times(times(plus(s(?x:Nat),s(?x:Nat)),s(s(?y:Nat))),s(?x:Nat)) = s(times(plus(?x:Nat,plus(?z:Nat,?z:Nat)),plus(s(?x:Nat),plus(?z:Nat,?y:Nat))))",
"s(times(times(s(?z:Nat),?y:Nat),plus(plus(?y:Nat,?y:Nat),times(?y:Nat,?y:Nat)))) = plus(s(?z:Nat),plus(s(plus(?y:Nat,?z:Nat)),s(times(?x:Nat,?y:Nat))))",
"s(times(times(plus(?z:Nat,?x:Nat),?y:Nat),plus(plus(?z:Nat,?x:Nat),s(?y:Nat)))) = plus(s(plus(?x:Nat,s(?y:Nat))),s(plus(plus(?z:Nat,?y:Nat),?y:Nat)))",
"plus(times(?x:Nat,s(times(?y:Nat,?y:Nat))),times(times(s(?y:Nat),?y:Nat),?y:Nat)) = plus(times(times(?y:Nat,s(?z:Nat)),?y:Nat),plus(?z:Nat,s(s(?y:Nat))))",
"times(plus(plus(plus(?x:Nat,?x:Nat),times(?y:Nat,?z:Nat)),s(s(?x:Nat))),?z:Nat) = plus(?x:Nat,plus(plus(times(?z:Nat,?x:Nat),?z:Nat),times(s(?x:Nat),?y:Nat)))",
"plus(?x:Nat,times(plus(plus(?x:Nat,?x:Nat),?z:Nat),times(plus(?x:Nat,?x:Nat),s(?y:Nat)))) = times(plus(plus(plus(?x:Nat,?x:Nat),times(?x:Nat,?x:Nat)),?z:Nat),?y:Nat)",
"plus(plus(?x:Nat,times(?y:Nat,times(?x:Nat,?x:Nat))),s(times(plus(?x:Nat,?x:Nat),?y:Nat))) = plus(plus(plus(?z:Nat,s(?y:Nat)),s(times(?z:Nat,?y:Nat))),?y:Nat)",
"plus(times(?z:Nat,s(s(?y:Nat))),plus(?z:Nat,plus(s(?z:Nat),plus(?y:Nat,?y:Nat)))) = s(times(plus(?z:Nat,plus(?z:Nat,?y:Nat)),s(times(?y:Nat,?y:Nat))))",
"plus(times(s(plus(?x:Nat,?z:Nat)),plus(times(?x:Nat,?z:Nat),times(?z:Nat,?y:Nat))),?x:Nat) = times(?z:Nat,plus(s(times(?z:Nat,?z:Nat)),plus(?y:Nat,s(?y:Nat))))",
"times(s(s(plus(?y:Nat,?y:Nat))),plus(s(plus(?x:Nat,?z:Nat)),times(?y:Nat,?x:Nat))) = times(?y:Nat,times(s(times(?z:Nat,?y:Nat)),s(times(?y:Nat,?y:Nat))))",
"times(times(?x:Nat,times(plus(?x:Nat,?z:Nat),?z:Nat)),plus(?x:Nat,plus(?x:Nat,s(?y:Nat)))) = plus(?x:Nat,plus(s(?x:Nat),plus(s(?x:Nat),times(?y:Nat,?z:Nat))))",
"times(times(?z:Nat,plus(?y:Nat,?x:Nat)),s(plus(times(?z:Nat,?z:Nat),times(?x:Nat,?z:Nat)))) = s(times(times(?x:Nat,?z:Nat),plus(times(?y:Nat,?y:Nat),s(?y:Nat))))",
"plus(plus(?z:Nat,plus(s(?y:Nat),?y:Nat)),plus(times(s(?z:Nat),s(?y:Nat)),s(?y:Nat))) = plus(times(s(?y:Nat),s(s(?z:Nat))),s(s(?y:Nat)))",
"plus(times(times(?z:Nat,s(?y:Nat)),plus(?z:Nat,?x:Nat)),plus(?y:Nat,plus(?x:Nat,s(?x:Nat)))) = plus(plus(times(?z:Nat,s(?y:Nat)),s(?x:Nat)),s(?z:Nat))",
"times(plus(?y:Nat,plus(s(?x:Nat),s(?x:Nat))),plus(?y:Nat,times(times(?z:Nat,?z:Nat),?y:Nat))) = plus(plus(plus(s(?x:Nat),?z:Nat),times(?z:Nat,?y:Nat)),?x:Nat)",
"times(plus(times(?z:Nat,plus(?x:Nat,?y:Nat)),times(times(?y:Nat,?z:Nat),times(?x:Nat,?x:Nat))),?x:Nat) = s(plus(?x:Nat,plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat))))",
"times(times(s(?y:Nat),times(s(?z:Nat),times(?y:Nat,?y:Nat))),times(times(?x:Nat,?z:Nat),?x:Nat)) = times(?z:Nat,s(times(times(?x:Nat,?z:Nat),plus(?y:Nat,?y:Nat))))",
"times(times(s(times(?z:Nat,?y:Nat)),s(s(?z:Nat))),s(times(s(?x:Nat),s(?x:Nat)))) = plus(s(plus(?x:Nat,?z:Nat)),plus(?y:Nat,s(s(?z:Nat))))",
"plus(plus(times(times(?x:Nat,?x:Nat),plus(?z:Nat,?z:Nat)),plus(s(?x:Nat),times(?x:Nat,?y:Nat))),?x:Nat) = plus(?x:Nat,plus(?x:Nat,times(plus(?z:Nat,?z:Nat),?z:Nat)))",
"times(plus(times(s(?z:Nat),?x:Nat),plus(?x:Nat,plus(?z:Nat,?y:Nat))),times(?x:Nat,times(?z:Nat,?y:Nat))) = s(times(times(?y:Nat,?z:Nat),s(times(?z:Nat,?x:Nat))))",
"times(times(s(?y:Nat),s(s(?z:Nat))),plus(?x:Nat,times(times(?x:Nat,?y:Nat),plus(?z:Nat,?x:Nat)))) = plus(?x:Nat,plus(?x:Nat,times(plus(?y:Nat,?z:Nat),?y:Nat)))",
"times(times(plus(plus(?z:Nat,?z:Nat),s(?z:Nat)),times(plus(?z:Nat,?y:Nat),times(?y:Nat,?y:Nat))),?x:Nat) = plus(times(times(times(?y:Nat,?x:Nat),?x:Nat),?z:Nat),?z:Nat)",
"times(?x:Nat,times(times(plus(?z:Nat,?y:Nat),times(?y:Nat,?x:Nat)),plus(plus(?z:Nat,?z:Nat),plus(?z:Nat,?y:Nat)))) = plus(?z:Nat,times(s(plus(?y:Nat,?x:Nat)),?x:Nat))",
"times(times(times(?x:Nat,?x:Nat),times(s(?y:Nat),times(?y:Nat,?y:Nat))),times(s(?y:Nat),plus(?x:Nat,?y:Nat))) = s(times(plus(s(?x:Nat),s(?y:Nat)),?z:Nat))",
"plus(plus(?x:Nat,plus(s(?y:Nat),plus(?y:Nat,?z:Nat))),plus(plus(plus(?x:Nat,?z:Nat),?y:Nat),times(?z:Nat,?y:Nat))) = s(s(plus(?x:Nat,times(?z:Nat,?x:Nat))))",
"plus(plus(times(?z:Nat,times(?y:Nat,?x:Nat)),?z:Nat),plus(plus(times(?y:Nat,?y:Nat),plus(?z:Nat,?x:Nat)),s(?y:Nat))) = plus(s(plus(s(?y:Nat),?z:Nat)),?z:Nat)",
"plus(times(plus(?z:Nat,plus(?y:Nat,?z:Nat)),plus(?x:Nat,s(?y:Nat))),plus(s(?x:Nat),s(plus(?z:Nat,?x:Nat)))) = times(s(s(times(?z:Nat,?z:Nat))),?z:Nat)",
"times(plus(plus(?y:Nat,s(?z:Nat)),s(?z:Nat)),plus(plus(?x:Nat,times(?y:Nat,?z:Nat)),times(plus(?x:Nat,?z:Nat),?x:Nat))) = times(?y:Nat,s(s(s(?x:Nat))))",
"times(times(s(times(?z:Nat,?x:Nat)),times(plus(?y:Nat,?z:Nat),plus(?z:Nat,?y:Nat))),times(plus(?x:Nat,s(?x:Nat)),?z:Nat)) = plus(?y:Nat,plus(s(?y:Nat),?z:Nat))",
"times(times(s(times(?z:Nat,?y:Nat)),plus(plus(?y:Nat,?x:Nat),times(?x:Nat,?y:Nat))),plus(s(plus(?y:Nat,?z:Nat)),?x:Nat)) = s(s(s(times(?z:Nat,?x:Nat))))",
"times(times(plus(s(?z:Nat),times(?z:Nat,?z:Nat)),?y:Nat),times(times(plus(?x:Nat,?x:Nat),plus(?x:Nat,?y:Nat)),s(?z:Nat))) = plus(?y:Nat,s(s(s(?x:Nat))))",
"plus(times(s(times(?y:Nat,?x:Nat)),times(?y:Nat,times(?z:Nat,?x:Nat))),plus(plus(times(?y:Nat,?z:Nat),times(?z:Nat,?x:Nat)),?y:Nat)) = times(plus(?x:Nat,?x:Nat),?y:Nat)",
"times(plus(times(s(?x:Nat),plus(?y:Nat,?y:Nat)),?y:Nat),times(s(plus(?x:Nat,?y:Nat)),plus(plus(?z:Nat,?y:Nat),times(?z:Nat,?z:Nat)))) = times(s(?y:Nat),?z:Nat)",
"plus(times(?x:Nat,times(s(?z:Nat),times(?x:Nat,?y:Nat))),plus(times(times(?y:Nat,?x:Nat),plus(?y:Nat,?y:Nat)),times(times(?z:Nat,?z:Nat),?y:Nat))) = plus(?x:Nat,?x:Nat)",
"times(times(s(times(?y:Nat,?z:Nat)),plus(?y:Nat,times(?y:Nat,?y:Nat))),plus(plus(times(?y:Nat,?y:Nat),?x:Nat),plus(s(?z:Nat),s(?y:Nat)))) = times(?x:Nat,?x:Nat)",
"s(?y:Nat) = plus(times(times(?z:Nat,times(?x:Nat,?z:Nat)),times(s(?z:Nat),?x:Nat)),plus(plus(?x:Nat,times(?x:Nat,?y:Nat)),plus(times(?z:Nat,?y:Nat),plus(?x:Nat,?x:Nat))))",
"plus(s(s(?x:Nat)),?x:Nat) = plus(plus(s(times(?x:Nat,?y:Nat)),times(plus(?y:Nat,?x:Nat),?y:Nat)),plus(plus(?x:Nat,s(?x:Nat)),times(times(?x:Nat,?z:Nat),?y:Nat)))",
"times(s(?y:Nat),s(s(?y:Nat))) = plus(times(?y:Nat,times(plus(?y:Nat,?z:Nat),?z:Nat)),plus(times(s(?x:Nat),plus(?y:Nat,?z:Nat)),plus(?z:Nat,plus(?z:Nat,?z:Nat))))",
"s(plus(?x:Nat,s(times(?y:Nat,?z:Nat)))) = times(plus(s(?z:Nat),times(times(?x:Nat,?x:Nat),times(?z:Nat,?y:Nat))),s(times(plus(?y:Nat,?y:Nat),plus(?y:Nat,?y:Nat))))",
"s(plus(s(plus(?z:Nat,?z:Nat)),?x:Nat)) = times(plus(?x:Nat,plus(plus(?y:Nat,?z:Nat),times(?z:Nat,?x:Nat))),plus(plus(times(?z:Nat,?z:Nat),times(?z:Nat,?x:Nat)),?y:Nat))",
"s(plus(plus(times(?z:Nat,?z:Nat),?z:Nat),?z:Nat)) = plus(times(times(s(?x:Nat),times(?y:Nat,?x:Nat)),?z:Nat),plus(?x:Nat,plus(plus(?x:Nat,?z:Nat),times(?x:Nat,?y:Nat))))",
"s(times(times(plus(?z:Nat,?z:Nat),?y:Nat),?z:Nat)) = times(times(?y:Nat,times(times(?z:Nat,?y:Nat),plus(?z:Nat,?y:Nat))),times(?z:Nat,times(plus(?y:Nat,?z:Nat),s(?y:Nat))))",
"plus(s(?z:Nat),s(s(times(?z:Nat,?y:Nat)))) = times(plus(s(s(?x:Nat)),plus(plus(?y:Nat,?z:Nat),plus(?x:Nat,?x:Nat))),s(times(?y:Nat,times(?x:Nat,?z:Nat))))",
"times(?z:Nat,plus(?y:Nat,s(times(?y:Nat,?z:Nat)))) = times(plus(plus(s(?x:Nat),times(?z:Nat,?z:Nat)),?z:Nat),times(s(times(?z:Nat,?y:Nat)),s(times(?z:Nat,?y:Nat))))",
"s(plus(times(times(?z:Nat,?y:Nat),?x:Nat),s(?y:Nat))) = plus(s(s(s(?y:Nat))),plus(times(s(?y:Nat),?z:Nat),times(plus(?z:Nat,?z:Nat),plus(?y:Nat,?z:Nat))))",
"plus(?y:Nat,times(plus(s(?y:Nat),times(?x:Nat,?z:Nat)),?y:Nat)) = s(times(times(plus(?y:Nat,?z:Nat),plus(?x:Nat,?x:Nat)),times(plus(?y:Nat,?x:Nat),times(?x:Nat,?z:Nat))))",
"plus(plus(s(?z:Nat),plus(?y:Nat,plus(?y:Nat,?z:Nat))),?x:Nat) = plus(times(s(plus(?z:Nat,?y:Nat)),s(times(?y:Nat,?z:Nat))),s(times(times(?x:Nat,?x:Nat),?x:Nat)))",
"plus(plus(times(times(?z:Nat,?y:Nat),s(?y:Nat)),?x:Nat),?x:Nat) = plus(plus(times(plus(?z:Nat,?x:Nat),plus(?y:Nat,?x:Nat)),?x:Nat),times(?y:Nat,s(plus(?y:Nat,?x:Nat))))",
"times(?y:Nat,s(plus(times(?y:Nat,?x:Nat),plus(?y:Nat,?x:Nat)))) = times(?y:Nat,plus(times(plus(?x:Nat,?z:Nat),s(?z:Nat)),times(times(?y:Nat,?z:Nat),times(?x:Nat,?y:Nat))))",
"times(s(s(times(?z:Nat,?z:Nat))),plus(?y:Nat,s(?y:Nat))) = times(s(?z:Nat),plus(plus(plus(?x:Nat,?y:Nat),?x:Nat),times(plus(?y:Nat,?x:Nat),plus(?z:Nat,?z:Nat))))",
"times(plus(?x:Nat,s(plus(?z:Nat,?z:Nat))),s(s(?z:Nat))) = times(times(plus(s(?x:Nat),?z:Nat),times(s(?z:Nat),times(?y:Nat,?z:Nat))),s(times(?y:Nat,?x:Nat)))",
"times(times(plus(?z:Nat,s(?z:Nat)),s(s(?z:Nat))),?y:Nat) = times(times(?y:Nat,?y:Nat),plus(s(plus(?z:Nat,?x:Nat)),plus(plus(?x:Nat,?x:Nat),plus(?z:Nat,?x:Nat))))",
"s(times(plus(?x:Nat,plus(?z:Nat,?x:Nat)),times(?z:Nat,s(?z:Nat)))) = plus(times(?y:Nat,plus(plus(?z:Nat,?y:Nat),times(?y:Nat,?x:Nat))),s(plus(?y:Nat,s(?y:Nat))))",
"plus(?x:Nat,times(times(?y:Nat,s(?y:Nat)),s(plus(?y:Nat,?z:Nat)))) = plus(s(times(plus(?y:Nat,?z:Nat),?x:Nat)),plus(?y:Nat,times(times(?x:Nat,?x:Nat),s(?z:Nat))))",
"times(plus(times(plus(?y:Nat,?x:Nat),?y:Nat),plus(?x:Nat,?z:Nat)),?x:Nat) = plus(s(s(plus(?z:Nat,?x:Nat))),times(?x:Nat,times(times(?z:Nat,?z:Nat),plus(?x:Nat,?x:Nat))))",
"s(plus(times(?y:Nat,times(?y:Nat,?x:Nat)),times(plus(?y:Nat,?x:Nat),?x:Nat))) = times(times(plus(times(?z:Nat,?z:Nat),?x:Nat),?z:Nat),plus(?x:Nat,s(times(?z:Nat,?z:Nat))))",
"plus(times(s(plus(?z:Nat,?y:Nat)),s(s(?z:Nat))),plus(?y:Nat,?x:Nat)) = plus(s(s(times(?z:Nat,?x:Nat))),times(s(s(?x:Nat)),s(plus(?x:Nat,?y:Nat))))",
"plus(times(plus(?x:Nat,plus(?y:Nat,?z:Nat)),times(s(?y:Nat),?x:Nat)),?y:Nat) = plus(plus(times(s(?x:Nat),?x:Nat),?x:Nat),times(times(plus(?y:Nat,?x:Nat),?x:Nat),?x:Nat))",
"times(plus(plus(?z:Nat,times(?x:Nat,?y:Nat)),s(times(?z:Nat,?y:Nat))),?z:Nat) = plus(plus(?z:Nat,plus(s(?z:Nat),s(?x:Nat))),times(times(?z:Nat,s(?x:Nat)),?y:Nat))",
"times(times(s(s(?x:Nat)),?y:Nat),times(?z:Nat,times(s(?y:Nat),?x:Nat))) = times(?z:Nat,times(times(s(?y:Nat),plus(?y:Nat,?x:Nat)),times(?x:Nat,times(?z:Nat,?x:Nat))))",
"s(plus(s(plus(?z:Nat,?z:Nat)),times(plus(?y:Nat,?x:Nat),plus(?x:Nat,?y:Nat)))) = plus(times(plus(times(?y:Nat,?y:Nat),times(?z:Nat,?y:Nat)),plus(?x:Nat,?y:Nat)),?x:Nat)",
"s(times(plus(plus(?z:Nat,?z:Nat),?x:Nat),times(times(?z:Nat,?z:Nat),s(?x:Nat)))) = plus(plus(s(?y:Nat),?z:Nat),s(times(times(?y:Nat,?x:Nat),times(?z:Nat,?x:Nat))))",
"times(s(s(plus(?x:Nat,?z:Nat))),times(plus(?y:Nat,times(?z:Nat,?y:Nat)),?x:Nat)) = plus(times(plus(s(?y:Nat),plus(?y:Nat,?z:Nat)),s(times(?y:Nat,?z:Nat))),?z:Nat)",
"plus(plus(plus(s(?y:Nat),plus(?x:Nat,?z:Nat)),plus(plus(?x:Nat,?z:Nat),?z:Nat)),?x:Nat) = s(times(times(times(?z:Nat,?x:Nat),?y:Nat),plus(?z:Nat,plus(?z:Nat,?z:Nat))))",
"times(plus(times(s(?x:Nat),plus(?z:Nat,?y:Nat)),plus(?x:Nat,times(?y:Nat,?y:Nat))),?x:Nat) = times(plus(times(s(?z:Nat),s(?y:Nat)),s(plus(?x:Nat,?z:Nat))),?z:Nat)",
"times(times(?y:Nat,?x:Nat),plus(plus(?z:Nat,plus(?z:Nat,?z:Nat)),times(s(?x:Nat),?z:Nat))) = plus(s(times(?z:Nat,s(?y:Nat))),times(?x:Nat,times(?y:Nat,s(?z:Nat))))",
"s(times(times(s(?y:Nat),plus(?y:Nat,?y:Nat)),plus(times(?x:Nat,?y:Nat),times(?z:Nat,?x:Nat)))) = plus(times(?y:Nat,plus(plus(?z:Nat,?z:Nat),plus(?x:Nat,?y:Nat))),?x:Nat)",
"s(times(times(times(?x:Nat,?y:Nat),times(?y:Nat,?x:Nat)),plus(s(?y:Nat),times(?y:Nat,?y:Nat)))) = plus(plus(?y:Nat,plus(times(?x:Nat,?z:Nat),times(?x:Nat,?x:Nat))),?x:Nat)",
"plus(s(plus(times(?y:Nat,?z:Nat),times(?x:Nat,?z:Nat))),s(times(?z:Nat,plus(?y:Nat,?z:Nat)))) = plus(plus(times(?x:Nat,?x:Nat),plus(plus(?y:Nat,?z:Nat),?x:Nat)),?z:Nat)",
"times(s(times(?z:Nat,plus(?z:Nat,?y:Nat))),plus(plus(s(?y:Nat),plus(?z:Nat,?x:Nat)),?y:Nat)) = plus(times(s(s(?x:Nat)),s(?x:Nat)),times(?z:Nat,s(?z:Nat)))",
"times(plus(s(plus(?y:Nat,?z:Nat)),times(s(?y:Nat),?z:Nat)),s(s(plus(?y:Nat,?z:Nat)))) = plus(plus(?x:Nat,s(plus(?x:Nat,?z:Nat))),s(plus(?z:Nat,?x:Nat)))",
"times(plus(plus(plus(?x:Nat,?x:Nat),times(?y:Nat,?y:Nat)),?x:Nat),s(s(times(?z:Nat,?x:Nat)))) = plus(times(?y:Nat,?x:Nat),s(times(plus(?z:Nat,?x:Nat),s(?x:Nat))))",
"times(plus(times(?y:Nat,plus(?x:Nat,?x:Nat)),s(s(?x:Nat))),s(s(plus(?y:Nat,?z:Nat)))) = s(plus(s(plus(?x:Nat,?y:Nat)),times(plus(?x:Nat,?x:Nat),?y:Nat)))",
"plus(plus(?z:Nat,s(times(?y:Nat,?x:Nat))),times(times(?z:Nat,plus(?z:Nat,?x:Nat)),s(s(?y:Nat)))) = times(s(times(times(?x:Nat,?z:Nat),plus(?z:Nat,?y:Nat))),?x:Nat)",
"times(plus(plus(plus(?z:Nat,?z:Nat),?y:Nat),times(s(?z:Nat),times(?y:Nat,?y:Nat))),s(s(?x:Nat))) = plus(s(plus(plus(?x:Nat,?z:Nat),plus(?z:Nat,?y:Nat))),?x:Nat)",
"times(plus(times(times(?x:Nat,?y:Nat),s(?z:Nat)),s(plus(?x:Nat,?z:Nat))),s(s(s(?x:Nat)))) = plus(times(plus(s(?y:Nat),plus(?x:Nat,?x:Nat)),?y:Nat),?y:Nat)",
"plus(plus(plus(s(?z:Nat),?z:Nat),s(s(?y:Nat))),times(?x:Nat,plus(s(?z:Nat),times(?y:Nat,?z:Nat)))) = times(times(times(?x:Nat,s(?z:Nat)),?z:Nat),s(?y:Nat))",
"times(s(plus(s(?y:Nat),?y:Nat)),times(times(?z:Nat,s(?x:Nat)),plus(s(?z:Nat),times(?y:Nat,?y:Nat)))) = times(times(plus(plus(?y:Nat,?x:Nat),?z:Nat),?z:Nat),?z:Nat)",
"times(plus(?y:Nat,times(?x:Nat,plus(?z:Nat,?x:Nat))),plus(?z:Nat,times(times(?y:Nat,?x:Nat),times(?z:Nat,?z:Nat)))) = s(plus(plus(?x:Nat,?y:Nat),plus(s(?y:Nat),?x:Nat)))",
"times(plus(s(times(?y:Nat,?x:Nat)),times(times(?z:Nat,?y:Nat),s(?y:Nat))),times(s(plus(?z:Nat,?y:Nat)),?x:Nat)) = s(times(times(times(?y:Nat,?y:Nat),?y:Nat),?z:Nat))",
"times(plus(plus(?z:Nat,?y:Nat),plus(plus(?x:Nat,?x:Nat),times(?y:Nat,?z:Nat))),plus(?z:Nat,s(plus(?z:Nat,?y:Nat)))) = plus(times(s(plus(?z:Nat,?z:Nat)),?x:Nat),?x:Nat)",
"plus(times(times(s(?x:Nat),?x:Nat),s(times(?x:Nat,?z:Nat))),plus(plus(s(?x:Nat),?z:Nat),s(plus(?x:Nat,?y:Nat)))) = s(times(s(times(?y:Nat,?z:Nat)),?y:Nat))",
"times(times(plus(plus(?x:Nat,?y:Nat),times(?z:Nat,?y:Nat)),times(plus(?y:Nat,?x:Nat),s(?y:Nat))),plus(s(times(?x:Nat,?y:Nat)),?z:Nat)) = times(?y:Nat,times(?z:Nat,?z:Nat))",
"times(plus(s(times(?x:Nat,?y:Nat)),times(s(?y:Nat),?z:Nat)),plus(times(plus(?x:Nat,?z:Nat),s(?z:Nat)),plus(?z:Nat,times(?x:Nat,?y:Nat)))) = plus(s(?y:Nat),?z:Nat)",
"times(times(?y:Nat,times(s(?z:Nat),times(?y:Nat,?y:Nat))),plus(plus(s(?x:Nat),plus(?y:Nat,?x:Nat)),times(plus(?z:Nat,?x:Nat),s(?x:Nat)))) = s(s(s(?x:Nat)))",
"times(plus(s(plus(?x:Nat,?y:Nat)),times(?x:Nat,plus(?x:Nat,?x:Nat))),times(plus(times(?x:Nat,?y:Nat),?x:Nat),plus(times(?x:Nat,?z:Nat),times(?x:Nat,?y:Nat)))) = s(?y:Nat)",
"times(plus(times(plus(?x:Nat,?y:Nat),plus(?y:Nat,?y:Nat)),plus(plus(?z:Nat,?z:Nat),times(?z:Nat,?x:Nat))),plus(s(plus(?y:Nat,?x:Nat)),s(s(?y:Nat)))) = s(?z:Nat)",
"plus(?y:Nat,s(?x:Nat)) = plus(times(plus(?y:Nat,times(?z:Nat,?z:Nat)),times(?x:Nat,times(?x:Nat,?z:Nat))),times(s(s(?y:Nat)),plus(plus(?z:Nat,?z:Nat),times(?z:Nat,?z:Nat))))",
"times(?y:Nat,plus(?y:Nat,?y:Nat)) = times(plus(s(times(?z:Nat,?z:Nat)),plus(?y:Nat,times(?x:Nat,?y:Nat))),times(s(s(?y:Nat)),times(plus(?x:Nat,?y:Nat),times(?y:Nat,?z:Nat))))",
"plus(?x:Nat,s(s(s(?z:Nat)))) = times(plus(s(?x:Nat),s(s(?y:Nat))),plus(times(plus(?y:Nat,?y:Nat),times(?y:Nat,?z:Nat)),times(s(?x:Nat),plus(?z:Nat,?x:Nat))))",
"times(?x:Nat,s(s(plus(?y:Nat,?y:Nat)))) = plus(plus(plus(plus(?y:Nat,?x:Nat),s(?y:Nat)),?z:Nat),times(plus(times(?y:Nat,?x:Nat),s(?y:Nat)),plus(?x:Nat,s(?y:Nat))))",
"s(plus(times(?x:Nat,times(?z:Nat,?x:Nat)),?z:Nat)) = times(times(?z:Nat,?z:Nat),plus(plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),plus(times(?z:Nat,?z:Nat),times(?x:Nat,?y:Nat))))",
"plus(?x:Nat,s(times(times(?z:Nat,?z:Nat),?z:Nat))) = times(times(?y:Nat,s(times(?z:Nat,?x:Nat))),times(times(?z:Nat,times(?x:Nat,?z:Nat)),plus(s(?y:Nat),times(?x:Nat,?z:Nat))))",
"plus(?x:Nat,times(?z:Nat,times(?x:Nat,s(?x:Nat)))) = times(plus(s(?x:Nat),plus(s(?z:Nat),times(?y:Nat,?z:Nat))),times(plus(?z:Nat,?x:Nat),times(?y:Nat,times(?x:Nat,?y:Nat))))",
"plus(?y:Nat,s(plus(times(?y:Nat,?z:Nat),?z:Nat))) = times(times(plus(plus(?x:Nat,?z:Nat),times(?z:Nat,?x:Nat)),?x:Nat),plus(?z:Nat,plus(times(?y:Nat,?z:Nat),times(?x:Nat,?y:Nat))))",
"times(?z:Nat,plus(?x:Nat,s(times(?y:Nat,?y:Nat)))) = plus(plus(times(?z:Nat,?y:Nat),times(?z:Nat,?x:Nat)),times(plus(s(?y:Nat),?x:Nat),plus(s(?y:Nat),times(?x:Nat,?y:Nat))))",
"s(plus(plus(plus(?x:Nat,?y:Nat),s(?y:Nat)),?y:Nat)) = plus(s(times(times(?x:Nat,?z:Nat),times(?x:Nat,?x:Nat))),times(?x:Nat,times(plus(?x:Nat,?x:Nat),plus(?y:Nat,?x:Nat))))",
"times(s(?x:Nat),times(s(?y:Nat),plus(?x:Nat,?y:Nat))) = plus(times(?x:Nat,plus(times(?z:Nat,?z:Nat),times(?z:Nat,?x:Nat))),times(?y:Nat,plus(times(?z:Nat,?x:Nat),s(?x:Nat))))",
"times(s(?y:Nat),s(times(?y:Nat,times(?x:Nat,?y:Nat)))) = times(plus(times(s(?y:Nat),?z:Nat),s(s(?z:Nat))),times(s(?x:Nat),plus(plus(?y:Nat,?z:Nat),s(?y:Nat))))",
"plus(s(s(s(?x:Nat))),s(times(?x:Nat,s(?y:Nat)))) = times(plus(?y:Nat,?z:Nat),plus(times(plus(?x:Nat,?z:Nat),times(?x:Nat,?z:Nat)),times(?z:Nat,plus(?y:Nat,?z:Nat))))",
"plus(plus(?x:Nat,plus(times(?y:Nat,?x:Nat),s(?y:Nat))),?x:Nat) = times(plus(?y:Nat,times(s(?x:Nat),?y:Nat)),times(times(?x:Nat,s(?y:Nat)),plus(s(?y:Nat),s(?x:Nat))))",
"times(?y:Nat,s(times(plus(?z:Nat,?x:Nat),times(?y:Nat,?y:Nat)))) = times(s(plus(times(?x:Nat,?y:Nat),s(?y:Nat))),plus(s(times(?x:Nat,?z:Nat)),s(times(?z:Nat,?z:Nat))))",
"times(plus(?y:Nat,?y:Nat),plus(s(times(?z:Nat,?x:Nat)),?x:Nat)) = times(plus(?x:Nat,?z:Nat),times(times(plus(?y:Nat,?z:Nat),?z:Nat),plus(times(?y:Nat,?z:Nat),times(?x:Nat,?x:Nat))))",
"s(times(plus(s(?y:Nat),?x:Nat),times(?x:Nat,times(?x:Nat,?z:Nat)))) = times(times(s(?y:Nat),times(plus(?z:Nat,?x:Nat),plus(?z:Nat,?x:Nat))),s(s(plus(?x:Nat,?x:Nat))))",
"s(times(times(?y:Nat,plus(?z:Nat,?z:Nat)),plus(?z:Nat,s(?x:Nat)))) = plus(times(?x:Nat,plus(plus(?z:Nat,?z:Nat),?y:Nat)),times(times(times(?x:Nat,?x:Nat),s(?z:Nat)),?y:Nat))",
"times(?z:Nat,plus(times(s(?z:Nat),times(?x:Nat,?z:Nat)),s(?z:Nat))) = times(s(?x:Nat),times(times(plus(?z:Nat,?z:Nat),s(?y:Nat)),times(s(?x:Nat),plus(?z:Nat,?z:Nat))))",
"times(s(s(s(?z:Nat))),plus(s(times(?x:Nat,?y:Nat)),?y:Nat)) = plus(times(?z:Nat,plus(s(?x:Nat),times(?z:Nat,?y:Nat))),times(plus(?z:Nat,plus(?x:Nat,?y:Nat)),?z:Nat))",
"times(times(?x:Nat,times(times(?y:Nat,?z:Nat),times(?z:Nat,?y:Nat))),?x:Nat) = plus(s(plus(?y:Nat,plus(?x:Nat,?x:Nat))),times(s(plus(?x:Nat,?z:Nat)),s(times(?z:Nat,?y:Nat))))",
"s(plus(plus(times(?z:Nat,?y:Nat),plus(?y:Nat,?x:Nat)),s(s(?z:Nat)))) = plus(?x:Nat,times(times(plus(?x:Nat,?z:Nat),plus(?y:Nat,?y:Nat)),plus(?z:Nat,plus(?y:Nat,?y:Nat))))",
"plus(?x:Nat,plus(times(?z:Nat,?x:Nat),times(plus(?z:Nat,?x:Nat),s(?x:Nat)))) = times(plus(s(?x:Nat),times(?y:Nat,s(?z:Nat))),plus(?y:Nat,plus(plus(?z:Nat,?x:Nat),?z:Nat)))",
"plus(?z:Nat,times(plus(times(?x:Nat,?z:Nat),?y:Nat),s(times(?y:Nat,?x:Nat)))) = times(plus(?z:Nat,?z:Nat),times(s(s(?x:Nat)),times(plus(?y:Nat,?x:Nat),plus(?y:Nat,?y:Nat))))",
"plus(plus(?z:Nat,?y:Nat),plus(?x:Nat,times(times(?x:Nat,?y:Nat),s(?y:Nat)))) = times(s(times(plus(?z:Nat,?z:Nat),?x:Nat)),times(times(plus(?y:Nat,?y:Nat),?x:Nat),s(?z:Nat)))",
"plus(times(?y:Nat,plus(s(?y:Nat),?x:Nat)),s(s(times(?y:Nat,?y:Nat)))) = times(plus(times(plus(?z:Nat,?z:Nat),?z:Nat),?y:Nat),s(times(s(?y:Nat),times(?x:Nat,?z:Nat))))",
"times(s(plus(s(?x:Nat),?z:Nat)),s(times(plus(?z:Nat,?z:Nat),?y:Nat))) = plus(times(times(plus(?z:Nat,?x:Nat),?x:Nat),?y:Nat),times(times(plus(?z:Nat,?z:Nat),?x:Nat),?z:Nat))",
"times(times(plus(?x:Nat,s(?y:Nat)),plus(plus(?x:Nat,?z:Nat),?x:Nat)),?x:Nat) = times(times(?y:Nat,?x:Nat),times(times(?x:Nat,s(?x:Nat)),times(s(?y:Nat),times(?y:Nat,?x:Nat))))",
"plus(times(s(times(?y:Nat,?z:Nat)),s(?z:Nat)),times(s(s(?z:Nat)),?y:Nat)) = plus(?x:Nat,times(times(plus(?y:Nat,?x:Nat),plus(?z:Nat,?x:Nat)),s(plus(?y:Nat,?x:Nat))))",
"times(plus(s(?y:Nat),times(times(?y:Nat,?x:Nat),plus(?y:Nat,?x:Nat))),s(?z:Nat)) = times(plus(plus(times(?z:Nat,?y:Nat),?y:Nat),plus(s(?z:Nat),times(?x:Nat,?x:Nat))),?z:Nat)",
"plus(plus(?z:Nat,?z:Nat),times(times(?x:Nat,s(?z:Nat)),times(?x:Nat,plus(?z:Nat,?y:Nat)))) = plus(plus(s(?z:Nat),?y:Nat),s(times(plus(?y:Nat,?x:Nat),plus(?z:Nat,?y:Nat))))",
"plus(times(times(s(?z:Nat),s(?z:Nat)),?y:Nat),times(?x:Nat,s(plus(?y:Nat,?y:Nat)))) = plus(?z:Nat,plus(times(plus(?z:Nat,?y:Nat),times(?y:Nat,?z:Nat)),times(?y:Nat,?y:Nat)))",
"times(plus(plus(times(?x:Nat,?y:Nat),?z:Nat),?x:Nat),times(s(s(?y:Nat)),s(?y:Nat))) = times(times(plus(times(?y:Nat,?z:Nat),times(?z:Nat,?z:Nat)),s(s(?x:Nat))),?z:Nat)",
"times(times(?y:Nat,plus(?x:Nat,plus(?y:Nat,?z:Nat))),s(plus(?z:Nat,plus(?z:Nat,?z:Nat)))) = s(times(plus(plus(?y:Nat,?y:Nat),s(?z:Nat)),times(plus(?y:Nat,?z:Nat),?x:Nat)))",
"times(times(times(plus(?z:Nat,?y:Nat),?x:Nat),plus(s(?y:Nat),times(?x:Nat,?z:Nat))),?y:Nat) = times(s(s(?z:Nat)),plus(?z:Nat,times(plus(?z:Nat,?x:Nat),times(?y:Nat,?z:Nat))))",
"times(plus(?y:Nat,?x:Nat),plus(s(plus(?x:Nat,?y:Nat)),times(times(?y:Nat,?y:Nat),s(?x:Nat)))) = plus(s(plus(?x:Nat,times(?x:Nat,?z:Nat))),s(s(plus(?y:Nat,?z:Nat))))",
"times(plus(times(s(?x:Nat),s(?y:Nat)),times(?z:Nat,plus(?x:Nat,?z:Nat))),s(s(?y:Nat))) = plus(plus(s(?x:Nat),plus(times(?y:Nat,?z:Nat),plus(?z:Nat,?z:Nat))),?x:Nat)",
"plus(times(?z:Nat,times(plus(?z:Nat,?z:Nat),?y:Nat)),times(plus(s(?x:Nat),s(?y:Nat)),s(?z:Nat))) = times(?y:Nat,plus(?x:Nat,plus(times(?z:Nat,?z:Nat),plus(?y:Nat,?y:Nat))))",
"times(s(plus(times(?y:Nat,?x:Nat),s(?z:Nat))),plus(plus(plus(?y:Nat,?y:Nat),s(?z:Nat)),?x:Nat)) = times(times(?z:Nat,?z:Nat),times(plus(?x:Nat,plus(?y:Nat,?z:Nat)),?z:Nat))",
"times(plus(plus(times(?y:Nat,?z:Nat),times(?x:Nat,?z:Nat)),s(?z:Nat)),plus(?x:Nat,s(s(?y:Nat)))) = plus(s(times(times(?x:Nat,?z:Nat),times(?y:Nat,?x:Nat))),s(?x:Nat))",
"plus(times(s(times(?x:Nat,?y:Nat)),s(?y:Nat)),times(times(s(?z:Nat),?x:Nat),times(s(?z:Nat),?y:Nat))) = plus(s(times(plus(?x:Nat,?z:Nat),times(?y:Nat,?z:Nat))),?x:Nat)",
"plus(times(plus(?x:Nat,?y:Nat),?y:Nat),times(times(times(?z:Nat,?z:Nat),plus(?z:Nat,?z:Nat)),s(s(?x:Nat)))) = times(plus(plus(?y:Nat,s(?y:Nat)),?y:Nat),s(s(?x:Nat)))",
"times(plus(?z:Nat,?z:Nat),times(times(s(?y:Nat),times(?x:Nat,?x:Nat)),plus(plus(?y:Nat,?z:Nat),s(?z:Nat)))) = times(times(?z:Nat,?x:Nat),times(s(plus(?x:Nat,?z:Nat)),?x:Nat))",
"times(plus(plus(?z:Nat,plus(?y:Nat,?x:Nat)),?x:Nat),times(?x:Nat,times(times(?x:Nat,?y:Nat),plus(?z:Nat,?x:Nat)))) = times(times(plus(?x:Nat,?x:Nat),s(times(?z:Nat,?x:Nat))),?x:Nat)",
"plus(s(?x:Nat),plus(times(plus(?y:Nat,?y:Nat),times(?x:Nat,?y:Nat)),plus(times(?x:Nat,?x:Nat),times(?x:Nat,?y:Nat)))) = plus(times(?z:Nat,s(?y:Nat)),s(s(s(?z:Nat))))",
"plus(s(s(s(?z:Nat))),plus(plus(?y:Nat,times(?z:Nat,?x:Nat)),plus(times(?y:Nat,?y:Nat),plus(?x:Nat,?x:Nat)))) = times(s(?z:Nat),times(s(plus(?x:Nat,?z:Nat)),?z:Nat))",
"plus(times(?x:Nat,plus(?z:Nat,?y:Nat)),plus(s(plus(?z:Nat,?y:Nat)),plus(times(?x:Nat,?y:Nat),plus(?z:Nat,?z:Nat)))) = times(times(plus(s(?z:Nat),s(?z:Nat)),?z:Nat),?y:Nat)",
"plus(times(?y:Nat,times(times(?y:Nat,?z:Nat),times(?z:Nat,?z:Nat))),plus(?y:Nat,times(s(?x:Nat),times(?x:Nat,?z:Nat)))) = s(plus(plus(times(?x:Nat,?x:Nat),s(?z:Nat)),?y:Nat))",
"times(plus(plus(s(?x:Nat),plus(?x:Nat,?z:Nat)),s(times(?y:Nat,?y:Nat))),plus(?y:Nat,plus(s(?z:Nat),?z:Nat))) = times(s(s(times(?x:Nat,?z:Nat))),plus(?z:Nat,?y:Nat))",
"times(plus(times(plus(?x:Nat,?y:Nat),s(?z:Nat)),s(s(?z:Nat))),s(plus(times(?z:Nat,?x:Nat),times(?y:Nat,?x:Nat)))) = s(times(?z:Nat,plus(s(?x:Nat),s(?y:Nat))))",
"times(times(s(plus(?z:Nat,?z:Nat)),times(plus(?z:Nat,?y:Nat),plus(?x:Nat,?z:Nat))),times(times(s(?z:Nat),?z:Nat),?y:Nat)) = s(times(?x:Nat,plus(times(?z:Nat,?z:Nat),?x:Nat)))",
"plus(plus(plus(?z:Nat,times(?x:Nat,?z:Nat)),times(times(?z:Nat,?x:Nat),plus(?z:Nat,?x:Nat))),plus(plus(times(?x:Nat,?z:Nat),?x:Nat),plus(?y:Nat,plus(?x:Nat,?x:Nat)))) = s(?z:Nat)",
"times(plus(plus(?y:Nat,times(?x:Nat,?z:Nat)),times(times(?z:Nat,?x:Nat),times(?z:Nat,?x:Nat))),times(s(times(?z:Nat,?y:Nat)),times(s(?y:Nat),times(?y:Nat,?y:Nat)))) = s(?z:Nat)",
"s(?x:Nat) = plus(times(plus(times(?y:Nat,?z:Nat),?z:Nat),plus(times(?x:Nat,?z:Nat),times(?z:Nat,?x:Nat))),times(times(?x:Nat,s(?y:Nat)),times(plus(?x:Nat,?x:Nat),plus(?y:Nat,?y:Nat))))",
"times(?y:Nat,?y:Nat) = plus(plus(times(plus(?y:Nat,?y:Nat),plus(?y:Nat,?y:Nat)),times(times(?y:Nat,?x:Nat),plus(?z:Nat,?x:Nat))),plus(?y:Nat,plus(times(?y:Nat,?y:Nat),plus(?z:Nat,?z:Nat))))",
"s(s(times(times(?x:Nat,?x:Nat),s(?x:Nat)))) = times(plus(s(?y:Nat),s(s(?y:Nat))),plus(times(?y:Nat,times(?y:Nat,?z:Nat)),times(times(?y:Nat,?z:Nat),plus(?x:Nat,?x:Nat))))",
"s(times(?x:Nat,plus(?x:Nat,plus(?y:Nat,?z:Nat)))) = plus(plus(s(plus(?y:Nat,?x:Nat)),?z:Nat),plus(plus(times(?y:Nat,?x:Nat),times(?y:Nat,?y:Nat)),times(?z:Nat,times(?z:Nat,?x:Nat))))",
"s(plus(?z:Nat,plus(times(?z:Nat,?z:Nat),s(?y:Nat)))) = times(plus(s(times(?x:Nat,?x:Nat)),times(plus(?x:Nat,?y:Nat),plus(?x:Nat,?x:Nat))),plus(plus(?z:Nat,s(?y:Nat)),?x:Nat))",
"times(?z:Nat,times(?z:Nat,times(s(?y:Nat),s(?y:Nat)))) = times(times(plus(times(?y:Nat,?x:Nat),s(?y:Nat)),?x:Nat),times(plus(times(?y:Nat,?y:Nat),?z:Nat),times(?z:Nat,s(?x:Nat))))",
"s(plus(plus(times(?z:Nat,?z:Nat),plus(?y:Nat,?z:Nat)),?z:Nat)) = times(times(plus(?x:Nat,plus(?x:Nat,?x:Nat)),times(plus(?y:Nat,?y:Nat),plus(?x:Nat,?x:Nat))),s(times(?z:Nat,?z:Nat)))",
"s(plus(times(plus(?x:Nat,?z:Nat),plus(?x:Nat,?z:Nat)),?y:Nat)) = plus(plus(?x:Nat,plus(s(?y:Nat),?x:Nat)),plus(plus(s(?y:Nat),s(?z:Nat)),plus(times(?y:Nat,?z:Nat),?x:Nat)))",
"plus(plus(?x:Nat,plus(times(?z:Nat,?y:Nat),s(?z:Nat))),?x:Nat) = times(times(s(times(?z:Nat,?x:Nat)),plus(?y:Nat,times(?y:Nat,?z:Nat))),plus(times(s(?y:Nat),s(?x:Nat)),?y:Nat))",
"times(plus(plus(times(?z:Nat,?y:Nat),?z:Nat),?x:Nat),s(?x:Nat)) = times(plus(s(plus(?x:Nat,?x:Nat)),plus(times(?x:Nat,?x:Nat),?z:Nat)),plus(s(plus(?z:Nat,?x:Nat)),s(?z:Nat)))",
"times(?y:Nat,times(?x:Nat,times(times(?x:Nat,?z:Nat),times(?z:Nat,?z:Nat)))) = times(s(plus(plus(?z:Nat,?z:Nat),?z:Nat)),times(times(?x:Nat,?y:Nat),plus(plus(?y:Nat,?z:Nat),s(?y:Nat))))",
"times(times(times(?x:Nat,s(?y:Nat)),times(s(?x:Nat),?x:Nat)),?y:Nat) = times(times(plus(s(?x:Nat),?x:Nat),s(plus(?y:Nat,?x:Nat))),s(plus(times(?y:Nat,?y:Nat),s(?y:Nat))))",
"plus(s(plus(times(?y:Nat,?x:Nat),?y:Nat)),times(s(s(?z:Nat)),?y:Nat)) = times(plus(times(times(?x:Nat,?z:Nat),?y:Nat),times(?y:Nat,?z:Nat)),s(times(s(?z:Nat),s(?z:Nat))))",
"plus(s(times(s(?z:Nat),?y:Nat)),s(plus(?y:Nat,plus(?x:Nat,?y:Nat)))) = plus(plus(?z:Nat,s(s(?x:Nat))),times(s(plus(?y:Nat,?x:Nat)),plus(plus(?y:Nat,?x:Nat),?x:Nat)))",
"times(plus(plus(s(?x:Nat),times(?x:Nat,?x:Nat)),?z:Nat),s(s(?z:Nat))) = plus(times(?y:Nat,plus(?z:Nat,plus(?y:Nat,?x:Nat))),times(s(?x:Nat),times(s(?x:Nat),s(?z:Nat))))",
"plus(?z:Nat,plus(s(s(?x:Nat)),plus(times(?z:Nat,?z:Nat),times(?z:Nat,?z:Nat)))) = plus(?z:Nat,plus(plus(plus(?x:Nat,?z:Nat),times(?z:Nat,?x:Nat)),plus(?x:Nat,plus(?y:Nat,?y:Nat))))",
"plus(s(times(plus(?y:Nat,?y:Nat),?z:Nat)),s(plus(times(?z:Nat,?x:Nat),?z:Nat))) = times(s(s(s(?z:Nat))),plus(s(s(?z:Nat)),times(plus(?y:Nat,?x:Nat),s(?y:Nat))))",
"times(s(times(plus(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat))),times(s(?z:Nat),?z:Nat)) = s(times(times(s(?z:Nat),times(?y:Nat,?x:Nat)),plus(plus(?z:Nat,?x:Nat),times(?x:Nat,?x:Nat))))",
"times(plus(times(?z:Nat,times(?x:Nat,?x:Nat)),s(s(?z:Nat))),plus(?x:Nat,?z:Nat)) = times(plus(plus(times(?z:Nat,?x:Nat),s(?z:Nat)),?z:Nat),times(s(plus(?z:Nat,?x:Nat)),?x:Nat))",
"times(?z:Nat,times(times(times(?z:Nat,?z:Nat),?y:Nat),plus(times(?x:Nat,?z:Nat),s(?x:Nat)))) = plus(s(s(plus(?z:Nat,?y:Nat))),plus(times(s(?x:Nat),plus(?x:Nat,?z:Nat)),?y:Nat))",
"times(plus(?y:Nat,?z:Nat),plus(times(?z:Nat,s(?y:Nat)),plus(s(?x:Nat),s(?y:Nat)))) = plus(times(s(plus(?x:Nat,?x:Nat)),times(times(?x:Nat,?y:Nat),times(?x:Nat,?z:Nat))),?y:Nat)",
"times(plus(plus(times(?x:Nat,?z:Nat),s(?z:Nat)),times(times(?z:Nat,?x:Nat),?y:Nat)),?z:Nat) = times(s(plus(times(?z:Nat,?x:Nat),?x:Nat)),times(?x:Nat,plus(?y:Nat,times(?z:Nat,?z:Nat))))",
"plus(s(plus(times(?y:Nat,?z:Nat),?z:Nat)),plus(?y:Nat,plus(plus(?z:Nat,?z:Nat),s(?y:Nat)))) = times(plus(plus(s(?x:Nat),?y:Nat),s(times(?y:Nat,?z:Nat))),s(s(?y:Nat)))",
"plus(times(plus(?x:Nat,?z:Nat),times(plus(?y:Nat,?y:Nat),?x:Nat)),times(?x:Nat,s(s(?x:Nat)))) = times(times(s(?z:Nat),s(times(?x:Nat,?z:Nat))),times(s(s(?y:Nat)),?x:Nat))",
"times(plus(?x:Nat,plus(times(?x:Nat,?x:Nat),?z:Nat)),plus(s(s(?z:Nat)),times(?z:Nat,?x:Nat))) = times(s(s(plus(?x:Nat,?z:Nat))),times(plus(times(?x:Nat,?x:Nat),?x:Nat),?x:Nat))",
"s(times(times(times(?z:Nat,?x:Nat),plus(?y:Nat,?x:Nat)),plus(times(?z:Nat,?y:Nat),times(?z:Nat,?x:Nat)))) = plus(plus(?y:Nat,?z:Nat),s(plus(plus(?x:Nat,?x:Nat),times(?y:Nat,?y:Nat))))",
"plus(plus(plus(s(?z:Nat),times(?x:Nat,?z:Nat)),?z:Nat),times(plus(s(?x:Nat),s(?x:Nat)),?x:Nat)) = times(plus(times(?x:Nat,times(?y:Nat,?z:Nat)),s(plus(?z:Nat,?y:Nat))),?y:Nat)",
"times(times(s(plus(?x:Nat,?x:Nat)),s(s(?z:Nat))),plus(?y:Nat,plus(s(?z:Nat),s(?y:Nat)))) = plus(?y:Nat,plus(s(?z:Nat),plus(times(?x:Nat,?y:Nat),times(?x:Nat,?x:Nat))))",
"times(times(times(s(?y:Nat),s(?z:Nat)),times(s(?x:Nat),?y:Nat)),plus(?x:Nat,s(s(?z:Nat)))) = s(plus(times(?x:Nat,s(?x:Nat)),times(plus(?y:Nat,?y:Nat),s(?y:Nat))))",
"times(plus(plus(s(?z:Nat),times(?y:Nat,?x:Nat)),times(times(?z:Nat,?z:Nat),plus(?z:Nat,?y:Nat))),s(?x:Nat)) = plus(times(?z:Nat,plus(?y:Nat,?x:Nat)),s(plus(s(?z:Nat),?x:Nat)))",
"times(times(plus(plus(?y:Nat,?y:Nat),s(?z:Nat)),s(?z:Nat)),times(plus(?x:Nat,plus(?y:Nat,?y:Nat)),?z:Nat)) = plus(times(?x:Nat,times(?x:Nat,plus(?x:Nat,?z:Nat))),times(?y:Nat,?y:Nat))",
"plus(times(plus(plus(?z:Nat,?z:Nat),plus(?y:Nat,?z:Nat)),?z:Nat),times(s(times(?y:Nat,?z:Nat)),s(s(?x:Nat)))) = plus(plus(s(s(?y:Nat)),plus(?x:Nat,s(?z:Nat))),?x:Nat)",
"plus(plus(?y:Nat,plus(?y:Nat,s(?z:Nat))),plus(times(?y:Nat,times(?x:Nat,?z:Nat)),times(times(?x:Nat,?y:Nat),s(?z:Nat)))) = plus(s(?z:Nat),plus(times(s(?z:Nat),?z:Nat),?y:Nat))",
"plus(plus(s(times(?x:Nat,?z:Nat)),plus(times(?x:Nat,?x:Nat),?z:Nat)),times(s(?y:Nat),times(?y:Nat,times(?x:Nat,?z:Nat)))) = s(s(plus(times(?z:Nat,?x:Nat),plus(?x:Nat,?y:Nat))))",
"plus(plus(plus(?z:Nat,times(?y:Nat,?y:Nat)),times(plus(?y:Nat,?y:Nat),?z:Nat)),s(times(plus(?z:Nat,?z:Nat),s(?x:Nat)))) = s(plus(plus(?z:Nat,times(?y:Nat,?x:Nat)),s(?x:Nat)))",
"times(times(times(plus(?z:Nat,?y:Nat),?y:Nat),times(plus(?x:Nat,?z:Nat),plus(?x:Nat,?y:Nat))),times(times(times(?y:Nat,?y:Nat),s(?z:Nat)),s(times(?z:Nat,?y:Nat)))) = times(?x:Nat,?y:Nat)",
"times(plus(plus(times(?z:Nat,?z:Nat),s(?x:Nat)),times(plus(?x:Nat,?z:Nat),times(?y:Nat,?z:Nat))),plus(times(?x:Nat,s(?z:Nat)),plus(times(?z:Nat,?y:Nat),s(?z:Nat)))) = s(?x:Nat)",
"plus(s(times(?x:Nat,plus(?z:Nat,?x:Nat))),?y:Nat) = times(plus(plus(?z:Nat,s(?y:Nat)),s(?y:Nat)),times(times(s(?z:Nat),plus(?x:Nat,?z:Nat)),plus(times(?x:Nat,?y:Nat),s(?x:Nat))))",
"times(?y:Nat,plus(plus(?z:Nat,plus(?z:Nat,?y:Nat)),?z:Nat)) = plus(times(times(s(?z:Nat),s(?y:Nat)),?x:Nat),plus(times(plus(?x:Nat,?y:Nat),?x:Nat),times(s(?z:Nat),times(?z:Nat,?z:Nat))))",
"times(?z:Nat,times(plus(?y:Nat,plus(?y:Nat,?y:Nat)),?x:Nat)) = plus(plus(times(?y:Nat,plus(?z:Nat,?x:Nat)),times(?y:Nat,plus(?x:Nat,?x:Nat))),plus(s(s(?y:Nat)),s(plus(?x:Nat,?z:Nat))))",
"s(plus(times(s(?z:Nat),plus(?y:Nat,?x:Nat)),plus(?x:Nat,?x:Nat))) = times(times(s(?z:Nat),?y:Nat),times(times(s(?y:Nat),plus(?x:Nat,?x:Nat)),times(s(?z:Nat),plus(?y:Nat,?x:Nat))))",
"plus(s(s(s(?y:Nat))),plus(times(?y:Nat,s(?y:Nat)),?y:Nat)) = times(times(s(times(?z:Nat,?y:Nat)),times(plus(?z:Nat,?z:Nat),times(?y:Nat,?z:Nat))),s(s(times(?y:Nat,?z:Nat))))",
"plus(plus(s(plus(?x:Nat,?y:Nat)),?z:Nat),s(times(?z:Nat,?z:Nat))) = times(plus(plus(times(?x:Nat,?x:Nat),s(?y:Nat)),times(?x:Nat,s(?x:Nat))),s(plus(s(?x:Nat),s(?z:Nat))))",
"times(s(s(s(?y:Nat))),times(?y:Nat,s(plus(?y:Nat,?z:Nat)))) = times(plus(s(?z:Nat),?x:Nat),plus(plus(plus(?z:Nat,?x:Nat),?y:Nat),times(times(?z:Nat,?y:Nat),plus(?x:Nat,?z:Nat))))",
"plus(?x:Nat,times(plus(s(?z:Nat),s(?x:Nat)),plus(?y:Nat,times(?z:Nat,?z:Nat)))) = times(plus(s(?x:Nat),s(s(?z:Nat))),times(s(times(?x:Nat,?x:Nat)),times(?x:Nat,s(?y:Nat))))",
"plus(?y:Nat,times(plus(?x:Nat,?y:Nat),times(plus(?z:Nat,?z:Nat),plus(?z:Nat,?z:Nat)))) = plus(times(s(?y:Nat),?x:Nat),plus(plus(times(?z:Nat,?y:Nat),?y:Nat),plus(plus(?x:Nat,?y:Nat),?y:Nat)))",
"plus(plus(plus(s(?y:Nat),s(?z:Nat)),?x:Nat),plus(plus(?y:Nat,?y:Nat),?y:Nat)) = plus(s(?y:Nat),times(plus(plus(?z:Nat,?y:Nat),plus(?x:Nat,?y:Nat)),plus(plus(?x:Nat,?y:Nat),?x:Nat)))",
"plus(times(times(s(?x:Nat),s(?z:Nat)),plus(plus(?x:Nat,?y:Nat),?x:Nat)),?x:Nat) = times(s(?x:Nat),times(plus(?z:Nat,plus(?z:Nat,?x:Nat)),times(times(?y:Nat,?z:Nat),times(?x:Nat,?y:Nat))))",
"times(s(times(plus(?x:Nat,?x:Nat),s(?z:Nat))),s(s(plus(?x:Nat,?y:Nat)))) = plus(plus(times(times(?y:Nat,?z:Nat),s(?y:Nat)),s(plus(?x:Nat,?z:Nat))),s(s(s(?y:Nat))))",
"times(times(s(?y:Nat),s(?z:Nat)),plus(plus(plus(?y:Nat,?x:Nat),?x:Nat),?x:Nat)) = times(times(?y:Nat,times(?y:Nat,s(?y:Nat))),plus(plus(?z:Nat,s(?y:Nat)),s(times(?x:Nat,?y:Nat))))",
"plus(times(?x:Nat,s(plus(?z:Nat,?z:Nat))),s(plus(plus(?z:Nat,?z:Nat),s(?x:Nat)))) = times(times(times(s(?x:Nat),times(?x:Nat,?x:Nat)),times(plus(?z:Nat,?y:Nat),?x:Nat)),s(?x:Nat))",
"plus(times(plus(s(?y:Nat),times(?z:Nat,?y:Nat)),times(?y:Nat,plus(?y:Nat,?y:Nat))),?z:Nat) = plus(times(times(?x:Nat,?x:Nat),times(times(?z:Nat,?y:Nat),plus(?y:Nat,?y:Nat))),plus(?y:Nat,?x:Nat))",
"times(times(plus(?z:Nat,plus(?z:Nat,?z:Nat)),plus(times(?x:Nat,?y:Nat),times(?y:Nat,?z:Nat))),?z:Nat) = plus(plus(?x:Nat,times(plus(?x:Nat,?z:Nat),s(?x:Nat))),s(times(?z:Nat,s(?y:Nat))))",
"plus(s(plus(times(?z:Nat,?x:Nat),s(?z:Nat))),plus(?x:Nat,plus(s(?y:Nat),plus(?x:Nat,?y:Nat)))) = plus(plus(plus(?x:Nat,?z:Nat),?z:Nat),plus(plus(?z:Nat,plus(?z:Nat,?x:Nat)),?x:Nat))",
"plus(plus(s(times(?y:Nat,?z:Nat)),s(?x:Nat)),s(plus(times(?z:Nat,?y:Nat),plus(?x:Nat,?z:Nat)))) = times(plus(s(times(?y:Nat,?y:Nat)),s(?y:Nat)),times(?x:Nat,plus(?z:Nat,?z:Nat)))",
"plus(plus(plus(?x:Nat,s(?x:Nat)),?y:Nat),times(times(times(?z:Nat,?x:Nat),times(?y:Nat,?z:Nat)),?x:Nat)) = plus(?z:Nat,plus(s(times(?x:Nat,?y:Nat)),plus(s(?z:Nat),times(?y:Nat,?z:Nat))))",
"times(?x:Nat,plus(times(times(?y:Nat,?x:Nat),s(?z:Nat)),plus(plus(?z:Nat,?y:Nat),plus(?z:Nat,?y:Nat)))) = times(s(?z:Nat),plus(s(s(?x:Nat)),plus(times(?x:Nat,?z:Nat),s(?z:Nat))))",
"times(s(plus(plus(?x:Nat,?x:Nat),?z:Nat)),times(?y:Nat,plus(times(?z:Nat,?z:Nat),plus(?x:Nat,?x:Nat)))) = s(times(times(times(?y:Nat,?y:Nat),s(?x:Nat)),times(s(?x:Nat),s(?x:Nat))))",
"times(s(times(times(?z:Nat,?z:Nat),?x:Nat)),plus(times(plus(?y:Nat,?y:Nat),plus(?y:Nat,?y:Nat)),?x:Nat)) = times(s(s(times(?x:Nat,?x:Nat))),plus(?x:Nat,times(s(?z:Nat),s(?x:Nat))))",
"times(plus(times(times(?y:Nat,?y:Nat),plus(?x:Nat,?z:Nat)),plus(plus(?x:Nat,?z:Nat),?x:Nat)),s(?x:Nat)) = times(plus(plus(s(?z:Nat),plus(?x:Nat,?z:Nat)),plus(?y:Nat,s(?z:Nat))),?x:Nat)",
"times(times(times(plus(?z:Nat,?y:Nat),?y:Nat),plus(?y:Nat,s(?x:Nat))),plus(?z:Nat,s(s(?z:Nat)))) = times(plus(?x:Nat,?z:Nat),plus(s(s(?y:Nat)),times(plus(?z:Nat,?x:Nat),?z:Nat)))",
"times(?y:Nat,times(plus(plus(?x:Nat,?z:Nat),plus(?z:Nat,?x:Nat)),times(times(?z:Nat,?z:Nat),plus(?y:Nat,?y:Nat)))) = plus(times(?y:Nat,s(?x:Nat)),plus(times(?z:Nat,times(?y:Nat,?x:Nat)),?y:Nat))",
"times(times(?x:Nat,s(plus(?y:Nat,?x:Nat))),times(s(plus(?z:Nat,?z:Nat)),plus(plus(?y:Nat,?x:Nat),?z:Nat))) = plus(s(s(s(?x:Nat))),plus(times(?y:Nat,s(?x:Nat)),s(?y:Nat)))",
"plus(s(times(s(?x:Nat),plus(?y:Nat,?z:Nat))),times(times(plus(?x:Nat,?y:Nat),?y:Nat),s(plus(?z:Nat,?y:Nat)))) = s(times(plus(plus(?y:Nat,?z:Nat),?y:Nat),s(plus(?y:Nat,?y:Nat))))",
"plus(plus(s(?z:Nat),times(times(?z:Nat,?x:Nat),plus(?x:Nat,?z:Nat))),s(times(s(?y:Nat),plus(?z:Nat,?z:Nat)))) = times(plus(times(times(?x:Nat,?y:Nat),times(?x:Nat,?z:Nat)),?z:Nat),?z:Nat)",
"plus(s(times(s(?z:Nat),times(?x:Nat,?z:Nat))),times(s(times(?y:Nat,?x:Nat)),plus(s(?x:Nat),times(?z:Nat,?z:Nat)))) = plus(s(times(plus(?y:Nat,?x:Nat),plus(?y:Nat,?z:Nat))),?y:Nat)",
"times(times(times(?x:Nat,?y:Nat),plus(times(?x:Nat,?y:Nat),times(?y:Nat,?z:Nat))),plus(times(?x:Nat,times(?y:Nat,?y:Nat)),?y:Nat)) = plus(s(?x:Nat),plus(times(plus(?y:Nat,?z:Nat),?z:Nat),?z:Nat))",
"plus(times(times(s(?x:Nat),?y:Nat),plus(?y:Nat,times(?x:Nat,?y:Nat))),times(times(?z:Nat,?z:Nat),times(times(?x:Nat,?z:Nat),?z:Nat))) = times(?z:Nat,plus(plus(times(?x:Nat,?z:Nat),?x:Nat),?x:Nat))",
"plus(times(plus(plus(?x:Nat,?z:Nat),plus(?y:Nat,?y:Nat)),times(plus(?z:Nat,?z:Nat),times(?y:Nat,?x:Nat))),times(plus(times(?z:Nat,?z:Nat),s(?y:Nat)),s(?y:Nat))) = s(times(?y:Nat,?y:Nat))",
"plus(times(plus(?x:Nat,plus(?y:Nat,?z:Nat)),times(plus(?y:Nat,?y:Nat),times(?y:Nat,?y:Nat))),times(times(plus(?z:Nat,?y:Nat),s(?x:Nat)),plus(plus(?x:Nat,?y:Nat),?x:Nat))) = plus(?x:Nat,?x:Nat)",
"plus(s(?z:Nat),plus(plus(?y:Nat,s(?y:Nat)),?y:Nat)) = plus(times(plus(plus(?x:Nat,?x:Nat),?z:Nat),times(s(?y:Nat),s(?z:Nat))),times(times(plus(?x:Nat,?y:Nat),?y:Nat),plus(?x:Nat,?y:Nat)))",
"times(times(s(plus(?x:Nat,?z:Nat)),?z:Nat),s(?x:Nat)) = times(plus(plus(times(?z:Nat,?y:Nat),?x:Nat),s(?z:Nat)),plus(times(s(?z:Nat),plus(?z:Nat,?y:Nat)),times(?z:Nat,plus(?y:Nat,?z:Nat))))",
"times(?x:Nat,times(times(s(?y:Nat),plus(?y:Nat,?z:Nat)),?z:Nat)) = plus(times(times(?z:Nat,s(?y:Nat)),plus(plus(?x:Nat,?x:Nat),?z:Nat)),times(?x:Nat,plus(times(?z:Nat,?y:Nat),times(?x:Nat,?x:Nat))))",
"plus(plus(s(?z:Nat),s(times(?z:Nat,?x:Nat))),plus(?z:Nat,s(?z:Nat))) = times(times(plus(times(?x:Nat,?z:Nat),s(?x:Nat)),s(?z:Nat)),plus(times(times(?y:Nat,?y:Nat),s(?z:Nat)),?z:Nat))",
"plus(plus(?z:Nat,s(plus(?z:Nat,?z:Nat))),plus(s(plus(?y:Nat,?z:Nat)),?x:Nat)) = plus(plus(?z:Nat,s(s(?x:Nat))),times(times(s(?x:Nat),s(?z:Nat)),plus(plus(?z:Nat,?y:Nat),?y:Nat)))",
"s(plus(times(s(?y:Nat),plus(?x:Nat,?y:Nat)),times(s(?y:Nat),times(?z:Nat,?z:Nat)))) = times(s(s(plus(?y:Nat,?x:Nat))),times(plus(s(?z:Nat),s(?z:Nat)),plus(s(?x:Nat),?y:Nat)))",
"plus(s(s(s(?z:Nat))),times(plus(plus(?z:Nat,?z:Nat),times(?y:Nat,?z:Nat)),?y:Nat)) = plus(plus(plus(s(?x:Nat),plus(?z:Nat,?x:Nat)),plus(times(?x:Nat,?x:Nat),plus(?x:Nat,?y:Nat))),?y:Nat)",
"plus(s(s(plus(?x:Nat,?z:Nat))),times(?z:Nat,plus(times(?x:Nat,?y:Nat),plus(?y:Nat,?x:Nat)))) = times(s(times(times(?z:Nat,?x:Nat),?y:Nat)),times(times(times(?z:Nat,?y:Nat),s(?x:Nat)),?z:Nat))",
"times(s(times(times(?y:Nat,?z:Nat),plus(?x:Nat,?x:Nat))),s(plus(s(?y:Nat),s(?z:Nat)))) = plus(s(times(?y:Nat,times(?z:Nat,?z:Nat))),times(times(?y:Nat,?z:Nat),s(plus(?z:Nat,?x:Nat))))",
"plus(?y:Nat,plus(plus(plus(?x:Nat,?y:Nat),s(?z:Nat)),times(times(?y:Nat,?x:Nat),plus(?y:Nat,?x:Nat)))) = plus(?y:Nat,plus(plus(s(?y:Nat),times(?z:Nat,?x:Nat)),plus(times(?y:Nat,?z:Nat),?z:Nat)))",
"plus(plus(s(times(?x:Nat,?y:Nat)),plus(s(?x:Nat),times(?y:Nat,?z:Nat))),times(s(?y:Nat),?z:Nat)) = times(times(plus(s(?x:Nat),?z:Nat),plus(s(?y:Nat),?y:Nat)),s(s(s(?y:Nat))))",
"plus(plus(plus(times(?y:Nat,?x:Nat),s(?z:Nat)),plus(plus(?y:Nat,?z:Nat),?z:Nat)),times(?x:Nat,?y:Nat)) = s(times(plus(s(?z:Nat),times(?y:Nat,?y:Nat)),times(s(?y:Nat),plus(?y:Nat,?z:Nat))))",
"plus(times(s(plus(?x:Nat,?z:Nat)),s(plus(?y:Nat,?y:Nat))),times(?y:Nat,s(plus(?z:Nat,?z:Nat)))) = times(plus(s(plus(?z:Nat,?y:Nat)),times(s(?z:Nat),?y:Nat)),s(times(?x:Nat,?y:Nat)))",
"plus(times(?z:Nat,plus(?z:Nat,?y:Nat)),times(plus(s(?x:Nat),s(?y:Nat)),times(?y:Nat,times(?x:Nat,?z:Nat)))) = times(times(plus(?y:Nat,plus(?y:Nat,?z:Nat)),plus(?x:Nat,plus(?z:Nat,?z:Nat))),?x:Nat)",
"plus(times(times(s(?y:Nat),s(?z:Nat)),s(plus(?x:Nat,?x:Nat))),times(s(?y:Nat),s(s(?x:Nat)))) = plus(times(plus(s(?z:Nat),plus(?x:Nat,?z:Nat)),s(s(?x:Nat))),s(?z:Nat))",
"times(times(plus(plus(?z:Nat,?x:Nat),s(?x:Nat)),times(s(?y:Nat),?z:Nat)),s(s(plus(?x:Nat,?x:Nat)))) = times(?y:Nat,plus(plus(s(?y:Nat),times(?y:Nat,?z:Nat)),plus(?x:Nat,s(?y:Nat))))",
"times(times(times(plus(?x:Nat,?y:Nat),times(?z:Nat,?y:Nat)),s(times(?x:Nat,?x:Nat))),plus(?x:Nat,s(?x:Nat))) = s(plus(plus(s(?y:Nat),times(?y:Nat,?z:Nat)),plus(plus(?x:Nat,?x:Nat),?z:Nat)))",
"plus(plus(plus(?z:Nat,?x:Nat),?x:Nat),times(times(s(?x:Nat),times(?z:Nat,?x:Nat)),times(s(?x:Nat),plus(?y:Nat,?z:Nat)))) = s(plus(s(?x:Nat),times(times(?x:Nat,?y:Nat),plus(?x:Nat,?z:Nat))))",
"plus(plus(times(?y:Nat,?x:Nat),plus(plus(?x:Nat,?z:Nat),s(?x:Nat))),times(s(times(?x:Nat,?y:Nat)),s(s(?y:Nat)))) = s(times(s(plus(?x:Nat,?x:Nat)),plus(times(?z:Nat,?y:Nat),?z:Nat)))",
"times(s(plus(times(?y:Nat,?z:Nat),times(?z:Nat,?x:Nat))),plus(times(plus(?z:Nat,?x:Nat),s(?x:Nat)),s(s(?z:Nat)))) = times(times(times(?z:Nat,?x:Nat),plus(s(?z:Nat),?y:Nat)),s(?z:Nat))",
"times(times(plus(?z:Nat,plus(?y:Nat,?z:Nat)),plus(plus(?y:Nat,?z:Nat),?z:Nat)),times(plus(?z:Nat,plus(?z:Nat,?x:Nat)),?z:Nat)) = times(times(?x:Nat,times(?y:Nat,?z:Nat)),s(s(plus(?z:Nat,?y:Nat))))",
"plus(plus(?x:Nat,plus(s(?y:Nat),times(?y:Nat,?z:Nat))),times(plus(?x:Nat,times(?z:Nat,?x:Nat)),times(times(?z:Nat,?z:Nat),s(?z:Nat)))) = s(plus(plus(times(?y:Nat,?x:Nat),s(?x:Nat)),?y:Nat))",
"plus(plus(plus(plus(?y:Nat,?x:Nat),plus(?y:Nat,?y:Nat)),?x:Nat),plus(plus(?z:Nat,s(?x:Nat)),plus(plus(?z:Nat,?x:Nat),s(?z:Nat)))) = plus(?z:Nat,s(times(plus(?z:Nat,?y:Nat),s(?x:Nat))))",
"plus(plus(times(plus(?x:Nat,?x:Nat),?z:Nat),plus(plus(?x:Nat,?y:Nat),s(?y:Nat))),plus(s(s(?y:Nat)),times(s(?x:Nat),?x:Nat))) = s(plus(plus(?z:Nat,?y:Nat),s(plus(?y:Nat,?x:Nat))))",
"times(plus(times(s(?z:Nat),plus(?y:Nat,?x:Nat)),plus(times(?z:Nat,?y:Nat),times(?y:Nat,?y:Nat))),times(s(times(?z:Nat,?y:Nat)),s(?x:Nat))) = plus(plus(?z:Nat,s(plus(?z:Nat,?y:Nat))),?y:Nat)",
"times(times(s(s(?z:Nat)),times(times(?y:Nat,?x:Nat),times(?y:Nat,?y:Nat))),plus(times(times(?x:Nat,?y:Nat),?x:Nat),plus(?y:Nat,s(?z:Nat)))) = plus(?y:Nat,times(times(s(?x:Nat),?y:Nat),?x:Nat))",
"plus(?y:Nat,plus(?z:Nat,?y:Nat)) = times(plus(s(times(?z:Nat,?x:Nat)),times(plus(?x:Nat,?x:Nat),times(?y:Nat,?x:Nat))),plus(plus(times(?x:Nat,?z:Nat),?x:Nat),times(times(?x:Nat,?z:Nat),plus(?z:Nat,?y:Nat))))",
"s(plus(plus(plus(?y:Nat,?x:Nat),?y:Nat),?x:Nat)) = plus(plus(times(?x:Nat,plus(?z:Nat,?z:Nat)),s(times(?x:Nat,?x:Nat))),times(times(times(?x:Nat,?z:Nat),times(?x:Nat,?x:Nat)),plus(s(?y:Nat),?x:Nat)))",
"s(times(plus(?y:Nat,?x:Nat),s(plus(?z:Nat,?z:Nat)))) = plus(plus(s(plus(?x:Nat,?x:Nat)),times(?z:Nat,times(?z:Nat,?z:Nat))),plus(s(times(?z:Nat,?x:Nat)),plus(s(?y:Nat),plus(?z:Nat,?x:Nat))))",
"s(plus(plus(times(?z:Nat,?z:Nat),?y:Nat),times(s(?z:Nat),?y:Nat))) = plus(times(times(plus(?x:Nat,?z:Nat),plus(?y:Nat,?x:Nat)),plus(s(?x:Nat),?x:Nat)),plus(s(?x:Nat),s(times(?x:Nat,?y:Nat))))",
"times(plus(?y:Nat,?y:Nat),times(?x:Nat,plus(plus(?x:Nat,?z:Nat),s(?x:Nat)))) = plus(times(plus(s(?y:Nat),times(?x:Nat,?y:Nat)),?z:Nat),plus(s(plus(?z:Nat,?z:Nat)),plus(times(?y:Nat,?z:Nat),?y:Nat)))",
"plus(times(?x:Nat,times(?y:Nat,s(?y:Nat))),plus(?y:Nat,s(plus(?y:Nat,?y:Nat)))) = times(times(plus(plus(?x:Nat,?z:Nat),s(?x:Nat)),plus(plus(?z:Nat,?y:Nat),times(?x:Nat,?y:Nat))),s(s(?x:Nat)))",
"times(?y:Nat,plus(times(?x:Nat,s(?x:Nat)),plus(s(?z:Nat),plus(?y:Nat,?x:Nat)))) = plus(plus(times(times(?y:Nat,?y:Nat),?y:Nat),times(?x:Nat,?x:Nat)),times(?z:Nat,plus(s(?z:Nat),times(?y:Nat,?x:Nat))))",
"s(plus(plus(times(?y:Nat,?z:Nat),plus(?z:Nat,?z:Nat)),plus(s(?y:Nat),plus(?x:Nat,?x:Nat)))) = plus(times(times(plus(?x:Nat,?y:Nat),times(?y:Nat,?y:Nat)),?z:Nat),times(s(times(?y:Nat,?y:Nat)),?x:Nat))",
"plus(times(plus(?y:Nat,times(?y:Nat,?x:Nat)),plus(plus(?x:Nat,?x:Nat),times(?x:Nat,?y:Nat))),?y:Nat) = times(plus(times(times(?x:Nat,?y:Nat),plus(?x:Nat,?y:Nat)),?y:Nat),plus(?y:Nat,s(plus(?y:Nat,?x:Nat))))",
"times(plus(plus(s(?x:Nat),?x:Nat),plus(?x:Nat,times(?y:Nat,?y:Nat))),s(s(s(?y:Nat)))) = s(plus(plus(plus(?y:Nat,?z:Nat),times(?y:Nat,?z:Nat)),plus(times(?x:Nat,?z:Nat),times(?z:Nat,?y:Nat))))",
"times(times(plus(s(?x:Nat),?z:Nat),?y:Nat),times(times(times(?y:Nat,?z:Nat),plus(?z:Nat,?x:Nat)),?z:Nat)) = times(plus(?z:Nat,times(?x:Nat,?z:Nat)),times(plus(times(?y:Nat,?z:Nat),plus(?z:Nat,?z:Nat)),?z:Nat))",
"plus(times(s(?x:Nat),s(s(?z:Nat))),times(s(?x:Nat),times(times(?y:Nat,?x:Nat),times(?z:Nat,?z:Nat)))) = plus(plus(plus(?y:Nat,?x:Nat),plus(plus(?y:Nat,?y:Nat),times(?x:Nat,?z:Nat))),s(?y:Nat))",
"plus(times(plus(?z:Nat,?y:Nat),?y:Nat),plus(times(?y:Nat,times(?y:Nat,?y:Nat)),times(plus(?z:Nat,?z:Nat),?x:Nat))) = plus(s(times(times(?x:Nat,?z:Nat),plus(?y:Nat,?z:Nat))),times(?y:Nat,s(s(?z:Nat))))",
"times(times(plus(?x:Nat,plus(?y:Nat,?x:Nat)),times(?z:Nat,?x:Nat)),times(?y:Nat,plus(?y:Nat,times(?z:Nat,?y:Nat)))) = plus(times(s(?x:Nat),plus(plus(?x:Nat,?y:Nat),times(?z:Nat,?z:Nat))),plus(?x:Nat,?z:Nat))",
"times(plus(plus(?z:Nat,?z:Nat),plus(s(?y:Nat),plus(?x:Nat,?y:Nat))),times(?x:Nat,plus(?y:Nat,times(?x:Nat,?x:Nat)))) = times(?y:Nat,plus(times(times(?z:Nat,?x:Nat),?y:Nat),plus(plus(?z:Nat,?y:Nat),?x:Nat)))",
"times(plus(times(times(?y:Nat,?z:Nat),times(?x:Nat,?y:Nat)),?y:Nat),plus(times(?y:Nat,times(?z:Nat,?x:Nat)),s(?x:Nat))) = times(plus(?z:Nat,s(plus(?y:Nat,?x:Nat))),s(times(plus(?x:Nat,?x:Nat),?y:Nat)))",
"plus(s(s(plus(?z:Nat,?z:Nat))),times(times(plus(?x:Nat,?x:Nat),plus(?y:Nat,?z:Nat)),times(s(?y:Nat),plus(?y:Nat,?y:Nat)))) = plus(times(?y:Nat,times(s(?x:Nat),times(?y:Nat,?y:Nat))),s(?x:Nat))",
"plus(plus(plus(plus(?y:Nat,?x:Nat),times(?x:Nat,?y:Nat)),plus(plus(?x:Nat,?y:Nat),s(?x:Nat))),times(s(times(?x:Nat,?z:Nat)),?z:Nat)) = plus(s(?x:Nat),plus(?z:Nat,times(plus(?x:Nat,?x:Nat),?z:Nat)))",
"times(s(plus(plus(?y:Nat,?y:Nat),times(?y:Nat,?x:Nat))),times(plus(plus(?x:Nat,?y:Nat),times(?z:Nat,?z:Nat)),times(s(?y:Nat),?x:Nat))) = plus(s(times(s(?x:Nat),?y:Nat)),s(s(s(?y:Nat))))",
"times(plus(s(times(?x:Nat,?z:Nat)),s(s(?x:Nat))),plus(plus(times(?x:Nat,?x:Nat),times(?z:Nat,?z:Nat)),plus(?x:Nat,plus(?y:Nat,?z:Nat)))) = s(plus(?z:Nat,plus(times(?z:Nat,?z:Nat),s(?z:Nat))))",
"times(times(times(plus(?x:Nat,?z:Nat),plus(?x:Nat,?z:Nat)),?x:Nat),times(times(plus(?y:Nat,?x:Nat),plus(?x:Nat,?z:Nat)),s(times(?z:Nat,?z:Nat)))) = times(times(s(?x:Nat),?z:Nat),s(plus(?y:Nat,?z:Nat)))",
"plus(times(plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),times(plus(?z:Nat,?y:Nat),times(?z:Nat,?y:Nat))),times(times(plus(?x:Nat,?y:Nat),times(?z:Nat,?z:Nat)),?x:Nat)) = times(plus(?x:Nat,s(?z:Nat)),?z:Nat)",
"times(s(plus(s(?x:Nat),s(?z:Nat))),?x:Nat) = plus(plus(times(s(?x:Nat),plus(?x:Nat,?x:Nat)),plus(plus(?y:Nat,?x:Nat),s(?z:Nat))),plus(plus(times(?x:Nat,?x:Nat),times(?z:Nat,?z:Nat)),s(?z:Nat)))",
"times(s(plus(times(?y:Nat,?y:Nat),times(?z:Nat,?z:Nat))),?x:Nat) = plus(times(times(s(?y:Nat),times(?z:Nat,?y:Nat)),?z:Nat),times(plus(times(?y:Nat,?y:Nat),s(?x:Nat)),plus(plus(?z:Nat,?y:Nat),s(?z:Nat))))",
"s(plus(times(?x:Nat,times(?x:Nat,?y:Nat)),s(plus(?x:Nat,?z:Nat)))) = times(times(?y:Nat,times(s(?y:Nat),times(?y:Nat,?x:Nat))),times(times(plus(?z:Nat,?z:Nat),?z:Nat),times(times(?x:Nat,?y:Nat),s(?z:Nat))))",
"plus(s(s(times(?z:Nat,?x:Nat))),times(s(times(?x:Nat,?y:Nat)),s(s(?z:Nat)))) = plus(s(times(times(?y:Nat,?y:Nat),times(?y:Nat,?z:Nat))),plus(times(times(?y:Nat,?y:Nat),?z:Nat),s(s(?x:Nat))))",
"plus(times(s(?x:Nat),s(s(?z:Nat))),plus(?y:Nat,times(plus(?z:Nat,?z:Nat),?y:Nat))) = plus(plus(times(?y:Nat,s(?y:Nat)),s(?x:Nat)),plus(times(plus(?x:Nat,?z:Nat),s(?y:Nat)),s(s(?y:Nat))))",
"times(plus(?y:Nat,?z:Nat),times(plus(?y:Nat,plus(?y:Nat,?x:Nat)),s(plus(?y:Nat,?z:Nat)))) = times(plus(times(times(?y:Nat,?z:Nat),?x:Nat),plus(?x:Nat,times(?z:Nat,?z:Nat))),plus(s(plus(?y:Nat,?y:Nat)),?x:Nat))",
"times(plus(plus(plus(?z:Nat,?y:Nat),s(?z:Nat)),times(?y:Nat,?x:Nat)),s(plus(?x:Nat,?z:Nat))) = plus(times(?y:Nat,?y:Nat),plus(times(plus(?x:Nat,?z:Nat),times(?z:Nat,?x:Nat)),plus(?z:Nat,plus(?x:Nat,?z:Nat))))",
"plus(plus(?z:Nat,times(plus(?y:Nat,?y:Nat),plus(?z:Nat,?z:Nat))),times(?x:Nat,plus(s(?z:Nat),s(?x:Nat)))) = plus(times(?y:Nat,?z:Nat),times(times(times(?x:Nat,?z:Nat),?y:Nat),times(?y:Nat,plus(?z:Nat,?z:Nat))))",
"plus(times(?y:Nat,plus(s(?y:Nat),?y:Nat)),plus(s(?z:Nat),times(times(?x:Nat,?z:Nat),plus(?y:Nat,?y:Nat)))) = plus(times(times(plus(?y:Nat,?y:Nat),?z:Nat),plus(?x:Nat,plus(?y:Nat,?x:Nat))),s(s(?z:Nat)))",
"times(times(s(s(?z:Nat)),plus(times(?x:Nat,?y:Nat),plus(?z:Nat,?y:Nat))),times(?x:Nat,times(?y:Nat,?x:Nat))) = plus(times(times(?z:Nat,plus(?z:Nat,?x:Nat)),?z:Nat),times(s(times(?y:Nat,?x:Nat)),s(?z:Nat)))",
"plus(times(times(times(?y:Nat,?x:Nat),s(?y:Nat)),?z:Nat),times(plus(s(?x:Nat),?y:Nat),plus(?x:Nat,s(?z:Nat)))) = plus(times(?x:Nat,plus(times(?x:Nat,?z:Nat),times(?z:Nat,?x:Nat))),plus(s(?z:Nat),?y:Nat))",
"times(plus(times(?x:Nat,times(?x:Nat,?y:Nat)),?z:Nat),plus(plus(times(?z:Nat,?z:Nat),plus(?z:Nat,?z:Nat)),s(?x:Nat))) = times(times(s(?y:Nat),?z:Nat),plus(plus(plus(?z:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),?y:Nat))",
"times(s(s(s(?z:Nat))),plus(times(plus(?z:Nat,?z:Nat),times(?y:Nat,?z:Nat)),times(times(?x:Nat,?z:Nat),s(?z:Nat)))) = plus(plus(times(s(?y:Nat),?x:Nat),?x:Nat),s(plus(s(?x:Nat),s(?z:Nat))))",
"times(plus(times(?z:Nat,plus(?z:Nat,?x:Nat)),times(times(?y:Nat,?z:Nat),plus(?x:Nat,?z:Nat))),s(s(plus(?x:Nat,?y:Nat)))) = plus(times(times(s(?z:Nat),times(?x:Nat,?z:Nat)),?x:Nat),s(plus(?z:Nat,?y:Nat)))",
"times(times(plus(times(?y:Nat,?x:Nat),times(?z:Nat,?z:Nat)),times(?z:Nat,times(?x:Nat,?z:Nat))),s(times(?z:Nat,s(?y:Nat)))) = plus(times(?y:Nat,?x:Nat),plus(?z:Nat,plus(times(?y:Nat,?z:Nat),times(?x:Nat,?z:Nat))))",
"times(plus(plus(times(?z:Nat,?x:Nat),times(?y:Nat,?x:Nat)),plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,?y:Nat))),s(s(s(?y:Nat)))) = s(times(times(plus(?y:Nat,?z:Nat),?z:Nat),plus(?x:Nat,plus(?y:Nat,?x:Nat))))",
"plus(plus(times(times(?y:Nat,?z:Nat),?y:Nat),plus(plus(?z:Nat,?y:Nat),s(?z:Nat))),s(plus(plus(?z:Nat,?x:Nat),times(?y:Nat,?x:Nat)))) = plus(plus(plus(?x:Nat,?y:Nat),plus(plus(?z:Nat,?x:Nat),?z:Nat)),?y:Nat)",
"times(?y:Nat,plus(plus(times(?x:Nat,?x:Nat),plus(?z:Nat,?x:Nat)),?y:Nat)) = times(plus(times(times(?y:Nat,?x:Nat),?x:Nat),times(times(?y:Nat,?y:Nat),?y:Nat)),times(times(?x:Nat,plus(?x:Nat,?y:Nat)),times(?x:Nat,s(?z:Nat))))",
"s(times(plus(plus(?y:Nat,?x:Nat),times(?x:Nat,?x:Nat)),plus(plus(?y:Nat,?z:Nat),?z:Nat))) = times(plus(plus(plus(?z:Nat,?y:Nat),?y:Nat),times(s(?y:Nat),s(?x:Nat))),plus(s(times(?z:Nat,?z:Nat)),s(?z:Nat)))",
"times(times(s(plus(?x:Nat,?y:Nat)),plus(plus(?y:Nat,?x:Nat),plus(?y:Nat,?x:Nat))),?x:Nat) = plus(s(times(times(?x:Nat,?z:Nat),?x:Nat)),times(plus(plus(?y:Nat,?z:Nat),?y:Nat),plus(plus(?y:Nat,?y:Nat),s(?x:Nat))))",
"plus(plus(s(times(?z:Nat,?y:Nat)),?y:Nat),plus(?y:Nat,plus(s(?x:Nat),times(?x:Nat,?y:Nat)))) = times(times(plus(times(?z:Nat,?y:Nat),times(?x:Nat,?z:Nat)),times(times(?z:Nat,?y:Nat),?y:Nat)),s(plus(?y:Nat,?x:Nat)))",
"plus(s(plus(plus(?y:Nat,?z:Nat),s(?z:Nat))),s(times(times(?x:Nat,?x:Nat),plus(?x:Nat,?z:Nat)))) = plus(s(times(?x:Nat,times(?x:Nat,?y:Nat))),times(times(?y:Nat,times(?z:Nat,?z:Nat)),plus(s(?y:Nat),?z:Nat)))",
"times(times(plus(s(?y:Nat),times(?x:Nat,?x:Nat)),plus(?y:Nat,times(?x:Nat,?y:Nat))),times(?x:Nat,plus(?x:Nat,?x:Nat))) = plus(times(s(times(?y:Nat,?x:Nat)),plus(?y:Nat,s(?y:Nat))),plus(?x:Nat,times(?y:Nat,?z:Nat)))",
"plus(s(s(s(?y:Nat))),plus(plus(s(?x:Nat),plus(?z:Nat,?y:Nat)),plus(plus(?z:Nat,?y:Nat),times(?z:Nat,?x:Nat)))) = times(?x:Nat,times(times(times(?x:Nat,?y:Nat),plus(?z:Nat,?y:Nat)),plus(?y:Nat,s(?y:Nat))))",
"times(plus(times(s(?x:Nat),times(?x:Nat,?y:Nat)),plus(s(?y:Nat),plus(?y:Nat,?z:Nat))),s(times(times(?x:Nat,?x:Nat),?y:Nat))) = times(?z:Nat,times(times(times(?x:Nat,?y:Nat),s(?z:Nat)),s(times(?y:Nat,?y:Nat))))",
"times(times(plus(?x:Nat,?y:Nat),times(s(?y:Nat),?z:Nat)),plus(times(?x:Nat,?y:Nat),times(times(?x:Nat,?x:Nat),plus(?x:Nat,?z:Nat)))) = plus(times(s(times(?z:Nat,?x:Nat)),times(s(?y:Nat),plus(?x:Nat,?x:Nat))),?y:Nat)",
"times(times(plus(plus(?x:Nat,?z:Nat),?z:Nat),s(plus(?x:Nat,?z:Nat))),plus(plus(plus(?z:Nat,?z:Nat),times(?x:Nat,?x:Nat)),?x:Nat)) = times(s(plus(times(?y:Nat,?x:Nat),?z:Nat)),times(?y:Nat,times(s(?x:Nat),?x:Nat)))",
"plus(s(plus(s(?z:Nat),s(?x:Nat))),plus(times(plus(?y:Nat,?y:Nat),plus(?y:Nat,?z:Nat)),plus(times(?z:Nat,?x:Nat),s(?y:Nat)))) = times(plus(plus(?z:Nat,?y:Nat),?y:Nat),plus(s(times(?y:Nat,?z:Nat)),?y:Nat))",
"times(plus(s(s(?y:Nat)),plus(plus(?y:Nat,?z:Nat),?y:Nat)),plus(times(plus(?x:Nat,?y:Nat),?y:Nat),plus(?x:Nat,plus(?z:Nat,?z:Nat)))) = plus(times(?x:Nat,?y:Nat),s(plus(times(?z:Nat,?y:Nat),plus(?y:Nat,?z:Nat))))",
"plus(times(plus(plus(?x:Nat,?y:Nat),?x:Nat),plus(times(?x:Nat,?y:Nat),plus(?x:Nat,?z:Nat))),times(times(?x:Nat,times(?x:Nat,?x:Nat)),times(?x:Nat,s(?x:Nat)))) = times(s(plus(times(?z:Nat,?x:Nat),?x:Nat)),s(?z:Nat))",
"plus(plus(times(?y:Nat,?x:Nat),times(times(?y:Nat,?z:Nat),plus(?z:Nat,?y:Nat))),?z:Nat) = times(plus(s(times(?x:Nat,?x:Nat)),times(plus(?x:Nat,?x:Nat),times(?z:Nat,?y:Nat))),plus(?y:Nat,times(s(?z:Nat),times(?z:Nat,?z:Nat))))",
"times(s(times(?y:Nat,s(?x:Nat))),times(?z:Nat,plus(plus(?z:Nat,?z:Nat),plus(?x:Nat,?y:Nat)))) = times(plus(plus(s(?z:Nat),plus(?x:Nat,?x:Nat)),?x:Nat),plus(s(?z:Nat),plus(times(?y:Nat,?x:Nat),plus(?z:Nat,?y:Nat))))",
"plus(plus(?y:Nat,times(?y:Nat,plus(?y:Nat,?y:Nat))),times(plus(?z:Nat,?y:Nat),times(times(?y:Nat,?z:Nat),?x:Nat))) = times(times(times(times(?z:Nat,?z:Nat),s(?z:Nat)),?x:Nat),times(?x:Nat,times(plus(?y:Nat,?z:Nat),s(?x:Nat))))",
"plus(times(?y:Nat,plus(s(?y:Nat),times(?z:Nat,?y:Nat))),plus(plus(s(?z:Nat),times(?x:Nat,?y:Nat)),?z:Nat)) = times(plus(s(?z:Nat),plus(times(?z:Nat,?x:Nat),times(?z:Nat,?z:Nat))),times(?x:Nat,s(times(?z:Nat,?x:Nat))))",
"plus(times(plus(s(?y:Nat),times(?x:Nat,?z:Nat)),plus(?x:Nat,s(?z:Nat))),s(s(times(?z:Nat,?y:Nat)))) = times(s(s(plus(?x:Nat,?x:Nat))),plus(plus(s(?z:Nat),?y:Nat),plus(s(?x:Nat),times(?z:Nat,?x:Nat))))",
"plus(plus(times(s(?y:Nat),?y:Nat),s(plus(?z:Nat,?z:Nat))),s(plus(times(?z:Nat,?x:Nat),plus(?z:Nat,?y:Nat)))) = times(plus(s(?z:Nat),plus(?x:Nat,plus(?z:Nat,?x:Nat))),s(plus(times(?y:Nat,?x:Nat),s(?x:Nat))))",
"plus(times(plus(plus(?y:Nat,?x:Nat),?z:Nat),s(plus(?z:Nat,?z:Nat))),s(plus(times(?x:Nat,?x:Nat),s(?x:Nat)))) = times(times(times(s(?x:Nat),?y:Nat),?z:Nat),plus(s(times(?z:Nat,?x:Nat)),times(s(?z:Nat),?y:Nat)))",
"times(times(s(?x:Nat),times(?z:Nat,plus(?x:Nat,?x:Nat))),plus(plus(s(?x:Nat),?z:Nat),plus(times(?y:Nat,?y:Nat),?z:Nat))) = times(plus(times(?y:Nat,?y:Nat),?x:Nat),times(times(plus(?x:Nat,?z:Nat),times(?y:Nat,?y:Nat)),?x:Nat))",
"times(plus(times(times(?y:Nat,?x:Nat),plus(?x:Nat,?y:Nat)),?x:Nat),plus(plus(?x:Nat,s(?x:Nat)),times(times(?z:Nat,?y:Nat),s(?z:Nat)))) = times(s(?z:Nat),plus(plus(s(?z:Nat),times(?y:Nat,?y:Nat)),times(?y:Nat,?y:Nat)))",
"times(?y:Nat,plus(s(?x:Nat),times(plus(?y:Nat,?x:Nat),plus(?x:Nat,?z:Nat)))) = times(times(?y:Nat,plus(times(?z:Nat,?z:Nat),times(?x:Nat,?z:Nat))),times(times(?y:Nat,times(?x:Nat,?x:Nat)),times(times(?z:Nat,?y:Nat),times(?z:Nat,?x:Nat))))",
"plus(s(?y:Nat),times(s(plus(?z:Nat,?x:Nat)),times(times(?y:Nat,?y:Nat),s(?y:Nat)))) = times(plus(plus(plus(?y:Nat,?y:Nat),times(?y:Nat,?x:Nat)),times(times(?z:Nat,?y:Nat),?y:Nat)),plus(times(?z:Nat,plus(?y:Nat,?z:Nat)),?z:Nat))",
"times(plus(times(times(?y:Nat,?z:Nat),s(?z:Nat)),?x:Nat),plus(times(?y:Nat,s(?x:Nat)),s(?z:Nat))) = times(plus(?y:Nat,s(plus(?z:Nat,?z:Nat))),plus(s(times(?y:Nat,?y:Nat)),times(plus(?z:Nat,?z:Nat),plus(?y:Nat,?z:Nat))))",
"times(times(times(?x:Nat,s(?z:Nat)),s(times(?x:Nat,?x:Nat))),plus(?z:Nat,plus(?z:Nat,plus(?y:Nat,?z:Nat)))) = plus(s(s(times(?z:Nat,?y:Nat))),times(times(times(?x:Nat,?z:Nat),s(?x:Nat)),plus(times(?z:Nat,?z:Nat),?z:Nat)))",
"times(times(times(times(?x:Nat,?x:Nat),plus(?z:Nat,?x:Nat)),s(?y:Nat)),times(?y:Nat,s(plus(?x:Nat,?x:Nat)))) = times(plus(s(times(?y:Nat,?x:Nat)),times(s(?z:Nat),s(?x:Nat))),plus(plus(?y:Nat,s(?z:Nat)),s(?y:Nat)))",
"times(plus(?z:Nat,times(plus(?z:Nat,?y:Nat),plus(?z:Nat,?x:Nat))),plus(times(times(?z:Nat,?y:Nat),times(?z:Nat,?x:Nat)),?x:Nat)) = s(times(plus(plus(?y:Nat,?y:Nat),times(?x:Nat,?x:Nat)),plus(times(?y:Nat,?y:Nat),plus(?y:Nat,?z:Nat))))",
"plus(plus(plus(plus(?z:Nat,?y:Nat),?x:Nat),s(plus(?x:Nat,?z:Nat))),plus(times(s(?x:Nat),times(?x:Nat,?y:Nat)),s(?y:Nat))) = plus(s(s(plus(?z:Nat,?z:Nat))),plus(s(plus(?x:Nat,?x:Nat)),plus(s(?y:Nat),?z:Nat)))",
"plus(plus(times(plus(?x:Nat,?z:Nat),?x:Nat),times(?x:Nat,times(?y:Nat,?y:Nat))),plus(plus(s(?z:Nat),times(?z:Nat,?x:Nat)),plus(s(?z:Nat),?y:Nat))) = plus(times(s(s(?z:Nat)),plus(s(?x:Nat),?y:Nat)),times(?y:Nat,?y:Nat))",
"times(times(times(?x:Nat,plus(?x:Nat,?x:Nat)),times(plus(?z:Nat,?x:Nat),s(?x:Nat))),times(s(times(?z:Nat,?z:Nat)),plus(s(?x:Nat),times(?y:Nat,?x:Nat)))) = times(times(s(s(?x:Nat)),s(s(?y:Nat))),plus(?y:Nat,?x:Nat))",
"plus(plus(times(?z:Nat,s(?z:Nat)),times(?z:Nat,times(?x:Nat,?x:Nat))),plus(times(plus(?z:Nat,?y:Nat),times(?y:Nat,?x:Nat)),plus(times(?z:Nat,?y:Nat),plus(?z:Nat,?z:Nat)))) = plus(plus(s(times(?y:Nat,?x:Nat)),s(?x:Nat)),?z:Nat)",
"plus(s(plus(?x:Nat,?y:Nat)),plus(times(s(?z:Nat),s(?z:Nat)),s(times(?y:Nat,?y:Nat)))) = plus(times(s(?y:Nat),s(plus(?x:Nat,?y:Nat))),times(times(plus(?y:Nat,?z:Nat),s(?z:Nat)),times(s(?y:Nat),plus(?z:Nat,?x:Nat))))",
"plus(plus(plus(s(?x:Nat),?x:Nat),s(plus(?y:Nat,?y:Nat))),s(plus(?z:Nat,plus(?z:Nat,?y:Nat)))) = times(plus(plus(s(?y:Nat),plus(?y:Nat,?x:Nat)),plus(plus(?z:Nat,?x:Nat),s(?y:Nat))),times(s(times(?z:Nat,?z:Nat)),?x:Nat))",
"times(plus(plus(s(?z:Nat),s(?x:Nat)),plus(plus(?x:Nat,?z:Nat),?y:Nat)),times(times(times(?z:Nat,?x:Nat),s(?y:Nat)),times(?x:Nat,?z:Nat))) = s(plus(times(plus(?y:Nat,?y:Nat),s(?z:Nat)),plus(plus(?z:Nat,?x:Nat),s(?x:Nat))))",
"plus(plus(plus(times(?y:Nat,?y:Nat),times(?z:Nat,?z:Nat)),times(plus(?x:Nat,?y:Nat),s(?y:Nat))),times(times(s(?y:Nat),plus(?y:Nat,?x:Nat)),times(s(?x:Nat),times(?z:Nat,?x:Nat)))) = s(times(times(?x:Nat,plus(?y:Nat,?y:Nat)),?y:Nat))",
"times(times(?x:Nat,s(s(?y:Nat))),plus(?z:Nat,times(?y:Nat,s(?x:Nat)))) = times(plus(times(plus(?y:Nat,?z:Nat),times(?x:Nat,?z:Nat)),plus(times(?x:Nat,?z:Nat),?x:Nat)),plus(times(times(?y:Nat,?x:Nat),plus(?x:Nat,?z:Nat)),s(s(?y:Nat))))",
"plus(s(plus(s(?z:Nat),plus(?z:Nat,?z:Nat))),plus(times(s(?y:Nat),?y:Nat),?y:Nat)) = plus(plus(?x:Nat,plus(s(?x:Nat),times(?x:Nat,?x:Nat))),times(plus(plus(?x:Nat,?z:Nat),times(?y:Nat,?x:Nat)),times(s(?y:Nat),times(?y:Nat,?x:Nat))))",
"s(plus(times(plus(?y:Nat,?x:Nat),times(?x:Nat,?y:Nat)),times(times(?x:Nat,?y:Nat),plus(?y:Nat,?z:Nat)))) = times(times(plus(times(?z:Nat,?y:Nat),plus(?z:Nat,?x:Nat)),?z:Nat),times(times(times(?y:Nat,?x:Nat),times(?z:Nat,?y:Nat)),s(s(?y:Nat))))",
"plus(s(plus(times(?z:Nat,?y:Nat),plus(?x:Nat,?y:Nat))),plus(times(?z:Nat,s(?x:Nat)),plus(s(?z:Nat),times(?y:Nat,?z:Nat)))) = times(plus(times(plus(?x:Nat,?y:Nat),times(?x:Nat,?x:Nat)),plus(plus(?y:Nat,?y:Nat),plus(?y:Nat,?y:Nat))),?z:Nat)",
"plus(plus(plus(?z:Nat,times(?y:Nat,?y:Nat)),times(plus(?z:Nat,?x:Nat),plus(?x:Nat,?y:Nat))),times(?z:Nat,times(?z:Nat,times(?z:Nat,?y:Nat)))) = plus(times(times(times(?z:Nat,?y:Nat),?y:Nat),?y:Nat),plus(?z:Nat,times(s(?z:Nat),times(?x:Nat,?y:Nat))))",
"times(times(times(plus(?x:Nat,?z:Nat),plus(?z:Nat,?y:Nat)),plus(times(?y:Nat,?y:Nat),?y:Nat)),plus(s(?y:Nat),s(plus(?y:Nat,?z:Nat)))) = plus(times(times(?y:Nat,s(?z:Nat)),plus(plus(?y:Nat,?y:Nat),plus(?x:Nat,?z:Nat))),plus(?x:Nat,?y:Nat))",
"plus(times(plus(times(?x:Nat,?x:Nat),s(?z:Nat)),plus(?y:Nat,s(?y:Nat))),plus(times(?y:Nat,?y:Nat),?z:Nat)) = times(plus(times(times(?x:Nat,?z:Nat),?z:Nat),s(s(?x:Nat))),plus(times(s(?y:Nat),times(?y:Nat,?y:Nat)),s(plus(?x:Nat,?z:Nat))))",
"times(plus(s(?z:Nat),times(s(?x:Nat),s(?y:Nat))),s(plus(plus(?y:Nat,?z:Nat),times(?z:Nat,?y:Nat)))) = times(times(plus(times(?x:Nat,?x:Nat),?x:Nat),s(times(?z:Nat,?z:Nat))),plus(s(s(?x:Nat)),plus(times(?x:Nat,?x:Nat),s(?z:Nat))))",
"plus(times(plus(s(?z:Nat),times(?x:Nat,?x:Nat)),plus(?z:Nat,plus(?x:Nat,?x:Nat))),s(times(times(?x:Nat,?y:Nat),times(?x:Nat,?x:Nat)))) = plus(s(plus(s(?x:Nat),s(?y:Nat))),plus(times(?y:Nat,?y:Nat),times(plus(?x:Nat,?x:Nat),s(?x:Nat))))",
"times(times(times(?x:Nat,times(?y:Nat,?z:Nat)),s(s(?z:Nat))),plus(times(times(?z:Nat,?z:Nat),?y:Nat),plus(s(?z:Nat),times(?z:Nat,?x:Nat)))) = plus(?y:Nat,plus(times(times(?x:Nat,?x:Nat),s(?z:Nat)),times(plus(?x:Nat,?z:Nat),times(?y:Nat,?x:Nat))))",
"plus(plus(times(times(?z:Nat,?z:Nat),plus(?y:Nat,?z:Nat)),?y:Nat),s(plus(plus(?y:Nat,?z:Nat),times(?x:Nat,?x:Nat)))) = times(s(times(times(?z:Nat,?x:Nat),plus(?z:Nat,?y:Nat))),times(times(plus(?x:Nat,?y:Nat),times(?z:Nat,?y:Nat)),s(plus(?x:Nat,?z:Nat))))",
"times(plus(times(times(?x:Nat,?x:Nat),plus(?y:Nat,?x:Nat)),plus(?y:Nat,times(?y:Nat,?z:Nat))),s(times(plus(?x:Nat,?z:Nat),plus(?y:Nat,?z:Nat)))) = plus(times(?z:Nat,plus(times(?x:Nat,?x:Nat),plus(?y:Nat,?y:Nat))),plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),?z:Nat))",
"plus(plus(s(plus(?z:Nat,?x:Nat)),s(plus(?z:Nat,?y:Nat))),plus(times(s(?y:Nat),plus(?y:Nat,?y:Nat)),plus(times(?z:Nat,?x:Nat),times(?z:Nat,?z:Nat)))) = plus(times(?z:Nat,plus(?y:Nat,s(?z:Nat))),s(plus(times(?x:Nat,?x:Nat),times(?z:Nat,?x:Nat))))",
"plus(plus(plus(times(?x:Nat,?z:Nat),?y:Nat),plus(?y:Nat,times(?y:Nat,?x:Nat))),times(times(times(?z:Nat,?z:Nat),s(?x:Nat)),plus(times(?y:Nat,?x:Nat),?y:Nat))) = times(s(times(times(?x:Nat,?x:Nat),?z:Nat)),plus(times(?x:Nat,?y:Nat),s(plus(?x:Nat,?y:Nat))))",
"plus(times(plus(?z:Nat,plus(?x:Nat,?x:Nat)),plus(times(?z:Nat,?x:Nat),s(?y:Nat))),times(s(?y:Nat),s(times(?z:Nat,?y:Nat)))) = times(plus(plus(?x:Nat,times(?z:Nat,?y:Nat)),times(?x:Nat,plus(?z:Nat,?x:Nat))),times(s(times(?z:Nat,?x:Nat)),s(times(?y:Nat,?y:Nat))))",
"plus(times(plus(s(?x:Nat),times(?x:Nat,?x:Nat)),s(s(?z:Nat))),times(times(s(?y:Nat),times(?x:Nat,?z:Nat)),s(times(?y:Nat,?z:Nat)))) = plus(times(plus(s(?y:Nat),times(?z:Nat,?x:Nat)),s(times(?y:Nat,?z:Nat))),plus(times(?z:Nat,s(?z:Nat)),plus(?y:Nat,s(?x:Nat))))",
"plus(times(plus(s(?y:Nat),?z:Nat),plus(times(?x:Nat,?z:Nat),?y:Nat)),plus(plus(times(?z:Nat,?z:Nat),plus(?x:Nat,?x:Nat)),?z:Nat)) = times(plus(plus(times(?z:Nat,?x:Nat),plus(?z:Nat,?z:Nat)),times(s(?x:Nat),times(?x:Nat,?y:Nat))),times(times(?x:Nat,times(?y:Nat,?z:Nat)),plus(plus(?y:Nat,?x:Nat),?y:Nat)))"
]

val problemsPlusWithZeroYes =
IOMstrs.rdEqs [
"s(?y:Nat) = plus(s(?y:Nat),0)",
"s(?y:Nat) = plus(s(0),?y:Nat)",
"s(?z:Nat) = plus(?z:Nat,s(0))",
"plus(0,?x:Nat) = plus(?x:Nat,0)",
"plus(0,s(?y:Nat)) = s(?y:Nat)",
"plus(0,s(0)) = s(0)",
"plus(s(0),0) = s(0)",
"s(?x:Nat) = plus(?x:Nat,s(plus(0,0)))",
"s(?x:Nat) = plus(0,plus(?x:Nat,s(0)))",
"s(?x:Nat) = plus(0,plus(s(?x:Nat),0))",
"s(?x:Nat) = plus(plus(0,s(?x:Nat)),0)",
"s(?x:Nat) = plus(plus(s(0),?x:Nat),0)",
"s(?y:Nat) = plus(s(0),plus(?y:Nat,0))",
"s(?z:Nat) = plus(0,s(plus(?z:Nat,0)))",
"s(?z:Nat) = plus(0,plus(0,s(?z:Nat)))",
"plus(?x:Nat,?x:Nat) = plus(?x:Nat,plus(?x:Nat,0))",
"plus(?x:Nat,?x:Nat) = plus(0,plus(?x:Nat,?x:Nat))",
"plus(?x:Nat,?x:Nat) = plus(plus(?x:Nat,?x:Nat),0)",
"plus(?x:Nat,?x:Nat) = plus(plus(?x:Nat,0),?x:Nat)",
"plus(?x:Nat,?y:Nat) = plus(?x:Nat,plus(?y:Nat,0))",
"plus(?y:Nat,?x:Nat) = plus(?x:Nat,plus(0,?y:Nat))",
"plus(?z:Nat,?x:Nat) = plus(plus(?x:Nat,?z:Nat),0)",
"plus(?z:Nat,?y:Nat) = plus(0,plus(?y:Nat,?z:Nat))",
"plus(0,?y:Nat) = plus(0,plus(?y:Nat,0))",
"s(plus(0,?y:Nat)) = plus(s(?y:Nat),0)",
"plus(?z:Nat,s(0)) = plus(s(?z:Nat),0)",
"plus(0,s(0)) = plus(s(0),0)",
"plus(0,plus(?x:Nat,0)) = plus(?x:Nat,0)",
"plus(0,plus(?y:Nat,?x:Nat)) = plus(?y:Nat,?x:Nat)",
"plus(s(?z:Nat),s(0)) = s(s(?z:Nat))",
"plus(plus(?x:Nat,0),?y:Nat) = plus(?y:Nat,?x:Nat)",
"plus(plus(?x:Nat,0),?z:Nat) = plus(?x:Nat,?z:Nat)",
"plus(plus(0,?x:Nat),?y:Nat) = plus(?x:Nat,?y:Nat)",
"plus(plus(0,?y:Nat),?x:Nat) = plus(?x:Nat,?y:Nat)",
"plus(plus(0,?y:Nat),?y:Nat) = plus(?y:Nat,?y:Nat)",
"plus(plus(0,0),?x:Nat) = plus(0,?x:Nat)",
"plus(s(plus(?x:Nat,0)),0) = s(?x:Nat)",
"plus(plus(?x:Nat,0),s(0)) = s(?x:Nat)",
"plus(plus(?z:Nat,s(0)),0) = s(?z:Nat)",
"plus(plus(0,0),s(?x:Nat)) = s(?x:Nat)",
"s(?y:Nat) = plus(plus(s(0),0),plus(?y:Nat,0))",
"s(s(?y:Nat)) = plus(s(?y:Nat),s(plus(0,0)))",
"plus(?x:Nat,?y:Nat) = plus(0,plus(?y:Nat,plus(?x:Nat,0)))",
"plus(?y:Nat,?z:Nat) = plus(0,plus(?y:Nat,plus(?z:Nat,0)))",
"s(plus(?y:Nat,?x:Nat)) = plus(plus(0,s(?y:Nat)),?x:Nat)",
"s(plus(?y:Nat,?y:Nat)) = plus(s(plus(?y:Nat,0)),?y:Nat)",
"s(plus(0,?x:Nat)) = plus(plus(?x:Nat,0),s(0))",
"plus(?x:Nat,s(?x:Nat)) = s(plus(plus(?x:Nat,?x:Nat),0))",
"plus(?x:Nat,s(?x:Nat)) = plus(s(plus(?x:Nat,?x:Nat)),0)",
"plus(?x:Nat,s(?y:Nat)) = plus(?x:Nat,plus(?y:Nat,s(0)))",
"plus(?x:Nat,s(?y:Nat)) = plus(plus(?x:Nat,s(0)),?y:Nat)",
"plus(?y:Nat,s(?z:Nat)) = plus(plus(?y:Nat,0),s(?z:Nat))",
"plus(s(?x:Nat),?y:Nat) = s(plus(plus(?x:Nat,0),?y:Nat))",
"plus(s(?y:Nat),?x:Nat) = plus(plus(s(?x:Nat),?y:Nat),0)",
"plus(s(?y:Nat),?z:Nat) = s(plus(?y:Nat,plus(0,?z:Nat)))",
"plus(s(?y:Nat),?z:Nat) = plus(plus(?z:Nat,s(0)),?y:Nat)",
"plus(plus(0,?y:Nat),?x:Nat) = plus(plus(?y:Nat,0),?x:Nat)",
"s(plus(plus(?x:Nat,0),0)) = plus(s(?x:Nat),0)",
"plus(?y:Nat,s(plus(0,?x:Nat))) = plus(?y:Nat,s(?x:Nat))",
"plus(?y:Nat,plus(0,s(?x:Nat))) = s(plus(?y:Nat,?x:Nat))",
"plus(?y:Nat,plus(s(0),?z:Nat)) = plus(s(?z:Nat),?y:Nat)",
"plus(0,s(plus(?x:Nat,?x:Nat))) = plus(?x:Nat,s(?x:Nat))",
"plus(s(plus(0,?x:Nat)),?z:Nat) = s(plus(?x:Nat,?z:Nat))",
"plus(plus(?x:Nat,s(?x:Nat)),0) = plus(s(?x:Nat),?x:Nat)",
"plus(plus(?x:Nat,s(0)),0) = plus(?x:Nat,s(0))",
"plus(plus(?y:Nat,s(0)),?y:Nat) = plus(?y:Nat,s(?y:Nat))",
"plus(plus(0,?x:Nat),s(?y:Nat)) = s(plus(?y:Nat,?x:Nat))",
"plus(plus(s(0),?x:Nat),?y:Nat) = plus(?y:Nat,s(?x:Nat))",
"plus(plus(s(0),?y:Nat),?z:Nat) = plus(?y:Nat,s(?z:Nat))",
"plus(?z:Nat,plus(plus(0,?y:Nat),0)) = plus(?y:Nat,?z:Nat)",
"plus(0,plus(0,plus(0,?z:Nat))) = plus(0,?z:Nat)",
"plus(0,plus(plus(?x:Nat,0),?x:Nat)) = plus(?x:Nat,?x:Nat)",
"plus(plus(?y:Nat,0),plus(0,?x:Nat)) = plus(?y:Nat,?x:Nat)",
"plus(plus(?y:Nat,plus(0,?x:Nat)),0) = plus(?y:Nat,?x:Nat)",
"plus(plus(0,plus(?y:Nat,0)),?y:Nat) = plus(?y:Nat,?y:Nat)",
"plus(plus(s(0),plus(?x:Nat,0)),0) = s(?x:Nat)",
"plus(?x:Nat,s(?x:Nat)) = plus(plus(0,s(0)),plus(?x:Nat,?x:Nat))",
"plus(?y:Nat,s(?x:Nat)) = s(plus(plus(?y:Nat,0),plus(?x:Nat,0)))",
"plus(?x:Nat,plus(?x:Nat,?y:Nat)) = plus(0,plus(plus(?x:Nat,?x:Nat),?y:Nat))",
"plus(s(?x:Nat),s(?z:Nat)) = s(plus(plus(?z:Nat,?x:Nat),s(0)))",
"plus(s(?y:Nat),s(?x:Nat)) = plus(plus(s(?x:Nat),?y:Nat),s(0))",
"plus(0,plus(?x:Nat,s(?y:Nat))) = plus(plus(?x:Nat,s(0)),?y:Nat)",
"plus(s(?x:Nat),plus(0,?z:Nat)) = plus(?x:Nat,s(plus(0,?z:Nat)))",
"plus(s(plus(?x:Nat,0)),0) = s(plus(plus(?x:Nat,0),0))",
"plus(s(plus(?y:Nat,0)),?y:Nat) = plus(s(?y:Nat),plus(?y:Nat,0))",
"plus(plus(0,?z:Nat),s(?x:Nat)) = plus(plus(0,s(?x:Nat)),?z:Nat)",
"plus(plus(s(0),?x:Nat),?z:Nat) = s(plus(?z:Nat,plus(?x:Nat,0)))",
"plus(?x:Nat,plus(plus(?z:Nat,?y:Nat),0)) = plus(?x:Nat,plus(?z:Nat,?y:Nat))",
"plus(?x:Nat,plus(plus(0,?x:Nat),0)) = plus(plus(0,?x:Nat),?x:Nat)",
"plus(plus(s(?y:Nat),?z:Nat),s(0)) = s(plus(s(?z:Nat),?y:Nat))",
"plus(plus(plus(?x:Nat,0),s(?y:Nat)),0) = plus(?y:Nat,s(?x:Nat))",
"s(plus(?y:Nat,?x:Nat)) = plus(s(plus(?y:Nat,0)),plus(plus(?x:Nat,0),0))",
"plus(s(?x:Nat),plus(?z:Nat,0)) = plus(plus(s(0),plus(?z:Nat,0)),?x:Nat)",
"plus(s(?y:Nat),plus(?y:Nat,?x:Nat)) = s(plus(plus(?y:Nat,0),plus(?x:Nat,?y:Nat)))",
"plus(plus(s(?x:Nat),?y:Nat),0) = plus(plus(?y:Nat,0),plus(?x:Nat,s(0)))",
"plus(?y:Nat,plus(?x:Nat,plus(?x:Nat,0))) = plus(?x:Nat,plus(?y:Nat,plus(0,?x:Nat)))",
"plus(0,plus(plus(?x:Nat,?x:Nat),s(?y:Nat))) = s(plus(?x:Nat,plus(?y:Nat,?x:Nat)))",
"plus(s(0),plus(s(0),plus(?x:Nat,0))) = s(plus(?x:Nat,s(0)))",
"plus(s(plus(?y:Nat,?x:Nat)),?y:Nat) = plus(s(?x:Nat),plus(plus(0,0),plus(?y:Nat,?y:Nat)))",
"plus(?x:Nat,plus(0,plus(?x:Nat,?z:Nat))) = plus(plus(plus(0,?x:Nat),plus(?z:Nat,0)),?x:Nat)",
"plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),?z:Nat) = plus(plus(?z:Nat,plus(?x:Nat,?x:Nat)),plus(?x:Nat,0))"
]

val problemsPlusWithoutZeroYes =
IOMstrs.rdEqs [
"plus(?x:Nat,?z:Nat) = plus(?z:Nat,?x:Nat)",
"s(plus(?x:Nat,?y:Nat)) = plus(s(?x:Nat),?y:Nat)",
"s(plus(?x:Nat,?z:Nat)) = s(plus(?z:Nat,?x:Nat))",
"s(plus(?y:Nat,?x:Nat)) = plus(s(?x:Nat),?y:Nat)",
"plus(?x:Nat,s(?x:Nat)) = s(plus(?x:Nat,?x:Nat))",
"plus(?x:Nat,s(?y:Nat)) = plus(s(?x:Nat),?y:Nat)",
"plus(?x:Nat,s(?z:Nat)) = s(plus(?z:Nat,?x:Nat))",
"plus(?x:Nat,s(?z:Nat)) = plus(?z:Nat,s(?x:Nat))",
"plus(?y:Nat,s(?y:Nat)) = plus(s(?y:Nat),?y:Nat)",
"plus(?z:Nat,s(?x:Nat)) = s(plus(?z:Nat,?x:Nat))",
"plus(s(?x:Nat),?y:Nat) = plus(?y:Nat,s(?x:Nat))",
"plus(s(?x:Nat),?y:Nat) = plus(s(?y:Nat),?x:Nat)",
"plus(s(?y:Nat),?y:Nat) = s(plus(?y:Nat,?y:Nat))",
"s(s(plus(?x:Nat,?x:Nat))) = s(plus(s(?x:Nat),?x:Nat))",
"s(s(plus(?x:Nat,?z:Nat))) = s(s(plus(?z:Nat,?x:Nat)))",
"s(s(plus(?y:Nat,?x:Nat))) = s(plus(s(?y:Nat),?x:Nat))",
"s(s(plus(?y:Nat,?x:Nat))) = plus(s(s(?x:Nat)),?y:Nat)",
"s(s(plus(?y:Nat,?y:Nat))) = plus(?y:Nat,s(s(?y:Nat)))",
"s(plus(?x:Nat,s(?z:Nat))) = plus(?x:Nat,s(s(?z:Nat)))",
"s(plus(?y:Nat,s(?x:Nat))) = s(s(plus(?y:Nat,?x:Nat)))",
"s(plus(?y:Nat,s(?x:Nat))) = plus(?x:Nat,s(s(?y:Nat)))",
"s(plus(?y:Nat,s(?x:Nat))) = plus(s(s(?y:Nat)),?x:Nat)",
"s(plus(?y:Nat,s(?y:Nat))) = s(s(plus(?y:Nat,?y:Nat)))",
"s(plus(s(?x:Nat),?x:Nat)) = plus(?x:Nat,s(s(?x:Nat)))",
"s(plus(s(?x:Nat),?y:Nat)) = plus(s(?x:Nat),s(?y:Nat))",
"s(plus(s(?y:Nat),?z:Nat)) = s(s(plus(?z:Nat,?y:Nat)))",
"s(plus(s(?y:Nat),?z:Nat)) = s(plus(?z:Nat,s(?y:Nat)))",
"s(plus(s(?z:Nat),?y:Nat)) = s(plus(?z:Nat,s(?y:Nat)))",
"plus(?y:Nat,s(s(?x:Nat))) = s(s(plus(?x:Nat,?y:Nat)))",
"plus(?y:Nat,s(s(?x:Nat))) = plus(?x:Nat,s(s(?y:Nat)))",
"plus(?y:Nat,plus(?x:Nat,?z:Nat)) = plus(?z:Nat,plus(?x:Nat,?y:Nat))",
"plus(?y:Nat,plus(?z:Nat,?z:Nat)) = plus(?z:Nat,plus(?y:Nat,?z:Nat))",
"plus(?z:Nat,s(s(?x:Nat))) = s(s(plus(?z:Nat,?x:Nat)))",
"plus(?z:Nat,plus(?x:Nat,?y:Nat)) = plus(plus(?y:Nat,?x:Nat),?z:Nat)",
"plus(s(?y:Nat),s(?z:Nat)) = plus(s(?z:Nat),s(?y:Nat))",
"plus(s(?z:Nat),s(?x:Nat)) = plus(?z:Nat,s(s(?x:Nat)))",
"plus(s(s(?x:Nat)),?z:Nat) = s(plus(?z:Nat,s(?x:Nat)))",
"plus(s(s(?y:Nat)),?z:Nat) = s(plus(s(?z:Nat),?y:Nat))",
"plus(plus(?x:Nat,?z:Nat),?x:Nat) = plus(?x:Nat,plus(?x:Nat,?z:Nat))",
"plus(plus(?y:Nat,?x:Nat),?x:Nat) = plus(?y:Nat,plus(?x:Nat,?x:Nat))",
"plus(plus(?y:Nat,?z:Nat),?x:Nat) = plus(plus(?z:Nat,?x:Nat),?y:Nat)",
"s(plus(?x:Nat,plus(?y:Nat,?z:Nat))) = plus(?y:Nat,plus(s(?z:Nat),?x:Nat))",
"s(plus(?y:Nat,plus(?x:Nat,?z:Nat))) = plus(s(plus(?x:Nat,?y:Nat)),?z:Nat)",
"s(plus(?y:Nat,plus(?z:Nat,?x:Nat))) = plus(?z:Nat,plus(s(?y:Nat),?x:Nat))",
"s(plus(?z:Nat,plus(?y:Nat,?z:Nat))) = plus(?y:Nat,plus(s(?z:Nat),?z:Nat))",
"s(plus(plus(?x:Nat,?x:Nat),?y:Nat)) = s(plus(?x:Nat,plus(?x:Nat,?y:Nat)))",
"s(plus(plus(?x:Nat,?y:Nat),?x:Nat)) = plus(?x:Nat,s(plus(?y:Nat,?x:Nat)))",
"s(plus(plus(?x:Nat,?y:Nat),?z:Nat)) = plus(s(plus(?y:Nat,?z:Nat)),?x:Nat)",
"s(plus(plus(?x:Nat,?z:Nat),?x:Nat)) = plus(s(?x:Nat),plus(?z:Nat,?x:Nat))",
"s(plus(plus(?x:Nat,?z:Nat),?y:Nat)) = plus(?z:Nat,plus(?x:Nat,s(?y:Nat)))",
"s(plus(plus(?y:Nat,?x:Nat),?x:Nat)) = plus(s(plus(?x:Nat,?y:Nat)),?x:Nat)",
"s(plus(plus(?y:Nat,?z:Nat),?z:Nat)) = s(plus(?z:Nat,plus(?y:Nat,?z:Nat)))",
"s(plus(plus(?z:Nat,?y:Nat),?x:Nat)) = plus(?z:Nat,plus(s(?x:Nat),?y:Nat))",
"plus(?x:Nat,s(plus(?x:Nat,?y:Nat))) = s(plus(?y:Nat,plus(?x:Nat,?x:Nat)))",
"plus(?x:Nat,s(plus(?x:Nat,?y:Nat))) = plus(s(?x:Nat),plus(?y:Nat,?x:Nat))",
"plus(?x:Nat,plus(?x:Nat,s(?z:Nat))) = plus(?x:Nat,s(plus(?z:Nat,?x:Nat)))",
"plus(?x:Nat,plus(s(?y:Nat),?x:Nat)) = s(plus(?y:Nat,plus(?x:Nat,?x:Nat)))",
"plus(?x:Nat,plus(s(?y:Nat),?y:Nat)) = plus(plus(?y:Nat,?y:Nat),s(?x:Nat))",
"plus(?y:Nat,plus(s(?y:Nat),?x:Nat)) = plus(s(?y:Nat),plus(?y:Nat,?x:Nat))",
"plus(?y:Nat,plus(s(?z:Nat),?y:Nat)) = plus(?z:Nat,plus(s(?y:Nat),?y:Nat))",
"plus(?z:Nat,s(plus(?x:Nat,?z:Nat))) = plus(s(plus(?x:Nat,?z:Nat)),?z:Nat)",
"plus(?z:Nat,plus(s(?y:Nat),?y:Nat)) = plus(s(plus(?z:Nat,?y:Nat)),?y:Nat)",
"plus(s(?x:Nat),plus(?x:Nat,?x:Nat)) = plus(?x:Nat,s(plus(?x:Nat,?x:Nat)))",
"plus(s(?z:Nat),plus(?y:Nat,?x:Nat)) = plus(s(plus(?y:Nat,?z:Nat)),?x:Nat)",
"plus(s(s(?y:Nat)),s(?y:Nat)) = plus(s(?y:Nat),s(s(?y:Nat)))",
"plus(s(plus(?x:Nat,?y:Nat)),?y:Nat) = plus(plus(s(?x:Nat),?y:Nat),?y:Nat)",
"plus(s(plus(?y:Nat,?x:Nat)),?x:Nat) = plus(s(plus(?x:Nat,?y:Nat)),?x:Nat)",
"plus(s(plus(?y:Nat,?x:Nat)),?z:Nat) = plus(?z:Nat,plus(s(?y:Nat),?x:Nat))",
"plus(s(plus(?y:Nat,?y:Nat)),?x:Nat) = plus(s(?x:Nat),plus(?y:Nat,?y:Nat))",
"plus(s(plus(?z:Nat,?x:Nat)),?y:Nat) = plus(?z:Nat,plus(?x:Nat,s(?y:Nat)))",
"plus(s(plus(?z:Nat,?z:Nat)),?x:Nat) = plus(?z:Nat,plus(?x:Nat,s(?z:Nat)))",
"plus(plus(?x:Nat,?x:Nat),s(?y:Nat)) = plus(?y:Nat,s(plus(?x:Nat,?x:Nat)))",
"plus(plus(?x:Nat,?y:Nat),s(?y:Nat)) = plus(plus(s(?y:Nat),?x:Nat),?y:Nat)",
"plus(plus(?x:Nat,?y:Nat),s(?z:Nat)) = plus(s(plus(?x:Nat,?y:Nat)),?z:Nat)",
"plus(plus(?x:Nat,s(?y:Nat)),?x:Nat) = plus(s(plus(?x:Nat,?x:Nat)),?y:Nat)",
"plus(plus(?y:Nat,s(?x:Nat)),?x:Nat) = plus(?x:Nat,plus(?x:Nat,s(?y:Nat)))",
"plus(plus(?y:Nat,s(?x:Nat)),?y:Nat) = plus(?x:Nat,plus(s(?y:Nat),?y:Nat))",
"plus(plus(s(?x:Nat),?x:Nat),?y:Nat) = plus(?y:Nat,plus(s(?x:Nat),?x:Nat))",
"plus(plus(s(?x:Nat),?y:Nat),?y:Nat) = plus(?y:Nat,plus(?x:Nat,s(?y:Nat)))",
"plus(plus(s(?x:Nat),?y:Nat),?z:Nat) = s(plus(plus(?x:Nat,?y:Nat),?z:Nat))",
"plus(plus(s(?y:Nat),?z:Nat),?y:Nat) = plus(s(?y:Nat),plus(?y:Nat,?z:Nat))",
"s(plus(plus(?x:Nat,?x:Nat),s(?x:Nat))) = plus(plus(s(?x:Nat),s(?x:Nat)),?x:Nat)",
"s(plus(plus(?x:Nat,?x:Nat),s(?z:Nat))) = s(plus(s(?z:Nat),plus(?x:Nat,?x:Nat)))",
"s(plus(plus(?x:Nat,?y:Nat),s(?z:Nat))) = s(plus(s(?x:Nat),plus(?y:Nat,?z:Nat)))",
"plus(?x:Nat,plus(?x:Nat,plus(?y:Nat,?z:Nat))) = plus(?y:Nat,plus(plus(?z:Nat,?x:Nat),?x:Nat))",
"plus(?x:Nat,plus(s(?x:Nat),s(?x:Nat))) = plus(s(plus(?x:Nat,?x:Nat)),s(?x:Nat))",
"plus(?x:Nat,plus(plus(?z:Nat,?x:Nat),?z:Nat)) = plus(plus(?z:Nat,plus(?x:Nat,?x:Nat)),?z:Nat)",
"plus(s(?x:Nat),plus(s(?x:Nat),?y:Nat)) = s(plus(plus(?x:Nat,?y:Nat),s(?x:Nat)))",
"plus(s(?z:Nat),s(plus(?x:Nat,?x:Nat))) = plus(?z:Nat,plus(s(?x:Nat),s(?x:Nat)))",
"plus(plus(?x:Nat,plus(?y:Nat,?x:Nat)),?y:Nat) = plus(?x:Nat,plus(?x:Nat,plus(?y:Nat,?y:Nat)))",
"plus(plus(?x:Nat,plus(?y:Nat,?x:Nat)),?z:Nat) = plus(?x:Nat,plus(?x:Nat,plus(?z:Nat,?y:Nat)))",
"plus(plus(plus(?y:Nat,?x:Nat),?x:Nat),?z:Nat) = plus(plus(plus(?x:Nat,?y:Nat),?z:Nat),?x:Nat)",
"s(plus(plus(?x:Nat,?x:Nat),plus(?y:Nat,?x:Nat))) = plus(plus(?x:Nat,?x:Nat),s(plus(?y:Nat,?x:Nat)))",
"s(plus(plus(?z:Nat,?z:Nat),plus(?z:Nat,?x:Nat))) = plus(plus(plus(?z:Nat,?z:Nat),s(?x:Nat)),?z:Nat)",
"plus(?y:Nat,plus(s(?z:Nat),plus(?x:Nat,?x:Nat))) = plus(plus(?z:Nat,plus(?x:Nat,?x:Nat)),s(?y:Nat))",
"plus(s(?x:Nat),plus(plus(?y:Nat,?x:Nat),?z:Nat)) = plus(s(?y:Nat),plus(plus(?x:Nat,?x:Nat),?z:Nat))",
"plus(plus(?x:Nat,?x:Nat),plus(?y:Nat,s(?y:Nat))) = plus(plus(s(?x:Nat),plus(?x:Nat,?y:Nat)),?y:Nat)",
"plus(plus(?x:Nat,?y:Nat),s(plus(?x:Nat,?x:Nat))) = plus(plus(s(?y:Nat),plus(?x:Nat,?x:Nat)),?x:Nat)",
"plus(plus(s(?y:Nat),?x:Nat),plus(?y:Nat,?x:Nat)) = plus(plus(?x:Nat,plus(?y:Nat,?x:Nat)),s(?y:Nat))",
"plus(plus(plus(?y:Nat,?z:Nat),s(?x:Nat)),?x:Nat) = plus(s(plus(?y:Nat,?x:Nat)),plus(?x:Nat,?z:Nat))",
"plus(plus(plus(?y:Nat,?x:Nat),?z:Nat),plus(?x:Nat,?y:Nat)) = plus(plus(plus(?x:Nat,?y:Nat),plus(?z:Nat,?y:Nat)),?x:Nat)"
]

val problemsPlusWithoutZeroNo =
IOMstrs.rdEqs [
"s(?x:Nat) = plus(?x:Nat,?x:Nat)",
"s(?x:Nat) = plus(?x:Nat,?y:Nat)",
"s(?x:Nat) = plus(?y:Nat,?x:Nat)",
"s(?x:Nat) = plus(?y:Nat,?y:Nat)",
"s(?x:Nat) = plus(?y:Nat,?z:Nat)",
"s(?x:Nat) = plus(?x:Nat,s(?x:Nat))",
"s(?x:Nat) = plus(?x:Nat,s(?y:Nat))",
"s(?x:Nat) = plus(?y:Nat,s(?y:Nat))",
"s(?x:Nat) = plus(?y:Nat,s(?z:Nat))",
"s(?x:Nat) = plus(?z:Nat,s(?x:Nat))",
"s(?x:Nat) = plus(s(?x:Nat),?x:Nat)",
"s(?x:Nat) = plus(s(?x:Nat),?y:Nat)",
"s(?x:Nat) = plus(s(?y:Nat),?x:Nat)",
"s(?x:Nat) = plus(s(?y:Nat),?y:Nat)",
"s(?z:Nat) = plus(s(?x:Nat),?y:Nat)",
"s(s(?x:Nat)) = plus(?x:Nat,?x:Nat)",
"s(s(?x:Nat)) = plus(?z:Nat,?x:Nat)",
"s(s(?y:Nat)) = plus(?y:Nat,?x:Nat)",
"s(s(?y:Nat)) = plus(?z:Nat,?z:Nat)",
"s(s(?z:Nat)) = plus(?y:Nat,?x:Nat)",
"plus(?x:Nat,?x:Nat) = plus(?x:Nat,?y:Nat)",
"plus(?x:Nat,?x:Nat) = plus(?y:Nat,?x:Nat)",
"plus(?x:Nat,?x:Nat) = plus(?y:Nat,?y:Nat)",
"plus(?x:Nat,?y:Nat) = plus(?y:Nat,?z:Nat)",
"plus(?y:Nat,?x:Nat) = plus(?y:Nat,?z:Nat)",
"plus(?y:Nat,?x:Nat) = plus(?z:Nat,?z:Nat)",
"plus(?y:Nat,?z:Nat) = plus(?x:Nat,?z:Nat)",
"s(?x:Nat) = plus(?x:Nat,plus(?z:Nat,?y:Nat))",
"s(?x:Nat) = plus(?y:Nat,s(s(?x:Nat)))",
"s(?x:Nat) = plus(?y:Nat,plus(?x:Nat,?y:Nat))",
"s(?x:Nat) = plus(?z:Nat,s(s(?z:Nat)))",
"s(?x:Nat) = plus(?z:Nat,plus(?x:Nat,?y:Nat))",
"s(?x:Nat) = plus(?z:Nat,plus(?y:Nat,?x:Nat))",
"s(?x:Nat) = plus(s(?x:Nat),s(?y:Nat))",
"s(?x:Nat) = plus(s(s(?x:Nat)),?x:Nat)",
"s(?x:Nat) = plus(plus(?y:Nat,?y:Nat),?z:Nat)",
"s(?x:Nat) = plus(plus(?z:Nat,?x:Nat),?x:Nat)",
"s(?y:Nat) = plus(?y:Nat,plus(?y:Nat,?z:Nat))",
"s(?y:Nat) = plus(?z:Nat,plus(?y:Nat,?y:Nat))",
"s(?y:Nat) = plus(plus(?x:Nat,?y:Nat),?x:Nat)",
"s(?y:Nat) = plus(plus(?x:Nat,?y:Nat),?z:Nat)",
"s(?y:Nat) = plus(plus(?x:Nat,?z:Nat),?y:Nat)",
"s(?z:Nat) = plus(?z:Nat,plus(?x:Nat,?x:Nat))",
"s(?z:Nat) = plus(plus(?y:Nat,?y:Nat),?z:Nat)",
"s(?z:Nat) = plus(plus(?z:Nat,?x:Nat),?x:Nat)",
"s(s(?x:Nat)) = plus(?x:Nat,s(?x:Nat))",
"s(s(?x:Nat)) = plus(?x:Nat,s(?y:Nat))",
"plus(?x:Nat,?x:Nat) = s(plus(?x:Nat,?z:Nat))",
"plus(?x:Nat,?x:Nat) = plus(?y:Nat,s(?x:Nat))",
"plus(?x:Nat,?x:Nat) = plus(s(?x:Nat),?y:Nat)",
"plus(?x:Nat,?x:Nat) = plus(s(?y:Nat),?x:Nat)",
"plus(?x:Nat,?x:Nat) = plus(s(?y:Nat),?y:Nat)",
"plus(?x:Nat,?y:Nat) = s(plus(?x:Nat,?x:Nat))",
"plus(?x:Nat,?y:Nat) = plus(?x:Nat,s(?y:Nat))",
"plus(?x:Nat,?y:Nat) = plus(s(?x:Nat),?x:Nat)",
"plus(?x:Nat,?y:Nat) = plus(s(?x:Nat),?z:Nat)",
"plus(?x:Nat,?z:Nat) = s(plus(?x:Nat,?y:Nat))",
"plus(?x:Nat,?z:Nat) = plus(s(?y:Nat),?x:Nat)",
"plus(?x:Nat,?z:Nat) = plus(s(?y:Nat),?z:Nat)",
"plus(?y:Nat,?x:Nat) = s(s(s(?y:Nat)))",
"plus(?y:Nat,?x:Nat) = s(plus(?x:Nat,?x:Nat))",
"plus(?y:Nat,?x:Nat) = s(plus(?z:Nat,?x:Nat))",
"plus(?y:Nat,?x:Nat) = plus(?x:Nat,s(?x:Nat))",
"plus(?y:Nat,?x:Nat) = plus(?x:Nat,s(?z:Nat))",
"plus(?y:Nat,?y:Nat) = s(plus(?x:Nat,?x:Nat))",
"plus(?y:Nat,?y:Nat) = s(plus(?x:Nat,?z:Nat))",
"plus(?y:Nat,?y:Nat) = s(plus(?z:Nat,?y:Nat))",
"plus(?y:Nat,?z:Nat) = s(plus(?x:Nat,?y:Nat))",
"plus(?y:Nat,?z:Nat) = s(plus(?z:Nat,?x:Nat))",
"plus(?y:Nat,?z:Nat) = s(plus(?z:Nat,?y:Nat))",
"plus(?y:Nat,?z:Nat) = plus(?x:Nat,s(?x:Nat))",
"plus(?y:Nat,?z:Nat) = plus(?z:Nat,s(?y:Nat))",
"plus(?z:Nat,?x:Nat) = s(s(s(?x:Nat)))",
"plus(?z:Nat,?x:Nat) = plus(?y:Nat,s(?x:Nat))",
"plus(?z:Nat,?x:Nat) = plus(?z:Nat,s(?y:Nat))",
"plus(?z:Nat,?x:Nat) = plus(s(?x:Nat),?x:Nat)",
"s(s(s(?x:Nat))) = plus(?x:Nat,?x:Nat)",
"s(s(s(?x:Nat))) = plus(?y:Nat,?y:Nat)",
"s(plus(?x:Nat,?x:Nat)) = plus(?y:Nat,?z:Nat)",
"s(plus(?y:Nat,?x:Nat)) = plus(?y:Nat,?x:Nat)",
"plus(?x:Nat,s(?x:Nat)) = plus(?x:Nat,?x:Nat)",
"plus(?x:Nat,s(?y:Nat)) = s(s(?z:Nat))",
"plus(?x:Nat,s(?y:Nat)) = plus(?x:Nat,?x:Nat)",
"plus(?x:Nat,s(?y:Nat)) = plus(?z:Nat,?z:Nat)",
"plus(?z:Nat,s(?y:Nat)) = s(s(?y:Nat))",
"plus(s(?x:Nat),?x:Nat) = s(s(?x:Nat))",
"plus(s(?x:Nat),?z:Nat) = plus(?x:Nat,?z:Nat)",
"plus(s(?y:Nat),?x:Nat) = s(s(?x:Nat))",
"plus(s(?y:Nat),?x:Nat) = s(s(?z:Nat))",
"plus(s(?y:Nat),?x:Nat) = plus(?x:Nat,?y:Nat)",
"plus(s(?y:Nat),?z:Nat) = plus(?x:Nat,?x:Nat)",
"plus(?x:Nat,s(s(?z:Nat))) = s(?x:Nat)",
"plus(?x:Nat,plus(?y:Nat,?x:Nat)) = s(?x:Nat)",
"plus(?y:Nat,plus(?y:Nat,?x:Nat)) = s(?x:Nat)",
"plus(?y:Nat,plus(?y:Nat,?y:Nat)) = s(?x:Nat)",
"plus(s(?y:Nat),s(?y:Nat)) = s(?x:Nat)",
"plus(s(?y:Nat),s(?y:Nat)) = s(?y:Nat)",
"plus(s(?y:Nat),s(?z:Nat)) = s(?z:Nat)",
"plus(s(s(?y:Nat)),?z:Nat) = s(?x:Nat)",
"plus(s(s(?z:Nat)),?x:Nat) = s(?z:Nat)",
"plus(plus(?x:Nat,?z:Nat),?z:Nat) = s(?y:Nat)",
"plus(plus(?y:Nat,?y:Nat),?y:Nat) = s(?y:Nat)",
"s(?x:Nat) = plus(?x:Nat,plus(?z:Nat,s(?y:Nat)))",
"s(?x:Nat) = plus(?y:Nat,s(plus(?y:Nat,?x:Nat)))",
"s(?x:Nat) = plus(?y:Nat,s(plus(?y:Nat,?z:Nat)))",
"s(?x:Nat) = plus(s(?x:Nat),plus(?z:Nat,?y:Nat))",
"s(?x:Nat) = plus(s(?y:Nat),plus(?y:Nat,?z:Nat))",
"s(?x:Nat) = plus(s(?z:Nat),plus(?z:Nat,?x:Nat))",
"s(?x:Nat) = plus(s(plus(?y:Nat,?y:Nat)),?x:Nat)",
"s(?x:Nat) = plus(plus(?x:Nat,?x:Nat),s(?z:Nat))",
"s(?x:Nat) = plus(plus(?y:Nat,?x:Nat),s(?x:Nat))",
"s(?x:Nat) = plus(plus(s(?z:Nat),?y:Nat),?y:Nat)",
"s(?y:Nat) = plus(?x:Nat,s(plus(?y:Nat,?x:Nat)))",
"s(?y:Nat) = plus(?x:Nat,s(plus(?y:Nat,?z:Nat)))",
"s(?y:Nat) = plus(?x:Nat,plus(s(?z:Nat),?z:Nat))",
"s(?y:Nat) = plus(s(?y:Nat),s(s(?y:Nat)))",
"s(?y:Nat) = plus(s(s(?x:Nat)),s(?z:Nat))",
"s(?y:Nat) = plus(s(s(s(?z:Nat))),?x:Nat)",
"s(?y:Nat) = plus(s(plus(?x:Nat,?x:Nat)),?z:Nat)",
"s(?y:Nat) = plus(s(plus(?z:Nat,?y:Nat)),?x:Nat)",
"s(?y:Nat) = plus(plus(s(?x:Nat),?x:Nat),?x:Nat)",
"s(?y:Nat) = plus(plus(s(?y:Nat),?y:Nat),?x:Nat)",
"s(?z:Nat) = plus(?x:Nat,plus(?y:Nat,s(?y:Nat)))",
"s(?z:Nat) = plus(s(?x:Nat),plus(?y:Nat,?z:Nat))",
"s(?z:Nat) = plus(s(s(?x:Nat)),s(?x:Nat))",
"s(?z:Nat) = plus(plus(?z:Nat,?x:Nat),s(?x:Nat))",
"s(s(?x:Nat)) = plus(?z:Nat,plus(?y:Nat,?z:Nat))",
"s(s(?x:Nat)) = plus(s(?z:Nat),s(?y:Nat))",
"s(s(?y:Nat)) = plus(s(?x:Nat),s(?x:Nat))",
"s(s(?y:Nat)) = plus(plus(?z:Nat,?z:Nat),?y:Nat)",
"s(s(?z:Nat)) = plus(plus(?y:Nat,?x:Nat),?y:Nat)",
"plus(?x:Nat,?x:Nat) = s(plus(?y:Nat,s(?y:Nat)))",
"plus(?x:Nat,?x:Nat) = plus(?x:Nat,plus(?x:Nat,?x:Nat))",
"plus(?x:Nat,?x:Nat) = plus(?x:Nat,plus(?z:Nat,?z:Nat))",
"plus(?x:Nat,?x:Nat) = plus(?y:Nat,plus(?x:Nat,?y:Nat))",
"plus(?x:Nat,?x:Nat) = plus(s(?x:Nat),s(?x:Nat))",
"plus(?x:Nat,?x:Nat) = plus(s(s(?x:Nat)),?x:Nat)",
"plus(?x:Nat,?x:Nat) = plus(plus(?y:Nat,?x:Nat),?y:Nat)",
"plus(?x:Nat,?x:Nat) = plus(plus(?z:Nat,?z:Nat),?y:Nat)",
"plus(?x:Nat,?y:Nat) = s(s(s(s(?z:Nat))))",
"plus(?x:Nat,?y:Nat) = s(plus(?x:Nat,s(?y:Nat)))",
"plus(?x:Nat,?y:Nat) = plus(?x:Nat,plus(?z:Nat,?z:Nat))",
"plus(?x:Nat,?y:Nat) = plus(?y:Nat,plus(?x:Nat,?x:Nat))",
"plus(?x:Nat,?y:Nat) = plus(plus(?x:Nat,?y:Nat),?x:Nat)",
"plus(?x:Nat,?y:Nat) = plus(plus(?y:Nat,?x:Nat),?x:Nat)",
"plus(?x:Nat,?y:Nat) = plus(plus(?y:Nat,?x:Nat),?y:Nat)",
"plus(?x:Nat,?z:Nat) = s(plus(?y:Nat,s(?y:Nat)))",
"plus(?x:Nat,?z:Nat) = plus(s(?z:Nat),s(?y:Nat))",
"plus(?y:Nat,?x:Nat) = s(plus(s(?y:Nat),?x:Nat))",
"plus(?y:Nat,?x:Nat) = s(plus(s(?y:Nat),?y:Nat))",
"plus(?y:Nat,?x:Nat) = plus(?z:Nat,s(s(?z:Nat)))",
"plus(?y:Nat,?x:Nat) = plus(?z:Nat,plus(?y:Nat,?x:Nat))",
"plus(?y:Nat,?x:Nat) = plus(s(?z:Nat),s(?x:Nat))",
"plus(?y:Nat,?y:Nat) = plus(?x:Nat,s(s(?y:Nat)))",
"plus(?y:Nat,?y:Nat) = plus(s(s(?z:Nat)),?x:Nat)",
"plus(?y:Nat,?y:Nat) = plus(plus(?x:Nat,?y:Nat),?y:Nat)",
"plus(?z:Nat,?x:Nat) = plus(plus(?x:Nat,?y:Nat),?x:Nat)",
"plus(?z:Nat,?y:Nat) = s(s(plus(?z:Nat,?x:Nat)))",
"plus(?z:Nat,?y:Nat) = s(plus(?y:Nat,s(?x:Nat)))",
"plus(?z:Nat,?y:Nat) = plus(?y:Nat,plus(?z:Nat,?y:Nat))",
"plus(?z:Nat,?z:Nat) = plus(?y:Nat,plus(?x:Nat,?x:Nat))",
"plus(?z:Nat,?z:Nat) = plus(s(?z:Nat),s(?y:Nat))",
"s(s(s(?x:Nat))) = plus(?x:Nat,s(?y:Nat))",
"s(s(s(?z:Nat))) = plus(s(?y:Nat),?z:Nat)",
"s(plus(?x:Nat,?x:Nat)) = plus(?z:Nat,s(?y:Nat))",
"s(plus(?x:Nat,?y:Nat)) = plus(?x:Nat,s(?z:Nat))",
"s(plus(?y:Nat,?y:Nat)) = plus(s(?z:Nat),?y:Nat)",
"plus(?x:Nat,s(?x:Nat)) = s(s(s(?y:Nat)))",
"plus(?x:Nat,s(?y:Nat)) = plus(?x:Nat,s(?x:Nat))",
"plus(?x:Nat,s(?y:Nat)) = plus(s(?y:Nat),?y:Nat)",
"plus(?y:Nat,s(?y:Nat)) = s(plus(?z:Nat,?x:Nat))",
"plus(?y:Nat,s(?z:Nat)) = plus(?z:Nat,s(?x:Nat))",
"plus(?z:Nat,s(?z:Nat)) = s(plus(?x:Nat,?x:Nat))",
"plus(?z:Nat,s(?z:Nat)) = s(plus(?x:Nat,?z:Nat))",
"plus(s(?x:Nat),?x:Nat) = s(plus(?z:Nat,?z:Nat))",
"plus(s(?x:Nat),?x:Nat) = plus(?y:Nat,s(?z:Nat))",
"plus(s(?x:Nat),?x:Nat) = plus(?z:Nat,s(?z:Nat))",
"plus(s(?x:Nat),?x:Nat) = plus(s(?x:Nat),?y:Nat)",
"plus(s(?x:Nat),?z:Nat) = plus(s(?x:Nat),?y:Nat)",
"plus(s(?y:Nat),?z:Nat) = s(plus(?x:Nat,?z:Nat))",
"plus(s(?z:Nat),?x:Nat) = s(s(s(?y:Nat)))",
"plus(s(?z:Nat),?x:Nat) = plus(s(?x:Nat),?y:Nat)",
"s(s(plus(?x:Nat,?x:Nat))) = plus(?y:Nat,?x:Nat)",
"s(s(plus(?y:Nat,?x:Nat))) = plus(?z:Nat,?y:Nat)",
"s(s(plus(?z:Nat,?x:Nat))) = plus(?x:Nat,?x:Nat)",
"s(plus(?y:Nat,s(?z:Nat))) = plus(?z:Nat,?x:Nat)",
"s(plus(s(?y:Nat),?z:Nat)) = plus(?x:Nat,?y:Nat)",
"s(plus(s(?z:Nat),?x:Nat)) = plus(?x:Nat,?y:Nat)",
"plus(?x:Nat,s(s(?x:Nat))) = s(s(?z:Nat))",
"plus(?x:Nat,s(s(?x:Nat))) = plus(?z:Nat,?x:Nat)",
"plus(?x:Nat,s(s(?y:Nat))) = plus(?x:Nat,?x:Nat)",
"plus(?x:Nat,s(s(?z:Nat))) = s(s(?x:Nat))",
"plus(?x:Nat,s(s(?z:Nat))) = plus(?y:Nat,?z:Nat)",
"plus(?x:Nat,plus(?x:Nat,?y:Nat)) = plus(?x:Nat,?y:Nat)",
"plus(?x:Nat,plus(?x:Nat,?y:Nat)) = plus(?z:Nat,?x:Nat)",
"plus(?x:Nat,plus(?x:Nat,?z:Nat)) = plus(?y:Nat,?z:Nat)",
"plus(?x:Nat,plus(?y:Nat,?y:Nat)) = plus(?x:Nat,?y:Nat)",
"plus(?y:Nat,plus(?x:Nat,?x:Nat)) = s(s(?y:Nat))",
"plus(?y:Nat,plus(?x:Nat,?z:Nat)) = s(s(?x:Nat))",
"plus(?y:Nat,plus(?y:Nat,?x:Nat)) = plus(?x:Nat,?y:Nat)",
"plus(?y:Nat,plus(?y:Nat,?y:Nat)) = s(s(?z:Nat))",
"plus(?y:Nat,plus(?z:Nat,?x:Nat)) = plus(?y:Nat,?x:Nat)",
"plus(?z:Nat,s(s(?z:Nat))) = plus(?x:Nat,?x:Nat)",
"plus(?z:Nat,plus(?y:Nat,?x:Nat)) = plus(?x:Nat,?y:Nat)",
"plus(?z:Nat,plus(?y:Nat,?x:Nat)) = plus(?z:Nat,?z:Nat)",
"plus(s(s(?z:Nat)),?x:Nat) = plus(?x:Nat,?x:Nat)",
"plus(plus(?x:Nat,?x:Nat),?x:Nat) = s(s(?y:Nat))",
"plus(plus(?x:Nat,?x:Nat),?x:Nat) = plus(?y:Nat,?x:Nat)",
"plus(plus(?x:Nat,?x:Nat),?y:Nat) = plus(?y:Nat,?x:Nat)",
"plus(plus(?x:Nat,?z:Nat),?x:Nat) = plus(?x:Nat,?y:Nat)",
"plus(plus(?y:Nat,?z:Nat),?x:Nat) = s(s(?y:Nat))",
"plus(plus(?y:Nat,?z:Nat),?x:Nat) = plus(?z:Nat,?y:Nat)",
"plus(plus(?z:Nat,?y:Nat),?y:Nat) = s(s(?z:Nat))",
"plus(?x:Nat,plus(s(?x:Nat),?y:Nat)) = s(?z:Nat)",
"plus(?x:Nat,plus(s(?x:Nat),?z:Nat)) = s(?x:Nat)",
"plus(?x:Nat,plus(s(?y:Nat),?x:Nat)) = s(?x:Nat)",
"plus(?y:Nat,s(plus(?x:Nat,?x:Nat))) = s(?z:Nat)",
"plus(?y:Nat,s(plus(?y:Nat,?x:Nat))) = s(?y:Nat)",
"plus(?y:Nat,s(plus(?z:Nat,?z:Nat))) = s(?z:Nat)",
"plus(?y:Nat,plus(?y:Nat,s(?x:Nat))) = s(?z:Nat)",
"plus(?y:Nat,plus(?z:Nat,s(?x:Nat))) = s(?x:Nat)",
"plus(?z:Nat,plus(?z:Nat,s(?x:Nat))) = s(?z:Nat)",
"plus(s(?x:Nat),plus(?x:Nat,?x:Nat)) = s(?y:Nat)",
"plus(s(plus(?x:Nat,?y:Nat)),?x:Nat) = s(?x:Nat)",
"plus(s(plus(?y:Nat,?x:Nat)),?x:Nat) = s(?y:Nat)",
"plus(plus(?x:Nat,s(?x:Nat)),?x:Nat) = s(?z:Nat)",
"plus(plus(?x:Nat,s(?y:Nat)),?x:Nat) = s(?z:Nat)",
"plus(plus(?y:Nat,?x:Nat),s(?y:Nat)) = s(?z:Nat)",
"plus(plus(?y:Nat,s(?z:Nat)),?y:Nat) = s(?y:Nat)",
"plus(plus(?z:Nat,?x:Nat),s(?y:Nat)) = s(?z:Nat)",
"plus(plus(?z:Nat,s(?x:Nat)),?x:Nat) = s(?y:Nat)",
"plus(plus(s(?x:Nat),?z:Nat),?y:Nat) = s(?y:Nat)",
"plus(plus(s(?y:Nat),?x:Nat),?z:Nat) = s(?x:Nat)",
"s(?x:Nat) = plus(?x:Nat,s(plus(?y:Nat,s(?z:Nat))))",
"s(?x:Nat) = plus(?x:Nat,s(plus(s(?x:Nat),?x:Nat)))",
"s(?x:Nat) = plus(?x:Nat,s(plus(s(?z:Nat),?z:Nat)))",
"s(?x:Nat) = plus(s(s(plus(?y:Nat,?z:Nat))),?y:Nat)",
"s(?x:Nat) = plus(s(plus(s(?y:Nat),?y:Nat)),?z:Nat)",
"s(?x:Nat) = plus(plus(?x:Nat,?x:Nat),plus(?z:Nat,?y:Nat))",
"s(?x:Nat) = plus(plus(?y:Nat,?z:Nat),plus(?x:Nat,?z:Nat))",
"s(?x:Nat) = plus(plus(?y:Nat,s(?y:Nat)),s(?y:Nat))",
"s(?x:Nat) = plus(plus(s(?y:Nat),?y:Nat),s(?x:Nat))",
"s(?x:Nat) = plus(plus(s(s(?x:Nat)),?x:Nat),?z:Nat)",
"s(?x:Nat) = plus(plus(plus(?y:Nat,?x:Nat),?x:Nat),?x:Nat)",
"s(?y:Nat) = plus(s(?x:Nat),plus(s(?y:Nat),?y:Nat))",
"s(?y:Nat) = plus(plus(?y:Nat,?y:Nat),plus(?x:Nat,?x:Nat))",
"s(?z:Nat) = plus(plus(plus(?x:Nat,?z:Nat),?x:Nat),?y:Nat)",
"s(s(?x:Nat)) = plus(?x:Nat,plus(s(?y:Nat),?z:Nat))",
"s(s(?x:Nat)) = plus(s(?y:Nat),plus(?x:Nat,?x:Nat))",
"s(s(?x:Nat)) = plus(s(plus(?y:Nat,?y:Nat)),?x:Nat)",
"s(s(?x:Nat)) = plus(plus(?z:Nat,s(?x:Nat)),?y:Nat)",
"s(s(?y:Nat)) = plus(s(plus(?x:Nat,?y:Nat)),?x:Nat)",
"s(s(?z:Nat)) = plus(s(plus(?y:Nat,?y:Nat)),?x:Nat)",
"plus(?x:Nat,?x:Nat) = s(plus(plus(?y:Nat,?x:Nat),?y:Nat))",
"plus(?x:Nat,?x:Nat) = s(plus(plus(?y:Nat,?x:Nat),?z:Nat))",
"plus(?x:Nat,?x:Nat) = plus(s(?z:Nat),s(s(?y:Nat)))",
"plus(?x:Nat,?y:Nat) = s(s(s(s(s(?y:Nat)))))",
"plus(?x:Nat,?y:Nat) = plus(s(?z:Nat),plus(?z:Nat,?y:Nat))",
"plus(?x:Nat,?y:Nat) = plus(s(s(?y:Nat)),s(?y:Nat))",
"plus(?x:Nat,?z:Nat) = plus(?y:Nat,plus(?z:Nat,s(?x:Nat)))",
"plus(?x:Nat,?z:Nat) = plus(s(?x:Nat),plus(?z:Nat,?x:Nat))",
"plus(?x:Nat,?z:Nat) = plus(s(?y:Nat),plus(?y:Nat,?x:Nat))",
"plus(?y:Nat,?x:Nat) = s(s(s(plus(?x:Nat,?x:Nat))))",
"plus(?y:Nat,?x:Nat) = s(s(s(plus(?z:Nat,?y:Nat))))",
"plus(?y:Nat,?y:Nat) = plus(s(plus(?z:Nat,?y:Nat)),?x:Nat)",
"plus(?y:Nat,?z:Nat) = s(plus(?z:Nat,s(s(?x:Nat))))",
"plus(?y:Nat,?z:Nat) = plus(?x:Nat,s(plus(?x:Nat,?x:Nat)))",
"plus(?z:Nat,?x:Nat) = plus(plus(?y:Nat,s(?z:Nat)),?x:Nat)",
"plus(?z:Nat,?y:Nat) = plus(s(plus(?x:Nat,?x:Nat)),?x:Nat)",
"plus(?z:Nat,?z:Nat) = s(plus(s(?y:Nat),s(?x:Nat)))",
"plus(?z:Nat,?z:Nat) = s(plus(plus(?z:Nat,?z:Nat),?x:Nat))",
"s(s(s(?x:Nat))) = plus(?y:Nat,s(s(?y:Nat)))",
"s(plus(?y:Nat,?y:Nat)) = plus(s(s(?x:Nat)),?y:Nat)",
"s(plus(?z:Nat,?z:Nat)) = plus(plus(?y:Nat,?y:Nat),?x:Nat)",
"plus(?x:Nat,s(?x:Nat)) = plus(?x:Nat,plus(?x:Nat,?z:Nat))",
"plus(?x:Nat,s(?x:Nat)) = plus(plus(?x:Nat,?x:Nat),?y:Nat)",
"plus(?x:Nat,s(?y:Nat)) = plus(?x:Nat,plus(?y:Nat,?z:Nat))",
"plus(?x:Nat,s(?y:Nat)) = plus(plus(?x:Nat,?x:Nat),?x:Nat)",
"plus(?y:Nat,s(?y:Nat)) = plus(plus(?x:Nat,?x:Nat),?x:Nat)",
"plus(?z:Nat,s(?x:Nat)) = plus(?y:Nat,plus(?z:Nat,?x:Nat))",
"plus(?z:Nat,s(?x:Nat)) = plus(?z:Nat,s(s(?y:Nat)))",
"plus(s(?x:Nat),?x:Nat) = s(plus(?x:Nat,s(?x:Nat)))",
"plus(s(?x:Nat),?x:Nat) = s(plus(s(?y:Nat),?x:Nat))",
"plus(s(?x:Nat),?y:Nat) = s(s(plus(?x:Nat,?x:Nat)))",
"plus(s(?x:Nat),?z:Nat) = plus(?z:Nat,s(s(?x:Nat)))",
"plus(s(?y:Nat),?x:Nat) = plus(?x:Nat,plus(?x:Nat,?y:Nat))",
"plus(s(?y:Nat),?y:Nat) = plus(plus(?z:Nat,?y:Nat),?y:Nat)",
"s(plus(s(?x:Nat),?x:Nat)) = plus(?z:Nat,s(?x:Nat))",
"plus(?x:Nat,plus(?x:Nat,?z:Nat)) = s(s(s(?z:Nat)))",
"plus(?x:Nat,plus(?x:Nat,?z:Nat)) = plus(s(?x:Nat),?x:Nat)",
"plus(?x:Nat,plus(?y:Nat,?z:Nat)) = plus(s(?x:Nat),?x:Nat)",
"plus(?x:Nat,plus(?y:Nat,?z:Nat)) = plus(s(?y:Nat),?y:Nat)",
"plus(?x:Nat,plus(?z:Nat,?x:Nat)) = plus(s(?z:Nat),?z:Nat)",
"plus(?y:Nat,plus(?x:Nat,?z:Nat)) = s(s(s(?y:Nat)))",
"plus(?y:Nat,plus(?x:Nat,?z:Nat)) = plus(s(?z:Nat),?y:Nat)",
"plus(?y:Nat,plus(?y:Nat,?x:Nat)) = plus(?x:Nat,s(?y:Nat))",
"plus(?z:Nat,s(s(?y:Nat))) = s(plus(?y:Nat,?z:Nat))",
"plus(?z:Nat,plus(?z:Nat,?y:Nat)) = plus(s(?x:Nat),?y:Nat)",
"plus(?z:Nat,plus(?z:Nat,?z:Nat)) = plus(?y:Nat,s(?z:Nat))",
"plus(s(?z:Nat),s(?y:Nat)) = plus(?y:Nat,s(?x:Nat))",
"plus(plus(?x:Nat,?y:Nat),?x:Nat) = s(plus(?x:Nat,?y:Nat))",
"plus(plus(?x:Nat,?y:Nat),?y:Nat) = plus(s(?x:Nat),?z:Nat)",
"plus(plus(?z:Nat,?x:Nat),?x:Nat) = s(plus(?x:Nat,?x:Nat))",
"plus(plus(?z:Nat,?y:Nat),?x:Nat) = s(plus(?z:Nat,?z:Nat))",
"plus(plus(?z:Nat,?y:Nat),?x:Nat) = plus(s(?y:Nat),?y:Nat)",
"s(plus(?x:Nat,plus(?x:Nat,?z:Nat))) = plus(?y:Nat,?z:Nat)",
"s(plus(?y:Nat,plus(?x:Nat,?x:Nat))) = plus(?z:Nat,?z:Nat)",
"s(plus(?y:Nat,plus(?y:Nat,?x:Nat))) = plus(?y:Nat,?z:Nat)",
"s(plus(?y:Nat,plus(?z:Nat,?x:Nat))) = plus(?x:Nat,?y:Nat)",
"s(plus(s(?y:Nat),s(?x:Nat))) = plus(?y:Nat,?x:Nat)",
"s(plus(plus(?x:Nat,?y:Nat),?z:Nat)) = plus(?x:Nat,?z:Nat)",
"plus(?x:Nat,s(s(s(?z:Nat)))) = s(s(?x:Nat))",
"plus(?x:Nat,plus(s(?z:Nat),?z:Nat)) = plus(?y:Nat,?y:Nat)",
"plus(?y:Nat,s(plus(?x:Nat,?z:Nat))) = plus(?y:Nat,?y:Nat)",
"plus(?z:Nat,s(plus(?x:Nat,?z:Nat))) = plus(?x:Nat,?z:Nat)",
"plus(?z:Nat,plus(?y:Nat,s(?x:Nat))) = plus(?y:Nat,?y:Nat)",
"plus(s(?x:Nat),s(s(?y:Nat))) = plus(?y:Nat,?x:Nat)",
"plus(s(?x:Nat),plus(?x:Nat,?y:Nat)) = plus(?x:Nat,?z:Nat)",
"plus(s(?y:Nat),plus(?x:Nat,?y:Nat)) = s(s(?z:Nat))",
"plus(s(?y:Nat),plus(?y:Nat,?z:Nat)) = s(s(?x:Nat))",
"plus(s(s(s(?z:Nat))),?x:Nat) = plus(?z:Nat,?y:Nat)",
"plus(plus(?y:Nat,?z:Nat),s(?x:Nat)) = plus(?y:Nat,?x:Nat)",
"plus(plus(s(?x:Nat),?y:Nat),?x:Nat) = plus(?y:Nat,?y:Nat)",
"plus(plus(s(?z:Nat),?z:Nat),?y:Nat) = s(s(?z:Nat))",
"plus(?x:Nat,plus(plus(?x:Nat,?z:Nat),?z:Nat)) = s(?x:Nat)",
"plus(?y:Nat,s(s(plus(?x:Nat,?y:Nat)))) = s(?x:Nat)",
"plus(?y:Nat,s(plus(s(?x:Nat),?x:Nat))) = s(?x:Nat)",
"plus(?y:Nat,plus(?y:Nat,s(s(?y:Nat)))) = s(?x:Nat)",
"plus(?y:Nat,plus(plus(?z:Nat,?x:Nat),?y:Nat)) = s(?z:Nat)",
"plus(?z:Nat,plus(?y:Nat,plus(?x:Nat,?z:Nat))) = s(?x:Nat)",
"plus(?z:Nat,plus(s(?y:Nat),s(?z:Nat))) = s(?y:Nat)",
"plus(s(s(?z:Nat)),plus(?z:Nat,?x:Nat)) = s(?z:Nat)",
"plus(plus(?x:Nat,?y:Nat),plus(?y:Nat,?x:Nat)) = s(?z:Nat)",
"plus(plus(?x:Nat,?z:Nat),plus(?y:Nat,?x:Nat)) = s(?x:Nat)",
"plus(plus(?x:Nat,plus(?z:Nat,?y:Nat)),?x:Nat) = s(?x:Nat)",
"plus(plus(?y:Nat,?z:Nat),s(s(?z:Nat))) = s(?y:Nat)",
"plus(plus(?z:Nat,plus(?z:Nat,?y:Nat)),?y:Nat) = s(?y:Nat)",
"plus(plus(s(s(?z:Nat)),?x:Nat),?x:Nat) = s(?y:Nat)",
"plus(plus(plus(?x:Nat,?y:Nat),?y:Nat),?x:Nat) = s(?y:Nat)",
"plus(plus(plus(?y:Nat,?y:Nat),?x:Nat),?y:Nat) = s(?z:Nat)",
"s(?x:Nat) = plus(?x:Nat,s(s(plus(?x:Nat,s(?x:Nat)))))",
"s(?x:Nat) = plus(?z:Nat,s(plus(plus(?y:Nat,?x:Nat),?x:Nat)))",
"s(?x:Nat) = plus(?z:Nat,plus(s(plus(?y:Nat,?y:Nat)),?x:Nat))",
"s(?x:Nat) = plus(s(?x:Nat),s(s(plus(?x:Nat,?x:Nat))))",
"s(?x:Nat) = plus(s(s(?x:Nat)),plus(s(?x:Nat),?z:Nat))",
"s(?x:Nat) = plus(plus(s(?y:Nat),plus(?z:Nat,?x:Nat)),?x:Nat)",
"s(?y:Nat) = plus(?x:Nat,plus(s(?x:Nat),plus(?z:Nat,?y:Nat)))",
"s(?y:Nat) = plus(?y:Nat,plus(s(?y:Nat),plus(?x:Nat,?x:Nat)))",
"s(?y:Nat) = plus(s(?x:Nat),plus(plus(?z:Nat,?x:Nat),?x:Nat))",
"s(?y:Nat) = plus(s(plus(s(s(?x:Nat)),?x:Nat)),?y:Nat)",
"s(?y:Nat) = plus(plus(s(plus(?z:Nat,?z:Nat)),?x:Nat),?y:Nat)",
"s(?y:Nat) = plus(plus(plus(?y:Nat,s(?x:Nat)),?y:Nat),?x:Nat)",
"s(?y:Nat) = plus(plus(plus(s(?y:Nat),?z:Nat),?y:Nat),?z:Nat)",
"s(?z:Nat) = plus(?x:Nat,s(plus(?x:Nat,plus(?z:Nat,?x:Nat))))",
"s(?z:Nat) = plus(?z:Nat,s(s(s(plus(?x:Nat,?y:Nat)))))",
"s(?z:Nat) = plus(s(s(s(plus(?x:Nat,?x:Nat)))),?x:Nat)",
"s(?z:Nat) = plus(s(plus(?z:Nat,plus(?x:Nat,?y:Nat))),?x:Nat)",
"s(s(?x:Nat)) = plus(?x:Nat,plus(plus(?y:Nat,?z:Nat),?x:Nat))",
"s(s(?y:Nat)) = plus(s(s(?x:Nat)),s(s(?z:Nat)))",
"plus(?x:Nat,?x:Nat) = plus(plus(s(?y:Nat),?z:Nat),s(?z:Nat))",
"plus(?x:Nat,?y:Nat) = s(s(plus(plus(?x:Nat,?x:Nat),?x:Nat)))",
"plus(?x:Nat,?y:Nat) = plus(plus(plus(?y:Nat,?z:Nat),?y:Nat),?x:Nat)",
"plus(?x:Nat,?z:Nat) = s(plus(s(plus(?x:Nat,?x:Nat)),?y:Nat))",
"plus(?x:Nat,?z:Nat) = plus(?y:Nat,plus(?z:Nat,plus(?x:Nat,?x:Nat)))",
"plus(?y:Nat,?x:Nat) = plus(?x:Nat,s(plus(s(?y:Nat),?x:Nat)))",
"plus(?y:Nat,?y:Nat) = plus(?y:Nat,plus(?x:Nat,plus(?x:Nat,?z:Nat)))",
"plus(?y:Nat,?z:Nat) = plus(?y:Nat,plus(?x:Nat,s(s(?x:Nat))))",
"plus(?z:Nat,?x:Nat) = s(s(plus(?x:Nat,plus(?x:Nat,?y:Nat))))",
"plus(?z:Nat,?x:Nat) = s(s(plus(plus(?y:Nat,?y:Nat),?z:Nat)))",
"plus(?z:Nat,?x:Nat) = plus(s(?x:Nat),s(plus(?z:Nat,?y:Nat)))",
"plus(?z:Nat,?y:Nat) = plus(plus(?z:Nat,plus(?y:Nat,?x:Nat)),?x:Nat)",
"s(plus(?x:Nat,?x:Nat)) = plus(?y:Nat,plus(s(?x:Nat),?y:Nat))",
"s(plus(?x:Nat,?x:Nat)) = plus(plus(s(?y:Nat),?x:Nat),?z:Nat)",
"s(plus(?y:Nat,?y:Nat)) = plus(?x:Nat,s(plus(?x:Nat,?x:Nat)))",
"s(plus(?z:Nat,?y:Nat)) = plus(s(?y:Nat),plus(?x:Nat,?x:Nat))",
"s(plus(?z:Nat,?z:Nat)) = plus(plus(s(?z:Nat),?z:Nat),?x:Nat)",
"plus(?x:Nat,s(?z:Nat)) = plus(plus(?x:Nat,s(?y:Nat)),?z:Nat)",
"plus(?y:Nat,s(?x:Nat)) = plus(s(s(s(?x:Nat))),?x:Nat)",
"plus(?y:Nat,s(?y:Nat)) = plus(plus(?x:Nat,?y:Nat),s(?x:Nat))",
"plus(s(?x:Nat),?x:Nat) = s(plus(plus(?z:Nat,?x:Nat),?y:Nat))",
"plus(s(?x:Nat),?y:Nat) = s(plus(s(s(?y:Nat)),?y:Nat))",
"plus(s(?y:Nat),?x:Nat) = plus(?x:Nat,s(plus(?x:Nat,?x:Nat)))",
"plus(s(?z:Nat),?x:Nat) = plus(plus(?z:Nat,s(?z:Nat)),?z:Nat)",
"plus(s(?z:Nat),?y:Nat) = s(s(plus(?z:Nat,s(?z:Nat))))",
"s(s(plus(?y:Nat,?z:Nat))) = plus(?x:Nat,plus(?z:Nat,?x:Nat))",
"s(plus(s(?y:Nat),?x:Nat)) = plus(s(s(?x:Nat)),?z:Nat)",
"s(plus(s(?y:Nat),?y:Nat)) = plus(plus(?z:Nat,?x:Nat),?y:Nat)",
"s(plus(s(?z:Nat),?z:Nat)) = plus(plus(?y:Nat,?y:Nat),?z:Nat)",
"plus(?x:Nat,plus(?x:Nat,?z:Nat)) = plus(?z:Nat,s(s(?y:Nat)))",
"plus(?y:Nat,plus(?x:Nat,?y:Nat)) = plus(s(s(?x:Nat)),?x:Nat)",
"plus(?y:Nat,plus(?y:Nat,?z:Nat)) = plus(?z:Nat,plus(?y:Nat,?x:Nat))",
"plus(?y:Nat,plus(?z:Nat,?x:Nat)) = plus(?x:Nat,s(s(?z:Nat)))",
"plus(?z:Nat,s(s(?y:Nat))) = plus(?y:Nat,plus(?x:Nat,?y:Nat))",
"plus(?z:Nat,s(s(?z:Nat))) = s(s(s(s(?x:Nat))))",
"plus(?z:Nat,plus(?x:Nat,?x:Nat)) = plus(s(s(?x:Nat)),?y:Nat)",
"plus(?z:Nat,plus(?y:Nat,?x:Nat)) = plus(s(s(?z:Nat)),?y:Nat)",
"plus(s(?y:Nat),s(?z:Nat)) = plus(?y:Nat,plus(?z:Nat,?x:Nat))",
"plus(s(s(?x:Nat)),?x:Nat) = plus(?z:Nat,plus(?x:Nat,?y:Nat))",
"plus(s(s(?y:Nat)),?x:Nat) = s(plus(?x:Nat,s(?x:Nat)))",
"plus(s(s(?y:Nat)),?y:Nat) = s(plus(?y:Nat,s(?x:Nat)))",
"plus(plus(?x:Nat,?x:Nat),?x:Nat) = plus(?y:Nat,plus(?x:Nat,?x:Nat))",
"plus(plus(?x:Nat,?y:Nat),?x:Nat) = plus(?x:Nat,plus(?x:Nat,?z:Nat))",
"plus(plus(?x:Nat,?y:Nat),?x:Nat) = plus(plus(?z:Nat,?y:Nat),?x:Nat)",
"plus(plus(?y:Nat,?x:Nat),?x:Nat) = plus(plus(?x:Nat,?y:Nat),?z:Nat)",
"plus(plus(?y:Nat,?z:Nat),?y:Nat) = plus(plus(?x:Nat,?x:Nat),?x:Nat)",
"plus(plus(?y:Nat,?z:Nat),?z:Nat) = plus(?y:Nat,plus(?z:Nat,?y:Nat))",
"plus(plus(?z:Nat,?x:Nat),?y:Nat) = plus(plus(?x:Nat,?y:Nat),?x:Nat)",
"plus(plus(?z:Nat,?z:Nat),?y:Nat) = plus(plus(?y:Nat,?x:Nat),?x:Nat)",
"s(plus(?z:Nat,plus(?x:Nat,?z:Nat))) = plus(?z:Nat,s(?y:Nat))",
"s(plus(plus(?z:Nat,?x:Nat),?x:Nat)) = plus(s(?y:Nat),?x:Nat)",
"plus(?x:Nat,s(plus(?x:Nat,?z:Nat))) = s(plus(?y:Nat,?x:Nat))",
"plus(?x:Nat,s(plus(?y:Nat,?y:Nat))) = plus(?z:Nat,s(?y:Nat))",
"plus(?x:Nat,s(plus(?z:Nat,?y:Nat))) = s(s(s(?z:Nat)))",
"plus(?x:Nat,plus(?x:Nat,s(?z:Nat))) = plus(?x:Nat,s(?x:Nat))",
"plus(?x:Nat,plus(?x:Nat,s(?z:Nat))) = plus(s(?x:Nat),?z:Nat)",
"plus(?x:Nat,plus(?y:Nat,s(?x:Nat))) = plus(?y:Nat,s(?x:Nat))",
"plus(?y:Nat,plus(s(?z:Nat),?x:Nat)) = plus(?z:Nat,s(?x:Nat))",
"plus(s(?x:Nat),s(s(?z:Nat))) = s(plus(?y:Nat,?x:Nat))",
"plus(s(?y:Nat),plus(?y:Nat,?y:Nat)) = s(plus(?z:Nat,?z:Nat))",
"plus(s(s(?z:Nat)),s(?x:Nat)) = plus(s(?y:Nat),?x:Nat)",
"plus(plus(?x:Nat,?x:Nat),s(?y:Nat)) = s(s(s(?y:Nat)))",
"plus(plus(s(?x:Nat),?z:Nat),?y:Nat) = plus(s(?x:Nat),?y:Nat)",
"s(plus(plus(s(?y:Nat),?x:Nat),?x:Nat)) = plus(?z:Nat,?x:Nat)",
"plus(?x:Nat,plus(?y:Nat,plus(?x:Nat,?y:Nat))) = s(s(?x:Nat))",
"plus(?x:Nat,plus(plus(?x:Nat,?y:Nat),?x:Nat)) = s(s(?z:Nat))",
"plus(?y:Nat,plus(?x:Nat,plus(?x:Nat,?z:Nat))) = plus(?x:Nat,?z:Nat)",
"plus(?y:Nat,plus(plus(?y:Nat,?x:Nat),?z:Nat)) = plus(?x:Nat,?x:Nat)",
"plus(s(?y:Nat),s(plus(?y:Nat,?y:Nat))) = plus(?x:Nat,?x:Nat)",
"plus(s(?y:Nat),s(plus(?z:Nat,?y:Nat))) = plus(?x:Nat,?z:Nat)",
"plus(s(?z:Nat),plus(s(?x:Nat),?y:Nat)) = plus(?x:Nat,?z:Nat)",
"plus(s(s(plus(?y:Nat,?x:Nat))),?y:Nat) = plus(?y:Nat,?z:Nat)",
"plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),?z:Nat) = plus(?y:Nat,?z:Nat)",
"plus(plus(?x:Nat,plus(?y:Nat,?y:Nat)),?x:Nat) = s(s(?z:Nat))",
"plus(plus(?y:Nat,s(s(?x:Nat))),?x:Nat) = plus(?x:Nat,?z:Nat)",
"plus(plus(?y:Nat,plus(?y:Nat,?x:Nat)),?z:Nat) = plus(?x:Nat,?z:Nat)",
"plus(plus(s(s(?x:Nat)),?z:Nat),?x:Nat) = plus(?x:Nat,?y:Nat)",
"plus(plus(plus(?z:Nat,?x:Nat),?y:Nat),?x:Nat) = plus(?z:Nat,?y:Nat)",
"plus(?x:Nat,plus(?x:Nat,plus(s(?x:Nat),?x:Nat))) = s(?x:Nat)",
"plus(?x:Nat,plus(?y:Nat,plus(?x:Nat,s(?x:Nat)))) = s(?x:Nat)",
"plus(?x:Nat,plus(plus(s(?y:Nat),?x:Nat),?x:Nat)) = s(?z:Nat)",
"plus(?y:Nat,s(plus(?y:Nat,plus(?x:Nat,?z:Nat)))) = s(?x:Nat)",
"plus(?z:Nat,s(s(s(plus(?z:Nat,?x:Nat))))) = s(?y:Nat)",
"plus(s(?x:Nat),plus(?x:Nat,s(s(?y:Nat)))) = s(?x:Nat)",
"plus(s(s(plus(?y:Nat,s(?y:Nat)))),?y:Nat) = s(?z:Nat)",
"plus(s(s(plus(s(?z:Nat),?y:Nat))),?x:Nat) = s(?y:Nat)",
"plus(s(plus(?y:Nat,plus(?x:Nat,?y:Nat))),?z:Nat) = s(?x:Nat)",
"plus(s(plus(?z:Nat,s(s(?z:Nat)))),?z:Nat) = s(?x:Nat)",
"plus(plus(?x:Nat,s(plus(?x:Nat,?z:Nat))),?x:Nat) = s(?z:Nat)",
"plus(plus(?x:Nat,plus(?x:Nat,s(?y:Nat))),?x:Nat) = s(?z:Nat)",
"plus(plus(?x:Nat,plus(?y:Nat,?z:Nat)),s(?y:Nat)) = s(?x:Nat)",
"plus(plus(?y:Nat,?x:Nat),s(plus(?z:Nat,?x:Nat))) = s(?y:Nat)",
"plus(plus(?z:Nat,s(?z:Nat)),plus(?y:Nat,?x:Nat)) = s(?y:Nat)",
"plus(plus(s(?x:Nat),?y:Nat),plus(?x:Nat,?y:Nat)) = s(?x:Nat)",
"plus(plus(s(?x:Nat),plus(?y:Nat,?y:Nat)),?z:Nat) = s(?x:Nat)",
"plus(plus(s(s(?z:Nat)),s(?z:Nat)),?x:Nat) = s(?z:Nat)",
"plus(plus(s(plus(?x:Nat,?z:Nat)),?x:Nat),?z:Nat) = s(?x:Nat)",
"plus(plus(s(plus(?y:Nat,?x:Nat)),?x:Nat),?y:Nat) = s(?x:Nat)",
"s(?x:Nat) = plus(?x:Nat,plus(?x:Nat,s(plus(s(?x:Nat),?y:Nat))))",
"s(?x:Nat) = plus(?x:Nat,plus(s(s(?x:Nat)),plus(?x:Nat,?y:Nat)))",
"s(?x:Nat) = plus(?y:Nat,plus(plus(?x:Nat,s(s(?x:Nat))),?y:Nat))",
"s(?x:Nat) = plus(s(plus(plus(s(?x:Nat),?z:Nat),?z:Nat)),?z:Nat)",
"s(?x:Nat) = plus(plus(?z:Nat,plus(?x:Nat,plus(?z:Nat,?y:Nat))),?y:Nat)",
"s(?x:Nat) = plus(plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),?x:Nat),?x:Nat)",
"s(?y:Nat) = plus(s(?x:Nat),s(s(plus(s(?x:Nat),?y:Nat))))",
"s(?y:Nat) = plus(s(?y:Nat),plus(?x:Nat,s(s(s(?x:Nat)))))",
"s(?y:Nat) = plus(s(plus(?x:Nat,plus(s(?x:Nat),?z:Nat))),?x:Nat)",
"s(?y:Nat) = plus(plus(plus(?x:Nat,?z:Nat),plus(?y:Nat,?y:Nat)),?y:Nat)",
"s(?y:Nat) = plus(plus(plus(?z:Nat,s(?z:Nat)),s(?z:Nat)),?y:Nat)",
"s(?z:Nat) = plus(?y:Nat,plus(plus(s(?x:Nat),?y:Nat),s(?y:Nat)))",
"s(?z:Nat) = plus(s(?x:Nat),plus(plus(?y:Nat,?x:Nat),s(?x:Nat)))",
"s(?z:Nat) = plus(plus(plus(?x:Nat,plus(?x:Nat,?y:Nat)),?x:Nat),?y:Nat)",
"s(?z:Nat) = plus(plus(plus(?y:Nat,?x:Nat),plus(?z:Nat,?z:Nat)),?x:Nat)",
"s(?z:Nat) = plus(plus(plus(?z:Nat,?z:Nat),?y:Nat),plus(?x:Nat,?y:Nat))",
"s(s(?x:Nat)) = plus(s(plus(s(s(?x:Nat)),?y:Nat)),?y:Nat)",
"s(s(?x:Nat)) = plus(s(plus(plus(?z:Nat,?x:Nat),?x:Nat)),?y:Nat)",
"s(s(?z:Nat)) = plus(s(plus(plus(?z:Nat,?y:Nat),?z:Nat)),?x:Nat)",
"plus(?x:Nat,?x:Nat) = plus(s(s(plus(?x:Nat,?x:Nat))),s(?x:Nat))",
"plus(?x:Nat,?x:Nat) = plus(s(plus(plus(?x:Nat,?y:Nat),?y:Nat)),?x:Nat)",
"plus(?x:Nat,?x:Nat) = plus(plus(s(?x:Nat),plus(?y:Nat,?x:Nat)),?x:Nat)",
"plus(?x:Nat,?y:Nat) = s(plus(?x:Nat,plus(plus(?z:Nat,?z:Nat),?y:Nat)))",
"plus(?x:Nat,?y:Nat) = plus(plus(plus(?y:Nat,?x:Nat),?y:Nat),s(?y:Nat))",
"plus(?x:Nat,?z:Nat) = s(s(plus(?y:Nat,plus(s(?z:Nat),?z:Nat))))",
"plus(?x:Nat,?z:Nat) = plus(plus(s(plus(?y:Nat,?x:Nat)),?y:Nat),?y:Nat)",
"plus(?y:Nat,?y:Nat) = s(s(plus(s(?z:Nat),plus(?x:Nat,?y:Nat))))",
"plus(?y:Nat,?y:Nat) = s(plus(?x:Nat,s(plus(?x:Nat,s(?x:Nat)))))",
"plus(?y:Nat,?y:Nat) = s(plus(?y:Nat,s(plus(s(?x:Nat),?z:Nat))))",
"plus(?y:Nat,?y:Nat) = plus(?y:Nat,plus(?x:Nat,plus(?y:Nat,s(?x:Nat))))",
"plus(?y:Nat,?y:Nat) = plus(s(?y:Nat),plus(?y:Nat,plus(?z:Nat,?y:Nat)))",
"plus(?y:Nat,?z:Nat) = plus(plus(?y:Nat,s(plus(?y:Nat,?x:Nat))),?y:Nat)",
"plus(?z:Nat,?x:Nat) = s(plus(s(?z:Nat),plus(?y:Nat,s(?x:Nat))))",
"plus(?z:Nat,?y:Nat) = s(plus(plus(?z:Nat,?x:Nat),plus(?z:Nat,?x:Nat)))",
"plus(?z:Nat,?y:Nat) = plus(?z:Nat,s(plus(?x:Nat,s(s(?x:Nat)))))",
"plus(?z:Nat,?z:Nat) = plus(?x:Nat,s(s(s(plus(?y:Nat,?y:Nat)))))",
"s(s(s(?x:Nat))) = plus(s(s(s(s(?y:Nat)))),?y:Nat)",
"s(plus(?x:Nat,?x:Nat)) = plus(plus(s(s(?x:Nat)),?z:Nat),?z:Nat)",
"s(plus(?x:Nat,?y:Nat)) = plus(plus(?x:Nat,?y:Nat),s(s(?y:Nat)))",
"s(plus(?x:Nat,?z:Nat)) = plus(s(s(s(s(?y:Nat)))),?z:Nat)",
"s(plus(?y:Nat,?x:Nat)) = plus(plus(?x:Nat,s(?x:Nat)),s(?x:Nat))",
"s(plus(?z:Nat,?x:Nat)) = plus(plus(s(?y:Nat),?y:Nat),s(?y:Nat))",
"s(plus(?z:Nat,?z:Nat)) = plus(plus(?x:Nat,?y:Nat),s(s(?x:Nat)))",
"plus(?y:Nat,s(?z:Nat)) = plus(s(s(?y:Nat)),plus(?x:Nat,?x:Nat))",
"plus(s(?x:Nat),?y:Nat) = s(plus(plus(?x:Nat,s(?y:Nat)),?y:Nat))",
"plus(s(?z:Nat),?x:Nat) = s(plus(plus(?x:Nat,?x:Nat),s(?z:Nat)))",
"plus(s(?z:Nat),?y:Nat) = s(plus(s(?y:Nat),s(s(?x:Nat))))",
"s(plus(?y:Nat,s(?y:Nat))) = plus(plus(s(?x:Nat),?y:Nat),?x:Nat)",
"plus(?x:Nat,plus(?x:Nat,?z:Nat)) = plus(plus(s(?x:Nat),?y:Nat),?y:Nat)",
"plus(?x:Nat,plus(?y:Nat,?x:Nat)) = plus(plus(?x:Nat,?y:Nat),s(?x:Nat))",
"plus(?x:Nat,plus(?z:Nat,?x:Nat)) = plus(?z:Nat,plus(?x:Nat,s(?z:Nat)))",
"plus(?y:Nat,s(s(?y:Nat))) = plus(?x:Nat,plus(s(?x:Nat),?y:Nat))",
"plus(?z:Nat,plus(?x:Nat,?z:Nat)) = plus(s(?x:Nat),plus(?z:Nat,?x:Nat))",
"plus(s(?y:Nat),s(?y:Nat)) = plus(?z:Nat,plus(?x:Nat,s(?z:Nat)))",
"plus(s(s(?y:Nat)),?x:Nat) = s(s(plus(s(?y:Nat),?x:Nat)))",
"plus(plus(?x:Nat,?y:Nat),?x:Nat) = s(plus(s(s(?y:Nat)),?x:Nat))",
"plus(plus(?z:Nat,?x:Nat),?y:Nat) = plus(?y:Nat,plus(s(?y:Nat),?y:Nat))",
"s(s(s(plus(?y:Nat,?z:Nat)))) = plus(?x:Nat,s(s(?y:Nat)))",
"s(s(plus(?z:Nat,s(?x:Nat)))) = plus(?z:Nat,s(s(?x:Nat)))",
"s(plus(?y:Nat,s(s(?x:Nat)))) = plus(s(s(?z:Nat)),?y:Nat)",
"s(plus(plus(?y:Nat,?x:Nat),?z:Nat)) = plus(plus(?x:Nat,?y:Nat),?x:Nat)",
"plus(?x:Nat,plus(?z:Nat,s(?x:Nat))) = plus(s(s(?y:Nat)),?z:Nat)",
"plus(?x:Nat,plus(s(?y:Nat),?y:Nat)) = plus(plus(?x:Nat,?y:Nat),?z:Nat)",
"plus(?x:Nat,plus(s(?z:Nat),?z:Nat)) = plus(plus(?z:Nat,?x:Nat),?x:Nat)",
"plus(s(?x:Nat),plus(?y:Nat,?x:Nat)) = s(plus(s(?z:Nat),?x:Nat))",
"plus(s(?y:Nat),plus(?z:Nat,?x:Nat)) = plus(?x:Nat,plus(?y:Nat,?x:Nat))",
"plus(s(?z:Nat),s(s(?x:Nat))) = plus(s(?x:Nat),s(?y:Nat))",
"plus(s(?z:Nat),plus(?z:Nat,?y:Nat)) = plus(?z:Nat,s(s(?y:Nat)))",
"plus(s(plus(?y:Nat,?x:Nat)),?y:Nat) = plus(plus(?x:Nat,?x:Nat),?x:Nat)",
"plus(plus(?x:Nat,s(?z:Nat)),?y:Nat) = plus(plus(?y:Nat,?y:Nat),?x:Nat)",
"plus(plus(?y:Nat,?y:Nat),s(?y:Nat)) = s(s(plus(?z:Nat,?x:Nat)))",
"plus(plus(s(?x:Nat),?z:Nat),?y:Nat) = plus(?x:Nat,plus(?y:Nat,?x:Nat))",
"plus(?x:Nat,plus(?x:Nat,plus(?x:Nat,?x:Nat))) = s(plus(?y:Nat,?z:Nat))",
"plus(?y:Nat,s(plus(s(?x:Nat),?x:Nat))) = s(plus(?z:Nat,?y:Nat))",
"plus(?z:Nat,plus(?x:Nat,plus(?x:Nat,?x:Nat))) = plus(s(?y:Nat),?x:Nat)",
"plus(?z:Nat,plus(plus(?y:Nat,?x:Nat),?y:Nat)) = plus(?z:Nat,s(?x:Nat))",
"plus(s(?z:Nat),plus(s(?z:Nat),?x:Nat)) = plus(s(?y:Nat),?y:Nat)",
"plus(s(plus(?x:Nat,?x:Nat)),s(?y:Nat)) = plus(?y:Nat,s(?x:Nat))",
"plus(plus(?x:Nat,?x:Nat),s(s(?x:Nat))) = plus(s(?z:Nat),?x:Nat)",
"plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?y:Nat)) = plus(?x:Nat,s(?z:Nat))",
"plus(plus(s(?y:Nat),s(?x:Nat)),?y:Nat) = plus(?y:Nat,s(?z:Nat))",
"plus(plus(plus(?x:Nat,?y:Nat),?y:Nat),?x:Nat) = plus(?z:Nat,s(?x:Nat))",
"s(s(s(plus(?y:Nat,plus(?z:Nat,?z:Nat))))) = plus(?y:Nat,?y:Nat)",
"s(s(plus(s(?x:Nat),plus(?z:Nat,?x:Nat)))) = plus(?x:Nat,?y:Nat)",
"s(s(plus(s(plus(?x:Nat,?x:Nat)),?x:Nat))) = plus(?x:Nat,?z:Nat)",
"s(s(plus(plus(?z:Nat,s(?y:Nat)),?y:Nat))) = plus(?y:Nat,?y:Nat)",
"s(plus(?z:Nat,plus(s(?z:Nat),s(?x:Nat)))) = plus(?z:Nat,?y:Nat)",
"s(plus(s(s(?z:Nat)),plus(?x:Nat,?z:Nat))) = plus(?y:Nat,?x:Nat)",
"s(plus(s(s(plus(?x:Nat,?x:Nat))),?x:Nat)) = plus(?x:Nat,?y:Nat)",
"s(plus(s(plus(?x:Nat,s(?y:Nat))),?y:Nat)) = plus(?z:Nat,?x:Nat)",
"s(plus(plus(?x:Nat,?y:Nat),plus(?y:Nat,?y:Nat))) = plus(?z:Nat,?x:Nat)",
"plus(?x:Nat,plus(s(plus(?y:Nat,?x:Nat)),?z:Nat)) = plus(?x:Nat,?z:Nat)",
"plus(?x:Nat,plus(plus(s(?y:Nat),?y:Nat),?y:Nat)) = s(s(?x:Nat))",
"plus(?y:Nat,s(s(s(plus(?x:Nat,?x:Nat))))) = s(s(?y:Nat))",
"plus(?z:Nat,plus(plus(?x:Nat,?x:Nat),s(?y:Nat))) = plus(?y:Nat,?x:Nat)",
"plus(?z:Nat,plus(plus(?x:Nat,s(?z:Nat)),?y:Nat)) = plus(?y:Nat,?y:Nat)",
"plus(s(plus(s(s(?x:Nat)),?x:Nat)),?z:Nat) = plus(?x:Nat,?y:Nat)",
"plus(s(plus(s(s(?y:Nat)),?z:Nat)),?y:Nat) = plus(?z:Nat,?y:Nat)",
"plus(plus(?y:Nat,?y:Nat),s(plus(?y:Nat,?x:Nat))) = plus(?x:Nat,?x:Nat)",
"plus(plus(?y:Nat,plus(?z:Nat,?y:Nat)),s(?y:Nat)) = plus(?x:Nat,?y:Nat)",
"plus(plus(?z:Nat,plus(s(?z:Nat),?x:Nat)),?z:Nat) = plus(?z:Nat,?y:Nat)",
"plus(plus(plus(?y:Nat,s(?x:Nat)),?y:Nat),?x:Nat) = plus(?x:Nat,?z:Nat)",
"plus(?x:Nat,s(plus(?x:Nat,plus(?x:Nat,s(?y:Nat))))) = s(?x:Nat)",
"plus(?x:Nat,plus(?z:Nat,plus(plus(?z:Nat,?y:Nat),?y:Nat))) = s(?z:Nat)",
"plus(?z:Nat,plus(plus(?z:Nat,?y:Nat),s(s(?y:Nat)))) = s(?y:Nat)",
"plus(s(?y:Nat),s(plus(s(?y:Nat),s(?z:Nat)))) = s(?x:Nat)",
"plus(s(?y:Nat),plus(plus(?z:Nat,?y:Nat),s(?z:Nat))) = s(?x:Nat)",
"plus(s(plus(?x:Nat,?x:Nat)),s(plus(?x:Nat,?z:Nat))) = s(?x:Nat)",
"plus(s(plus(?x:Nat,s(plus(?x:Nat,?x:Nat)))),?y:Nat) = s(?x:Nat)",
"plus(s(plus(?x:Nat,s(plus(?z:Nat,?y:Nat)))),?y:Nat) = s(?y:Nat)",
"plus(plus(?x:Nat,s(?z:Nat)),s(plus(?y:Nat,?x:Nat))) = s(?y:Nat)",
"plus(plus(?z:Nat,?x:Nat),plus(plus(?z:Nat,?x:Nat),?x:Nat)) = s(?z:Nat)",
"plus(plus(s(s(?z:Nat)),s(?x:Nat)),s(?x:Nat)) = s(?y:Nat)",
"plus(plus(plus(s(?x:Nat),?x:Nat),s(?x:Nat)),?x:Nat) = s(?z:Nat)",
"plus(plus(plus(plus(?x:Nat,?y:Nat),?x:Nat),?x:Nat),?z:Nat) = s(?z:Nat)",
"s(?x:Nat) = plus(s(s(plus(?x:Nat,plus(?x:Nat,?y:Nat)))),s(?y:Nat))",
"s(?x:Nat) = plus(s(plus(?x:Nat,plus(plus(?x:Nat,?y:Nat),?y:Nat))),?z:Nat)",
"s(?x:Nat) = plus(plus(?y:Nat,plus(s(plus(?x:Nat,?y:Nat)),?x:Nat)),?x:Nat)",
"s(?x:Nat) = plus(plus(plus(?y:Nat,s(plus(?x:Nat,?y:Nat))),?y:Nat),?y:Nat)",
"s(?y:Nat) = plus(s(s(?x:Nat)),s(plus(s(s(?z:Nat)),?x:Nat)))",
"s(?y:Nat) = plus(plus(s(plus(?y:Nat,s(?z:Nat))),?x:Nat),s(?y:Nat))",
"s(?z:Nat) = plus(?z:Nat,plus(plus(?y:Nat,?y:Nat),s(s(s(?y:Nat)))))",
"s(?z:Nat) = plus(plus(?z:Nat,?z:Nat),s(plus(?z:Nat,s(s(?z:Nat)))))",
"s(s(?x:Nat)) = plus(?y:Nat,plus(plus(?y:Nat,plus(?y:Nat,?x:Nat)),?x:Nat))",
"s(s(?x:Nat)) = plus(?z:Nat,plus(s(plus(?z:Nat,s(?z:Nat))),?x:Nat))",
"s(s(?x:Nat)) = plus(s(s(?x:Nat)),plus(s(?y:Nat),s(?z:Nat)))",
"s(s(?x:Nat)) = plus(plus(plus(plus(?y:Nat,?y:Nat),?z:Nat),?y:Nat),?y:Nat)",
"s(s(?z:Nat)) = plus(plus(?z:Nat,s(s(plus(?y:Nat,?x:Nat)))),?x:Nat)",
"s(s(?z:Nat)) = plus(plus(?z:Nat,plus(plus(?y:Nat,?z:Nat),?x:Nat)),?x:Nat)",
"plus(?x:Nat,?x:Nat) = s(plus(s(?x:Nat),plus(plus(?x:Nat,?x:Nat),?x:Nat)))",
"plus(?x:Nat,?x:Nat) = plus(s(?z:Nat),plus(plus(?y:Nat,s(?z:Nat)),?y:Nat))",
"plus(?x:Nat,?x:Nat) = plus(s(plus(plus(?y:Nat,s(?x:Nat)),?x:Nat)),?y:Nat)",
"plus(?x:Nat,?x:Nat) = plus(plus(s(?z:Nat),plus(s(?y:Nat),?z:Nat)),?y:Nat)",
"plus(?x:Nat,?x:Nat) = plus(plus(plus(plus(?x:Nat,?x:Nat),?y:Nat),?z:Nat),?x:Nat)",
"plus(?x:Nat,?y:Nat) = plus(plus(?y:Nat,plus(?y:Nat,plus(?y:Nat,?x:Nat))),?x:Nat)",
"plus(?x:Nat,?z:Nat) = s(plus(plus(?z:Nat,s(plus(?y:Nat,?x:Nat))),?x:Nat))",
"plus(?x:Nat,?z:Nat) = plus(s(s(s(?y:Nat))),plus(?y:Nat,s(?z:Nat)))",
"plus(?y:Nat,?x:Nat) = plus(s(plus(plus(?y:Nat,s(?y:Nat)),?y:Nat)),?y:Nat)",
"plus(?y:Nat,?y:Nat) = s(s(plus(plus(plus(?x:Nat,?y:Nat),?y:Nat),?y:Nat)))",
"plus(?y:Nat,?y:Nat) = s(plus(?x:Nat,s(plus(plus(?z:Nat,?x:Nat),?x:Nat))))",
"plus(?y:Nat,?z:Nat) = plus(s(plus(s(s(?x:Nat)),?x:Nat)),s(?x:Nat))",
"plus(?z:Nat,?x:Nat) = s(plus(s(plus(plus(?x:Nat,?y:Nat),?x:Nat)),?x:Nat))",
"plus(?z:Nat,?x:Nat) = plus(?y:Nat,s(s(plus(plus(?x:Nat,?z:Nat),?x:Nat))))",
"plus(?z:Nat,?y:Nat) = s(plus(plus(?y:Nat,?x:Nat),s(plus(?x:Nat,?x:Nat))))",
"s(plus(?x:Nat,?x:Nat)) = plus(?x:Nat,plus(s(?z:Nat),plus(?x:Nat,?x:Nat)))",
"s(plus(?x:Nat,?z:Nat)) = plus(?x:Nat,s(s(plus(s(?z:Nat),?x:Nat))))",
"s(plus(?y:Nat,?y:Nat)) = plus(s(plus(?z:Nat,?x:Nat)),plus(?y:Nat,?y:Nat))",
"plus(?y:Nat,s(?z:Nat)) = plus(?x:Nat,plus(?x:Nat,s(plus(?x:Nat,?z:Nat))))",
"plus(s(?x:Nat),?y:Nat) = s(s(plus(s(plus(?x:Nat,?x:Nat)),?y:Nat)))",
"plus(s(?y:Nat),?x:Nat) = s(plus(?y:Nat,plus(?x:Nat,s(s(?x:Nat)))))",
"plus(s(?y:Nat),?y:Nat) = s(plus(s(?y:Nat),s(s(s(?y:Nat)))))",
"plus(s(?z:Nat),?x:Nat) = plus(s(?x:Nat),plus(s(s(?x:Nat)),?x:Nat))",
"plus(s(?y:Nat),s(?y:Nat)) = s(plus(?y:Nat,s(plus(?y:Nat,?x:Nat))))",
"plus(s(s(?x:Nat)),?x:Nat) = plus(plus(?x:Nat,?z:Nat),plus(?x:Nat,?y:Nat))",
"plus(plus(?x:Nat,?x:Nat),?x:Nat) = plus(?x:Nat,s(s(s(s(?z:Nat)))))",
"plus(plus(?x:Nat,?x:Nat),?y:Nat) = s(plus(s(plus(?z:Nat,?z:Nat)),?z:Nat))",
"plus(plus(?z:Nat,?y:Nat),?x:Nat) = plus(?y:Nat,plus(?z:Nat,plus(?y:Nat,?y:Nat)))",
"s(plus(?x:Nat,plus(?x:Nat,?z:Nat))) = plus(plus(s(?z:Nat),?y:Nat),?z:Nat)",
"plus(?x:Nat,s(plus(?z:Nat,?y:Nat))) = plus(s(?z:Nat),s(s(?y:Nat)))",
"plus(s(?x:Nat),s(s(?x:Nat))) = plus(plus(?z:Nat,?z:Nat),s(?x:Nat))",
"plus(s(s(s(?y:Nat))),?z:Nat) = s(plus(plus(?x:Nat,?x:Nat),?z:Nat))",
"plus(s(plus(?x:Nat,?x:Nat)),?x:Nat) = plus(plus(?y:Nat,s(?z:Nat)),?y:Nat)",
"plus(?y:Nat,plus(?x:Nat,s(s(?z:Nat)))) = plus(?z:Nat,s(s(?x:Nat)))",
"plus(?z:Nat,s(plus(?x:Nat,s(?x:Nat)))) = plus(?y:Nat,s(s(?y:Nat)))",
"plus(?z:Nat,plus(plus(?x:Nat,?x:Nat),?y:Nat)) = s(plus(s(?y:Nat),?y:Nat))",
"plus(s(s(s(?z:Nat))),s(?x:Nat)) = s(plus(?y:Nat,s(?y:Nat)))",
"plus(s(s(s(s(?x:Nat)))),?x:Nat) = plus(?y:Nat,plus(?x:Nat,?y:Nat))",
"plus(s(s(s(s(?z:Nat)))),?y:Nat) = plus(s(s(?x:Nat)),?y:Nat)",
"plus(s(s(plus(?x:Nat,?x:Nat))),?y:Nat) = plus(?z:Nat,plus(?x:Nat,?y:Nat))",
"plus(s(plus(?y:Nat,?z:Nat)),s(?z:Nat)) = s(plus(?x:Nat,s(?x:Nat)))",
"plus(s(plus(s(?z:Nat),?z:Nat)),?y:Nat) = s(plus(s(?x:Nat),?z:Nat))",
"plus(plus(?x:Nat,plus(?z:Nat,?y:Nat)),?z:Nat) = plus(?z:Nat,plus(?y:Nat,?y:Nat))",
"plus(plus(?y:Nat,plus(?z:Nat,?x:Nat)),?x:Nat) = plus(plus(?x:Nat,?y:Nat),?z:Nat)",
"plus(plus(?z:Nat,plus(?y:Nat,?x:Nat)),?z:Nat) = s(plus(?y:Nat,s(?x:Nat)))",
"s(plus(plus(?y:Nat,?z:Nat),plus(?z:Nat,?y:Nat))) = plus(s(?x:Nat),?x:Nat)",
"plus(?y:Nat,s(s(s(plus(?x:Nat,?x:Nat))))) = plus(s(?y:Nat),?x:Nat)",
"plus(?z:Nat,plus(plus(?x:Nat,?x:Nat),s(?y:Nat))) = s(plus(?z:Nat,?x:Nat))",
"plus(s(?y:Nat),plus(?z:Nat,plus(?x:Nat,?z:Nat))) = s(plus(?x:Nat,?x:Nat))",
"plus(s(plus(s(?x:Nat),s(?z:Nat))),?y:Nat) = s(plus(?y:Nat,?y:Nat))",
"plus(s(plus(plus(?x:Nat,?z:Nat),?x:Nat)),?y:Nat) = s(plus(?y:Nat,?x:Nat))",
"plus(s(plus(plus(?x:Nat,?z:Nat),?x:Nat)),?y:Nat) = s(plus(?z:Nat,?y:Nat))",
"s(s(plus(plus(s(?x:Nat),s(?x:Nat)),?x:Nat))) = plus(?z:Nat,?x:Nat)",
"s(plus(?x:Nat,plus(plus(?z:Nat,s(?x:Nat)),?x:Nat))) = plus(?x:Nat,?y:Nat)",
"s(plus(plus(?y:Nat,s(plus(?x:Nat,?x:Nat))),?x:Nat)) = plus(?x:Nat,?y:Nat)",
"plus(?x:Nat,plus(s(s(s(?y:Nat))),s(?x:Nat))) = plus(?y:Nat,?x:Nat)",
"plus(?y:Nat,s(plus(s(s(?x:Nat)),s(?x:Nat)))) = plus(?y:Nat,?x:Nat)",
"plus(?y:Nat,s(plus(s(s(?z:Nat)),s(?y:Nat)))) = plus(?x:Nat,?y:Nat)",
"plus(?y:Nat,plus(?x:Nat,plus(s(s(?x:Nat)),?z:Nat))) = plus(?x:Nat,?x:Nat)",
"plus(?y:Nat,plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),?y:Nat)) = s(s(?z:Nat))",
"plus(?z:Nat,plus(s(s(plus(?y:Nat,?x:Nat))),?x:Nat)) = plus(?x:Nat,?z:Nat)",
"plus(s(?y:Nat),plus(?x:Nat,plus(s(?x:Nat),?y:Nat))) = plus(?y:Nat,?y:Nat)",
"plus(s(?y:Nat),plus(plus(?z:Nat,s(?x:Nat)),?z:Nat)) = plus(?x:Nat,?x:Nat)",
"plus(s(?z:Nat),plus(plus(s(?y:Nat),?z:Nat),?x:Nat)) = s(s(?z:Nat))",
"plus(s(s(plus(?x:Nat,?x:Nat))),plus(?y:Nat,?z:Nat)) = plus(?y:Nat,?y:Nat)",
"plus(s(plus(?x:Nat,?x:Nat)),plus(s(?y:Nat),?z:Nat)) = plus(?x:Nat,?y:Nat)",
"plus(plus(?y:Nat,s(plus(s(?x:Nat),?y:Nat))),?x:Nat) = s(s(?z:Nat))",
"plus(plus(s(s(?x:Nat)),?z:Nat),s(s(?z:Nat))) = plus(?y:Nat,?x:Nat)",
"plus(plus(plus(?z:Nat,?z:Nat),plus(?z:Nat,?x:Nat)),?y:Nat) = plus(?y:Nat,?x:Nat)",
"plus(plus(plus(s(?x:Nat),s(?x:Nat)),?x:Nat),?x:Nat) = s(s(?y:Nat))",
"plus(plus(plus(s(s(?y:Nat)),?y:Nat),?y:Nat),?x:Nat) = s(s(?y:Nat))",
"plus(plus(plus(plus(?x:Nat,?x:Nat),?x:Nat),?y:Nat),?x:Nat) = plus(?x:Nat,?x:Nat)",
"plus(?x:Nat,s(plus(?y:Nat,plus(?y:Nat,plus(?x:Nat,?x:Nat))))) = s(?x:Nat)",
"plus(?x:Nat,s(plus(?z:Nat,plus(s(?x:Nat),s(?x:Nat))))) = s(?y:Nat)",
"plus(?x:Nat,plus(?y:Nat,plus(?y:Nat,plus(s(?z:Nat),?x:Nat)))) = s(?y:Nat)",
"plus(?x:Nat,plus(s(?x:Nat),plus(?y:Nat,s(s(?x:Nat))))) = s(?y:Nat)",
"plus(?x:Nat,plus(plus(?y:Nat,plus(s(?z:Nat),?x:Nat)),?z:Nat)) = s(?x:Nat)",
"plus(?y:Nat,plus(s(?y:Nat),plus(s(s(?z:Nat)),?x:Nat))) = s(?y:Nat)",
"plus(?z:Nat,plus(s(?y:Nat),plus(?y:Nat,plus(?x:Nat,?y:Nat)))) = s(?y:Nat)",
"plus(s(?x:Nat),s(plus(?x:Nat,plus(s(?y:Nat),?x:Nat)))) = s(?x:Nat)",
"plus(s(?y:Nat),s(plus(plus(?z:Nat,?x:Nat),s(?y:Nat)))) = s(?y:Nat)",
"plus(s(?y:Nat),plus(?x:Nat,s(s(plus(?x:Nat,?x:Nat))))) = s(?x:Nat)",
"plus(s(?y:Nat),plus(s(?x:Nat),s(plus(?x:Nat,?y:Nat)))) = s(?z:Nat)",
"plus(s(?z:Nat),plus(?y:Nat,s(plus(s(?x:Nat),?z:Nat)))) = s(?z:Nat)",
"plus(s(s(?x:Nat)),plus(?y:Nat,plus(?x:Nat,s(?z:Nat)))) = s(?y:Nat)",
"plus(s(plus(plus(plus(?x:Nat,?x:Nat),?x:Nat),?y:Nat)),?y:Nat) = s(?z:Nat)",
"plus(plus(?x:Nat,s(?y:Nat)),plus(?y:Nat,s(s(?y:Nat)))) = s(?y:Nat)",
"plus(plus(?y:Nat,plus(?y:Nat,s(plus(?x:Nat,?x:Nat)))),?y:Nat) = s(?x:Nat)",
"plus(plus(s(plus(?y:Nat,plus(?z:Nat,?x:Nat))),?x:Nat),?z:Nat) = s(?x:Nat)",
"plus(plus(s(plus(plus(?z:Nat,?z:Nat),?z:Nat)),?z:Nat),?y:Nat) = s(?x:Nat)",
"plus(plus(plus(plus(?x:Nat,?x:Nat),?x:Nat),?x:Nat),s(?y:Nat)) = s(?x:Nat)",
"s(?x:Nat) = plus(?x:Nat,plus(?x:Nat,plus(?y:Nat,plus(?z:Nat,plus(?z:Nat,?x:Nat)))))",
"s(?x:Nat) = plus(?x:Nat,plus(plus(plus(s(?y:Nat),s(?z:Nat)),?z:Nat),?x:Nat))",
"s(?x:Nat) = plus(?z:Nat,plus(s(?x:Nat),plus(s(?z:Nat),plus(?y:Nat,?z:Nat))))",
"s(?x:Nat) = plus(plus(plus(?y:Nat,plus(plus(?z:Nat,?x:Nat),?y:Nat)),?z:Nat),?x:Nat)",
"s(?y:Nat) = plus(?y:Nat,plus(?y:Nat,plus(plus(plus(?x:Nat,?z:Nat),?z:Nat),?x:Nat)))",
"s(?y:Nat) = plus(s(s(plus(s(?z:Nat),plus(?x:Nat,?y:Nat)))),s(?y:Nat))",
"s(?y:Nat) = plus(s(plus(plus(s(?z:Nat),plus(?x:Nat,?y:Nat)),?z:Nat)),?x:Nat)",
"s(?y:Nat) = plus(plus(plus(s(?z:Nat),s(plus(?y:Nat,?z:Nat))),?z:Nat),?y:Nat)",
"s(?z:Nat) = plus(plus(plus(plus(?z:Nat,?y:Nat),plus(?x:Nat,?y:Nat)),?x:Nat),?z:Nat)",
"s(s(?x:Nat)) = plus(plus(s(?x:Nat),?x:Nat),plus(plus(?x:Nat,?z:Nat),?x:Nat))",
"plus(?x:Nat,?x:Nat) = s(s(plus(s(s(?z:Nat)),plus(?y:Nat,s(?y:Nat)))))",
"plus(?x:Nat,?x:Nat) = s(plus(?y:Nat,plus(s(s(?z:Nat)),plus(?z:Nat,?y:Nat))))",
"plus(?x:Nat,?x:Nat) = plus(?y:Nat,plus(plus(?y:Nat,s(?z:Nat)),plus(?z:Nat,?y:Nat)))",
"plus(?x:Nat,?x:Nat) = plus(s(plus(s(?z:Nat),s(s(?x:Nat)))),s(?x:Nat))",
"plus(?x:Nat,?y:Nat) = s(plus(?y:Nat,plus(?x:Nat,plus(plus(?x:Nat,?y:Nat),?x:Nat))))",
"plus(?x:Nat,?y:Nat) = plus(plus(plus(?z:Nat,s(?x:Nat)),plus(?x:Nat,?x:Nat)),?x:Nat)",
"plus(?x:Nat,?y:Nat) = plus(plus(plus(s(?x:Nat),plus(?x:Nat,?x:Nat)),?x:Nat),?x:Nat)",
"plus(?x:Nat,?z:Nat) = plus(plus(s(plus(?x:Nat,plus(?x:Nat,?y:Nat))),?y:Nat),?y:Nat)",
"plus(?x:Nat,?z:Nat) = plus(plus(s(plus(plus(?x:Nat,?z:Nat),?y:Nat)),?x:Nat),?z:Nat)",
"plus(?y:Nat,?x:Nat) = plus(?z:Nat,s(plus(?z:Nat,plus(plus(?x:Nat,?y:Nat),?x:Nat))))",
"plus(?z:Nat,?x:Nat) = plus(?y:Nat,plus(?x:Nat,plus(plus(?x:Nat,?y:Nat),s(?x:Nat))))",
"plus(?z:Nat,?x:Nat) = plus(plus(plus(s(plus(?y:Nat,?x:Nat)),?y:Nat),?y:Nat),?y:Nat)",
"plus(?z:Nat,?x:Nat) = plus(plus(plus(s(plus(?z:Nat,?z:Nat)),?y:Nat),?z:Nat),?x:Nat)",
"plus(?z:Nat,?y:Nat) = plus(plus(plus(?x:Nat,?x:Nat),s(?x:Nat)),plus(?z:Nat,?z:Nat))",
"s(s(s(?y:Nat))) = plus(plus(plus(?z:Nat,?z:Nat),plus(?z:Nat,?x:Nat)),?y:Nat)",
"s(s(s(?z:Nat))) = plus(plus(?y:Nat,plus(?x:Nat,?y:Nat)),plus(?z:Nat,?y:Nat))",
"s(plus(?x:Nat,?z:Nat)) = plus(?x:Nat,s(plus(s(s(?y:Nat)),s(?x:Nat))))",
"s(plus(?z:Nat,?x:Nat)) = plus(?x:Nat,plus(plus(?z:Nat,?z:Nat),plus(?z:Nat,?z:Nat)))",
"plus(?x:Nat,s(?x:Nat)) = plus(plus(plus(?y:Nat,?y:Nat),?z:Nat),plus(?x:Nat,?x:Nat))",
"plus(s(?z:Nat),?y:Nat) = plus(plus(plus(?x:Nat,plus(?z:Nat,?y:Nat)),?x:Nat),?x:Nat)",
"s(s(plus(?x:Nat,?x:Nat))) = plus(?x:Nat,plus(plus(?y:Nat,s(?x:Nat)),?z:Nat))",
"plus(?x:Nat,plus(?x:Nat,?y:Nat)) = plus(plus(plus(?y:Nat,?y:Nat),?x:Nat),s(?x:Nat))",
"plus(?x:Nat,plus(?y:Nat,?x:Nat)) = plus(plus(?z:Nat,?y:Nat),s(plus(?x:Nat,?z:Nat)))",
"plus(?x:Nat,plus(?z:Nat,?x:Nat)) = s(s(plus(plus(?x:Nat,?x:Nat),s(?y:Nat))))",
"plus(s(?z:Nat),s(?z:Nat)) = plus(s(plus(s(?x:Nat),?x:Nat)),s(?y:Nat))",
"plus(plus(?y:Nat,?x:Nat),?x:Nat) = s(s(plus(s(?x:Nat),plus(?y:Nat,?y:Nat))))",
"plus(plus(?y:Nat,?y:Nat),?x:Nat) = s(s(s(plus(plus(?z:Nat,?y:Nat),?z:Nat))))",
"plus(?x:Nat,s(s(s(?y:Nat)))) = plus(plus(?x:Nat,plus(?x:Nat,?z:Nat)),?y:Nat)",
"plus(?y:Nat,plus(s(?x:Nat),?x:Nat)) = plus(?y:Nat,plus(s(s(?z:Nat)),?y:Nat))",
"plus(?y:Nat,plus(?x:Nat,plus(?y:Nat,?y:Nat))) = plus(plus(?x:Nat,s(?z:Nat)),?x:Nat)",
"plus(?y:Nat,plus(s(s(?x:Nat)),?x:Nat)) = plus(?z:Nat,plus(?x:Nat,s(?x:Nat)))",
"plus(?z:Nat,plus(?z:Nat,plus(?y:Nat,?y:Nat))) = plus(plus(?x:Nat,s(?y:Nat)),?x:Nat)",
"s(s(plus(s(?y:Nat),plus(?y:Nat,?z:Nat)))) = plus(?z:Nat,s(s(?y:Nat)))",
"s(s(plus(s(plus(?x:Nat,?z:Nat)),?y:Nat))) = plus(plus(?y:Nat,?y:Nat),?x:Nat)",
"s(plus(s(?x:Nat),plus(s(?z:Nat),?y:Nat))) = plus(plus(?y:Nat,?y:Nat),?y:Nat)",
"s(plus(s(s(s(?y:Nat))),s(?z:Nat))) = plus(plus(?z:Nat,?z:Nat),?x:Nat)",
"plus(?x:Nat,s(s(plus(s(?y:Nat),?z:Nat)))) = plus(plus(?x:Nat,?x:Nat),?x:Nat)",
"plus(?x:Nat,plus(s(?x:Nat),plus(?x:Nat,?z:Nat))) = plus(plus(?x:Nat,?x:Nat),?y:Nat)",
"plus(?x:Nat,plus(plus(?z:Nat,s(?x:Nat)),?y:Nat)) = s(plus(s(?z:Nat),?y:Nat))",
"plus(?y:Nat,plus(s(s(s(?y:Nat))),?x:Nat)) = plus(?y:Nat,s(s(?z:Nat)))",
"plus(plus(?y:Nat,s(plus(?x:Nat,?x:Nat))),?x:Nat) = plus(plus(?y:Nat,?x:Nat),?x:Nat)",
"plus(plus(s(?z:Nat),?z:Nat),plus(?x:Nat,?x:Nat)) = plus(plus(?z:Nat,?x:Nat),?z:Nat)",
"s(plus(?z:Nat,plus(?x:Nat,s(plus(?z:Nat,?x:Nat))))) = plus(?y:Nat,s(?x:Nat))",
"plus(?x:Nat,plus(?x:Nat,s(s(plus(?y:Nat,?x:Nat))))) = s(plus(?y:Nat,?x:Nat))",
"plus(?x:Nat,plus(plus(?x:Nat,plus(?y:Nat,?x:Nat)),?x:Nat)) = plus(?y:Nat,s(?x:Nat))",
"plus(?y:Nat,plus(plus(plus(?x:Nat,?x:Nat),?x:Nat),?z:Nat)) = plus(?x:Nat,s(?z:Nat))",
"plus(s(?z:Nat),plus(s(?x:Nat),plus(?x:Nat,?y:Nat))) = plus(?y:Nat,s(?x:Nat))",
"plus(plus(s(?y:Nat),plus(?y:Nat,s(?y:Nat))),?z:Nat) = s(plus(?y:Nat,?y:Nat))",
"plus(plus(plus(?x:Nat,?x:Nat),?x:Nat),plus(?y:Nat,?y:Nat)) = s(plus(?y:Nat,?y:Nat))",
"s(plus(s(plus(?y:Nat,plus(?y:Nat,?y:Nat))),s(?y:Nat))) = plus(?z:Nat,?z:Nat)",
"s(plus(plus(?x:Nat,plus(plus(?x:Nat,?z:Nat),?y:Nat)),?x:Nat)) = plus(?y:Nat,?x:Nat)",
"plus(?x:Nat,s(plus(?y:Nat,plus(?x:Nat,plus(?y:Nat,?y:Nat))))) = plus(?x:Nat,?y:Nat)",
"plus(?x:Nat,s(plus(s(plus(?z:Nat,?x:Nat)),s(?x:Nat)))) = plus(?x:Nat,?x:Nat)",
"plus(?x:Nat,plus(plus(?y:Nat,plus(s(?y:Nat),?y:Nat)),?z:Nat)) = plus(?z:Nat,?y:Nat)",
"plus(?y:Nat,s(plus(?x:Nat,plus(plus(?y:Nat,?x:Nat),?z:Nat)))) = plus(?x:Nat,?x:Nat)",
"plus(s(plus(?x:Nat,plus(s(?z:Nat),?y:Nat))),s(?x:Nat)) = plus(?x:Nat,?x:Nat)",
"plus(s(plus(plus(?x:Nat,?y:Nat),?x:Nat)),plus(?x:Nat,?y:Nat)) = s(s(?z:Nat))",
"plus(plus(?x:Nat,s(plus(plus(?y:Nat,?x:Nat),?y:Nat))),?x:Nat) = plus(?y:Nat,?x:Nat)",
"plus(plus(?y:Nat,plus(?x:Nat,?z:Nat)),s(plus(?x:Nat,?y:Nat))) = s(s(?x:Nat))",
"plus(plus(?z:Nat,plus(?x:Nat,?y:Nat)),s(plus(?y:Nat,?x:Nat))) = plus(?z:Nat,?z:Nat)",
"plus(plus(s(s(plus(?x:Nat,?z:Nat))),?y:Nat),s(?x:Nat)) = plus(?z:Nat,?z:Nat)",
"plus(plus(plus(?x:Nat,?x:Nat),s(?x:Nat)),s(s(?x:Nat))) = plus(?x:Nat,?y:Nat)",
"plus(?x:Nat,plus(s(plus(plus(?x:Nat,?y:Nat),?x:Nat)),s(?x:Nat))) = s(?x:Nat)",
"plus(?y:Nat,s(plus(plus(plus(?y:Nat,?y:Nat),s(?y:Nat)),?y:Nat))) = s(?x:Nat)",
"plus(?z:Nat,plus(s(plus(?y:Nat,?x:Nat)),s(plus(?z:Nat,?y:Nat)))) = s(?x:Nat)",
"plus(?z:Nat,plus(plus(plus(s(?y:Nat),?y:Nat),?x:Nat),s(?x:Nat))) = s(?x:Nat)",
"plus(plus(?x:Nat,?x:Nat),plus(?y:Nat,s(plus(s(?y:Nat),?x:Nat)))) = s(?z:Nat)",
"plus(plus(?x:Nat,?x:Nat),plus(plus(?y:Nat,plus(?x:Nat,?x:Nat)),?x:Nat)) = s(?x:Nat)",
"plus(plus(?x:Nat,s(plus(?y:Nat,?z:Nat))),plus(?x:Nat,s(?x:Nat))) = s(?x:Nat)",
"plus(plus(?x:Nat,plus(s(plus(?x:Nat,?z:Nat)),s(?x:Nat))),?y:Nat) = s(?y:Nat)",
"plus(plus(?y:Nat,plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,?y:Nat))),?x:Nat) = s(?y:Nat)",
"plus(plus(s(?y:Nat),?y:Nat),s(plus(?x:Nat,s(s(?z:Nat))))) = s(?z:Nat)",
"plus(plus(s(s(s(?x:Nat))),s(plus(?x:Nat,?y:Nat))),?z:Nat) = s(?y:Nat)",
"plus(plus(s(plus(s(?z:Nat),?y:Nat)),?y:Nat),s(s(?z:Nat))) = s(?x:Nat)",
"plus(plus(plus(?x:Nat,?y:Nat),?x:Nat),plus(plus(?y:Nat,?x:Nat),?y:Nat)) = s(?y:Nat)",
"plus(plus(plus(?x:Nat,s(?y:Nat)),plus(?y:Nat,s(?x:Nat))),?y:Nat) = s(?x:Nat)",
"plus(plus(plus(plus(?y:Nat,?x:Nat),?y:Nat),?x:Nat),s(s(?y:Nat))) = s(?x:Nat)",
"plus(plus(plus(plus(plus(?y:Nat,?z:Nat),?y:Nat),?x:Nat),?z:Nat),?z:Nat) = s(?x:Nat)",
"s(?x:Nat) = plus(plus(s(plus(plus(?x:Nat,?z:Nat),?y:Nat)),s(?x:Nat)),s(?x:Nat))",
"s(?x:Nat) = plus(plus(plus(s(?x:Nat),?z:Nat),plus(?y:Nat,s(?x:Nat))),s(?x:Nat))",
"s(?z:Nat) = plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,plus(s(?x:Nat),s(s(?y:Nat)))))",
"s(?z:Nat) = plus(plus(s(plus(plus(?x:Nat,?z:Nat),plus(?x:Nat,?x:Nat))),?y:Nat),?z:Nat)",
"s(s(?x:Nat)) = plus(s(s(plus(?x:Nat,?y:Nat))),s(plus(s(?x:Nat),?z:Nat)))",
"plus(?x:Nat,?x:Nat) = plus(?y:Nat,s(s(plus(plus(?y:Nat,?x:Nat),plus(?z:Nat,?z:Nat)))))",
"plus(?y:Nat,?y:Nat) = s(plus(plus(?y:Nat,?y:Nat),s(plus(plus(?z:Nat,?x:Nat),?x:Nat))))",
"plus(?y:Nat,?y:Nat) = plus(plus(plus(?z:Nat,plus(?x:Nat,s(?y:Nat))),s(?x:Nat)),?y:Nat)",
"plus(?z:Nat,?x:Nat) = plus(?x:Nat,plus(plus(s(plus(?z:Nat,?x:Nat)),?z:Nat),s(?y:Nat)))",
"plus(?z:Nat,?y:Nat) = plus(?y:Nat,plus(plus(?x:Nat,plus(plus(?x:Nat,?z:Nat),?y:Nat)),?y:Nat))",
"s(plus(?x:Nat,?x:Nat)) = plus(?x:Nat,plus(s(?x:Nat),plus(?x:Nat,plus(?y:Nat,?x:Nat))))",
"plus(s(?x:Nat),?x:Nat) = plus(plus(?x:Nat,s(plus(plus(?x:Nat,?x:Nat),?x:Nat))),?z:Nat)",
"plus(?x:Nat,plus(?x:Nat,?x:Nat)) = plus(s(?y:Nat),s(s(plus(s(?x:Nat),?x:Nat))))",
"plus(?x:Nat,plus(?x:Nat,?z:Nat)) = plus(plus(?z:Nat,plus(?y:Nat,?x:Nat)),plus(?y:Nat,?y:Nat))",
"plus(?x:Nat,plus(?z:Nat,?x:Nat)) = plus(?y:Nat,s(plus(plus(s(?x:Nat),?x:Nat),?y:Nat)))",
"plus(s(?x:Nat),s(?x:Nat)) = s(plus(s(plus(?z:Nat,?x:Nat)),plus(?y:Nat,?x:Nat)))",
"s(s(plus(?y:Nat,s(?x:Nat)))) = plus(s(s(s(plus(?y:Nat,?x:Nat)))),?x:Nat)",
"plus(s(s(?x:Nat)),s(?x:Nat)) = s(plus(plus(?y:Nat,?z:Nat),plus(?z:Nat,?x:Nat)))",
"plus(plus(?y:Nat,s(?x:Nat)),?z:Nat) = s(plus(plus(?x:Nat,plus(?z:Nat,?z:Nat)),?x:Nat))",
"s(s(plus(?z:Nat,plus(?x:Nat,?y:Nat)))) = plus(s(s(s(s(?x:Nat)))),?x:Nat)",
"s(s(plus(s(s(?x:Nat)),?y:Nat))) = plus(s(plus(?x:Nat,s(?z:Nat))),?x:Nat)",
"s(plus(s(plus(?x:Nat,?x:Nat)),?x:Nat)) = plus(plus(s(?x:Nat),?y:Nat),s(?x:Nat))",
"plus(?y:Nat,plus(?y:Nat,plus(?x:Nat,?x:Nat))) = s(plus(s(?y:Nat),plus(?y:Nat,?y:Nat)))",
"plus(?y:Nat,plus(?y:Nat,plus(?z:Nat,?x:Nat))) = plus(plus(plus(?x:Nat,?z:Nat),?y:Nat),?z:Nat)",
"plus(s(plus(?x:Nat,s(?z:Nat))),?x:Nat) = plus(plus(?z:Nat,plus(?x:Nat,?z:Nat)),?x:Nat)",
"plus(plus(?x:Nat,?y:Nat),s(s(?z:Nat))) = plus(?x:Nat,plus(plus(?y:Nat,?x:Nat),?z:Nat))",
"s(plus(?y:Nat,plus(?x:Nat,plus(?x:Nat,?x:Nat)))) = plus(s(s(?x:Nat)),s(?y:Nat))",
"plus(s(?y:Nat),plus(plus(?z:Nat,?y:Nat),?y:Nat)) = s(plus(s(?y:Nat),s(?x:Nat)))",
"plus(plus(?x:Nat,s(?y:Nat)),plus(?y:Nat,?x:Nat)) = plus(?x:Nat,s(plus(?z:Nat,?x:Nat)))",
"plus(?z:Nat,s(s(plus(?x:Nat,plus(?x:Nat,?z:Nat))))) = plus(?x:Nat,plus(?x:Nat,?x:Nat))",
"plus(?z:Nat,s(plus(plus(?z:Nat,s(?z:Nat)),?y:Nat))) = plus(?z:Nat,plus(?x:Nat,?y:Nat))",
"plus(?z:Nat,s(plus(plus(s(?x:Nat),?x:Nat),?y:Nat))) = plus(?y:Nat,plus(?z:Nat,?x:Nat))",
"plus(s(plus(s(plus(?x:Nat,?z:Nat)),?y:Nat)),?x:Nat) = plus(?y:Nat,plus(?y:Nat,?z:Nat))",
"plus(plus(?x:Nat,s(?x:Nat)),s(plus(?y:Nat,?x:Nat))) = plus(s(s(?z:Nat)),?y:Nat)",
"plus(plus(s(plus(?x:Nat,?x:Nat)),s(?y:Nat)),?y:Nat) = plus(?x:Nat,s(s(?z:Nat)))",
"s(plus(?x:Nat,plus(plus(?z:Nat,plus(?z:Nat,?y:Nat)),?y:Nat))) = plus(?z:Nat,s(?x:Nat))",
"s(plus(plus(plus(?x:Nat,plus(?z:Nat,?x:Nat)),?x:Nat),?y:Nat)) = plus(s(?y:Nat),?z:Nat)",
"plus(?y:Nat,plus(s(plus(?x:Nat,plus(?z:Nat,?x:Nat))),?x:Nat)) = plus(s(?x:Nat),?y:Nat)",
"plus(?y:Nat,plus(s(plus(s(?z:Nat),s(?x:Nat))),?x:Nat)) = plus(?x:Nat,s(?z:Nat))",
"plus(plus(?z:Nat,?y:Nat),s(plus(?y:Nat,plus(?x:Nat,?y:Nat)))) = s(plus(?x:Nat,?y:Nat))",
"plus(plus(plus(?z:Nat,s(plus(?z:Nat,?x:Nat))),?y:Nat),?y:Nat) = plus(?x:Nat,s(?z:Nat))",
"plus(plus(plus(s(s(?y:Nat)),s(?x:Nat)),?x:Nat),?x:Nat) = plus(?z:Nat,s(?z:Nat))",
"s(plus(s(plus(plus(?x:Nat,?y:Nat),plus(?x:Nat,?x:Nat))),?x:Nat)) = plus(?y:Nat,?z:Nat)",
"plus(?x:Nat,plus(?z:Nat,plus(plus(plus(?y:Nat,?z:Nat),?x:Nat),?y:Nat))) = plus(?y:Nat,?y:Nat)",
"plus(?y:Nat,plus(plus(?x:Nat,plus(plus(?y:Nat,?z:Nat),?z:Nat)),?z:Nat)) = s(s(?z:Nat))",
"plus(?z:Nat,plus(plus(?x:Nat,?x:Nat),plus(s(s(?x:Nat)),?x:Nat))) = plus(?x:Nat,?y:Nat)",
"plus(s(?y:Nat),plus(s(?z:Nat),s(s(plus(?x:Nat,?y:Nat))))) = plus(?x:Nat,?y:Nat)",
"plus(s(s(?y:Nat)),s(plus(s(plus(?y:Nat,?x:Nat)),?z:Nat))) = s(s(?y:Nat))",
"plus(s(plus(s(?z:Nat),plus(?y:Nat,plus(?x:Nat,?x:Nat)))),?x:Nat) = plus(?x:Nat,?x:Nat)",
"plus(plus(?x:Nat,?y:Nat),plus(?y:Nat,plus(s(s(?x:Nat)),?y:Nat))) = s(s(?y:Nat))",
"plus(plus(?z:Nat,plus(plus(?y:Nat,plus(?y:Nat,?z:Nat)),?y:Nat)),?y:Nat) = plus(?x:Nat,?x:Nat)",
"plus(plus(s(plus(?x:Nat,?y:Nat)),?z:Nat),plus(?x:Nat,s(?y:Nat))) = s(s(?z:Nat))",
"plus(plus(s(plus(plus(?z:Nat,?y:Nat),?z:Nat)),s(?x:Nat)),?x:Nat) = s(s(?y:Nat))",
"plus(?x:Nat,plus(plus(?x:Nat,plus(plus(?y:Nat,?y:Nat),?y:Nat)),s(?y:Nat))) = s(?x:Nat)",
"plus(?x:Nat,plus(plus(s(plus(?x:Nat,?y:Nat)),plus(?x:Nat,?z:Nat)),?x:Nat)) = s(?z:Nat)",
"plus(s(plus(?z:Nat,s(s(?y:Nat)))),plus(plus(?z:Nat,?x:Nat),?z:Nat)) = s(?y:Nat)",
"plus(plus(?x:Nat,plus(?x:Nat,?y:Nat)),s(s(s(plus(?x:Nat,?y:Nat))))) = s(?z:Nat)",
"plus(plus(?y:Nat,?y:Nat),plus(plus(plus(s(?y:Nat),?x:Nat),?z:Nat),?x:Nat)) = s(?y:Nat)",
"plus(plus(plus(s(?y:Nat),s(?x:Nat)),plus(?x:Nat,?x:Nat)),s(?x:Nat)) = s(?x:Nat)",
"plus(plus(plus(plus(?z:Nat,?y:Nat),plus(s(?x:Nat),?z:Nat)),?y:Nat),?x:Nat) = s(?x:Nat)",
"s(?x:Nat) = plus(?y:Nat,plus(plus(?x:Nat,plus(plus(?y:Nat,?x:Nat),?x:Nat)),s(s(?x:Nat))))",
"s(?x:Nat) = plus(s(plus(?x:Nat,?x:Nat)),plus(s(plus(plus(?y:Nat,?x:Nat),?z:Nat)),?y:Nat))",
"s(?x:Nat) = plus(s(plus(s(plus(?x:Nat,?x:Nat)),s(plus(?y:Nat,?x:Nat)))),s(?y:Nat))",
"s(?x:Nat) = plus(plus(s(plus(?y:Nat,?x:Nat)),plus(s(plus(?x:Nat,?y:Nat)),?x:Nat)),?x:Nat)",
"s(?y:Nat) = plus(plus(plus(plus(?z:Nat,plus(?x:Nat,?z:Nat)),plus(?z:Nat,?x:Nat)),?x:Nat),?z:Nat)",
"s(?z:Nat) = plus(s(?x:Nat),plus(plus(s(?y:Nat),?y:Nat),plus(plus(?y:Nat,?y:Nat),?y:Nat)))",
"s(s(?x:Nat)) = plus(plus(?z:Nat,s(plus(plus(?z:Nat,?z:Nat),?x:Nat))),plus(?x:Nat,?x:Nat))",
"s(s(?x:Nat)) = plus(plus(plus(plus(s(?y:Nat),?x:Nat),s(?y:Nat)),s(?x:Nat)),?x:Nat)",
"plus(?x:Nat,?z:Nat) = plus(plus(?z:Nat,plus(plus(?z:Nat,plus(?x:Nat,?x:Nat)),s(?z:Nat))),?x:Nat)",
"plus(?y:Nat,?x:Nat) = s(s(plus(plus(s(?x:Nat),?x:Nat),plus(?x:Nat,plus(?y:Nat,?x:Nat)))))",
"plus(?z:Nat,?x:Nat) = plus(?x:Nat,plus(plus(plus(?y:Nat,?z:Nat),?x:Nat),s(plus(?x:Nat,?y:Nat))))",
"plus(?z:Nat,?x:Nat) = plus(s(?z:Nat),plus(?x:Nat,plus(plus(plus(?x:Nat,?x:Nat),?x:Nat),?z:Nat)))",
"s(plus(?y:Nat,?z:Nat)) = plus(s(s(plus(?x:Nat,s(?z:Nat)))),plus(?x:Nat,s(?x:Nat)))",
"plus(?x:Nat,s(?x:Nat)) = plus(plus(s(s(s(?y:Nat))),plus(s(?x:Nat),?x:Nat)),?y:Nat)",
"plus(?z:Nat,s(?z:Nat)) = s(plus(plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,s(?x:Nat))),?x:Nat))",
"plus(s(?y:Nat),?x:Nat) = plus(s(plus(plus(s(?x:Nat),?z:Nat),?z:Nat)),plus(?y:Nat,?y:Nat))",
"plus(s(?y:Nat),?x:Nat) = plus(plus(plus(?x:Nat,plus(plus(?x:Nat,?x:Nat),?x:Nat)),?x:Nat),?y:Nat)",
"s(s(plus(?x:Nat,?x:Nat))) = plus(plus(?z:Nat,?x:Nat),plus(?z:Nat,plus(?y:Nat,s(?y:Nat))))",
"plus(?x:Nat,plus(?x:Nat,?x:Nat)) = plus(plus(plus(?y:Nat,?y:Nat),?y:Nat),s(plus(?z:Nat,?x:Nat)))",
"plus(?x:Nat,plus(?y:Nat,?x:Nat)) = plus(plus(s(plus(?y:Nat,?y:Nat)),?x:Nat),plus(?y:Nat,?z:Nat))",
"plus(?z:Nat,plus(?y:Nat,?z:Nat)) = plus(plus(?y:Nat,plus(?x:Nat,?y:Nat)),s(s(s(?x:Nat))))",
"plus(s(s(?y:Nat)),?y:Nat) = plus(s(plus(?y:Nat,plus(plus(?x:Nat,?x:Nat),?x:Nat))),?z:Nat)",
"plus(plus(?y:Nat,?y:Nat),?x:Nat) = plus(plus(plus(?z:Nat,s(plus(?z:Nat,?z:Nat))),?x:Nat),?x:Nat)",
"plus(plus(?y:Nat,?y:Nat),?y:Nat) = plus(s(?z:Nat),plus(?x:Nat,plus(?z:Nat,plus(?y:Nat,?y:Nat))))",
"plus(?x:Nat,s(plus(?x:Nat,?y:Nat))) = plus(plus(?z:Nat,plus(?z:Nat,plus(?y:Nat,?x:Nat))),?x:Nat)",
"plus(?x:Nat,s(plus(?y:Nat,?y:Nat))) = s(plus(s(s(?z:Nat)),plus(s(?x:Nat),?x:Nat)))",
"plus(?x:Nat,s(plus(?y:Nat,?z:Nat))) = s(s(plus(s(plus(?x:Nat,?x:Nat)),s(?z:Nat))))",
"plus(plus(?y:Nat,s(?z:Nat)),?z:Nat) = plus(plus(plus(?y:Nat,?z:Nat),?y:Nat),plus(?y:Nat,?z:Nat))",
"plus(plus(plus(?x:Nat,?x:Nat),?x:Nat),?z:Nat) = plus(?z:Nat,plus(?x:Nat,plus(s(?x:Nat),?z:Nat)))",
"s(s(plus(s(plus(?y:Nat,?x:Nat)),?y:Nat))) = plus(?x:Nat,plus(plus(?x:Nat,?y:Nat),?y:Nat))",
"s(plus(plus(plus(?y:Nat,?y:Nat),?z:Nat),?x:Nat)) = plus(plus(s(s(?x:Nat)),?z:Nat),?x:Nat)",
"plus(plus(?z:Nat,s(s(s(?y:Nat)))),?x:Nat) = s(s(plus(?z:Nat,plus(?z:Nat,?x:Nat))))",
"plus(plus(plus(s(?x:Nat),?x:Nat),?z:Nat),?y:Nat) = plus(plus(?x:Nat,?z:Nat),plus(?z:Nat,?y:Nat))",
"plus(?x:Nat,plus(?y:Nat,plus(s(s(?z:Nat)),?y:Nat))) = plus(s(s(?x:Nat)),s(?x:Nat))",
"plus(plus(?y:Nat,plus(plus(?y:Nat,?x:Nat),?x:Nat)),?x:Nat) = s(plus(?z:Nat,plus(?y:Nat,?y:Nat)))",
"plus(plus(s(?y:Nat),s(s(s(?x:Nat)))),?y:Nat) = plus(?y:Nat,plus(?y:Nat,s(?x:Nat)))",
"plus(plus(plus(s(s(?x:Nat)),?y:Nat),?x:Nat),?z:Nat) = plus(?y:Nat,s(plus(?x:Nat,?y:Nat)))",
"s(plus(?y:Nat,plus(plus(s(?z:Nat),?x:Nat),s(?y:Nat)))) = plus(?y:Nat,s(s(?x:Nat)))",
"s(plus(s(?y:Nat),plus(s(plus(?x:Nat,?y:Nat)),?y:Nat))) = plus(?x:Nat,plus(?x:Nat,?y:Nat))",
"s(plus(s(s(s(?x:Nat))),plus(?x:Nat,s(?x:Nat)))) = plus(?y:Nat,plus(?y:Nat,?z:Nat))",
"plus(s(?y:Nat),plus(?y:Nat,s(s(plus(?x:Nat,?x:Nat))))) = s(plus(?x:Nat,s(?x:Nat)))",
"plus(s(s(s(plus(?y:Nat,?x:Nat)))),plus(?y:Nat,?y:Nat)) = plus(?x:Nat,plus(?y:Nat,?y:Nat))",
"plus(plus(?z:Nat,s(?x:Nat)),plus(s(?z:Nat),s(?y:Nat))) = s(plus(s(?x:Nat),?y:Nat))",
"plus(plus(plus(?z:Nat,?z:Nat),s(?z:Nat)),s(s(?z:Nat))) = s(plus(s(?x:Nat),?y:Nat))",
"plus(plus(plus(plus(?x:Nat,s(?x:Nat)),?x:Nat),?y:Nat),?y:Nat) = plus(?y:Nat,plus(?y:Nat,?z:Nat))",
"s(plus(s(plus(?x:Nat,?z:Nat)),s(s(plus(?x:Nat,?y:Nat))))) = plus(?x:Nat,s(?z:Nat))",
"plus(?x:Nat,plus(plus(?y:Nat,plus(s(?z:Nat),s(?y:Nat))),?x:Nat)) = plus(s(?x:Nat),?z:Nat)",
"plus(s(s(plus(?x:Nat,plus(?x:Nat,?y:Nat)))),plus(?y:Nat,?x:Nat)) = s(plus(?x:Nat,?y:Nat))",
"plus(plus(plus(plus(s(?x:Nat),?y:Nat),?x:Nat),?z:Nat),s(?x:Nat)) = plus(?x:Nat,s(?z:Nat))",
"s(plus(plus(?x:Nat,s(s(?z:Nat))),plus(?y:Nat,plus(?y:Nat,?z:Nat)))) = plus(?y:Nat,?y:Nat)",
"s(plus(plus(?x:Nat,plus(plus(?z:Nat,?x:Nat),plus(?y:Nat,?y:Nat))),?x:Nat)) = plus(?x:Nat,?x:Nat)",
"plus(?y:Nat,plus(plus(?y:Nat,plus(?y:Nat,s(?x:Nat))),s(s(?x:Nat)))) = plus(?x:Nat,?z:Nat)",
"plus(?y:Nat,plus(plus(plus(plus(?x:Nat,?x:Nat),s(?y:Nat)),?z:Nat),?z:Nat)) = plus(?x:Nat,?x:Nat)",
"plus(?z:Nat,s(plus(?z:Nat,plus(plus(?y:Nat,?y:Nat),plus(?x:Nat,?x:Nat))))) = plus(?y:Nat,?y:Nat)",
"plus(s(plus(plus(s(?x:Nat),s(?x:Nat)),s(s(?y:Nat)))),?y:Nat) = plus(?x:Nat,?x:Nat)",
"plus(plus(?x:Nat,s(?x:Nat)),s(s(plus(?x:Nat,plus(?z:Nat,?z:Nat))))) = s(s(?x:Nat))",
"plus(plus(?x:Nat,s(plus(plus(?z:Nat,?z:Nat),plus(?y:Nat,?z:Nat)))),?z:Nat) = plus(?x:Nat,?y:Nat)",
"plus(plus(?z:Nat,?z:Nat),s(plus(plus(?z:Nat,plus(?x:Nat,?y:Nat)),?y:Nat))) = plus(?x:Nat,?x:Nat)",
"plus(plus(s(plus(?z:Nat,s(?y:Nat))),plus(?y:Nat,?x:Nat)),s(?x:Nat)) = s(s(?x:Nat))",
"plus(plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?z:Nat)),plus(s(?z:Nat),?x:Nat)) = plus(?x:Nat,?y:Nat)",
"plus(plus(plus(?x:Nat,plus(plus(?x:Nat,?x:Nat),s(?x:Nat))),?x:Nat),?y:Nat) = plus(?y:Nat,?x:Nat)",
"plus(plus(plus(?y:Nat,plus(?y:Nat,plus(?x:Nat,?y:Nat))),s(?y:Nat)),?y:Nat) = plus(?z:Nat,?x:Nat)",
"plus(plus(plus(plus(?z:Nat,plus(?x:Nat,?x:Nat)),?x:Nat),?y:Nat),s(?x:Nat)) = s(s(?z:Nat))",
"plus(plus(plus(plus(s(?z:Nat),?y:Nat),?x:Nat),plus(?z:Nat,?z:Nat)),?x:Nat) = plus(?x:Nat,?x:Nat)",
"plus(s(plus(plus(?x:Nat,?y:Nat),plus(?x:Nat,?x:Nat))),plus(?x:Nat,s(?y:Nat))) = s(?y:Nat)",
"plus(plus(?x:Nat,?x:Nat),plus(plus(plus(?x:Nat,plus(?x:Nat,?z:Nat)),?x:Nat),?z:Nat)) = s(?y:Nat)",
"plus(plus(plus(s(?x:Nat),plus(?y:Nat,?x:Nat)),s(?x:Nat)),plus(?x:Nat,?x:Nat)) = s(?x:Nat)",
"s(?x:Nat) = plus(plus(?x:Nat,s(?x:Nat)),plus(s(?y:Nat),plus(s(s(?y:Nat)),s(?x:Nat))))",
"s(?y:Nat) = plus(?x:Nat,plus(plus(s(s(?z:Nat)),plus(plus(?z:Nat,?x:Nat),s(?z:Nat))),?x:Nat))",
"s(s(?z:Nat)) = plus(?x:Nat,plus(s(?x:Nat),plus(plus(plus(?y:Nat,?z:Nat),s(?y:Nat)),?y:Nat)))",
"plus(?x:Nat,?x:Nat) = s(plus(?z:Nat,plus(plus(s(?z:Nat),plus(?y:Nat,?z:Nat)),plus(?y:Nat,?x:Nat))))",
"plus(?x:Nat,?x:Nat) = plus(?z:Nat,plus(plus(plus(?y:Nat,?z:Nat),?z:Nat),plus(?x:Nat,plus(?y:Nat,?x:Nat))))",
"plus(?y:Nat,?x:Nat) = s(s(plus(plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?y:Nat)),plus(?z:Nat,?y:Nat))))",
"plus(?y:Nat,?x:Nat) = s(plus(plus(plus(plus(?z:Nat,?z:Nat),?y:Nat),?z:Nat),s(s(s(?x:Nat)))))",
"plus(?y:Nat,?x:Nat) = plus(s(plus(s(plus(?x:Nat,?z:Nat)),plus(?x:Nat,?y:Nat))),plus(?x:Nat,?y:Nat))",
"plus(?y:Nat,?x:Nat) = plus(plus(plus(?z:Nat,?x:Nat),plus(plus(?x:Nat,?z:Nat),?y:Nat)),plus(?z:Nat,?x:Nat))",
"plus(?y:Nat,?z:Nat) = plus(s(plus(plus(s(?x:Nat),s(?z:Nat)),plus(s(?y:Nat),?x:Nat))),?y:Nat)",
"plus(?x:Nat,s(?z:Nat)) = s(plus(s(?y:Nat),s(plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,?x:Nat)))))",
"plus(?z:Nat,s(?x:Nat)) = plus(s(plus(plus(?x:Nat,?x:Nat),plus(plus(?y:Nat,?y:Nat),?z:Nat))),?y:Nat)",
"plus(s(?x:Nat),?x:Nat) = s(plus(plus(?x:Nat,s(s(?y:Nat))),plus(plus(?x:Nat,?z:Nat),?z:Nat)))",
"s(s(s(s(?y:Nat)))) = plus(plus(plus(?z:Nat,?x:Nat),plus(s(?x:Nat),?y:Nat)),s(?x:Nat))",
"s(plus(?z:Nat,s(?z:Nat))) = plus(plus(?x:Nat,?y:Nat),s(plus(plus(s(?x:Nat),?x:Nat),?z:Nat)))",
"s(plus(s(?x:Nat),?y:Nat)) = plus(?x:Nat,plus(?x:Nat,plus(?x:Nat,plus(plus(?y:Nat,?y:Nat),?y:Nat))))",
"plus(?x:Nat,s(s(?x:Nat))) = plus(plus(?z:Nat,s(plus(?z:Nat,plus(?y:Nat,?x:Nat)))),s(?z:Nat))",
"plus(?y:Nat,plus(?y:Nat,?z:Nat)) = s(plus(plus(plus(?z:Nat,?x:Nat),s(plus(?z:Nat,?x:Nat))),?x:Nat))",
"plus(?z:Nat,plus(?z:Nat,?y:Nat)) = plus(plus(?y:Nat,?x:Nat),plus(s(s(?z:Nat)),plus(?x:Nat,?y:Nat)))",
"plus(plus(?x:Nat,?x:Nat),?y:Nat) = plus(plus(plus(?y:Nat,?x:Nat),s(plus(?x:Nat,s(?z:Nat)))),?y:Nat)",
"plus(?y:Nat,s(plus(?x:Nat,?x:Nat))) = plus(plus(s(?y:Nat),plus(?x:Nat,plus(?x:Nat,?y:Nat))),?z:Nat)",
"plus(?y:Nat,s(plus(?y:Nat,?z:Nat))) = plus(plus(s(plus(plus(?z:Nat,?y:Nat),?z:Nat)),?y:Nat),?x:Nat)",
"plus(s(s(?x:Nat)),s(?z:Nat)) = s(plus(plus(plus(?x:Nat,?y:Nat),?y:Nat),plus(?x:Nat,?x:Nat)))",
"plus(s(s(?y:Nat)),s(?y:Nat)) = s(plus(s(?z:Nat),plus(?z:Nat,s(plus(?x:Nat,?x:Nat)))))",
"plus(plus(?z:Nat,s(?x:Nat)),?x:Nat) = plus(s(plus(plus(plus(?z:Nat,?x:Nat),?y:Nat),?x:Nat)),?x:Nat)",
"plus(?z:Nat,s(plus(s(?x:Nat),?x:Nat))) = s(plus(?y:Nat,plus(?y:Nat,s(plus(?z:Nat,?z:Nat)))))",
"plus(plus(?z:Nat,?x:Nat),plus(?z:Nat,?x:Nat)) = plus(plus(?z:Nat,s(s(plus(?y:Nat,?x:Nat)))),?z:Nat)",
"plus(plus(?z:Nat,plus(?z:Nat,?z:Nat)),?y:Nat) = s(plus(plus(?x:Nat,plus(s(?x:Nat),?x:Nat)),?x:Nat))",
"plus(plus(plus(?x:Nat,?x:Nat),?y:Nat),?y:Nat) = s(s(plus(plus(?y:Nat,s(?x:Nat)),s(?x:Nat))))",
"plus(plus(plus(?y:Nat,?x:Nat),?x:Nat),?x:Nat) = plus(?z:Nat,plus(?x:Nat,s(plus(?x:Nat,s(?y:Nat)))))",
"plus(?x:Nat,plus(?y:Nat,plus(s(?x:Nat),?x:Nat))) = s(s(plus(plus(?y:Nat,?x:Nat),s(?y:Nat))))",
"plus(plus(?y:Nat,?y:Nat),plus(s(?z:Nat),?z:Nat)) = s(plus(plus(plus(?y:Nat,?x:Nat),?z:Nat),?y:Nat))",
"plus(plus(s(s(?x:Nat)),?z:Nat),s(?z:Nat)) = plus(plus(?x:Nat,?y:Nat),s(plus(?x:Nat,?y:Nat)))",
"s(plus(plus(?z:Nat,s(plus(?x:Nat,?z:Nat))),?z:Nat)) = plus(plus(?x:Nat,plus(?x:Nat,?y:Nat)),?y:Nat)",
"plus(s(plus(?x:Nat,s(plus(?z:Nat,?y:Nat)))),?x:Nat) = plus(s(s(s(s(?z:Nat)))),?x:Nat)",
"plus(plus(plus(plus(?y:Nat,?y:Nat),?z:Nat),?z:Nat),?x:Nat) = plus(s(s(?y:Nat)),s(s(?y:Nat)))",
"plus(?y:Nat,plus(?x:Nat,s(plus(?y:Nat,plus(?x:Nat,?x:Nat))))) = plus(s(?x:Nat),plus(?x:Nat,?x:Nat))",
"plus(s(plus(?z:Nat,s(plus(?x:Nat,?y:Nat)))),s(?x:Nat)) = plus(plus(s(?z:Nat),?y:Nat),?x:Nat)",
"plus(s(plus(plus(?z:Nat,plus(?x:Nat,?z:Nat)),?y:Nat)),?x:Nat) = plus(plus(s(?x:Nat),?z:Nat),?y:Nat)",
"plus(plus(s(s(?x:Nat)),?z:Nat),plus(?y:Nat,s(?y:Nat))) = plus(plus(s(?x:Nat),?y:Nat),?y:Nat)",
"s(plus(plus(plus(?x:Nat,s(?z:Nat)),?y:Nat),plus(?x:Nat,?x:Nat))) = plus(s(?x:Nat),s(?x:Nat))",
"plus(?x:Nat,s(plus(plus(?x:Nat,?z:Nat),plus(?z:Nat,s(?y:Nat))))) = plus(?z:Nat,s(s(?x:Nat)))",
"plus(s(plus(?y:Nat,plus(s(?y:Nat),plus(?y:Nat,?y:Nat)))),?y:Nat) = plus(plus(?x:Nat,?x:Nat),?x:Nat)",
"plus(plus(?x:Nat,s(plus(?x:Nat,plus(?x:Nat,?x:Nat)))),s(?y:Nat)) = plus(plus(?x:Nat,?z:Nat),?z:Nat)",
"plus(plus(s(?x:Nat),plus(plus(?x:Nat,?y:Nat),?x:Nat)),s(?z:Nat)) = s(plus(s(?z:Nat),?x:Nat))",
"plus(plus(s(plus(?x:Nat,?x:Nat)),?x:Nat),s(plus(?y:Nat,?x:Nat))) = plus(s(?y:Nat),s(?z:Nat))",
"plus(plus(s(plus(?z:Nat,s(?y:Nat))),?x:Nat),s(s(?y:Nat))) = plus(plus(?z:Nat,?z:Nat),?z:Nat)",
"s(s(plus(plus(plus(?y:Nat,?x:Nat),plus(?y:Nat,?x:Nat)),s(?x:Nat)))) = plus(?y:Nat,s(?x:Nat))",
"plus(?z:Nat,plus(plus(plus(s(?x:Nat),?x:Nat),?x:Nat),plus(?y:Nat,?x:Nat))) = plus(?z:Nat,s(?x:Nat))",
"s(plus(plus(?x:Nat,s(?z:Nat)),plus(?x:Nat,plus(?x:Nat,plus(?y:Nat,?y:Nat))))) = plus(?z:Nat,?x:Nat)",
"s(plus(plus(plus(plus(?y:Nat,?z:Nat),?x:Nat),s(plus(?x:Nat,?x:Nat))),?z:Nat)) = plus(?z:Nat,?x:Nat)",
"plus(?x:Nat,plus(plus(plus(plus(?x:Nat,?z:Nat),?x:Nat),plus(?y:Nat,?z:Nat)),?z:Nat)) = plus(?z:Nat,?z:Nat)",
"plus(s(?z:Nat),plus(plus(plus(plus(?y:Nat,?z:Nat),?x:Nat),s(?y:Nat)),?x:Nat)) = plus(?y:Nat,?y:Nat)",
"plus(plus(?x:Nat,?z:Nat),plus(plus(?y:Nat,?z:Nat),plus(s(s(?z:Nat)),?z:Nat))) = plus(?x:Nat,?z:Nat)",
"plus(plus(?y:Nat,?x:Nat),s(plus(plus(plus(?x:Nat,?y:Nat),?y:Nat),s(?x:Nat)))) = plus(?x:Nat,?x:Nat)",
"plus(plus(?z:Nat,?y:Nat),plus(plus(plus(plus(?x:Nat,?y:Nat),?x:Nat),?z:Nat),?x:Nat)) = s(s(?x:Nat))",
"plus(s(plus(s(s(?z:Nat)),s(s(?x:Nat)))),plus(?z:Nat,s(s(?y:Nat)))) = s(?z:Nat)",
"plus(plus(s(s(?x:Nat)),plus(plus(s(?y:Nat),plus(?y:Nat,?x:Nat)),?x:Nat)),?y:Nat) = s(?x:Nat)",
"s(?y:Nat) = plus(plus(s(?x:Nat),plus(plus(?y:Nat,plus(?y:Nat,?z:Nat)),s(plus(?z:Nat,?y:Nat)))),?x:Nat)",
"s(?y:Nat) = plus(plus(plus(plus(?x:Nat,?x:Nat),?z:Nat),s(?y:Nat)),plus(s(s(s(?y:Nat))),?x:Nat))",
"s(s(?y:Nat)) = plus(plus(?z:Nat,s(s(?y:Nat))),s(plus(s(?z:Nat),s(plus(?z:Nat,?y:Nat)))))",
"plus(?y:Nat,?y:Nat) = plus(s(s(plus(s(?y:Nat),?x:Nat))),plus(s(s(?x:Nat)),plus(?x:Nat,?y:Nat)))",
"plus(?y:Nat,?y:Nat) = plus(s(plus(?z:Nat,plus(?y:Nat,?z:Nat))),s(plus(?y:Nat,plus(s(?y:Nat),?y:Nat))))",
"plus(?y:Nat,?z:Nat) = plus(plus(plus(?y:Nat,plus(?z:Nat,?y:Nat)),plus(?x:Nat,s(plus(?y:Nat,?x:Nat)))),?x:Nat)",
"plus(?z:Nat,?x:Nat) = plus(s(plus(s(plus(?x:Nat,?x:Nat)),?x:Nat)),plus(plus(?y:Nat,s(?z:Nat)),?x:Nat))",
"plus(?z:Nat,?y:Nat) = plus(plus(?y:Nat,plus(?z:Nat,plus(plus(?x:Nat,?x:Nat),?x:Nat))),plus(s(?y:Nat),?y:Nat))",
"plus(?y:Nat,s(?z:Nat)) = plus(plus(plus(s(s(?y:Nat)),plus(plus(?x:Nat,?x:Nat),?x:Nat)),?x:Nat),?x:Nat)",
"plus(?x:Nat,s(s(?z:Nat))) = plus(plus(plus(plus(plus(?x:Nat,?y:Nat),s(?x:Nat)),?x:Nat),?y:Nat),?y:Nat)",
"plus(plus(?y:Nat,?x:Nat),?y:Nat) = plus(?x:Nat,plus(s(plus(?y:Nat,plus(?x:Nat,?z:Nat))),plus(?y:Nat,?x:Nat)))",
"plus(?x:Nat,plus(s(?y:Nat),?y:Nat)) = plus(?x:Nat,plus(s(plus(?z:Nat,?x:Nat)),s(plus(?x:Nat,?z:Nat))))",
"s(plus(plus(s(?y:Nat),?z:Nat),?x:Nat)) = plus(?x:Nat,plus(?y:Nat,plus(plus(s(?y:Nat),?y:Nat),?x:Nat)))",
"plus(s(s(?x:Nat)),plus(?z:Nat,?z:Nat)) = plus(?z:Nat,s(s(plus(plus(?x:Nat,?x:Nat),s(?x:Nat)))))",
"plus(s(plus(s(?x:Nat),?z:Nat)),?x:Nat) = plus(s(plus(s(s(?x:Nat)),plus(?z:Nat,?x:Nat))),?x:Nat)",
"plus(plus(?x:Nat,s(?z:Nat)),s(?y:Nat)) = plus(plus(s(s(?z:Nat)),plus(?x:Nat,?x:Nat)),s(?z:Nat))",
"plus(?x:Nat,s(plus(?y:Nat,s(s(?z:Nat))))) = plus(s(plus(s(?y:Nat),?z:Nat)),plus(?x:Nat,?y:Nat))",
"plus(s(?y:Nat),plus(?x:Nat,plus(?y:Nat,?y:Nat))) = plus(plus(s(plus(?y:Nat,?x:Nat)),s(?y:Nat)),?x:Nat)",
"s(s(plus(plus(s(?x:Nat),s(?y:Nat)),?x:Nat))) = plus(plus(?x:Nat,?y:Nat),plus(?y:Nat,s(?x:Nat)))",
"plus(s(s(plus(plus(?z:Nat,?y:Nat),?x:Nat))),?x:Nat) = plus(s(?y:Nat),plus(s(s(?z:Nat)),?y:Nat))",
"plus(s(plus(?x:Nat,?y:Nat)),plus(s(?x:Nat),?x:Nat)) = plus(plus(?y:Nat,?z:Nat),s(plus(?z:Nat,?z:Nat)))",
"plus(plus(plus(s(?x:Nat),?z:Nat),?x:Nat),s(?x:Nat)) = plus(plus(?x:Nat,?y:Nat),plus(s(?x:Nat),?x:Nat))",
"s(plus(plus(plus(s(?z:Nat),?y:Nat),?y:Nat),s(?z:Nat))) = plus(s(s(?y:Nat)),plus(?x:Nat,?z:Nat))",
"plus(?x:Nat,plus(plus(?x:Nat,s(s(?y:Nat))),s(?y:Nat))) = plus(plus(?x:Nat,?y:Nat),plus(?z:Nat,?y:Nat))",
"plus(?y:Nat,s(s(plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat))))) = s(plus(?y:Nat,plus(?z:Nat,?x:Nat)))",
"plus(plus(?y:Nat,plus(plus(plus(?x:Nat,?x:Nat),?x:Nat),?z:Nat)),?x:Nat) = s(plus(plus(?x:Nat,?x:Nat),?x:Nat))",
"plus(plus(?z:Nat,?z:Nat),plus(?x:Nat,s(plus(?x:Nat,s(?y:Nat))))) = s(s(plus(?x:Nat,s(?z:Nat))))",
"plus(?x:Nat,plus(plus(?y:Nat,plus(s(?y:Nat),plus(?y:Nat,?y:Nat))),?x:Nat)) = plus(plus(?y:Nat,?x:Nat),?y:Nat)",
"plus(plus(s(plus(s(?x:Nat),plus(?x:Nat,?z:Nat))),?x:Nat),s(?x:Nat)) = s(s(plus(?x:Nat,?x:Nat)))",
"plus(s(plus(?y:Nat,plus(plus(?y:Nat,?y:Nat),?y:Nat))),plus(?y:Nat,s(?x:Nat))) = s(plus(?x:Nat,?x:Nat))",
"plus(plus(plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,plus(?x:Nat,?z:Nat))),?z:Nat),?x:Nat) = plus(s(?x:Nat),?z:Nat)",
"s(plus(plus(?z:Nat,plus(s(?x:Nat),plus(?z:Nat,?y:Nat))),s(plus(?y:Nat,?z:Nat)))) = plus(?y:Nat,?z:Nat)",
"plus(?x:Nat,plus(plus(s(s(?y:Nat)),?x:Nat),plus(s(?x:Nat),s(s(?x:Nat))))) = s(s(?y:Nat))",
"plus(s(plus(?y:Nat,?y:Nat)),s(plus(s(s(?y:Nat)),plus(s(?y:Nat),?x:Nat)))) = plus(?y:Nat,?x:Nat)",
"plus(plus(plus(s(plus(?y:Nat,?x:Nat)),plus(?x:Nat,?z:Nat)),?y:Nat),plus(?z:Nat,?x:Nat)) = s(s(?y:Nat))",
"plus(s(plus(?x:Nat,?y:Nat)),plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),s(s(s(?x:Nat))))) = s(?z:Nat)",
"plus(s(plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,plus(?x:Nat,?x:Nat)))),s(s(s(?x:Nat)))) = s(?x:Nat)",
"plus(plus(s(?x:Nat),?y:Nat),plus(plus(plus(s(?x:Nat),?z:Nat),s(s(?x:Nat))),?z:Nat)) = s(?x:Nat)",
"plus(plus(s(?x:Nat),plus(plus(plus(?y:Nat,?x:Nat),s(?z:Nat)),plus(?x:Nat,?x:Nat))),?x:Nat) = s(?y:Nat)",
"plus(plus(plus(?x:Nat,?y:Nat),s(?y:Nat)),plus(?x:Nat,plus(?y:Nat,plus(?x:Nat,s(?x:Nat))))) = s(?x:Nat)",
"s(?x:Nat) = plus(?z:Nat,plus(plus(s(plus(?z:Nat,?x:Nat)),plus(?x:Nat,plus(?y:Nat,?y:Nat))),plus(?x:Nat,?x:Nat)))",
"s(?x:Nat) = plus(s(plus(?x:Nat,plus(s(?y:Nat),?x:Nat))),plus(?y:Nat,plus(plus(?x:Nat,s(?x:Nat)),?y:Nat)))",
"s(?x:Nat) = plus(s(plus(plus(s(?x:Nat),plus(?y:Nat,?x:Nat)),plus(?x:Nat,plus(?y:Nat,?x:Nat)))),s(?x:Nat))",
"s(?y:Nat) = plus(?y:Nat,plus(plus(plus(plus(?y:Nat,?z:Nat),?x:Nat),?z:Nat),plus(s(plus(?x:Nat,?x:Nat)),?x:Nat)))",
"plus(?x:Nat,?z:Nat) = plus(plus(plus(s(plus(?z:Nat,?x:Nat)),plus(?y:Nat,s(?z:Nat))),plus(?y:Nat,?x:Nat)),?x:Nat)",
"s(plus(?x:Nat,?x:Nat)) = plus(?x:Nat,plus(plus(plus(?y:Nat,?y:Nat),plus(plus(?x:Nat,?x:Nat),s(?y:Nat))),?y:Nat))",
"s(plus(?z:Nat,?y:Nat)) = plus(?x:Nat,plus(plus(?y:Nat,s(?z:Nat)),plus(plus(plus(?z:Nat,?x:Nat),?x:Nat),?z:Nat)))",
"plus(?x:Nat,s(?x:Nat)) = plus(?y:Nat,plus(plus(s(?y:Nat),?y:Nat),plus(plus(?y:Nat,plus(?z:Nat,?z:Nat)),?x:Nat)))",
"plus(s(?x:Nat),?z:Nat) = plus(plus(plus(plus(?y:Nat,?y:Nat),?z:Nat),?x:Nat),plus(?y:Nat,s(plus(?x:Nat,?y:Nat))))",
"plus(?x:Nat,plus(?x:Nat,?y:Nat)) = s(plus(s(s(?y:Nat)),s(plus(plus(?x:Nat,?z:Nat),plus(?x:Nat,?y:Nat)))))",
"plus(plus(?z:Nat,?x:Nat),?y:Nat) = plus(plus(plus(s(?x:Nat),?x:Nat),?x:Nat),plus(?x:Nat,plus(s(?x:Nat),?x:Nat)))",
"s(plus(s(s(?z:Nat)),?x:Nat)) = plus(?z:Nat,plus(s(?x:Nat),plus(plus(?y:Nat,plus(?z:Nat,?x:Nat)),?y:Nat)))",
"plus(?x:Nat,plus(plus(?x:Nat,?y:Nat),?x:Nat)) = plus(plus(?x:Nat,plus(s(s(?y:Nat)),plus(?y:Nat,?x:Nat))),?y:Nat)",
"plus(?z:Nat,plus(s(s(s(?y:Nat))),?x:Nat)) = plus(plus(?x:Nat,plus(s(plus(?z:Nat,?y:Nat)),?z:Nat)),?x:Nat)",
"plus(s(plus(?x:Nat,s(s(?z:Nat)))),?x:Nat) = plus(s(plus(?z:Nat,plus(?z:Nat,plus(?x:Nat,?x:Nat)))),?y:Nat)",
"plus(plus(plus(?z:Nat,?x:Nat),s(?x:Nat)),?y:Nat) = s(plus(plus(s(?y:Nat),s(plus(?z:Nat,?x:Nat))),?z:Nat))",
"s(plus(?y:Nat,plus(s(plus(?y:Nat,?y:Nat)),?x:Nat))) = plus(plus(?x:Nat,plus(?x:Nat,?z:Nat)),s(s(?y:Nat)))",
"plus(s(?z:Nat),s(plus(?x:Nat,s(s(?y:Nat))))) = plus(plus(?x:Nat,plus(?x:Nat,?y:Nat)),plus(?z:Nat,?x:Nat))",
"plus(s(?z:Nat),plus(plus(?x:Nat,?x:Nat),s(?x:Nat))) = s(plus(plus(s(plus(?y:Nat,?y:Nat)),?y:Nat),?y:Nat))",
"plus(plus(?y:Nat,plus(s(s(?x:Nat)),?z:Nat)),?z:Nat) = plus(plus(?x:Nat,?y:Nat),plus(s(s(?z:Nat)),?y:Nat))",
"plus(?x:Nat,s(s(plus(plus(?y:Nat,?z:Nat),s(?z:Nat))))) = s(plus(s(s(plus(?x:Nat,?x:Nat))),?x:Nat))",
"plus(s(plus(plus(s(?x:Nat),s(?x:Nat)),?y:Nat)),?z:Nat) = plus(s(plus(?y:Nat,?x:Nat)),plus(?y:Nat,?y:Nat))",
"plus(plus(s(?y:Nat),s(s(plus(?z:Nat,?z:Nat)))),?x:Nat) = plus(plus(plus(?z:Nat,?y:Nat),?y:Nat),s(?x:Nat))",
"plus(plus(s(s(plus(?x:Nat,?x:Nat))),s(?y:Nat)),?x:Nat) = plus(?z:Nat,plus(s(plus(?y:Nat,?y:Nat)),?z:Nat))",
"plus(?y:Nat,plus(plus(?x:Nat,plus(?x:Nat,plus(?z:Nat,?x:Nat))),?x:Nat)) = plus(plus(?y:Nat,?z:Nat),s(s(?z:Nat)))",
"plus(s(plus(plus(s(?y:Nat),plus(?y:Nat,?x:Nat)),?y:Nat)),?y:Nat) = plus(?x:Nat,plus(?x:Nat,s(s(?y:Nat))))",
"plus(s(plus(s(plus(?y:Nat,?y:Nat)),?x:Nat)),plus(s(?z:Nat),?x:Nat)) = plus(s(plus(?z:Nat,?x:Nat)),?z:Nat)",
"plus(plus(s(?x:Nat),s(s(?x:Nat))),plus(?z:Nat,plus(s(?y:Nat),?x:Nat))) = plus(s(?x:Nat),s(?y:Nat))",
"plus(plus(plus(?x:Nat,?y:Nat),s(plus(?y:Nat,s(?y:Nat)))),plus(?x:Nat,?x:Nat)) = plus(?y:Nat,s(s(?z:Nat)))",
"plus(?z:Nat,s(plus(s(s(?z:Nat)),plus(plus(?x:Nat,?x:Nat),plus(?z:Nat,?x:Nat))))) = s(plus(?y:Nat,?y:Nat))",
"plus(s(?y:Nat),plus(plus(plus(plus(?y:Nat,?y:Nat),plus(?x:Nat,?x:Nat)),?x:Nat),?z:Nat)) = s(plus(?y:Nat,?z:Nat))",
"plus(plus(plus(s(?z:Nat),s(s(?y:Nat))),s(s(plus(?y:Nat,?z:Nat)))),?x:Nat) = plus(s(?x:Nat),?x:Nat)",
"s(plus(s(?x:Nat),plus(plus(?x:Nat,s(?z:Nat)),plus(s(?z:Nat),plus(?x:Nat,?x:Nat))))) = plus(?x:Nat,?x:Nat)",
"plus(?x:Nat,plus(plus(plus(s(?x:Nat),?y:Nat),plus(plus(?z:Nat,?x:Nat),s(?y:Nat))),?x:Nat)) = s(s(?y:Nat))",
"plus(s(?x:Nat),s(plus(plus(plus(?x:Nat,?y:Nat),?z:Nat),plus(plus(?y:Nat,?y:Nat),?z:Nat)))) = s(s(?x:Nat))",
"plus(plus(?x:Nat,plus(plus(plus(?x:Nat,?z:Nat),s(?z:Nat)),plus(?x:Nat,?z:Nat))),s(?y:Nat)) = plus(?x:Nat,?y:Nat)",
"plus(s(?y:Nat),plus(plus(?x:Nat,plus(plus(?y:Nat,?x:Nat),?x:Nat)),s(s(plus(?y:Nat,?z:Nat))))) = s(?z:Nat)",
"plus(plus(?x:Nat,plus(?x:Nat,s(?y:Nat))),plus(?y:Nat,plus(plus(?y:Nat,plus(?x:Nat,?z:Nat)),?x:Nat))) = s(?x:Nat)",
"plus(plus(plus(?y:Nat,?z:Nat),plus(plus(s(?z:Nat),?z:Nat),plus(plus(?z:Nat,?z:Nat),?x:Nat))),?z:Nat) = s(?x:Nat)",
"s(?y:Nat) = plus(plus(?x:Nat,plus(?y:Nat,s(?z:Nat))),s(plus(?y:Nat,plus(plus(?y:Nat,?z:Nat),plus(?z:Nat,?z:Nat)))))",
"s(?y:Nat) = plus(plus(s(plus(?z:Nat,s(?x:Nat))),?y:Nat),s(plus(plus(?x:Nat,?x:Nat),s(plus(?x:Nat,?x:Nat)))))",
"s(s(?z:Nat)) = plus(plus(?z:Nat,?z:Nat),plus(?z:Nat,plus(plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,?y:Nat)),s(?y:Nat))))",
"plus(?z:Nat,s(?x:Nat)) = s(plus(s(plus(plus(?z:Nat,?x:Nat),?x:Nat)),plus(?z:Nat,plus(plus(?y:Nat,?z:Nat),?z:Nat))))",
"s(plus(s(?z:Nat),?z:Nat)) = plus(?y:Nat,plus(plus(plus(?x:Nat,?z:Nat),?z:Nat),plus(?x:Nat,s(plus(?z:Nat,?x:Nat)))))",
"plus(s(?x:Nat),plus(?x:Nat,?y:Nat)) = s(plus(?y:Nat,plus(plus(s(?x:Nat),s(?y:Nat)),plus(s(?x:Nat),?y:Nat))))",
"plus(plus(?x:Nat,?z:Nat),s(?y:Nat)) = plus(plus(plus(plus(plus(?x:Nat,?x:Nat),?x:Nat),s(s(?y:Nat))),?x:Nat),?y:Nat)",
"plus(plus(s(?z:Nat),?z:Nat),?x:Nat) = plus(plus(plus(s(plus(?x:Nat,?y:Nat)),s(?z:Nat)),s(?y:Nat)),s(?x:Nat))",
"s(s(s(plus(?x:Nat,s(?z:Nat))))) = plus(plus(s(?y:Nat),s(s(plus(?z:Nat,?y:Nat)))),plus(?x:Nat,?y:Nat))",
"plus(?x:Nat,s(s(plus(?x:Nat,?x:Nat)))) = plus(plus(?x:Nat,?z:Nat),s(plus(?x:Nat,plus(?z:Nat,plus(?y:Nat,?x:Nat)))))",
"s(s(s(plus(?x:Nat,plus(?x:Nat,?y:Nat))))) = plus(plus(plus(plus(?x:Nat,?x:Nat),?y:Nat),plus(?z:Nat,?y:Nat)),?z:Nat)",
"s(s(plus(s(plus(?y:Nat,?z:Nat)),?z:Nat))) = plus(?x:Nat,plus(plus(plus(?y:Nat,plus(?y:Nat,?y:Nat)),?y:Nat),?z:Nat))",
"plus(?x:Nat,s(plus(?x:Nat,s(s(?x:Nat))))) = s(plus(?x:Nat,plus(s(?y:Nat),plus(?z:Nat,plus(?x:Nat,?y:Nat)))))",
"plus(?x:Nat,s(plus(plus(?z:Nat,?y:Nat),?x:Nat))) = plus(plus(plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),?y:Nat),?x:Nat),?x:Nat)",
"plus(plus(?x:Nat,plus(?z:Nat,s(?z:Nat))),?x:Nat) = plus(plus(?y:Nat,plus(?y:Nat,plus(s(?z:Nat),?y:Nat))),s(?y:Nat))",
"s(s(plus(s(?y:Nat),plus(?y:Nat,s(?y:Nat))))) = plus(?x:Nat,s(s(plus(plus(?x:Nat,?z:Nat),s(?x:Nat)))))",
"s(s(plus(plus(?x:Nat,plus(?x:Nat,?z:Nat)),?y:Nat))) = plus(plus(?z:Nat,s(plus(plus(?z:Nat,?y:Nat),?x:Nat))),?x:Nat)",
"plus(?y:Nat,s(plus(plus(?x:Nat,plus(?y:Nat,?x:Nat)),?y:Nat))) = s(s(s(plus(s(?z:Nat),plus(?y:Nat,?z:Nat)))))",
"plus(s(plus(plus(?x:Nat,?z:Nat),?y:Nat)),s(s(s(plus(?y:Nat,?z:Nat))))) = plus(?x:Nat,plus(s(?y:Nat),?x:Nat))",
"plus(plus(plus(?x:Nat,plus(plus(?y:Nat,?x:Nat),plus(?y:Nat,?z:Nat))),?x:Nat),?x:Nat) = s(s(plus(?x:Nat,s(?y:Nat))))",
"plus(plus(plus(?y:Nat,?y:Nat),s(?z:Nat)),plus(plus(?x:Nat,s(?y:Nat)),?z:Nat)) = plus(?x:Nat,plus(s(?y:Nat),?x:Nat))",
"s(plus(plus(s(?z:Nat),plus(s(?y:Nat),plus(?z:Nat,?z:Nat))),plus(?z:Nat,?y:Nat))) = plus(?x:Nat,plus(?x:Nat,?z:Nat))",
"plus(?x:Nat,plus(plus(plus(?y:Nat,plus(?z:Nat,?z:Nat)),s(s(?x:Nat))),s(?z:Nat))) = plus(s(?x:Nat),s(?y:Nat))",
"plus(plus(s(?y:Nat),plus(plus(plus(?x:Nat,?y:Nat),plus(?x:Nat,?x:Nat)),?x:Nat)),?x:Nat) = plus(plus(?x:Nat,?y:Nat),?y:Nat)",
"plus(plus(plus(plus(?y:Nat,plus(?z:Nat,?x:Nat)),?x:Nat),plus(?x:Nat,s(?y:Nat))),?z:Nat) = plus(plus(?x:Nat,?x:Nat),?x:Nat)",
"plus(plus(plus(plus(plus(?y:Nat,?y:Nat),plus(?y:Nat,?y:Nat)),?z:Nat),?y:Nat),s(?z:Nat)) = plus(?x:Nat,plus(?x:Nat,?y:Nat))",
"plus(?x:Nat,plus(plus(s(?y:Nat),?z:Nat),plus(plus(plus(?y:Nat,?x:Nat),s(?y:Nat)),?x:Nat))) = s(s(s(?x:Nat)))",
"plus(plus(plus(plus(s(?y:Nat),?y:Nat),plus(plus(?z:Nat,?y:Nat),?x:Nat)),?x:Nat),s(?x:Nat)) = plus(?x:Nat,s(?x:Nat))",
"plus(s(?y:Nat),plus(plus(s(plus(?y:Nat,?y:Nat)),?x:Nat),s(plus(?x:Nat,plus(?x:Nat,?x:Nat))))) = s(s(?y:Nat))",
"plus(plus(?x:Nat,?z:Nat),plus(s(plus(?y:Nat,?x:Nat)),plus(?z:Nat,plus(plus(?y:Nat,?z:Nat),?z:Nat)))) = plus(?y:Nat,?y:Nat)",
"plus(?y:Nat,plus(s(plus(s(?x:Nat),?z:Nat)),plus(plus(?x:Nat,?x:Nat),plus(s(?z:Nat),s(?y:Nat))))) = s(?y:Nat)",
"s(?x:Nat) = plus(plus(?x:Nat,plus(?x:Nat,s(s(?z:Nat)))),plus(plus(?z:Nat,plus(plus(?x:Nat,?z:Nat),s(?x:Nat))),?x:Nat))",
"s(?x:Nat) = plus(plus(plus(plus(plus(?x:Nat,?z:Nat),?y:Nat),?x:Nat),s(plus(plus(?x:Nat,?x:Nat),plus(?z:Nat,?z:Nat)))),?z:Nat)",
"s(?y:Nat) = plus(plus(plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,?y:Nat)),plus(s(plus(?x:Nat,?x:Nat)),s(s(?x:Nat)))),?x:Nat)",
"s(?y:Nat) = plus(plus(plus(plus(?y:Nat,?y:Nat),plus(s(?x:Nat),?x:Nat)),plus(?x:Nat,plus(?x:Nat,?x:Nat))),plus(?x:Nat,?z:Nat))",
"s(?y:Nat) = plus(plus(plus(plus(plus(?x:Nat,?x:Nat),s(?z:Nat)),plus(s(?x:Nat),s(?y:Nat))),plus(?z:Nat,?y:Nat)),?z:Nat)",
"plus(?x:Nat,?x:Nat) = plus(s(plus(plus(?x:Nat,s(?y:Nat)),plus(s(?y:Nat),s(?y:Nat)))),s(s(s(s(?x:Nat)))))",
"plus(?x:Nat,?x:Nat) = plus(s(plus(plus(s(?z:Nat),?z:Nat),?y:Nat)),plus(s(?z:Nat),s(plus(plus(?y:Nat,?z:Nat),?x:Nat))))",
"plus(?y:Nat,?y:Nat) = plus(?y:Nat,plus(plus(s(plus(?x:Nat,?x:Nat)),s(?x:Nat)),plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),?x:Nat)))",
"s(plus(?x:Nat,?z:Nat)) = plus(s(?y:Nat),plus(plus(plus(plus(?z:Nat,?x:Nat),s(?y:Nat)),?x:Nat),plus(?x:Nat,s(?x:Nat))))",
"s(plus(?x:Nat,?z:Nat)) = plus(plus(?x:Nat,?y:Nat),plus(plus(s(plus(?y:Nat,?z:Nat)),?z:Nat),plus(?x:Nat,plus(?y:Nat,?z:Nat))))",
"plus(?x:Nat,plus(?y:Nat,?z:Nat)) = plus(?y:Nat,plus(plus(?y:Nat,s(s(?y:Nat))),plus(?z:Nat,plus(?y:Nat,plus(?y:Nat,?z:Nat)))))",
"plus(plus(?x:Nat,?x:Nat),?x:Nat) = plus(?x:Nat,plus(plus(?x:Nat,plus(plus(?x:Nat,?y:Nat),?x:Nat)),s(plus(?z:Nat,s(?y:Nat)))))",
"plus(s(?z:Nat),s(s(?y:Nat))) = s(plus(s(plus(plus(?z:Nat,?z:Nat),?y:Nat)),plus(s(plus(?x:Nat,?z:Nat)),?x:Nat)))",
"plus(plus(?y:Nat,s(?y:Nat)),?x:Nat) = plus(s(plus(?x:Nat,plus(plus(?x:Nat,?x:Nat),plus(?y:Nat,?x:Nat)))),plus(?x:Nat,?y:Nat))",
"s(plus(s(plus(?y:Nat,?z:Nat)),?z:Nat)) = plus(plus(?x:Nat,plus(plus(plus(?z:Nat,?y:Nat),plus(?z:Nat,?z:Nat)),?z:Nat)),?y:Nat)",
"plus(?y:Nat,s(plus(s(?z:Nat),?y:Nat))) = plus(plus(?z:Nat,?y:Nat),plus(plus(?y:Nat,?y:Nat),plus(?y:Nat,plus(?y:Nat,?x:Nat))))",
"plus(plus(?z:Nat,?x:Nat),plus(?y:Nat,?z:Nat)) = s(s(plus(s(plus(?x:Nat,?x:Nat)),plus(s(?x:Nat),plus(?y:Nat,?y:Nat)))))",
"plus(s(plus(plus(?y:Nat,?x:Nat),?y:Nat)),?x:Nat) = plus(?y:Nat,plus(s(s(plus(?y:Nat,?z:Nat))),plus(?y:Nat,s(?x:Nat))))",
"plus(plus(s(plus(?z:Nat,?y:Nat)),?z:Nat),?y:Nat) = plus(s(plus(plus(plus(?y:Nat,?x:Nat),s(?y:Nat)),?y:Nat)),s(?x:Nat))",
"s(plus(?x:Nat,s(plus(s(?x:Nat),s(?x:Nat))))) = plus(?x:Nat,s(s(plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat)))))",
"s(plus(?y:Nat,plus(plus(s(?x:Nat),?y:Nat),?x:Nat))) = plus(s(plus(?y:Nat,?z:Nat)),s(plus(?y:Nat,plus(?y:Nat,?x:Nat))))",
"s(plus(?y:Nat,plus(plus(?y:Nat,?y:Nat),plus(?y:Nat,?y:Nat)))) = plus(plus(plus(s(plus(?x:Nat,?x:Nat)),?x:Nat),?y:Nat),?x:Nat)",
"s(plus(plus(plus(?y:Nat,plus(?z:Nat,?x:Nat)),?y:Nat),?z:Nat)) = plus(s(plus(?z:Nat,?y:Nat)),plus(plus(?y:Nat,?y:Nat),?x:Nat))",
"plus(?z:Nat,plus(plus(plus(?x:Nat,?z:Nat),s(?x:Nat)),?x:Nat)) = s(plus(?x:Nat,plus(plus(?y:Nat,?x:Nat),plus(?z:Nat,?x:Nat))))",
"plus(s(plus(?y:Nat,?x:Nat)),s(s(plus(?z:Nat,?x:Nat)))) = plus(?x:Nat,plus(plus(s(?x:Nat),?x:Nat),s(s(?z:Nat))))",
"plus(plus(?x:Nat,plus(?x:Nat,s(plus(?x:Nat,?x:Nat)))),?x:Nat) = s(plus(s(plus(s(?z:Nat),plus(?x:Nat,?x:Nat))),?x:Nat))",
"plus(plus(s(plus(?z:Nat,s(?x:Nat))),s(?x:Nat)),?x:Nat) = plus(?z:Nat,plus(plus(?y:Nat,s(plus(?x:Nat,?z:Nat))),?x:Nat))",
"plus(s(s(plus(plus(?x:Nat,?z:Nat),?x:Nat))),plus(?x:Nat,?x:Nat)) = s(plus(s(s(plus(?x:Nat,?x:Nat))),s(?y:Nat)))",
"s(plus(plus(s(plus(?x:Nat,?z:Nat)),s(plus(?x:Nat,?x:Nat))),?x:Nat)) = plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),s(?x:Nat))",
"plus(plus(s(s(?x:Nat)),s(?z:Nat)),plus(plus(?x:Nat,?y:Nat),?x:Nat)) = s(plus(s(?y:Nat),plus(s(?x:Nat),?y:Nat)))",
"plus(plus(plus(?x:Nat,s(plus(?x:Nat,?x:Nat))),?x:Nat),plus(?y:Nat,?x:Nat)) = s(plus(?z:Nat,plus(?y:Nat,plus(?x:Nat,?z:Nat))))",
"plus(?x:Nat,plus(s(plus(?x:Nat,?x:Nat)),plus(plus(?x:Nat,s(?x:Nat)),?x:Nat))) = plus(plus(s(?x:Nat),s(?x:Nat)),?x:Nat)",
"plus(s(?y:Nat),plus(?x:Nat,plus(s(?x:Nat),plus(plus(?y:Nat,?x:Nat),?z:Nat)))) = plus(s(plus(?y:Nat,s(?x:Nat))),?x:Nat)",
"plus(s(plus(s(?y:Nat),plus(?x:Nat,plus(?y:Nat,?x:Nat)))),plus(?x:Nat,?x:Nat)) = plus(s(plus(?y:Nat,?y:Nat)),s(?x:Nat))",
"plus(?z:Nat,plus(plus(?x:Nat,plus(?y:Nat,plus(?z:Nat,?z:Nat))),plus(plus(?x:Nat,?z:Nat),?x:Nat))) = plus(?y:Nat,plus(?z:Nat,?z:Nat))",
"plus(plus(plus(s(?x:Nat),plus(?x:Nat,?z:Nat)),plus(s(plus(?x:Nat,?x:Nat)),?x:Nat)),?x:Nat) = plus(?y:Nat,s(s(?x:Nat)))",
"plus(plus(plus(s(plus(?x:Nat,?x:Nat)),s(?y:Nat)),plus(?y:Nat,plus(?x:Nat,s(?z:Nat)))),?y:Nat) = s(plus(?y:Nat,?z:Nat))",
"plus(plus(?z:Nat,plus(?y:Nat,s(s(?x:Nat)))),plus(s(plus(plus(?x:Nat,?x:Nat),?x:Nat)),s(?x:Nat))) = plus(?x:Nat,?z:Nat)",
"plus(plus(s(plus(plus(?y:Nat,?x:Nat),s(?y:Nat))),?y:Nat),plus(s(plus(s(?x:Nat),?z:Nat)),?y:Nat)) = plus(?x:Nat,?x:Nat)",
"plus(?y:Nat,s(?y:Nat)) = plus(plus(s(plus(?x:Nat,?x:Nat)),plus(plus(?x:Nat,s(?y:Nat)),s(plus(?y:Nat,?x:Nat)))),s(?x:Nat))",
"plus(?z:Nat,s(?x:Nat)) = plus(plus(s(s(plus(?y:Nat,?y:Nat))),?y:Nat),plus(?y:Nat,plus(s(?z:Nat),s(plus(?x:Nat,?z:Nat)))))",
"plus(?z:Nat,s(?y:Nat)) = plus(?x:Nat,plus(plus(plus(plus(?x:Nat,?x:Nat),?x:Nat),plus(plus(?x:Nat,?x:Nat),plus(?z:Nat,?z:Nat))),?z:Nat))",
"plus(?y:Nat,plus(plus(?z:Nat,?y:Nat),s(?y:Nat))) = plus(s(plus(plus(?z:Nat,?y:Nat),s(?y:Nat))),plus(plus(?z:Nat,?y:Nat),?y:Nat))",
"s(plus(plus(plus(s(?x:Nat),?x:Nat),?y:Nat),?x:Nat)) = plus(plus(?z:Nat,s(?x:Nat)),plus(plus(s(?z:Nat),?x:Nat),s(?z:Nat)))",
"plus(s(plus(?x:Nat,plus(?x:Nat,s(?y:Nat)))),?x:Nat) = plus(s(plus(plus(s(?x:Nat),?y:Nat),s(plus(?x:Nat,?x:Nat)))),?x:Nat)",
"plus(plus(s(s(?x:Nat)),s(?z:Nat)),s(?z:Nat)) = plus(plus(s(s(s(?z:Nat))),plus(?x:Nat,plus(?x:Nat,?x:Nat))),?x:Nat)",
"plus(plus(plus(?z:Nat,?x:Nat),?x:Nat),plus(?x:Nat,?x:Nat)) = plus(plus(plus(s(plus(?y:Nat,?z:Nat)),s(?y:Nat)),s(?x:Nat)),?x:Nat)",
"plus(?x:Nat,plus(?x:Nat,plus(s(?y:Nat),plus(?x:Nat,?z:Nat)))) = plus(plus(plus(?y:Nat,plus(s(?x:Nat),s(?x:Nat))),?x:Nat),?z:Nat)",
"plus(s(plus(plus(?x:Nat,?y:Nat),s(plus(?z:Nat,?x:Nat)))),?y:Nat) = plus(?z:Nat,plus(?x:Nat,plus(plus(?y:Nat,?x:Nat),s(?x:Nat))))",
"plus(plus(?y:Nat,plus(plus(s(?y:Nat),?x:Nat),?y:Nat)),plus(?y:Nat,s(?z:Nat))) = plus(plus(plus(?x:Nat,?x:Nat),s(?y:Nat)),?z:Nat)",
"plus(?z:Nat,plus(plus(?z:Nat,?x:Nat),plus(plus(?z:Nat,plus(?z:Nat,?x:Nat)),s(?x:Nat)))) = plus(plus(?x:Nat,plus(?y:Nat,?y:Nat)),?x:Nat)",
"plus(plus(plus(?y:Nat,?y:Nat),plus(s(plus(?x:Nat,?y:Nat)),s(s(?y:Nat)))),?y:Nat) = plus(?x:Nat,plus(?z:Nat,plus(?y:Nat,?x:Nat)))",
"plus(?y:Nat,plus(plus(plus(s(?y:Nat),plus(?z:Nat,?x:Nat)),plus(?x:Nat,s(?y:Nat))),s(?y:Nat))) = plus(plus(?z:Nat,?x:Nat),?x:Nat)",
"plus(?y:Nat,plus(plus(plus(?z:Nat,?y:Nat),plus(?x:Nat,?y:Nat)),plus(plus(s(?x:Nat),?y:Nat),s(?y:Nat)))) = s(s(s(?y:Nat)))",
"plus(s(plus(?z:Nat,plus(?z:Nat,plus(?x:Nat,?x:Nat)))),plus(s(plus(?x:Nat,plus(?x:Nat,?z:Nat))),?y:Nat)) = s(plus(?x:Nat,?x:Nat))",
"plus(plus(plus(plus(?z:Nat,s(?y:Nat)),?y:Nat),plus(?x:Nat,plus(plus(?x:Nat,?x:Nat),plus(?y:Nat,?y:Nat)))),?x:Nat) = plus(?x:Nat,?x:Nat)",
"plus(?x:Nat,?z:Nat) = plus(plus(plus(?x:Nat,plus(?z:Nat,plus(?x:Nat,?x:Nat))),plus(s(?y:Nat),plus(s(?y:Nat),?x:Nat))),plus(?z:Nat,?x:Nat))",
"plus(s(?x:Nat),?x:Nat) = plus(s(?x:Nat),plus(plus(s(?x:Nat),plus(?y:Nat,plus(?x:Nat,?x:Nat))),plus(?x:Nat,s(plus(?z:Nat,?z:Nat)))))",
"s(plus(s(?x:Nat),plus(?z:Nat,?y:Nat))) = plus(s(plus(?y:Nat,plus(s(?y:Nat),?x:Nat))),s(s(plus(s(?x:Nat),s(?y:Nat)))))",
"s(plus(plus(plus(?x:Nat,?z:Nat),?x:Nat),?z:Nat)) = plus(s(plus(?x:Nat,s(s(?y:Nat)))),plus(?y:Nat,s(s(plus(?x:Nat,?y:Nat)))))",
"plus(plus(plus(plus(?x:Nat,?x:Nat),?x:Nat),?y:Nat),?y:Nat) = plus(plus(?x:Nat,plus(?z:Nat,plus(?y:Nat,plus(?z:Nat,?y:Nat)))),plus(?y:Nat,?z:Nat))",
"s(plus(s(?x:Nat),s(plus(plus(?y:Nat,?x:Nat),?y:Nat)))) = plus(?y:Nat,plus(plus(s(?x:Nat),?x:Nat),plus(?x:Nat,s(s(?x:Nat)))))",
"plus(s(plus(?y:Nat,?y:Nat)),plus(s(?z:Nat),s(?z:Nat))) = plus(plus(s(s(?x:Nat)),plus(s(?y:Nat),s(?z:Nat))),s(?x:Nat))",
"plus(plus(plus(?x:Nat,plus(?x:Nat,plus(?x:Nat,?y:Nat))),?x:Nat),?x:Nat) = plus(s(s(?x:Nat)),plus(s(?y:Nat),plus(s(?z:Nat),?y:Nat)))",
"s(s(plus(plus(plus(?y:Nat,?z:Nat),?y:Nat),plus(?z:Nat,s(?y:Nat))))) = plus(?y:Nat,plus(?x:Nat,plus(s(plus(?z:Nat,?x:Nat)),?y:Nat)))",
"plus(s(plus(plus(?x:Nat,plus(?y:Nat,?x:Nat)),?x:Nat)),plus(?x:Nat,?y:Nat)) = plus(plus(?x:Nat,?z:Nat),plus(plus(?x:Nat,?x:Nat),s(?z:Nat)))",
"plus(plus(?x:Nat,?z:Nat),plus(?x:Nat,plus(plus(s(?y:Nat),?x:Nat),?x:Nat))) = plus(s(s(plus(plus(?y:Nat,?x:Nat),s(?y:Nat)))),?x:Nat)",
"plus(plus(plus(s(?y:Nat),?x:Nat),plus(?y:Nat,?y:Nat)),plus(?z:Nat,?x:Nat)) = plus(s(?z:Nat),plus(?x:Nat,s(plus(?x:Nat,s(?y:Nat)))))",
"plus(plus(plus(?x:Nat,s(plus(?x:Nat,?x:Nat))),s(?z:Nat)),plus(?y:Nat,?z:Nat)) = s(plus(?y:Nat,plus(plus(s(?z:Nat),?y:Nat),?y:Nat)))",
"plus(?y:Nat,plus(plus(?y:Nat,?y:Nat),plus(plus(?x:Nat,plus(?z:Nat,?x:Nat)),s(?x:Nat)))) = s(plus(plus(?x:Nat,plus(?z:Nat,?x:Nat)),?z:Nat))",
"plus(plus(?x:Nat,plus(?x:Nat,plus(plus(?x:Nat,?x:Nat),plus(?y:Nat,?y:Nat)))),plus(?x:Nat,?y:Nat)) = plus(s(plus(s(?z:Nat),?y:Nat)),?y:Nat)",
"plus(plus(plus(plus(plus(?x:Nat,?x:Nat),?x:Nat),plus(s(?x:Nat),?z:Nat)),plus(?x:Nat,?x:Nat)),?x:Nat) = s(plus(?z:Nat,plus(?x:Nat,?x:Nat)))",
"plus(plus(?x:Nat,plus(plus(s(?y:Nat),?x:Nat),plus(?y:Nat,plus(?y:Nat,?x:Nat)))),plus(?y:Nat,s(s(plus(?z:Nat,?x:Nat))))) = s(?x:Nat)",
"plus(plus(plus(?x:Nat,?z:Nat),plus(?x:Nat,?y:Nat)),plus(plus(?z:Nat,s(?x:Nat)),plus(plus(plus(?y:Nat,?x:Nat),?y:Nat),?x:Nat))) = s(?x:Nat)",
"s(plus(?z:Nat,?x:Nat)) = plus(plus(plus(?z:Nat,?y:Nat),plus(?y:Nat,plus(s(?y:Nat),?y:Nat))),s(plus(plus(plus(?z:Nat,?y:Nat),?x:Nat),?y:Nat)))",
"plus(plus(?x:Nat,?x:Nat),?z:Nat) = s(plus(plus(plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),?z:Nat),plus(plus(s(?z:Nat),s(?z:Nat)),?x:Nat)))",
"s(plus(s(?x:Nat),s(?x:Nat))) = plus(?y:Nat,plus(plus(?y:Nat,plus(?x:Nat,?y:Nat)),plus(plus(plus(?x:Nat,?y:Nat),plus(?z:Nat,?x:Nat)),?y:Nat)))",
"plus(s(plus(?x:Nat,s(s(?z:Nat)))),?y:Nat) = s(plus(plus(plus(s(?x:Nat),?z:Nat),?x:Nat),plus(?x:Nat,plus(?y:Nat,plus(?y:Nat,?x:Nat)))))",
"s(plus(?y:Nat,s(plus(plus(?y:Nat,?y:Nat),?y:Nat)))) = plus(s(plus(?x:Nat,?z:Nat)),s(plus(plus(s(?x:Nat),plus(?x:Nat,?y:Nat)),?x:Nat)))",
"s(plus(?x:Nat,plus(?x:Nat,plus(s(?x:Nat),plus(?z:Nat,?x:Nat))))) = plus(?y:Nat,plus(plus(plus(?x:Nat,?x:Nat),s(?y:Nat)),plus(?y:Nat,?z:Nat)))",
"plus(plus(plus(plus(s(?z:Nat),s(?x:Nat)),?x:Nat),?x:Nat),?y:Nat) = plus(?x:Nat,plus(plus(s(plus(?x:Nat,?z:Nat)),s(?z:Nat)),s(?z:Nat)))",
"plus(s(plus(plus(?z:Nat,plus(?x:Nat,?y:Nat)),s(?z:Nat))),plus(?z:Nat,?x:Nat)) = s(plus(?x:Nat,plus(plus(?x:Nat,?x:Nat),plus(?y:Nat,?x:Nat))))",
"plus(?y:Nat,plus(?y:Nat,plus(plus(s(?x:Nat),?x:Nat),plus(plus(?y:Nat,?y:Nat),plus(?z:Nat,?y:Nat))))) = plus(s(?x:Nat),s(s(s(?z:Nat))))",
"plus(plus(plus(s(plus(?x:Nat,?x:Nat)),?y:Nat),plus(?x:Nat,s(?x:Nat))),plus(?y:Nat,plus(?x:Nat,?x:Nat))) = plus(plus(?x:Nat,s(?x:Nat)),?z:Nat)",
"plus(plus(s(plus(?x:Nat,?z:Nat)),plus(plus(?x:Nat,s(?x:Nat)),?x:Nat)),plus(?x:Nat,plus(?x:Nat,plus(?z:Nat,?y:Nat)))) = plus(s(?y:Nat),?y:Nat)",
"plus(plus(plus(?z:Nat,plus(?x:Nat,?z:Nat)),plus(plus(?x:Nat,?z:Nat),?x:Nat)),plus(plus(?z:Nat,?y:Nat),plus(s(?x:Nat),?y:Nat))) = plus(?x:Nat,?y:Nat)",
"plus(plus(plus(s(s(?y:Nat)),s(plus(?y:Nat,?y:Nat))),?z:Nat),plus(plus(plus(plus(?x:Nat,?x:Nat),s(?x:Nat)),?z:Nat),?x:Nat)) = s(?z:Nat)",
"s(?y:Nat) = plus(plus(?z:Nat,plus(s(?y:Nat),?x:Nat)),plus(s(?x:Nat),plus(plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?y:Nat)),plus(s(?y:Nat),?z:Nat))))",
"plus(?x:Nat,plus(s(?x:Nat),?x:Nat)) = plus(plus(?x:Nat,?z:Nat),s(plus(plus(?z:Nat,plus(?x:Nat,?y:Nat)),plus(plus(?x:Nat,?x:Nat),plus(?z:Nat,?z:Nat)))))",
"plus(?y:Nat,plus(s(plus(s(?x:Nat),s(?z:Nat))),s(s(plus(?z:Nat,?z:Nat))))) = plus(plus(?y:Nat,s(s(plus(?x:Nat,?x:Nat)))),s(?z:Nat))",
"plus(?z:Nat,plus(s(?z:Nat),plus(plus(?x:Nat,s(?y:Nat)),plus(?x:Nat,plus(?x:Nat,?z:Nat))))) = plus(s(plus(plus(?z:Nat,s(?y:Nat)),?y:Nat)),?x:Nat)",
"plus(plus(plus(?z:Nat,s(plus(?x:Nat,?z:Nat))),plus(?y:Nat,s(plus(?x:Nat,?z:Nat)))),?x:Nat) = plus(plus(?x:Nat,plus(?x:Nat,plus(?z:Nat,?z:Nat))),?x:Nat)",
"plus(s(plus(plus(plus(?x:Nat,?z:Nat),plus(?x:Nat,?x:Nat)),plus(plus(?y:Nat,?x:Nat),s(?y:Nat)))),plus(?z:Nat,plus(?z:Nat,?z:Nat))) = plus(?x:Nat,?x:Nat)",
"plus(plus(plus(?y:Nat,?x:Nat),?y:Nat),plus(plus(?y:Nat,?z:Nat),plus(plus(s(?x:Nat),plus(?x:Nat,?x:Nat)),s(plus(?x:Nat,?z:Nat))))) = plus(?z:Nat,?y:Nat)",
"plus(plus(plus(plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),plus(?y:Nat,?x:Nat)),plus(?z:Nat,s(?z:Nat))),s(plus(?x:Nat,?x:Nat))) = plus(?z:Nat,?y:Nat)",
"s(s(s(s(plus(?y:Nat,?z:Nat))))) = plus(plus(plus(s(plus(?x:Nat,?x:Nat)),?z:Nat),?x:Nat),plus(?z:Nat,plus(?x:Nat,plus(?y:Nat,plus(?y:Nat,?y:Nat)))))",
"plus(plus(plus(?x:Nat,?y:Nat),?z:Nat),?x:Nat) = plus(s(plus(?y:Nat,plus(plus(?x:Nat,?x:Nat),?y:Nat))),plus(?x:Nat,plus(plus(plus(?x:Nat,?x:Nat),?y:Nat),?y:Nat)))",
"plus(?y:Nat,plus(plus(plus(s(?z:Nat),?y:Nat),plus(?y:Nat,?x:Nat)),?z:Nat)) = plus(s(?y:Nat),plus(plus(?x:Nat,?x:Nat),plus(plus(?z:Nat,?x:Nat),s(?z:Nat))))",
"plus(s(plus(s(s(?x:Nat)),s(plus(?y:Nat,?y:Nat)))),s(s(plus(?x:Nat,?x:Nat)))) = plus(s(?y:Nat),s(plus(?z:Nat,plus(?z:Nat,s(?y:Nat)))))",
"plus(plus(plus(?y:Nat,plus(?z:Nat,plus(?y:Nat,?y:Nat))),plus(?x:Nat,plus(?x:Nat,s(?x:Nat)))),?y:Nat) = plus(s(plus(?y:Nat,s(plus(?z:Nat,?z:Nat)))),?x:Nat)",
"plus(?y:Nat,?z:Nat) = plus(plus(?z:Nat,plus(?x:Nat,?x:Nat)),plus(plus(?y:Nat,plus(plus(?x:Nat,?y:Nat),?z:Nat)),plus(plus(?y:Nat,plus(?z:Nat,?y:Nat)),plus(?x:Nat,?x:Nat))))",
"plus(plus(?x:Nat,?x:Nat),?x:Nat) = s(plus(plus(plus(plus(?x:Nat,?z:Nat),?y:Nat),plus(?y:Nat,?y:Nat)),plus(s(plus(?z:Nat,?x:Nat)),plus(plus(?x:Nat,?x:Nat),?z:Nat))))",
"plus(plus(plus(plus(?x:Nat,?x:Nat),?y:Nat),?x:Nat),?y:Nat) = s(plus(plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),s(?z:Nat)),plus(plus(plus(?x:Nat,?y:Nat),?y:Nat),?y:Nat)))",
"plus(plus(?x:Nat,s(plus(?y:Nat,plus(?x:Nat,?y:Nat)))),?y:Nat) = plus(?z:Nat,plus(s(plus(?x:Nat,plus(?x:Nat,?x:Nat))),plus(s(?y:Nat),s(plus(?z:Nat,?x:Nat)))))",
"plus(plus(s(s(s(?y:Nat))),?y:Nat),plus(?x:Nat,?z:Nat)) = plus(s(s(plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,?x:Nat)))),s(s(plus(?x:Nat,s(?x:Nat)))))",
"plus(plus(?y:Nat,s(plus(plus(?y:Nat,?y:Nat),s(?x:Nat)))),?y:Nat) = plus(plus(plus(plus(?y:Nat,plus(?y:Nat,?y:Nat)),s(?x:Nat)),plus(s(?y:Nat),?x:Nat)),?x:Nat)",
"plus(plus(s(plus(?z:Nat,?x:Nat)),s(?y:Nat)),plus(?x:Nat,?x:Nat)) = plus(plus(s(plus(plus(?x:Nat,?x:Nat),?x:Nat)),plus(?x:Nat,s(?x:Nat))),plus(?x:Nat,?y:Nat))",
"plus(plus(plus(s(plus(?y:Nat,?x:Nat)),?x:Nat),s(?x:Nat)),?x:Nat) = plus(plus(s(?x:Nat),plus(s(plus(?x:Nat,?z:Nat)),?x:Nat)),s(plus(s(?y:Nat),?z:Nat)))",
"plus(?y:Nat,plus(plus(s(s(?x:Nat)),plus(plus(?x:Nat,?y:Nat),?y:Nat)),?x:Nat)) = plus(plus(?z:Nat,plus(?z:Nat,plus(?x:Nat,plus(?x:Nat,?y:Nat)))),plus(?x:Nat,?x:Nat))",
"plus(plus(?x:Nat,plus(plus(plus(?z:Nat,?y:Nat),?y:Nat),plus(?x:Nat,s(?x:Nat)))),s(?x:Nat)) = plus(?z:Nat,s(plus(plus(plus(?x:Nat,?x:Nat),?z:Nat),s(?x:Nat))))",
"plus(plus(?y:Nat,plus(s(plus(?y:Nat,?x:Nat)),s(plus(?x:Nat,?x:Nat)))),s(s(s(?x:Nat)))) = plus(s(?y:Nat),s(s(plus(?z:Nat,plus(?x:Nat,?y:Nat)))))",
"plus(?y:Nat,plus(plus(plus(?x:Nat,plus(?y:Nat,?x:Nat)),plus(?x:Nat,?y:Nat)),s(plus(plus(?y:Nat,?y:Nat),?z:Nat)))) = s(plus(?y:Nat,plus(plus(?z:Nat,?x:Nat),?x:Nat)))",
"plus(plus(s(?z:Nat),plus(?y:Nat,plus(s(?x:Nat),?z:Nat))),s(plus(plus(plus(?x:Nat,?z:Nat),?x:Nat),?y:Nat))) = plus(?z:Nat,plus(?x:Nat,s(plus(?x:Nat,?x:Nat))))",
"plus(plus(s(plus(?y:Nat,?y:Nat)),plus(plus(?y:Nat,plus(?x:Nat,?x:Nat)),?x:Nat)),s(s(plus(plus(?x:Nat,?y:Nat),?y:Nat)))) = plus(plus(?y:Nat,s(?y:Nat)),?y:Nat)",
"plus(plus(s(plus(s(?y:Nat),?x:Nat)),?x:Nat),?y:Nat) = plus(plus(plus(s(s(?z:Nat)),s(?x:Nat)),plus(plus(plus(?y:Nat,?y:Nat),?y:Nat),s(?x:Nat))),s(?y:Nat))",
"s(plus(plus(?y:Nat,plus(s(?z:Nat),s(?y:Nat))),?z:Nat)) = plus(?z:Nat,plus(plus(?y:Nat,plus(plus(?z:Nat,?x:Nat),?x:Nat)),plus(?x:Nat,plus(?x:Nat,plus(?z:Nat,?x:Nat)))))",
"plus(?z:Nat,plus(?x:Nat,plus(?x:Nat,plus(?z:Nat,plus(?y:Nat,?y:Nat))))) = plus(plus(plus(?y:Nat,?x:Nat),?x:Nat),plus(plus(?y:Nat,plus(s(?x:Nat),plus(?x:Nat,?x:Nat))),?x:Nat))",
"s(plus(plus(?y:Nat,plus(s(?x:Nat),plus(?x:Nat,?x:Nat))),s(?x:Nat))) = plus(plus(?x:Nat,?x:Nat),plus(plus(?y:Nat,plus(?x:Nat,s(?x:Nat))),s(plus(?z:Nat,?x:Nat))))",
"plus(plus(plus(?x:Nat,?y:Nat),?x:Nat),plus(plus(s(s(?x:Nat)),?x:Nat),?y:Nat)) = plus(plus(?x:Nat,plus(s(s(?y:Nat)),s(?y:Nat))),plus(plus(?y:Nat,?x:Nat),?y:Nat))",
"plus(s(plus(s(?x:Nat),plus(s(?x:Nat),plus(?z:Nat,?y:Nat)))),s(plus(?y:Nat,?z:Nat))) = plus(plus(plus(s(?x:Nat),plus(?y:Nat,?z:Nat)),plus(?x:Nat,?y:Nat)),?x:Nat)",
"plus(?x:Nat,plus(s(plus(?z:Nat,plus(?y:Nat,?z:Nat))),plus(?y:Nat,plus(plus(?x:Nat,?y:Nat),?z:Nat)))) = plus(?x:Nat,plus(plus(?x:Nat,s(s(?y:Nat))),plus(?z:Nat,?x:Nat)))",
"plus(?x:Nat,plus(plus(?y:Nat,s(?y:Nat)),plus(plus(plus(?y:Nat,?z:Nat),s(?x:Nat)),s(?y:Nat)))) = plus(s(plus(plus(?y:Nat,?y:Nat),s(plus(?x:Nat,?x:Nat)))),?y:Nat)",
"s(s(?y:Nat)) = plus(plus(plus(plus(plus(?x:Nat,?z:Nat),s(?z:Nat)),plus(?x:Nat,s(?z:Nat))),s(s(s(?y:Nat)))),plus(plus(?x:Nat,?x:Nat),plus(?z:Nat,s(?x:Nat))))",
"plus(plus(s(?x:Nat),plus(s(?y:Nat),s(plus(?x:Nat,?x:Nat)))),?y:Nat) = plus(?x:Nat,plus(s(plus(plus(?y:Nat,?x:Nat),plus(?z:Nat,?y:Nat))),plus(plus(?y:Nat,?x:Nat),?z:Nat)))",
"plus(plus(plus(s(?z:Nat),?y:Nat),plus(?z:Nat,plus(?y:Nat,?y:Nat))),plus(?y:Nat,?x:Nat)) = plus(plus(s(?z:Nat),plus(?z:Nat,plus(plus(?z:Nat,?x:Nat),plus(?z:Nat,?x:Nat)))),?y:Nat)",
"plus(plus(plus(s(?z:Nat),plus(?y:Nat,plus(?x:Nat,?x:Nat))),plus(s(plus(?x:Nat,?x:Nat)),plus(?x:Nat,?x:Nat))),s(s(?y:Nat))) = s(plus(?x:Nat,plus(s(?z:Nat),?x:Nat)))",
"plus(?x:Nat,plus(plus(plus(plus(?z:Nat,?x:Nat),s(?x:Nat)),s(plus(?x:Nat,?y:Nat))),?x:Nat)) = plus(s(plus(plus(s(?y:Nat),?x:Nat),?y:Nat)),plus(s(?y:Nat),s(s(?y:Nat))))",
"plus(plus(plus(s(?x:Nat),plus(?x:Nat,plus(?z:Nat,?x:Nat))),s(s(plus(?x:Nat,?x:Nat)))),?x:Nat) = plus(plus(plus(plus(?y:Nat,?x:Nat),s(s(?x:Nat))),plus(?z:Nat,?y:Nat)),?z:Nat)",
"plus(plus(?x:Nat,?x:Nat),plus(plus(?x:Nat,plus(?z:Nat,plus(?z:Nat,?z:Nat))),plus(s(?x:Nat),plus(s(?y:Nat),?z:Nat)))) = plus(s(plus(?x:Nat,plus(plus(?x:Nat,?x:Nat),?x:Nat))),?z:Nat)",
"plus(plus(plus(plus(?z:Nat,s(?z:Nat)),?z:Nat),?z:Nat),plus(plus(plus(s(?x:Nat),s(?x:Nat)),s(?z:Nat)),plus(s(plus(?z:Nat,?x:Nat)),?x:Nat))) = plus(?y:Nat,s(s(?x:Nat)))",
"plus(plus(plus(?x:Nat,s(s(?z:Nat))),?x:Nat),?z:Nat) = plus(plus(plus(?z:Nat,s(?z:Nat)),plus(plus(s(?z:Nat),plus(?x:Nat,?y:Nat)),plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?y:Nat)))),?z:Nat)",
"plus(plus(plus(?y:Nat,?y:Nat),plus(s(plus(?z:Nat,?x:Nat)),?x:Nat)),plus(?z:Nat,?x:Nat)) = plus(plus(plus(s(plus(?z:Nat,?x:Nat)),?x:Nat),?x:Nat),plus(plus(plus(?z:Nat,?y:Nat),?z:Nat),?x:Nat))",
"plus(plus(s(?x:Nat),s(plus(?x:Nat,s(?y:Nat)))),plus(?x:Nat,plus(?z:Nat,plus(plus(?y:Nat,?x:Nat),s(?y:Nat))))) = plus(plus(s(s(plus(?y:Nat,?y:Nat))),?y:Nat),s(s(s(?x:Nat))))",
"plus(plus(?x:Nat,?y:Nat),s(plus(plus(plus(?x:Nat,?y:Nat),plus(?y:Nat,?x:Nat)),?y:Nat))) = plus(s(s(plus(?x:Nat,s(?y:Nat)))),s(plus(plus(plus(?x:Nat,?z:Nat),?y:Nat),plus(plus(?x:Nat,?z:Nat),?x:Nat))))",
"s(plus(s(s(plus(?z:Nat,?x:Nat))),plus(s(?x:Nat),plus(?z:Nat,plus(?z:Nat,?x:Nat))))) = plus(plus(plus(plus(?z:Nat,?x:Nat),s(plus(?z:Nat,?z:Nat))),plus(plus(?x:Nat,?z:Nat),?x:Nat)),plus(?x:Nat,?y:Nat))",
"s(plus(plus(s(?z:Nat),plus(s(?z:Nat),s(?x:Nat))),plus(?y:Nat,plus(?y:Nat,?y:Nat)))) = plus(plus(plus(?x:Nat,plus(plus(?z:Nat,?x:Nat),s(?y:Nat))),plus(plus(?z:Nat,plus(?x:Nat,?y:Nat)),?z:Nat)),?z:Nat)",
"plus(plus(plus(plus(s(?x:Nat),?z:Nat),plus(?y:Nat,?y:Nat)),plus(plus(?z:Nat,?z:Nat),plus(?x:Nat,plus(?y:Nat,?y:Nat)))),plus(plus(?x:Nat,plus(s(?x:Nat),?y:Nat)),?y:Nat)) = s(s(plus(plus(plus(?z:Nat,?x:Nat),plus(?x:Nat,?y:Nat)),s(?x:Nat))))"]

val problemsPlusWithZeroNo =
IOMstrs.rdEqs [
"s(?x:Nat) = plus(?x:Nat,0)",
"s(?x:Nat) = plus(?y:Nat,0)",
"s(?x:Nat) = plus(0,?x:Nat)",
"s(?x:Nat) = plus(0,?y:Nat)",
"s(0) = plus(?x:Nat,?x:Nat)",
"s(0) = plus(?x:Nat,?y:Nat)",
"s(0) = plus(0,0)",
"plus(?y:Nat,0) = s(0)",
"plus(0,?y:Nat) = s(0)",
"s(?x:Nat) = plus(0,s(?z:Nat))",
"s(?x:Nat) = plus(s(0),?y:Nat)",
"s(?y:Nat) = plus(?x:Nat,s(0))",
"s(0) = plus(?x:Nat,s(?z:Nat))",
"s(s(?x:Nat)) = plus(0,?x:Nat)",
"s(s(0)) = plus(?x:Nat,?z:Nat)",
"s(s(0)) = plus(?x:Nat,0)",
"s(s(0)) = plus(?y:Nat,?y:Nat)",
"plus(?x:Nat,?x:Nat) = plus(?x:Nat,0)",
"plus(?x:Nat,?x:Nat) = plus(?z:Nat,0)",
"plus(?x:Nat,?y:Nat) = plus(?y:Nat,0)",
"plus(?y:Nat,?x:Nat) = plus(0,?x:Nat)",
"plus(?y:Nat,?y:Nat) = plus(0,?x:Nat)",
"plus(?y:Nat,?z:Nat) = plus(?x:Nat,0)",
"plus(?y:Nat,?z:Nat) = plus(0,?y:Nat)",
"plus(?z:Nat,0) = s(s(?z:Nat))",
"plus(0,?x:Nat) = s(s(?y:Nat))",
"plus(0,?x:Nat) = plus(?y:Nat,0)",
"plus(?y:Nat,s(?y:Nat)) = s(0)",
"plus(s(?y:Nat),0) = s(?x:Nat)",
"plus(s(?z:Nat),0) = s(0)",
"plus(s(0),?z:Nat) = s(0)",
"s(?x:Nat) = plus(0,plus(?x:Nat,?y:Nat))",
"s(?x:Nat) = plus(plus(?y:Nat,0),?x:Nat)",
"s(?x:Nat) = plus(plus(?z:Nat,?x:Nat),0)",
"s(?y:Nat) = plus(0,plus(?z:Nat,?x:Nat))",
"s(?y:Nat) = plus(plus(?x:Nat,0),?z:Nat)",
"s(?z:Nat) = plus(?x:Nat,plus(?y:Nat,0))",
"s(0) = plus(?x:Nat,s(s(?x:Nat)))",
"s(0) = plus(?x:Nat,plus(?x:Nat,?x:Nat))",
"s(0) = plus(?y:Nat,plus(?x:Nat,?z:Nat))",
"s(0) = plus(s(?x:Nat),s(?x:Nat))",
"s(0) = plus(plus(?y:Nat,?y:Nat),?z:Nat)",
"s(0) = plus(plus(0,?z:Nat),?x:Nat)",
"s(s(0)) = plus(?y:Nat,s(?x:Nat))",
"plus(?x:Nat,?x:Nat) = s(plus(0,?z:Nat))",
"plus(?x:Nat,?x:Nat) = plus(0,s(?x:Nat))",
"plus(?x:Nat,?x:Nat) = plus(s(?y:Nat),0)",
"plus(?x:Nat,?x:Nat) = plus(s(0),?x:Nat)",
"plus(?x:Nat,?x:Nat) = plus(s(0),?y:Nat)",
"plus(?x:Nat,?y:Nat) = s(plus(?z:Nat,0))",
"plus(?x:Nat,?y:Nat) = s(plus(0,?y:Nat))",
"plus(?x:Nat,?y:Nat) = s(plus(0,?z:Nat))",
"plus(?x:Nat,?y:Nat) = plus(?x:Nat,s(0))",
"plus(?x:Nat,?z:Nat) = plus(0,s(?x:Nat))",
"plus(?x:Nat,0) = s(plus(?x:Nat,?z:Nat))",
"plus(?y:Nat,0) = plus(?y:Nat,s(?x:Nat))",
"plus(?y:Nat,0) = plus(s(?z:Nat),?y:Nat)",
"plus(?z:Nat,?x:Nat) = s(plus(?x:Nat,0))",
"plus(?z:Nat,0) = plus(s(?y:Nat),?y:Nat)",
"plus(0,?x:Nat) = plus(?x:Nat,s(?x:Nat))",
"plus(0,?y:Nat) = s(s(s(0)))",
"plus(0,?y:Nat) = plus(?z:Nat,s(?x:Nat))",
"s(plus(?x:Nat,?x:Nat)) = plus(?z:Nat,0)",
"s(plus(?x:Nat,?y:Nat)) = plus(?z:Nat,0)",
"s(plus(?x:Nat,0)) = plus(?x:Nat,?x:Nat)",
"s(plus(?y:Nat,?y:Nat)) = plus(0,?y:Nat)",
"plus(?x:Nat,s(0)) = plus(?x:Nat,?x:Nat)",
"plus(?x:Nat,s(0)) = plus(?z:Nat,?y:Nat)",
"plus(?y:Nat,s(?x:Nat)) = plus(0,?y:Nat)",
"plus(?y:Nat,s(?y:Nat)) = plus(0,?x:Nat)",
"plus(?y:Nat,s(0)) = plus(?x:Nat,?x:Nat)",
"plus(?y:Nat,s(0)) = plus(?x:Nat,?y:Nat)",
"plus(0,s(0)) = plus(?x:Nat,?x:Nat)",
"plus(s(?x:Nat),?x:Nat) = plus(0,?y:Nat)",
"plus(s(?x:Nat),0) = plus(?x:Nat,?x:Nat)",
"plus(s(?y:Nat),0) = plus(?x:Nat,?y:Nat)",
"plus(s(?z:Nat),?x:Nat) = plus(0,?x:Nat)",
"plus(s(?z:Nat),0) = plus(?x:Nat,?y:Nat)",
"plus(s(0),?x:Nat) = plus(?z:Nat,?x:Nat)",
"plus(s(0),?x:Nat) = plus(0,?y:Nat)",
"plus(s(0),?y:Nat) = plus(?x:Nat,0)",
"plus(?z:Nat,plus(?y:Nat,?y:Nat)) = s(0)",
"plus(0,s(s(?y:Nat))) = s(?y:Nat)",
"plus(plus(?y:Nat,0),?y:Nat) = s(0)",
"plus(plus(?z:Nat,?y:Nat),?x:Nat) = s(0)",
"plus(plus(0,?y:Nat),0) = s(?y:Nat)",
"plus(plus(0,0),?x:Nat) = s(?z:Nat)",
"s(?x:Nat) = plus(s(plus(?z:Nat,?y:Nat)),0)",
"s(?x:Nat) = plus(plus(s(0),?x:Nat),?z:Nat)",
"s(?y:Nat) = plus(0,s(plus(?x:Nat,?x:Nat)))",
"s(?y:Nat) = plus(s(s(0)),s(?x:Nat))",
"s(?y:Nat) = plus(plus(?x:Nat,s(0)),?z:Nat)",
"s(0) = plus(plus(?z:Nat,?x:Nat),s(?y:Nat))",
"s(s(?x:Nat)) = plus(?y:Nat,plus(?x:Nat,0))",
"s(s(?x:Nat)) = plus(s(?y:Nat),s(0))",
"plus(?x:Nat,?x:Nat) = plus(s(?x:Nat),s(0))",
"plus(?x:Nat,0) = plus(?z:Nat,plus(?x:Nat,?y:Nat))",
"plus(?y:Nat,?y:Nat) = plus(s(s(?x:Nat)),0)",
"plus(?y:Nat,0) = plus(?z:Nat,s(s(?x:Nat)))",
"plus(?y:Nat,0) = plus(s(s(?x:Nat)),?x:Nat)",
"plus(?z:Nat,?z:Nat) = plus(s(?x:Nat),s(0))",
"plus(?z:Nat,?z:Nat) = plus(plus(0,?y:Nat),0)",
"plus(0,?y:Nat) = plus(?z:Nat,plus(?x:Nat,?y:Nat))",
"plus(0,?y:Nat) = plus(plus(?x:Nat,?z:Nat),?y:Nat)",
"plus(0,?y:Nat) = plus(plus(?z:Nat,?x:Nat),?x:Nat)",
"plus(0,?z:Nat) = plus(s(s(0)),?z:Nat)",
"s(plus(?y:Nat,0)) = plus(s(?x:Nat),?x:Nat)",
"s(plus(?z:Nat,?z:Nat)) = plus(s(0),?x:Nat)",
"s(plus(0,?x:Nat)) = plus(?z:Nat,s(?z:Nat))",
"plus(0,s(?x:Nat)) = s(plus(?y:Nat,?x:Nat))",
"plus(s(?x:Nat),0) = plus(s(?y:Nat),?x:Nat)",
"plus(s(?y:Nat),?x:Nat) = plus(s(0),?x:Nat)",
"s(plus(s(?x:Nat),?y:Nat)) = plus(?z:Nat,0)",
"plus(?x:Nat,plus(?x:Nat,0)) = plus(?y:Nat,?x:Nat)",
"plus(?x:Nat,plus(?y:Nat,0)) = plus(?x:Nat,0)",
"plus(?x:Nat,plus(0,?y:Nat)) = s(s(?z:Nat))",
"plus(?y:Nat,s(s(?y:Nat))) = plus(?x:Nat,0)",
"plus(?z:Nat,plus(?y:Nat,?y:Nat)) = plus(0,?x:Nat)",
"plus(s(0),s(?z:Nat)) = plus(?z:Nat,?x:Nat)",
"plus(s(s(?z:Nat)),0) = s(s(?y:Nat))",
"plus(plus(?x:Nat,?x:Nat),?x:Nat) = plus(?y:Nat,0)",
"plus(plus(?x:Nat,0),?y:Nat) = s(s(?z:Nat))",
"plus(plus(?y:Nat,?z:Nat),?x:Nat) = s(s(0))",
"plus(plus(?z:Nat,?x:Nat),0) = plus(?y:Nat,?y:Nat)",
"plus(plus(?z:Nat,0),?x:Nat) = plus(?x:Nat,?x:Nat)",
"plus(plus(0,?y:Nat),0) = plus(?y:Nat,?z:Nat)",
"plus(?x:Nat,plus(s(0),?z:Nat)) = s(?y:Nat)",
"plus(?y:Nat,s(plus(0,?x:Nat))) = s(0)",
"plus(?z:Nat,s(plus(?y:Nat,?z:Nat))) = s(0)",
"plus(?z:Nat,s(plus(?y:Nat,0))) = s(?x:Nat)",
"plus(0,s(plus(0,?z:Nat))) = s(?y:Nat)",
"plus(s(?y:Nat),s(s(?x:Nat))) = s(0)",
"plus(s(?y:Nat),plus(0,?z:Nat)) = s(?x:Nat)",
"plus(s(s(0)),s(0)) = s(?x:Nat)",
"plus(s(plus(?y:Nat,?y:Nat)),?y:Nat) = s(0)",
"plus(plus(?y:Nat,s(?x:Nat)),?y:Nat) = s(0)",
"plus(plus(s(0),?y:Nat),?y:Nat) = s(?y:Nat)",
"s(?x:Nat) = plus(s(?x:Nat),plus(s(?y:Nat),0))",
"s(?x:Nat) = plus(plus(0,0),plus(?x:Nat,?y:Nat))",
"s(?y:Nat) = plus(s(?x:Nat),s(plus(0,?y:Nat)))",
"s(?z:Nat) = plus(0,plus(plus(?x:Nat,0),?y:Nat))",
"s(0) = plus(?y:Nat,plus(plus(?y:Nat,?y:Nat),?x:Nat))",
"s(0) = plus(s(plus(s(0),?x:Nat)),?x:Nat)",
"s(0) = plus(plus(?y:Nat,plus(?y:Nat,?y:Nat)),?x:Nat)",
"plus(?x:Nat,0) = s(plus(?x:Nat,s(s(?x:Nat))))",
"plus(?y:Nat,?x:Nat) = plus(?x:Nat,s(s(s(0))))",
"plus(?y:Nat,?y:Nat) = plus(s(plus(?y:Nat,0)),?x:Nat)",
"plus(?y:Nat,?z:Nat) = plus(s(plus(?y:Nat,0)),?x:Nat)",
"plus(0,?x:Nat) = s(s(s(plus(?y:Nat,?x:Nat))))",
"s(plus(?x:Nat,0)) = plus(?x:Nat,plus(?z:Nat,?y:Nat))",
"s(plus(?y:Nat,0)) = plus(plus(?x:Nat,?x:Nat),?y:Nat)",
"s(plus(0,?x:Nat)) = plus(plus(?x:Nat,?y:Nat),?x:Nat)",
"s(plus(0,?z:Nat)) = plus(plus(?y:Nat,?x:Nat),?x:Nat)",
"plus(?x:Nat,s(?x:Nat)) = plus(0,s(s(?y:Nat)))",
"plus(?x:Nat,s(?x:Nat)) = plus(s(0),s(0))",
"plus(s(0),?z:Nat) = plus(plus(?y:Nat,?x:Nat),?y:Nat)",
"s(plus(s(?x:Nat),0)) = plus(s(?y:Nat),?y:Nat)",
"s(plus(s(0),?x:Nat)) = plus(s(?x:Nat),?x:Nat)",
"plus(?x:Nat,plus(?x:Nat,?y:Nat)) = plus(s(?x:Nat),0)",
"plus(?y:Nat,s(s(0))) = plus(s(?y:Nat),?y:Nat)",
"plus(?y:Nat,plus(?y:Nat,?z:Nat)) = plus(s(?z:Nat),0)",
"plus(s(0),s(?z:Nat)) = s(plus(?z:Nat,?z:Nat))",
"plus(plus(?z:Nat,?x:Nat),?y:Nat) = plus(0,s(?y:Nat))",
"plus(plus(?z:Nat,?z:Nat),?x:Nat) = s(plus(0,?z:Nat))",
"s(s(s(plus(0,?y:Nat)))) = plus(?y:Nat,?y:Nat)",
"plus(0,plus(?z:Nat,s(?x:Nat))) = plus(?x:Nat,?x:Nat)",
"plus(plus(?z:Nat,?z:Nat),s(?y:Nat)) = plus(0,?x:Nat)",
"plus(plus(0,?z:Nat),s(?y:Nat)) = plus(?x:Nat,?z:Nat)",
"plus(plus(s(0),?x:Nat),?y:Nat) = plus(0,?x:Nat)",
"plus(plus(s(0),?x:Nat),?z:Nat) = s(s(?x:Nat))",
"plus(?x:Nat,plus(s(s(?x:Nat)),?z:Nat)) = s(0)",
"plus(?y:Nat,plus(s(s(?x:Nat)),?x:Nat)) = s(0)",
"plus(?z:Nat,plus(s(s(?x:Nat)),?z:Nat)) = s(0)",
"plus(0,plus(plus(?z:Nat,?x:Nat),?x:Nat)) = s(0)",
"plus(s(?x:Nat),plus(0,s(?z:Nat))) = s(?y:Nat)",
"plus(s(plus(0,0)),s(?x:Nat)) = s(0)",
"plus(plus(?y:Nat,plus(0,?x:Nat)),?x:Nat) = s(?x:Nat)",
"s(?x:Nat) = plus(plus(plus(s(?x:Nat),?x:Nat),?x:Nat),0)",
"s(0) = plus(s(plus(?x:Nat,plus(?z:Nat,?x:Nat))),?z:Nat)",
"plus(?x:Nat,?y:Nat) = plus(plus(?x:Nat,?y:Nat),plus(?y:Nat,0))",
"plus(?x:Nat,0) = plus(plus(?y:Nat,?y:Nat),s(s(0)))",
"plus(?y:Nat,0) = plus(s(?x:Nat),plus(s(?x:Nat),0))",
"plus(?z:Nat,s(0)) = plus(?x:Nat,plus(?x:Nat,s(?z:Nat)))",
"plus(0,s(?z:Nat)) = plus(s(?y:Nat),plus(?y:Nat,?x:Nat))",
"plus(s(?z:Nat),0) = plus(?y:Nat,plus(?x:Nat,s(?z:Nat)))",
"plus(?x:Nat,s(s(?y:Nat))) = plus(?z:Nat,plus(0,?z:Nat))",
"plus(?x:Nat,plus(?x:Nat,?y:Nat)) = plus(plus(?y:Nat,0),?x:Nat)",
"plus(?x:Nat,plus(0,?y:Nat)) = plus(plus(?y:Nat,?y:Nat),?z:Nat)",
"plus(?y:Nat,s(s(0))) = plus(s(s(?x:Nat)),?x:Nat)",
"plus(?y:Nat,s(s(0))) = plus(s(s(?y:Nat)),?z:Nat)",
"plus(?y:Nat,plus(?z:Nat,?z:Nat)) = plus(plus(?z:Nat,0),?y:Nat)",
"plus(s(s(0)),?y:Nat) = s(plus(?y:Nat,s(?y:Nat)))",
"plus(plus(?x:Nat,?x:Nat),?z:Nat) = plus(plus(0,?x:Nat),?y:Nat)",
"plus(plus(?x:Nat,0),?x:Nat) = s(plus(s(?z:Nat),?y:Nat))",
"plus(plus(?y:Nat,?x:Nat),0) = plus(0,plus(?y:Nat,?z:Nat))",
"plus(plus(?y:Nat,?x:Nat),0) = plus(plus(?x:Nat,?y:Nat),?y:Nat)",
"plus(plus(0,?x:Nat),0) = plus(?z:Nat,s(s(0)))",
"s(plus(?x:Nat,plus(?x:Nat,?z:Nat))) = plus(0,s(?x:Nat))",
"plus(s(0),plus(?x:Nat,?x:Nat)) = s(plus(?y:Nat,?y:Nat))",
"plus(plus(?z:Nat,?y:Nat),s(?x:Nat)) = plus(0,s(?z:Nat))",
"plus(plus(0,?y:Nat),s(?y:Nat)) = plus(0,s(?z:Nat))",
"plus(plus(0,?z:Nat),s(?z:Nat)) = plus(?x:Nat,s(?y:Nat))",
"plus(plus(s(?x:Nat),?x:Nat),?x:Nat) = plus(s(0),?x:Nat)",
"s(s(plus(plus(?x:Nat,0),?x:Nat))) = plus(?x:Nat,?y:Nat)",
"plus(?x:Nat,plus(plus(?y:Nat,0),?x:Nat)) = plus(?y:Nat,?y:Nat)",
"plus(?y:Nat,s(s(s(s(0))))) = plus(?y:Nat,?z:Nat)",
"plus(s(?x:Nat),plus(?y:Nat,s(?x:Nat))) = plus(?x:Nat,0)",
"plus(?y:Nat,plus(plus(s(0),?x:Nat),?y:Nat)) = s(?z:Nat)",
"plus(s(0),plus(s(?y:Nat),s(?z:Nat))) = s(?x:Nat)",
"plus(s(plus(?x:Nat,s(s(?y:Nat)))),0) = s(0)",
"plus(plus(?y:Nat,s(plus(?x:Nat,?x:Nat))),0) = s(0)",
"s(?x:Nat) = plus(plus(plus(?y:Nat,s(?y:Nat)),0),s(?z:Nat))",
"s(?y:Nat) = plus(s(s(plus(?y:Nat,plus(?x:Nat,?x:Nat)))),0)",
"s(?y:Nat) = plus(plus(s(?y:Nat),s(s(?y:Nat))),s(0))",
"s(?y:Nat) = plus(plus(plus(plus(0,?z:Nat),?y:Nat),?z:Nat),?z:Nat)",
"s(0) = plus(0,plus(?z:Nat,plus(plus(?z:Nat,?x:Nat),?x:Nat)))",
"s(s(?y:Nat)) = plus(plus(0,?y:Nat),plus(s(?z:Nat),?z:Nat))",
"s(s(?z:Nat)) = plus(s(?x:Nat),plus(plus(?x:Nat,0),?z:Nat))",
"plus(?x:Nat,?x:Nat) = plus(?y:Nat,plus(?x:Nat,plus(0,s(0))))",
"plus(?x:Nat,?x:Nat) = plus(s(plus(plus(?x:Nat,?y:Nat),0)),?x:Nat)",
"plus(?x:Nat,?x:Nat) = plus(plus(?y:Nat,plus(0,s(?y:Nat))),?z:Nat)",
"plus(?y:Nat,0) = plus(plus(plus(?z:Nat,s(?y:Nat)),?y:Nat),?x:Nat)",
"plus(?z:Nat,?z:Nat) = plus(plus(?y:Nat,?y:Nat),s(plus(0,?x:Nat)))",
"plus(0,?y:Nat) = plus(0,plus(?x:Nat,s(plus(?y:Nat,?x:Nat))))",
"plus(0,?z:Nat) = plus(?y:Nat,s(s(plus(s(?y:Nat),0))))",
"s(plus(?z:Nat,?y:Nat)) = plus(plus(s(0),?y:Nat),s(?y:Nat))",
"plus(s(?z:Nat),0) = plus(plus(plus(?x:Nat,?x:Nat),?x:Nat),?y:Nat)",
"plus(s(?z:Nat),s(?y:Nat)) = plus(?x:Nat,s(plus(?x:Nat,0)))",
"plus(plus(?x:Nat,?z:Nat),?x:Nat) = plus(?x:Nat,s(s(s(0))))",
"s(s(plus(s(?y:Nat),?x:Nat))) = plus(?z:Nat,s(s(0)))",
"plus(s(?z:Nat),plus(?x:Nat,?y:Nat)) = plus(?x:Nat,plus(?z:Nat,0))",
"plus(plus(s(?x:Nat),?z:Nat),0) = plus(?z:Nat,s(s(?x:Nat)))",
"plus(plus(s(0),0),?x:Nat) = s(s(plus(?y:Nat,?z:Nat)))",
"s(s(plus(0,plus(?y:Nat,?z:Nat)))) = plus(s(?x:Nat),?z:Nat)",
"s(plus(plus(s(0),0),?y:Nat)) = plus(s(?x:Nat),?x:Nat)",
"plus(s(?x:Nat),s(plus(?y:Nat,0))) = plus(s(?z:Nat),?y:Nat)",
"plus(plus(s(?x:Nat),s(0)),?x:Nat) = s(plus(?y:Nat,?x:Nat))",
"s(plus(?x:Nat,s(s(plus(?z:Nat,?x:Nat))))) = plus(0,?x:Nat)",
"s(plus(plus(s(s(?y:Nat)),?x:Nat),?y:Nat)) = plus(?x:Nat,0)",
"plus(?x:Nat,plus(s(?y:Nat),plus(0,?x:Nat))) = plus(?z:Nat,?y:Nat)",
"plus(?z:Nat,plus(?x:Nat,s(plus(?y:Nat,?z:Nat)))) = s(s(0))",
"plus(0,plus(s(s(?x:Nat)),s(?x:Nat))) = s(s(0))",
"plus(s(s(?z:Nat)),s(s(s(0)))) = plus(?x:Nat,?y:Nat)",
"plus(plus(s(0),plus(?x:Nat,?y:Nat)),?y:Nat) = plus(?z:Nat,?z:Nat)",
"plus(plus(plus(0,?x:Nat),s(?x:Nat)),0) = plus(?y:Nat,0)",
"plus(?x:Nat,s(plus(plus(?x:Nat,s(?x:Nat)),0))) = s(?y:Nat)",
"plus(?x:Nat,plus(s(plus(s(?x:Nat),?y:Nat)),0)) = s(?x:Nat)",
"plus(s(plus(s(s(?x:Nat)),?x:Nat)),s(?y:Nat)) = s(0)",
"plus(s(plus(s(plus(?x:Nat,?z:Nat)),?x:Nat)),0) = s(?z:Nat)",
"plus(s(plus(plus(?x:Nat,s(?x:Nat)),0)),?y:Nat) = s(?y:Nat)",
"plus(plus(?x:Nat,s(s(plus(?z:Nat,?z:Nat)))),0) = s(?y:Nat)",
"plus(plus(?x:Nat,s(plus(s(0),?y:Nat))),?x:Nat) = s(?z:Nat)",
"plus(plus(?y:Nat,s(?y:Nat)),s(plus(0,?x:Nat))) = s(?x:Nat)",
"plus(plus(?y:Nat,plus(plus(?z:Nat,?x:Nat),0)),?z:Nat) = s(?y:Nat)",
"plus(plus(?z:Nat,plus(?y:Nat,plus(?x:Nat,0))),?x:Nat) = s(?x:Nat)",
"s(?x:Nat) = plus(plus(plus(?y:Nat,0),s(?z:Nat)),plus(?x:Nat,?x:Nat))",
"s(?y:Nat) = plus(?y:Nat,plus(plus(?y:Nat,s(plus(?z:Nat,0))),?y:Nat))",
"s(?y:Nat) = plus(s(s(plus(s(?z:Nat),plus(?x:Nat,?z:Nat)))),0)",
"s(0) = plus(?y:Nat,plus(?x:Nat,plus(plus(s(?z:Nat),?x:Nat),?z:Nat)))",
"s(0) = plus(s(?x:Nat),s(s(plus(s(?x:Nat),s(?x:Nat)))))",
"s(s(?y:Nat)) = plus(plus(plus(plus(?y:Nat,?y:Nat),?z:Nat),0),?y:Nat)",
"plus(?x:Nat,?z:Nat) = plus(plus(s(plus(?x:Nat,?z:Nat)),0),s(?y:Nat))",
"plus(?y:Nat,?x:Nat) = plus(?z:Nat,plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),0))",
"plus(?z:Nat,?x:Nat) = plus(plus(plus(0,?y:Nat),?y:Nat),s(s(?z:Nat)))",
"plus(?z:Nat,?y:Nat) = plus(plus(0,?z:Nat),plus(?x:Nat,plus(?x:Nat,?z:Nat)))",
"plus(0,s(?x:Nat)) = plus(?x:Nat,plus(s(plus(?x:Nat,?z:Nat)),?x:Nat))",
"s(plus(?z:Nat,s(0))) = plus(?z:Nat,plus(0,plus(?x:Nat,?x:Nat)))",
"plus(?y:Nat,s(s(0))) = plus(plus(?x:Nat,?z:Nat),plus(?y:Nat,?z:Nat))",
"plus(plus(?y:Nat,0),?y:Nat) = plus(?x:Nat,plus(plus(?y:Nat,?x:Nat),?z:Nat))",
"s(s(plus(?y:Nat,s(?x:Nat)))) = plus(s(plus(0,?y:Nat)),?x:Nat)",
"s(s(plus(0,plus(?x:Nat,?x:Nat)))) = plus(plus(?x:Nat,?x:Nat),?z:Nat)",
"plus(?z:Nat,s(s(plus(?x:Nat,0)))) = plus(?y:Nat,plus(?x:Nat,?x:Nat))",
"plus(0,s(s(plus(?y:Nat,?z:Nat)))) = plus(plus(?x:Nat,?x:Nat),?z:Nat)",
"plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,0)) = s(plus(?x:Nat,s(?z:Nat)))",
"plus(plus(?x:Nat,plus(?x:Nat,0)),?z:Nat) = plus(plus(?x:Nat,?y:Nat),?y:Nat)",
"plus(plus(?z:Nat,?x:Nat),plus(?x:Nat,?x:Nat)) = plus(?z:Nat,plus(?y:Nat,0))",
"plus(plus(plus(?x:Nat,0),?y:Nat),?y:Nat) = s(s(s(s(?x:Nat))))",
"plus(s(0),s(plus(?x:Nat,s(?z:Nat)))) = plus(s(?x:Nat),?x:Nat)",
"s(plus(s(?z:Nat),s(plus(s(?z:Nat),0)))) = plus(0,?y:Nat)",
"plus(?x:Nat,s(plus(?y:Nat,plus(s(0),0)))) = s(s(?y:Nat))",
"plus(0,plus(?x:Nat,s(plus(s(?y:Nat),?x:Nat)))) = plus(?x:Nat,?z:Nat)",
"plus(s(plus(plus(?x:Nat,s(0)),?x:Nat)),?z:Nat) = s(s(?z:Nat))",
"plus(plus(0,s(s(plus(?y:Nat,?y:Nat)))),?x:Nat) = plus(?z:Nat,?z:Nat)",
"plus(plus(?x:Nat,plus(s(plus(0,?x:Nat)),?x:Nat)),?z:Nat) = s(?z:Nat)",
"plus(plus(s(?x:Nat),plus(?z:Nat,plus(?z:Nat,?z:Nat))),0) = s(?x:Nat)",
"plus(plus(s(plus(0,plus(?x:Nat,?x:Nat))),?z:Nat),?x:Nat) = s(?x:Nat)",
"s(?x:Nat) = plus(plus(0,plus(?x:Nat,?y:Nat)),plus(plus(?x:Nat,?y:Nat),?y:Nat))",
"s(?y:Nat) = plus(s(s(s(plus(?z:Nat,?y:Nat)))),plus(0,s(?y:Nat)))",
"s(?z:Nat) = plus(0,s(plus(plus(?x:Nat,plus(?y:Nat,?x:Nat)),s(?x:Nat))))",
"s(?z:Nat) = plus(plus(s(plus(?z:Nat,?y:Nat)),s(0)),plus(?y:Nat,?x:Nat))",
"s(0) = plus(?y:Nat,plus(plus(?x:Nat,s(plus(?x:Nat,0))),s(?x:Nat)))",
"s(s(?y:Nat)) = plus(plus(?z:Nat,plus(s(plus(?x:Nat,?z:Nat)),0)),?z:Nat)",
"s(s(?z:Nat)) = plus(plus(0,plus(?x:Nat,s(plus(?y:Nat,?x:Nat)))),?x:Nat)",
"s(s(0)) = plus(?x:Nat,plus(plus(s(plus(?y:Nat,0)),?y:Nat),?x:Nat))",
"plus(?x:Nat,?x:Nat) = plus(?x:Nat,plus(plus(s(plus(?y:Nat,0)),?x:Nat),?x:Nat))",
"plus(?y:Nat,?y:Nat) = s(plus(plus(?y:Nat,plus(plus(?x:Nat,0),?x:Nat)),?y:Nat))",
"plus(?y:Nat,?z:Nat) = plus(plus(s(plus(plus(?y:Nat,?x:Nat),?y:Nat)),?y:Nat),0)",
"plus(?y:Nat,0) = s(plus(plus(s(?y:Nat),?x:Nat),plus(?x:Nat,s(?x:Nat))))",
"plus(?y:Nat,0) = plus(plus(?z:Nat,0),s(s(s(plus(0,?y:Nat)))))",
"plus(0,?z:Nat) = plus(?y:Nat,plus(plus(?x:Nat,?z:Nat),s(plus(?y:Nat,?x:Nat))))",
"plus(0,0) = plus(plus(plus(plus(?x:Nat,?x:Nat),s(0)),?z:Nat),?x:Nat)",
"plus(?z:Nat,s(?x:Nat)) = plus(plus(s(s(plus(?z:Nat,0))),?y:Nat),0)",
"plus(?x:Nat,plus(0,?x:Nat)) = plus(plus(s(plus(?y:Nat,?y:Nat)),?y:Nat),?x:Nat)",
"plus(s(s(?y:Nat)),?y:Nat) = plus(0,plus(s(plus(?y:Nat,?x:Nat)),?y:Nat))",
"plus(plus(?z:Nat,?y:Nat),?x:Nat) = plus(plus(s(s(s(?y:Nat))),0),0)",
"plus(plus(s(?z:Nat),?z:Nat),?z:Nat) = plus(plus(?y:Nat,0),plus(?x:Nat,?x:Nat))",
"s(plus(plus(?z:Nat,s(0)),?y:Nat)) = plus(s(?y:Nat),s(s(?z:Nat)))",
"plus(plus(?x:Nat,plus(?x:Nat,0)),?x:Nat) = plus(plus(?y:Nat,s(?y:Nat)),?x:Nat)",
"plus(plus(plus(?y:Nat,?x:Nat),?x:Nat),?y:Nat) = s(plus(?y:Nat,plus(0,?x:Nat)))",
"plus(s(s(s(?x:Nat))),plus(0,s(?z:Nat))) = plus(?z:Nat,s(?z:Nat))",
"plus(plus(0,s(?z:Nat)),s(s(s(?y:Nat)))) = plus(s(0),?z:Nat)",
"plus(plus(plus(?z:Nat,0),plus(?x:Nat,?y:Nat)),?x:Nat) = plus(?x:Nat,s(?y:Nat))",
"plus(plus(plus(plus(?y:Nat,?y:Nat),?z:Nat),?x:Nat),?x:Nat) = s(plus(?x:Nat,0))",
"s(s(plus(plus(s(?z:Nat),plus(?x:Nat,?z:Nat)),?y:Nat))) = plus(0,0)",
"s(plus(s(?z:Nat),plus(?z:Nat,s(plus(?x:Nat,?y:Nat))))) = plus(0,?x:Nat)",
"s(plus(plus(?x:Nat,?x:Nat),plus(plus(?x:Nat,?x:Nat),?z:Nat))) = plus(0,?x:Nat)",
"s(plus(plus(plus(?x:Nat,plus(?z:Nat,0)),?x:Nat),?x:Nat)) = plus(?x:Nat,?y:Nat)",
"plus(s(s(plus(plus(?z:Nat,?y:Nat),?x:Nat))),s(?x:Nat)) = plus(?z:Nat,0)",
"plus(s(plus(?z:Nat,plus(s(?x:Nat),s(?z:Nat)))),0) = plus(?x:Nat,?x:Nat)",
"plus(plus(plus(?y:Nat,s(?x:Nat)),plus(?x:Nat,0)),?y:Nat) = plus(?x:Nat,?y:Nat)",
"plus(plus(?x:Nat,s(?x:Nat)),plus(s(plus(?x:Nat,0)),?y:Nat)) = s(?x:Nat)",
"plus(plus(?x:Nat,plus(?z:Nat,plus(?z:Nat,plus(?y:Nat,?y:Nat)))),0) = s(?x:Nat)",
"plus(plus(s(?x:Nat),?y:Nat),plus(s(plus(0,?x:Nat)),0)) = s(?z:Nat)",
"plus(plus(plus(?x:Nat,0),?x:Nat),plus(plus(?y:Nat,?y:Nat),0)) = s(0)",
"plus(plus(plus(?x:Nat,plus(?y:Nat,plus(0,?z:Nat))),?y:Nat),?x:Nat) = s(?z:Nat)",
"plus(plus(plus(plus(?x:Nat,?x:Nat),0),?x:Nat),s(s(?y:Nat))) = s(?y:Nat)",
"s(?x:Nat) = plus(s(plus(s(plus(?x:Nat,?x:Nat)),s(plus(0,?x:Nat)))),?z:Nat)",
"s(0) = plus(?y:Nat,plus(plus(s(plus(?x:Nat,?x:Nat)),?z:Nat),plus(?x:Nat,?x:Nat)))",
"plus(?x:Nat,?x:Nat) = s(plus(plus(s(s(?y:Nat)),s(plus(?z:Nat,?x:Nat))),0))",
"plus(?x:Nat,?x:Nat) = plus(plus(s(0),?z:Nat),s(plus(?x:Nat,plus(?x:Nat,?z:Nat))))",
"plus(?x:Nat,?z:Nat) = plus(plus(?y:Nat,plus(plus(?x:Nat,plus(?z:Nat,0)),?x:Nat)),?z:Nat)",
"plus(?x:Nat,0) = plus(plus(plus(s(?x:Nat),s(plus(?y:Nat,?x:Nat))),?x:Nat),?z:Nat)",
"plus(?y:Nat,?y:Nat) = plus(s(s(plus(plus(?z:Nat,?z:Nat),?x:Nat))),plus(?x:Nat,0))",
"plus(?y:Nat,?z:Nat) = plus(plus(s(plus(?x:Nat,0)),?z:Nat),plus(?y:Nat,s(?y:Nat)))",
"plus(?z:Nat,?y:Nat) = plus(?y:Nat,plus(?x:Nat,plus(?x:Nat,plus(plus(?y:Nat,0),?x:Nat))))",
"s(plus(0,?x:Nat)) = plus(s(s(plus(?y:Nat,?y:Nat))),s(s(s(?x:Nat))))",
"plus(?z:Nat,s(?x:Nat)) = plus(?z:Nat,plus(?x:Nat,s(plus(?x:Nat,plus(0,?x:Nat)))))",
"plus(s(?x:Nat),?x:Nat) = plus(s(plus(plus(?x:Nat,plus(?y:Nat,?x:Nat)),?x:Nat)),0)",
"plus(s(?x:Nat),0) = s(s(plus(plus(?z:Nat,?x:Nat),s(plus(?x:Nat,?z:Nat)))))",
"plus(?x:Nat,plus(?x:Nat,?x:Nat)) = plus(plus(s(0),?z:Nat),s(plus(?z:Nat,?z:Nat)))",
"s(plus(plus(?y:Nat,?z:Nat),0)) = plus(?z:Nat,plus(plus(?x:Nat,?x:Nat),s(0)))",
"plus(?x:Nat,s(s(s(?x:Nat)))) = s(s(plus(0,plus(s(?y:Nat),?y:Nat))))",
"plus(s(?x:Nat),plus(?y:Nat,?x:Nat)) = plus(s(s(plus(0,?z:Nat))),s(0))",
"plus(?x:Nat,plus(?y:Nat,s(s(s(0))))) = s(plus(plus(?y:Nat,?y:Nat),0))",
"plus(s(?x:Nat),plus(plus(?z:Nat,?x:Nat),0)) = s(plus(?y:Nat,s(s(?x:Nat))))",
"s(s(plus(s(plus(?y:Nat,?x:Nat)),s(?y:Nat)))) = plus(plus(?x:Nat,0),?z:Nat)",
"plus(0,plus(plus(s(?x:Nat),?y:Nat),s(?y:Nat))) = plus(?y:Nat,plus(?y:Nat,?z:Nat))",
"plus(plus(s(plus(?x:Nat,0)),s(?y:Nat)),0) = s(s(plus(0,?z:Nat)))",
"plus(?x:Nat,plus(plus(?y:Nat,s(s(?y:Nat))),s(0))) = s(plus(?y:Nat,?y:Nat))",
"plus(plus(?z:Nat,plus(s(?x:Nat),?x:Nat)),plus(0,?x:Nat)) = plus(?x:Nat,s(?x:Nat))",
"s(plus(s(0),plus(plus(plus(?x:Nat,?y:Nat),?z:Nat),?z:Nat))) = plus(?x:Nat,?y:Nat)",
"plus(?y:Nat,plus(plus(s(0),s(plus(0,?z:Nat))),?x:Nat)) = plus(?z:Nat,?x:Nat)",
"plus(?z:Nat,plus(s(s(plus(?y:Nat,?y:Nat))),plus(?y:Nat,0))) = plus(?x:Nat,?y:Nat)",
"plus(s(?x:Nat),plus(plus(?y:Nat,s(plus(?z:Nat,0))),?z:Nat)) = plus(?z:Nat,?x:Nat)",
"plus(plus(?y:Nat,?x:Nat),plus(plus(s(?y:Nat),s(?y:Nat)),0)) = plus(?z:Nat,?z:Nat)",
"plus(?x:Nat,s(plus(plus(plus(?x:Nat,0),?y:Nat),plus(?x:Nat,?x:Nat)))) = s(?y:Nat)",
"plus(plus(plus(?z:Nat,?x:Nat),s(0)),s(plus(?z:Nat,s(?y:Nat)))) = s(?z:Nat)",
"plus(plus(plus(s(0),?y:Nat),s(plus(s(?z:Nat),?x:Nat))),?z:Nat) = s(?y:Nat)",
"s(?y:Nat) = plus(plus(plus(?x:Nat,0),plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),0)),?y:Nat)",
"s(?y:Nat) = plus(plus(plus(?x:Nat,plus(plus(?y:Nat,0),plus(?x:Nat,?x:Nat))),?x:Nat),?z:Nat)",
"s(?z:Nat) = plus(0,plus(plus(?z:Nat,plus(plus(?x:Nat,?z:Nat),?x:Nat)),plus(?x:Nat,?y:Nat)))",
"s(0) = plus(?x:Nat,plus(plus(plus(?x:Nat,plus(?y:Nat,?z:Nat)),plus(?z:Nat,?y:Nat)),?x:Nat))",
"s(s(?x:Nat)) = plus(plus(plus(?y:Nat,?z:Nat),?x:Nat),plus(?x:Nat,plus(0,s(?x:Nat))))",
"plus(?x:Nat,?x:Nat) = plus(s(plus(plus(s(?z:Nat),0),plus(?z:Nat,?z:Nat))),s(?x:Nat))",
"plus(?x:Nat,?z:Nat) = plus(plus(?y:Nat,plus(0,plus(s(0),plus(?z:Nat,?z:Nat)))),?z:Nat)",
"plus(?x:Nat,0) = s(s(plus(plus(?x:Nat,s(?y:Nat)),plus(?x:Nat,plus(?z:Nat,0)))))",
"plus(?y:Nat,?z:Nat) = plus(?x:Nat,plus(plus(s(s(?x:Nat)),plus(?y:Nat,0)),s(0)))",
"s(plus(?x:Nat,?x:Nat)) = plus(?y:Nat,plus(plus(plus(0,?y:Nat),plus(?x:Nat,?x:Nat)),?y:Nat))",
"s(plus(?z:Nat,?x:Nat)) = plus(?x:Nat,plus(plus(?z:Nat,plus(plus(?z:Nat,?z:Nat),?x:Nat)),0))",
"plus(s(s(?y:Nat)),0) = plus(plus(plus(plus(?x:Nat,s(?x:Nat)),?x:Nat),?z:Nat),?z:Nat)",
"plus(plus(0,?y:Nat),?y:Nat) = plus(plus(plus(?z:Nat,s(?x:Nat)),plus(?z:Nat,?x:Nat)),?x:Nat)",
"s(plus(plus(0,?y:Nat),?y:Nat)) = plus(s(s(plus(s(?y:Nat),s(0)))),?z:Nat)",
"plus(?x:Nat,plus(s(?x:Nat),0)) = plus(?x:Nat,plus(plus(s(s(?y:Nat)),?x:Nat),?x:Nat))",
"plus(0,plus(s(?y:Nat),?z:Nat)) = plus(s(plus(s(s(?z:Nat)),s(?x:Nat))),?x:Nat)",
"plus(s(s(?x:Nat)),s(?x:Nat)) = s(plus(plus(0,plus(s(?z:Nat),?y:Nat)),?x:Nat))",
"plus(plus(?y:Nat,s(0)),?x:Nat) = plus(s(plus(plus(?z:Nat,s(?y:Nat)),?z:Nat)),?y:Nat)",
"s(plus(plus(plus(?y:Nat,0),?z:Nat),?x:Nat)) = plus(plus(s(s(?x:Nat)),?z:Nat),?y:Nat)",
"plus(?x:Nat,plus(?x:Nat,plus(plus(0,?y:Nat),?x:Nat))) = plus(s(plus(?x:Nat,?y:Nat)),?z:Nat)",
"s(plus(?x:Nat,plus(s(s(?x:Nat)),plus(?x:Nat,0)))) = plus(plus(?x:Nat,?x:Nat),?y:Nat)",
"plus(?y:Nat,plus(plus(0,s(s(?z:Nat))),s(?y:Nat))) = s(plus(s(?x:Nat),?y:Nat))",
"plus(?z:Nat,plus(s(0),plus(?z:Nat,plus(?y:Nat,?x:Nat)))) = plus(s(s(?y:Nat)),?y:Nat)",
"plus(s(plus(?z:Nat,s(plus(?x:Nat,?y:Nat)))),s(?x:Nat)) = plus(plus(?x:Nat,?z:Nat),0)",
"plus(plus(plus(?x:Nat,0),s(?y:Nat)),plus(?y:Nat,?x:Nat)) = s(plus(?y:Nat,s(?x:Nat)))",
"s(plus(s(?y:Nat),plus(?y:Nat,plus(plus(?y:Nat,?z:Nat),?z:Nat)))) = plus(s(?x:Nat),0)",
"plus(s(plus(plus(plus(?x:Nat,?y:Nat),?x:Nat),0)),s(?x:Nat)) = plus(s(0),?x:Nat)",
"plus(s(plus(0,plus(plus(?y:Nat,?z:Nat),plus(?z:Nat,?z:Nat)))),?y:Nat) = s(s(?z:Nat))",
"plus(s(plus(plus(?z:Nat,?y:Nat),plus(?y:Nat,plus(?x:Nat,?x:Nat)))),?x:Nat) = plus(0,?y:Nat)",
"plus(plus(?x:Nat,s(plus(plus(?x:Nat,?y:Nat),plus(?z:Nat,?y:Nat)))),?z:Nat) = plus(?x:Nat,0)",
"plus(plus(?y:Nat,s(0)),s(plus(plus(?x:Nat,s(?x:Nat)),?x:Nat))) = s(s(?z:Nat))",
"plus(plus(0,plus(plus(?y:Nat,?z:Nat),plus(?y:Nat,s(?y:Nat)))),?x:Nat) = plus(?z:Nat,?x:Nat)",
"plus(plus(?z:Nat,?z:Nat),plus(s(?z:Nat),s(plus(?y:Nat,plus(?y:Nat,0))))) = s(?x:Nat)",
"plus(?x:Nat,?y:Nat) = plus(s(?x:Nat),plus(s(plus(plus(?x:Nat,?y:Nat),plus(?x:Nat,?x:Nat))),0))",
"plus(?x:Nat,0) = plus(plus(?y:Nat,plus(0,plus(s(?x:Nat),?y:Nat))),plus(?x:Nat,s(?x:Nat)))",
"plus(?y:Nat,?x:Nat) = plus(s(?y:Nat),s(plus(plus(plus(?z:Nat,?x:Nat),plus(0,?x:Nat)),?y:Nat)))",
"plus(?y:Nat,?y:Nat) = plus(?x:Nat,plus(plus(plus(plus(?y:Nat,?x:Nat),plus(?y:Nat,?z:Nat)),0),?y:Nat))",
"plus(?y:Nat,?y:Nat) = plus(s(?x:Nat),plus(plus(s(?y:Nat),0),plus(plus(0,?y:Nat),0)))",
"plus(0,0) = s(plus(plus(plus(plus(0,?z:Nat),?y:Nat),s(plus(?x:Nat,?x:Nat))),?z:Nat))",
"s(s(s(0))) = plus(s(s(plus(plus(?y:Nat,?x:Nat),?z:Nat))),plus(s(?x:Nat),?x:Nat))",
"plus(?z:Nat,s(?y:Nat)) = plus(plus(plus(?y:Nat,s(?z:Nat)),plus(0,s(s(?y:Nat)))),?x:Nat)",
"plus(s(?y:Nat),?z:Nat) = plus(plus(plus(?x:Nat,s(?z:Nat)),s(0)),s(plus(?x:Nat,?y:Nat)))",
"s(s(plus(?y:Nat,?y:Nat))) = plus(s(0),s(plus(plus(plus(?z:Nat,?z:Nat),?y:Nat),?y:Nat)))",
"plus(?x:Nat,s(s(?z:Nat))) = plus(?x:Nat,plus(plus(s(s(0)),?y:Nat),plus(?y:Nat,?z:Nat)))",
"plus(plus(?x:Nat,0),?z:Nat) = plus(s(s(plus(s(?y:Nat),?x:Nat))),plus(s(?z:Nat),?x:Nat))",
"plus(s(plus(?y:Nat,0)),?x:Nat) = plus(plus(s(plus(?x:Nat,?y:Nat)),plus(?y:Nat,0)),?x:Nat)",
"plus(plus(s(?x:Nat),?x:Nat),?x:Nat) = plus(?x:Nat,plus(plus(plus(s(?z:Nat),?x:Nat),0),?x:Nat))",
"plus(s(s(s(?z:Nat))),s(?x:Nat)) = s(plus(s(s(0)),plus(?x:Nat,s(?x:Nat))))",
"plus(s(plus(s(?x:Nat),?y:Nat)),?y:Nat) = plus(plus(plus(?z:Nat,0),?x:Nat),plus(?y:Nat,0))",
"s(plus(s(plus(s(?z:Nat),?y:Nat)),?x:Nat)) = plus(plus(plus(?x:Nat,?y:Nat),0),s(?x:Nat))",
"plus(?x:Nat,plus(plus(?y:Nat,0),s(?y:Nat))) = plus(s(?y:Nat),plus(s(?x:Nat),s(?x:Nat)))",
"plus(0,plus(plus(s(?x:Nat),?y:Nat),?x:Nat)) = plus(0,plus(s(plus(?y:Nat,?x:Nat)),?y:Nat))",
"plus(s(plus(plus(?z:Nat,?z:Nat),?y:Nat)),0) = plus(?x:Nat,plus(plus(?y:Nat,?x:Nat),s(?y:Nat)))",
"plus(?z:Nat,plus(s(s(?y:Nat)),plus(?x:Nat,0))) = plus(?y:Nat,plus(?x:Nat,plus(?x:Nat,?x:Nat)))",
"plus(s(s(plus(?y:Nat,s(?z:Nat)))),s(?x:Nat)) = plus(0,plus(plus(?z:Nat,?x:Nat),?y:Nat))",
"s(plus(plus(s(0),s(plus(?y:Nat,?z:Nat))),?x:Nat)) = plus(s(?x:Nat),plus(?x:Nat,?y:Nat))",
"plus(?x:Nat,s(plus(plus(?z:Nat,?y:Nat),plus(0,?x:Nat)))) = plus(plus(s(?z:Nat),?z:Nat),?y:Nat)",
"plus(plus(plus(0,?x:Nat),?y:Nat),plus(?y:Nat,s(?x:Nat))) = plus(plus(?x:Nat,?y:Nat),s(?z:Nat))",
"plus(s(plus(plus(?x:Nat,?x:Nat),s(plus(?y:Nat,0)))),plus(?y:Nat,?x:Nat)) = plus(?z:Nat,?y:Nat)",
"plus(s(plus(plus(?y:Nat,?x:Nat),0)),plus(plus(?x:Nat,?y:Nat),s(0))) = plus(?z:Nat,?x:Nat)",
"plus(s(plus(plus(?z:Nat,plus(?x:Nat,?z:Nat)),plus(s(0),0))),?x:Nat) = plus(?y:Nat,?x:Nat)",
"plus(?y:Nat,plus(s(s(?x:Nat)),plus(?z:Nat,plus(plus(?x:Nat,0),s(?z:Nat))))) = s(?y:Nat)",
"plus(plus(plus(plus(0,?y:Nat),plus(plus(?x:Nat,?x:Nat),?x:Nat)),?x:Nat),s(?x:Nat)) = s(?y:Nat)",
"plus(0,?z:Nat) = plus(?x:Nat,plus(plus(plus(plus(0,0),plus(?x:Nat,?z:Nat)),?x:Nat),s(?y:Nat)))",
"s(plus(?y:Nat,?y:Nat)) = plus(plus(plus(?z:Nat,?x:Nat),0),s(s(plus(plus(?y:Nat,?x:Nat),0))))",
"s(plus(?y:Nat,?z:Nat)) = plus(plus(s(plus(s(?z:Nat),?z:Nat)),plus(?x:Nat,0)),plus(?y:Nat,?x:Nat))",
"s(plus(0,?z:Nat)) = plus(s(plus(plus(plus(?y:Nat,0),s(?z:Nat)),s(s(?x:Nat)))),?y:Nat)",
"plus(s(?x:Nat),?z:Nat) = plus(plus(plus(?z:Nat,plus(?z:Nat,0)),?z:Nat),s(s(s(s(?x:Nat)))))",
"plus(plus(?y:Nat,?y:Nat),?z:Nat) = plus(plus(?z:Nat,plus(plus(0,?x:Nat),?z:Nat)),plus(s(?y:Nat),0))",
"plus(s(?x:Nat),plus(?x:Nat,?x:Nat)) = plus(0,plus(s(?y:Nat),plus(?z:Nat,plus(s(?x:Nat),0))))",
"s(s(s(s(plus(?y:Nat,?x:Nat))))) = plus(plus(?z:Nat,s(plus(?x:Nat,plus(0,?x:Nat)))),?y:Nat)",
"plus(?x:Nat,s(s(plus(?x:Nat,plus(?y:Nat,?y:Nat))))) = plus(0,plus(0,plus(s(?z:Nat),?y:Nat)))",
"plus(plus(?y:Nat,plus(0,s(plus(?y:Nat,?y:Nat)))),?z:Nat) = plus(?z:Nat,plus(?z:Nat,plus(?y:Nat,?x:Nat)))",
"plus(?x:Nat,plus(plus(0,0),plus(s(s(0)),?y:Nat))) = s(plus(plus(?x:Nat,?x:Nat),?x:Nat))",
"plus(?z:Nat,s(plus(plus(plus(0,?z:Nat),?y:Nat),s(?x:Nat)))) = plus(?y:Nat,plus(s(?y:Nat),?y:Nat))",
"plus(plus(?x:Nat,plus(?x:Nat,plus(0,plus(?y:Nat,?x:Nat)))),?y:Nat) = plus(0,s(s(s(?z:Nat))))",
"plus(0,plus(?y:Nat,plus(s(plus(?x:Nat,?y:Nat)),s(s(?x:Nat))))) = plus(?z:Nat,plus(?y:Nat,?z:Nat))",
"plus(0,plus(plus(plus(plus(?x:Nat,?y:Nat),?z:Nat),?z:Nat),s(?y:Nat))) = s(plus(s(?y:Nat),?y:Nat))",
"plus(plus(plus(?z:Nat,?x:Nat),0),plus(plus(plus(?z:Nat,?x:Nat),s(s(?y:Nat))),?x:Nat)) = s(?z:Nat)",
"s(s(0)) = plus(plus(plus(?x:Nat,plus(plus(0,?z:Nat),plus(0,0))),plus(?x:Nat,?x:Nat)),0)",
"plus(?x:Nat,?y:Nat) = s(plus(?y:Nat,plus(plus(s(?y:Nat),s(0)),plus(s(?z:Nat),plus(?z:Nat,?x:Nat)))))",
"plus(?z:Nat,s(?y:Nat)) = plus(?y:Nat,plus(plus(?x:Nat,?z:Nat),plus(0,plus(plus(0,?x:Nat),s(?z:Nat)))))",
"plus(0,s(?y:Nat)) = plus(plus(?z:Nat,?x:Nat),plus(plus(?x:Nat,plus(plus(?z:Nat,?y:Nat),s(?y:Nat))),?z:Nat))",
"plus(plus(?y:Nat,0),0) = plus(?y:Nat,plus(plus(?y:Nat,plus(?y:Nat,plus(?z:Nat,?z:Nat))),plus(?y:Nat,?x:Nat)))",
"plus(plus(?y:Nat,plus(s(?x:Nat),?z:Nat)),?x:Nat) = plus(plus(s(?x:Nat),s(s(plus(?y:Nat,0)))),?x:Nat)",
"plus(?y:Nat,plus(s(plus(s(?x:Nat),?y:Nat)),?x:Nat)) = s(plus(?y:Nat,plus(?x:Nat,plus(0,s(?z:Nat)))))",
"plus(plus(?y:Nat,0),plus(s(s(?x:Nat)),?y:Nat)) = s(plus(plus(?x:Nat,s(?z:Nat)),plus(?y:Nat,?x:Nat)))",
"plus(plus(?x:Nat,?z:Nat),plus(?x:Nat,s(plus(0,?y:Nat)))) = plus(?y:Nat,s(s(s(plus(?x:Nat,?y:Nat)))))",
"plus(plus(?y:Nat,?y:Nat),plus(s(plus(s(?z:Nat),0)),0)) = plus(plus(plus(?x:Nat,?x:Nat),?z:Nat),?x:Nat)",
"plus(?x:Nat,plus(plus(plus(?z:Nat,0),plus(?y:Nat,?x:Nat)),s(s(?z:Nat)))) = s(plus(s(?x:Nat),?z:Nat))",
"plus(s(?z:Nat),plus(?y:Nat,plus(plus(s(0),plus(?y:Nat,?y:Nat)),?x:Nat))) = plus(?y:Nat,plus(?x:Nat,?z:Nat))",
"plus(s(plus(?x:Nat,plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,?z:Nat)))),s(?x:Nat)) = plus(?x:Nat,plus(?y:Nat,0))",
"plus(?y:Nat,plus(plus(s(?x:Nat),plus(plus(?y:Nat,0),plus(?y:Nat,?x:Nat))),?x:Nat)) = plus(?y:Nat,s(?y:Nat))",
"plus(plus(?y:Nat,?z:Nat),plus(s(?x:Nat),plus(plus(s(?x:Nat),s(?z:Nat)),?y:Nat))) = s(plus(?x:Nat,0))",
"plus(plus(?y:Nat,s(plus(s(?x:Nat),?y:Nat))),plus(0,s(plus(plus(?y:Nat,?y:Nat),?z:Nat)))) = s(0)",
"plus(?y:Nat,0) = plus(s(plus(plus(?x:Nat,plus(?z:Nat,?y:Nat)),0)),plus(?x:Nat,plus(s(s(?y:Nat)),?x:Nat)))",
"plus(plus(?y:Nat,?z:Nat),0) = s(plus(?x:Nat,plus(plus(plus(?y:Nat,?x:Nat),plus(0,?x:Nat)),plus(?x:Nat,0))))",
"s(plus(0,s(plus(?x:Nat,?x:Nat)))) = plus(plus(plus(0,plus(plus(0,?x:Nat),?x:Nat)),?x:Nat),s(?x:Nat))",
"plus(plus(s(plus(?z:Nat,?x:Nat)),0),plus(?z:Nat,?y:Nat)) = plus(plus(s(?y:Nat),s(s(s(?z:Nat)))),?x:Nat)",
"s(s(plus(plus(plus(?x:Nat,0),?x:Nat),plus(s(?z:Nat),s(?z:Nat))))) = plus(plus(?x:Nat,s(?x:Nat)),?y:Nat)",
"plus(?x:Nat,plus(plus(?y:Nat,plus(?x:Nat,plus(0,?z:Nat))),plus(plus(?y:Nat,s(?y:Nat)),?x:Nat))) = plus(?z:Nat,?y:Nat)",
"plus(?x:Nat,plus(plus(plus(?y:Nat,plus(?x:Nat,0)),?y:Nat),plus(s(?z:Nat),plus(?z:Nat,0)))) = s(s(?x:Nat))",
"plus(s(?x:Nat),plus(plus(?z:Nat,plus(?y:Nat,?y:Nat)),s(plus(s(0),plus(?x:Nat,?y:Nat))))) = s(s(?x:Nat))",
"plus(plus(plus(?z:Nat,plus(plus(?x:Nat,0),s(?z:Nat))),plus(plus(?y:Nat,?y:Nat),?y:Nat)),?z:Nat) = plus(?y:Nat,?y:Nat)",
"plus(?z:Nat,plus(s(plus(s(?x:Nat),plus(?x:Nat,?y:Nat))),plus(0,plus(plus(?z:Nat,?x:Nat),?y:Nat)))) = s(0)",
"plus(plus(plus(?y:Nat,?x:Nat),plus(plus(plus(?x:Nat,?y:Nat),plus(?y:Nat,?z:Nat)),s(s(?z:Nat)))),?x:Nat) = s(0)",
"s(?z:Nat) = plus(?y:Nat,plus(plus(s(?x:Nat),plus(s(?y:Nat),plus(?x:Nat,?z:Nat))),plus(plus(?x:Nat,s(0)),?y:Nat)))",
"s(?z:Nat) = plus(plus(plus(plus(s(?x:Nat),s(0)),?z:Nat),s(plus(plus(?z:Nat,?x:Nat),plus(0,?z:Nat)))),?x:Nat)",
"plus(?x:Nat,0) = plus(plus(?x:Nat,s(?x:Nat)),plus(plus(?x:Nat,?z:Nat),plus(s(?y:Nat),plus(?y:Nat,plus(?z:Nat,?z:Nat)))))",
"plus(s(?x:Nat),0) = plus(plus(plus(s(?y:Nat),plus(s(?x:Nat),plus(?y:Nat,?z:Nat))),?x:Nat),s(plus(?y:Nat,?x:Nat)))",
"s(plus(plus(?z:Nat,?x:Nat),?y:Nat)) = plus(plus(s(plus(plus(0,?x:Nat),plus(?x:Nat,?y:Nat))),plus(?z:Nat,?x:Nat)),?x:Nat)",
"plus(?y:Nat,s(plus(0,?z:Nat))) = s(plus(plus(s(s(?y:Nat)),plus(?z:Nat,?y:Nat)),plus(?x:Nat,s(s(0)))))",
"plus(s(s(plus(?y:Nat,?y:Nat))),?x:Nat) = s(plus(plus(?z:Nat,s(?y:Nat)),plus(0,plus(?z:Nat,plus(?z:Nat,?x:Nat)))))",
"s(s(plus(plus(?y:Nat,?y:Nat),s(?y:Nat)))) = plus(?x:Nat,plus(plus(?y:Nat,plus(s(?x:Nat),?y:Nat)),plus(0,?z:Nat)))",
"s(plus(s(s(plus(?z:Nat,?x:Nat))),?y:Nat)) = plus(plus(s(s(plus(?x:Nat,0))),s(s(s(?x:Nat)))),?z:Nat)",
"plus(plus(0,plus(s(?x:Nat),?x:Nat)),?x:Nat) = plus(plus(plus(s(plus(?x:Nat,?y:Nat)),0),plus(?x:Nat,?z:Nat)),?y:Nat)",
"plus(plus(?y:Nat,s(s(plus(0,?x:Nat)))),?x:Nat) = plus(plus(plus(?y:Nat,?x:Nat),plus(?y:Nat,plus(?x:Nat,0))),?y:Nat)",
"plus(?y:Nat,s(plus(s(?y:Nat),s(plus(0,?x:Nat))))) = s(s(plus(s(0),plus(?x:Nat,plus(?y:Nat,?x:Nat)))))",
"s(plus(s(?x:Nat),plus(plus(s(?y:Nat),s(?y:Nat)),?y:Nat))) = plus(s(plus(0,plus(s(?y:Nat),?x:Nat))),0)",
"plus(?y:Nat,plus(plus(0,?z:Nat),s(s(plus(?x:Nat,?x:Nat))))) = plus(s(s(?x:Nat)),plus(plus(?x:Nat,0),?x:Nat))",
"plus(plus(?x:Nat,plus(s(plus(?z:Nat,?z:Nat)),?x:Nat)),plus(?z:Nat,?x:Nat)) = plus(plus(plus(s(?x:Nat),?z:Nat),?z:Nat),0)",
"plus(plus(s(plus(plus(?x:Nat,?x:Nat),?x:Nat)),s(?z:Nat)),plus(s(?y:Nat),?x:Nat)) = s(plus(?y:Nat,plus(?z:Nat,0)))",
"s(?x:Nat) = plus(plus(0,plus(s(plus(0,?x:Nat)),s(plus(?x:Nat,?z:Nat)))),plus(s(?x:Nat),s(s(s(?x:Nat)))))",
"plus(0,plus(0,?z:Nat)) = s(plus(plus(plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),plus(?z:Nat,plus(?x:Nat,?z:Nat))),?z:Nat))",
"plus(0,plus(s(s(?z:Nat)),?y:Nat)) = plus(plus(plus(s(?x:Nat),plus(s(?z:Nat),s(?x:Nat))),plus(?z:Nat,?x:Nat)),?y:Nat)",
"s(plus(s(plus(?x:Nat,?z:Nat)),s(?x:Nat))) = plus(?z:Nat,plus(plus(?y:Nat,?x:Nat),plus(?z:Nat,plus(?z:Nat,plus(?x:Nat,0)))))",
"plus(s(s(plus(?x:Nat,s(?y:Nat)))),?z:Nat) = plus(?z:Nat,plus(s(0),plus(plus(s(0),plus(?z:Nat,?x:Nat)),?z:Nat)))",
"plus(plus(?x:Nat,plus(?y:Nat,s(?y:Nat))),?x:Nat) = plus(plus(plus(?z:Nat,s(?x:Nat)),plus(?y:Nat,plus(s(?z:Nat),?x:Nat))),0)",
"plus(plus(plus(?x:Nat,s(s(?y:Nat))),?y:Nat),?z:Nat) = plus(plus(plus(?z:Nat,0),plus(?x:Nat,s(s(?x:Nat)))),s(?y:Nat))",
"plus(?x:Nat,plus(?x:Nat,plus(plus(plus(?y:Nat,?x:Nat),plus(?z:Nat,0)),?z:Nat))) = plus(plus(s(s(s(?x:Nat))),?x:Nat),?x:Nat)",
"plus(plus(?x:Nat,s(0)),plus(s(plus(plus(?z:Nat,?x:Nat),?x:Nat)),plus(?z:Nat,?x:Nat))) = plus(?x:Nat,plus(s(?x:Nat),?z:Nat))",
"plus(?z:Nat,plus(plus(plus(s(?y:Nat),plus(?z:Nat,?z:Nat)),plus(plus(?y:Nat,?z:Nat),plus(?y:Nat,?x:Nat))),?x:Nat)) = plus(0,?x:Nat)",
"s(?z:Nat) = plus(plus(plus(s(plus(?y:Nat,?z:Nat)),plus(s(?y:Nat),plus(?x:Nat,?y:Nat))),plus(?z:Nat,plus(s(?z:Nat),?x:Nat))),0)",
"plus(0,?y:Nat) = plus(?y:Nat,plus(s(plus(plus(?x:Nat,?x:Nat),?x:Nat)),plus(plus(?y:Nat,s(?y:Nat)),plus(?x:Nat,plus(?y:Nat,?x:Nat)))))",
"s(plus(plus(s(0),s(?x:Nat)),?y:Nat)) = plus(s(plus(plus(?z:Nat,?y:Nat),?y:Nat)),plus(?y:Nat,s(s(plus(?x:Nat,?z:Nat)))))",
"plus(plus(?y:Nat,plus(?x:Nat,s(s(0)))),?x:Nat) = plus(plus(plus(?z:Nat,?y:Nat),?x:Nat),plus(0,s(s(plus(?z:Nat,?x:Nat)))))",
"plus(s(?x:Nat),plus(plus(?z:Nat,plus(plus(0,?x:Nat),?x:Nat)),?z:Nat)) = s(plus(?x:Nat,plus(plus(?y:Nat,plus(?x:Nat,?z:Nat)),?y:Nat)))",
"plus(plus(s(plus(s(?x:Nat),0)),plus(?x:Nat,?x:Nat)),s(plus(?y:Nat,?y:Nat))) = s(s(plus(plus(?x:Nat,s(?y:Nat)),?x:Nat)))",
"plus(plus(plus(plus(?y:Nat,?y:Nat),?z:Nat),plus(?x:Nat,?y:Nat)),plus(s(s(s(?z:Nat))),s(?y:Nat))) = plus(plus(?x:Nat,0),?y:Nat)",
"plus(s(plus(?x:Nat,?z:Nat)),?z:Nat) = plus(plus(s(plus(?x:Nat,plus(?x:Nat,?x:Nat))),plus(0,?y:Nat)),plus(s(0),plus(?y:Nat,?z:Nat)))",
"plus(plus(s(plus(0,?x:Nat)),?x:Nat),?z:Nat) = plus(plus(plus(s(?x:Nat),plus(s(?x:Nat),?z:Nat)),?y:Nat),s(plus(s(?y:Nat),?z:Nat)))",
"plus(s(plus(?x:Nat,s(plus(?x:Nat,?y:Nat)))),s(?y:Nat)) = plus(plus(plus(?y:Nat,?x:Nat),s(s(plus(0,?x:Nat)))),plus(?y:Nat,?z:Nat))",
"plus(s(plus(plus(s(?y:Nat),?x:Nat),plus(?x:Nat,?y:Nat))),0) = plus(?y:Nat,plus(plus(0,plus(s(?x:Nat),?x:Nat)),plus(?y:Nat,?z:Nat)))",
"plus(plus(?x:Nat,plus(plus(?x:Nat,?x:Nat),plus(?y:Nat,?x:Nat))),?z:Nat) = plus(plus(s(plus(plus(0,?z:Nat),s(?y:Nat))),s(?x:Nat)),?x:Nat)",
"plus(plus(?y:Nat,plus(s(?x:Nat),plus(?y:Nat,s(0)))),s(?y:Nat)) = plus(?x:Nat,s(plus(s(s(0)),plus(?x:Nat,s(?x:Nat)))))",
"plus(plus(s(?z:Nat),plus(s(0),plus(?z:Nat,?z:Nat))),plus(?z:Nat,?y:Nat)) = plus(?x:Nat,plus(?y:Nat,plus(?z:Nat,plus(?x:Nat,s(?y:Nat)))))",
"plus(plus(plus(plus(?z:Nat,s(0)),0),plus(?x:Nat,plus(?x:Nat,plus(?x:Nat,?x:Nat)))),?y:Nat) = plus(s(?y:Nat),plus(s(?y:Nat),?x:Nat))",
"plus(plus(?x:Nat,plus(s(s(?y:Nat)),plus(?y:Nat,?z:Nat))),plus(s(plus(s(?x:Nat),0)),?x:Nat)) = plus(?x:Nat,plus(?x:Nat,s(?y:Nat)))",
"plus(plus(0,s(s(s(?x:Nat)))),plus(plus(plus(?y:Nat,plus(?y:Nat,?y:Nat)),plus(plus(?y:Nat,?x:Nat),?y:Nat)),?y:Nat)) = plus(?z:Nat,?x:Nat)",
"s(?y:Nat) = plus(plus(plus(plus(?y:Nat,?y:Nat),s(?z:Nat)),plus(plus(?x:Nat,s(0)),?z:Nat)),plus(?z:Nat,plus(?y:Nat,plus(s(?y:Nat),?x:Nat))))",
"plus(s(plus(?z:Nat,0)),?x:Nat) = plus(plus(plus(plus(?x:Nat,?x:Nat),plus(?z:Nat,?x:Nat)),?x:Nat),s(plus(?x:Nat,plus(plus(?x:Nat,?z:Nat),?x:Nat))))",
"plus(?y:Nat,plus(plus(0,plus(plus(?x:Nat,?y:Nat),?x:Nat)),?x:Nat)) = plus(?z:Nat,plus(s(plus(?x:Nat,plus(?x:Nat,?y:Nat))),plus(s(0),?z:Nat)))",
"s(plus(s(?z:Nat),plus(?y:Nat,plus(plus(0,?x:Nat),plus(?y:Nat,?x:Nat))))) = plus(plus(?z:Nat,s(?x:Nat)),plus(s(?z:Nat),s(s(?y:Nat))))",
"plus(plus(plus(s(plus(?y:Nat,?x:Nat)),s(?x:Nat)),plus(plus(?y:Nat,?x:Nat),s(?z:Nat))),s(?y:Nat)) = plus(s(?z:Nat),plus(0,s(?z:Nat)))",
"plus(?z:Nat,s(?x:Nat)) = plus(plus(plus(s(?z:Nat),plus(?x:Nat,?z:Nat)),plus(?y:Nat,0)),plus(?x:Nat,plus(plus(?x:Nat,?y:Nat),s(plus(?y:Nat,?x:Nat)))))",
"plus(?x:Nat,plus(?y:Nat,0)) = plus(plus(plus(plus(?y:Nat,s(?x:Nat)),s(?x:Nat)),plus(s(?y:Nat),?z:Nat)),s(plus(?x:Nat,s(plus(?x:Nat,?z:Nat)))))",
"plus(plus(plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),?y:Nat),plus(?z:Nat,?y:Nat)),?x:Nat) = plus(plus(?z:Nat,0),s(plus(plus(plus(?x:Nat,?y:Nat),?x:Nat),?z:Nat)))",
"plus(plus(s(plus(plus(?x:Nat,?x:Nat),?y:Nat)),0),plus(s(plus(plus(?y:Nat,?y:Nat),?x:Nat)),s(plus(?x:Nat,?x:Nat)))) = plus(plus(?x:Nat,?x:Nat),?x:Nat)",
"plus(plus(s(plus(?z:Nat,?z:Nat)),plus(plus(?x:Nat,0),s(plus(?y:Nat,?y:Nat)))),s(plus(?x:Nat,s(plus(?x:Nat,?z:Nat))))) = plus(plus(?x:Nat,?z:Nat),?z:Nat)",
"plus(plus(plus(s(?y:Nat),s(plus(0,?y:Nat))),?z:Nat),plus(plus(plus(?x:Nat,?y:Nat),plus(plus(?y:Nat,0),?y:Nat)),s(?x:Nat))) = plus(s(?y:Nat),?y:Nat)",
"plus(plus(plus(?x:Nat,?z:Nat),s(?z:Nat)),?x:Nat) = plus(plus(plus(?z:Nat,s(plus(?z:Nat,?z:Nat))),plus(?x:Nat,s(0))),plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),?x:Nat))",
"s(plus(0,plus(plus(?z:Nat,plus(?x:Nat,?x:Nat)),plus(plus(?x:Nat,?z:Nat),plus(?x:Nat,?x:Nat))))) = plus(plus(s(?y:Nat),plus(plus(?x:Nat,s(?x:Nat)),?x:Nat)),?z:Nat)",
"plus(plus(s(plus(plus(0,?y:Nat),s(?z:Nat))),plus(s(plus(?y:Nat,?x:Nat)),?z:Nat)),s(plus(?z:Nat,?x:Nat))) = plus(?x:Nat,plus(plus(s(?y:Nat),?x:Nat),?z:Nat))",
"plus(plus(s(plus(plus(?x:Nat,?y:Nat),?x:Nat)),s(plus(plus(?x:Nat,?x:Nat),?x:Nat))),plus(plus(?x:Nat,plus(?z:Nat,?y:Nat)),plus(?z:Nat,?x:Nat))) = s(plus(0,?z:Nat))",
"plus(plus(?y:Nat,plus(s(?x:Nat),s(?x:Nat))),?x:Nat) = plus(?x:Nat,plus(plus(plus(?y:Nat,?y:Nat),s(?x:Nat)),plus(plus(?x:Nat,?x:Nat),plus(plus(0,?y:Nat),s(?y:Nat)))))",
"plus(plus(?x:Nat,plus(plus(?y:Nat,0),s(0))),s(s(?x:Nat))) = plus(plus(s(plus(?y:Nat,plus(?x:Nat,?x:Nat))),plus(s(s(?x:Nat)),?y:Nat)),plus(?x:Nat,?x:Nat))",
"s(plus(s(plus(plus(?x:Nat,?y:Nat),plus(?x:Nat,?y:Nat))),plus(plus(?z:Nat,?x:Nat),?x:Nat))) = plus(plus(s(0),s(plus(?y:Nat,?z:Nat))),plus(plus(?y:Nat,?x:Nat),?x:Nat))",
"plus(s(plus(s(plus(?y:Nat,?z:Nat)),plus(s(0),plus(?y:Nat,?x:Nat)))),?z:Nat) = plus(plus(plus(s(plus(?x:Nat,?y:Nat)),s(?y:Nat)),plus(?y:Nat,plus(?y:Nat,?z:Nat))),?z:Nat)",
"plus(plus(?x:Nat,plus(plus(plus(?y:Nat,?z:Nat),plus(0,?y:Nat)),?z:Nat)),s(?y:Nat)) = plus(0,plus(plus(?x:Nat,plus(s(?z:Nat),s(?y:Nat))),s(s(plus(?z:Nat,?y:Nat)))))",
"plus(s(s(plus(?y:Nat,?x:Nat))),plus(plus(s(0),plus(?x:Nat,plus(?x:Nat,?y:Nat))),?z:Nat)) = plus(plus(plus(plus(?y:Nat,?x:Nat),s(?x:Nat)),s(s(?y:Nat))),s(?y:Nat))",
"plus(s(plus(s(s(?x:Nat)),plus(?z:Nat,s(?z:Nat)))),plus(?y:Nat,plus(s(0),plus(?x:Nat,?y:Nat)))) = plus(plus(?x:Nat,plus(?x:Nat,plus(0,plus(?x:Nat,?z:Nat)))),?x:Nat)",
"plus(plus(plus(?x:Nat,?y:Nat),?x:Nat),plus(?y:Nat,plus(?x:Nat,?y:Nat))) = plus(s(s(plus(?y:Nat,s(?y:Nat)))),s(plus(plus(plus(?y:Nat,?x:Nat),?x:Nat),plus(plus(0,?z:Nat),?z:Nat))))",
"s(plus(plus(s(plus(?x:Nat,0)),plus(s(?x:Nat),s(?x:Nat))),plus(plus(?z:Nat,?y:Nat),?x:Nat))) = plus(0,plus(0,plus(plus(?z:Nat,plus(?z:Nat,?z:Nat)),s(s(?y:Nat)))))",
"plus(?x:Nat,?z:Nat) = plus(plus(plus(s(plus(?z:Nat,?x:Nat)),s(s(?x:Nat))),?x:Nat),plus(plus(plus(0,plus(?x:Nat,?y:Nat)),s(?y:Nat)),plus(plus(?z:Nat,?y:Nat),plus(plus(?y:Nat,0),?x:Nat))))",
"plus(plus(?z:Nat,plus(?y:Nat,?x:Nat)),plus(plus(s(?x:Nat),plus(plus(?x:Nat,?y:Nat),?z:Nat)),plus(plus(?y:Nat,s(?x:Nat)),plus(?y:Nat,plus(?x:Nat,?y:Nat))))) = plus(plus(plus(?z:Nat,0),?x:Nat),?x:Nat)",
"s(plus(plus(plus(?y:Nat,s(?x:Nat)),0),plus(plus(?x:Nat,plus(?y:Nat,?z:Nat)),?z:Nat))) = plus(plus(?x:Nat,plus(plus(?x:Nat,s(?z:Nat)),s(s(?y:Nat)))),plus(plus(?x:Nat,?y:Nat),plus(?x:Nat,?z:Nat)))",
"plus(plus(s(?y:Nat),s(?x:Nat)),plus(?x:Nat,plus(plus(plus(?y:Nat,?y:Nat),?y:Nat),?x:Nat))) = plus(plus(s(?x:Nat),?y:Nat),plus(plus(plus(?z:Nat,?z:Nat),plus(?y:Nat,?z:Nat)),plus(0,plus(?x:Nat,?x:Nat))))",
"plus(plus(s(plus(?x:Nat,?y:Nat)),?x:Nat),s(plus(s(?x:Nat),plus(s(?y:Nat),?z:Nat)))) = plus(s(plus(plus(plus(?x:Nat,?x:Nat),?y:Nat),plus(plus(?y:Nat,?x:Nat),s(?z:Nat)))),plus(plus(plus(?y:Nat,plus(?x:Nat,?y:Nat)),0),plus(s(plus(?x:Nat,?x:Nat)),s(?y:Nat))))"]


val problemsListLenYes =
IOMstrs.rdEqs [
"rev(rev(rev(?ys:List))) = rev(?ys:List)",
"app(?ys:List,?ys:List) = rev(rev(app(?ys:List,?ys:List)))",
"cons(?x:Nat,?ys:List) = cons(?x:Nat,rev(rev(?ys:List)))",
"app(?ys:List,rev(rev(?zs:List))) = app(?ys:List,?zs:List)",
"app(rev(rev(?xs:List)),?ys:List) = app(?xs:List,?ys:List)",
"rev(rev(app(?zs:List,?ys:List))) = app(?zs:List,?ys:List)",
"rev(rev(cons(?z:Nat,?xs:List))) = cons(?z:Nat,?xs:List)",
"cons(0,?xs:List) = cons(times(0,plus(?x:Nat,?x:Nat)),?xs:List)",
"cons(0,?xs:List) = cons(times(plus(?x:Nat,?y:Nat),0),?xs:List)",
"rev(cons(0,?xs:List)) = rev(cons(times(0,?y:Nat),?xs:List))",
"app(?xs:List,rev(rev(?xs:List))) = rev(rev(app(?xs:List,?xs:List)))",
"app(?xs:List,rev(rev(?ys:List))) = app(rev(rev(?xs:List)),?ys:List)",
"rev(rev(cons(?y:Nat,?ys:List))) = cons(?y:Nat,rev(rev(?ys:List)))",
"rev(rev(cons(0,?ys:List))) = cons(times(0,?x:Nat),?ys:List)",
"cons(plus(0,len(?xs:List)),?ys:List) = cons(len(?xs:List),?ys:List)",
"cons(plus(times(0,?x:Nat),0),?ys:List) = rev(rev(cons(0,?ys:List)))",
"cons(plus(times(?x:Nat,0),plus(?x:Nat,0)),?xs:List) = cons(?x:Nat,?xs:List)",
"cons(len(cons(0,?ys:List)),?zs:List) = cons(plus(len(?ys:List),s(0)),?zs:List)",
"rev(rev(rev(?ys:List))) = rev(?ys:List)",
"app(?xs:List,?xs:List) = app(?xs:List,rev(rev(?xs:List)))",
"app(?xs:List,?ys:List) = app(rev(rev(?xs:List)),?ys:List)",
"app(?ys:List,?ys:List) = rev(rev(app(?ys:List,?ys:List)))",
"cons(?x:Nat,?xs:List) = cons(?x:Nat,rev(rev(?xs:List)))",
"cons(?z:Nat,?zs:List) = rev(rev(cons(?z:Nat,?zs:List)))",
"rev(rev(app(?xs:List,?zs:List))) = app(?xs:List,?zs:List)",
"cons(0,rev(rev(?xs:List))) = cons(0,?xs:List)",
"cons(plus(0,?x:Nat),?xs:List) = cons(?x:Nat,?xs:List)",
"cons(len(?ys:List),?xs:List) = cons(len(rev(?ys:List)),?xs:List)",
"cons(?z:Nat,?xs:List) = cons(plus(?z:Nat,times(?y:Nat,0)),?xs:List)",
"app(rev(rev(?xs:List)),?xs:List) = app(?xs:List,rev(rev(?xs:List)))",
"app(cons(?y:Nat,?xs:List),?xs:List) = cons(?y:Nat,app(?xs:List,?xs:List))",
"rev(rev(cons(0,?ys:List))) = cons(times(0,?x:Nat),?ys:List)",
"cons(0,?ys:List) = cons(times(?y:Nat,times(0,?x:Nat)),rev(rev(?ys:List)))",
"rev(rev(cons(0,?xs:List))) = cons(times(0,plus(?y:Nat,?x:Nat)),?xs:List)",
"app(rev(app(?xs:List,?xs:List)),?zs:List) = app(app(rev(?xs:List),rev(?xs:List)),?zs:List)",
"cons(?x:Nat,cons(0,rev(?xs:List))) = cons(?x:Nat,cons(times(?x:Nat,0),rev(?xs:List)))",
"rev(?ys:List) = rev(rev(rev(?ys:List)))",
"app(?zs:List,?xs:List) = app(?zs:List,rev(rev(?xs:List)))",
"rev(rev(?xs:List)) = rev(rev(rev(rev(?xs:List))))",
"rev(rev(app(?xs:List,?xs:List))) = app(?xs:List,?xs:List)",
"rev(rev(app(?xs:List,?ys:List))) = app(?xs:List,?ys:List)",
"rev(rev(cons(?x:Nat,?ys:List))) = cons(?x:Nat,?ys:List)",
"rev(rev(cons(0,?ys:List))) = cons(0,?ys:List)",
"app(rev(?xs:List),rev(?ys:List)) = rev(app(?ys:List,?xs:List))",
"cons(len(rev(rev(?xs:List))),?xs:List) = cons(len(?xs:List),?xs:List)",
"cons(times(0,len(rev(?ys:List))),?zs:List) = cons(0,?zs:List)",
"rev(rev(app(rev(?ys:List),rev(?ys:List)))) = rev(app(?ys:List,?ys:List))",
"app(cons(?x:Nat,?ys:List),?ys:List) = app(rev(rev(cons(?x:Nat,?ys:List))),?ys:List)",
"cons(?x:Nat,app(?xs:List,?xs:List)) = cons(?x:Nat,rev(rev(app(?xs:List,?xs:List))))",
"cons(times(times(len(?xs:List),plus(?y:Nat,?z:Nat)),0),?ys:List) = cons(0,?ys:List)",
"cons(plus(?y:Nat,plus(0,s(?x:Nat))),?xs:List) = cons(s(plus(?x:Nat,?y:Nat)),?xs:List)",
"cons(times(0,?x:Nat),?zs:List) = cons(times(times(plus(?z:Nat,?z:Nat),0),s(times(0,0))),?zs:List)",
"rev(?ys:List) = rev(rev(rev(?ys:List)))",
"app(?zs:List,?xs:List) = app(?zs:List,rev(rev(?xs:List)))",
"rev(rev(?xs:List)) = rev(rev(rev(rev(?xs:List))))",
"rev(rev(app(?xs:List,?xs:List))) = app(?xs:List,?xs:List)",
"rev(rev(app(?xs:List,?ys:List))) = app(?xs:List,?ys:List)",
"rev(rev(cons(?x:Nat,?ys:List))) = cons(?x:Nat,?ys:List)",
"rev(rev(cons(0,?ys:List))) = cons(0,?ys:List)",
"app(rev(?xs:List),rev(?ys:List)) = rev(app(?ys:List,?xs:List))",
"cons(len(rev(rev(?xs:List))),?xs:List) = cons(len(?xs:List),?xs:List)",
"cons(times(0,len(rev(?ys:List))),?zs:List) = cons(0,?zs:List)",
"rev(rev(app(rev(?ys:List),rev(?ys:List)))) = rev(app(?ys:List,?ys:List))",
"app(cons(?x:Nat,?ys:List),?ys:List) = app(rev(rev(cons(?x:Nat,?ys:List))),?ys:List)",
"cons(?x:Nat,app(?xs:List,?xs:List)) = cons(?x:Nat,rev(rev(app(?xs:List,?xs:List))))",
"cons(times(times(len(?xs:List),plus(?y:Nat,?z:Nat)),0),?ys:List) = cons(0,?ys:List)",
"cons(plus(?y:Nat,plus(0,s(?x:Nat))),?xs:List) = cons(s(plus(?x:Nat,?y:Nat)),?xs:List)",
"cons(times(0,?x:Nat),?zs:List) = cons(times(times(plus(?z:Nat,?z:Nat),0),s(times(0,0))),?zs:List)",
"rev(?ys:List) = rev(rev(rev(?ys:List)))",
"cons(?y:Nat,?ys:List) = cons(plus(?y:Nat,0),?ys:List)",
"rev(rev(app(?xs:List,?ys:List))) = app(?xs:List,?ys:List)",
"rev(rev(cons(?x:Nat,?ys:List))) = cons(?x:Nat,?ys:List)",
"cons(len(?xs:List),?zs:List) = cons(plus(len(?xs:List),0),?zs:List)",
"cons(times(0,0),?ys:List) = cons(times(0,?y:Nat),?ys:List)",
"rev(rev(cons(times(?x:Nat,0),?xs:List))) = cons(times(0,?x:Nat),?xs:List)",
"cons(times(len(cons(?y:Nat,?xs:List)),0),?ys:List) = cons(times(0,s(len(?zs:List))),?ys:List)",
"cons(len(cons(times(?x:Nat,?x:Nat),?ys:List)),?xs:List) = cons(plus(len(cons(?y:Nat,?ys:List)),0),?xs:List)",
"app(?ys:List,?ys:List) = app(rev(rev(?ys:List)),?ys:List)",
"app(?zs:List,?xs:List) = app(?zs:List,rev(rev(?xs:List)))",
"app(?ys:List,rev(rev(?ys:List))) = app(?ys:List,?ys:List)",
"rev(rev(app(?ys:List,?zs:List))) = app(?ys:List,?zs:List)",
"cons(?x:Nat,rev(rev(?xs:List))) = cons(?x:Nat,?xs:List)",
"cons(plus(0,0),?xs:List) = cons(0,?xs:List)",
"app(?ys:List,app(?xs:List,?xs:List)) = app(app(?ys:List,?xs:List),?xs:List)",
"app(?ys:List,app(?xs:List,?ys:List)) = app(app(?ys:List,?xs:List),?ys:List)",
"app(app(?xs:List,?xs:List),?xs:List) = app(?xs:List,app(?xs:List,?xs:List))",
"app(app(?xs:List,?xs:List),?xs:List) = app(?xs:List,rev(rev(app(?xs:List,?xs:List))))",
"rev(?zs:List) = rev(rev(rev(?zs:List)))",
"app(?ys:List,?zs:List) = rev(rev(app(?ys:List,?zs:List)))",
"app(?xs:List,rev(rev(?xs:List))) = app(?xs:List,?xs:List)",
"app(?xs:List,rev(rev(?ys:List))) = app(?xs:List,?ys:List)",
"app(rev(rev(?xs:List)),?xs:List) = app(?xs:List,?xs:List)",
"app(rev(rev(?ys:List)),?zs:List) = app(?ys:List,?zs:List)",
"rev(rev(app(?ys:List,?ys:List))) = app(?ys:List,?ys:List)",
"app(?ys:List,rev(?xs:List)) = rev(app(?xs:List,rev(?ys:List)))",
"rev(app(?xs:List,?ys:List)) = app(rev(?ys:List),rev(?xs:List))",
"rev(app(rev(?xs:List),?ys:List)) = app(rev(?ys:List),?xs:List)",
"app(rev(?xs:List),?ys:List) = app(rev(rev(rev(?xs:List))),?ys:List)",
"app(?xs:List,app(?xs:List,?xs:List)) = app(app(?xs:List,?xs:List),?xs:List)",
"app(?xs:List,app(?xs:List,?zs:List)) = app(app(?xs:List,?xs:List),?zs:List)",
"app(?ys:List,app(?xs:List,?ys:List)) = app(app(?ys:List,?xs:List),?ys:List)",
"app(app(?xs:List,?zs:List),?zs:List) = app(?xs:List,app(?zs:List,?zs:List))",
"app(app(?zs:List,?xs:List),?ys:List) = app(?zs:List,app(?xs:List,?ys:List))",
"app(?ys:List,rev(rev(app(?zs:List,?xs:List)))) = app(?ys:List,app(?zs:List,?xs:List))",
"app(rev(rev(?xs:List)),app(?ys:List,?xs:List)) = app(?xs:List,app(?ys:List,?xs:List))",
"app(?xs:List,app(app(?zs:List,?xs:List),?ys:List)) = app(?xs:List,app(?zs:List,app(?xs:List,?ys:List)))",
"app(app(?xs:List,?zs:List),app(?zs:List,?ys:List)) = app(app(app(?xs:List,?zs:List),?zs:List),?ys:List)",
"app(app(?xs:List,app(?ys:List,?xs:List)),?zs:List) = app(app(app(?xs:List,?ys:List),?xs:List),?zs:List)",
"app(app(?ys:List,app(?zs:List,?xs:List)),?zs:List) = app(?ys:List,app(?zs:List,app(?xs:List,?zs:List)))",
"app(?xs:List,app(?ys:List,app(?ys:List,app(?ys:List,?zs:List)))) = app(?xs:List,app(?ys:List,app(app(?ys:List,?ys:List),?zs:List)))",
"app(?zs:List,app(?ys:List,app(?ys:List,app(?ys:List,?ys:List)))) = app(app(app(app(?zs:List,?ys:List),?ys:List),?ys:List),?ys:List)",
"rev(rev(rev(?xs:List))) = rev(?xs:List)",
"app(?xs:List,?xs:List) = app(?xs:List,rev(rev(?xs:List)))",
"app(?ys:List,?xs:List) = app(?ys:List,rev(rev(?xs:List)))",
"app(rev(rev(?xs:List)),?xs:List) = app(?xs:List,?xs:List)",
"app(rev(rev(?xs:List)),?ys:List) = app(?xs:List,?ys:List)",
"rev(rev(app(?xs:List,?xs:List))) = app(?xs:List,?xs:List)",
"app(?xs:List,rev(?xs:List)) = rev(app(?xs:List,rev(?xs:List)))",
"app(rev(?xs:List),rev(?xs:List)) = rev(app(?xs:List,?xs:List))",
"rev(app(rev(?ys:List),?xs:List)) = app(rev(?xs:List),?ys:List)",
"rev(app(rev(?xs:List),rev(?xs:List))) = app(?xs:List,?xs:List)",
"app(?xs:List,app(?ys:List,?xs:List)) = app(app(?xs:List,?ys:List),?xs:List)",
"app(app(?xs:List,?xs:List),?xs:List) = app(?xs:List,app(?xs:List,?xs:List))",
"app(app(?xs:List,?xs:List),?ys:List) = app(?xs:List,app(?xs:List,?ys:List))",
"app(app(?ys:List,?xs:List),?xs:List) = app(?ys:List,app(?xs:List,?xs:List))",
"app(?xs:List,rev(rev(rev(?xs:List)))) = app(?xs:List,rev(?xs:List))",
"app(rev(rev(rev(?xs:List))),?xs:List) = app(rev(?xs:List),?xs:List)",
"rev(app(rev(rev(?xs:List)),?xs:List)) = rev(app(?xs:List,?xs:List))",
"app(?xs:List,app(?xs:List,?xs:List)) = app(app(?xs:List,rev(rev(?xs:List))),?xs:List)",
"app(?ys:List,app(?xs:List,?xs:List)) = app(?ys:List,app(?xs:List,rev(rev(?xs:List))))",
"app(app(?xs:List,?xs:List),?xs:List) = app(app(?xs:List,rev(rev(?xs:List))),?xs:List)",
"app(app(?xs:List,?xs:List),?xs:List) = app(app(rev(rev(?xs:List)),?xs:List),?xs:List)",
"app(app(?xs:List,?xs:List),?xs:List) = app(rev(rev(app(?xs:List,?xs:List))),?xs:List)",
"app(app(?xs:List,?xs:List),?ys:List) = app(app(?xs:List,rev(rev(?xs:List))),?ys:List)",
"app(app(?xs:List,?xs:List),?ys:List) = app(rev(rev(app(?xs:List,?xs:List))),?ys:List)",
"app(app(?xs:List,?xs:List),?ys:List) = rev(rev(app(app(?xs:List,?xs:List),?ys:List)))",
"app(app(?xs:List,?ys:List),?xs:List) = app(app(?xs:List,rev(rev(?ys:List))),?xs:List)",
"rev(app(app(?xs:List,?xs:List),?xs:List)) = rev(app(?xs:List,app(?xs:List,?xs:List)))",
"rev(rev(rev(app(?xs:List,?xs:List)))) = rev(app(rev(rev(?xs:List)),?xs:List))",
"app(?xs:List,app(rev(rev(?xs:List)),?xs:List)) = app(app(?xs:List,?xs:List),?xs:List)",
"app(?xs:List,rev(rev(app(?xs:List,?xs:List)))) = app(app(?xs:List,?xs:List),?xs:List)",
"app(app(?xs:List,?xs:List),rev(rev(?xs:List))) = app(?xs:List,app(?xs:List,?xs:List))",
"app(rev(rev(?xs:List)),app(?xs:List,?xs:List)) = app(app(?xs:List,?xs:List),?xs:List)",
"app(rev(rev(app(?xs:List,?xs:List))),?xs:List) = app(?xs:List,app(?xs:List,?xs:List))",
"rev(rev(app(?xs:List,app(?xs:List,?xs:List)))) = app(app(?xs:List,?xs:List),?xs:List)",
"rev(rev(app(?xs:List,app(?ys:List,?xs:List)))) = app(app(?xs:List,?ys:List),?xs:List)",
"rev(rev(app(?ys:List,app(?xs:List,?xs:List)))) = app(?ys:List,app(?xs:List,?xs:List))",
"rev(rev(app(app(?xs:List,?xs:List),?xs:List))) = app(?xs:List,app(?xs:List,?xs:List))",
"rev(rev(app(app(?ys:List,?xs:List),?xs:List))) = app(app(?ys:List,?xs:List),?xs:List)",
"app(app(?xs:List,?xs:List),?ys:List) = app(rev(app(rev(?xs:List),rev(?xs:List))),?ys:List)",
"app(?xs:List,app(rev(?xs:List),?xs:List)) = app(?xs:List,rev(app(rev(?xs:List),?xs:List)))",
"app(?xs:List,rev(app(rev(?xs:List),rev(?xs:List)))) = app(app(?xs:List,?xs:List),?xs:List)",
"rev(app(rev(app(?xs:List,?xs:List)),rev(?xs:List))) = app(?xs:List,app(?xs:List,?xs:List))",
"app(?xs:List,app(?xs:List,app(?xs:List,?xs:List))) = app(app(?xs:List,?xs:List),app(?xs:List,?xs:List))",
"app(?xs:List,app(?xs:List,app(?xs:List,?xs:List))) = app(app(?xs:List,app(?xs:List,?xs:List)),?xs:List)",
"app(?xs:List,app(?ys:List,app(?xs:List,?xs:List))) = app(app(?xs:List,?ys:List),app(?xs:List,?xs:List))",
"app(?xs:List,app(app(?xs:List,?xs:List),?xs:List)) = app(?xs:List,app(?xs:List,app(?xs:List,?xs:List)))",
"app(?xs:List,rev(app(?ys:List,rev(?xs:List)))) = rev(app(?ys:List,rev(app(?xs:List,?xs:List))))",
"app(?xs:List,rev(rev(app(?xs:List,?xs:List)))) = app(rev(rev(app(?xs:List,?xs:List))),?xs:List)",
"app(app(?xs:List,?xs:List),app(?ys:List,?xs:List)) = app(app(app(?xs:List,?xs:List),?ys:List),?xs:List)",
"app(app(?xs:List,app(?xs:List,?xs:List)),?xs:List) = app(?xs:List,app(app(?xs:List,?xs:List),?xs:List))",
"app(app(app(?xs:List,?xs:List),?xs:List),?xs:List) = app(?xs:List,app(app(?xs:List,?xs:List),?xs:List))",
"app(app(rev(rev(?xs:List)),?xs:List),?xs:List) = app(?xs:List,rev(rev(app(?xs:List,?xs:List))))",
"rev(rev(app(?xs:List,app(?xs:List,?xs:List)))) = app(?xs:List,rev(rev(app(?xs:List,?xs:List))))",
"rev(rev(app(app(?xs:List,?xs:List),?xs:List))) = rev(rev(app(?xs:List,app(?xs:List,?xs:List))))",
"app(?xs:List,app(app(?xs:List,?xs:List),?xs:List)) = app(app(rev(rev(?xs:List)),?xs:List),app(?xs:List,?xs:List))",
"app(app(app(?xs:List,?xs:List),?xs:List),?xs:List) = app(?xs:List,app(app(?xs:List,?xs:List),rev(rev(?xs:List))))",
"app(?xs:List,app(?xs:List,app(?xs:List,rev(?xs:List)))) = app(app(app(?xs:List,?xs:List),?xs:List),rev(?xs:List))",
"app(?ys:List,app(?xs:List,app(rev(?ys:List),?xs:List))) = app(app(app(?ys:List,?xs:List),rev(?ys:List)),?xs:List)",
"rev(app(app(?xs:List,?ys:List),app(?xs:List,?xs:List))) = rev(app(app(app(?xs:List,?ys:List),?xs:List),?xs:List))",
"rev(app(app(app(?xs:List,?xs:List),?xs:List),?xs:List)) = rev(app(app(?xs:List,app(?xs:List,?xs:List)),?xs:List))",
"app(?xs:List,app(app(?xs:List,?xs:List),rev(rev(?xs:List)))) = app(app(?xs:List,?xs:List),app(?xs:List,?xs:List))",
"app(rev(rev(app(?xs:List,?ys:List))),app(?xs:List,?xs:List)) = app(?xs:List,app(?ys:List,app(?xs:List,?xs:List)))",
"app(?xs:List,app(rev(app(?xs:List,?xs:List)),?ys:List)) = app(rev(app(app(?xs:List,?xs:List),rev(?xs:List))),?ys:List)",
"app(?xs:List,app(app(?xs:List,rev(?xs:List)),rev(?xs:List))) = app(?xs:List,app(?xs:List,rev(app(?xs:List,?xs:List))))",
"app(rev(?xs:List),rev(app(?ys:List,app(?xs:List,?xs:List)))) = rev(app(app(?ys:List,app(?xs:List,?xs:List)),?xs:List))",
"app(?xs:List,app(?xs:List,app(?xs:List,app(?xs:List,?xs:List)))) = app(app(app(?xs:List,app(?xs:List,?xs:List)),?xs:List),?xs:List)",
"app(?xs:List,app(app(?xs:List,?xs:List),app(?xs:List,?xs:List))) = app(app(?xs:List,app(?xs:List,app(?xs:List,?xs:List))),?xs:List)",
"app(?xs:List,app(app(?xs:List,app(?xs:List,?xs:List)),?xs:List)) = app(app(app(app(?xs:List,?xs:List),?xs:List),?xs:List),?xs:List)",
"app(?xs:List,app(app(?ys:List,app(?xs:List,?xs:List)),?xs:List)) = app(app(app(app(?xs:List,?ys:List),?xs:List),?xs:List),?xs:List)",
"app(app(?xs:List,app(?xs:List,app(?xs:List,?xs:List))),?xs:List) = app(app(app(app(?xs:List,?xs:List),?xs:List),?xs:List),?xs:List)",
"app(app(?xs:List,app(?xs:List,app(?xs:List,?ys:List))),?xs:List) = app(?xs:List,app(app(?xs:List,?xs:List),app(?ys:List,?xs:List)))",
"app(app(?xs:List,app(?xs:List,app(?ys:List,?xs:List))),?xs:List) = app(?xs:List,app(?xs:List,app(?ys:List,app(?xs:List,?xs:List))))",
"app(app(?xs:List,app(app(?xs:List,?xs:List),?xs:List)),?xs:List) = app(?xs:List,app(?xs:List,app(?xs:List,app(?xs:List,?xs:List))))",
"app(app(?xs:List,app(app(?ys:List,?xs:List),?xs:List)),?xs:List) = app(?xs:List,app(app(?ys:List,?xs:List),app(?xs:List,?xs:List)))",
"app(app(?ys:List,app(app(?xs:List,?xs:List),?xs:List)),?xs:List) = app(app(?ys:List,app(?xs:List,app(?xs:List,?xs:List))),?xs:List)",
"app(app(app(?xs:List,app(?xs:List,?xs:List)),?xs:List),?xs:List) = app(?xs:List,app(app(app(?xs:List,?xs:List),?xs:List),?xs:List))",
"app(app(app(?xs:List,app(?xs:List,?ys:List)),?xs:List),?xs:List) = app(?xs:List,app(?xs:List,app(?ys:List,app(?xs:List,?xs:List))))",
"app(app(app(?ys:List,?xs:List),?xs:List),app(?xs:List,?xs:List)) = app(?ys:List,app(app(app(?xs:List,?xs:List),?xs:List),?xs:List))",
"app(app(app(?ys:List,?xs:List),app(?xs:List,?xs:List)),?xs:List) = app(?ys:List,app(app(app(?xs:List,?xs:List),?xs:List),?xs:List))",
"app(app(app(app(?xs:List,?xs:List),?xs:List),?xs:List),?xs:List) = app(?xs:List,app(?xs:List,app(?xs:List,app(?xs:List,?xs:List))))",
"app(app(app(app(?ys:List,?xs:List),?xs:List),?xs:List),?xs:List) = app(?ys:List,app(app(app(?xs:List,?xs:List),?xs:List),?xs:List))",
"app(?xs:List,app(app(app(?xs:List,?xs:List),?xs:List),?xs:List)) = app(?xs:List,app(app(?xs:List,app(?xs:List,?xs:List)),rev(rev(?xs:List))))",
"app(app(?xs:List,app(app(?xs:List,?xs:List),rev(?xs:List))),?xs:List) = app(app(?xs:List,?xs:List),app(?xs:List,app(rev(?xs:List),?xs:List)))",
"app(app(rev(?xs:List),app(app(?xs:List,?xs:List),?xs:List)),?xs:List) = app(rev(?xs:List),app(?xs:List,app(app(?xs:List,?xs:List),?xs:List)))",
"app(app(app(?ys:List,app(?xs:List,?xs:List)),?xs:List),rev(rev(?xs:List))) = app(app(app(app(?ys:List,?xs:List),?xs:List),?xs:List),?xs:List)",
"app(?xs:List,app(app(?xs:List,?xs:List),app(app(?xs:List,?xs:List),?xs:List))) = app(app(?xs:List,?xs:List),app(?xs:List,app(?xs:List,app(?xs:List,?xs:List))))",
"app(?xs:List,app(app(app(?xs:List,?xs:List),?xs:List),app(?xs:List,?xs:List))) = app(app(?xs:List,app(app(?xs:List,?xs:List),app(?xs:List,?xs:List))),?xs:List)",
"app(?xs:List,app(app(app(?xs:List,?xs:List),app(?xs:List,?xs:List)),?xs:List)) = app(?xs:List,app(app(?xs:List,?xs:List),app(?xs:List,app(?xs:List,?xs:List))))",
"app(?xs:List,app(app(app(?xs:List,?ys:List),app(?xs:List,?xs:List)),?xs:List)) = app(app(?xs:List,?xs:List),app(?ys:List,app(app(?xs:List,?xs:List),?xs:List)))",
"app(app(?xs:List,?xs:List),app(?xs:List,app(app(?xs:List,?xs:List),?xs:List))) = app(?xs:List,app(app(?xs:List,?xs:List),app(app(?xs:List,?xs:List),?xs:List)))",
"app(app(?xs:List,?xs:List),app(app(?xs:List,app(?xs:List,?xs:List)),?xs:List)) = app(app(?xs:List,?xs:List),app(?xs:List,app(?xs:List,app(?xs:List,?xs:List))))",
"app(app(?xs:List,?xs:List),app(app(app(?xs:List,?xs:List),?ys:List),?xs:List)) = app(?xs:List,app(app(app(?xs:List,?xs:List),app(?xs:List,?ys:List)),?xs:List))",
"app(app(?xs:List,?xs:List),rev(app(app(?xs:List,?xs:List),rev(?xs:List)))) = app(?xs:List,app(app(?xs:List,?xs:List),app(rev(?xs:List),rev(?xs:List))))",
"app(app(?xs:List,?ys:List),app(?ys:List,app(?xs:List,app(?ys:List,?xs:List)))) = app(app(app(app(?xs:List,?ys:List),app(?ys:List,?xs:List)),?ys:List),?xs:List)",
"app(app(?xs:List,app(?ys:List,app(?xs:List,?xs:List))),app(?xs:List,?xs:List)) = app(?xs:List,app(app(app(?ys:List,?xs:List),app(?xs:List,?xs:List)),?xs:List))",
"app(app(app(?xs:List,?xs:List),app(app(?xs:List,?xs:List),?xs:List)),?xs:List) = app(app(app(?xs:List,app(?xs:List,?xs:List)),?xs:List),app(?xs:List,?xs:List))",
"app(app(app(app(?xs:List,?xs:List),app(?xs:List,?xs:List)),?xs:List),?xs:List) = app(app(app(app(?xs:List,?xs:List),?xs:List),?xs:List),app(?xs:List,?xs:List))",
"app(app(app(app(?xs:List,?ys:List),?xs:List),app(?xs:List,?xs:List)),?xs:List) = app(?xs:List,app(app(?ys:List,?xs:List),app(?xs:List,app(?xs:List,?xs:List))))",
"app(app(app(app(?ys:List,?xs:List),app(?xs:List,?xs:List)),?xs:List),?xs:List) = app(app(app(?ys:List,app(?xs:List,?xs:List)),?xs:List),app(?xs:List,?xs:List))",
"app(app(?xs:List,app(app(?xs:List,?xs:List),app(?ys:List,?xs:List))),?xs:List) = app(app(rev(rev(?xs:List)),?xs:List),app(?xs:List,app(app(?ys:List,?xs:List),?xs:List)))",
"app(?xs:List,app(app(app(?xs:List,?xs:List),?xs:List),app(app(?ys:List,?xs:List),app(?xs:List,?xs:List)))) = app(app(?xs:List,app(?xs:List,?xs:List)),app(app(?xs:List,?ys:List),app(?xs:List,app(?xs:List,?xs:List))))",
"app(app(app(?xs:List,app(?xs:List,?ys:List)),app(app(?xs:List,?xs:List),app(?xs:List,?xs:List))),rev(rev(?xs:List))) = app(app(?xs:List,app(?xs:List,app(?ys:List,?xs:List))),app(app(?xs:List,?xs:List),app(?xs:List,?xs:List)))"
]

val problemsListLenNo =
IOMstrs.rdEqs [
"rev(?xs:List) = rev(?ys:List)",
"rev(?xs:List) = app(?xs:List,?xs:List)",
"rev(?xs:List) = app(?xs:List,?ys:List)",
"rev(?xs:List) = app(?ys:List,?xs:List)",
"rev(?xs:List) = app(?zs:List,?ys:List)",
"rev(?xs:List) = app(?zs:List,?zs:List)",
"rev(?xs:List) = cons(?x:Nat,?ys:List)",
"rev(?xs:List) = cons(0,?xs:List)",
"rev(?xs:List) = cons(0,?zs:List)",
"rev(?ys:List) = rev(rev(?zs:List))",
"rev(?zs:List) = cons(?x:Nat,?zs:List)",
"rev(?xs:List) = app(?xs:List,rev(?xs:List))",
"rev(?xs:List) = app(?ys:List,rev(?zs:List))",
"rev(?xs:List) = app(?zs:List,rev(?xs:List))",
"rev(?xs:List) = app(rev(?zs:List),?xs:List)",
"rev(?xs:List) = cons(s(?x:Nat),?xs:List)",
"rev(?ys:List) = app(rev(?ys:List),?xs:List)",
"rev(?ys:List) = rev(app(?xs:List,?xs:List))",
"rev(?ys:List) = rev(app(?zs:List,?ys:List))",
"rev(?ys:List) = cons(len(?zs:List),?zs:List)",
"rev(?zs:List) = rev(cons(?y:Nat,?xs:List))",
"rev(?zs:List) = cons(len(?ys:List),?xs:List)",
"app(?xs:List,?xs:List) = app(?xs:List,?ys:List)",
"app(?xs:List,?xs:List) = app(?ys:List,?xs:List)",
"app(?xs:List,?xs:List) = app(?ys:List,?ys:List)",
"app(?xs:List,?xs:List) = cons(?x:Nat,?xs:List)",
"app(?xs:List,?xs:List) = cons(0,?xs:List)",
"app(?xs:List,?xs:List) = cons(0,?zs:List)",
"app(?xs:List,?ys:List) = app(?xs:List,?zs:List)",
"app(?xs:List,?ys:List) = app(?zs:List,?xs:List)",
"app(?xs:List,?ys:List) = app(?zs:List,?ys:List)",
"app(?xs:List,?ys:List) = app(?zs:List,?zs:List)",
"app(?xs:List,?ys:List) = cons(?x:Nat,?zs:List)",
"app(?xs:List,?ys:List) = cons(0,?xs:List)",
"app(?xs:List,?zs:List) = rev(rev(?zs:List))",
"app(?xs:List,?zs:List) = cons(?x:Nat,?xs:List)",
"app(?xs:List,?zs:List) = cons(0,?ys:List)",
"app(?xs:List,?zs:List) = cons(0,?zs:List)",
"app(?ys:List,?xs:List) = rev(rev(?zs:List))",
"app(?ys:List,?xs:List) = cons(?x:Nat,?xs:List)",
"app(?ys:List,?zs:List) = app(?zs:List,?ys:List)",
"app(?zs:List,?zs:List) = cons(?x:Nat,?ys:List)",
"rev(rev(?xs:List)) = app(?zs:List,?zs:List)",
"rev(rev(?ys:List)) = cons(?y:Nat,?ys:List)",
"rev(rev(?zs:List)) = app(?zs:List,?zs:List)",
"rev(rev(?zs:List)) = cons(0,?zs:List)",
"cons(?x:Nat,?zs:List) = cons(?y:Nat,?xs:List)",
"cons(0,?ys:List) = cons(?x:Nat,?ys:List)",
"cons(0,?ys:List) = cons(0,?xs:List)",
"app(?xs:List,rev(?xs:List)) = rev(?ys:List)",
"app(rev(?ys:List),?ys:List) = rev(?zs:List)",
"rev(app(?ys:List,?xs:List)) = rev(?zs:List)",
"rev(cons(?x:Nat,?xs:List)) = rev(?xs:List)",
"cons(len(?zs:List),?ys:List) = rev(?zs:List)",
"rev(?xs:List) = app(?ys:List,app(?xs:List,?xs:List))",
"rev(?xs:List) = app(?ys:List,cons(0,?ys:List))",
"rev(?xs:List) = app(?zs:List,app(?ys:List,?ys:List))",
"rev(?xs:List) = app(app(?ys:List,?xs:List),?zs:List)",
"rev(?xs:List) = app(app(?ys:List,?zs:List),?xs:List)",
"rev(?xs:List) = app(app(?ys:List,?zs:List),?ys:List)",
"rev(?xs:List) = rev(app(rev(?xs:List),?xs:List))",
"rev(?xs:List) = cons(plus(?y:Nat,?x:Nat),?xs:List)",
"rev(?xs:List) = cons(times(?x:Nat,?x:Nat),?zs:List)",
"rev(?xs:List) = cons(times(?y:Nat,?x:Nat),?ys:List)",
"rev(?ys:List) = app(?ys:List,app(?xs:List,?xs:List))",
"rev(?ys:List) = app(app(?xs:List,?xs:List),?xs:List)",
"rev(?ys:List) = app(app(?ys:List,?xs:List),?xs:List)",
"rev(?ys:List) = cons(?y:Nat,app(?zs:List,?xs:List))",
"rev(?zs:List) = app(?zs:List,app(?zs:List,?zs:List))",
"rev(?zs:List) = app(app(?zs:List,?xs:List),?zs:List)",
"rev(?zs:List) = app(app(?zs:List,?zs:List),?ys:List)",
"rev(?zs:List) = app(cons(?x:Nat,?xs:List),?xs:List)",
"app(?xs:List,?xs:List) = app(?xs:List,rev(?zs:List))",
"app(?xs:List,?xs:List) = app(?zs:List,rev(?ys:List))",
"app(?xs:List,?xs:List) = app(rev(?ys:List),?ys:List)",
"app(?xs:List,?xs:List) = rev(app(?xs:List,?ys:List))",
"app(?xs:List,?xs:List) = rev(app(?ys:List,?xs:List))",
"app(?xs:List,?zs:List) = rev(cons(?y:Nat,?xs:List))",
"app(?xs:List,?zs:List) = cons(?x:Nat,rev(?xs:List))",
"app(?xs:List,?zs:List) = cons(len(?ys:List),?xs:List)",
"app(?ys:List,?xs:List) = cons(s(0),?zs:List)",
"app(?ys:List,?ys:List) = cons(len(?zs:List),?zs:List)",
"app(?ys:List,?zs:List) = rev(rev(rev(?xs:List)))",
"app(?ys:List,?zs:List) = cons(len(?xs:List),?xs:List)",
"app(?zs:List,?xs:List) = app(rev(?xs:List),?ys:List)",
"app(?zs:List,?xs:List) = rev(app(?xs:List,?ys:List))",
"app(?zs:List,?xs:List) = rev(app(?zs:List,?xs:List))",
"app(?zs:List,?ys:List) = app(rev(?ys:List),?ys:List)",
"app(?zs:List,?ys:List) = rev(app(?xs:List,?zs:List))",
"app(?zs:List,?ys:List) = cons(0,rev(?ys:List))",
"app(?zs:List,?ys:List) = cons(s(?y:Nat),?ys:List)",
"cons(?x:Nat,?xs:List) = app(rev(?zs:List),?ys:List)",
"cons(?x:Nat,?xs:List) = rev(app(?ys:List,?zs:List))",
"cons(?x:Nat,?xs:List) = cons(?x:Nat,rev(?ys:List))",
"cons(?x:Nat,?zs:List) = app(?xs:List,rev(?ys:List))",
"cons(?y:Nat,?xs:List) = app(?zs:List,rev(?xs:List))",
"cons(0,?zs:List) = rev(app(?xs:List,?ys:List))",
"app(?xs:List,rev(?xs:List)) = app(?ys:List,?zs:List)",
"app(?xs:List,rev(?ys:List)) = app(?zs:List,?ys:List)",
"app(?ys:List,rev(?xs:List)) = cons(?y:Nat,?ys:List)",
"app(rev(?xs:List),?xs:List) = app(?xs:List,?ys:List)",
"app(rev(?ys:List),?ys:List) = app(?zs:List,?xs:List)",
"app(rev(?ys:List),?ys:List) = cons(?x:Nat,?ys:List)",
"app(rev(?ys:List),?zs:List) = app(?ys:List,?xs:List)",
"app(rev(?zs:List),?xs:List) = app(?xs:List,?xs:List)",
"app(rev(?zs:List),?xs:List) = app(?zs:List,?xs:List)",
"rev(app(?xs:List,?xs:List)) = app(?xs:List,?zs:List)",
"rev(app(?xs:List,?ys:List)) = app(?ys:List,?xs:List)",
"rev(app(?xs:List,?ys:List)) = app(?zs:List,?ys:List)",
"rev(app(?ys:List,?xs:List)) = app(?zs:List,?zs:List)",
"rev(app(?ys:List,?ys:List)) = cons(?y:Nat,?xs:List)",
"rev(app(?ys:List,?ys:List)) = cons(0,?ys:List)",
"rev(rev(rev(?ys:List))) = app(?xs:List,?xs:List)",
"rev(cons(?x:Nat,?ys:List)) = cons(?y:Nat,?ys:List)",
"rev(cons(?y:Nat,?xs:List)) = app(?ys:List,?zs:List)",
"cons(?x:Nat,rev(?xs:List)) = app(?zs:List,?zs:List)",
"cons(?x:Nat,rev(?ys:List)) = cons(0,?ys:List)",
"cons(0,rev(?xs:List)) = cons(?y:Nat,?xs:List)",
"cons(s(0),?zs:List) = app(?zs:List,?xs:List)",
"cons(len(?xs:List),?ys:List) = app(?xs:List,?xs:List)",
"app(?xs:List,app(?xs:List,?xs:List)) = rev(?ys:List)",
"app(?xs:List,rev(rev(?ys:List))) = rev(?xs:List)",
"app(?xs:List,cons(?x:Nat,?xs:List)) = rev(?ys:List)",
"app(?xs:List,cons(0,?ys:List)) = rev(?xs:List)",
"app(?ys:List,app(?zs:List,?ys:List)) = rev(?zs:List)",
"app(?zs:List,app(?zs:List,?xs:List)) = rev(?xs:List)",
"app(?zs:List,app(?zs:List,?xs:List)) = rev(?ys:List)",
"app(app(?xs:List,?xs:List),?xs:List) = rev(?xs:List)",
"app(app(?xs:List,?ys:List),?ys:List) = rev(?zs:List)",
"app(app(?xs:List,?ys:List),?zs:List) = rev(?xs:List)",
"app(app(?xs:List,?zs:List),?xs:List) = rev(?zs:List)",
"app(app(?zs:List,?zs:List),?xs:List) = rev(?xs:List)",
"app(rev(rev(?zs:List)),?ys:List) = rev(?xs:List)",
"app(cons(?y:Nat,?xs:List),?ys:List) = rev(?ys:List)",
"app(cons(?y:Nat,?xs:List),?zs:List) = rev(?xs:List)",
"app(cons(0,?ys:List),?xs:List) = rev(?xs:List)",
"cons(?x:Nat,app(?zs:List,?ys:List)) = rev(?ys:List)",
"rev(?xs:List) = app(?zs:List,app(rev(?zs:List),?ys:List))",
"rev(?xs:List) = app(rev(?ys:List),app(?xs:List,?ys:List))",
"rev(?ys:List) = rev(cons(0,app(?ys:List,?xs:List)))",
"rev(?zs:List) = app(app(?xs:List,?zs:List),rev(?ys:List))",
"rev(?zs:List) = rev(app(app(?xs:List,?xs:List),?xs:List))",
"rev(?zs:List) = cons(s(len(?xs:List)),rev(?xs:List))",
"app(?xs:List,?xs:List) = app(app(?ys:List,?zs:List),?ys:List)",
"app(?xs:List,?xs:List) = app(app(?zs:List,?xs:List),?xs:List)",
"app(?xs:List,?xs:List) = app(app(?zs:List,?zs:List),?xs:List)",
"app(?xs:List,?xs:List) = cons(?y:Nat,app(?xs:List,?xs:List))",
"app(?xs:List,?xs:List) = cons(plus(?y:Nat,?x:Nat),?ys:List)",
"app(?xs:List,?xs:List) = cons(plus(0,?y:Nat),?xs:List)",
"app(?xs:List,?xs:List) = cons(times(?x:Nat,?x:Nat),?xs:List)",
"app(?xs:List,?ys:List) = app(?xs:List,app(?xs:List,?xs:List))",
"app(?xs:List,?ys:List) = app(?xs:List,app(?zs:List,?xs:List))",
"app(?xs:List,?ys:List) = cons(?y:Nat,app(?xs:List,?ys:List))",
"app(?xs:List,?zs:List) = app(?xs:List,app(?zs:List,?zs:List))",
"app(?xs:List,?zs:List) = app(?ys:List,app(?xs:List,?ys:List))",
"app(?xs:List,?zs:List) = app(?ys:List,cons(0,?zs:List))",
"app(?xs:List,?zs:List) = app(app(?zs:List,?xs:List),?xs:List)",
"app(?ys:List,?xs:List) = app(?zs:List,app(?xs:List,?xs:List))",
"app(?ys:List,?xs:List) = app(app(?xs:List,?zs:List),?zs:List)",
"app(?ys:List,?xs:List) = app(app(?ys:List,?ys:List),?xs:List)",
"app(?ys:List,?xs:List) = app(app(?zs:List,?xs:List),?zs:List)",
"app(?ys:List,?xs:List) = cons(?y:Nat,cons(?y:Nat,?zs:List))",
"app(?ys:List,?ys:List) = app(?xs:List,app(?xs:List,?ys:List))",
"app(?ys:List,?ys:List) = app(app(?zs:List,?zs:List),?xs:List)",
"app(?ys:List,?zs:List) = app(?xs:List,app(?ys:List,?zs:List))",
"app(?ys:List,?zs:List) = app(app(?ys:List,?xs:List),?zs:List)",
"app(?ys:List,?zs:List) = cons(plus(?x:Nat,?y:Nat),?ys:List)",
"app(?zs:List,?xs:List) = app(app(?xs:List,?zs:List),?xs:List)",
"app(?zs:List,?xs:List) = app(app(?xs:List,?zs:List),?ys:List)",
"app(?zs:List,?ys:List) = app(?xs:List,app(?zs:List,?zs:List))",
"app(?zs:List,?ys:List) = app(app(?xs:List,?xs:List),?xs:List)",
"app(?zs:List,?ys:List) = app(app(?ys:List,?ys:List),?xs:List)",
"app(?zs:List,?ys:List) = app(app(?zs:List,?ys:List),?xs:List)",
"app(?zs:List,?ys:List) = cons(?y:Nat,app(?xs:List,?zs:List))",
"app(?zs:List,?zs:List) = app(cons(?y:Nat,?xs:List),?zs:List)",
"rev(rev(?xs:List)) = app(?xs:List,cons(0,?ys:List))",
"rev(rev(?xs:List)) = app(rev(?xs:List),rev(?zs:List))",
"rev(rev(?ys:List)) = app(app(?ys:List,?xs:List),?zs:List)",
"cons(?x:Nat,?xs:List) = app(?xs:List,app(?xs:List,?xs:List))",
"cons(?x:Nat,?xs:List) = app(app(?xs:List,?ys:List),?xs:List)",
"cons(?x:Nat,?ys:List) = app(?xs:List,app(?ys:List,?zs:List))",
"cons(?x:Nat,?zs:List) = app(app(?zs:List,?zs:List),?xs:List)",
"cons(?y:Nat,?xs:List) = app(?zs:List,app(?zs:List,?xs:List))",
"cons(?y:Nat,?xs:List) = app(app(?ys:List,?xs:List),?xs:List)",
"cons(?y:Nat,?xs:List) = cons(0,cons(?x:Nat,?ys:List))",
"cons(?y:Nat,?ys:List) = cons(0,app(?xs:List,?xs:List))",
"cons(0,?xs:List) = app(?xs:List,app(?xs:List,?ys:List))",
"cons(0,?zs:List) = app(?ys:List,cons(0,?xs:List))",
"cons(0,?zs:List) = app(app(?xs:List,?xs:List),?zs:List)",
"cons(0,?zs:List) = app(app(?zs:List,?zs:List),?xs:List)",
"app(?xs:List,rev(?xs:List)) = app(?ys:List,rev(?ys:List))",
"app(?xs:List,rev(?xs:List)) = app(rev(?ys:List),?xs:List)",
"app(?xs:List,rev(?zs:List)) = rev(app(?zs:List,?xs:List))",
"app(rev(?xs:List),?xs:List) = cons(len(?ys:List),?xs:List)",
"rev(app(?xs:List,?xs:List)) = app(?ys:List,rev(?zs:List))",
"rev(cons(?y:Nat,?zs:List)) = cons(s(?y:Nat),?zs:List)",
"cons(?x:Nat,rev(?xs:List)) = cons(0,rev(?xs:List))",
"cons(s(?y:Nat),?zs:List) = app(?xs:List,rev(?zs:List))",
"cons(len(?ys:List),?xs:List) = rev(app(?zs:List,?ys:List))",
"app(?xs:List,app(?xs:List,?ys:List)) = app(?zs:List,?xs:List)",
"app(?xs:List,app(?xs:List,?zs:List)) = app(?xs:List,?ys:List)",
"app(?xs:List,app(?ys:List,?zs:List)) = app(?ys:List,?ys:List)",
"app(?xs:List,app(?zs:List,?xs:List)) = app(?zs:List,?xs:List)",
"app(?xs:List,app(?zs:List,?ys:List)) = cons(0,?ys:List)",
"app(?xs:List,app(?zs:List,?zs:List)) = app(?xs:List,?xs:List)",
"app(?zs:List,app(?xs:List,?xs:List)) = cons(?x:Nat,?xs:List)",
"app(?zs:List,app(?xs:List,?ys:List)) = cons(?y:Nat,?ys:List)",
"app(?zs:List,app(?xs:List,?ys:List)) = cons(0,?xs:List)",
"app(app(?xs:List,?xs:List),?xs:List) = app(?xs:List,?xs:List)",
"app(app(?xs:List,?xs:List),?xs:List) = rev(rev(?ys:List))",
"app(app(?xs:List,?ys:List),?zs:List) = app(?zs:List,?xs:List)",
"app(app(?xs:List,?ys:List),?zs:List) = cons(?x:Nat,?ys:List)",
"app(app(?xs:List,?zs:List),?xs:List) = app(?xs:List,?xs:List)",
"app(app(?xs:List,?zs:List),?xs:List) = app(?xs:List,?zs:List)",
"app(app(?zs:List,?xs:List),?xs:List) = app(?zs:List,?xs:List)",
"app(app(?zs:List,?ys:List),?xs:List) = app(?xs:List,?xs:List)",
"app(rev(?xs:List),rev(?ys:List)) = app(?ys:List,?xs:List)",
"app(rev(?ys:List),rev(?xs:List)) = app(?ys:List,?xs:List)",
"app(rev(rev(?zs:List)),?xs:List) = app(?zs:List,?ys:List)",
"app(cons(?x:Nat,?xs:List),?ys:List) = app(?xs:List,?xs:List)",
"app(cons(?y:Nat,?zs:List),?xs:List) = app(?zs:List,?ys:List)",
"rev(app(?xs:List,rev(?xs:List))) = app(?ys:List,?xs:List)",
"rev(app(?xs:List,rev(?xs:List))) = app(?zs:List,?zs:List)",
"rev(app(?zs:List,rev(?xs:List))) = app(?ys:List,?zs:List)",
"rev(app(rev(?xs:List),?xs:List)) = app(?ys:List,?zs:List)",
"rev(rev(app(?xs:List,?xs:List))) = cons(?x:Nat,?ys:List)",
"rev(rev(app(?ys:List,?ys:List))) = app(?xs:List,?ys:List)",
"rev(rev(app(?ys:List,?ys:List))) = app(?zs:List,?xs:List)",
"rev(cons(?y:Nat,rev(?xs:List))) = app(?xs:List,?xs:List)",
"cons(?x:Nat,app(?xs:List,?xs:List)) = app(?zs:List,?xs:List)",
"cons(?y:Nat,app(?xs:List,?xs:List)) = rev(rev(?zs:List))",
"cons(0,app(?ys:List,?xs:List)) = app(?zs:List,?zs:List)",
"cons(0,app(?ys:List,?xs:List)) = cons(?x:Nat,?xs:List)",
"cons(s(?x:Nat),rev(?xs:List)) = app(?ys:List,?xs:List)",
"cons(len(?zs:List),rev(?ys:List)) = app(?xs:List,?xs:List)",
"rev(app(?xs:List,app(?zs:List,?zs:List))) = rev(?ys:List)",
"rev(app(rev(?zs:List),rev(?zs:List))) = rev(?ys:List)",
"cons(s(0),app(?xs:List,?xs:List)) = rev(?zs:List)",
"cons(len(app(?ys:List,?xs:List)),?zs:List) = rev(?xs:List)",
"cons(len(app(?zs:List,?xs:List)),?xs:List) = rev(?zs:List)",
"rev(?xs:List) = app(?xs:List,app(?xs:List,cons(?y:Nat,?zs:List)))",
"rev(?xs:List) = app(app(?xs:List,?xs:List),rev(rev(?xs:List)))",
"rev(?xs:List) = app(app(?xs:List,?xs:List),cons(0,?ys:List))",
"rev(?xs:List) = app(app(app(?xs:List,?zs:List),?zs:List),?ys:List)",
"rev(?ys:List) = cons(?x:Nat,cons(?y:Nat,cons(0,?xs:List)))",
"rev(?zs:List) = app(?zs:List,app(?xs:List,app(?ys:List,?ys:List)))",
"rev(?zs:List) = app(?zs:List,app(?ys:List,cons(?y:Nat,?xs:List)))",
"rev(?zs:List) = app(?zs:List,cons(?x:Nat,app(?xs:List,?ys:List)))",
"rev(?zs:List) = app(cons(0,cons(?x:Nat,?xs:List)),?xs:List)",
"rev(?zs:List) = cons(?y:Nat,rev(app(rev(?zs:List),?xs:List)))",
"app(?xs:List,?xs:List) = app(app(?ys:List,?xs:List),rev(?zs:List))",
"app(?xs:List,?xs:List) = app(app(rev(?ys:List),?zs:List),?xs:List)",
"app(?xs:List,?xs:List) = app(cons(len(?zs:List),?zs:List),?xs:List)",
"app(?xs:List,?xs:List) = rev(app(?xs:List,cons(?x:Nat,?xs:List)))",
"app(?xs:List,?xs:List) = cons(s(?y:Nat),app(?zs:List,?xs:List))",
"app(?xs:List,?xs:List) = cons(times(?y:Nat,?y:Nat),rev(?zs:List))",
"app(?xs:List,?ys:List) = app(app(rev(?ys:List),?xs:List),?zs:List)",
"app(?xs:List,?ys:List) = cons(len(app(?zs:List,?ys:List)),?xs:List)",
"app(?xs:List,?zs:List) = app(?zs:List,rev(cons(?y:Nat,?zs:List)))",
"app(?xs:List,?zs:List) = rev(app(app(?ys:List,?xs:List),?xs:List))",
"app(?xs:List,?zs:List) = cons(?y:Nat,app(?ys:List,rev(?zs:List)))",
"app(?ys:List,?xs:List) = app(?xs:List,app(rev(?xs:List),?ys:List))",
"app(?ys:List,?xs:List) = app(?ys:List,cons(len(?zs:List),?xs:List))",
"app(?ys:List,?ys:List) = app(app(?zs:List,?ys:List),rev(?ys:List))",
"app(?ys:List,?ys:List) = cons(len(?ys:List),app(?zs:List,?ys:List))",
"app(?zs:List,?xs:List) = app(?xs:List,rev(app(?zs:List,?zs:List)))",
"app(?zs:List,?xs:List) = app(app(?xs:List,rev(?ys:List)),?xs:List)",
"app(?zs:List,?xs:List) = app(app(?zs:List,?xs:List),rev(?ys:List))",
"app(?zs:List,?ys:List) = app(app(?ys:List,?zs:List),rev(?ys:List))",
"app(?zs:List,?ys:List) = cons(s(?y:Nat),cons(?x:Nat,?ys:List))",
"cons(?x:Nat,?xs:List) = app(app(rev(?xs:List),?ys:List),?xs:List)",
"cons(?x:Nat,?xs:List) = cons(?x:Nat,app(rev(?zs:List),?xs:List))",
"cons(?x:Nat,?xs:List) = cons(?y:Nat,cons(len(?ys:List),?ys:List))",
"cons(0,?xs:List) = app(?xs:List,cons(len(?xs:List),?ys:List))",
"cons(0,?xs:List) = app(rev(?zs:List),app(?ys:List,?ys:List))",
"cons(0,?ys:List) = app(app(?xs:List,?xs:List),rev(?zs:List))",
"cons(0,?zs:List) = rev(app(cons(?x:Nat,?zs:List),?xs:List))",
"app(?xs:List,rev(?ys:List)) = app(?zs:List,app(?ys:List,?xs:List))",
"app(?ys:List,rev(?xs:List)) = app(?xs:List,app(?xs:List,?ys:List))",
"app(?ys:List,rev(?xs:List)) = app(cons(0,?zs:List),?zs:List)",
"app(?zs:List,rev(?xs:List)) = app(?xs:List,app(?ys:List,?ys:List))",
"app(rev(?xs:List),?xs:List) = app(app(?zs:List,?xs:List),?zs:List)",
"app(rev(?xs:List),?ys:List) = app(?xs:List,cons(?x:Nat,?ys:List))",
"app(rev(?ys:List),?xs:List) = cons(?y:Nat,app(?xs:List,?xs:List))",
"app(rev(?ys:List),?ys:List) = cons(s(len(?xs:List)),?zs:List)",
"app(rev(?ys:List),?zs:List) = cons(0,app(?ys:List,?xs:List))",
"rev(app(?xs:List,?xs:List)) = app(app(?xs:List,?ys:List),?xs:List)",
"rev(app(?xs:List,?ys:List)) = app(cons(0,?ys:List),?xs:List)",
"rev(app(?ys:List,?xs:List)) = cons(?x:Nat,cons(?y:Nat,?zs:List))",
"rev(cons(?x:Nat,?ys:List)) = app(?ys:List,app(?ys:List,?xs:List))",
"rev(cons(?x:Nat,?zs:List)) = app(app(?zs:List,?ys:List),?ys:List)",
"rev(cons(?x:Nat,?zs:List)) = app(cons(?y:Nat,?ys:List),?xs:List)",
"rev(cons(?y:Nat,?ys:List)) = cons(?x:Nat,app(?zs:List,?xs:List))",
"cons(?x:Nat,rev(?ys:List)) = app(?ys:List,app(?xs:List,?ys:List))",
"cons(?x:Nat,rev(?zs:List)) = app(cons(?x:Nat,?xs:List),?xs:List)",
"cons(len(?zs:List),?xs:List) = app(app(?zs:List,?zs:List),?ys:List)",
"app(?xs:List,app(?ys:List,?zs:List)) = app(rev(?ys:List),?ys:List)",
"app(?ys:List,app(?xs:List,?ys:List)) = app(rev(?ys:List),?xs:List)",
"app(?ys:List,cons(?x:Nat,?ys:List)) = app(rev(?xs:List),?zs:List)",
"app(?zs:List,app(?xs:List,?xs:List)) = cons(0,rev(?zs:List))",
"app(?zs:List,cons(?x:Nat,?zs:List)) = app(rev(?xs:List),?xs:List)",
"app(app(?xs:List,?xs:List),?ys:List) = cons(len(?zs:List),?ys:List)",
"app(app(?xs:List,?zs:List),?ys:List) = cons(?y:Nat,rev(?zs:List))",
"app(app(?ys:List,?zs:List),?ys:List) = app(?xs:List,rev(?ys:List))",
"app(rev(rev(?ys:List)),?zs:List) = rev(app(?ys:List,?ys:List))",
"app(cons(?x:Nat,?zs:List),?ys:List) = app(?ys:List,rev(?ys:List))",
"app(cons(?y:Nat,?xs:List),?zs:List) = app(?ys:List,rev(?zs:List))",
"app(cons(0,?ys:List),?xs:List) = rev(app(?xs:List,?zs:List))",
"app(?xs:List,cons(len(?xs:List),?xs:List)) = cons(0,?xs:List)",
"app(?ys:List,rev(app(?xs:List,?zs:List))) = app(?xs:List,?zs:List)",
"app(?ys:List,cons(?y:Nat,rev(?zs:List))) = app(?ys:List,?ys:List)",
"app(?zs:List,app(rev(?xs:List),?ys:List)) = app(?zs:List,?ys:List)",
"app(app(?xs:List,rev(?ys:List)),?ys:List) = cons(?y:Nat,?ys:List)",
"app(app(?ys:List,?zs:List),rev(?ys:List)) = app(?xs:List,?xs:List)",
"app(app(?ys:List,rev(?xs:List)),?zs:List) = app(?ys:List,?ys:List)",
"app(app(rev(?zs:List),?xs:List),?ys:List) = cons(?x:Nat,?ys:List)",
"app(rev(app(?xs:List,?xs:List)),?xs:List) = cons(0,?zs:List)",
"app(rev(app(?xs:List,?ys:List)),?ys:List) = app(?xs:List,?ys:List)",
"rev(app(?ys:List,app(?zs:List,?ys:List))) = app(?xs:List,?ys:List)",
"rev(app(app(?zs:List,?xs:List),?ys:List)) = cons(?y:Nat,?ys:List)",
"rev(app(cons(?y:Nat,?xs:List),?xs:List)) = app(?ys:List,?zs:List)",
"rev(cons(?x:Nat,app(?ys:List,?zs:List))) = app(?zs:List,?ys:List)",
"cons(?y:Nat,app(rev(?ys:List),?xs:List)) = app(?xs:List,?zs:List)",
"cons(s(s(?x:Nat)),rev(?ys:List)) = app(?ys:List,?ys:List)",
"cons(s(s(len(?xs:List))),?xs:List) = app(?ys:List,?ys:List)",
"cons(len(app(?ys:List,?ys:List)),?xs:List) = app(?xs:List,?ys:List)",
"cons(len(cons(0,?zs:List)),?ys:List) = cons(?x:Nat,?ys:List)",
"cons(plus(0,s(0)),?ys:List) = app(?xs:List,?xs:List)",
"app(?xs:List,app(?zs:List,app(?ys:List,?xs:List))) = rev(?zs:List)",
"app(?ys:List,app(?xs:List,app(?xs:List,?ys:List))) = rev(?xs:List)",
"app(?ys:List,app(app(?zs:List,?xs:List),?ys:List)) = rev(?zs:List)",
"app(app(?zs:List,?ys:List),app(?ys:List,?xs:List)) = rev(?xs:List)",
"app(app(?zs:List,app(?zs:List,?ys:List)),?xs:List) = rev(?xs:List)",
"rev(cons(times(len(?xs:List),?y:Nat),?xs:List)) = rev(?ys:List)",
"rev(?xs:List) = app(app(?zs:List,?xs:List),rev(app(?xs:List,?zs:List)))",
"rev(?xs:List) = app(rev(?zs:List),app(?xs:List,app(?zs:List,?zs:List)))",
"rev(?xs:List) = cons(s(plus(?x:Nat,?x:Nat)),app(?ys:List,?xs:List))",
"rev(?xs:List) = cons(plus(len(cons(0,?xs:List)),?x:Nat),?zs:List)",
"rev(?ys:List) = app(app(?zs:List,?zs:List),app(rev(?ys:List),?ys:List))",
"app(?xs:List,?xs:List) = app(?xs:List,app(app(?ys:List,?xs:List),?ys:List))",
"app(?xs:List,?xs:List) = app(?ys:List,app(?xs:List,app(?ys:List,?xs:List)))",
"app(?xs:List,?xs:List) = rev(app(?zs:List,app(?zs:List,rev(?ys:List))))",
"app(?xs:List,?ys:List) = app(?xs:List,app(cons(?x:Nat,?zs:List),?zs:List))",
"app(?xs:List,?ys:List) = app(app(app(?xs:List,?zs:List),?xs:List),?ys:List)",
"app(?xs:List,?ys:List) = app(cons(s(s(?x:Nat)),?xs:List),?xs:List)",
"app(?xs:List,?zs:List) = app(?xs:List,app(?xs:List,app(?ys:List,?xs:List)))",
"app(?xs:List,?zs:List) = app(?zs:List,app(app(?ys:List,?zs:List),?xs:List))",
"app(?xs:List,?zs:List) = rev(app(app(rev(?xs:List),?zs:List),?zs:List))",
"app(?ys:List,?xs:List) = app(?xs:List,app(?ys:List,app(?xs:List,?xs:List)))",
"app(?ys:List,?xs:List) = app(?xs:List,app(?zs:List,app(?zs:List,?xs:List)))",
"app(?ys:List,?xs:List) = app(?xs:List,app(cons(?x:Nat,?ys:List),?xs:List))",
"app(?ys:List,?xs:List) = app(app(?zs:List,app(?xs:List,?xs:List)),?ys:List)",
"app(?ys:List,?xs:List) = app(app(?zs:List,cons(?x:Nat,?ys:List)),?xs:List)",
"app(?ys:List,?xs:List) = app(app(app(?ys:List,?ys:List),?ys:List),?xs:List)",
"app(?ys:List,?xs:List) = app(rev(app(rev(?ys:List),?ys:List)),?xs:List)",
"app(?ys:List,?zs:List) = app(app(app(?xs:List,?xs:List),?xs:List),?ys:List)",
"app(?zs:List,?xs:List) = app(?xs:List,app(app(?zs:List,?xs:List),?zs:List))",
"app(?zs:List,?xs:List) = app(app(?xs:List,?ys:List),app(?xs:List,?ys:List))",
"app(?zs:List,?xs:List) = app(app(?xs:List,?zs:List),app(?xs:List,?xs:List))",
"app(?zs:List,?xs:List) = app(app(?zs:List,?xs:List),app(?xs:List,?xs:List))",
"app(?zs:List,?xs:List) = app(app(app(?ys:List,?ys:List),?xs:List),?zs:List)",
"app(?zs:List,?xs:List) = app(cons(times(0,?x:Nat),?xs:List),?xs:List)",
"app(?zs:List,?ys:List) = app(?ys:List,app(?zs:List,app(?xs:List,?zs:List)))",
"app(?zs:List,?ys:List) = app(app(?xs:List,app(?zs:List,?zs:List)),?xs:List)",
"app(?zs:List,?ys:List) = app(app(?zs:List,?xs:List),app(?zs:List,?xs:List))",
"app(?zs:List,?zs:List) = app(app(?zs:List,app(?xs:List,?xs:List)),?xs:List)",
"cons(?x:Nat,?xs:List) = app(?xs:List,app(?xs:List,app(?xs:List,?xs:List)))",
"cons(?y:Nat,?xs:List) = app(app(app(?zs:List,?ys:List),?ys:List),?xs:List)",
"cons(?y:Nat,?zs:List) = app(?zs:List,cons(?y:Nat,app(?xs:List,?xs:List)))",
"cons(0,?xs:List) = app(app(app(?xs:List,?zs:List),?ys:List),?xs:List)",
"app(rev(?zs:List),?xs:List) = app(app(?zs:List,?zs:List),rev(?zs:List))",
"rev(app(?xs:List,?zs:List)) = app(?zs:List,cons(len(?zs:List),?zs:List))",
"rev(app(?ys:List,?ys:List)) = app(rev(cons(?y:Nat,?xs:List)),?ys:List)",
"cons(?x:Nat,rev(?xs:List)) = rev(app(?ys:List,app(?xs:List,?ys:List)))",
"cons(len(?xs:List),?xs:List) = cons(plus(s(0),?x:Nat),?ys:List)",
"cons(len(?xs:List),?zs:List) = app(rev(?ys:List),cons(?y:Nat,?zs:List))",
"app(?xs:List,app(?ys:List,?ys:List)) = cons(?x:Nat,app(?zs:List,?xs:List))",
"app(?xs:List,app(?zs:List,?ys:List)) = cons(s(s(?y:Nat)),?ys:List)",
"app(?xs:List,app(?zs:List,?zs:List)) = app(?xs:List,app(?zs:List,?xs:List))",
"app(?xs:List,app(?zs:List,?zs:List)) = app(?ys:List,app(?ys:List,?zs:List))",
"app(?ys:List,app(?xs:List,?xs:List)) = app(?xs:List,app(?ys:List,?ys:List))",
"app(?ys:List,app(?xs:List,?xs:List)) = cons(times(?y:Nat,0),?xs:List)",
"app(?ys:List,app(?xs:List,?ys:List)) = app(?ys:List,app(?zs:List,?ys:List))",
"app(?ys:List,app(?xs:List,?zs:List)) = cons(?y:Nat,app(?ys:List,?zs:List))",
"app(?ys:List,cons(?y:Nat,?xs:List)) = app(rev(rev(?ys:List)),?xs:List)",
"app(app(?xs:List,?xs:List),?xs:List) = cons(times(0,?y:Nat),?ys:List)",
"app(app(?xs:List,?zs:List),?ys:List) = app(?zs:List,cons(?y:Nat,?xs:List))",
"app(app(?ys:List,?xs:List),?xs:List) = app(app(?xs:List,?xs:List),?ys:List)",
"app(app(?ys:List,?zs:List),?xs:List) = app(app(?xs:List,?xs:List),?xs:List)",
"app(cons(?x:Nat,?xs:List),?ys:List) = cons(len(?ys:List),rev(?xs:List))",
"app(cons(0,?xs:List),?zs:List) = app(cons(?x:Nat,?zs:List),?xs:List)",
"app(cons(0,?ys:List),?ys:List) = app(cons(0,?xs:List),?xs:List)",
"app(cons(0,?zs:List),?ys:List) = app(?ys:List,app(?ys:List,?xs:List))",
"rev(app(?xs:List,rev(?zs:List))) = cons(?x:Nat,app(?zs:List,?xs:List))",
"rev(rev(app(?xs:List,?ys:List))) = app(app(?zs:List,?xs:List),?ys:List)",
"cons(?x:Nat,app(?xs:List,?xs:List)) = app(app(?xs:List,?zs:List),?ys:List)",
"cons(?x:Nat,cons(?x:Nat,?zs:List)) = app(app(?xs:List,?xs:List),?zs:List)",
"cons(?x:Nat,cons(0,?zs:List)) = app(app(?zs:List,?xs:List),?zs:List)",
"app(?zs:List,rev(cons(0,?ys:List))) = app(?xs:List,rev(?xs:List))",
"app(app(?xs:List,?xs:List),rev(?xs:List)) = app(rev(?ys:List),?ys:List)",
"app(rev(?xs:List),cons(0,?xs:List)) = app(?xs:List,rev(?zs:List))",
"app(rev(app(?zs:List,?xs:List)),?ys:List) = app(rev(?ys:List),?zs:List)",
"app(cons(s(0),?xs:List),?ys:List) = rev(cons(0,?ys:List))",
"rev(app(app(?zs:List,?zs:List),?xs:List)) = rev(cons(0,?xs:List))",
"rev(app(app(?zs:List,?zs:List),?ys:List)) = app(?xs:List,rev(?ys:List))",
"rev(rev(rev(app(?zs:List,?zs:List)))) = rev(app(?ys:List,?xs:List))",
"cons(?y:Nat,app(?xs:List,rev(?ys:List))) = rev(app(?ys:List,?ys:List))",
"cons(s(0),app(?xs:List,?zs:List)) = cons(s(?x:Nat),?xs:List)",
"app(?xs:List,app(?xs:List,app(?xs:List,?ys:List))) = app(?zs:List,?ys:List)",
"app(?xs:List,app(?xs:List,app(?ys:List,?xs:List))) = app(?ys:List,?xs:List)",
"app(?xs:List,app(?xs:List,app(?zs:List,?ys:List))) = cons(?x:Nat,?zs:List)",
"app(?xs:List,app(app(?xs:List,?xs:List),?xs:List)) = cons(?x:Nat,?xs:List)",
"app(?xs:List,app(app(?xs:List,?ys:List),?ys:List)) = app(?ys:List,?ys:List)",
"app(?xs:List,cons(?x:Nat,app(?ys:List,?xs:List))) = app(?xs:List,?xs:List)",
"app(?xs:List,cons(times(?x:Nat,0),?zs:List)) = cons(0,?zs:List)",
"app(?ys:List,app(?xs:List,app(?xs:List,?zs:List))) = app(?xs:List,?ys:List)",
"app(?ys:List,app(?xs:List,app(?ys:List,?zs:List))) = app(?zs:List,?xs:List)",
"app(?zs:List,app(app(?xs:List,?ys:List),?xs:List)) = app(?zs:List,?zs:List)",
"app(?zs:List,app(app(?ys:List,?zs:List),?ys:List)) = app(?xs:List,?xs:List)",
"app(?zs:List,cons(?x:Nat,app(?xs:List,?xs:List))) = app(?xs:List,?ys:List)",
"app(app(?xs:List,app(?xs:List,?zs:List)),?zs:List) = app(?zs:List,?xs:List)",
"app(app(?xs:List,app(?ys:List,?ys:List)),?xs:List) = app(?zs:List,?xs:List)",
"app(app(?ys:List,cons(?x:Nat,?ys:List)),?ys:List) = app(?zs:List,?zs:List)",
"app(app(?zs:List,app(?xs:List,?xs:List)),?xs:List) = cons(?y:Nat,?xs:List)",
"app(app(?zs:List,app(?zs:List,?xs:List)),?xs:List) = app(?ys:List,?xs:List)",
"app(app(app(?xs:List,?xs:List),?xs:List),?xs:List) = app(?xs:List,?xs:List)",
"app(app(app(?xs:List,?xs:List),?xs:List),?zs:List) = app(?xs:List,?xs:List)",
"app(app(app(?xs:List,?zs:List),?ys:List),?ys:List) = app(?ys:List,?ys:List)",
"app(app(app(?xs:List,?zs:List),?zs:List),?zs:List) = app(?zs:List,?zs:List)",
"app(rev(app(?ys:List,?zs:List)),rev(?zs:List)) = app(?xs:List,?zs:List)",
"app(cons(?y:Nat,?ys:List),app(?ys:List,?zs:List)) = cons(0,?zs:List)",
"app(cons(s(?x:Nat),?ys:List),rev(?ys:List)) = app(?xs:List,?ys:List)",
"app(cons(plus(?x:Nat,?x:Nat),?xs:List),?ys:List) = app(?zs:List,?ys:List)",
"app(cons(times(?x:Nat,?x:Nat),?ys:List),?xs:List) = app(?ys:List,?xs:List)",
"rev(rev(app(cons(?x:Nat,?ys:List),?zs:List))) = app(?zs:List,?zs:List)",
"cons(?y:Nat,app(app(?zs:List,?xs:List),?ys:List)) = app(?ys:List,?xs:List)",
"app(cons(len(app(?zs:List,?zs:List)),?xs:List),?zs:List) = rev(?zs:List)",
"cons(times(?x:Nat,len(?xs:List)),app(?xs:List,?xs:List)) = rev(?xs:List)",
"rev(?xs:List) = app(app(app(?ys:List,app(?xs:List,?zs:List)),?xs:List),?xs:List)",
"rev(?xs:List) = app(rev(rev(cons(len(?ys:List),rev(?zs:List)))),?ys:List)",
"rev(?ys:List) = app(?xs:List,app(rev(app(?xs:List,rev(?zs:List))),?xs:List))",
"rev(?ys:List) = app(rev(cons(times(len(?zs:List),?x:Nat),?zs:List)),?ys:List)",
"rev(?zs:List) = rev(cons(s(times(plus(?x:Nat,?x:Nat),?x:Nat)),?zs:List))",
"app(?xs:List,?xs:List) = rev(cons(plus(times(?x:Nat,?x:Nat),0),?zs:List))",
"app(?xs:List,?ys:List) = app(?zs:List,app(?xs:List,cons(s(0),?ys:List)))",
"app(?xs:List,?zs:List) = app(app(?xs:List,?xs:List),app(rev(?xs:List),?xs:List))",
"app(?xs:List,?zs:List) = app(app(app(?ys:List,?xs:List),rev(?xs:List)),?zs:List)",
"app(?xs:List,?zs:List) = rev(app(app(?xs:List,?xs:List),cons(?x:Nat,?zs:List)))",
"app(?xs:List,?zs:List) = cons(s(len(?xs:List)),cons(s(?x:Nat),?xs:List))",
"app(?ys:List,?xs:List) = app(app(?zs:List,app(?xs:List,rev(?zs:List))),?xs:List)",
"app(?ys:List,?xs:List) = cons(?y:Nat,cons(?y:Nat,rev(app(?xs:List,?xs:List))))",
"app(?ys:List,?xs:List) = cons(s(?y:Nat),app(?ys:List,app(?ys:List,?ys:List)))",
"app(?ys:List,?xs:List) = cons(len(rev(?ys:List)),app(?ys:List,rev(?zs:List)))",
"app(?ys:List,?zs:List) = rev(cons(s(len(app(?xs:List,?zs:List))),?zs:List))",
"app(?zs:List,?ys:List) = cons(s(?x:Nat),app(app(?ys:List,?zs:List),?ys:List))",
"app(?zs:List,?zs:List) = cons(times(len(app(?zs:List,?xs:List)),?x:Nat),?xs:List)",
"cons(?x:Nat,?xs:List) = app(?xs:List,app(?ys:List,rev(cons(?x:Nat,?ys:List))))",
"cons(?x:Nat,?zs:List) = cons(?x:Nat,app(rev(?zs:List),cons(?x:Nat,?xs:List)))",
"cons(?y:Nat,?zs:List) = app(rev(?zs:List),app(?ys:List,app(?zs:List,?zs:List)))",
"app(rev(?ys:List),?zs:List) = rev(app(app(?xs:List,?zs:List),rev(?zs:List)))",
"rev(app(?ys:List,?xs:List)) = app(cons(?y:Nat,?ys:List),app(?zs:List,?ys:List))",
"rev(app(?zs:List,?ys:List)) = cons(len(app(?ys:List,rev(?ys:List))),?ys:List)",
"cons(0,rev(?ys:List)) = app(app(app(?zs:List,?ys:List),?ys:List),?ys:List)",
"cons(s(?x:Nat),?xs:List) = app(?zs:List,app(app(?zs:List,?ys:List),?xs:List))",
"cons(len(?zs:List),?xs:List) = app(app(?zs:List,app(?xs:List,?ys:List)),?xs:List)",
"cons(len(?zs:List),?zs:List) = rev(cons(len(app(?zs:List,?xs:List)),?xs:List))",
"app(?xs:List,app(?ys:List,?zs:List)) = app(?zs:List,app(rev(?ys:List),?zs:List))",
"app(?xs:List,app(?zs:List,?xs:List)) = app(?ys:List,app(?xs:List,rev(?xs:List)))",
"app(?zs:List,app(?ys:List,?xs:List)) = app(?ys:List,rev(app(?xs:List,?zs:List)))",
"app(app(?xs:List,?ys:List),?xs:List) = rev(app(app(?xs:List,?ys:List),?xs:List))",
"app(app(?xs:List,?zs:List),?ys:List) = app(?xs:List,cons(len(?ys:List),?zs:List))",
"rev(app(rev(?zs:List),?xs:List)) = app(app(?ys:List,rev(?zs:List)),?xs:List)",
"cons(?x:Nat,app(?xs:List,?zs:List)) = cons(s(times(?x:Nat,?x:Nat)),?zs:List)",
"app(?ys:List,app(?zs:List,rev(?xs:List))) = app(?ys:List,app(?zs:List,?ys:List))",
"app(?zs:List,app(?xs:List,rev(?ys:List))) = app(?ys:List,app(?xs:List,?ys:List))",
"app(?zs:List,app(rev(?xs:List),?xs:List)) = cons(times(?x:Nat,0),?xs:List)",
"app(?zs:List,app(rev(?xs:List),?ys:List)) = app(?ys:List,cons(0,?ys:List))",
"app(?zs:List,cons(?x:Nat,rev(?xs:List))) = cons(?x:Nat,app(?xs:List,?zs:List))",
"app(rev(app(?zs:List,?ys:List)),?ys:List) = app(?ys:List,app(?ys:List,?xs:List))",
"rev(app(app(?ys:List,?xs:List),?xs:List)) = cons(plus(?y:Nat,?x:Nat),?xs:List)",
"cons(?y:Nat,app(rev(?zs:List),?xs:List)) = app(?xs:List,app(?xs:List,?xs:List))",
"app(?xs:List,app(app(?xs:List,?xs:List),?xs:List)) = app(rev(?xs:List),?xs:List)",
"app(?ys:List,app(?zs:List,app(?zs:List,?xs:List))) = app(?zs:List,rev(?ys:List))",
"app(app(?ys:List,?xs:List),cons(?x:Nat,?xs:List)) = app(rev(?ys:List),?ys:List)",
"app(app(?zs:List,app(?zs:List,?xs:List)),?ys:List) = app(rev(?xs:List),?xs:List)",
"rev(app(cons(s(?x:Nat),?zs:List),?ys:List)) = rev(cons(?x:Nat,?ys:List))",
"cons(?y:Nat,app(app(?zs:List,?xs:List),?zs:List)) = cons(len(?xs:List),?xs:List)",
"app(?xs:List,app(?zs:List,rev(app(?ys:List,?xs:List)))) = app(?xs:List,?xs:List)",
"app(?xs:List,app(cons(0,rev(?ys:List)),?xs:List)) = app(?zs:List,?xs:List)",
"app(?xs:List,cons(plus(s(?x:Nat),?x:Nat),?ys:List)) = app(?ys:List,?ys:List)",
"app(?ys:List,app(?ys:List,app(rev(?zs:List),?zs:List))) = rev(rev(?zs:List))",
"app(?zs:List,app(?xs:List,rev(app(?ys:List,?xs:List)))) = app(?xs:List,?ys:List)",
"app(?zs:List,app(rev(app(?zs:List,?xs:List)),?xs:List)) = app(?ys:List,?ys:List)",
"app(app(?xs:List,app(rev(?ys:List),?xs:List)),?ys:List) = app(?ys:List,?xs:List)",
"app(app(?xs:List,rev(?xs:List)),app(?ys:List,?zs:List)) = app(?zs:List,?ys:List)",
"app(app(?zs:List,?ys:List),app(rev(?xs:List),?zs:List)) = cons(?x:Nat,?ys:List)",
"app(app(app(?xs:List,?ys:List),rev(?zs:List)),?zs:List) = app(?xs:List,?xs:List)",
"app(app(rev(?ys:List),?zs:List),app(?zs:List,?xs:List)) = app(?zs:List,?xs:List)",
"app(app(rev(?zs:List),?xs:List),app(?zs:List,?xs:List)) = app(?zs:List,?xs:List)",
"app(rev(?ys:List),app(?ys:List,app(?ys:List,?zs:List))) = cons(?y:Nat,?xs:List)",
"app(cons(?y:Nat,?ys:List),app(rev(?xs:List),?zs:List)) = cons(?x:Nat,?xs:List)",
"app(cons(s(?x:Nat),?zs:List),app(?xs:List,?xs:List)) = app(?zs:List,?zs:List)",
"app(cons(s(s(s(0))),?xs:List),?xs:List) = app(?zs:List,?ys:List)",
"app(cons(len(?xs:List),?zs:List),app(?zs:List,?zs:List)) = app(?ys:List,?xs:List)",
"rev(app(app(?ys:List,?xs:List),app(?zs:List,?xs:List))) = app(?xs:List,?xs:List)",
"rev(app(app(?ys:List,?ys:List),app(?xs:List,?zs:List))) = app(?zs:List,?zs:List)",
"rev(app(app(?ys:List,?ys:List),app(?ys:List,?xs:List))) = app(?zs:List,?zs:List)",
"cons(?x:Nat,app(?xs:List,rev(app(?ys:List,?zs:List)))) = app(?xs:List,?zs:List)",
"cons(s(plus(?y:Nat,s(len(?zs:List)))),?zs:List) = app(?xs:List,?xs:List)",
"cons(len(app(?xs:List,?xs:List)),app(?ys:List,?xs:List)) = app(?ys:List,?xs:List)",
"cons(len(app(?xs:List,app(?zs:List,?zs:List))),?zs:List) = app(?xs:List,?xs:List)",
"cons(plus(?x:Nat,len(cons(?y:Nat,?xs:List))),?zs:List) = cons(?x:Nat,?ys:List)",
"app(?zs:List,app(app(cons(?y:Nat,?zs:List),?xs:List),?zs:List)) = rev(?xs:List)",
"app(?zs:List,cons(?x:Nat,app(?zs:List,app(?xs:List,?ys:List)))) = rev(?xs:List)",
"app(app(?xs:List,?zs:List),app(?xs:List,app(?xs:List,?zs:List))) = rev(?xs:List)",
"app(app(?ys:List,app(app(?ys:List,?xs:List),?ys:List)),?zs:List) = rev(?ys:List)",
"app(app(cons(s(len(?zs:List)),?xs:List),?xs:List),?ys:List) = rev(?zs:List)",
"app(cons(0,app(app(?xs:List,?xs:List),?zs:List)),?zs:List) = rev(?ys:List)",
"app(cons(times(plus(?x:Nat,0),?y:Nat),?xs:List),?zs:List) = rev(?xs:List)",
"rev(app(?ys:List,rev(cons(0,app(?ys:List,?zs:List))))) = rev(?zs:List)",
"cons(plus(?x:Nat,times(?y:Nat,times(?x:Nat,?y:Nat))),?xs:List) = rev(?ys:List)",
"rev(?xs:List) = app(?zs:List,app(rev(?ys:List),cons(?x:Nat,rev(rev(?xs:List)))))",
"rev(?xs:List) = app(rev(app(?xs:List,cons(plus(?y:Nat,?x:Nat),?xs:List))),?ys:List)",
"rev(?xs:List) = app(cons(?x:Nat,rev(?xs:List)),app(?zs:List,app(?xs:List,?xs:List)))",
"rev(?xs:List) = app(cons(s(plus(0,?y:Nat)),app(?xs:List,?ys:List)),?zs:List)",
"rev(?xs:List) = rev(app(app(app(?xs:List,?zs:List),app(?zs:List,?ys:List)),?xs:List))",
"rev(?ys:List) = app(app(?zs:List,app(?xs:List,cons(s(0),?xs:List))),?xs:List)",
"rev(?ys:List) = cons(times(len(app(?zs:List,?xs:List)),times(?x:Nat,?x:Nat)),?xs:List)",
"rev(?zs:List) = app(app(?xs:List,?xs:List),rev(app(?zs:List,app(?zs:List,?zs:List))))",
"rev(?zs:List) = app(app(cons(len(cons(?y:Nat,?xs:List)),?xs:List),?xs:List),?ys:List)",
"rev(?zs:List) = rev(app(app(app(rev(?xs:List),?xs:List),?xs:List),rev(?xs:List)))",
"rev(?zs:List) = cons(len(?xs:List),app(?xs:List,cons(?x:Nat,app(?zs:List,?zs:List))))",
"app(?xs:List,?xs:List) = app(app(app(?xs:List,app(?zs:List,?zs:List)),?xs:List),?zs:List)",
"app(?xs:List,?xs:List) = rev(cons(plus(plus(?x:Nat,len(?ys:List)),?y:Nat),?zs:List))",
"app(?xs:List,?xs:List) = cons(times(?y:Nat,len(rev(?ys:List))),app(?xs:List,?zs:List))",
"app(?xs:List,?ys:List) = app(app(?xs:List,app(app(?xs:List,?ys:List),?ys:List)),?ys:List)",
"app(?xs:List,?zs:List) = app(?ys:List,cons(times(?y:Nat,times(?x:Nat,?x:Nat)),?xs:List))",
"app(?xs:List,?zs:List) = app(app(app(?xs:List,?xs:List),app(?zs:List,?xs:List)),?xs:List)",
"app(?xs:List,?zs:List) = app(app(app(?ys:List,?ys:List),app(?xs:List,?xs:List)),?xs:List)",
"app(?xs:List,?zs:List) = app(cons(?x:Nat,app(app(?zs:List,?zs:List),?zs:List)),?ys:List)",
"app(?ys:List,?xs:List) = app(rev(?xs:List),app(app(?xs:List,?xs:List),rev(?xs:List)))",
"app(?ys:List,?xs:List) = cons(?x:Nat,app(cons(0,?xs:List),app(?xs:List,?xs:List)))",
"app(?zs:List,?xs:List) = app(app(?ys:List,app(?xs:List,app(?xs:List,?xs:List))),?xs:List)",
"app(?zs:List,?xs:List) = app(app(?ys:List,rev(rev(app(?xs:List,?ys:List)))),?xs:List)",
"app(?zs:List,?xs:List) = app(app(cons(0,cons(0,?zs:List)),?xs:List),?ys:List)",
"app(?zs:List,?xs:List) = app(cons(?x:Nat,app(?xs:List,?xs:List)),app(?ys:List,?xs:List))",
"app(?zs:List,?xs:List) = app(cons(?y:Nat,?ys:List),app(app(?xs:List,?xs:List),?ys:List))",
"app(?zs:List,?xs:List) = rev(cons(?x:Nat,cons(s(?y:Nat),app(?xs:List,?zs:List))))",
"app(?zs:List,?zs:List) = app(app(cons(times(0,?x:Nat),?zs:List),?ys:List),?zs:List)",
"app(?zs:List,?zs:List) = rev(app(app(rev(app(?xs:List,?xs:List)),?xs:List),?xs:List))",
"rev(rev(?xs:List)) = app(?xs:List,app(app(app(?xs:List,?zs:List),?zs:List),?zs:List))",
"rev(rev(?zs:List)) = app(app(?xs:List,app(?zs:List,?zs:List)),app(?ys:List,?ys:List))",
"cons(?x:Nat,?xs:List) = app(?zs:List,cons(?y:Nat,app(app(?xs:List,?xs:List),?xs:List)))",
"cons(?x:Nat,?xs:List) = app(app(app(?ys:List,app(?zs:List,?ys:List)),?ys:List),?ys:List)",
"cons(?y:Nat,?xs:List) = app(?ys:List,app(?xs:List,app(?xs:List,cons(0,?xs:List))))",
"cons(0,?ys:List) = app(app(?zs:List,app(app(?xs:List,?zs:List),?xs:List)),?xs:List)",
"cons(0,?zs:List) = app(?ys:List,app(?xs:List,app(?ys:List,app(?zs:List,?xs:List))))",
"cons(0,?zs:List) = app(?zs:List,cons(0,app(app(?xs:List,?xs:List),?xs:List)))",
"app(?zs:List,rev(?ys:List)) = app(app(app(?xs:List,?ys:List),rev(?zs:List)),?zs:List)",
"app(rev(?xs:List),?ys:List) = app(app(?xs:List,?xs:List),app(?zs:List,rev(?ys:List)))",
"rev(app(?ys:List,?ys:List)) = app(app(?xs:List,?zs:List),app(rev(?xs:List),?xs:List))",
"app(?xs:List,app(?ys:List,?ys:List)) = app(rev(?zs:List),app(?xs:List,rev(?ys:List)))",
"app(?xs:List,cons(?x:Nat,?ys:List)) = app(?ys:List,cons(len(rev(?xs:List)),?ys:List))",
"app(?ys:List,app(?xs:List,?zs:List)) = app(?xs:List,app(rev(rev(?ys:List)),?ys:List))",
"app(?ys:List,app(?zs:List,?xs:List)) = app(?xs:List,app(?xs:List,app(?xs:List,?xs:List)))",
"app(?ys:List,app(?zs:List,?zs:List)) = app(?xs:List,app(?zs:List,cons(0,?xs:List)))",
"app(?ys:List,cons(?x:Nat,?ys:List)) = app(app(?xs:List,app(?zs:List,?zs:List)),?xs:List)",
"app(?zs:List,app(?ys:List,?zs:List)) = app(?ys:List,app(?xs:List,app(?ys:List,?zs:List)))",
"app(app(?xs:List,?xs:List),?zs:List) = app(?ys:List,app(?xs:List,app(?zs:List,?zs:List)))",
"app(app(?xs:List,?xs:List),?zs:List) = app(?zs:List,app(?ys:List,app(?xs:List,?zs:List)))",
"app(rev(?zs:List),rev(?zs:List)) = app(cons(?x:Nat,?zs:List),app(?xs:List,?ys:List))",
"cons(0,cons(0,?xs:List)) = cons(times(s(0),len(?ys:List)),?xs:List)",
"app(app(rev(?xs:List),?xs:List),?zs:List) = rev(cons(0,cons(?y:Nat,?zs:List)))",
"app(rev(app(?ys:List,?zs:List)),?ys:List) = app(app(?ys:List,?zs:List),rev(?ys:List))",
"cons(len(?ys:List),app(?ys:List,?zs:List)) = cons(s(?x:Nat),app(?zs:List,?xs:List))",
"app(?xs:List,app(app(?xs:List,?ys:List),?ys:List)) = app(app(?xs:List,?ys:List),?zs:List)",
"app(?xs:List,app(app(?xs:List,?zs:List),?xs:List)) = cons(?y:Nat,cons(0,?ys:List))",
"app(?xs:List,app(app(?ys:List,?ys:List),?zs:List)) = app(app(?zs:List,?ys:List),?zs:List)",
"app(?ys:List,app(?xs:List,app(?xs:List,?xs:List))) = app(app(?ys:List,?ys:List),?zs:List)",
"app(?ys:List,cons(times(?x:Nat,?y:Nat),?ys:List)) = app(app(?zs:List,?ys:List),?xs:List)",
"app(?zs:List,app(?xs:List,app(?zs:List,?ys:List))) = rev(rev(app(?zs:List,?zs:List)))",
"app(?zs:List,app(app(?xs:List,?xs:List),?xs:List)) = app(?xs:List,app(?ys:List,?ys:List))",
"app(?zs:List,app(app(?xs:List,?zs:List),?zs:List)) = app(app(?zs:List,?xs:List),?zs:List)",
"app(app(?zs:List,?xs:List),app(?ys:List,?zs:List)) = app(app(?zs:List,?zs:List),?ys:List)",
"cons(?x:Nat,app(rev(rev(?ys:List)),?ys:List)) = app(app(?xs:List,?ys:List),?zs:List)",
"cons(times(0,?x:Nat),rev(rev(?xs:List))) = app(app(?ys:List,?ys:List),?zs:List)",
"app(?ys:List,app(?zs:List,app(rev(?xs:List),?ys:List))) = rev(app(?zs:List,?xs:List))",
"app(?xs:List,app(?xs:List,app(app(?zs:List,?zs:List),?zs:List))) = cons(?x:Nat,?zs:List)",
"app(?xs:List,app(?zs:List,app(?xs:List,app(?xs:List,?xs:List)))) = app(?xs:List,?xs:List)",
"app(?xs:List,rev(app(?xs:List,app(?ys:List,rev(?ys:List))))) = cons(?y:Nat,?ys:List)",
"app(?ys:List,app(?ys:List,app(?ys:List,rev(rev(?zs:List))))) = cons(?x:Nat,?zs:List)",
"app(?ys:List,app(app(app(?ys:List,?xs:List),?xs:List),?zs:List)) = app(?xs:List,?zs:List)",
"app(?ys:List,cons(s(s(times(?y:Nat,?x:Nat))),?zs:List)) = app(?xs:List,?zs:List)",
"app(?ys:List,cons(times(?x:Nat,plus(?y:Nat,?y:Nat)),?zs:List)) = app(?ys:List,?xs:List)",
"app(?zs:List,app(app(?ys:List,app(?xs:List,?xs:List)),?xs:List)) = app(?zs:List,?xs:List)",
"app(app(?ys:List,app(app(?xs:List,?zs:List),?zs:List)),?xs:List) = app(?zs:List,?zs:List)",
"app(app(?zs:List,?xs:List),cons(0,cons(0,?xs:List))) = app(?ys:List,?zs:List)",
"app(app(?zs:List,app(?xs:List,?zs:List)),app(?ys:List,?xs:List)) = app(?zs:List,?ys:List)",
"app(app(app(?ys:List,?ys:List),app(?xs:List,?xs:List)),?xs:List) = app(?xs:List,?xs:List)",
"app(app(app(?ys:List,?zs:List),app(?xs:List,?xs:List)),?zs:List) = app(?zs:List,?ys:List)",
"app(app(app(app(?xs:List,?xs:List),?ys:List),?xs:List),?zs:List) = app(?ys:List,?ys:List)",
"app(app(app(app(?zs:List,?xs:List),?xs:List),?xs:List),?xs:List) = app(?xs:List,?ys:List)",
"app(app(cons(?y:Nat,app(?xs:List,?xs:List)),?ys:List),?xs:List) = app(?xs:List,?ys:List)",
"app(rev(app(?xs:List,rev(app(?zs:List,?ys:List)))),?zs:List) = app(?zs:List,?xs:List)",
"app(rev(app(?zs:List,app(?zs:List,rev(?zs:List)))),?xs:List) = app(?xs:List,?zs:List)",
"app(cons(plus(s(len(?xs:List)),?x:Nat),?ys:List),?ys:List) = app(?xs:List,?ys:List)",
"rev(app(?zs:List,cons(s(plus(?x:Nat,0)),?xs:List))) = app(?xs:List,?xs:List)",
"rev(app(rev(app(cons(?x:Nat,?xs:List),?zs:List)),?xs:List)) = cons(0,?zs:List)",
"cons(?x:Nat,app(app(?ys:List,app(?ys:List,?zs:List)),?xs:List)) = app(?ys:List,?ys:List)",
"cons(s(0),app(?ys:List,app(rev(?zs:List),?ys:List))) = app(?ys:List,?xs:List)",
"cons(len(app(app(rev(?xs:List),?zs:List),?xs:List)),?xs:List) = app(?zs:List,?xs:List)",
"cons(plus(?y:Nat,s(0)),cons(len(?ys:List),?xs:List)) = app(?ys:List,?ys:List)",
"app(?xs:List,app(?xs:List,rev(app(?zs:List,app(?xs:List,?ys:List))))) = rev(?ys:List)",
"app(?xs:List,app(?zs:List,app(?xs:List,app(?ys:List,rev(?xs:List))))) = rev(?zs:List)",
"app(?ys:List,cons(times(s(?y:Nat),s(len(?ys:List))),?xs:List)) = rev(?xs:List)",
"app(?zs:List,app(rev(app(cons(0,?ys:List),?xs:List)),?ys:List)) = rev(?ys:List)",
"app(app(?ys:List,rev(?xs:List)),app(?ys:List,cons(?y:Nat,?ys:List))) = rev(?ys:List)",
"app(app(?zs:List,app(rev(cons(?x:Nat,?ys:List)),?xs:List)),?xs:List) = rev(?zs:List)",
"app(rev(app(?xs:List,rev(app(?ys:List,?ys:List)))),rev(?ys:List)) = rev(?zs:List)",
"app(rev(rev(?xs:List)),app(?xs:List,rev(app(?ys:List,?xs:List)))) = rev(?ys:List)",
"cons(?x:Nat,rev(app(app(?ys:List,app(?ys:List,?ys:List)),?ys:List))) = rev(?zs:List)",
"rev(?xs:List) = app(app(app(app(app(?zs:List,?zs:List),?zs:List),?xs:List),?zs:List),?xs:List)",
"rev(?xs:List) = app(cons(times(s(?y:Nat),times(?x:Nat,s(?y:Nat))),?xs:List),?xs:List)",
"rev(?xs:List) = cons(0,app(app(?ys:List,app(app(?xs:List,?xs:List),?ys:List)),?xs:List))",
"rev(?ys:List) = app(app(app(?ys:List,cons(?y:Nat,app(?ys:List,?xs:List))),?zs:List),?xs:List)",
"rev(?ys:List) = app(app(cons(times(plus(?x:Nat,?x:Nat),?x:Nat),?ys:List),?ys:List),?xs:List)",
"rev(?zs:List) = app(app(?ys:List,?xs:List),app(cons(s(len(?ys:List)),?xs:List),?zs:List))",
"rev(?zs:List) = app(app(cons(plus(plus(?x:Nat,?x:Nat),0),?xs:List),?xs:List),?ys:List)",
"rev(?zs:List) = app(rev(?xs:List),cons(?x:Nat,cons(s(times(0,?y:Nat)),?xs:List)))",
"rev(?zs:List) = cons(?x:Nat,app(?xs:List,rev(cons(s(0),app(?ys:List,?xs:List)))))",
"rev(?zs:List) = cons(times(times(plus(?x:Nat,times(0,?x:Nat)),?y:Nat),?x:Nat),?zs:List)",
"app(?xs:List,?xs:List) = rev(app(?xs:List,app(cons(len(?ys:List),rev(?ys:List)),?zs:List)))",
"app(?xs:List,?ys:List) = app(app(?zs:List,?zs:List),app(app(?xs:List,rev(?xs:List)),?zs:List))",
"app(?xs:List,?ys:List) = rev(app(?xs:List,cons(0,app(cons(?x:Nat,?ys:List),?xs:List))))",
"app(?xs:List,?ys:List) = cons(times(plus(len(?xs:List),times(0,0)),?x:Nat),?ys:List)",
"app(?xs:List,?zs:List) = rev(app(?ys:List,app(?xs:List,app(app(?ys:List,?xs:List),?zs:List))))",
"app(?ys:List,?xs:List) = app(?xs:List,rev(app(cons(?y:Nat,app(?ys:List,?zs:List)),?xs:List)))",
"app(?ys:List,?xs:List) = app(?zs:List,cons(times(?x:Nat,len(cons(0,?zs:List))),?zs:List))",
"app(?ys:List,?xs:List) = rev(app(cons(?x:Nat,app(?ys:List,app(?ys:List,?zs:List))),?zs:List))",
"app(?ys:List,?xs:List) = rev(rev(app(cons(len(?xs:List),app(?ys:List,?ys:List)),?ys:List)))",
"app(?ys:List,?ys:List) = rev(app(?ys:List,app(?xs:List,app(?xs:List,app(?ys:List,?xs:List)))))",
"app(?zs:List,?ys:List) = rev(app(app(?xs:List,app(?xs:List,app(?zs:List,?xs:List))),?xs:List))",
"app(?zs:List,?zs:List) = rev(app(?xs:List,app(?ys:List,cons(?y:Nat,app(?ys:List,?ys:List)))))",
"cons(?x:Nat,?ys:List) = app(rev(app(?zs:List,app(app(?ys:List,?xs:List),?zs:List))),?zs:List)",
"cons(?x:Nat,?ys:List) = cons(s(0),app(?xs:List,app(?xs:List,app(?xs:List,?ys:List))))",
"cons(?y:Nat,?ys:List) = app(?ys:List,app(app(rev(app(?zs:List,?xs:List)),?ys:List),?ys:List))",
"cons(0,?zs:List) = app(cons(s(plus(plus(0,?x:Nat),?x:Nat)),?xs:List),?xs:List)",
"app(?xs:List,rev(?xs:List)) = app(app(?zs:List,app(?ys:List,app(?ys:List,?zs:List))),?ys:List)",
"app(?xs:List,rev(?ys:List)) = cons(s(?y:Nat),app(rev(?zs:List),app(?zs:List,?ys:List)))",
"app(?zs:List,rev(?ys:List)) = rev(rev(rev(app(app(?ys:List,?ys:List),rev(?xs:List)))))",
"rev(app(?ys:List,?ys:List)) = app(cons(?x:Nat,app(?xs:List,app(?ys:List,?zs:List))),?ys:List)",
"cons(len(?ys:List),?xs:List) = app(app(app(?ys:List,?ys:List),app(?xs:List,?zs:List)),?ys:List)",
"app(?xs:List,app(?xs:List,?zs:List)) = app(?ys:List,app(app(rev(?ys:List),?ys:List),?xs:List))",
"app(?ys:List,app(?ys:List,?xs:List)) = app(app(?zs:List,rev(?xs:List)),app(?ys:List,?ys:List))",
"app(?ys:List,app(?ys:List,?xs:List)) = cons(?x:Nat,app(?ys:List,rev(app(?xs:List,?xs:List))))",
"app(?zs:List,cons(0,?ys:List)) = cons(?x:Nat,app(app(?xs:List,rev(?ys:List)),?xs:List))",
"app(app(?xs:List,?xs:List),?zs:List) = app(?xs:List,app(app(?zs:List,?xs:List),rev(?ys:List)))",
"app(app(?xs:List,?ys:List),?xs:List) = cons(len(?xs:List),app(cons(?y:Nat,?xs:List),?zs:List))",
"app(app(?ys:List,?ys:List),?ys:List) = app(app(?ys:List,cons(s(0),?xs:List)),?xs:List)",
"cons(?y:Nat,app(?ys:List,?xs:List)) = app(cons(?x:Nat,rev(?xs:List)),app(?xs:List,?zs:List))",
"cons(?y:Nat,app(?zs:List,?ys:List)) = cons(0,cons(0,rev(app(?zs:List,?zs:List))))",
"app(?xs:List,app(?ys:List,rev(?zs:List))) = app(app(?ys:List,cons(0,?xs:List)),?xs:List)",
"app(?xs:List,rev(app(?ys:List,?zs:List))) = app(app(app(?zs:List,?ys:List),?zs:List),?xs:List)",
"app(?ys:List,app(rev(?zs:List),?xs:List)) = app(rev(rev(?xs:List)),app(?ys:List,?ys:List))",
"app(app(?zs:List,?xs:List),rev(?xs:List)) = rev(app(rev(?xs:List),app(?xs:List,?xs:List)))",
"app(rev(?xs:List),app(?zs:List,?xs:List)) = app(app(?zs:List,app(?zs:List,?zs:List)),?ys:List)",
"rev(app(app(?ys:List,?zs:List),?zs:List)) = cons(?x:Nat,app(?xs:List,cons(0,?xs:List)))",
"cons(s(?x:Nat),app(?ys:List,?xs:List)) = app(app(?zs:List,cons(?x:Nat,?xs:List)),?xs:List)",
"app(?xs:List,rev(cons(len(?xs:List),?xs:List))) = cons(times(?y:Nat,s(0)),?zs:List)",
"app(app(?xs:List,?ys:List),app(?ys:List,?xs:List)) = app(?ys:List,rev(app(?xs:List,?xs:List)))",
"rev(cons(0,cons(0,rev(?xs:List)))) = rev(app(app(?ys:List,?ys:List),?ys:List))",
"app(app(?ys:List,rev(cons(0,?xs:List))),?zs:List) = app(?xs:List,app(?zs:List,?zs:List))",
"app(app(?zs:List,rev(app(?ys:List,?xs:List))),?xs:List) = app(app(?xs:List,?xs:List),?zs:List)",
"app(app(?zs:List,rev(rev(?xs:List))),rev(?ys:List)) = app(?zs:List,rev(rev(?xs:List)))",
"app(app(rev(?xs:List),app(?xs:List,?xs:List)),?zs:List) = app(?xs:List,app(?ys:List,?xs:List))",
"rev(app(?ys:List,cons(?x:Nat,app(?ys:List,?ys:List)))) = app(app(?xs:List,?zs:List),?xs:List)",
"cons(plus(s(times(0,?x:Nat)),?y:Nat),?xs:List) = app(rev(?ys:List),rev(?ys:List))",
"app(?zs:List,app(app(?xs:List,rev(rev(?xs:List))),?zs:List)) = rev(cons(?x:Nat,?ys:List))",
"app(app(?xs:List,?ys:List),cons(times(?x:Nat,0),?ys:List)) = rev(rev(rev(?xs:List)))",
"app(app(?ys:List,?zs:List),app(?zs:List,cons(?y:Nat,?ys:List))) = rev(app(?xs:List,?ys:List))",
"app(app(app(app(?zs:List,?ys:List),?xs:List),?xs:List),?zs:List) = rev(cons(?y:Nat,?xs:List))",
"app(app(rev(cons(?x:Nat,rev(?ys:List))),?zs:List),?xs:List) = app(rev(?zs:List),?xs:List)",
"cons(len(rev(app(?xs:List,app(?ys:List,?zs:List)))),?xs:List) = app(?ys:List,rev(?xs:List))",
"app(?xs:List,app(cons(times(s(?x:Nat),?x:Nat),?ys:List),?zs:List)) = app(?xs:List,?xs:List)",
"app(?ys:List,app(?ys:List,rev(app(app(?ys:List,?zs:List),?zs:List)))) = app(?zs:List,?zs:List)",
"app(?zs:List,app(?xs:List,cons(plus(s(?y:Nat),?x:Nat),?ys:List))) = app(?zs:List,?xs:List)",
"app(app(app(app(rev(?zs:List),?zs:List),?ys:List),?ys:List),?zs:List) = app(?zs:List,?xs:List)",
"app(app(rev(?xs:List),?ys:List),app(app(?xs:List,?xs:List),?ys:List)) = rev(rev(?zs:List))",
"app(app(rev(?xs:List),app(?xs:List,cons(?y:Nat,?xs:List))),?xs:List) = cons(?x:Nat,?zs:List)",
"app(app(rev(app(cons(?x:Nat,?ys:List),?ys:List)),?xs:List),?xs:List) = app(?zs:List,?ys:List)",
"app(cons(0,app(?ys:List,rev(app(?zs:List,?zs:List)))),?zs:List) = app(?ys:List,?xs:List)",
"app(cons(plus(?x:Nat,times(len(?xs:List),?y:Nat)),?xs:List),?xs:List) = app(?ys:List,?xs:List)",
"rev(app(app(?xs:List,app(app(?zs:List,?xs:List),?zs:List)),?ys:List)) = app(?xs:List,?zs:List)",
"cons(0,app(?ys:List,app(rev(app(?ys:List,?ys:List)),?xs:List))) = cons(?x:Nat,?zs:List)",
"cons(plus(0,s(?x:Nat)),app(app(?zs:List,?zs:List),?ys:List)) = app(?ys:List,?xs:List)",
"app(?xs:List,app(app(?zs:List,app(app(?xs:List,?xs:List),?xs:List)),?xs:List)) = rev(?xs:List)",
"app(?xs:List,cons(plus(len(?xs:List),s(plus(?y:Nat,?y:Nat))),?zs:List)) = rev(?xs:List)",
"app(?zs:List,app(cons(?y:Nat,app(?ys:List,app(?zs:List,?zs:List))),?xs:List)) = rev(?xs:List)",
"app(app(?xs:List,app(?xs:List,cons(0,app(?ys:List,?ys:List)))),?xs:List) = rev(?xs:List)",
"app(app(?xs:List,app(app(?xs:List,app(?xs:List,?xs:List)),?xs:List)),?xs:List) = rev(?zs:List)",
"app(app(?xs:List,rev(app(?xs:List,?ys:List))),cons(s(?x:Nat),?xs:List)) = rev(?zs:List)",
"app(app(?ys:List,app(app(app(?zs:List,?ys:List),?ys:List),?xs:List)),?ys:List) = rev(?ys:List)",
"cons(len(cons(len(?ys:List),app(app(?ys:List,?ys:List),?zs:List))),?zs:List) = rev(?ys:List)",
"cons(plus(0,times(plus(plus(?x:Nat,?x:Nat),?y:Nat),?x:Nat)),?zs:List) = rev(?xs:List)",
"rev(?xs:List) = app(?xs:List,app(?xs:List,app(?zs:List,app(rev(?xs:List),cons(?x:Nat,?ys:List)))))",
"rev(?xs:List) = app(?ys:List,app(app(?ys:List,?xs:List),rev(cons(plus(?x:Nat,?y:Nat),?zs:List))))",
"rev(?xs:List) = app(?zs:List,rev(app(cons(?x:Nat,?ys:List),cons(?y:Nat,app(?xs:List,?ys:List)))))",
"rev(?ys:List) = app(app(?zs:List,?xs:List),cons(times(?x:Nat,len(cons(?y:Nat,?xs:List))),?xs:List))",
"rev(?ys:List) = app(rev(rev(rev(?xs:List))),app(?xs:List,app(app(?xs:List,?ys:List),?xs:List)))",
"app(?xs:List,?xs:List) = app(app(?xs:List,app(?ys:List,app(?ys:List,cons(?x:Nat,?ys:List)))),?xs:List)",
"app(?xs:List,?xs:List) = rev(app(cons(len(?zs:List),app(cons(?y:Nat,?ys:List),?zs:List)),?ys:List))",
"app(?xs:List,?xs:List) = cons(?x:Nat,app(?xs:List,app(?xs:List,app(?zs:List,app(?ys:List,?xs:List)))))",
"app(?xs:List,?ys:List) = app(?ys:List,app(?ys:List,app(?zs:List,app(?ys:List,cons(0,?xs:List)))))",
"app(?xs:List,?ys:List) = app(?zs:List,app(?xs:List,app(app(?xs:List,app(?zs:List,?ys:List)),?xs:List)))",
"app(?xs:List,?ys:List) = app(app(?xs:List,rev(app(?zs:List,?ys:List))),rev(app(?zs:List,?xs:List)))",
"app(?xs:List,?ys:List) = app(app(app(?ys:List,?xs:List),app(?xs:List,?ys:List)),app(?xs:List,?zs:List))",
"app(?xs:List,?zs:List) = app(?xs:List,app(rev(cons(len(?xs:List),?xs:List)),cons(?y:Nat,?xs:List)))",
"app(?xs:List,?zs:List) = app(app(app(?zs:List,?xs:List),cons(?x:Nat,?ys:List)),app(?xs:List,?xs:List))",
"app(?xs:List,?zs:List) = cons(plus(s(times(len(?ys:List),?x:Nat)),0),cons(?x:Nat,?xs:List))",
"app(?zs:List,?xs:List) = app(?xs:List,cons(plus(0,times(0,?x:Nat)),app(?zs:List,?xs:List)))",
"app(?zs:List,?xs:List) = app(app(rev(app(rev(?xs:List),?ys:List)),app(?zs:List,?xs:List)),?zs:List)",
"app(?zs:List,?xs:List) = app(app(cons(times(0,plus(?y:Nat,?x:Nat)),?ys:List),?zs:List),?xs:List)",
"app(?zs:List,?xs:List) = cons(times(len(rev(?ys:List)),len(app(rev(?xs:List),?zs:List))),?zs:List)",
"rev(rev(?xs:List)) = rev(app(cons(0,app(rev(?ys:List),app(?xs:List,?zs:List))),?ys:List))",
"cons(?x:Nat,?xs:List) = app(app(?zs:List,app(?ys:List,?zs:List)),app(app(?xs:List,?xs:List),?xs:List))",
"cons(?x:Nat,?ys:List) = app(rev(app(rev(?zs:List),app(app(?ys:List,?xs:List),?zs:List))),?xs:List)",
"cons(?y:Nat,?xs:List) = app(cons(?x:Nat,app(rev(?zs:List),cons(s(0),?ys:List))),?xs:List)",
"cons(?y:Nat,?ys:List) = app(app(?xs:List,?xs:List),app(app(cons(?y:Nat,?xs:List),?xs:List),?xs:List))",
"cons(0,?ys:List) = app(app(cons(plus(plus(0,0),0),?zs:List),?ys:List),?ys:List)",
"cons(0,?zs:List) = app(app(cons(times(0,times(?y:Nat,0)),?xs:List),?xs:List),?xs:List)",
"app(?xs:List,rev(?ys:List)) = cons(times(len(app(?ys:List,?xs:List)),times(0,?x:Nat)),?zs:List)",
"app(rev(?xs:List),?xs:List) = app(rev(rev(?xs:List)),cons(times(s(0),?y:Nat),?xs:List))",
"app(rev(?xs:List),?zs:List) = rev(cons(?x:Nat,app(app(cons(?x:Nat,?xs:List),?xs:List),?zs:List)))",
"app(rev(?xs:List),?zs:List) = cons(plus(plus(?y:Nat,0),?y:Nat),app(rev(?zs:List),?zs:List))",
"rev(rev(rev(?zs:List))) = cons(plus(times(?x:Nat,?y:Nat),?x:Nat),app(rev(?zs:List),?zs:List))",
"cons(?y:Nat,rev(?zs:List)) = rev(cons(times(?x:Nat,times(s(0),s(?x:Nat))),?ys:List))",
"cons(s(?x:Nat),?xs:List) = app(rev(app(app(?zs:List,?zs:List),app(?zs:List,?ys:List))),?ys:List)",
"cons(len(?xs:List),?ys:List) = cons(?y:Nat,cons(?x:Nat,app(?xs:List,app(rev(?ys:List),?zs:List))))",
"app(?xs:List,app(?ys:List,?ys:List)) = app(app(app(?zs:List,?xs:List),?ys:List),cons(?x:Nat,?xs:List))",
"app(?ys:List,app(?zs:List,?ys:List)) = app(app(?xs:List,?xs:List),app(app(?xs:List,?xs:List),?ys:List))",
"app(?zs:List,app(?xs:List,?xs:List)) = app(rev(app(app(rev(?xs:List),?xs:List),?ys:List)),?ys:List)",
"app(?zs:List,app(?xs:List,?zs:List)) = app(rev(app(?xs:List,app(?zs:List,rev(?xs:List)))),?xs:List)",
"app(?zs:List,app(?ys:List,?xs:List)) = app(app(cons(?x:Nat,app(?ys:List,?zs:List)),?zs:List),?zs:List)",
"app(app(?ys:List,?ys:List),?ys:List) = app(cons(s(s(0)),?ys:List),app(?zs:List,?xs:List))",
"app(rev(rev(?zs:List)),?xs:List) = app(cons(?x:Nat,?xs:List),app(?xs:List,app(?zs:List,?xs:List)))",
"app(cons(?x:Nat,?xs:List),?zs:List) = app(?zs:List,app(rev(?xs:List),app(rev(?xs:List),?ys:List)))",
"cons(times(?x:Nat,?x:Nat),?ys:List) = cons(?y:Nat,app(app(?zs:List,?ys:List),cons(?x:Nat,?xs:List)))",
"cons(times(?x:Nat,?x:Nat),?zs:List) = app(app(app(?zs:List,?ys:List),cons(?x:Nat,?xs:List)),?ys:List)",
"app(?xs:List,cons(s(?x:Nat),?xs:List)) = app(rev(cons(?x:Nat,?xs:List)),app(?zs:List,?xs:List))",
"app(?xs:List,cons(len(?xs:List),?xs:List)) = app(cons(?y:Nat,?ys:List),app(rev(?zs:List),?xs:List))",
"app(?ys:List,rev(app(?xs:List,?ys:List))) = rev(app(app(?xs:List,app(?ys:List,?zs:List)),?xs:List))",
"app(?zs:List,app(?xs:List,rev(?ys:List))) = app(?ys:List,rev(app(app(?xs:List,?ys:List),?ys:List)))",
"rev(app(?zs:List,app(?ys:List,?zs:List))) = app(app(?xs:List,rev(app(?ys:List,?xs:List))),?zs:List)",
"cons(plus(len(?xs:List),0),?ys:List) = app(app(?xs:List,app(rev(?zs:List),?xs:List)),?ys:List)",
"cons(plus(len(?zs:List),?y:Nat),?xs:List) = cons(s(0),app(?xs:List,app(?xs:List,?zs:List)))",
"app(?ys:List,app(app(?xs:List,?zs:List),?zs:List)) = cons(0,app(app(?xs:List,?zs:List),?zs:List))",
"app(app(?xs:List,cons(?x:Nat,?xs:List)),?ys:List) = app(app(?ys:List,app(?zs:List,?xs:List)),?ys:List)",
"app(app(?ys:List,?xs:List),app(?xs:List,?zs:List)) = app(?xs:List,app(app(?ys:List,?zs:List),?zs:List))",
"app(cons(plus(0,?y:Nat),?ys:List),?ys:List) = app(?ys:List,app(?xs:List,cons(?x:Nat,?xs:List)))",
"rev(app(app(?ys:List,?ys:List),rev(?ys:List))) = app(?zs:List,app(app(?zs:List,?xs:List),?ys:List))",
"cons(0,app(app(?xs:List,?zs:List),?zs:List)) = app(?zs:List,rev(cons(len(?zs:List),?xs:List)))",
"cons(plus(times(?y:Nat,?x:Nat),?x:Nat),?ys:List) = app(app(?xs:List,cons(?y:Nat,?ys:List)),?ys:List)",
"cons(times(0,0),app(?zs:List,?ys:List)) = app(app(cons(?y:Nat,?ys:List),?xs:List),?xs:List)",
"app(app(rev(?xs:List),?ys:List),app(?xs:List,?ys:List)) = app(rev(app(?ys:List,?xs:List)),?xs:List)",
"rev(app(rev(app(rev(?xs:List),?ys:List)),?xs:List)) = app(app(?xs:List,rev(?ys:List)),?ys:List)",
"cons(?x:Nat,app(?zs:List,app(?xs:List,rev(?ys:List)))) = rev(app(?ys:List,app(?zs:List,?zs:List)))",
"app(?xs:List,cons(?x:Nat,app(?zs:List,app(?zs:List,?zs:List)))) = app(?xs:List,cons(?y:Nat,?zs:List))",
"app(?xs:List,cons(len(cons(s(?x:Nat),?xs:List)),?xs:List)) = app(?zs:List,app(?ys:List,?zs:List))",
"app(?zs:List,app(app(app(?zs:List,?zs:List),?zs:List),?zs:List)) = app(?xs:List,app(?ys:List,?xs:List))",
"app(app(?xs:List,app(?xs:List,app(?xs:List,?ys:List))),?xs:List) = app(app(?xs:List,?xs:List),?xs:List)",
"app(app(?xs:List,cons(times(?x:Nat,0),?zs:List)),?xs:List) = app(?ys:List,app(?xs:List,?ys:List))",
"app(app(?zs:List,app(?xs:List,cons(?y:Nat,?xs:List))),?ys:List) = app(?xs:List,cons(0,?xs:List))",
"app(app(?zs:List,cons(?x:Nat,?xs:List)),app(?xs:List,?zs:List)) = app(?zs:List,app(?xs:List,?zs:List))",
"app(app(app(?xs:List,?ys:List),app(?xs:List,?xs:List)),?zs:List) = app(cons(?x:Nat,?ys:List),?zs:List)",
"app(app(app(?zs:List,?xs:List),?xs:List),app(?xs:List,?xs:List)) = cons(?x:Nat,app(?zs:List,?xs:List))",
"app(app(app(?zs:List,?xs:List),app(?ys:List,?xs:List)),?zs:List) = app(?xs:List,app(?xs:List,?xs:List))",
"rev(rev(app(app(app(?zs:List,?zs:List),?zs:List),?zs:List))) = app(cons(?x:Nat,?ys:List),?xs:List)",
"cons(?y:Nat,app(?zs:List,app(app(?xs:List,?xs:List),?xs:List))) = app(app(?ys:List,?zs:List),?ys:List)",
"cons(len(?zs:List),cons(?x:Nat,app(?zs:List,rev(?xs:List)))) = app(?ys:List,app(?zs:List,?ys:List))",
"app(app(?ys:List,app(?ys:List,cons(len(?xs:List),?xs:List))),?zs:List) = app(?zs:List,rev(?zs:List))",
"app(rev(?xs:List),app(app(rev(rev(?xs:List)),?zs:List),?xs:List)) = rev(app(?zs:List,?zs:List))",
"cons(s(0),app(app(app(?ys:List,?ys:List),?ys:List),?ys:List)) = app(rev(?ys:List),?zs:List)",
"cons(times(?y:Nat,len(app(?zs:List,?xs:List))),app(?ys:List,?ys:List)) = cons(len(?xs:List),?zs:List)",
"app(?xs:List,app(?xs:List,app(?xs:List,app(cons(?x:Nat,?zs:List),?xs:List)))) = app(?zs:List,?zs:List)",
"app(?xs:List,app(app(?xs:List,app(?zs:List,app(?ys:List,?xs:List))),?xs:List)) = app(?xs:List,?xs:List)",
"app(?xs:List,app(app(app(app(?ys:List,?xs:List),?ys:List),?xs:List),?zs:List)) = app(?zs:List,?ys:List)",
"app(?ys:List,app(app(app(?xs:List,?zs:List),app(?zs:List,?xs:List)),?ys:List)) = app(?xs:List,?xs:List)",
"app(?ys:List,cons(s(times(s(?x:Nat),times(?x:Nat,?x:Nat))),?xs:List)) = app(?ys:List,?ys:List)",
"app(?zs:List,app(app(app(?xs:List,app(?zs:List,?zs:List)),?zs:List),?xs:List)) = app(?ys:List,?zs:List)",
"app(?zs:List,app(app(app(?zs:List,?xs:List),app(?ys:List,?ys:List)),?xs:List)) = app(?zs:List,?xs:List)",
"app(?zs:List,app(app(app(?zs:List,?zs:List),app(?ys:List,?ys:List)),?xs:List)) = app(?xs:List,?xs:List)",
"app(app(?xs:List,?xs:List),app(app(?xs:List,?xs:List),app(?xs:List,?zs:List))) = app(?xs:List,?zs:List)",
"app(app(?xs:List,?ys:List),app(cons(s(len(?zs:List)),?ys:List),?xs:List)) = app(?zs:List,?zs:List)",
"app(app(?ys:List,app(rev(?xs:List),cons(len(?ys:List),?zs:List))),?zs:List) = app(?ys:List,?xs:List)",
"app(app(?zs:List,app(?xs:List,?ys:List)),app(?xs:List,app(?zs:List,?xs:List))) = app(?ys:List,?xs:List)",
"app(app(?zs:List,cons(?y:Nat,cons(plus(?y:Nat,?x:Nat),?xs:List))),?xs:List) = cons(?y:Nat,?xs:List)",
"app(app(app(?ys:List,?ys:List),rev(?xs:List)),cons(?x:Nat,rev(?ys:List))) = app(?zs:List,?xs:List)",
"app(app(app(app(app(?zs:List,?ys:List),?zs:List),?xs:List),?xs:List),?ys:List) = app(?xs:List,?zs:List)",
"rev(app(app(?ys:List,app(rev(?xs:List),app(?ys:List,?ys:List))),?xs:List)) = app(?xs:List,?xs:List)",
"rev(app(app(?zs:List,?zs:List),app(?zs:List,app(?zs:List,rev(?xs:List))))) = cons(?y:Nat,?xs:List)",
"rev(app(cons(?x:Nat,?zs:List),app(rev(app(?zs:List,?ys:List)),?xs:List))) = cons(?x:Nat,?xs:List)",
"rev(cons(plus(len(?zs:List),plus(plus(?x:Nat,0),?y:Nat)),?xs:List)) = app(?xs:List,?zs:List)",
"cons(len(app(?zs:List,?ys:List)),app(app(?ys:List,rev(?xs:List)),?zs:List)) = rev(rev(?xs:List))",
"app(?zs:List,app(cons(plus(times(?x:Nat,0),s(?y:Nat)),?zs:List),?xs:List)) = rev(?zs:List)",
"app(app(?xs:List,rev(app(app(?xs:List,?ys:List),cons(0,?xs:List)))),?zs:List) = rev(?xs:List)",
"app(app(?ys:List,?xs:List),app(cons(len(?zs:List),?ys:List),cons(0,?xs:List))) = rev(?xs:List)",
"app(app(rev(app(?zs:List,rev(?xs:List))),app(?ys:List,rev(?xs:List))),?zs:List) = rev(?zs:List)",
"app(app(cons(times(?x:Nat,times(?x:Nat,?y:Nat)),?zs:List),rev(?ys:List)),?ys:List) = rev(?xs:List)",
"app(rev(cons(times(plus(0,?x:Nat),plus(?x:Nat,?x:Nat)),?xs:List)),?ys:List) = rev(?xs:List)",
"rev(app(app(app(app(?xs:List,?xs:List),?ys:List),cons(?x:Nat,?zs:List)),?zs:List)) = rev(?ys:List)",
"cons(?x:Nat,app(app(?xs:List,app(app(?ys:List,?ys:List),rev(?ys:List))),?xs:List)) = rev(?zs:List)",
"cons(len(?ys:List),app(?xs:List,app(?xs:List,app(app(?zs:List,?zs:List),?ys:List)))) = rev(?xs:List)",
"cons(plus(len(app(app(?xs:List,?zs:List),?ys:List)),0),cons(?x:Nat,?xs:List)) = rev(?ys:List)",
"rev(?xs:List) = app(?xs:List,app(app(cons(times(?y:Nat,?y:Nat),?zs:List),cons(0,?xs:List)),?xs:List))",
"rev(?xs:List) = app(?ys:List,app(app(?xs:List,?zs:List),cons(0,app(?xs:List,app(?xs:List,?ys:List)))))",
"rev(?xs:List) = app(app(app(?ys:List,cons(?x:Nat,app(?xs:List,?xs:List))),?ys:List),app(?xs:List,?xs:List))",
"rev(?xs:List) = app(app(app(?zs:List,?xs:List),cons(0,app(app(?ys:List,?zs:List),?xs:List))),?zs:List)",
"rev(?xs:List) = cons(?x:Nat,app(app(app(?xs:List,?zs:List),app(app(?zs:List,?ys:List),?ys:List)),?ys:List))",
"rev(?xs:List) = cons(?x:Nat,app(cons(times(times(?x:Nat,0),?y:Nat),app(?xs:List,?xs:List)),?xs:List))",
"rev(?xs:List) = cons(s(?x:Nat),app(?ys:List,cons(times(?x:Nat,?x:Nat),rev(app(?ys:List,?xs:List)))))",
"rev(?ys:List) = app(cons(times(plus(?x:Nat,?y:Nat),?x:Nat),app(cons(0,?xs:List),?xs:List)),?ys:List)",
"rev(?zs:List) = app(cons(len(?zs:List),app(app(rev(?ys:List),?zs:List),?zs:List)),app(?ys:List,?zs:List))",
"rev(?zs:List) = rev(app(?xs:List,app(app(?zs:List,app(?xs:List,?xs:List)),rev(app(?ys:List,?ys:List)))))",
"app(?xs:List,?xs:List) = app(app(?ys:List,?xs:List),app(rev(cons(0,?zs:List)),app(?ys:List,?xs:List)))",
"app(?xs:List,?xs:List) = app(app(app(?ys:List,rev(app(?xs:List,?xs:List))),?xs:List),cons(?x:Nat,?xs:List))",
"app(?xs:List,?xs:List) = app(app(app(rev(?xs:List),rev(?xs:List)),app(?xs:List,rev(?ys:List))),?xs:List)",
"app(?xs:List,?ys:List) = app(app(app(?ys:List,cons(s(?y:Nat),?zs:List)),?zs:List),cons(?x:Nat,?xs:List))",
"app(?ys:List,?xs:List) = app(?xs:List,app(?ys:List,app(?xs:List,app(app(?zs:List,?xs:List),rev(?ys:List)))))",
"app(?ys:List,?xs:List) = app(app(?ys:List,app(cons(s(0),?zs:List),?xs:List)),rev(rev(?zs:List)))",
"app(?ys:List,?zs:List) = app(app(?zs:List,?xs:List),app(cons(len(?zs:List),app(?ys:List,?zs:List)),?zs:List))",
"app(?ys:List,?zs:List) = cons(plus(len(app(app(?xs:List,?zs:List),?xs:List)),times(0,?y:Nat)),?ys:List)",
"app(?zs:List,?xs:List) = app(?xs:List,app(rev(rev(?xs:List)),app(cons(?x:Nat,rev(?zs:List)),?xs:List)))",
"app(?zs:List,?zs:List) = app(app(?xs:List,app(app(?zs:List,?xs:List),?xs:List)),cons(len(?xs:List),?xs:List))",
"cons(?x:Nat,?xs:List) = app(app(app(app(rev(?ys:List),app(?ys:List,?ys:List)),?zs:List),?zs:List),?ys:List)",
"cons(?x:Nat,?zs:List) = app(?zs:List,cons(plus(times(?x:Nat,?x:Nat),s(s(?y:Nat))),rev(?xs:List)))",
"cons(?x:Nat,?zs:List) = app(cons(?y:Nat,app(?ys:List,cons(s(0),?xs:List))),app(?xs:List,?xs:List))",
"cons(0,?ys:List) = app(app(?xs:List,app(?xs:List,?zs:List)),app(app(?zs:List,rev(?zs:List)),?zs:List))",
"cons(0,?zs:List) = app(rev(app(?xs:List,?ys:List)),cons(0,app(?ys:List,app(?ys:List,?zs:List))))",
"app(?ys:List,rev(?zs:List)) = app(app(app(cons(?x:Nat,app(?ys:List,?ys:List)),?xs:List),?ys:List),?xs:List)",
"app(rev(?zs:List),?zs:List) = cons(?x:Nat,cons(times(?x:Nat,?x:Nat),app(?xs:List,app(?ys:List,?zs:List))))",
"rev(app(?ys:List,?xs:List)) = rev(app(cons(times(0,?y:Nat),rev(app(?xs:List,?xs:List))),?xs:List))",
"rev(app(?zs:List,?zs:List)) = app(app(?xs:List,?zs:List),app(app(app(?xs:List,?xs:List),?xs:List),?xs:List))",
"cons(s(?x:Nat),?ys:List) = rev(app(app(?xs:List,rev(?ys:List)),app(app(?xs:List,?xs:List),?ys:List)))",
"cons(len(?zs:List),?xs:List) = app(app(?zs:List,cons(times(plus(?y:Nat,?x:Nat),?y:Nat),?xs:List)),?ys:List)",
"app(?xs:List,app(?xs:List,?xs:List)) = app(?xs:List,app(?ys:List,cons(?x:Nat,app(?zs:List,rev(?ys:List)))))",
"app(?ys:List,app(?ys:List,?xs:List)) = cons(?x:Nat,cons(plus(?x:Nat,len(cons(?x:Nat,?xs:List))),?xs:List))",
"app(?zs:List,cons(?x:Nat,?xs:List)) = app(?xs:List,app(?xs:List,cons(s(plus(0,?x:Nat)),?xs:List)))",
"app(app(?xs:List,?xs:List),?ys:List) = app(rev(app(?xs:List,?zs:List)),app(?xs:List,app(?xs:List,?zs:List)))",
"app(app(?xs:List,?ys:List),?xs:List) = rev(app(?xs:List,app(?ys:List,app(app(?xs:List,?xs:List),?ys:List))))",
"app(app(?xs:List,?zs:List),?xs:List) = cons(times(?x:Nat,times(0,plus(s(0),?x:Nat))),?xs:List)",
"app(app(?ys:List,?xs:List),?zs:List) = app(?zs:List,app(?xs:List,rev(app(?zs:List,app(?zs:List,?xs:List)))))",
"rev(app(?ys:List,rev(?zs:List))) = cons(?y:Nat,app(cons(len(cons(?x:Nat,?zs:List)),?ys:List),?xs:List))",
"cons(?x:Nat,app(?ys:List,?xs:List)) = cons(times(times(0,s(?x:Nat)),plus(?y:Nat,?x:Nat)),?xs:List)",
"cons(plus(?y:Nat,0),?ys:List) = cons(times(?x:Nat,?x:Nat),app(?zs:List,cons(s(?x:Nat),?zs:List)))",
"app(?xs:List,rev(app(?zs:List,?zs:List))) = cons(plus(s(?x:Nat),?x:Nat),app(?ys:List,rev(?ys:List)))",
"app(app(?zs:List,rev(?ys:List)),?xs:List) = app(?xs:List,app(rev(?zs:List),app(rev(?xs:List),?zs:List)))",
"rev(app(app(?ys:List,?ys:List),?ys:List)) = app(cons(plus(?y:Nat,?x:Nat),app(?xs:List,?zs:List)),?zs:List)",
"cons(s(times(?x:Nat,0)),?zs:List) = app(cons(?y:Nat,app(?xs:List,app(?xs:List,?xs:List))),?xs:List)",
"cons(times(len(?xs:List),?x:Nat),?xs:List) = app(?zs:List,rev(app(app(?zs:List,rev(?ys:List)),?ys:List)))",
"app(?xs:List,app(app(?ys:List,?xs:List),?xs:List)) = cons(?x:Nat,app(rev(?xs:List),app(?xs:List,?ys:List)))",
"app(app(?ys:List,?ys:List),app(?xs:List,?zs:List)) = cons(0,cons(len(cons(?y:Nat,?zs:List)),?xs:List))",
"app(app(?zs:List,cons(?x:Nat,?zs:List)),?zs:List) = app(cons(s(plus(?y:Nat,?x:Nat)),?ys:List),?xs:List)",
"cons(?x:Nat,app(?ys:List,app(?zs:List,?ys:List))) = cons(s(s(times(?x:Nat,s(?y:Nat)))),?zs:List)",
"app(?xs:List,rev(app(rev(rev(?xs:List)),?ys:List))) = app(app(?xs:List,?xs:List),app(?ys:List,?xs:List))",
"app(?xs:List,cons(0,app(?zs:List,rev(?zs:List)))) = app(cons(plus(?x:Nat,?y:Nat),?zs:List),?ys:List)",
"app(?ys:List,app(rev(?xs:List),app(?ys:List,?ys:List))) = app(app(?zs:List,app(?xs:List,?xs:List)),?xs:List)",
"app(?zs:List,app(app(rev(?xs:List),?zs:List),?zs:List)) = rev(app(app(?xs:List,?xs:List),rev(?ys:List)))",
"app(app(?zs:List,app(?xs:List,rev(?zs:List))),?xs:List) = app(app(?zs:List,?ys:List),app(?xs:List,?xs:List))",
"cons(len(app(?xs:List,?xs:List)),rev(rev(?xs:List))) = app(app(?xs:List,?ys:List),app(?xs:List,?xs:List))",
"app(?zs:List,app(rev(app(rev(?xs:List),?xs:List)),?ys:List)) = rev(cons(s(len(?xs:List)),?ys:List))",
"app(app(app(?ys:List,?xs:List),app(?zs:List,?ys:List)),?zs:List) = app(app(?xs:List,?xs:List),rev(?xs:List))",
"rev(cons(plus(0,len(app(?zs:List,?ys:List))),?zs:List)) = app(?ys:List,app(rev(?ys:List),?xs:List))",
"cons(?x:Nat,app(cons(?x:Nat,?zs:List),app(?ys:List,?xs:List))) = app(rev(?xs:List),rev(rev(?xs:List)))",
"app(app(?zs:List,app(?zs:List,app(?zs:List,?zs:List))),rev(?zs:List)) = app(?zs:List,app(?ys:List,?xs:List))",
"app(rev(?xs:List),app(?xs:List,rev(app(?xs:List,rev(?ys:List))))) = cons(?x:Nat,app(?zs:List,?xs:List))",
"app(cons(len(?ys:List),?xs:List),cons(plus(?x:Nat,?y:Nat),?xs:List)) = rev(cons(len(?ys:List),?ys:List))",
"cons(0,app(app(app(?ys:List,?xs:List),?zs:List),rev(?zs:List))) = app(app(?zs:List,?ys:List),?ys:List)",
"cons(plus(len(cons(?x:Nat,app(?zs:List,?xs:List))),?x:Nat),?xs:List) = cons(plus(?x:Nat,0),?ys:List)",
"app(?zs:List,app(app(app(?xs:List,app(?ys:List,?xs:List)),?ys:List),?ys:List)) = app(rev(?xs:List),?xs:List)",
"app(app(?xs:List,cons(times(?x:Nat,times(?y:Nat,0)),?zs:List)),?zs:List) = cons(s(?x:Nat),?xs:List)",
"app(rev(rev(app(?zs:List,app(?xs:List,?ys:List)))),app(?xs:List,?ys:List)) = cons(len(?ys:List),?xs:List)",
"cons(times(times(times(s(?x:Nat),s(?y:Nat)),?x:Nat),?x:Nat),?ys:List) = app(?ys:List,rev(?xs:List))",
"app(?xs:List,app(app(?xs:List,app(rev(?zs:List),?xs:List)),app(?xs:List,?xs:List))) = app(?zs:List,?xs:List)",
"app(?xs:List,cons(?y:Nat,app(app(rev(?zs:List),?ys:List),app(?xs:List,?ys:List)))) = app(?zs:List,?xs:List)",
"app(app(?xs:List,app(rev(app(?xs:List,?xs:List)),?ys:List)),app(?ys:List,?ys:List)) = app(?ys:List,?xs:List)",
"app(app(?ys:List,?xs:List),app(app(?xs:List,app(rev(?xs:List),?xs:List)),?ys:List)) = app(?zs:List,?xs:List)",
"app(app(?zs:List,?zs:List),rev(app(app(app(?zs:List,?ys:List),?xs:List),?xs:List))) = app(?xs:List,?xs:List)",
"app(app(?zs:List,rev(rev(rev(?xs:List)))),app(?xs:List,app(?xs:List,?xs:List))) = app(?xs:List,?ys:List)",
"app(app(app(app(?ys:List,rev(?zs:List)),?xs:List),app(?ys:List,?ys:List)),?zs:List) = cons(?x:Nat,?xs:List)",
"app(app(app(app(rev(?zs:List),?ys:List),?zs:List),?ys:List),cons(?y:Nat,?ys:List)) = app(?zs:List,?xs:List)",
"app(app(rev(?ys:List),app(?xs:List,?xs:List)),app(app(?ys:List,?ys:List),?xs:List)) = app(?zs:List,?zs:List)",
"rev(cons(0,app(app(app(?xs:List,?zs:List),cons(?x:Nat,?xs:List)),?zs:List))) = app(?zs:List,?xs:List)",
"cons(plus(?x:Nat,len(app(app(?ys:List,?ys:List),rev(?xs:List)))),rev(?xs:List)) = app(?xs:List,?xs:List)",
"cons(plus(s(s(times(?x:Nat,?x:Nat))),times(?y:Nat,s(0))),?xs:List) = app(?xs:List,?xs:List)",
"cons(plus(len(cons(times(0,?y:Nat),app(?zs:List,?zs:List))),?x:Nat),?ys:List) = app(?xs:List,?ys:List)",
"cons(plus(plus(len(app(?xs:List,?xs:List)),?y:Nat),?x:Nat),cons(0,?xs:List)) = app(?zs:List,?ys:List)",
"cons(times(times(times(?x:Nat,len(?xs:List)),?x:Nat),plus(0,0)),?xs:List) = app(?zs:List,?zs:List)",
"app(?ys:List,app(?xs:List,cons(?x:Nat,app(app(?ys:List,?ys:List),app(?xs:List,?zs:List))))) = rev(?zs:List)",
"app(?zs:List,app(app(?ys:List,?zs:List),app(rev(?xs:List),rev(cons(?x:Nat,?xs:List))))) = rev(?zs:List)",
"app(app(?xs:List,?ys:List),rev(app(rev(app(?ys:List,?zs:List)),app(?zs:List,?xs:List)))) = rev(?xs:List)",
"app(app(?ys:List,?ys:List),app(app(app(?zs:List,app(?xs:List,?zs:List)),?xs:List),?zs:List)) = rev(?xs:List)",
"app(app(?ys:List,app(app(?xs:List,?xs:List),app(app(?xs:List,?zs:List),?xs:List))),?zs:List) = rev(?xs:List)",
"app(app(?zs:List,app(?xs:List,app(?xs:List,app(?ys:List,?xs:List)))),app(?xs:List,?zs:List)) = rev(?xs:List)",
"app(app(app(?zs:List,app(app(?xs:List,?xs:List),cons(0,?xs:List))),?zs:List),?xs:List) = rev(?ys:List)",
"cons(plus(0,0),app(?xs:List,app(app(app(?xs:List,?xs:List),?zs:List),?ys:List))) = rev(?zs:List)",
"rev(?xs:List) = app(?xs:List,cons(times(len(app(?zs:List,?xs:List)),0),app(?ys:List,app(?xs:List,?xs:List))))",
"rev(?xs:List) = cons(times(?x:Nat,times(0,len(?ys:List))),app(app(rev(rev(?xs:List)),?xs:List),?xs:List))",
"rev(?ys:List) = app(?ys:List,app(app(app(?xs:List,?xs:List),?ys:List),cons(len(?zs:List),app(?zs:List,?zs:List))))",
"rev(?ys:List) = app(app(cons(s(?x:Nat),?ys:List),?xs:List),app(?xs:List,app(app(?xs:List,?xs:List),?ys:List)))",
"rev(?zs:List) = app(app(app(?xs:List,?ys:List),?zs:List),app(app(?zs:List,cons(?y:Nat,rev(?xs:List))),?ys:List))",
"rev(?zs:List) = rev(app(rev(rev(?xs:List)),app(app(app(?zs:List,?ys:List),?zs:List),app(?xs:List,?xs:List))))",
"app(?xs:List,?xs:List) = app(app(app(?xs:List,app(app(?xs:List,?xs:List),?ys:List)),?xs:List),app(?xs:List,?xs:List))",
"app(?xs:List,?xs:List) = cons(len(?xs:List),cons(len(app(?zs:List,?ys:List)),app(cons(?x:Nat,?zs:List),?xs:List)))",
"app(?xs:List,?zs:List) = app(app(cons(s(times(?x:Nat,?y:Nat)),rev(cons(0,?xs:List))),?xs:List),?xs:List)",
"app(?ys:List,?xs:List) = app(app(app(?xs:List,app(?zs:List,?xs:List)),app(app(?zs:List,?ys:List),?xs:List)),?ys:List)",
"app(?ys:List,?xs:List) = cons(?y:Nat,app(app(?ys:List,app(app(?zs:List,?xs:List),app(?xs:List,?xs:List))),?ys:List))",
"app(?ys:List,?zs:List) = app(app(app(?ys:List,?zs:List),app(?xs:List,app(app(?xs:List,?xs:List),?ys:List))),?xs:List)",
"app(?ys:List,?zs:List) = cons(s(s(times(len(?xs:List),?y:Nat))),rev(rev(app(rev(?zs:List),?ys:List))))",
"app(?zs:List,?xs:List) = cons(s(len(app(?xs:List,?ys:List))),cons(s(s(times(?x:Nat,?y:Nat))),?zs:List))",
"app(?zs:List,?zs:List) = app(app(?xs:List,cons(s(len(?xs:List)),app(?xs:List,app(?xs:List,?zs:List)))),?xs:List)",
"app(?zs:List,?zs:List) = app(app(app(?zs:List,app(?xs:List,?xs:List)),?zs:List),app(cons(?y:Nat,?xs:List),?ys:List))",
"app(?zs:List,?zs:List) = app(cons(?x:Nat,?xs:List),cons(len(rev(app(?xs:List,?xs:List))),app(?zs:List,?ys:List)))",
"app(?zs:List,?zs:List) = cons(s(times(?x:Nat,len(?ys:List))),app(app(?zs:List,app(?xs:List,?xs:List)),?xs:List))",
"cons(?x:Nat,?ys:List) = app(app(app(?xs:List,app(rev(?ys:List),?zs:List)),rev(?ys:List)),app(?xs:List,?xs:List))",
"cons(?x:Nat,?zs:List) = cons(0,cons(plus(plus(?x:Nat,?y:Nat),plus(?x:Nat,plus(?x:Nat,?x:Nat))),?xs:List))",
"cons(?y:Nat,?xs:List) = app(app(app(?xs:List,app(app(?ys:List,?xs:List),app(?zs:List,?zs:List))),?xs:List),?xs:List)",
"cons(?y:Nat,?zs:List) = app(app(app(?xs:List,?zs:List),app(app(?ys:List,?ys:List),rev(rev(?zs:List)))),?xs:List)",
"cons(0,?xs:List) = app(app(?zs:List,?xs:List),app(app(?ys:List,app(?ys:List,cons(?y:Nat,?zs:List))),?xs:List))",
"cons(0,?ys:List) = app(app(app(app(cons(0,?zs:List),?xs:List),app(?zs:List,?ys:List)),?ys:List),?ys:List)",
"cons(0,?zs:List) = cons(times(s(plus(s(?y:Nat),plus(?y:Nat,?x:Nat))),0),app(?zs:List,?zs:List))",
"app(?xs:List,rev(?xs:List)) = app(?xs:List,rev(app(?zs:List,app(app(?xs:List,?zs:List),app(?xs:List,?zs:List)))))",
"app(?xs:List,rev(?zs:List)) = cons(plus(times(?y:Nat,plus(s(0),times(?x:Nat,?x:Nat))),?y:Nat),?zs:List)",
"app(?zs:List,rev(?xs:List)) = app(app(?ys:List,app(?xs:List,rev(?xs:List))),app(app(?zs:List,?xs:List),?xs:List))",
"app(?zs:List,rev(?zs:List)) = app(cons(?y:Nat,cons(0,?ys:List)),rev(app(cons(0,?ys:List),?zs:List)))",
"rev(app(?ys:List,?zs:List)) = app(?xs:List,app(app(app(app(?zs:List,?zs:List),?xs:List),?ys:List),rev(?ys:List)))",
"rev(cons(?x:Nat,?xs:List)) = cons(plus(?x:Nat,times(?x:Nat,?x:Nat)),app(?ys:List,app(?ys:List,rev(?zs:List))))",
"app(?xs:List,app(?xs:List,?xs:List)) = app(app(?xs:List,app(app(app(?xs:List,?ys:List),?xs:List),?zs:List)),?xs:List)",
"app(?xs:List,app(?zs:List,?xs:List)) = app(app(?ys:List,app(app(rev(?xs:List),rev(?xs:List)),?xs:List)),?xs:List)",
"app(?xs:List,rev(rev(?zs:List))) = cons(s(times(?x:Nat,s(?y:Nat))),app(app(?xs:List,?zs:List),?xs:List))",
"app(?zs:List,cons(0,?zs:List)) = app(?xs:List,app(app(?xs:List,?zs:List),app(app(?xs:List,?xs:List),?ys:List)))",
"app(app(?ys:List,?xs:List),?zs:List) = cons(plus(len(?xs:List),s(len(rev(?xs:List)))),app(?ys:List,?zs:List))",
"cons(0,app(?zs:List,?ys:List)) = app(?xs:List,app(?zs:List,app(?ys:List,app(app(?zs:List,?zs:List),?zs:List))))",
"cons(plus(?y:Nat,?x:Nat),?zs:List) = app(app(?xs:List,rev(?xs:List)),cons(len(?xs:List),cons(0,?xs:List)))",
"cons(times(?y:Nat,0),?zs:List) = app(app(?zs:List,cons(times(0,0),app(?xs:List,?xs:List))),?zs:List)",
"app(?xs:List,rev(app(?ys:List,?ys:List))) = app(rev(app(app(?zs:List,app(?xs:List,?ys:List)),?xs:List)),?xs:List)",
"app(?zs:List,app(?xs:List,rev(?xs:List))) = app(?zs:List,app(app(?zs:List,?ys:List),rev(app(?xs:List,?xs:List))))",
"app(cons(len(?ys:List),?xs:List),?zs:List) = app(cons(?x:Nat,app(cons(s(?y:Nat),?xs:List),?ys:List)),?ys:List)",
"app(?ys:List,app(app(?xs:List,?ys:List),?xs:List)) = app(app(app(?xs:List,rev(?ys:List)),?ys:List),rev(?xs:List))",
"app(app(?ys:List,?ys:List),app(?xs:List,?xs:List)) = cons(plus(0,times(?y:Nat,times(?y:Nat,?x:Nat))),?ys:List)",
"app(app(?ys:List,?ys:List),cons(?y:Nat,?ys:List)) = app(app(?xs:List,app(?ys:List,?xs:List)),cons(0,?zs:List))",
"app(app(app(?zs:List,?ys:List),?xs:List),?xs:List) = app(app(?zs:List,cons(?y:Nat,app(?zs:List,?xs:List))),?ys:List)",
"cons(len(?xs:List),app(rev(?ys:List),?xs:List)) = app(app(cons(0,?ys:List),?ys:List),cons(?x:Nat,?zs:List))",
"app(app(?xs:List,app(?xs:List,rev(?zs:List))),?ys:List) = rev(app(app(?xs:List,app(?xs:List,?zs:List)),?xs:List))",
"app(app(?xs:List,cons(len(?ys:List),?ys:List)),?xs:List) = app(?ys:List,cons(times(s(0),?x:Nat),?zs:List))",
"cons(len(?zs:List),app(?xs:List,cons(?x:Nat,?xs:List))) = cons(plus(0,len(?ys:List)),app(?xs:List,?xs:List))",
"app(app(?ys:List,?ys:List),cons(?x:Nat,cons(?x:Nat,?xs:List))) = app(app(app(?zs:List,?ys:List),?xs:List),?zs:List)",
"app(app(?ys:List,app(cons(?y:Nat,?xs:List),?zs:List)),?xs:List) = app(app(?xs:List,app(?zs:List,?ys:List)),?ys:List)",
"app(rev(?zs:List),app(?ys:List,app(?ys:List,rev(?xs:List)))) = app(?ys:List,app(?zs:List,cons(?x:Nat,?ys:List)))",
"app(rev(app(app(?zs:List,rev(?xs:List)),?ys:List)),?zs:List) = app(app(?ys:List,?xs:List),app(?zs:List,?zs:List))",
"app(cons(?y:Nat,app(app(?xs:List,?zs:List),?ys:List)),?xs:List) = app(rev(cons(len(?xs:List),?zs:List)),?xs:List)",
"rev(rev(cons(times(times(?x:Nat,?x:Nat),0),?xs:List))) = app(app(cons(?x:Nat,?zs:List),?zs:List),?zs:List)",
"cons(?x:Nat,app(?ys:List,cons(?x:Nat,cons(?y:Nat,?xs:List)))) = app(app(app(?ys:List,?ys:List),?xs:List),?xs:List)",
"app(?xs:List,app(?xs:List,cons(times(?y:Nat,s(?y:Nat)),?xs:List))) = cons(?x:Nat,app(rev(?zs:List),?xs:List))",
"app(?xs:List,app(cons(times(0,0),rev(?xs:List)),?xs:List)) = app(?xs:List,app(?zs:List,rev(?ys:List)))",
"app(?xs:List,cons(len(app(rev(?xs:List),?ys:List)),cons(?y:Nat,?zs:List))) = rev(rev(cons(?y:Nat,?xs:List)))",
"app(app(?ys:List,app(?xs:List,app(app(?zs:List,?ys:List),?xs:List))),?xs:List) = app(app(?ys:List,?ys:List),?xs:List)",
"app(app(?ys:List,app(app(cons(0,?zs:List),?xs:List),?ys:List)),?zs:List) = app(cons(?x:Nat,?xs:List),?ys:List)",
"app(app(app(?ys:List,cons(plus(?y:Nat,?x:Nat),?ys:List)),?xs:List),?xs:List) = app(?zs:List,app(?zs:List,?xs:List))",
"app(app(rev(rev(?zs:List)),app(?xs:List,?xs:List)),cons(0,?xs:List)) = app(app(?ys:List,?ys:List),?xs:List)",
"app(app(cons(?x:Nat,app(?xs:List,rev(?xs:List))),?zs:List),rev(?xs:List)) = app(cons(0,?xs:List),?xs:List)",
"app(rev(?xs:List),app(rev(app(?xs:List,app(?ys:List,?xs:List))),?ys:List)) = app(?zs:List,cons(?x:Nat,?zs:List))",
"cons(?x:Nat,app(?zs:List,cons(?x:Nat,app(cons(?x:Nat,?ys:List),?ys:List)))) = app(?xs:List,app(?xs:List,?xs:List))",
"app(?xs:List,app(rev(app(app(?ys:List,?xs:List),?zs:List)),app(?zs:List,?ys:List))) = rev(app(?ys:List,?ys:List))",
"app(app(?xs:List,app(?ys:List,app(app(?zs:List,?xs:List),rev(?xs:List)))),?zs:List) = app(?xs:List,rev(?zs:List))",
"app(app(?xs:List,cons(0,rev(?xs:List))),app(app(?ys:List,?zs:List),?zs:List)) = cons(?x:Nat,rev(?ys:List))",
"app(app(?ys:List,app(?xs:List,?xs:List)),app(?ys:List,app(?xs:List,rev(?xs:List)))) = cons(len(?zs:List),?xs:List)",
"app(app(app(?ys:List,?xs:List),?xs:List),app(?ys:List,app(?ys:List,rev(?xs:List)))) = rev(app(?ys:List,?ys:List))",
"app(rev(app(?zs:List,cons(plus(?x:Nat,?x:Nat),app(?ys:List,?zs:List)))),?xs:List) = app(?zs:List,rev(?xs:List))",
"app(cons(plus(plus(plus(?y:Nat,0),len(?zs:List)),?x:Nat),?xs:List),?ys:List) = app(rev(?ys:List),?zs:List)",
"rev(app(app(cons(plus(?y:Nat,?y:Nat),?ys:List),?zs:List),cons(?y:Nat,?zs:List))) = app(?xs:List,rev(?zs:List))",
"cons(len(rev(app(?xs:List,app(?xs:List,?ys:List)))),rev(app(?xs:List,?ys:List))) = cons(?x:Nat,rev(?ys:List))",
"app(?xs:List,app(cons(s(times(0,?x:Nat)),rev(cons(?y:Nat,?ys:List))),?ys:List)) = app(?ys:List,?xs:List)",
"app(?ys:List,app(?zs:List,cons(?y:Nat,app(app(?xs:List,?ys:List),app(?xs:List,?zs:List))))) = cons(?y:Nat,?zs:List)",
"app(?zs:List,app(?ys:List,app(app(cons(?x:Nat,?xs:List),app(?xs:List,?ys:List)),?zs:List))) = cons(?x:Nat,?zs:List)",
"app(?zs:List,rev(cons(len(?ys:List),app(app(?ys:List,?xs:List),app(?zs:List,?xs:List))))) = app(?xs:List,?ys:List)",
"app(?zs:List,cons(?x:Nat,app(?xs:List,app(cons(?x:Nat,?xs:List),app(?xs:List,?ys:List))))) = app(?zs:List,?xs:List)",
"app(?zs:List,cons(?y:Nat,app(app(?xs:List,app(?zs:List,?ys:List)),app(?xs:List,?xs:List)))) = app(?zs:List,?xs:List)",
"app(app(?xs:List,app(?xs:List,app(?ys:List,app(?ys:List,?xs:List)))),app(?xs:List,?ys:List)) = app(?ys:List,?xs:List)",
"app(app(?zs:List,?xs:List),app(?xs:List,cons(plus(?x:Nat,plus(?x:Nat,?x:Nat)),?xs:List))) = app(?xs:List,?xs:List)",
"app(app(?zs:List,app(app(?zs:List,app(?ys:List,?zs:List)),?xs:List)),app(?xs:List,?zs:List)) = app(?ys:List,?zs:List)",
"app(app(app(app(?ys:List,?ys:List),app(?xs:List,?xs:List)),app(?xs:List,?ys:List)),?zs:List) = cons(?y:Nat,?zs:List)",
"app(app(app(cons(?x:Nat,?ys:List),?zs:List),cons(plus(?x:Nat,?x:Nat),?xs:List)),?xs:List) = app(?zs:List,?ys:List)",
"app(app(cons(plus(times(?y:Nat,?x:Nat),s(?y:Nat)),?zs:List),rev(?ys:List)),?ys:List) = app(?xs:List,?ys:List)",
"app(rev(?ys:List),rev(cons(?x:Nat,cons(times(?x:Nat,?x:Nat),app(?xs:List,?xs:List))))) = app(?xs:List,?ys:List)",
"cons(times(times(s(plus(0,?y:Nat)),0),len(?xs:List)),app(?xs:List,?ys:List)) = cons(?x:Nat,?zs:List)",
"app(app(?xs:List,?xs:List),app(?ys:List,app(cons(?y:Nat,?zs:List),app(rev(?zs:List),?xs:List)))) = rev(?xs:List)",
"app(app(app(?xs:List,?zs:List),?ys:List),app(rev(app(app(?ys:List,?zs:List),?ys:List)),?ys:List)) = rev(?zs:List)",
"app(app(app(app(app(?zs:List,?zs:List),?xs:List),rev(app(?xs:List,?xs:List))),?ys:List),?xs:List) = rev(?ys:List)",
"app(app(rev(?ys:List),?ys:List),app(app(app(app(?zs:List,?ys:List),?xs:List),?ys:List),?zs:List)) = rev(?xs:List)",
"cons(times(len(app(?ys:List,?xs:List)),0),app(app(app(?xs:List,?zs:List),?xs:List),?xs:List)) = rev(?xs:List)",
"rev(?ys:List) = app(?xs:List,app(app(?zs:List,app(app(?ys:List,?zs:List),app(?zs:List,?xs:List))),app(?zs:List,?zs:List)))",
"rev(?zs:List) = app(app(?xs:List,app(app(?ys:List,app(?ys:List,?zs:List)),app(?zs:List,app(?zs:List,?zs:List)))),?ys:List)",
"app(?xs:List,?xs:List) = app(app(?ys:List,app(app(app(?xs:List,?ys:List),?zs:List),rev(?zs:List))),cons(?x:Nat,?zs:List))",
"app(?xs:List,?xs:List) = app(app(?zs:List,?zs:List),app(app(rev(cons(0,?zs:List)),?xs:List),cons(?y:Nat,?ys:List)))",
"app(?xs:List,?xs:List) = app(cons(times(0,len(?xs:List)),app(?xs:List,app(?xs:List,app(?zs:List,?xs:List)))),?zs:List)",
"app(?ys:List,?xs:List) = app(rev(?xs:List),app(app(?zs:List,?ys:List),app(app(?ys:List,app(?zs:List,?zs:List)),?zs:List)))",
"app(?ys:List,?ys:List) = app(app(?xs:List,rev(?ys:List)),app(app(?xs:List,app(app(?xs:List,?xs:List),?ys:List)),?zs:List))",
"app(?zs:List,?xs:List) = app(?xs:List,app(?ys:List,app(cons(s(0),app(?xs:List,?ys:List)),rev(rev(?xs:List)))))",
"app(?zs:List,?xs:List) = app(?zs:List,app(app(?ys:List,app(app(?zs:List,?xs:List),app(?zs:List,?ys:List))),rev(?xs:List)))",
"app(?zs:List,?xs:List) = app(?zs:List,rev(app(app(?ys:List,app(?ys:List,?xs:List)),app(?xs:List,app(?ys:List,?ys:List)))))",
"app(?zs:List,?ys:List) = app(rev(?zs:List),app(rev(?zs:List),app(cons(len(?zs:List),app(?zs:List,?xs:List)),?ys:List)))",
"app(?zs:List,?ys:List) = cons(plus(s(0),0),app(app(?ys:List,app(rev(?ys:List),rev(?ys:List))),?xs:List))",
"app(?zs:List,?ys:List) = cons(plus(times(0,s(s(?y:Nat))),len(cons(times(?x:Nat,?x:Nat),?xs:List))),?zs:List)",
"app(?zs:List,?zs:List) = cons(plus(times(plus(?x:Nat,len(?zs:List)),plus(?y:Nat,0)),?x:Nat),app(?xs:List,?zs:List))",
"rev(rev(?xs:List)) = rev(cons(s(times(?x:Nat,len(?zs:List))),app(rev(?xs:List),rev(app(?xs:List,?ys:List)))))",
"cons(0,?zs:List) = rev(app(app(app(?ys:List,?zs:List),app(?xs:List,app(?zs:List,?xs:List))),app(?zs:List,?xs:List)))",
"rev(app(?zs:List,?zs:List)) = app(app(app(app(?xs:List,app(?ys:List,?xs:List)),?xs:List),?xs:List),cons(?x:Nat,?xs:List))",
"cons(s(?y:Nat),?ys:List) = cons(?x:Nat,app(app(?xs:List,app(app(?xs:List,?xs:List),app(?xs:List,?zs:List))),?zs:List))",
"cons(len(?ys:List),?xs:List) = app(?xs:List,app(cons(?x:Nat,?xs:List),app(app(?xs:List,?xs:List),rev(rev(?zs:List)))))",
"cons(len(?zs:List),?xs:List) = app(app(?xs:List,?xs:List),app(?ys:List,app(app(?xs:List,app(?xs:List,?xs:List)),?ys:List)))",
"app(?xs:List,app(?xs:List,?xs:List)) = app(app(?xs:List,app(app(app(?zs:List,?zs:List),?zs:List),?ys:List)),rev(?xs:List))",
"app(?xs:List,app(?xs:List,?xs:List)) = app(cons(?x:Nat,?xs:List),app(?ys:List,app(rev(cons(?x:Nat,?ys:List)),?xs:List)))",
"app(?xs:List,cons(?y:Nat,?ys:List)) = app(cons(plus(?y:Nat,plus(?x:Nat,plus(?y:Nat,0))),?zs:List),rev(?xs:List))",
"app(?zs:List,cons(0,?xs:List)) = app(app(app(cons(?x:Nat,app(?ys:List,?zs:List)),rev(?xs:List)),?ys:List),?ys:List)",
"app(app(?xs:List,?ys:List),?xs:List) = app(cons(len(?xs:List),rev(?zs:List)),app(?xs:List,cons(s(?x:Nat),?ys:List)))",
"app(app(?ys:List,?xs:List),?ys:List) = rev(rev(app(app(app(?zs:List,?ys:List),?ys:List),rev(cons(?x:Nat,?xs:List)))))",
"app(app(?zs:List,?ys:List),?xs:List) = app(?zs:List,rev(app(app(?ys:List,?zs:List),app(cons(?y:Nat,?zs:List),?xs:List))))",
"app(cons(?x:Nat,?ys:List),?ys:List) = app(app(?zs:List,?zs:List),cons(plus(?y:Nat,len(app(?zs:List,?xs:List))),?zs:List))",
"app(cons(?x:Nat,?zs:List),?xs:List) = app(cons(times(len(?xs:List),?y:Nat),app(rev(?xs:List),?zs:List)),rev(?xs:List))",
"cons(?x:Nat,app(?xs:List,?ys:List)) = app(?xs:List,cons(plus(?x:Nat,times(s(?y:Nat),plus(0,?x:Nat))),?ys:List))",
"cons(len(rev(?zs:List)),?xs:List) = cons(len(?ys:List),app(?zs:List,app(app(?zs:List,cons(?x:Nat,?xs:List)),?ys:List)))",
"cons(times(?y:Nat,?y:Nat),?xs:List) = app(?ys:List,app(app(rev(?xs:List),app(app(?ys:List,?zs:List),?zs:List)),?zs:List))",
"app(?zs:List,app(rev(?xs:List),?ys:List)) = app(?ys:List,cons(?x:Nat,cons(plus(0,0),app(?xs:List,?xs:List))))",
"app(app(?ys:List,?zs:List),rev(?xs:List)) = cons(?x:Nat,app(app(?ys:List,cons(?x:Nat,app(?xs:List,?ys:List))),?xs:List))",
"app(?xs:List,app(?ys:List,app(?zs:List,?zs:List))) = rev(app(?zs:List,app(rev(?zs:List),app(rev(?zs:List),?ys:List))))",
"app(app(?xs:List,?xs:List),app(?ys:List,?ys:List)) = app(?xs:List,app(?zs:List,cons(?x:Nat,rev(app(?zs:List,?xs:List)))))",
"app(app(app(?zs:List,?ys:List),?xs:List),?xs:List) = app(app(?xs:List,rev(app(cons(?x:Nat,?zs:List),?zs:List))),?zs:List)",
"app(cons(0,cons(0,?xs:List)),?xs:List) = cons(len(app(cons(plus(0,0),?xs:List),?xs:List)),?xs:List)",
"cons(?x:Nat,app(?xs:List,app(?zs:List,?ys:List))) = app(?zs:List,app(?xs:List,app(cons(?x:Nat,rev(?ys:List)),?ys:List)))",
"cons(times(?x:Nat,plus(?y:Nat,?x:Nat)),?ys:List) = cons(len(app(app(?xs:List,app(?xs:List,?zs:List)),?ys:List)),?zs:List)",
"app(?xs:List,app(app(?ys:List,?xs:List),rev(?ys:List))) = app(app(?ys:List,app(app(?xs:List,?xs:List),?xs:List)),?xs:List)",
"rev(cons(len(?xs:List),app(rev(?ys:List),?xs:List))) = app(app(?xs:List,?xs:List),app(app(?zs:List,?xs:List),?xs:List))",
"cons(plus(len(?ys:List),len(rev(?ys:List))),?xs:List) = app(?xs:List,app(app(?xs:List,?zs:List),app(?xs:List,?ys:List)))",
"app(?xs:List,app(?zs:List,app(app(?xs:List,?zs:List),?xs:List))) = app(app(?xs:List,rev(app(?xs:List,?xs:List))),?xs:List)",
"app(cons(?x:Nat,app(?xs:List,app(?xs:List,?xs:List))),?ys:List) = app(?zs:List,app(?ys:List,rev(rev(rev(?zs:List)))))",
"app(cons(s(?x:Nat),app(app(?xs:List,?ys:List),?zs:List)),?zs:List) = app(app(cons(?y:Nat,?xs:List),?zs:List),?zs:List)",
"cons(len(rev(cons(len(?xs:List),app(?zs:List,?ys:List)))),?zs:List) = app(?xs:List,cons(times(?y:Nat,?x:Nat),?xs:List))",
"app(?ys:List,app(?xs:List,app(?zs:List,cons(?x:Nat,app(?ys:List,?xs:List))))) = app(app(rev(?xs:List),?zs:List),?ys:List)",
"app(app(rev(app(?xs:List,?ys:List)),app(?xs:List,?ys:List)),rev(?ys:List)) = app(rev(?zs:List),app(?xs:List,?ys:List))",
"cons(?y:Nat,app(app(cons(plus(?y:Nat,0),?zs:List),?xs:List),?ys:List)) = app(?xs:List,cons(?y:Nat,rev(?xs:List)))",
"app(?xs:List,app(cons(s(plus(?x:Nat,?x:Nat)),?zs:List),cons(0,?zs:List))) = app(?xs:List,app(?xs:List,?ys:List))",
"app(app(rev(app(?xs:List,app(?zs:List,?ys:List))),?xs:List),app(?xs:List,?ys:List)) = app(app(?zs:List,?ys:List),?ys:List)",
"app(rev(app(?xs:List,app(cons(?y:Nat,?zs:List),cons(?x:Nat,?ys:List)))),?xs:List) = app(?xs:List,app(?zs:List,?zs:List))",
"app(cons(len(?xs:List),app(?ys:List,app(?ys:List,app(?zs:List,?xs:List)))),?xs:List) = app(app(?zs:List,?xs:List),?ys:List)",
"rev(app(cons(s(len(?ys:List)),app(?xs:List,app(?xs:List,?zs:List))),?zs:List)) = app(rev(rev(?xs:List)),?zs:List)",
"cons(s(plus(plus(?x:Nat,len(?zs:List)),plus(len(?xs:List),?x:Nat))),?xs:List) = app(?zs:List,app(?xs:List,?ys:List))",
"cons(plus(plus(len(?xs:List),?x:Nat),?x:Nat),app(app(?xs:List,?zs:List),?ys:List)) = app(?xs:List,app(?xs:List,?xs:List))",
"app(?xs:List,app(app(?xs:List,app(app(?ys:List,?xs:List),?ys:List)),app(?zs:List,?zs:List))) = app(?ys:List,rev(?ys:List))",
"app(?zs:List,app(rev(rev(?xs:List)),app(app(?ys:List,cons(?x:Nat,?xs:List)),?ys:List))) = app(rev(?xs:List),?ys:List)",
"app(app(?xs:List,app(app(cons(?y:Nat,?ys:List),app(?zs:List,?zs:List)),?ys:List)),?ys:List) = rev(app(?ys:List,?ys:List))",
"app(app(?ys:List,app(?ys:List,cons(0,?zs:List))),cons(plus(0,?y:Nat),?ys:List)) = cons(len(?xs:List),?xs:List)",
"app(app(?ys:List,cons(times(len(?zs:List),len(?xs:List)),app(?ys:List,?xs:List))),?xs:List) = rev(app(?xs:List,?xs:List))",
"app(app(app(app(app(?xs:List,?ys:List),app(?xs:List,?xs:List)),?zs:List),?xs:List),?zs:List) = app(?xs:List,rev(?zs:List))",
"app(app(app(app(app(?zs:List,?ys:List),app(?zs:List,?ys:List)),?ys:List),?xs:List),?xs:List) = app(rev(?zs:List),?xs:List)",
"cons(0,rev(app(app(rev(?ys:List),?zs:List),app(app(?zs:List,?ys:List),?xs:List)))) = app(?zs:List,rev(?xs:List))",
"app(?xs:List,cons(s(plus(?x:Nat,?y:Nat)),app(app(?ys:List,app(?ys:List,?ys:List)),?xs:List))) = app(?xs:List,?xs:List)",
"app(?ys:List,app(rev(cons(?x:Nat,cons(?x:Nat,?xs:List))),app(cons(?x:Nat,?xs:List),?ys:List))) = app(?xs:List,?zs:List)",
"app(app(?xs:List,?xs:List),app(cons(?x:Nat,app(?xs:List,?xs:List)),cons(len(?ys:List),?xs:List))) = cons(0,?zs:List)",
"app(app(?xs:List,?ys:List),app(app(cons(s(0),cons(0,?ys:List)),?ys:List),?ys:List)) = cons(0,?xs:List)",
"app(app(?xs:List,app(app(?xs:List,cons(?x:Nat,?zs:List)),?xs:List)),rev(app(?zs:List,?xs:List))) = app(?xs:List,?ys:List)",
"app(app(app(rev(?xs:List),?ys:List),?zs:List),app(?xs:List,app(cons(?y:Nat,?ys:List),?xs:List))) = app(?xs:List,?xs:List)",
"app(app(cons(?y:Nat,?ys:List),rev(app(app(?ys:List,?zs:List),?xs:List))),cons(?x:Nat,?ys:List)) = cons(?x:Nat,?zs:List)",
"app(app(cons(s(0),?ys:List),cons(?y:Nat,app(app(?ys:List,?xs:List),?xs:List))),?ys:List) = app(?ys:List,?zs:List)",
"app(app(cons(times(?x:Nat,len(?xs:List)),app(app(?ys:List,?zs:List),?xs:List)),?zs:List),?zs:List) = app(?xs:List,?xs:List)",
"app(rev(?ys:List),app(app(app(app(?xs:List,?ys:List),app(?xs:List,?zs:List)),?xs:List),?xs:List)) = cons(?x:Nat,?ys:List)",
"app(rev(cons(len(cons(?x:Nat,?xs:List)),app(?ys:List,app(?zs:List,?xs:List)))),rev(?xs:List)) = app(?xs:List,?xs:List)",
"app(cons(plus(times(?x:Nat,?x:Nat),0),rev(?xs:List)),app(?zs:List,app(?xs:List,?xs:List))) = app(?ys:List,?xs:List)",
"rev(app(?ys:List,app(cons(plus(?x:Nat,?x:Nat),?ys:List),app(app(?ys:List,?zs:List),?xs:List)))) = app(?ys:List,?xs:List)",
"rev(app(app(app(?xs:List,?xs:List),app(app(?xs:List,?ys:List),app(?zs:List,?ys:List))),?zs:List)) = app(?zs:List,?xs:List)",
"rev(app(app(app(?xs:List,rev(?ys:List)),?zs:List),app(cons(len(?ys:List),?zs:List),?zs:List))) = app(?zs:List,?ys:List)",
"cons(len(app(?xs:List,app(?xs:List,app(?xs:List,?xs:List)))),app(app(?xs:List,?xs:List),?xs:List)) = app(?zs:List,?zs:List)",
"cons(len(cons(0,rev(?xs:List))),app(rev(app(?ys:List,app(?xs:List,?zs:List))),?zs:List)) = app(?zs:List,?xs:List)",
"cons(times(times(len(app(?xs:List,?ys:List)),times(s(?x:Nat),?y:Nat)),len(?zs:List)),?ys:List) = app(?zs:List,?zs:List)",
"app(rev(app(?ys:List,app(rev(?xs:List),app(?xs:List,?xs:List)))),app(app(?zs:List,?zs:List),?zs:List)) = rev(?zs:List)",
"app(cons(plus(?y:Nat,len(app(?xs:List,?ys:List))),?zs:List),app(?zs:List,cons(0,rev(?zs:List)))) = rev(?ys:List)",
"rev(?xs:List) = app(app(?zs:List,app(app(app(?xs:List,?zs:List),?xs:List),cons(0,rev(?zs:List)))),cons(?x:Nat,?xs:List))",
"app(?xs:List,?xs:List) = app(app(?ys:List,app(?xs:List,cons(times(?x:Nat,0),?xs:List))),app(app(?zs:List,?zs:List),?zs:List))",
"app(?xs:List,?xs:List) = app(app(app(?zs:List,?zs:List),?xs:List),app(app(?ys:List,cons(times(?x:Nat,0),?zs:List)),?ys:List))",
"app(?xs:List,?xs:List) = cons(s(s(plus(times(?y:Nat,?y:Nat),len(?ys:List)))),app(app(?xs:List,rev(?xs:List)),?xs:List))",
"app(?xs:List,?ys:List) = app(app(app(app(app(?ys:List,?zs:List),?ys:List),app(app(?zs:List,?zs:List),?ys:List)),?xs:List),?zs:List)",
"app(?xs:List,?ys:List) = cons(times(plus(?y:Nat,s(?x:Nat)),plus(0,0)),app(cons(s(?x:Nat),?zs:List),?zs:List))",
"app(?xs:List,?zs:List) = app(cons(s(plus(times(0,0),?x:Nat)),app(?ys:List,app(rev(?ys:List),?xs:List))),?xs:List)",
"app(?ys:List,?xs:List) = app(?ys:List,app(app(?xs:List,app(app(?zs:List,?ys:List),app(?xs:List,?xs:List))),app(?xs:List,?ys:List)))",
"app(?ys:List,?zs:List) = app(?xs:List,app(app(app(app(?zs:List,?zs:List),?ys:List),app(cons(?x:Nat,?zs:List),?xs:List)),?zs:List))",
"app(?zs:List,?xs:List) = app(app(?xs:List,?ys:List),cons(?x:Nat,app(cons(?x:Nat,?zs:List),app(?xs:List,app(?ys:List,?ys:List)))))",
"app(?zs:List,?xs:List) = app(app(?ys:List,app(?ys:List,app(?ys:List,?ys:List))),cons(times(s(len(?xs:List)),0),?xs:List))",
"rev(rev(?ys:List)) = app(app(app(cons(?x:Nat,?xs:List),app(?zs:List,app(?ys:List,?zs:List))),app(?zs:List,?xs:List)),?xs:List)",
"cons(?y:Nat,?zs:List) = app(app(app(?zs:List,app(app(?ys:List,?ys:List),app(?zs:List,?xs:List))),?zs:List),app(?xs:List,?zs:List))",
"rev(app(?zs:List,?ys:List)) = app(app(app(?zs:List,app(rev(?zs:List),?zs:List)),?ys:List),cons(?x:Nat,app(?zs:List,?xs:List)))",
"rev(cons(?x:Nat,?ys:List)) = app(?zs:List,cons(len(?xs:List),app(cons(?x:Nat,rev(?zs:List)),cons(s(?x:Nat),?xs:List))))",
"cons(len(?ys:List),?xs:List) = cons(len(?xs:List),rev(app(app(?ys:List,app(?ys:List,?xs:List)),rev(app(?zs:List,?xs:List)))))",
"cons(len(?zs:List),?zs:List) = app(app(?zs:List,app(app(?ys:List,?xs:List),cons(?x:Nat,app(?ys:List,?zs:List)))),rev(?zs:List))",
"app(?xs:List,app(?xs:List,?xs:List)) = app(app(app(app(cons(?y:Nat,?xs:List),rev(?ys:List)),?ys:List),?xs:List),rev(?xs:List))",
"app(?zs:List,app(?xs:List,?zs:List)) = cons(plus(plus(len(?ys:List),s(0)),?y:Nat),rev(rev(app(?zs:List,?ys:List))))",
"app(app(?xs:List,?xs:List),?ys:List) = app(app(?xs:List,app(?ys:List,app(app(?zs:List,?ys:List),?ys:List))),cons(0,?zs:List))",
"app(app(?ys:List,?xs:List),?xs:List) = cons(?y:Nat,app(app(app(app(?ys:List,?xs:List),cons(0,?xs:List)),?xs:List),?xs:List))",
"app(rev(?ys:List),rev(?xs:List)) = app(app(?zs:List,cons(?x:Nat,app(?xs:List,?zs:List))),app(?zs:List,cons(?y:Nat,?ys:List)))",
"app(cons(0,?ys:List),?ys:List) = app(app(app(app(app(?zs:List,?zs:List),app(?zs:List,?ys:List)),?ys:List),?xs:List),?zs:List)",
"cons(?x:Nat,app(?ys:List,?xs:List)) = app(app(app(cons(times(?x:Nat,?x:Nat),?ys:List),?zs:List),?ys:List),cons(?x:Nat,?zs:List))",
"app(app(?zs:List,?ys:List),rev(?ys:List)) = app(cons(?x:Nat,rev(cons(plus(?y:Nat,?x:Nat),app(?ys:List,?xs:List)))),?xs:List)",
"app(app(rev(?xs:List),?ys:List),?zs:List) = app(?xs:List,app(app(rev(?zs:List),?zs:List),cons(times(?x:Nat,?x:Nat),?xs:List)))",
"app(rev(?xs:List),app(?xs:List,?zs:List)) = cons(len(?xs:List),cons(plus(plus(?x:Nat,?x:Nat),times(?x:Nat,?y:Nat)),?ys:List))",
"app(rev(?xs:List),rev(rev(?xs:List))) = app(rev(app(?xs:List,app(?ys:List,?ys:List))),app(?zs:List,cons(?x:Nat,?xs:List)))",
"app(cons(len(?xs:List),?ys:List),?ys:List) = rev(app(app(?xs:List,app(app(?ys:List,?xs:List),?xs:List)),app(?ys:List,?ys:List)))",
"rev(app(app(?ys:List,?ys:List),?ys:List)) = cons(?y:Nat,app(?xs:List,app(app(?ys:List,app(?zs:List,?ys:List)),rev(?xs:List))))",
"cons(times(len(?xs:List),s(?x:Nat)),?ys:List) = app(rev(app(?ys:List,app(rev(?zs:List),cons(?x:Nat,?zs:List)))),?xs:List)",
"app(?xs:List,cons(times(0,len(?xs:List)),?xs:List)) = app(?zs:List,app(?ys:List,cons(len(?zs:List),cons(?x:Nat,?zs:List))))",
"app(?ys:List,rev(app(app(?xs:List,?xs:List),?zs:List))) = app(app(rev(?xs:List),app(app(?xs:List,?xs:List),?ys:List)),?ys:List)",
"app(app(?xs:List,app(?xs:List,rev(?zs:List))),?zs:List) = app(app(?xs:List,?zs:List),cons(s(0),app(?ys:List,?zs:List)))",
"app(rev(rev(?zs:List)),rev(app(?xs:List,?xs:List))) = app(?xs:List,app(app(?zs:List,rev(cons(?x:Nat,?xs:List))),?zs:List))",
"rev(app(cons(plus(?y:Nat,0),?ys:List),?xs:List)) = app(?ys:List,app(?xs:List,app(?ys:List,app(?ys:List,rev(?ys:List)))))",
"cons(s(times(?x:Nat,?x:Nat)),cons(?x:Nat,?xs:List)) = app(?xs:List,app(app(rev(app(?xs:List,?zs:List)),?xs:List),?ys:List))",
"app(app(?xs:List,?ys:List),cons(len(rev(?zs:List)),?ys:List)) = app(rev(app(rev(cons(?y:Nat,?ys:List)),?ys:List)),?zs:List)",
"app(app(app(?xs:List,?xs:List),app(?xs:List,?ys:List)),?xs:List) = app(app(cons(?y:Nat,?zs:List),?zs:List),app(?ys:List,?ys:List))",
"rev(app(rev(app(?ys:List,app(?xs:List,?xs:List))),?zs:List)) = app(rev(cons(len(app(?xs:List,?zs:List)),?zs:List)),?ys:List)",
"rev(cons(times(len(?ys:List),s(s(?y:Nat))),?xs:List)) = rev(app(?xs:List,app(app(?xs:List,rev(?xs:List)),?xs:List)))",
"cons(len(rev(?xs:List)),app(?xs:List,app(?zs:List,?ys:List))) = app(app(app(?ys:List,?ys:List),rev(?ys:List)),rev(?ys:List))",
"app(cons(plus(?x:Nat,0),rev(?ys:List)),app(?ys:List,?ys:List)) = app(?xs:List,app(rev(rev(rev(?xs:List))),?zs:List))",
"rev(app(?xs:List,app(?xs:List,app(?zs:List,app(?zs:List,?zs:List))))) = app(?ys:List,app(?xs:List,cons(s(?x:Nat),?ys:List)))",
"app(?xs:List,app(?ys:List,cons(?x:Nat,app(?ys:List,app(?xs:List,?xs:List))))) = cons(times(?x:Nat,plus(0,?y:Nat)),?xs:List)",
"app(app(?xs:List,app(?ys:List,app(?zs:List,app(?zs:List,?zs:List)))),?xs:List) = app(?xs:List,app(?ys:List,app(?ys:List,?zs:List)))",
"app(app(?xs:List,cons(0,?zs:List)),app(?zs:List,app(?xs:List,?zs:List))) = cons(?y:Nat,cons(?x:Nat,app(?zs:List,?xs:List)))",
"app(app(app(?xs:List,?xs:List),?xs:List),app(?ys:List,app(?xs:List,?xs:List))) = app(app(?zs:List,app(?zs:List,?xs:List)),?zs:List)",
"app(?xs:List,app(?zs:List,app(?zs:List,app(rev(?zs:List),app(?xs:List,?xs:List))))) = rev(app(app(?zs:List,?xs:List),?xs:List))",
"app(app(rev(?xs:List),cons(plus(len(?ys:List),?y:Nat),rev(?zs:List))),?ys:List) = app(app(rev(?xs:List),?ys:List),?ys:List)",
"app(cons(len(app(cons(?x:Nat,?zs:List),app(?xs:List,?ys:List))),?ys:List),?zs:List) = app(?ys:List,rev(app(?xs:List,?xs:List)))",
"app(?zs:List,app(app(?ys:List,cons(s(?x:Nat),?xs:List)),rev(app(?xs:List,?xs:List)))) = app(app(?xs:List,?xs:List),?ys:List)",
"app(app(app(?zs:List,?ys:List),cons(?y:Nat,cons(plus(?y:Nat,?x:Nat),?ys:List))),?zs:List) = app(?zs:List,app(?zs:List,?xs:List))",
"app(app(app(app(?xs:List,app(?xs:List,?ys:List)),app(?xs:List,?ys:List)),?ys:List),?xs:List) = app(cons(?x:Nat,?zs:List),?xs:List)",
"app(app(app(app(app(?xs:List,?ys:List),?ys:List),rev(?xs:List)),rev(?xs:List)),?zs:List) = cons(0,app(?ys:List,?ys:List))",
"app(app(app(rev(rev(?xs:List)),app(?xs:List,app(?zs:List,?xs:List))),?zs:List),?ys:List) = app(app(?xs:List,?xs:List),?ys:List)",
"app(app(cons(times(?x:Nat,?x:Nat),?xs:List),rev(app(?xs:List,rev(?zs:List)))),?ys:List) = cons(?y:Nat,app(?xs:List,?xs:List))",
"app(cons(?x:Nat,app(?xs:List,app(app(?ys:List,?zs:List),app(?xs:List,?zs:List)))),?xs:List) = rev(rev(app(?xs:List,?ys:List)))",
"app(cons(times(times(len(?zs:List),times(0,?x:Nat)),0),?zs:List),rev(?xs:List)) = cons(times(?y:Nat,?y:Nat),?xs:List)",
"cons(?x:Nat,cons(times(s(0),plus(len(?xs:List),times(0,?y:Nat))),?xs:List)) = cons(s(0),rev(?zs:List))",
"cons(?y:Nat,app(app(app(app(?ys:List,?xs:List),app(?xs:List,?xs:List)),?ys:List),?xs:List)) = app(?xs:List,app(?zs:List,?xs:List))",
"app(?xs:List,app(app(?ys:List,?xs:List),cons(s(?x:Nat),app(app(?xs:List,?xs:List),?xs:List)))) = rev(app(?zs:List,?xs:List))",
"app(app(?xs:List,?zs:List),cons(plus(?x:Nat,len(app(?ys:List,?zs:List))),app(?xs:List,?ys:List))) = app(?ys:List,rev(?xs:List))",
"app(app(cons(s(?y:Nat),?xs:List),?xs:List),app(?xs:List,app(?zs:List,app(?zs:List,?ys:List)))) = app(rev(?xs:List),?xs:List)",
"cons(plus(0,0),app(?ys:List,cons(len(?ys:List),app(?zs:List,app(?ys:List,?xs:List))))) = cons(?y:Nat,rev(?zs:List))",
"app(?xs:List,app(app(app(?xs:List,?xs:List),?xs:List),app(app(?zs:List,cons(0,?ys:List)),?xs:List))) = app(?ys:List,?xs:List)",
"app(?xs:List,app(app(cons(times(?x:Nat,?x:Nat),?xs:List),?xs:List),app(?xs:List,cons(?x:Nat,?xs:List)))) = cons(?x:Nat,?xs:List)",
"app(app(?xs:List,app(cons(plus(?x:Nat,0),app(?zs:List,?zs:List)),app(?ys:List,?xs:List))),?ys:List) = app(?xs:List,?xs:List)",
"app(app(?xs:List,rev(app(rev(?xs:List),rev(?xs:List)))),app(rev(app(?xs:List,?ys:List)),?ys:List)) = app(?ys:List,?ys:List)",
"app(app(app(?ys:List,?zs:List),app(?xs:List,app(app(?zs:List,?zs:List),app(?xs:List,?ys:List)))),?ys:List) = app(?zs:List,?xs:List)",
"app(app(app(app(?zs:List,?zs:List),cons(len(?xs:List),rev(?xs:List))),app(?zs:List,?ys:List)),?xs:List) = cons(0,?ys:List)",
"app(rev(?xs:List),rev(app(rev(rev(?xs:List)),app(app(?zs:List,?zs:List),cons(0,?zs:List))))) = app(?zs:List,?xs:List)",
"app(cons(plus(?x:Nat,times(times(0,?y:Nat),?y:Nat)),cons(?x:Nat,app(?zs:List,?zs:List))),?xs:List) = app(?zs:List,?xs:List)",
"rev(app(app(cons(?y:Nat,cons(?x:Nat,?xs:List)),cons(?y:Nat,app(?zs:List,?ys:List))),rev(?zs:List))) = app(?xs:List,?zs:List)",
"cons(len(app(rev(?ys:List),?zs:List)),cons(?y:Nat,app(app(?zs:List,?zs:List),app(?ys:List,?ys:List)))) = app(?zs:List,?zs:List)",
"app(?xs:List,app(cons(times(?y:Nat,s(?x:Nat)),?xs:List),app(app(?ys:List,app(?zs:List,?ys:List)),?zs:List))) = rev(?xs:List)",
"app(app(?xs:List,?zs:List),app(app(?ys:List,app(app(?zs:List,?zs:List),?xs:List)),rev(app(?ys:List,?xs:List)))) = rev(?xs:List)",
"app(app(app(?xs:List,?zs:List),cons(times(s(?x:Nat),?y:Nat),app(app(?ys:List,?xs:List),?zs:List))),?zs:List) = rev(?zs:List)",
"app(app(app(cons(len(?zs:List),?xs:List),app(app(?ys:List,?zs:List),?xs:List)),?xs:List),app(?zs:List,?xs:List)) = rev(?ys:List)",
"app(cons(?x:Nat,?zs:List),app(cons(times(s(?y:Nat),?x:Nat),app(?xs:List,?ys:List)),cons(0,?ys:List))) = rev(?zs:List)",
"app(cons(times(?y:Nat,?y:Nat),?ys:List),app(rev(?xs:List),app(?zs:List,app(?xs:List,cons(?x:Nat,?xs:List))))) = rev(?ys:List)",
"rev(app(app(cons(s(0),app(?xs:List,?zs:List)),cons(0,cons(0,?xs:List))),rev(?ys:List))) = rev(?zs:List)",
"rev(?xs:List) = app(app(app(?xs:List,?zs:List),cons(times(0,len(?ys:List)),rev(?xs:List))),app(?ys:List,cons(0,?xs:List)))",
"rev(?xs:List) = app(app(app(app(app(?ys:List,?ys:List),app(?xs:List,?xs:List)),app(cons(?y:Nat,?ys:List),?ys:List)),?ys:List),?ys:List)",
"rev(?ys:List) = cons(s(?x:Nat),app(app(app(?xs:List,cons(0,?xs:List)),?xs:List),app(rev(app(?xs:List,?xs:List)),?ys:List)))",
"rev(?zs:List) = app(?xs:List,app(app(app(?xs:List,app(?xs:List,?xs:List)),cons(plus(0,?x:Nat),?ys:List)),app(?ys:List,?ys:List)))",
"app(?xs:List,?xs:List) = app(?zs:List,app(rev(?ys:List),app(app(?zs:List,app(?zs:List,?ys:List)),app(rev(?xs:List),rev(?xs:List)))))",
"app(?xs:List,?xs:List) = app(app(rev(app(app(?zs:List,?ys:List),?xs:List)),app(cons(0,app(?xs:List,?xs:List)),?ys:List)),?ys:List)",
"app(?xs:List,?ys:List) = cons(?x:Nat,app(app(app(?ys:List,app(?zs:List,?zs:List)),app(app(?zs:List,?ys:List),?xs:List)),rev(?xs:List)))",
"app(?xs:List,?zs:List) = app(app(?zs:List,?xs:List),app(rev(app(cons(?x:Nat,?xs:List),app(?zs:List,?xs:List))),cons(?y:Nat,?ys:List)))",
"app(?ys:List,?ys:List) = app(app(app(app(app(?zs:List,?xs:List),app(?xs:List,?ys:List)),cons(len(?zs:List),?ys:List)),?xs:List),?xs:List)",
"app(?zs:List,?xs:List) = cons(times(s(times(times(?x:Nat,?x:Nat),plus(?y:Nat,?x:Nat))),?x:Nat),app(?ys:List,app(?xs:List,?zs:List)))",
"app(?zs:List,?zs:List) = app(app(app(?ys:List,rev(?ys:List)),app(app(?xs:List,app(?zs:List,?xs:List)),app(?zs:List,?xs:List))),?xs:List)",
"app(?zs:List,?zs:List) = app(app(rev(app(?xs:List,?ys:List)),app(?zs:List,rev(?xs:List))),app(?zs:List,rev(app(?zs:List,?xs:List))))",
"app(?zs:List,?zs:List) = app(cons(plus(plus(s(?x:Nat),?y:Nat),len(app(?zs:List,?xs:List))),rev(?zs:List)),app(?xs:List,?ys:List))",
"rev(rev(?ys:List)) = app(app(cons(s(?y:Nat),?xs:List),app(?xs:List,?zs:List)),app(?zs:List,app(?ys:List,app(?ys:List,?ys:List))))",
"app(rev(?ys:List),?zs:List) = app(app(?ys:List,?zs:List),app(app(cons(len(?xs:List),rev(?xs:List)),?ys:List),app(?ys:List,?xs:List)))",
"rev(cons(?x:Nat,?zs:List)) = app(?ys:List,cons(times(?x:Nat,times(?y:Nat,times(0,?x:Nat))),app(app(?xs:List,?xs:List),?xs:List)))",
"cons(s(?x:Nat),?xs:List) = app(?xs:List,cons(0,app(app(?xs:List,app(?ys:List,?ys:List)),cons(0,app(?ys:List,?xs:List)))))",
"cons(s(?x:Nat),?xs:List) = app(app(?zs:List,rev(app(?ys:List,?zs:List))),rev(app(app(?xs:List,?xs:List),app(?ys:List,?xs:List))))",
"cons(s(0),?zs:List) = app(rev(?xs:List),app(?zs:List,cons(s(plus(0,?x:Nat)),app(app(?ys:List,?xs:List),?xs:List))))",
"app(rev(?xs:List),rev(?xs:List)) = app(?xs:List,cons(times(len(rev(?xs:List)),?x:Nat),app(?ys:List,rev(app(?ys:List,?ys:List)))))",
"cons(?y:Nat,app(?zs:List,?xs:List)) = app(app(?ys:List,app(app(?ys:List,?zs:List),app(rev(?ys:List),cons(?x:Nat,?ys:List)))),?ys:List)",
"cons(times(?x:Nat,?x:Nat),?zs:List) = app(?xs:List,rev(app(app(rev(?xs:List),app(?zs:List,?xs:List)),cons(s(?y:Nat),?ys:List))))",
"app(app(?xs:List,?ys:List),rev(?ys:List)) = app(app(app(app(app(?zs:List,?xs:List),app(?ys:List,?ys:List)),?zs:List),?xs:List),?xs:List)",
"rev(app(?ys:List,app(?zs:List,?ys:List))) = app(cons(?x:Nat,app(?zs:List,cons(plus(?x:Nat,?x:Nat),?ys:List))),app(?ys:List,?ys:List))",
"app(?zs:List,app(?xs:List,app(?ys:List,?ys:List))) = app(app(?ys:List,app(app(cons(0,?xs:List),rev(?zs:List)),?xs:List)),?ys:List)",
"app(app(?xs:List,app(?ys:List,?xs:List)),?xs:List) = cons(?x:Nat,cons(plus(?y:Nat,plus(plus(?x:Nat,?x:Nat),len(?ys:List))),?zs:List))",
"app(app(?xs:List,app(?xs:List,rev(?xs:List))),?xs:List) = app(cons(?x:Nat,app(?xs:List,cons(s(?x:Nat),rev(?zs:List)))),?ys:List)",
"app(app(?zs:List,app(rev(?ys:List),?zs:List)),?xs:List) = app(?xs:List,app(app(?xs:List,cons(?x:Nat,app(?xs:List,?xs:List))),?zs:List))",
"app(app(app(?xs:List,?xs:List),?zs:List),rev(?xs:List)) = app(app(?xs:List,rev(?zs:List)),app(?xs:List,app(rev(?ys:List),?ys:List)))",
"app(rev(?ys:List),rev(app(rev(?ys:List),?zs:List))) = cons(0,app(?zs:List,app(rev(rev(?zs:List)),app(?zs:List,?ys:List))))",
"app(?xs:List,app(app(rev(rev(?ys:List)),?xs:List),?xs:List)) = app(?zs:List,app(?zs:List,app(cons(?x:Nat,rev(?xs:List)),?zs:List)))",
"app(?zs:List,app(cons(0,?zs:List),cons(?x:Nat,?ys:List))) = cons(len(cons(?y:Nat,app(app(?xs:List,?xs:List),?zs:List))),?xs:List)",
"app(app(?zs:List,?xs:List),app(app(?xs:List,?zs:List),?xs:List)) = app(app(app(?ys:List,app(rev(?xs:List),?xs:List)),?xs:List),?xs:List)",
"cons(times(?x:Nat,len(rev(?xs:List))),app(?xs:List,?xs:List)) = app(rev(app(?zs:List,app(cons(?y:Nat,?ys:List),?xs:List))),?xs:List)",
"app(?xs:List,app(rev(app(app(?zs:List,?ys:List),?xs:List)),?zs:List)) = app(app(app(?xs:List,app(?ys:List,?zs:List)),?ys:List),?xs:List)",
"app(?ys:List,cons(len(app(?xs:List,app(?ys:List,?zs:List))),?ys:List)) = app(?zs:List,app(app(?zs:List,app(?zs:List,?ys:List)),?ys:List))",
"app(?zs:List,app(app(?xs:List,rev(app(?zs:List,?ys:List))),?zs:List)) = cons(plus(times(0,times(?x:Nat,?y:Nat)),?x:Nat),?zs:List)",
"app(app(cons(?x:Nat,rev(?ys:List)),?zs:List),app(?zs:List,?zs:List)) = app(?xs:List,app(app(?zs:List,app(?xs:List,?xs:List)),?zs:List))",
"app(cons(?x:Nat,cons(?x:Nat,app(?ys:List,?ys:List))),rev(?ys:List)) = cons(times(?x:Nat,plus(?x:Nat,0)),app(?zs:List,?xs:List))",
"cons(plus(s(?x:Nat),plus(?x:Nat,s(len(?xs:List)))),?zs:List) = cons(s(?x:Nat),rev(app(cons(?x:Nat,?xs:List),?xs:List)))",
"app(rev(?ys:List),app(?zs:List,app(?xs:List,rev(cons(0,?xs:List))))) = cons(0,app(cons(len(?ys:List),?zs:List),?xs:List))",
"app(?xs:List,cons(?x:Nat,app(app(?zs:List,?xs:List),rev(app(?zs:List,?ys:List))))) = cons(times(0,?x:Nat),cons(?y:Nat,?xs:List))",
"app(?xs:List,cons(len(?ys:List),app(app(app(?xs:List,?xs:List),?xs:List),?xs:List))) = app(?ys:List,app(app(?xs:List,?ys:List),?xs:List))",
"app(app(?xs:List,app(rev(app(?ys:List,?xs:List)),?xs:List)),app(?zs:List,?zs:List)) = cons(0,app(?zs:List,app(?xs:List,?zs:List)))",
"app(app(?zs:List,app(?xs:List,app(?xs:List,rev(?zs:List)))),app(?ys:List,?zs:List)) = app(app(?ys:List,app(?xs:List,?xs:List)),?xs:List)",
"app(app(app(?zs:List,app(?xs:List,?xs:List)),rev(app(?zs:List,?xs:List))),?ys:List) = app(?zs:List,app(app(?xs:List,?xs:List),?xs:List))",
"app(cons(plus(times(s(?x:Nat),plus(?x:Nat,?y:Nat)),?y:Nat),?xs:List),?zs:List) = rev(cons(?y:Nat,app(?zs:List,rev(?xs:List))))",
"cons(s(len(rev(app(?ys:List,?ys:List)))),cons(0,app(?ys:List,?ys:List))) = rev(rev(cons(times(0,?x:Nat),?zs:List)))",
"cons(plus(times(0,plus(0,s(0))),?y:Nat),app(?zs:List,?zs:List)) = cons(len(?xs:List),app(rev(?xs:List),?xs:List))",
"app(app(?ys:List,app(app(app(?xs:List,?ys:List),?ys:List),cons(?x:Nat,?xs:List))),?xs:List) = rev(cons(?x:Nat,rev(rev(?zs:List))))",
"app(app(?zs:List,?xs:List),app(?ys:List,app(?ys:List,cons(s(?x:Nat),app(?zs:List,?ys:List))))) = cons(?x:Nat,app(?zs:List,?zs:List))",
"rev(app(?xs:List,app(cons(?x:Nat,app(?xs:List,?ys:List)),app(?ys:List,app(?xs:List,?xs:List))))) = app(?zs:List,app(?zs:List,?xs:List))",
"rev(app(app(app(rev(?ys:List),?xs:List),app(?ys:List,?zs:List)),app(?xs:List,rev(?zs:List)))) = app(?xs:List,app(?xs:List,?xs:List))",
"app(?ys:List,app(?zs:List,cons(plus(?x:Nat,?x:Nat),app(cons(0,?xs:List),app(?xs:List,?xs:List))))) = rev(app(?xs:List,?zs:List))",
"app(?zs:List,app(app(rev(app(?zs:List,?xs:List)),rev(?xs:List)),app(cons(0,?xs:List),?xs:List))) = cons(len(?zs:List),?ys:List)",
"app(app(?ys:List,?zs:List),cons(?y:Nat,cons(times(?x:Nat,len(?ys:List)),cons(s(?y:Nat),?ys:List)))) = cons(?y:Nat,rev(?zs:List))",
"app(app(?zs:List,app(?ys:List,app(?ys:List,?xs:List))),app(app(?ys:List,?zs:List),cons(?x:Nat,?ys:List))) = app(rev(?zs:List),?zs:List)",
"app(?xs:List,cons(s(0),app(app(?xs:List,?ys:List),app(app(?ys:List,?zs:List),app(?xs:List,?ys:List))))) = cons(?x:Nat,?xs:List)",
"app(app(?xs:List,app(cons(0,?xs:List),cons(?x:Nat,app(?xs:List,?ys:List)))),rev(app(?ys:List,?xs:List))) = cons(0,?ys:List)",
"app(app(?ys:List,?ys:List),app(app(app(?xs:List,?ys:List),app(?ys:List,?ys:List)),rev(app(?xs:List,?ys:List)))) = app(?ys:List,?xs:List)",
"app(app(?zs:List,app(cons(?x:Nat,app(?ys:List,?xs:List)),app(app(?ys:List,?ys:List),rev(?zs:List)))),?xs:List) = cons(0,?xs:List)",
"app(app(app(?xs:List,rev(?xs:List)),?ys:List),app(?zs:List,app(app(?xs:List,app(?ys:List,?zs:List)),?zs:List))) = cons(?x:Nat,?xs:List)",
"app(app(app(cons(0,app(?ys:List,?xs:List)),app(?xs:List,?ys:List)),?ys:List),app(?ys:List,rev(?ys:List))) = app(?ys:List,?xs:List)",
"app(app(rev(cons(?x:Nat,app(?xs:List,?ys:List))),?ys:List),app(?zs:List,cons(times(?x:Nat,0),?xs:List))) = app(?xs:List,?zs:List)",
"app(rev(app(app(?xs:List,?ys:List),cons(len(?xs:List),?zs:List))),app(app(rev(?ys:List),?xs:List),?zs:List)) = app(?xs:List,?ys:List)",
"app(cons(len(app(?xs:List,?ys:List)),app(app(?xs:List,app(?zs:List,?ys:List)),?xs:List)),cons(?y:Nat,?ys:List)) = app(?xs:List,?ys:List)",
"rev(app(app(?ys:List,cons(times(?y:Nat,?x:Nat),?xs:List)),app(?xs:List,cons(?x:Nat,app(?xs:List,?ys:List))))) = cons(?y:Nat,?ys:List)",
"app(?xs:List,app(app(?zs:List,?xs:List),cons(times(s(0),plus(0,?x:Nat)),rev(app(?ys:List,?ys:List))))) = rev(?xs:List)",
"cons(?x:Nat,app(app(?ys:List,app(app(?zs:List,?xs:List),?xs:List)),app(app(?zs:List,app(?xs:List,?zs:List)),?xs:List))) = rev(?zs:List)",
"app(?xs:List,?xs:List) = app(app(cons(times(s(?y:Nat),?x:Nat),?zs:List),?zs:List),cons(?y:Nat,rev(app(?ys:List,app(?xs:List,?xs:List)))))",
"app(?xs:List,?zs:List) = cons(?x:Nat,app(app(cons(len(?xs:List),?xs:List),rev(?xs:List)),app(?xs:List,app(app(?xs:List,?ys:List),?ys:List))))",
"app(?ys:List,?xs:List) = app(?ys:List,cons(s(plus(times(?x:Nat,?y:Nat),times(?y:Nat,?x:Nat))),app(app(?xs:List,?ys:List),rev(?zs:List))))",
"app(?ys:List,?zs:List) = app(app(app(?ys:List,app(?zs:List,?ys:List)),cons(times(len(?zs:List),s(0)),rev(?zs:List))),rev(?zs:List))",
"app(?zs:List,rev(?ys:List)) = cons(s(?x:Nat),cons(s(s(s(0))),cons(s(plus(?x:Nat,?x:Nat)),app(?zs:List,?ys:List))))",
"app(rev(?ys:List),?xs:List) = app(rev(app(app(app(?xs:List,?xs:List),cons(?x:Nat,?zs:List)),app(?ys:List,?xs:List))),app(?zs:List,?zs:List))",
"rev(app(?zs:List,?xs:List)) = cons(?y:Nat,app(app(?xs:List,app(app(?xs:List,?xs:List),?xs:List)),cons(?x:Nat,app(rev(?ys:List),?xs:List))))",
"app(app(?xs:List,?xs:List),?xs:List) = app(app(?zs:List,app(app(app(?zs:List,?zs:List),app(?zs:List,?ys:List)),?zs:List)),app(?xs:List,?xs:List))",
"app(cons(?x:Nat,?xs:List),?ys:List) = app(?xs:List,app(cons(plus(len(?xs:List),0),?xs:List),app(?xs:List,app(?ys:List,rev(?zs:List)))))",
"app(?xs:List,cons(s(?x:Nat),?ys:List)) = app(app(?xs:List,app(app(app(?zs:List,?zs:List),rev(?zs:List)),?ys:List)),cons(?y:Nat,?xs:List))",
"app(?zs:List,cons(len(?xs:List),?zs:List)) = app(?zs:List,rev(app(app(app(?ys:List,?zs:List),app(?ys:List,?zs:List)),app(?xs:List,?ys:List))))",
"app(rev(?xs:List),app(?zs:List,?zs:List)) = app(app(?xs:List,app(cons(s(?x:Nat),?ys:List),app(app(?xs:List,?zs:List),?ys:List))),?xs:List)",
"app(cons(len(?zs:List),?zs:List),?ys:List) = app(app(app(?xs:List,?xs:List),app(app(?zs:List,?zs:List),rev(?ys:List))),app(?xs:List,?zs:List))",
"app(app(?ys:List,?ys:List),app(?ys:List,?ys:List)) = app(?xs:List,app(app(rev(?xs:List),app(?xs:List,rev(?xs:List))),app(?ys:List,?ys:List)))",
"app(app(?ys:List,app(?ys:List,?ys:List)),?zs:List) = app(app(?ys:List,app(app(?ys:List,?xs:List),app(?zs:List,?xs:List))),cons(?x:Nat,?zs:List))",
"app(app(?zs:List,?zs:List),app(?xs:List,?zs:List)) = app(app(app(app(?ys:List,?xs:List),?ys:List),app(?zs:List,?ys:List)),app(?ys:List,?ys:List))",
"cons(len(rev(app(?ys:List,?ys:List))),?ys:List) = app(?zs:List,app(?xs:List,app(app(app(?ys:List,?zs:List),app(?ys:List,?ys:List)),?zs:List)))",
"cons(len(cons(s(?x:Nat),?xs:List)),?ys:List) = rev(cons(times(?y:Nat,s(s(0))),app(rev(app(?ys:List,?ys:List)),?xs:List)))",
"app(cons(0,app(?xs:List,?xs:List)),rev(?xs:List)) = app(app(rev(?zs:List),app(app(app(?ys:List,?zs:List),?xs:List),?xs:List)),?xs:List)",
"rev(app(app(app(?xs:List,?ys:List),?xs:List),?ys:List)) = app(app(?xs:List,app(rev(cons(0,?xs:List)),?xs:List)),app(?xs:List,?xs:List))",
"rev(app(rev(cons(s(?x:Nat),?xs:List)),?xs:List)) = cons(times(?x:Nat,s(plus(plus(0,?y:Nat),?x:Nat))),app(?zs:List,?xs:List))",
"app(app(app(?ys:List,?zs:List),?zs:List),app(?xs:List,?ys:List)) = cons(?x:Nat,app(app(app(?ys:List,cons(?x:Nat,?xs:List)),?zs:List),?ys:List))",
"app(app(app(?ys:List,app(?xs:List,?xs:List)),?ys:List),?ys:List) = cons(?x:Nat,app(rev(app(app(?xs:List,?xs:List),?xs:List)),rev(?xs:List)))",
"cons(len(cons(?x:Nat,rev(app(?zs:List,?ys:List)))),?ys:List) = rev(rev(cons(times(plus(?x:Nat,?x:Nat),?x:Nat),app(?ys:List,?zs:List))))",
"app(?xs:List,cons(?y:Nat,app(app(rev(?xs:List),?zs:List),?zs:List))) = app(?xs:List,app(cons(plus(s(?y:Nat),?x:Nat),?xs:List),?ys:List))",
"app(?zs:List,app(app(?xs:List,rev(cons(?x:Nat,?xs:List))),?ys:List)) = app(cons(0,cons(len(app(?xs:List,?ys:List)),?xs:List)),?ys:List)",
"app(app(cons(len(app(?zs:List,?zs:List)),?xs:List),?xs:List),?ys:List) = app(app(?ys:List,?ys:List),app(?xs:List,rev(app(?zs:List,?ys:List))))",
"rev(app(app(?ys:List,app(app(?ys:List,?xs:List),?ys:List)),?zs:List)) = app(app(rev(cons(?x:Nat,?zs:List)),?xs:List),app(?zs:List,?zs:List))",
"rev(app(app(?zs:List,app(cons(0,?zs:List),?xs:List)),?zs:List)) = app(app(app(cons(?y:Nat,?zs:List),rev(?ys:List)),?ys:List),?xs:List)",
"app(?xs:List,app(?xs:List,cons(0,app(cons(?y:Nat,?ys:List),?zs:List)))) = cons(s(len(app(?zs:List,app(?ys:List,?zs:List)))),?xs:List)",
"app(app(?xs:List,app(?xs:List,app(app(?zs:List,?ys:List),?xs:List))),?xs:List) = app(app(?xs:List,app(?zs:List,?ys:List)),app(?xs:List,?ys:List))",
"app(app(?ys:List,app(?xs:List,?xs:List)),app(rev(rev(?zs:List)),?xs:List)) = app(?zs:List,app(app(?xs:List,?ys:List),cons(?y:Nat,?xs:List)))",
"app(app(app(?ys:List,?zs:List),app(?zs:List,app(?xs:List,?zs:List))),?xs:List) = app(?zs:List,rev(rev(app(cons(0,?xs:List),?xs:List))))",
"app(app(cons(?y:Nat,?ys:List),app(?xs:List,cons(?y:Nat,?ys:List))),?zs:List) = cons(0,app(app(app(?ys:List,?xs:List),?xs:List),?zs:List))",
"app(app(cons(0,?zs:List),app(?ys:List,?xs:List)),app(?xs:List,?ys:List)) = rev(app(app(?ys:List,cons(?x:Nat,rev(?ys:List))),?xs:List))",
"app(rev(?zs:List),app(?ys:List,rev(app(cons(?y:Nat,?ys:List),?xs:List)))) = cons(0,app(?xs:List,app(?ys:List,app(?xs:List,?ys:List))))",
"app(cons(times(?x:Nat,s(?x:Nat)),rev(?xs:List)),rev(rev(?zs:List))) = app(app(?xs:List,?ys:List),cons(?x:Nat,cons(0,?zs:List)))",
"app(app(app(app(?ys:List,cons(?y:Nat,?xs:List)),rev(?xs:List)),?zs:List),?ys:List) = cons(len(app(?zs:List,?zs:List)),app(?xs:List,?xs:List))",
"app(app(app(cons(s(0),app(?xs:List,?zs:List)),?xs:List),?ys:List),?xs:List) = rev(app(rev(app(?zs:List,rev(?xs:List))),?xs:List))",
"cons(plus(plus(s(times(?x:Nat,?y:Nat)),?x:Nat),?y:Nat),app(?zs:List,?zs:List)) = cons(len(?ys:List),app(app(?ys:List,?xs:List),?ys:List))",
"app(app(?xs:List,app(app(?xs:List,?zs:List),app(?ys:List,app(?xs:List,?ys:List)))),?zs:List) = app(rev(rev(app(?ys:List,?zs:List))),?xs:List)",
"app(app(?ys:List,app(?xs:List,rev(app(?xs:List,?zs:List)))),app(?zs:List,rev(?xs:List))) = cons(?x:Nat,app(?zs:List,app(?xs:List,?xs:List)))",
"app(rev(app(?zs:List,?ys:List)),cons(s(times(len(?zs:List),len(?xs:List))),?xs:List)) = cons(times(s(?x:Nat),?y:Nat),rev(?zs:List))",
"app(?ys:List,app(app(rev(?ys:List),app(?ys:List,?zs:List)),app(cons(?x:Nat,?zs:List),?xs:List))) = rev(app(?ys:List,app(?zs:List,?xs:List)))",
"app(?xs:List,app(app(cons(len(?zs:List),?xs:List),app(cons(?x:Nat,?xs:List),rev(?xs:List))),?ys:List)) = app(?ys:List,app(?ys:List,?zs:List))",
"app(?ys:List,app(?ys:List,app(app(app(?ys:List,?ys:List),?ys:List),cons(?x:Nat,app(?ys:List,?zs:List))))) = cons(plus(?y:Nat,0),?xs:List)",
"app(?zs:List,app(app(?ys:List,rev(app(?ys:List,?zs:List))),cons(plus(s(?x:Nat),?x:Nat),?zs:List))) = app(?zs:List,app(?ys:List,?xs:List))",
"app(?zs:List,app(cons(times(0,?y:Nat),app(cons(?x:Nat,?xs:List),app(?ys:List,?ys:List))),?ys:List)) = app(app(?xs:List,?zs:List),?ys:List)",
"app(?zs:List,rev(app(app(?xs:List,app(?xs:List,?zs:List)),app(app(?ys:List,?xs:List),rev(?xs:List))))) = app(?xs:List,app(?zs:List,?ys:List))",
"app(app(rev(?zs:List),app(app(app(?zs:List,?zs:List),?zs:List),?xs:List)),app(rev(?xs:List),?zs:List)) = app(app(?zs:List,?xs:List),?xs:List)",
"app(rev(?zs:List),app(app(?xs:List,?xs:List),app(?xs:List,app(rev(?zs:List),app(?xs:List,?zs:List))))) = app(app(?xs:List,?xs:List),?ys:List)",
"cons(len(cons(s(times(?y:Nat,?y:Nat)),?ys:List)),app(?xs:List,cons(?x:Nat,cons(?x:Nat,?zs:List)))) = app(cons(?x:Nat,?ys:List),?zs:List)",
"cons(plus(s(len(?zs:List)),times(len(cons(?x:Nat,?ys:List)),s(times(?y:Nat,?x:Nat)))),?xs:List) = app(app(?ys:List,?zs:List),?xs:List)",
"app(app(?zs:List,?xs:List),app(app(app(rev(?zs:List),app(?zs:List,?ys:List)),?zs:List),app(?xs:List,?zs:List))) = rev(app(?xs:List,?xs:List))",
"app(app(app(app(app(?zs:List,?xs:List),?ys:List),app(?ys:List,?zs:List)),?zs:List),app(?zs:List,rev(?xs:List))) = rev(app(?zs:List,?xs:List))",
"app(rev(app(app(app(?ys:List,?zs:List),?zs:List),app(app(?xs:List,?ys:List),app(?zs:List,?xs:List)))),?zs:List) = app(?zs:List,rev(?ys:List))",
"app(rev(cons(times(?y:Nat,plus(?y:Nat,?y:Nat)),app(app(?zs:List,?zs:List),app(?ys:List,?zs:List)))),?xs:List) = cons(?y:Nat,rev(?ys:List))",
"app(?xs:List,rev(app(app(app(?ys:List,?zs:List),cons(?y:Nat,?xs:List)),app(app(?ys:List,?xs:List),rev(?xs:List))))) = app(?xs:List,?xs:List)",
"app(?xs:List,cons(times(plus(0,len(?zs:List)),plus(?x:Nat,?y:Nat)),cons(?y:Nat,rev(cons(0,?xs:List))))) = cons(?y:Nat,?ys:List)",
"app(?ys:List,app(app(?xs:List,rev(app(?zs:List,?xs:List))),app(app(rev(?ys:List),app(?zs:List,?ys:List)),?zs:List))) = cons(?x:Nat,?ys:List)",
"app(?ys:List,app(app(?ys:List,app(?xs:List,cons(0,?ys:List))),app(app(?zs:List,?xs:List),app(?zs:List,?xs:List)))) = rev(rev(?ys:List))",
"app(app(app(?xs:List,?ys:List),?xs:List),cons(plus(len(app(?xs:List,?ys:List)),0),cons(?x:Nat,rev(?zs:List)))) = app(?zs:List,?ys:List)",
"app(app(app(?zs:List,app(?xs:List,rev(?ys:List))),cons(0,?xs:List)),cons(times(?y:Nat,?y:Nat),rev(?zs:List))) = app(?xs:List,?xs:List)",
"app(app(app(rev(app(?zs:List,?ys:List)),?zs:List),?ys:List),rev(app(app(?ys:List,app(?xs:List,?zs:List)),?xs:List))) = app(?xs:List,?ys:List)",
"app(app(rev(?xs:List),?ys:List),app(cons(?x:Nat,?xs:List),app(rev(app(?xs:List,?ys:List)),app(?xs:List,?ys:List)))) = app(?zs:List,?xs:List)",
"app(rev(app(?zs:List,app(app(?xs:List,?zs:List),?zs:List))),app(rev(app(?ys:List,app(?zs:List,?ys:List))),?ys:List)) = app(?xs:List,?xs:List)",
"app(cons(?x:Nat,?xs:List),app(app(?zs:List,cons(times(0,?x:Nat),app(?xs:List,?zs:List))),app(?xs:List,?xs:List))) = app(?xs:List,?xs:List)",
"app(cons(times(times(times(0,?y:Nat),len(?xs:List)),len(app(?xs:List,?ys:List))),?xs:List),app(?zs:List,?ys:List)) = app(?xs:List,?ys:List)",
"cons(len(app(cons(plus(?x:Nat,0),?zs:List),rev(cons(?x:Nat,?zs:List)))),cons(s(len(?ys:List)),?xs:List)) = app(?xs:List,?xs:List)",
"cons(s(len(cons(?x:Nat,?xs:List))),app(app(app(app(?zs:List,?xs:List),?xs:List),?ys:List),app(rev(?ys:List),?zs:List))) = rev(?ys:List)",
"rev(?xs:List) = app(rev(?zs:List),app(app(cons(plus(?x:Nat,0),app(?ys:List,?xs:List)),?xs:List),rev(app(app(?ys:List,?xs:List),?ys:List))))",
"app(?xs:List,?xs:List) = app(rev(app(cons(times(?y:Nat,0),app(?xs:List,?zs:List)),app(?xs:List,app(?ys:List,?ys:List)))),app(?xs:List,?ys:List))",
"app(?xs:List,?zs:List) = app(?ys:List,cons(plus(?x:Nat,len(?xs:List)),app(app(app(?ys:List,?ys:List),cons(?x:Nat,?ys:List)),app(?ys:List,?xs:List))))",
"app(?xs:List,?zs:List) = app(app(app(?ys:List,cons(len(?xs:List),rev(?zs:List))),app(app(?ys:List,rev(?xs:List)),app(?ys:List,?ys:List))),?ys:List)",
"app(?ys:List,?xs:List) = cons(len(app(app(?xs:List,?ys:List),?zs:List)),app(app(app(rev(?zs:List),app(?xs:List,?xs:List)),?xs:List),rev(?xs:List)))",
"app(?zs:List,?zs:List) = app(app(cons(s(s(0)),app(?xs:List,app(?ys:List,?xs:List))),?xs:List),rev(cons(?y:Nat,app(?xs:List,?xs:List))))",
"app(?zs:List,?zs:List) = app(rev(app(app(app(?zs:List,?zs:List),app(?xs:List,?zs:List)),?zs:List)),rev(app(rev(cons(?y:Nat,?ys:List)),?zs:List)))",
"cons(?x:Nat,?ys:List) = app(app(?xs:List,?zs:List),app(?zs:List,cons(plus(times(?x:Nat,?x:Nat),len(?ys:List)),cons(?y:Nat,cons(?x:Nat,?xs:List)))))",
"cons(0,?zs:List) = app(app(cons(s(0),?xs:List),app(app(?zs:List,?zs:List),app(app(?xs:List,?xs:List),?xs:List))),app(?xs:List,?zs:List))",
"rev(rev(rev(?zs:List))) = app(?xs:List,app(cons(s(times(?x:Nat,0)),?xs:List),cons(plus(times(?y:Nat,?x:Nat),s(?x:Nat)),?xs:List)))",
"app(?zs:List,app(?zs:List,?ys:List)) = app(app(app(?xs:List,cons(times(?x:Nat,?x:Nat),?ys:List)),cons(?y:Nat,cons(?x:Nat,rev(?ys:List)))),?ys:List)",
"app(app(?xs:List,?xs:List),?xs:List) = app(app(app(cons(len(?xs:List),cons(?y:Nat,?ys:List)),app(?ys:List,app(?xs:List,?xs:List))),?zs:List),?ys:List)",
"app(app(?zs:List,?ys:List),?zs:List) = app(app(?xs:List,?xs:List),app(app(app(?xs:List,?xs:List),app(cons(?x:Nat,?ys:List),?zs:List)),rev(?zs:List)))",
"cons(?y:Nat,rev(rev(?zs:List))) = app(app(rev(?xs:List),app(?zs:List,?xs:List)),app(app(app(?zs:List,app(?zs:List,?ys:List)),?xs:List),?ys:List))",
"cons(times(?y:Nat,?x:Nat),?xs:List) = app(rev(?xs:List),app(app(?xs:List,?ys:List),cons(len(rev(?xs:List)),app(rev(?ys:List),rev(?xs:List)))))",
"cons(times(?y:Nat,?y:Nat),?ys:List) = app(cons(plus(?y:Nat,len(?xs:List)),app(cons(?y:Nat,?zs:List),app(?xs:List,app(?ys:List,?xs:List)))),?ys:List)",
"cons(times(?y:Nat,?y:Nat),?zs:List) = app(app(app(?ys:List,app(rev(?zs:List),?zs:List)),?zs:List),app(app(app(?zs:List,?xs:List),?xs:List),?xs:List))",
"app(app(?zs:List,rev(?ys:List)),?zs:List) = cons(times(times(?x:Nat,0),plus(0,0)),app(?zs:List,cons(plus(?y:Nat,?x:Nat),?ys:List)))",
"app(cons(?y:Nat,?ys:List),rev(?zs:List)) = app(?xs:List,app(?zs:List,app(app(rev(?ys:List),?xs:List),app(rev(?zs:List),app(?xs:List,?ys:List)))))",
"rev(cons(s(len(?xs:List)),?xs:List)) = cons(s(times(?x:Nat,?x:Nat)),app(rev(app(app(?ys:List,?xs:List),app(?xs:List,?xs:List))),?zs:List))",
"app(?xs:List,app(?ys:List,app(?ys:List,?ys:List))) = rev(app(?xs:List,app(cons(times(?y:Nat,?x:Nat),app(?zs:List,?ys:List)),cons(?y:Nat,?zs:List))))",
"app(?xs:List,cons(times(?y:Nat,?x:Nat),?xs:List)) = app(app(app(app(?xs:List,app(?zs:List,?xs:List)),?ys:List),cons(len(?ys:List),?xs:List)),?xs:List)",
"app(app(?xs:List,?xs:List),rev(rev(?ys:List))) = app(app(?zs:List,?ys:List),cons(?x:Nat,app(cons(s(0),app(?zs:List,?xs:List)),?ys:List)))",
"app(app(?xs:List,app(?zs:List,?xs:List)),?ys:List) = app(app(cons(?x:Nat,rev(?xs:List)),?xs:List),cons(?x:Nat,cons(s(len(?xs:List)),?xs:List)))",
"app(app(app(?xs:List,?zs:List),?ys:List),?ys:List) = app(cons(?x:Nat,rev(app(app(?zs:List,?xs:List),app(?zs:List,?ys:List)))),app(?xs:List,?zs:List))",
"app(cons(?x:Nat,app(?xs:List,?xs:List)),?ys:List) = app(cons(times(?y:Nat,plus(?x:Nat,len(?xs:List))),?ys:List),app(app(?zs:List,?xs:List),?ys:List))",
"cons(0,app(app(?zs:List,?ys:List),?xs:List)) = app(app(rev(app(?xs:List,cons(0,?zs:List))),app(cons(?y:Nat,?xs:List),?xs:List)),?xs:List)",
"app(?xs:List,app(?xs:List,app(?xs:List,rev(?xs:List)))) = app(app(?zs:List,app(?ys:List,?zs:List)),app(?ys:List,app(app(?zs:List,?ys:List),?xs:List)))",
"app(?zs:List,app(app(?xs:List,rev(?ys:List)),?zs:List)) = app(cons(plus(?x:Nat,0),app(?xs:List,cons(?x:Nat,?ys:List))),app(?xs:List,?zs:List))",
"app(app(?ys:List,app(?xs:List,rev(?zs:List))),?ys:List) = cons(plus(?x:Nat,?y:Nat),app(app(?xs:List,app(?ys:List,app(?zs:List,?zs:List))),?ys:List))",
"app(?xs:List,rev(cons(?y:Nat,app(rev(?xs:List),?xs:List)))) = app(?xs:List,cons(?x:Nat,cons(s(?x:Nat),app(?xs:List,app(?xs:List,?xs:List)))))",
"app(app(?ys:List,app(cons(?x:Nat,?xs:List),?zs:List)),?zs:List) = app(app(?ys:List,rev(app(app(?zs:List,?xs:List),app(?xs:List,?xs:List)))),?xs:List)",
"app(app(app(?xs:List,?xs:List),cons(?y:Nat,?zs:List)),?zs:List) = cons(plus(times(len(app(?xs:List,?zs:List)),plus(?y:Nat,0)),?y:Nat),?xs:List)",
"app(app(cons(?x:Nat,?xs:List),?zs:List),cons(?x:Nat,?xs:List)) = app(cons(0,app(cons(len(?ys:List),app(?zs:List,?zs:List)),?zs:List)),?xs:List)",
"rev(app(rev(app(?ys:List,app(?ys:List,?xs:List))),?xs:List)) = cons(s(plus(plus(?x:Nat,?x:Nat),plus(plus(?y:Nat,?x:Nat),?y:Nat))),?zs:List)",
"cons(times(plus(?y:Nat,0),?x:Nat),app(?ys:List,?zs:List)) = app(rev(app(cons(?y:Nat,?ys:List),app(app(?ys:List,?ys:List),?xs:List))),?xs:List)",
"app(?xs:List,app(cons(s(?x:Nat),app(?ys:List,?xs:List)),?zs:List)) = app(?xs:List,rev(app(?xs:List,cons(len(?xs:List),app(?zs:List,?xs:List)))))",
"app(app(?xs:List,rev(?xs:List)),app(?ys:List,app(?zs:List,?xs:List))) = app(app(app(?ys:List,?ys:List),app(?xs:List,app(?ys:List,?ys:List))),?ys:List)",
"app(rev(app(app(app(?zs:List,?ys:List),?zs:List),?ys:List)),?zs:List) = rev(app(?xs:List,app(?xs:List,app(app(?ys:List,?zs:List),rev(?xs:List)))))",
"rev(app(?ys:List,app(app(rev(?zs:List),?ys:List),rev(?xs:List)))) = app(app(cons(?x:Nat,?ys:List),app(?zs:List,cons(?x:Nat,?zs:List))),?zs:List)",
"app(app(?xs:List,app(?ys:List,app(?xs:List,app(?zs:List,?xs:List)))),?xs:List) = cons(len(app(app(?ys:List,app(?xs:List,?ys:List)),?ys:List)),?xs:List)",
"app(?ys:List,app(?ys:List,app(app(cons(?y:Nat,?xs:List),?xs:List),rev(?zs:List)))) = app(?zs:List,app(rev(rev(app(?xs:List,?ys:List))),?xs:List))",
"app(app(?xs:List,app(app(?ys:List,?xs:List),rev(app(?xs:List,?ys:List)))),?zs:List) = app(app(app(app(?xs:List,?xs:List),?ys:List),?zs:List),?ys:List)",
"app(cons(s(times(times(0,?y:Nat),plus(0,?y:Nat))),?xs:List),?ys:List) = cons(times(times(len(rev(?ys:List)),0),?y:Nat),?xs:List)",
"app(cons(plus(plus(times(?x:Nat,?x:Nat),0),s(?x:Nat)),?xs:List),?xs:List) = cons(s(plus(s(?x:Nat),s(s(0)))),?xs:List)",
"cons(times(len(rev(?xs:List)),len(?xs:List)),app(?ys:List,app(?xs:List,?ys:List))) = app(app(rev(?ys:List),?zs:List),rev(app(?xs:List,?xs:List)))",
"app(?xs:List,cons(plus(0,plus(times(0,?x:Nat),times(?x:Nat,?x:Nat))),?ys:List)) = app(app(app(?ys:List,rev(?xs:List)),?zs:List),?zs:List)",
"app(app(?zs:List,cons(0,app(cons(?x:Nat,?ys:List),app(?zs:List,?xs:List)))),?zs:List) = app(?ys:List,app(cons(s(?x:Nat),?ys:List),?ys:List))",
"app(app(app(?zs:List,?ys:List),app(app(?zs:List,?ys:List),cons(0,rev(?xs:List)))),?zs:List) = app(?ys:List,app(app(?ys:List,?xs:List),?xs:List))",
"app(rev(?zs:List),app(app(?ys:List,?xs:List),app(app(?zs:List,app(?xs:List,?xs:List)),?zs:List))) = app(app(?xs:List,app(?xs:List,?zs:List)),?xs:List)",
"app(cons(?y:Nat,rev(rev(?zs:List))),app(app(?ys:List,cons(len(?zs:List),?xs:List)),?xs:List)) = app(?ys:List,app(?zs:List,app(?zs:List,?zs:List)))",
"app(app(?ys:List,app(app(?ys:List,cons(?x:Nat,?xs:List)),app(cons(0,?ys:List),?xs:List))),?xs:List) = app(app(?zs:List,rev(?xs:List)),?zs:List)",
"rev(app(app(app(?xs:List,rev(?zs:List)),app(app(?zs:List,?zs:List),app(?ys:List,?xs:List))),?xs:List)) = rev(app(app(?ys:List,?ys:List),?ys:List))",
"cons(0,app(?ys:List,app(app(cons(?x:Nat,?xs:List),?ys:List),app(app(?ys:List,?zs:List),?zs:List)))) = app(rev(?xs:List),app(?xs:List,?xs:List))",
"cons(times(plus(s(?x:Nat),?y:Nat),?x:Nat),app(app(app(?zs:List,?xs:List),?ys:List),rev(?zs:List))) = cons(plus(?y:Nat,?x:Nat),rev(?xs:List))",
"app(app(app(app(?xs:List,app(?xs:List,?xs:List)),app(?ys:List,?ys:List)),rev(app(?ys:List,?xs:List))),?zs:List) = app(?xs:List,app(?ys:List,?xs:List))",
"rev(app(app(?xs:List,app(app(?xs:List,?xs:List),?ys:List)),cons(0,app(?zs:List,app(?xs:List,?xs:List))))) = app(app(?xs:List,?zs:List),?xs:List)",
"app(app(?xs:List,app(rev(rev(?xs:List)),app(?zs:List,rev(?xs:List)))),app(?xs:List,cons(len(?zs:List),?xs:List))) = cons(len(?zs:List),?xs:List)",
"app(app(app(?ys:List,?zs:List),app(app(app(?ys:List,?xs:List),?xs:List),?ys:List)),cons(?x:Nat,app(?xs:List,?xs:List))) = cons(?x:Nat,rev(?ys:List))",
"app(app(app(app(app(?ys:List,?ys:List),app(?xs:List,?xs:List)),?ys:List),rev(app(rev(?xs:List),?xs:List))),?xs:List) = app(rev(?xs:List),?ys:List)",
"app(app(app(app(cons(?x:Nat,?zs:List),app(?xs:List,?xs:List)),?zs:List),app(app(?xs:List,?zs:List),?xs:List)),?zs:List) = cons(len(?ys:List),?zs:List)",
"app(cons(plus(times(len(?xs:List),?y:Nat),?y:Nat),app(rev(app(?zs:List,?zs:List)),cons(0,?ys:List))),?xs:List) = rev(app(?ys:List,?ys:List))",
"rev(rev(app(app(app(?ys:List,?xs:List),app(?xs:List,?xs:List)),cons(plus(?x:Nat,?y:Nat),cons(?y:Nat,?zs:List))))) = rev(app(?xs:List,?zs:List))",
"cons(len(app(cons(times(?x:Nat,0),?zs:List),?ys:List)),app(?zs:List,rev(app(?ys:List,app(?zs:List,?xs:List))))) = cons(s(0),?ys:List)",
"app(?xs:List,app(app(cons(plus(?y:Nat,0),app(?xs:List,?zs:List)),rev(app(?zs:List,?ys:List))),app(?xs:List,?zs:List))) = app(?xs:List,?xs:List)",
"app(?ys:List,app(app(cons(s(?y:Nat),?xs:List),app(?xs:List,app(?ys:List,?zs:List))),cons(?y:Nat,cons(?x:Nat,?xs:List)))) = app(?xs:List,?zs:List)",
"app(app(app(?ys:List,?zs:List),?xs:List),cons(len(app(?xs:List,?xs:List)),cons(plus(s(?x:Nat),s(?x:Nat)),?xs:List))) = app(?ys:List,?zs:List)",
"app(app(app(cons(len(?zs:List),?xs:List),?ys:List),app(app(?zs:List,?ys:List),app(?xs:List,app(?xs:List,?zs:List)))),?zs:List) = app(?ys:List,?xs:List)",
"app(app(cons(len(?xs:List),?zs:List),app(app(?xs:List,cons(?x:Nat,?ys:List)),app(app(?zs:List,?ys:List),?xs:List))),?xs:List) = app(?xs:List,?xs:List)",
"app(rev(?xs:List),app(app(rev(?xs:List),app(app(?zs:List,?ys:List),?xs:List)),app(?zs:List,rev(cons(?x:Nat,?xs:List))))) = app(?xs:List,?xs:List)",
"app(rev(app(?xs:List,app(?xs:List,app(?zs:List,?ys:List)))),app(?xs:List,app(app(?zs:List,?ys:List),app(?xs:List,?xs:List)))) = app(?ys:List,?xs:List)",
"app(?xs:List,app(app(?xs:List,rev(?xs:List)),cons(times(?y:Nat,s(0)),app(cons(?x:Nat,?zs:List),app(?ys:List,?xs:List))))) = rev(?zs:List)",
"app(app(?ys:List,app(app(rev(?xs:List),?ys:List),app(app(?zs:List,?ys:List),app(?xs:List,?xs:List)))),app(rev(?xs:List),?xs:List)) = rev(?xs:List)",
"app(rev(app(app(app(?ys:List,?xs:List),?xs:List),?xs:List)),cons(times(0,plus(len(?zs:List),?x:Nat)),app(?xs:List,?xs:List))) = rev(?ys:List)",
"rev(?xs:List) = app(cons(times(plus(len(?xs:List),s(?x:Nat)),len(?zs:List)),app(?zs:List,app(?ys:List,app(?ys:List,?xs:List)))),app(?ys:List,?zs:List))",
"rev(?ys:List) = app(app(cons(plus(times(?x:Nat,?y:Nat),?x:Nat),app(rev(?zs:List),?xs:List)),app(?ys:List,app(?xs:List,cons(?y:Nat,?xs:List)))),?xs:List)",
"rev(?ys:List) = app(cons(times(?y:Nat,0),app(?ys:List,rev(app(?ys:List,?ys:List)))),app(?ys:List,app(app(?zs:List,app(?zs:List,?ys:List)),?ys:List)))",
"rev(?zs:List) = app(?ys:List,app(cons(len(?ys:List),app(?zs:List,?xs:List)),app(app(rev(?zs:List),app(?zs:List,?ys:List)),app(rev(?ys:List),?ys:List))))",
"rev(?zs:List) = app(cons(?x:Nat,?zs:List),app(app(?xs:List,app(?xs:List,cons(0,?zs:List))),app(app(?xs:List,?zs:List),cons(s(?x:Nat),?xs:List))))",
"app(?xs:List,?xs:List) = app(app(app(cons(?x:Nat,cons(?x:Nat,?xs:List)),app(rev(?xs:List),?ys:List)),app(app(rev(?zs:List),?xs:List),?zs:List)),?zs:List)",
"app(?xs:List,?xs:List) = app(app(cons(?x:Nat,cons(plus(?y:Nat,0),cons(?x:Nat,?zs:List))),app(app(?zs:List,app(?zs:List,?zs:List)),?xs:List)),?zs:List)",
"app(?xs:List,?ys:List) = app(app(?xs:List,app(rev(app(?ys:List,?xs:List)),app(cons(?x:Nat,?ys:List),?zs:List))),cons(times(?x:Nat,s(?x:Nat)),?xs:List))",
"app(?xs:List,?ys:List) = app(app(app(app(?xs:List,app(?zs:List,?xs:List)),cons(times(0,?y:Nat),?ys:List)),cons(0,cons(?y:Nat,?zs:List))),?xs:List)",
"app(?ys:List,?xs:List) = app(app(?xs:List,app(cons(times(?y:Nat,?x:Nat),app(?xs:List,?xs:List)),app(?xs:List,app(?ys:List,?xs:List)))),app(?zs:List,?xs:List))",
"cons(?y:Nat,?ys:List) = rev(app(cons(times(times(?x:Nat,?x:Nat),len(?ys:List)),?xs:List),app(rev(?xs:List),app(rev(?xs:List),app(?xs:List,?zs:List)))))",
"app(?xs:List,rev(?zs:List)) = app(app(app(?ys:List,app(?zs:List,rev(?xs:List))),app(?zs:List,cons(times(?x:Nat,?y:Nat),app(?xs:List,?zs:List)))),?ys:List)",
"app(rev(?zs:List),?xs:List) = app(app(cons(times(plus(?x:Nat,?x:Nat),0),app(cons(?x:Nat,?ys:List),?ys:List)),app(rev(?ys:List),?ys:List)),?xs:List)",
"app(?xs:List,cons(0,?xs:List)) = cons(plus(len(app(app(?zs:List,?zs:List),app(?zs:List,?zs:List))),times(s(?x:Nat),times(?x:Nat,?x:Nat))),?ys:List)",
"app(app(?xs:List,?xs:List),?xs:List) = app(app(cons(?x:Nat,cons(?y:Nat,?zs:List)),rev(cons(?y:Nat,app(?zs:List,?xs:List)))),cons(len(?xs:List),?xs:List))",
"app(app(?ys:List,?ys:List),?xs:List) = app(?ys:List,app(cons(?x:Nat,app(?zs:List,?xs:List)),app(?ys:List,app(cons(?x:Nat,?ys:List),app(?zs:List,?xs:List)))))",
"app(app(?ys:List,?ys:List),?zs:List) = app(app(?ys:List,?xs:List),app(app(app(?ys:List,app(?xs:List,?xs:List)),app(?xs:List,?ys:List)),app(?ys:List,?ys:List)))",
"app(app(?xs:List,?zs:List),rev(?ys:List)) = app(app(app(?zs:List,?xs:List),app(cons(times(0,?x:Nat),app(?ys:List,?xs:List)),rev(?ys:List))),?xs:List)",
"app(app(?zs:List,?ys:List),rev(?ys:List)) = app(app(cons(?y:Nat,?zs:List),app(rev(app(?ys:List,?xs:List)),?zs:List)),app(app(?zs:List,?xs:List),?xs:List))",
"app(app(app(?zs:List,?zs:List),?ys:List),?ys:List) = rev(app(app(app(app(?xs:List,?xs:List),?xs:List),rev(app(?zs:List,?xs:List))),app(?zs:List,?xs:List)))",
"rev(cons(plus(s(0),?x:Nat),?zs:List)) = cons(len(app(?xs:List,?xs:List)),cons(plus(len(app(?zs:List,?xs:List)),plus(0,?x:Nat)),?xs:List))",
"cons(s(times(0,len(?ys:List))),?xs:List) = app(app(app(?ys:List,?ys:List),?zs:List),app(app(app(app(?zs:List,?xs:List),?ys:List),?zs:List),?xs:List))",
"app(?zs:List,cons(times(?x:Nat,s(?x:Nat)),?xs:List)) = cons(times(times(0,?y:Nat),times(times(len(?zs:List),len(?xs:List)),len(?ys:List))),?zs:List)",
"cons(?y:Nat,app(rev(?xs:List),rev(rev(?ys:List)))) = cons(times(0,len(?zs:List)),cons(0,app(app(?ys:List,app(?ys:List,?zs:List)),?xs:List)))",
"app(?ys:List,cons(?x:Nat,app(app(?ys:List,?xs:List),?xs:List))) = app(rev(?ys:List),app(cons(?x:Nat,app(cons(?y:Nat,?xs:List),rev(?ys:List))),?zs:List))",
"app(app(app(?xs:List,?xs:List),app(?ys:List,?xs:List)),?zs:List) = rev(cons(times(0,?x:Nat),app(app(?xs:List,?zs:List),rev(app(?xs:List,?zs:List)))))",
"app(app(app(rev(?ys:List),?zs:List),rev(?xs:List)),?xs:List) = app(?zs:List,app(app(?xs:List,app(app(?ys:List,?ys:List),app(?xs:List,?zs:List))),?xs:List))",
"app(?ys:List,app(app(rev(?ys:List),?xs:List),app(?ys:List,?zs:List))) = app(app(?xs:List,app(?xs:List,app(?xs:List,cons(?y:Nat,?zs:List)))),rev(?ys:List))",
"app(cons(times(s(times(?x:Nat,?x:Nat)),?y:Nat),?zs:List),?zs:List) = cons(len(app(?xs:List,app(app(?zs:List,?xs:List),app(?xs:List,?xs:List)))),?xs:List)",
"cons(?x:Nat,app(?xs:List,rev(app(?zs:List,app(?ys:List,?zs:List))))) = cons(plus(times(0,plus(plus(?y:Nat,?x:Nat),?x:Nat)),len(?xs:List)),?xs:List)",
"app(app(app(app(cons(?y:Nat,?ys:List),?xs:List),?zs:List),?xs:List),?ys:List) = app(app(cons(0,?zs:List),app(app(?ys:List,?zs:List),?xs:List)),?zs:List)",
"app(cons(len(cons(?y:Nat,?zs:List)),?zs:List),cons(len(?zs:List),?zs:List)) = app(app(?zs:List,?ys:List),app(rev(cons(s(?x:Nat),?zs:List)),?xs:List))",
"rev(app(app(?ys:List,app(rev(?xs:List),app(?zs:List,?zs:List))),?xs:List)) = app(app(?xs:List,app(app(?xs:List,app(?zs:List,?ys:List)),?ys:List)),?ys:List)",
"cons(?y:Nat,app(?ys:List,cons(?y:Nat,app(app(?zs:List,?zs:List),?ys:List)))) = app(app(?zs:List,app(?xs:List,?xs:List)),app(app(?ys:List,?xs:List),?zs:List))",
"cons(0,app(app(app(app(?zs:List,?ys:List),?ys:List),?zs:List),?ys:List)) = rev(app(app(rev(app(?ys:List,?ys:List)),app(?zs:List,?zs:List)),?ys:List))",
"app(?xs:List,rev(app(?ys:List,app(app(?ys:List,?xs:List),cons(0,?xs:List))))) = cons(times(?x:Nat,times(plus(?x:Nat,?x:Nat),len(?ys:List))),?xs:List)",
"app(rev(?ys:List),app(?ys:List,app(app(app(?xs:List,?ys:List),?zs:List),?ys:List))) = app(cons(s(len(?zs:List)),cons(?x:Nat,?zs:List)),rev(?ys:List))",
"app(rev(?zs:List),app(app(app(app(?ys:List,?xs:List),?ys:List),?ys:List),?zs:List)) = cons(times(?x:Nat,?x:Nat),cons(s(0),app(?zs:List,?ys:List)))",
"app(cons(?x:Nat,rev(?ys:List)),app(?zs:List,app(app(?xs:List,?xs:List),?ys:List))) = cons(s(times(times(plus(?y:Nat,0),0),?y:Nat)),?zs:List)",
"cons(s(len(cons(len(?zs:List),?ys:List))),app(app(?xs:List,?xs:List),?xs:List)) = app(?zs:List,cons(plus(plus(?y:Nat,len(?xs:List)),?x:Nat),?zs:List))",
"app(?ys:List,app(cons(plus(len(?ys:List),s(?x:Nat)),?zs:List),app(?xs:List,?zs:List))) = app(cons(0,app(app(?zs:List,?xs:List),?xs:List)),?xs:List)",
"app(?zs:List,rev(app(rev(app(?xs:List,?xs:List)),app(app(?ys:List,?ys:List),?ys:List)))) = app(?ys:List,app(?xs:List,app(app(?zs:List,?zs:List),?ys:List)))",
"app(app(?zs:List,?xs:List),app(rev(rev(cons(?x:Nat,?xs:List))),app(?xs:List,?zs:List))) = cons(plus(?y:Nat,?x:Nat),cons(?y:Nat,app(?zs:List,?zs:List)))",
"app(cons(0,rev(app(rev(?ys:List),?ys:List))),app(?zs:List,app(?xs:List,?xs:List))) = rev(rev(app(?zs:List,app(?zs:List,app(?xs:List,?xs:List)))))",
"rev(cons(plus(len(rev(?zs:List)),s(?x:Nat)),app(?xs:List,app(?ys:List,?zs:List)))) = app(cons(plus(?x:Nat,?y:Nat),app(?zs:List,?xs:List)),?xs:List)",
"app(cons(plus(plus(?y:Nat,?x:Nat),times(len(?ys:List),plus(0,?x:Nat))),?zs:List),rev(?xs:List)) = app(app(?zs:List,app(?zs:List,?ys:List)),?xs:List)",
"app(app(?ys:List,rev(app(app(?xs:List,?xs:List),app(?xs:List,?ys:List)))),app(?zs:List,app(?xs:List,?xs:List))) = app(?ys:List,app(?ys:List,rev(?zs:List)))",
"app(app(app(app(?xs:List,app(?zs:List,?ys:List)),?ys:List),app(rev(?zs:List),cons(?x:Nat,?zs:List))),?ys:List) = app(app(rev(?ys:List),?zs:List),?xs:List)",
"app(app(app(app(?ys:List,app(?zs:List,?xs:List)),app(?ys:List,app(?xs:List,?xs:List))),?zs:List),app(?zs:List,?xs:List)) = app(app(?xs:List,?ys:List),?xs:List)",
"app(cons(?y:Nat,?ys:List),app(app(?ys:List,cons(times(?x:Nat,?x:Nat),?zs:List)),cons(?y:Nat,app(?xs:List,?xs:List)))) = cons(plus(?x:Nat,?x:Nat),?xs:List)",
"app(cons(0,?zs:List),cons(plus(plus(times(?x:Nat,?x:Nat),?x:Nat),0),rev(app(?ys:List,rev(?zs:List))))) = app(app(?ys:List,?xs:List),?xs:List)",
"rev(app(cons(times(0,len(?zs:List)),?zs:List),app(app(app(?ys:List,?xs:List),app(?xs:List,?xs:List)),?xs:List))) = cons(?y:Nat,app(?zs:List,?xs:List))",
"cons(len(app(?zs:List,rev(?zs:List))),app(?zs:List,cons(?x:Nat,app(app(?xs:List,?xs:List),app(?ys:List,?zs:List))))) = app(app(?ys:List,?ys:List),?xs:List)",
"app(app(?xs:List,cons(?x:Nat,app(app(?xs:List,?zs:List),rev(?ys:List)))),app(app(rev(rev(?xs:List)),?ys:List),?xs:List)) = app(rev(?ys:List),?zs:List)",
"app(app(app(app(app(?zs:List,?ys:List),?zs:List),app(?xs:List,?xs:List)),app(cons(s(?x:Nat),?xs:List),?zs:List)),?ys:List) = cons(len(?xs:List),?ys:List)",
"app(app(?xs:List,?xs:List),app(?ys:List,app(app(?xs:List,app(?ys:List,?xs:List)),app(cons(0,?zs:List),app(?xs:List,?ys:List))))) = app(?zs:List,?ys:List)",
"app(app(app(?xs:List,?xs:List),app(app(app(?zs:List,?zs:List),?xs:List),cons(plus(?x:Nat,?y:Nat),?zs:List))),app(?ys:List,?xs:List)) = rev(rev(?ys:List))",
"app(app(app(app(?xs:List,?zs:List),app(?xs:List,?zs:List)),?ys:List),app(?xs:List,app(?xs:List,app(app(?zs:List,?xs:List),?xs:List)))) = app(?zs:List,?xs:List)",
"app(app(app(app(app(?zs:List,?xs:List),?zs:List),?zs:List),app(?ys:List,app(app(?zs:List,?ys:List),cons(0,?ys:List)))),?xs:List) = app(?ys:List,?ys:List)",
"app(app(rev(?xs:List),?xs:List),app(?zs:List,app(cons(s(0),rev(?xs:List)),app(app(?zs:List,?xs:List),rev(?ys:List))))) = app(?zs:List,?xs:List)",
"app(app(?ys:List,?ys:List),app(cons(plus(len(?ys:List),?x:Nat),app(app(?xs:List,?ys:List),?ys:List)),app(?xs:List,app(?ys:List,?zs:List)))) = rev(?ys:List)",
"app(app(app(app(cons(?x:Nat,?xs:List),cons(0,?xs:List)),?zs:List),cons(s(0),cons(0,app(?ys:List,?ys:List)))),?zs:List) = rev(?ys:List)",
"app(app(cons(s(len(?zs:List)),?xs:List),app(?xs:List,app(app(?xs:List,?ys:List),rev(?zs:List)))),cons(?x:Nat,app(?zs:List,?zs:List))) = rev(?ys:List)",
"app(?xs:List,?xs:List) = app(app(app(?zs:List,?xs:List),app(app(?xs:List,app(?xs:List,?xs:List)),?zs:List)),app(rev(cons(plus(?y:Nat,0),?zs:List)),?ys:List))",
"app(?xs:List,?xs:List) = cons(times(len(app(?zs:List,?xs:List)),?x:Nat),app(app(?xs:List,app(app(?xs:List,?ys:List),?zs:List)),app(?ys:List,app(?xs:List,?zs:List))))",
"app(?xs:List,?ys:List) = app(app(app(?ys:List,cons(?x:Nat,app(?zs:List,?zs:List))),?xs:List),app(app(app(?zs:List,rev(?ys:List)),app(?xs:List,?xs:List)),?ys:List))",
"app(?xs:List,?zs:List) = app(app(?ys:List,app(app(?ys:List,app(?zs:List,?xs:List)),?xs:List)),app(app(app(app(?ys:List,?ys:List),?xs:List),rev(?zs:List)),?xs:List))",
"app(?ys:List,?xs:List) = app(app(cons(len(?xs:List),app(app(?xs:List,?xs:List),?xs:List)),?zs:List),app(?xs:List,app(app(app(?ys:List,?ys:List),?ys:List),?zs:List)))",
"app(?zs:List,?xs:List) = app(app(app(app(?ys:List,rev(?ys:List)),?xs:List),app(cons(plus(?y:Nat,0),?ys:List),app(?xs:List,app(?xs:List,?xs:List)))),?xs:List)",
"cons(0,?xs:List) = app(app(app(?xs:List,?xs:List),?ys:List),app(rev(?xs:List),app(app(?xs:List,cons(?x:Nat,?xs:List)),app(app(?xs:List,?zs:List),?xs:List))))",
"app(?zs:List,rev(?xs:List)) = app(app(cons(0,app(app(?zs:List,?ys:List),?ys:List)),rev(rev(cons(?y:Nat,?xs:List)))),app(cons(?y:Nat,?xs:List),?xs:List))",
"app(rev(?zs:List),?ys:List) = app(app(app(?ys:List,?ys:List),?zs:List),app(?zs:List,app(cons(?x:Nat,cons(?y:Nat,?ys:List)),app(app(?ys:List,?zs:List),?xs:List))))",
"app(?xs:List,app(?zs:List,?ys:List)) = app(app(app(?zs:List,?xs:List),?ys:List),cons(s(?x:Nat),cons(s(?x:Nat),cons(plus(0,?x:Nat),rev(?xs:List)))))",
"cons(0,app(?ys:List,?zs:List)) = app(app(cons(plus(s(?x:Nat),times(?x:Nat,?y:Nat)),app(app(?zs:List,?zs:List),app(?zs:List,?xs:List))),?xs:List),?ys:List)",
"rev(app(app(?xs:List,?xs:List),?zs:List)) = app(app(cons(len(app(?zs:List,?xs:List)),?zs:List),cons(plus(?y:Nat,plus(0,0)),?xs:List)),rev(?ys:List))",
"cons(s(s(len(?xs:List))),?xs:List) = app(cons(s(times(0,times(?y:Nat,?y:Nat))),app(app(?zs:List,?xs:List),app(rev(?xs:List),?xs:List))),?ys:List)",
"app(?ys:List,app(?ys:List,app(?zs:List,?xs:List))) = app(?xs:List,app(app(app(?zs:List,app(?xs:List,?xs:List)),?xs:List),rev(app(?zs:List,app(?xs:List,?zs:List)))))",
"app(?ys:List,cons(len(rev(?ys:List)),?ys:List)) = app(cons(0,?ys:List),app(app(rev(app(?zs:List,?ys:List)),cons(0,?zs:List)),app(?xs:List,?ys:List)))",
"app(?xs:List,app(?xs:List,cons(?x:Nat,app(?xs:List,?xs:List)))) = cons(len(app(?ys:List,?zs:List)),app(app(app(?xs:List,cons(?y:Nat,?xs:List)),?zs:List),?xs:List))",
"app(app(app(cons(?x:Nat,?ys:List),?xs:List),?zs:List),rev(?ys:List)) = cons(?x:Nat,app(?xs:List,app(app(?xs:List,app(?ys:List,?xs:List)),app(?ys:List,?zs:List))))",
"cons(?x:Nat,app(app(rev(cons(?x:Nat,?xs:List)),?zs:List),?ys:List)) = app(?xs:List,app(?xs:List,app(app(app(?zs:List,?zs:List),app(?zs:List,?xs:List)),?zs:List)))",
"cons(s(times(times(times(?y:Nat,?y:Nat),?x:Nat),?x:Nat)),?xs:List) = app(?xs:List,app(cons(len(?xs:List),?ys:List),app(?zs:List,rev(app(?ys:List,?xs:List)))))",
"app(?xs:List,app(app(?zs:List,?ys:List),app(app(?xs:List,?ys:List),?ys:List))) = rev(app(cons(?y:Nat,cons(plus(?y:Nat,?x:Nat),app(?xs:List,?xs:List))),?ys:List))",
"cons(times(times(times(times(?x:Nat,?x:Nat),?y:Nat),0),?x:Nat),?xs:List) = rev(cons(plus(?x:Nat,s(0)),app(?ys:List,app(rev(?ys:List),?ys:List))))",
"app(app(app(app(?ys:List,?xs:List),rev(app(?xs:List,?ys:List))),?zs:List),?xs:List) = app(rev(?xs:List),app(rev(app(app(?xs:List,?ys:List),?zs:List)),?zs:List))",
"app(app(app(rev(?ys:List),?xs:List),?xs:List),app(app(?xs:List,?xs:List),?xs:List)) = cons(s(plus(times(plus(?x:Nat,0),?y:Nat),s(?x:Nat))),?zs:List)",
"app(app(rev(app(app(?xs:List,?xs:List),?xs:List)),?xs:List),app(?zs:List,?zs:List)) = app(app(?zs:List,?xs:List),app(cons(?x:Nat,app(?xs:List,?zs:List)),?zs:List))",
"app(app(?xs:List,app(cons(times(?x:Nat,?x:Nat),app(?ys:List,?zs:List)),?xs:List)),?xs:List) = app(rev(app(?ys:List,?xs:List)),app(app(?xs:List,?xs:List),?xs:List))",
"app(rev(?xs:List),app(cons(s(plus(0,?x:Nat)),app(?xs:List,?zs:List)),?ys:List)) = app(cons(plus(plus(len(?ys:List),?x:Nat),0),?zs:List),?ys:List)",
"app(cons(?x:Nat,app(app(app(?ys:List,?zs:List),?ys:List),?xs:List)),cons(?y:Nat,?xs:List)) = app(?ys:List,app(?ys:List,app(cons(len(?xs:List),?xs:List),?xs:List)))",
"app(cons(times(?x:Nat,times(0,0)),cons(?y:Nat,app(?zs:List,?zs:List))),?zs:List) = app(app(?xs:List,app(app(?xs:List,?zs:List),?xs:List)),rev(?zs:List))",
"cons(0,app(?xs:List,app(app(app(?zs:List,?ys:List),?xs:List),app(rev(?xs:List),?xs:List)))) = app(?zs:List,app(?xs:List,cons(plus(0,?y:Nat),?ys:List)))",
"app(app(?xs:List,app(rev(?ys:List),app(app(?ys:List,?xs:List),?xs:List))),cons(len(?zs:List),?ys:List)) = cons(s(plus(0,?y:Nat)),app(?xs:List,?zs:List))",
"app(app(app(rev(?zs:List),rev(app(?xs:List,?xs:List))),cons(len(?xs:List),rev(?zs:List))),?zs:List) = cons(times(?x:Nat,plus(len(?xs:List),?y:Nat)),?xs:List)",
"app(?xs:List,app(cons(len(rev(?xs:List)),app(?ys:List,?ys:List)),app(?zs:List,app(?ys:List,app(?zs:List,?zs:List))))) = app(app(?ys:List,?xs:List),rev(?zs:List))",
"app(app(app(app(cons(?x:Nat,?ys:List),?zs:List),app(cons(?y:Nat,?zs:List),?ys:List)),?zs:List),app(?xs:List,?zs:List)) = app(rev(?xs:List),cons(?y:Nat,?ys:List))",
"app(app(cons(?y:Nat,?ys:List),app(app(?xs:List,cons(0,?zs:List)),cons(?x:Nat,rev(?xs:List)))),rev(?zs:List)) = app(app(?xs:List,?zs:List),rev(?xs:List))",
"app(?xs:List,app(app(app(?ys:List,?ys:List),app(rev(?xs:List),app(?ys:List,?xs:List))),app(app(?xs:List,?ys:List),?xs:List))) = app(rev(rev(?zs:List)),?xs:List)",
"app(app(?xs:List,app(?xs:List,rev(app(?xs:List,?ys:List)))),app(?xs:List,app(?ys:List,app(app(?ys:List,?zs:List),?xs:List)))) = app(app(?xs:List,?ys:List),?ys:List)",
"app(app(?zs:List,app(app(?ys:List,?xs:List),?zs:List)),app(?xs:List,app(app(rev(?ys:List),?zs:List),rev(rev(?zs:List))))) = app(?xs:List,app(?zs:List,?xs:List))",
"app(app(cons(len(?ys:List),app(app(?xs:List,?xs:List),?xs:List)),?ys:List),app(?xs:List,app(?xs:List,app(?xs:List,?ys:List)))) = cons(times(?x:Nat,?y:Nat),?ys:List)",
"app(app(app(rev(app(?xs:List,?xs:List)),?ys:List),app(?xs:List,app(rev(?xs:List),app(?zs:List,?ys:List)))),app(?xs:List,?ys:List)) = app(rev(?xs:List),?xs:List)",
"app(app(?ys:List,app(rev(app(?zs:List,?zs:List)),?zs:List)),app(app(?xs:List,app(app(?xs:List,?zs:List),app(?zs:List,?zs:List))),?ys:List)) = cons(?y:Nat,?xs:List)",
"app(app(app(app(?xs:List,app(?xs:List,?xs:List)),?xs:List),app(cons(s(0),app(?zs:List,?xs:List)),app(?xs:List,?xs:List))),?xs:List) = app(?ys:List,?xs:List)",
"app(app(cons(plus(?x:Nat,s(?y:Nat)),?ys:List),cons(0,?zs:List)),app(app(?zs:List,app(?ys:List,?zs:List)),rev(rev(?xs:List)))) = app(?xs:List,?xs:List)",
"rev(app(app(cons(len(?ys:List),app(?xs:List,?ys:List)),?ys:List),app(app(app(?ys:List,?zs:List),cons(?x:Nat,?ys:List)),rev(?xs:List)))) = cons(0,?xs:List)",
"rev(?zs:List) = cons(len(app(cons(?x:Nat,?ys:List),?xs:List)),cons(times(times(len(?xs:List),?x:Nat),plus(times(?x:Nat,?y:Nat),s(?x:Nat))),cons(?x:Nat,?zs:List)))",
"app(?ys:List,?xs:List) = app(cons(times(0,s(len(?zs:List))),?zs:List),cons(times(len(app(?zs:List,?ys:List)),times(len(?zs:List),times(?x:Nat,?x:Nat))),?zs:List))",
"app(?ys:List,?ys:List) = app(cons(?x:Nat,cons(times(?x:Nat,?x:Nat),app(?xs:List,?zs:List))),cons(times(s(?x:Nat),len(?xs:List)),rev(app(rev(?xs:List),?xs:List))))",
"app(?zs:List,?xs:List) = app(app(app(?xs:List,?zs:List),app(?xs:List,?xs:List)),rev(app(app(rev(?zs:List),cons(?x:Nat,?ys:List)),app(?xs:List,app(?zs:List,?xs:List)))))",
"app(?zs:List,?zs:List) = app(?zs:List,app(app(app(?zs:List,cons(?x:Nat,?ys:List)),app(?ys:List,?zs:List)),app(?xs:List,app(app(?ys:List,?xs:List),app(?zs:List,?zs:List)))))",
"app(app(?ys:List,?xs:List),?ys:List) = app(app(rev(app(cons(?y:Nat,?zs:List),?zs:List)),app(?xs:List,cons(plus(?x:Nat,0),rev(?xs:List)))),app(?zs:List,?zs:List))",
"app(cons(?y:Nat,?zs:List),?xs:List) = app(app(app(app(?xs:List,?xs:List),cons(?x:Nat,app(?zs:List,?xs:List))),?ys:List),rev(app(cons(len(?zs:List),?ys:List),?xs:List)))",
"rev(app(?xs:List,rev(?zs:List))) = app(rev(?xs:List),app(app(app(?zs:List,rev(?zs:List)),?xs:List),app(?zs:List,app(app(?ys:List,?zs:List),app(?xs:List,?xs:List)))))",
"cons(plus(?x:Nat,?y:Nat),?zs:List) = cons(len(app(rev(?xs:List),app(rev(?zs:List),cons(?y:Nat,?ys:List)))),cons(len(app(?zs:List,app(?xs:List,?xs:List))),?xs:List))",
"app(app(?zs:List,?ys:List),cons(?x:Nat,?zs:List)) = app(?xs:List,app(?xs:List,app(app(?ys:List,app(?ys:List,?xs:List)),app(cons(0,?zs:List),app(?zs:List,?ys:List)))))",
"app(app(app(?xs:List,?xs:List),?xs:List),?ys:List) = app(?xs:List,app(app(app(?zs:List,app(?xs:List,?ys:List)),cons(times(0,?y:Nat),?zs:List)),app(?zs:List,?ys:List)))",
"app(?xs:List,app(app(?xs:List,rev(?xs:List)),?zs:List)) = app(cons(len(app(?zs:List,app(?ys:List,?ys:List))),app(app(?xs:List,app(?ys:List,?zs:List)),?xs:List)),?xs:List)",
"app(app(rev(?zs:List),app(?zs:List,?xs:List)),rev(?xs:List)) = app(app(?ys:List,app(app(app(?zs:List,?ys:List),app(?xs:List,?xs:List)),?zs:List)),app(?zs:List,?xs:List))",
"app(?ys:List,app(rev(?xs:List),app(app(?xs:List,?ys:List),?zs:List))) = app(app(?zs:List,rev(app(app(?zs:List,?xs:List),?xs:List))),app(app(?ys:List,?xs:List),?ys:List))",
"app(app(?ys:List,app(?xs:List,cons(s(?x:Nat),?ys:List))),?zs:List) = app(app(app(?zs:List,app(?xs:List,?xs:List)),?xs:List),app(app(?ys:List,?ys:List),rev(?zs:List)))",
"app(rev(app(app(?xs:List,?xs:List),?xs:List)),app(?ys:List,?ys:List)) = app(app(?zs:List,?zs:List),app(?xs:List,app(?xs:List,app(cons(?x:Nat,?ys:List),rev(?ys:List)))))",
"app(?ys:List,app(app(app(?xs:List,app(?xs:List,?zs:List)),?ys:List),?zs:List)) = cons(0,cons(?x:Nat,app(app(?zs:List,app(?zs:List,?zs:List)),app(?xs:List,?xs:List))))",
"app(?ys:List,rev(app(?xs:List,cons(len(?xs:List),cons(?x:Nat,?xs:List))))) = app(app(?xs:List,app(app(cons(?y:Nat,?ys:List),cons(0,?xs:List)),?xs:List)),?zs:List)",
"app(app(?ys:List,rev(app(?zs:List,app(?xs:List,?zs:List)))),rev(?xs:List)) = app(app(rev(?xs:List),?xs:List),app(app(?zs:List,rev(app(?zs:List,?xs:List))),?xs:List))",
"app(app(cons(?x:Nat,app(?zs:List,app(?ys:List,?zs:List))),?ys:List),?xs:List) = app(?zs:List,app(app(?xs:List,app(app(?xs:List,?xs:List),?xs:List)),app(?ys:List,?xs:List)))",
"cons(times(s(s(len(?xs:List))),plus(s(0),?x:Nat)),?zs:List) = app(?xs:List,app(app(app(app(?zs:List,?zs:List),cons(?x:Nat,?xs:List)),?zs:List),?xs:List))",
"app(?zs:List,app(app(?zs:List,?ys:List),app(?zs:List,app(?zs:List,app(?xs:List,?zs:List))))) = app(cons(?x:Nat,app(?xs:List,app(app(?xs:List,?zs:List),?xs:List))),?ys:List)",
"app(app(?ys:List,app(?ys:List,app(cons(?y:Nat,?xs:List),app(?xs:List,?ys:List)))),?ys:List) = app(rev(app(?zs:List,rev(?ys:List))),app(?xs:List,app(?zs:List,?ys:List)))",
"app(app(app(?zs:List,?xs:List),cons(?x:Nat,app(?ys:List,app(?ys:List,?xs:List)))),?xs:List) = app(app(app(app(?xs:List,app(?xs:List,?xs:List)),?xs:List),?xs:List),?xs:List)",
"app(app(rev(app(rev(?ys:List),app(?xs:List,?zs:List))),?zs:List),app(?xs:List,?ys:List)) = app(app(?xs:List,app(?xs:List,app(?ys:List,app(?xs:List,?xs:List)))),?ys:List)",
"cons(times(plus(len(?ys:List),?y:Nat),?x:Nat),app(?xs:List,app(?ys:List,rev(?zs:List)))) = app(cons(times(?x:Nat,?x:Nat),app(?zs:List,?xs:List)),app(?zs:List,?zs:List))",
"app(?ys:List,app(?ys:List,app(app(?xs:List,cons(?y:Nat,?xs:List)),rev(app(?xs:List,?xs:List))))) = app(?xs:List,app(?xs:List,rev(app(?xs:List,app(?xs:List,?ys:List)))))",
"app(app(?zs:List,app(app(rev(?ys:List),app(?zs:List,?xs:List)),?xs:List)),app(?zs:List,?xs:List)) = app(?xs:List,cons(len(app(app(?zs:List,?ys:List),?xs:List)),?ys:List))",
"cons(times(len(rev(?xs:List)),plus(s(0),s(len(?zs:List)))),app(?xs:List,?ys:List)) = rev(app(?xs:List,app(app(app(?xs:List,?xs:List),?zs:List),?xs:List)))",
"app(?ys:List,cons(plus(len(rev(?xs:List)),plus(0,times(0,?x:Nat))),app(?xs:List,?ys:List))) = rev(app(app(?xs:List,app(?xs:List,rev(?xs:List))),?ys:List))",
"cons(times(?y:Nat,len(rev(rev(?xs:List)))),app(cons(?x:Nat,app(app(?ys:List,?xs:List),?ys:List)),?xs:List)) = rev(app(app(?xs:List,cons(?x:Nat,?ys:List)),?ys:List))",
"app(app(?xs:List,rev(app(?ys:List,app(?zs:List,?ys:List)))),app(app(?ys:List,app(?ys:List,app(?ys:List,?ys:List))),?ys:List)) = app(app(rev(?zs:List),?xs:List),?xs:List)",
"cons(plus(times(times(len(?xs:List),0),times(plus(?x:Nat,?x:Nat),?x:Nat)),plus(s(len(?xs:List)),0)),?ys:List) = cons(?x:Nat,app(rev(?ys:List),?xs:List))",
"app(app(cons(?x:Nat,app(app(?ys:List,?ys:List),app(?zs:List,?ys:List))),app(app(app(?xs:List,?ys:List),?zs:List),?zs:List)),?ys:List) = rev(rev(app(?ys:List,?xs:List)))",
"app(app(app(app(app(?xs:List,?zs:List),rev(?ys:List)),app(app(?xs:List,?xs:List),app(?zs:List,?zs:List))),app(?zs:List,?ys:List)),?ys:List) = cons(s(0),?ys:List)",
"app(cons(?y:Nat,app(?xs:List,app(cons(0,?zs:List),app(?xs:List,?zs:List)))),app(?zs:List,app(rev(app(?zs:List,?ys:List)),?zs:List))) = cons(?x:Nat,rev(?xs:List))",
"app(app(?xs:List,?zs:List),app(app(rev(app(?xs:List,?zs:List)),app(?zs:List,?ys:List)),app(rev(?ys:List),app(?zs:List,app(?ys:List,?ys:List))))) = app(?xs:List,?ys:List)",
"app(app(?xs:List,app(cons(?x:Nat,?xs:List),?zs:List)),cons(plus(plus(plus(?x:Nat,?y:Nat),times(0,?x:Nat)),0),app(?ys:List,?ys:List))) = app(?xs:List,?xs:List)",
"app(app(rev(app(?zs:List,?xs:List)),rev(cons(plus(?y:Nat,?x:Nat),rev(?xs:List)))),app(cons(len(?xs:List),?ys:List),app(?ys:List,?xs:List))) = app(?xs:List,?xs:List)",
"app(app(rev(app(?zs:List,app(?ys:List,?ys:List))),app(?zs:List,?xs:List)),app(?zs:List,app(?xs:List,app(rev(?zs:List),app(?xs:List,?ys:List))))) = app(?zs:List,?zs:List)",
"app(app(cons(?x:Nat,?zs:List),app(?ys:List,app(app(?ys:List,?zs:List),app(?zs:List,?xs:List)))),cons(s(?x:Nat),cons(s(?x:Nat),?xs:List))) = app(?xs:List,?xs:List)",
"app(rev(?zs:List),cons(times(times(len(?ys:List),?x:Nat),s(times(?x:Nat,?x:Nat))),app(app(?zs:List,?ys:List),cons(?x:Nat,rev(?xs:List))))) = app(?xs:List,?ys:List)",
"app(rev(app(rev(cons(?x:Nat,?xs:List)),?zs:List)),app(?zs:List,app(cons(0,rev(?zs:List)),app(rev(?xs:List),app(?zs:List,?ys:List))))) = app(?zs:List,?zs:List)",
"cons(plus(plus(0,?x:Nat),times(plus(0,?x:Nat),?y:Nat)),app(app(?xs:List,?ys:List),cons(times(plus(?x:Nat,0),?x:Nat),?ys:List))) = app(?ys:List,?ys:List)",
"cons(times(0,plus(s(?x:Nat),plus(plus(0,?y:Nat),len(?ys:List)))),rev(app(app(app(?zs:List,?zs:List),rev(?xs:List)),?zs:List))) = app(?ys:List,?ys:List)",
"rev(?xs:List) = app(app(?ys:List,?zs:List),app(app(app(?ys:List,rev(?xs:List)),rev(app(?zs:List,?xs:List))),app(app(app(?xs:List,?zs:List),cons(?x:Nat,?zs:List)),?xs:List)))",
"rev(?ys:List) = app(cons(times(times(?x:Nat,plus(0,0)),?y:Nat),app(?xs:List,cons(plus(?y:Nat,?x:Nat),app(?xs:List,?ys:List)))),app(?xs:List,app(?xs:List,?zs:List)))",
"rev(?zs:List) = app(app(app(?zs:List,rev(app(?xs:List,?zs:List))),?ys:List),app(app(app(?ys:List,rev(?zs:List)),?ys:List),app(?xs:List,app(?xs:List,app(?zs:List,?xs:List)))))",
"app(?xs:List,rev(?zs:List)) = app(cons(plus(len(?zs:List),?x:Nat),app(rev(rev(?ys:List)),app(?xs:List,rev(?ys:List)))),app(app(?xs:List,cons(0,?xs:List)),?xs:List))",
"rev(app(?ys:List,?zs:List)) = app(app(?xs:List,?xs:List),app(app(app(app(?xs:List,?xs:List),app(?zs:List,?xs:List)),app(app(?zs:List,?ys:List),app(?zs:List,?xs:List))),?xs:List))",
"rev(app(?zs:List,?zs:List)) = app(cons(s(plus(?y:Nat,len(?ys:List))),?xs:List),app(?xs:List,app(app(app(?zs:List,?xs:List),?ys:List),app(app(?xs:List,?xs:List),?zs:List))))",
"app(?zs:List,cons(s(?x:Nat),?zs:List)) = app(cons(plus(?y:Nat,?x:Nat),?xs:List),app(cons(times(times(?x:Nat,?x:Nat),s(?x:Nat)),?xs:List),cons(s(?x:Nat),?ys:List)))",
"app(rev(rev(?xs:List)),app(?zs:List,?ys:List)) = app(cons(?y:Nat,app(?xs:List,?zs:List)),app(app(?xs:List,?zs:List),app(app(?xs:List,?ys:List),rev(app(?zs:List,?ys:List)))))",
"app(app(?ys:List,rev(app(?ys:List,?xs:List))),?xs:List) = app(app(app(app(app(?ys:List,?ys:List),?zs:List),app(?zs:List,?ys:List)),?xs:List),app(app(?ys:List,?ys:List),?xs:List))",
"app(?xs:List,rev(app(?ys:List,app(app(?xs:List,?xs:List),?xs:List)))) = cons(?x:Nat,app(app(?ys:List,?ys:List),app(?xs:List,app(app(?ys:List,?xs:List),app(?xs:List,?ys:List)))))",
"app(?ys:List,app(?ys:List,cons(?x:Nat,cons(len(?ys:List),?ys:List)))) = cons(plus(len(?ys:List),len(?xs:List)),app(?xs:List,cons(?x:Nat,app(app(?ys:List,?xs:List),?xs:List))))",
"app(?zs:List,app(rev(app(app(?ys:List,?xs:List),?ys:List)),?zs:List)) = app(app(?xs:List,app(app(app(?ys:List,?ys:List),app(?xs:List,?xs:List)),app(?ys:List,?ys:List))),?zs:List)",
"app(app(?ys:List,cons(0,rev(cons(?x:Nat,?xs:List)))),?ys:List) = app(app(app(app(?xs:List,cons(?x:Nat,?ys:List)),app(?zs:List,cons(0,?ys:List))),?xs:List),?zs:List)",
"app(app(?xs:List,app(?xs:List,app(cons(?y:Nat,?ys:List),?zs:List))),?ys:List) = app(app(app(?xs:List,cons(s(0),?zs:List)),app(?xs:List,app(?xs:List,?xs:List))),?xs:List)",
"app(app(?ys:List,?ys:List),app(app(?ys:List,?ys:List),app(?ys:List,?xs:List))) = app(app(app(?ys:List,?xs:List),app(?ys:List,app(app(?xs:List,?ys:List),rev(?xs:List)))),?zs:List)",
"app(rev(app(?ys:List,?ys:List)),app(?xs:List,rev(app(?ys:List,?zs:List)))) = app(app(?ys:List,app(?xs:List,?xs:List)),cons(plus(?y:Nat,times(0,?y:Nat)),rev(?zs:List)))",
"cons(plus(?x:Nat,0),app(app(?xs:List,?ys:List),app(?ys:List,?zs:List))) = app(?xs:List,app(app(app(cons(0,?zs:List),app(?zs:List,?xs:List)),?zs:List),rev(?ys:List)))",
"app(?xs:List,app(?zs:List,app(app(?zs:List,?ys:List),rev(app(?ys:List,?ys:List))))) = cons(plus(len(cons(?x:Nat,app(?xs:List,?zs:List))),0),rev(app(?ys:List,?zs:List)))",
"app(?ys:List,app(app(app(?xs:List,cons(?x:Nat,?zs:List)),?ys:List),rev(?xs:List))) = app(?xs:List,app(?zs:List,app(app(app(?ys:List,?xs:List),app(?ys:List,?xs:List)),?xs:List)))",
"app(app(?xs:List,?ys:List),cons(?x:Nat,app(cons(len(?xs:List),?zs:List),?xs:List))) = app(cons(times(?x:Nat,?y:Nat),app(app(app(?xs:List,?xs:List),?ys:List),?xs:List)),?zs:List)",
"app(app(rev(cons(times(0,?y:Nat),app(?xs:List,?xs:List))),?zs:List),?zs:List) = app(app(?zs:List,app(?zs:List,cons(0,?xs:List))),cons(?y:Nat,app(?ys:List,?zs:List)))",
"app(rev(app(app(app(?zs:List,?ys:List),?zs:List),?xs:List)),app(?xs:List,?xs:List)) = app(app(?ys:List,app(app(app(?ys:List,?xs:List),?xs:List),app(?xs:List,?ys:List))),?ys:List)",
"cons(times(?x:Nat,plus(?y:Nat,plus(len(?ys:List),plus(?y:Nat,?x:Nat)))),?xs:List) = cons(len(?zs:List),app(app(?xs:List,?zs:List),app(app(?xs:List,?xs:List),rev(?ys:List))))",
"app(app(app(?xs:List,?xs:List),app(?ys:List,app(app(?xs:List,?xs:List),?xs:List))),?xs:List) = app(app(app(?zs:List,?zs:List),?zs:List),app(app(?xs:List,rev(?xs:List)),?ys:List))",
"app(app(app(app(cons(?y:Nat,?ys:List),?xs:List),app(?ys:List,?zs:List)),?xs:List),?ys:List) = app(rev(?zs:List),app(?xs:List,app(app(app(?xs:List,?zs:List),?zs:List),?xs:List)))",
"app(app(rev(?xs:List),app(cons(s(?x:Nat),?ys:List),app(?xs:List,?ys:List))),?ys:List) = app(rev(app(?zs:List,app(app(?zs:List,?zs:List),?xs:List))),app(?xs:List,?zs:List))",
"cons(s(times(s(plus(?x:Nat,?x:Nat)),times(?y:Nat,plus(?x:Nat,?x:Nat)))),?xs:List) = rev(cons(?y:Nat,app(?ys:List,app(app(?zs:List,?zs:List),app(?zs:List,?zs:List)))))",
"app(app(app(app(app(?ys:List,?ys:List),?ys:List),cons(?y:Nat,rev(?xs:List))),?xs:List),?zs:List) = rev(app(app(app(?ys:List,?zs:List),rev(?zs:List)),app(?zs:List,?zs:List)))",
"app(rev(app(?xs:List,?ys:List)),app(cons(times(plus(?x:Nat,?y:Nat),0),?xs:List),?ys:List)) = app(app(app(app(?ys:List,cons(0,?xs:List)),?ys:List),?xs:List),?zs:List)",
"app(app(app(?xs:List,cons(0,app(?zs:List,?xs:List))),cons(s(plus(?x:Nat,?x:Nat)),?ys:List)),?zs:List) = rev(app(?xs:List,rev(app(app(?zs:List,?ys:List),?zs:List))))",
"app(app(app(app(?xs:List,?xs:List),cons(plus(?x:Nat,0),app(?zs:List,?xs:List))),?ys:List),app(?zs:List,?zs:List)) = app(?xs:List,rev(app(app(?xs:List,?ys:List),?zs:List)))",
"app(rev(app(cons(times(0,0),?ys:List),?ys:List)),app(?xs:List,rev(app(app(?xs:List,?ys:List),?xs:List)))) = app(app(rev(?ys:List),?zs:List),app(?zs:List,?xs:List))",
"app(app(rev(app(?zs:List,app(?ys:List,?ys:List))),app(app(?zs:List,?xs:List),app(?xs:List,cons(?x:Nat,?xs:List)))),?zs:List) = cons(?x:Nat,app(cons(?y:Nat,?xs:List),?ys:List))",
"app(cons(s(0),app(?xs:List,?ys:List)),app(?xs:List,app(cons(plus(?x:Nat,?x:Nat),?xs:List),app(?zs:List,?zs:List)))) = app(?zs:List,cons(?x:Nat,app(?xs:List,?xs:List)))",
"cons(?x:Nat,app(app(app(cons(0,?ys:List),app(?zs:List,?zs:List)),?zs:List),app(cons(len(?zs:List),?xs:List),?ys:List))) = app(rev(rev(app(?zs:List,?ys:List))),?ys:List)",
"app(app(?xs:List,app(app(?zs:List,?zs:List),?ys:List)),app(app(?xs:List,app(?xs:List,?xs:List)),app(?xs:List,rev(rev(?ys:List))))) = app(app(?xs:List,rev(?zs:List)),?ys:List)",
"app(app(app(?zs:List,?zs:List),app(app(rev(?ys:List),?xs:List),?xs:List)),app(app(?xs:List,?ys:List),app(?ys:List,rev(app(?xs:List,?xs:List))))) = app(?zs:List,rev(?ys:List))",
"app(rev(app(?zs:List,app(?zs:List,?xs:List))),app(app(?ys:List,?ys:List),app(rev(rev(?ys:List)),app(app(?zs:List,?ys:List),rev(?xs:List))))) = rev(app(?ys:List,?xs:List))",
"app(?xs:List,app(app(app(?zs:List,app(?zs:List,?xs:List)),app(rev(?zs:List),app(?zs:List,?xs:List))),app(app(?ys:List,app(?zs:List,?xs:List)),?xs:List))) = app(?ys:List,?ys:List)",
"app(app(?xs:List,app(cons(?x:Nat,app(?xs:List,?xs:List)),app(app(?ys:List,?ys:List),?xs:List))),app(?ys:List,app(cons(?x:Nat,rev(?xs:List)),?zs:List))) = app(?zs:List,?ys:List)",
"app(app(?xs:List,cons(len(app(?ys:List,?ys:List)),app(?ys:List,cons(0,?xs:List)))),app(rev(?ys:List),rev(cons(plus(?x:Nat,0),?ys:List)))) = app(?ys:List,?zs:List)",
"app(app(cons(times(s(?y:Nat),?x:Nat),cons(times(?y:Nat,?y:Nat),?xs:List)),app(app(cons(0,?xs:List),?ys:List),?ys:List)),app(?ys:List,?zs:List)) = app(?xs:List,?zs:List)",
"cons(s(len(rev(?xs:List))),cons(plus(times(0,len(?xs:List)),len(app(?xs:List,?ys:List))),app(?xs:List,app(cons(?x:Nat,?xs:List),?xs:List)))) = app(?ys:List,?xs:List)",
"app(app(?xs:List,cons(?x:Nat,app(app(?xs:List,?xs:List),?xs:List))),app(app(?ys:List,app(app(?ys:List,?ys:List),app(?xs:List,?xs:List))),app(?zs:List,?xs:List))) = rev(?xs:List)",
"rev(?ys:List) = app(cons(0,?xs:List),app(app(app(?xs:List,cons(0,?ys:List)),cons(len(?zs:List),cons(?x:Nat,?xs:List))),app(app(cons(?x:Nat,?xs:List),?xs:List),?xs:List)))",
"app(?zs:List,?xs:List) = app(app(cons(times(0,times(?x:Nat,0)),?zs:List),app(?ys:List,cons(0,?xs:List))),rev(cons(0,cons(s(?x:Nat),cons(?y:Nat,?xs:List)))))",
"app(?zs:List,?zs:List) = app(app(?xs:List,rev(app(?zs:List,?xs:List))),app(app(app(app(?zs:List,?ys:List),app(?xs:List,?zs:List)),app(app(?ys:List,?xs:List),rev(?zs:List))),?ys:List))",
"app(?xs:List,app(?zs:List,?xs:List)) = app(cons(?x:Nat,?xs:List),cons(plus(plus(times(?x:Nat,0),len(?zs:List)),?x:Nat),cons(plus(plus(0,0),len(?xs:List)),?xs:List)))",
"app(?xs:List,cons(0,?ys:List)) = app(app(app(app(rev(?ys:List),?zs:List),app(rev(?ys:List),?xs:List)),?xs:List),app(app(app(app(?ys:List,?zs:List),?zs:List),?xs:List),?ys:List))",
"app(app(rev(?xs:List),?zs:List),cons(?y:Nat,?ys:List)) = app(app(?ys:List,?zs:List),app(app(?ys:List,app(?zs:List,?zs:List)),app(app(cons(?y:Nat,?ys:List),rev(?zs:List)),?zs:List)))",
"app(app(?xs:List,app(app(?zs:List,rev(?xs:List)),?zs:List)),?xs:List) = rev(app(cons(?y:Nat,app(app(?ys:List,?xs:List),app(?xs:List,?xs:List))),cons(times(0,?x:Nat),?ys:List)))",
"app(app(app(?zs:List,app(?ys:List,app(?zs:List,?xs:List))),?zs:List),?ys:List) = cons(times(len(cons(plus(?y:Nat,?x:Nat),?zs:List)),len(app(app(?zs:List,?xs:List),?ys:List))),?ys:List)",
"app(app(app(app(?zs:List,?zs:List),?zs:List),app(?zs:List,?xs:List)),?xs:List) = app(?ys:List,app(?xs:List,cons(times(times(?y:Nat,0),?x:Nat),app(app(?zs:List,?ys:List),?ys:List))))",
"app(app(?xs:List,cons(?x:Nat,app(app(?ys:List,?xs:List),rev(?xs:List)))),?xs:List) = app(app(?xs:List,app(app(cons(?y:Nat,?xs:List),app(?zs:List,?xs:List)),rev(?ys:List))),?zs:List)",
"app(rev(cons(?x:Nat,app(app(?zs:List,?zs:List),app(?xs:List,?xs:List)))),?ys:List) = app(rev(app(app(app(?xs:List,?zs:List),?xs:List),cons(?y:Nat,?xs:List))),rev(rev(?xs:List)))",
"app(?xs:List,app(app(app(app(?xs:List,?xs:List),?xs:List),?xs:List),app(?ys:List,?zs:List))) = app(cons(len(?xs:List),app(?ys:List,app(rev(?zs:List),?zs:List))),cons(0,?zs:List))",
"app(?ys:List,app(?ys:List,app(app(?zs:List,?ys:List),app(?ys:List,cons(?x:Nat,?zs:List))))) = app(app(app(app(app(?zs:List,?zs:List),app(?xs:List,?ys:List)),?ys:List),?xs:List),?xs:List)",
"app(?ys:List,app(app(?xs:List,app(?xs:List,?ys:List)),app(app(?zs:List,?ys:List),?xs:List))) = app(cons(0,app(?xs:List,app(app(?xs:List,?zs:List),?ys:List))),cons(?x:Nat,?ys:List))",
"app(rev(app(app(?xs:List,app(?xs:List,?zs:List)),?xs:List)),cons(?x:Nat,rev(?zs:List))) = app(app(app(app(app(?zs:List,?xs:List),app(?xs:List,?xs:List)),?xs:List),?zs:List),?zs:List)",
"app(rev(app(app(rev(?zs:List),app(?ys:List,?ys:List)),rev(?xs:List))),rev(?zs:List)) = app(?zs:List,cons(0,app(app(?ys:List,rev(?xs:List)),cons(s(?x:Nat),?xs:List))))",
"app(?zs:List,app(app(?ys:List,app(?xs:List,?zs:List)),app(?ys:List,app(?xs:List,rev(?ys:List))))) = app(?ys:List,app(rev(?zs:List),app(?xs:List,app(cons(?x:Nat,?ys:List),?xs:List))))",
"app(app(?zs:List,rev(app(?ys:List,app(?xs:List,?zs:List)))),app(?ys:List,app(?ys:List,?ys:List))) = app(app(?xs:List,cons(plus(times(?x:Nat,?y:Nat),?x:Nat),rev(?xs:List))),?xs:List)",
"app(?xs:List,app(app(app(app(?zs:List,?ys:List),?xs:List),app(app(?xs:List,?zs:List),?zs:List)),?xs:List)) = app(?xs:List,app(?zs:List,app(?ys:List,cons(plus(?y:Nat,?y:Nat),?ys:List))))",
"app(app(app(?ys:List,?ys:List),?zs:List),app(?ys:List,app(?ys:List,app(?ys:List,app(?xs:List,?xs:List))))) = app(rev(rev(?zs:List)),app(?zs:List,app(?xs:List,app(?xs:List,?xs:List))))",
"app(app(cons(plus(s(?x:Nat),len(?xs:List)),?zs:List),rev(app(?xs:List,rev(?ys:List)))),?zs:List) = rev(app(?xs:List,app(rev(?xs:List),cons(?x:Nat,app(?xs:List,?ys:List)))))",
"app(cons(0,?xs:List),app(app(app(app(?zs:List,?ys:List),app(?xs:List,?ys:List)),?zs:List),?xs:List)) = app(cons(?y:Nat,?zs:List),app(app(?xs:List,app(?zs:List,?xs:List)),?xs:List))",
"app(app(?xs:List,app(rev(cons(?x:Nat,?zs:List)),?zs:List)),app(app(app(app(?zs:List,?xs:List),rev(?xs:List)),?ys:List),?zs:List)) = cons(?x:Nat,app(?zs:List,app(?xs:List,?xs:List)))",
"app(app(?ys:List,app(?ys:List,?xs:List)),app(rev(rev(?xs:List)),app(app(?zs:List,rev(?xs:List)),app(rev(?zs:List),?xs:List)))) = app(?ys:List,app(cons(0,?xs:List),?zs:List))",
"app(?zs:List,app(app(app(cons(?y:Nat,?zs:List),app(?ys:List,?xs:List)),app(?ys:List,app(?xs:List,?zs:List))),app(app(?xs:List,?xs:List),?zs:List))) = app(cons(?y:Nat,?xs:List),?ys:List)",
"app(app(?zs:List,app(app(rev(?ys:List),cons(0,?xs:List)),app(?xs:List,?xs:List))),cons(?x:Nat,cons(?x:Nat,app(?xs:List,rev(?xs:List))))) = app(?xs:List,app(?zs:List,?zs:List))",
"app(app(cons(times(len(?xs:List),?x:Nat),?xs:List),app(?ys:List,cons(times(?x:Nat,0),?zs:List))),app(cons(s(?y:Nat),?xs:List),?ys:List)) = app(?xs:List,app(?zs:List,?xs:List))",
"app(?xs:List,cons(len(app(cons(?x:Nat,?ys:List),app(?ys:List,?xs:List))),cons(times(0,plus(?x:Nat,?x:Nat)),cons(0,app(?ys:List,?xs:List))))) = cons(s(?y:Nat),?zs:List)",
"app(?ys:List,app(app(?ys:List,app(?zs:List,?ys:List)),app(app(app(?xs:List,?xs:List),rev(?xs:List)),cons(plus(?x:Nat,?y:Nat),app(?ys:List,?ys:List))))) = app(?xs:List,rev(?ys:List))",
"app(app(app(app(?zs:List,app(?ys:List,?xs:List)),app(app(?ys:List,?xs:List),app(?zs:List,?xs:List))),app(cons(?y:Nat,app(?zs:List,?ys:List)),?ys:List)),?ys:List) = app(?xs:List,?zs:List)",
"app(app(cons(s(?x:Nat),cons(len(?ys:List),app(?xs:List,?zs:List))),cons(plus(?x:Nat,plus(?y:Nat,?x:Nat)),?zs:List)),app(app(?xs:List,?zs:List),?zs:List)) = app(?ys:List,?ys:List)",
"cons(plus(plus(plus(len(?xs:List),?y:Nat),len(app(?ys:List,?ys:List))),times(plus(?x:Nat,len(?xs:List)),0)),app(?ys:List,app(app(?xs:List,?xs:List),?xs:List))) = rev(?xs:List)",
"rev(?xs:List) = cons(plus(plus(?y:Nat,len(?xs:List)),plus(s(?y:Nat),0)),cons(s(plus(len(?ys:List),times(0,?x:Nat))),app(app(?ys:List,app(?xs:List,?xs:List)),?xs:List)))",
"cons(0,?xs:List) = app(?zs:List,app(app(rev(app(?ys:List,?xs:List)),app(app(?zs:List,?xs:List),?xs:List)),app(app(?zs:List,app(?xs:List,?ys:List)),cons(plus(?x:Nat,0),?ys:List))))",
"app(?ys:List,cons(?y:Nat,?ys:List)) = app(?zs:List,app(app(app(app(?zs:List,?xs:List),?zs:List),app(rev(?xs:List),app(?ys:List,?xs:List))),app(?xs:List,app(?ys:List,app(?xs:List,?zs:List)))))",
"app(app(rev(?ys:List),?ys:List),?ys:List) = app(?zs:List,app(app(app(?ys:List,app(?ys:List,?xs:List)),?zs:List),app(app(app(?zs:List,?ys:List),cons(?x:Nat,?ys:List)),app(?xs:List,?xs:List))))",
"app(rev(?xs:List),app(?xs:List,?zs:List)) = app(app(app(app(app(?zs:List,?xs:List),cons(0,?xs:List)),app(?xs:List,app(?zs:List,?xs:List))),?zs:List),app(app(?xs:List,?xs:List),?ys:List))",
"app(app(?xs:List,?ys:List),cons(len(?zs:List),?zs:List)) = cons(times(s(times(0,0)),plus(0,len(app(?ys:List,?ys:List)))),app(app(app(?xs:List,?zs:List),?zs:List),?ys:List))",
"app(app(?ys:List,?xs:List),app(rev(?xs:List),app(?zs:List,?zs:List))) = app(?ys:List,app(app(app(app(?zs:List,?xs:List),?zs:List),rev(?xs:List)),app(rev(app(?xs:List,?zs:List)),?ys:List)))",
"app(app(?zs:List,app(?zs:List,rev(?xs:List))),cons(0,?zs:List)) = app(app(?ys:List,app(app(app(?zs:List,?xs:List),app(?xs:List,?xs:List)),app(app(?xs:List,?xs:List),?ys:List))),?xs:List)",
"rev(app(rev(app(?xs:List,?zs:List)),app(?xs:List,app(?zs:List,rev(?zs:List))))) = app(app(app(?ys:List,?ys:List),app(?xs:List,?xs:List)),app(?xs:List,app(cons(?x:Nat,?ys:List),?xs:List)))",
"app(?xs:List,app(app(?ys:List,?ys:List),app(?ys:List,app(?zs:List,cons(0,?zs:List))))) = app(?ys:List,cons(?y:Nat,app(app(?xs:List,?ys:List),app(rev(?ys:List),app(?xs:List,?zs:List)))))",
"app(app(?xs:List,rev(app(?ys:List,rev(?xs:List)))),app(?xs:List,cons(?x:Nat,?xs:List))) = app(cons(0,cons(plus(?x:Nat,times(?x:Nat,0)),?zs:List)),app(rev(?ys:List),?xs:List))",
"app(app(?zs:List,cons(plus(s(?x:Nat),times(?x:Nat,?x:Nat)),rev(?zs:List))),?ys:List) = app(app(app(?xs:List,?zs:List),rev(?xs:List)),cons(len(app(?zs:List,?zs:List)),rev(?ys:List)))",
"app(app(app(app(app(?ys:List,?xs:List),?zs:List),?xs:List),?zs:List),app(?xs:List,?ys:List)) = cons(?y:Nat,app(rev(app(?ys:List,?ys:List)),app(?zs:List,app(?xs:List,app(?xs:List,?ys:List)))))",
"cons(plus(0,plus(times(plus(0,?x:Nat),?x:Nat),len(rev(?zs:List)))),?xs:List) = app(?xs:List,cons(len(?ys:List),cons(plus(plus(?y:Nat,?x:Nat),?y:Nat),rev(rev(?xs:List)))))",
"rev(app(?ys:List,app(app(?zs:List,app(?xs:List,?zs:List)),app(?ys:List,app(?ys:List,?ys:List))))) = app(app(?zs:List,?zs:List),app(app(?ys:List,app(rev(?xs:List),rev(?xs:List))),?ys:List))",
"app(?ys:List,app(app(app(?xs:List,app(?ys:List,?xs:List)),rev(app(?xs:List,?zs:List))),rev(?xs:List))) = app(rev(app(app(?ys:List,rev(?xs:List)),rev(app(?zs:List,?xs:List)))),?xs:List)",
"app(app(app(app(rev(?ys:List),?ys:List),app(?xs:List,?xs:List)),rev(app(?xs:List,?xs:List))),?xs:List) = app(app(app(?zs:List,rev(app(?xs:List,?xs:List))),?xs:List),app(?zs:List,?xs:List))",
"app(app(app(app(rev(?xs:List),?ys:List),?ys:List),?zs:List),app(app(?xs:List,?ys:List),app(?xs:List,?zs:List))) = cons(s(0),app(rev(app(app(?xs:List,?zs:List),?xs:List)),?ys:List))",
"app(app(app(rev(rev(?ys:List)),app(cons(?x:Nat,?xs:List),cons(?x:Nat,?zs:List))),cons(0,?zs:List)),?ys:List) = app(?xs:List,cons(times(s(?x:Nat),0),app(?xs:List,?ys:List)))",
"app(cons(?x:Nat,app(cons(?x:Nat,app(?xs:List,?ys:List)),app(app(?zs:List,?xs:List),app(?xs:List,?zs:List)))),?xs:List) = app(app(?zs:List,app(cons(s(?x:Nat),?zs:List),?zs:List)),?xs:List)",
"cons(times(plus(times(plus(0,?y:Nat),s(?y:Nat)),times(0,times(0,?x:Nat))),?x:Nat),app(?xs:List,?zs:List)) = app(app(cons(0,?zs:List),?zs:List),app(?xs:List,?zs:List))",
"app(cons(times(plus(0,?x:Nat),times(plus(?x:Nat,?x:Nat),0)),?ys:List),rev(app(app(rev(?ys:List),?xs:List),?ys:List))) = cons(0,rev(app(?ys:List,app(?ys:List,?zs:List))))",
"app(app(?xs:List,?xs:List),app(app(app(?xs:List,rev(?xs:List)),?ys:List),app(cons(0,app(?xs:List,?zs:List)),cons(?y:Nat,rev(?zs:List))))) = cons(times(?x:Nat,?x:Nat),rev(?xs:List))",
"app(app(rev(app(?xs:List,?xs:List)),app(cons(plus(?x:Nat,?x:Nat),?ys:List),?xs:List)),app(app(?xs:List,?ys:List),app(?xs:List,app(?xs:List,?xs:List)))) = cons(?x:Nat,app(?xs:List,?ys:List))",
"app(app(app(?xs:List,cons(?x:Nat,rev(?xs:List))),app(?xs:List,app(app(?xs:List,?ys:List),?zs:List))),app(app(cons(?y:Nat,cons(?x:Nat,?zs:List)),?ys:List),?xs:List)) = app(?ys:List,?xs:List)",
"cons(len(app(app(app(?ys:List,?xs:List),?xs:List),app(app(?xs:List,?zs:List),app(?ys:List,?ys:List)))),app(?zs:List,app(app(rev(?zs:List),rev(?zs:List)),?xs:List))) = app(?ys:List,?zs:List)",
"rev(?ys:List) = app(app(app(app(?xs:List,app(?xs:List,?xs:List)),app(?xs:List,rev(?zs:List))),app(app(cons(0,?zs:List),app(?zs:List,?zs:List)),app(app(?xs:List,?xs:List),?xs:List))),?ys:List)",
"app(app(?zs:List,app(?xs:List,?xs:List)),?xs:List) = app(app(app(?ys:List,app(app(?ys:List,?zs:List),rev(?zs:List))),rev(app(app(?ys:List,?xs:List),?xs:List))),app(?ys:List,app(?xs:List,?zs:List)))",
"cons(?y:Nat,app(app(?ys:List,?xs:List),?zs:List)) = app(app(rev(app(app(?xs:List,?zs:List),app(?xs:List,?xs:List))),app(app(?zs:List,app(?ys:List,?xs:List)),?zs:List)),cons(s(?x:Nat),?zs:List))",
"app(app(app(?zs:List,?ys:List),app(?xs:List,?zs:List)),?zs:List) = cons(s(0),app(app(app(app(?ys:List,?zs:List),?xs:List),?zs:List),app(?xs:List,cons(times(?y:Nat,?x:Nat),rev(?zs:List)))))",
"rev(app(rev(app(app(?ys:List,?zs:List),?xs:List)),?ys:List)) = app(?xs:List,cons(plus(times(?y:Nat,?x:Nat),?y:Nat),app(app(?xs:List,?ys:List),app(cons(?x:Nat,?xs:List),app(?xs:List,?zs:List)))))",
"cons(plus(times(len(app(?xs:List,?xs:List)),?x:Nat),?x:Nat),?ys:List) = app(?xs:List,app(rev(?ys:List),app(app(cons(0,?xs:List),cons(0,?ys:List)),cons(?x:Nat,app(?xs:List,?xs:List)))))",
"rev(app(app(rev(?zs:List),app(cons(?y:Nat,?ys:List),?ys:List)),?xs:List)) = app(app(app(app(?xs:List,?zs:List),?zs:List),?xs:List),app(cons(0,app(app(?ys:List,?ys:List),?xs:List)),?xs:List))",
"cons(0,app(app(?xs:List,app(?xs:List,app(?xs:List,?ys:List))),?zs:List)) = app(app(app(app(?xs:List,cons(0,?zs:List)),cons(plus(0,0),app(?zs:List,?ys:List))),?zs:List),?xs:List)",
"app(rev(rev(app(app(?xs:List,?xs:List),?zs:List))),cons(?y:Nat,app(?ys:List,?ys:List))) = rev(cons(times(?x:Nat,0),app(rev(?zs:List),app(app(?xs:List,?zs:List),app(?xs:List,?zs:List)))))",
"app(cons(times(plus(s(0),?x:Nat),len(?xs:List)),cons(?x:Nat,?xs:List)),?ys:List) = cons(?x:Nat,app(app(app(?zs:List,app(?xs:List,?ys:List)),app(?xs:List,app(?xs:List,?ys:List))),?ys:List))",
"rev(app(rev(?xs:List),app(?ys:List,app(app(?ys:List,?xs:List),app(?ys:List,?ys:List))))) = app(rev(app(app(rev(?xs:List),cons(0,?xs:List)),?zs:List)),app(?xs:List,app(?xs:List,?xs:List)))",
"cons(?y:Nat,app(app(rev(rev(?xs:List)),?zs:List),cons(?y:Nat,app(?zs:List,?zs:List)))) = app(app(cons(s(?x:Nat),?xs:List),rev(?xs:List)),app(app(?ys:List,app(?zs:List,?xs:List)),?ys:List))",
"app(cons(times(?x:Nat,plus(times(?y:Nat,0),s(?x:Nat))),app(?xs:List,?xs:List)),?zs:List) = app(rev(app(?xs:List,app(?zs:List,app(?zs:List,?zs:List)))),app(app(?ys:List,?xs:List),?zs:List))",
"app(?xs:List,app(cons(times(?x:Nat,plus(?y:Nat,?y:Nat)),app(?ys:List,app(?xs:List,?xs:List))),?xs:List)) = app(?zs:List,app(cons(len(?zs:List),?zs:List),app(rev(app(?ys:List,?xs:List)),?xs:List)))",
"app(app(?ys:List,app(app(?ys:List,?zs:List),app(app(?xs:List,?zs:List),app(?ys:List,?ys:List)))),?xs:List) = app(app(?xs:List,?ys:List),app(app(?ys:List,app(?xs:List,cons(0,?zs:List))),?xs:List))",
"app(app(?ys:List,rev(cons(s(?x:Nat),rev(?zs:List)))),rev(app(?xs:List,rev(app(?xs:List,?zs:List))))) = app(?zs:List,app(?xs:List,cons(times(s(0),0),app(?ys:List,?xs:List))))",
"cons(plus(s(0),?y:Nat),app(?xs:List,cons(?x:Nat,app(app(?ys:List,?zs:List),app(?xs:List,?zs:List))))) = app(rev(app(?ys:List,app(app(?zs:List,?xs:List),?zs:List))),app(?zs:List,?xs:List))",
"app(app(?ys:List,app(?ys:List,app(?xs:List,app(?xs:List,?ys:List)))),app(app(?xs:List,app(?ys:List,?zs:List)),?ys:List)) = app(rev(cons(times(len(?zs:List),s(?x:Nat)),rev(?xs:List))),?zs:List)",
"app(cons(s(plus(0,len(?zs:List))),app(?zs:List,?ys:List)),app(?xs:List,rev(app(app(?ys:List,?zs:List),?ys:List)))) = rev(app(?zs:List,app(app(cons(0,?xs:List),?xs:List),?ys:List)))",
"app(app(app(?ys:List,app(app(?xs:List,?zs:List),?xs:List)),app(?xs:List,app(app(?xs:List,?xs:List),app(?xs:List,?ys:List)))),?ys:List) = app(?zs:List,cons(s(len(cons(0,?xs:List))),?xs:List))",
"rev(app(app(?zs:List,app(app(?zs:List,?xs:List),app(?xs:List,?zs:List))),app(?ys:List,app(app(?xs:List,?zs:List),rev(?ys:List))))) = cons(?x:Nat,rev(cons(plus(?x:Nat,s(?x:Nat)),?xs:List)))",
"app(app(app(?xs:List,rev(cons(0,?zs:List))),app(app(?ys:List,app(?xs:List,?xs:List)),cons(times(?y:Nat,0),app(?xs:List,?xs:List)))),?xs:List) = app(rev(app(?zs:List,?zs:List)),?ys:List)",
"cons(times(plus(times(s(?x:Nat),0),len(cons(?x:Nat,?ys:List))),times(s(0),?x:Nat)),app(?xs:List,app(?xs:List,app(?xs:List,app(?xs:List,?ys:List))))) = rev(app(?zs:List,?xs:List))",
"rev(?xs:List) = app(app(cons(plus(s(0),?x:Nat),?xs:List),?xs:List),app(app(rev(app(?zs:List,?zs:List)),?ys:List),app(rev(rev(?ys:List)),app(app(?xs:List,?xs:List),app(?xs:List,?ys:List)))))",
"cons(?x:Nat,app(?xs:List,?zs:List)) = app(app(?xs:List,app(app(app(?xs:List,?ys:List),?xs:List),cons(len(?xs:List),app(?xs:List,?ys:List)))),app(app(?ys:List,cons(0,cons(0,?zs:List))),?zs:List))",
"app(?xs:List,cons(?y:Nat,rev(?xs:List))) = app(app(app(?ys:List,?xs:List),app(app(rev(?zs:List),?ys:List),?zs:List)),app(rev(?ys:List),app(?xs:List,cons(plus(0,?y:Nat),app(?xs:List,?ys:List)))))",
"app(?xs:List,app(app(?xs:List,?zs:List),app(?ys:List,?ys:List))) = app(app(?ys:List,cons(plus(0,plus(?x:Nat,?x:Nat)),rev(?xs:List))),app(app(app(app(?ys:List,?zs:List),?xs:List),?ys:List),?ys:List))",
"app(app(app(?xs:List,app(?xs:List,cons(0,?xs:List))),?zs:List),?ys:List) = app(app(?zs:List,?xs:List),app(app(rev(?xs:List),?xs:List),app(cons(plus(?x:Nat,?x:Nat),app(?ys:List,?zs:List)),?xs:List)))",
"rev(app(app(app(rev(?ys:List),?zs:List),app(?xs:List,?xs:List)),?ys:List)) = app(cons(?x:Nat,?ys:List),app(cons(len(app(?zs:List,?xs:List)),?ys:List),app(app(?xs:List,app(?ys:List,?xs:List)),?zs:List)))",
"rev(app(?ys:List,app(app(cons(?x:Nat,?xs:List),?ys:List),app(cons(0,?xs:List),rev(?xs:List))))) = app(app(rev(?xs:List),app(rev(?zs:List),?zs:List)),rev(app(?zs:List,app(?zs:List,?zs:List))))",
"app(app(cons(len(?xs:List),app(app(?xs:List,?ys:List),?zs:List)),?ys:List),app(app(?xs:List,?zs:List),?xs:List)) = app(cons(plus(len(?xs:List),times(?x:Nat,times(?x:Nat,?y:Nat))),rev(?xs:List)),?ys:List)",
"app(?ys:List,cons(times(plus(?y:Nat,?x:Nat),len(cons(?x:Nat,?ys:List))),app(rev(?xs:List),app(?ys:List,?zs:List)))) = app(?xs:List,app(app(?xs:List,app(app(?ys:List,?zs:List),rev(?zs:List))),?xs:List))",
"app(?zs:List,app(app(?zs:List,rev(?xs:List)),app(cons(plus(?x:Nat,?x:Nat),?ys:List),app(app(?zs:List,?zs:List),?xs:List)))) = app(app(cons(s(len(?xs:List)),app(?ys:List,?xs:List)),?xs:List),?ys:List)",
"app(?zs:List,app(cons(times(0,len(?zs:List)),app(?zs:List,app(?xs:List,?xs:List))),cons(times(?x:Nat,?x:Nat),?ys:List))) = app(?ys:List,cons(s(?y:Nat),rev(app(app(?xs:List,?zs:List),?zs:List))))",
"app(app(?ys:List,app(?xs:List,?zs:List)),app(app(?xs:List,?xs:List),app(app(?xs:List,?ys:List),app(app(?zs:List,?xs:List),?xs:List)))) = app(app(?xs:List,app(?xs:List,rev(?ys:List))),app(?zs:List,?zs:List))",
"app(app(rev(app(rev(?xs:List),?zs:List)),app(cons(0,app(?zs:List,?zs:List)),?ys:List)),app(app(?xs:List,?xs:List),?ys:List)) = app(app(cons(s(?x:Nat),?xs:List),?zs:List),app(?xs:List,?ys:List))",
"app(app(?xs:List,?zs:List),app(app(app(?ys:List,app(?ys:List,?xs:List)),app(app(?xs:List,?xs:List),?xs:List)),app(?ys:List,cons(0,?ys:List)))) = app(app(?ys:List,rev(?zs:List)),app(?ys:List,?xs:List))",
"app(app(rev(?ys:List),app(app(?ys:List,app(?xs:List,?ys:List)),?xs:List)),app(app(?xs:List,?xs:List),cons(plus(s(?x:Nat),0),?xs:List))) = rev(rev(app(?ys:List,cons(s(?x:Nat),?xs:List))))",
"app(app(?zs:List,app(app(?ys:List,rev(?ys:List)),app(app(?xs:List,?xs:List),cons(?x:Nat,?xs:List)))),app(?zs:List,cons(s(?y:Nat),cons(?x:Nat,app(?xs:List,?ys:List))))) = rev(app(?xs:List,?ys:List))",
"app(cons(s(len(cons(0,?xs:List))),app(?zs:List,?ys:List)),app(app(rev(cons(?x:Nat,?xs:List)),app(app(?zs:List,?xs:List),?zs:List)),cons(s(s(?x:Nat)),?xs:List))) = app(?zs:List,?ys:List)",
"rev(?ys:List) = app(app(app(?ys:List,app(app(?xs:List,?ys:List),rev(?xs:List))),app(app(app(?xs:List,?zs:List),app(?ys:List,?xs:List)),?zs:List)),rev(app(app(?xs:List,?ys:List),app(?ys:List,rev(?zs:List)))))",
"app(?xs:List,rev(?ys:List)) = app(app(cons(plus(times(?y:Nat,?y:Nat),0),app(app(?zs:List,?zs:List),cons(0,?xs:List))),?xs:List),app(rev(app(app(?ys:List,?xs:List),app(?xs:List,?ys:List))),?xs:List))",
"app(?zs:List,rev(app(?xs:List,?xs:List))) = app(app(?xs:List,rev(?xs:List)),app(app(cons(s(?x:Nat),app(?ys:List,?xs:List)),app(rev(?xs:List),?ys:List)),app(app(cons(0,?xs:List),?zs:List),?zs:List)))",
"app(?ys:List,app(?xs:List,rev(app(?ys:List,app(?xs:List,?xs:List))))) = app(rev(app(?xs:List,?ys:List)),app(app(app(app(?zs:List,?xs:List),?xs:List),app(app(?zs:List,?xs:List),?xs:List)),app(?xs:List,?zs:List)))",
"cons(s(plus(len(?zs:List),s(plus(?x:Nat,?y:Nat)))),cons(0,?zs:List)) = app(?zs:List,app(app(?ys:List,?zs:List),app(rev(cons(?y:Nat,?ys:List)),app(cons(?x:Nat,?ys:List),app(?zs:List,?xs:List)))))",
"app(app(app(app(cons(?x:Nat,?xs:List),cons(?y:Nat,?xs:List)),?xs:List),?xs:List),?ys:List) = app(?ys:List,app(cons(plus(?x:Nat,0),app(app(?xs:List,?xs:List),?xs:List)),rev(app(?xs:List,rev(?xs:List)))))",
"app(app(?xs:List,cons(plus(plus(?x:Nat,?y:Nat),len(?zs:List)),app(?zs:List,?ys:List))),?ys:List) = rev(app(rev(app(rev(?zs:List),app(?xs:List,?xs:List))),cons(?x:Nat,app(?xs:List,app(?xs:List,?ys:List)))))",
"app(app(cons(?x:Nat,?zs:List),app(app(?zs:List,?xs:List),?zs:List)),cons(s(?x:Nat),?xs:List)) = app(rev(rev(cons(?y:Nat,rev(?ys:List)))),app(app(app(?xs:List,cons(?y:Nat,?xs:List)),?xs:List),?xs:List))",
"rev(app(?xs:List,app(app(?xs:List,app(?zs:List,?xs:List)),app(app(?zs:List,?ys:List),?zs:List)))) = app(app(app(cons(0,?xs:List),cons(times(?y:Nat,?x:Nat),?ys:List)),app(rev(?zs:List),?ys:List)),?xs:List)",
"rev(cons(times(?x:Nat,len(?zs:List)),app(app(app(?ys:List,?xs:List),?zs:List),rev(?zs:List)))) = cons(times(s(len(app(?xs:List,?xs:List))),s(times(?y:Nat,plus(0,?y:Nat)))),cons(?y:Nat,?ys:List))",
"app(?zs:List,app(app(rev(app(?xs:List,?xs:List)),?zs:List),app(?xs:List,app(rev(?zs:List),?zs:List)))) = app(app(?xs:List,cons(plus(times(?y:Nat,0),?x:Nat),app(app(?xs:List,?ys:List),?xs:List))),?zs:List)",
"app(app(app(?zs:List,?zs:List),app(?xs:List,app(app(?zs:List,?zs:List),app(?ys:List,?ys:List)))),?ys:List) = cons(times(plus(0,s(s(?y:Nat))),times(s(?x:Nat),times(len(?zs:List),?x:Nat))),?xs:List)",
"app(app(app(app(?ys:List,app(?ys:List,?xs:List)),app(?zs:List,?xs:List)),?xs:List),cons(?x:Nat,?xs:List)) = app(app(cons(?x:Nat,?xs:List),?zs:List),app(?ys:List,app(?zs:List,app(app(?ys:List,?zs:List),?xs:List))))",
"app(app(app(app(?zs:List,?ys:List),app(?xs:List,?xs:List)),?ys:List),cons(s(plus(0,0)),?xs:List)) = rev(app(app(?xs:List,?ys:List),cons(0,cons(plus(0,?x:Nat),cons(?x:Nat,?zs:List)))))",
"app(app(app(cons(?x:Nat,?zs:List),?zs:List),app(app(?xs:List,app(?xs:List,?xs:List)),?zs:List)),app(?xs:List,?xs:List)) = app(rev(app(?zs:List,app(rev(?zs:List),app(?ys:List,?xs:List)))),app(?zs:List,?xs:List))",
"app(cons(len(?xs:List),?zs:List),app(app(app(app(?ys:List,?ys:List),app(?ys:List,?ys:List)),rev(app(?ys:List,?ys:List))),?zs:List)) = app(rev(app(app(app(?zs:List,?zs:List),rev(?xs:List)),?zs:List)),?xs:List)",
"app(app(app(app(app(?zs:List,?xs:List),rev(?ys:List)),app(rev(?ys:List),?xs:List)),?zs:List),app(?xs:List,app(app(?ys:List,?ys:List),app(?xs:List,?xs:List)))) = app(app(?ys:List,?ys:List),app(?zs:List,?xs:List))",
"app(cons(len(cons(?y:Nat,?zs:List)),app(?ys:List,?xs:List)),cons(?x:Nat,app(cons(s(?x:Nat),app(?zs:List,?zs:List)),app(cons(?x:Nat,?ys:List),?xs:List)))) = app(?xs:List,cons(plus(?x:Nat,?x:Nat),?ys:List))",
"app(rev(app(app(?ys:List,app(?ys:List,?zs:List)),app(?ys:List,?zs:List))),cons(plus(?x:Nat,times(s(?x:Nat),len(?xs:List))),cons(times(?x:Nat,0),?zs:List))) = app(?xs:List,cons(s(?y:Nat),?xs:List))",
"app(cons(times(len(?xs:List),?x:Nat),app(app(?xs:List,app(?ys:List,?xs:List)),?ys:List)),app(?ys:List,app(app(cons(?x:Nat,?zs:List),app(?xs:List,?zs:List)),?ys:List))) = app(cons(len(?xs:List),?zs:List),?ys:List)",
"cons(times(s(len(app(?zs:List,?ys:List))),len(rev(rev(?ys:List)))),app(app(?ys:List,app(?ys:List,app(?xs:List,?xs:List))),app(?zs:List,rev(app(?ys:List,?zs:List))))) = app(?xs:List,app(?xs:List,?zs:List))",
"cons(len(app(cons(s(0),?zs:List),app(?ys:List,app(?xs:List,?zs:List)))),app(rev(?xs:List),cons(plus(times(?x:Nat,?x:Nat),len(?zs:List)),app(?ys:List,app(?zs:List,?xs:List))))) = app(?ys:List,?ys:List)",
"app(app(rev(?zs:List),?ys:List),app(?zs:List,?zs:List)) = app(app(?xs:List,?zs:List),cons(times(times(?y:Nat,?x:Nat),plus(?x:Nat,times(0,?x:Nat))),cons(?y:Nat,cons(plus(?x:Nat,?y:Nat),app(?xs:List,?zs:List)))))",
"rev(app(app(app(?ys:List,?ys:List),?zs:List),?zs:List)) = app(app(app(?xs:List,app(rev(?xs:List),?ys:List)),app(?ys:List,app(cons(?x:Nat,?zs:List),?zs:List))),cons(plus(?x:Nat,s(times(0,?x:Nat))),?ys:List))",
"app(rev(?ys:List),app(app(cons(len(?ys:List),?xs:List),?xs:List),?ys:List)) = cons(plus(plus(plus(times(?x:Nat,?y:Nat),times(?x:Nat,?y:Nat)),?x:Nat),plus(times(times(?x:Nat,?y:Nat),?y:Nat),s(?x:Nat))),?ys:List)",
"cons(times(plus(?x:Nat,times(plus(?y:Nat,?x:Nat),?y:Nat)),?x:Nat),app(?xs:List,?xs:List)) = app(cons(len(app(?ys:List,?xs:List)),?zs:List),cons(?x:Nat,cons(len(?xs:List),app(app(?xs:List,?xs:List),rev(?ys:List)))))",
"cons(plus(?y:Nat,?x:Nat),app(rev(app(app(?ys:List,?zs:List),?ys:List)),rev(app(rev(?xs:List),app(?ys:List,?ys:List))))) = app(app(?xs:List,rev(?zs:List)),app(?ys:List,app(app(?xs:List,rev(?ys:List)),?xs:List)))",
"cons(plus(0,times(plus(0,0),plus(0,s(0)))),app(app(?ys:List,?ys:List),app(?xs:List,?zs:List))) = rev(cons(?x:Nat,app(rev(app(?ys:List,?ys:List)),app(app(?xs:List,?zs:List),?ys:List))))",
"app(?ys:List,app(app(app(?ys:List,?xs:List),app(app(?xs:List,?xs:List),?xs:List)),app(app(app(?xs:List,?ys:List),?ys:List),?ys:List))) = app(rev(app(app(?xs:List,app(?xs:List,?ys:List)),app(?xs:List,?zs:List))),?xs:List)",
"app(app(app(app(?xs:List,?xs:List),app(?ys:List,app(?zs:List,?zs:List))),app(?ys:List,?xs:List)),app(?zs:List,rev(app(?ys:List,?zs:List)))) = app(app(?zs:List,app(cons(times(?x:Nat,?x:Nat),?xs:List),?zs:List)),?xs:List)",
"app(app(rev(?zs:List),cons(plus(s(?x:Nat),0),cons(times(?x:Nat,?x:Nat),?ys:List))),app(app(?xs:List,rev(?zs:List)),?xs:List)) = app(app(?xs:List,cons(?y:Nat,?xs:List)),app(?xs:List,cons(0,?zs:List)))",
"app(rev(app(app(?zs:List,app(?xs:List,?zs:List)),app(?ys:List,app(?xs:List,?xs:List)))),rev(cons(?x:Nat,cons(0,app(?xs:List,?xs:List))))) = app(cons(?x:Nat,app(?xs:List,app(?zs:List,?ys:List))),rev(?zs:List))",
"cons(times(plus(?x:Nat,?x:Nat),0),app(app(app(app(?ys:List,?zs:List),?xs:List),cons(0,app(?zs:List,?ys:List))),app(?xs:List,?zs:List))) = cons(s(times(?x:Nat,times(times(?y:Nat,?y:Nat),0))),?xs:List)",
"cons(len(app(cons(?y:Nat,app(?xs:List,?xs:List)),cons(plus(?y:Nat,?y:Nat),app(?xs:List,?zs:List)))),app(app(app(?ys:List,?zs:List),cons(times(?x:Nat,0),?xs:List)),?xs:List)) = app(app(?ys:List,?ys:List),?xs:List)",
"app(app(app(app(?zs:List,app(?xs:List,?zs:List)),app(rev(?xs:List),?xs:List)),app(?xs:List,rev(app(?ys:List,?ys:List)))),app(app(?zs:List,?xs:List),app(app(cons(?x:Nat,?xs:List),?zs:List),?ys:List))) = rev(?ys:List)",
"app(cons(plus(0,?x:Nat),?zs:List),?zs:List) = app(app(rev(app(app(?xs:List,?ys:List),app(?xs:List,?xs:List))),app(?zs:List,app(?ys:List,?xs:List))),cons(plus(s(?y:Nat),plus(?x:Nat,times(?x:Nat,?y:Nat))),?xs:List))",
"app(app(app(?ys:List,rev(rev(?xs:List))),?ys:List),?ys:List) = app(?ys:List,app(app(rev(?xs:List),app(app(?ys:List,?zs:List),app(?ys:List,?xs:List))),app(app(rev(?zs:List),?zs:List),cons(times(?y:Nat,?x:Nat),?xs:List))))",
"app(app(?ys:List,?xs:List),rev(app(?xs:List,cons(?y:Nat,?xs:List)))) = app(app(app(cons(?x:Nat,app(?ys:List,?ys:List)),?xs:List),cons(?y:Nat,app(?zs:List,?xs:List))),rev(cons(?x:Nat,app(?ys:List,app(?ys:List,?xs:List)))))",
"app(rev(?xs:List),app(?zs:List,app(?zs:List,cons(0,?ys:List)))) = cons(plus(plus(times(len(?xs:List),plus(?x:Nat,?y:Nat)),len(app(?ys:List,?zs:List))),times(s(?x:Nat),?y:Nat)),app(app(?ys:List,?xs:List),?ys:List))",
"rev(app(app(app(?ys:List,cons(?y:Nat,?xs:List)),?ys:List),?xs:List)) = app(app(?zs:List,app(rev(app(?ys:List,?ys:List)),app(rev(?xs:List),app(?ys:List,?ys:List)))),app(cons(len(?xs:List),?zs:List),rev(rev(?xs:List))))",
"cons(times(times(s(plus(?x:Nat,?x:Nat)),0),s(?y:Nat)),?zs:List) = rev(app(app(app(?ys:List,cons(?x:Nat,?ys:List)),app(app(?zs:List,?ys:List),?ys:List)),app(app(?xs:List,?xs:List),app(?xs:List,rev(?zs:List)))))",
"app(app(?xs:List,app(cons(times(?x:Nat,?x:Nat),rev(?zs:List)),cons(0,?xs:List))),?xs:List) = app(cons(plus(len(app(?xs:List,?zs:List)),?x:Nat),app(?ys:List,app(?xs:List,app(?xs:List,?xs:List)))),app(?xs:List,?xs:List))",
"cons(?x:Nat,cons(times(?x:Nat,len(?xs:List)),app(cons(len(?xs:List),app(?xs:List,?xs:List)),?zs:List))) = app(app(rev(cons(s(0),?ys:List)),?zs:List),app(app(rev(app(?ys:List,?zs:List)),?xs:List),rev(?ys:List)))",
"app(app(app(app(app(?xs:List,?ys:List),?zs:List),?xs:List),app(app(?zs:List,rev(?zs:List)),?zs:List)),?xs:List) = app(?xs:List,app(?xs:List,app(app(app(?zs:List,?xs:List),rev(?ys:List)),app(cons(?x:Nat,?xs:List),?xs:List))))",
"app(app(app(rev(app(?zs:List,?zs:List)),?ys:List),?zs:List),app(cons(0,?zs:List),app(?ys:List,?zs:List))) = app(?xs:List,app(app(app(?zs:List,cons(?x:Nat,?zs:List)),?zs:List),app(rev(?xs:List),app(?ys:List,?zs:List))))",
"app(app(?ys:List,app(app(?ys:List,app(?xs:List,?ys:List)),rev(app(?xs:List,?ys:List)))),app(app(?zs:List,?xs:List),?zs:List)) = cons(len(rev(cons(len(?xs:List),rev(?xs:List)))),app(app(?ys:List,?zs:List),rev(?xs:List)))",
"app(app(app(app(app(?ys:List,?zs:List),cons(?y:Nat,?xs:List)),app(?xs:List,?ys:List)),rev(?ys:List)),app(cons(?x:Nat,?ys:List),cons(?x:Nat,?xs:List))) = app(app(app(cons(0,rev(?xs:List)),?xs:List),?xs:List),?ys:List)",
"app(app(rev(app(?xs:List,?xs:List)),?zs:List),app(app(app(cons(0,?zs:List),app(?xs:List,?zs:List)),cons(times(0,0),?zs:List)),?ys:List)) = app(?xs:List,app(?ys:List,cons(0,cons(s(0),?xs:List))))",
"app(app(?xs:List,?xs:List),app(app(?zs:List,app(app(?xs:List,?zs:List),app(?ys:List,?zs:List))),app(rev(app(?zs:List,?ys:List)),app(?xs:List,app(?xs:List,?zs:List))))) = app(rev(app(?ys:List,?xs:List)),app(?ys:List,?zs:List))",
"app(app(app(?xs:List,app(app(?zs:List,?ys:List),?zs:List)),app(app(app(?zs:List,?zs:List),app(?ys:List,?xs:List)),?zs:List)),cons(plus(?x:Nat,times(?y:Nat,plus(?x:Nat,0))),app(?xs:List,?zs:List))) = cons(0,?zs:List)",
"app(?xs:List,?xs:List) = app(cons(times(plus(0,plus(?x:Nat,?x:Nat)),?x:Nat),app(app(?xs:List,app(?xs:List,?xs:List)),?zs:List)),app(app(app(cons(?x:Nat,?zs:List),app(?xs:List,?zs:List)),app(?ys:List,rev(?xs:List))),?ys:List))",
"app(?ys:List,?zs:List) = app(app(app(app(cons(?x:Nat,?xs:List),app(?zs:List,?xs:List)),?ys:List),cons(plus(0,plus(?x:Nat,0)),?ys:List)),cons(plus(?x:Nat,plus(0,?x:Nat)),rev(app(app(?ys:List,?zs:List),?xs:List))))",
"app(app(?xs:List,?zs:List),cons(len(?ys:List),app(?zs:List,?ys:List))) = app(?xs:List,app(cons(times(times(0,?x:Nat),plus(?x:Nat,?y:Nat)),app(?xs:List,?zs:List)),app(cons(?x:Nat,app(?ys:List,?zs:List)),app(?xs:List,?zs:List))))",
"cons(?x:Nat,cons(?x:Nat,app(rev(app(?xs:List,?zs:List)),rev(?ys:List)))) = app(rev(cons(plus(times(?y:Nat,?x:Nat),s(?x:Nat)),rev(?zs:List))),app(app(?zs:List,app(app(?zs:List,?zs:List),app(?ys:List,?zs:List))),?zs:List))",
"app(app(?zs:List,app(cons(len(?xs:List),app(?ys:List,?xs:List)),?xs:List)),app(rev(?ys:List),?ys:List)) = cons(s(?x:Nat),app(app(app(app(?ys:List,?xs:List),cons(?y:Nat,?ys:List)),app(app(?ys:List,?ys:List),?xs:List)),?ys:List))",
"cons(plus(plus(?y:Nat,s(?x:Nat)),times(times(plus(0,?x:Nat),0),times(?y:Nat,?x:Nat))),rev(?xs:List)) = app(cons(?x:Nat,?zs:List),app(?xs:List,cons(len(app(?zs:List,?ys:List)),app(?ys:List,cons(0,?zs:List)))))",
"app(app(app(app(?ys:List,app(?xs:List,?zs:List)),app(?ys:List,app(?zs:List,?xs:List))),app(?ys:List,app(?zs:List,?ys:List))),?ys:List) = app(app(rev(app(?zs:List,app(?zs:List,?xs:List))),?ys:List),app(?xs:List,app(?ys:List,?zs:List)))",
"app(?xs:List,app(app(cons(times(0,0),?ys:List),app(app(?xs:List,?zs:List),?zs:List)),rev(app(app(?zs:List,?zs:List),?zs:List)))) = app(app(?zs:List,?xs:List),app(app(app(app(?ys:List,?ys:List),?ys:List),?xs:List),?xs:List))",
"app(app(cons(s(?x:Nat),?xs:List),app(app(app(?xs:List,?zs:List),app(?xs:List,?zs:List)),cons(times(?x:Nat,?x:Nat),?ys:List))),?ys:List) = app(app(?zs:List,app(?ys:List,?ys:List)),app(cons(s(len(?ys:List)),?ys:List),?ys:List))",
"app(app(rev(?zs:List),?zs:List),app(app(app(rev(?ys:List),app(?xs:List,?xs:List)),cons(plus(?x:Nat,?x:Nat),app(?ys:List,?ys:List))),app(?xs:List,rev(app(?ys:List,?xs:List))))) = cons(0,app(app(?zs:List,?zs:List),?xs:List))",
"app(app(app(app(app(?ys:List,?ys:List),?xs:List),?zs:List),app(?xs:List,rev(app(?xs:List,?ys:List)))),app(cons(?y:Nat,app(app(?xs:List,?ys:List),?ys:List)),app(?ys:List,app(?xs:List,?ys:List)))) = app(app(?xs:List,?zs:List),?ys:List)",
"app(?zs:List,app(app(?xs:List,?xs:List),?zs:List)) = app(app(app(app(?xs:List,?ys:List),app(app(?xs:List,?xs:List),?zs:List)),cons(s(plus(?x:Nat,0)),rev(?ys:List))),app(?xs:List,app(?xs:List,app(?xs:List,app(?xs:List,?xs:List)))))",
"cons(times(s(times(?x:Nat,plus(?y:Nat,?y:Nat))),plus(?x:Nat,0)),?zs:List) = app(app(rev(?ys:List),app(?xs:List,app(?xs:List,app(?ys:List,?zs:List)))),cons(0,app(app(app(?ys:List,?xs:List),app(?ys:List,?zs:List)),?ys:List)))",
"app(app(?xs:List,app(?zs:List,app(app(?xs:List,?zs:List),app(?xs:List,?ys:List)))),?xs:List) = app(?ys:List,app(app(app(app(?ys:List,?zs:List),app(?ys:List,?ys:List)),app(app(?ys:List,?zs:List),?ys:List)),cons(plus(?x:Nat,?y:Nat),?zs:List)))",
"app(app(?ys:List,app(?ys:List,?ys:List)),app(app(?ys:List,?zs:List),app(?ys:List,app(cons(?x:Nat,?ys:List),rev(?zs:List))))) = cons(plus(?x:Nat,plus(plus(plus(?y:Nat,?y:Nat),0),s(times(?y:Nat,0)))),cons(?y:Nat,?zs:List))",
"app(cons(plus(?x:Nat,plus(times(?y:Nat,?x:Nat),len(?ys:List))),app(app(app(?zs:List,?ys:List),app(?zs:List,?zs:List)),?xs:List)),?zs:List) = app(cons(?x:Nat,app(app(?xs:List,cons(?x:Nat,?zs:List)),cons(?x:Nat,rev(?xs:List)))),?ys:List)",
"cons(?y:Nat,?xs:List) = app(app(app(rev(?ys:List),?xs:List),app(app(?xs:List,?xs:List),app(?xs:List,app(?xs:List,?zs:List)))),cons(plus(plus(times(?x:Nat,?x:Nat),times(?x:Nat,?y:Nat)),plus(?y:Nat,plus(?y:Nat,?x:Nat))),app(?xs:List,?xs:List)))",
"rev(app(app(rev(rev(?ys:List)),?xs:List),?ys:List)) = app(cons(times(times(plus(?y:Nat,?x:Nat),times(?x:Nat,?y:Nat)),?y:Nat),app(app(?zs:List,?xs:List),app(rev(?ys:List),?zs:List))),app(?xs:List,rev(app(app(?ys:List,?xs:List),?xs:List))))",
"rev(app(cons(times(s(?x:Nat),plus(?x:Nat,?x:Nat)),?ys:List),rev(cons(?x:Nat,?xs:List)))) = app(app(app(?xs:List,?xs:List),cons(?x:Nat,app(?ys:List,app(?zs:List,?zs:List)))),app(cons(s(len(?zs:List)),rev(rev(?xs:List))),?zs:List))",
"cons(plus(plus(len(?ys:List),0),?y:Nat),app(?zs:List,app(?xs:List,app(?xs:List,rev(?zs:List))))) = app(app(?xs:List,app(cons(plus(?y:Nat,?x:Nat),?xs:List),?zs:List)),app(rev(?ys:List),app(cons(plus(?x:Nat,0),?ys:List),?ys:List)))",
"app(app(app(?xs:List,?ys:List),app(?xs:List,app(app(?zs:List,?xs:List),cons(0,?xs:List)))),app(?xs:List,app(?zs:List,?xs:List))) = app(app(app(?xs:List,app(?xs:List,app(?xs:List,?ys:List))),?xs:List),app(?xs:List,app(?ys:List,rev(?ys:List))))",
"cons(plus(s(plus(0,plus(?x:Nat,?x:Nat))),s(0)),app(app(app(?xs:List,app(?ys:List,?ys:List)),cons(len(?xs:List),?xs:List)),?zs:List)) = app(cons(times(len(app(?ys:List,?zs:List)),times(?x:Nat,s(?y:Nat))),?zs:List),?xs:List)",
"app(cons(?x:Nat,rev(?xs:List)),cons(times(s(len(?zs:List)),?x:Nat),cons(s(times(?y:Nat,?x:Nat)),app(app(?xs:List,?zs:List),cons(?y:Nat,?zs:List))))) = app(?ys:List,cons(s(?x:Nat),app(app(?ys:List,cons(0,?xs:List)),?xs:List)))",
"app(app(?xs:List,app(app(app(?zs:List,?xs:List),?xs:List),rev(app(?xs:List,?ys:List)))),app(app(?xs:List,app(app(?xs:List,?xs:List),app(?zs:List,?ys:List))),?ys:List)) = app(?xs:List,app(app(?ys:List,app(?ys:List,app(?zs:List,?ys:List))),?ys:List))",
"cons(plus(0,times(len(rev(?xs:List)),?y:Nat)),?xs:List) = app(app(app(app(app(?ys:List,?ys:List),app(?zs:List,?zs:List)),app(app(?xs:List,?zs:List),?zs:List)),cons(s(len(?ys:List)),app(app(?zs:List,?xs:List),app(?xs:List,?ys:List)))),?xs:List)",
"app(app(app(app(app(?xs:List,?ys:List),cons(?y:Nat,?xs:List)),?ys:List),?zs:List),?zs:List) = app(app(app(rev(?xs:List),?xs:List),app(app(app(?zs:List,?zs:List),?ys:List),app(?ys:List,app(?xs:List,?zs:List)))),app(?zs:List,rev(rev(rev(?xs:List)))))",
"app(app(app(?xs:List,app(?ys:List,app(?zs:List,?ys:List))),rev(app(?zs:List,?ys:List))),cons(times(times(?x:Nat,?x:Nat),s(?x:Nat)),?ys:List)) = app(?xs:List,app(app(?ys:List,app(app(?ys:List,?ys:List),app(?ys:List,?zs:List))),app(?xs:List,?zs:List)))",
"app(app(?zs:List,app(?zs:List,app(app(?zs:List,?xs:List),app(?xs:List,?zs:List)))),app(app(app(app(?ys:List,?xs:List),?zs:List),app(rev(?xs:List),app(?zs:List,?ys:List))),app(rev(app(?zs:List,?ys:List)),?ys:List))) = app(app(?zs:List,?ys:List),?xs:List)",
"rev(app(app(app(?xs:List,rev(?ys:List)),?zs:List),?ys:List)) = app(app(?ys:List,rev(?zs:List)),app(app(cons(s(?x:Nat),app(?xs:List,?ys:List)),app(rev(?xs:List),cons(?x:Nat,?xs:List))),app(cons(?x:Nat,?ys:List),app(app(?xs:List,?ys:List),?xs:List))))",
"app(app(?xs:List,?zs:List),app(?xs:List,app(?xs:List,app(?xs:List,app(?zs:List,?zs:List))))) = app(app(app(?xs:List,?zs:List),app(?zs:List,?xs:List)),app(rev(app(app(?xs:List,?ys:List),cons(?x:Nat,?xs:List))),rev(app(rev(?zs:List),cons(?y:Nat,?ys:List)))))",
"cons(times(times(s(len(?xs:List)),?x:Nat),len(rev(app(?ys:List,?ys:List)))),?zs:List) = app(app(app(app(?zs:List,app(?zs:List,?ys:List)),app(?zs:List,app(?zs:List,?ys:List))),app(?xs:List,?zs:List)),app(?xs:List,app(cons(len(?zs:List),?zs:List),?ys:List)))",
"app(app(app(app(?xs:List,?ys:List),cons(0,app(?xs:List,?xs:List))),?ys:List),rev(app(?ys:List,?ys:List))) = app(app(?xs:List,app(?xs:List,app(?xs:List,app(?ys:List,?xs:List)))),app(?xs:List,cons(0,app(cons(?x:Nat,?zs:List),app(?zs:List,?ys:List)))))",
"cons(plus(plus(len(app(?xs:List,?ys:List)),?y:Nat),times(?y:Nat,s(plus(?x:Nat,?x:Nat)))),app(?xs:List,rev(?ys:List))) = app(cons(s(?y:Nat),app(rev(app(?zs:List,?ys:List)),?ys:List)),cons(0,app(?zs:List,app(?xs:List,app(?xs:List,?xs:List)))))",
"app(app(app(app(app(?zs:List,?xs:List),app(?xs:List,?xs:List)),rev(cons(0,?ys:List))),app(rev(app(?xs:List,?ys:List)),app(?ys:List,app(?ys:List,?ys:List)))),?ys:List) = app(app(?xs:List,app(?zs:List,rev(?xs:List))),app(?zs:List,app(?xs:List,?xs:List)))",
"app(app(app(cons(plus(?y:Nat,?x:Nat),app(?xs:List,?zs:List)),app(?ys:List,?xs:List)),?ys:List),app(?xs:List,cons(times(len(?zs:List),?x:Nat),rev(app(?zs:List,?zs:List))))) = app(cons(times(times(plus(0,0),?x:Nat),0),?xs:List),rev(?ys:List))",
"app(app(cons(times(0,times(?y:Nat,?x:Nat)),?zs:List),app(app(?xs:List,?ys:List),?zs:List)),cons(s(s(0)),app(app(app(?xs:List,?xs:List),?xs:List),?zs:List))) = rev(app(cons(plus(s(0),s(0)),app(?zs:List,?zs:List)),?zs:List))",
"app(app(cons(plus(?y:Nat,?y:Nat),app(?ys:List,?xs:List)),app(?zs:List,app(?xs:List,cons(?x:Nat,?xs:List)))),?zs:List) = app(app(app(?zs:List,?zs:List),?xs:List),app(app(app(app(?xs:List,?zs:List),app(?xs:List,?ys:List)),app(?xs:List,?zs:List)),app(?zs:List,?ys:List)))",
"app(app(app(app(cons(?y:Nat,?xs:List),?xs:List),app(cons(?x:Nat,?ys:List),?ys:List)),?ys:List),rev(app(rev(app(?ys:List,?zs:List)),?zs:List))) = app(cons(plus(times(len(?xs:List),s(?x:Nat)),times(?x:Nat,?y:Nat)),app(?xs:List,app(?xs:List,?xs:List))),?xs:List)",
"app(app(app(?ys:List,?xs:List),app(?zs:List,app(app(?zs:List,?ys:List),app(?xs:List,?zs:List)))),?ys:List) = app(rev(app(cons(times(?y:Nat,0),cons(?x:Nat,?xs:List)),?ys:List)),cons(s(s(?x:Nat)),app(?ys:List,app(app(?zs:List,?xs:List),app(?ys:List,?zs:List)))))",
"cons(?x:Nat,app(app(cons(?y:Nat,rev(?zs:List)),?zs:List),app(app(app(?ys:List,?zs:List),?xs:List),?xs:List))) = app(rev(app(cons(plus(?x:Nat,?x:Nat),?ys:List),?xs:List)),app(app(app(?ys:List,?zs:List),cons(s(?x:Nat),app(?xs:List,?ys:List))),app(?xs:List,?ys:List)))",
"app(app(cons(s(plus(0,?y:Nat)),cons(plus(0,0),?xs:List)),app(cons(0,?xs:List),?xs:List)),cons(plus(?y:Nat,?x:Nat),cons(?x:Nat,?xs:List))) = app(?xs:List,cons(?x:Nat,app(cons(0,?ys:List),app(app(?ys:List,?xs:List),app(?xs:List,?xs:List)))))",
"cons(plus(times(?x:Nat,plus(s(?x:Nat),?x:Nat)),s(plus(s(?x:Nat),s(?x:Nat)))),app(cons(s(len(?xs:List)),app(app(?zs:List,?ys:List),?zs:List)),?xs:List)) = cons(0,app(app(?ys:List,app(app(?xs:List,?zs:List),?zs:List)),rev(cons(?y:Nat,?xs:List))))",
"app(app(cons(?x:Nat,app(?zs:List,cons(?x:Nat,?xs:List))),cons(s(?y:Nat),?xs:List)),app(app(cons(times(?x:Nat,0),?ys:List),?ys:List),app(app(?xs:List,?ys:List),rev(?zs:List)))) = app(rev(app(?xs:List,app(app(?xs:List,?xs:List),cons(?x:Nat,?ys:List)))),?ys:List)",
"app(app(app(app(?xs:List,app(?ys:List,?xs:List)),?zs:List),?zs:List),app(app(app(app(?zs:List,?xs:List),?ys:List),app(app(?xs:List,?xs:List),app(?xs:List,?xs:List))),app(rev(?ys:List),?xs:List))) = app(?ys:List,cons(times(s(times(?x:Nat,?y:Nat)),s(?x:Nat)),?ys:List))",
"app(app(?xs:List,cons(times(s(0),s(?x:Nat)),rev(?ys:List))),rev(rev(?ys:List))) = app(app(app(cons(s(0),rev(?xs:List)),app(rev(?xs:List),app(?xs:List,?zs:List))),app(app(cons(?x:Nat,?ys:List),?zs:List),app(?ys:List,cons(0,?ys:List)))),?zs:List)",
"app(?ys:List,app(app(rev(?xs:List),?xs:List),?xs:List)) = app(cons(len(?ys:List),cons(plus(times(0,?x:Nat),times(?y:Nat,0)),cons(len(?zs:List),cons(0,?zs:List)))),app(app(app(?ys:List,?xs:List),app(?xs:List,?xs:List)),rev(app(rev(?zs:List),app(?ys:List,?zs:List)))))",
"app(cons(s(0),rev(?ys:List)),rev(app(app(?zs:List,?xs:List),?xs:List))) = app(app(cons(plus(?y:Nat,len(?zs:List)),?ys:List),rev(cons(times(?y:Nat,?x:Nat),app(?zs:List,?xs:List)))),app(?ys:List,app(cons(plus(0,?x:Nat),?xs:List),app(?ys:List,app(?ys:List,?ys:List)))))",
"app(rev(app(?ys:List,rev(app(?ys:List,?zs:List)))),rev(app(app(?zs:List,rev(?zs:List)),rev(?xs:List)))) = app(app(?xs:List,cons(0,?ys:List)),app(cons(len(?ys:List),app(?xs:List,app(?zs:List,?xs:List))),app(rev(app(?xs:List,?ys:List)),app(?zs:List,app(?zs:List,?xs:List)))))",
"cons(plus(s(?x:Nat),times(times(?y:Nat,times(?x:Nat,0)),?x:Nat)),app(app(rev(?xs:List),?zs:List),?xs:List)) = app(app(app(app(?xs:List,app(?ys:List,?xs:List)),?xs:List),?zs:List),app(app(cons(0,app(?ys:List,?xs:List)),rev(?xs:List)),app(?ys:List,app(?ys:List,?ys:List))))",
"app(app(app(rev(?xs:List),?zs:List),app(?ys:List,app(?xs:List,?zs:List))),rev(app(?zs:List,rev(app(?xs:List,?xs:List))))) = app(?zs:List,app(app(app(app(?ys:List,?ys:List),?xs:List),app(app(?xs:List,?xs:List),app(?xs:List,?xs:List))),rev(cons(s(0),app(?zs:List,?zs:List)))))",
"app(rev(rev(?zs:List)),app(?ys:List,app(app(?xs:List,app(?zs:List,?xs:List)),app(?xs:List,cons(?y:Nat,?zs:List))))) = app(app(app(?zs:List,?xs:List),app(rev(app(?zs:List,?xs:List)),app(?ys:List,cons(?y:Nat,?xs:List)))),app(rev(app(cons(?x:Nat,?ys:List),app(?xs:List,?zs:List))),?zs:List))",
"app(app(app(?zs:List,?xs:List),app(app(app(?zs:List,?zs:List),?xs:List),cons(s(?x:Nat),?xs:List))),app(app(?xs:List,app(?ys:List,rev(?zs:List))),app(app(?zs:List,cons(0,?zs:List)),?ys:List))) = cons(?x:Nat,app(cons(?y:Nat,cons(len(?zs:List),rev(?ys:List))),app(?xs:List,?zs:List)))",
"cons(len(app(app(?ys:List,app(?xs:List,?xs:List)),?ys:List)),app(app(app(app(?xs:List,?zs:List),rev(?ys:List)),app(app(?ys:List,?xs:List),?xs:List)),?ys:List)) = app(app(cons(0,?zs:List),?ys:List),app(app(app(?zs:List,?xs:List),rev(app(?xs:List,?xs:List))),app(app(?ys:List,?xs:List),?xs:List)))",
"app(app(app(app(cons(?x:Nat,?ys:List),app(?ys:List,?ys:List)),app(app(?ys:List,?xs:List),?xs:List)),rev(app(?zs:List,?xs:List))),app(?xs:List,app(app(rev(?ys:List),app(?zs:List,?xs:List)),app(?ys:List,?zs:List)))) = app(cons(plus(plus(0,len(?zs:List)),?y:Nat),?xs:List),app(?ys:List,?zs:List))"
]


val problemsMaxMinYes =
IOMstrs.rdEqs [
"max(?x:Nat,?x:Nat) = min(?x:Nat,?x:Nat)",
"max(?y:Nat,0) = min(?y:Nat,?y:Nat)",
"max(0,?x:Nat) = min(?x:Nat,?x:Nat)",
"max(0,0) = min(0,?x:Nat)",
"max(?x:Nat,0) = max(?x:Nat,min(?y:Nat,?x:Nat))",
"min(?x:Nat,?x:Nat) = max(0,min(?x:Nat,?x:Nat))",
"min(?y:Nat,?y:Nat) = min(?y:Nat,max(?x:Nat,?y:Nat))",
"min(0,?x:Nat) = min(max(?x:Nat,0),0)",
"max(min(?y:Nat,?y:Nat),?x:Nat) = max(?x:Nat,?y:Nat)",
"max(min(?y:Nat,?z:Nat),0) = min(?z:Nat,?y:Nat)",
"max(min(0,?y:Nat),?z:Nat) = max(?z:Nat,0)",
"min(?x:Nat,max(?x:Nat,?x:Nat)) = max(?x:Nat,?x:Nat)",
"min(?x:Nat,max(0,?x:Nat)) = max(?x:Nat,?x:Nat)",
"min(?x:Nat,max(0,?y:Nat)) = min(?x:Nat,?y:Nat)",
"min(?y:Nat,max(?x:Nat,?y:Nat)) = max(?y:Nat,?y:Nat)",
"min(?y:Nat,max(?y:Nat,?x:Nat)) = max(?y:Nat,?y:Nat)",
"min(max(?y:Nat,?y:Nat),?x:Nat) = min(?x:Nat,?y:Nat)",
"min(max(?y:Nat,?y:Nat),0) = min(?x:Nat,0)",
"min(max(?y:Nat,?z:Nat),?y:Nat) = max(?y:Nat,0)",
"min(max(?z:Nat,?y:Nat),?y:Nat) = min(?y:Nat,?y:Nat)",
"min(max(0,?x:Nat),?x:Nat) = min(?x:Nat,?x:Nat)",
"max(?x:Nat,0) = max(0,max(min(?x:Nat,?x:Nat),?x:Nat))",
"max(?y:Nat,?y:Nat) = max(?y:Nat,min(?y:Nat,max(?y:Nat,?z:Nat)))",
"max(?y:Nat,?y:Nat) = max(min(?x:Nat,min(0,0)),?y:Nat)",
"min(?y:Nat,?x:Nat) = max(min(max(0,?y:Nat),?x:Nat),0)",
"min(?y:Nat,?y:Nat) = max(min(min(?y:Nat,0),0),?y:Nat)",
"min(?z:Nat,0) = min(max(?y:Nat,min(?x:Nat,?x:Nat)),0)",
"max(?x:Nat,min(?x:Nat,?z:Nat)) = max(max(0,?x:Nat),0)",
"max(min(0,?x:Nat),?x:Nat) = max(?x:Nat,min(?x:Nat,?z:Nat))",
"min(?x:Nat,max(?x:Nat,?y:Nat)) = max(min(?x:Nat,?y:Nat),?x:Nat)",
"min(max(?y:Nat,0),?x:Nat) = min(min(?y:Nat,?x:Nat),?x:Nat)",
"max(?x:Nat,max(?z:Nat,min(0,?y:Nat))) = max(?x:Nat,?z:Nat)",
"max(?x:Nat,min(?y:Nat,min(?y:Nat,?x:Nat))) = max(?x:Nat,0)",
"max(max(?y:Nat,min(?x:Nat,?y:Nat)),?y:Nat) = max(?y:Nat,?y:Nat)",
"max(min(?x:Nat,?x:Nat),min(?x:Nat,?y:Nat)) = max(?x:Nat,0)",
"max(min(min(?y:Nat,?y:Nat),?y:Nat),?x:Nat) = max(?x:Nat,?y:Nat)",
"min(?x:Nat,max(?y:Nat,max(?y:Nat,?y:Nat))) = min(?x:Nat,?y:Nat)",
"min(0,max(?z:Nat,max(?y:Nat,?x:Nat))) = min(0,?y:Nat)",
"min(max(?y:Nat,?x:Nat),min(?x:Nat,?y:Nat)) = min(?y:Nat,?x:Nat)",
"max(?x:Nat,?x:Nat) = min(?x:Nat,min(max(?z:Nat,?x:Nat),min(?x:Nat,?x:Nat)))",
"max(?x:Nat,?y:Nat) = max(?y:Nat,min(max(max(?x:Nat,?z:Nat),?x:Nat),?x:Nat))",
"max(?x:Nat,0) = max(?x:Nat,min(?x:Nat,max(?y:Nat,min(0,?y:Nat))))",
"max(?z:Nat,?z:Nat) = max(0,min(max(?z:Nat,max(?x:Nat,?z:Nat)),?z:Nat))",
"min(?y:Nat,?y:Nat) = max(min(?y:Nat,min(?y:Nat,min(?x:Nat,?z:Nat))),?y:Nat)",
"min(?z:Nat,?y:Nat) = min(?z:Nat,max(?y:Nat,max(0,min(0,?y:Nat))))",
"max(?x:Nat,min(?x:Nat,?y:Nat)) = max(?x:Nat,min(?x:Nat,min(?x:Nat,?x:Nat)))",
"max(?y:Nat,max(0,?y:Nat)) = min(max(max(?y:Nat,?x:Nat),0),?y:Nat)",
"min(?z:Nat,min(?x:Nat,?x:Nat)) = min(?z:Nat,min(max(?y:Nat,?x:Nat),?x:Nat))",
"max(min(?x:Nat,0),min(?y:Nat,?x:Nat)) = max(min(?x:Nat,?y:Nat),0)",
"max(min(?y:Nat,min(0,?x:Nat)),?y:Nat) = max(?y:Nat,min(?x:Nat,?y:Nat))",
"min(?z:Nat,max(?z:Nat,max(?y:Nat,?x:Nat))) = min(max(?z:Nat,0),?z:Nat)",
"max(max(min(max(?z:Nat,?x:Nat),?y:Nat),?y:Nat),?x:Nat) = max(?x:Nat,?y:Nat)",
"max(min(0,?x:Nat),max(min(?x:Nat,?x:Nat),?y:Nat)) = max(?y:Nat,?x:Nat)",
"max(min(max(?y:Nat,min(0,?y:Nat)),0),?y:Nat) = min(?y:Nat,?y:Nat)",
"max(0,0) = max(min(min(?x:Nat,min(0,?y:Nat)),?x:Nat),min(0,?z:Nat))",
"min(?x:Nat,0) = min(min(?y:Nat,?z:Nat),min(max(min(?x:Nat,0),0),?z:Nat))",
"min(?y:Nat,?x:Nat) = min(?y:Nat,min(max(min(0,max(?x:Nat,?x:Nat)),?x:Nat),?y:Nat))",
"min(?y:Nat,?x:Nat) = min(min(?x:Nat,min(?y:Nat,max(?y:Nat,max(?x:Nat,?x:Nat)))),?x:Nat)",
"min(?y:Nat,?y:Nat) = max(?y:Nat,min(min(min(max(?y:Nat,?y:Nat),0),?z:Nat),?y:Nat))",
"max(min(?y:Nat,?x:Nat),?y:Nat) = max(?y:Nat,min(max(min(?z:Nat,?y:Nat),0),0))",
"min(max(?x:Nat,0),0) = min(max(0,0),min(max(?y:Nat,?z:Nat),?y:Nat))",
"max(?x:Nat,min(?z:Nat,max(0,min(?x:Nat,?x:Nat)))) = min(?x:Nat,max(?x:Nat,?z:Nat))",
"max(min(min(?x:Nat,0),max(?y:Nat,?z:Nat)),?y:Nat) = max(?y:Nat,min(?y:Nat,?x:Nat))",
"max(?y:Nat,min(min(0,?x:Nat),min(?x:Nat,max(?y:Nat,?y:Nat)))) = max(?y:Nat,?y:Nat)",
"max(?y:Nat,min(min(0,min(max(?x:Nat,?y:Nat),?z:Nat)),0)) = min(?y:Nat,?y:Nat)",
"max(max(?x:Nat,min(min(?y:Nat,?x:Nat),max(0,?z:Nat))),0) = max(0,?x:Nat)",
"max(max(?y:Nat,min(0,max(?x:Nat,min(?y:Nat,0)))),0) = max(?y:Nat,?y:Nat)",
"max(max(min(max(min(0,?y:Nat),?y:Nat),?z:Nat),?y:Nat),?x:Nat) = max(?x:Nat,?y:Nat)",
"max(min(max(?z:Nat,?z:Nat),?x:Nat),max(min(?z:Nat,?z:Nat),?x:Nat)) = max(?x:Nat,?z:Nat)",
"max(min(min(?z:Nat,0),min(?x:Nat,max(?y:Nat,?x:Nat))),?x:Nat) = max(?x:Nat,?x:Nat)",
"min(?y:Nat,max(?x:Nat,max(?x:Nat,max(max(?x:Nat,?y:Nat),?y:Nat)))) = max(?y:Nat,0)",
"min(0,max(min(min(?y:Nat,max(?x:Nat,?x:Nat)),?x:Nat),0)) = min(0,?z:Nat)",
"max(?x:Nat,?x:Nat) = max(min(min(?y:Nat,min(?z:Nat,min(?z:Nat,?y:Nat))),?x:Nat),max(?x:Nat,0))",
"min(?y:Nat,?y:Nat) = min(max(min(?z:Nat,min(max(0,?z:Nat),?y:Nat)),?y:Nat),max(?y:Nat,?x:Nat))",
"min(0,?y:Nat) = min(min(max(?y:Nat,?y:Nat),min(?x:Nat,max(min(0,?y:Nat),?y:Nat))),0)",
"max(?x:Nat,max(?x:Nat,min(0,?x:Nat))) = max(?x:Nat,min(max(max(?x:Nat,?x:Nat),0),?y:Nat))",
"max(min(?y:Nat,max(?y:Nat,?y:Nat)),?z:Nat) = max(min(0,?y:Nat),max(?y:Nat,max(0,?z:Nat)))",
"max(min(?x:Nat,max(max(?y:Nat,?x:Nat),?x:Nat)),0) = max(min(max(?y:Nat,?y:Nat),0),?x:Nat)",
"max(?x:Nat,max(max(0,min(max(?x:Nat,?y:Nat),?x:Nat)),?y:Nat)) = max(?x:Nat,max(?y:Nat,?y:Nat))",
"max(max(max(min(?y:Nat,min(?x:Nat,?z:Nat)),?x:Nat),?y:Nat),?x:Nat) = max(min(?y:Nat,?y:Nat),?x:Nat)",
"min(?y:Nat,max(min(?z:Nat,min(min(?y:Nat,0),?x:Nat)),?x:Nat)) = min(min(?x:Nat,?x:Nat),?y:Nat)",
"min(?y:Nat,max(min(max(?x:Nat,max(?y:Nat,?y:Nat)),?x:Nat),?y:Nat)) = max(max(?y:Nat,?y:Nat),?y:Nat)",
"min(min(0,max(?x:Nat,0)),min(0,min(?y:Nat,?x:Nat))) = max(0,min(?y:Nat,0))",
"min(min(max(?y:Nat,min(?x:Nat,?y:Nat)),?y:Nat),min(0,?y:Nat)) = min(min(0,?x:Nat),0)",
"max(?z:Nat,max(min(max(0,?x:Nat),max(?y:Nat,max(?x:Nat,?z:Nat))),?x:Nat)) = max(?x:Nat,?z:Nat)",
"max(max(min(min(?y:Nat,?z:Nat),min(?x:Nat,max(?z:Nat,0))),?x:Nat),?x:Nat) = max(?x:Nat,?x:Nat)",
"max(min(?x:Nat,max(min(?y:Nat,min(?x:Nat,?y:Nat)),min(?y:Nat,?x:Nat))),?y:Nat) = min(?y:Nat,?y:Nat)",
"min(?y:Nat,max(?x:Nat,max(max(min(?y:Nat,?y:Nat),min(?z:Nat,?x:Nat)),?x:Nat))) = min(?y:Nat,?y:Nat)",
"min(0,min(0,max(min(?y:Nat,?x:Nat),min(max(?y:Nat,0),?x:Nat)))) = min(0,?x:Nat)",
"min(min(0,max(max(?z:Nat,max(?z:Nat,?y:Nat)),?y:Nat)),min(?z:Nat,?y:Nat)) = min(?x:Nat,0)",
"min(min(max(?x:Nat,min(max(?z:Nat,?x:Nat),?y:Nat)),max(?y:Nat,?z:Nat)),0) = min(?x:Nat,0)",
"max(?x:Nat,?x:Nat) = max(min(?x:Nat,max(min(max(?y:Nat,?y:Nat),?z:Nat),min(?x:Nat,0))),min(?x:Nat,?x:Nat))",
"max(?x:Nat,?x:Nat) = max(min(?y:Nat,?x:Nat),max(?x:Nat,min(0,max(min(?y:Nat,?x:Nat),max(?y:Nat,?x:Nat)))))",
"max(?x:Nat,?y:Nat) = max(max(?y:Nat,max(?y:Nat,0)),max(max(min(min(?x:Nat,?x:Nat),0),?y:Nat),?x:Nat))",
"max(?y:Nat,0) = min(?y:Nat,max(min(min(?y:Nat,?y:Nat),?y:Nat),min(max(max(?y:Nat,?x:Nat),?y:Nat),?x:Nat)))",
"min(?x:Nat,0) = min(?x:Nat,min(0,min(max(min(?x:Nat,?x:Nat),?x:Nat),max(min(0,?y:Nat),?x:Nat))))",
"min(?y:Nat,0) = min(min(?z:Nat,0),min(?y:Nat,min(max(?x:Nat,?y:Nat),min(?x:Nat,max(0,?y:Nat)))))",
"max(?x:Nat,max(?x:Nat,0)) = min(?x:Nat,min(min(?x:Nat,max(max(?y:Nat,?y:Nat),min(?x:Nat,?x:Nat))),?x:Nat))",
"max(min(?x:Nat,?y:Nat),?x:Nat) = max(min(?z:Nat,max(min(?x:Nat,0),min(max(0,0),?x:Nat))),?x:Nat)",
"max(?z:Nat,min(?z:Nat,max(?z:Nat,min(min(?y:Nat,?y:Nat),?y:Nat)))) = max(?z:Nat,min(0,min(0,0)))",
"max(0,max(?x:Nat,min(?x:Nat,max(?x:Nat,max(?x:Nat,0))))) = max(min(?y:Nat,min(0,?y:Nat)),?x:Nat)",
"max(max(max(max(min(?x:Nat,?z:Nat),?x:Nat),?x:Nat),?y:Nat),?y:Nat) = max(?x:Nat,max(min(?x:Nat,?x:Nat),?y:Nat))",
"min(?y:Nat,max(?y:Nat,max(?y:Nat,min(min(?x:Nat,?z:Nat),0)))) = max(min(min(0,0),?x:Nat),?y:Nat)",
"max(?x:Nat,max(max(?x:Nat,?y:Nat),min(min(?x:Nat,max(?y:Nat,?z:Nat)),max(?z:Nat,?y:Nat)))) = max(?x:Nat,?y:Nat)",
"max(max(?y:Nat,?y:Nat),max(min(?x:Nat,?x:Nat),min(?y:Nat,max(?z:Nat,min(?y:Nat,?x:Nat))))) = max(?x:Nat,?y:Nat)",
"max(max(0,?y:Nat),max(?x:Nat,min(max(max(0,?x:Nat),min(?x:Nat,?z:Nat)),?y:Nat))) = max(?x:Nat,?y:Nat)",
"max(max(max(?y:Nat,?y:Nat),?y:Nat),max(max(min(min(?y:Nat,?y:Nat),?y:Nat),?y:Nat),?x:Nat)) = max(?y:Nat,?x:Nat)",
"max(min(min(min(?y:Nat,?y:Nat),?x:Nat),min(min(0,?x:Nat),?x:Nat)),min(?x:Nat,?x:Nat)) = min(?x:Nat,?x:Nat)",
"min(max(max(?x:Nat,0),min(max(max(?y:Nat,?z:Nat),?x:Nat),?x:Nat)),max(?z:Nat,?x:Nat)) = max(?x:Nat,?x:Nat)",
"min(min(0,max(max(min(?x:Nat,?y:Nat),min(0,?x:Nat)),max(?y:Nat,0))),?x:Nat) = min(0,?x:Nat)",
"max(?y:Nat,0) = max(?y:Nat,min(max(max(max(0,0),?x:Nat),max(?y:Nat,?y:Nat)),min(min(0,?z:Nat),?y:Nat)))",
"max(0,?y:Nat) = max(max(?y:Nat,min(min(?y:Nat,max(?y:Nat,?y:Nat)),max(?x:Nat,min(0,?y:Nat)))),min(?y:Nat,?z:Nat))",
"min(?y:Nat,max(?x:Nat,?x:Nat)) = min(min(?x:Nat,?x:Nat),max(min(min(max(?x:Nat,?x:Nat),?x:Nat),?y:Nat),min(?x:Nat,?y:Nat)))",
"max(?y:Nat,min(0,min(min(?y:Nat,0),?y:Nat))) = max(max(max(0,min(max(0,?y:Nat),?x:Nat)),0),?y:Nat)",
"min(max(0,?x:Nat),max(?y:Nat,max(?x:Nat,?y:Nat))) = min(?x:Nat,max(max(min(0,max(0,0)),?x:Nat),?x:Nat))",
"min(max(min(max(?z:Nat,min(?y:Nat,0)),?y:Nat),?x:Nat),?x:Nat) = min(?x:Nat,min(?x:Nat,max(?x:Nat,max(0,?y:Nat))))",
"min(min(?x:Nat,?y:Nat),min(max(max(?y:Nat,?x:Nat),?x:Nat),?y:Nat)) = max(min(?y:Nat,?x:Nat),min(min(?y:Nat,?x:Nat),?y:Nat))",
"min(min(min(max(?z:Nat,?x:Nat),max(0,0)),?z:Nat),0) = min(min(min(?x:Nat,max(?x:Nat,?y:Nat)),0),?z:Nat)",
"max(min(?z:Nat,max(0,max(max(?z:Nat,0),min(?y:Nat,0)))),?y:Nat) = max(min(max(?y:Nat,?x:Nat),?y:Nat),?z:Nat)",
"min(?z:Nat,max(max(max(min(0,0),min(?y:Nat,?x:Nat)),?y:Nat),min(?x:Nat,?y:Nat))) = min(min(?z:Nat,?z:Nat),?y:Nat)",
"min(?y:Nat,max(?x:Nat,min(min(min(?x:Nat,?y:Nat),min(?y:Nat,?y:Nat)),max(max(?x:Nat,?x:Nat),0)))) = min(?x:Nat,?y:Nat)",
"max(?x:Nat,?x:Nat) = max(min(max(0,max(max(?z:Nat,?x:Nat),?x:Nat)),0),max(max(?x:Nat,min(0,max(?y:Nat,?z:Nat))),0))",
"max(?y:Nat,?x:Nat) = max(max(max(max(?x:Nat,?y:Nat),?y:Nat),max(min(?x:Nat,?y:Nat),max(?y:Nat,?x:Nat))),min(?z:Nat,min(?z:Nat,?y:Nat)))",
"min(?y:Nat,0) = min(max(?x:Nat,min(max(?x:Nat,?y:Nat),max(?x:Nat,0))),min(0,min(?y:Nat,max(?x:Nat,min(0,?x:Nat)))))",
"min(0,?z:Nat) = min(max(max(?y:Nat,min(min(0,?y:Nat),max(?x:Nat,0))),max(max(?z:Nat,?z:Nat),max(0,?x:Nat))),0)",
"min(max(?y:Nat,?y:Nat),?x:Nat) = min(max(min(min(?y:Nat,?x:Nat),0),max(min(min(?y:Nat,?z:Nat),max(?x:Nat,?z:Nat)),?y:Nat)),?x:Nat)",
"min(max(?x:Nat,min(max(min(?z:Nat,?y:Nat),0),min(?x:Nat,0))),?x:Nat) = max(?x:Nat,min(?y:Nat,min(?z:Nat,min(?y:Nat,?x:Nat))))",
"min(max(?y:Nat,?x:Nat),min(max(?y:Nat,min(?y:Nat,max(?x:Nat,?x:Nat))),max(?y:Nat,?x:Nat))) = max(min(min(?z:Nat,?y:Nat),?x:Nat),?y:Nat)",
"min(min(min(min(min(0,?z:Nat),max(?y:Nat,0)),0),?x:Nat),max(0,?x:Nat)) = min(0,max(max(?y:Nat,?x:Nat),?x:Nat))",
"min(max(max(min(?y:Nat,0),?y:Nat),max(max(?x:Nat,max(?y:Nat,?z:Nat)),?x:Nat)),max(0,?x:Nat)) = max(?x:Nat,min(0,?x:Nat))",
"max(?y:Nat,max(min(min(?x:Nat,max(?y:Nat,?y:Nat)),max(max(?y:Nat,?x:Nat),?x:Nat)),max(min(?y:Nat,?y:Nat),0))) = max(?y:Nat,?y:Nat)",
"min(min(min(0,max(max(?z:Nat,?y:Nat),?y:Nat)),min(max(?x:Nat,?y:Nat),max(min(0,?y:Nat),?x:Nat))),max(?y:Nat,0)) = min(0,?y:Nat)",
"min(max(?x:Nat,?x:Nat),0) = min(min(min(0,min(0,?y:Nat)),max(min(?y:Nat,?z:Nat),?x:Nat)),max(min(?x:Nat,min(?x:Nat,max(?y:Nat,0))),?x:Nat))",
"min(0,max(min(0,max(?x:Nat,?z:Nat)),?y:Nat)) = min(min(min(min(max(?x:Nat,?y:Nat),?y:Nat),max(max(?x:Nat,?x:Nat),max(?y:Nat,?x:Nat))),?y:Nat),0)",
"min(?x:Nat,max(min(?y:Nat,min(?x:Nat,?y:Nat)),min(?y:Nat,max(0,0)))) = max(min(min(?y:Nat,?x:Nat),max(?y:Nat,?z:Nat)),min(?y:Nat,min(0,0)))",
"max(max(?x:Nat,?y:Nat),min(min(min(0,min(?y:Nat,0)),max(?x:Nat,?z:Nat)),?x:Nat)) = max(0,max(max(min(min(?x:Nat,?y:Nat),?x:Nat),?x:Nat),?y:Nat))",
"max(min(min(min(min(?z:Nat,0),min(?y:Nat,?x:Nat)),max(min(?x:Nat,?x:Nat),?z:Nat)),0),?y:Nat) = max(?y:Nat,min(0,max(max(?z:Nat,?x:Nat),0)))",
"min(min(max(?y:Nat,min(?x:Nat,?x:Nat)),0),max(0,min(max(max(?z:Nat,0),min(?x:Nat,?z:Nat)),min(?y:Nat,?y:Nat)))) = min(max(?z:Nat,?y:Nat),0)",
"min(?y:Nat,?y:Nat) = min(max(max(min(?x:Nat,0),max(min(?z:Nat,?y:Nat),?y:Nat)),max(min(?y:Nat,?y:Nat),min(?y:Nat,min(?z:Nat,?x:Nat)))),min(max(?x:Nat,?y:Nat),?y:Nat))",
"min(?x:Nat,min(?y:Nat,?x:Nat)) = min(?y:Nat,min(max(min(?x:Nat,?x:Nat),min(max(?y:Nat,?z:Nat),max(0,?x:Nat))),min(min(max(?x:Nat,?y:Nat),max(?y:Nat,?y:Nat)),?y:Nat)))",
"min(?x:Nat,min(?y:Nat,min(?y:Nat,?z:Nat))) = min(min(min(min(min(?x:Nat,?x:Nat),max(?z:Nat,?z:Nat)),?y:Nat),min(min(?x:Nat,max(?x:Nat,?x:Nat)),min(?y:Nat,?z:Nat))),?x:Nat)",
"min(max(?y:Nat,max(?y:Nat,min(min(0,?x:Nat),?y:Nat))),min(min(?y:Nat,min(?y:Nat,min(?x:Nat,?y:Nat))),?x:Nat)) = min(?x:Nat,min(max(?y:Nat,min(?x:Nat,?x:Nat)),?y:Nat))",
"min(min(?x:Nat,min(?x:Nat,max(?y:Nat,?y:Nat))),min(max(min(?y:Nat,?x:Nat),max(min(0,?x:Nat),min(?x:Nat,?y:Nat))),?y:Nat)) = min(?x:Nat,min(?y:Nat,min(?x:Nat,?y:Nat)))",
"min(?y:Nat,max(max(min(?y:Nat,min(?y:Nat,?x:Nat)),max(max(0,?z:Nat),max(?y:Nat,?y:Nat))),max(?y:Nat,min(max(?y:Nat,0),min(?y:Nat,?x:Nat))))) = min(?y:Nat,?y:Nat)",
"max(0,max(min(?x:Nat,max(?y:Nat,max(?x:Nat,?x:Nat))),0)) = max(min(max(?y:Nat,max(max(?y:Nat,?y:Nat),?y:Nat)),min(?x:Nat,max(max(?y:Nat,?x:Nat),min(0,?y:Nat)))),?x:Nat)",
"min(min(max(?x:Nat,min(0,min(?x:Nat,0))),max(?y:Nat,0)),max(?x:Nat,min(max(?x:Nat,min(?y:Nat,?x:Nat)),?y:Nat))) = min(max(?z:Nat,?x:Nat),max(0,min(?x:Nat,min(?x:Nat,?y:Nat))))",
"max(min(?z:Nat,?y:Nat),min(?y:Nat,max(?y:Nat,?y:Nat))) = min(max(max(min(?x:Nat,max(?y:Nat,?y:Nat)),min(?x:Nat,max(?x:Nat,0))),?y:Nat),min(max(?y:Nat,max(?y:Nat,max(?y:Nat,?x:Nat))),min(?y:Nat,?y:Nat)))",
"min(min(?z:Nat,min(min(min(?x:Nat,?x:Nat),min(?x:Nat,0)),?x:Nat)),?y:Nat) = min(min(min(max(min(?x:Nat,?x:Nat),max(?z:Nat,?z:Nat)),0),?x:Nat),max(0,max(min(?y:Nat,max(?x:Nat,?z:Nat)),?x:Nat)))",
"min(max(0,0),max(?z:Nat,?y:Nat)) = min(max(min(min(max(?x:Nat,?x:Nat),0),min(min(?x:Nat,?z:Nat),?z:Nat)),0),max(min(?x:Nat,min(?y:Nat,max(?y:Nat,?x:Nat))),max(?y:Nat,min(?y:Nat,max(0,0)))))",
"max(min(min(min(max(?x:Nat,?x:Nat),min(?y:Nat,?x:Nat)),0),min(?z:Nat,?z:Nat)),0) = min(min(0,min(min(?z:Nat,?y:Nat),max(min(0,?x:Nat),0))),max(max(?x:Nat,max(?z:Nat,max(?y:Nat,?x:Nat))),?x:Nat))",
"min(min(min(min(max(?x:Nat,?y:Nat),max(?y:Nat,0)),0),max(max(min(0,?z:Nat),0),min(?x:Nat,?x:Nat))),max(min(max(min(?y:Nat,?y:Nat),min(?x:Nat,?y:Nat)),?y:Nat),0)) = min(?x:Nat,min(min(?z:Nat,?x:Nat),0))",
"max(max(max(?y:Nat,?y:Nat),?x:Nat),min(max(0,?x:Nat),min(?x:Nat,min(?y:Nat,?y:Nat)))) = max(max(min(max(max(?x:Nat,?x:Nat),max(?y:Nat,0)),max(max(?y:Nat,?z:Nat),?y:Nat)),min(?x:Nat,?x:Nat)),max(min(?x:Nat,?y:Nat),min(min(?y:Nat,?y:Nat),?x:Nat)))",
"max(?y:Nat,max(max(max(min(?y:Nat,?x:Nat),?x:Nat),min(min(?z:Nat,?y:Nat),0)),max(?y:Nat,?x:Nat))) = max(max(max(min(0,max(?x:Nat,?x:Nat)),min(?x:Nat,?y:Nat)),max(min(?y:Nat,max(?y:Nat,?x:Nat)),max(?x:Nat,max(?y:Nat,?x:Nat)))),min(min(?y:Nat,?x:Nat),min(max(?x:Nat,min(?z:Nat,?x:Nat)),?y:Nat)))"]

val problemsMaxMinNo =
IOMstrs.rdEqs [
"max(?x:Nat,?x:Nat) = min(?x:Nat,?y:Nat)",
"max(?x:Nat,?x:Nat) = min(?x:Nat,0)",
"max(?x:Nat,?x:Nat) = min(?y:Nat,?x:Nat)",
"max(?x:Nat,?x:Nat) = min(?z:Nat,?y:Nat)",
"max(?x:Nat,?x:Nat) = min(0,?y:Nat)",
"max(?x:Nat,?y:Nat) = min(?x:Nat,?y:Nat)",
"max(?x:Nat,?y:Nat) = min(?x:Nat,0)",
"max(?x:Nat,?y:Nat) = min(?y:Nat,?x:Nat)",
"max(?x:Nat,?y:Nat) = min(?y:Nat,?z:Nat)",
"max(?x:Nat,?y:Nat) = min(?y:Nat,0)",
"max(?x:Nat,?y:Nat) = min(0,?y:Nat)",
"max(?x:Nat,?z:Nat) = min(?x:Nat,?x:Nat)",
"max(?x:Nat,?z:Nat) = min(?y:Nat,0)",
"max(?x:Nat,0) = min(?x:Nat,?y:Nat)",
"max(?x:Nat,0) = min(?x:Nat,0)",
"max(?x:Nat,0) = min(?z:Nat,?x:Nat)",
"max(?x:Nat,0) = min(0,0)",
"max(?y:Nat,?x:Nat) = min(?x:Nat,?x:Nat)",
"max(?y:Nat,?x:Nat) = min(?y:Nat,?z:Nat)",
"max(?y:Nat,?x:Nat) = min(?z:Nat,?z:Nat)",
"max(?y:Nat,?y:Nat) = min(?x:Nat,0)",
"max(?y:Nat,?y:Nat) = min(0,?y:Nat)",
"max(?y:Nat,?z:Nat) = min(?x:Nat,?y:Nat)",
"max(?z:Nat,?x:Nat) = min(?y:Nat,?x:Nat)",
"max(?z:Nat,?z:Nat) = min(0,0)",
"max(?z:Nat,0) = min(?x:Nat,?y:Nat)",
"max(?z:Nat,0) = min(?y:Nat,?y:Nat)",
"max(?z:Nat,0) = min(?y:Nat,0)",
"max(?z:Nat,0) = min(0,?y:Nat)",
"max(0,?x:Nat) = min(?x:Nat,?y:Nat)",
"max(0,?x:Nat) = min(?y:Nat,?x:Nat)",
"max(0,?x:Nat) = min(?z:Nat,?z:Nat)",
"max(0,?y:Nat) = min(?x:Nat,0)",
"max(0,?y:Nat) = min(0,?x:Nat)",
"max(0,?y:Nat) = min(0,?y:Nat)",
"min(?x:Nat,?x:Nat) = max(?y:Nat,?y:Nat)",
"min(?y:Nat,?z:Nat) = max(0,?x:Nat)",
"min(?y:Nat,?z:Nat) = max(0,0)",
"min(?y:Nat,0) = max(0,?y:Nat)",
"min(0,?x:Nat) = max(?x:Nat,?y:Nat)",
"min(0,?z:Nat) = max(?y:Nat,?x:Nat)",
"max(?x:Nat,?x:Nat) = max(min(?x:Nat,?x:Nat),?y:Nat)",
"max(?x:Nat,?x:Nat) = min(?y:Nat,min(?x:Nat,0))",
"max(?x:Nat,?x:Nat) = min(?z:Nat,min(?y:Nat,?z:Nat))",
"max(?x:Nat,?y:Nat) = max(min(?y:Nat,?x:Nat),0)",
"max(?x:Nat,?y:Nat) = min(?y:Nat,max(?x:Nat,0))",
"max(?x:Nat,?y:Nat) = min(max(?z:Nat,0),?x:Nat)",
"max(?x:Nat,?y:Nat) = min(max(0,?x:Nat),?z:Nat)",
"max(?x:Nat,0) = max(?x:Nat,min(?y:Nat,?y:Nat))",
"max(?x:Nat,0) = min(?y:Nat,max(0,0))",
"max(?x:Nat,0) = min(0,max(?y:Nat,0))",
"max(?y:Nat,?x:Nat) = max(min(?y:Nat,0),?x:Nat)",
"max(?y:Nat,?x:Nat) = min(?x:Nat,min(?y:Nat,?y:Nat))",
"max(?y:Nat,?x:Nat) = min(?y:Nat,min(?x:Nat,?y:Nat))",
"max(?y:Nat,?x:Nat) = min(?y:Nat,min(?z:Nat,?z:Nat))",
"max(?y:Nat,?x:Nat) = min(?z:Nat,max(?z:Nat,?x:Nat))",
"max(?y:Nat,?x:Nat) = min(min(?x:Nat,?y:Nat),0)",
"max(?y:Nat,?y:Nat) = max(0,min(?x:Nat,?y:Nat))",
"max(?y:Nat,?y:Nat) = max(min(?z:Nat,?x:Nat),?y:Nat)",
"max(?y:Nat,?y:Nat) = min(0,max(?y:Nat,?y:Nat))",
"max(?y:Nat,?z:Nat) = min(?x:Nat,min(0,?x:Nat))",
"max(?y:Nat,?z:Nat) = min(max(0,?y:Nat),?z:Nat)",
"max(?y:Nat,0) = max(min(?x:Nat,?y:Nat),0)",
"max(?y:Nat,0) = min(?y:Nat,max(?x:Nat,0))",
"max(?z:Nat,?x:Nat) = min(?y:Nat,min(?y:Nat,?z:Nat))",
"max(?z:Nat,?y:Nat) = min(max(?x:Nat,0),?y:Nat)",
"max(?z:Nat,?y:Nat) = min(min(?x:Nat,?y:Nat),?z:Nat)",
"max(?z:Nat,?z:Nat) = max(min(?x:Nat,?y:Nat),0)",
"max(?z:Nat,?z:Nat) = max(min(?y:Nat,?y:Nat),?x:Nat)",
"max(?z:Nat,?z:Nat) = min(max(?y:Nat,?x:Nat),?z:Nat)",
"max(?z:Nat,0) = min(?y:Nat,max(?y:Nat,?x:Nat))",
"max(?z:Nat,0) = min(?y:Nat,max(0,?x:Nat))",
"max(?z:Nat,0) = min(0,max(?z:Nat,?x:Nat))",
"max(0,?x:Nat) = min(max(?y:Nat,?z:Nat),0)",
"max(0,?x:Nat) = min(min(?y:Nat,?z:Nat),?x:Nat)",
"max(0,?y:Nat) = min(?z:Nat,max(?x:Nat,?x:Nat))",
"max(0,?y:Nat) = min(min(?y:Nat,0),?z:Nat)",
"max(0,0) = min(max(?x:Nat,?x:Nat),?x:Nat)",
"max(0,0) = min(min(?x:Nat,?z:Nat),?x:Nat)",
"min(?x:Nat,?x:Nat) = max(?y:Nat,max(?x:Nat,?z:Nat))",
"min(?x:Nat,?x:Nat) = max(?z:Nat,min(?x:Nat,?y:Nat))",
"min(?x:Nat,?x:Nat) = max(0,min(?y:Nat,?x:Nat))",
"min(?x:Nat,?x:Nat) = min(max(?x:Nat,?x:Nat),?y:Nat)",
"min(?x:Nat,?x:Nat) = min(max(0,?x:Nat),?z:Nat)",
"min(?x:Nat,?y:Nat) = max(max(?x:Nat,?y:Nat),?y:Nat)",
"min(?x:Nat,?y:Nat) = max(min(?x:Nat,?x:Nat),?y:Nat)",
"min(?x:Nat,?y:Nat) = min(?y:Nat,max(?y:Nat,?y:Nat))",
"min(?x:Nat,?y:Nat) = min(max(?x:Nat,?x:Nat),?z:Nat)",
"min(?x:Nat,?y:Nat) = min(max(0,?z:Nat),?y:Nat)",
"min(?x:Nat,?z:Nat) = max(?y:Nat,min(?y:Nat,?z:Nat))",
"min(?x:Nat,?z:Nat) = max(?z:Nat,max(?y:Nat,?x:Nat))",
"min(?x:Nat,?z:Nat) = max(?z:Nat,min(0,?y:Nat))",
"min(?x:Nat,?z:Nat) = min(0,max(?y:Nat,?z:Nat))",
"min(?x:Nat,0) = max(min(?x:Nat,?x:Nat),?z:Nat)",
"min(?x:Nat,0) = max(min(?y:Nat,?y:Nat),?x:Nat)",
"min(?y:Nat,?x:Nat) = max(?y:Nat,min(?y:Nat,?x:Nat))",
"min(?y:Nat,?x:Nat) = max(max(?z:Nat,0),0)",
"min(?y:Nat,?x:Nat) = min(?y:Nat,max(?y:Nat,?x:Nat))",
"min(?y:Nat,?y:Nat) = max(?y:Nat,min(?x:Nat,?x:Nat))",
"min(?y:Nat,?y:Nat) = max(max(?x:Nat,?x:Nat),?y:Nat)",
"min(?y:Nat,?y:Nat) = max(max(?x:Nat,?z:Nat),?y:Nat)",
"min(?y:Nat,?y:Nat) = max(max(?y:Nat,?x:Nat),?x:Nat)",
"min(?y:Nat,?y:Nat) = min(max(?x:Nat,?y:Nat),?x:Nat)",
"min(?y:Nat,?z:Nat) = min(?x:Nat,max(?z:Nat,?z:Nat))",
"min(?y:Nat,0) = max(?x:Nat,max(0,?x:Nat))",
"min(?y:Nat,0) = min(?x:Nat,max(?z:Nat,?y:Nat))",
"min(?y:Nat,0) = min(?y:Nat,max(?y:Nat,?x:Nat))",
"min(?z:Nat,?x:Nat) = max(max(?z:Nat,?x:Nat),?y:Nat)",
"min(?z:Nat,?x:Nat) = min(max(?z:Nat,?y:Nat),?x:Nat)",
"min(?z:Nat,?y:Nat) = max(?x:Nat,max(?y:Nat,?y:Nat))",
"min(?z:Nat,?y:Nat) = max(?x:Nat,min(?x:Nat,0))",
"min(?z:Nat,?y:Nat) = max(max(0,0),?x:Nat)",
"min(?z:Nat,?z:Nat) = max(?x:Nat,min(?y:Nat,?x:Nat))",
"min(?z:Nat,?z:Nat) = max(0,max(?y:Nat,0))",
"min(?z:Nat,?z:Nat) = min(max(?y:Nat,?y:Nat),?x:Nat)",
"min(?z:Nat,0) = max(?z:Nat,min(?x:Nat,?x:Nat))",
"min(?z:Nat,0) = max(max(?y:Nat,?y:Nat),?x:Nat)",
"min(0,?x:Nat) = max(?y:Nat,max(?z:Nat,?x:Nat))",
"min(0,?x:Nat) = max(max(?x:Nat,?x:Nat),?y:Nat)",
"min(0,?x:Nat) = max(min(0,?x:Nat),?y:Nat)",
"min(0,?y:Nat) = max(0,min(?y:Nat,?x:Nat))",
"min(0,?y:Nat) = max(min(?y:Nat,?y:Nat),?x:Nat)",
"min(0,?y:Nat) = min(?x:Nat,max(?y:Nat,?x:Nat))",
"min(0,?y:Nat) = min(max(?z:Nat,0),?z:Nat)",
"min(0,0) = max(max(0,?z:Nat),?y:Nat)",
"min(0,0) = min(?x:Nat,max(?z:Nat,?x:Nat))",
"max(?x:Nat,max(?x:Nat,?y:Nat)) = min(?y:Nat,?z:Nat)",
"max(?x:Nat,max(?x:Nat,?z:Nat)) = min(?x:Nat,?x:Nat)",
"max(?x:Nat,max(0,?z:Nat)) = min(?y:Nat,?y:Nat)",
"max(?x:Nat,max(0,?z:Nat)) = min(0,?x:Nat)",
"max(?x:Nat,min(?y:Nat,?x:Nat)) = max(?y:Nat,0)",
"max(?x:Nat,min(?y:Nat,?y:Nat)) = min(?z:Nat,?x:Nat)",
"max(?y:Nat,max(?x:Nat,0)) = min(?y:Nat,?z:Nat)",
"max(?y:Nat,max(?y:Nat,?z:Nat)) = min(?y:Nat,?x:Nat)",
"max(?y:Nat,min(?y:Nat,?y:Nat)) = min(?z:Nat,?x:Nat)",
"max(?y:Nat,min(?y:Nat,?y:Nat)) = min(0,0)",
"max(?y:Nat,min(?y:Nat,0)) = max(?x:Nat,?x:Nat)",
"max(?z:Nat,min(?y:Nat,?x:Nat)) = min(?y:Nat,0)",
"max(?z:Nat,min(?z:Nat,?y:Nat)) = min(?y:Nat,?z:Nat)",
"max(?z:Nat,min(0,?z:Nat)) = min(?y:Nat,?x:Nat)",
"max(0,max(0,?y:Nat)) = min(?x:Nat,?y:Nat)",
"max(0,min(?x:Nat,?y:Nat)) = max(0,?x:Nat)",
"max(0,min(?y:Nat,?x:Nat)) = max(0,?x:Nat)",
"max(0,min(?y:Nat,?y:Nat)) = min(?x:Nat,?x:Nat)",
"max(max(?x:Nat,?x:Nat),?x:Nat) = min(?y:Nat,?y:Nat)",
"max(max(?x:Nat,?x:Nat),?y:Nat) = min(?x:Nat,?x:Nat)",
"max(max(?x:Nat,?y:Nat),?x:Nat) = min(?y:Nat,?x:Nat)",
"max(max(?x:Nat,?z:Nat),?x:Nat) = min(?z:Nat,0)",
"max(max(?y:Nat,?y:Nat),0) = min(?z:Nat,?z:Nat)",
"max(max(?z:Nat,?x:Nat),?x:Nat) = min(?y:Nat,0)",
"max(max(?z:Nat,?x:Nat),?x:Nat) = min(0,?y:Nat)",
"max(max(0,?x:Nat),?z:Nat) = min(?y:Nat,?y:Nat)",
"max(max(0,?z:Nat),?y:Nat) = min(?x:Nat,0)",
"max(min(?x:Nat,?x:Nat),?x:Nat) = max(?x:Nat,?z:Nat)",
"max(min(?x:Nat,?x:Nat),?x:Nat) = min(?y:Nat,?z:Nat)",
"max(min(?x:Nat,?x:Nat),0) = max(?y:Nat,?y:Nat)",
"max(min(?y:Nat,?x:Nat),?x:Nat) = max(0,?z:Nat)",
"max(min(?y:Nat,?x:Nat),?y:Nat) = min(?x:Nat,?x:Nat)",
"max(min(?y:Nat,?x:Nat),?y:Nat) = min(?x:Nat,0)",
"max(min(?y:Nat,?y:Nat),?y:Nat) = max(?z:Nat,?x:Nat)",
"min(?x:Nat,max(?x:Nat,?y:Nat)) = min(?z:Nat,?y:Nat)",
"min(?x:Nat,max(?x:Nat,?z:Nat)) = min(0,?x:Nat)",
"min(?x:Nat,max(?y:Nat,?x:Nat)) = min(?z:Nat,?y:Nat)",
"min(?x:Nat,max(?y:Nat,?y:Nat)) = max(?x:Nat,?x:Nat)",
"min(?x:Nat,max(?y:Nat,?y:Nat)) = max(?x:Nat,?y:Nat)",
"min(?x:Nat,max(?y:Nat,0)) = min(0,?y:Nat)",
"min(?x:Nat,max(0,0)) = min(?z:Nat,?z:Nat)",
"min(?y:Nat,max(?x:Nat,?y:Nat)) = min(?y:Nat,?x:Nat)",
"min(?y:Nat,max(?x:Nat,0)) = min(?z:Nat,?y:Nat)",
"min(?y:Nat,max(?y:Nat,?y:Nat)) = min(?x:Nat,?x:Nat)",
"min(?y:Nat,max(0,?x:Nat)) = min(0,?y:Nat)",
"min(?y:Nat,min(?x:Nat,?x:Nat)) = max(?z:Nat,?y:Nat)",
"min(?y:Nat,min(?y:Nat,?x:Nat)) = max(?z:Nat,?y:Nat)",
"min(?z:Nat,max(?y:Nat,?y:Nat)) = min(?y:Nat,0)",
"min(?z:Nat,max(?z:Nat,?y:Nat)) = min(?x:Nat,?x:Nat)",
"min(?z:Nat,max(0,?z:Nat)) = min(?z:Nat,?x:Nat)",
"min(?z:Nat,min(?x:Nat,?z:Nat)) = max(0,?y:Nat)",
"min(?z:Nat,min(0,?y:Nat)) = max(?y:Nat,?x:Nat)",
"min(?z:Nat,min(0,?z:Nat)) = max(?x:Nat,0)",
"min(0,max(?x:Nat,?x:Nat)) = min(?x:Nat,?z:Nat)",
"min(0,max(?y:Nat,?x:Nat)) = max(?x:Nat,?y:Nat)",
"min(0,max(?z:Nat,0)) = max(?y:Nat,?x:Nat)",
"min(0,max(0,?z:Nat)) = min(?z:Nat,?x:Nat)",
"min(max(?x:Nat,?x:Nat),?x:Nat) = max(?y:Nat,?y:Nat)",
"min(max(?x:Nat,?z:Nat),?y:Nat) = max(?z:Nat,?x:Nat)",
"min(max(?y:Nat,?x:Nat),?x:Nat) = min(0,?y:Nat)",
"min(max(?y:Nat,?y:Nat),?y:Nat) = min(?x:Nat,?x:Nat)",
"min(max(?y:Nat,?y:Nat),?y:Nat) = min(?x:Nat,0)",
"min(max(?y:Nat,0),?y:Nat) = max(?x:Nat,?y:Nat)",
"min(max(?y:Nat,0),0) = max(?x:Nat,?y:Nat)",
"min(max(?z:Nat,?x:Nat),?x:Nat) = min(0,?y:Nat)",
"min(max(?z:Nat,?x:Nat),?y:Nat) = min(?y:Nat,?x:Nat)",
"min(max(?z:Nat,?y:Nat),?y:Nat) = max(?x:Nat,?y:Nat)",
"min(max(?z:Nat,?y:Nat),?y:Nat) = max(?z:Nat,?y:Nat)",
"min(max(?z:Nat,?y:Nat),0) = min(?z:Nat,?z:Nat)",
"min(max(?z:Nat,?z:Nat),?x:Nat) = min(0,?z:Nat)",
"min(max(?z:Nat,0),0) = max(?x:Nat,?y:Nat)",
"min(max(?z:Nat,0),0) = min(?y:Nat,?x:Nat)",
"min(min(?x:Nat,?x:Nat),?y:Nat) = max(?y:Nat,?x:Nat)",
"min(min(?y:Nat,?x:Nat),?z:Nat) = max(?x:Nat,?z:Nat)",
"min(min(0,?y:Nat),?y:Nat) = max(?y:Nat,?y:Nat)",
"min(min(0,?z:Nat),?x:Nat) = max(?x:Nat,0)",
"max(?x:Nat,?x:Nat) = max(max(?x:Nat,min(?z:Nat,?y:Nat)),?x:Nat)",
"max(?x:Nat,?x:Nat) = min(?x:Nat,max(?y:Nat,max(?y:Nat,?z:Nat)))",
"max(?x:Nat,?x:Nat) = min(?y:Nat,max(?y:Nat,min(?y:Nat,?z:Nat)))",
"max(?x:Nat,?x:Nat) = min(max(?y:Nat,max(?z:Nat,?z:Nat)),?x:Nat)",
"max(?x:Nat,?y:Nat) = max(?x:Nat,min(min(?y:Nat,?y:Nat),?x:Nat))",
"max(?x:Nat,?y:Nat) = max(?y:Nat,min(min(0,?y:Nat),?z:Nat))",
"max(?x:Nat,?y:Nat) = max(min(?x:Nat,max(?x:Nat,?x:Nat)),?z:Nat)",
"max(?x:Nat,?y:Nat) = min(?x:Nat,min(max(?x:Nat,0),?x:Nat))",
"max(?x:Nat,?y:Nat) = min(?y:Nat,min(min(?x:Nat,0),?x:Nat))",
"max(?x:Nat,?y:Nat) = min(min(0,?y:Nat),min(?z:Nat,0))",
"max(?x:Nat,?y:Nat) = min(min(max(?x:Nat,?x:Nat),?x:Nat),?y:Nat)",
"max(?x:Nat,?y:Nat) = min(min(min(?y:Nat,?y:Nat),0),?x:Nat)",
"max(?x:Nat,?z:Nat) = min(min(max(?x:Nat,?y:Nat),?z:Nat),?y:Nat)",
"max(?x:Nat,0) = max(max(0,?y:Nat),min(?x:Nat,?y:Nat))",
"max(?x:Nat,0) = min(?x:Nat,max(min(0,?y:Nat),0))",
"max(?x:Nat,0) = min(max(min(?z:Nat,?y:Nat),0),?y:Nat)",
"max(?x:Nat,0) = min(min(?y:Nat,max(?x:Nat,?y:Nat)),?z:Nat)",
"max(?y:Nat,?x:Nat) = max(min(0,?z:Nat),max(?z:Nat,?x:Nat))",
"max(?y:Nat,?x:Nat) = min(?x:Nat,max(max(?y:Nat,?y:Nat),?y:Nat))",
"max(?y:Nat,?x:Nat) = min(max(?y:Nat,?y:Nat),max(?z:Nat,?z:Nat))",
"max(?y:Nat,?x:Nat) = min(max(0,min(?y:Nat,?x:Nat)),0)",
"max(?y:Nat,?x:Nat) = min(max(min(?y:Nat,?x:Nat),?x:Nat),?y:Nat)",
"max(?y:Nat,?y:Nat) = max(min(max(?x:Nat,?y:Nat),0),?x:Nat)",
"max(?y:Nat,?y:Nat) = max(min(max(?y:Nat,?x:Nat),?x:Nat),0)",
"max(?y:Nat,?y:Nat) = min(?x:Nat,max(?y:Nat,min(?x:Nat,?x:Nat)))",
"max(?y:Nat,?z:Nat) = max(0,max(0,min(0,?y:Nat)))",
"max(?y:Nat,?z:Nat) = max(min(?y:Nat,min(?x:Nat,?z:Nat)),?x:Nat)",
"max(?y:Nat,?z:Nat) = max(min(?z:Nat,?z:Nat),max(?x:Nat,?x:Nat))",
"max(?y:Nat,?z:Nat) = min(max(max(0,?x:Nat),0),?x:Nat)",
"max(?y:Nat,?z:Nat) = min(min(0,?y:Nat),min(?x:Nat,?y:Nat))",
"max(?y:Nat,0) = min(?z:Nat,min(?x:Nat,min(?z:Nat,?x:Nat)))",
"max(?z:Nat,?x:Nat) = min(max(?x:Nat,?z:Nat),min(?z:Nat,?y:Nat))",
"max(?z:Nat,?y:Nat) = max(?x:Nat,max(?y:Nat,min(0,0)))",
"max(?z:Nat,?z:Nat) = min(min(min(0,?y:Nat),?y:Nat),0)",
"max(?z:Nat,0) = max(min(min(?x:Nat,?y:Nat),?x:Nat),?z:Nat)",
"max(0,?y:Nat) = max(min(?x:Nat,?x:Nat),min(?y:Nat,0))",
"max(0,?y:Nat) = min(0,max(?x:Nat,min(?z:Nat,?y:Nat)))",
"max(0,?y:Nat) = min(max(max(?x:Nat,0),?x:Nat),?z:Nat)",
"max(0,?z:Nat) = min(min(?x:Nat,min(?z:Nat,0)),?y:Nat)",
"max(0,0) = max(?x:Nat,min(max(0,?x:Nat),0))",
"min(?x:Nat,?x:Nat) = max(?y:Nat,min(?x:Nat,min(?x:Nat,?y:Nat)))",
"min(?x:Nat,?x:Nat) = max(max(?y:Nat,?y:Nat),min(?y:Nat,?y:Nat))",
"min(?x:Nat,?x:Nat) = max(max(?z:Nat,?x:Nat),min(?x:Nat,?z:Nat))",
"min(?x:Nat,?y:Nat) = max(?y:Nat,max(?x:Nat,max(?z:Nat,?z:Nat)))",
"min(?x:Nat,?y:Nat) = max(?z:Nat,min(max(?x:Nat,?y:Nat),?y:Nat))",
"min(?x:Nat,?y:Nat) = min(?x:Nat,min(max(?x:Nat,?x:Nat),0))",
"min(?x:Nat,?y:Nat) = min(0,max(min(?y:Nat,?y:Nat),0))",
"min(?x:Nat,?z:Nat) = min(?y:Nat,max(?z:Nat,max(?y:Nat,?y:Nat)))",
"min(?x:Nat,?z:Nat) = min(min(?y:Nat,?z:Nat),max(?y:Nat,?y:Nat))",
"min(?x:Nat,0) = min(?x:Nat,max(max(?x:Nat,?y:Nat),?x:Nat))",
"min(?y:Nat,?x:Nat) = max(max(max(?x:Nat,?y:Nat),?z:Nat),0)",
"min(?y:Nat,?x:Nat) = min(?y:Nat,max(min(?z:Nat,?y:Nat),0))",
"min(?y:Nat,?y:Nat) = max(?y:Nat,max(max(?x:Nat,?x:Nat),?x:Nat))",
"min(?y:Nat,?y:Nat) = max(min(?x:Nat,?x:Nat),max(0,?z:Nat))",
"min(?y:Nat,?y:Nat) = min(max(min(0,?x:Nat),0),?x:Nat)",
"min(?y:Nat,?z:Nat) = max(min(?x:Nat,?y:Nat),max(?z:Nat,?y:Nat))",
"min(?y:Nat,?z:Nat) = min(max(?x:Nat,?y:Nat),min(?x:Nat,0))",
"min(?y:Nat,?z:Nat) = min(max(0,max(?x:Nat,?z:Nat)),?y:Nat)",
"min(?z:Nat,?x:Nat) = max(?x:Nat,max(?y:Nat,min(?y:Nat,?x:Nat)))",
"min(?z:Nat,?x:Nat) = max(max(max(?y:Nat,?y:Nat),?x:Nat),0)",
"min(?z:Nat,?y:Nat) = max(max(min(0,?y:Nat),?x:Nat),?x:Nat)",
"min(?z:Nat,?y:Nat) = min(max(max(0,?x:Nat),?x:Nat),?y:Nat)",
"min(0,?x:Nat) = max(?x:Nat,max(0,min(?x:Nat,0)))",
"min(0,?x:Nat) = max(?x:Nat,min(?z:Nat,max(0,?y:Nat)))",
"min(0,?x:Nat) = min(?x:Nat,max(min(?x:Nat,?y:Nat),?z:Nat))",
"min(0,?x:Nat) = min(min(max(?y:Nat,?y:Nat),?x:Nat),?x:Nat)",
"min(0,?x:Nat) = min(min(max(?z:Nat,?z:Nat),?z:Nat),?x:Nat)",
"min(0,?y:Nat) = max(min(min(?z:Nat,?x:Nat),?x:Nat),?x:Nat)",
"min(0,?y:Nat) = min(?x:Nat,min(?x:Nat,max(?y:Nat,?z:Nat)))",
"min(0,?y:Nat) = min(max(0,?x:Nat),max(?x:Nat,?y:Nat))",
"min(0,?z:Nat) = max(?x:Nat,max(?y:Nat,max(?x:Nat,?x:Nat)))",
"min(0,?z:Nat) = max(max(?x:Nat,?z:Nat),max(?y:Nat,0))",
"min(0,0) = max(?y:Nat,min(?y:Nat,min(?x:Nat,0)))",
"max(?x:Nat,max(?y:Nat,?z:Nat)) = min(min(?x:Nat,?z:Nat),?y:Nat)",
"max(?x:Nat,max(0,?x:Nat)) = max(?y:Nat,min(?y:Nat,?x:Nat))",
"max(?y:Nat,max(?y:Nat,?y:Nat)) = max(min(?x:Nat,0),0)",
"max(?y:Nat,max(?z:Nat,0)) = max(0,min(0,?x:Nat))",
"max(?y:Nat,min(?x:Nat,?y:Nat)) = max(?x:Nat,max(?y:Nat,?z:Nat))",
"max(?y:Nat,min(?z:Nat,0)) = min(?y:Nat,min(?y:Nat,0))",
"max(?z:Nat,max(0,?y:Nat)) = min(?y:Nat,min(?x:Nat,?y:Nat))",
"max(0,min(?y:Nat,?y:Nat)) = min(min(?z:Nat,?y:Nat),?z:Nat)",
"max(max(?x:Nat,?x:Nat),?y:Nat) = max(?y:Nat,min(0,?y:Nat))",
"max(max(?x:Nat,?y:Nat),0) = min(max(?z:Nat,?y:Nat),0)",
"max(max(?x:Nat,?z:Nat),?z:Nat) = min(max(?y:Nat,?y:Nat),?y:Nat)",
"max(max(?y:Nat,?x:Nat),?x:Nat) = min(min(?x:Nat,0),?y:Nat)",
"max(max(?y:Nat,?y:Nat),?y:Nat) = min(?y:Nat,min(?x:Nat,?z:Nat))",
"max(max(?y:Nat,0),?y:Nat) = max(min(?z:Nat,?y:Nat),?x:Nat)",
"max(max(?y:Nat,0),?z:Nat) = min(?x:Nat,min(?y:Nat,?z:Nat))",
"max(max(?z:Nat,?x:Nat),?x:Nat) = min(?x:Nat,max(?x:Nat,?x:Nat))",
"max(min(?x:Nat,?y:Nat),?z:Nat) = min(?y:Nat,max(?y:Nat,?x:Nat))",
"max(min(?y:Nat,?x:Nat),0) = max(max(?x:Nat,?x:Nat),?y:Nat)",
"max(min(?y:Nat,?x:Nat),0) = min(?y:Nat,max(?z:Nat,0))",
"max(min(?y:Nat,?y:Nat),?z:Nat) = max(max(0,?x:Nat),?x:Nat)",
"max(min(?y:Nat,0),?z:Nat) = max(max(?x:Nat,?y:Nat),0)",
"max(min(0,0),?x:Nat) = max(min(?x:Nat,?x:Nat),?y:Nat)",
"min(?x:Nat,max(?x:Nat,?y:Nat)) = min(0,max(?z:Nat,0))",
"min(?x:Nat,min(?y:Nat,?x:Nat)) = max(max(?x:Nat,?x:Nat),?x:Nat)",
"min(?y:Nat,max(0,?x:Nat)) = max(?y:Nat,min(?x:Nat,?y:Nat))",
"min(?y:Nat,min(?z:Nat,?x:Nat)) = max(max(?z:Nat,?x:Nat),?x:Nat)",
"min(?y:Nat,min(?z:Nat,?y:Nat)) = max(min(?x:Nat,?y:Nat),?x:Nat)",
"min(0,max(?x:Nat,?y:Nat)) = min(min(?x:Nat,?y:Nat),?y:Nat)",
"min(0,max(?y:Nat,?x:Nat)) = max(0,max(?z:Nat,?x:Nat))",
"min(0,max(?z:Nat,?x:Nat)) = min(?z:Nat,max(?y:Nat,0))",
"min(0,min(?x:Nat,?z:Nat)) = min(max(?y:Nat,?z:Nat),?z:Nat)",
"min(0,min(?y:Nat,?z:Nat)) = min(?x:Nat,max(?y:Nat,?y:Nat))",
"min(max(?x:Nat,?x:Nat),0) = max(?z:Nat,min(0,?y:Nat))",
"min(max(?x:Nat,?y:Nat),?x:Nat) = max(?z:Nat,max(?y:Nat,?y:Nat))",
"min(max(?x:Nat,?y:Nat),?x:Nat) = min(min(?x:Nat,?y:Nat),0)",
"min(max(?x:Nat,?y:Nat),?y:Nat) = max(min(?y:Nat,?y:Nat),?x:Nat)",
"min(max(?x:Nat,?y:Nat),0) = max(?x:Nat,max(?y:Nat,?x:Nat))",
"min(max(?x:Nat,?z:Nat),?y:Nat) = max(?y:Nat,max(?x:Nat,?x:Nat))",
"min(max(?y:Nat,0),?y:Nat) = min(max(?x:Nat,?x:Nat),?y:Nat)",
"min(max(?y:Nat,0),?y:Nat) = min(max(?y:Nat,?y:Nat),0)",
"min(max(?z:Nat,?x:Nat),?x:Nat) = max(max(?y:Nat,?y:Nat),?y:Nat)",
"min(min(?y:Nat,?x:Nat),?y:Nat) = max(?y:Nat,min(?x:Nat,?y:Nat))",
"min(min(?y:Nat,?x:Nat),?z:Nat) = max(min(0,?x:Nat),?y:Nat)",
"min(min(?y:Nat,?y:Nat),0) = max(?x:Nat,max(?z:Nat,?y:Nat))",
"min(min(?y:Nat,?z:Nat),?z:Nat) = min(?x:Nat,max(?y:Nat,0))",
"min(min(?z:Nat,?x:Nat),?x:Nat) = max(max(?x:Nat,?x:Nat),?z:Nat)",
"min(min(?z:Nat,0),?z:Nat) = max(max(?y:Nat,?z:Nat),?x:Nat)",
"min(min(0,?x:Nat),?x:Nat) = max(max(0,?y:Nat),?z:Nat)",
"min(min(0,0),?x:Nat) = min(max(?x:Nat,?y:Nat),?x:Nat)",
"max(?x:Nat,max(?z:Nat,min(?x:Nat,?x:Nat))) = max(?y:Nat,?x:Nat)",
"max(?x:Nat,min(?y:Nat,min(?y:Nat,?x:Nat))) = min(?z:Nat,?x:Nat)",
"max(?y:Nat,max(?x:Nat,min(?z:Nat,0))) = max(?z:Nat,?y:Nat)",
"max(?y:Nat,max(?z:Nat,min(?z:Nat,0))) = min(?x:Nat,?y:Nat)",
"max(?y:Nat,max(min(?z:Nat,?y:Nat),?y:Nat)) = max(?y:Nat,?z:Nat)",
"max(?y:Nat,min(max(?x:Nat,?x:Nat),?y:Nat)) = max(?x:Nat,?y:Nat)",
"max(?z:Nat,max(0,max(0,?x:Nat))) = min(?x:Nat,?z:Nat)",
"max(0,min(min(?x:Nat,?x:Nat),?x:Nat)) = max(?x:Nat,?z:Nat)",
"max(0,min(min(0,?z:Nat),?y:Nat)) = max(0,?x:Nat)",
"max(max(?x:Nat,?x:Nat),max(0,?x:Nat)) = min(?y:Nat,?y:Nat)",
"max(max(?x:Nat,max(?y:Nat,?x:Nat)),?x:Nat) = min(?y:Nat,?y:Nat)",
"max(max(?x:Nat,min(0,0)),?x:Nat) = min(0,0)",
"max(max(?y:Nat,?x:Nat),min(?y:Nat,?z:Nat)) = min(0,?y:Nat)",
"max(max(?y:Nat,max(?x:Nat,?y:Nat)),?x:Nat) = min(?x:Nat,?x:Nat)",
"max(max(?z:Nat,?x:Nat),max(?z:Nat,?z:Nat)) = min(0,0)",
"max(max(?z:Nat,0),max(?z:Nat,?x:Nat)) = min(0,?y:Nat)",
"max(max(?z:Nat,0),max(0,?y:Nat)) = min(?x:Nat,0)",
"max(max(0,max(?y:Nat,0)),?x:Nat) = min(0,?y:Nat)",
"max(max(max(?x:Nat,?y:Nat),0),?y:Nat) = min(?x:Nat,?z:Nat)",
"max(max(max(?y:Nat,0),?x:Nat),0) = min(?z:Nat,?y:Nat)",
"max(max(min(?z:Nat,?y:Nat),?z:Nat),?x:Nat) = min(?y:Nat,?y:Nat)",
"max(max(min(?z:Nat,0),?y:Nat),?x:Nat) = min(?x:Nat,?x:Nat)",
"max(max(min(0,?y:Nat),?x:Nat),?x:Nat) = max(0,?y:Nat)",
"max(min(?x:Nat,min(?y:Nat,?z:Nat)),?z:Nat) = min(?x:Nat,?y:Nat)",
"max(min(?y:Nat,?x:Nat),max(?x:Nat,?x:Nat)) = min(?x:Nat,?z:Nat)",
"max(min(max(?z:Nat,?x:Nat),?y:Nat),?x:Nat) = max(?x:Nat,?x:Nat)",
"max(min(min(?y:Nat,0),?y:Nat),?x:Nat) = max(?y:Nat,?x:Nat)",
"min(?x:Nat,max(0,max(?x:Nat,?x:Nat))) = max(?x:Nat,?y:Nat)",
"min(?x:Nat,max(max(0,?y:Nat),?z:Nat)) = max(?x:Nat,?x:Nat)",
"min(?x:Nat,min(?y:Nat,max(?z:Nat,?y:Nat))) = min(?z:Nat,?y:Nat)",
"min(?x:Nat,min(max(?x:Nat,0),?y:Nat)) = max(?y:Nat,?y:Nat)",
"min(?y:Nat,max(?y:Nat,max(?z:Nat,?y:Nat))) = max(0,0)",
"min(?y:Nat,max(max(?y:Nat,?z:Nat),?x:Nat)) = min(?y:Nat,?z:Nat)",
"min(?z:Nat,max(max(?x:Nat,?z:Nat),?x:Nat)) = max(?y:Nat,?z:Nat)",
"min(?z:Nat,max(min(?y:Nat,?y:Nat),?y:Nat)) = max(?x:Nat,?x:Nat)",
"min(?z:Nat,min(?y:Nat,max(?x:Nat,?x:Nat))) = min(?y:Nat,?x:Nat)",
"min(?z:Nat,min(?y:Nat,max(?y:Nat,?y:Nat))) = max(?x:Nat,?y:Nat)",
"min(0,max(max(?y:Nat,0),?y:Nat)) = min(?y:Nat,?z:Nat)",
"min(max(?x:Nat,?x:Nat),min(?z:Nat,?x:Nat)) = min(?y:Nat,?x:Nat)",
"min(max(?y:Nat,min(?x:Nat,?x:Nat)),?y:Nat) = max(?x:Nat,?y:Nat)",
"min(max(?y:Nat,min(0,?y:Nat)),?y:Nat) = min(?x:Nat,?y:Nat)",
"min(max(max(?x:Nat,0),?y:Nat),?x:Nat) = min(?y:Nat,?y:Nat)",
"min(max(max(?x:Nat,0),?y:Nat),?y:Nat) = min(?y:Nat,?z:Nat)",
"min(min(?x:Nat,max(0,?y:Nat)),?y:Nat) = max(?y:Nat,?y:Nat)",
"min(min(?y:Nat,0),max(?x:Nat,?y:Nat)) = min(?x:Nat,?x:Nat)",
"min(min(?z:Nat,?y:Nat),max(0,?x:Nat)) = min(?x:Nat,0)",
"min(min(?z:Nat,min(?y:Nat,?x:Nat)),?x:Nat) = max(0,0)",
"min(min(0,max(?z:Nat,0)),?x:Nat) = max(?z:Nat,?y:Nat)",
"min(min(max(?x:Nat,0),?z:Nat),?y:Nat) = max(?y:Nat,?y:Nat)",
"min(min(max(?y:Nat,0),0),?z:Nat) = max(?x:Nat,?y:Nat)",
"min(min(max(?y:Nat,0),0),?z:Nat) = max(0,?x:Nat)",
"min(min(min(?x:Nat,?x:Nat),0),?x:Nat) = max(?y:Nat,?y:Nat)",
"min(min(min(?z:Nat,?y:Nat),?y:Nat),?z:Nat) = max(?x:Nat,?z:Nat)",
"max(?x:Nat,?x:Nat) = max(max(max(?x:Nat,?z:Nat),min(?x:Nat,0)),?y:Nat)",
"max(?x:Nat,?x:Nat) = max(max(min(?z:Nat,?z:Nat),max(?y:Nat,?y:Nat)),?x:Nat)",
"max(?x:Nat,?x:Nat) = min(min(?z:Nat,max(min(0,0),?y:Nat)),?z:Nat)",
"max(?x:Nat,?y:Nat) = min(?y:Nat,max(max(min(?x:Nat,?y:Nat),0),?y:Nat))",
"max(?x:Nat,?y:Nat) = min(?y:Nat,min(min(max(?y:Nat,?x:Nat),0),?y:Nat))",
"max(?x:Nat,?z:Nat) = max(min(max(0,0),max(?z:Nat,?x:Nat)),?x:Nat)",
"max(?x:Nat,0) = max(max(?x:Nat,min(max(?y:Nat,0),?z:Nat)),?y:Nat)",
"max(?y:Nat,?x:Nat) = max(min(?y:Nat,max(?z:Nat,0)),max(?y:Nat,?z:Nat))",
"max(?y:Nat,?x:Nat) = max(min(max(min(?z:Nat,?y:Nat),0),?y:Nat),0)",
"max(?y:Nat,?x:Nat) = min(min(?z:Nat,?y:Nat),min(min(?z:Nat,?z:Nat),?x:Nat))",
"max(?y:Nat,?y:Nat) = min(?x:Nat,min(min(0,min(?z:Nat,?z:Nat)),?x:Nat))",
"max(?y:Nat,?y:Nat) = min(0,max(?x:Nat,min(max(?y:Nat,?y:Nat),0)))",
"max(?y:Nat,?y:Nat) = min(min(max(?x:Nat,?x:Nat),?x:Nat),min(?x:Nat,?y:Nat))",
"max(?y:Nat,0) = min(max(max(0,?x:Nat),?y:Nat),min(0,?y:Nat))",
"max(?z:Nat,?x:Nat) = max(max(?z:Nat,min(?x:Nat,?x:Nat)),max(0,?y:Nat))",
"max(?z:Nat,?y:Nat) = max(?x:Nat,min(max(?y:Nat,max(?y:Nat,?y:Nat)),?z:Nat))",
"max(?z:Nat,?y:Nat) = max(min(min(0,?y:Nat),min(?y:Nat,?z:Nat)),?y:Nat)",
"max(?z:Nat,?y:Nat) = min(?y:Nat,min(?x:Nat,min(?x:Nat,max(?x:Nat,?y:Nat))))",
"max(?z:Nat,?z:Nat) = max(?x:Nat,max(?y:Nat,min(min(0,?x:Nat),?x:Nat)))",
"max(?z:Nat,?z:Nat) = min(0,min(max(max(?x:Nat,?x:Nat),?y:Nat),?y:Nat))",
"max(?z:Nat,0) = min(max(max(?x:Nat,?x:Nat),max(0,?x:Nat)),?y:Nat)",
"max(0,?x:Nat) = min(?y:Nat,max(min(?y:Nat,?x:Nat),max(?y:Nat,?x:Nat)))",
"max(0,?y:Nat) = min(?z:Nat,max(min(max(0,?x:Nat),0),?y:Nat))",
"max(0,?y:Nat) = min(max(min(?z:Nat,?y:Nat),min(?y:Nat,0)),?y:Nat)",
"max(0,?y:Nat) = min(min(min(min(?x:Nat,0),?z:Nat),?z:Nat),0)",
"max(0,?z:Nat) = min(max(max(min(?y:Nat,?z:Nat),?y:Nat),?z:Nat),0)",
"min(?x:Nat,?x:Nat) = min(max(?x:Nat,?x:Nat),min(0,max(?x:Nat,?x:Nat)))",
"min(?x:Nat,?x:Nat) = min(min(max(max(?x:Nat,?z:Nat),?x:Nat),?z:Nat),?x:Nat)",
"min(?x:Nat,?y:Nat) = max(max(max(?y:Nat,?y:Nat),min(?x:Nat,?z:Nat)),?x:Nat)",
"min(?x:Nat,?z:Nat) = max(?y:Nat,min(min(?z:Nat,max(?x:Nat,?y:Nat)),?x:Nat))",
"min(?x:Nat,0) = max(?y:Nat,max(0,max(?z:Nat,max(?z:Nat,?x:Nat))))",
"min(?x:Nat,0) = min(min(max(?z:Nat,0),?z:Nat),max(?y:Nat,0))",
"min(?y:Nat,?x:Nat) = min(?y:Nat,max(?z:Nat,min(?z:Nat,min(?y:Nat,0))))",
"min(?y:Nat,?x:Nat) = min(?z:Nat,min(max(?z:Nat,?z:Nat),max(0,?x:Nat)))",
"min(?y:Nat,?x:Nat) = min(max(?x:Nat,?z:Nat),max(?x:Nat,max(?y:Nat,?y:Nat)))",
"min(?y:Nat,?y:Nat) = max(0,min(min(?y:Nat,?x:Nat),max(?x:Nat,?z:Nat)))",
"min(?y:Nat,?y:Nat) = min(max(?z:Nat,min(0,min(?y:Nat,?y:Nat))),?y:Nat)",
"min(?y:Nat,?y:Nat) = min(min(?z:Nat,?y:Nat),min(max(?y:Nat,?y:Nat),?y:Nat))",
"min(?y:Nat,?z:Nat) = min(max(?x:Nat,?y:Nat),min(?y:Nat,max(?z:Nat,?x:Nat)))",
"min(?y:Nat,0) = max(?y:Nat,min(?y:Nat,min(min(?y:Nat,?y:Nat),?y:Nat)))",
"min(?y:Nat,0) = max(max(?y:Nat,min(?x:Nat,min(?y:Nat,?x:Nat))),0)",
"min(?y:Nat,0) = min(?y:Nat,max(?y:Nat,min(max(?x:Nat,?x:Nat),?x:Nat)))",
"min(?z:Nat,?x:Nat) = max(?y:Nat,max(?y:Nat,max(min(?z:Nat,?y:Nat),?y:Nat)))",
"min(?z:Nat,?x:Nat) = max(?z:Nat,max(max(0,?x:Nat),min(?y:Nat,0)))",
"min(?z:Nat,?x:Nat) = max(min(max(min(?y:Nat,0),0),0),?x:Nat)",
"min(?z:Nat,?x:Nat) = min(?y:Nat,min(max(min(?x:Nat,0),0),?x:Nat))",
"min(?z:Nat,?y:Nat) = min(0,max(max(max(?y:Nat,?x:Nat),?x:Nat),?y:Nat))",
"min(?z:Nat,?y:Nat) = min(max(?y:Nat,0),min(min(?y:Nat,?y:Nat),?x:Nat))",
"min(?z:Nat,0) = min(max(max(?z:Nat,min(?x:Nat,?y:Nat)),?y:Nat),?x:Nat)",
"min(0,?x:Nat) = max(?y:Nat,max(?y:Nat,min(max(?y:Nat,?z:Nat),?y:Nat)))",
"min(0,?x:Nat) = min(max(?z:Nat,min(?x:Nat,min(?x:Nat,0))),?z:Nat)",
"min(0,?y:Nat) = max(0,max(min(?x:Nat,max(?x:Nat,0)),?y:Nat))",
"min(0,?y:Nat) = min(max(?x:Nat,0),max(max(0,?y:Nat),?z:Nat))",
"min(0,?y:Nat) = min(min(?y:Nat,min(?x:Nat,?x:Nat)),max(0,?y:Nat))",
"max(?x:Nat,min(?y:Nat,?y:Nat)) = max(?x:Nat,max(max(?z:Nat,?z:Nat),0))",
"max(?y:Nat,max(?y:Nat,?y:Nat)) = min(min(?y:Nat,?x:Nat),max(?z:Nat,?x:Nat))",
"max(?y:Nat,min(?x:Nat,?x:Nat)) = min(min(?y:Nat,?y:Nat),max(?y:Nat,?y:Nat))",
"max(0,min(?z:Nat,0)) = max(min(?x:Nat,min(?y:Nat,?z:Nat)),?x:Nat)",
"max(max(?x:Nat,0),?x:Nat) = min(min(min(?z:Nat,?x:Nat),?y:Nat),?y:Nat)",
"max(max(?y:Nat,?y:Nat),?z:Nat) = max(min(?x:Nat,?x:Nat),min(?y:Nat,?x:Nat))",
"max(max(?z:Nat,0),0) = max(?x:Nat,min(?y:Nat,max(?x:Nat,?z:Nat)))",
"max(max(0,?x:Nat),?y:Nat) = max(?y:Nat,min(?y:Nat,max(?x:Nat,?y:Nat)))",
"max(min(?y:Nat,0),?y:Nat) = max(max(?x:Nat,max(?y:Nat,?x:Nat)),?y:Nat)",
"max(min(?z:Nat,?y:Nat),?x:Nat) = min(0,min(max(?z:Nat,?x:Nat),?x:Nat))",
"max(min(?z:Nat,0),0) = min(max(?x:Nat,min(?y:Nat,?y:Nat)),?x:Nat)",
"min(?x:Nat,max(?y:Nat,?x:Nat)) = min(?x:Nat,max(min(?z:Nat,?y:Nat),?y:Nat))",
"min(?x:Nat,min(0,?x:Nat)) = min(?y:Nat,max(?x:Nat,min(?y:Nat,0)))",
"min(?y:Nat,min(?y:Nat,?x:Nat)) = max(max(?y:Nat,max(?z:Nat,0)),?z:Nat)",
"min(?y:Nat,min(?y:Nat,0)) = max(?x:Nat,min(?y:Nat,min(?z:Nat,?y:Nat)))",
"min(?y:Nat,min(?z:Nat,?z:Nat)) = max(?y:Nat,max(?x:Nat,min(?y:Nat,0)))",
"min(?z:Nat,max(?x:Nat,0)) = min(max(?y:Nat,?x:Nat),min(?y:Nat,?x:Nat))",
"min(0,max(?x:Nat,?y:Nat)) = max(min(max(?x:Nat,?x:Nat),0),?x:Nat)",
"min(max(?x:Nat,?x:Nat),?x:Nat) = min(?y:Nat,min(?y:Nat,max(?y:Nat,?y:Nat)))",
"min(max(?x:Nat,?x:Nat),?x:Nat) = min(?y:Nat,min(max(?z:Nat,?x:Nat),?z:Nat))",
"min(max(?x:Nat,?x:Nat),?y:Nat) = max(?x:Nat,min(min(?z:Nat,?y:Nat),?y:Nat))",
"min(max(?y:Nat,?x:Nat),0) = min(max(?y:Nat,max(?z:Nat,0)),?y:Nat)",
"min(max(0,?y:Nat),?y:Nat) = min(min(?y:Nat,max(?y:Nat,?y:Nat)),?x:Nat)",
"min(max(0,?y:Nat),0) = max(max(?y:Nat,?y:Nat),min(?x:Nat,?x:Nat))",
"min(min(?x:Nat,?x:Nat),?x:Nat) = max(min(?x:Nat,?x:Nat),max(?y:Nat,?y:Nat))",
"max(?x:Nat,max(max(?y:Nat,?x:Nat),?x:Nat)) = max(?x:Nat,min(?z:Nat,?z:Nat))",
"max(?y:Nat,min(min(?x:Nat,?y:Nat),0)) = min(min(?y:Nat,?y:Nat),?x:Nat)",
"max(?z:Nat,min(?y:Nat,min(?z:Nat,?z:Nat))) = max(min(?y:Nat,?y:Nat),?y:Nat)",
"max(0,max(max(?z:Nat,?y:Nat),?y:Nat)) = min(max(?z:Nat,?y:Nat),0)",
"max(max(?x:Nat,min(0,?x:Nat)),?x:Nat) = max(0,max(?y:Nat,?z:Nat))",
"max(max(?y:Nat,?y:Nat),max(?y:Nat,?x:Nat)) = min(0,max(?y:Nat,?x:Nat))",
"max(max(?z:Nat,?x:Nat),max(?x:Nat,?x:Nat)) = max(min(?y:Nat,?y:Nat),?x:Nat)",
"max(max(min(?x:Nat,?z:Nat),?z:Nat),?x:Nat) = min(max(?x:Nat,0),?x:Nat)",
"max(min(?y:Nat,0),min(?y:Nat,?z:Nat)) = max(min(?z:Nat,0),?y:Nat)",
"max(min(?y:Nat,min(?z:Nat,?x:Nat)),0) = max(?y:Nat,max(?y:Nat,?x:Nat))",
"max(min(?y:Nat,min(?z:Nat,?y:Nat)),?z:Nat) = min(0,max(0,?y:Nat))",
"max(min(?z:Nat,max(?y:Nat,?y:Nat)),0) = max(?x:Nat,max(?y:Nat,?x:Nat))",
"max(min(0,min(?x:Nat,?x:Nat)),0) = max(?x:Nat,min(?y:Nat,?x:Nat))",
"max(min(max(?y:Nat,?y:Nat),?x:Nat),?z:Nat) = max(max(?x:Nat,?z:Nat),?z:Nat)",
"max(min(max(?z:Nat,?x:Nat),?x:Nat),?y:Nat) = min(?y:Nat,max(?x:Nat,?y:Nat))",
"min(?x:Nat,max(?z:Nat,max(?x:Nat,?y:Nat))) = min(?z:Nat,min(?x:Nat,?z:Nat))",
"min(?x:Nat,min(?y:Nat,min(?x:Nat,?x:Nat))) = min(max(?x:Nat,?z:Nat),?x:Nat)",
"min(?x:Nat,min(0,max(?x:Nat,?x:Nat))) = max(max(?x:Nat,0),?y:Nat)",
"min(?y:Nat,max(?x:Nat,max(?x:Nat,?x:Nat))) = min(?x:Nat,min(0,?y:Nat))",
"min(?y:Nat,max(?z:Nat,max(?y:Nat,?x:Nat))) = max(?x:Nat,min(?z:Nat,?y:Nat))",
"min(?z:Nat,min(?y:Nat,min(0,?x:Nat))) = max(max(?x:Nat,?y:Nat),?y:Nat)",
"min(0,max(max(?x:Nat,?y:Nat),?x:Nat)) = max(?x:Nat,max(?z:Nat,?x:Nat))",
"min(0,min(?x:Nat,min(?x:Nat,?x:Nat))) = max(min(?x:Nat,?y:Nat),?z:Nat)",
"min(max(?x:Nat,min(?x:Nat,?x:Nat)),?x:Nat) = min(min(?x:Nat,?x:Nat),?y:Nat)",
"min(max(?z:Nat,?x:Nat),max(0,?x:Nat)) = min(max(?y:Nat,?x:Nat),?y:Nat)",
"min(max(?z:Nat,?y:Nat),max(?x:Nat,?x:Nat)) = max(max(?x:Nat,?x:Nat),?x:Nat)",
"min(max(?z:Nat,0),max(?x:Nat,0)) = min(min(?z:Nat,?x:Nat),?y:Nat)",
"min(max(?z:Nat,min(?y:Nat,?y:Nat)),?y:Nat) = min(?x:Nat,max(?x:Nat,?x:Nat))",
"min(max(max(?y:Nat,0),?y:Nat),?x:Nat) = min(?x:Nat,min(0,?x:Nat))",
"min(max(min(?x:Nat,?z:Nat),?y:Nat),?y:Nat) = min(max(0,0),?y:Nat)",
"min(max(min(?y:Nat,?x:Nat),?y:Nat),?x:Nat) = max(min(?y:Nat,?y:Nat),0)",
"min(max(min(?y:Nat,0),?x:Nat),?y:Nat) = min(min(?x:Nat,0),0)",
"min(min(?y:Nat,?x:Nat),max(?x:Nat,?z:Nat)) = max(?y:Nat,min(?z:Nat,?y:Nat))",
"min(min(?y:Nat,?y:Nat),min(?x:Nat,0)) = max(?x:Nat,max(?x:Nat,?x:Nat))",
"min(min(max(?y:Nat,0),?x:Nat),?y:Nat) = max(?y:Nat,max(?y:Nat,?x:Nat))",
"max(?x:Nat,min(max(min(?x:Nat,?y:Nat),?y:Nat),?y:Nat)) = max(?z:Nat,?y:Nat)",
"max(?x:Nat,min(min(max(0,0),?z:Nat),?z:Nat)) = min(0,?x:Nat)",
"max(?y:Nat,max(?x:Nat,min(max(0,?y:Nat),?x:Nat))) = max(?x:Nat,?z:Nat)",
"max(?y:Nat,min(min(min(?z:Nat,?x:Nat),?x:Nat),?y:Nat)) = max(?x:Nat,?y:Nat)",
"max(?z:Nat,max(max(min(?x:Nat,?z:Nat),?x:Nat),0)) = min(?x:Nat,0)",
"max(?z:Nat,max(min(?x:Nat,min(?y:Nat,?y:Nat)),?x:Nat)) = max(?z:Nat,?y:Nat)",
"max(0,max(max(max(?x:Nat,0),?z:Nat),?x:Nat)) = min(?x:Nat,?x:Nat)",
"max(max(?y:Nat,min(min(?y:Nat,?x:Nat),?x:Nat)),0) = min(?y:Nat,?z:Nat)",
"max(max(0,?y:Nat),max(min(?x:Nat,?y:Nat),?y:Nat)) = max(?z:Nat,?y:Nat)",
"max(max(0,min(?y:Nat,min(?y:Nat,?y:Nat))),0) = min(?x:Nat,0)",
"max(min(?x:Nat,?x:Nat),max(max(?y:Nat,?y:Nat),?x:Nat)) = min(0,?x:Nat)",
"max(min(?x:Nat,?y:Nat),min(?y:Nat,min(?y:Nat,?x:Nat))) = min(?y:Nat,?y:Nat)",
"max(min(?x:Nat,?z:Nat),min(min(?x:Nat,?y:Nat),?y:Nat)) = max(?x:Nat,?z:Nat)",
"max(min(?y:Nat,max(?y:Nat,min(?z:Nat,?x:Nat))),?x:Nat) = min(0,0)",
"max(min(0,min(max(?y:Nat,?z:Nat),?y:Nat)),?z:Nat) = min(?y:Nat,?y:Nat)",
"max(min(max(?y:Nat,?y:Nat),?x:Nat),min(0,?y:Nat)) = max(?z:Nat,?y:Nat)",
"max(min(max(?y:Nat,?z:Nat),?x:Nat),max(0,?z:Nat)) = max(?x:Nat,?x:Nat)",
"max(min(max(?y:Nat,max(?z:Nat,?y:Nat)),?y:Nat),?y:Nat) = max(?z:Nat,?y:Nat)",
"max(min(min(?x:Nat,min(?z:Nat,?y:Nat)),?y:Nat),0) = max(?x:Nat,?x:Nat)",
"min(?x:Nat,max(?y:Nat,min(max(?y:Nat,?y:Nat),?y:Nat))) = max(?z:Nat,?y:Nat)",
"min(?x:Nat,max(max(?z:Nat,?z:Nat),min(?z:Nat,?z:Nat))) = min(?z:Nat,0)",
"min(?x:Nat,min(max(?x:Nat,0),min(?y:Nat,?x:Nat))) = max(?z:Nat,?x:Nat)",
"min(?y:Nat,max(max(?y:Nat,min(?x:Nat,?x:Nat)),?x:Nat)) = min(?z:Nat,?y:Nat)",
"min(?y:Nat,max(max(max(?y:Nat,?y:Nat),0),?y:Nat)) = min(?x:Nat,?z:Nat)",
"min(?y:Nat,max(min(?x:Nat,?y:Nat),min(?y:Nat,?y:Nat))) = min(?y:Nat,?x:Nat)",
"min(?z:Nat,max(?x:Nat,max(min(?x:Nat,0),?z:Nat))) = min(?x:Nat,0)",
"min(?z:Nat,max(?y:Nat,min(?y:Nat,min(0,0)))) = min(0,?z:Nat)",
"min(?z:Nat,min(?x:Nat,max(?z:Nat,max(?x:Nat,?z:Nat)))) = min(?x:Nat,?y:Nat)",
"min(max(?x:Nat,?z:Nat),min(?x:Nat,max(?y:Nat,?z:Nat))) = max(?x:Nat,?x:Nat)",
"min(max(?x:Nat,0),max(max(?x:Nat,?y:Nat),?y:Nat)) = min(?y:Nat,?y:Nat)",
"min(max(?x:Nat,max(?x:Nat,?x:Nat)),min(?x:Nat,?x:Nat)) = max(?y:Nat,?x:Nat)",
"min(max(0,?x:Nat),max(min(?x:Nat,?y:Nat),0)) = max(0,?x:Nat)",
"min(max(0,?y:Nat),min(?y:Nat,max(?x:Nat,?x:Nat))) = max(?z:Nat,?x:Nat)",
"min(max(0,min(max(?x:Nat,?x:Nat),0)),?y:Nat) = min(?x:Nat,?x:Nat)",
"min(max(max(?z:Nat,?y:Nat),?x:Nat),min(?y:Nat,?z:Nat)) = max(?x:Nat,?y:Nat)",
"min(max(max(?z:Nat,?y:Nat),max(?x:Nat,0)),0) = min(?y:Nat,?y:Nat)",
"min(max(max(?z:Nat,min(?z:Nat,?x:Nat)),0),?y:Nat) = min(?y:Nat,?x:Nat)",
"min(max(max(min(?x:Nat,?z:Nat),?z:Nat),?y:Nat),0) = min(?x:Nat,?z:Nat)",
"min(max(min(?x:Nat,0),max(?x:Nat,?y:Nat)),?y:Nat) = max(?z:Nat,?y:Nat)",
"min(max(min(0,?y:Nat),?z:Nat),min(0,?z:Nat)) = min(?y:Nat,?y:Nat)",
"min(min(?x:Nat,max(?x:Nat,min(?x:Nat,?y:Nat))),?y:Nat) = max(0,?z:Nat)",
"min(min(?y:Nat,max(?y:Nat,max(?y:Nat,?z:Nat))),?y:Nat) = max(?z:Nat,?y:Nat)",
"min(min(?y:Nat,min(max(?x:Nat,?y:Nat),?x:Nat)),?y:Nat) = max(?y:Nat,?x:Nat)",
"min(min(?z:Nat,?y:Nat),max(max(?x:Nat,?y:Nat),?x:Nat)) = min(0,?y:Nat)",
"min(min(min(?x:Nat,?x:Nat),?x:Nat),max(?x:Nat,?x:Nat)) = min(0,?y:Nat)",
"min(min(min(?y:Nat,max(?x:Nat,?x:Nat)),?y:Nat),?y:Nat) = min(0,?y:Nat)",
"min(min(min(0,0),max(?y:Nat,0)),?x:Nat) = min(?x:Nat,?x:Nat)",
"max(?x:Nat,?x:Nat) = min(?y:Nat,max(?x:Nat,min(max(?x:Nat,min(?x:Nat,?x:Nat)),?x:Nat)))",
"max(?x:Nat,?x:Nat) = min(min(max(?y:Nat,max(?y:Nat,0)),max(?y:Nat,0)),?x:Nat)",
"max(?x:Nat,?y:Nat) = max(?x:Nat,min(?x:Nat,max(?x:Nat,max(max(?z:Nat,?z:Nat),0))))",
"max(?x:Nat,?y:Nat) = min(?y:Nat,min(?x:Nat,max(max(0,max(?y:Nat,?x:Nat)),?y:Nat)))",
"max(?x:Nat,?y:Nat) = min(0,max(max(?y:Nat,0),max(0,min(?x:Nat,0))))",
"max(?x:Nat,?y:Nat) = min(max(min(max(?y:Nat,max(?x:Nat,0)),?y:Nat),?x:Nat),?y:Nat)",
"max(?x:Nat,?y:Nat) = min(min(max(max(?x:Nat,min(0,0)),?z:Nat),?x:Nat),0)",
"max(?x:Nat,?y:Nat) = min(min(max(min(?y:Nat,max(?x:Nat,?x:Nat)),?x:Nat),?z:Nat),?y:Nat)",
"max(?x:Nat,?z:Nat) = min(max(?y:Nat,max(?y:Nat,max(?y:Nat,0))),max(?z:Nat,0))",
"max(?x:Nat,?z:Nat) = min(min(0,?x:Nat),min(?y:Nat,min(?x:Nat,max(?z:Nat,?x:Nat))))",
"max(?x:Nat,0) = max(?z:Nat,min(?y:Nat,max(max(min(?x:Nat,?z:Nat),?x:Nat),?y:Nat)))",
"max(?y:Nat,?x:Nat) = max(?x:Nat,max(?z:Nat,min(max(min(?y:Nat,?x:Nat),?y:Nat),?x:Nat)))",
"max(?y:Nat,?x:Nat) = max(min(max(min(?x:Nat,?y:Nat),?x:Nat),?y:Nat),min(?y:Nat,?z:Nat))",
"max(?y:Nat,?x:Nat) = min(max(?z:Nat,min(?x:Nat,max(min(0,?x:Nat),?y:Nat))),?x:Nat)",
"max(?y:Nat,?y:Nat) = max(max(?y:Nat,min(?y:Nat,?y:Nat)),max(min(?y:Nat,?y:Nat),?x:Nat))",
"max(?y:Nat,?y:Nat) = min(max(?y:Nat,max(max(max(0,?z:Nat),?x:Nat),?x:Nat)),0)",
"max(?y:Nat,0) = min(0,max(min(?x:Nat,?x:Nat),max(0,max(?x:Nat,?z:Nat))))",
"max(?z:Nat,?x:Nat) = min(min(max(?x:Nat,min(max(?y:Nat,?y:Nat),?x:Nat)),0),?x:Nat)",
"max(?z:Nat,?x:Nat) = min(min(max(0,max(max(?z:Nat,?y:Nat),0)),?y:Nat),?x:Nat)",
"max(?z:Nat,?x:Nat) = min(min(min(?z:Nat,0),max(?x:Nat,max(0,?y:Nat))),?z:Nat)",
"max(?z:Nat,?y:Nat) = max(max(min(?x:Nat,max(?x:Nat,?x:Nat)),?x:Nat),max(?y:Nat,?z:Nat))",
"max(?z:Nat,?z:Nat) = max(?x:Nat,min(?x:Nat,max(?x:Nat,min(max(?x:Nat,?x:Nat),0))))",
"max(?z:Nat,?z:Nat) = max(0,max(min(?z:Nat,max(min(?y:Nat,?y:Nat),?x:Nat)),?y:Nat))",
"max(0,?y:Nat) = max(max(?x:Nat,0),min(min(?x:Nat,min(?x:Nat,?z:Nat)),?y:Nat))",
"max(0,?y:Nat) = min(?y:Nat,min(0,min(0,max(min(?y:Nat,?y:Nat),?x:Nat))))",
"max(0,?y:Nat) = min(min(?x:Nat,max(?y:Nat,min(0,?y:Nat))),max(?z:Nat,?y:Nat))",
"max(0,?z:Nat) = min(?x:Nat,max(0,min(min(?y:Nat,0),min(?x:Nat,?y:Nat))))",
"max(0,0) = min(max(max(?x:Nat,min(max(?x:Nat,?x:Nat),?x:Nat)),?z:Nat),?z:Nat)",
"max(0,0) = min(min(max(0,min(max(?y:Nat,?z:Nat),?x:Nat)),?y:Nat),?y:Nat)",
"min(?x:Nat,?x:Nat) = max(max(min(?y:Nat,?x:Nat),?x:Nat),max(max(?x:Nat,?y:Nat),?z:Nat))",
"min(?x:Nat,?x:Nat) = max(max(min(?y:Nat,max(?y:Nat,?y:Nat)),?x:Nat),max(?y:Nat,?y:Nat))",
"min(?x:Nat,?y:Nat) = max(?y:Nat,min(max(?y:Nat,max(?y:Nat,?x:Nat)),min(?y:Nat,?x:Nat)))",
"min(?x:Nat,?y:Nat) = max(?z:Nat,max(max(?x:Nat,min(?x:Nat,max(0,?x:Nat))),?z:Nat))",
"min(?x:Nat,?y:Nat) = max(max(?x:Nat,max(?x:Nat,max(?x:Nat,min(?z:Nat,?y:Nat)))),?x:Nat)",
"min(?x:Nat,?y:Nat) = max(max(max(?z:Nat,?y:Nat),min(min(?x:Nat,?x:Nat),?x:Nat)),?x:Nat)",
"min(?x:Nat,?y:Nat) = min(?x:Nat,min(?z:Nat,min(max(?z:Nat,0),min(?x:Nat,0))))",
"min(?x:Nat,?y:Nat) = min(?z:Nat,max(min(?y:Nat,min(max(?z:Nat,?x:Nat),?y:Nat)),?z:Nat))",
"min(?x:Nat,?y:Nat) = min(?z:Nat,max(min(min(?x:Nat,max(?y:Nat,?z:Nat)),?x:Nat),?y:Nat))",
"min(?x:Nat,?y:Nat) = min(0,min(?y:Nat,min(?z:Nat,min(0,max(?z:Nat,?z:Nat)))))",
"min(?x:Nat,?y:Nat) = min(max(?y:Nat,min(?x:Nat,max(max(0,?z:Nat),0))),?y:Nat)",
"min(?x:Nat,?y:Nat) = min(max(0,min(?y:Nat,max(?y:Nat,min(?x:Nat,?x:Nat)))),?y:Nat)",
"min(?x:Nat,0) = max(max(?y:Nat,min(?x:Nat,max(max(?y:Nat,0),?y:Nat))),?x:Nat)",
"min(?x:Nat,0) = min(?x:Nat,max(max(max(?x:Nat,?y:Nat),?z:Nat),max(0,?x:Nat)))",
"min(?y:Nat,?x:Nat) = max(?y:Nat,max(min(min(min(?y:Nat,?x:Nat),0),?y:Nat),?x:Nat))",
"min(?y:Nat,?x:Nat) = max(0,min(?x:Nat,min(min(?y:Nat,min(?x:Nat,?y:Nat)),?z:Nat)))",
"min(?y:Nat,?x:Nat) = min(?z:Nat,min(?x:Nat,max(min(?y:Nat,min(?x:Nat,?x:Nat)),0)))",
"min(?y:Nat,?x:Nat) = min(max(?x:Nat,max(?x:Nat,?y:Nat)),max(max(?y:Nat,?z:Nat),?y:Nat))",
"min(?y:Nat,?y:Nat) = max(0,max(?x:Nat,min(min(min(?x:Nat,?z:Nat),?y:Nat),?x:Nat)))",
"min(?y:Nat,?y:Nat) = max(min(max(?y:Nat,min(0,max(?y:Nat,?z:Nat))),?x:Nat),?x:Nat)",
"min(?y:Nat,?y:Nat) = min(min(min(max(0,min(?x:Nat,?x:Nat)),?z:Nat),?y:Nat),0)",
"min(?y:Nat,?z:Nat) = min(?z:Nat,max(min(?y:Nat,min(?z:Nat,?x:Nat)),max(?z:Nat,?y:Nat)))",
"min(?y:Nat,0) = max(min(?y:Nat,min(?z:Nat,max(max(?z:Nat,?z:Nat),?y:Nat))),?z:Nat)",
"min(?y:Nat,0) = min(max(?y:Nat,min(max(?x:Nat,?y:Nat),?z:Nat)),max(?y:Nat,?y:Nat))",
"min(?z:Nat,?x:Nat) = max(?x:Nat,max(max(?x:Nat,max(?x:Nat,?x:Nat)),max(?z:Nat,?x:Nat)))",
"min(?z:Nat,?x:Nat) = max(min(max(?y:Nat,min(0,?z:Nat)),?x:Nat),min(?z:Nat,?y:Nat))",
"min(?z:Nat,?x:Nat) = min(?y:Nat,max(max(max(min(?x:Nat,0),0),0),0))",
"min(?z:Nat,?x:Nat) = min(max(?y:Nat,min(?z:Nat,max(max(?x:Nat,?x:Nat),?x:Nat))),?z:Nat)",
"min(?z:Nat,?x:Nat) = min(min(?x:Nat,?x:Nat),max(?y:Nat,min(max(?y:Nat,?x:Nat),0)))",
"min(?z:Nat,?y:Nat) = max(max(?x:Nat,?z:Nat),max(0,min(0,min(0,?y:Nat))))",
"min(?z:Nat,?y:Nat) = min(max(min(?y:Nat,max(0,max(?z:Nat,0))),?y:Nat),?x:Nat)",
"min(?z:Nat,?y:Nat) = min(min(?x:Nat,max(max(?y:Nat,max(?x:Nat,0)),?y:Nat)),?y:Nat)",
"min(?z:Nat,?z:Nat) = max(min(?x:Nat,max(max(max(?y:Nat,?y:Nat),?x:Nat),0)),?x:Nat)",
"min(0,?x:Nat) = max(min(max(?x:Nat,max(min(?y:Nat,?z:Nat),?y:Nat)),0),?x:Nat)",
"min(0,?z:Nat) = max(?y:Nat,max(max(?y:Nat,max(0,min(?x:Nat,?x:Nat))),0))",
"min(0,?z:Nat) = max(max(min(max(?y:Nat,min(0,?z:Nat)),?z:Nat),?z:Nat),?y:Nat)",
"min(0,?z:Nat) = min(?y:Nat,min(min(?z:Nat,?z:Nat),max(min(?z:Nat,?x:Nat),?z:Nat)))",
"max(?x:Nat,max(?z:Nat,?z:Nat)) = min(min(max(max(?y:Nat,?y:Nat),?x:Nat),?y:Nat),?z:Nat)",
"max(?x:Nat,min(?x:Nat,?y:Nat)) = max(max(?y:Nat,max(?z:Nat,min(?x:Nat,?x:Nat))),?y:Nat)",
"max(?y:Nat,max(?z:Nat,?x:Nat)) = max(max(min(0,?z:Nat),min(?y:Nat,?z:Nat)),?x:Nat)",
"max(?y:Nat,min(?z:Nat,?x:Nat)) = max(max(min(max(?z:Nat,?y:Nat),?x:Nat),0),?x:Nat)",
"max(?z:Nat,min(?y:Nat,?y:Nat)) = min(?x:Nat,min(?x:Nat,min(0,min(?y:Nat,?z:Nat))))",
"max(max(?z:Nat,?y:Nat),?z:Nat) = max(min(?x:Nat,?x:Nat),min(0,min(?z:Nat,?x:Nat)))",
"max(min(?x:Nat,?y:Nat),?x:Nat) = min(?y:Nat,max(?y:Nat,max(min(?z:Nat,?z:Nat),?z:Nat)))",
"max(min(?x:Nat,?z:Nat),?z:Nat) = min(max(?x:Nat,min(max(?x:Nat,?x:Nat),?z:Nat)),?y:Nat)",
"max(min(?z:Nat,0),?y:Nat) = max(?x:Nat,min(?x:Nat,min(0,max(?y:Nat,?z:Nat))))",
"max(min(0,?x:Nat),?x:Nat) = max(max(?y:Nat,?z:Nat),min(?x:Nat,min(?y:Nat,0)))",
"max(min(0,?y:Nat),?y:Nat) = min(?y:Nat,min(min(max(?y:Nat,?y:Nat),0),?y:Nat))",
"min(?x:Nat,max(?y:Nat,?y:Nat)) = max(min(?x:Nat,max(?x:Nat,?z:Nat)),min(0,?y:Nat))",
"min(?x:Nat,max(?z:Nat,0)) = min(max(max(?y:Nat,max(?x:Nat,?z:Nat)),?y:Nat),?z:Nat)",
"min(?x:Nat,min(0,?z:Nat)) = max(min(0,max(?y:Nat,0)),max(?y:Nat,?y:Nat))",
"min(?y:Nat,max(?x:Nat,0)) = min(max(?x:Nat,min(?z:Nat,max(?x:Nat,?y:Nat))),?y:Nat)",
"min(?y:Nat,max(?y:Nat,?x:Nat)) = min(?x:Nat,max(min(?x:Nat,?x:Nat),min(?x:Nat,?y:Nat)))",
"min(?y:Nat,min(?x:Nat,?y:Nat)) = max(max(?x:Nat,?z:Nat),min(max(?z:Nat,?x:Nat),?y:Nat))",
"min(?z:Nat,min(?y:Nat,?y:Nat)) = min(max(max(?x:Nat,?x:Nat),min(?y:Nat,0)),?y:Nat)",
"min(0,min(?y:Nat,?y:Nat)) = min(max(max(?x:Nat,?x:Nat),min(0,?x:Nat)),?y:Nat)",
"min(max(?x:Nat,?y:Nat),?y:Nat) = min(min(max(0,?y:Nat),?y:Nat),min(?x:Nat,?y:Nat))",
"min(max(?x:Nat,0),?y:Nat) = max(min(max(?x:Nat,0),?y:Nat),min(?x:Nat,?z:Nat))",
"min(max(?y:Nat,?x:Nat),?x:Nat) = max(max(?y:Nat,max(?y:Nat,max(0,0))),0)",
"min(max(?y:Nat,?z:Nat),?x:Nat) = min(?z:Nat,min(min(?x:Nat,?x:Nat),min(0,?z:Nat)))",
"min(max(0,?y:Nat),?x:Nat) = max(min(?x:Nat,?y:Nat),min(max(?x:Nat,?x:Nat),?z:Nat))",
"min(max(0,?y:Nat),?z:Nat) = max(?x:Nat,max(?z:Nat,max(0,min(?y:Nat,?x:Nat))))",
"min(min(?x:Nat,?y:Nat),?z:Nat) = max(?z:Nat,max(min(?x:Nat,max(?z:Nat,?y:Nat)),0))",
"min(min(?z:Nat,?y:Nat),?y:Nat) = max(min(?y:Nat,?y:Nat),max(max(0,0),?y:Nat))",
"min(min(0,?y:Nat),?z:Nat) = min(?z:Nat,max(max(max(?y:Nat,?y:Nat),?z:Nat),?x:Nat))",
"min(min(0,?y:Nat),0) = min(min(?z:Nat,?x:Nat),max(0,min(?z:Nat,?x:Nat)))",
"max(?x:Nat,max(?x:Nat,min(?y:Nat,?y:Nat))) = max(?x:Nat,min(max(?x:Nat,?y:Nat),?x:Nat))",
"max(?x:Nat,max(?y:Nat,min(0,?x:Nat))) = max(?y:Nat,min(max(0,?x:Nat),0))",
"max(?x:Nat,min(?y:Nat,min(?z:Nat,?x:Nat))) = max(max(min(?z:Nat,0),?y:Nat),?x:Nat)",
"max(?z:Nat,min(max(?x:Nat,?y:Nat),0)) = min(min(?x:Nat,?y:Nat),max(?x:Nat,?y:Nat))",
"max(max(?x:Nat,?y:Nat),max(0,?x:Nat)) = min(max(?z:Nat,?y:Nat),max(?y:Nat,?z:Nat))",
"max(max(?x:Nat,max(?y:Nat,?y:Nat)),?x:Nat) = min(min(max(0,?y:Nat),?z:Nat),?z:Nat)",
"max(max(?x:Nat,min(?x:Nat,?x:Nat)),?z:Nat) = max(?x:Nat,max(min(0,?z:Nat),0))",
"max(max(?x:Nat,min(?z:Nat,?y:Nat)),0) = max(max(max(?y:Nat,0),?x:Nat),?y:Nat)",
"max(max(?y:Nat,max(?y:Nat,?x:Nat)),?y:Nat) = min(0,min(min(?y:Nat,?x:Nat),?z:Nat))",
"max(max(min(?x:Nat,?z:Nat),0),?y:Nat) = min(max(0,0),max(?y:Nat,?x:Nat))",
"max(min(?y:Nat,?x:Nat),max(?y:Nat,?x:Nat)) = max(?y:Nat,max(?y:Nat,min(?x:Nat,?y:Nat)))",
"max(min(?y:Nat,max(?x:Nat,0)),?z:Nat) = max(min(?z:Nat,max(?y:Nat,?y:Nat)),?y:Nat)",
"max(min(min(?y:Nat,?y:Nat),?x:Nat),?y:Nat) = max(?x:Nat,min(min(?x:Nat,?z:Nat),?y:Nat))",
"max(min(min(?z:Nat,?y:Nat),?x:Nat),?z:Nat) = min(0,min(?y:Nat,max(?x:Nat,?z:Nat)))",
"min(?x:Nat,max(min(?x:Nat,?y:Nat),?y:Nat)) = min(?x:Nat,max(?x:Nat,min(?y:Nat,0)))",
"min(?x:Nat,max(min(?x:Nat,0),?y:Nat)) = min(max(?z:Nat,min(?x:Nat,?y:Nat)),?x:Nat)",
"min(max(0,?x:Nat),min(?x:Nat,0)) = max(?y:Nat,max(?y:Nat,max(?y:Nat,?z:Nat)))",
"min(max(max(?z:Nat,?x:Nat),?x:Nat),?y:Nat) = min(min(max(?y:Nat,?y:Nat),?x:Nat),?y:Nat)",
"min(min(?x:Nat,min(?z:Nat,?x:Nat)),?y:Nat) = min(max(?y:Nat,max(?x:Nat,?z:Nat)),?x:Nat)",
"min(min(?y:Nat,?y:Nat),min(?z:Nat,0)) = max(max(max(?y:Nat,?y:Nat),?x:Nat),?y:Nat)",
"min(min(?y:Nat,max(?y:Nat,0)),?z:Nat) = min(max(?z:Nat,?y:Nat),max(?x:Nat,?x:Nat))",
"min(min(?y:Nat,min(?y:Nat,?x:Nat)),?z:Nat) = min(min(?y:Nat,max(?y:Nat,?y:Nat)),?x:Nat)",
"max(?z:Nat,min(?x:Nat,max(min(?z:Nat,?x:Nat),?y:Nat))) = max(?x:Nat,max(?y:Nat,?x:Nat))",
"max(?z:Nat,min(0,min(max(?y:Nat,?y:Nat),?x:Nat))) = max(max(?y:Nat,?x:Nat),?x:Nat)",
"max(max(?x:Nat,max(?x:Nat,?x:Nat)),max(?y:Nat,?z:Nat)) = min(min(?y:Nat,0),?z:Nat)",
"max(min(?x:Nat,?y:Nat),max(max(?y:Nat,?z:Nat),?x:Nat)) = max(?z:Nat,min(?y:Nat,?y:Nat))",
"min(?x:Nat,max(min(?y:Nat,?z:Nat),max(?x:Nat,?x:Nat))) = min(min(?x:Nat,0),?x:Nat)",
"min(?y:Nat,min(0,min(?x:Nat,max(?z:Nat,0)))) = max(max(?y:Nat,0),?y:Nat)",
"min(?z:Nat,max(?y:Nat,min(min(?x:Nat,?x:Nat),?z:Nat))) = min(max(?x:Nat,0),0)",
"min(0,min(min(max(0,?z:Nat),?x:Nat),?z:Nat)) = min(max(?x:Nat,?z:Nat),?y:Nat)",
"min(max(?x:Nat,max(?x:Nat,0)),max(?y:Nat,?y:Nat)) = min(?x:Nat,max(?x:Nat,?z:Nat))",
"min(max(?x:Nat,max(max(?z:Nat,0),0)),?y:Nat) = max(max(?z:Nat,?x:Nat),?z:Nat)",
"min(max(min(min(?x:Nat,?y:Nat),0),?z:Nat),?y:Nat) = min(?x:Nat,min(?x:Nat,0))",
"min(min(?x:Nat,?x:Nat),max(?x:Nat,min(?x:Nat,?y:Nat))) = min(max(0,0),?y:Nat)",
"min(min(?x:Nat,max(?z:Nat,?x:Nat)),min(?x:Nat,?x:Nat)) = min(min(?x:Nat,0),?x:Nat)",
"min(min(?y:Nat,min(max(?x:Nat,?z:Nat),?x:Nat)),0) = max(?z:Nat,max(?y:Nat,?y:Nat))",
"min(min(max(0,min(?x:Nat,?y:Nat)),?y:Nat),?x:Nat) = max(max(?x:Nat,?y:Nat),?y:Nat)",
"min(min(min(?y:Nat,max(0,?x:Nat)),?x:Nat),?y:Nat) = min(?x:Nat,max(0,?x:Nat))",
"max(?x:Nat,max(0,max(min(?x:Nat,max(?x:Nat,?x:Nat)),?z:Nat))) = min(0,?x:Nat)",
"max(?x:Nat,min(?y:Nat,max(?y:Nat,max(min(?y:Nat,?y:Nat),?y:Nat)))) = min(?y:Nat,?x:Nat)",
"max(?x:Nat,min(?y:Nat,min(?z:Nat,max(min(0,?y:Nat),?y:Nat)))) = min(0,?y:Nat)",
"max(?x:Nat,min(0,min(max(min(?x:Nat,?y:Nat),?x:Nat),0))) = min(?x:Nat,?y:Nat)",
"max(?x:Nat,min(min(0,min(0,max(?x:Nat,?y:Nat))),?y:Nat)) = max(?y:Nat,?x:Nat)",
"max(?x:Nat,min(min(0,min(min(0,?y:Nat),?x:Nat)),?y:Nat)) = min(?y:Nat,?x:Nat)",
"max(?y:Nat,max(?y:Nat,min(min(?x:Nat,?y:Nat),min(?z:Nat,0)))) = min(?x:Nat,?z:Nat)",
"max(?y:Nat,max(max(max(min(?x:Nat,?y:Nat),?z:Nat),?z:Nat),0)) = max(?y:Nat,?x:Nat)",
"max(?y:Nat,max(min(max(0,?x:Nat),min(?y:Nat,?x:Nat)),?y:Nat)) = max(?x:Nat,?z:Nat)",
"max(?z:Nat,min(?x:Nat,max(0,min(min(?z:Nat,?y:Nat),?x:Nat)))) = min(0,?y:Nat)",
"max(max(?y:Nat,min(?y:Nat,max(?x:Nat,max(?x:Nat,?z:Nat)))),0) = min(?z:Nat,?z:Nat)",
"max(max(?y:Nat,min(max(?z:Nat,max(?y:Nat,?y:Nat)),?z:Nat)),?x:Nat) = min(?z:Nat,?y:Nat)",
"max(max(?y:Nat,min(min(max(?z:Nat,?x:Nat),0),?x:Nat)),?y:Nat) = max(0,?x:Nat)",
"max(max(0,max(min(?y:Nat,?y:Nat),?y:Nat)),min(?z:Nat,?y:Nat)) = max(?x:Nat,0)",
"max(max(max(min(max(?x:Nat,0),?y:Nat),0),?y:Nat),0) = max(?x:Nat,?x:Nat)",
"max(max(min(min(?x:Nat,min(?x:Nat,?x:Nat)),?y:Nat),0),0) = min(0,?x:Nat)",
"max(min(?z:Nat,max(?y:Nat,min(max(?x:Nat,?y:Nat),?x:Nat))),?y:Nat) = max(?z:Nat,0)",
"max(min(?z:Nat,max(max(?x:Nat,?z:Nat),?y:Nat)),min(?y:Nat,0)) = min(?y:Nat,?z:Nat)",
"max(min(max(?x:Nat,?x:Nat),?y:Nat),max(min(?y:Nat,?z:Nat),?z:Nat)) = min(0,?x:Nat)",
"max(min(min(?y:Nat,min(?x:Nat,?x:Nat)),?y:Nat),min(?z:Nat,?x:Nat)) = max(?y:Nat,?y:Nat)",
"max(min(min(?z:Nat,min(?y:Nat,max(?x:Nat,?x:Nat))),?z:Nat),0) = min(?x:Nat,0)",
"max(min(min(?z:Nat,min(max(?y:Nat,0),?z:Nat)),0),?y:Nat) = min(?z:Nat,0)",
"max(min(min(max(?z:Nat,0),?y:Nat),0),max(?x:Nat,?x:Nat)) = max(?z:Nat,0)",
"max(min(min(max(?z:Nat,0),0),max(?x:Nat,?y:Nat)),?y:Nat) = min(0,?x:Nat)",
"min(?x:Nat,max(min(0,min(0,max(?y:Nat,?y:Nat))),?x:Nat)) = min(0,?x:Nat)",
"min(?x:Nat,min(min(max(?y:Nat,min(0,?y:Nat)),?x:Nat),?y:Nat)) = max(0,?y:Nat)",
"min(?x:Nat,min(min(min(?y:Nat,?y:Nat),max(?x:Nat,0)),?z:Nat)) = min(?x:Nat,?x:Nat)",
"min(?y:Nat,max(min(?y:Nat,min(?x:Nat,min(?y:Nat,?x:Nat))),?z:Nat)) = max(?x:Nat,?x:Nat)",
"min(?y:Nat,max(min(0,max(?x:Nat,?z:Nat)),min(?x:Nat,?x:Nat))) = max(?z:Nat,?x:Nat)",
"min(?y:Nat,min(?z:Nat,min(max(min(0,0),0),0))) = max(?z:Nat,?x:Nat)",
"min(?z:Nat,min(min(?y:Nat,max(?z:Nat,0)),max(?x:Nat,?y:Nat))) = max(?x:Nat,0)",
"min(0,max(min(0,max(0,min(?x:Nat,?y:Nat))),?x:Nat)) = max(?x:Nat,?x:Nat)",
"min(max(?x:Nat,max(?y:Nat,max(?y:Nat,?z:Nat))),max(?y:Nat,?y:Nat)) = min(?y:Nat,?x:Nat)",
"min(max(?x:Nat,max(?y:Nat,min(?z:Nat,min(?z:Nat,?x:Nat)))),?z:Nat) = max(?y:Nat,?x:Nat)",
"min(max(?y:Nat,min(max(0,?x:Nat),max(?x:Nat,?x:Nat))),?y:Nat) = min(0,?y:Nat)",
"min(max(?y:Nat,min(max(0,?y:Nat),?z:Nat)),max(?y:Nat,?y:Nat)) = min(?x:Nat,?y:Nat)",
"min(max(?z:Nat,max(max(max(?x:Nat,?y:Nat),?z:Nat),?x:Nat)),?y:Nat) = max(?x:Nat,?y:Nat)",
"min(max(?z:Nat,min(min(?x:Nat,max(?y:Nat,?y:Nat)),0)),?x:Nat) = min(?z:Nat,?y:Nat)",
"min(max(max(?y:Nat,0),?x:Nat),min(?y:Nat,min(?z:Nat,?x:Nat))) = min(?x:Nat,?y:Nat)",
"min(max(max(?y:Nat,0),max(max(0,?x:Nat),?x:Nat)),?y:Nat) = min(?y:Nat,0)",
"min(max(max(0,?x:Nat),max(?z:Nat,?y:Nat)),min(?y:Nat,0)) = max(?x:Nat,?x:Nat)",
"min(max(min(?y:Nat,max(min(?x:Nat,?x:Nat),0)),?z:Nat),?x:Nat) = min(0,?y:Nat)",
"min(max(min(max(?x:Nat,?z:Nat),max(?x:Nat,?x:Nat)),?y:Nat),?x:Nat) = max(?y:Nat,?x:Nat)",
"min(max(min(min(0,min(0,?y:Nat)),?z:Nat),?z:Nat),?y:Nat) = min(?x:Nat,?z:Nat)",
"min(min(?x:Nat,max(?x:Nat,max(max(0,?z:Nat),?y:Nat))),0) = min(?y:Nat,?y:Nat)",
"min(min(?y:Nat,?y:Nat),max(0,min(max(?x:Nat,?x:Nat),?x:Nat))) = min(?y:Nat,?z:Nat)",
"min(min(0,0),min(min(?x:Nat,max(?z:Nat,?x:Nat)),?x:Nat)) = min(?x:Nat,?y:Nat)",
"min(min(0,max(min(0,?y:Nat),?z:Nat)),max(0,?x:Nat)) = min(?y:Nat,?x:Nat)",
"min(min(0,min(min(?z:Nat,?y:Nat),max(0,?x:Nat))),0) = max(?x:Nat,?x:Nat)",
"min(min(0,min(min(?z:Nat,0),max(?z:Nat,?y:Nat))),?z:Nat) = max(?y:Nat,?y:Nat)",
"min(min(max(?x:Nat,?z:Nat),0),max(min(0,?x:Nat),0)) = max(?y:Nat,0)",
"min(min(min(?z:Nat,max(0,?y:Nat)),max(?z:Nat,?y:Nat)),?y:Nat) = max(?z:Nat,?x:Nat)",
"min(min(min(0,min(?y:Nat,min(?x:Nat,0))),?x:Nat),?z:Nat) = max(?x:Nat,?y:Nat)",
"max(?x:Nat,?y:Nat) = max(?y:Nat,max(min(?y:Nat,?y:Nat),max(min(0,min(?x:Nat,?x:Nat)),?z:Nat)))",
"max(?x:Nat,?y:Nat) = min(max(max(?y:Nat,?y:Nat),max(?y:Nat,min(?x:Nat,max(?y:Nat,?x:Nat)))),?y:Nat)",
"max(?x:Nat,?z:Nat) = max(max(min(?y:Nat,max(min(?x:Nat,0),min(?x:Nat,?y:Nat))),?z:Nat),?z:Nat)",
"max(?x:Nat,?z:Nat) = max(min(max(max(?x:Nat,min(?x:Nat,?y:Nat)),?x:Nat),min(?x:Nat,?z:Nat)),?x:Nat)",
"max(?y:Nat,?x:Nat) = max(min(?x:Nat,max(max(max(?x:Nat,0),?x:Nat),0)),min(?z:Nat,?y:Nat))",
"max(?y:Nat,?x:Nat) = min(max(max(max(min(?z:Nat,?x:Nat),min(?x:Nat,?y:Nat)),?y:Nat),?z:Nat),?x:Nat)",
"max(?y:Nat,?x:Nat) = min(min(?x:Nat,min(?x:Nat,min(max(?z:Nat,?y:Nat),min(0,?y:Nat)))),?x:Nat)",
"max(?y:Nat,?y:Nat) = max(max(?y:Nat,max(?y:Nat,max(min(?x:Nat,?z:Nat),max(0,?y:Nat)))),?y:Nat)",
"max(?y:Nat,?y:Nat) = min(min(?z:Nat,?y:Nat),max(max(max(?z:Nat,max(?z:Nat,?x:Nat)),?x:Nat),?z:Nat))",
"max(?y:Nat,?y:Nat) = min(min(?z:Nat,min(max(max(?y:Nat,?y:Nat),min(?y:Nat,?x:Nat)),0)),?x:Nat)",
"max(?y:Nat,?y:Nat) = min(min(max(?x:Nat,max(min(?y:Nat,0),?y:Nat)),?y:Nat),min(?x:Nat,0))",
"max(?y:Nat,?z:Nat) = max(?y:Nat,min(max(?x:Nat,max(max(?y:Nat,?y:Nat),max(?z:Nat,?y:Nat))),?y:Nat))",
"max(?y:Nat,?z:Nat) = min(?z:Nat,max(max(min(min(?y:Nat,?x:Nat),min(?x:Nat,?y:Nat)),?z:Nat),?x:Nat))",
"max(?y:Nat,?z:Nat) = min(min(0,max(min(?y:Nat,?x:Nat),max(min(?x:Nat,?y:Nat),?x:Nat))),?y:Nat)",
"max(?y:Nat,0) = max(max(min(max(?x:Nat,?x:Nat),0),min(0,?y:Nat)),max(?y:Nat,?x:Nat))",
"max(?y:Nat,0) = min(max(max(min(0,max(?z:Nat,0)),?y:Nat),max(?y:Nat,?y:Nat)),?x:Nat)",
"max(?z:Nat,?x:Nat) = min(max(?y:Nat,max(max(min(?x:Nat,?x:Nat),max(?y:Nat,?z:Nat)),?y:Nat)),?y:Nat)",
"max(?z:Nat,?x:Nat) = min(min(max(min(0,0),max(?z:Nat,min(?y:Nat,?y:Nat))),?x:Nat),?x:Nat)",
"max(?z:Nat,0) = min(0,min(max(min(max(?z:Nat,?z:Nat),min(?y:Nat,?z:Nat)),0),0))",
"max(0,?x:Nat) = min(?x:Nat,max(max(0,min(min(?z:Nat,0),min(?x:Nat,?x:Nat))),?y:Nat))",
"max(0,?y:Nat) = max(max(?x:Nat,?y:Nat),min(max(?z:Nat,max(0,?x:Nat)),min(?z:Nat,?z:Nat)))",
"max(0,?y:Nat) = min(max(?z:Nat,?z:Nat),min(max(max(?x:Nat,0),max(?x:Nat,?y:Nat)),?z:Nat))",
"max(0,?z:Nat) = max(?x:Nat,max(max(min(max(0,0),max(?y:Nat,?x:Nat)),0),?y:Nat))",
"min(?x:Nat,?x:Nat) = min(max(?y:Nat,min(min(max(?z:Nat,?x:Nat),?y:Nat),max(?y:Nat,?y:Nat))),?y:Nat)",
"min(?x:Nat,?y:Nat) = max(?z:Nat,max(max(?y:Nat,max(?x:Nat,min(?x:Nat,?x:Nat))),max(?x:Nat,?z:Nat)))",
"min(?x:Nat,?y:Nat) = max(min(max(?x:Nat,?y:Nat),?y:Nat),max(min(min(?z:Nat,0),?y:Nat),?y:Nat))",
"min(?x:Nat,?y:Nat) = min(min(?x:Nat,min(min(min(?z:Nat,?z:Nat),?y:Nat),max(?x:Nat,0))),?z:Nat)",
"min(?x:Nat,?z:Nat) = min(min(max(max(max(0,0),?y:Nat),?y:Nat),max(?y:Nat,?x:Nat)),0)",
"min(?x:Nat,0) = max(?y:Nat,min(max(min(max(?y:Nat,?x:Nat),?y:Nat),0),min(0,?y:Nat)))",
"min(?x:Nat,0) = min(max(max(min(min(?z:Nat,?z:Nat),?x:Nat),?x:Nat),?x:Nat),max(?y:Nat,0))",
"min(?y:Nat,?y:Nat) = min(?y:Nat,min(min(min(max(?y:Nat,?z:Nat),?x:Nat),?x:Nat),min(?y:Nat,?y:Nat)))",
"min(?y:Nat,?y:Nat) = min(min(max(max(0,?y:Nat),min(?z:Nat,?y:Nat)),min(?x:Nat,0)),?z:Nat)",
"min(?y:Nat,?z:Nat) = max(min(max(min(?x:Nat,0),?z:Nat),?x:Nat),min(?x:Nat,max(?y:Nat,?x:Nat)))",
"min(?y:Nat,?z:Nat) = min(min(max(max(min(?y:Nat,?y:Nat),?z:Nat),?y:Nat),min(?x:Nat,?x:Nat)),?y:Nat)",
"min(?z:Nat,?x:Nat) = min(max(?y:Nat,?z:Nat),max(?x:Nat,min(min(min(?z:Nat,?y:Nat),?x:Nat),?y:Nat)))",
"min(?z:Nat,?x:Nat) = min(max(max(max(0,?x:Nat),max(min(0,?y:Nat),0)),?z:Nat),0)",
"min(?z:Nat,?y:Nat) = max(min(0,min(min(max(?y:Nat,0),max(?x:Nat,0)),?z:Nat)),?y:Nat)",
"min(?z:Nat,?z:Nat) = min(min(min(max(0,?y:Nat),max(0,min(?y:Nat,?x:Nat))),?z:Nat),?z:Nat)",
"min(0,?x:Nat) = max(max(max(?y:Nat,?y:Nat),min(?x:Nat,min(max(?x:Nat,?x:Nat),?x:Nat))),?y:Nat)",
"max(?y:Nat,max(?y:Nat,?x:Nat)) = max(max(0,min(?y:Nat,?y:Nat)),max(min(?y:Nat,?y:Nat),?z:Nat))",
"max(?y:Nat,min(?x:Nat,0)) = max(min(min(?y:Nat,?x:Nat),?x:Nat),min(?x:Nat,max(?y:Nat,?y:Nat)))",
"max(?y:Nat,min(0,?y:Nat)) = min(max(0,min(?x:Nat,max(max(?y:Nat,0),?x:Nat))),?x:Nat)",
"max(?z:Nat,min(?y:Nat,0)) = max(min(min(?x:Nat,min(min(0,?y:Nat),0)),?y:Nat),?x:Nat)",
"max(?z:Nat,min(?z:Nat,?y:Nat)) = min(?x:Nat,min(min(min(?y:Nat,?z:Nat),max(?y:Nat,?x:Nat)),?x:Nat))",
"max(?z:Nat,min(0,?x:Nat)) = max(max(0,max(?z:Nat,?y:Nat)),max(?x:Nat,max(?y:Nat,?y:Nat)))",
"max(0,max(?y:Nat,?x:Nat)) = min(min(?y:Nat,max(min(0,min(?y:Nat,?y:Nat)),?z:Nat)),?x:Nat)",
"max(max(?x:Nat,?z:Nat),?x:Nat) = max(?y:Nat,min(?z:Nat,max(0,min(0,min(?y:Nat,?x:Nat)))))",
"max(max(?z:Nat,?y:Nat),?y:Nat) = min(0,max(?x:Nat,max(min(0,max(0,0)),?x:Nat)))",
"max(min(?x:Nat,?y:Nat),?x:Nat) = max(max(?y:Nat,min(min(0,?x:Nat),min(?y:Nat,?x:Nat))),?z:Nat)",
"max(min(?x:Nat,?y:Nat),?y:Nat) = min(min(min(min(?x:Nat,min(0,?z:Nat)),?x:Nat),?y:Nat),?y:Nat)",
"max(min(?x:Nat,?y:Nat),?z:Nat) = max(min(?y:Nat,min(?x:Nat,min(?x:Nat,max(?z:Nat,0)))),?z:Nat)",
"max(min(?x:Nat,?z:Nat),0) = max(?x:Nat,max(max(max(?x:Nat,min(?x:Nat,?y:Nat)),0),0))",
"max(min(?x:Nat,0),?z:Nat) = min(min(min(min(?y:Nat,max(?x:Nat,?x:Nat)),?z:Nat),?y:Nat),?y:Nat)",
"max(min(?y:Nat,?y:Nat),?y:Nat) = min(min(?x:Nat,?y:Nat),max(min(max(?x:Nat,?y:Nat),?y:Nat),0))",
"max(min(?y:Nat,0),?z:Nat) = max(?x:Nat,min(?z:Nat,max(?x:Nat,max(min(?y:Nat,?y:Nat),?y:Nat))))",
"min(?x:Nat,max(?y:Nat,?y:Nat)) = max(?y:Nat,max(?y:Nat,max(min(max(?y:Nat,?x:Nat),0),?y:Nat)))",
"min(?x:Nat,max(0,?y:Nat)) = min(?x:Nat,max(min(?y:Nat,min(?x:Nat,?x:Nat)),min(?x:Nat,?x:Nat)))",
"min(?x:Nat,min(?y:Nat,?x:Nat)) = min(?x:Nat,max(?z:Nat,min(?y:Nat,max(max(?y:Nat,?y:Nat),?y:Nat))))",
"min(?y:Nat,min(?z:Nat,?y:Nat)) = min(?x:Nat,max(max(max(?x:Nat,max(?y:Nat,?z:Nat)),?x:Nat),0))",
"min(max(?y:Nat,?x:Nat),?z:Nat) = max(?y:Nat,min(min(?x:Nat,min(max(?x:Nat,?y:Nat),?y:Nat)),?z:Nat))",
"min(max(?y:Nat,0),?x:Nat) = max(max(max(?x:Nat,?y:Nat),?x:Nat),min(?y:Nat,min(?y:Nat,?x:Nat)))",
"min(min(?y:Nat,?z:Nat),0) = max(?z:Nat,max(min(0,max(max(?x:Nat,0),?z:Nat)),?x:Nat))",
"min(min(?z:Nat,?y:Nat),?y:Nat) = max(min(?y:Nat,min(?y:Nat,?x:Nat)),min(?y:Nat,min(?y:Nat,?y:Nat)))",
"min(min(0,?y:Nat),?y:Nat) = min(max(min(min(?z:Nat,max(?y:Nat,?x:Nat)),?x:Nat),?z:Nat),?y:Nat)",
"max(?y:Nat,max(min(?y:Nat,0),?y:Nat)) = min(?z:Nat,min(min(?x:Nat,min(?x:Nat,0)),?x:Nat))",
"max(?z:Nat,min(max(0,0),?x:Nat)) = min(max(?x:Nat,min(?y:Nat,0)),max(0,0))",
"max(0,max(?x:Nat,min(?x:Nat,?x:Nat))) = min(max(min(?z:Nat,0),min(0,0)),?x:Nat)",
"max(0,min(max(0,?y:Nat),0)) = min(?z:Nat,max(min(?y:Nat,min(?z:Nat,?y:Nat)),?z:Nat))",
"max(max(?y:Nat,?y:Nat),max(?x:Nat,0)) = min(?x:Nat,min(min(?x:Nat,?z:Nat),min(?z:Nat,?x:Nat)))",
"max(max(?y:Nat,min(?z:Nat,0)),?y:Nat) = min(min(?x:Nat,max(min(?y:Nat,?y:Nat),?x:Nat)),?x:Nat)",
"max(min(0,min(?y:Nat,?y:Nat)),?z:Nat) = min(max(?x:Nat,max(?x:Nat,min(?x:Nat,?z:Nat))),?y:Nat)",
"max(min(max(0,?y:Nat),?x:Nat),0) = max(max(max(?x:Nat,min(?y:Nat,?y:Nat)),0),?z:Nat)",
"min(max(?z:Nat,min(?y:Nat,?y:Nat)),?y:Nat) = max(max(?x:Nat,max(?x:Nat,max(?y:Nat,?y:Nat))),0)",
"min(max(max(?z:Nat,?z:Nat),?y:Nat),?x:Nat) = max(max(0,min(?x:Nat,?x:Nat)),min(?z:Nat,?z:Nat))",
"min(min(max(?z:Nat,?x:Nat),0),0) = max(min(?x:Nat,0),min(?z:Nat,max(?x:Nat,?x:Nat)))",
"max(?x:Nat,max(max(?x:Nat,max(?y:Nat,?x:Nat)),?x:Nat)) = min(?x:Nat,min(min(0,?x:Nat),?x:Nat))",
"max(?y:Nat,min(min(?x:Nat,max(0,?x:Nat)),?x:Nat)) = max(min(?z:Nat,max(?x:Nat,?x:Nat)),?y:Nat)",
"max(max(min(?y:Nat,min(?z:Nat,?y:Nat)),?x:Nat),?z:Nat) = max(?z:Nat,max(?x:Nat,max(?y:Nat,?z:Nat)))",
"max(min(?y:Nat,?z:Nat),max(min(?y:Nat,0),?y:Nat)) = max(max(?y:Nat,max(?y:Nat,?x:Nat)),?y:Nat)",
"max(min(min(?x:Nat,?y:Nat),?x:Nat),min(?z:Nat,?x:Nat)) = min(min(max(0,0),?x:Nat),?x:Nat)",
"max(min(min(?y:Nat,min(0,?x:Nat)),?x:Nat),?z:Nat) = min(?z:Nat,max(?x:Nat,min(?y:Nat,?y:Nat)))",
"min(?x:Nat,min(?x:Nat,min(?y:Nat,min(?y:Nat,?x:Nat)))) = min(max(?y:Nat,max(0,0)),?y:Nat)",
"min(?y:Nat,max(max(0,?x:Nat),min(?y:Nat,?y:Nat))) = max(?z:Nat,max(?y:Nat,min(?z:Nat,0)))",
"min(?y:Nat,min(max(?x:Nat,min(?x:Nat,?z:Nat)),?z:Nat)) = max(max(min(?z:Nat,?x:Nat),?y:Nat),?x:Nat)",
"min(0,max(min(max(0,0),?y:Nat),?x:Nat)) = max(?z:Nat,max(min(?y:Nat,?y:Nat),?y:Nat))",
"min(0,min(0,min(min(?y:Nat,0),?x:Nat))) = max(max(min(?x:Nat,?x:Nat),?y:Nat),?y:Nat)",
"min(max(?y:Nat,?y:Nat),min(max(?x:Nat,?x:Nat),?y:Nat)) = max(?y:Nat,min(?y:Nat,max(0,?y:Nat)))",
"min(max(?z:Nat,0),max(min(?x:Nat,?x:Nat),0)) = max(?y:Nat,max(?y:Nat,min(?x:Nat,?y:Nat)))",
"min(max(min(0,min(?x:Nat,0)),?x:Nat),0) = max(?x:Nat,max(min(?x:Nat,?x:Nat),?x:Nat))",
"min(min(?z:Nat,min(?z:Nat,?y:Nat)),min(0,?x:Nat)) = max(min(min(?y:Nat,?y:Nat),?y:Nat),?y:Nat)",
"min(min(min(?y:Nat,?x:Nat),?y:Nat),min(?z:Nat,?y:Nat)) = max(min(?x:Nat,min(?x:Nat,?x:Nat)),?x:Nat)",
"max(?x:Nat,max(min(max(min(?x:Nat,?y:Nat),?z:Nat),?z:Nat),0)) = max(max(?z:Nat,?y:Nat),?y:Nat)",
"max(?x:Nat,min(max(?y:Nat,min(max(?y:Nat,?y:Nat),?z:Nat)),?y:Nat)) = min(?y:Nat,max(?y:Nat,?y:Nat))",
"max(?x:Nat,min(max(max(?x:Nat,0),min(?y:Nat,?y:Nat)),0)) = min(max(?x:Nat,?x:Nat),0)",
"max(?y:Nat,min(?z:Nat,max(?z:Nat,min(min(?y:Nat,?y:Nat),?x:Nat)))) = max(min(?x:Nat,?z:Nat),?y:Nat)",
"max(?y:Nat,min(?z:Nat,max(0,min(?x:Nat,min(?x:Nat,?z:Nat))))) = max(min(?z:Nat,?z:Nat),?x:Nat)",
"max(?z:Nat,max(?x:Nat,max(min(0,?y:Nat),max(?y:Nat,0)))) = max(max(?x:Nat,?x:Nat),?x:Nat)",
"max(max(?x:Nat,min(?y:Nat,min(0,min(?x:Nat,?y:Nat)))),0) = max(max(?y:Nat,?y:Nat),?y:Nat)",
"max(max(?z:Nat,max(0,?x:Nat)),min(?y:Nat,min(?z:Nat,?x:Nat))) = min(?y:Nat,max(?x:Nat,?y:Nat))",
"max(max(0,max(min(?z:Nat,?z:Nat),min(?z:Nat,0))),?y:Nat) = max(min(?x:Nat,?z:Nat),?z:Nat)",
"max(max(max(min(?x:Nat,?y:Nat),?z:Nat),max(0,?x:Nat)),0) = max(?x:Nat,max(?z:Nat,?y:Nat))",
"max(max(min(min(?x:Nat,min(?x:Nat,0)),?x:Nat),?x:Nat),?y:Nat) = min(min(?y:Nat,?y:Nat),?z:Nat)",
"max(min(max(?x:Nat,0),?x:Nat),min(max(?x:Nat,?x:Nat),?y:Nat)) = max(max(?y:Nat,?z:Nat),?x:Nat)",
"min(?x:Nat,max(?x:Nat,min(max(?y:Nat,?x:Nat),min(0,0)))) = min(?y:Nat,min(?y:Nat,?x:Nat))",
"min(?x:Nat,max(min(?x:Nat,min(0,max(?y:Nat,?z:Nat))),?y:Nat)) = max(?y:Nat,min(?x:Nat,?x:Nat))",
"min(?x:Nat,min(?y:Nat,max(min(min(?z:Nat,?y:Nat),?z:Nat),?y:Nat))) = min(0,min(?y:Nat,?z:Nat))",
"min(?x:Nat,min(min(max(?x:Nat,max(?x:Nat,?y:Nat)),?y:Nat),?y:Nat)) = max(min(?z:Nat,?y:Nat),?z:Nat)",
"min(?y:Nat,min(min(?y:Nat,min(?x:Nat,max(0,?x:Nat))),?y:Nat)) = min(?y:Nat,max(?z:Nat,?y:Nat))",
"min(?z:Nat,max(max(?y:Nat,min(min(?x:Nat,?y:Nat),?z:Nat)),?y:Nat)) = max(min(?x:Nat,?y:Nat),?y:Nat)",
"min(max(?y:Nat,?y:Nat),min(max(?x:Nat,?x:Nat),max(?y:Nat,?x:Nat))) = max(max(?y:Nat,?x:Nat),?y:Nat)",
"min(max(0,max(?z:Nat,max(?z:Nat,max(0,?y:Nat)))),0) = max(max(0,?x:Nat),?z:Nat)",
"min(max(max(?z:Nat,?x:Nat),?y:Nat),max(?z:Nat,max(?x:Nat,0))) = min(?y:Nat,max(?x:Nat,?x:Nat))",
"min(max(min(max(min(?y:Nat,?z:Nat),?z:Nat),?x:Nat),?z:Nat),?y:Nat) = max(?x:Nat,min(?x:Nat,?x:Nat))",
"min(max(min(min(?x:Nat,min(?y:Nat,0)),?y:Nat),?z:Nat),0) = min(?z:Nat,min(?x:Nat,?x:Nat))",
"max(?x:Nat,max(max(?x:Nat,?y:Nat),max(min(max(?x:Nat,0),?y:Nat),?y:Nat))) = max(?y:Nat,?z:Nat)",
"max(?x:Nat,min(?x:Nat,max(?x:Nat,max(min(?x:Nat,?y:Nat),max(?z:Nat,0))))) = min(?x:Nat,0)",
"max(?y:Nat,min(?y:Nat,max(min(max(?x:Nat,?y:Nat),min(0,?y:Nat)),0))) = min(?z:Nat,?z:Nat)",
"max(?z:Nat,max(max(?x:Nat,0),max(min(min(?y:Nat,?x:Nat),0),?y:Nat))) = max(?x:Nat,0)",
"max(max(?x:Nat,?y:Nat),max(?y:Nat,min(?y:Nat,max(min(?y:Nat,0),?y:Nat)))) = min(?y:Nat,0)",
"max(max(?x:Nat,min(?y:Nat,max(min(?y:Nat,?z:Nat),min(?x:Nat,?x:Nat)))),?x:Nat) = min(?x:Nat,0)",
"max(max(0,0),min(0,max(min(?x:Nat,max(?x:Nat,?z:Nat)),?z:Nat))) = max(?y:Nat,?z:Nat)",
"max(max(max(?y:Nat,min(min(?y:Nat,?z:Nat),min(?x:Nat,?x:Nat))),?x:Nat),?x:Nat) = max(?x:Nat,?x:Nat)",
"max(max(max(min(?z:Nat,?y:Nat),max(?y:Nat,?y:Nat)),max(?x:Nat,?x:Nat)),?y:Nat) = min(?y:Nat,?y:Nat)",
"max(max(min(0,?y:Nat),?y:Nat),max(?y:Nat,min(min(?x:Nat,0),?y:Nat))) = min(?y:Nat,0)",
"max(min(?x:Nat,?x:Nat),min(min(min(?y:Nat,max(?x:Nat,?y:Nat)),?z:Nat),?z:Nat)) = min(?y:Nat,0)",
"max(min(?y:Nat,?x:Nat),max(max(max(?x:Nat,max(?z:Nat,?x:Nat)),?y:Nat),?z:Nat)) = max(?x:Nat,?y:Nat)",
"max(min(?y:Nat,min(max(min(?x:Nat,?x:Nat),0),?y:Nat)),min(?y:Nat,?y:Nat)) = min(?y:Nat,?x:Nat)",
"max(min(?z:Nat,max(?y:Nat,?y:Nat)),max(max(?y:Nat,min(?x:Nat,?z:Nat)),?x:Nat)) = max(0,0)",
"max(min(0,min(?x:Nat,max(min(?x:Nat,?x:Nat),?z:Nat))),max(?y:Nat,?y:Nat)) = max(?z:Nat,?x:Nat)",
"max(min(max(?x:Nat,?y:Nat),min(?x:Nat,min(min(?y:Nat,?x:Nat),?y:Nat))),?y:Nat) = max(?y:Nat,?x:Nat)",
"max(min(max(?x:Nat,min(?y:Nat,min(?y:Nat,?y:Nat))),?y:Nat),max(?y:Nat,?x:Nat)) = min(?x:Nat,0)",
"max(min(max(?y:Nat,?y:Nat),min(min(min(?x:Nat,?y:Nat),?z:Nat),0)),0) = max(?x:Nat,?y:Nat)",
"max(min(min(min(?y:Nat,min(?x:Nat,0)),?x:Nat),max(?x:Nat,0)),?x:Nat) = max(?y:Nat,?y:Nat)",
"min(?x:Nat,min(min(min(?y:Nat,?x:Nat),min(max(?x:Nat,?y:Nat),0)),?x:Nat)) = max(?z:Nat,0)",
"min(?y:Nat,max(?x:Nat,max(max(?z:Nat,?x:Nat),min(min(?y:Nat,?z:Nat),?z:Nat)))) = max(?y:Nat,?y:Nat)",
"min(?y:Nat,max(?z:Nat,max(max(?z:Nat,max(0,?y:Nat)),min(?y:Nat,0)))) = min(?y:Nat,0)",
"min(?y:Nat,min(?x:Nat,max(max(0,?x:Nat),max(min(?y:Nat,?y:Nat),?x:Nat)))) = min(?y:Nat,?y:Nat)",
"min(?y:Nat,min(min(0,min(?y:Nat,?y:Nat)),max(?x:Nat,max(?z:Nat,?y:Nat)))) = min(?z:Nat,?x:Nat)",
"min(?z:Nat,min(max(?z:Nat,max(max(?x:Nat,?y:Nat),max(?x:Nat,?x:Nat))),?x:Nat)) = min(0,?x:Nat)",
"min(?z:Nat,min(max(?z:Nat,max(min(0,?y:Nat),max(?x:Nat,?y:Nat))),?y:Nat)) = max(?y:Nat,?y:Nat)",
"min(0,min(max(min(min(?y:Nat,?x:Nat),max(?x:Nat,?x:Nat)),?y:Nat),?y:Nat)) = max(?y:Nat,?x:Nat)",
"min(max(?x:Nat,0),max(?y:Nat,min(max(?z:Nat,max(?y:Nat,?x:Nat)),?y:Nat))) = min(0,?y:Nat)",
"min(max(?x:Nat,max(min(0,max(?y:Nat,?y:Nat)),min(?y:Nat,0))),0) = max(?x:Nat,0)",
"min(max(0,max(min(min(?x:Nat,?y:Nat),max(?x:Nat,?y:Nat)),?y:Nat)),?y:Nat) = min(?y:Nat,?z:Nat)",
"min(max(max(?z:Nat,?x:Nat),min(0,min(?y:Nat,?x:Nat))),max(?z:Nat,0)) = min(?y:Nat,?x:Nat)",
"min(max(max(min(?z:Nat,?x:Nat),0),?y:Nat),max(?y:Nat,min(?x:Nat,?y:Nat))) = max(0,0)",
"min(max(min(?x:Nat,?y:Nat),max(min(?y:Nat,max(?x:Nat,?x:Nat)),?z:Nat)),?x:Nat) = max(?x:Nat,?y:Nat)",
"min(max(min(?z:Nat,?x:Nat),min(min(min(?z:Nat,?y:Nat),?x:Nat),?y:Nat)),?x:Nat) = min(?x:Nat,?x:Nat)",
"min(min(?x:Nat,max(max(?x:Nat,max(0,?x:Nat)),max(0,?y:Nat))),?y:Nat) = min(0,?x:Nat)",
"min(min(?x:Nat,max(min(min(?y:Nat,?x:Nat),0),?z:Nat)),min(?y:Nat,?z:Nat)) = max(0,?z:Nat)",
"min(min(?z:Nat,min(min(min(?x:Nat,?x:Nat),min(?z:Nat,?y:Nat)),?x:Nat)),?x:Nat) = max(?z:Nat,?y:Nat)",
"min(min(0,max(max(?z:Nat,0),min(?z:Nat,max(?y:Nat,?y:Nat)))),?x:Nat) = min(?x:Nat,?y:Nat)",
"min(min(max(?y:Nat,min(min(?y:Nat,?x:Nat),?y:Nat)),min(?y:Nat,?x:Nat)),0) = max(?x:Nat,?y:Nat)",
"min(min(max(?z:Nat,?y:Nat),min(max(?z:Nat,?y:Nat),min(?z:Nat,?z:Nat))),?x:Nat) = min(?z:Nat,?z:Nat)",
"min(min(max(min(?z:Nat,?x:Nat),max(?z:Nat,min(?y:Nat,?x:Nat))),?x:Nat),?y:Nat) = max(?y:Nat,?y:Nat)",
"min(min(min(?y:Nat,min(?y:Nat,?x:Nat)),min(max(?x:Nat,0),?z:Nat)),?y:Nat) = max(?y:Nat,?x:Nat)",
"min(min(min(?y:Nat,min(max(?x:Nat,?y:Nat),min(?x:Nat,?z:Nat))),?x:Nat),?y:Nat) = max(?x:Nat,0)",
"min(min(min(0,0),min(?z:Nat,max(0,?y:Nat))),max(?y:Nat,?x:Nat)) = min(?z:Nat,?x:Nat)",
"max(?x:Nat,?x:Nat) = max(min(?y:Nat,?x:Nat),max(?z:Nat,max(?x:Nat,min(max(?y:Nat,0),min(?z:Nat,0)))))",
"max(?x:Nat,?x:Nat) = max(min(?y:Nat,max(min(max(?y:Nat,0),min(?z:Nat,?y:Nat)),0)),min(?y:Nat,?x:Nat))",
"max(?x:Nat,?x:Nat) = max(min(max(max(max(0,?y:Nat),?x:Nat),min(?y:Nat,max(?x:Nat,?x:Nat))),?x:Nat),?y:Nat)",
"max(?x:Nat,?x:Nat) = min(?z:Nat,min(max(0,?x:Nat),min(min(min(?y:Nat,0),?y:Nat),min(?y:Nat,0))))",
"max(?x:Nat,?y:Nat) = min(max(?x:Nat,min(max(?z:Nat,min(?z:Nat,?y:Nat)),min(?x:Nat,min(?x:Nat,?x:Nat)))),?x:Nat)",
"max(?x:Nat,?y:Nat) = min(min(?z:Nat,?x:Nat),max(min(0,max(max(?x:Nat,0),max(?x:Nat,?z:Nat))),0))",
"max(?x:Nat,?z:Nat) = max(max(?x:Nat,max(min(?x:Nat,min(?x:Nat,?z:Nat)),min(max(?z:Nat,?y:Nat),?x:Nat))),?x:Nat)",
"max(?x:Nat,?z:Nat) = max(min(?x:Nat,max(max(?x:Nat,?y:Nat),?y:Nat)),min(min(?y:Nat,?z:Nat),max(?x:Nat,?x:Nat)))",
"max(?y:Nat,?x:Nat) = min(min(0,?y:Nat),min(max(max(?z:Nat,?y:Nat),min(min(0,?z:Nat),?x:Nat)),?x:Nat))",
"max(?y:Nat,?x:Nat) = min(min(max(?z:Nat,?x:Nat),?y:Nat),max(?y:Nat,min(?y:Nat,max(min(?y:Nat,?z:Nat),?y:Nat))))",
"max(?y:Nat,?y:Nat) = max(?z:Nat,min(max(min(?y:Nat,?y:Nat),max(min(?z:Nat,?y:Nat),?x:Nat)),min(0,?x:Nat)))",
"max(?y:Nat,?z:Nat) = max(max(?z:Nat,min(min(?y:Nat,min(0,?z:Nat)),max(?x:Nat,min(?z:Nat,?x:Nat)))),0)",
"max(?y:Nat,0) = min(max(?x:Nat,?y:Nat),min(max(max(max(?y:Nat,?y:Nat),?x:Nat),?y:Nat),min(?y:Nat,0)))",
"max(?z:Nat,?x:Nat) = min(min(max(max(0,min(?y:Nat,?x:Nat)),max(?y:Nat,min(?x:Nat,?y:Nat))),0),?y:Nat)",
"max(?z:Nat,?z:Nat) = min(?y:Nat,max(max(max(?x:Nat,max(?z:Nat,?z:Nat)),min(?x:Nat,?z:Nat)),max(0,?x:Nat)))",
"max(?z:Nat,0) = max(max(max(max(max(?y:Nat,?x:Nat),?y:Nat),max(?x:Nat,?z:Nat)),min(?z:Nat,?x:Nat)),?y:Nat)",
"max(0,?x:Nat) = max(0,max(max(0,min(?y:Nat,?x:Nat)),min(min(max(?x:Nat,?y:Nat),?x:Nat),0)))",
"max(0,?y:Nat) = min(max(?y:Nat,0),max(min(max(?x:Nat,?x:Nat),?y:Nat),min(min(?y:Nat,?z:Nat),?x:Nat)))",
"max(0,?y:Nat) = min(min(max(max(min(?x:Nat,?x:Nat),?x:Nat),?z:Nat),min(?z:Nat,?y:Nat)),min(0,0))",
"min(?x:Nat,?x:Nat) = max(?y:Nat,max(min(?x:Nat,?y:Nat),max(max(min(?y:Nat,?x:Nat),?y:Nat),min(?y:Nat,?z:Nat))))",
"min(?x:Nat,?y:Nat) = max(max(?z:Nat,min(?z:Nat,?x:Nat)),min(?y:Nat,min(min(max(0,?y:Nat),?z:Nat),?y:Nat)))",
"min(?x:Nat,?y:Nat) = max(max(max(?x:Nat,max(?y:Nat,?y:Nat)),max(min(?x:Nat,?y:Nat),?y:Nat)),max(?x:Nat,?x:Nat))",
"min(?x:Nat,?z:Nat) = max(min(min(min(?x:Nat,max(0,?x:Nat)),?x:Nat),?x:Nat),min(?z:Nat,min(?z:Nat,0)))",
"min(?y:Nat,?y:Nat) = max(max(?x:Nat,?x:Nat),min(?x:Nat,min(min(?y:Nat,?z:Nat),max(max(?x:Nat,?x:Nat),0))))",
"min(?y:Nat,?y:Nat) = min(min(max(max(?y:Nat,max(0,0)),max(?x:Nat,?y:Nat)),max(?y:Nat,0)),?z:Nat)",
"min(0,?x:Nat) = min(?x:Nat,max(max(max(?z:Nat,max(?x:Nat,?x:Nat)),max(0,min(0,?x:Nat))),0))",
"min(0,?x:Nat) = min(min(?z:Nat,?z:Nat),max(?y:Nat,min(min(?z:Nat,?y:Nat),min(?x:Nat,min(?x:Nat,?x:Nat)))))",
"min(0,?z:Nat) = max(max(min(?x:Nat,min(?y:Nat,?x:Nat)),min(?x:Nat,min(0,min(?z:Nat,?y:Nat)))),?y:Nat)",
"min(0,?z:Nat) = min(?y:Nat,min(min(max(min(?x:Nat,?x:Nat),0),max(max(?z:Nat,?x:Nat),?y:Nat)),?y:Nat))",
"min(0,0) = min(min(max(min(max(?z:Nat,?y:Nat),max(?x:Nat,?x:Nat)),0),max(?z:Nat,?y:Nat)),?y:Nat)",
"max(?x:Nat,min(?z:Nat,?y:Nat)) = min(?z:Nat,max(min(max(?x:Nat,max(0,?y:Nat)),?y:Nat),max(0,?y:Nat)))",
"max(?x:Nat,min(0,?y:Nat)) = max(max(min(?z:Nat,max(min(0,0),?y:Nat)),?x:Nat),min(0,?z:Nat))",
"max(?y:Nat,min(?x:Nat,0)) = max(min(max(?z:Nat,?x:Nat),min(?x:Nat,max(min(?y:Nat,0),?x:Nat))),?z:Nat)",
"max(?z:Nat,max(?x:Nat,?z:Nat)) = min(max(min(?y:Nat,min(?y:Nat,?y:Nat)),max(0,?y:Nat)),min(?x:Nat,?x:Nat))",
"max(?z:Nat,max(?y:Nat,0)) = max(0,min(max(?y:Nat,0),min(max(?z:Nat,?x:Nat),min(?x:Nat,?x:Nat))))",
"max(max(?y:Nat,?y:Nat),?x:Nat) = max(min(0,min(max(?z:Nat,0),min(max(?y:Nat,0),?z:Nat))),?y:Nat)",
"max(min(?x:Nat,?x:Nat),?x:Nat) = max(max(min(max(min(?x:Nat,?z:Nat),?x:Nat),?y:Nat),?y:Nat),min(0,?z:Nat))",
"max(min(?x:Nat,?z:Nat),?x:Nat) = min(0,min(?z:Nat,min(max(min(?y:Nat,?x:Nat),min(?y:Nat,?y:Nat)),?x:Nat)))",
"max(min(?y:Nat,?x:Nat),?x:Nat) = min(min(0,min(min(?y:Nat,0),min(min(?y:Nat,0),?x:Nat))),0)",
"max(min(?y:Nat,?y:Nat),?x:Nat) = max(max(?z:Nat,max(?z:Nat,max(max(?y:Nat,?x:Nat),max(?x:Nat,?z:Nat)))),?x:Nat)",
"max(min(?y:Nat,?y:Nat),?y:Nat) = min(?x:Nat,max(?y:Nat,max(?x:Nat,max(min(?z:Nat,?y:Nat),min(0,?y:Nat)))))",
"max(min(?y:Nat,?z:Nat),?x:Nat) = max(max(?y:Nat,max(?y:Nat,min(0,?x:Nat))),max(max(?y:Nat,?y:Nat),?z:Nat))",
"max(min(?z:Nat,?x:Nat),0) = max(?z:Nat,min(max(0,?z:Nat),max(max(?z:Nat,0),min(?x:Nat,0))))",
"max(min(?z:Nat,?y:Nat),?x:Nat) = min(max(?x:Nat,max(?y:Nat,min(?x:Nat,max(?z:Nat,?y:Nat)))),min(?x:Nat,?x:Nat))",
"min(?x:Nat,max(?x:Nat,?x:Nat)) = max(max(min(min(min(?z:Nat,?x:Nat),?x:Nat),?x:Nat),min(0,?x:Nat)),0)",
"min(?y:Nat,min(?x:Nat,?y:Nat)) = max(max(min(min(?y:Nat,?z:Nat),?z:Nat),?x:Nat),min(min(?y:Nat,0),?x:Nat))",
"min(0,min(?z:Nat,?y:Nat)) = max(min(?x:Nat,max(min(max(?x:Nat,0),min(?x:Nat,0)),?x:Nat)),?x:Nat)",
"min(min(?y:Nat,0),?x:Nat) = min(max(?x:Nat,min(max(min(?x:Nat,0),min(0,?x:Nat)),?y:Nat)),?z:Nat)",
"max(?x:Nat,min(max(?y:Nat,?y:Nat),?y:Nat)) = max(0,min(min(0,max(?y:Nat,max(?z:Nat,?y:Nat))),?y:Nat))",
"max(?x:Nat,min(min(?y:Nat,?y:Nat),?y:Nat)) = max(0,min(?y:Nat,max(min(?x:Nat,min(?x:Nat,?x:Nat)),?y:Nat)))",
"max(?z:Nat,min(?y:Nat,max(?y:Nat,?y:Nat))) = max(min(min(?z:Nat,?x:Nat),max(max(0,?y:Nat),?x:Nat)),?x:Nat)",
"max(0,max(min(?x:Nat,?y:Nat),?y:Nat)) = max(max(max(?y:Nat,max(max(?y:Nat,?y:Nat),?y:Nat)),?y:Nat),?z:Nat)",
"max(0,min(?z:Nat,min(?z:Nat,?x:Nat))) = min(?y:Nat,max(?y:Nat,min(min(max(?y:Nat,?y:Nat),?x:Nat),?x:Nat)))",
"max(max(?x:Nat,?z:Nat),max(?y:Nat,?z:Nat)) = min(max(?y:Nat,min(?y:Nat,min(?y:Nat,?y:Nat))),max(?z:Nat,0))",
"max(max(?x:Nat,max(?y:Nat,0)),?x:Nat) = min(max(max(?x:Nat,?z:Nat),max(?y:Nat,min(?z:Nat,?x:Nat))),0)",
"max(max(?y:Nat,?x:Nat),min(?x:Nat,?y:Nat)) = min(?x:Nat,max(min(?x:Nat,min(0,max(?x:Nat,?y:Nat))),?y:Nat))",
"max(min(?y:Nat,max(?y:Nat,?x:Nat)),?z:Nat) = min(?x:Nat,min(min(min(?y:Nat,max(?x:Nat,?x:Nat)),?y:Nat),?x:Nat))",
"max(min(?z:Nat,min(?z:Nat,?y:Nat)),?x:Nat) = max(min(max(?y:Nat,?y:Nat),0),min(0,min(0,?y:Nat)))",
"max(min(min(?z:Nat,0),0),0) = min(max(max(min(min(?y:Nat,?z:Nat),?y:Nat),?x:Nat),?y:Nat),?x:Nat)",
"min(?x:Nat,max(?x:Nat,max(?z:Nat,?x:Nat))) = max(min(?z:Nat,max(max(?x:Nat,min(0,0)),0)),?y:Nat)",
"min(?x:Nat,max(max(0,?y:Nat),?y:Nat)) = max(?y:Nat,min(0,max(?y:Nat,min(?y:Nat,max(0,?y:Nat)))))",
"min(?x:Nat,min(max(?y:Nat,?z:Nat),?z:Nat)) = max(min(?x:Nat,?y:Nat),max(max(0,max(?y:Nat,?y:Nat)),0))",
"min(?y:Nat,max(0,max(?x:Nat,?y:Nat))) = max(min(?x:Nat,min(?x:Nat,0)),min(?y:Nat,max(?z:Nat,0)))",
"min(max(?y:Nat,min(?z:Nat,?y:Nat)),?x:Nat) = min(min(0,max(?x:Nat,min(0,min(0,0)))),0)",
"min(max(0,max(?z:Nat,?y:Nat)),?y:Nat) = min(max(max(max(?z:Nat,0),?x:Nat),?y:Nat),min(0,?x:Nat))",
"min(min(?x:Nat,?z:Nat),max(0,?x:Nat)) = max(?z:Nat,min(?y:Nat,min(max(?x:Nat,min(?y:Nat,?y:Nat)),?y:Nat)))",
"min(min(min(0,0),0),?z:Nat) = max(?x:Nat,max(max(?z:Nat,?x:Nat),max(?x:Nat,min(?x:Nat,?z:Nat))))",
"max(max(?y:Nat,?x:Nat),max(?z:Nat,min(0,0))) = min(0,min(min(?x:Nat,max(?x:Nat,?y:Nat)),?z:Nat))",
"max(min(min(?y:Nat,?x:Nat),max(?y:Nat,?y:Nat)),?x:Nat) = min(?z:Nat,min(?y:Nat,min(min(?z:Nat,?x:Nat),?x:Nat)))",
"max(min(min(0,min(?y:Nat,?x:Nat)),?z:Nat),?x:Nat) = max(max(?z:Nat,min(?y:Nat,?x:Nat)),max(?y:Nat,?z:Nat))",
"min(max(?x:Nat,0),min(?z:Nat,max(?z:Nat,?x:Nat))) = min(min(max(0,0),max(?x:Nat,?x:Nat)),?x:Nat)",
"min(max(0,min(?y:Nat,?y:Nat)),max(?y:Nat,?x:Nat)) = min(max(0,min(min(?x:Nat,?x:Nat),?z:Nat)),0)",
"min(max(min(?y:Nat,max(?z:Nat,?y:Nat)),?y:Nat),?x:Nat) = max(?z:Nat,max(?x:Nat,min(max(?x:Nat,?x:Nat),?y:Nat)))",
"max(?x:Nat,min(?x:Nat,min(max(?z:Nat,max(0,0)),0))) = max(?y:Nat,max(min(?z:Nat,?x:Nat),?y:Nat))",
"max(?x:Nat,min(min(?x:Nat,min(?y:Nat,max(?y:Nat,?x:Nat))),?y:Nat)) = max(?x:Nat,max(max(?x:Nat,?x:Nat),?y:Nat))",
"max(?y:Nat,min(max(max(?y:Nat,max(0,?y:Nat)),?y:Nat),?x:Nat)) = min(?x:Nat,min(?y:Nat,max(0,?y:Nat)))",
"max(max(?x:Nat,max(?x:Nat,min(min(?y:Nat,?z:Nat),?y:Nat))),?y:Nat) = min(?z:Nat,max(min(?x:Nat,?x:Nat),?z:Nat))",
"max(max(?x:Nat,min(0,0)),min(?y:Nat,max(?y:Nat,0))) = min(max(min(0,?y:Nat),?x:Nat),?x:Nat)",
"max(max(?x:Nat,min(min(max(?z:Nat,?y:Nat),?x:Nat),?y:Nat)),?x:Nat) = max(?y:Nat,max(?x:Nat,max(0,0)))",
"max(max(?y:Nat,?x:Nat),max(min(max(?y:Nat,0),?x:Nat),?x:Nat)) = max(max(?z:Nat,min(0,?y:Nat)),?y:Nat)",
"max(max(?y:Nat,max(0,?z:Nat)),min(min(?x:Nat,?y:Nat),?y:Nat)) = min(max(?z:Nat,0),min(?x:Nat,?y:Nat))",
"max(max(?y:Nat,min(?y:Nat,?y:Nat)),max(max(?y:Nat,0),?y:Nat)) = max(0,min(?y:Nat,min(?x:Nat,?x:Nat)))",
"max(max(max(0,?z:Nat),max(min(?z:Nat,?z:Nat),?y:Nat)),?x:Nat) = max(?x:Nat,min(min(?x:Nat,0),?z:Nat))",
"max(min(?y:Nat,min(?x:Nat,max(?x:Nat,min(?z:Nat,?y:Nat)))),0) = min(?x:Nat,min(?x:Nat,max(?z:Nat,?x:Nat)))",
"max(min(min(?z:Nat,?y:Nat),?y:Nat),min(?y:Nat,max(0,?x:Nat))) = min(?x:Nat,min(min(?x:Nat,?x:Nat),?y:Nat))",
"min(?y:Nat,max(?y:Nat,min(max(min(?z:Nat,0),?x:Nat),?x:Nat))) = min(?y:Nat,min(max(?x:Nat,?y:Nat),?z:Nat))",
"min(max(0,max(?y:Nat,max(min(?z:Nat,?y:Nat),?x:Nat))),?x:Nat) = min(max(min(?x:Nat,?y:Nat),?x:Nat),?z:Nat)",
"min(max(max(?y:Nat,min(?z:Nat,max(?x:Nat,?y:Nat))),?x:Nat),0) = min(min(?x:Nat,max(?x:Nat,?y:Nat)),?x:Nat)",
"min(min(max(?z:Nat,?y:Nat),min(?x:Nat,max(?z:Nat,?x:Nat))),?y:Nat) = min(min(min(?z:Nat,?z:Nat),?x:Nat),?y:Nat)",
"min(min(max(?z:Nat,?z:Nat),max(max(?y:Nat,?y:Nat),0)),?x:Nat) = max(max(?y:Nat,?y:Nat),max(0,?y:Nat))",
"max(?x:Nat,min(min(?x:Nat,min(min(?y:Nat,?y:Nat),min(?y:Nat,?x:Nat))),?x:Nat)) = max(?x:Nat,min(?y:Nat,?y:Nat))",
"max(?y:Nat,max(?x:Nat,max(0,min(max(?y:Nat,?z:Nat),max(?x:Nat,?y:Nat))))) = max(min(?x:Nat,?x:Nat),?x:Nat)",
"max(?y:Nat,max(min(?x:Nat,max(0,?y:Nat)),max(?x:Nat,min(0,0)))) = max(?y:Nat,min(?z:Nat,?y:Nat))",
"max(?z:Nat,min(?y:Nat,max(min(min(?z:Nat,?y:Nat),max(?y:Nat,?x:Nat)),0))) = max(min(?y:Nat,?x:Nat),?y:Nat)",
"max(max(?y:Nat,min(max(min(?y:Nat,?z:Nat),?y:Nat),min(?x:Nat,0))),?x:Nat) = max(?x:Nat,min(?y:Nat,?x:Nat))",
"max(max(max(max(max(?x:Nat,?z:Nat),max(?x:Nat,?y:Nat)),?z:Nat),?x:Nat),?y:Nat) = min(min(?x:Nat,?x:Nat),?z:Nat)",
"max(max(max(min(max(0,0),min(?x:Nat,?y:Nat)),?x:Nat),?y:Nat),0) = min(?y:Nat,min(?x:Nat,?y:Nat))",
"max(max(min(?x:Nat,min(max(?z:Nat,?y:Nat),max(?z:Nat,?y:Nat))),?x:Nat),0) = max(?y:Nat,max(?z:Nat,0))",
"max(min(?y:Nat,max(?x:Nat,min(max(?x:Nat,0),?y:Nat))),max(?z:Nat,?z:Nat)) = max(max(?y:Nat,0),?y:Nat)",
"max(min(0,?z:Nat),max(?y:Nat,max(?y:Nat,min(0,max(?x:Nat,?y:Nat))))) = max(max(?x:Nat,?x:Nat),?y:Nat)",
"max(min(max(?z:Nat,?x:Nat),max(?x:Nat,max(min(?x:Nat,0),?x:Nat))),?y:Nat) = min(max(0,0),?y:Nat)",
"max(min(max(0,?x:Nat),min(?x:Nat,max(?y:Nat,max(?y:Nat,?y:Nat)))),?x:Nat) = min(min(0,0),?y:Nat)",
"max(min(max(max(0,?y:Nat),?y:Nat),max(0,?y:Nat)),min(0,?y:Nat)) = min(max(0,?x:Nat),?x:Nat)",
"min(?x:Nat,min(?x:Nat,max(max(min(?x:Nat,?x:Nat),min(?y:Nat,?y:Nat)),?y:Nat))) = max(max(?y:Nat,?z:Nat),?x:Nat)",
"min(?y:Nat,max(min(?y:Nat,min(min(0,?y:Nat),max(?z:Nat,?x:Nat))),?y:Nat)) = max(?z:Nat,min(?x:Nat,?z:Nat))",
"min(?z:Nat,max(?y:Nat,max(min(0,?z:Nat),max(?y:Nat,max(?y:Nat,?y:Nat))))) = min(min(?x:Nat,?y:Nat),?x:Nat)",
"min(0,min(max(?y:Nat,max(min(?y:Nat,?x:Nat),max(?y:Nat,?z:Nat))),?y:Nat)) = max(?x:Nat,max(?y:Nat,0))",
"min(max(?y:Nat,min(?y:Nat,min(max(?y:Nat,?x:Nat),max(?y:Nat,?y:Nat)))),?y:Nat) = min(?z:Nat,min(?y:Nat,?y:Nat))",
"min(max(min(0,min(min(?x:Nat,?z:Nat),min(0,?z:Nat))),?x:Nat),0) = max(?x:Nat,max(?y:Nat,0))",
"min(max(min(min(?x:Nat,0),min(?y:Nat,?z:Nat)),0),max(0,?y:Nat)) = max(0,min(?y:Nat,?z:Nat))",
"min(min(?x:Nat,?x:Nat),min(0,max(?z:Nat,max(?x:Nat,min(?z:Nat,?z:Nat))))) = max(min(?x:Nat,?x:Nat),?y:Nat)",
"min(min(?x:Nat,min(min(?z:Nat,?x:Nat),min(min(?y:Nat,?y:Nat),?y:Nat))),?x:Nat) = min(max(?y:Nat,?z:Nat),?y:Nat)",
"min(min(?y:Nat,max(?z:Nat,0)),min(max(?x:Nat,?y:Nat),min(?z:Nat,0))) = max(max(?y:Nat,?x:Nat),?y:Nat)",
"min(min(?z:Nat,min(max(min(?z:Nat,?y:Nat),max(?y:Nat,?x:Nat)),0)),0) = min(?x:Nat,max(?x:Nat,?x:Nat))",
"min(min(0,min(max(?z:Nat,?z:Nat),min(?x:Nat,min(?x:Nat,?y:Nat)))),?x:Nat) = min(min(?y:Nat,?x:Nat),?y:Nat)",
"min(min(min(?x:Nat,min(?x:Nat,?y:Nat)),?y:Nat),min(max(?y:Nat,?y:Nat),?x:Nat)) = max(?y:Nat,min(?y:Nat,?x:Nat))",
"min(min(min(0,?y:Nat),?x:Nat),max(min(min(?x:Nat,0),?x:Nat),?x:Nat)) = min(?x:Nat,min(?y:Nat,?z:Nat))",
"max(?x:Nat,max(min(?y:Nat,?y:Nat),max(min(max(?y:Nat,?z:Nat),max(?y:Nat,?x:Nat)),?z:Nat))) = min(?x:Nat,?x:Nat)",
"max(?x:Nat,max(min(min(?y:Nat,max(?x:Nat,?x:Nat)),?y:Nat),max(?y:Nat,min(0,?x:Nat)))) = min(?y:Nat,?y:Nat)",
"max(?x:Nat,min(max(?z:Nat,?z:Nat),min(min(min(?x:Nat,?z:Nat),0),max(?x:Nat,?x:Nat)))) = min(?y:Nat,?y:Nat)",
"max(?y:Nat,max(max(min(?x:Nat,?y:Nat),?y:Nat),max(?y:Nat,min(?x:Nat,max(?z:Nat,?y:Nat))))) = min(?y:Nat,?x:Nat)",
"max(?y:Nat,max(min(max(min(?z:Nat,0),max(?x:Nat,?y:Nat)),0),max(?z:Nat,?x:Nat))) = min(?x:Nat,?x:Nat)",
"max(?y:Nat,max(min(min(?y:Nat,?x:Nat),?y:Nat),min(max(?z:Nat,0),min(?z:Nat,?z:Nat)))) = max(?y:Nat,?y:Nat)",
"max(?z:Nat,min(max(min(?z:Nat,?x:Nat),min(?y:Nat,?x:Nat)),min(?x:Nat,min(?z:Nat,?x:Nat)))) = max(?x:Nat,?y:Nat)",
"max(max(?x:Nat,min(?x:Nat,?x:Nat)),max(0,max(?y:Nat,max(max(?x:Nat,0),?y:Nat)))) = min(?x:Nat,0)",
"max(max(min(?x:Nat,?x:Nat),?y:Nat),max(min(max(max(?z:Nat,?x:Nat),?z:Nat),?x:Nat),0)) = max(?y:Nat,?z:Nat)",
"max(max(min(?y:Nat,?z:Nat),?x:Nat),min(?y:Nat,max(min(max(0,?z:Nat),?y:Nat),?x:Nat))) = max(?y:Nat,?y:Nat)",
"max(max(min(max(?z:Nat,?y:Nat),?z:Nat),?x:Nat),max(?z:Nat,max(?y:Nat,min(?z:Nat,?y:Nat)))) = max(?y:Nat,?y:Nat)",
"max(max(min(max(0,min(?x:Nat,0)),0),?z:Nat),max(?z:Nat,min(?y:Nat,?y:Nat))) = max(?x:Nat,?x:Nat)",
"max(max(min(max(min(0,?x:Nat),min(?y:Nat,?x:Nat)),?y:Nat),?y:Nat),max(?x:Nat,?z:Nat)) = max(?y:Nat,?y:Nat)",
"max(min(?z:Nat,max(max(?x:Nat,max(?x:Nat,?x:Nat)),max(?z:Nat,max(?y:Nat,?x:Nat)))),?x:Nat) = max(?y:Nat,?x:Nat)",
"max(min(max(?y:Nat,0),?x:Nat),max(min(?z:Nat,min(0,min(0,?z:Nat))),?y:Nat)) = max(?y:Nat,?z:Nat)",
"max(min(max(min(min(?y:Nat,?x:Nat),?x:Nat),min(0,max(?y:Nat,?x:Nat))),?y:Nat),?y:Nat) = min(?z:Nat,0)",
"max(min(min(?y:Nat,min(max(0,?y:Nat),max(?y:Nat,?y:Nat))),min(?y:Nat,?x:Nat)),?y:Nat) = max(?x:Nat,?y:Nat)",
"min(?z:Nat,max(max(?x:Nat,min(?y:Nat,min(?x:Nat,?z:Nat))),min(?x:Nat,min(?x:Nat,0)))) = max(?y:Nat,?y:Nat)",
"min(0,min(min(max(max(?x:Nat,?x:Nat),?x:Nat),?y:Nat),max(min(?z:Nat,?y:Nat),?x:Nat))) = max(?x:Nat,?x:Nat)",
"min(max(?y:Nat,min(max(min(0,0),max(?x:Nat,?x:Nat)),?x:Nat)),min(?y:Nat,?y:Nat)) = min(0,?z:Nat)",
"min(max(?z:Nat,?y:Nat),max(?z:Nat,min(?y:Nat,min(min(?y:Nat,0),min(?x:Nat,?x:Nat))))) = min(?y:Nat,?x:Nat)",
"min(max(max(max(?x:Nat,min(?y:Nat,?y:Nat)),min(?y:Nat,0)),max(?x:Nat,?x:Nat)),?z:Nat) = max(0,?x:Nat)",
"min(max(min(?x:Nat,max(?y:Nat,max(?x:Nat,?x:Nat))),min(max(?y:Nat,?z:Nat),?y:Nat)),?y:Nat) = min(?y:Nat,0)",
"min(max(min(min(max(?z:Nat,?y:Nat),max(?z:Nat,?z:Nat)),0),max(?y:Nat,?x:Nat)),?y:Nat) = max(0,?x:Nat)",
"min(min(?x:Nat,max(max(max(?x:Nat,0),?x:Nat),min(?x:Nat,min(?x:Nat,?y:Nat)))),?x:Nat) = max(?x:Nat,?z:Nat)",
"min(min(?y:Nat,?x:Nat),max(max(?x:Nat,max(min(0,?x:Nat),max(?x:Nat,0))),?y:Nat)) = min(?y:Nat,0)",
"min(min(?z:Nat,min(?y:Nat,max(?x:Nat,?x:Nat))),max(min(min(?z:Nat,?y:Nat),?z:Nat),?y:Nat)) = max(?z:Nat,?y:Nat)",
"min(min(min(max(?y:Nat,?x:Nat),max(min(?x:Nat,?x:Nat),min(?y:Nat,?x:Nat))),0),?x:Nat) = min(?x:Nat,?z:Nat)",
"min(min(min(min(?y:Nat,0),max(min(?x:Nat,0),max(?x:Nat,?x:Nat))),?x:Nat),0) = max(?x:Nat,?x:Nat)",
"min(min(min(min(?z:Nat,max(0,?x:Nat)),min(?x:Nat,0)),min(?y:Nat,?x:Nat)),?y:Nat) = max(?z:Nat,?x:Nat)",
"max(?x:Nat,?x:Nat) = min(min(max(?y:Nat,?z:Nat),max(min(?y:Nat,max(?x:Nat,?y:Nat)),max(?x:Nat,?x:Nat))),max(?x:Nat,?y:Nat))",
"max(?x:Nat,0) = max(min(?x:Nat,min(?z:Nat,?y:Nat)),max(?y:Nat,max(min(min(?z:Nat,0),?z:Nat),max(?y:Nat,?z:Nat))))",
"max(?y:Nat,?x:Nat) = min(max(0,max(min(min(?x:Nat,?z:Nat),0),min(?y:Nat,max(?z:Nat,?x:Nat)))),min(?x:Nat,?x:Nat))",
"max(?y:Nat,?y:Nat) = max(max(?z:Nat,?x:Nat),min(0,max(min(max(?x:Nat,?z:Nat),0),min(?z:Nat,min(?y:Nat,?z:Nat)))))",
"max(?y:Nat,?z:Nat) = min(0,min(?y:Nat,max(min(max(?y:Nat,?z:Nat),min(?x:Nat,?x:Nat)),min(?y:Nat,max(?x:Nat,?x:Nat)))))",
"max(?z:Nat,?x:Nat) = min(max(?x:Nat,min(max(0,?x:Nat),max(min(0,?z:Nat),?z:Nat))),max(max(?y:Nat,0),?x:Nat))",
"max(?z:Nat,0) = max(max(max(max(min(0,0),?y:Nat),0),max(?x:Nat,min(?x:Nat,max(?x:Nat,?y:Nat)))),?y:Nat)",
"max(0,?z:Nat) = max(min(max(min(?z:Nat,?y:Nat),min(min(?y:Nat,?x:Nat),max(?x:Nat,?x:Nat))),max(?z:Nat,0)),?y:Nat)",
"max(0,0) = min(?y:Nat,min(min(?y:Nat,min(?z:Nat,?y:Nat)),max(?x:Nat,min(min(?y:Nat,0),min(?y:Nat,?y:Nat)))))",
"min(?x:Nat,?x:Nat) = max(min(?z:Nat,min(max(?y:Nat,min(0,?y:Nat)),min(0,?x:Nat))),min(?y:Nat,max(0,?x:Nat)))",
"min(?x:Nat,?y:Nat) = max(min(?x:Nat,min(max(max(?y:Nat,?y:Nat),max(?x:Nat,?y:Nat)),max(?x:Nat,0))),min(?x:Nat,?y:Nat))",
"min(?y:Nat,?x:Nat) = min(?y:Nat,min(max(?y:Nat,min(min(?x:Nat,?x:Nat),0)),max(?x:Nat,max(min(?x:Nat,?z:Nat),?y:Nat))))",
"min(?y:Nat,?x:Nat) = min(max(?y:Nat,?x:Nat),max(max(?y:Nat,min(0,max(?y:Nat,?z:Nat))),max(?x:Nat,max(?x:Nat,?y:Nat))))",
"min(?z:Nat,?x:Nat) = min(max(?z:Nat,min(min(max(?y:Nat,?y:Nat),max(0,?x:Nat)),min(min(0,?z:Nat),?y:Nat))),?y:Nat)",
"min(?z:Nat,?y:Nat) = min(max(max(min(?x:Nat,min(?x:Nat,?x:Nat)),?z:Nat),min(min(min(?z:Nat,?z:Nat),?x:Nat),?y:Nat)),?z:Nat)",
"min(0,?z:Nat) = max(?y:Nat,max(min(?x:Nat,min(?x:Nat,min(?y:Nat,?x:Nat))),min(min(?x:Nat,min(0,0)),?x:Nat)))",
"max(?x:Nat,min(0,?x:Nat)) = min(?x:Nat,min(max(?z:Nat,min(max(?y:Nat,?y:Nat),0)),max(?z:Nat,max(?z:Nat,?y:Nat))))",
"max(0,max(?y:Nat,?z:Nat)) = max(?x:Nat,min(max(max(?y:Nat,?y:Nat),0),max(?x:Nat,max(?y:Nat,min(?x:Nat,?y:Nat)))))",
"max(max(?y:Nat,?x:Nat),?y:Nat) = min(max(min(?x:Nat,max(max(?z:Nat,0),0)),max(min(?x:Nat,?y:Nat),?y:Nat)),?x:Nat)",
"max(min(0,?y:Nat),?y:Nat) = min(min(min(?x:Nat,min(max(?y:Nat,?z:Nat),?x:Nat)),?z:Nat),min(max(?y:Nat,?y:Nat),?x:Nat))",
"min(?y:Nat,max(?y:Nat,?y:Nat)) = min(0,max(min(?y:Nat,max(min(?x:Nat,?y:Nat),max(?y:Nat,?y:Nat))),min(?x:Nat,?x:Nat)))",
"min(?y:Nat,max(?z:Nat,?x:Nat)) = min(min(?x:Nat,?x:Nat),min(?z:Nat,min(max(?x:Nat,min(?x:Nat,?y:Nat)),max(?x:Nat,?z:Nat))))",
"min(max(?x:Nat,?y:Nat),?y:Nat) = max(?y:Nat,max(?x:Nat,max(max(?z:Nat,min(?x:Nat,0)),min(?z:Nat,max(0,?y:Nat)))))",
"min(max(?x:Nat,?y:Nat),?z:Nat) = min(min(min(max(0,max(?x:Nat,?y:Nat)),0),?x:Nat),max(max(?y:Nat,?x:Nat),0))",
"min(max(?x:Nat,?z:Nat),?x:Nat) = min(?x:Nat,min(?x:Nat,min(min(?y:Nat,?y:Nat),max(min(?x:Nat,?x:Nat),min(0,?x:Nat)))))",
"min(min(?x:Nat,?x:Nat),?x:Nat) = min(min(min(min(?y:Nat,?x:Nat),?x:Nat),min(?x:Nat,max(max(?x:Nat,?y:Nat),0))),?x:Nat)",
"max(?x:Nat,max(max(?y:Nat,?x:Nat),?z:Nat)) = max(min(max(?y:Nat,min(min(?y:Nat,?y:Nat),?x:Nat)),min(0,?x:Nat)),?y:Nat)",
"max(?y:Nat,max(min(?y:Nat,0),?z:Nat)) = min(min(min(0,min(max(?x:Nat,?x:Nat),?y:Nat)),min(?z:Nat,?x:Nat)),?x:Nat)",
"max(?y:Nat,max(min(?z:Nat,0),?x:Nat)) = max(min(min(max(?z:Nat,max(?y:Nat,0)),min(?z:Nat,?y:Nat)),?z:Nat),?x:Nat)",
"max(?y:Nat,min(?z:Nat,min(?z:Nat,?y:Nat))) = max(min(max(max(?z:Nat,?y:Nat),min(?y:Nat,min(0,0))),?x:Nat),0)",
"max(?y:Nat,min(max(?x:Nat,?x:Nat),?x:Nat)) = max(?x:Nat,max(min(?z:Nat,min(min(?x:Nat,?x:Nat),min(0,?z:Nat))),?x:Nat))",
"max(?z:Nat,min(max(0,?x:Nat),?x:Nat)) = max(?y:Nat,max(?y:Nat,min(min(max(?x:Nat,?y:Nat),max(?x:Nat,?y:Nat)),0)))",
"max(0,max(?y:Nat,min(?x:Nat,?z:Nat))) = min(?y:Nat,min(?x:Nat,max(max(min(0,?y:Nat),min(0,?z:Nat)),?x:Nat)))",
"max(max(0,?x:Nat),max(?x:Nat,?y:Nat)) = max(min(max(?x:Nat,?x:Nat),max(max(?x:Nat,?y:Nat),min(0,0))),?x:Nat)",
"max(min(?z:Nat,min(?x:Nat,?y:Nat)),?z:Nat) = min(?y:Nat,max(max(min(min(?z:Nat,?x:Nat),0),min(?z:Nat,?z:Nat)),?x:Nat))",
"max(min(min(?x:Nat,?x:Nat),?x:Nat),?x:Nat) = max(max(max(max(max(?x:Nat,?x:Nat),max(?x:Nat,?y:Nat)),0),?x:Nat),?y:Nat)",
"min(?x:Nat,max(max(?x:Nat,?z:Nat),?x:Nat)) = min(max(min(min(?y:Nat,?y:Nat),min(?y:Nat,max(?y:Nat,?x:Nat))),?z:Nat),0)",
"min(?y:Nat,max(?y:Nat,min(?y:Nat,?y:Nat))) = min(min(?x:Nat,?x:Nat),min(?x:Nat,max(?z:Nat,max(?y:Nat,min(?x:Nat,?x:Nat)))))",
"min(?y:Nat,max(0,min(?y:Nat,?x:Nat))) = max(min(?x:Nat,max(max(?y:Nat,min(?x:Nat,?y:Nat)),min(0,0))),?y:Nat)",
"min(?y:Nat,min(?y:Nat,max(0,0))) = max(?x:Nat,min(max(?x:Nat,max(min(?z:Nat,?y:Nat),min(?x:Nat,0))),?z:Nat))",
"min(?y:Nat,min(min(0,?z:Nat),?y:Nat)) = max(?x:Nat,max(max(?z:Nat,0),min(max(?y:Nat,?z:Nat),min(?z:Nat,?x:Nat))))",
"min(max(?z:Nat,?y:Nat),max(?y:Nat,0)) = max(?z:Nat,min(min(?y:Nat,?y:Nat),min(min(max(?y:Nat,?z:Nat),?x:Nat),?x:Nat)))",
"min(min(max(?y:Nat,?x:Nat),0),?x:Nat) = min(?z:Nat,min(max(?y:Nat,min(0,?y:Nat)),max(max(?x:Nat,0),?z:Nat)))",
"min(min(min(?x:Nat,?y:Nat),0),0) = min(max(?x:Nat,?x:Nat),max(min(?y:Nat,?y:Nat),max(?y:Nat,min(?z:Nat,?y:Nat))))",
"max(?x:Nat,min(max(?x:Nat,min(?x:Nat,?y:Nat)),?z:Nat)) = max(?x:Nat,max(?x:Nat,max(?x:Nat,max(?y:Nat,min(?z:Nat,?x:Nat)))))",
"max(0,min(min(?x:Nat,max(?z:Nat,?z:Nat)),?y:Nat)) = max(max(?x:Nat,min(min(?x:Nat,min(?x:Nat,?x:Nat)),?z:Nat)),?x:Nat)",
"max(max(?z:Nat,min(?y:Nat,0)),min(?y:Nat,?z:Nat)) = min(min(?y:Nat,max(0,max(max(?x:Nat,0),0))),?z:Nat)",
"max(min(?x:Nat,max(max(?y:Nat,?y:Nat),?y:Nat)),?z:Nat) = min(?y:Nat,min(?y:Nat,max(min(?y:Nat,max(?x:Nat,?z:Nat)),0)))",
"min(?x:Nat,min(?y:Nat,min(min(0,?y:Nat),?x:Nat))) = min(max(max(?y:Nat,max(min(?x:Nat,?x:Nat),?y:Nat)),0),?z:Nat)",
"min(?x:Nat,min(max(min(?x:Nat,?z:Nat),?z:Nat),?z:Nat)) = max(?z:Nat,min(?y:Nat,min(?x:Nat,min(min(?z:Nat,?y:Nat),?y:Nat))))",
"min(0,max(?z:Nat,min(max(?y:Nat,?y:Nat),0))) = max(0,max(?y:Nat,max(max(max(0,?z:Nat),?y:Nat),?x:Nat)))",
"min(min(0,max(?y:Nat,min(?z:Nat,?y:Nat))),?y:Nat) = min(?x:Nat,min(min(?x:Nat,min(?z:Nat,?y:Nat)),max(?z:Nat,?z:Nat)))",
"min(min(max(min(0,0),?x:Nat),?y:Nat),?z:Nat) = max(?y:Nat,max(max(min(?z:Nat,min(?y:Nat,?y:Nat)),?z:Nat),?x:Nat))",
"max(?x:Nat,max(min(max(min(?x:Nat,?x:Nat),?y:Nat),?x:Nat),?x:Nat)) = min(min(?z:Nat,max(?x:Nat,?x:Nat)),min(?x:Nat,?y:Nat))",
"max(?x:Nat,min(min(?x:Nat,max(min(?x:Nat,0),0)),?y:Nat)) = max(max(?y:Nat,min(?y:Nat,min(?x:Nat,?z:Nat))),?z:Nat)",
"max(?z:Nat,max(max(?y:Nat,max(?y:Nat,min(?y:Nat,?x:Nat))),?y:Nat)) = min(max(min(0,max(0,?z:Nat)),?x:Nat),?z:Nat)",
"max(?z:Nat,max(max(max(max(?x:Nat,?x:Nat),?x:Nat),?z:Nat),?x:Nat)) = max(?x:Nat,min(?y:Nat,min(max(?x:Nat,?x:Nat),?x:Nat)))",
"max(?z:Nat,max(min(0,max(min(?y:Nat,?x:Nat),?y:Nat)),?y:Nat)) = max(?z:Nat,min(min(?x:Nat,?y:Nat),min(0,?y:Nat)))",
"max(0,max(0,max(?z:Nat,min(?y:Nat,min(?x:Nat,?z:Nat))))) = max(max(max(?x:Nat,0),min(?y:Nat,?x:Nat)),?x:Nat)",
"max(min(?x:Nat,?x:Nat),max(max(?y:Nat,0),max(?x:Nat,?y:Nat))) = min(max(min(?y:Nat,?y:Nat),?x:Nat),min(0,0))",
"min(min(?x:Nat,max(?x:Nat,min(?x:Nat,max(?y:Nat,?x:Nat)))),?z:Nat) = min(max(max(?y:Nat,min(?z:Nat,?x:Nat)),0),?x:Nat)",
"max(?x:Nat,max(max(?x:Nat,?y:Nat),max(max(max(?z:Nat,?y:Nat),?z:Nat),?y:Nat))) = max(?z:Nat,min(?z:Nat,max(?y:Nat,?x:Nat)))",
"max(?y:Nat,max(max(min(?z:Nat,max(?y:Nat,0)),min(?x:Nat,?x:Nat)),?x:Nat)) = min(?y:Nat,min(max(?y:Nat,?y:Nat),?z:Nat))",
"max(?z:Nat,max(min(min(max(?x:Nat,?y:Nat),?y:Nat),max(0,?z:Nat)),0)) = min(max(?y:Nat,max(?z:Nat,?z:Nat)),?y:Nat)",
"max(max(?x:Nat,max(?x:Nat,max(max(0,0),?y:Nat))),max(0,?y:Nat)) = min(?x:Nat,max(min(?z:Nat,?x:Nat),?y:Nat))",
"max(max(?x:Nat,min(max(min(?x:Nat,?z:Nat),0),max(?x:Nat,?z:Nat))),?y:Nat) = min(max(?y:Nat,?y:Nat),max(?x:Nat,?y:Nat))",
"max(min(?x:Nat,max(min(0,?z:Nat),max(0,min(?x:Nat,?x:Nat)))),?y:Nat) = min(?y:Nat,min(0,max(?y:Nat,?x:Nat)))",
"max(min(max(?x:Nat,?x:Nat),min(max(max(0,?x:Nat),?y:Nat),?x:Nat)),0) = min(max(?x:Nat,?x:Nat),min(?x:Nat,?y:Nat))",
"max(min(min(max(?y:Nat,?x:Nat),max(?x:Nat,max(?x:Nat,?z:Nat))),?z:Nat),?y:Nat) = max(max(0,?z:Nat),max(?x:Nat,0))",
"min(?x:Nat,max(max(?y:Nat,min(max(?z:Nat,0),max(0,?x:Nat))),?y:Nat)) = max(max(?y:Nat,0),max(?y:Nat,?x:Nat))",
"min(min(?y:Nat,min(max(min(?x:Nat,?z:Nat),?x:Nat),?z:Nat)),min(?y:Nat,?y:Nat)) = min(max(?x:Nat,max(?y:Nat,0)),0)",
"max(?y:Nat,min(max(?y:Nat,?z:Nat),min(min(?y:Nat,?x:Nat),max(max(?z:Nat,?y:Nat),?x:Nat)))) = max(max(?z:Nat,?y:Nat),?x:Nat)",
"max(max(?x:Nat,min(min(min(?x:Nat,?x:Nat),min(?x:Nat,0)),min(?z:Nat,?y:Nat))),?y:Nat) = max(max(?y:Nat,0),?z:Nat)",
"max(max(0,min(?z:Nat,?x:Nat)),max(max(max(max(?y:Nat,?y:Nat),?z:Nat),?y:Nat),0)) = max(min(?z:Nat,?y:Nat),?x:Nat)",
"max(max(min(max(?x:Nat,min(?z:Nat,?x:Nat)),max(?y:Nat,?y:Nat)),max(?x:Nat,?x:Nat)),?x:Nat) = max(max(?y:Nat,?y:Nat),?y:Nat)",
"max(min(?y:Nat,max(0,min(min(?y:Nat,?z:Nat),?z:Nat))),max(?y:Nat,min(0,?x:Nat))) = max(0,min(?y:Nat,?x:Nat))",
"max(min(max(?x:Nat,?x:Nat),?x:Nat),min(min(?x:Nat,min(?x:Nat,max(?x:Nat,?z:Nat))),?y:Nat)) = min(max(?x:Nat,?y:Nat),?z:Nat)",
"max(min(min(max(0,max(?z:Nat,?y:Nat)),?x:Nat),?z:Nat),min(?y:Nat,max(?x:Nat,0))) = min(min(?y:Nat,?x:Nat),?x:Nat)",
"min(?x:Nat,max(min(?x:Nat,?z:Nat),max(0,min(max(?x:Nat,?y:Nat),min(?y:Nat,?z:Nat))))) = min(min(?y:Nat,0),?y:Nat)",
"min(?x:Nat,max(min(min(?x:Nat,min(?x:Nat,?x:Nat)),?y:Nat),min(?y:Nat,min(?y:Nat,?x:Nat)))) = max(min(0,?y:Nat),?z:Nat)",
"min(?x:Nat,max(min(min(0,min(?x:Nat,?z:Nat)),min(min(?y:Nat,0),?y:Nat)),?y:Nat)) = max(max(0,?y:Nat),?y:Nat)",
"min(?x:Nat,min(max(0,0),max(max(?x:Nat,?x:Nat),max(min(0,?y:Nat),?z:Nat)))) = max(?x:Nat,min(?y:Nat,0))",
"min(?y:Nat,max(max(?x:Nat,0),min(max(?z:Nat,?x:Nat),min(?y:Nat,min(?z:Nat,?x:Nat))))) = max(?z:Nat,max(?x:Nat,?y:Nat))",
"max(?x:Nat,min(max(min(min(?x:Nat,?x:Nat),?z:Nat),?z:Nat),max(0,min(?x:Nat,min(?z:Nat,?x:Nat))))) = max(?x:Nat,?z:Nat)",
"max(?y:Nat,max(min(?x:Nat,min(max(?y:Nat,?y:Nat),?x:Nat)),max(min(?y:Nat,?z:Nat),min(?x:Nat,?y:Nat)))) = min(?x:Nat,?y:Nat)",
"max(0,min(min(max(max(?y:Nat,?z:Nat),min(?y:Nat,0)),?x:Nat),max(0,min(?y:Nat,?y:Nat)))) = max(?x:Nat,0)",
"max(max(?y:Nat,min(?x:Nat,min(max(0,?x:Nat),min(?x:Nat,?z:Nat)))),max(?z:Nat,min(?z:Nat,?x:Nat))) = max(0,?x:Nat)",
"max(max(?y:Nat,min(min(0,max(?x:Nat,?x:Nat)),max(min(?x:Nat,?x:Nat),?x:Nat))),min(?y:Nat,?y:Nat)) = max(?x:Nat,?x:Nat)",
"max(max(?z:Nat,max(min(?x:Nat,min(?z:Nat,0)),max(?y:Nat,min(?y:Nat,0)))),min(?y:Nat,?x:Nat)) = min(?y:Nat,?x:Nat)",
"max(max(0,min(?y:Nat,?x:Nat)),min(0,max(min(?y:Nat,0),max(max(?z:Nat,?x:Nat),?y:Nat)))) = max(?x:Nat,?x:Nat)",
"max(max(min(?x:Nat,?y:Nat),min(min(min(?x:Nat,?y:Nat),min(?x:Nat,?x:Nat)),max(?z:Nat,?z:Nat))),?z:Nat) = min(?y:Nat,?y:Nat)",
"max(max(min(max(min(?y:Nat,?y:Nat),min(?y:Nat,?x:Nat)),0),?x:Nat),min(min(0,?x:Nat),?y:Nat)) = min(?z:Nat,0)",
"max(min(max(?y:Nat,?x:Nat),min(min(?y:Nat,?x:Nat),?x:Nat)),max(max(min(?z:Nat,?x:Nat),?x:Nat),?x:Nat)) = max(?y:Nat,?y:Nat)",
"max(min(max(?y:Nat,min(?y:Nat,?x:Nat)),?y:Nat),max(max(?y:Nat,max(?y:Nat,?y:Nat)),max(0,?x:Nat))) = max(?z:Nat,?y:Nat)",
"max(min(max(?z:Nat,0),min(min(0,min(?x:Nat,0)),min(min(?z:Nat,?x:Nat),?z:Nat))),?y:Nat) = min(0,0)",
"max(min(max(0,min(?x:Nat,?y:Nat)),?y:Nat),min(0,max(max(0,0),max(?x:Nat,?x:Nat)))) = min(?z:Nat,?y:Nat)",
"min(?x:Nat,max(?x:Nat,min(max(max(?x:Nat,?y:Nat),min(?y:Nat,?z:Nat)),max(min(?y:Nat,?x:Nat),?x:Nat)))) = max(?x:Nat,?y:Nat)",
"min(?x:Nat,max(max(max(?z:Nat,max(?y:Nat,?y:Nat)),?x:Nat),max(min(max(0,0),?z:Nat),?y:Nat))) = min(0,0)",
"min(?y:Nat,max(min(?y:Nat,max(?x:Nat,min(?z:Nat,?y:Nat))),min(min(0,min(?y:Nat,?x:Nat)),?z:Nat))) = min(?y:Nat,?x:Nat)",
"min(?y:Nat,min(0,max(max(min(0,?x:Nat),max(?x:Nat,0)),max(min(?z:Nat,?x:Nat),?y:Nat)))) = min(?x:Nat,?y:Nat)",
"min(?y:Nat,min(min(min(?x:Nat,?y:Nat),min(min(0,0),?x:Nat)),max(max(?x:Nat,?y:Nat),0))) = max(?x:Nat,0)",
"min(?z:Nat,min(min(max(max(?y:Nat,?x:Nat),?z:Nat),min(?x:Nat,?x:Nat)),min(max(?x:Nat,?x:Nat),?y:Nat))) = max(?y:Nat,?z:Nat)",
"min(0,min(min(?y:Nat,max(?y:Nat,min(?y:Nat,0))),min(min(0,max(?y:Nat,0)),?x:Nat))) = min(?x:Nat,?x:Nat)",
"min(max(max(min(?y:Nat,?x:Nat),max(min(?x:Nat,?y:Nat),?y:Nat)),?x:Nat),max(?y:Nat,min(?x:Nat,?x:Nat))) = min(?y:Nat,0)",
"min(max(min(0,?y:Nat),min(?x:Nat,min(?z:Nat,0))),max(min(?y:Nat,?y:Nat),min(?x:Nat,?z:Nat))) = min(?y:Nat,?y:Nat)",
"min(min(max(max(min(?x:Nat,?x:Nat),0),max(max(0,?y:Nat),0)),max(?z:Nat,0)),?x:Nat) = max(0,?y:Nat)",
"min(min(min(max(?z:Nat,?z:Nat),0),?x:Nat),max(?z:Nat,max(?x:Nat,min(max(?y:Nat,?z:Nat),?z:Nat)))) = min(?x:Nat,?z:Nat)",
"max(?x:Nat,?x:Nat) = max(max(max(0,?x:Nat),max(min(?y:Nat,max(?y:Nat,?x:Nat)),max(min(?x:Nat,0),max(?y:Nat,?x:Nat)))),?z:Nat)",
"max(?x:Nat,?x:Nat) = min(min(?x:Nat,?y:Nat),max(max(max(?y:Nat,min(?y:Nat,?x:Nat)),?z:Nat),min(min(max(?y:Nat,?z:Nat),?y:Nat),?x:Nat)))",
"max(?x:Nat,?x:Nat) = min(min(0,min(max(0,?z:Nat),max(min(0,?x:Nat),?y:Nat))),max(max(min(?y:Nat,?y:Nat),0),0))",
"max(?x:Nat,?y:Nat) = min(min(min(?y:Nat,0),?y:Nat),max(max(min(min(0,0),max(?x:Nat,?y:Nat)),?x:Nat),max(?z:Nat,?x:Nat)))",
"max(?x:Nat,?z:Nat) = max(min(min(min(?x:Nat,min(?x:Nat,?z:Nat)),?z:Nat),?x:Nat),min(max(max(?z:Nat,?x:Nat),?x:Nat),max(?z:Nat,?z:Nat)))",
"max(?y:Nat,?x:Nat) = max(max(min(?z:Nat,?y:Nat),max(?z:Nat,?z:Nat)),max(max(?y:Nat,max(min(0,?x:Nat),min(0,?y:Nat))),?x:Nat))",
"max(?y:Nat,?z:Nat) = min(min(?y:Nat,min(min(?y:Nat,0),max(0,max(?x:Nat,?x:Nat)))),max(min(?y:Nat,?y:Nat),min(0,?y:Nat)))",
"max(?z:Nat,?x:Nat) = min(?x:Nat,max(min(?x:Nat,max(max(0,?x:Nat),min(0,?z:Nat))),max(min(min(?y:Nat,?x:Nat),?x:Nat),?y:Nat)))",
"max(?z:Nat,?x:Nat) = min(?z:Nat,max(max(?x:Nat,?x:Nat),max(min(max(?x:Nat,?x:Nat),max(?y:Nat,?z:Nat)),max(?x:Nat,max(0,?z:Nat)))))",
"max(?z:Nat,0) = min(min(?x:Nat,max(?x:Nat,min(min(0,?z:Nat),min(?x:Nat,0)))),max(?y:Nat,max(min(?x:Nat,?x:Nat),?x:Nat)))",
"max(0,?z:Nat) = min(max(?x:Nat,?x:Nat),min(max(?x:Nat,max(?x:Nat,?y:Nat)),min(max(min(?y:Nat,?y:Nat),min(?y:Nat,?x:Nat)),?y:Nat)))",
"min(?x:Nat,?y:Nat) = max(min(max(max(min(?y:Nat,?x:Nat),0),max(max(?y:Nat,?y:Nat),min(?y:Nat,?y:Nat))),?y:Nat),max(0,?x:Nat))",
"min(?x:Nat,?z:Nat) = min(max(min(max(?x:Nat,?z:Nat),max(?y:Nat,?y:Nat)),max(max(max(?z:Nat,?x:Nat),?x:Nat),0)),max(?z:Nat,?x:Nat))",
"min(?x:Nat,0) = min(max(0,max(min(?x:Nat,max(?y:Nat,?x:Nat)),?y:Nat)),max(min(min(?x:Nat,?y:Nat),min(?x:Nat,?x:Nat)),?x:Nat))",
"min(?y:Nat,?x:Nat) = max(max(min(?x:Nat,?x:Nat),min(?y:Nat,min(0,?z:Nat))),min(min(?y:Nat,min(?x:Nat,min(?y:Nat,?y:Nat))),0))",
"min(?y:Nat,?x:Nat) = min(max(?z:Nat,max(0,0)),max(min(min(?y:Nat,?x:Nat),?z:Nat),max(?x:Nat,max(min(?z:Nat,?x:Nat),?x:Nat))))",
"min(?y:Nat,?y:Nat) = max(min(max(?x:Nat,min(min(?y:Nat,0),max(?y:Nat,?y:Nat))),max(?z:Nat,?x:Nat)),min(?y:Nat,max(?z:Nat,0)))",
"min(?z:Nat,?x:Nat) = min(max(max(max(?z:Nat,max(0,?z:Nat)),?z:Nat),max(max(max(?x:Nat,?z:Nat),max(?z:Nat,?y:Nat)),0)),?z:Nat)",
"min(?z:Nat,?x:Nat) = min(min(?z:Nat,min(?y:Nat,max(min(?y:Nat,?y:Nat),?x:Nat))),min(max(max(?x:Nat,?y:Nat),max(?x:Nat,0)),?y:Nat))",
"min(?z:Nat,0) = max(max(0,min(0,min(min(?z:Nat,?x:Nat),max(?x:Nat,?y:Nat)))),max(min(min(?y:Nat,?x:Nat),?x:Nat),?x:Nat))",
"min(0,?x:Nat) = max(?y:Nat,max(min(min(?y:Nat,min(?z:Nat,?y:Nat)),?z:Nat),max(?y:Nat,min(min(?z:Nat,?y:Nat),max(?x:Nat,?x:Nat)))))",
"min(0,?y:Nat) = max(max(max(min(?x:Nat,min(?y:Nat,?x:Nat)),?y:Nat),min(max(max(?x:Nat,?x:Nat),max(?y:Nat,?y:Nat)),0)),?y:Nat)",
"min(0,?y:Nat) = max(min(min(max(?y:Nat,?z:Nat),max(max(?x:Nat,?x:Nat),?y:Nat)),min(max(?z:Nat,min(0,?x:Nat)),0)),?y:Nat)",
"max(?z:Nat,min(0,?y:Nat)) = min(max(max(?z:Nat,max(?x:Nat,min(?y:Nat,?x:Nat))),?z:Nat),max(?x:Nat,max(?z:Nat,min(?y:Nat,0))))",
"max(max(?x:Nat,?x:Nat),?y:Nat) = min(max(?y:Nat,min(max(?y:Nat,max(?x:Nat,?x:Nat)),min(min(?y:Nat,?z:Nat),0))),max(?x:Nat,?y:Nat))",
"min(?z:Nat,max(?y:Nat,?y:Nat)) = min(min(max(min(min(0,?x:Nat),min(?y:Nat,0)),0),min(?x:Nat,?x:Nat)),max(?y:Nat,?y:Nat))",
"min(max(?z:Nat,?z:Nat),?z:Nat) = max(?y:Nat,min(max(?x:Nat,min(?y:Nat,min(0,?y:Nat))),max(0,max(?y:Nat,max(?y:Nat,?y:Nat)))))",
"min(max(0,?z:Nat),0) = max(max(min(max(0,min(?z:Nat,0)),min(?x:Nat,?z:Nat)),min(?z:Nat,?y:Nat)),min(?x:Nat,?x:Nat))",
"min(min(?y:Nat,?x:Nat),?x:Nat) = max(?y:Nat,max(min(min(?y:Nat,max(?y:Nat,?y:Nat)),max(max(?y:Nat,?z:Nat),min(?y:Nat,?x:Nat))),?x:Nat))",
"min(min(?z:Nat,?z:Nat),?y:Nat) = max(min(0,?x:Nat),max(max(min(min(?y:Nat,?x:Nat),min(?y:Nat,?x:Nat)),?y:Nat),max(?z:Nat,?z:Nat)))",
"max(max(?y:Nat,max(?y:Nat,0)),?x:Nat) = max(min(max(max(?x:Nat,?z:Nat),max(min(?y:Nat,?x:Nat),?y:Nat)),?x:Nat),min(0,0))",
"max(max(0,max(?y:Nat,?y:Nat)),0) = min(max(min(?z:Nat,min(0,max(?x:Nat,?y:Nat))),min(?x:Nat,?y:Nat)),min(?x:Nat,?x:Nat))",
"max(min(?x:Nat,min(?x:Nat,?y:Nat)),?z:Nat) = min(?y:Nat,max(max(min(?x:Nat,?y:Nat),max(?y:Nat,?y:Nat)),min(?x:Nat,max(?x:Nat,?y:Nat))))",
"max(min(?z:Nat,?y:Nat),min(?x:Nat,?y:Nat)) = max(max(?x:Nat,?x:Nat),min(?x:Nat,max(min(max(0,0),max(0,?y:Nat)),?x:Nat)))",
"max(min(max(?x:Nat,?y:Nat),?x:Nat),?y:Nat) = min(min(min(max(?y:Nat,?z:Nat),?x:Nat),?y:Nat),max(max(max(?y:Nat,?z:Nat),?y:Nat),0))",
"min(max(max(?y:Nat,?x:Nat),?z:Nat),0) = max(max(min(max(?x:Nat,min(?y:Nat,?y:Nat)),min(min(?z:Nat,?x:Nat),?z:Nat)),?z:Nat),?y:Nat)",
"min(min(?y:Nat,?z:Nat),max(0,?x:Nat)) = min(?y:Nat,min(max(min(?x:Nat,max(?y:Nat,0)),max(max(?z:Nat,?y:Nat),?y:Nat)),?y:Nat))",
"max(?x:Nat,min(?y:Nat,max(max(?y:Nat,?z:Nat),?x:Nat))) = max(min(?x:Nat,min(?z:Nat,?x:Nat)),min(?y:Nat,min(?y:Nat,min(?x:Nat,?z:Nat))))",
"max(min(?x:Nat,?x:Nat),max(min(?x:Nat,?z:Nat),0)) = min(max(max(max(?x:Nat,?y:Nat),?x:Nat),min(?x:Nat,?z:Nat)),max(0,?y:Nat))",
"min(?x:Nat,max(?x:Nat,max(max(?x:Nat,?y:Nat),?y:Nat))) = min(min(?x:Nat,min(0,min(max(0,?y:Nat),max(?z:Nat,?x:Nat)))),?x:Nat)",
"min(?x:Nat,min(max(?z:Nat,0),max(?y:Nat,?y:Nat))) = min(max(0,min(min(min(?x:Nat,?y:Nat),min(?y:Nat,0)),?y:Nat)),?x:Nat)",
"min(?y:Nat,min(0,min(?x:Nat,max(?y:Nat,?z:Nat)))) = max(min(?x:Nat,?y:Nat),max(?x:Nat,max(?y:Nat,max(?z:Nat,max(?x:Nat,?y:Nat)))))",
"min(min(?x:Nat,min(min(0,?y:Nat),?y:Nat)),?z:Nat) = max(?y:Nat,max(min(?y:Nat,?x:Nat),max(?x:Nat,min(min(?y:Nat,?x:Nat),?z:Nat))))",
"max(0,max(?x:Nat,max(?y:Nat,max(max(?y:Nat,?y:Nat),?x:Nat)))) = max(min(?x:Nat,min(min(?x:Nat,min(?y:Nat,0)),?y:Nat)),?y:Nat)",
"max(0,max(min(min(max(?y:Nat,?z:Nat),?y:Nat),?x:Nat),?z:Nat)) = max(min(min(?y:Nat,min(?y:Nat,?y:Nat)),?x:Nat),min(?z:Nat,?x:Nat))",
"max(max(?x:Nat,max(?y:Nat,max(max(?x:Nat,?y:Nat),?y:Nat))),0) = min(?z:Nat,min(?y:Nat,min(?y:Nat,max(max(?y:Nat,?y:Nat),?x:Nat))))",
"max(max(max(?y:Nat,min(?z:Nat,?x:Nat)),0),max(?x:Nat,?x:Nat)) = min(max(max(?x:Nat,?x:Nat),0),max(?z:Nat,min(?x:Nat,?y:Nat)))",
"max(max(max(max(min(?y:Nat,?y:Nat),0),?z:Nat),?y:Nat),?x:Nat) = min(max(?x:Nat,max(0,0)),max(min(0,?y:Nat),?x:Nat))",
"max(max(min(?x:Nat,0),?y:Nat),max(max(?x:Nat,?y:Nat),?y:Nat)) = min(0,min(max(min(0,?z:Nat),0),min(?y:Nat,?y:Nat)))",
"min(?x:Nat,max(0,min(min(max(?z:Nat,?y:Nat),?y:Nat),?y:Nat))) = min(max(max(min(min(?z:Nat,?y:Nat),?x:Nat),?x:Nat),?z:Nat),?y:Nat)",
"min(?z:Nat,max(?y:Nat,min(min(?x:Nat,max(?x:Nat,?y:Nat)),?y:Nat))) = max(max(max(max(min(?y:Nat,?y:Nat),?z:Nat),?x:Nat),?x:Nat),?z:Nat)",
"min(max(min(max(min(0,?y:Nat),0),?y:Nat),?z:Nat),?z:Nat) = max(min(max(0,min(?x:Nat,min(?x:Nat,?y:Nat))),?x:Nat),?y:Nat)",
"min(min(max(?x:Nat,max(?x:Nat,?x:Nat)),?y:Nat),max(?z:Nat,?y:Nat)) = min(max(?x:Nat,max(max(?y:Nat,min(?x:Nat,?z:Nat)),0)),?y:Nat)",
"max(min(max(?x:Nat,?x:Nat),max(?y:Nat,max(?y:Nat,max(?y:Nat,0)))),?y:Nat) = min(?x:Nat,max(?x:Nat,min(min(?y:Nat,?x:Nat),?z:Nat)))",
"max(min(max(min(0,?x:Nat),min(?x:Nat,min(?x:Nat,?y:Nat))),?y:Nat),?z:Nat) = min(?z:Nat,max(0,max(min(0,?y:Nat),0)))",
"max(min(max(min(min(?y:Nat,?z:Nat),?z:Nat),max(?y:Nat,?y:Nat)),0),?x:Nat) = min(?y:Nat,max(0,max(max(?y:Nat,?x:Nat),?y:Nat)))",
"min(0,max(?y:Nat,min(max(max(?x:Nat,?x:Nat),?y:Nat),max(0,?x:Nat)))) = min(min(min(?x:Nat,min(?y:Nat,?z:Nat)),?x:Nat),?y:Nat)",
"min(max(?x:Nat,max(max(min(?y:Nat,?y:Nat),?y:Nat),min(?y:Nat,0))),?y:Nat) = min(?x:Nat,max(min(max(?z:Nat,?x:Nat),?y:Nat),?y:Nat))",
"min(max(0,max(min(max(?y:Nat,?y:Nat),max(?y:Nat,?y:Nat)),?y:Nat)),?x:Nat) = max(0,min(?x:Nat,max(max(?y:Nat,?z:Nat),?z:Nat)))",
"min(min(min(0,min(min(?x:Nat,?z:Nat),max(?x:Nat,?x:Nat))),?y:Nat),?x:Nat) = min(max(?x:Nat,min(max(?x:Nat,?x:Nat),?y:Nat)),?y:Nat)",
"max(?x:Nat,max(?y:Nat,max(max(?y:Nat,?x:Nat),min(max(?z:Nat,0),max(?z:Nat,?z:Nat))))) = min(?x:Nat,min(min(?x:Nat,?y:Nat),?y:Nat))",
"max(0,max(min(?z:Nat,max(?y:Nat,?x:Nat)),max(max(?y:Nat,min(?x:Nat,0)),?y:Nat))) = min(min(min(?x:Nat,?y:Nat),?z:Nat),?z:Nat)",
"max(0,max(min(max(min(?x:Nat,?z:Nat),max(?z:Nat,?x:Nat)),max(?y:Nat,?y:Nat)),?x:Nat)) = max(?y:Nat,min(min(?z:Nat,?x:Nat),?y:Nat))",
"max(max(max(?z:Nat,max(0,?y:Nat)),?y:Nat),max(?x:Nat,min(?z:Nat,max(?x:Nat,?y:Nat)))) = max(?y:Nat,min(max(0,?x:Nat),0))",
"max(max(max(min(min(0,?x:Nat),0),min(0,0)),?y:Nat),min(?z:Nat,?x:Nat)) = max(?y:Nat,max(min(?z:Nat,?x:Nat),?x:Nat))",
"max(max(min(?x:Nat,?x:Nat),min(max(min(?y:Nat,?y:Nat),?x:Nat),min(?y:Nat,?x:Nat))),?z:Nat) = max(min(min(?z:Nat,?z:Nat),0),?z:Nat)",
"max(max(min(max(?y:Nat,min(0,?x:Nat)),?y:Nat),min(max(?x:Nat,?z:Nat),?z:Nat)),?y:Nat) = max(min(max(?x:Nat,?z:Nat),?y:Nat),?y:Nat)",
"max(min(min(?x:Nat,min(?x:Nat,?z:Nat)),?x:Nat),max(max(?x:Nat,?y:Nat),min(0,?y:Nat))) = min(min(?x:Nat,min(?y:Nat,?y:Nat)),?z:Nat)",
"min(max(?x:Nat,min(max(?y:Nat,min(?x:Nat,?y:Nat)),min(min(?y:Nat,?x:Nat),?z:Nat))),?x:Nat) = min(?z:Nat,min(?y:Nat,max(?z:Nat,?x:Nat)))",
"min(max(min(?x:Nat,?y:Nat),min(?z:Nat,min(max(?x:Nat,?x:Nat),?y:Nat))),max(?y:Nat,?y:Nat)) = min(?y:Nat,min(min(0,?x:Nat),?z:Nat))",
"min(max(min(max(?x:Nat,max(?z:Nat,?z:Nat)),?y:Nat),min(?y:Nat,max(?y:Nat,0))),?x:Nat) = min(min(?x:Nat,0),max(0,?y:Nat))",
"max(max(min(min(0,min(?x:Nat,?y:Nat)),?z:Nat),0),min(?x:Nat,max(0,min(?y:Nat,?z:Nat)))) = min(min(?y:Nat,?z:Nat),?y:Nat)",
"max(min(?x:Nat,?y:Nat),max(min(0,0),min(min(?x:Nat,?x:Nat),min(min(?y:Nat,?x:Nat),?x:Nat)))) = min(max(?y:Nat,?y:Nat),?y:Nat)",
"max(min(max(min(?y:Nat,max(?x:Nat,?z:Nat)),?x:Nat),?y:Nat),min(min(max(?x:Nat,?y:Nat),0),?y:Nat)) = min(min(0,?z:Nat),0)",
"max(min(min(min(min(?y:Nat,?x:Nat),?x:Nat),min(?z:Nat,max(?x:Nat,?y:Nat))),?y:Nat),max(?x:Nat,?x:Nat)) = min(?y:Nat,max(?x:Nat,?z:Nat))",
"min(?x:Nat,min(max(0,min(?x:Nat,max(?x:Nat,?x:Nat))),min(?y:Nat,max(min(?y:Nat,?x:Nat),?x:Nat)))) = max(max(?z:Nat,?x:Nat),?x:Nat)",
"min(max(?y:Nat,?y:Nat),max(max(max(max(?y:Nat,0),?y:Nat),?y:Nat),min(0,max(0,?y:Nat)))) = max(min(?x:Nat,?x:Nat),0)",
"min(max(max(max(0,?z:Nat),?x:Nat),0),min(min(?y:Nat,?y:Nat),max(max(?y:Nat,?x:Nat),?x:Nat))) = min(?x:Nat,min(?x:Nat,?x:Nat))",
"min(min(?x:Nat,max(?x:Nat,?y:Nat)),max(max(min(max(?x:Nat,?x:Nat),min(?x:Nat,0)),?z:Nat),?y:Nat)) = max(?z:Nat,max(?x:Nat,?x:Nat))",
"min(min(max(min(?z:Nat,min(?x:Nat,?x:Nat)),max(min(?x:Nat,?x:Nat),min(?y:Nat,?y:Nat))),?y:Nat),?x:Nat) = max(min(?z:Nat,?x:Nat),?y:Nat)",
"max(?x:Nat,min(max(max(min(?x:Nat,?x:Nat),max(?y:Nat,?x:Nat)),?y:Nat),max(max(?y:Nat,max(?x:Nat,0)),?x:Nat))) = max(0,?x:Nat)",
"max(?z:Nat,max(max(min(min(?z:Nat,?y:Nat),max(?y:Nat,0)),?x:Nat),max(max(?x:Nat,0),max(0,0)))) = max(?z:Nat,?y:Nat)",
"max(max(0,max(0,max(?y:Nat,?x:Nat))),min(min(?y:Nat,min(?x:Nat,?y:Nat)),min(min(0,?x:Nat),?y:Nat))) = min(?y:Nat,?z:Nat)",
"max(max(max(?y:Nat,?z:Nat),?y:Nat),max(min(?y:Nat,min(min(?z:Nat,?y:Nat),max(?y:Nat,?y:Nat))),min(?y:Nat,0))) = min(?x:Nat,?y:Nat)",
"max(max(min(?x:Nat,max(?z:Nat,?x:Nat)),max(min(?x:Nat,?x:Nat),?y:Nat)),max(max(?x:Nat,min(0,?x:Nat)),?y:Nat)) = max(?x:Nat,?x:Nat)",
"max(min(?x:Nat,0),min(?x:Nat,max(min(max(0,?z:Nat),min(?y:Nat,?z:Nat)),min(max(?x:Nat,?x:Nat),?x:Nat)))) = min(?x:Nat,0)",
"max(min(?x:Nat,min(?x:Nat,min(0,max(?x:Nat,?y:Nat)))),min(?y:Nat,min(min(?y:Nat,max(?x:Nat,?y:Nat)),?z:Nat))) = min(?y:Nat,?x:Nat)",
"max(min(?z:Nat,?x:Nat),max(max(0,min(max(?x:Nat,0),?y:Nat)),max(max(max(?z:Nat,0),?z:Nat),0))) = min(0,?y:Nat)",
"min(?y:Nat,min(min(?z:Nat,max(max(?y:Nat,?z:Nat),?x:Nat)),min(max(0,?x:Nat),max(?z:Nat,min(?y:Nat,?y:Nat))))) = min(?y:Nat,?y:Nat)",
"min(?z:Nat,max(max(min(min(?x:Nat,?z:Nat),min(?x:Nat,?x:Nat)),?y:Nat),max(max(0,?x:Nat),max(?z:Nat,?x:Nat)))) = max(0,?x:Nat)",
"min(?z:Nat,min(max(max(max(?z:Nat,?z:Nat),?y:Nat),?x:Nat),min(?x:Nat,max(min(?y:Nat,?y:Nat),min(?y:Nat,?y:Nat))))) = min(?z:Nat,0)",
"min(max(?x:Nat,min(max(?y:Nat,?x:Nat),?x:Nat)),min(min(?y:Nat,?y:Nat),min(max(?y:Nat,?x:Nat),max(?z:Nat,?y:Nat)))) = min(?z:Nat,?y:Nat)",
"min(max(min(?x:Nat,?z:Nat),max(max(min(?x:Nat,?y:Nat),?y:Nat),min(max(?y:Nat,0),max(?x:Nat,0)))),?x:Nat) = min(0,?x:Nat)",
"min(max(min(0,max(max(?y:Nat,0),min(0,0))),?x:Nat),max(max(?y:Nat,0),max(?z:Nat,?y:Nat))) = max(?x:Nat,?z:Nat)",
"min(min(0,min(?x:Nat,?z:Nat)),max(?x:Nat,max(min(?x:Nat,max(?z:Nat,0)),min(min(?z:Nat,?z:Nat),?x:Nat)))) = min(?y:Nat,?x:Nat)",
"min(min(min(?y:Nat,min(?z:Nat,min(?x:Nat,?y:Nat))),?y:Nat),min(?z:Nat,min(?y:Nat,max(?y:Nat,min(?y:Nat,?x:Nat))))) = min(?z:Nat,?x:Nat)",
"max(?x:Nat,?x:Nat) = max(max(min(max(min(0,?x:Nat),min(?y:Nat,?y:Nat)),min(?x:Nat,?y:Nat)),min(min(?y:Nat,max(?x:Nat,?x:Nat)),?y:Nat)),?y:Nat)",
"max(?y:Nat,?y:Nat) = min(min(max(?x:Nat,min(max(?z:Nat,?y:Nat),?z:Nat)),?x:Nat),min(min(?z:Nat,max(max(?x:Nat,?y:Nat),max(?y:Nat,0))),?y:Nat))",
"max(?y:Nat,?y:Nat) = min(min(max(?y:Nat,0),min(?y:Nat,?x:Nat)),max(min(min(max(?x:Nat,?y:Nat),?x:Nat),min(0,min(?x:Nat,?x:Nat))),?y:Nat))",
"max(?z:Nat,?y:Nat) = max(min(min(?y:Nat,?x:Nat),?x:Nat),max(max(max(max(?x:Nat,?z:Nat),max(?z:Nat,?y:Nat)),max(min(0,?z:Nat),?y:Nat)),?x:Nat))",
"min(?x:Nat,?x:Nat) = max(min(0,?y:Nat),max(max(?y:Nat,max(0,min(0,?y:Nat))),min(min(?z:Nat,?z:Nat),max(min(?z:Nat,?y:Nat),?x:Nat))))",
"min(?x:Nat,?x:Nat) = min(max(0,max(max(min(0,?z:Nat),0),?y:Nat)),max(max(min(?y:Nat,0),max(?y:Nat,max(?x:Nat,?y:Nat))),?x:Nat))",
"min(?x:Nat,?z:Nat) = min(max(max(?z:Nat,?y:Nat),?z:Nat),min(min(min(?z:Nat,0),0),max(min(min(?y:Nat,?z:Nat),min(?x:Nat,0)),?x:Nat)))",
"min(?x:Nat,?z:Nat) = min(max(min(min(max(?x:Nat,?x:Nat),?y:Nat),?y:Nat),?y:Nat),min(max(max(?y:Nat,max(?y:Nat,?x:Nat)),0),min(?z:Nat,?x:Nat)))",
"min(?x:Nat,0) = max(max(?x:Nat,max(max(max(?x:Nat,?x:Nat),max(?x:Nat,?y:Nat)),?y:Nat)),min(min(min(max(?x:Nat,0),?y:Nat),?y:Nat),?z:Nat))",
"min(?y:Nat,?x:Nat) = max(max(?x:Nat,max(?x:Nat,max(max(?y:Nat,?y:Nat),?y:Nat))),min(?x:Nat,min(min(min(?y:Nat,?z:Nat),max(?x:Nat,?x:Nat)),?z:Nat)))",
"min(?y:Nat,?x:Nat) = max(max(max(?x:Nat,max(?x:Nat,?z:Nat)),min(max(min(?y:Nat,?z:Nat),min(0,?x:Nat)),max(?z:Nat,?y:Nat))),max(?z:Nat,?x:Nat))",
"min(?y:Nat,?x:Nat) = min(max(max(?x:Nat,min(min(?x:Nat,0),?y:Nat)),max(max(?z:Nat,min(?z:Nat,?z:Nat)),?x:Nat)),min(?x:Nat,max(?z:Nat,?x:Nat)))",
"min(?y:Nat,?z:Nat) = min(min(max(max(min(?y:Nat,?z:Nat),min(0,?z:Nat)),max(?y:Nat,min(?z:Nat,?x:Nat))),min(max(?x:Nat,?x:Nat),?x:Nat)),?x:Nat)",
"min(?y:Nat,0) = min(?z:Nat,max(max(max(?x:Nat,min(?x:Nat,?x:Nat)),max(max(?y:Nat,?x:Nat),min(?x:Nat,?y:Nat))),max(?x:Nat,min(0,?x:Nat))))",
"min(0,?z:Nat) = min(max(min(?y:Nat,max(max(?x:Nat,?z:Nat),max(?z:Nat,?y:Nat))),?z:Nat),max(max(?z:Nat,0),max(?z:Nat,min(?x:Nat,?z:Nat))))",
"max(?x:Nat,max(?x:Nat,?z:Nat)) = min(max(max(min(0,?z:Nat),min(?z:Nat,min(?y:Nat,?y:Nat))),?x:Nat),min(?z:Nat,max(?z:Nat,min(?y:Nat,?y:Nat))))",
"max(?z:Nat,max(?x:Nat,?x:Nat)) = min(max(max(min(max(?x:Nat,?z:Nat),max(?y:Nat,?y:Nat)),min(min(?y:Nat,?x:Nat),max(?x:Nat,?x:Nat))),0),0)",
"max(?z:Nat,max(0,?x:Nat)) = max(min(max(max(?y:Nat,0),?x:Nat),min(max(?x:Nat,max(?x:Nat,?y:Nat)),max(min(0,?x:Nat),?x:Nat))),0)",
"max(0,min(?y:Nat,?y:Nat)) = min(max(max(max(?x:Nat,min(?x:Nat,?y:Nat)),max(?x:Nat,?x:Nat)),min(?x:Nat,max(?x:Nat,0))),min(?y:Nat,?z:Nat))",
"max(max(max(?z:Nat,?z:Nat),?x:Nat),?x:Nat) = min(min(min(max(min(0,?y:Nat),?x:Nat),?z:Nat),max(max(?x:Nat,0),min(?y:Nat,?y:Nat))),?y:Nat)",
"max(min(?y:Nat,max(?y:Nat,?x:Nat)),?y:Nat) = min(min(max(?x:Nat,max(max(?x:Nat,?y:Nat),min(?y:Nat,?x:Nat))),min(min(0,?x:Nat),?y:Nat)),?z:Nat)",
"min(0,min(max(?z:Nat,?y:Nat),?y:Nat)) = min(?z:Nat,max(max(min(0,max(?y:Nat,?y:Nat)),0),max(?x:Nat,min(?z:Nat,max(0,0)))))",
"min(max(?z:Nat,?x:Nat),max(?y:Nat,?x:Nat)) = min(min(0,min(?x:Nat,min(min(?y:Nat,?y:Nat),max(0,?y:Nat)))),max(max(?y:Nat,?z:Nat),?z:Nat))",
"min(min(?y:Nat,min(?x:Nat,?z:Nat)),?y:Nat) = max(max(min(min(?x:Nat,?x:Nat),?x:Nat),min(?x:Nat,min(min(?z:Nat,?y:Nat),max(?x:Nat,?y:Nat)))),?x:Nat)",
"min(min(?y:Nat,min(?x:Nat,0)),?y:Nat) = min(min(?x:Nat,max(?y:Nat,?x:Nat)),min(?x:Nat,min(min(max(?x:Nat,?y:Nat),?y:Nat),max(?z:Nat,?y:Nat))))",
"max(?x:Nat,min(max(?x:Nat,?z:Nat),min(?x:Nat,?y:Nat))) = min(?x:Nat,max(max(?z:Nat,max(?z:Nat,?y:Nat)),min(min(max(?y:Nat,?x:Nat),?y:Nat),?z:Nat)))",
"max(?z:Nat,max(min(max(?x:Nat,?x:Nat),?y:Nat),?z:Nat)) = max(max(0,max(max(?x:Nat,?y:Nat),max(min(?z:Nat,?x:Nat),min(?z:Nat,?y:Nat)))),?y:Nat)",
"max(min(min(?x:Nat,min(?y:Nat,0)),?z:Nat),0) = min(max(max(min(?x:Nat,?x:Nat),max(max(?y:Nat,?x:Nat),?y:Nat)),?y:Nat),min(?y:Nat,?x:Nat))",
"min(?x:Nat,max(?x:Nat,min(min(?x:Nat,?x:Nat),?z:Nat))) = max(max(min(?x:Nat,min(min(?y:Nat,?y:Nat),min(?y:Nat,?x:Nat))),min(?y:Nat,?x:Nat)),?y:Nat)",
"min(max(0,max(min(?y:Nat,?y:Nat),?y:Nat)),?y:Nat) = min(max(min(?y:Nat,min(?z:Nat,?y:Nat)),min(min(0,min(?y:Nat,?z:Nat)),0)),0)",
"max(?x:Nat,max(?y:Nat,min(min(?x:Nat,0),max(?y:Nat,?y:Nat)))) = min(?y:Nat,max(min(max(min(?x:Nat,?z:Nat),min(?y:Nat,?y:Nat)),?z:Nat),0))",
"max(?x:Nat,max(?z:Nat,max(max(0,max(0,?x:Nat)),?x:Nat))) = max(min(min(0,?x:Nat),min(?y:Nat,max(?y:Nat,min(?y:Nat,?z:Nat)))),?z:Nat)",
"max(max(?x:Nat,min(min(?z:Nat,min(?y:Nat,?x:Nat)),?y:Nat)),0) = min(min(min(?z:Nat,min(min(?z:Nat,?x:Nat),max(?x:Nat,?x:Nat))),?y:Nat),0)",
"max(max(max(min(min(?y:Nat,?y:Nat),?x:Nat),?y:Nat),?x:Nat),?z:Nat) = max(min(min(?y:Nat,max(0,?x:Nat)),max(min(?y:Nat,?x:Nat),?y:Nat)),?y:Nat)",
"max(max(min(max(max(?y:Nat,?y:Nat),?x:Nat),?x:Nat),?y:Nat),0) = min(max(min(max(?x:Nat,min(?x:Nat,?z:Nat)),max(?x:Nat,?x:Nat)),?x:Nat),?z:Nat)",
"max(max(min(min(?z:Nat,?x:Nat),?x:Nat),0),min(?y:Nat,?x:Nat)) = min(min(min(?y:Nat,max(min(?y:Nat,?z:Nat),min(?z:Nat,?x:Nat))),0),?x:Nat)",
"max(min(?x:Nat,max(?y:Nat,min(min(0,0),?z:Nat))),?x:Nat) = min(?x:Nat,max(min(min(0,?x:Nat),max(min(?z:Nat,?y:Nat),0)),?y:Nat))",
"max(min(?x:Nat,min(?z:Nat,max(?y:Nat,min(?x:Nat,0)))),?y:Nat) = max(min(?y:Nat,?z:Nat),max(min(min(max(?y:Nat,?x:Nat),?y:Nat),?x:Nat),?x:Nat))",
"max(min(?z:Nat,min(?x:Nat,?z:Nat)),max(?z:Nat,max(?z:Nat,?x:Nat))) = max(max(min(?y:Nat,?x:Nat),max(?x:Nat,?y:Nat)),max(min(?z:Nat,?x:Nat),?x:Nat))",
"max(min(0,?x:Nat),min(?y:Nat,min(min(?y:Nat,?y:Nat),?y:Nat))) = min(?z:Nat,max(?y:Nat,max(max(max(0,?z:Nat),min(0,0)),0)))",
"min(?x:Nat,min(max(min(min(?y:Nat,0),?y:Nat),0),?x:Nat)) = min(max(max(?y:Nat,min(max(?x:Nat,?y:Nat),?y:Nat)),?y:Nat),min(?x:Nat,?y:Nat))",
"min(max(min(max(?y:Nat,?x:Nat),?z:Nat),?x:Nat),min(0,?x:Nat)) = min(min(min(?x:Nat,max(?y:Nat,max(?z:Nat,0))),?x:Nat),min(?y:Nat,?y:Nat))",
"max(?x:Nat,max(min(?x:Nat,?y:Nat),max(0,min(?y:Nat,min(?x:Nat,?x:Nat))))) = max(min(min(?y:Nat,max(?y:Nat,min(?y:Nat,0))),0),?y:Nat)",
"max(max(max(max(?x:Nat,?x:Nat),?y:Nat),max(?x:Nat,?z:Nat)),min(?y:Nat,?y:Nat)) = max(min(0,max(?y:Nat,min(min(?x:Nat,?y:Nat),?y:Nat))),?x:Nat)",
"max(max(min(max(min(?z:Nat,?y:Nat),?y:Nat),min(?x:Nat,?x:Nat)),?y:Nat),?y:Nat) = max(?z:Nat,min(min(?z:Nat,max(0,max(?x:Nat,?x:Nat))),?z:Nat))",
"max(min(?y:Nat,min(?z:Nat,max(max(?z:Nat,?y:Nat),min(?y:Nat,?x:Nat)))),?x:Nat) = max(min(min(max(?y:Nat,?z:Nat),?z:Nat),?y:Nat),min(?y:Nat,?y:Nat))",
"max(min(min(min(min(0,?x:Nat),min(?z:Nat,?y:Nat)),?y:Nat),?x:Nat),?y:Nat) = min(max(max(?y:Nat,min(?y:Nat,min(?x:Nat,0))),?y:Nat),0)",
"min(?x:Nat,min(max(max(max(?y:Nat,?y:Nat),min(?y:Nat,?y:Nat)),?y:Nat),?y:Nat)) = min(?z:Nat,min(min(max(?y:Nat,?y:Nat),?y:Nat),min(0,?z:Nat)))",
"min(?y:Nat,min(max(max(max(?x:Nat,0),?y:Nat),min(?y:Nat,?y:Nat)),0)) = min(?y:Nat,max(?x:Nat,max(?y:Nat,min(0,max(?z:Nat,?x:Nat)))))",
"min(max(max(?y:Nat,min(?x:Nat,max(?y:Nat,?y:Nat))),?y:Nat),max(?z:Nat,?z:Nat)) = min(max(max(?x:Nat,max(?z:Nat,max(?y:Nat,0))),0),?x:Nat)",
"min(max(min(0,max(min(?y:Nat,?x:Nat),?y:Nat)),min(?x:Nat,?z:Nat)),?x:Nat) = min(min(?y:Nat,?z:Nat),min(max(?x:Nat,max(0,?x:Nat)),?x:Nat))",
"min(min(0,max(?x:Nat,?y:Nat)),max(min(max(0,?x:Nat),?y:Nat),?z:Nat)) = max(max(max(?x:Nat,max(min(?x:Nat,0),?x:Nat)),?z:Nat),?y:Nat)",
"min(min(min(max(max(?y:Nat,?y:Nat),?y:Nat),min(0,?y:Nat)),0),0) = max(max(min(?y:Nat,max(?x:Nat,max(0,?x:Nat))),?y:Nat),?x:Nat)",
"min(?y:Nat,max(0,min(min(min(?x:Nat,?y:Nat),?y:Nat),min(0,max(?z:Nat,?z:Nat))))) = min(min(?x:Nat,max(?z:Nat,max(0,?y:Nat))),?y:Nat)",
"min(max(?y:Nat,?x:Nat),min(max(max(?z:Nat,?z:Nat),max(min(0,?x:Nat),?x:Nat)),?y:Nat)) = min(0,min(?x:Nat,max(0,min(0,?x:Nat))))",
"max(?x:Nat,min(min(?x:Nat,max(min(?y:Nat,?z:Nat),?x:Nat)),max(max(0,max(?y:Nat,?z:Nat)),?x:Nat))) = max(?y:Nat,max(?x:Nat,max(?z:Nat,?y:Nat)))",
"max(max(max(?z:Nat,?y:Nat),?z:Nat),min(max(min(max(?z:Nat,?y:Nat),max(?z:Nat,?y:Nat)),?z:Nat),?y:Nat)) = max(?x:Nat,min(min(?y:Nat,?y:Nat),?y:Nat))",
"max(min(?y:Nat,?z:Nat),max(?x:Nat,min(min(?z:Nat,min(?x:Nat,0)),min(max(?x:Nat,?x:Nat),?y:Nat)))) = min(min(?y:Nat,max(?z:Nat,?x:Nat)),?z:Nat)",
"max(min(?z:Nat,?y:Nat),max(max(min(?y:Nat,max(?x:Nat,0)),min(?z:Nat,0)),max(?x:Nat,0))) = min(min(?z:Nat,min(?y:Nat,0)),?y:Nat)",
"min(?x:Nat,min(max(min(min(?y:Nat,?z:Nat),min(?y:Nat,?y:Nat)),min(?x:Nat,min(?z:Nat,?x:Nat))),?x:Nat)) = max(max(0,0),min(?x:Nat,?x:Nat))",
"min(?y:Nat,min(min(?x:Nat,max(0,?x:Nat)),min(max(min(0,?y:Nat),min(?x:Nat,?x:Nat)),?z:Nat))) = min(?x:Nat,max(min(?x:Nat,?y:Nat),?x:Nat))",
"min(max(max(max(max(?y:Nat,?x:Nat),max(?x:Nat,0)),min(?y:Nat,max(?x:Nat,?z:Nat))),?x:Nat),?z:Nat) = max(0,max(?y:Nat,min(?y:Nat,?y:Nat)))",
"max(max(?y:Nat,min(?y:Nat,max(max(0,?y:Nat),?z:Nat))),max(max(?z:Nat,max(?x:Nat,min(?x:Nat,?x:Nat))),?z:Nat)) = max(min(?y:Nat,?z:Nat),0)",
"max(max(max(?z:Nat,?y:Nat),min(?x:Nat,max(0,min(?x:Nat,?y:Nat)))),max(0,max(?z:Nat,min(?x:Nat,?z:Nat)))) = min(min(?z:Nat,?x:Nat),0)",
"max(min(max(?z:Nat,?z:Nat),max(max(max(?y:Nat,?x:Nat),0),max(?y:Nat,max(?y:Nat,?y:Nat)))),min(0,?x:Nat)) = min(min(?y:Nat,?x:Nat),0)",
"min(max(?y:Nat,min(?z:Nat,?y:Nat)),min(max(min(?x:Nat,min(?x:Nat,?z:Nat)),max(?y:Nat,max(0,?y:Nat))),0)) = min(?y:Nat,max(?z:Nat,?x:Nat))",
"min(max(max(0,min(min(?z:Nat,?x:Nat),?y:Nat)),0),min(?x:Nat,max(?y:Nat,max(max(?y:Nat,?x:Nat),?z:Nat)))) = min(max(?x:Nat,?y:Nat),0)",
"min(min(min(min(?x:Nat,?y:Nat),min(max(?z:Nat,0),?y:Nat)),max(min(min(?x:Nat,?x:Nat),?y:Nat),0)),0) = max(?x:Nat,max(?z:Nat,?y:Nat))",
"max(max(?x:Nat,max(max(min(?x:Nat,0),0),?x:Nat)),min(min(min(min(?z:Nat,?y:Nat),min(?x:Nat,?y:Nat)),?y:Nat),?y:Nat)) = max(?y:Nat,?z:Nat)",
"max(max(?z:Nat,max(0,min(max(?y:Nat,?z:Nat),max(0,?z:Nat)))),max(?z:Nat,min(?z:Nat,max(min(?z:Nat,?y:Nat),?z:Nat)))) = min(?y:Nat,0)",
"max(max(max(?x:Nat,?y:Nat),max(0,0)),max(0,max(max(min(?y:Nat,?x:Nat),max(?x:Nat,0)),max(0,0)))) = min(0,?x:Nat)",
"max(max(max(max(?x:Nat,min(?z:Nat,?y:Nat)),min(?y:Nat,?z:Nat)),min(?x:Nat,?x:Nat)),max(?z:Nat,max(min(?z:Nat,?y:Nat),0))) = max(?x:Nat,?y:Nat)",
"max(max(min(?x:Nat,max(?z:Nat,?z:Nat)),min(?x:Nat,min(0,?y:Nat))),min(max(?z:Nat,min(0,min(?x:Nat,?x:Nat))),?x:Nat)) = min(?y:Nat,?y:Nat)",
"max(min(?z:Nat,0),max(min(?y:Nat,min(min(?x:Nat,?y:Nat),?x:Nat)),min(max(?x:Nat,min(?x:Nat,?y:Nat)),min(0,?y:Nat)))) = min(?x:Nat,0)",
"max(min(min(max(min(?x:Nat,?z:Nat),?y:Nat),0),max(max(?x:Nat,min(?y:Nat,?z:Nat)),min(max(?x:Nat,?x:Nat),?x:Nat))),0) = max(?y:Nat,?z:Nat)",
"min(max(min(min(min(0,?z:Nat),?z:Nat),0),max(max(?y:Nat,?x:Nat),min(min(?y:Nat,?y:Nat),min(?y:Nat,?y:Nat)))),?x:Nat) = max(?z:Nat,?y:Nat)",
"min(max(min(min(min(0,0),?z:Nat),?y:Nat),0),min(min(?x:Nat,max(?y:Nat,?x:Nat)),max(max(?y:Nat,?x:Nat),?y:Nat))) = max(?y:Nat,?y:Nat)",
"min(min(max(max(?x:Nat,max(?z:Nat,0)),min(max(?x:Nat,?y:Nat),?y:Nat)),min(?y:Nat,?y:Nat)),min(?x:Nat,min(?x:Nat,?x:Nat))) = max(?y:Nat,?x:Nat)",
"min(min(min(?y:Nat,max(max(?y:Nat,?x:Nat),0)),max(min(0,max(?x:Nat,?z:Nat)),max(min(?x:Nat,?y:Nat),?y:Nat))),?x:Nat) = max(?z:Nat,?x:Nat)",
"min(min(min(0,min(?x:Nat,max(?z:Nat,?y:Nat))),max(min(?x:Nat,?y:Nat),max(max(?x:Nat,?x:Nat),max(?y:Nat,?z:Nat)))),?y:Nat) = min(?x:Nat,?x:Nat)",
"min(?x:Nat,?y:Nat) = max(max(max(min(?x:Nat,min(?x:Nat,?x:Nat)),min(max(?x:Nat,?y:Nat),?z:Nat)),min(?x:Nat,min(?y:Nat,max(?z:Nat,0)))),min(?y:Nat,?y:Nat))",
"min(?x:Nat,?y:Nat) = max(min(max(?x:Nat,?y:Nat),min(max(min(?x:Nat,?x:Nat),?y:Nat),?z:Nat)),max(?x:Nat,max(min(?x:Nat,?z:Nat),min(?y:Nat,min(?y:Nat,?x:Nat)))))",
"max(?x:Nat,max(0,?x:Nat)) = min(min(?x:Nat,min(0,?y:Nat)),max(max(max(0,?x:Nat),min(min(?x:Nat,?x:Nat),min(?x:Nat,?y:Nat))),min(?z:Nat,?z:Nat)))",
"max(max(0,?y:Nat),?x:Nat) = min(max(max(?x:Nat,min(?x:Nat,?y:Nat)),?x:Nat),max(min(min(?y:Nat,?y:Nat),max(max(?z:Nat,0),min(?z:Nat,?y:Nat))),?z:Nat))",
"max(min(?x:Nat,?x:Nat),?y:Nat) = max(?x:Nat,min(min(min(?y:Nat,min(?x:Nat,?x:Nat)),min(?y:Nat,min(?x:Nat,?y:Nat))),min(?y:Nat,min(?x:Nat,max(?x:Nat,?x:Nat)))))",
"max(min(0,0),?x:Nat) = max(min(min(min(?x:Nat,?x:Nat),max(0,?x:Nat)),min(?x:Nat,0)),min(?z:Nat,min(?x:Nat,max(max(?x:Nat,?y:Nat),?y:Nat))))",
"min(?z:Nat,min(?z:Nat,?z:Nat)) = max(min(min(max(min(?y:Nat,?y:Nat),?x:Nat),max(min(?x:Nat,?y:Nat),?y:Nat)),min(min(min(?y:Nat,?x:Nat),?y:Nat),0)),?y:Nat)",
"max(min(0,?x:Nat),min(?y:Nat,?y:Nat)) = max(min(max(max(max(0,?y:Nat),max(?x:Nat,?x:Nat)),0),?x:Nat),max(?x:Nat,max(?x:Nat,min(?x:Nat,?y:Nat))))",
"min(min(?x:Nat,?y:Nat),min(0,?y:Nat)) = max(?y:Nat,min(max(max(min(?x:Nat,?x:Nat),0),?x:Nat),max(min(max(?z:Nat,?y:Nat),max(?y:Nat,?z:Nat)),0)))",
"max(max(?y:Nat,?z:Nat),max(?y:Nat,max(0,?z:Nat))) = max(max(max(min(max(?y:Nat,0),0),min(?y:Nat,?x:Nat)),?y:Nat),max(?y:Nat,max(?z:Nat,?x:Nat)))",
"max(min(max(?x:Nat,0),max(?x:Nat,?y:Nat)),?y:Nat) = min(min(min(min(min(?x:Nat,?x:Nat),?y:Nat),?x:Nat),min(?y:Nat,max(?y:Nat,max(0,?x:Nat)))),0)",
"min(max(?z:Nat,min(min(?y:Nat,?y:Nat),?z:Nat)),?x:Nat) = max(min(max(?y:Nat,?z:Nat),max(max(0,max(0,?x:Nat)),max(?y:Nat,min(?y:Nat,?x:Nat)))),?x:Nat)",
"min(min(max(?y:Nat,?x:Nat),?x:Nat),min(?x:Nat,?x:Nat)) = min(min(?x:Nat,min(min(?y:Nat,min(?y:Nat,?y:Nat)),max(max(?z:Nat,?x:Nat),min(0,?y:Nat)))),?x:Nat)",
"max(?x:Nat,max(?y:Nat,min(max(0,?z:Nat),max(?y:Nat,?x:Nat)))) = max(min(?y:Nat,?x:Nat),max(?z:Nat,max(min(0,max(?y:Nat,?x:Nat)),max(?x:Nat,?z:Nat))))",
"max(?x:Nat,max(max(?x:Nat,max(?x:Nat,min(?x:Nat,?z:Nat))),?y:Nat)) = max(max(?x:Nat,?z:Nat),min(min(min(0,?z:Nat),min(min(0,?y:Nat),?x:Nat)),?y:Nat))",
"max(?y:Nat,min(min(max(?y:Nat,max(?z:Nat,?z:Nat)),?y:Nat),0)) = min(max(max(min(min(?x:Nat,?x:Nat),?y:Nat),min(max(?y:Nat,?x:Nat),0)),?x:Nat),?z:Nat)",
"max(max(?x:Nat,max(max(0,?y:Nat),?y:Nat)),max(?y:Nat,?z:Nat)) = min(max(max(max(0,?y:Nat),0),max(max(?z:Nat,?x:Nat),max(?x:Nat,?y:Nat))),?z:Nat)",
"max(min(?z:Nat,max(min(max(?y:Nat,?y:Nat),?x:Nat),?x:Nat)),?x:Nat) = max(?y:Nat,max(max(max(?x:Nat,max(?x:Nat,?x:Nat)),min(?y:Nat,max(?y:Nat,?x:Nat))),?x:Nat))",
"min(?y:Nat,min(?y:Nat,min(max(min(?x:Nat,?z:Nat),0),?y:Nat))) = max(min(max(min(?y:Nat,min(?x:Nat,?x:Nat)),max(?y:Nat,?y:Nat)),?x:Nat),max(?y:Nat,?y:Nat))",
"min(max(?x:Nat,min(?y:Nat,min(?y:Nat,min(?y:Nat,?y:Nat)))),?x:Nat) = max(min(min(?x:Nat,?x:Nat),max(max(?y:Nat,max(?z:Nat,?y:Nat)),?y:Nat)),max(?x:Nat,?z:Nat))",
"min(max(?z:Nat,?z:Nat),max(min(?z:Nat,?z:Nat),max(?z:Nat,?y:Nat))) = min(max(min(?x:Nat,?x:Nat),?z:Nat),min(?z:Nat,max(max(?x:Nat,min(?y:Nat,?y:Nat)),?y:Nat)))",
"max(?x:Nat,min(min(max(?z:Nat,?x:Nat),0),max(?x:Nat,max(?y:Nat,?z:Nat)))) = min(?x:Nat,max(?y:Nat,max(min(min(?x:Nat,?y:Nat),max(?z:Nat,?y:Nat)),0)))",
"max(max(?z:Nat,min(?y:Nat,max(max(?y:Nat,?x:Nat),?y:Nat))),min(?y:Nat,?y:Nat)) = min(max(0,?y:Nat),min(?z:Nat,min(min(max(?y:Nat,?x:Nat),?x:Nat),?x:Nat)))",
"max(min(min(max(?z:Nat,?y:Nat),?x:Nat),?x:Nat),min(?y:Nat,max(?y:Nat,?z:Nat))) = min(0,min(min(max(min(0,?y:Nat),?y:Nat),max(?x:Nat,?x:Nat)),0))",
"min(?y:Nat,max(max(min(?x:Nat,0),max(min(?x:Nat,?z:Nat),?y:Nat)),0)) = min(min(min(0,max(max(0,?y:Nat),max(0,?y:Nat))),?y:Nat),?z:Nat)",
"min(0,min(max(?x:Nat,min(max(?z:Nat,?y:Nat),min(?x:Nat,0))),?x:Nat)) = min(min(?y:Nat,?y:Nat),max(max(min(?x:Nat,min(?x:Nat,?x:Nat)),?x:Nat),?y:Nat))",
"min(max(max(max(?x:Nat,?y:Nat),max(max(?z:Nat,?y:Nat),?y:Nat)),?x:Nat),?y:Nat) = min(max(?x:Nat,?x:Nat),max(0,min(max(?y:Nat,max(?x:Nat,?y:Nat)),0)))",
"max(?x:Nat,max(min(max(min(0,?z:Nat),0),min(max(?z:Nat,?x:Nat),0)),?x:Nat)) = max(min(min(?y:Nat,?y:Nat),max(?x:Nat,max(?z:Nat,?z:Nat))),?x:Nat)",
"max(?y:Nat,max(max(0,0),min(?y:Nat,max(max(0,0),min(?x:Nat,?x:Nat))))) = max(max(?x:Nat,max(0,?y:Nat)),max(?x:Nat,max(?x:Nat,?z:Nat)))",
"max(?z:Nat,max(max(?y:Nat,max(0,?y:Nat)),min(min(min(?y:Nat,?y:Nat),?y:Nat),?y:Nat))) = max(max(max(?x:Nat,?y:Nat),max(?x:Nat,?y:Nat)),min(?y:Nat,?x:Nat))",
"max(0,min(min(?y:Nat,min(?y:Nat,?y:Nat)),min(max(min(?x:Nat,?y:Nat),?x:Nat),?y:Nat))) = min(?y:Nat,min(?z:Nat,min(min(?z:Nat,?z:Nat),max(0,?z:Nat))))",
"max(max(min(0,min(?x:Nat,max(0,?z:Nat))),max(?z:Nat,?y:Nat)),min(?z:Nat,?x:Nat)) = min(min(min(?z:Nat,max(0,min(?x:Nat,?y:Nat))),?x:Nat),?y:Nat)",
"max(min(min(?y:Nat,max(?x:Nat,?z:Nat)),0),max(max(?x:Nat,0),max(?z:Nat,0))) = max(max(min(?z:Nat,?x:Nat),max(max(0,?x:Nat),?x:Nat)),?x:Nat)",
"max(min(min(min(?x:Nat,?y:Nat),max(min(?y:Nat,?x:Nat),?z:Nat)),max(?x:Nat,?x:Nat)),0) = min(?y:Nat,max(?x:Nat,max(?z:Nat,max(0,min(?x:Nat,?x:Nat)))))",
"min(?y:Nat,max(max(max(max(?y:Nat,?z:Nat),?y:Nat),max(?x:Nat,?x:Nat)),min(?x:Nat,?y:Nat))) = max(max(?y:Nat,max(max(min(?x:Nat,?z:Nat),0),?x:Nat)),0)",
"min(min(?z:Nat,?x:Nat),max(max(?y:Nat,min(min(?y:Nat,?x:Nat),min(?x:Nat,?x:Nat))),?y:Nat)) = min(max(max(?y:Nat,max(?z:Nat,min(?x:Nat,?z:Nat))),?x:Nat),?y:Nat)",
"min(min(min(?y:Nat,max(?z:Nat,?z:Nat)),min(min(?x:Nat,?x:Nat),?x:Nat)),max(?z:Nat,?z:Nat)) = max(min(max(max(?y:Nat,min(?x:Nat,?y:Nat)),?x:Nat),?y:Nat),?y:Nat)",
"min(min(min(?z:Nat,?z:Nat),max(?x:Nat,0)),min(min(?y:Nat,?x:Nat),max(?z:Nat,?y:Nat))) = max(min(?z:Nat,max(min(max(?y:Nat,?z:Nat),0),0)),?x:Nat)",
"max(min(?x:Nat,max(?y:Nat,?y:Nat)),max(max(?z:Nat,?z:Nat),max(min(?x:Nat,?x:Nat),min(?z:Nat,?x:Nat)))) = min(?y:Nat,min(?z:Nat,min(min(0,?y:Nat),?x:Nat)))",
"min(?x:Nat,min(min(max(?y:Nat,?x:Nat),max(min(?y:Nat,?x:Nat),max(0,?x:Nat))),min(?x:Nat,?y:Nat))) = min(?x:Nat,max(min(max(0,?x:Nat),?z:Nat),?y:Nat))",
"min(?z:Nat,max(?y:Nat,max(max(?x:Nat,min(?y:Nat,?y:Nat)),min(max(?y:Nat,?y:Nat),min(?y:Nat,?z:Nat))))) = min(?x:Nat,max(min(max(?y:Nat,?z:Nat),0),?y:Nat))",
"min(max(min(min(min(?z:Nat,?x:Nat),max(?z:Nat,?y:Nat)),max(min(?z:Nat,?y:Nat),?y:Nat)),?x:Nat),?y:Nat) = min(max(min(max(0,?y:Nat),?z:Nat),?y:Nat),?y:Nat)",
"max(?x:Nat,min(min(?y:Nat,min(min(0,?x:Nat),?y:Nat)),min(?y:Nat,min(max(?x:Nat,?y:Nat),min(?x:Nat,0))))) = min(min(?x:Nat,max(?z:Nat,?x:Nat)),?y:Nat)",
"max(0,max(min(max(max(?y:Nat,?z:Nat),?x:Nat),max(min(?x:Nat,?y:Nat),min(?y:Nat,0))),min(?y:Nat,0))) = min(max(max(?x:Nat,?y:Nat),?x:Nat),?y:Nat)",
"max(max(max(max(max(?z:Nat,?y:Nat),?x:Nat),?z:Nat),max(min(?x:Nat,min(?z:Nat,?x:Nat)),?x:Nat)),min(?z:Nat,?z:Nat)) = max(0,max(?x:Nat,max(?z:Nat,?x:Nat)))",
"min(max(0,?x:Nat),min(min(min(max(0,?x:Nat),?y:Nat),?y:Nat),max(max(?x:Nat,?y:Nat),min(0,0)))) = min(max(?z:Nat,?z:Nat),max(?x:Nat,?y:Nat))",
"min(min(max(?y:Nat,min(?y:Nat,?y:Nat)),min(max(0,min(?x:Nat,?y:Nat)),min(?x:Nat,min(0,?z:Nat)))),?x:Nat) = min(max(?y:Nat,max(?x:Nat,?y:Nat)),?y:Nat)",
"max(max(max(?y:Nat,min(max(?y:Nat,?x:Nat),max(?z:Nat,?x:Nat))),min(?x:Nat,?x:Nat)),max(min(?x:Nat,min(0,?y:Nat)),?x:Nat)) = min(?y:Nat,min(?x:Nat,?z:Nat))",
"min(max(?y:Nat,min(min(min(?y:Nat,?y:Nat),?y:Nat),min(0,?y:Nat))),max(max(?x:Nat,max(max(?y:Nat,?x:Nat),0)),?x:Nat)) = min(?y:Nat,min(?y:Nat,?x:Nat))",
"min(max(max(?x:Nat,?z:Nat),min(?z:Nat,max(?y:Nat,max(?x:Nat,?z:Nat)))),max(?y:Nat,max(max(?z:Nat,min(0,?z:Nat)),?x:Nat))) = max(max(?x:Nat,?y:Nat),?y:Nat)",
"min(min(min(max(?z:Nat,?y:Nat),max(max(?x:Nat,?y:Nat),max(?x:Nat,?y:Nat))),max(max(0,0),?x:Nat)),min(?y:Nat,?y:Nat)) = max(?x:Nat,min(?y:Nat,?y:Nat))",
"max(max(max(?y:Nat,min(min(?y:Nat,?y:Nat),?x:Nat)),min(min(max(?z:Nat,?z:Nat),min(?y:Nat,?z:Nat)),min(max(?x:Nat,?x:Nat),?x:Nat))),?x:Nat) = max(?z:Nat,?x:Nat)",
"max(max(max(max(max(?z:Nat,?y:Nat),max(?x:Nat,?x:Nat)),max(max(?x:Nat,0),min(?z:Nat,?y:Nat))),min(max(?x:Nat,?z:Nat),?x:Nat)),0) = min(?y:Nat,?y:Nat)",
"max(min(?x:Nat,?x:Nat),min(min(0,max(?z:Nat,max(?x:Nat,?z:Nat))),max(max(?x:Nat,max(?x:Nat,?x:Nat)),max(?x:Nat,max(?y:Nat,0))))) = max(?z:Nat,?z:Nat)",
"max(min(0,max(min(?y:Nat,0),max(?y:Nat,0))),max(max(min(?x:Nat,?x:Nat),max(max(?y:Nat,0),max(?y:Nat,?x:Nat))),?x:Nat)) = max(?x:Nat,?x:Nat)",
"max(min(max(?y:Nat,?x:Nat),min(max(0,max(?z:Nat,0)),min(?z:Nat,?x:Nat))),max(min(?z:Nat,max(?y:Nat,?x:Nat)),max(?x:Nat,0))) = max(?x:Nat,?y:Nat)",
"min(max(min(min(0,min(0,?y:Nat)),?y:Nat),min(max(?y:Nat,max(?y:Nat,?y:Nat)),?x:Nat)),min(max(0,max(?y:Nat,?z:Nat)),?y:Nat)) = max(?z:Nat,?z:Nat)",
"min(max(min(min(min(?z:Nat,?x:Nat),?z:Nat),min(min(?y:Nat,?y:Nat),?z:Nat)),?z:Nat),max(max(?y:Nat,min(?y:Nat,max(?y:Nat,?y:Nat))),?y:Nat)) = min(?x:Nat,?y:Nat)",
"min(min(?z:Nat,max(?y:Nat,min(max(0,?y:Nat),max(?z:Nat,?y:Nat)))),min(min(?x:Nat,?y:Nat),min(0,min(max(?y:Nat,?y:Nat),?z:Nat)))) = max(?z:Nat,?y:Nat)",
"max(?y:Nat,?y:Nat) = min(min(?y:Nat,max(max(?y:Nat,max(?z:Nat,0)),max(?x:Nat,min(0,?z:Nat)))),max(min(min(?z:Nat,max(?y:Nat,?y:Nat)),?x:Nat),min(?z:Nat,?x:Nat)))",
"max(0,?x:Nat) = max(min(?x:Nat,min(max(min(?x:Nat,?x:Nat),?z:Nat),?z:Nat)),max(min(min(?y:Nat,?x:Nat),min(?y:Nat,?x:Nat)),min(0,max(?y:Nat,max(?z:Nat,0)))))",
"min(?x:Nat,?x:Nat) = max(max(min(max(?y:Nat,?y:Nat),?y:Nat),min(max(max(?y:Nat,?x:Nat),max(0,?x:Nat)),max(max(?x:Nat,?y:Nat),?y:Nat))),max(?z:Nat,min(0,?z:Nat)))",
"min(?x:Nat,?x:Nat) = min(max(?z:Nat,min(min(?z:Nat,min(?x:Nat,0)),min(max(?z:Nat,0),?x:Nat))),max(max(0,min(max(?x:Nat,?y:Nat),min(?x:Nat,?z:Nat))),?y:Nat))",
"min(?y:Nat,?x:Nat) = min(max(max(min(min(?z:Nat,?y:Nat),?z:Nat),min(?y:Nat,?z:Nat)),0),max(max(min(?y:Nat,?x:Nat),min(min(?x:Nat,0),min(?z:Nat,?x:Nat))),0))",
"min(?y:Nat,0) = min(max(min(max(?x:Nat,max(?y:Nat,?y:Nat)),?x:Nat),?y:Nat),max(max(min(?y:Nat,?x:Nat),min(?x:Nat,?y:Nat)),min(max(?x:Nat,?x:Nat),min(?y:Nat,0))))",
"min(?z:Nat,0) = max(min(min(min(?z:Nat,?z:Nat),max(?x:Nat,0)),?y:Nat),min(min(min(min(?y:Nat,?x:Nat),0),min(?x:Nat,?x:Nat)),min(min(0,?y:Nat),?x:Nat)))",
"max(max(0,?z:Nat),?x:Nat) = max(min(max(max(min(?z:Nat,?x:Nat),0),max(?x:Nat,min(?x:Nat,0))),?x:Nat),min(min(?y:Nat,?y:Nat),max(?y:Nat,max(0,?z:Nat))))",
"min(?y:Nat,max(?z:Nat,0)) = max(max(max(?x:Nat,min(max(0,?y:Nat),?y:Nat)),min(max(?x:Nat,max(?z:Nat,?x:Nat)),min(?y:Nat,0))),min(?y:Nat,max(?y:Nat,?y:Nat)))",
"max(max(?z:Nat,min(?z:Nat,0)),?y:Nat) = max(?z:Nat,min(max(?x:Nat,?x:Nat),min(max(max(?x:Nat,?y:Nat),max(?y:Nat,?y:Nat)),max(min(?y:Nat,?z:Nat),max(?x:Nat,?y:Nat)))))",
"min(max(0,?z:Nat),min(?y:Nat,?y:Nat)) = min(max(min(?x:Nat,min(?x:Nat,?x:Nat)),0),min(max(?y:Nat,?x:Nat),min(max(min(?y:Nat,?y:Nat),min(?z:Nat,?x:Nat)),?y:Nat)))",
"min(max(min(?y:Nat,?x:Nat),?y:Nat),?y:Nat) = min(max(min(?z:Nat,?z:Nat),?y:Nat),max(max(min(max(0,?x:Nat),max(?x:Nat,0)),max(?y:Nat,min(?z:Nat,?x:Nat))),0))",
"min(min(?y:Nat,max(?y:Nat,0)),?x:Nat) = min(min(max(min(?y:Nat,?x:Nat),?x:Nat),?z:Nat),max(max(max(max(0,?y:Nat),min(?x:Nat,?z:Nat)),?y:Nat),max(?y:Nat,?y:Nat)))",
"min(?z:Nat,min(?y:Nat,max(?y:Nat,max(max(?x:Nat,?y:Nat),?z:Nat)))) = min(max(0,max(?x:Nat,?x:Nat)),max(?x:Nat,max(?z:Nat,max(min(?z:Nat,?x:Nat),min(?y:Nat,?x:Nat)))))",
"min(0,min(min(?z:Nat,?x:Nat),min(?x:Nat,max(?x:Nat,?y:Nat)))) = min(min(max(?x:Nat,min(?x:Nat,min(0,?y:Nat))),?y:Nat),max(max(?y:Nat,?y:Nat),min(?y:Nat,?x:Nat)))",
"max(?x:Nat,max(0,min(min(max(?x:Nat,?x:Nat),min(?z:Nat,0)),?z:Nat))) = min(min(min(?x:Nat,?y:Nat),max(?y:Nat,?x:Nat)),max(?z:Nat,min(?y:Nat,min(?x:Nat,?y:Nat))))",
"max(0,max(min(?x:Nat,max(min(?x:Nat,?y:Nat),min(?y:Nat,?z:Nat))),?y:Nat)) = max(min(?y:Nat,0),max(min(?z:Nat,max(max(?z:Nat,?x:Nat),0)),max(?y:Nat,?y:Nat)))",
"max(max(?y:Nat,0),max(0,min(min(max(?z:Nat,?x:Nat),?y:Nat),?y:Nat))) = min(min(0,min(max(min(0,0),max(0,?x:Nat)),?x:Nat)),max(?x:Nat,?z:Nat))",
"max(max(?y:Nat,max(max(?y:Nat,?x:Nat),?y:Nat)),max(?z:Nat,max(?x:Nat,?y:Nat))) = max(min(?x:Nat,min(max(max(?z:Nat,0),?y:Nat),max(max(?y:Nat,?y:Nat),?y:Nat))),?y:Nat)",
"min(max(0,min(?y:Nat,?x:Nat)),min(?y:Nat,max(?x:Nat,max(0,?x:Nat)))) = max(max(min(0,?y:Nat),max(?y:Nat,?x:Nat)),max(min(?y:Nat,min(?x:Nat,?x:Nat)),?x:Nat))",
"min(max(max(0,?y:Nat),min(min(0,min(?z:Nat,?z:Nat)),?y:Nat)),?x:Nat) = min(min(max(?z:Nat,?x:Nat),max(max(max(?x:Nat,?x:Nat),max(?z:Nat,0)),0)),0)",
"max(?x:Nat,max(max(min(max(?x:Nat,?x:Nat),max(?y:Nat,?y:Nat)),max(0,?x:Nat)),?y:Nat)) = min(max(0,max(max(max(?y:Nat,0),?x:Nat),max(0,0))),?x:Nat)",
"max(max(?x:Nat,?z:Nat),min(?y:Nat,max(?z:Nat,max(min(?z:Nat,?x:Nat),max(?x:Nat,?z:Nat))))) = min(max(?z:Nat,?z:Nat),min(?y:Nat,max(max(?y:Nat,max(?z:Nat,?x:Nat)),?x:Nat)))",
"max(max(?x:Nat,min(max(max(?y:Nat,?y:Nat),?x:Nat),max(?x:Nat,?x:Nat))),max(0,?y:Nat)) = max(?x:Nat,min(min(0,max(max(?z:Nat,0),max(?y:Nat,?x:Nat))),0))",
"max(max(?y:Nat,min(min(?y:Nat,max(?y:Nat,?x:Nat)),min(min(?x:Nat,?y:Nat),?y:Nat))),0) = max(min(min(max(?y:Nat,0),max(0,0)),max(?z:Nat,?y:Nat)),?x:Nat)",
"max(max(min(max(min(0,?y:Nat),max(?x:Nat,0)),?y:Nat),?z:Nat),max(?z:Nat,?x:Nat)) = min(max(0,max(max(max(?x:Nat,?x:Nat),max(?y:Nat,?x:Nat)),?x:Nat)),0)",
"max(min(?x:Nat,max(min(0,max(?x:Nat,0)),min(max(?y:Nat,?x:Nat),?x:Nat))),?z:Nat) = min(?x:Nat,min(?x:Nat,min(?y:Nat,min(max(?y:Nat,?z:Nat),min(?x:Nat,?z:Nat)))))",
"min(max(min(0,?x:Nat),?z:Nat),min(?y:Nat,min(?x:Nat,max(max(?x:Nat,?y:Nat),?z:Nat)))) = max(max(?y:Nat,max(min(?x:Nat,min(?y:Nat,?x:Nat)),max(0,?y:Nat))),?x:Nat)",
"min(min(?x:Nat,0),min(min(min(?x:Nat,?z:Nat),max(0,?z:Nat)),max(?y:Nat,?z:Nat))) = min(?z:Nat,min(min(?x:Nat,max(max(?y:Nat,?x:Nat),max(?x:Nat,?y:Nat))),?z:Nat))",
"min(min(?z:Nat,?x:Nat),min(?x:Nat,max(0,max(min(?x:Nat,0),min(?y:Nat,0))))) = min(min(?y:Nat,max(max(?y:Nat,?z:Nat),max(0,min(?y:Nat,?x:Nat)))),?x:Nat)",
"max(?z:Nat,min(max(min(?y:Nat,?x:Nat),min(?x:Nat,?z:Nat)),min(max(min(?y:Nat,?x:Nat),?x:Nat),?y:Nat))) = max(max(min(?x:Nat,0),?x:Nat),max(?x:Nat,max(?x:Nat,?z:Nat)))",
"max(max(?z:Nat,min(?x:Nat,?z:Nat)),min(max(min(min(0,?y:Nat),max(?y:Nat,?x:Nat)),0),?x:Nat)) = max(min(min(max(0,max(?x:Nat,?y:Nat)),?x:Nat),?x:Nat),?x:Nat)",
"max(max(max(?z:Nat,max(max(0,?x:Nat),?y:Nat)),min(min(?y:Nat,?x:Nat),?x:Nat)),min(?y:Nat,?y:Nat)) = min(min(?x:Nat,0),max(min(?y:Nat,min(?x:Nat,0)),?y:Nat))",
"min(max(max(max(?z:Nat,?y:Nat),?z:Nat),?z:Nat),min(max(?y:Nat,?y:Nat),max(?x:Nat,min(?x:Nat,0)))) = min(min(max(min(0,min(?x:Nat,0)),?y:Nat),?z:Nat),?x:Nat)",
"min(min(min(max(max(0,?x:Nat),max(?x:Nat,?y:Nat)),?x:Nat),?z:Nat),min(max(0,?x:Nat),?y:Nat)) = min(max(?x:Nat,max(max(?z:Nat,max(?z:Nat,0)),?y:Nat)),?y:Nat)",
"max(max(?y:Nat,max(max(max(?y:Nat,?x:Nat),?y:Nat),?y:Nat)),max(min(min(min(?x:Nat,?x:Nat),?x:Nat),?x:Nat),?z:Nat)) = max(max(min(0,?x:Nat),0),min(?y:Nat,?x:Nat))",
"max(?x:Nat,min(max(max(?z:Nat,min(?y:Nat,?x:Nat)),?y:Nat),min(min(0,?z:Nat),max(min(?y:Nat,?x:Nat),min(?x:Nat,?x:Nat))))) = min(max(?z:Nat,min(0,?x:Nat)),?y:Nat)",
"min(max(min(max(?y:Nat,min(0,?y:Nat)),?x:Nat),max(?y:Nat,?x:Nat)),min(max(max(0,min(?y:Nat,0)),?x:Nat),?y:Nat)) = max(?x:Nat,max(?x:Nat,max(?z:Nat,?x:Nat)))",
"max(min(0,min(min(?x:Nat,min(?x:Nat,?y:Nat)),min(?y:Nat,max(?y:Nat,?y:Nat)))),min(min(max(?y:Nat,min(?y:Nat,?z:Nat)),?x:Nat),?y:Nat)) = min(0,max(?y:Nat,?x:Nat))",
"min(?x:Nat,min(min(min(max(?x:Nat,?x:Nat),min(?y:Nat,?x:Nat)),min(max(0,?y:Nat),?x:Nat)),min(?z:Nat,max(0,max(?z:Nat,0))))) = min(?y:Nat,max(?x:Nat,?x:Nat))",
"min(?z:Nat,min(min(max(?y:Nat,max(?y:Nat,?y:Nat)),min(?z:Nat,max(?z:Nat,?y:Nat))),max(?x:Nat,min(min(?y:Nat,?z:Nat),max(?y:Nat,0))))) = max(min(?y:Nat,0),?z:Nat)",
"min(min(min(?z:Nat,?y:Nat),min(?y:Nat,min(?y:Nat,?y:Nat))),min(?x:Nat,max(min(?x:Nat,max(?x:Nat,?y:Nat)),min(max(?y:Nat,?y:Nat),?y:Nat)))) = max(?x:Nat,min(0,?x:Nat))",
"max(?x:Nat,min(max(min(?x:Nat,min(0,?x:Nat)),min(min(?y:Nat,?y:Nat),min(0,?y:Nat))),min(min(min(?y:Nat,?x:Nat),max(?z:Nat,?x:Nat)),?z:Nat))) = min(0,?z:Nat)",
"min(?x:Nat,max(min(min(min(?x:Nat,0),?x:Nat),min(max(?y:Nat,?y:Nat),0)),max(min(?x:Nat,?y:Nat),min(max(?y:Nat,?x:Nat),max(?y:Nat,?y:Nat))))) = min(?y:Nat,?y:Nat)",
"max(0,?y:Nat) = max(max(max(?z:Nat,max(?y:Nat,max(?x:Nat,0))),max(?x:Nat,min(min(?x:Nat,?y:Nat),0))),max(min(min(?z:Nat,?z:Nat),max(min(?y:Nat,?x:Nat),0)),?y:Nat))",
"max(max(?x:Nat,?x:Nat),0) = min(max(max(max(?y:Nat,?x:Nat),min(min(?y:Nat,?y:Nat),min(?z:Nat,0))),max(min(min(?y:Nat,?x:Nat),?z:Nat),max(?y:Nat,min(?x:Nat,?x:Nat)))),0)",
"max(max(?y:Nat,?y:Nat),?z:Nat) = max(max(min(max(?y:Nat,min(?y:Nat,?x:Nat)),min(max(?y:Nat,?x:Nat),min(?y:Nat,?x:Nat))),min(max(max(?x:Nat,?x:Nat),max(?x:Nat,?x:Nat)),?y:Nat)),0)",
"max(max(?z:Nat,?y:Nat),max(max(?x:Nat,?z:Nat),min(?x:Nat,?z:Nat))) = min(0,min(max(max(0,min(?y:Nat,?y:Nat)),?y:Nat),max(max(min(0,?x:Nat),max(?z:Nat,?x:Nat)),?z:Nat)))",
"min(max(?z:Nat,max(?z:Nat,?x:Nat)),min(?x:Nat,max(?y:Nat,0))) = min(max(min(?x:Nat,min(?x:Nat,max(?y:Nat,?x:Nat))),?z:Nat),min(min(?x:Nat,min(?z:Nat,min(?x:Nat,?x:Nat))),?x:Nat))",
"min(min(?y:Nat,max(?x:Nat,min(0,max(?y:Nat,?y:Nat)))),?x:Nat) = min(min(max(min(?x:Nat,?x:Nat),max(0,min(?x:Nat,?x:Nat))),min(max(?z:Nat,min(?z:Nat,?x:Nat)),?x:Nat)),?y:Nat)",
"min(min(max(?y:Nat,max(min(0,0),?y:Nat)),?z:Nat),?x:Nat) = min(min(max(max(min(?x:Nat,?y:Nat),min(?y:Nat,?x:Nat)),max(max(?x:Nat,?x:Nat),?y:Nat)),?y:Nat),min(?z:Nat,?y:Nat))",
"min(min(min(min(min(?x:Nat,?y:Nat),?x:Nat),?y:Nat),?y:Nat),?x:Nat) = min(?y:Nat,min(min(min(?y:Nat,min(?y:Nat,?x:Nat)),?z:Nat),min(?x:Nat,min(min(?y:Nat,?y:Nat),max(?x:Nat,?x:Nat)))))",
"max(0,min(max(?y:Nat,max(min(?y:Nat,?x:Nat),max(?x:Nat,?x:Nat))),?x:Nat)) = max(min(min(min(?z:Nat,0),max(max(0,?x:Nat),?x:Nat)),max(?z:Nat,?z:Nat)),min(?z:Nat,0))",
"max(max(?x:Nat,?y:Nat),min(?y:Nat,min(min(?y:Nat,max(0,?z:Nat)),0))) = max(max(max(min(min(?x:Nat,?z:Nat),?x:Nat),min(min(?x:Nat,?z:Nat),?y:Nat)),max(0,?y:Nat)),?z:Nat)",
"max(min(max(?x:Nat,min(0,max(?z:Nat,0))),max(?y:Nat,0)),0) = max(max(max(?y:Nat,?y:Nat),max(max(max(?y:Nat,?x:Nat),min(?x:Nat,?x:Nat)),max(?x:Nat,?x:Nat))),?x:Nat)",
"min(min(?x:Nat,min(min(?x:Nat,?z:Nat),min(0,?x:Nat))),max(?y:Nat,0)) = max(min(min(max(min(?y:Nat,?y:Nat),min(?y:Nat,?x:Nat)),min(?x:Nat,max(?x:Nat,?x:Nat))),?x:Nat),?y:Nat)",
"min(min(?y:Nat,min(max(min(?x:Nat,?z:Nat),max(?y:Nat,0)),?y:Nat)),?x:Nat) = min(max(?z:Nat,?x:Nat),min(min(min(?y:Nat,max(0,?x:Nat)),min(max(?z:Nat,0),?z:Nat)),?x:Nat))",
"max(0,max(?x:Nat,min(max(max(?y:Nat,0),?x:Nat),max(?z:Nat,min(?x:Nat,0))))) = min(max(?y:Nat,min(max(min(?y:Nat,0),?x:Nat),max(?y:Nat,min(?z:Nat,?x:Nat)))),?y:Nat)",
"min(max(min(max(min(?x:Nat,?y:Nat),max(?x:Nat,?y:Nat)),0),min(?x:Nat,?y:Nat)),?y:Nat) = max(max(?z:Nat,min(?y:Nat,min(0,min(?x:Nat,?x:Nat)))),min(?z:Nat,max(?y:Nat,?y:Nat)))",
"max(max(max(?x:Nat,max(min(?x:Nat,?y:Nat),?y:Nat)),max(?x:Nat,min(?y:Nat,?y:Nat))),max(?x:Nat,?y:Nat)) = min(?y:Nat,min(max(min(max(?y:Nat,?x:Nat),min(?y:Nat,?y:Nat)),?x:Nat),?y:Nat))",
"max(max(min(min(?z:Nat,min(?x:Nat,?z:Nat)),?x:Nat),min(?x:Nat,max(?x:Nat,min(?x:Nat,?x:Nat)))),?y:Nat) = min(min(?y:Nat,max(min(max(?y:Nat,?y:Nat),?x:Nat),?y:Nat)),min(?z:Nat,?y:Nat))",
"min(?y:Nat,max(min(?z:Nat,?y:Nat),min(max(0,max(?y:Nat,0)),min(min(?x:Nat,?z:Nat),?x:Nat)))) = max(max(?x:Nat,min(max(?y:Nat,max(?y:Nat,0)),max(?y:Nat,?y:Nat))),?y:Nat)",
"min(?y:Nat,min(min(min(max(?x:Nat,?y:Nat),max(?y:Nat,?y:Nat)),?y:Nat),max(max(?x:Nat,?x:Nat),0))) = max(max(max(min(max(0,0),max(?z:Nat,?x:Nat)),0),?z:Nat),?y:Nat)",
"min(max(min(max(?z:Nat,?y:Nat),0),min(max(min(?x:Nat,?z:Nat),max(?x:Nat,0)),?z:Nat)),?z:Nat) = min(max(max(?y:Nat,max(max(?z:Nat,?x:Nat),?x:Nat)),min(?z:Nat,?y:Nat)),?z:Nat)",
"max(max(?y:Nat,max(max(min(?x:Nat,?z:Nat),?x:Nat),?z:Nat)),max(?x:Nat,min(?z:Nat,min(max(?y:Nat,?z:Nat),?x:Nat)))) = min(min(?y:Nat,max(max(?x:Nat,?y:Nat),max(?x:Nat,?y:Nat))),?z:Nat)",
"max(min(max(min(?x:Nat,?x:Nat),?z:Nat),min(min(0,max(0,?y:Nat)),min(?z:Nat,0))),min(?z:Nat,?y:Nat)) = max(?y:Nat,max(max(0,min(min(?x:Nat,?y:Nat),0)),?y:Nat))",
"min(?x:Nat,min(min(?y:Nat,min(?y:Nat,min(?x:Nat,0))),max(min(min(0,?x:Nat),?y:Nat),min(?x:Nat,?x:Nat)))) = max(?x:Nat,min(0,max(max(0,max(?x:Nat,?z:Nat)),?x:Nat)))",
"min(0,min(max(max(?y:Nat,?z:Nat),?x:Nat),min(max(?y:Nat,0),min(max(?z:Nat,?z:Nat),max(0,?x:Nat))))) = max(max(0,?y:Nat),min(?x:Nat,min(0,max(?y:Nat,0))))",
"max(max(?y:Nat,max(max(max(?y:Nat,?x:Nat),min(?z:Nat,?y:Nat)),?z:Nat)),max(max(max(?x:Nat,?y:Nat),min(?x:Nat,?x:Nat)),0)) = min(max(min(?z:Nat,0),max(?y:Nat,0)),?y:Nat)",
"min(min(?x:Nat,min(?x:Nat,max(max(?x:Nat,?y:Nat),min(?y:Nat,?y:Nat)))),max(max(min(max(?z:Nat,0),?z:Nat),?y:Nat),?x:Nat)) = min(max(max(?x:Nat,?x:Nat),?z:Nat),min(?x:Nat,0))",
"min(max(max(min(?z:Nat,max(?z:Nat,?y:Nat)),min(?y:Nat,max(?y:Nat,?z:Nat))),min(min(?y:Nat,?y:Nat),min(?y:Nat,max(0,?y:Nat)))),?y:Nat) = max(max(max(0,?x:Nat),?y:Nat),0)",
"min(max(?y:Nat,max(?z:Nat,min(max(?x:Nat,?x:Nat),min(?x:Nat,?y:Nat)))),max(min(max(max(?z:Nat,?x:Nat),?y:Nat),max(?x:Nat,max(?z:Nat,?x:Nat))),0)) = min(?x:Nat,min(?y:Nat,?y:Nat))",
"min(min(max(?y:Nat,min(min(0,?z:Nat),max(0,?y:Nat))),?z:Nat),max(min(?z:Nat,max(min(?x:Nat,?x:Nat),min(?y:Nat,?y:Nat))),min(?x:Nat,?x:Nat))) = min(0,min(?y:Nat,?y:Nat))",
"max(min(max(?y:Nat,?y:Nat),max(max(max(?y:Nat,?z:Nat),min(?y:Nat,?y:Nat)),?x:Nat)),min(max(max(min(?y:Nat,0),max(?z:Nat,?x:Nat)),max(?x:Nat,?y:Nat)),?x:Nat)) = min(?y:Nat,?z:Nat)",
"min(max(max(?y:Nat,min(max(?x:Nat,?y:Nat),max(?y:Nat,?y:Nat))),min(min(min(?x:Nat,?x:Nat),0),min(?y:Nat,?y:Nat))),max(max(0,max(?z:Nat,?y:Nat)),?y:Nat)) = min(?x:Nat,?y:Nat)",
"min(?y:Nat,?y:Nat) = min(max(0,max(max(max(?x:Nat,?y:Nat),min(?x:Nat,0)),max(?y:Nat,0))),min(min(min(?x:Nat,min(?y:Nat,?y:Nat)),min(?y:Nat,max(?y:Nat,?x:Nat))),max(?y:Nat,?z:Nat)))",
"min(0,?y:Nat) = max(?z:Nat,min(max(min(min(?y:Nat,?z:Nat),min(?z:Nat,?z:Nat)),min(max(?x:Nat,?z:Nat),min(?z:Nat,?z:Nat))),max(max(?y:Nat,max(?z:Nat,?z:Nat)),min(?y:Nat,min(?x:Nat,?x:Nat)))))",
"max(max(?z:Nat,max(?x:Nat,0)),0) = min(max(?y:Nat,?y:Nat),max(min(max(max(?x:Nat,?x:Nat),max(?x:Nat,0)),max(max(?y:Nat,0),?x:Nat)),min(max(max(?y:Nat,?x:Nat),?z:Nat),0)))",
"max(min(max(?y:Nat,?y:Nat),?z:Nat),?x:Nat) = min(min(max(max(max(0,?y:Nat),?y:Nat),min(max(0,?y:Nat),?x:Nat)),min(?y:Nat,min(max(0,?x:Nat),?y:Nat))),min(0,max(?z:Nat,?y:Nat)))",
"max(?y:Nat,max(min(?y:Nat,max(min(?y:Nat,?y:Nat),max(0,?y:Nat))),?y:Nat)) = min(?x:Nat,max(min(max(0,min(?y:Nat,?y:Nat)),min(min(?y:Nat,?z:Nat),min(?y:Nat,?y:Nat))),max(?x:Nat,?y:Nat)))",
"max(max(max(max(0,?y:Nat),max(?x:Nat,max(0,0))),?x:Nat),?y:Nat) = min(max(max(?x:Nat,?x:Nat),max(max(?x:Nat,max(?x:Nat,?y:Nat)),max(max(0,0),max(?y:Nat,?y:Nat)))),?z:Nat)",
"min(min(max(?y:Nat,min(?x:Nat,?x:Nat)),max(?y:Nat,?y:Nat)),max(0,?x:Nat)) = min(max(min(min(max(?x:Nat,0),0),?x:Nat),max(max(?z:Nat,min(?x:Nat,0)),?y:Nat)),max(?y:Nat,?x:Nat))",
"min(min(max(max(max(?y:Nat,?y:Nat),?z:Nat),min(?z:Nat,?z:Nat)),?y:Nat),?z:Nat) = min(max(min(0,?x:Nat),min(0,max(max(?z:Nat,?z:Nat),max(?z:Nat,?z:Nat)))),max(min(?y:Nat,?z:Nat),?x:Nat))",
"max(min(0,max(max(max(?x:Nat,?y:Nat),min(0,?x:Nat)),max(?y:Nat,?y:Nat))),?x:Nat) = min(min(max(max(?z:Nat,max(?x:Nat,0)),0),min(max(?x:Nat,?z:Nat),0)),max(?z:Nat,?y:Nat))",
"min(min(max(?y:Nat,0),max(max(min(?x:Nat,?x:Nat),max(0,0)),?x:Nat)),?x:Nat) = max(min(min(max(max(?y:Nat,?x:Nat),min(?x:Nat,?y:Nat)),min(?x:Nat,min(?y:Nat,?z:Nat))),?x:Nat),?y:Nat)",
"max(min(max(min(max(?y:Nat,?y:Nat),0),max(?z:Nat,?x:Nat)),?x:Nat),min(min(?x:Nat,?x:Nat),?y:Nat)) = min(max(max(min(min(?z:Nat,?y:Nat),max(?y:Nat,?y:Nat)),max(?x:Nat,?y:Nat)),?y:Nat),?y:Nat)",
"min(max(max(?z:Nat,?y:Nat),min(?y:Nat,?x:Nat)),min(max(min(min(?x:Nat,?y:Nat),0),?x:Nat),?y:Nat)) = min(max(max(?y:Nat,?x:Nat),?z:Nat),min(?z:Nat,max(?z:Nat,max(?x:Nat,max(?y:Nat,?y:Nat)))))",
"max(min(max(?x:Nat,0),min(max(?y:Nat,?x:Nat),?x:Nat)),max(min(max(min(?y:Nat,?z:Nat),?z:Nat),?x:Nat),?y:Nat)) = min(min(?x:Nat,?x:Nat),max(max(max(0,?y:Nat),min(?z:Nat,?x:Nat)),?x:Nat))",
"min(max(max(min(?x:Nat,?y:Nat),max(?x:Nat,?z:Nat)),?x:Nat),min(max(min(?y:Nat,?x:Nat),max(?y:Nat,0)),?y:Nat)) = max(max(0,min(?y:Nat,?z:Nat)),max(?x:Nat,min(?y:Nat,max(0,?x:Nat))))",
"max(max(?y:Nat,min(min(0,?x:Nat),min(?x:Nat,max(?z:Nat,?y:Nat)))),max(min(min(?x:Nat,max(?x:Nat,?y:Nat)),?z:Nat),0)) = max(?x:Nat,min(min(max(max(0,?z:Nat),0),?x:Nat),?x:Nat))",
"max(max(max(max(?y:Nat,max(?y:Nat,?y:Nat)),min(min(?x:Nat,?z:Nat),?y:Nat)),min(?z:Nat,max(?y:Nat,?z:Nat))),min(?x:Nat,?y:Nat)) = max(0,min(max(?x:Nat,min(?x:Nat,min(?y:Nat,?x:Nat))),?z:Nat))",
"max(?y:Nat,min(min(min(?y:Nat,max(?y:Nat,?z:Nat)),max(min(0,?y:Nat),?x:Nat)),min(max(?x:Nat,min(?y:Nat,?x:Nat)),min(?x:Nat,?y:Nat)))) = max(?x:Nat,max(max(?y:Nat,min(?x:Nat,0)),?y:Nat))",
"max(min(max(max(max(?z:Nat,?x:Nat),0),max(?y:Nat,0)),max(max(?z:Nat,max(?z:Nat,0)),min(?y:Nat,min(0,?x:Nat)))),?x:Nat) = min(min(?x:Nat,?y:Nat),min(?y:Nat,max(?x:Nat,?z:Nat)))",
"max(max(min(?x:Nat,max(max(?x:Nat,?z:Nat),0)),min(min(min(?x:Nat,?x:Nat),?x:Nat),?z:Nat)),min(?x:Nat,min(?y:Nat,max(max(0,?x:Nat),max(?x:Nat,?x:Nat))))) = max(?y:Nat,max(?y:Nat,?z:Nat))",
"min(min(max(?z:Nat,?y:Nat),max(0,min(min(0,?x:Nat),max(?x:Nat,?y:Nat)))),min(?y:Nat,max(max(?z:Nat,?z:Nat),min(min(?x:Nat,?x:Nat),min(?y:Nat,?x:Nat))))) = max(max(?x:Nat,?x:Nat),?y:Nat)",
"min(min(min(max(min(?y:Nat,?x:Nat),max(?z:Nat,?y:Nat)),0),max(max(?x:Nat,?x:Nat),min(0,?x:Nat))),max(min(?z:Nat,min(?y:Nat,max(?x:Nat,?y:Nat))),?y:Nat)) = max(?y:Nat,max(?x:Nat,?y:Nat))",
"max(max(min(?y:Nat,min(max(?x:Nat,?x:Nat),max(?x:Nat,?x:Nat))),?x:Nat),max(max(min(max(?z:Nat,0),min(?x:Nat,0)),max(min(?x:Nat,?x:Nat),max(?y:Nat,?y:Nat))),0)) = min(0,?z:Nat)",
"max(min(min(?z:Nat,min(?y:Nat,?y:Nat)),?x:Nat),max(min(min(min(0,?x:Nat),min(0,?x:Nat)),min(min(?y:Nat,?y:Nat),?y:Nat)),max(min(0,max(?x:Nat,0)),?x:Nat))) = min(?y:Nat,0)",
"min(?y:Nat,?x:Nat) = min(max(max(max(?x:Nat,?y:Nat),max(?z:Nat,min(0,?x:Nat))),min(max(min(?y:Nat,?x:Nat),min(?z:Nat,?z:Nat)),?z:Nat)),min(max(min(?z:Nat,?x:Nat),min(min(?x:Nat,?z:Nat),?y:Nat)),?x:Nat))",
"min(max(?z:Nat,?y:Nat),?z:Nat) = max(min(min(max(max(?x:Nat,0),min(0,?x:Nat)),max(0,max(?z:Nat,?y:Nat))),0),min(max(min(0,max(?x:Nat,?y:Nat)),max(?y:Nat,?y:Nat)),min(?y:Nat,?z:Nat)))",
"max(max(?x:Nat,?y:Nat),min(?z:Nat,0)) = min(?y:Nat,max(min(max(max(?y:Nat,?x:Nat),max(?z:Nat,?y:Nat)),min(min(?y:Nat,?y:Nat),min(?y:Nat,?x:Nat))),max(max(max(0,0),max(?x:Nat,?x:Nat)),0)))",
"max(?z:Nat,min(min(?x:Nat,max(0,min(?y:Nat,?x:Nat))),?z:Nat)) = min(min(max(?z:Nat,max(max(?y:Nat,?x:Nat),min(?x:Nat,?x:Nat))),?y:Nat),max(min(min(0,min(?z:Nat,0)),max(?x:Nat,?y:Nat)),?y:Nat))",
"min(?y:Nat,max(?z:Nat,min(?y:Nat,min(max(?x:Nat,?y:Nat),min(?y:Nat,?y:Nat))))) = max(min(min(max(?y:Nat,min(?y:Nat,?y:Nat)),0),?y:Nat),min(?x:Nat,max(max(?y:Nat,?x:Nat),min(?z:Nat,max(?x:Nat,?y:Nat)))))",
"min(min(max(?x:Nat,min(min(?y:Nat,?y:Nat),max(?x:Nat,?y:Nat))),0),?y:Nat) = min(max(max(?z:Nat,?x:Nat),min(?x:Nat,min(0,min(?y:Nat,?y:Nat)))),min(min(?x:Nat,min(max(?y:Nat,0),?z:Nat)),?z:Nat))",
"max(max(?x:Nat,max(max(max(?y:Nat,?x:Nat),0),max(min(?z:Nat,?y:Nat),?x:Nat))),max(?y:Nat,0)) = max(max(min(0,min(max(?y:Nat,?x:Nat),min(?x:Nat,?x:Nat))),min(min(?y:Nat,?y:Nat),?z:Nat)),?x:Nat)",
"max(max(?y:Nat,min(?z:Nat,min(?y:Nat,?y:Nat))),max(?y:Nat,max(max(?y:Nat,max(0,?y:Nat)),?x:Nat))) = min(max(min(min(min(?y:Nat,0),?y:Nat),?y:Nat),?z:Nat),max(max(?x:Nat,?y:Nat),max(?y:Nat,?x:Nat)))",
"max(max(min(min(?y:Nat,min(0,?x:Nat)),min(max(?x:Nat,?y:Nat),max(0,?y:Nat))),max(min(?z:Nat,?y:Nat),?y:Nat)),?y:Nat) = min(?y:Nat,max(?x:Nat,min(max(0,?y:Nat),max(?x:Nat,min(?y:Nat,?x:Nat)))))",
"max(min(max(?x:Nat,max(?x:Nat,max(0,?x:Nat))),?y:Nat),max(?y:Nat,min(0,min(min(?y:Nat,0),min(0,?x:Nat))))) = max(min(?x:Nat,min(max(?z:Nat,?y:Nat),max(?x:Nat,min(?x:Nat,?y:Nat)))),0)",
"min(min(?y:Nat,min(?x:Nat,max(?z:Nat,min(?y:Nat,0)))),max(max(min(max(?x:Nat,?y:Nat),max(?y:Nat,0)),?x:Nat),0)) = max(min(max(?x:Nat,min(?y:Nat,?y:Nat)),?y:Nat),min(?y:Nat,max(?y:Nat,?z:Nat)))",
"min(max(min(max(?x:Nat,min(?y:Nat,?y:Nat)),max(min(?x:Nat,?x:Nat),min(?y:Nat,0))),?z:Nat),max(min(?x:Nat,min(?y:Nat,?z:Nat)),0)) = min(?y:Nat,max(?y:Nat,min(min(min(?z:Nat,?x:Nat),?z:Nat),?y:Nat)))",
"max(min(?y:Nat,min(?x:Nat,min(?y:Nat,max(?y:Nat,?y:Nat)))),min(?y:Nat,min(min(min(?y:Nat,?x:Nat),max(?y:Nat,?x:Nat)),min(min(?x:Nat,?y:Nat),?x:Nat)))) = min(max(0,min(min(?z:Nat,0),?x:Nat)),?y:Nat)",
"max(min(max(?x:Nat,min(max(0,?x:Nat),max(?z:Nat,?x:Nat))),max(?y:Nat,max(min(?z:Nat,0),min(0,?z:Nat)))),min(?x:Nat,min(min(?y:Nat,?y:Nat),?x:Nat))) = min(max(min(?y:Nat,?z:Nat),?z:Nat),?z:Nat)",
"max(max(max(?y:Nat,min(0,?y:Nat)),min(max(?x:Nat,max(?x:Nat,?x:Nat)),max(min(0,?y:Nat),0))),min(max(min(max(?x:Nat,?x:Nat),min(?x:Nat,?y:Nat)),?x:Nat),?y:Nat)) = max(?x:Nat,min(?y:Nat,?z:Nat))",
"max(max(max(min(max(0,?y:Nat),?y:Nat),0),max(?y:Nat,max(max(?x:Nat,?z:Nat),min(?y:Nat,0)))),min(max(?y:Nat,?x:Nat),max(min(?x:Nat,?x:Nat),max(min(?z:Nat,?y:Nat),?y:Nat)))) = max(?x:Nat,?y:Nat)",
"max(?z:Nat,?x:Nat) = min(min(max(max(max(?x:Nat,?x:Nat),?z:Nat),min(max(?x:Nat,?x:Nat),max(?z:Nat,?z:Nat))),min(?y:Nat,?y:Nat)),max(max(max(0,0),?z:Nat),max(min(min(0,?z:Nat),min(?z:Nat,?y:Nat)),?x:Nat)))",
"min(?z:Nat,min(0,?y:Nat)) = min(max(max(max(?y:Nat,?x:Nat),min(?y:Nat,max(?y:Nat,?y:Nat))),0),min(max(min(?z:Nat,?x:Nat),max(min(?y:Nat,?y:Nat),?z:Nat)),max(min(max(?y:Nat,?x:Nat),min(?x:Nat,?y:Nat)),?x:Nat)))",
"min(max(?x:Nat,?z:Nat),0) = max(?y:Nat,max(max(max(max(?x:Nat,0),max(?x:Nat,?z:Nat)),max(max(0,0),min(?y:Nat,?z:Nat))),max(min(max(?y:Nat,?x:Nat),max(?x:Nat,?z:Nat)),min(?y:Nat,min(?x:Nat,?y:Nat)))))",
"min(max(?y:Nat,?z:Nat),?y:Nat) = max(min(?x:Nat,max(max(min(?z:Nat,?y:Nat),min(?y:Nat,?z:Nat)),min(max(?x:Nat,0),?x:Nat))),max(min(min(?y:Nat,max(?z:Nat,?x:Nat)),0),min(max(max(?x:Nat,?x:Nat),?x:Nat),?y:Nat)))",
"min(min(min(?z:Nat,?x:Nat),?y:Nat),?y:Nat) = max(max(max(min(max(0,?x:Nat),?y:Nat),?x:Nat),max(?z:Nat,max(0,?x:Nat))),min(min(min(min(?x:Nat,?y:Nat),max(?z:Nat,?y:Nat)),?x:Nat),min(max(0,?z:Nat),?y:Nat)))",
"min(min(?x:Nat,?y:Nat),max(?x:Nat,max(?y:Nat,max(?y:Nat,?z:Nat)))) = max(max(max(?y:Nat,max(max(?y:Nat,?x:Nat),?y:Nat)),0),max(max(min(max(?y:Nat,?x:Nat),?z:Nat),max(min(?x:Nat,?x:Nat),?x:Nat)),min(?y:Nat,?x:Nat)))",
"min(min(?z:Nat,max(max(max(?x:Nat,?x:Nat),?x:Nat),?x:Nat)),?x:Nat) = min(max(?z:Nat,max(max(max(0,?x:Nat),min(?y:Nat,?z:Nat)),min(?z:Nat,max(?x:Nat,?z:Nat)))),max(max(min(min(?x:Nat,?z:Nat),?x:Nat),?y:Nat),?x:Nat))",
"max(?y:Nat,min(min(min(?y:Nat,?y:Nat),min(max(?x:Nat,?z:Nat),min(0,?y:Nat))),?x:Nat)) = min(?z:Nat,min(max(max(?y:Nat,0),min(max(?y:Nat,?y:Nat),max(?x:Nat,?y:Nat))),max(min(0,?y:Nat),min(?x:Nat,?z:Nat))))",
"min(?y:Nat,max(max(?y:Nat,max(?x:Nat,max(?z:Nat,?x:Nat))),max(0,min(?y:Nat,?x:Nat)))) = min(min(min(?x:Nat,?x:Nat),max(max(max(?x:Nat,?x:Nat),min(?x:Nat,0)),min(?y:Nat,min(?x:Nat,0)))),min(?y:Nat,?x:Nat))",
"min(max(min(max(?x:Nat,max(?x:Nat,?x:Nat)),min(?y:Nat,0)),min(0,min(?z:Nat,?x:Nat))),?x:Nat) = max(max(0,max(0,min(?y:Nat,min(0,?x:Nat)))),min(?z:Nat,max(max(?x:Nat,?x:Nat),max(?y:Nat,?y:Nat))))",
"min(?y:Nat,max(max(min(min(?y:Nat,?z:Nat),?x:Nat),min(?y:Nat,?x:Nat)),min(min(min(?x:Nat,?x:Nat),?y:Nat),?x:Nat))) = min(?x:Nat,max(?y:Nat,max(min(max(?x:Nat,?x:Nat),min(?y:Nat,?y:Nat)),min(?z:Nat,max(?z:Nat,?x:Nat)))))",
"min(max(min(?y:Nat,?x:Nat),min(min(max(?x:Nat,?x:Nat),min(?x:Nat,0)),?x:Nat)),min(max(max(?x:Nat,0),?z:Nat),?x:Nat)) = min(?z:Nat,min(min(max(0,?z:Nat),?y:Nat),max(min(?x:Nat,max(0,0)),0)))",
"min(max(min(min(min(0,?x:Nat),?x:Nat),min(min(?y:Nat,?x:Nat),?x:Nat)),min(0,min(max(?y:Nat,?x:Nat),max(?x:Nat,?y:Nat)))),?y:Nat) = max(max(?x:Nat,0),min(max(max(max(?z:Nat,?z:Nat),?z:Nat),?y:Nat),?y:Nat))",
"min(min(?x:Nat,0),max(max(?z:Nat,max(max(0,?x:Nat),0)),min(max(max(?y:Nat,?y:Nat),?y:Nat),max(max(?y:Nat,0),max(?y:Nat,?y:Nat))))) = max(max(min(min(max(?z:Nat,?z:Nat),?x:Nat),?z:Nat),0),?y:Nat)",
"min(max(max(min(?x:Nat,max(?x:Nat,?y:Nat)),max(?x:Nat,max(0,?x:Nat))),max(max(max(?x:Nat,?y:Nat),?z:Nat),max(max(?z:Nat,?z:Nat),min(0,?x:Nat)))),max(?z:Nat,?y:Nat)) = min(max(max(?z:Nat,?y:Nat),?x:Nat),?y:Nat)",
"max(?x:Nat,max(max(?x:Nat,?z:Nat),max(max(?z:Nat,?z:Nat),?x:Nat))) = min(max(?x:Nat,min(min(min(?x:Nat,?z:Nat),max(?z:Nat,?z:Nat)),max(min(?y:Nat,?z:Nat),min(0,0)))),max(min(?x:Nat,min(?y:Nat,min(?x:Nat,?y:Nat))),?z:Nat))",
"max(min(max(max(0,?z:Nat),max(?z:Nat,max(?y:Nat,?y:Nat))),?y:Nat),?x:Nat) = max(max(max(max(min(?x:Nat,?y:Nat),0),?z:Nat),max(?z:Nat,max(?x:Nat,0))),max(?y:Nat,max(0,min(min(?z:Nat,?y:Nat),max(?x:Nat,0)))))",
"min(max(min(?x:Nat,?y:Nat),max(?x:Nat,0)),max(min(min(min(0,?y:Nat),min(?z:Nat,?y:Nat)),0),?x:Nat)) = min(?y:Nat,max(?x:Nat,max(min(max(?y:Nat,0),max(?y:Nat,?z:Nat)),max(min(?x:Nat,?x:Nat),max(0,0)))))",
"min(max(min(max(min(?y:Nat,?x:Nat),0),?y:Nat),max(min(0,max(0,?y:Nat)),min(?x:Nat,?y:Nat))),?x:Nat) = max(max(?y:Nat,?y:Nat),max(0,max(max(min(?y:Nat,?y:Nat),max(?x:Nat,?x:Nat)),min(?y:Nat,max(?x:Nat,?y:Nat)))))",
"min(max(min(?x:Nat,min(?x:Nat,max(?x:Nat,0))),min(min(min(?y:Nat,?x:Nat),0),min(max(?z:Nat,0),?y:Nat))),0) = max(?x:Nat,min(max(?z:Nat,max(?y:Nat,max(?x:Nat,?x:Nat))),min(max(max(?y:Nat,0),?y:Nat),?x:Nat)))",
"min(min(?z:Nat,?z:Nat),min(max(0,max(max(0,?x:Nat),?x:Nat)),min(max(min(?y:Nat,?y:Nat),min(?z:Nat,?x:Nat)),?x:Nat))) = min(min(min(0,max(?y:Nat,min(?z:Nat,?x:Nat))),min(?y:Nat,?x:Nat)),min(0,min(?x:Nat,?x:Nat)))",
"max(max(min(max(?z:Nat,min(?z:Nat,?x:Nat)),?y:Nat),?x:Nat),max(0,max(min(?x:Nat,min(?x:Nat,?y:Nat)),max(?y:Nat,min(?x:Nat,?x:Nat))))) = min(max(min(min(min(?y:Nat,?x:Nat),?y:Nat),min(?y:Nat,min(?x:Nat,?y:Nat))),?z:Nat),0)",
"max(min(max(min(?x:Nat,max(?y:Nat,?z:Nat)),?z:Nat),max(max(max(?x:Nat,?y:Nat),max(?x:Nat,0)),min(?y:Nat,max(?x:Nat,?x:Nat)))),min(max(min(?x:Nat,?y:Nat),?y:Nat),?x:Nat)) = min(min(?y:Nat,min(min(?x:Nat,?x:Nat),?x:Nat)),?z:Nat)",
"min(?y:Nat,min(min(max(max(?x:Nat,?y:Nat),min(?x:Nat,0)),0),min(0,min(min(0,?x:Nat),?y:Nat)))) = max(?x:Nat,min(max(?x:Nat,min(max(?z:Nat,?y:Nat),max(?x:Nat,?x:Nat))),min(min(min(?z:Nat,?z:Nat),0),min(0,0))))",
"min(max(min(min(max(?x:Nat,?x:Nat),max(?z:Nat,?x:Nat)),?y:Nat),max(?y:Nat,max(?x:Nat,min(?y:Nat,?x:Nat)))),max(?y:Nat,?x:Nat)) = min(min(min(min(?y:Nat,min(?z:Nat,?z:Nat)),min(0,max(0,?z:Nat))),max(?y:Nat,?y:Nat)),max(?x:Nat,?y:Nat))",
"max(max(max(?z:Nat,max(?x:Nat,?y:Nat)),?x:Nat),min(max(?y:Nat,min(?x:Nat,?z:Nat)),min(max(max(?y:Nat,?y:Nat),min(?y:Nat,?z:Nat)),?z:Nat))) = max(max(max(?x:Nat,?y:Nat),?y:Nat),min(?x:Nat,max(min(max(?y:Nat,0),min(?x:Nat,?z:Nat)),?x:Nat)))",
"max(min(max(?z:Nat,?z:Nat),min(min(min(?x:Nat,0),?y:Nat),?y:Nat)),min(min(?y:Nat,min(?z:Nat,min(?y:Nat,?y:Nat))),max(max(max(?x:Nat,?z:Nat),min(0,?x:Nat)),min(min(?y:Nat,?y:Nat),?x:Nat)))) = max(min(?y:Nat,?z:Nat),max(?z:Nat,?x:Nat))",
"max(?y:Nat,max(min(min(0,0),?y:Nat),?y:Nat)) = max(min(max(max(max(?x:Nat,?z:Nat),?z:Nat),min(max(?y:Nat,?z:Nat),max(?x:Nat,?x:Nat))),max(?z:Nat,max(?x:Nat,max(?y:Nat,?y:Nat)))),min(?y:Nat,max(0,max(max(?y:Nat,?y:Nat),max(?z:Nat,?x:Nat)))))",
"max(min(?y:Nat,min(min(max(?x:Nat,?y:Nat),?y:Nat),?x:Nat)),min(min(?z:Nat,?y:Nat),?x:Nat)) = max(min(max(max(?y:Nat,?y:Nat),?z:Nat),max(?x:Nat,?x:Nat)),min(min(max(min(?y:Nat,0),min(0,?x:Nat)),max(max(?z:Nat,?x:Nat),min(?x:Nat,?x:Nat))),?x:Nat))",
"max(max(min(max(?y:Nat,?y:Nat),?x:Nat),min(min(max(?x:Nat,0),0),?x:Nat)),max(0,max(0,max(max(?x:Nat,?z:Nat),min(?x:Nat,?y:Nat))))) = max(?x:Nat,max(max(?x:Nat,max(max(?y:Nat,0),?z:Nat)),max(min(0,max(0,0)),?z:Nat)))",
"min(?x:Nat,max(min(?y:Nat,?x:Nat),min(min(max(?x:Nat,0),max(?x:Nat,0)),?y:Nat))) = max(max(max(?y:Nat,max(max(0,?y:Nat),0)),min(min(?y:Nat,?x:Nat),min(?x:Nat,?x:Nat))),max(max(max(max(?z:Nat,?x:Nat),min(0,?y:Nat)),min(0,?y:Nat)),?y:Nat))",
"max(?y:Nat,min(min(?y:Nat,min(?z:Nat,?z:Nat)),min(min(min(?y:Nat,?y:Nat),?y:Nat),min(0,?y:Nat)))) = max(min(max(min(min(0,?x:Nat),?y:Nat),min(?x:Nat,?x:Nat)),?y:Nat),min(max(?y:Nat,min(min(?z:Nat,?y:Nat),?y:Nat)),max(min(max(?x:Nat,?x:Nat),?z:Nat),?x:Nat)))",
"max(?x:Nat,max(max(max(max(?y:Nat,?y:Nat),min(0,?y:Nat)),max(max(?x:Nat,0),?y:Nat)),max(?y:Nat,?y:Nat))) = max(?z:Nat,max(min(max(min(?y:Nat,?z:Nat),max(?x:Nat,?x:Nat)),min(?x:Nat,?z:Nat)),min(max(?y:Nat,?x:Nat),max(max(?x:Nat,?z:Nat),max(?y:Nat,?z:Nat)))))",
"max(min(min(max(max(?x:Nat,?x:Nat),max(?z:Nat,?y:Nat)),max(0,max(?y:Nat,?x:Nat))),max(?x:Nat,max(?y:Nat,min(?x:Nat,?z:Nat)))),?y:Nat) = max(max(max(?z:Nat,?x:Nat),min(?z:Nat,min(0,max(?y:Nat,?y:Nat)))),max(?y:Nat,min(0,min(max(?z:Nat,?y:Nat),?y:Nat))))",
"min(max(0,max(min(?x:Nat,max(0,?y:Nat)),?y:Nat)),max(?y:Nat,?y:Nat)) = min(min(min(max(?x:Nat,?x:Nat),min(max(?y:Nat,?y:Nat),min(?y:Nat,?x:Nat))),max(?y:Nat,max(?x:Nat,0))),max(max(max(max(?y:Nat,?x:Nat),max(?x:Nat,?x:Nat)),max(0,0)),max(?y:Nat,?z:Nat)))",
"min(max(?y:Nat,max(max(max(?y:Nat,0),?z:Nat),min(?y:Nat,?x:Nat))),max(?x:Nat,?y:Nat)) = max(min(max(?x:Nat,max(0,max(?y:Nat,0))),min(max(max(0,?y:Nat),min(?y:Nat,?x:Nat)),max(max(?x:Nat,?y:Nat),?x:Nat))),max(0,min(max(?x:Nat,max(?y:Nat,?y:Nat)),?y:Nat)))",
"max(?x:Nat,max(min(min(max(?x:Nat,?y:Nat),?x:Nat),max(0,0)),max(?y:Nat,max(max(?x:Nat,?y:Nat),?y:Nat)))) = min(min(?x:Nat,max(max(min(?x:Nat,0),max(?y:Nat,?y:Nat)),?y:Nat)),max(?x:Nat,max(min(max(?y:Nat,?y:Nat),min(?y:Nat,?x:Nat)),max(min(?y:Nat,?x:Nat),?z:Nat))))",
"min(min(max(max(min(?y:Nat,0),max(?y:Nat,0)),max(max(0,?x:Nat),min(?x:Nat,0))),?y:Nat),0) = max(min(max(min(?x:Nat,max(?y:Nat,?y:Nat)),min(?x:Nat,?x:Nat)),max(?x:Nat,?y:Nat)),max(?z:Nat,min(min(max(0,?x:Nat),?y:Nat),max(0,max(?x:Nat,?x:Nat)))))",
"min(min(?y:Nat,min(min(?x:Nat,min(?y:Nat,?x:Nat)),0)),min(?x:Nat,max(min(min(?y:Nat,0),max(0,?z:Nat)),?x:Nat))) = max(min(max(max(min(?z:Nat,?y:Nat),?z:Nat),max(min(?z:Nat,?z:Nat),max(0,?z:Nat))),?y:Nat),max(0,max(min(0,?z:Nat),min(0,?x:Nat))))",
"max(max(0,min(max(?y:Nat,max(?y:Nat,?z:Nat)),min(?x:Nat,?x:Nat))),max(?x:Nat,?z:Nat)) = max(min(max(max(max(?y:Nat,?z:Nat),?z:Nat),?x:Nat),min(max(?x:Nat,?y:Nat),?y:Nat)),max(max(min(min(?x:Nat,?x:Nat),?x:Nat),min(?x:Nat,max(?x:Nat,?y:Nat))),max(max(min(?z:Nat,?z:Nat),?y:Nat),?x:Nat)))",
"min(min(max(min(?x:Nat,max(?y:Nat,0)),min(max(?x:Nat,?y:Nat),0)),?y:Nat),max(max(?z:Nat,0),?z:Nat)) = max(min(max(max(max(?z:Nat,?y:Nat),?y:Nat),max(?z:Nat,?y:Nat)),max(min(?z:Nat,?x:Nat),min(min(?z:Nat,?x:Nat),?z:Nat))),min(max(?y:Nat,max(max(0,?y:Nat),?x:Nat)),max(0,?y:Nat)))",
"min(max(max(?x:Nat,?z:Nat),0),min(min(?y:Nat,?x:Nat),min(0,max(min(?x:Nat,?y:Nat),max(?y:Nat,?y:Nat))))) = max(max(min(?y:Nat,min(max(?x:Nat,?x:Nat),?x:Nat)),min(max(?x:Nat,?y:Nat),min(max(0,?y:Nat),?y:Nat))),min(min(max(?y:Nat,?z:Nat),min(min(?y:Nat,?x:Nat),min(?x:Nat,0))),max(max(min(?x:Nat,?z:Nat),?y:Nat),?x:Nat)))",
"min(min(max(?x:Nat,max(?x:Nat,max(?z:Nat,0))),max(?x:Nat,?z:Nat)),min(min(min(?y:Nat,min(?x:Nat,0)),?x:Nat),min(max(?x:Nat,min(?z:Nat,?x:Nat)),max(?y:Nat,0)))) = max(max(max(min(max(0,?z:Nat),min(?y:Nat,?y:Nat)),min(min(?x:Nat,?y:Nat),0)),?z:Nat),max(max(?y:Nat,max(max(?x:Nat,?y:Nat),min(?y:Nat,?x:Nat))),0))"]

val problemsMaxYes =
IOMstrs.rdEqs [
"s(?x:Nat) = max(?x:Nat,s(?x:Nat))",
"s(?x:Nat) = max(s(?x:Nat),?x:Nat)",
"s(?x:Nat) = max(s(?x:Nat),0)",
"s(?y:Nat) = max(0,s(?y:Nat))",
"s(0) = max(0,s(0))",
"max(?y:Nat,?x:Nat) = max(?x:Nat,?y:Nat)",
"max(?y:Nat,?y:Nat) = max(?y:Nat,0)",
"max(?y:Nat,?y:Nat) = max(0,?y:Nat)",
"max(?y:Nat,0) = max(0,?y:Nat)",
"s(?x:Nat) = max(s(?x:Nat),s(0))",
"s(?y:Nat) = max(s(?y:Nat),s(?y:Nat))",
"s(?x:Nat) = max(?x:Nat,s(max(?x:Nat,?x:Nat)))",
"s(?x:Nat) = max(0,max(?x:Nat,s(?x:Nat)))",
"s(?x:Nat) = max(max(?x:Nat,s(?x:Nat)),?x:Nat)",
"max(?x:Nat,?y:Nat) = max(?x:Nat,max(?y:Nat,?y:Nat))",
"max(?x:Nat,?y:Nat) = max(max(?x:Nat,?x:Nat),?y:Nat)",
"max(?x:Nat,?y:Nat) = max(max(?x:Nat,?y:Nat),0)",
"max(?x:Nat,?z:Nat) = max(max(0,?x:Nat),?z:Nat)",
"max(?y:Nat,?x:Nat) = max(?x:Nat,max(?x:Nat,?y:Nat))",
"max(?y:Nat,?x:Nat) = max(?y:Nat,max(0,?x:Nat))",
"max(?y:Nat,?x:Nat) = max(max(?x:Nat,?y:Nat),?x:Nat)",
"max(?y:Nat,?y:Nat) = max(max(0,?y:Nat),?y:Nat)",
"max(?y:Nat,0) = max(?y:Nat,max(0,?y:Nat))",
"max(?z:Nat,0) = max(0,max(?z:Nat,0))",
"max(0,?y:Nat) = max(max(0,?y:Nat),?y:Nat)",
"s(max(?y:Nat,0)) = max(?y:Nat,s(?y:Nat))",
"max(?x:Nat,s(?x:Nat)) = s(max(?x:Nat,?x:Nat))",
"max(s(?x:Nat),?x:Nat) = s(max(?x:Nat,0))",
"max(s(?x:Nat),0) = s(max(?x:Nat,?x:Nat))",
"max(s(?x:Nat),0) = max(0,s(?x:Nat))",
"max(s(?y:Nat),0) = max(?y:Nat,s(?y:Nat))",
"max(s(0),?y:Nat) = max(?y:Nat,s(0))",
"max(?x:Nat,max(?x:Nat,0)) = max(?x:Nat,0)",
"max(?x:Nat,max(?z:Nat,0)) = max(?x:Nat,?z:Nat)",
"max(?z:Nat,max(?z:Nat,?z:Nat)) = max(0,?z:Nat)",
"max(0,s(s(0))) = s(s(0))",
"max(0,max(?x:Nat,?x:Nat)) = max(?x:Nat,?x:Nat)",
"max(0,max(?x:Nat,?z:Nat)) = max(?z:Nat,?x:Nat)",
"max(max(?x:Nat,?x:Nat),?z:Nat) = max(?z:Nat,?x:Nat)",
"max(max(?x:Nat,?y:Nat),?y:Nat) = max(?x:Nat,?y:Nat)",
"max(max(?x:Nat,0),?y:Nat) = max(?x:Nat,?y:Nat)",
"max(max(?y:Nat,?y:Nat),?y:Nat) = max(?y:Nat,?y:Nat)",
"max(max(0,?x:Nat),0) = max(?x:Nat,0)",
"max(?x:Nat,max(?x:Nat,s(?x:Nat))) = s(?x:Nat)",
"max(?y:Nat,max(0,s(?y:Nat))) = s(?y:Nat)",
"max(s(?x:Nat),max(?x:Nat,?x:Nat)) = s(?x:Nat)",
"max(max(s(?y:Nat),0),0) = s(?y:Nat)",
"s(s(?y:Nat)) = max(s(s(?y:Nat)),s(?y:Nat))",
"max(s(0),max(s(?x:Nat),0)) = s(?x:Nat)",
"max(max(?x:Nat,s(?x:Nat)),s(?x:Nat)) = s(?x:Nat)",
"max(max(s(?x:Nat),s(?x:Nat)),0) = s(?x:Nat)",
"s(?x:Nat) = max(max(?x:Nat,max(s(?x:Nat),?x:Nat)),?x:Nat)",
"s(s(?y:Nat)) = max(?y:Nat,s(s(max(0,?y:Nat))))",
"max(?x:Nat,?x:Nat) = max(0,max(?x:Nat,max(?x:Nat,?x:Nat)))",
"max(?x:Nat,?y:Nat) = max(max(max(?x:Nat,?x:Nat),?y:Nat),?x:Nat)",
"max(?z:Nat,?y:Nat) = max(max(?y:Nat,?y:Nat),max(?z:Nat,?y:Nat))",
"max(0,?x:Nat) = max(max(?x:Nat,?x:Nat),max(?x:Nat,?x:Nat))",
"s(max(?x:Nat,?y:Nat)) = max(s(max(?y:Nat,?x:Nat)),?x:Nat)",
"max(s(?x:Nat),?x:Nat) = max(s(max(?x:Nat,?x:Nat)),?x:Nat)",
"max(s(0),?y:Nat) = max(max(?y:Nat,0),s(0))",
"max(?y:Nat,max(?x:Nat,?x:Nat)) = max(?x:Nat,max(?y:Nat,?x:Nat))",
"max(?y:Nat,max(?x:Nat,?y:Nat)) = max(?x:Nat,max(0,?y:Nat))",
"max(0,max(?z:Nat,0)) = max(max(?z:Nat,0),?z:Nat)",
"max(max(?y:Nat,?z:Nat),?y:Nat) = max(max(?y:Nat,?z:Nat),0)",
"max(?x:Nat,s(max(?x:Nat,?x:Nat))) = s(max(0,?x:Nat))",
"max(?x:Nat,max(?x:Nat,s(?x:Nat))) = s(max(?x:Nat,?x:Nat))",
"max(?z:Nat,max(?x:Nat,s(?z:Nat))) = max(s(?z:Nat),?x:Nat)",
"max(0,max(?z:Nat,s(?y:Nat))) = max(?z:Nat,s(?y:Nat))",
"max(s(max(?y:Nat,?x:Nat)),?y:Nat) = s(max(?x:Nat,?y:Nat))",
"max(max(?z:Nat,s(?z:Nat)),?x:Nat) = max(?x:Nat,s(?z:Nat))",
"max(max(s(?y:Nat),?x:Nat),?y:Nat) = max(s(?y:Nat),?x:Nat)",
"max(?x:Nat,max(max(?x:Nat,0),0)) = max(0,?x:Nat)",
"max(?y:Nat,max(max(?x:Nat,?y:Nat),?y:Nat)) = max(?x:Nat,?y:Nat)",
"max(?z:Nat,max(max(0,?z:Nat),?y:Nat)) = max(?y:Nat,?z:Nat)",
"max(max(?x:Nat,?x:Nat),max(?y:Nat,0)) = max(?y:Nat,?x:Nat)",
"max(max(?y:Nat,?x:Nat),max(?y:Nat,?x:Nat)) = max(?x:Nat,?y:Nat)",
"max(max(max(?y:Nat,?y:Nat),?y:Nat),?x:Nat) = max(?x:Nat,?y:Nat)",
"max(s(max(max(0,?x:Nat),0)),?x:Nat) = s(?x:Nat)",
"s(?x:Nat) = max(max(max(s(?x:Nat),s(0)),0),?x:Nat)",
"s(?y:Nat) = max(s(?y:Nat),max(0,s(max(0,?y:Nat))))",
"s(max(?x:Nat,?y:Nat)) = max(s(?y:Nat),s(max(0,?x:Nat)))",
"s(max(?y:Nat,?x:Nat)) = max(max(s(?x:Nat),?x:Nat),s(?y:Nat))",
"max(s(?y:Nat),max(max(0,?y:Nat),s(0))) = s(?y:Nat)",
"max(max(s(0),s(max(0,0))),0) = s(0)",
"s(max(?y:Nat,?x:Nat)) = max(s(max(0,max(?x:Nat,?y:Nat))),?y:Nat)",
"max(max(0,?y:Nat),?x:Nat) = max(?x:Nat,max(?y:Nat,max(?x:Nat,0)))",
"s(max(0,max(0,?y:Nat))) = max(max(s(?y:Nat),0),?y:Nat)",
"max(?y:Nat,max(?x:Nat,max(?y:Nat,?y:Nat))) = max(?x:Nat,max(0,?y:Nat))",
"max(max(?y:Nat,?x:Nat),max(max(?y:Nat,?y:Nat),?y:Nat)) = max(?y:Nat,?x:Nat)",
"max(max(?y:Nat,max(?x:Nat,?y:Nat)),max(?x:Nat,?x:Nat)) = max(?x:Nat,?y:Nat)",
"max(max(max(?x:Nat,0),max(0,0)),?y:Nat) = max(?x:Nat,?y:Nat)",
"max(0,max(s(max(max(?x:Nat,0),?x:Nat)),?x:Nat)) = s(?x:Nat)",
"max(max(0,s(max(?y:Nat,max(0,0)))),?y:Nat) = s(?y:Nat)",
"max(s(s(?y:Nat)),s(max(0,?y:Nat))) = s(s(max(?y:Nat,?y:Nat)))",
"max(?x:Nat,?y:Nat) = max(?x:Nat,max(max(?y:Nat,max(max(0,0),?x:Nat)),?y:Nat))",
"max(s(?y:Nat),?x:Nat) = max(?x:Nat,max(max(?x:Nat,max(?x:Nat,?x:Nat)),s(?y:Nat)))",
"max(?z:Nat,max(?x:Nat,?y:Nat)) = max(max(max(max(?y:Nat,?z:Nat),?x:Nat),?z:Nat),?z:Nat)",
"max(?z:Nat,max(?y:Nat,?x:Nat)) = max(?x:Nat,max(max(0,max(?z:Nat,?x:Nat)),?y:Nat))",
"max(max(?z:Nat,?x:Nat),?x:Nat) = max(max(?x:Nat,max(max(?z:Nat,?x:Nat),?x:Nat)),?x:Nat)",
"max(?y:Nat,max(?x:Nat,max(?z:Nat,?x:Nat))) = max(max(max(?z:Nat,?y:Nat),?x:Nat),?y:Nat)",
"max(max(s(?y:Nat),s(?z:Nat)),?y:Nat) = max(s(?y:Nat),max(?y:Nat,s(?z:Nat)))",
"max(max(?z:Nat,0),max(?z:Nat,s(?x:Nat))) = max(?z:Nat,max(?x:Nat,s(?x:Nat)))",
"max(?x:Nat,max(max(max(s(?x:Nat),0),0),0)) = s(max(?x:Nat,?x:Nat))",
"max(?x:Nat,max(s(max(?y:Nat,0)),?x:Nat)) = max(?x:Nat,max(max(?y:Nat,?x:Nat),s(?y:Nat)))",
"max(?x:Nat,max(max(s(?x:Nat),?z:Nat),0)) = max(?z:Nat,max(max(?z:Nat,?x:Nat),s(?x:Nat)))",
"max(s(max(max(max(?x:Nat,?y:Nat),?x:Nat),?y:Nat)),?x:Nat) = s(max(0,max(?x:Nat,?y:Nat)))",
"max(max(max(0,max(?z:Nat,max(?y:Nat,?z:Nat))),0),?x:Nat) = max(max(?z:Nat,?y:Nat),?x:Nat)",
"max(?z:Nat,s(s(?x:Nat))) = max(s(0),max(s(max(max(?x:Nat,0),s(?x:Nat))),?z:Nat))",
"max(?x:Nat,max(max(max(?z:Nat,?x:Nat),0),max(s(?x:Nat),s(?x:Nat)))) = max(?z:Nat,s(?x:Nat))",
"max(?y:Nat,max(?x:Nat,?x:Nat)) = max(?x:Nat,max(max(max(?y:Nat,max(0,?y:Nat)),max(?y:Nat,?y:Nat)),?x:Nat))",
"max(s(max(?x:Nat,?y:Nat)),max(s(?x:Nat),?x:Nat)) = max(s(max(?y:Nat,?x:Nat)),max(s(?y:Nat),?y:Nat))",
"max(max(?y:Nat,?x:Nat),max(?y:Nat,max(max(?y:Nat,?y:Nat),0))) = max(max(max(?y:Nat,?y:Nat),?x:Nat),?y:Nat)",
"max(s(max(0,max(max(0,0),max(?y:Nat,?y:Nat)))),?x:Nat) = max(s(?y:Nat),max(?x:Nat,s(?y:Nat)))",
"max(max(s(max(?y:Nat,?x:Nat)),?z:Nat),?z:Nat) = max(s(?x:Nat),max(max(?z:Nat,s(?y:Nat)),s(max(0,?y:Nat))))",
"max(?x:Nat,max(max(?y:Nat,max(max(?y:Nat,?y:Nat),?x:Nat)),?z:Nat)) = max(max(0,max(0,max(max(?x:Nat,?z:Nat),?x:Nat))),?y:Nat)",
"s(max(max(max(s(?x:Nat),max(?y:Nat,?x:Nat)),0),max(?x:Nat,max(max(?y:Nat,?z:Nat),max(?y:Nat,?x:Nat))))) = max(s(max(max(max(?y:Nat,?x:Nat),s(?x:Nat)),?z:Nat)),max(max(0,?y:Nat),max(?x:Nat,?y:Nat)))"]


val problemsMinYes =
IOMstrs.rdEqs [
"min(?x:Nat,0) = min(0,?y:Nat)",
"min(?y:Nat,?x:Nat) = min(?x:Nat,?y:Nat)",
"min(?y:Nat,0) = min(0,?y:Nat)",
"min(0,?x:Nat) = min(0,?z:Nat)",
"min(0,?x:Nat) = min(0,0)",
"min(0,0) = min(?z:Nat,0)",
"s(0) = min(s(?y:Nat),s(0))",
"min(?x:Nat,?x:Nat) = min(?x:Nat,s(?x:Nat))",
"min(0,s(?y:Nat)) = min(0,?x:Nat)",
"min(s(?x:Nat),0) = min(0,?y:Nat)",
"min(s(0),s(?y:Nat)) = s(0)",
"min(?x:Nat,?x:Nat) = min(s(s(?x:Nat)),?x:Nat)",
"min(?x:Nat,?y:Nat) = min(min(?x:Nat,?x:Nat),?y:Nat)",
"min(?y:Nat,0) = min(0,s(s(?x:Nat)))",
"min(?z:Nat,0) = min(0,min(?y:Nat,?y:Nat))",
"min(?z:Nat,0) = min(0,min(?z:Nat,0))",
"min(0,?y:Nat) = min(min(?x:Nat,0),?x:Nat)",
"min(0,?z:Nat) = min(?y:Nat,min(0,?z:Nat))",
"min(0,?z:Nat) = min(min(?x:Nat,0),?y:Nat)",
"min(0,s(?y:Nat)) = min(0,s(?z:Nat))",
"min(0,s(?z:Nat)) = min(0,s(0))",
"min(s(0),?y:Nat) = min(?y:Nat,s(0))",
"min(min(?x:Nat,?x:Nat),?z:Nat) = min(?z:Nat,?x:Nat)",
"min(min(?x:Nat,?y:Nat),?y:Nat) = min(?x:Nat,?y:Nat)",
"min(min(0,?x:Nat),0) = min(?x:Nat,0)",
"min(min(0,?y:Nat),?z:Nat) = min(?z:Nat,0)",
"min(?x:Nat,0) = min(s(min(?y:Nat,?z:Nat)),0)",
"min(?y:Nat,?x:Nat) = min(min(?y:Nat,?x:Nat),s(?x:Nat))",
"min(?z:Nat,0) = min(min(0,s(?y:Nat)),?x:Nat)",
"min(0,s(?x:Nat)) = min(?x:Nat,min(0,?x:Nat))",
"min(s(s(0)),0) = min(0,s(?y:Nat))",
"min(min(?x:Nat,?y:Nat),0) = min(0,s(?x:Nat))",
"min(?y:Nat,min(s(0),0)) = min(?y:Nat,0)",
"min(?z:Nat,min(?y:Nat,s(?z:Nat))) = min(?z:Nat,?y:Nat)",
"min(s(?x:Nat),min(0,?y:Nat)) = min(0,?y:Nat)",
"min(min(0,s(?y:Nat)),?y:Nat) = min(?x:Nat,0)",
"min(?x:Nat,?y:Nat) = min(min(min(?x:Nat,?x:Nat),?y:Nat),?x:Nat)",
"min(0,?x:Nat) = min(min(s(s(?x:Nat)),0),0)",
"min(?y:Nat,min(0,?z:Nat)) = min(0,s(s(?x:Nat)))",
"min(min(0,?y:Nat),?x:Nat) = min(min(0,0),?x:Nat)",
"min(?y:Nat,s(min(?y:Nat,0))) = min(?y:Nat,s(0))",
"min(?x:Nat,min(0,min(?z:Nat,?y:Nat))) = min(?y:Nat,0)",
"min(?x:Nat,min(min(?x:Nat,0),0)) = min(0,?x:Nat)",
"min(?y:Nat,min(min(?x:Nat,?y:Nat),?y:Nat)) = min(?x:Nat,?y:Nat)",
"min(0,s(min(?x:Nat,s(?y:Nat)))) = min(0,?x:Nat)",
"min(0,min(?x:Nat,min(?y:Nat,0))) = min(0,?x:Nat)",
"min(min(min(?y:Nat,0),?x:Nat),?y:Nat) = min(0,?x:Nat)",
"min(min(min(?y:Nat,0),?y:Nat),?x:Nat) = min(0,?z:Nat)",
"min(?x:Nat,?x:Nat) = min(min(?x:Nat,s(?x:Nat)),min(?x:Nat,?x:Nat))",
"min(?x:Nat,?y:Nat) = min(?x:Nat,min(min(?x:Nat,s(?y:Nat)),?y:Nat))",
"min(?x:Nat,0) = min(?y:Nat,min(min(s(?y:Nat),0),?x:Nat))",
"min(?y:Nat,0) = min(s(min(?x:Nat,?x:Nat)),min(?y:Nat,0))",
"min(?y:Nat,min(0,?z:Nat)) = min(s(min(?y:Nat,0)),0)",
"min(0,s(s(0))) = min(0,min(s(?z:Nat),?y:Nat))",
"min(0,min(?y:Nat,s(s(0)))) = min(0,s(0))",
"min(?z:Nat,min(s(?x:Nat),min(?y:Nat,0))) = min(0,?z:Nat)",
"min(?y:Nat,?x:Nat) = min(min(?y:Nat,min(s(?x:Nat),s(?x:Nat))),?x:Nat)",
"min(?x:Nat,min(0,?z:Nat)) = min(min(0,?y:Nat),min(?x:Nat,0))",
"min(min(0,?y:Nat),?x:Nat) = min(?x:Nat,min(?y:Nat,min(?x:Nat,0)))",
"min(min(0,?y:Nat),?y:Nat) = min(s(s(?x:Nat)),min(?z:Nat,0))",
"min(min(s(?x:Nat),?z:Nat),0) = min(0,min(s(0),?x:Nat))",
"min(s(s(?x:Nat)),min(0,0)) = min(min(0,?x:Nat),0)",
"min(min(?y:Nat,min(?x:Nat,?y:Nat)),min(?x:Nat,?x:Nat)) = min(?x:Nat,?y:Nat)",
"min(min(0,min(min(?z:Nat,?x:Nat),?x:Nat)),?x:Nat) = min(0,?y:Nat)",
"min(min(s(min(s(?y:Nat),?x:Nat)),?y:Nat),?x:Nat) = min(?y:Nat,?x:Nat)",
"min(min(s(min(s(?y:Nat),?y:Nat)),?y:Nat),?z:Nat) = min(?z:Nat,?y:Nat)",
"min(s(?y:Nat),s(min(s(min(?x:Nat,?z:Nat)),0))) = s(0)",
"min(?x:Nat,0) = min(?y:Nat,min(min(?y:Nat,0),min(0,s(?z:Nat))))",
"min(?z:Nat,?y:Nat) = min(min(min(?z:Nat,min(?z:Nat,?z:Nat)),s(?z:Nat)),?y:Nat)",
"min(0,?y:Nat) = min(min(?x:Nat,0),s(s(min(s(?z:Nat),?y:Nat))))",
"min(0,s(?z:Nat)) = min(0,min(min(?x:Nat,s(?x:Nat)),s(?y:Nat)))",
"min(0,min(?y:Nat,?x:Nat)) = min(?y:Nat,min(0,s(s(s(?x:Nat)))))",
"min(0,min(0,s(?y:Nat))) = min(min(min(?y:Nat,?y:Nat),0),?y:Nat)",
"min(min(s(?x:Nat),?x:Nat),?y:Nat) = min(?x:Nat,min(?y:Nat,s(s(?x:Nat))))",
"min(min(0,?x:Nat),min(?x:Nat,?z:Nat)) = min(s(0),min(0,?z:Nat))",
"min(?x:Nat,min(?z:Nat,min(0,min(?x:Nat,s(?z:Nat))))) = min(0,?y:Nat)",
"min(s(s(s(?y:Nat))),min(min(?x:Nat,?y:Nat),?y:Nat)) = min(?x:Nat,?y:Nat)",
"min(?x:Nat,0) = min(min(min(s(0),?x:Nat),s(min(?y:Nat,?z:Nat))),0)",
"min(?y:Nat,0) = min(s(?y:Nat),min(min(?x:Nat,s(min(?x:Nat,?x:Nat))),0))",
"min(?z:Nat,0) = min(?y:Nat,min(?y:Nat,min(min(min(0,?x:Nat),?y:Nat),0)))",
"min(0,?x:Nat) = min(min(0,min(min(min(?x:Nat,?y:Nat),?x:Nat),?y:Nat)),?y:Nat)",
"min(?z:Nat,min(?x:Nat,?y:Nat)) = min(min(min(min(?y:Nat,?z:Nat),?x:Nat),?z:Nat),?z:Nat)",
"min(min(?z:Nat,?x:Nat),?x:Nat) = min(min(?x:Nat,min(min(?z:Nat,?x:Nat),?x:Nat)),?x:Nat)",
"min(?y:Nat,min(?x:Nat,min(?z:Nat,?x:Nat))) = min(min(min(?z:Nat,?y:Nat),?x:Nat),?y:Nat)",
"min(min(s(?y:Nat),s(?z:Nat)),?y:Nat) = min(s(?y:Nat),min(?y:Nat,s(?z:Nat)))",
"min(0,s(s(s(min(?y:Nat,?x:Nat))))) = min(min(0,s(?x:Nat)),?z:Nat)",
"min(?y:Nat,min(0,?y:Nat)) = min(?z:Nat,min(0,min(min(?x:Nat,s(?y:Nat)),?x:Nat)))",
"min(s(?z:Nat),s(0)) = min(min(s(0),s(s(s(?z:Nat)))),s(?y:Nat))",
"min(s(min(?z:Nat,?z:Nat)),0) = min(s(s(0)),min(min(?x:Nat,?x:Nat),0))",
"min(min(s(?z:Nat),?x:Nat),min(?x:Nat,0)) = min(?x:Nat,min(min(0,?y:Nat),0))",
"min(min(min(0,?y:Nat),s(?z:Nat)),?z:Nat) = min(0,s(s(min(?z:Nat,?x:Nat))))",
"min(min(?z:Nat,min(min(0,min(?y:Nat,0)),?z:Nat)),s(?y:Nat)) = min(0,?y:Nat)",
"min(min(min(?y:Nat,min(min(?x:Nat,?y:Nat),s(?y:Nat))),?y:Nat),?y:Nat) = min(?y:Nat,?x:Nat)",
"min(0,0) = min(min(s(min(?z:Nat,s(?z:Nat))),min(?y:Nat,?x:Nat)),min(0,?x:Nat))",
"min(min(?x:Nat,?x:Nat),min(min(min(?x:Nat,?y:Nat),?z:Nat),0)) = min(min(?x:Nat,0),?z:Nat)",
"min(0,min(s(min(?z:Nat,?x:Nat)),?y:Nat)) = min(?y:Nat,min(?x:Nat,min(0,s(s(?y:Nat)))))",
"min(0,s(min(min(?z:Nat,s(?z:Nat)),?y:Nat))) = min(min(?x:Nat,min(?y:Nat,s(?z:Nat))),0)",
"min(min(?y:Nat,?z:Nat),min(min(?x:Nat,s(?z:Nat)),s(s(?y:Nat)))) = min(?x:Nat,min(?z:Nat,?y:Nat))",
"min(min(min(min(s(?x:Nat),s(0)),?z:Nat),0),s(?x:Nat)) = min(0,min(?y:Nat,?x:Nat))",
"min(min(0,?y:Nat),min(min(?x:Nat,0),min(s(?y:Nat),s(s(?y:Nat))))) = min(?x:Nat,0)",
"min(0,min(?x:Nat,?x:Nat)) = min(0,min(min(?y:Nat,min(min(?x:Nat,?x:Nat),min(?x:Nat,0))),0))",
"min(min(?x:Nat,0),s(?y:Nat)) = min(min(?x:Nat,?z:Nat),min(0,s(min(min(?y:Nat,?y:Nat),s(?x:Nat)))))",
"min(min(min(?x:Nat,s(?y:Nat)),0),min(?x:Nat,min(min(s(0),s(?z:Nat)),?y:Nat))) = min(0,?x:Nat)",
"min(0,?y:Nat) = min(min(0,?y:Nat),min(min(min(s(?y:Nat),?x:Nat),s(?x:Nat)),s(min(s(?y:Nat),0))))",
"min(?x:Nat,min(s(min(s(?y:Nat),0)),0)) = min(min(0,0),min(0,min(?x:Nat,s(s(0)))))",
"min(0,s(min(min(s(0),?y:Nat),min(min(?x:Nat,?y:Nat),s(?z:Nat))))) = min(min(s(?z:Nat),0),?y:Nat)",
"min(min(0,min(min(min(?x:Nat,?z:Nat),?y:Nat),s(s(?y:Nat)))),s(s(0))) = min(0,min(?y:Nat,?z:Nat))",
"min(min(?z:Nat,min(?z:Nat,min(0,?y:Nat))),?y:Nat) = min(min(min(min(s(0),s(0)),s(0)),0),?y:Nat)",
"min(min(?x:Nat,min(0,min(0,0))),min(?z:Nat,?y:Nat)) = min(?y:Nat,min(s(?x:Nat),min(min(0,?z:Nat),?y:Nat)))",
"min(s(min(min(min(?y:Nat,?y:Nat),min(?x:Nat,?x:Nat)),?x:Nat)),0) = min(min(min(0,s(0)),s(?x:Nat)),?y:Nat)",
"min(s(min(s(?y:Nat),min(min(?x:Nat,?y:Nat),?y:Nat))),min(?y:Nat,min(?x:Nat,min(s(?x:Nat),?x:Nat)))) = min(?x:Nat,?y:Nat)",
"min(min(?z:Nat,s(?z:Nat)),min(min(s(s(?z:Nat)),s(?x:Nat)),min(?y:Nat,min(min(0,?x:Nat),?z:Nat)))) = min(?y:Nat,0)",
"min(min(min(min(?z:Nat,0),?z:Nat),min(?x:Nat,min(s(?z:Nat),?y:Nat))),min(?y:Nat,?x:Nat)) = min(0,min(?x:Nat,min(0,?x:Nat)))",
"min(min(min(?y:Nat,0),s(s(s(0)))),0) = min(min(min(?x:Nat,?z:Nat),?x:Nat),min(min(s(min(?z:Nat,?y:Nat)),0),0))",
"min(0,min(min(min(min(?y:Nat,?y:Nat),s(0)),s(?x:Nat)),?x:Nat)) = min(0,min(min(0,min(?x:Nat,min(?y:Nat,?y:Nat))),?x:Nat))"]

val problemsMaxNo =
IOMstrs.rdEqs [
"s(?x:Nat) = max(?x:Nat,?x:Nat)",
"s(?x:Nat) = max(?x:Nat,?z:Nat)",
"s(?x:Nat) = max(?x:Nat,0)",
"s(?x:Nat) = max(?y:Nat,?x:Nat)",
"s(?x:Nat) = max(?y:Nat,?y:Nat)",
"s(?x:Nat) = max(?y:Nat,?z:Nat)",
"s(?x:Nat) = max(0,?x:Nat)",
"s(?x:Nat) = max(0,?y:Nat)",
"s(?y:Nat) = max(?x:Nat,0)",
"s(?y:Nat) = max(0,0)",
"s(0) = max(?x:Nat,?x:Nat)",
"s(0) = max(?x:Nat,?y:Nat)",
"s(0) = max(?x:Nat,0)",
"s(0) = max(0,?x:Nat)",
"s(?x:Nat) = max(?y:Nat,s(?x:Nat))",
"s(?x:Nat) = max(?z:Nat,s(?z:Nat))",
"s(?x:Nat) = max(?z:Nat,s(0))",
"s(?x:Nat) = max(0,s(?y:Nat))",
"s(?x:Nat) = max(s(?x:Nat),?y:Nat)",
"s(?x:Nat) = max(s(?y:Nat),?x:Nat)",
"s(?x:Nat) = max(s(?y:Nat),?y:Nat)",
"s(?x:Nat) = max(s(0),?x:Nat)",
"s(?y:Nat) = max(?y:Nat,s(?x:Nat))",
"s(?y:Nat) = max(?z:Nat,s(?x:Nat))",
"s(0) = max(?x:Nat,s(?x:Nat))",
"s(0) = max(0,s(?x:Nat))",
"s(0) = max(s(?x:Nat),?x:Nat)",
"s(0) = max(s(0),?z:Nat)",
"s(s(?x:Nat)) = max(?y:Nat,?y:Nat)",
"s(s(?x:Nat)) = max(?y:Nat,?z:Nat)",
"s(s(?y:Nat)) = max(?x:Nat,?y:Nat)",
"s(s(?y:Nat)) = max(?y:Nat,?x:Nat)",
"s(s(?y:Nat)) = max(?z:Nat,0)",
"s(s(?y:Nat)) = max(0,?y:Nat)",
"s(s(?y:Nat)) = max(0,0)",
"s(s(?z:Nat)) = max(?z:Nat,?z:Nat)",
"s(s(?z:Nat)) = max(0,?y:Nat)",
"s(s(0)) = max(?x:Nat,?z:Nat)",
"max(?x:Nat,?x:Nat) = max(?x:Nat,?y:Nat)",
"max(?x:Nat,?x:Nat) = max(?y:Nat,?y:Nat)",
"max(?x:Nat,?x:Nat) = max(?y:Nat,0)",
"max(?x:Nat,?x:Nat) = max(0,?y:Nat)",
"max(?x:Nat,?y:Nat) = max(?x:Nat,?z:Nat)",
"max(?x:Nat,?y:Nat) = max(?x:Nat,0)",
"max(?x:Nat,?y:Nat) = max(?z:Nat,?x:Nat)",
"max(?x:Nat,?y:Nat) = max(?z:Nat,?y:Nat)",
"max(?x:Nat,?y:Nat) = max(0,?y:Nat)",
"max(?x:Nat,?z:Nat) = max(0,?x:Nat)",
"max(?x:Nat,0) = max(?y:Nat,?x:Nat)",
"max(?x:Nat,0) = max(0,?y:Nat)",
"max(?y:Nat,?x:Nat) = max(?x:Nat,?x:Nat)",
"max(?y:Nat,?x:Nat) = max(0,0)",
"max(?y:Nat,?y:Nat) = s(s(0))",
"max(?y:Nat,?z:Nat) = max(?x:Nat,0)",
"max(?y:Nat,0) = s(s(0))",
"max(?z:Nat,?y:Nat) = max(0,?x:Nat)",
"max(?z:Nat,?z:Nat) = max(?y:Nat,?x:Nat)",
"max(0,?x:Nat) = s(s(0))",
"max(0,?x:Nat) = max(0,?z:Nat)",
"max(0,0) = max(?y:Nat,?y:Nat)",
"max(?x:Nat,s(?y:Nat)) = s(0)",
"max(?x:Nat,s(0)) = s(?x:Nat)",
"max(s(?x:Nat),?y:Nat) = s(?z:Nat)",
"max(s(?y:Nat),?x:Nat) = s(0)",
"max(s(?y:Nat),0) = s(?z:Nat)",
"max(s(0),?x:Nat) = s(?y:Nat)",
"s(?x:Nat) = max(?y:Nat,s(s(?x:Nat)))",
"s(?x:Nat) = max(?y:Nat,max(?x:Nat,?x:Nat))",
"s(?x:Nat) = max(?y:Nat,max(?y:Nat,?z:Nat))",
"s(?x:Nat) = max(?y:Nat,max(0,?y:Nat))",
"s(?x:Nat) = max(s(?y:Nat),s(?y:Nat))",
"s(?x:Nat) = max(max(?y:Nat,?x:Nat),?z:Nat)",
"s(?x:Nat) = max(max(?y:Nat,?z:Nat),?y:Nat)",
"s(?x:Nat) = max(max(0,?x:Nat),?y:Nat)",
"s(?x:Nat) = max(max(0,?z:Nat),?z:Nat)",
"s(?y:Nat) = max(?x:Nat,s(s(0)))",
"s(?y:Nat) = max(?x:Nat,max(?z:Nat,?y:Nat))",
"s(?y:Nat) = max(?y:Nat,s(s(?x:Nat)))",
"s(?y:Nat) = max(?y:Nat,max(?y:Nat,?y:Nat))",
"s(?y:Nat) = max(?y:Nat,max(?z:Nat,0))",
"s(?y:Nat) = max(?z:Nat,max(?x:Nat,?x:Nat))",
"s(?y:Nat) = max(0,s(s(?y:Nat)))",
"s(?y:Nat) = max(0,s(s(0)))",
"s(?y:Nat) = max(s(s(0)),?x:Nat)",
"s(?y:Nat) = max(max(?y:Nat,?x:Nat),?y:Nat)",
"s(?y:Nat) = max(max(?y:Nat,?x:Nat),0)",
"s(?y:Nat) = max(max(?y:Nat,?y:Nat),?y:Nat)",
"s(?y:Nat) = max(max(?y:Nat,?z:Nat),?x:Nat)",
"s(?y:Nat) = max(max(?y:Nat,0),?y:Nat)",
"s(?z:Nat) = max(?x:Nat,s(s(?y:Nat)))",
"s(?z:Nat) = max(?z:Nat,s(s(?z:Nat)))",
"s(?z:Nat) = max(s(s(?x:Nat)),?z:Nat)",
"s(0) = max(?x:Nat,s(s(0)))",
"s(0) = max(?y:Nat,s(s(?x:Nat)))",
"s(0) = max(?y:Nat,max(?x:Nat,?x:Nat))",
"s(0) = max(?y:Nat,max(?x:Nat,?y:Nat))",
"s(0) = max(s(?y:Nat),s(0))",
"s(0) = max(s(s(?y:Nat)),?x:Nat)",
"s(s(?x:Nat)) = max(?x:Nat,s(?y:Nat))",
"s(s(?x:Nat)) = max(?x:Nat,s(0))",
"s(s(?x:Nat)) = max(s(?x:Nat),?x:Nat)",
"s(s(?y:Nat)) = max(?x:Nat,s(?x:Nat))",
"s(s(?y:Nat)) = max(s(?x:Nat),0)",
"s(s(?y:Nat)) = max(s(?y:Nat),?z:Nat)",
"s(s(?y:Nat)) = max(s(?z:Nat),?x:Nat)",
"s(s(?z:Nat)) = max(?x:Nat,s(?y:Nat))",
"s(s(?z:Nat)) = max(s(?x:Nat),?x:Nat)",
"s(s(0)) = max(?y:Nat,s(?z:Nat))",
"s(s(0)) = max(s(?y:Nat),?x:Nat)",
"s(s(0)) = max(s(?z:Nat),?z:Nat)",
"max(?x:Nat,?x:Nat) = s(max(?y:Nat,?z:Nat))",
"max(?x:Nat,?x:Nat) = max(?x:Nat,s(0))",
"max(?x:Nat,?x:Nat) = max(?y:Nat,s(0))",
"max(?x:Nat,?x:Nat) = max(0,s(?y:Nat))",
"max(?x:Nat,?x:Nat) = max(s(?x:Nat),?z:Nat)",
"max(?x:Nat,?x:Nat) = max(s(0),0)",
"max(?x:Nat,?y:Nat) = max(?y:Nat,s(?x:Nat))",
"max(?x:Nat,?y:Nat) = max(?z:Nat,s(0))",
"max(?x:Nat,?y:Nat) = max(s(?y:Nat),0)",
"max(?x:Nat,?y:Nat) = max(s(?z:Nat),?y:Nat)",
"max(?x:Nat,?z:Nat) = max(?x:Nat,s(?x:Nat))",
"max(?x:Nat,0) = s(max(?y:Nat,?z:Nat))",
"max(?y:Nat,?x:Nat) = s(max(?x:Nat,?x:Nat))",
"max(?y:Nat,?x:Nat) = s(max(0,?y:Nat))",
"max(?y:Nat,?x:Nat) = max(s(?y:Nat),?x:Nat)",
"max(?y:Nat,?x:Nat) = max(s(0),?z:Nat)",
"max(?y:Nat,?y:Nat) = max(?x:Nat,s(?y:Nat))",
"max(?y:Nat,?y:Nat) = max(?y:Nat,s(?x:Nat))",
"max(?y:Nat,?y:Nat) = max(?z:Nat,s(?z:Nat))",
"max(?y:Nat,?z:Nat) = max(?y:Nat,s(?x:Nat))",
"max(?y:Nat,?z:Nat) = max(s(?y:Nat),?y:Nat)",
"max(?y:Nat,0) = s(s(s(?y:Nat)))",
"max(?y:Nat,0) = max(?x:Nat,s(?z:Nat))",
"max(?y:Nat,0) = max(?y:Nat,s(?x:Nat))",
"max(?z:Nat,?x:Nat) = s(max(?x:Nat,?y:Nat))",
"max(?z:Nat,?x:Nat) = s(max(?y:Nat,?x:Nat))",
"max(?z:Nat,?x:Nat) = max(?y:Nat,s(?z:Nat))",
"max(?z:Nat,?y:Nat) = s(max(?x:Nat,?x:Nat))",
"max(?z:Nat,?y:Nat) = max(?x:Nat,s(?x:Nat))",
"max(?z:Nat,?y:Nat) = max(s(?x:Nat),?x:Nat)",
"max(0,?x:Nat) = max(?x:Nat,s(?y:Nat))",
"max(0,?x:Nat) = max(s(?x:Nat),?x:Nat)",
"max(0,?y:Nat) = s(s(s(?z:Nat)))",
"s(s(s(?x:Nat))) = max(?y:Nat,?x:Nat)",
"s(s(s(?y:Nat))) = max(?y:Nat,?z:Nat)",
"s(s(s(?y:Nat))) = max(?z:Nat,?x:Nat)",
"s(s(s(0))) = max(?y:Nat,0)",
"s(max(?x:Nat,?x:Nat)) = max(?x:Nat,?x:Nat)",
"s(max(?x:Nat,?z:Nat)) = max(?x:Nat,?x:Nat)",
"s(max(?y:Nat,?x:Nat)) = max(?x:Nat,?x:Nat)",
"s(max(?y:Nat,?x:Nat)) = max(?y:Nat,?z:Nat)",
"s(max(?y:Nat,?x:Nat)) = max(0,?y:Nat)",
"s(max(?y:Nat,?z:Nat)) = max(?y:Nat,0)",
"s(max(?y:Nat,0)) = max(?x:Nat,0)",
"s(max(0,?x:Nat)) = max(?z:Nat,?y:Nat)",
"s(max(0,?y:Nat)) = max(?x:Nat,?y:Nat)",
"s(max(0,?y:Nat)) = max(?y:Nat,?y:Nat)",
"max(?y:Nat,s(?z:Nat)) = max(?x:Nat,?y:Nat)",
"max(0,s(?x:Nat)) = s(s(?y:Nat))",
"max(0,s(?y:Nat)) = max(?y:Nat,?y:Nat)",
"max(s(?x:Nat),?x:Nat) = max(?x:Nat,0)",
"max(s(?x:Nat),?x:Nat) = max(?z:Nat,0)",
"max(s(?x:Nat),?y:Nat) = max(?y:Nat,0)",
"max(s(?x:Nat),0) = max(0,?y:Nat)",
"max(s(?y:Nat),?x:Nat) = s(s(?x:Nat))",
"max(s(?y:Nat),?x:Nat) = max(?x:Nat,?x:Nat)",
"max(s(?y:Nat),?x:Nat) = max(?x:Nat,?y:Nat)",
"max(s(?y:Nat),?y:Nat) = max(?x:Nat,?x:Nat)",
"max(s(?y:Nat),?y:Nat) = max(?x:Nat,?y:Nat)",
"max(s(?y:Nat),0) = max(?y:Nat,?z:Nat)",
"max(s(?z:Nat),?z:Nat) = max(0,0)",
"max(s(0),?x:Nat) = s(s(?x:Nat))",
"max(s(0),?x:Nat) = max(?y:Nat,0)",
"max(s(0),0) = s(s(?x:Nat))",
"max(s(0),0) = max(?x:Nat,?z:Nat)",
"max(?x:Nat,max(?x:Nat,?y:Nat)) = s(?y:Nat)",
"max(?x:Nat,max(?y:Nat,?x:Nat)) = s(?z:Nat)",
"max(?x:Nat,max(?z:Nat,0)) = s(?y:Nat)",
"max(?y:Nat,s(s(?y:Nat))) = s(0)",
"max(?z:Nat,max(?x:Nat,?z:Nat)) = s(?x:Nat)",
"max(?z:Nat,max(?z:Nat,?y:Nat)) = s(0)",
"max(0,max(?z:Nat,?y:Nat)) = s(?x:Nat)",
"max(s(?x:Nat),s(?y:Nat)) = s(?x:Nat)",
"max(s(0),s(?y:Nat)) = s(0)",
"max(s(s(?x:Nat)),?x:Nat) = s(?y:Nat)",
"max(s(s(?x:Nat)),?y:Nat) = s(?z:Nat)",
"max(s(s(?z:Nat)),?x:Nat) = s(?z:Nat)",
"max(max(?x:Nat,?x:Nat),?x:Nat) = s(?y:Nat)",
"max(max(?x:Nat,?x:Nat),?z:Nat) = s(?x:Nat)",
"max(max(?x:Nat,?z:Nat),?y:Nat) = s(0)",
"max(max(?x:Nat,0),?x:Nat) = s(?y:Nat)",
"max(max(?y:Nat,?y:Nat),0) = s(?x:Nat)",
"max(max(?z:Nat,?y:Nat),?x:Nat) = s(?x:Nat)",
"max(max(?z:Nat,0),?x:Nat) = s(0)",
"max(max(0,?z:Nat),?y:Nat) = s(?x:Nat)",
"s(?x:Nat) = max(?y:Nat,s(max(?x:Nat,?x:Nat)))",
"s(?x:Nat) = max(?y:Nat,max(s(0),?y:Nat))",
"s(?x:Nat) = max(0,max(s(0),?z:Nat))",
"s(?x:Nat) = max(s(?y:Nat),max(?x:Nat,?x:Nat))",
"s(?x:Nat) = max(s(max(?x:Nat,?x:Nat)),?y:Nat)",
"s(?y:Nat) = max(?x:Nat,s(s(s(?y:Nat))))",
"s(?y:Nat) = max(?y:Nat,s(max(?x:Nat,?y:Nat)))",
"s(?y:Nat) = max(0,max(s(?y:Nat),?x:Nat))",
"s(?y:Nat) = max(s(0),max(?y:Nat,?y:Nat))",
"s(?y:Nat) = max(max(0,?x:Nat),s(?y:Nat))",
"s(?z:Nat) = max(s(?z:Nat),max(?y:Nat,?x:Nat))",
"s(0) = max(?y:Nat,max(s(?y:Nat),?y:Nat))",
"s(0) = max(0,s(max(?z:Nat,?x:Nat)))",
"s(0) = max(s(?z:Nat),max(?y:Nat,0))",
"s(0) = max(s(max(?y:Nat,?x:Nat)),?y:Nat)",
"s(0) = max(max(s(?z:Nat),?y:Nat),?y:Nat)",
"s(s(?x:Nat)) = max(?x:Nat,s(s(0)))",
"s(s(?x:Nat)) = max(s(?x:Nat),s(?z:Nat))",
"s(s(?y:Nat)) = max(0,max(?x:Nat,?y:Nat))",
"s(s(?y:Nat)) = max(s(0),s(?x:Nat))",
"s(s(?y:Nat)) = max(max(?y:Nat,?y:Nat),?z:Nat)",
"s(s(?y:Nat)) = max(max(0,?x:Nat),?y:Nat)",
"s(s(?z:Nat)) = max(?y:Nat,max(0,?x:Nat))",
"s(s(?z:Nat)) = max(0,s(s(?y:Nat)))",
"s(s(?z:Nat)) = max(s(?x:Nat),s(?z:Nat))",
"s(s(0)) = max(?x:Nat,max(?y:Nat,?x:Nat))",
"s(s(0)) = max(?x:Nat,max(?z:Nat,?y:Nat))",
"s(s(0)) = max(0,max(?x:Nat,?z:Nat))",
"max(?x:Nat,?x:Nat) = s(s(s(s(?x:Nat))))",
"max(?x:Nat,?x:Nat) = s(max(?x:Nat,s(0)))",
"max(?x:Nat,?x:Nat) = s(max(s(?y:Nat),?x:Nat))",
"max(?x:Nat,?y:Nat) = s(s(max(?y:Nat,?z:Nat)))",
"max(?x:Nat,?y:Nat) = s(s(max(?z:Nat,?x:Nat)))",
"max(?x:Nat,?y:Nat) = max(0,s(s(?x:Nat)))",
"max(?x:Nat,?y:Nat) = max(s(s(?y:Nat)),?x:Nat)",
"max(?x:Nat,?y:Nat) = max(max(?y:Nat,?y:Nat),0)",
"max(?x:Nat,?y:Nat) = max(max(?z:Nat,0),?y:Nat)",
"max(?x:Nat,?y:Nat) = max(max(0,?x:Nat),0)",
"max(?x:Nat,?z:Nat) = max(?y:Nat,max(?x:Nat,?x:Nat))",
"max(?x:Nat,0) = max(?x:Nat,s(s(?x:Nat)))",
"max(?x:Nat,0) = max(s(s(?x:Nat)),?y:Nat)",
"max(?y:Nat,?x:Nat) = max(s(s(?x:Nat)),?z:Nat)",
"max(?y:Nat,?x:Nat) = max(max(?z:Nat,0),0)",
"max(?y:Nat,?y:Nat) = s(max(?x:Nat,s(0)))",
"max(?y:Nat,?y:Nat) = max(?y:Nat,max(?x:Nat,?x:Nat))",
"max(?y:Nat,?y:Nat) = max(?z:Nat,max(?y:Nat,?y:Nat))",
"max(?y:Nat,?z:Nat) = s(max(0,s(?y:Nat)))",
"max(?y:Nat,0) = s(s(max(0,?x:Nat)))",
"max(?y:Nat,0) = max(?y:Nat,max(?x:Nat,?y:Nat))",
"max(?y:Nat,0) = max(s(?x:Nat),s(0))",
"max(?z:Nat,?y:Nat) = s(max(?y:Nat,s(?z:Nat)))",
"max(?z:Nat,?y:Nat) = max(s(s(0)),?y:Nat)",
"max(?z:Nat,0) = max(?y:Nat,s(s(0)))",
"max(0,?x:Nat) = s(max(s(?x:Nat),?x:Nat))",
"max(0,?y:Nat) = max(?x:Nat,s(s(?z:Nat)))",
"max(0,?y:Nat) = max(max(?x:Nat,?x:Nat),?y:Nat)",
"max(0,?y:Nat) = max(max(?x:Nat,0),?x:Nat)",
"max(0,?z:Nat) = max(?y:Nat,max(0,?z:Nat))",
"max(0,?z:Nat) = max(s(?x:Nat),s(?z:Nat))",
"s(s(s(?x:Nat))) = max(s(0),?y:Nat)",
"s(s(s(?y:Nat))) = max(?y:Nat,s(?z:Nat))",
"s(max(?x:Nat,?y:Nat)) = max(?z:Nat,s(?z:Nat))",
"s(max(?x:Nat,?y:Nat)) = max(0,s(?z:Nat))",
"max(?x:Nat,s(?x:Nat)) = s(max(0,?z:Nat))",
"max(?x:Nat,s(?y:Nat)) = max(s(0),?y:Nat)",
"max(?x:Nat,s(?z:Nat)) = max(s(?x:Nat),0)",
"max(?x:Nat,s(0)) = s(max(?y:Nat,?y:Nat))",
"max(?y:Nat,s(?x:Nat)) = s(max(?x:Nat,?x:Nat))",
"max(?y:Nat,s(?y:Nat)) = max(?x:Nat,s(?y:Nat))",
"max(?y:Nat,s(?z:Nat)) = max(s(?x:Nat),?x:Nat)",
"max(?z:Nat,s(?y:Nat)) = s(s(s(?y:Nat)))",
"max(0,s(?x:Nat)) = s(max(0,?y:Nat))",
"max(0,s(?y:Nat)) = max(s(?x:Nat),?x:Nat)",
"max(0,s(?z:Nat)) = s(s(s(0)))",
"max(0,s(0)) = s(max(?y:Nat,?z:Nat))",
"max(s(?x:Nat),?x:Nat) = max(s(?x:Nat),?z:Nat)",
"max(s(?x:Nat),?x:Nat) = max(s(?y:Nat),?z:Nat)",
"max(s(?x:Nat),?y:Nat) = s(max(?x:Nat,?y:Nat))",
"max(s(?x:Nat),0) = s(max(?y:Nat,?z:Nat))",
"max(s(?y:Nat),?x:Nat) = max(s(0),?x:Nat)",
"max(s(?y:Nat),?y:Nat) = max(?x:Nat,s(?y:Nat))",
"max(s(?z:Nat),?y:Nat) = max(?y:Nat,s(?x:Nat))",
"max(s(?z:Nat),?y:Nat) = max(s(0),?x:Nat)",
"max(s(?z:Nat),?z:Nat) = s(max(?x:Nat,?x:Nat))",
"max(s(?z:Nat),?z:Nat) = s(max(?y:Nat,?x:Nat))",
"max(s(?z:Nat),0) = s(s(s(?z:Nat)))",
"max(s(0),?y:Nat) = max(?y:Nat,s(?z:Nat))",
"max(s(0),0) = max(s(0),?x:Nat)",
"s(s(s(s(?z:Nat)))) = max(0,?x:Nat)",
"s(s(max(?y:Nat,0))) = max(?x:Nat,?y:Nat)",
"s(max(?x:Nat,s(?z:Nat))) = max(?y:Nat,0)",
"s(max(0,s(?x:Nat))) = max(?z:Nat,?x:Nat)",
"s(max(s(?y:Nat),?y:Nat)) = max(?z:Nat,0)",
"s(max(s(?y:Nat),?z:Nat)) = max(?x:Nat,?z:Nat)",
"s(max(s(?z:Nat),?x:Nat)) = max(?x:Nat,0)",
"s(max(s(0),0)) = max(?x:Nat,?x:Nat)",
"max(?x:Nat,s(s(?x:Nat))) = max(?x:Nat,?y:Nat)",
"max(?x:Nat,s(s(0))) = s(s(0))",
"max(?x:Nat,s(s(0))) = max(?x:Nat,?z:Nat)",
"max(?x:Nat,max(?z:Nat,?x:Nat)) = max(?x:Nat,?y:Nat)",
"max(?x:Nat,max(?z:Nat,?x:Nat)) = max(0,0)",
"max(?y:Nat,s(s(?x:Nat))) = max(?y:Nat,?y:Nat)",
"max(?y:Nat,s(s(?z:Nat))) = max(?x:Nat,?x:Nat)",
"max(?y:Nat,max(0,0)) = max(?z:Nat,?y:Nat)",
"max(0,s(s(?y:Nat))) = max(?x:Nat,?z:Nat)",
"max(s(?x:Nat),s(?y:Nat)) = s(s(?z:Nat))",
"max(s(?x:Nat),s(?y:Nat)) = max(?x:Nat,?y:Nat)",
"max(s(?y:Nat),s(?z:Nat)) = max(?x:Nat,?z:Nat)",
"max(s(s(?x:Nat)),?x:Nat) = max(0,?y:Nat)",
"max(s(s(?x:Nat)),?z:Nat) = max(?z:Nat,0)",
"max(s(s(?y:Nat)),?z:Nat) = max(?x:Nat,?x:Nat)",
"max(s(s(?z:Nat)),?x:Nat) = s(s(?y:Nat))",
"max(max(?x:Nat,?z:Nat),?x:Nat) = max(?x:Nat,?y:Nat)",
"max(max(?x:Nat,?z:Nat),?x:Nat) = max(?y:Nat,?z:Nat)",
"max(max(?x:Nat,?z:Nat),?y:Nat) = max(?y:Nat,?y:Nat)",
"max(max(?y:Nat,?x:Nat),?z:Nat) = max(?y:Nat,?y:Nat)",
"max(max(?y:Nat,?y:Nat),?x:Nat) = max(?x:Nat,?x:Nat)",
"max(max(?y:Nat,?z:Nat),?z:Nat) = s(s(?y:Nat))",
"max(max(?z:Nat,?x:Nat),0) = max(?x:Nat,?y:Nat)",
"max(max(?z:Nat,?y:Nat),?y:Nat) = max(?x:Nat,?y:Nat)",
"max(max(?z:Nat,0),?x:Nat) = max(?x:Nat,?x:Nat)",
"max(?x:Nat,s(s(s(?x:Nat)))) = s(?x:Nat)",
"max(?x:Nat,max(s(?y:Nat),?y:Nat)) = s(?y:Nat)",
"max(?y:Nat,s(s(s(0)))) = s(?y:Nat)",
"max(?y:Nat,s(max(0,?x:Nat))) = s(?x:Nat)",
"max(?y:Nat,max(?x:Nat,s(?y:Nat))) = s(?z:Nat)",
"max(?y:Nat,max(0,s(?x:Nat))) = s(?x:Nat)",
"max(?y:Nat,max(s(?y:Nat),?z:Nat)) = s(?z:Nat)",
"max(0,s(s(s(?x:Nat)))) = s(?y:Nat)",
"max(0,s(max(?y:Nat,?x:Nat))) = s(?y:Nat)",
"max(0,max(s(0),?z:Nat)) = s(0)",
"max(s(?x:Nat),max(?x:Nat,?z:Nat)) = s(?y:Nat)",
"max(s(?x:Nat),max(?x:Nat,?z:Nat)) = s(?z:Nat)",
"max(s(?x:Nat),max(0,?y:Nat)) = s(?x:Nat)",
"max(s(?y:Nat),max(?x:Nat,?x:Nat)) = s(0)",
"max(s(max(?x:Nat,0)),?y:Nat) = s(?x:Nat)",
"max(s(max(?y:Nat,?y:Nat)),0) = s(?x:Nat)",
"max(s(max(0,0)),?y:Nat) = s(?z:Nat)",
"max(max(?x:Nat,?z:Nat),s(?y:Nat)) = s(?y:Nat)",
"max(max(?x:Nat,s(?y:Nat)),?y:Nat) = s(?x:Nat)",
"max(max(?y:Nat,?x:Nat),s(?z:Nat)) = s(?x:Nat)",
"max(max(?y:Nat,?y:Nat),s(0)) = s(?y:Nat)",
"max(max(?y:Nat,0),s(?y:Nat)) = s(?x:Nat)",
"max(max(?y:Nat,s(?y:Nat)),?x:Nat) = s(?x:Nat)",
"max(max(?y:Nat,s(?y:Nat)),?y:Nat) = s(?z:Nat)",
"max(max(?y:Nat,s(?z:Nat)),0) = s(?z:Nat)",
"max(max(0,s(?z:Nat)),?x:Nat) = s(?x:Nat)",
"max(max(s(?x:Nat),?x:Nat),0) = s(?y:Nat)",
"s(?x:Nat) = max(?y:Nat,max(max(?z:Nat,?y:Nat),?z:Nat))",
"s(?x:Nat) = max(s(?x:Nat),max(s(?y:Nat),?x:Nat))",
"s(?x:Nat) = max(s(s(?x:Nat)),s(s(?x:Nat)))",
"s(?x:Nat) = max(s(max(s(?x:Nat),0)),?y:Nat)",
"s(?x:Nat) = max(max(?z:Nat,0),s(s(?y:Nat)))",
"s(?x:Nat) = max(max(0,?y:Nat),max(?y:Nat,?y:Nat))",
"s(?y:Nat) = max(s(?x:Nat),max(s(?y:Nat),0))",
"s(?y:Nat) = max(s(s(?z:Nat)),max(?z:Nat,?z:Nat))",
"s(?y:Nat) = max(s(max(?x:Nat,?y:Nat)),s(?x:Nat))",
"s(?y:Nat) = max(s(max(?x:Nat,?y:Nat)),s(0))",
"s(?y:Nat) = max(max(?y:Nat,max(?z:Nat,?z:Nat)),?z:Nat)",
"s(?y:Nat) = max(max(?z:Nat,?x:Nat),max(0,?y:Nat))",
"s(?y:Nat) = max(max(max(?z:Nat,0),?x:Nat),?y:Nat)",
"s(?y:Nat) = max(max(max(0,0),?z:Nat),?x:Nat)",
"s(?z:Nat) = max(max(?y:Nat,?y:Nat),s(s(?y:Nat)))",
"s(?z:Nat) = max(max(s(?y:Nat),s(?x:Nat)),?x:Nat)",
"s(0) = max(?x:Nat,max(?y:Nat,max(?y:Nat,?x:Nat)))",
"s(0) = max(?z:Nat,max(max(?x:Nat,?x:Nat),?y:Nat))",
"s(0) = max(max(0,s(s(?x:Nat))),?x:Nat)",
"s(s(?x:Nat)) = max(max(?x:Nat,s(?x:Nat)),?y:Nat)",
"s(s(?y:Nat)) = max(s(0),max(?x:Nat,?y:Nat))",
"s(s(0)) = max(max(?x:Nat,s(?z:Nat)),?y:Nat)",
"max(?x:Nat,?x:Nat) = max(max(s(?y:Nat),?y:Nat),?x:Nat)",
"max(?x:Nat,?y:Nat) = s(max(?z:Nat,max(?x:Nat,?x:Nat)))",
"max(?x:Nat,?y:Nat) = s(max(max(?y:Nat,0),?x:Nat))",
"max(?x:Nat,?y:Nat) = max(s(s(?x:Nat)),s(?z:Nat))",
"max(?x:Nat,0) = s(max(max(?x:Nat,?y:Nat),?x:Nat))",
"max(?x:Nat,0) = max(s(max(?y:Nat,?z:Nat)),0)",
"max(?y:Nat,?y:Nat) = s(s(s(max(?x:Nat,?z:Nat))))",
"max(?y:Nat,0) = max(?x:Nat,max(?x:Nat,s(?y:Nat)))",
"max(?y:Nat,0) = max(s(0),s(s(?x:Nat)))",
"max(?z:Nat,?y:Nat) = max(s(max(?y:Nat,?x:Nat)),?z:Nat)",
"max(?z:Nat,?y:Nat) = max(max(0,s(?y:Nat)),?x:Nat)",
"max(?z:Nat,0) = max(max(0,s(?y:Nat)),?x:Nat)",
"max(0,?y:Nat) = max(max(?z:Nat,s(0)),?y:Nat)",
"max(0,0) = max(?x:Nat,max(s(?z:Nat),?x:Nat))",
"s(s(s(?x:Nat))) = max(max(?y:Nat,0),?z:Nat)",
"s(max(?x:Nat,?x:Nat)) = max(s(s(0)),?y:Nat)",
"s(max(?y:Nat,?x:Nat)) = max(max(?y:Nat,?z:Nat),?y:Nat)",
"s(max(?y:Nat,?y:Nat)) = max(max(?x:Nat,?x:Nat),?y:Nat)",
"s(max(?z:Nat,?x:Nat)) = max(s(s(?y:Nat)),?x:Nat)",
"s(max(0,?x:Nat)) = max(?x:Nat,max(?z:Nat,?x:Nat))",
"s(max(0,?z:Nat)) = max(?y:Nat,max(?x:Nat,?x:Nat))",
"max(?x:Nat,s(?x:Nat)) = max(0,s(s(0)))",
"max(?y:Nat,s(?y:Nat)) = max(max(?y:Nat,0),?y:Nat)",
"max(?z:Nat,s(?x:Nat)) = s(s(max(?x:Nat,?z:Nat)))",
"max(0,s(?y:Nat)) = s(max(s(?x:Nat),?z:Nat))",
"max(s(?x:Nat),?x:Nat) = max(max(0,0),?y:Nat)",
"max(s(?x:Nat),?y:Nat) = max(?x:Nat,max(?z:Nat,?z:Nat))",
"max(s(?x:Nat),?z:Nat) = max(max(?x:Nat,?x:Nat),?y:Nat)",
"max(s(?y:Nat),?x:Nat) = s(max(?y:Nat,s(?z:Nat)))",
"max(s(?y:Nat),?x:Nat) = max(?y:Nat,s(s(?x:Nat)))",
"s(s(max(?y:Nat,?x:Nat))) = max(?z:Nat,s(?y:Nat))",
"s(s(max(?y:Nat,0))) = max(s(?x:Nat),?x:Nat)",
"s(max(s(?x:Nat),?x:Nat)) = max(s(?x:Nat),?y:Nat)",
"max(?x:Nat,s(s(?x:Nat))) = s(s(s(?x:Nat)))",
"max(?x:Nat,max(?y:Nat,?y:Nat)) = max(s(?z:Nat),?y:Nat)",
"max(?y:Nat,max(?x:Nat,?y:Nat)) = s(s(s(?y:Nat)))",
"max(?y:Nat,max(?z:Nat,?x:Nat)) = s(max(?y:Nat,?z:Nat))",
"max(?y:Nat,max(?z:Nat,?y:Nat)) = s(max(0,?x:Nat))",
"max(?z:Nat,max(?y:Nat,?x:Nat)) = max(?z:Nat,s(?x:Nat))",
"max(?z:Nat,max(?y:Nat,?y:Nat)) = s(s(s(?z:Nat)))",
"max(0,max(?y:Nat,?y:Nat)) = s(max(0,0))",
"max(0,max(?z:Nat,?x:Nat)) = s(max(0,?x:Nat))",
"max(s(?x:Nat),s(?z:Nat)) = s(s(s(?z:Nat)))",
"max(s(s(0)),0) = max(0,s(?y:Nat))",
"max(max(?x:Nat,?x:Nat),?x:Nat) = max(?z:Nat,s(0))",
"max(max(?x:Nat,?z:Nat),0) = max(?y:Nat,s(?y:Nat))",
"max(max(?x:Nat,0),?x:Nat) = max(?y:Nat,s(?y:Nat))",
"max(max(?y:Nat,?y:Nat),?z:Nat) = max(s(?x:Nat),?y:Nat)",
"max(max(?z:Nat,?x:Nat),?y:Nat) = max(?y:Nat,s(?x:Nat))",
"max(max(0,0),?y:Nat) = max(?x:Nat,s(0))",
"s(s(s(s(s(?x:Nat))))) = max(?y:Nat,?x:Nat)",
"s(max(?x:Nat,max(?x:Nat,?z:Nat))) = max(?x:Nat,?x:Nat)",
"s(max(?z:Nat,max(?z:Nat,?x:Nat))) = max(?y:Nat,?y:Nat)",
"max(?x:Nat,max(?x:Nat,s(?x:Nat))) = max(?y:Nat,?x:Nat)",
"max(?x:Nat,max(s(0),?x:Nat)) = max(?x:Nat,?x:Nat)",
"max(?y:Nat,s(s(s(?z:Nat)))) = s(s(?z:Nat))",
"max(?y:Nat,s(max(?y:Nat,0))) = s(s(0))",
"max(?y:Nat,s(max(?y:Nat,0))) = max(?x:Nat,?z:Nat)",
"max(?y:Nat,max(?y:Nat,s(?x:Nat))) = s(s(?x:Nat))",
"max(?y:Nat,max(s(?x:Nat),?x:Nat)) = max(?z:Nat,?y:Nat)",
"max(?y:Nat,max(s(?y:Nat),?z:Nat)) = max(?x:Nat,?z:Nat)",
"max(?z:Nat,s(s(s(0)))) = s(s(?y:Nat))",
"max(?z:Nat,max(?y:Nat,s(?z:Nat))) = max(?z:Nat,?y:Nat)",
"max(?z:Nat,max(?z:Nat,s(?y:Nat))) = max(?x:Nat,?y:Nat)",
"max(s(?x:Nat),max(?x:Nat,?y:Nat)) = max(?y:Nat,0)",
"max(s(?x:Nat),max(0,?y:Nat)) = max(0,?y:Nat)",
"max(s(s(?x:Nat)),s(?x:Nat)) = max(?z:Nat,?y:Nat)",
"max(s(s(?x:Nat)),s(?y:Nat)) = s(s(?x:Nat))",
"max(s(s(?x:Nat)),s(?y:Nat)) = max(0,?z:Nat)",
"max(s(max(?x:Nat,?x:Nat)),?y:Nat) = max(?y:Nat,?y:Nat)",
"max(max(?x:Nat,?x:Nat),s(?z:Nat)) = max(?z:Nat,?y:Nat)",
"max(max(?x:Nat,s(?x:Nat)),?y:Nat) = max(?z:Nat,0)",
"max(max(?z:Nat,s(?x:Nat)),?x:Nat) = max(?y:Nat,0)",
"max(max(0,s(?y:Nat)),?y:Nat) = max(?x:Nat,0)",
"max(?x:Nat,s(s(max(?y:Nat,?z:Nat)))) = s(?y:Nat)",
"max(?x:Nat,s(max(s(?x:Nat),0))) = s(?x:Nat)",
"max(?x:Nat,s(max(s(?z:Nat),0))) = s(?y:Nat)",
"max(?x:Nat,max(?x:Nat,max(?y:Nat,?x:Nat))) = s(?x:Nat)",
"max(?x:Nat,max(?y:Nat,max(?x:Nat,?y:Nat))) = s(?x:Nat)",
"max(?x:Nat,max(s(?y:Nat),s(?z:Nat))) = s(0)",
"max(?x:Nat,max(max(?x:Nat,0),?x:Nat)) = s(?y:Nat)",
"max(?x:Nat,max(max(0,?z:Nat),0)) = s(?x:Nat)",
"max(?y:Nat,s(s(max(?y:Nat,?y:Nat)))) = s(?y:Nat)",
"max(?y:Nat,s(max(s(?x:Nat),?y:Nat))) = s(?x:Nat)",
"max(?y:Nat,max(?y:Nat,s(s(0)))) = s(?x:Nat)",
"max(?y:Nat,max(?z:Nat,max(?x:Nat,0))) = s(?z:Nat)",
"max(?y:Nat,max(max(?y:Nat,?y:Nat),?y:Nat)) = s(?x:Nat)",
"max(?z:Nat,max(?x:Nat,max(?z:Nat,?y:Nat))) = s(?x:Nat)",
"max(?z:Nat,max(?y:Nat,s(s(?x:Nat)))) = s(0)",
"max(?z:Nat,max(s(s(?y:Nat)),?x:Nat)) = s(?x:Nat)",
"max(?z:Nat,max(s(s(?z:Nat)),?x:Nat)) = s(?x:Nat)",
"max(0,max(?x:Nat,max(?y:Nat,?x:Nat))) = s(?z:Nat)",
"max(s(?z:Nat),max(s(?x:Nat),?z:Nat)) = s(?x:Nat)",
"max(s(?z:Nat),max(s(?x:Nat),0)) = s(?y:Nat)",
"max(s(0),max(?x:Nat,s(?z:Nat))) = s(?x:Nat)",
"max(s(s(max(?x:Nat,?z:Nat))),?x:Nat) = s(?y:Nat)",
"max(s(max(?y:Nat,s(0))),?y:Nat) = s(?y:Nat)",
"max(max(0,?x:Nat),s(s(?x:Nat))) = s(?y:Nat)",
"max(max(s(0),?z:Nat),s(?x:Nat)) = s(?y:Nat)",
"max(max(max(?y:Nat,?y:Nat),0),0) = s(?z:Nat)",
"max(max(max(?y:Nat,?z:Nat),?y:Nat),?y:Nat) = s(?z:Nat)",
"max(max(max(?y:Nat,0),0),0) = s(?z:Nat)",
"s(?x:Nat) = max(s(?y:Nat),max(s(?y:Nat),s(?x:Nat)))",
"s(?x:Nat) = max(s(s(s(max(?x:Nat,0)))),?x:Nat)",
"s(?x:Nat) = max(s(s(max(s(?z:Nat),?y:Nat))),?y:Nat)",
"s(?x:Nat) = max(s(max(?x:Nat,?y:Nat)),max(?x:Nat,0))",
"s(?x:Nat) = max(s(max(?x:Nat,max(?x:Nat,?z:Nat))),?z:Nat)",
"s(?x:Nat) = max(s(max(0,s(?y:Nat))),s(0))",
"s(?x:Nat) = max(max(0,s(max(?y:Nat,?z:Nat))),?x:Nat)",
"s(?y:Nat) = max(s(?x:Nat),max(max(?y:Nat,?y:Nat),?x:Nat))",
"s(?y:Nat) = max(s(s(max(?y:Nat,s(?y:Nat)))),?y:Nat)",
"s(?y:Nat) = max(max(?x:Nat,?x:Nat),max(s(0),?y:Nat))",
"s(?y:Nat) = max(max(0,s(?x:Nat)),s(s(?y:Nat)))",
"s(?z:Nat) = max(?x:Nat,s(max(max(?z:Nat,?y:Nat),?x:Nat)))",
"s(?z:Nat) = max(0,max(0,s(s(s(?z:Nat)))))",
"s(?z:Nat) = max(s(s(s(max(?y:Nat,?y:Nat)))),?x:Nat)",
"s(?z:Nat) = max(max(?x:Nat,0),max(s(?y:Nat),?y:Nat))",
"s(?z:Nat) = max(max(?x:Nat,s(0)),s(s(0)))",
"s(?z:Nat) = max(max(s(?y:Nat),0),max(?x:Nat,?x:Nat))",
"s(0) = max(?x:Nat,max(s(max(?z:Nat,?y:Nat)),?z:Nat))",
"s(0) = max(s(0),max(0,s(s(?x:Nat))))",
"s(s(?y:Nat)) = max(s(0),max(?y:Nat,s(?y:Nat)))",
"s(s(?y:Nat)) = max(max(?y:Nat,?x:Nat),max(0,?y:Nat))",
"s(s(?y:Nat)) = max(max(max(?x:Nat,?x:Nat),?z:Nat),0)",
"s(s(?z:Nat)) = max(?x:Nat,max(?y:Nat,s(s(?y:Nat))))",
"s(s(0)) = max(?y:Nat,max(?x:Nat,max(?y:Nat,?y:Nat)))",
"s(s(0)) = max(s(?y:Nat),max(?y:Nat,s(?x:Nat)))",
"max(?x:Nat,?x:Nat) = s(s(max(?y:Nat,max(?z:Nat,?y:Nat))))",
"max(?x:Nat,?x:Nat) = s(max(s(max(?z:Nat,?x:Nat)),0))",
"max(?x:Nat,?y:Nat) = max(?y:Nat,s(max(s(?x:Nat),?x:Nat)))",
"max(?x:Nat,?z:Nat) = max(s(0),max(s(?y:Nat),0))",
"max(?x:Nat,0) = max(?y:Nat,max(s(?x:Nat),s(0)))",
"max(?y:Nat,?y:Nat) = max(?z:Nat,s(max(s(?y:Nat),?y:Nat)))",
"max(?y:Nat,?y:Nat) = max(max(s(s(?y:Nat)),?y:Nat),?x:Nat)",
"max(?y:Nat,0) = s(s(max(?y:Nat,max(?x:Nat,?y:Nat))))",
"max(?y:Nat,0) = s(s(max(max(0,?y:Nat),?x:Nat)))",
"max(?z:Nat,?x:Nat) = s(max(?x:Nat,s(max(?y:Nat,?y:Nat))))",
"max(?z:Nat,?x:Nat) = max(?y:Nat,s(max(s(?x:Nat),0)))",
"max(?z:Nat,?z:Nat) = s(max(0,max(?y:Nat,s(0))))",
"max(0,?x:Nat) = s(max(s(s(?x:Nat)),s(?y:Nat)))",
"max(0,?x:Nat) = s(max(s(s(s(0))),?x:Nat))",
"max(0,?x:Nat) = max(max(s(s(?x:Nat)),0),0)",
"max(0,?y:Nat) = s(max(max(s(?y:Nat),0),0))",
"max(0,?y:Nat) = max(?z:Nat,max(max(?y:Nat,?x:Nat),?y:Nat))",
"max(0,?z:Nat) = s(max(max(?x:Nat,s(?x:Nat)),?y:Nat))",
"s(max(?x:Nat,?y:Nat)) = max(?z:Nat,s(max(?z:Nat,?x:Nat)))",
"max(?y:Nat,s(?y:Nat)) = s(max(s(s(?x:Nat)),?y:Nat))",
"max(?y:Nat,s(?y:Nat)) = max(s(max(?y:Nat,?z:Nat)),?y:Nat)",
"max(0,s(?x:Nat)) = max(?y:Nat,s(max(?z:Nat,?y:Nat)))",
"max(0,s(0)) = max(?x:Nat,s(max(0,?y:Nat)))",
"max(s(?x:Nat),?z:Nat) = max(max(?x:Nat,?y:Nat),s(0))",
"max(s(?z:Nat),?z:Nat) = max(s(s(s(?y:Nat))),?y:Nat)",
"max(s(0),?y:Nat) = max(max(?z:Nat,0),s(?x:Nat))",
"max(s(0),?z:Nat) = max(?x:Nat,s(max(?y:Nat,?y:Nat)))",
"s(s(max(?x:Nat,?x:Nat))) = max(max(?y:Nat,?x:Nat),?z:Nat)",
"s(s(max(?y:Nat,?y:Nat))) = max(?x:Nat,max(?y:Nat,0))",
"s(s(max(?z:Nat,0))) = max(max(?y:Nat,?y:Nat),0)",
"s(s(max(0,?z:Nat))) = max(max(?y:Nat,?z:Nat),?y:Nat)",
"s(max(?z:Nat,s(?y:Nat))) = max(?y:Nat,max(?x:Nat,?y:Nat))",
"s(max(s(?x:Nat),?x:Nat)) = max(max(0,?y:Nat),?y:Nat)",
"max(?y:Nat,s(s(?x:Nat))) = max(s(?x:Nat),s(?x:Nat))",
"max(?z:Nat,s(s(?z:Nat))) = max(max(?x:Nat,?z:Nat),0)",
"max(s(?x:Nat),s(?y:Nat)) = max(max(?y:Nat,0),?y:Nat)",
"max(s(s(?y:Nat)),?y:Nat) = max(?x:Nat,max(?z:Nat,?z:Nat))",
"max(max(?x:Nat,?y:Nat),?y:Nat) = max(max(?y:Nat,0),?z:Nat)",
"max(max(?x:Nat,0),?y:Nat) = max(?x:Nat,max(?x:Nat,?z:Nat))",
"max(max(?y:Nat,?y:Nat),?x:Nat) = max(max(?x:Nat,?y:Nat),?z:Nat)",
"max(max(?y:Nat,?y:Nat),?y:Nat) = max(?x:Nat,s(s(?y:Nat)))",
"max(max(?y:Nat,?z:Nat),?y:Nat) = s(s(max(?z:Nat,?y:Nat)))",
"max(max(?z:Nat,0),?y:Nat) = max(max(?y:Nat,?z:Nat),?x:Nat)",
"s(s(max(s(?x:Nat),0))) = max(s(0),?z:Nat)",
"s(max(0,max(0,?y:Nat))) = max(s(?x:Nat),?z:Nat)",
"s(max(s(?y:Nat),s(?x:Nat))) = max(s(?y:Nat),?y:Nat)",
"max(?y:Nat,s(max(?y:Nat,0))) = max(?y:Nat,s(0))",
"max(?y:Nat,max(?z:Nat,s(?z:Nat))) = s(max(?y:Nat,?z:Nat))",
"max(0,s(max(?y:Nat,?y:Nat))) = max(?z:Nat,s(?y:Nat))",
"max(s(?z:Nat),s(s(?x:Nat))) = max(s(0),?z:Nat)",
"max(s(s(s(?y:Nat))),?y:Nat) = s(max(?y:Nat,?z:Nat))",
"max(s(max(?y:Nat,?y:Nat)),?x:Nat) = max(?y:Nat,s(?x:Nat))",
"max(max(?x:Nat,s(?x:Nat)),?z:Nat) = s(max(?x:Nat,0))",
"max(max(?y:Nat,s(0)),?x:Nat) = max(s(?y:Nat),0)",
"max(max(?z:Nat,?z:Nat),s(0)) = s(max(?x:Nat,?x:Nat))",
"max(max(?z:Nat,s(?z:Nat)),?x:Nat) = max(?z:Nat,s(?y:Nat))",
"s(s(max(s(s(0)),?y:Nat))) = max(?y:Nat,?x:Nat)",
"s(max(?z:Nat,s(s(s(?y:Nat))))) = max(0,?y:Nat)",
"s(max(s(s(s(?y:Nat))),?y:Nat)) = max(0,?z:Nat)",
"s(max(max(?z:Nat,?x:Nat),s(?x:Nat))) = max(?x:Nat,?z:Nat)",
"max(?x:Nat,s(s(max(?z:Nat,?x:Nat)))) = max(?y:Nat,?z:Nat)",
"max(?x:Nat,s(s(max(?z:Nat,?z:Nat)))) = max(?z:Nat,0)",
"max(?x:Nat,s(max(s(?x:Nat),?x:Nat))) = max(?y:Nat,?x:Nat)",
"max(?x:Nat,max(?y:Nat,max(?x:Nat,?y:Nat))) = s(s(0))",
"max(?x:Nat,max(s(?y:Nat),s(?x:Nat))) = max(0,?y:Nat)",
"max(?x:Nat,max(max(?x:Nat,?x:Nat),?y:Nat)) = s(s(?y:Nat))",
"max(?y:Nat,s(s(max(?x:Nat,?x:Nat)))) = max(?x:Nat,?x:Nat)",
"max(0,max(?x:Nat,max(?y:Nat,0))) = max(0,?x:Nat)",
"max(s(?x:Nat),max(s(?x:Nat),0)) = max(?z:Nat,?y:Nat)",
"max(s(?y:Nat),s(max(?x:Nat,?z:Nat))) = max(?y:Nat,?y:Nat)",
"max(s(?y:Nat),s(max(?y:Nat,?x:Nat))) = max(0,?y:Nat)",
"max(s(?y:Nat),max(?x:Nat,s(?x:Nat))) = s(s(0))",
"max(s(s(s(?x:Nat))),s(?x:Nat)) = max(?x:Nat,?x:Nat)",
"max(s(max(s(?x:Nat),?x:Nat)),?z:Nat) = s(s(0))",
"max(max(?x:Nat,?x:Nat),max(0,?x:Nat)) = max(?x:Nat,?z:Nat)",
"max(max(?x:Nat,?y:Nat),max(?y:Nat,0)) = max(?z:Nat,?x:Nat)",
"max(max(s(s(?y:Nat)),?z:Nat),?z:Nat) = s(s(?y:Nat))",
"max(max(max(?x:Nat,?x:Nat),0),?y:Nat) = max(?y:Nat,?y:Nat)",
"max(max(max(?y:Nat,0),?x:Nat),?y:Nat) = max(0,?x:Nat)",
"max(?x:Nat,s(s(max(s(?y:Nat),?y:Nat)))) = s(?z:Nat)",
"max(?x:Nat,max(?y:Nat,s(s(s(?x:Nat))))) = s(?x:Nat)",
"max(?x:Nat,max(?z:Nat,max(0,s(?y:Nat)))) = s(?y:Nat)",
"max(?x:Nat,max(0,s(max(?z:Nat,0)))) = s(?y:Nat)",
"max(?x:Nat,max(s(max(?y:Nat,?x:Nat)),?z:Nat)) = s(?x:Nat)",
"max(?x:Nat,max(s(max(?y:Nat,?z:Nat)),?y:Nat)) = s(?y:Nat)",
"max(?y:Nat,s(max(?x:Nat,max(?y:Nat,?y:Nat)))) = s(?z:Nat)",
"max(?y:Nat,s(max(s(?x:Nat),s(0)))) = s(?y:Nat)",
"max(?y:Nat,max(?x:Nat,max(0,s(?y:Nat)))) = s(?z:Nat)",
"max(?z:Nat,s(max(s(s(?x:Nat)),?y:Nat))) = s(?z:Nat)",
"max(s(0),max(?y:Nat,max(?y:Nat,?y:Nat))) = s(?z:Nat)",
"max(s(s(?y:Nat)),s(max(?x:Nat,?y:Nat))) = s(?z:Nat)",
"max(s(s(s(?y:Nat))),max(?x:Nat,?y:Nat)) = s(?x:Nat)",
"max(max(?y:Nat,max(?x:Nat,s(?x:Nat))),?x:Nat) = s(?x:Nat)",
"max(max(0,s(0)),max(?y:Nat,0)) = s(?y:Nat)",
"max(max(s(?y:Nat),s(?x:Nat)),s(?y:Nat)) = s(?y:Nat)",
"max(max(s(?z:Nat),max(0,?y:Nat)),?y:Nat) = s(?x:Nat)",
"max(max(max(?x:Nat,?x:Nat),s(?z:Nat)),?x:Nat) = s(?x:Nat)",
"max(max(max(?y:Nat,?x:Nat),s(?z:Nat)),0) = s(?z:Nat)",
"max(max(max(?z:Nat,?y:Nat),?y:Nat),s(?z:Nat)) = s(?x:Nat)",
"s(?x:Nat) = max(s(?x:Nat),max(max(?y:Nat,s(?x:Nat)),?y:Nat))",
"s(?x:Nat) = max(max(?y:Nat,max(?y:Nat,max(0,?x:Nat))),?y:Nat)",
"s(?x:Nat) = max(max(?z:Nat,s(max(?y:Nat,s(?z:Nat)))),?z:Nat)",
"s(?y:Nat) = max(?x:Nat,max(?z:Nat,max(?y:Nat,s(s(?x:Nat)))))",
"s(?y:Nat) = max(?z:Nat,s(s(max(0,max(?x:Nat,?z:Nat)))))",
"s(?y:Nat) = max(?z:Nat,max(?z:Nat,max(s(s(?x:Nat)),?x:Nat)))",
"s(?y:Nat) = max(0,max(0,s(max(?x:Nat,s(?x:Nat)))))",
"s(?y:Nat) = max(s(max(?x:Nat,s(max(?y:Nat,?x:Nat)))),?y:Nat)",
"s(?y:Nat) = max(max(?y:Nat,s(s(?y:Nat))),s(s(?y:Nat)))",
"s(?y:Nat) = max(max(s(0),?x:Nat),max(s(?x:Nat),?y:Nat))",
"s(?y:Nat) = max(max(max(s(?y:Nat),?x:Nat),s(?x:Nat)),?y:Nat)",
"s(?z:Nat) = max(?x:Nat,max(s(s(?y:Nat)),max(?x:Nat,?x:Nat)))",
"s(?z:Nat) = max(?x:Nat,max(max(max(?y:Nat,?y:Nat),0),0))",
"s(?z:Nat) = max(?y:Nat,max(max(0,s(s(?x:Nat))),?y:Nat))",
"s(0) = max(max(max(?x:Nat,?x:Nat),?x:Nat),max(?x:Nat,?x:Nat))",
"s(s(?x:Nat)) = max(?x:Nat,max(?z:Nat,s(max(?y:Nat,?x:Nat))))",
"max(?x:Nat,?x:Nat) = max(max(?x:Nat,s(?x:Nat)),max(?x:Nat,?x:Nat))",
"max(?x:Nat,?x:Nat) = max(max(?z:Nat,?y:Nat),max(?x:Nat,s(?x:Nat)))",
"max(?x:Nat,?y:Nat) = s(max(max(0,max(0,?z:Nat)),?x:Nat))",
"max(?x:Nat,?y:Nat) = max(max(?x:Nat,s(max(?z:Nat,?x:Nat))),?x:Nat)",
"max(?x:Nat,?z:Nat) = max(max(?y:Nat,s(max(?z:Nat,?x:Nat))),?y:Nat)",
"max(?x:Nat,?z:Nat) = max(max(s(s(s(?x:Nat))),0),?y:Nat)",
"max(?x:Nat,0) = max(?y:Nat,max(max(s(?y:Nat),0),?x:Nat))",
"max(?x:Nat,0) = max(max(?x:Nat,max(?z:Nat,?y:Nat)),s(?x:Nat))",
"max(?y:Nat,?x:Nat) = max(max(s(max(?y:Nat,?y:Nat)),?x:Nat),?x:Nat)",
"max(?y:Nat,?y:Nat) = max(s(?z:Nat),s(max(?z:Nat,s(?z:Nat))))",
"max(?y:Nat,?y:Nat) = max(s(max(s(s(?y:Nat)),?x:Nat)),?x:Nat)",
"max(?y:Nat,0) = max(s(max(?x:Nat,?x:Nat)),max(?y:Nat,0))",
"max(0,?x:Nat) = s(max(?y:Nat,max(max(?z:Nat,?z:Nat),?x:Nat)))",
"max(0,0) = s(s(max(0,s(max(?y:Nat,?z:Nat)))))",
"s(max(?x:Nat,?x:Nat)) = max(s(s(?x:Nat)),max(?y:Nat,?y:Nat))",
"s(max(?x:Nat,?x:Nat)) = max(max(max(?y:Nat,?y:Nat),?x:Nat),?y:Nat)",
"s(max(?x:Nat,?z:Nat)) = max(max(?x:Nat,s(?z:Nat)),s(?y:Nat))",
"s(max(?y:Nat,?y:Nat)) = max(max(max(?x:Nat,?x:Nat),?z:Nat),?y:Nat)",
"s(max(0,?x:Nat)) = max(max(?y:Nat,s(s(?x:Nat))),?y:Nat)",
"max(?y:Nat,s(?y:Nat)) = max(?z:Nat,max(max(?z:Nat,0),?z:Nat))",
"max(s(?y:Nat),?x:Nat) = max(?z:Nat,max(?z:Nat,max(?y:Nat,?y:Nat)))",
"s(s(max(?y:Nat,?z:Nat))) = max(s(?y:Nat),max(?x:Nat,?x:Nat))",
"max(?y:Nat,max(0,?z:Nat)) = max(s(max(?y:Nat,0)),0)",
"max(0,s(s(0))) = max(0,max(s(?z:Nat),?y:Nat))",
"max(0,max(?y:Nat,?x:Nat)) = max(?y:Nat,s(max(?x:Nat,?x:Nat)))",
"max(max(?x:Nat,?y:Nat),?x:Nat) = max(max(s(?z:Nat),0),?y:Nat)",
"s(max(?y:Nat,max(?z:Nat,?y:Nat))) = max(max(?y:Nat,?y:Nat),0)",
"s(max(?z:Nat,s(s(?x:Nat)))) = max(s(0),s(?x:Nat))",
"max(?x:Nat,max(s(0),?z:Nat)) = max(max(0,?y:Nat),?x:Nat)",
"max(0,max(s(?y:Nat),?y:Nat)) = max(max(?x:Nat,?x:Nat),?z:Nat)",
"max(s(?y:Nat),s(s(?x:Nat))) = max(max(?y:Nat,?z:Nat),?y:Nat)",
"max(s(s(s(?y:Nat))),?x:Nat) = max(s(0),s(?x:Nat))",
"max(s(max(0,?z:Nat)),?y:Nat) = max(?x:Nat,s(s(?x:Nat)))",
"s(max(?x:Nat,max(?x:Nat,s(?x:Nat)))) = max(s(?y:Nat),?x:Nat)",
"s(max(s(0),max(?x:Nat,?y:Nat))) = max(s(?x:Nat),0)",
"max(?x:Nat,s(max(?y:Nat,s(?x:Nat)))) = s(max(?x:Nat,?x:Nat))",
"max(?x:Nat,s(max(?y:Nat,s(0)))) = max(s(?z:Nat),?y:Nat)",
"max(?x:Nat,max(max(?x:Nat,0),0)) = s(max(?x:Nat,?x:Nat))",
"max(?y:Nat,s(s(max(?x:Nat,?y:Nat)))) = max(s(?x:Nat),?y:Nat)",
"max(?y:Nat,max(max(0,?z:Nat),?x:Nat)) = max(s(?y:Nat),?z:Nat)",
"max(?z:Nat,max(max(?y:Nat,?x:Nat),0)) = s(s(s(?y:Nat)))",
"max(0,max(?y:Nat,s(s(0)))) = max(0,s(0))",
"max(s(?z:Nat),max(0,s(0))) = max(s(?x:Nat),?z:Nat)",
"max(s(max(s(?x:Nat),?x:Nat)),?y:Nat) = s(s(s(0)))",
"max(max(0,max(0,?y:Nat)),?x:Nat) = max(?y:Nat,s(?z:Nat))",
"s(max(?x:Nat,max(s(s(?y:Nat)),0))) = max(?y:Nat,?x:Nat)",
"s(max(?z:Nat,max(max(?x:Nat,?y:Nat),?y:Nat))) = max(?y:Nat,?x:Nat)",
"max(?x:Nat,max(?y:Nat,s(s(s(?y:Nat))))) = max(0,?y:Nat)",
"max(?x:Nat,max(?z:Nat,s(max(?z:Nat,0)))) = max(?y:Nat,?x:Nat)",
"max(?y:Nat,max(max(?z:Nat,0),s(?y:Nat))) = s(s(0))",
"max(?z:Nat,max(max(?x:Nat,?y:Nat),s(0))) = max(?x:Nat,0)",
"max(s(max(max(?y:Nat,?z:Nat),?y:Nat)),?y:Nat) = max(?x:Nat,?x:Nat)",
"max(max(?x:Nat,?z:Nat),max(?y:Nat,s(?x:Nat))) = max(0,?y:Nat)",
"max(max(?x:Nat,s(max(0,?y:Nat))),?x:Nat) = max(?x:Nat,?y:Nat)",
"max(max(?x:Nat,max(?y:Nat,?y:Nat)),s(0)) = max(?y:Nat,?x:Nat)",
"max(max(?x:Nat,max(?y:Nat,s(?x:Nat))),0) = max(?y:Nat,?x:Nat)",
"max(max(?y:Nat,s(0)),max(?x:Nat,0)) = max(?y:Nat,?x:Nat)",
"max(max(?y:Nat,s(s(?y:Nat))),s(?y:Nat)) = max(?z:Nat,?y:Nat)",
"max(max(0,s(max(?y:Nat,?z:Nat))),?y:Nat) = max(?z:Nat,?x:Nat)",
"max(max(s(?y:Nat),max(0,?x:Nat)),?z:Nat) = max(?x:Nat,?x:Nat)",
"max(max(s(?z:Nat),?x:Nat),max(?x:Nat,?x:Nat)) = max(?x:Nat,?x:Nat)",
"max(?x:Nat,s(s(max(max(?z:Nat,?z:Nat),0)))) = s(0)",
"max(?x:Nat,s(max(max(?z:Nat,s(?y:Nat)),?y:Nat))) = s(?y:Nat)",
"max(?x:Nat,max(s(max(?x:Nat,?y:Nat)),s(?x:Nat))) = s(?y:Nat)",
"max(s(?y:Nat),max(max(?x:Nat,s(0)),?x:Nat)) = s(?y:Nat)",
"max(s(s(s(?x:Nat))),max(?y:Nat,s(?y:Nat))) = s(0)",
"max(s(s(s(max(0,0)))),s(?z:Nat)) = s(?z:Nat)",
"max(s(max(?y:Nat,?y:Nat)),s(s(s(?y:Nat)))) = s(?x:Nat)",
"max(s(max(?z:Nat,0)),max(?z:Nat,s(?x:Nat))) = s(?z:Nat)",
"max(max(?y:Nat,max(s(?y:Nat),s(?x:Nat))),?x:Nat) = s(?x:Nat)",
"max(max(?z:Nat,?y:Nat),max(?y:Nat,max(?x:Nat,?x:Nat))) = s(?x:Nat)",
"max(max(?z:Nat,s(s(max(?x:Nat,?y:Nat)))),?x:Nat) = s(?x:Nat)",
"max(max(0,?y:Nat),s(s(s(s(?y:Nat))))) = s(?x:Nat)",
"max(max(s(?x:Nat),0),s(s(s(?x:Nat)))) = s(?z:Nat)",
"max(max(max(?x:Nat,s(s(?y:Nat))),0),0) = s(?y:Nat)",
"s(?x:Nat) = max(0,s(max(max(s(?y:Nat),s(?z:Nat)),?x:Nat)))",
"s(?x:Nat) = max(max(max(s(?x:Nat),?y:Nat),0),max(?x:Nat,?x:Nat))",
"s(?y:Nat) = max(?y:Nat,max(s(max(max(?z:Nat,0),?z:Nat)),?z:Nat))",
"s(?y:Nat) = max(?z:Nat,max(s(?x:Nat),max(max(0,?y:Nat),?y:Nat)))",
"s(?z:Nat) = max(?y:Nat,max(max(s(?y:Nat),0),max(?x:Nat,?y:Nat)))",
"s(?z:Nat) = max(max(?y:Nat,max(?x:Nat,s(?z:Nat))),max(?y:Nat,?z:Nat))",
"s(0) = max(?y:Nat,s(max(max(?y:Nat,?x:Nat),max(?x:Nat,?y:Nat))))",
"s(0) = max(max(?x:Nat,max(?x:Nat,s(s(?y:Nat)))),s(?z:Nat))",
"s(s(?y:Nat)) = max(?z:Nat,max(?z:Nat,s(s(max(?x:Nat,0)))))",
"s(s(?y:Nat)) = max(0,s(max(max(?x:Nat,?x:Nat),s(?y:Nat))))",
"s(s(?z:Nat)) = max(max(?y:Nat,?x:Nat),max(?z:Nat,max(?x:Nat,?x:Nat)))",
"s(s(0)) = max(max(max(0,max(0,?z:Nat)),0),?z:Nat)",
"max(?x:Nat,?x:Nat) = max(s(s(max(?y:Nat,0))),max(0,0))",
"max(?x:Nat,?z:Nat) = s(max(s(max(max(?x:Nat,?z:Nat),?x:Nat)),?y:Nat))",
"max(?y:Nat,?x:Nat) = max(s(?y:Nat),s(max(max(?x:Nat,?x:Nat),?z:Nat)))",
"max(?y:Nat,?x:Nat) = max(max(?y:Nat,max(s(?x:Nat),s(?x:Nat))),?x:Nat)",
"max(?y:Nat,?y:Nat) = max(0,max(0,s(s(max(?y:Nat,0)))))",
"max(?y:Nat,?y:Nat) = max(s(max(max(?x:Nat,?z:Nat),s(0))),?y:Nat)",
"max(0,?y:Nat) = s(max(max(s(?z:Nat),max(?y:Nat,?y:Nat)),0))",
"s(max(?x:Nat,?z:Nat)) = max(max(?z:Nat,max(s(0),0)),?x:Nat)",
"max(?x:Nat,s(?x:Nat)) = max(0,max(max(s(?y:Nat),?x:Nat),?y:Nat))",
"max(s(?x:Nat),?x:Nat) = max(max(?y:Nat,s(s(s(?x:Nat)))),?z:Nat)",
"max(s(?z:Nat),?x:Nat) = s(max(?z:Nat,max(max(0,0),0)))",
"s(s(max(?y:Nat,0))) = max(max(?y:Nat,?y:Nat),max(0,?y:Nat))",
"max(?x:Nat,max(?x:Nat,0)) = max(s(s(?x:Nat)),max(?y:Nat,?x:Nat))",
"max(?y:Nat,s(s(0))) = s(max(s(max(?y:Nat,?z:Nat)),0))",
"max(?y:Nat,max(?x:Nat,?x:Nat)) = max(?z:Nat,max(?x:Nat,max(0,?y:Nat)))",
"max(0,s(s(0))) = max(max(?z:Nat,?y:Nat),max(?y:Nat,?y:Nat))",
"max(s(0),s(0)) = max(s(?y:Nat),max(s(?x:Nat),?y:Nat))",
"s(max(max(?x:Nat,?y:Nat),0)) = max(max(?z:Nat,s(0)),?y:Nat)",
"max(?z:Nat,max(s(?y:Nat),?z:Nat)) = max(?y:Nat,max(s(?z:Nat),?y:Nat))",
"max(s(?z:Nat),s(s(?x:Nat))) = max(s(max(?y:Nat,?x:Nat)),?z:Nat)",
"max(max(?y:Nat,s(?z:Nat)),?x:Nat) = max(s(s(?z:Nat)),s(?y:Nat))",
"max(max(s(?x:Nat),?z:Nat),0) = max(0,max(s(0),?x:Nat))",
"max(max(s(0),?y:Nat),?x:Nat) = max(?y:Nat,max(0,s(0)))",
"s(max(s(max(0,?x:Nat)),0)) = max(max(?y:Nat,0),?x:Nat)",
"max(?x:Nat,max(max(?y:Nat,?z:Nat),0)) = max(s(0),s(?x:Nat))",
"max(s(?x:Nat),s(max(?y:Nat,?x:Nat))) = max(max(0,?y:Nat),?x:Nat)",
"max(s(s(?x:Nat)),max(0,0)) = max(max(0,?x:Nat),0)",
"max(s(s(max(?x:Nat,?y:Nat))),?y:Nat) = max(?y:Nat,max(?x:Nat,?x:Nat))",
"max(max(max(?x:Nat,?x:Nat),?y:Nat),?x:Nat) = max(s(s(?z:Nat)),?y:Nat)",
"max(max(max(?z:Nat,?y:Nat),?y:Nat),?y:Nat) = max(0,s(s(?x:Nat)))",
"s(s(max(s(max(?x:Nat,?x:Nat)),?x:Nat))) = max(?x:Nat,s(?y:Nat))",
"s(max(s(max(?x:Nat,?z:Nat)),s(?z:Nat))) = max(s(?y:Nat),?y:Nat)",
"s(max(max(max(?y:Nat,?x:Nat),?y:Nat),?y:Nat)) = max(s(?y:Nat),?x:Nat)",
"max(s(max(?y:Nat,s(?y:Nat))),s(?y:Nat)) = max(s(?y:Nat),?y:Nat)",
"max(s(max(max(?z:Nat,?x:Nat),?x:Nat)),?x:Nat) = max(?x:Nat,s(?x:Nat))",
"max(max(?y:Nat,s(?y:Nat)),max(?x:Nat,0)) = s(max(?y:Nat,?z:Nat))",
"max(max(?z:Nat,?x:Nat),max(?z:Nat,s(0))) = s(max(?z:Nat,?z:Nat))",
"max(max(s(?x:Nat),max(?y:Nat,?x:Nat)),?z:Nat) = max(s(?x:Nat),?x:Nat)",
"max(max(s(max(?x:Nat,?z:Nat)),?y:Nat),?z:Nat) = s(s(s(?z:Nat)))",
"s(max(max(max(0,s(?x:Nat)),?y:Nat),?x:Nat)) = max(0,?x:Nat)",
"max(?x:Nat,s(max(max(?y:Nat,0),s(?z:Nat)))) = max(?x:Nat,?y:Nat)",
"max(?x:Nat,max(?x:Nat,max(?x:Nat,max(?y:Nat,?x:Nat)))) = max(?x:Nat,?x:Nat)",
"max(s(s(max(?y:Nat,max(?x:Nat,?x:Nat)))),?y:Nat) = max(?x:Nat,?y:Nat)",
"max(s(max(?x:Nat,max(?x:Nat,?y:Nat))),s(?x:Nat)) = s(s(?x:Nat))",
"max(s(max(s(?y:Nat),?y:Nat)),max(?x:Nat,?x:Nat)) = s(s(?z:Nat))",
"max(max(?x:Nat,s(max(?z:Nat,?x:Nat))),s(?x:Nat)) = s(s(?x:Nat))",
"max(max(?y:Nat,max(max(0,0),0)),?z:Nat) = max(?y:Nat,?x:Nat)",
"max(max(s(?x:Nat),?x:Nat),max(0,s(?x:Nat))) = max(?y:Nat,?x:Nat)",
"max(max(s(max(s(?y:Nat),?x:Nat)),?y:Nat),?x:Nat) = max(?y:Nat,?x:Nat)",
"max(max(max(?y:Nat,?x:Nat),max(0,?y:Nat)),0) = s(s(?y:Nat))",
"max(max(max(0,0),max(?z:Nat,0)),?x:Nat) = max(?y:Nat,?z:Nat)",
"max(?x:Nat,s(max(s(s(?y:Nat)),max(?x:Nat,0)))) = s(?x:Nat)",
"max(?x:Nat,s(max(max(?y:Nat,?x:Nat),max(?y:Nat,?y:Nat)))) = s(?y:Nat)",
"max(?x:Nat,s(max(max(s(?x:Nat),s(?y:Nat)),?y:Nat))) = s(?y:Nat)",
"max(?x:Nat,max(max(?x:Nat,?x:Nat),max(?z:Nat,s(?x:Nat)))) = s(?x:Nat)",
"max(?y:Nat,s(max(0,max(max(?x:Nat,?z:Nat),?x:Nat)))) = s(?x:Nat)",
"max(?y:Nat,max(max(0,?z:Nat),max(s(?x:Nat),?y:Nat))) = s(?x:Nat)",
"max(?y:Nat,max(max(s(max(?z:Nat,0)),?y:Nat),?y:Nat)) = s(?x:Nat)",
"max(?z:Nat,s(max(max(0,max(?x:Nat,?x:Nat)),?x:Nat))) = s(0)",
"max(s(?x:Nat),max(max(s(?z:Nat),?y:Nat),s(?x:Nat))) = s(?x:Nat)",
"max(s(?y:Nat),max(?y:Nat,s(max(s(?y:Nat),?z:Nat)))) = s(?y:Nat)",
"max(max(?z:Nat,s(?y:Nat)),max(max(?y:Nat,?x:Nat),?y:Nat)) = s(?z:Nat)",
"max(max(0,max(s(max(?y:Nat,?x:Nat)),0)),?y:Nat) = s(?x:Nat)",
"max(max(s(?x:Nat),?y:Nat),max(max(?y:Nat,?y:Nat),?x:Nat)) = s(?y:Nat)",
"max(max(s(max(?y:Nat,max(?y:Nat,0))),?x:Nat),?y:Nat) = s(?z:Nat)",
"s(?y:Nat) = max(?x:Nat,max(max(0,s(s(0))),s(s(?y:Nat))))",
"s(?y:Nat) = max(?x:Nat,max(max(s(s(?z:Nat)),?x:Nat),max(?x:Nat,?x:Nat)))",
"s(?y:Nat) = max(s(max(?x:Nat,s(s(?z:Nat)))),s(max(?y:Nat,?z:Nat)))",
"s(?y:Nat) = max(max(0,s(max(?x:Nat,s(?z:Nat)))),max(?x:Nat,?y:Nat))",
"s(?y:Nat) = max(max(s(?x:Nat),max(max(0,s(0)),?y:Nat)),?x:Nat)",
"s(?z:Nat) = max(s(s(?z:Nat)),s(max(s(max(?x:Nat,?z:Nat)),?x:Nat)))",
"s(?z:Nat) = max(max(?x:Nat,max(s(?y:Nat),s(max(?x:Nat,0)))),0)",
"s(?z:Nat) = max(max(?y:Nat,s(max(?y:Nat,max(?y:Nat,0)))),s(?y:Nat))",
"s(0) = max(max(max(?x:Nat,max(0,?y:Nat)),?x:Nat),s(s(?y:Nat)))",
"max(?x:Nat,?z:Nat) = max(?x:Nat,max(s(0),max(max(?y:Nat,?x:Nat),0)))",
"max(?x:Nat,?z:Nat) = max(max(?z:Nat,s(s(s(?x:Nat)))),max(?y:Nat,?y:Nat))",
"max(?x:Nat,0) = max(?y:Nat,max(max(?y:Nat,0),max(0,s(?z:Nat))))",
"max(?y:Nat,?x:Nat) = max(max(?x:Nat,?x:Nat),s(s(max(s(?x:Nat),?x:Nat))))",
"max(?y:Nat,?y:Nat) = s(max(s(max(?x:Nat,s(?x:Nat))),max(?y:Nat,?y:Nat)))",
"max(?y:Nat,?y:Nat) = max(s(max(max(s(?y:Nat),?x:Nat),?z:Nat)),s(?y:Nat))",
"max(?y:Nat,?y:Nat) = max(max(0,s(max(?x:Nat,max(?z:Nat,?x:Nat)))),?x:Nat)",
"max(?y:Nat,?z:Nat) = max(?x:Nat,max(max(0,s(max(?x:Nat,?x:Nat))),?x:Nat))",
"max(?z:Nat,?y:Nat) = max(max(max(?z:Nat,max(?z:Nat,?z:Nat)),s(?z:Nat)),?y:Nat)",
"max(0,?y:Nat) = max(max(?x:Nat,0),s(s(max(s(?z:Nat),?y:Nat))))",
"max(0,?z:Nat) = max(max(s(max(s(0),?x:Nat)),?x:Nat),s(?y:Nat))",
"max(0,0) = max(s(?z:Nat),max(s(max(s(?y:Nat),?y:Nat)),?x:Nat))",
"s(max(?z:Nat,?x:Nat)) = max(max(max(?y:Nat,?z:Nat),?y:Nat),max(?y:Nat,?x:Nat))",
"max(?x:Nat,s(?x:Nat)) = max(?x:Nat,max(max(max(?y:Nat,?z:Nat),?x:Nat),?z:Nat))",
"max(0,s(?x:Nat)) = max(max(s(max(?y:Nat,s(?x:Nat))),?x:Nat),?x:Nat)",
"max(s(?y:Nat),?z:Nat) = max(max(?x:Nat,?x:Nat),s(s(max(0,?x:Nat))))",
"max(s(?z:Nat),?x:Nat) = max(max(s(?x:Nat),max(s(?x:Nat),?x:Nat)),?y:Nat)",
"max(s(0),?x:Nat) = max(s(max(?z:Nat,s(max(?x:Nat,0)))),?z:Nat)",
"s(s(max(?z:Nat,?x:Nat))) = max(max(s(max(?x:Nat,?x:Nat)),?x:Nat),?z:Nat)",
"max(?x:Nat,max(?y:Nat,?x:Nat)) = max(max(0,?z:Nat),max(?x:Nat,s(?x:Nat)))",
"max(?z:Nat,max(?y:Nat,?y:Nat)) = max(s(max(?x:Nat,s(s(?x:Nat)))),?z:Nat)",
"max(0,max(?y:Nat,?x:Nat)) = max(?y:Nat,max(0,s(s(s(?x:Nat)))))",
"max(s(?z:Nat),s(?z:Nat)) = s(s(max(s(s(?x:Nat)),s(?y:Nat))))",
"max(s(s(?x:Nat)),?y:Nat) = max(?z:Nat,max(s(s(?y:Nat)),s(?y:Nat)))",
"max(max(0,?x:Nat),?y:Nat) = max(s(max(?x:Nat,?x:Nat)),max(?x:Nat,?y:Nat))",
"s(max(max(?y:Nat,?x:Nat),?y:Nat)) = max(?y:Nat,max(?z:Nat,max(?z:Nat,?x:Nat)))",
"max(?z:Nat,max(s(?x:Nat),?y:Nat)) = max(?z:Nat,max(?x:Nat,max(?z:Nat,?z:Nat)))",
"max(0,max(?y:Nat,s(?x:Nat))) = s(s(max(max(?y:Nat,?x:Nat),0)))",
"max(0,max(0,s(?y:Nat))) = max(max(max(?y:Nat,?y:Nat),0),?y:Nat)",
"max(max(s(?x:Nat),?x:Nat),?y:Nat) = max(?x:Nat,max(?y:Nat,s(s(?x:Nat))))",
"max(s(max(s(0),?x:Nat)),?y:Nat) = max(?y:Nat,s(max(?x:Nat,?z:Nat)))",
"max(max(?z:Nat,s(s(?z:Nat))),?z:Nat) = max(s(0),max(?x:Nat,?x:Nat))",
"max(max(s(s(?y:Nat)),?x:Nat),?x:Nat) = max(?y:Nat,s(max(?y:Nat,?y:Nat)))",
"max(max(max(?z:Nat,0),0),?x:Nat) = max(s(?y:Nat),max(?x:Nat,?z:Nat))",
"s(s(max(?y:Nat,max(s(?x:Nat),0)))) = max(s(s(?x:Nat)),?x:Nat)",
"max(?x:Nat,max(?y:Nat,max(s(?z:Nat),?x:Nat))) = max(?y:Nat,max(?y:Nat,?y:Nat))",
"max(?y:Nat,s(max(0,max(?x:Nat,?x:Nat)))) = max(s(s(?y:Nat)),?x:Nat)",
"max(s(s(max(?x:Nat,s(0)))),?x:Nat) = max(max(0,?y:Nat),0)",
"max(s(max(?z:Nat,0)),s(s(?x:Nat))) = max(s(?x:Nat),s(?x:Nat))",
"max(max(s(s(?x:Nat)),s(?z:Nat)),?y:Nat) = max(s(?x:Nat),s(?z:Nat))",
"s(max(s(s(s(0))),max(0,?y:Nat))) = max(?z:Nat,s(0))",
"max(0,s(s(max(max(?x:Nat,?z:Nat),0)))) = s(s(s(?x:Nat)))",
"s(s(max(?x:Nat,max(max(?z:Nat,?x:Nat),s(?x:Nat))))) = max(?x:Nat,?x:Nat)",
"s(max(max(?x:Nat,max(s(?x:Nat),?z:Nat)),s(?z:Nat))) = max(?x:Nat,0)",
"max(?x:Nat,max(?z:Nat,max(0,max(?x:Nat,s(?z:Nat))))) = max(0,?y:Nat)",
"max(0,s(max(?y:Nat,max(?x:Nat,max(?z:Nat,?y:Nat))))) = max(?y:Nat,?z:Nat)",
"max(s(?x:Nat),s(max(max(0,s(?y:Nat)),?x:Nat))) = max(?z:Nat,?x:Nat)",
"max(s(?x:Nat),max(?y:Nat,max(s(s(?x:Nat)),?y:Nat))) = max(?z:Nat,?y:Nat)",
"max(s(s(s(?y:Nat))),max(max(?x:Nat,?y:Nat),?y:Nat)) = max(?x:Nat,?y:Nat)",
"max(s(max(?y:Nat,max(?x:Nat,max(?x:Nat,?z:Nat)))),?y:Nat) = s(s(?x:Nat))",
"max(max(?x:Nat,s(max(max(?y:Nat,?y:Nat),?z:Nat))),?y:Nat) = max(?y:Nat,?y:Nat)",
"max(max(?z:Nat,max(s(s(?x:Nat)),?x:Nat)),s(0)) = s(s(0))",
"max(max(s(?x:Nat),s(max(?y:Nat,s(?y:Nat)))),0) = max(?z:Nat,?x:Nat)",
"max(max(s(?y:Nat),?y:Nat),max(max(?z:Nat,?y:Nat),?y:Nat)) = max(?x:Nat,?x:Nat)",
"max(max(s(s(?z:Nat)),max(?x:Nat,0)),s(?y:Nat)) = max(?y:Nat,?x:Nat)",
"max(max(max(?y:Nat,?y:Nat),s(?y:Nat)),s(s(?x:Nat))) = max(0,?y:Nat)",
"max(?x:Nat,max(?z:Nat,max(max(s(?z:Nat),?y:Nat),s(?y:Nat)))) = s(?x:Nat)",
"max(?z:Nat,max(max(?y:Nat,max(?y:Nat,max(?y:Nat,?y:Nat))),?x:Nat)) = s(0)",
"max(?z:Nat,max(max(?y:Nat,max(max(0,?y:Nat),0)),?y:Nat)) = s(?y:Nat)",
"max(s(?y:Nat),s(max(s(max(0,?y:Nat)),s(?y:Nat)))) = s(?y:Nat)",
"max(max(?y:Nat,?x:Nat),max(?x:Nat,s(s(max(?y:Nat,?x:Nat))))) = s(?y:Nat)",
"max(max(0,s(max(0,max(?z:Nat,0)))),s(?y:Nat)) = s(?z:Nat)",
"max(max(s(s(?y:Nat)),?z:Nat),s(max(?x:Nat,s(?z:Nat)))) = s(0)",
"max(max(s(s(?y:Nat)),s(s(0))),max(?x:Nat,?y:Nat)) = s(?y:Nat)",
"max(max(max(?z:Nat,s(?y:Nat)),?x:Nat),max(?y:Nat,s(?x:Nat))) = s(?y:Nat)",
"s(?x:Nat) = max(?z:Nat,max(max(?y:Nat,?z:Nat),s(max(?z:Nat,max(?x:Nat,0)))))",
"s(?x:Nat) = max(s(s(?y:Nat)),max(?x:Nat,max(s(?x:Nat),max(0,?x:Nat))))",
"s(?x:Nat) = max(max(s(0),max(0,0)),max(?x:Nat,s(s(?y:Nat))))",
"s(0) = max(s(max(max(max(?y:Nat,?y:Nat),max(?z:Nat,?y:Nat)),?y:Nat)),?y:Nat)",
"s(0) = max(max(?y:Nat,?x:Nat),s(max(?x:Nat,max(0,max(?y:Nat,?x:Nat)))))",
"s(s(?z:Nat)) = max(s(?y:Nat),max(?y:Nat,max(s(max(?x:Nat,0)),?x:Nat)))",
"max(?x:Nat,?y:Nat) = max(max(max(max(0,max(?z:Nat,0)),?x:Nat),?x:Nat),?z:Nat)",
"max(?z:Nat,?z:Nat) = max(max(s(?x:Nat),?y:Nat),s(s(s(max(?x:Nat,?y:Nat)))))",
"max(0,?y:Nat) = max(max(?x:Nat,?x:Nat),max(max(?z:Nat,s(?x:Nat)),s(?y:Nat)))",
"max(?x:Nat,s(?y:Nat)) = s(max(s(max(s(?x:Nat),max(?x:Nat,?x:Nat))),?x:Nat))",
"max(s(?x:Nat),?x:Nat) = s(s(max(max(s(?y:Nat),max(0,?z:Nat)),0)))",
"max(s(?x:Nat),?x:Nat) = max(?y:Nat,s(max(s(?y:Nat),s(max(0,?y:Nat)))))",
"s(s(max(?x:Nat,?z:Nat))) = max(?y:Nat,s(max(?y:Nat,max(?x:Nat,s(?x:Nat)))))",
"s(max(?x:Nat,s(?x:Nat))) = max(max(?x:Nat,max(?x:Nat,max(?x:Nat,?z:Nat))),0)",
"max(?z:Nat,max(?z:Nat,0)) = max(?x:Nat,s(max(s(max(0,?x:Nat)),?y:Nat)))",
"max(s(s(?x:Nat)),?z:Nat) = max(0,max(?x:Nat,max(max(?y:Nat,?y:Nat),?x:Nat)))",
"max(s(s(?y:Nat)),?z:Nat) = max(s(?y:Nat),max(s(s(s(?x:Nat))),?x:Nat))",
"max(max(?x:Nat,?x:Nat),?x:Nat) = max(?x:Nat,max(max(?y:Nat,0),max(?y:Nat,?x:Nat)))",
"max(max(?y:Nat,?y:Nat),?y:Nat) = max(s(s(s(?z:Nat))),max(0,s(?x:Nat)))",
"max(max(0,?x:Nat),?y:Nat) = s(s(max(max(max(?y:Nat,0),?y:Nat),?y:Nat)))",
"max(max(0,?x:Nat),0) = s(max(max(s(?x:Nat),?y:Nat),max(?x:Nat,?x:Nat)))",
"s(max(?x:Nat,max(?z:Nat,?x:Nat))) = max(?y:Nat,s(max(?z:Nat,max(0,?x:Nat))))",
"max(s(?y:Nat),max(?y:Nat,?y:Nat)) = max(max(s(0),?y:Nat),max(?y:Nat,?x:Nat))",
"max(?x:Nat,s(max(0,s(?z:Nat)))) = max(0,s(max(?z:Nat,s(?z:Nat))))",
"max(0,s(s(s(max(?y:Nat,?x:Nat))))) = max(max(0,s(?x:Nat)),?z:Nat)",
"max(s(s(?x:Nat)),s(max(?z:Nat,?x:Nat))) = max(max(?x:Nat,?y:Nat),s(0))",
"max(max(s(s(0)),?z:Nat),s(0)) = max(max(0,0),s(?y:Nat))",
"max(max(max(?y:Nat,?y:Nat),s(?y:Nat)),?y:Nat) = max(?y:Nat,s(max(?x:Nat,?x:Nat)))",
"max(max(max(?z:Nat,0),?x:Nat),s(?x:Nat)) = s(s(max(?y:Nat,s(0))))",
"s(s(max(?x:Nat,max(?y:Nat,max(?y:Nat,?x:Nat))))) = max(?y:Nat,s(s(?y:Nat)))",
"max(?x:Nat,max(?x:Nat,s(max(s(?y:Nat),?z:Nat)))) = max(?x:Nat,max(0,?y:Nat))",
"max(?y:Nat,max(s(max(0,?x:Nat)),s(?y:Nat))) = max(max(?z:Nat,0),?y:Nat)",
"max(s(?x:Nat),s(s(max(s(?z:Nat),?y:Nat)))) = max(max(?x:Nat,?y:Nat),?x:Nat)",
"max(s(?y:Nat),max(s(s(?y:Nat)),s(?y:Nat))) = s(max(?x:Nat,s(?x:Nat)))",
"max(s(s(?x:Nat)),max(max(0,?y:Nat),?y:Nat)) = s(max(s(?x:Nat),0))",
"max(max(0,max(?z:Nat,max(?y:Nat,?x:Nat))),?x:Nat) = max(max(?x:Nat,?y:Nat),?x:Nat)",
"max(?x:Nat,max(?y:Nat,max(?z:Nat,max(s(?x:Nat),?y:Nat)))) = s(max(?y:Nat,?z:Nat))",
"max(?x:Nat,max(?y:Nat,max(s(?y:Nat),s(s(?x:Nat))))) = s(max(0,0))",
"max(s(?x:Nat),s(s(max(max(0,0),?x:Nat)))) = max(s(?y:Nat),?y:Nat)",
"max(s(?x:Nat),max(s(?y:Nat),s(max(?x:Nat,?y:Nat)))) = s(max(?z:Nat,?x:Nat))",
"max(max(?x:Nat,max(?y:Nat,s(?x:Nat))),max(?y:Nat,?y:Nat)) = max(s(?y:Nat),?y:Nat)",
"max(max(s(max(?y:Nat,?x:Nat)),max(?y:Nat,?x:Nat)),?z:Nat) = s(max(?x:Nat,?y:Nat))",
"s(max(s(?x:Nat),max(max(?y:Nat,?y:Nat),s(s(?y:Nat))))) = max(?x:Nat,?y:Nat)",
"s(max(s(?x:Nat),max(max(s(?x:Nat),s(?x:Nat)),?y:Nat))) = max(?x:Nat,?y:Nat)",
"max(?y:Nat,max(s(?z:Nat),max(?y:Nat,s(max(?x:Nat,?y:Nat))))) = max(?z:Nat,0)",
"max(?y:Nat,max(max(?x:Nat,max(?x:Nat,s(?z:Nat))),s(?z:Nat))) = max(?x:Nat,?z:Nat)",
"max(?y:Nat,max(max(?z:Nat,max(?z:Nat,?z:Nat)),max(?y:Nat,?z:Nat))) = max(?z:Nat,?x:Nat)",
"max(s(?x:Nat),s(max(?x:Nat,max(?z:Nat,max(?x:Nat,?z:Nat))))) = max(0,?y:Nat)",
"max(s(?y:Nat),max(?z:Nat,max(?x:Nat,max(?z:Nat,s(?y:Nat))))) = max(?y:Nat,?y:Nat)",
"max(max(?x:Nat,?x:Nat),max(?y:Nat,s(s(max(?x:Nat,?y:Nat))))) = s(s(?z:Nat))",
"max(max(max(?z:Nat,s(?x:Nat)),max(s(?x:Nat),?z:Nat)),?x:Nat) = s(s(?z:Nat))",
"max(max(max(s(?y:Nat),max(0,s(?z:Nat))),?z:Nat),0) = s(s(0))",
"max(?y:Nat,max(s(?x:Nat),max(max(s(?z:Nat),?y:Nat),s(?y:Nat)))) = s(?y:Nat)",
"max(s(?y:Nat),s(s(max(max(?z:Nat,?y:Nat),max(?y:Nat,?y:Nat))))) = s(?y:Nat)",
"max(s(max(max(max(?x:Nat,?y:Nat),max(?y:Nat,?y:Nat)),?x:Nat)),0) = s(?x:Nat)",
"max(max(?x:Nat,max(s(0),?z:Nat)),max(?y:Nat,max(?x:Nat,?x:Nat))) = s(?z:Nat)",
"max(max(?z:Nat,max(max(s(?x:Nat),max(?z:Nat,?y:Nat)),?z:Nat)),0) = s(?y:Nat)",
"max(max(s(?z:Nat),?y:Nat),max(s(?z:Nat),s(max(?y:Nat,?z:Nat)))) = s(?y:Nat)",
"max(max(max(s(max(?y:Nat,0)),max(?y:Nat,0)),?x:Nat),?x:Nat) = s(?y:Nat)",
"s(?y:Nat) = max(s(max(max(max(?x:Nat,?y:Nat),max(?z:Nat,0)),?x:Nat)),s(?x:Nat))",
"s(?z:Nat) = max(?x:Nat,max(?z:Nat,max(max(0,max(0,?y:Nat)),s(s(?x:Nat)))))",
"s(0) = max(s(max(?x:Nat,s(?y:Nat))),max(s(s(max(0,?x:Nat))),0))",
"s(0) = max(max(?y:Nat,max(max(?x:Nat,?y:Nat),max(?y:Nat,?y:Nat))),s(s(?y:Nat)))",
"s(s(?x:Nat)) = max(max(?z:Nat,s(0)),max(max(0,max(?y:Nat,0)),?y:Nat))",
"s(s(?y:Nat)) = max(?x:Nat,s(max(s(max(?y:Nat,?x:Nat)),s(max(?z:Nat,?x:Nat)))))",
"s(s(?y:Nat)) = max(0,max(s(s(?y:Nat)),max(s(?z:Nat),max(?x:Nat,?z:Nat))))",
"s(s(?y:Nat)) = max(0,max(max(max(?x:Nat,max(?x:Nat,?y:Nat)),?y:Nat),s(0)))",
"max(?x:Nat,?x:Nat) = max(max(max(?y:Nat,s(?z:Nat)),?y:Nat),max(s(s(?y:Nat)),?y:Nat))",
"max(?y:Nat,?y:Nat) = s(max(max(max(s(?x:Nat),max(?x:Nat,0)),s(?x:Nat)),?y:Nat))",
"max(0,?x:Nat) = max(s(?x:Nat),s(max(max(max(?z:Nat,?z:Nat),?y:Nat),s(?y:Nat))))",
"max(0,?z:Nat) = s(max(s(max(?z:Nat,?x:Nat)),s(max(max(?z:Nat,?x:Nat),?z:Nat))))",
"s(max(?x:Nat,?y:Nat)) = max(s(max(s(0),max(?y:Nat,max(?x:Nat,?z:Nat)))),?x:Nat)",
"max(s(?y:Nat),?z:Nat) = max(s(?y:Nat),max(s(max(s(0),s(?z:Nat))),?y:Nat))",
"max(s(0),?x:Nat) = max(max(?y:Nat,?z:Nat),max(?x:Nat,s(max(?y:Nat,s(?x:Nat)))))",
"max(max(?x:Nat,?x:Nat),?x:Nat) = max(max(0,s(?x:Nat)),max(max(?y:Nat,?y:Nat),?y:Nat))",
"max(max(?y:Nat,0),?x:Nat) = max(s(0),s(max(s(max(?y:Nat,0)),?z:Nat)))",
"s(s(max(?y:Nat,s(?y:Nat)))) = max(?x:Nat,max(?x:Nat,s(max(s(?y:Nat),?y:Nat))))",
"s(max(max(?y:Nat,?y:Nat),?y:Nat)) = max(max(max(max(?x:Nat,?x:Nat),0),0),0)",
"max(?y:Nat,s(max(?y:Nat,?x:Nat))) = max(max(?x:Nat,max(?x:Nat,?y:Nat)),max(?x:Nat,?x:Nat))",
"max(s(?x:Nat),max(?y:Nat,?x:Nat)) = max(0,s(s(max(?x:Nat,max(?y:Nat,?y:Nat)))))",
"max(s(max(?z:Nat,?z:Nat)),0) = max(s(s(0)),max(max(?x:Nat,?x:Nat),0))",
"s(max(s(s(s(?y:Nat))),?z:Nat)) = max(max(max(?z:Nat,?x:Nat),?z:Nat),s(?x:Nat))",
"s(max(s(?z:Nat),max(?x:Nat,max(?y:Nat,?z:Nat)))) = max(max(0,0),s(?x:Nat))",
"max(s(?x:Nat),max(s(max(?z:Nat,?y:Nat)),?z:Nat)) = max(s(s(?z:Nat)),s(?z:Nat))",
"max(s(0),max(s(?y:Nat),s(s(?y:Nat)))) = max(s(?z:Nat),max(?y:Nat,?x:Nat))",
"max(max(max(?y:Nat,s(s(?y:Nat))),?y:Nat),0) = max(s(max(?y:Nat,?y:Nat)),?z:Nat)",
"s(max(s(max(?y:Nat,max(?y:Nat,?x:Nat))),s(?x:Nat))) = max(?y:Nat,max(?x:Nat,?y:Nat))",
"max(?x:Nat,max(max(max(?z:Nat,s(0)),?y:Nat),?x:Nat)) = max(0,s(s(?y:Nat)))",
"max(max(max(s(max(0,?z:Nat)),?y:Nat),?z:Nat),?y:Nat) = s(s(s(s(?y:Nat))))",
"max(?y:Nat,s(max(max(0,?y:Nat),max(?z:Nat,s(?x:Nat))))) = max(s(?y:Nat),0)",
"max(s(max(max(s(0),max(?z:Nat,?x:Nat)),?x:Nat)),0) = s(max(?x:Nat,?y:Nat))",
"max(max(s(s(?z:Nat)),s(max(0,s(?x:Nat)))),?x:Nat) = s(max(?y:Nat,?z:Nat))",
"max(max(s(s(s(?x:Nat))),0),max(?y:Nat,s(?x:Nat))) = s(max(?z:Nat,?z:Nat))",
"max(?x:Nat,max(max(?x:Nat,?y:Nat),max(s(?x:Nat),max(0,?y:Nat)))) = max(?y:Nat,?x:Nat)",
"max(0,max(s(max(?y:Nat,s(?x:Nat))),max(s(?x:Nat),0))) = max(?y:Nat,?x:Nat)",
"max(s(?y:Nat),max(s(0),max(?y:Nat,max(s(?z:Nat),0)))) = max(?y:Nat,?x:Nat)",
"max(s(max(s(?y:Nat),max(?y:Nat,max(?x:Nat,?y:Nat)))),s(?y:Nat)) = max(0,0)",
"max(max(s(?y:Nat),?y:Nat),s(s(max(max(?x:Nat,?x:Nat),0)))) = max(?y:Nat,?y:Nat)",
"max(s(max(?x:Nat,?x:Nat)),max(?z:Nat,s(max(0,max(?y:Nat,0))))) = s(?y:Nat)",
"max(max(?y:Nat,s(?y:Nat)),max(max(max(?x:Nat,s(?z:Nat)),0),?x:Nat)) = s(?z:Nat)",
"max(max(max(max(?y:Nat,max(?z:Nat,?x:Nat)),?y:Nat),s(s(?y:Nat))),?y:Nat) = s(0)",
"s(?x:Nat) = max(max(max(max(?x:Nat,?x:Nat),s(max(?x:Nat,?x:Nat))),?z:Nat),max(?y:Nat,?x:Nat))",
"s(?y:Nat) = max(max(max(?x:Nat,?y:Nat),s(s(?z:Nat))),max(max(s(0),?x:Nat),?x:Nat))",
"s(?z:Nat) = max(s(max(s(0),max(s(0),max(?x:Nat,?x:Nat)))),max(?z:Nat,?z:Nat))",
"s(0) = max(max(s(?z:Nat),?x:Nat),max(s(?z:Nat),max(0,s(max(0,?z:Nat)))))",
"max(?z:Nat,?z:Nat) = s(s(max(max(?x:Nat,max(?x:Nat,0)),max(s(?x:Nat),s(?x:Nat)))))",
"s(max(?x:Nat,?y:Nat)) = max(?y:Nat,s(max(max(max(0,?x:Nat),max(?x:Nat,?x:Nat)),?z:Nat)))",
"s(max(?z:Nat,s(?z:Nat))) = max(max(?y:Nat,max(max(?z:Nat,?z:Nat),s(?y:Nat))),s(?x:Nat))",
"max(s(max(?y:Nat,?x:Nat)),?y:Nat) = s(max(max(max(?y:Nat,s(?z:Nat)),s(?z:Nat)),?x:Nat))",
"max(s(max(0,?x:Nat)),?y:Nat) = max(max(max(?x:Nat,max(s(0),?x:Nat)),?z:Nat),?y:Nat)",
"max(max(?x:Nat,?z:Nat),s(?x:Nat)) = max(max(s(s(max(?x:Nat,?y:Nat))),s(?z:Nat)),?x:Nat)",
"s(s(s(max(s(?x:Nat),?x:Nat)))) = max(max(s(0),max(0,s(?y:Nat))),?x:Nat)",
"max(0,s(max(s(?z:Nat),?z:Nat))) = max(max(s(?x:Nat),s(0)),max(?x:Nat,?x:Nat))",
"max(s(s(max(0,?x:Nat))),?x:Nat) = max(max(max(?y:Nat,?x:Nat),0),max(?y:Nat,?z:Nat))",
"max(max(?x:Nat,?y:Nat),max(?x:Nat,?y:Nat)) = max(s(max(max(s(?x:Nat),0),?x:Nat)),?x:Nat)",
"max(max(0,?x:Nat),max(?x:Nat,?x:Nat)) = max(?y:Nat,max(max(s(?y:Nat),?x:Nat),s(?z:Nat)))",
"max(max(s(?z:Nat),s(?z:Nat)),0) = s(max(s(max(?y:Nat,max(0,0))),?y:Nat))",
"max(?x:Nat,max(s(?y:Nat),max(?x:Nat,?x:Nat))) = s(max(max(s(?y:Nat),s(0)),?z:Nat))",
"max(s(s(s(max(?x:Nat,0)))),?x:Nat) = max(max(max(?y:Nat,s(0)),0),?z:Nat)",
"max(max(max(s(?x:Nat),0),?y:Nat),?y:Nat) = max(?y:Nat,max(max(?x:Nat,?z:Nat),s(?y:Nat)))",
"max(?x:Nat,s(max(?z:Nat,max(?y:Nat,s(?y:Nat))))) = max(max(0,?y:Nat),s(s(?z:Nat)))",
"max(s(max(s(?x:Nat),s(s(?y:Nat)))),?y:Nat) = max(s(max(?x:Nat,s(0))),?x:Nat)",
"s(max(0,max(0,max(?x:Nat,max(?x:Nat,?y:Nat))))) = max(s(?z:Nat),max(?y:Nat,?z:Nat))",
"s(max(s(?y:Nat),max(max(0,?x:Nat),s(?x:Nat)))) = max(max(?x:Nat,s(?x:Nat)),?y:Nat)",
"max(max(s(max(s(0),?z:Nat)),?y:Nat),s(?z:Nat)) = max(s(s(?z:Nat)),s(?y:Nat))",
"max(?z:Nat,max(max(?y:Nat,max(?y:Nat,s(?y:Nat))),s(?x:Nat))) = s(max(s(0),?x:Nat))",
"max(s(0),max(s(max(max(?y:Nat,?y:Nat),0)),?z:Nat)) = max(max(?z:Nat,?y:Nat),?x:Nat)",
"max(max(?x:Nat,?x:Nat),max(max(max(?x:Nat,?y:Nat),?z:Nat),0)) = max(max(?x:Nat,0),?z:Nat)",
"max(max(?x:Nat,s(?z:Nat)),s(s(max(max(?x:Nat,0),?x:Nat)))) = s(max(?z:Nat,?x:Nat))",
"max(max(max(0,s(?x:Nat)),s(max(0,s(?x:Nat)))),?y:Nat) = max(s(?y:Nat),?y:Nat)",
"s(max(max(max(0,s(?x:Nat)),max(s(?x:Nat),?z:Nat)),s(?z:Nat))) = max(?x:Nat,?y:Nat)",
"s(max(max(max(max(?x:Nat,?z:Nat),max(?y:Nat,?y:Nat)),s(0)),?z:Nat)) = max(?x:Nat,?x:Nat)",
"max(s(max(max(?z:Nat,?x:Nat),max(max(?z:Nat,?z:Nat),?x:Nat))),s(?x:Nat)) = max(0,0)",
"max(max(?x:Nat,max(max(max(0,?z:Nat),s(?x:Nat)),s(?z:Nat))),0) = max(?y:Nat,?x:Nat)",
"max(max(s(?y:Nat),s(?z:Nat)),s(max(s(max(?x:Nat,0)),0))) = max(?y:Nat,?z:Nat)",
"max(max(max(?y:Nat,s(0)),max(max(?z:Nat,?x:Nat),s(?y:Nat))),?z:Nat) = max(?y:Nat,?y:Nat)",
"max(s(?z:Nat),max(s(max(?x:Nat,?x:Nat)),max(s(?x:Nat),max(?y:Nat,?y:Nat)))) = s(?x:Nat)",
"max(max(?x:Nat,?y:Nat),max(max(s(max(0,?y:Nat)),?x:Nat),max(?x:Nat,?y:Nat))) = s(0)",
"max(max(max(?y:Nat,s(max(?x:Nat,?y:Nat))),s(max(?x:Nat,s(?y:Nat)))),?z:Nat) = s(?y:Nat)",
"max(max(max(max(?x:Nat,?y:Nat),?y:Nat),max(?y:Nat,s(max(?x:Nat,?x:Nat)))),?x:Nat) = s(?y:Nat)",
"max(max(max(max(?z:Nat,max(0,?y:Nat)),?z:Nat),max(0,s(?z:Nat))),?x:Nat) = s(?z:Nat)",
"max(max(max(max(0,s(?z:Nat)),s(?y:Nat)),s(?x:Nat)),max(?z:Nat,?x:Nat)) = s(?y:Nat)",
"s(?x:Nat) = max(max(max(?x:Nat,max(max(?y:Nat,0),?z:Nat)),max(s(0),?x:Nat)),s(0))",
"s(?y:Nat) = max(max(?y:Nat,max(max(?y:Nat,s(?x:Nat)),max(max(?z:Nat,?x:Nat),s(?y:Nat)))),?y:Nat)",
"s(?y:Nat) = max(max(s(s(max(?y:Nat,0))),s(?x:Nat)),max(max(?y:Nat,s(?x:Nat)),0))",
"s(?y:Nat) = max(max(max(max(s(?x:Nat),0),max(?y:Nat,?y:Nat)),s(max(?x:Nat,?x:Nat))),?x:Nat)",
"s(s(?y:Nat)) = max(max(max(?x:Nat,max(max(?x:Nat,?z:Nat),max(?x:Nat,?x:Nat))),s(?z:Nat)),?y:Nat)",
"s(s(?z:Nat)) = max(max(?x:Nat,max(?x:Nat,?y:Nat)),max(0,max(s(?z:Nat),s(s(?y:Nat)))))",
"max(0,?z:Nat) = max(max(max(max(max(?x:Nat,?z:Nat),?y:Nat),?x:Nat),s(s(s(?z:Nat)))),?z:Nat)",
"s(max(s(?z:Nat),?z:Nat)) = max(max(max(max(?x:Nat,s(0)),s(?x:Nat)),s(?x:Nat)),?y:Nat)",
"max(?y:Nat,s(s(?y:Nat))) = max(max(max(s(s(?y:Nat)),s(s(?x:Nat))),s(?y:Nat)),?y:Nat)",
"max(max(0,?y:Nat),0) = s(max(max(?y:Nat,s(?x:Nat)),max(max(s(?y:Nat),?x:Nat),?y:Nat)))",
"max(max(s(?y:Nat),?x:Nat),?y:Nat) = max(s(0),s(s(max(s(0),max(?y:Nat,?y:Nat)))))",
"s(s(max(s(?z:Nat),s(?x:Nat)))) = max(max(?x:Nat,s(max(s(?x:Nat),s(?y:Nat)))),?y:Nat)",
"s(s(max(s(0),s(s(0))))) = max(s(s(max(?z:Nat,?z:Nat))),max(?x:Nat,?y:Nat))",
"s(max(0,max(s(s(0)),?x:Nat))) = max(max(?z:Nat,max(?x:Nat,s(s(?x:Nat)))),?z:Nat)",
"s(max(max(?z:Nat,s(?z:Nat)),s(?y:Nat))) = max(?y:Nat,max(max(s(?x:Nat),s(?y:Nat)),?z:Nat))",
"max(?y:Nat,max(?x:Nat,s(max(?y:Nat,?y:Nat)))) = max(?y:Nat,s(max(max(?z:Nat,?y:Nat),s(?y:Nat))))",
"max(0,s(max(0,s(s(?y:Nat))))) = max(s(max(?x:Nat,max(?x:Nat,?x:Nat))),s(?z:Nat))",
"max(0,max(s(max(?z:Nat,?x:Nat)),?y:Nat)) = max(?y:Nat,max(?x:Nat,max(0,s(s(?y:Nat)))))",
"max(max(max(0,0),0),s(?z:Nat)) = max(?x:Nat,max(?x:Nat,s(s(max(?x:Nat,?y:Nat)))))",
"max(?z:Nat,s(s(max(?y:Nat,max(0,?z:Nat))))) = max(s(?z:Nat),s(max(?y:Nat,s(0))))",
"max(0,s(max(max(?z:Nat,s(?z:Nat)),?y:Nat))) = max(max(?x:Nat,max(?y:Nat,s(?z:Nat))),0)",
"max(max(?x:Nat,max(?z:Nat,max(?x:Nat,?x:Nat))),?y:Nat) = max(max(max(0,0),s(?x:Nat)),?z:Nat)",
"max(max(?y:Nat,?x:Nat),max(?y:Nat,max(?z:Nat,?y:Nat))) = max(?x:Nat,s(max(?y:Nat,max(?y:Nat,?y:Nat))))",
"max(max(?y:Nat,s(s(max(?x:Nat,?z:Nat)))),?x:Nat) = s(max(max(?x:Nat,s(s(?x:Nat))),0))",
"max(max(?z:Nat,s(max(?z:Nat,s(?x:Nat)))),0) = s(max(max(?z:Nat,?y:Nat),max(?y:Nat,?x:Nat)))",
"s(max(?x:Nat,max(max(?y:Nat,?y:Nat),s(s(?y:Nat))))) = max(max(max(?x:Nat,?x:Nat),0),0)",
"s(max(max(s(?y:Nat),s(max(?x:Nat,0))),?x:Nat)) = max(max(0,max(0,?x:Nat)),?x:Nat)",
"max(?y:Nat,s(max(?x:Nat,max(max(0,?x:Nat),?y:Nat)))) = max(s(s(max(?y:Nat,0))),?y:Nat)",
"max(s(max(?x:Nat,max(?y:Nat,?x:Nat))),max(?x:Nat,0)) = max(?x:Nat,max(max(?x:Nat,?y:Nat),?z:Nat))",
"max(max(0,max(?y:Nat,?x:Nat)),max(?x:Nat,s(?y:Nat))) = s(max(?x:Nat,s(max(0,?x:Nat))))",
"s(max(max(s(max(?x:Nat,0)),?x:Nat),max(?z:Nat,?x:Nat))) = max(?y:Nat,max(s(?z:Nat),?x:Nat))",
"max(s(s(?x:Nat)),s(max(max(?x:Nat,s(?x:Nat)),?z:Nat))) = max(s(?x:Nat),max(?y:Nat,?z:Nat))",
"max(max(max(max(?y:Nat,?y:Nat),?x:Nat),s(?z:Nat)),s(?y:Nat)) = s(max(?y:Nat,max(?z:Nat,?x:Nat)))",
"max(s(max(max(?z:Nat,?y:Nat),max(s(?x:Nat),s(?x:Nat)))),?x:Nat) = s(s(max(?x:Nat,0)))",
"max(s(max(max(max(0,?y:Nat),max(?z:Nat,?z:Nat)),0)),?z:Nat) = max(max(?y:Nat,0),?y:Nat)",
"max(max(s(?x:Nat),s(max(?z:Nat,?y:Nat))),max(?z:Nat,s(?y:Nat))) = max(?y:Nat,max(?z:Nat,0))",
"max(max(s(?y:Nat),max(s(s(?x:Nat)),?x:Nat)),s(s(?z:Nat))) = max(?y:Nat,max(?x:Nat,?x:Nat))",
"max(max(max(max(s(?x:Nat),s(0)),?z:Nat),0),s(?x:Nat)) = max(0,max(?y:Nat,?x:Nat))",
"max(?x:Nat,max(max(s(?z:Nat),max(s(?x:Nat),s(?x:Nat))),s(?z:Nat))) = s(max(?y:Nat,?y:Nat))",
"max(s(s(0)),max(s(max(0,?z:Nat)),max(s(0),0))) = s(max(0,?y:Nat))",
"max(s(max(?z:Nat,?z:Nat)),max(max(s(?z:Nat),?z:Nat),max(?x:Nat,?x:Nat))) = max(s(?x:Nat),?z:Nat)",
"s(max(0,max(max(max(?x:Nat,?x:Nat),max(?y:Nat,?x:Nat)),max(?x:Nat,0)))) = max(0,?z:Nat)",
"max(?y:Nat,max(max(max(max(0,?y:Nat),s(?x:Nat)),0),s(s(?x:Nat)))) = s(s(?z:Nat))",
"max(?z:Nat,max(max(?y:Nat,s(s(0))),max(s(max(?x:Nat,?y:Nat)),?y:Nat))) = max(?x:Nat,?x:Nat)",
"max(s(s(?x:Nat)),max(?y:Nat,s(max(max(?z:Nat,?y:Nat),max(?y:Nat,?y:Nat))))) = max(?x:Nat,?y:Nat)",
"max(s(max(?x:Nat,?y:Nat)),s(max(s(s(?y:Nat)),s(max(?y:Nat,?y:Nat))))) = max(?y:Nat,?x:Nat)",
"max(?y:Nat,max(s(?x:Nat),max(max(max(?y:Nat,?y:Nat),?y:Nat),s(max(?y:Nat,?x:Nat))))) = s(0)",
"max(s(?x:Nat),max(max(?x:Nat,?x:Nat),max(max(?z:Nat,?x:Nat),s(max(0,?y:Nat))))) = s(0)",
"max(s(max(s(?x:Nat),max(?y:Nat,?y:Nat))),max(max(s(?y:Nat),?z:Nat),s(?x:Nat))) = s(?x:Nat)",
"max(?x:Nat,?x:Nat) = max(?x:Nat,s(max(max(?x:Nat,s(?x:Nat)),max(max(0,?y:Nat),max(?y:Nat,?x:Nat)))))",
"s(max(?x:Nat,?y:Nat)) = max(?x:Nat,max(s(max(?x:Nat,?x:Nat)),max(max(?z:Nat,0),s(s(?x:Nat)))))",
"max(0,max(?x:Nat,?x:Nat)) = max(0,max(max(?y:Nat,max(max(?x:Nat,?x:Nat),max(?x:Nat,0))),0))",
"max(s(s(?y:Nat)),?x:Nat) = max(?y:Nat,max(?y:Nat,max(s(s(?y:Nat)),max(?y:Nat,max(?y:Nat,?y:Nat)))))",
"max(?x:Nat,s(s(s(?y:Nat)))) = max(s(s(0)),max(?x:Nat,s(max(max(0,?x:Nat),?x:Nat))))",
"max(?y:Nat,s(max(?y:Nat,?z:Nat))) = s(max(0,max(max(?y:Nat,?x:Nat),max(?y:Nat,max(?y:Nat,?y:Nat)))))",
"max(s(max(?x:Nat,?x:Nat)),?y:Nat) = max(max(s(max(?y:Nat,?x:Nat)),s(?y:Nat)),max(s(?y:Nat),?x:Nat))",
"max(s(max(?x:Nat,?x:Nat)),0) = s(max(max(?y:Nat,?z:Nat),max(?x:Nat,max(?z:Nat,max(?x:Nat,?y:Nat)))))",
"s(max(?x:Nat,s(max(s(?y:Nat),?z:Nat)))) = max(max(s(0),?x:Nat),s(s(max(?x:Nat,?x:Nat))))",
"max(?z:Nat,max(max(s(s(?z:Nat)),?x:Nat),?y:Nat)) = max(s(max(?y:Nat,?x:Nat)),s(max(?x:Nat,0)))",
"s(max(max(?y:Nat,0),max(?x:Nat,max(?z:Nat,?z:Nat)))) = max(?x:Nat,max(s(?y:Nat),max(0,?x:Nat)))",
"s(s(max(max(max(?z:Nat,?x:Nat),?y:Nat),max(?x:Nat,?x:Nat)))) = max(max(max(?y:Nat,?y:Nat),0),?z:Nat)",
"max(?y:Nat,max(max(?x:Nat,?y:Nat),s(max(s(?y:Nat),s(?z:Nat))))) = s(max(0,max(0,?y:Nat)))",
"max(?x:Nat,max(max(max(max(?z:Nat,?y:Nat),?y:Nat),?y:Nat),s(max(?y:Nat,0)))) = s(max(?y:Nat,?y:Nat))",
"max(?z:Nat,max(max(?z:Nat,?x:Nat),max(max(s(?y:Nat),?y:Nat),max(s(?y:Nat),?z:Nat)))) = max(?y:Nat,?x:Nat)",
"max(max(?y:Nat,max(?y:Nat,s(max(0,0)))),max(?y:Nat,max(s(?x:Nat),?x:Nat))) = s(s(?x:Nat))",
"max(max(s(?x:Nat),max(?x:Nat,max(s(?x:Nat),max(?y:Nat,?z:Nat)))),s(s(0))) = max(0,?y:Nat)",
"max(max(max(s(0),s(max(?y:Nat,?x:Nat))),max(s(s(?y:Nat)),?y:Nat)),?z:Nat) = max(?z:Nat,?y:Nat)",
"max(max(s(?z:Nat),max(?y:Nat,?z:Nat)),max(max(max(?x:Nat,?y:Nat),max(?y:Nat,?x:Nat)),?z:Nat)) = s(?y:Nat)",
"max(max(s(max(?z:Nat,max(0,?y:Nat))),s(max(0,max(0,?y:Nat)))),s(?y:Nat)) = s(?x:Nat)",
"max(?x:Nat,0) = max(max(s(max(?y:Nat,max(?y:Nat,?x:Nat))),max(max(max(?z:Nat,?x:Nat),?y:Nat),?z:Nat)),?y:Nat)",
"s(max(?z:Nat,?x:Nat)) = max(max(?y:Nat,s(max(max(?y:Nat,?y:Nat),?y:Nat))),max(s(max(?z:Nat,?y:Nat)),?z:Nat))",
"max(?y:Nat,max(?y:Nat,s(0))) = max(?x:Nat,max(s(s(?y:Nat)),max(max(max(?x:Nat,?y:Nat),?y:Nat),?x:Nat)))",
"max(s(s(?y:Nat)),s(?x:Nat)) = max(?x:Nat,max(max(max(s(?z:Nat),0),?x:Nat),max(s(0),?x:Nat)))",
"max(max(?x:Nat,0),s(?y:Nat)) = max(max(?x:Nat,?z:Nat),max(0,s(max(max(?y:Nat,?y:Nat),s(?x:Nat)))))",
"max(?x:Nat,max(s(max(?y:Nat,0)),?z:Nat)) = max(?z:Nat,max(?x:Nat,max(max(?x:Nat,max(?x:Nat,0)),0)))",
"s(max(0,max(max(?x:Nat,?x:Nat),s(0)))) = max(max(s(?x:Nat),s(s(s(?y:Nat)))),s(?y:Nat))",
"max(?y:Nat,max(max(s(max(?y:Nat,?y:Nat)),?x:Nat),?z:Nat)) = max(s(max(?y:Nat,max(s(?x:Nat),0))),?x:Nat)",
"max(?z:Nat,max(s(max(s(?z:Nat),s(?z:Nat))),?z:Nat)) = max(max(0,max(max(?x:Nat,?z:Nat),?x:Nat)),0)",
"max(s(?x:Nat),max(?y:Nat,s(max(s(?y:Nat),?y:Nat)))) = s(max(?y:Nat,s(max(max(?y:Nat,?z:Nat),?x:Nat))))",
"s(s(max(max(max(?y:Nat,?y:Nat),?x:Nat),s(s(?z:Nat))))) = max(max(s(max(?x:Nat,?y:Nat)),?y:Nat),?z:Nat)",
"max(s(max(max(?y:Nat,max(?x:Nat,?y:Nat)),?y:Nat)),s(?z:Nat)) = max(?z:Nat,s(s(max(s(?x:Nat),0))))",
"max(max(?z:Nat,max(s(?x:Nat),s(max(?x:Nat,?z:Nat)))),0) = max(?y:Nat,s(s(max(s(?y:Nat),0))))",
"max(max(max(max(?x:Nat,?y:Nat),max(0,0)),?x:Nat),0) = max(max(s(?y:Nat),max(?x:Nat,?x:Nat)),?x:Nat)",
"max(?x:Nat,max(max(?x:Nat,s(s(0))),s(max(?z:Nat,?x:Nat)))) = max(s(?x:Nat),max(?z:Nat,s(0)))",
"max(s(s(s(0))),s(max(max(?x:Nat,?y:Nat),s(?z:Nat)))) = max(max(?y:Nat,?x:Nat),max(?z:Nat,0))",
"max(max(?y:Nat,s(max(?y:Nat,s(?x:Nat)))),s(max(?y:Nat,max(?x:Nat,s(?y:Nat))))) = s(max(?y:Nat,?x:Nat))",
"max(s(s(s(0))),max(max(max(?y:Nat,?x:Nat),s(max(?y:Nat,?x:Nat))),s(?x:Nat))) = max(?y:Nat,?x:Nat)",
"max(max(s(max(max(?x:Nat,?x:Nat),0)),max(s(?x:Nat),max(?y:Nat,max(?y:Nat,?x:Nat)))),?x:Nat) = s(?y:Nat)",
"max(max(max(s(?y:Nat),s(?z:Nat)),max(s(?x:Nat),0)),max(max(?z:Nat,s(?y:Nat)),?y:Nat)) = s(?y:Nat)",
"s(?x:Nat) = max(max(s(?y:Nat),?z:Nat),max(?y:Nat,max(max(?y:Nat,max(?y:Nat,0)),max(max(?x:Nat,?y:Nat),?y:Nat))))",
"max(0,?y:Nat) = max(max(0,?y:Nat),max(max(max(s(?y:Nat),?x:Nat),s(?x:Nat)),s(max(s(?y:Nat),0))))",
"max(s(0),max(s(s(?x:Nat)),0)) = s(max(?y:Nat,max(max(max(?z:Nat,?z:Nat),s(?y:Nat)),s(?x:Nat))))",
"s(s(s(max(max(0,?y:Nat),s(?x:Nat))))) = max(max(s(max(?y:Nat,max(0,0))),?y:Nat),s(?y:Nat))",
"max(max(?z:Nat,0),max(max(?z:Nat,s(s(?y:Nat))),s(?x:Nat))) = max(s(max(?x:Nat,s(s(?z:Nat)))),?x:Nat)",
"max(0,s(max(max(s(0),?y:Nat),max(max(?x:Nat,?y:Nat),s(?z:Nat))))) = max(max(s(?z:Nat),0),?y:Nat)",
"max(?y:Nat,max(s(max(?y:Nat,max(?x:Nat,?z:Nat))),max(max(s(?y:Nat),0),?y:Nat))) = max(?y:Nat,max(?y:Nat,?z:Nat))",
"max(max(0,max(max(max(?x:Nat,?z:Nat),?y:Nat),s(s(?y:Nat)))),s(s(0))) = max(0,max(?y:Nat,?z:Nat))",
"s(max(max(?y:Nat,max(?x:Nat,0)),max(max(?x:Nat,0),max(s(?z:Nat),max(?y:Nat,?y:Nat))))) = max(?x:Nat,?x:Nat)",
"max(max(max(?y:Nat,max(max(?x:Nat,?z:Nat),max(?x:Nat,?z:Nat))),max(max(?x:Nat,?y:Nat),?x:Nat)),?y:Nat) = max(?x:Nat,0)",
"max(s(max(max(s(0),?z:Nat),max(0,s(?x:Nat)))),s(max(max(?x:Nat,s(?x:Nat)),?y:Nat))) = s(?y:Nat)",
"s(max(s(?z:Nat),max(?y:Nat,0))) = max(s(?y:Nat),max(max(max(s(?x:Nat),?y:Nat),s(0)),max(?y:Nat,?z:Nat)))",
"max(max(?y:Nat,max(s(?z:Nat),?y:Nat)),?z:Nat) = max(max(max(?x:Nat,?y:Nat),s(?x:Nat)),max(0,max(?z:Nat,s(?x:Nat))))",
"max(?y:Nat,max(?y:Nat,max(s(max(?x:Nat,?z:Nat)),?x:Nat))) = max(s(s(s(?y:Nat))),s(max(0,max(?x:Nat,?x:Nat))))",
"max(max(?x:Nat,max(0,max(0,0))),max(?z:Nat,?y:Nat)) = max(?y:Nat,max(s(?x:Nat),max(max(0,?z:Nat),?y:Nat)))",
"max(max(max(?x:Nat,s(?y:Nat)),s(?x:Nat)),max(s(?x:Nat),s(?x:Nat))) = max(?y:Nat,max(?y:Nat,max(s(?y:Nat),?y:Nat)))",
"s(max(max(max(max(?x:Nat,?x:Nat),?y:Nat),max(max(?x:Nat,0),?z:Nat)),?y:Nat)) = max(?x:Nat,max(max(?y:Nat,0),?x:Nat))",
"max(?x:Nat,max(max(s(?x:Nat),?z:Nat),max(max(max(0,?y:Nat),s(?z:Nat)),s(?z:Nat)))) = max(max(?x:Nat,?z:Nat),?x:Nat)",
"max(max(max(max(s(?x:Nat),?z:Nat),max(0,0)),max(s(s(?y:Nat)),?x:Nat)),?z:Nat) = s(s(max(?z:Nat,?x:Nat)))",
"max(max(max(max(s(?y:Nat),?x:Nat),s(s(?y:Nat))),?z:Nat),s(s(max(max(?x:Nat,0),?x:Nat)))) = max(?z:Nat,?x:Nat)",
"max(?y:Nat,max(?z:Nat,max(max(?y:Nat,0),?x:Nat))) = max(max(max(s(max(?y:Nat,?y:Nat)),max(?x:Nat,?x:Nat)),?y:Nat),s(?x:Nat))",
"max(s(max(max(?x:Nat,?z:Nat),max(0,?z:Nat))),?z:Nat) = max(?x:Nat,max(max(s(?z:Nat),max(s(?x:Nat),s(?z:Nat))),?y:Nat))",
"max(0,max(max(s(s(?y:Nat)),max(?z:Nat,?x:Nat)),0)) = s(max(max(s(max(?x:Nat,?x:Nat)),max(?y:Nat,?y:Nat)),?y:Nat))",
"s(s(max(s(max(?y:Nat,?y:Nat)),s(?y:Nat)))) = max(?x:Nat,s(max(max(?x:Nat,?x:Nat),max(max(?y:Nat,?x:Nat),max(?y:Nat,?z:Nat)))))",
"max(max(?x:Nat,0),s(s(max(?x:Nat,?z:Nat)))) = max(max(max(s(s(?x:Nat)),?z:Nat),?x:Nat),s(s(max(?y:Nat,s(?y:Nat)))))",
"s(max(?z:Nat,max(s(max(?z:Nat,?y:Nat)),max(s(?z:Nat),s(?z:Nat))))) = max(?y:Nat,max(?y:Nat,max(max(0,?x:Nat),s(0))))",
"max(max(?x:Nat,?x:Nat),max(max(max(max(?x:Nat,0),?x:Nat),0),max(?z:Nat,s(?y:Nat)))) = s(s(s(max(?x:Nat,s(?z:Nat)))))",
"max(max(max(max(?z:Nat,0),?z:Nat),max(?x:Nat,max(s(?z:Nat),?y:Nat))),max(?y:Nat,?x:Nat)) = max(0,max(?x:Nat,max(0,?x:Nat)))",
"max(max(max(0,?z:Nat),?y:Nat),max(max(s(?z:Nat),max(0,max(?x:Nat,?z:Nat))),max(0,s(?y:Nat)))) = s(max(?y:Nat,?x:Nat))",
"max(max(max(?y:Nat,0),s(s(s(0)))),0) = max(max(max(?x:Nat,?z:Nat),?x:Nat),max(max(s(max(?z:Nat,?y:Nat)),0),0))",
"max(max(max(?y:Nat,?x:Nat),max(?x:Nat,max(?z:Nat,max(?x:Nat,?x:Nat)))),max(s(s(s(?y:Nat))),max(?x:Nat,max(?y:Nat,?z:Nat)))) = s(?y:Nat)",
"max(s(?y:Nat),s(?x:Nat)) = max(max(max(max(s(?x:Nat),s(?y:Nat)),?z:Nat),?z:Nat),max(s(max(?y:Nat,?x:Nat)),max(0,max(?y:Nat,?y:Nat))))",
"max(s(s(max(s(?x:Nat),max(?y:Nat,?x:Nat)))),?y:Nat) = max(max(?x:Nat,s(?x:Nat)),max(max(max(?y:Nat,0),max(?x:Nat,?y:Nat)),s(?x:Nat)))",
"max(0,s(max(s(s(?x:Nat)),max(s(?x:Nat),?x:Nat)))) = max(?y:Nat,max(max(s(?y:Nat),max(max(?y:Nat,?x:Nat),s(?y:Nat))),s(?z:Nat)))",
"max(max(max(max(?x:Nat,?x:Nat),?x:Nat),s(max(?x:Nat,?z:Nat))),?y:Nat) = max(max(max(?z:Nat,s(s(0))),?z:Nat),s(s(s(s(?x:Nat)))))",
"s(max(0,s(max(?x:Nat,max(?y:Nat,?x:Nat))))) = max(max(max(0,?y:Nat),?x:Nat),s(max(max(?x:Nat,s(?y:Nat)),max(max(?y:Nat,?y:Nat),?x:Nat))))",
"max(s(?y:Nat),max(max(max(max(?x:Nat,?y:Nat),max(?y:Nat,?y:Nat)),?x:Nat),?x:Nat)) = max(s(max(max(s(?y:Nat),s(?y:Nat)),max(?x:Nat,?y:Nat))),0)",
"max(max(max(s(max(?y:Nat,?x:Nat)),?y:Nat),s(max(?x:Nat,max(0,?x:Nat)))),?z:Nat) = max(max(0,max(?y:Nat,max(s(?x:Nat),s(?x:Nat)))),?z:Nat)",
"max(max(max(s(s(?y:Nat)),s(max(?x:Nat,?x:Nat))),s(max(max(?z:Nat,?z:Nat),s(?y:Nat)))),?y:Nat) = s(max(max(max(?y:Nat,0),0),0))",
"max(max(max(max(?y:Nat,?z:Nat),max(s(?x:Nat),?y:Nat)),?y:Nat),s(s(s(?x:Nat)))) = max(max(?y:Nat,?z:Nat),max(?x:Nat,max(?x:Nat,max(s(?x:Nat),s(0)))))",
"max(max(s(max(max(?x:Nat,?x:Nat),0)),s(?y:Nat)),max(max(s(max(0,?x:Nat)),?y:Nat),max(?x:Nat,0))) = max(max(s(max(0,?y:Nat)),?y:Nat),?z:Nat)",
"max(max(s(max(?x:Nat,0)),0),s(?x:Nat)) = max(s(max(max(?y:Nat,s(?x:Nat)),s(max(?y:Nat,?z:Nat)))),max(?x:Nat,max(s(max(?z:Nat,?y:Nat)),s(?x:Nat))))",
"max(max(max(0,?x:Nat),s(s(s(?x:Nat)))),s(max(?y:Nat,s(?z:Nat)))) = max(?x:Nat,max(max(s(?x:Nat),max(?y:Nat,?y:Nat)),max(max(s(?y:Nat),?y:Nat),?x:Nat)))"]

val problemsMinNo =
IOMstrs.rdEqs [
"s(?x:Nat) = min(?x:Nat,?x:Nat)",
"s(?x:Nat) = min(?x:Nat,?z:Nat)",
"s(?x:Nat) = min(?x:Nat,0)",
"s(?x:Nat) = min(?y:Nat,?x:Nat)",
"s(?x:Nat) = min(?y:Nat,?y:Nat)",
"s(?x:Nat) = min(?y:Nat,?z:Nat)",
"s(?x:Nat) = min(0,?x:Nat)",
"s(?x:Nat) = min(0,?y:Nat)",
"s(?y:Nat) = min(?x:Nat,0)",
"s(?y:Nat) = min(0,0)",
"s(0) = min(?x:Nat,?x:Nat)",
"s(0) = min(?x:Nat,?y:Nat)",
"s(0) = min(?x:Nat,0)",
"s(0) = min(0,?x:Nat)",
"s(?x:Nat) = min(?x:Nat,s(?x:Nat))",
"s(?x:Nat) = min(?y:Nat,s(?x:Nat))",
"s(?x:Nat) = min(?z:Nat,s(?z:Nat))",
"s(?x:Nat) = min(?z:Nat,s(0))",
"s(?x:Nat) = min(0,s(?y:Nat))",
"s(?x:Nat) = min(s(?x:Nat),?x:Nat)",
"s(?x:Nat) = min(s(?x:Nat),?y:Nat)",
"s(?x:Nat) = min(s(?x:Nat),0)",
"s(?x:Nat) = min(s(?y:Nat),?x:Nat)",
"s(?x:Nat) = min(s(?y:Nat),?y:Nat)",
"s(?x:Nat) = min(s(0),?x:Nat)",
"s(?y:Nat) = min(?y:Nat,s(?x:Nat))",
"s(?y:Nat) = min(?z:Nat,s(?x:Nat))",
"s(?y:Nat) = min(0,s(?y:Nat))",
"s(0) = min(?x:Nat,s(?x:Nat))",
"s(0) = min(0,s(?x:Nat))",
"s(0) = min(s(?x:Nat),?x:Nat)",
"s(0) = min(s(0),?z:Nat)",
"s(s(?x:Nat)) = min(?y:Nat,?y:Nat)",
"s(s(?x:Nat)) = min(?y:Nat,?z:Nat)",
"s(s(?y:Nat)) = min(?x:Nat,?y:Nat)",
"s(s(?y:Nat)) = min(?y:Nat,?x:Nat)",
"s(s(?y:Nat)) = min(?z:Nat,0)",
"s(s(?y:Nat)) = min(0,?y:Nat)",
"s(s(?y:Nat)) = min(0,0)",
"s(s(?z:Nat)) = min(?z:Nat,?z:Nat)",
"s(s(?z:Nat)) = min(0,?y:Nat)",
"s(s(0)) = min(?x:Nat,?z:Nat)",
"min(?x:Nat,?x:Nat) = min(?x:Nat,?y:Nat)",
"min(?x:Nat,?x:Nat) = min(?y:Nat,?y:Nat)",
"min(?x:Nat,?x:Nat) = min(?y:Nat,0)",
"min(?x:Nat,?x:Nat) = min(0,?y:Nat)",
"min(?x:Nat,?y:Nat) = min(?x:Nat,?z:Nat)",
"min(?x:Nat,?y:Nat) = min(?x:Nat,0)",
"min(?x:Nat,?y:Nat) = min(?z:Nat,?x:Nat)",
"min(?x:Nat,?y:Nat) = min(?z:Nat,?y:Nat)",
"min(?x:Nat,?y:Nat) = min(0,?y:Nat)",
"min(?x:Nat,?z:Nat) = min(0,?x:Nat)",
"min(?x:Nat,0) = min(?y:Nat,?x:Nat)",
"min(?y:Nat,?x:Nat) = min(?x:Nat,?x:Nat)",
"min(?y:Nat,?x:Nat) = min(0,0)",
"min(?y:Nat,?y:Nat) = s(s(0))",
"min(?y:Nat,?y:Nat) = min(?y:Nat,0)",
"min(?y:Nat,?y:Nat) = min(0,?y:Nat)",
"min(?y:Nat,?z:Nat) = min(?x:Nat,0)",
"min(?y:Nat,0) = s(s(0))",
"min(?z:Nat,?y:Nat) = min(0,?x:Nat)",
"min(?z:Nat,?z:Nat) = min(?y:Nat,?x:Nat)",
"min(0,?x:Nat) = s(s(0))",
"min(0,0) = min(?y:Nat,?y:Nat)",
"min(?x:Nat,s(?y:Nat)) = s(0)",
"min(?x:Nat,s(0)) = s(?x:Nat)",
"min(s(?x:Nat),?y:Nat) = s(?z:Nat)",
"min(s(?y:Nat),?x:Nat) = s(0)",
"min(s(?y:Nat),0) = s(?z:Nat)",
"min(s(0),?x:Nat) = s(?y:Nat)",
"s(?x:Nat) = min(?y:Nat,s(s(?x:Nat)))",
"s(?x:Nat) = min(?y:Nat,min(?x:Nat,?x:Nat))",
"s(?x:Nat) = min(?y:Nat,min(?y:Nat,?z:Nat))",
"s(?x:Nat) = min(?y:Nat,min(0,?y:Nat))",
"s(?x:Nat) = min(s(?y:Nat),s(?y:Nat))",
"s(?x:Nat) = min(min(?y:Nat,?x:Nat),?z:Nat)",
"s(?x:Nat) = min(min(?y:Nat,?z:Nat),?y:Nat)",
"s(?x:Nat) = min(min(0,?x:Nat),?y:Nat)",
"s(?x:Nat) = min(min(0,?z:Nat),?z:Nat)",
"s(?y:Nat) = min(?x:Nat,s(s(0)))",
"s(?y:Nat) = min(?x:Nat,min(?z:Nat,?y:Nat))",
"s(?y:Nat) = min(?y:Nat,s(s(?x:Nat)))",
"s(?y:Nat) = min(?y:Nat,min(?y:Nat,?y:Nat))",
"s(?y:Nat) = min(?y:Nat,min(?z:Nat,0))",
"s(?y:Nat) = min(?z:Nat,min(?x:Nat,?x:Nat))",
"s(?y:Nat) = min(0,s(s(?y:Nat)))",
"s(?y:Nat) = min(0,s(s(0)))",
"s(?y:Nat) = min(s(s(0)),?x:Nat)",
"s(?y:Nat) = min(min(?y:Nat,?x:Nat),?y:Nat)",
"s(?y:Nat) = min(min(?y:Nat,?x:Nat),0)",
"s(?y:Nat) = min(min(?y:Nat,?y:Nat),?y:Nat)",
"s(?y:Nat) = min(min(?y:Nat,?z:Nat),?x:Nat)",
"s(?y:Nat) = min(min(?y:Nat,0),?y:Nat)",
"s(?z:Nat) = min(?x:Nat,s(s(?y:Nat)))",
"s(?z:Nat) = min(?z:Nat,s(s(?z:Nat)))",
"s(?z:Nat) = min(s(s(?x:Nat)),?z:Nat)",
"s(0) = min(?x:Nat,s(s(0)))",
"s(0) = min(?y:Nat,s(s(?x:Nat)))",
"s(0) = min(?y:Nat,min(?x:Nat,?x:Nat))",
"s(0) = min(?y:Nat,min(?x:Nat,?y:Nat))",
"s(0) = min(s(s(?y:Nat)),?x:Nat)",
"s(s(?x:Nat)) = min(?x:Nat,s(?y:Nat))",
"s(s(?x:Nat)) = min(?x:Nat,s(0))",
"s(s(?x:Nat)) = min(s(?x:Nat),?x:Nat)",
"s(s(?y:Nat)) = min(?x:Nat,s(?x:Nat))",
"s(s(?y:Nat)) = min(s(?x:Nat),0)",
"s(s(?y:Nat)) = min(s(?y:Nat),?z:Nat)",
"s(s(?y:Nat)) = min(s(?z:Nat),?x:Nat)",
"s(s(?z:Nat)) = min(?x:Nat,s(?y:Nat))",
"s(s(?z:Nat)) = min(s(?x:Nat),?x:Nat)",
"s(s(0)) = min(?y:Nat,s(?z:Nat))",
"s(s(0)) = min(s(?y:Nat),?x:Nat)",
"s(s(0)) = min(s(?z:Nat),?z:Nat)",
"min(?x:Nat,?x:Nat) = s(min(?y:Nat,?z:Nat))",
"min(?x:Nat,?x:Nat) = min(?x:Nat,s(0))",
"min(?x:Nat,?x:Nat) = min(?y:Nat,s(0))",
"min(?x:Nat,?x:Nat) = min(0,s(?y:Nat))",
"min(?x:Nat,?x:Nat) = min(s(?x:Nat),?z:Nat)",
"min(?x:Nat,?x:Nat) = min(s(0),0)",
"min(?x:Nat,?y:Nat) = min(?y:Nat,s(?x:Nat))",
"min(?x:Nat,?y:Nat) = min(?z:Nat,s(0))",
"min(?x:Nat,?y:Nat) = min(s(?y:Nat),0)",
"min(?x:Nat,?y:Nat) = min(s(?z:Nat),?y:Nat)",
"min(?x:Nat,?z:Nat) = min(?x:Nat,s(?x:Nat))",
"min(?x:Nat,0) = s(min(?y:Nat,?z:Nat))",
"min(?y:Nat,?x:Nat) = s(min(?x:Nat,?x:Nat))",
"min(?y:Nat,?x:Nat) = s(min(0,?y:Nat))",
"min(?y:Nat,?x:Nat) = min(s(?y:Nat),?x:Nat)",
"min(?y:Nat,?x:Nat) = min(s(0),?z:Nat)",
"min(?y:Nat,?y:Nat) = min(?x:Nat,s(?y:Nat))",
"min(?y:Nat,?y:Nat) = min(?y:Nat,s(?x:Nat))",
"min(?y:Nat,?y:Nat) = min(?z:Nat,s(?z:Nat))",
"min(?y:Nat,?z:Nat) = min(?y:Nat,s(?x:Nat))",
"min(?y:Nat,?z:Nat) = min(s(?y:Nat),?y:Nat)",
"min(?y:Nat,0) = s(s(s(?y:Nat)))",
"min(?y:Nat,0) = min(?x:Nat,s(?z:Nat))",
"min(?y:Nat,0) = min(?y:Nat,s(?x:Nat))",
"min(?z:Nat,?x:Nat) = s(min(?x:Nat,?y:Nat))",
"min(?z:Nat,?x:Nat) = s(min(?y:Nat,?x:Nat))",
"min(?z:Nat,?x:Nat) = min(?y:Nat,s(?z:Nat))",
"min(?z:Nat,?y:Nat) = s(min(?x:Nat,?x:Nat))",
"min(?z:Nat,?y:Nat) = min(?x:Nat,s(?x:Nat))",
"min(?z:Nat,?y:Nat) = min(s(?x:Nat),?x:Nat)",
"min(0,?x:Nat) = min(?x:Nat,s(?y:Nat))",
"min(0,?x:Nat) = min(s(?x:Nat),?x:Nat)",
"min(0,?y:Nat) = s(s(s(?z:Nat)))",
"s(s(s(?x:Nat))) = min(?y:Nat,?x:Nat)",
"s(s(s(?y:Nat))) = min(?y:Nat,?z:Nat)",
"s(s(s(?y:Nat))) = min(?z:Nat,?x:Nat)",
"s(s(s(0))) = min(?y:Nat,0)",
"s(min(?x:Nat,?x:Nat)) = min(?x:Nat,?x:Nat)",
"s(min(?x:Nat,?z:Nat)) = min(?x:Nat,?x:Nat)",
"s(min(?y:Nat,?x:Nat)) = min(?x:Nat,?x:Nat)",
"s(min(?y:Nat,?x:Nat)) = min(?y:Nat,?z:Nat)",
"s(min(?y:Nat,?x:Nat)) = min(0,?y:Nat)",
"s(min(?y:Nat,?z:Nat)) = min(?y:Nat,0)",
"s(min(?y:Nat,0)) = min(?x:Nat,0)",
"s(min(0,?x:Nat)) = min(?z:Nat,?y:Nat)",
"s(min(0,?y:Nat)) = min(?x:Nat,?y:Nat)",
"s(min(0,?y:Nat)) = min(?y:Nat,?y:Nat)",
"min(?y:Nat,s(?z:Nat)) = min(?x:Nat,?y:Nat)",
"min(0,s(?x:Nat)) = s(s(?y:Nat))",
"min(0,s(?y:Nat)) = min(?y:Nat,?y:Nat)",
"min(s(?x:Nat),?x:Nat) = min(?x:Nat,0)",
"min(s(?x:Nat),?x:Nat) = min(?z:Nat,0)",
"min(s(?x:Nat),?y:Nat) = min(?y:Nat,0)",
"min(s(?y:Nat),?x:Nat) = s(s(?x:Nat))",
"min(s(?y:Nat),?x:Nat) = min(?x:Nat,?x:Nat)",
"min(s(?y:Nat),?x:Nat) = min(?x:Nat,?y:Nat)",
"min(s(?y:Nat),?y:Nat) = min(?x:Nat,?x:Nat)",
"min(s(?y:Nat),?y:Nat) = min(?x:Nat,?y:Nat)",
"min(s(?y:Nat),0) = min(?y:Nat,?z:Nat)",
"min(s(?z:Nat),?z:Nat) = min(0,0)",
"min(s(0),?x:Nat) = s(s(?x:Nat))",
"min(s(0),?x:Nat) = min(?y:Nat,0)",
"min(s(0),0) = s(s(?x:Nat))",
"min(s(0),0) = min(?x:Nat,?z:Nat)",
"min(?x:Nat,min(?x:Nat,?y:Nat)) = s(?y:Nat)",
"min(?x:Nat,min(?y:Nat,?x:Nat)) = s(?z:Nat)",
"min(?x:Nat,min(?z:Nat,0)) = s(?y:Nat)",
"min(?y:Nat,s(s(?y:Nat))) = s(0)",
"min(?z:Nat,min(?x:Nat,?z:Nat)) = s(?x:Nat)",
"min(?z:Nat,min(?z:Nat,?y:Nat)) = s(0)",
"min(0,min(?z:Nat,?y:Nat)) = s(?x:Nat)",
"min(s(?x:Nat),s(?y:Nat)) = s(?x:Nat)",
"min(s(s(?x:Nat)),?x:Nat) = s(?y:Nat)",
"min(s(s(?x:Nat)),?y:Nat) = s(?z:Nat)",
"min(s(s(?z:Nat)),?x:Nat) = s(?z:Nat)",
"min(min(?x:Nat,?x:Nat),?x:Nat) = s(?y:Nat)",
"min(min(?x:Nat,?x:Nat),?z:Nat) = s(?x:Nat)",
"min(min(?x:Nat,?z:Nat),?y:Nat) = s(0)",
"min(min(?x:Nat,0),?x:Nat) = s(?y:Nat)",
"min(min(?y:Nat,?y:Nat),0) = s(?x:Nat)",
"min(min(?z:Nat,?y:Nat),?x:Nat) = s(?x:Nat)",
"min(min(?z:Nat,0),?x:Nat) = s(0)",
"min(min(0,?z:Nat),?y:Nat) = s(?x:Nat)",
"s(?x:Nat) = min(?x:Nat,s(min(?x:Nat,?x:Nat)))",
"s(?x:Nat) = min(?y:Nat,s(min(?x:Nat,?x:Nat)))",
"s(?x:Nat) = min(?y:Nat,min(s(0),?y:Nat))",
"s(?x:Nat) = min(0,min(?x:Nat,s(?x:Nat)))",
"s(?x:Nat) = min(0,min(s(0),?z:Nat))",
"s(?x:Nat) = min(s(?y:Nat),min(?x:Nat,?x:Nat))",
"s(?x:Nat) = min(s(min(?x:Nat,?x:Nat)),?y:Nat)",
"s(?x:Nat) = min(min(?x:Nat,s(?x:Nat)),?x:Nat)",
"s(?y:Nat) = min(?x:Nat,s(s(s(?y:Nat))))",
"s(?y:Nat) = min(?y:Nat,s(min(?x:Nat,?y:Nat)))",
"s(?y:Nat) = min(0,min(s(?y:Nat),?x:Nat))",
"s(?y:Nat) = min(s(0),min(?y:Nat,?y:Nat))",
"s(?y:Nat) = min(min(0,?x:Nat),s(?y:Nat))",
"s(?z:Nat) = min(s(?z:Nat),min(?y:Nat,?x:Nat))",
"s(0) = min(?y:Nat,min(s(?y:Nat),?y:Nat))",
"s(0) = min(0,s(min(?z:Nat,?x:Nat)))",
"s(0) = min(s(?z:Nat),min(?y:Nat,0))",
"s(0) = min(s(min(?y:Nat,?x:Nat)),?y:Nat)",
"s(0) = min(min(s(?z:Nat),?y:Nat),?y:Nat)",
"s(s(?x:Nat)) = min(?x:Nat,s(s(0)))",
"s(s(?x:Nat)) = min(s(?x:Nat),s(?z:Nat))",
"s(s(?y:Nat)) = min(0,min(?x:Nat,?y:Nat))",
"s(s(?y:Nat)) = min(s(0),s(?x:Nat))",
"s(s(?y:Nat)) = min(min(?y:Nat,?y:Nat),?z:Nat)",
"s(s(?y:Nat)) = min(min(0,?x:Nat),?y:Nat)",
"s(s(?z:Nat)) = min(?y:Nat,min(0,?x:Nat))",
"s(s(?z:Nat)) = min(0,s(s(?y:Nat)))",
"s(s(?z:Nat)) = min(s(?x:Nat),s(?z:Nat))",
"s(s(0)) = min(?x:Nat,min(?y:Nat,?x:Nat))",
"s(s(0)) = min(?x:Nat,min(?z:Nat,?y:Nat))",
"s(s(0)) = min(0,min(?x:Nat,?z:Nat))",
"min(?x:Nat,?x:Nat) = s(s(s(s(?x:Nat))))",
"min(?x:Nat,?x:Nat) = s(min(?x:Nat,s(0)))",
"min(?x:Nat,?x:Nat) = s(min(s(?y:Nat),?x:Nat))",
"min(?x:Nat,?y:Nat) = s(s(min(?y:Nat,?z:Nat)))",
"min(?x:Nat,?y:Nat) = s(s(min(?z:Nat,?x:Nat)))",
"min(?x:Nat,?y:Nat) = min(0,s(s(?x:Nat)))",
"min(?x:Nat,?y:Nat) = min(s(s(?y:Nat)),?x:Nat)",
"min(?x:Nat,?y:Nat) = min(min(?y:Nat,?y:Nat),0)",
"min(?x:Nat,?y:Nat) = min(min(?z:Nat,0),?y:Nat)",
"min(?x:Nat,?y:Nat) = min(min(0,?x:Nat),0)",
"min(?x:Nat,?z:Nat) = min(?y:Nat,min(?x:Nat,?x:Nat))",
"min(?x:Nat,0) = min(?x:Nat,s(s(?x:Nat)))",
"min(?x:Nat,0) = min(s(s(?x:Nat)),?y:Nat)",
"min(?y:Nat,?x:Nat) = min(s(s(?x:Nat)),?z:Nat)",
"min(?y:Nat,?x:Nat) = min(min(?z:Nat,0),0)",
"min(?y:Nat,?y:Nat) = s(min(?x:Nat,s(0)))",
"min(?y:Nat,?y:Nat) = min(?y:Nat,min(?x:Nat,?x:Nat))",
"min(?y:Nat,?y:Nat) = min(?z:Nat,min(?y:Nat,?y:Nat))",
"min(?y:Nat,?z:Nat) = s(min(0,s(?y:Nat)))",
"min(?y:Nat,0) = s(s(min(0,?x:Nat)))",
"min(?y:Nat,0) = min(?y:Nat,min(?x:Nat,?y:Nat))",
"min(?y:Nat,0) = min(s(?x:Nat),s(0))",
"min(?z:Nat,?y:Nat) = s(min(?y:Nat,s(?z:Nat)))",
"min(?z:Nat,?y:Nat) = min(s(s(0)),?y:Nat)",
"min(?z:Nat,0) = min(?y:Nat,s(s(0)))",
"min(0,?x:Nat) = s(min(s(?x:Nat),?x:Nat))",
"min(0,?y:Nat) = min(?x:Nat,s(s(?z:Nat)))",
"min(0,?y:Nat) = min(min(?x:Nat,?x:Nat),?y:Nat)",
"min(0,?z:Nat) = min(s(?x:Nat),s(?z:Nat))",
"s(s(s(?x:Nat))) = min(s(0),?y:Nat)",
"s(s(s(?y:Nat))) = min(?y:Nat,s(?z:Nat))",
"s(min(?x:Nat,?y:Nat)) = min(?z:Nat,s(?z:Nat))",
"s(min(?x:Nat,?y:Nat)) = min(0,s(?z:Nat))",
"min(?x:Nat,s(?x:Nat)) = s(min(?x:Nat,?x:Nat))",
"min(?x:Nat,s(?x:Nat)) = s(min(0,?z:Nat))",
"min(?x:Nat,s(?y:Nat)) = min(s(0),?y:Nat)",
"min(?x:Nat,s(?z:Nat)) = min(s(?x:Nat),0)",
"min(?x:Nat,s(0)) = s(min(?y:Nat,?y:Nat))",
"min(?y:Nat,s(?x:Nat)) = s(min(?x:Nat,?x:Nat))",
"min(?y:Nat,s(?y:Nat)) = min(?x:Nat,s(?y:Nat))",
"min(?y:Nat,s(?z:Nat)) = min(s(?x:Nat),?x:Nat)",
"min(?z:Nat,s(?y:Nat)) = s(s(s(?y:Nat)))",
"min(0,s(?x:Nat)) = s(min(0,?y:Nat))",
"min(0,s(?y:Nat)) = min(s(?x:Nat),?x:Nat)",
"min(0,s(?z:Nat)) = s(s(s(0)))",
"min(0,s(0)) = s(min(?y:Nat,?z:Nat))",
"min(s(?x:Nat),?x:Nat) = min(s(?x:Nat),?z:Nat)",
"min(s(?x:Nat),?x:Nat) = min(s(?y:Nat),?z:Nat)",
"min(s(?x:Nat),?y:Nat) = s(min(?x:Nat,?y:Nat))",
"min(s(?x:Nat),0) = s(min(?y:Nat,?z:Nat))",
"min(s(?y:Nat),?x:Nat) = min(s(0),?x:Nat)",
"min(s(?y:Nat),?y:Nat) = min(?x:Nat,s(?y:Nat))",
"min(s(?z:Nat),?y:Nat) = min(?y:Nat,s(?x:Nat))",
"min(s(?z:Nat),?y:Nat) = min(s(0),?x:Nat)",
"min(s(?z:Nat),?z:Nat) = s(min(?x:Nat,?x:Nat))",
"min(s(?z:Nat),?z:Nat) = s(min(?y:Nat,?x:Nat))",
"min(s(?z:Nat),0) = s(s(s(?z:Nat)))",
"min(s(0),?y:Nat) = min(?y:Nat,s(?z:Nat))",
"min(s(0),0) = min(s(0),?x:Nat)",
"s(s(s(s(?z:Nat)))) = min(0,?x:Nat)",
"s(s(min(?y:Nat,0))) = min(?x:Nat,?y:Nat)",
"s(min(?x:Nat,s(?z:Nat))) = min(?y:Nat,0)",
"s(min(0,s(?x:Nat))) = min(?z:Nat,?x:Nat)",
"s(min(s(?y:Nat),?y:Nat)) = min(?z:Nat,0)",
"s(min(s(?y:Nat),?z:Nat)) = min(?x:Nat,?z:Nat)",
"s(min(s(?z:Nat),?x:Nat)) = min(?x:Nat,0)",
"s(min(s(0),0)) = min(?x:Nat,?x:Nat)",
"min(?x:Nat,s(s(?x:Nat))) = min(?x:Nat,?y:Nat)",
"min(?x:Nat,s(s(0))) = s(s(0))",
"min(?x:Nat,s(s(0))) = min(?x:Nat,?z:Nat)",
"min(?x:Nat,min(?z:Nat,?x:Nat)) = min(?x:Nat,?y:Nat)",
"min(?x:Nat,min(?z:Nat,?x:Nat)) = min(0,0)",
"min(?x:Nat,min(?z:Nat,0)) = min(?x:Nat,?z:Nat)",
"min(?y:Nat,s(s(?x:Nat))) = min(?y:Nat,?y:Nat)",
"min(?y:Nat,s(s(?z:Nat))) = min(?x:Nat,?x:Nat)",
"min(?y:Nat,min(0,0)) = min(?z:Nat,?y:Nat)",
"min(0,s(s(?y:Nat))) = min(?x:Nat,?z:Nat)",
"min(s(?x:Nat),s(?y:Nat)) = s(s(?z:Nat))",
"min(s(?x:Nat),s(?y:Nat)) = min(?x:Nat,?y:Nat)",
"min(s(?y:Nat),s(?z:Nat)) = min(?x:Nat,?z:Nat)",
"min(s(s(?x:Nat)),?x:Nat) = min(0,?y:Nat)",
"min(s(s(?x:Nat)),?z:Nat) = min(?z:Nat,0)",
"min(s(s(?y:Nat)),?z:Nat) = min(?x:Nat,?x:Nat)",
"min(s(s(?z:Nat)),?x:Nat) = s(s(?y:Nat))",
"min(min(?x:Nat,?z:Nat),?x:Nat) = min(?x:Nat,?y:Nat)",
"min(min(?x:Nat,?z:Nat),?x:Nat) = min(?y:Nat,?z:Nat)",
"min(min(?x:Nat,?z:Nat),?y:Nat) = min(?y:Nat,?y:Nat)",
"min(min(?y:Nat,?x:Nat),?z:Nat) = min(?y:Nat,?y:Nat)",
"min(min(?y:Nat,?y:Nat),?x:Nat) = min(?x:Nat,?x:Nat)",
"min(min(?y:Nat,?z:Nat),?z:Nat) = s(s(?y:Nat))",
"min(min(?z:Nat,?x:Nat),0) = min(?x:Nat,?y:Nat)",
"min(min(?z:Nat,?y:Nat),?y:Nat) = min(?x:Nat,?y:Nat)",
"min(min(?z:Nat,0),?x:Nat) = min(?x:Nat,?x:Nat)",
"min(?x:Nat,s(s(s(?x:Nat)))) = s(?x:Nat)",
"min(?x:Nat,min(s(?y:Nat),?y:Nat)) = s(?y:Nat)",
"min(?y:Nat,s(s(s(0)))) = s(?y:Nat)",
"min(?y:Nat,s(min(0,?x:Nat))) = s(?x:Nat)",
"min(?y:Nat,min(?x:Nat,s(?y:Nat))) = s(?z:Nat)",
"min(?y:Nat,min(0,s(?x:Nat))) = s(?x:Nat)",
"min(?y:Nat,min(s(?y:Nat),?z:Nat)) = s(?z:Nat)",
"min(0,s(s(s(?x:Nat)))) = s(?y:Nat)",
"min(0,s(min(?y:Nat,?x:Nat))) = s(?y:Nat)",
"min(0,min(s(0),?z:Nat)) = s(0)",
"min(s(?x:Nat),min(?x:Nat,?z:Nat)) = s(?y:Nat)",
"min(s(?x:Nat),min(?x:Nat,?z:Nat)) = s(?z:Nat)",
"min(s(?x:Nat),min(0,?y:Nat)) = s(?x:Nat)",
"min(s(?y:Nat),min(?x:Nat,?x:Nat)) = s(0)",
"min(s(min(?x:Nat,0)),?y:Nat) = s(?x:Nat)",
"min(s(min(?y:Nat,?y:Nat)),0) = s(?x:Nat)",
"min(s(min(0,0)),?y:Nat) = s(?z:Nat)",
"min(min(?x:Nat,?z:Nat),s(?y:Nat)) = s(?y:Nat)",
"min(min(?x:Nat,s(?y:Nat)),?y:Nat) = s(?x:Nat)",
"min(min(?y:Nat,?x:Nat),s(?z:Nat)) = s(?x:Nat)",
"min(min(?y:Nat,?y:Nat),s(0)) = s(?y:Nat)",
"min(min(?y:Nat,0),s(?y:Nat)) = s(?x:Nat)",
"min(min(?y:Nat,s(?y:Nat)),?x:Nat) = s(?x:Nat)",
"min(min(?y:Nat,s(?y:Nat)),?y:Nat) = s(?z:Nat)",
"min(min(?y:Nat,s(?z:Nat)),0) = s(?z:Nat)",
"min(min(0,s(?z:Nat)),?x:Nat) = s(?x:Nat)",
"min(min(s(?x:Nat),?x:Nat),0) = s(?y:Nat)",
"min(min(s(?y:Nat),0),0) = s(?y:Nat)",
"s(?x:Nat) = min(?y:Nat,min(min(?z:Nat,?y:Nat),?z:Nat))",
"s(?x:Nat) = min(s(?x:Nat),min(s(?y:Nat),?x:Nat))",
"s(?x:Nat) = min(s(s(?x:Nat)),s(s(?x:Nat)))",
"s(?x:Nat) = min(s(min(s(?x:Nat),0)),?y:Nat)",
"s(?x:Nat) = min(min(?z:Nat,0),s(s(?y:Nat)))",
"s(?x:Nat) = min(min(0,?y:Nat),min(?y:Nat,?y:Nat))",
"s(?y:Nat) = min(s(?x:Nat),min(s(?y:Nat),0))",
"s(?y:Nat) = min(s(s(?z:Nat)),min(?z:Nat,?z:Nat))",
"s(?y:Nat) = min(s(min(?x:Nat,?y:Nat)),s(?x:Nat))",
"s(?y:Nat) = min(s(min(?x:Nat,?y:Nat)),s(0))",
"s(?y:Nat) = min(min(?y:Nat,min(?z:Nat,?z:Nat)),?z:Nat)",
"s(?y:Nat) = min(min(?z:Nat,?x:Nat),min(0,?y:Nat))",
"s(?y:Nat) = min(min(min(?z:Nat,0),?x:Nat),?y:Nat)",
"s(?y:Nat) = min(min(min(0,0),?z:Nat),?x:Nat)",
"s(?z:Nat) = min(min(?y:Nat,?y:Nat),s(s(?y:Nat)))",
"s(?z:Nat) = min(min(s(?y:Nat),s(?x:Nat)),?x:Nat)",
"s(0) = min(?x:Nat,min(?y:Nat,min(?y:Nat,?x:Nat)))",
"s(0) = min(?z:Nat,min(min(?x:Nat,?x:Nat),?y:Nat))",
"s(0) = min(min(0,s(s(?x:Nat))),?x:Nat)",
"s(s(?x:Nat)) = min(min(?x:Nat,s(?x:Nat)),?y:Nat)",
"s(s(?y:Nat)) = min(s(0),min(?x:Nat,?y:Nat))",
"s(s(0)) = min(min(?x:Nat,s(?z:Nat)),?y:Nat)",
"min(?x:Nat,?x:Nat) = min(min(s(?y:Nat),?y:Nat),?x:Nat)",
"min(?x:Nat,?y:Nat) = s(min(?z:Nat,min(?x:Nat,?x:Nat)))",
"min(?x:Nat,?y:Nat) = s(min(min(?y:Nat,0),?x:Nat))",
"min(?x:Nat,?y:Nat) = min(s(s(?x:Nat)),s(?z:Nat))",
"min(?x:Nat,0) = s(min(min(?x:Nat,?y:Nat),?x:Nat))",
"min(?y:Nat,?y:Nat) = s(s(s(min(?x:Nat,?z:Nat))))",
"min(?y:Nat,0) = min(?x:Nat,min(?x:Nat,s(?y:Nat)))",
"min(?y:Nat,0) = min(s(0),s(s(?x:Nat)))",
"min(?z:Nat,?y:Nat) = min(s(min(?y:Nat,?x:Nat)),?z:Nat)",
"min(?z:Nat,?y:Nat) = min(min(0,s(?y:Nat)),?x:Nat)",
"min(0,?y:Nat) = min(min(?z:Nat,s(0)),?y:Nat)",
"min(0,0) = min(?x:Nat,min(s(?z:Nat),?x:Nat))",
"s(s(s(?x:Nat))) = min(min(?y:Nat,0),?z:Nat)",
"s(min(?x:Nat,?x:Nat)) = min(s(s(0)),?y:Nat)",
"s(min(?y:Nat,?x:Nat)) = min(min(?y:Nat,?z:Nat),?y:Nat)",
"s(min(?y:Nat,?y:Nat)) = min(min(?x:Nat,?x:Nat),?y:Nat)",
"s(min(?z:Nat,?x:Nat)) = min(s(s(?y:Nat)),?x:Nat)",
"s(min(0,?x:Nat)) = min(?x:Nat,min(?z:Nat,?x:Nat))",
"s(min(0,?z:Nat)) = min(?y:Nat,min(?x:Nat,?x:Nat))",
"min(?x:Nat,s(?x:Nat)) = min(0,s(s(0)))",
"min(?y:Nat,s(?y:Nat)) = min(min(?y:Nat,0),?y:Nat)",
"min(?z:Nat,s(?x:Nat)) = s(s(min(?x:Nat,?z:Nat)))",
"min(0,s(?y:Nat)) = s(min(s(?x:Nat),?z:Nat))",
"min(s(?x:Nat),?x:Nat) = min(min(0,0),?y:Nat)",
"min(s(?x:Nat),?y:Nat) = min(?x:Nat,min(?z:Nat,?z:Nat))",
"min(s(?x:Nat),?z:Nat) = min(min(?x:Nat,?x:Nat),?y:Nat)",
"min(s(?y:Nat),?x:Nat) = s(min(?y:Nat,s(?z:Nat)))",
"min(s(?y:Nat),?x:Nat) = min(?y:Nat,s(s(?x:Nat)))",
"s(s(min(?y:Nat,?x:Nat))) = min(?z:Nat,s(?y:Nat))",
"s(s(min(?y:Nat,0))) = min(s(?x:Nat),?x:Nat)",
"s(min(s(?x:Nat),?x:Nat)) = min(s(?x:Nat),?y:Nat)",
"min(?x:Nat,s(s(?x:Nat))) = s(s(s(?x:Nat)))",
"min(?x:Nat,min(?y:Nat,?y:Nat)) = min(s(?z:Nat),?y:Nat)",
"min(?y:Nat,min(?x:Nat,?y:Nat)) = s(s(s(?y:Nat)))",
"min(?y:Nat,min(?z:Nat,?x:Nat)) = s(min(?y:Nat,?z:Nat))",
"min(?y:Nat,min(?z:Nat,?y:Nat)) = s(min(0,?x:Nat))",
"min(?z:Nat,min(?y:Nat,?x:Nat)) = min(?z:Nat,s(?x:Nat))",
"min(?z:Nat,min(?y:Nat,?y:Nat)) = s(s(s(?z:Nat)))",
"min(0,min(?y:Nat,?y:Nat)) = s(min(0,0))",
"min(0,min(?z:Nat,?x:Nat)) = s(min(0,?x:Nat))",
"min(s(?x:Nat),s(?z:Nat)) = s(s(s(?z:Nat)))",
"min(min(?x:Nat,?x:Nat),?x:Nat) = min(?z:Nat,s(0))",
"min(min(?x:Nat,?z:Nat),0) = min(?y:Nat,s(?y:Nat))",
"min(min(?x:Nat,0),?x:Nat) = min(?y:Nat,s(?y:Nat))",
"min(min(?y:Nat,?y:Nat),?z:Nat) = min(s(?x:Nat),?y:Nat)",
"min(min(?z:Nat,?x:Nat),?y:Nat) = min(?y:Nat,s(?x:Nat))",
"min(min(0,0),?y:Nat) = min(?x:Nat,s(0))",
"s(s(s(s(s(?x:Nat))))) = min(?y:Nat,?x:Nat)",
"s(min(?x:Nat,min(?x:Nat,?z:Nat))) = min(?x:Nat,?x:Nat)",
"s(min(?z:Nat,min(?z:Nat,?x:Nat))) = min(?y:Nat,?y:Nat)",
"min(?x:Nat,min(?x:Nat,s(?x:Nat))) = min(?y:Nat,?x:Nat)",
"min(?x:Nat,min(s(0),?x:Nat)) = min(?x:Nat,?x:Nat)",
"min(?y:Nat,s(s(s(?z:Nat)))) = s(s(?z:Nat))",
"min(?y:Nat,s(min(?y:Nat,0))) = s(s(0))",
"min(?y:Nat,s(min(?y:Nat,0))) = min(?x:Nat,?z:Nat)",
"min(?y:Nat,min(?y:Nat,s(?x:Nat))) = s(s(?x:Nat))",
"min(?y:Nat,min(s(?x:Nat),?x:Nat)) = min(?z:Nat,?y:Nat)",
"min(?y:Nat,min(s(?y:Nat),?z:Nat)) = min(?x:Nat,?z:Nat)",
"min(?z:Nat,s(s(s(0)))) = s(s(?y:Nat))",
"min(?z:Nat,min(?z:Nat,s(?y:Nat))) = min(?x:Nat,?y:Nat)",
"min(s(?x:Nat),min(?x:Nat,?y:Nat)) = min(?y:Nat,0)",
"min(s(s(?x:Nat)),s(?x:Nat)) = min(?z:Nat,?y:Nat)",
"min(s(s(?x:Nat)),s(?y:Nat)) = s(s(?x:Nat))",
"min(s(s(?x:Nat)),s(?y:Nat)) = min(0,?z:Nat)",
"min(s(min(?x:Nat,?x:Nat)),?y:Nat) = min(?y:Nat,?y:Nat)",
"min(min(?x:Nat,?x:Nat),s(?z:Nat)) = min(?z:Nat,?y:Nat)",
"min(min(?x:Nat,s(?x:Nat)),?y:Nat) = min(?z:Nat,0)",
"min(min(?z:Nat,s(?x:Nat)),?x:Nat) = min(?y:Nat,0)",
"min(?x:Nat,s(s(min(?y:Nat,?z:Nat)))) = s(?y:Nat)",
"min(?x:Nat,s(min(s(?x:Nat),0))) = s(?x:Nat)",
"min(?x:Nat,s(min(s(?z:Nat),0))) = s(?y:Nat)",
"min(?x:Nat,min(?x:Nat,min(?y:Nat,?x:Nat))) = s(?x:Nat)",
"min(?x:Nat,min(?y:Nat,min(?x:Nat,?y:Nat))) = s(?x:Nat)",
"min(?x:Nat,min(s(?y:Nat),s(?z:Nat))) = s(0)",
"min(?x:Nat,min(min(?x:Nat,0),?x:Nat)) = s(?y:Nat)",
"min(?x:Nat,min(min(0,?z:Nat),0)) = s(?x:Nat)",
"min(?y:Nat,s(s(min(?y:Nat,?y:Nat)))) = s(?y:Nat)",
"min(?y:Nat,s(min(s(?x:Nat),?y:Nat))) = s(?x:Nat)",
"min(?y:Nat,min(?y:Nat,s(s(0)))) = s(?x:Nat)",
"min(?y:Nat,min(?z:Nat,min(?x:Nat,0))) = s(?z:Nat)",
"min(?y:Nat,min(min(?y:Nat,?y:Nat),?y:Nat)) = s(?x:Nat)",
"min(?z:Nat,min(?x:Nat,min(?z:Nat,?y:Nat))) = s(?x:Nat)",
"min(?z:Nat,min(?y:Nat,s(s(?x:Nat)))) = s(0)",
"min(?z:Nat,min(s(s(?y:Nat)),?x:Nat)) = s(?x:Nat)",
"min(?z:Nat,min(s(s(?z:Nat)),?x:Nat)) = s(?x:Nat)",
"min(0,min(?x:Nat,min(?y:Nat,?x:Nat))) = s(?z:Nat)",
"min(s(?z:Nat),min(s(?x:Nat),?z:Nat)) = s(?x:Nat)",
"min(s(?z:Nat),min(s(?x:Nat),0)) = s(?y:Nat)",
"min(s(0),min(?x:Nat,s(?z:Nat))) = s(?x:Nat)",
"min(s(s(min(?x:Nat,?z:Nat))),?x:Nat) = s(?y:Nat)",
"min(s(min(?y:Nat,s(0))),?y:Nat) = s(?y:Nat)",
"min(min(0,?x:Nat),s(s(?x:Nat))) = s(?y:Nat)",
"min(min(s(0),?z:Nat),s(?x:Nat)) = s(?y:Nat)",
"min(min(min(?y:Nat,?y:Nat),0),0) = s(?z:Nat)",
"min(min(min(?y:Nat,?z:Nat),?y:Nat),?y:Nat) = s(?z:Nat)",
"min(min(min(?y:Nat,0),0),0) = s(?z:Nat)",
"s(?x:Nat) = min(s(?y:Nat),min(s(?y:Nat),s(?x:Nat)))",
"s(?x:Nat) = min(s(s(s(min(?x:Nat,0)))),?x:Nat)",
"s(?x:Nat) = min(s(s(min(s(?z:Nat),?y:Nat))),?y:Nat)",
"s(?x:Nat) = min(s(min(?x:Nat,?y:Nat)),min(?x:Nat,0))",
"s(?x:Nat) = min(s(min(?x:Nat,min(?x:Nat,?z:Nat))),?z:Nat)",
"s(?x:Nat) = min(s(min(0,s(?y:Nat))),s(0))",
"s(?x:Nat) = min(min(?x:Nat,min(s(?x:Nat),?x:Nat)),?x:Nat)",
"s(?x:Nat) = min(min(0,s(min(?y:Nat,?z:Nat))),?x:Nat)",
"s(?y:Nat) = min(s(?x:Nat),min(min(?y:Nat,?y:Nat),?x:Nat))",
"s(?y:Nat) = min(s(s(min(?y:Nat,s(?y:Nat)))),?y:Nat)",
"s(?y:Nat) = min(min(?x:Nat,?x:Nat),min(s(0),?y:Nat))",
"s(?y:Nat) = min(min(0,s(?x:Nat)),s(s(?y:Nat)))",
"s(?z:Nat) = min(?x:Nat,s(min(min(?z:Nat,?y:Nat),?x:Nat)))",
"s(?z:Nat) = min(0,min(0,s(s(s(?z:Nat)))))",
"s(?z:Nat) = min(s(s(s(min(?y:Nat,?y:Nat)))),?x:Nat)",
"s(?z:Nat) = min(min(?x:Nat,0),min(s(?y:Nat),?y:Nat))",
"s(?z:Nat) = min(min(?x:Nat,s(0)),s(s(0)))",
"s(?z:Nat) = min(min(s(?y:Nat),0),min(?x:Nat,?x:Nat))",
"s(0) = min(?x:Nat,min(s(min(?z:Nat,?y:Nat)),?z:Nat))",
"s(0) = min(s(0),min(0,s(s(?x:Nat))))",
"s(s(?y:Nat)) = min(s(0),min(?y:Nat,s(?y:Nat)))",
"s(s(?y:Nat)) = min(min(?y:Nat,?x:Nat),min(0,?y:Nat))",
"s(s(?y:Nat)) = min(min(min(?x:Nat,?x:Nat),?z:Nat),0)",
"s(s(?z:Nat)) = min(?x:Nat,min(?y:Nat,s(s(?y:Nat))))",
"s(s(0)) = min(?y:Nat,min(?x:Nat,min(?y:Nat,?y:Nat)))",
"s(s(0)) = min(s(?y:Nat),min(?y:Nat,s(?x:Nat)))",
"min(?x:Nat,?x:Nat) = s(s(min(?y:Nat,min(?z:Nat,?y:Nat))))",
"min(?x:Nat,?x:Nat) = s(min(s(min(?z:Nat,?x:Nat)),0))",
"min(?x:Nat,?y:Nat) = min(?y:Nat,s(min(s(?x:Nat),?x:Nat)))",
"min(?x:Nat,?z:Nat) = min(s(0),min(s(?y:Nat),0))",
"min(?x:Nat,0) = min(?y:Nat,min(s(?x:Nat),s(0)))",
"min(?y:Nat,?y:Nat) = min(?z:Nat,s(min(s(?y:Nat),?y:Nat)))",
"min(?y:Nat,?y:Nat) = min(min(s(s(?y:Nat)),?y:Nat),?x:Nat)",
"min(?y:Nat,0) = s(s(min(?y:Nat,min(?x:Nat,?y:Nat))))",
"min(?y:Nat,0) = s(s(min(min(0,?y:Nat),?x:Nat)))",
"min(?z:Nat,?x:Nat) = s(min(?x:Nat,s(min(?y:Nat,?y:Nat))))",
"min(?z:Nat,?x:Nat) = min(?y:Nat,s(min(s(?x:Nat),0)))",
"min(?z:Nat,?z:Nat) = s(min(0,min(?y:Nat,s(0))))",
"min(0,?x:Nat) = s(min(s(s(?x:Nat)),s(?y:Nat)))",
"min(0,?x:Nat) = s(min(s(s(s(0))),?x:Nat))",
"min(0,?y:Nat) = s(min(min(s(?y:Nat),0),0))",
"min(0,?y:Nat) = min(?z:Nat,min(min(?y:Nat,?x:Nat),?y:Nat))",
"min(0,?z:Nat) = s(min(min(?x:Nat,s(?x:Nat)),?y:Nat))",
"s(min(?x:Nat,?y:Nat)) = min(?z:Nat,s(min(?z:Nat,?x:Nat)))",
"min(?y:Nat,s(?y:Nat)) = s(min(s(s(?x:Nat)),?y:Nat))",
"min(?y:Nat,s(?y:Nat)) = min(s(min(?y:Nat,?z:Nat)),?y:Nat)",
"min(0,s(?x:Nat)) = min(?y:Nat,s(min(?z:Nat,?y:Nat)))",
"min(0,s(0)) = min(?x:Nat,s(min(0,?y:Nat)))",
"min(s(?x:Nat),?z:Nat) = min(min(?x:Nat,?y:Nat),s(0))",
"min(s(?z:Nat),?z:Nat) = min(s(s(s(?y:Nat))),?y:Nat)",
"min(s(0),?y:Nat) = min(min(?y:Nat,0),s(0))",
"min(s(0),?y:Nat) = min(min(?z:Nat,0),s(?x:Nat))",
"min(s(0),?z:Nat) = min(?x:Nat,s(min(?y:Nat,?y:Nat)))",
"s(s(min(?x:Nat,?x:Nat))) = min(min(?y:Nat,?x:Nat),?z:Nat)",
"s(s(min(?y:Nat,?y:Nat))) = min(?x:Nat,min(?y:Nat,0))",
"s(s(min(?z:Nat,0))) = min(min(?y:Nat,?y:Nat),0)",
"s(s(min(0,?z:Nat))) = min(min(?y:Nat,?z:Nat),?y:Nat)",
"s(min(?z:Nat,s(?y:Nat))) = min(?y:Nat,min(?x:Nat,?y:Nat))",
"s(min(s(?x:Nat),?x:Nat)) = min(min(0,?y:Nat),?y:Nat)",
"min(?y:Nat,s(s(?x:Nat))) = min(s(?x:Nat),s(?x:Nat))",
"min(?z:Nat,s(s(?z:Nat))) = min(min(?x:Nat,?z:Nat),0)",
"min(s(?x:Nat),s(?y:Nat)) = min(min(?y:Nat,0),?y:Nat)",
"min(s(s(?y:Nat)),?y:Nat) = min(?x:Nat,min(?z:Nat,?z:Nat))",
"min(min(?x:Nat,?y:Nat),?y:Nat) = min(min(?y:Nat,0),?z:Nat)",
"min(min(?x:Nat,0),?y:Nat) = min(?x:Nat,min(?x:Nat,?z:Nat))",
"min(min(?y:Nat,?y:Nat),?x:Nat) = min(min(?x:Nat,?y:Nat),?z:Nat)",
"min(min(?y:Nat,?y:Nat),?y:Nat) = min(?x:Nat,s(s(?y:Nat)))",
"min(min(?y:Nat,?z:Nat),?y:Nat) = s(s(min(?z:Nat,?y:Nat)))",
"min(min(?z:Nat,0),?y:Nat) = min(min(?y:Nat,?z:Nat),?x:Nat)",
"s(s(min(s(?x:Nat),0))) = min(s(0),?z:Nat)",
"s(min(0,min(0,?y:Nat))) = min(s(?x:Nat),?z:Nat)",
"s(min(s(?y:Nat),s(?x:Nat))) = min(s(?y:Nat),?y:Nat)",
"min(?y:Nat,min(?z:Nat,s(?z:Nat))) = s(min(?y:Nat,?z:Nat))",
"min(?z:Nat,min(?x:Nat,s(?z:Nat))) = min(s(?z:Nat),?x:Nat)",
"min(0,s(min(?y:Nat,?y:Nat))) = min(?z:Nat,s(?y:Nat))",
"min(s(?z:Nat),s(s(?x:Nat))) = min(s(0),?z:Nat)",
"min(s(s(s(?y:Nat))),?y:Nat) = s(min(?y:Nat,?z:Nat))",
"min(s(min(?y:Nat,?y:Nat)),?x:Nat) = min(?y:Nat,s(?x:Nat))",
"min(min(?x:Nat,s(?x:Nat)),?z:Nat) = s(min(?x:Nat,0))",
"min(min(?y:Nat,s(0)),?x:Nat) = min(s(?y:Nat),0)",
"min(min(?z:Nat,?z:Nat),s(0)) = s(min(?x:Nat,?x:Nat))",
"min(min(?z:Nat,s(?z:Nat)),?x:Nat) = min(?x:Nat,s(?z:Nat))",
"min(min(?z:Nat,s(?z:Nat)),?x:Nat) = min(?z:Nat,s(?y:Nat))",
"s(s(min(s(s(0)),?y:Nat))) = min(?y:Nat,?x:Nat)",
"s(min(?z:Nat,s(s(s(?y:Nat))))) = min(0,?y:Nat)",
"s(min(s(s(s(?y:Nat))),?y:Nat)) = min(0,?z:Nat)",
"s(min(min(?z:Nat,?x:Nat),s(?x:Nat))) = min(?x:Nat,?z:Nat)",
"min(?x:Nat,s(s(min(?z:Nat,?x:Nat)))) = min(?y:Nat,?z:Nat)",
"min(?x:Nat,s(s(min(?z:Nat,?z:Nat)))) = min(?z:Nat,0)",
"min(?x:Nat,s(min(s(?x:Nat),?x:Nat))) = min(?y:Nat,?x:Nat)",
"min(?x:Nat,min(?y:Nat,min(?x:Nat,?y:Nat))) = s(s(0))",
"min(?x:Nat,min(s(?y:Nat),s(?x:Nat))) = min(0,?y:Nat)",
"min(?x:Nat,min(min(?x:Nat,?x:Nat),?y:Nat)) = s(s(?y:Nat))",
"min(?y:Nat,s(s(min(?x:Nat,?x:Nat)))) = min(?x:Nat,?x:Nat)",
"min(?z:Nat,min(min(0,?z:Nat),?y:Nat)) = min(?y:Nat,?z:Nat)",
"min(s(?x:Nat),min(s(?x:Nat),0)) = min(?z:Nat,?y:Nat)",
"min(s(?y:Nat),s(min(?x:Nat,?z:Nat))) = min(?y:Nat,?y:Nat)",
"min(s(?y:Nat),s(min(?y:Nat,?x:Nat))) = min(0,?y:Nat)",
"min(s(?y:Nat),min(?x:Nat,s(?x:Nat))) = s(s(0))",
"min(s(s(s(?x:Nat))),s(?x:Nat)) = min(?x:Nat,?x:Nat)",
"min(s(min(s(?x:Nat),?x:Nat)),?z:Nat) = s(s(0))",
"min(min(?x:Nat,?x:Nat),min(0,?x:Nat)) = min(?x:Nat,?z:Nat)",
"min(min(?x:Nat,?y:Nat),min(?y:Nat,0)) = min(?z:Nat,?x:Nat)",
"min(min(s(s(?y:Nat)),?z:Nat),?z:Nat) = s(s(?y:Nat))",
"min(min(min(?x:Nat,?x:Nat),0),?y:Nat) = min(?y:Nat,?y:Nat)",
"min(?x:Nat,s(s(min(s(?y:Nat),?y:Nat)))) = s(?z:Nat)",
"min(?x:Nat,min(?y:Nat,s(s(s(?x:Nat))))) = s(?x:Nat)",
"min(?x:Nat,min(?z:Nat,min(0,s(?y:Nat)))) = s(?y:Nat)",
"min(?x:Nat,min(0,s(min(?z:Nat,0)))) = s(?y:Nat)",
"min(?x:Nat,min(s(min(?y:Nat,?x:Nat)),?z:Nat)) = s(?x:Nat)",
"min(?x:Nat,min(s(min(?y:Nat,?z:Nat)),?y:Nat)) = s(?y:Nat)",
"min(?y:Nat,s(min(?x:Nat,min(?y:Nat,?y:Nat)))) = s(?z:Nat)",
"min(?y:Nat,s(min(s(?x:Nat),s(0)))) = s(?y:Nat)",
"min(?y:Nat,min(?x:Nat,min(0,s(?y:Nat)))) = s(?z:Nat)",
"min(?z:Nat,s(min(s(s(?x:Nat)),?y:Nat))) = s(?z:Nat)",
"min(s(0),min(?y:Nat,min(?y:Nat,?y:Nat))) = s(?z:Nat)",
"min(s(s(?y:Nat)),s(min(?x:Nat,?y:Nat))) = s(?z:Nat)",
"min(s(s(s(?y:Nat))),min(?x:Nat,?y:Nat)) = s(?x:Nat)",
"min(min(?y:Nat,min(?x:Nat,s(?x:Nat))),?x:Nat) = s(?x:Nat)",
"min(min(0,s(0)),min(?y:Nat,0)) = s(?y:Nat)",
"min(min(s(?y:Nat),s(?x:Nat)),s(?y:Nat)) = s(?y:Nat)",
"min(min(s(?z:Nat),min(0,?y:Nat)),?y:Nat) = s(?x:Nat)",
"min(min(min(?x:Nat,?x:Nat),s(?z:Nat)),?x:Nat) = s(?x:Nat)",
"min(min(min(?y:Nat,?x:Nat),s(?z:Nat)),0) = s(?z:Nat)",
"min(min(min(?z:Nat,?y:Nat),?y:Nat),s(?z:Nat)) = s(?x:Nat)",
"s(?x:Nat) = min(s(?x:Nat),min(min(?y:Nat,s(?x:Nat)),?y:Nat))",
"s(?x:Nat) = min(min(?y:Nat,min(?y:Nat,min(0,?x:Nat))),?y:Nat)",
"s(?x:Nat) = min(min(?z:Nat,s(min(?y:Nat,s(?z:Nat)))),?z:Nat)",
"s(?x:Nat) = min(min(min(s(?x:Nat),s(0)),0),?x:Nat)",
"s(?y:Nat) = min(?x:Nat,min(?z:Nat,min(?y:Nat,s(s(?x:Nat)))))",
"s(?y:Nat) = min(?z:Nat,s(s(min(0,min(?x:Nat,?z:Nat)))))",
"s(?y:Nat) = min(?z:Nat,min(?z:Nat,min(s(s(?x:Nat)),?x:Nat)))",
"s(?y:Nat) = min(0,min(0,s(min(?x:Nat,s(?x:Nat)))))",
"s(?y:Nat) = min(s(min(?x:Nat,s(min(?y:Nat,?x:Nat)))),?y:Nat)",
"s(?y:Nat) = min(min(?y:Nat,s(s(?y:Nat))),s(s(?y:Nat)))",
"s(?y:Nat) = min(min(s(0),?x:Nat),min(s(?x:Nat),?y:Nat))",
"s(?y:Nat) = min(min(min(s(?y:Nat),?x:Nat),s(?x:Nat)),?y:Nat)",
"s(?z:Nat) = min(?x:Nat,min(s(s(?y:Nat)),min(?x:Nat,?x:Nat)))",
"s(?z:Nat) = min(?x:Nat,min(min(min(?y:Nat,?y:Nat),0),0))",
"s(?z:Nat) = min(?y:Nat,min(min(0,s(s(?x:Nat))),?y:Nat))",
"s(0) = min(min(min(?x:Nat,?x:Nat),?x:Nat),min(?x:Nat,?x:Nat))",
"s(s(?x:Nat)) = min(?x:Nat,min(?z:Nat,s(min(?y:Nat,?x:Nat))))",
"min(?x:Nat,?x:Nat) = min(min(?z:Nat,?y:Nat),min(?x:Nat,s(?x:Nat)))",
"min(?x:Nat,?y:Nat) = s(min(min(0,min(0,?z:Nat)),?x:Nat))",
"min(?x:Nat,?y:Nat) = min(min(?x:Nat,s(min(?z:Nat,?x:Nat))),?x:Nat)",
"min(?x:Nat,?z:Nat) = min(min(?y:Nat,s(min(?z:Nat,?x:Nat))),?y:Nat)",
"min(?x:Nat,?z:Nat) = min(min(s(s(s(?x:Nat))),0),?y:Nat)",
"min(?x:Nat,0) = min(min(?x:Nat,min(?z:Nat,?y:Nat)),s(?x:Nat))",
"min(?y:Nat,?x:Nat) = min(min(s(min(?y:Nat,?y:Nat)),?x:Nat),?x:Nat)",
"min(?y:Nat,?y:Nat) = min(s(?z:Nat),s(min(?z:Nat,s(?z:Nat))))",
"min(?y:Nat,?y:Nat) = min(s(min(s(s(?y:Nat)),?x:Nat)),?x:Nat)",
"min(0,?x:Nat) = s(min(?y:Nat,min(min(?z:Nat,?z:Nat),?x:Nat)))",
"min(0,0) = s(s(min(0,s(min(?y:Nat,?z:Nat)))))",
"s(min(?x:Nat,?x:Nat)) = min(s(s(?x:Nat)),min(?y:Nat,?y:Nat))",
"s(min(?x:Nat,?x:Nat)) = min(min(min(?y:Nat,?y:Nat),?x:Nat),?y:Nat)",
"s(min(?x:Nat,?z:Nat)) = min(min(?x:Nat,s(?z:Nat)),s(?y:Nat))",
"s(min(?y:Nat,?y:Nat)) = min(min(min(?x:Nat,?x:Nat),?z:Nat),?y:Nat)",
"s(min(0,?x:Nat)) = min(min(?y:Nat,s(s(?x:Nat))),?y:Nat)",
"min(?y:Nat,s(?y:Nat)) = min(?z:Nat,min(min(?z:Nat,0),?z:Nat))",
"min(s(?y:Nat),?x:Nat) = min(?z:Nat,min(?z:Nat,min(?y:Nat,?y:Nat)))",
"s(s(min(?y:Nat,?z:Nat))) = min(s(?y:Nat),min(?x:Nat,?x:Nat))",
"min(0,min(?y:Nat,?x:Nat)) = min(?y:Nat,s(min(?x:Nat,?x:Nat)))",
"min(min(?x:Nat,?y:Nat),?x:Nat) = min(min(s(?z:Nat),0),?y:Nat)",
"s(min(?y:Nat,min(?z:Nat,?y:Nat))) = min(min(?y:Nat,?y:Nat),0)",
"s(min(?z:Nat,s(s(?x:Nat)))) = min(s(0),s(?x:Nat))",
"min(?x:Nat,min(s(0),?z:Nat)) = min(min(0,?y:Nat),?x:Nat)",
"min(0,min(s(?y:Nat),?y:Nat)) = min(min(?x:Nat,?x:Nat),?z:Nat)",
"min(s(?y:Nat),s(s(?x:Nat))) = min(min(?y:Nat,?z:Nat),?y:Nat)",
"min(s(s(s(?y:Nat))),?x:Nat) = min(s(0),s(?x:Nat))",
"min(s(min(0,?z:Nat)),?y:Nat) = min(?x:Nat,s(s(?x:Nat)))",
"s(min(?x:Nat,min(?x:Nat,s(?x:Nat)))) = min(s(?y:Nat),?x:Nat)",
"s(min(s(0),min(?x:Nat,?y:Nat))) = min(s(?x:Nat),0)",
"min(?x:Nat,s(min(?y:Nat,s(?x:Nat)))) = s(min(?x:Nat,?x:Nat))",
"min(?x:Nat,s(min(?y:Nat,s(0)))) = min(s(?z:Nat),?y:Nat)",
"min(?x:Nat,min(min(?x:Nat,0),0)) = s(min(?x:Nat,?x:Nat))",
"min(?y:Nat,s(s(min(?x:Nat,?y:Nat)))) = min(s(?x:Nat),?y:Nat)",
"min(?y:Nat,min(min(0,?z:Nat),?x:Nat)) = min(s(?y:Nat),?z:Nat)",
"min(?z:Nat,min(min(?y:Nat,?x:Nat),0)) = s(s(s(?y:Nat)))",
"min(s(?z:Nat),min(0,s(0))) = min(s(?x:Nat),?z:Nat)",
"min(s(min(s(?x:Nat),?x:Nat)),?y:Nat) = s(s(s(0)))",
"min(min(0,min(0,?y:Nat)),?x:Nat) = min(?y:Nat,s(?z:Nat))",
"s(min(?x:Nat,min(s(s(?y:Nat)),0))) = min(?y:Nat,?x:Nat)",
"s(min(?z:Nat,min(min(?x:Nat,?y:Nat),?y:Nat))) = min(?y:Nat,?x:Nat)",
"min(?x:Nat,min(?y:Nat,s(s(s(?y:Nat))))) = min(0,?y:Nat)",
"min(?x:Nat,min(?z:Nat,s(min(?z:Nat,0)))) = min(?y:Nat,?x:Nat)",
"min(?y:Nat,min(min(?z:Nat,0),s(?y:Nat))) = s(s(0))",
"min(?z:Nat,min(min(?x:Nat,?y:Nat),s(0))) = min(?x:Nat,0)",
"min(s(min(min(?y:Nat,?z:Nat),?y:Nat)),?y:Nat) = min(?x:Nat,?x:Nat)",
"min(min(?x:Nat,?z:Nat),min(?y:Nat,s(?x:Nat))) = min(0,?y:Nat)",
"min(min(?x:Nat,s(min(0,?y:Nat))),?x:Nat) = min(?x:Nat,?y:Nat)",
"min(min(?x:Nat,min(?y:Nat,?y:Nat)),s(0)) = min(?y:Nat,?x:Nat)",
"min(min(?x:Nat,min(?y:Nat,s(?x:Nat))),0) = min(?y:Nat,?x:Nat)",
"min(min(?y:Nat,s(0)),min(?x:Nat,0)) = min(?y:Nat,?x:Nat)",
"min(min(?y:Nat,s(s(?y:Nat))),s(?y:Nat)) = min(?z:Nat,?y:Nat)",
"min(min(0,s(min(?y:Nat,?z:Nat))),?y:Nat) = min(?z:Nat,?x:Nat)",
"min(min(s(?y:Nat),min(0,?x:Nat)),?z:Nat) = min(?x:Nat,?x:Nat)",
"min(min(s(?z:Nat),?x:Nat),min(?x:Nat,?x:Nat)) = min(?x:Nat,?x:Nat)",
"min(?x:Nat,s(s(min(min(?z:Nat,?z:Nat),0)))) = s(0)",
"min(?x:Nat,s(min(min(?z:Nat,s(?y:Nat)),?y:Nat))) = s(?y:Nat)",
"min(?x:Nat,min(s(min(?x:Nat,?y:Nat)),s(?x:Nat))) = s(?y:Nat)",
"min(s(?y:Nat),min(min(?x:Nat,s(0)),?x:Nat)) = s(?y:Nat)",
"min(s(s(s(?x:Nat))),min(?y:Nat,s(?y:Nat))) = s(0)",
"min(s(s(s(min(0,0)))),s(?z:Nat)) = s(?z:Nat)",
"min(s(min(?y:Nat,?y:Nat)),s(s(s(?y:Nat)))) = s(?x:Nat)",
"min(s(min(?z:Nat,0)),min(?z:Nat,s(?x:Nat))) = s(?z:Nat)",
"min(min(?y:Nat,min(s(?y:Nat),s(?x:Nat))),?x:Nat) = s(?x:Nat)",
"min(min(?z:Nat,?y:Nat),min(?y:Nat,min(?x:Nat,?x:Nat))) = s(?x:Nat)",
"min(min(?z:Nat,s(s(min(?x:Nat,?y:Nat)))),?x:Nat) = s(?x:Nat)",
"min(min(0,?y:Nat),s(s(s(s(?y:Nat))))) = s(?x:Nat)",
"min(min(s(?x:Nat),0),s(s(s(?x:Nat)))) = s(?z:Nat)",
"min(min(min(?x:Nat,s(s(?y:Nat))),0),0) = s(?y:Nat)",
"s(?x:Nat) = min(0,s(min(min(s(?y:Nat),s(?z:Nat)),?x:Nat)))",
"s(?x:Nat) = min(min(min(s(?x:Nat),?y:Nat),0),min(?x:Nat,?x:Nat))",
"s(?y:Nat) = min(?y:Nat,min(s(min(min(?z:Nat,0),?z:Nat)),?z:Nat))",
"s(?y:Nat) = min(?z:Nat,min(s(?x:Nat),min(min(0,?y:Nat),?y:Nat)))",
"s(?z:Nat) = min(?y:Nat,min(min(s(?y:Nat),0),min(?x:Nat,?y:Nat)))",
"s(?z:Nat) = min(min(?y:Nat,min(?x:Nat,s(?z:Nat))),min(?y:Nat,?z:Nat))",
"s(0) = min(?y:Nat,s(min(min(?y:Nat,?x:Nat),min(?x:Nat,?y:Nat))))",
"s(0) = min(min(?x:Nat,min(?x:Nat,s(s(?y:Nat)))),s(?z:Nat))",
"s(s(?y:Nat)) = min(?z:Nat,min(?z:Nat,s(s(min(?x:Nat,0)))))",
"s(s(?y:Nat)) = min(0,s(min(min(?x:Nat,?x:Nat),s(?y:Nat))))",
"s(s(?z:Nat)) = min(min(?y:Nat,?x:Nat),min(?z:Nat,min(?x:Nat,?x:Nat)))",
"s(s(0)) = min(min(min(0,min(0,?z:Nat)),0),?z:Nat)",
"min(?x:Nat,?x:Nat) = min(s(s(min(?y:Nat,0))),min(0,0))",
"min(?x:Nat,?z:Nat) = s(min(s(min(min(?x:Nat,?z:Nat),?x:Nat)),?y:Nat))",
"min(?y:Nat,?x:Nat) = min(s(?y:Nat),s(min(min(?x:Nat,?x:Nat),?z:Nat)))",
"min(?y:Nat,?y:Nat) = min(0,min(0,s(s(min(?y:Nat,0)))))",
"min(?y:Nat,?y:Nat) = min(s(min(min(?x:Nat,?z:Nat),s(0))),?y:Nat)",
"min(0,?y:Nat) = s(min(min(s(?z:Nat),min(?y:Nat,?y:Nat)),0))",
"s(min(?x:Nat,?z:Nat)) = min(min(?z:Nat,min(s(0),0)),?x:Nat)",
"min(?x:Nat,s(?x:Nat)) = min(0,min(min(s(?y:Nat),?x:Nat),?y:Nat))",
"min(s(?x:Nat),?x:Nat) = min(min(?y:Nat,s(s(s(?x:Nat)))),?z:Nat)",
"min(s(?z:Nat),?x:Nat) = s(min(?z:Nat,min(min(0,0),0)))",
"s(s(min(?y:Nat,0))) = min(min(?y:Nat,?y:Nat),min(0,?y:Nat))",
"min(?x:Nat,min(?x:Nat,0)) = min(s(s(?x:Nat)),min(?y:Nat,?x:Nat))",
"min(?y:Nat,s(s(0))) = s(min(s(min(?y:Nat,?z:Nat)),0))",
"min(?y:Nat,min(?x:Nat,?x:Nat)) = min(?z:Nat,min(?x:Nat,min(0,?y:Nat)))",
"min(0,s(s(0))) = min(min(?z:Nat,?y:Nat),min(?y:Nat,?y:Nat))",
"min(s(0),s(0)) = min(s(?y:Nat),min(s(?x:Nat),?y:Nat))",
"s(min(min(?x:Nat,?y:Nat),0)) = min(min(?z:Nat,s(0)),?y:Nat)",
"min(?z:Nat,min(s(?y:Nat),?z:Nat)) = min(?y:Nat,min(s(?z:Nat),?y:Nat))",
"min(s(?z:Nat),s(s(?x:Nat))) = min(s(min(?y:Nat,?x:Nat)),?z:Nat)",
"min(min(?y:Nat,s(?z:Nat)),?x:Nat) = min(s(s(?z:Nat)),s(?y:Nat))",
"min(min(s(0),?y:Nat),?x:Nat) = min(?y:Nat,min(0,s(0)))",
"s(min(s(min(0,?x:Nat)),0)) = min(min(?y:Nat,0),?x:Nat)",
"min(?x:Nat,min(min(?y:Nat,?z:Nat),0)) = min(s(0),s(?x:Nat))",
"min(?y:Nat,min(?x:Nat,min(?y:Nat,?y:Nat))) = min(?x:Nat,min(0,?y:Nat))",
"min(s(?x:Nat),s(min(?y:Nat,?x:Nat))) = min(min(0,?y:Nat),?x:Nat)",
"min(s(s(min(?x:Nat,?y:Nat))),?y:Nat) = min(?y:Nat,min(?x:Nat,?x:Nat))",
"min(min(min(?x:Nat,?x:Nat),?y:Nat),?x:Nat) = min(s(s(?z:Nat)),?y:Nat)",
"min(min(min(?z:Nat,?y:Nat),?y:Nat),?y:Nat) = min(0,s(s(?x:Nat)))",
"s(s(min(s(min(?x:Nat,?x:Nat)),?x:Nat))) = min(?x:Nat,s(?y:Nat))",
"s(min(s(min(?x:Nat,?z:Nat)),s(?z:Nat))) = min(s(?y:Nat),?y:Nat)",
"s(min(min(min(?y:Nat,?x:Nat),?y:Nat),?y:Nat)) = min(s(?y:Nat),?x:Nat)",
"min(s(min(?y:Nat,s(?y:Nat))),s(?y:Nat)) = min(s(?y:Nat),?y:Nat)",
"min(s(min(min(?z:Nat,?x:Nat),?x:Nat)),?x:Nat) = min(?x:Nat,s(?x:Nat))",
"min(min(?y:Nat,s(?y:Nat)),min(?x:Nat,0)) = s(min(?y:Nat,?z:Nat))",
"min(min(?z:Nat,?x:Nat),min(?z:Nat,s(0))) = s(min(?z:Nat,?z:Nat))",
"min(min(s(?x:Nat),min(?y:Nat,?x:Nat)),?z:Nat) = min(s(?x:Nat),?x:Nat)",
"min(min(s(min(?x:Nat,?z:Nat)),?y:Nat),?z:Nat) = s(s(s(?z:Nat)))",
"s(min(min(min(0,s(?x:Nat)),?y:Nat),?x:Nat)) = min(0,?x:Nat)",
"min(?x:Nat,s(min(min(?y:Nat,0),s(?z:Nat)))) = min(?x:Nat,?y:Nat)",
"min(?x:Nat,min(?x:Nat,min(?x:Nat,min(?y:Nat,?x:Nat)))) = min(?x:Nat,?x:Nat)",
"min(s(s(min(?y:Nat,min(?x:Nat,?x:Nat)))),?y:Nat) = min(?x:Nat,?y:Nat)",
"min(s(min(?x:Nat,min(?x:Nat,?y:Nat))),s(?x:Nat)) = s(s(?x:Nat))",
"min(s(min(s(?y:Nat),?y:Nat)),min(?x:Nat,?x:Nat)) = s(s(?z:Nat))",
"min(min(?x:Nat,s(min(?z:Nat,?x:Nat))),s(?x:Nat)) = s(s(?x:Nat))",
"min(min(?y:Nat,min(min(0,0),0)),?z:Nat) = min(?y:Nat,?x:Nat)",
"min(min(s(?x:Nat),?x:Nat),min(0,s(?x:Nat))) = min(?y:Nat,?x:Nat)",
"min(min(min(?y:Nat,?x:Nat),min(0,?y:Nat)),0) = s(s(?y:Nat))",
"min(min(min(0,0),min(?z:Nat,0)),?x:Nat) = min(?y:Nat,?z:Nat)",
"min(?x:Nat,s(min(s(s(?y:Nat)),min(?x:Nat,0)))) = s(?x:Nat)",
"min(?x:Nat,s(min(min(?y:Nat,?x:Nat),min(?y:Nat,?y:Nat)))) = s(?y:Nat)",
"min(?x:Nat,s(min(min(s(?x:Nat),s(?y:Nat)),?y:Nat))) = s(?y:Nat)",
"min(?x:Nat,min(min(?x:Nat,?x:Nat),min(?z:Nat,s(?x:Nat)))) = s(?x:Nat)",
"min(?y:Nat,s(min(0,min(min(?x:Nat,?z:Nat),?x:Nat)))) = s(?x:Nat)",
"min(?y:Nat,min(min(0,?z:Nat),min(s(?x:Nat),?y:Nat))) = s(?x:Nat)",
"min(?y:Nat,min(min(s(min(?z:Nat,0)),?y:Nat),?y:Nat)) = s(?x:Nat)",
"min(?z:Nat,s(min(min(0,min(?x:Nat,?x:Nat)),?x:Nat))) = s(0)",
"min(s(?x:Nat),min(min(s(?z:Nat),?y:Nat),s(?x:Nat))) = s(?x:Nat)",
"min(s(?y:Nat),min(?y:Nat,s(min(s(?y:Nat),?z:Nat)))) = s(?y:Nat)",
"min(min(?z:Nat,s(?y:Nat)),min(min(?y:Nat,?x:Nat),?y:Nat)) = s(?z:Nat)",
"min(min(0,min(s(min(?y:Nat,?x:Nat)),0)),?y:Nat) = s(?x:Nat)",
"min(min(s(?x:Nat),?y:Nat),min(min(?y:Nat,?y:Nat),?x:Nat)) = s(?y:Nat)",
"min(min(s(min(?y:Nat,min(?y:Nat,0))),?x:Nat),?y:Nat) = s(?z:Nat)",
"s(?y:Nat) = min(?x:Nat,min(min(0,s(s(0))),s(s(?y:Nat))))",
"s(?y:Nat) = min(?x:Nat,min(min(s(s(?z:Nat)),?x:Nat),min(?x:Nat,?x:Nat)))",
"s(?y:Nat) = min(s(min(?x:Nat,s(s(?z:Nat)))),s(min(?y:Nat,?z:Nat)))",
"s(?y:Nat) = min(min(0,s(min(?x:Nat,s(?z:Nat)))),min(?x:Nat,?y:Nat))",
"s(?y:Nat) = min(min(s(?x:Nat),min(min(0,s(0)),?y:Nat)),?x:Nat)",
"s(?z:Nat) = min(s(s(?z:Nat)),s(min(s(min(?x:Nat,?z:Nat)),?x:Nat)))",
"s(?z:Nat) = min(min(?x:Nat,min(s(?y:Nat),s(min(?x:Nat,0)))),0)",
"s(?z:Nat) = min(min(?y:Nat,s(min(?y:Nat,min(?y:Nat,0)))),s(?y:Nat))",
"s(0) = min(min(min(?x:Nat,min(0,?y:Nat)),?x:Nat),s(s(?y:Nat)))",
"min(?x:Nat,?z:Nat) = min(?x:Nat,min(s(0),min(min(?y:Nat,?x:Nat),0)))",
"min(?x:Nat,?z:Nat) = min(min(?z:Nat,s(s(s(?x:Nat)))),min(?y:Nat,?y:Nat))",
"min(?y:Nat,?x:Nat) = min(min(?x:Nat,?x:Nat),s(s(min(s(?x:Nat),?x:Nat))))",
"min(?y:Nat,?y:Nat) = s(min(s(min(?x:Nat,s(?x:Nat))),min(?y:Nat,?y:Nat)))",
"min(?y:Nat,?y:Nat) = min(s(min(min(s(?y:Nat),?x:Nat),?z:Nat)),s(?y:Nat))",
"min(?y:Nat,?y:Nat) = min(min(0,s(min(?x:Nat,min(?z:Nat,?x:Nat)))),?x:Nat)",
"min(?y:Nat,?z:Nat) = min(?x:Nat,min(min(0,s(min(?x:Nat,?x:Nat))),?x:Nat))",
"min(0,?z:Nat) = min(min(s(min(s(0),?x:Nat)),?x:Nat),s(?y:Nat))",
"min(0,0) = min(s(?z:Nat),min(s(min(s(?y:Nat),?y:Nat)),?x:Nat))",
"s(min(?z:Nat,?x:Nat)) = min(min(min(?y:Nat,?z:Nat),?y:Nat),min(?y:Nat,?x:Nat))",
"min(?x:Nat,s(?x:Nat)) = min(?x:Nat,min(min(min(?y:Nat,?z:Nat),?x:Nat),?z:Nat))",
"min(0,s(?x:Nat)) = min(min(s(min(?y:Nat,s(?x:Nat))),?x:Nat),?x:Nat)",
"min(s(?y:Nat),?z:Nat) = min(min(?x:Nat,?x:Nat),s(s(min(0,?x:Nat))))",
"min(s(?z:Nat),?x:Nat) = min(min(s(?x:Nat),min(s(?x:Nat),?x:Nat)),?y:Nat)",
"min(s(0),?x:Nat) = min(s(min(?z:Nat,s(min(?x:Nat,0)))),?z:Nat)",
"s(s(min(?z:Nat,?x:Nat))) = min(min(s(min(?x:Nat,?x:Nat)),?x:Nat),?z:Nat)",
"min(?x:Nat,min(?y:Nat,?x:Nat)) = min(min(0,?z:Nat),min(?x:Nat,s(?x:Nat)))",
"min(?z:Nat,min(?y:Nat,?y:Nat)) = min(s(min(?x:Nat,s(s(?x:Nat)))),?z:Nat)",
"min(s(?z:Nat),s(?z:Nat)) = s(s(min(s(s(?x:Nat)),s(?y:Nat))))",
"min(s(s(?x:Nat)),?y:Nat) = min(?z:Nat,min(s(s(?y:Nat)),s(?y:Nat)))",
"min(min(0,?x:Nat),?y:Nat) = min(s(min(?x:Nat,?x:Nat)),min(?x:Nat,?y:Nat))",
"s(min(min(?y:Nat,?x:Nat),?y:Nat)) = min(?y:Nat,min(?z:Nat,min(?z:Nat,?x:Nat)))",
"min(?z:Nat,min(s(?x:Nat),?y:Nat)) = min(?z:Nat,min(?x:Nat,min(?z:Nat,?z:Nat)))",
"min(0,min(?y:Nat,s(?x:Nat))) = s(s(min(min(?y:Nat,?x:Nat),0)))",
"min(s(min(s(0),?x:Nat)),?y:Nat) = min(?y:Nat,s(min(?x:Nat,?z:Nat)))",
"min(min(?z:Nat,s(s(?z:Nat))),?z:Nat) = min(s(0),min(?x:Nat,?x:Nat))",
"min(min(s(s(?y:Nat)),?x:Nat),?x:Nat) = min(?y:Nat,s(min(?y:Nat,?y:Nat)))",
"min(min(min(?z:Nat,0),0),?x:Nat) = min(s(?y:Nat),min(?x:Nat,?z:Nat))",
"s(s(min(?y:Nat,min(s(?x:Nat),0)))) = min(s(s(?x:Nat)),?x:Nat)",
"min(?x:Nat,min(?y:Nat,min(s(?z:Nat),?x:Nat))) = min(?y:Nat,min(?y:Nat,?y:Nat))",
"min(?y:Nat,s(min(0,min(?x:Nat,?x:Nat)))) = min(s(s(?y:Nat)),?x:Nat)",
"min(s(s(min(?x:Nat,s(0)))),?x:Nat) = min(min(0,?y:Nat),0)",
"min(s(min(?z:Nat,0)),s(s(?x:Nat))) = min(s(?x:Nat),s(?x:Nat))",
"min(min(s(s(?x:Nat)),s(?z:Nat)),?y:Nat) = min(s(?x:Nat),s(?z:Nat))",
"s(min(s(s(s(0))),min(0,?y:Nat))) = min(?z:Nat,s(0))",
"min(0,s(s(min(min(?x:Nat,?z:Nat),0)))) = s(s(s(?x:Nat)))",
"s(s(min(?x:Nat,min(min(?z:Nat,?x:Nat),s(?x:Nat))))) = min(?x:Nat,?x:Nat)",
"s(min(min(?x:Nat,min(s(?x:Nat),?z:Nat)),s(?z:Nat))) = min(?x:Nat,0)",
"min(0,s(min(?y:Nat,min(?x:Nat,min(?z:Nat,?y:Nat))))) = min(?y:Nat,?z:Nat)",
"min(s(?x:Nat),s(min(min(0,s(?y:Nat)),?x:Nat))) = min(?z:Nat,?x:Nat)",
"min(s(?x:Nat),min(?y:Nat,min(s(s(?x:Nat)),?y:Nat))) = min(?z:Nat,?y:Nat)",
"min(s(min(?y:Nat,min(?x:Nat,min(?x:Nat,?z:Nat)))),?y:Nat) = s(s(?x:Nat))",
"min(min(?x:Nat,s(min(min(?y:Nat,?y:Nat),?z:Nat))),?y:Nat) = min(?y:Nat,?y:Nat)",
"min(min(?z:Nat,min(s(s(?x:Nat)),?x:Nat)),s(0)) = s(s(0))",
"min(min(s(?x:Nat),s(min(?y:Nat,s(?y:Nat)))),0) = min(?z:Nat,?x:Nat)",
"min(min(s(?y:Nat),?y:Nat),min(min(?z:Nat,?y:Nat),?y:Nat)) = min(?x:Nat,?x:Nat)",
"min(min(s(s(?z:Nat)),min(?x:Nat,0)),s(?y:Nat)) = min(?y:Nat,?x:Nat)",
"min(min(min(?y:Nat,?y:Nat),s(?y:Nat)),s(s(?x:Nat))) = min(0,?y:Nat)",
"min(?x:Nat,min(?z:Nat,min(min(s(?z:Nat),?y:Nat),s(?y:Nat)))) = s(?x:Nat)",
"min(?z:Nat,min(min(?y:Nat,min(?y:Nat,min(?y:Nat,?y:Nat))),?x:Nat)) = s(0)",
"min(?z:Nat,min(min(?y:Nat,min(min(0,?y:Nat),0)),?y:Nat)) = s(?y:Nat)",
"min(s(?y:Nat),s(min(s(min(0,?y:Nat)),s(?y:Nat)))) = s(?y:Nat)",
"min(min(?y:Nat,?x:Nat),min(?x:Nat,s(s(min(?y:Nat,?x:Nat))))) = s(?y:Nat)",
"min(min(0,s(min(0,min(?z:Nat,0)))),s(?y:Nat)) = s(?z:Nat)",
"min(min(s(s(?y:Nat)),?z:Nat),s(min(?x:Nat,s(?z:Nat)))) = s(0)",
"min(min(s(s(?y:Nat)),s(s(0))),min(?x:Nat,?y:Nat)) = s(?y:Nat)",
"min(min(min(?z:Nat,s(?y:Nat)),?x:Nat),min(?y:Nat,s(?x:Nat))) = s(?y:Nat)",
"s(?x:Nat) = min(?z:Nat,min(min(?y:Nat,?z:Nat),s(min(?z:Nat,min(?x:Nat,0)))))",
"s(?x:Nat) = min(s(s(?y:Nat)),min(?x:Nat,min(s(?x:Nat),min(0,?x:Nat))))",
"s(?x:Nat) = min(min(s(0),min(0,0)),min(?x:Nat,s(s(?y:Nat))))",
"s(0) = min(s(min(min(min(?y:Nat,?y:Nat),min(?z:Nat,?y:Nat)),?y:Nat)),?y:Nat)",
"s(0) = min(min(?y:Nat,?x:Nat),s(min(?x:Nat,min(0,min(?y:Nat,?x:Nat)))))",
"s(s(?z:Nat)) = min(s(?y:Nat),min(?y:Nat,min(s(min(?x:Nat,0)),?x:Nat)))",
"min(?x:Nat,?y:Nat) = min(?x:Nat,min(min(?y:Nat,min(min(0,0),?x:Nat)),?y:Nat))",
"min(?x:Nat,?y:Nat) = min(min(min(min(0,min(?z:Nat,0)),?x:Nat),?x:Nat),?z:Nat)",
"min(?z:Nat,?z:Nat) = min(min(s(?x:Nat),?y:Nat),s(s(s(min(?x:Nat,?y:Nat)))))",
"min(0,?y:Nat) = min(min(?x:Nat,?x:Nat),min(min(?z:Nat,s(?x:Nat)),s(?y:Nat)))",
"min(?x:Nat,s(?y:Nat)) = s(min(s(min(s(?x:Nat),min(?x:Nat,?x:Nat))),?x:Nat))",
"min(s(?x:Nat),?x:Nat) = s(s(min(min(s(?y:Nat),min(0,?z:Nat)),0)))",
"min(s(?x:Nat),?x:Nat) = min(?y:Nat,s(min(s(?y:Nat),s(min(0,?y:Nat)))))",
"s(s(min(?x:Nat,?z:Nat))) = min(?y:Nat,s(min(?y:Nat,min(?x:Nat,s(?x:Nat)))))",
"s(min(?x:Nat,s(?x:Nat))) = min(min(?x:Nat,min(?x:Nat,min(?x:Nat,?z:Nat))),0)",
"min(?z:Nat,min(?z:Nat,0)) = min(?x:Nat,s(min(s(min(0,?x:Nat)),?y:Nat)))",
"min(s(s(?x:Nat)),?z:Nat) = min(0,min(?x:Nat,min(min(?y:Nat,?y:Nat),?x:Nat)))",
"min(s(s(?y:Nat)),?z:Nat) = min(s(?y:Nat),min(s(s(s(?x:Nat))),?x:Nat))",
"min(min(?x:Nat,?x:Nat),?x:Nat) = min(?x:Nat,min(min(?y:Nat,0),min(?y:Nat,?x:Nat)))",
"min(min(?y:Nat,?y:Nat),?y:Nat) = min(s(s(s(?z:Nat))),min(0,s(?x:Nat)))",
"min(min(0,?x:Nat),?y:Nat) = s(s(min(min(min(?y:Nat,0),?y:Nat),?y:Nat)))",
"min(min(0,?x:Nat),0) = s(min(min(s(?x:Nat),?y:Nat),min(?x:Nat,?x:Nat)))",
"s(min(?x:Nat,min(?z:Nat,?x:Nat))) = min(?y:Nat,s(min(?z:Nat,min(0,?x:Nat))))",
"min(s(?y:Nat),min(?y:Nat,?y:Nat)) = min(min(s(0),?y:Nat),min(?y:Nat,?x:Nat))",
"min(?x:Nat,s(min(0,s(?z:Nat)))) = min(0,s(min(?z:Nat,s(?z:Nat))))",
"min(s(s(?x:Nat)),s(min(?z:Nat,?x:Nat))) = min(min(?x:Nat,?y:Nat),s(0))",
"min(min(?z:Nat,0),min(?z:Nat,s(?x:Nat))) = min(?z:Nat,min(?x:Nat,s(?x:Nat)))",
"min(min(s(s(0)),?z:Nat),s(0)) = min(min(0,0),s(?y:Nat))",
"min(min(min(?y:Nat,?y:Nat),s(?y:Nat)),?y:Nat) = min(?y:Nat,s(min(?x:Nat,?x:Nat)))",
"min(min(min(?z:Nat,0),?x:Nat),s(?x:Nat)) = s(s(min(?y:Nat,s(0))))",
"s(s(min(?x:Nat,min(?y:Nat,min(?y:Nat,?x:Nat))))) = min(?y:Nat,s(s(?y:Nat)))",
"min(?x:Nat,min(?x:Nat,s(min(s(?y:Nat),?z:Nat)))) = min(?x:Nat,min(0,?y:Nat))",
"min(?y:Nat,min(s(min(0,?x:Nat)),s(?y:Nat))) = min(min(?z:Nat,0),?y:Nat)",
"min(s(?x:Nat),s(s(min(s(?z:Nat),?y:Nat)))) = min(min(?x:Nat,?y:Nat),?x:Nat)",
"min(s(?y:Nat),min(s(s(?y:Nat)),s(?y:Nat))) = s(min(?x:Nat,s(?x:Nat)))",
"min(s(s(?x:Nat)),min(min(0,?y:Nat),?y:Nat)) = s(min(s(?x:Nat),0))",
"min(min(0,min(?z:Nat,min(?y:Nat,?x:Nat))),?x:Nat) = min(min(?x:Nat,?y:Nat),?x:Nat)",
"min(?x:Nat,min(?y:Nat,min(?z:Nat,min(s(?x:Nat),?y:Nat)))) = s(min(?y:Nat,?z:Nat))",
"min(?x:Nat,min(?y:Nat,min(s(?y:Nat),s(s(?x:Nat))))) = s(min(0,0))",
"min(s(?x:Nat),s(s(min(min(0,0),?x:Nat)))) = min(s(?y:Nat),?y:Nat)",
"min(s(?x:Nat),min(s(?y:Nat),s(min(?x:Nat,?y:Nat)))) = s(min(?z:Nat,?x:Nat))",
"min(min(?x:Nat,min(?y:Nat,s(?x:Nat))),min(?y:Nat,?y:Nat)) = min(s(?y:Nat),?y:Nat)",
"min(min(s(min(?y:Nat,?x:Nat)),min(?y:Nat,?x:Nat)),?z:Nat) = s(min(?x:Nat,?y:Nat))",
"s(min(s(?x:Nat),min(min(?y:Nat,?y:Nat),s(s(?y:Nat))))) = min(?x:Nat,?y:Nat)",
"s(min(s(?x:Nat),min(min(s(?x:Nat),s(?x:Nat)),?y:Nat))) = min(?x:Nat,?y:Nat)",
"min(?y:Nat,min(s(?z:Nat),min(?y:Nat,s(min(?x:Nat,?y:Nat))))) = min(?z:Nat,0)",
"min(?y:Nat,min(min(?x:Nat,min(?x:Nat,s(?z:Nat))),s(?z:Nat))) = min(?x:Nat,?z:Nat)",
"min(?y:Nat,min(min(?z:Nat,min(?z:Nat,?z:Nat)),min(?y:Nat,?z:Nat))) = min(?z:Nat,?x:Nat)",
"min(s(?x:Nat),s(min(?x:Nat,min(?z:Nat,min(?x:Nat,?z:Nat))))) = min(0,?y:Nat)",
"min(s(?y:Nat),min(?z:Nat,min(?x:Nat,min(?z:Nat,s(?y:Nat))))) = min(?y:Nat,?y:Nat)",
"min(min(?x:Nat,?x:Nat),min(?y:Nat,s(s(min(?x:Nat,?y:Nat))))) = s(s(?z:Nat))",
"min(min(min(?z:Nat,s(?x:Nat)),min(s(?x:Nat),?z:Nat)),?x:Nat) = s(s(?z:Nat))",
"min(min(min(s(?y:Nat),min(0,s(?z:Nat))),?z:Nat),0) = s(s(0))",
"min(?y:Nat,min(s(?x:Nat),min(min(s(?z:Nat),?y:Nat),s(?y:Nat)))) = s(?y:Nat)",
"min(s(?y:Nat),s(s(min(min(?z:Nat,?y:Nat),min(?y:Nat,?y:Nat))))) = s(?y:Nat)",
"min(s(min(min(min(?x:Nat,?y:Nat),min(?y:Nat,?y:Nat)),?x:Nat)),0) = s(?x:Nat)",
"min(min(?x:Nat,min(s(0),?z:Nat)),min(?y:Nat,min(?x:Nat,?x:Nat))) = s(?z:Nat)",
"min(min(?z:Nat,min(min(s(?x:Nat),min(?z:Nat,?y:Nat)),?z:Nat)),0) = s(?y:Nat)",
"min(min(s(?z:Nat),?y:Nat),min(s(?z:Nat),s(min(?y:Nat,?z:Nat)))) = s(?y:Nat)",
"min(min(min(s(min(?y:Nat,0)),min(?y:Nat,0)),?x:Nat),?x:Nat) = s(?y:Nat)",
"s(?y:Nat) = min(s(min(min(min(?x:Nat,?y:Nat),min(?z:Nat,0)),?x:Nat)),s(?x:Nat))",
"s(?z:Nat) = min(?x:Nat,min(?z:Nat,min(min(0,min(0,?y:Nat)),s(s(?x:Nat)))))",
"s(0) = min(s(min(?x:Nat,s(?y:Nat))),min(s(s(min(0,?x:Nat))),0))",
"s(0) = min(min(?y:Nat,min(min(?x:Nat,?y:Nat),min(?y:Nat,?y:Nat))),s(s(?y:Nat)))",
"s(s(?x:Nat)) = min(min(?z:Nat,s(0)),min(min(0,min(?y:Nat,0)),?y:Nat))",
"s(s(?y:Nat)) = min(?x:Nat,s(min(s(min(?y:Nat,?x:Nat)),s(min(?z:Nat,?x:Nat)))))",
"s(s(?y:Nat)) = min(0,min(s(s(?y:Nat)),min(s(?z:Nat),min(?x:Nat,?z:Nat))))",
"s(s(?y:Nat)) = min(0,min(min(min(?x:Nat,min(?x:Nat,?y:Nat)),?y:Nat),s(0)))",
"min(?x:Nat,?x:Nat) = min(min(min(?y:Nat,s(?z:Nat)),?y:Nat),min(s(s(?y:Nat)),?y:Nat))",
"min(?y:Nat,?y:Nat) = s(min(min(min(s(?x:Nat),min(?x:Nat,0)),s(?x:Nat)),?y:Nat))",
"min(0,?x:Nat) = min(s(?x:Nat),s(min(min(min(?z:Nat,?z:Nat),?y:Nat),s(?y:Nat))))",
"min(0,?z:Nat) = s(min(s(min(?z:Nat,?x:Nat)),s(min(min(?z:Nat,?x:Nat),?z:Nat))))",
"s(min(?x:Nat,?y:Nat)) = min(s(min(s(0),min(?y:Nat,min(?x:Nat,?z:Nat)))),?x:Nat)",
"min(s(?y:Nat),?z:Nat) = min(s(?y:Nat),min(s(min(s(0),s(?z:Nat))),?y:Nat))",
"min(s(0),?x:Nat) = min(min(?y:Nat,?z:Nat),min(?x:Nat,s(min(?y:Nat,s(?x:Nat)))))",
"min(min(?x:Nat,?x:Nat),?x:Nat) = min(min(0,s(?x:Nat)),min(min(?y:Nat,?y:Nat),?y:Nat))",
"min(min(?y:Nat,0),?x:Nat) = min(s(0),s(min(s(min(?y:Nat,0)),?z:Nat)))",
"s(s(min(?y:Nat,s(?y:Nat)))) = min(?x:Nat,min(?x:Nat,s(min(s(?y:Nat),?y:Nat))))",
"s(min(min(?y:Nat,?y:Nat),?y:Nat)) = min(min(min(min(?x:Nat,?x:Nat),0),0),0)",
"min(?y:Nat,s(min(?y:Nat,?x:Nat))) = min(min(?x:Nat,min(?x:Nat,?y:Nat)),min(?x:Nat,?x:Nat))",
"min(s(?x:Nat),min(?y:Nat,?x:Nat)) = min(0,s(s(min(?x:Nat,min(?y:Nat,?y:Nat)))))",
"s(min(s(s(s(?y:Nat))),?z:Nat)) = min(min(min(?z:Nat,?x:Nat),?z:Nat),s(?x:Nat))",
"s(min(s(?z:Nat),min(?x:Nat,min(?y:Nat,?z:Nat)))) = min(min(0,0),s(?x:Nat))",
"min(s(?x:Nat),min(s(min(?z:Nat,?y:Nat)),?z:Nat)) = min(s(s(?z:Nat)),s(?z:Nat))",
"min(s(0),min(s(?y:Nat),s(s(?y:Nat)))) = min(s(?z:Nat),min(?y:Nat,?x:Nat))",
"min(min(min(?y:Nat,s(s(?y:Nat))),?y:Nat),0) = min(s(min(?y:Nat,?y:Nat)),?z:Nat)",
"s(min(s(min(?y:Nat,min(?y:Nat,?x:Nat))),s(?x:Nat))) = min(?y:Nat,min(?x:Nat,?y:Nat))",
"min(?x:Nat,min(min(min(?z:Nat,s(0)),?y:Nat),?x:Nat)) = min(0,s(s(?y:Nat)))",
"min(min(min(s(min(0,?z:Nat)),?y:Nat),?z:Nat),?y:Nat) = s(s(s(s(?y:Nat))))",
"min(?y:Nat,s(min(min(0,?y:Nat),min(?z:Nat,s(?x:Nat))))) = min(s(?y:Nat),0)",
"min(s(min(min(s(0),min(?z:Nat,?x:Nat)),?x:Nat)),0) = s(min(?x:Nat,?y:Nat))",
"min(min(s(s(?z:Nat)),s(min(0,s(?x:Nat)))),?x:Nat) = s(min(?y:Nat,?z:Nat))",
"min(min(s(s(s(?x:Nat))),0),min(?y:Nat,s(?x:Nat))) = s(min(?z:Nat,?z:Nat))",
"min(?x:Nat,min(min(?x:Nat,?y:Nat),min(s(?x:Nat),min(0,?y:Nat)))) = min(?y:Nat,?x:Nat)",
"min(0,min(s(min(?y:Nat,s(?x:Nat))),min(s(?x:Nat),0))) = min(?y:Nat,?x:Nat)",
"min(s(?y:Nat),min(s(0),min(?y:Nat,min(s(?z:Nat),0)))) = min(?y:Nat,?x:Nat)",
"min(s(min(s(?y:Nat),min(?y:Nat,min(?x:Nat,?y:Nat)))),s(?y:Nat)) = min(0,0)",
"min(min(s(?y:Nat),?y:Nat),s(s(min(min(?x:Nat,?x:Nat),0)))) = min(?y:Nat,?y:Nat)",
"min(s(min(?x:Nat,?x:Nat)),min(?z:Nat,s(min(0,min(?y:Nat,0))))) = s(?y:Nat)",
"min(min(?y:Nat,s(?y:Nat)),min(min(min(?x:Nat,s(?z:Nat)),0),?x:Nat)) = s(?z:Nat)",
"min(min(min(min(?y:Nat,min(?z:Nat,?x:Nat)),?y:Nat),s(s(?y:Nat))),?y:Nat) = s(0)",
"s(?x:Nat) = min(min(min(min(?x:Nat,?x:Nat),s(min(?x:Nat,?x:Nat))),?z:Nat),min(?y:Nat,?x:Nat))",
"s(?y:Nat) = min(min(min(?x:Nat,?y:Nat),s(s(?z:Nat))),min(min(s(0),?x:Nat),?x:Nat))",
"s(?z:Nat) = min(s(min(s(0),min(s(0),min(?x:Nat,?x:Nat)))),min(?z:Nat,?z:Nat))",
"s(0) = min(min(s(?z:Nat),?x:Nat),min(s(?z:Nat),min(0,s(min(0,?z:Nat)))))",
"min(?z:Nat,?z:Nat) = s(s(min(min(?x:Nat,min(?x:Nat,0)),min(s(?x:Nat),s(?x:Nat)))))",
"s(min(?x:Nat,?y:Nat)) = min(?y:Nat,s(min(min(min(0,?x:Nat),min(?x:Nat,?x:Nat)),?z:Nat)))",
"s(min(?z:Nat,s(?z:Nat))) = min(min(?y:Nat,min(min(?z:Nat,?z:Nat),s(?y:Nat))),s(?x:Nat))",
"min(s(min(?y:Nat,?x:Nat)),?y:Nat) = s(min(min(min(?y:Nat,s(?z:Nat)),s(?z:Nat)),?x:Nat))",
"min(s(min(0,?x:Nat)),?y:Nat) = min(min(min(?x:Nat,min(s(0),?x:Nat)),?z:Nat),?y:Nat)",
"min(min(?x:Nat,?z:Nat),s(?x:Nat)) = min(min(s(s(min(?x:Nat,?y:Nat))),s(?z:Nat)),?x:Nat)",
"s(s(s(min(s(?x:Nat),?x:Nat)))) = min(min(s(0),min(0,s(?y:Nat))),?x:Nat)",
"min(0,s(min(s(?z:Nat),?z:Nat))) = min(min(s(?x:Nat),s(0)),min(?x:Nat,?x:Nat))",
"min(s(s(min(0,?x:Nat))),?x:Nat) = min(min(min(?y:Nat,?x:Nat),0),min(?y:Nat,?z:Nat))",
"min(min(?x:Nat,?y:Nat),min(?x:Nat,?y:Nat)) = min(s(min(min(s(?x:Nat),0),?x:Nat)),?x:Nat)",
"min(min(0,?x:Nat),min(?x:Nat,?x:Nat)) = min(?y:Nat,min(min(s(?y:Nat),?x:Nat),s(?z:Nat)))",
"min(min(s(?z:Nat),s(?z:Nat)),0) = s(min(s(min(?y:Nat,min(0,0))),?y:Nat))",
"min(?x:Nat,min(s(?y:Nat),min(?x:Nat,?x:Nat))) = s(min(min(s(?y:Nat),s(0)),?z:Nat))",
"min(s(s(s(min(?x:Nat,0)))),?x:Nat) = min(min(min(?y:Nat,s(0)),0),?z:Nat)",
"min(min(min(s(?x:Nat),0),?y:Nat),?y:Nat) = min(?y:Nat,min(min(?x:Nat,?z:Nat),s(?y:Nat)))",
"min(?x:Nat,s(min(?z:Nat,min(?y:Nat,s(?y:Nat))))) = min(min(0,?y:Nat),s(s(?z:Nat)))",
"min(s(min(s(?x:Nat),s(s(?y:Nat)))),?y:Nat) = min(s(min(?x:Nat,s(0))),?x:Nat)",
"s(min(0,min(0,min(?x:Nat,min(?x:Nat,?y:Nat))))) = min(s(?z:Nat),min(?y:Nat,?z:Nat))",
"s(min(s(?y:Nat),min(min(0,?x:Nat),s(?x:Nat)))) = min(min(?x:Nat,s(?x:Nat)),?y:Nat)",
"min(min(s(min(s(0),?z:Nat)),?y:Nat),s(?z:Nat)) = min(s(s(?z:Nat)),s(?y:Nat))",
"min(?z:Nat,min(min(?y:Nat,min(?y:Nat,s(?y:Nat))),s(?x:Nat))) = s(min(s(0),?x:Nat))",
"min(s(0),min(s(min(min(?y:Nat,?y:Nat),0)),?z:Nat)) = min(min(?z:Nat,?y:Nat),?x:Nat)",
"min(min(?x:Nat,s(?z:Nat)),s(s(min(min(?x:Nat,0),?x:Nat)))) = s(min(?z:Nat,?x:Nat))",
"min(min(min(0,s(?x:Nat)),s(min(0,s(?x:Nat)))),?y:Nat) = min(s(?y:Nat),?y:Nat)",
"s(min(min(min(0,s(?x:Nat)),min(s(?x:Nat),?z:Nat)),s(?z:Nat))) = min(?x:Nat,?y:Nat)",
"s(min(min(min(min(?x:Nat,?z:Nat),min(?y:Nat,?y:Nat)),s(0)),?z:Nat)) = min(?x:Nat,?x:Nat)",
"min(s(min(min(?z:Nat,?x:Nat),min(min(?z:Nat,?z:Nat),?x:Nat))),s(?x:Nat)) = min(0,0)",
"min(min(?x:Nat,min(min(min(0,?z:Nat),s(?x:Nat)),s(?z:Nat))),0) = min(?y:Nat,?x:Nat)",
"min(min(s(?y:Nat),s(?z:Nat)),s(min(s(min(?x:Nat,0)),0))) = min(?y:Nat,?z:Nat)",
"min(min(min(?y:Nat,s(0)),min(min(?z:Nat,?x:Nat),s(?y:Nat))),?z:Nat) = min(?y:Nat,?y:Nat)",
"min(s(?z:Nat),min(s(min(?x:Nat,?x:Nat)),min(s(?x:Nat),min(?y:Nat,?y:Nat)))) = s(?x:Nat)",
"min(min(?x:Nat,?y:Nat),min(min(s(min(0,?y:Nat)),?x:Nat),min(?x:Nat,?y:Nat))) = s(0)",
"min(min(min(?y:Nat,s(min(?x:Nat,?y:Nat))),s(min(?x:Nat,s(?y:Nat)))),?z:Nat) = s(?y:Nat)",
"min(min(min(min(?x:Nat,?y:Nat),?y:Nat),min(?y:Nat,s(min(?x:Nat,?x:Nat)))),?x:Nat) = s(?y:Nat)",
"min(min(min(min(?z:Nat,min(0,?y:Nat)),?z:Nat),min(0,s(?z:Nat))),?x:Nat) = s(?z:Nat)",
"min(min(min(min(0,s(?z:Nat)),s(?y:Nat)),s(?x:Nat)),min(?z:Nat,?x:Nat)) = s(?y:Nat)",
"s(?x:Nat) = min(min(min(?x:Nat,min(min(?y:Nat,0),?z:Nat)),min(s(0),?x:Nat)),s(0))",
"s(?y:Nat) = min(min(?y:Nat,min(min(?y:Nat,s(?x:Nat)),min(min(?z:Nat,?x:Nat),s(?y:Nat)))),?y:Nat)",
"s(?y:Nat) = min(min(s(s(min(?y:Nat,0))),s(?x:Nat)),min(min(?y:Nat,s(?x:Nat)),0))",
"s(?y:Nat) = min(min(min(min(s(?x:Nat),0),min(?y:Nat,?y:Nat)),s(min(?x:Nat,?x:Nat))),?x:Nat)",
"s(s(?y:Nat)) = min(min(min(?x:Nat,min(min(?x:Nat,?z:Nat),min(?x:Nat,?x:Nat))),s(?z:Nat)),?y:Nat)",
"s(s(?z:Nat)) = min(min(?x:Nat,min(?x:Nat,?y:Nat)),min(0,min(s(?z:Nat),s(s(?y:Nat)))))",
"min(0,?z:Nat) = min(min(min(min(min(?x:Nat,?z:Nat),?y:Nat),?x:Nat),s(s(s(?z:Nat)))),?z:Nat)",
"s(min(s(?z:Nat),?z:Nat)) = min(min(min(min(?x:Nat,s(0)),s(?x:Nat)),s(?x:Nat)),?y:Nat)",
"min(?y:Nat,s(s(?y:Nat))) = min(min(min(s(s(?y:Nat)),s(s(?x:Nat))),s(?y:Nat)),?y:Nat)",
"min(min(0,?y:Nat),0) = s(min(min(?y:Nat,s(?x:Nat)),min(min(s(?y:Nat),?x:Nat),?y:Nat)))",
"min(min(s(?y:Nat),?x:Nat),?y:Nat) = min(s(0),s(s(min(s(0),min(?y:Nat,?y:Nat)))))",
"s(s(min(s(?z:Nat),s(?x:Nat)))) = min(min(?x:Nat,s(min(s(?x:Nat),s(?y:Nat)))),?y:Nat)",
"s(s(min(s(0),s(s(0))))) = min(s(s(min(?z:Nat,?z:Nat))),min(?x:Nat,?y:Nat))",
"s(min(0,min(s(s(0)),?x:Nat))) = min(min(?z:Nat,min(?x:Nat,s(s(?x:Nat)))),?z:Nat)",
"s(min(min(?z:Nat,s(?z:Nat)),s(?y:Nat))) = min(?y:Nat,min(min(s(?x:Nat),s(?y:Nat)),?z:Nat))",
"min(?y:Nat,min(?x:Nat,s(min(?y:Nat,?y:Nat)))) = min(?y:Nat,s(min(min(?z:Nat,?y:Nat),s(?y:Nat))))",
"min(0,s(min(0,s(s(?y:Nat))))) = min(s(min(?x:Nat,min(?x:Nat,?x:Nat))),s(?z:Nat))",
"min(min(min(0,0),0),s(?z:Nat)) = min(?x:Nat,min(?x:Nat,s(s(min(?x:Nat,?y:Nat)))))",
"min(?z:Nat,s(s(min(?y:Nat,min(0,?z:Nat))))) = min(s(?z:Nat),s(min(?y:Nat,s(0))))",
"min(min(?x:Nat,min(?z:Nat,min(?x:Nat,?x:Nat))),?y:Nat) = min(min(min(0,0),s(?x:Nat)),?z:Nat)",
"min(min(?y:Nat,?x:Nat),min(?y:Nat,min(?z:Nat,?y:Nat))) = min(?x:Nat,s(min(?y:Nat,min(?y:Nat,?y:Nat))))",
"min(min(?y:Nat,s(s(min(?x:Nat,?z:Nat)))),?x:Nat) = s(min(min(?x:Nat,s(s(?x:Nat))),0))",
"min(min(?z:Nat,s(min(?z:Nat,s(?x:Nat)))),0) = s(min(min(?z:Nat,?y:Nat),min(?y:Nat,?x:Nat)))",
"s(min(?x:Nat,min(min(?y:Nat,?y:Nat),s(s(?y:Nat))))) = min(min(min(?x:Nat,?x:Nat),0),0)",
"s(min(min(s(?y:Nat),s(min(?x:Nat,0))),?x:Nat)) = min(min(0,min(0,?x:Nat)),?x:Nat)",
"min(?y:Nat,s(min(?x:Nat,min(min(0,?x:Nat),?y:Nat)))) = min(s(s(min(?y:Nat,0))),?y:Nat)",
"min(s(min(?x:Nat,min(?y:Nat,?x:Nat))),min(?x:Nat,0)) = min(?x:Nat,min(min(?x:Nat,?y:Nat),?z:Nat))",
"min(min(0,min(?y:Nat,?x:Nat)),min(?x:Nat,s(?y:Nat))) = s(min(?x:Nat,s(min(0,?x:Nat))))",
"s(min(min(s(min(?x:Nat,0)),?x:Nat),min(?z:Nat,?x:Nat))) = min(?y:Nat,min(s(?z:Nat),?x:Nat))",
"min(s(s(?x:Nat)),s(min(min(?x:Nat,s(?x:Nat)),?z:Nat))) = min(s(?x:Nat),min(?y:Nat,?z:Nat))",
"min(min(min(min(?y:Nat,?y:Nat),?x:Nat),s(?z:Nat)),s(?y:Nat)) = s(min(?y:Nat,min(?z:Nat,?x:Nat)))",
"min(s(min(min(?z:Nat,?y:Nat),min(s(?x:Nat),s(?x:Nat)))),?x:Nat) = s(s(min(?x:Nat,0)))",
"min(s(min(min(min(0,?y:Nat),min(?z:Nat,?z:Nat)),0)),?z:Nat) = min(min(?y:Nat,0),?y:Nat)",
"min(min(s(?x:Nat),s(min(?z:Nat,?y:Nat))),min(?z:Nat,s(?y:Nat))) = min(?y:Nat,min(?z:Nat,0))",
"min(min(s(?y:Nat),min(s(s(?x:Nat)),?x:Nat)),s(s(?z:Nat))) = min(?y:Nat,min(?x:Nat,?x:Nat))",
"min(?x:Nat,min(min(s(?z:Nat),min(s(?x:Nat),s(?x:Nat))),s(?z:Nat))) = s(min(?y:Nat,?y:Nat))",
"min(s(s(0)),min(s(min(0,?z:Nat)),min(s(0),0))) = s(min(0,?y:Nat))",
"min(s(min(?z:Nat,?z:Nat)),min(min(s(?z:Nat),?z:Nat),min(?x:Nat,?x:Nat))) = min(s(?x:Nat),?z:Nat)",
"s(min(0,min(min(min(?x:Nat,?x:Nat),min(?y:Nat,?x:Nat)),min(?x:Nat,0)))) = min(0,?z:Nat)",
"min(?y:Nat,min(min(min(min(0,?y:Nat),s(?x:Nat)),0),s(s(?x:Nat)))) = s(s(?z:Nat))",
"min(?z:Nat,min(min(?y:Nat,s(s(0))),min(s(min(?x:Nat,?y:Nat)),?y:Nat))) = min(?x:Nat,?x:Nat)",
"min(s(s(?x:Nat)),min(?y:Nat,s(min(min(?z:Nat,?y:Nat),min(?y:Nat,?y:Nat))))) = min(?x:Nat,?y:Nat)",
"min(s(min(?x:Nat,?y:Nat)),s(min(s(s(?y:Nat)),s(min(?y:Nat,?y:Nat))))) = min(?y:Nat,?x:Nat)",
"min(?y:Nat,min(s(?x:Nat),min(min(min(?y:Nat,?y:Nat),?y:Nat),s(min(?y:Nat,?x:Nat))))) = s(0)",
"min(s(?x:Nat),min(min(?x:Nat,?x:Nat),min(min(?z:Nat,?x:Nat),s(min(0,?y:Nat))))) = s(0)",
"min(s(min(s(?x:Nat),min(?y:Nat,?y:Nat))),min(min(s(?y:Nat),?z:Nat),s(?x:Nat))) = s(?x:Nat)",
"min(?x:Nat,?x:Nat) = min(?x:Nat,s(min(min(?x:Nat,s(?x:Nat)),min(min(0,?y:Nat),min(?y:Nat,?x:Nat)))))",
"s(min(?x:Nat,?y:Nat)) = min(?x:Nat,min(s(min(?x:Nat,?x:Nat)),min(min(?z:Nat,0),s(s(?x:Nat)))))",
"min(s(s(?y:Nat)),?x:Nat) = min(?y:Nat,min(?y:Nat,min(s(s(?y:Nat)),min(?y:Nat,min(?y:Nat,?y:Nat)))))",
"min(?x:Nat,s(s(s(?y:Nat)))) = min(s(s(0)),min(?x:Nat,s(min(min(0,?x:Nat),?x:Nat))))",
"min(?y:Nat,s(min(?y:Nat,?z:Nat))) = s(min(0,min(min(?y:Nat,?x:Nat),min(?y:Nat,min(?y:Nat,?y:Nat)))))",
"min(s(min(?x:Nat,?x:Nat)),?y:Nat) = min(min(s(min(?y:Nat,?x:Nat)),s(?y:Nat)),min(s(?y:Nat),?x:Nat))",
"min(s(min(?x:Nat,?x:Nat)),0) = s(min(min(?y:Nat,?z:Nat),min(?x:Nat,min(?z:Nat,min(?x:Nat,?y:Nat)))))",
"s(min(?x:Nat,s(min(s(?y:Nat),?z:Nat)))) = min(min(s(0),?x:Nat),s(s(min(?x:Nat,?x:Nat))))",
"min(?z:Nat,min(min(s(s(?z:Nat)),?x:Nat),?y:Nat)) = min(s(min(?y:Nat,?x:Nat)),s(min(?x:Nat,0)))",
"s(min(min(?y:Nat,0),min(?x:Nat,min(?z:Nat,?z:Nat)))) = min(?x:Nat,min(s(?y:Nat),min(0,?x:Nat)))",
"s(s(min(min(min(?z:Nat,?x:Nat),?y:Nat),min(?x:Nat,?x:Nat)))) = min(min(min(?y:Nat,?y:Nat),0),?z:Nat)",
"min(?y:Nat,min(min(?x:Nat,?y:Nat),s(min(s(?y:Nat),s(?z:Nat))))) = s(min(0,min(0,?y:Nat)))",
"min(?x:Nat,min(min(min(min(?z:Nat,?y:Nat),?y:Nat),?y:Nat),s(min(?y:Nat,0)))) = s(min(?y:Nat,?y:Nat))",
"min(?z:Nat,min(min(?z:Nat,?x:Nat),min(min(s(?y:Nat),?y:Nat),min(s(?y:Nat),?z:Nat)))) = min(?y:Nat,?x:Nat)",
"min(min(?y:Nat,min(?y:Nat,s(min(0,0)))),min(?y:Nat,min(s(?x:Nat),?x:Nat))) = s(s(?x:Nat))",
"min(min(s(?x:Nat),min(?x:Nat,min(s(?x:Nat),min(?y:Nat,?z:Nat)))),s(s(0))) = min(0,?y:Nat)",
"min(min(min(s(0),s(min(?y:Nat,?x:Nat))),min(s(s(?y:Nat)),?y:Nat)),?z:Nat) = min(?z:Nat,?y:Nat)",
"min(min(s(?z:Nat),min(?y:Nat,?z:Nat)),min(min(min(?x:Nat,?y:Nat),min(?y:Nat,?x:Nat)),?z:Nat)) = s(?y:Nat)",
"min(min(s(min(?z:Nat,min(0,?y:Nat))),s(min(0,min(0,?y:Nat)))),s(?y:Nat)) = s(?x:Nat)",
"min(?x:Nat,0) = min(min(s(min(?y:Nat,min(?y:Nat,?x:Nat))),min(min(min(?z:Nat,?x:Nat),?y:Nat),?z:Nat)),?y:Nat)",
"s(min(?z:Nat,?x:Nat)) = min(min(?y:Nat,s(min(min(?y:Nat,?y:Nat),?y:Nat))),min(s(min(?z:Nat,?y:Nat)),?z:Nat))",
"min(?y:Nat,min(?y:Nat,s(0))) = min(?x:Nat,min(s(s(?y:Nat)),min(min(min(?x:Nat,?y:Nat),?y:Nat),?x:Nat)))",
"min(s(s(?y:Nat)),s(?x:Nat)) = min(?x:Nat,min(min(min(s(?z:Nat),0),?x:Nat),min(s(0),?x:Nat)))",
"min(?x:Nat,min(s(min(?y:Nat,0)),?z:Nat)) = min(?z:Nat,min(?x:Nat,min(min(?x:Nat,min(?x:Nat,0)),0)))",
"s(min(0,min(min(?x:Nat,?x:Nat),s(0)))) = min(min(s(?x:Nat),s(s(s(?y:Nat)))),s(?y:Nat))",
"min(?y:Nat,min(min(s(min(?y:Nat,?y:Nat)),?x:Nat),?z:Nat)) = min(s(min(?y:Nat,min(s(?x:Nat),0))),?x:Nat)",
"min(?z:Nat,min(s(min(s(?z:Nat),s(?z:Nat))),?z:Nat)) = min(min(0,min(min(?x:Nat,?z:Nat),?x:Nat)),0)",
"min(s(?x:Nat),min(?y:Nat,s(min(s(?y:Nat),?y:Nat)))) = s(min(?y:Nat,s(min(min(?y:Nat,?z:Nat),?x:Nat))))",
"s(s(min(min(min(?y:Nat,?y:Nat),?x:Nat),s(s(?z:Nat))))) = min(min(s(min(?x:Nat,?y:Nat)),?y:Nat),?z:Nat)",
"min(s(min(min(?y:Nat,min(?x:Nat,?y:Nat)),?y:Nat)),s(?z:Nat)) = min(?z:Nat,s(s(min(s(?x:Nat),0))))",
"min(min(?z:Nat,min(s(?x:Nat),s(min(?x:Nat,?z:Nat)))),0) = min(?y:Nat,s(s(min(s(?y:Nat),0))))",
"min(min(min(min(?x:Nat,?y:Nat),min(0,0)),?x:Nat),0) = min(min(s(?y:Nat),min(?x:Nat,?x:Nat)),?x:Nat)",
"min(?x:Nat,min(min(?x:Nat,s(s(0))),s(min(?z:Nat,?x:Nat)))) = min(s(?x:Nat),min(?z:Nat,s(0)))",
"min(s(s(s(0))),s(min(min(?x:Nat,?y:Nat),s(?z:Nat)))) = min(min(?y:Nat,?x:Nat),min(?z:Nat,0))",
"min(min(?y:Nat,s(min(?y:Nat,s(?x:Nat)))),s(min(?y:Nat,min(?x:Nat,s(?y:Nat))))) = s(min(?y:Nat,?x:Nat))",
"min(s(s(s(0))),min(min(min(?y:Nat,?x:Nat),s(min(?y:Nat,?x:Nat))),s(?x:Nat))) = min(?y:Nat,?x:Nat)",
"min(min(s(min(min(?x:Nat,?x:Nat),0)),min(s(?x:Nat),min(?y:Nat,min(?y:Nat,?x:Nat)))),?x:Nat) = s(?y:Nat)",
"min(min(min(s(?y:Nat),s(?z:Nat)),min(s(?x:Nat),0)),min(min(?z:Nat,s(?y:Nat)),?y:Nat)) = s(?y:Nat)",
"s(?x:Nat) = min(min(s(?y:Nat),?z:Nat),min(?y:Nat,min(min(?y:Nat,min(?y:Nat,0)),min(min(?x:Nat,?y:Nat),?y:Nat))))",
"min(s(0),min(s(s(?x:Nat)),0)) = s(min(?y:Nat,min(min(min(?z:Nat,?z:Nat),s(?y:Nat)),s(?x:Nat))))",
"s(s(s(min(min(0,?y:Nat),s(?x:Nat))))) = min(min(s(min(?y:Nat,min(0,0))),?y:Nat),s(?y:Nat))",
"min(min(?z:Nat,0),min(min(?z:Nat,s(s(?y:Nat))),s(?x:Nat))) = min(s(min(?x:Nat,s(s(?z:Nat)))),?x:Nat)",
"min(?y:Nat,min(s(min(?y:Nat,min(?x:Nat,?z:Nat))),min(min(s(?y:Nat),0),?y:Nat))) = min(?y:Nat,min(?y:Nat,?z:Nat))",
"s(min(min(?y:Nat,min(?x:Nat,0)),min(min(?x:Nat,0),min(s(?z:Nat),min(?y:Nat,?y:Nat))))) = min(?x:Nat,?x:Nat)",
"min(min(min(?y:Nat,min(min(?x:Nat,?z:Nat),min(?x:Nat,?z:Nat))),min(min(?x:Nat,?y:Nat),?x:Nat)),?y:Nat) = min(?x:Nat,0)",
"min(s(min(min(s(0),?z:Nat),min(0,s(?x:Nat)))),s(min(min(?x:Nat,s(?x:Nat)),?y:Nat))) = s(?y:Nat)",
"s(min(s(?z:Nat),min(?y:Nat,0))) = min(s(?y:Nat),min(min(min(s(?x:Nat),?y:Nat),s(0)),min(?y:Nat,?z:Nat)))",
"min(min(?y:Nat,min(s(?z:Nat),?y:Nat)),?z:Nat) = min(min(min(?x:Nat,?y:Nat),s(?x:Nat)),min(0,min(?z:Nat,s(?x:Nat))))",
"min(?y:Nat,min(?y:Nat,min(s(min(?x:Nat,?z:Nat)),?x:Nat))) = min(s(s(s(?y:Nat))),s(min(0,min(?x:Nat,?x:Nat))))",
"min(min(min(?x:Nat,s(?y:Nat)),s(?x:Nat)),min(s(?x:Nat),s(?x:Nat))) = min(?y:Nat,min(?y:Nat,min(s(?y:Nat),?y:Nat)))",
"s(min(min(min(min(?x:Nat,?x:Nat),?y:Nat),min(min(?x:Nat,0),?z:Nat)),?y:Nat)) = min(?x:Nat,min(min(?y:Nat,0),?x:Nat))",
"min(?x:Nat,min(min(s(?x:Nat),?z:Nat),min(min(min(0,?y:Nat),s(?z:Nat)),s(?z:Nat)))) = min(min(?x:Nat,?z:Nat),?x:Nat)",
"min(min(min(min(s(?x:Nat),?z:Nat),min(0,0)),min(s(s(?y:Nat)),?x:Nat)),?z:Nat) = s(s(min(?z:Nat,?x:Nat)))",
"min(min(min(min(s(?y:Nat),?x:Nat),s(s(?y:Nat))),?z:Nat),s(s(min(min(?x:Nat,0),?x:Nat)))) = min(?z:Nat,?x:Nat)",
"min(?y:Nat,min(?z:Nat,min(min(?y:Nat,0),?x:Nat))) = min(min(min(s(min(?y:Nat,?y:Nat)),min(?x:Nat,?x:Nat)),?y:Nat),s(?x:Nat))",
"min(s(min(min(?x:Nat,?z:Nat),min(0,?z:Nat))),?z:Nat) = min(?x:Nat,min(min(s(?z:Nat),min(s(?x:Nat),s(?z:Nat))),?y:Nat))",
"min(?x:Nat,min(min(?y:Nat,min(min(?y:Nat,?y:Nat),?x:Nat)),?z:Nat)) = min(min(0,min(0,min(min(?x:Nat,?z:Nat),?x:Nat))),?y:Nat)",
"min(0,min(min(s(s(?y:Nat)),min(?z:Nat,?x:Nat)),0)) = s(min(min(s(min(?x:Nat,?x:Nat)),min(?y:Nat,?y:Nat)),?y:Nat))",
"s(s(min(s(min(?y:Nat,?y:Nat)),s(?y:Nat)))) = min(?x:Nat,s(min(min(?x:Nat,?x:Nat),min(min(?y:Nat,?x:Nat),min(?y:Nat,?z:Nat)))))",
"min(min(?x:Nat,0),s(s(min(?x:Nat,?z:Nat)))) = min(min(min(s(s(?x:Nat)),?z:Nat),?x:Nat),s(s(min(?y:Nat,s(?y:Nat)))))",
"s(min(?z:Nat,min(s(min(?z:Nat,?y:Nat)),min(s(?z:Nat),s(?z:Nat))))) = min(?y:Nat,min(?y:Nat,min(min(0,?x:Nat),s(0))))",
"min(min(?x:Nat,?x:Nat),min(min(min(min(?x:Nat,0),?x:Nat),0),min(?z:Nat,s(?y:Nat)))) = s(s(s(min(?x:Nat,s(?z:Nat)))))",
"min(min(min(0,?z:Nat),?y:Nat),min(min(s(?z:Nat),min(0,min(?x:Nat,?z:Nat))),min(0,s(?y:Nat)))) = s(min(?y:Nat,?x:Nat))",
"min(min(min(?y:Nat,?x:Nat),min(?x:Nat,min(?z:Nat,min(?x:Nat,?x:Nat)))),min(s(s(s(?y:Nat))),min(?x:Nat,min(?y:Nat,?z:Nat)))) = s(?y:Nat)",
"min(s(?y:Nat),s(?x:Nat)) = min(min(min(min(s(?x:Nat),s(?y:Nat)),?z:Nat),?z:Nat),min(s(min(?y:Nat,?x:Nat)),min(0,min(?y:Nat,?y:Nat))))",
"min(s(s(min(s(?x:Nat),min(?y:Nat,?x:Nat)))),?y:Nat) = min(min(?x:Nat,s(?x:Nat)),min(min(min(?y:Nat,0),min(?x:Nat,?y:Nat)),s(?x:Nat)))",
"min(0,s(min(s(s(?x:Nat)),min(s(?x:Nat),?x:Nat)))) = min(?y:Nat,min(min(s(?y:Nat),min(min(?y:Nat,?x:Nat),s(?y:Nat))),s(?z:Nat)))",
"min(min(min(min(?x:Nat,?x:Nat),?x:Nat),s(min(?x:Nat,?z:Nat))),?y:Nat) = min(min(min(?z:Nat,s(s(0))),?z:Nat),s(s(s(s(?x:Nat)))))",
"s(min(0,s(min(?x:Nat,min(?y:Nat,?x:Nat))))) = min(min(min(0,?y:Nat),?x:Nat),s(min(min(?x:Nat,s(?y:Nat)),min(min(?y:Nat,?y:Nat),?x:Nat))))",
"min(s(?y:Nat),min(min(min(min(?x:Nat,?y:Nat),min(?y:Nat,?y:Nat)),?x:Nat),?x:Nat)) = min(s(min(min(s(?y:Nat),s(?y:Nat)),min(?x:Nat,?y:Nat))),0)",
"min(min(min(s(min(?y:Nat,?x:Nat)),?y:Nat),s(min(?x:Nat,min(0,?x:Nat)))),?z:Nat) = min(min(0,min(?y:Nat,min(s(?x:Nat),s(?x:Nat)))),?z:Nat)",
"min(min(min(s(s(?y:Nat)),s(min(?x:Nat,?x:Nat))),s(min(min(?z:Nat,?z:Nat),s(?y:Nat)))),?y:Nat) = s(min(min(min(?y:Nat,0),0),0))",
"min(min(min(min(?y:Nat,?z:Nat),min(s(?x:Nat),?y:Nat)),?y:Nat),s(s(s(?x:Nat)))) = min(min(?y:Nat,?z:Nat),min(?x:Nat,min(?x:Nat,min(s(?x:Nat),s(0)))))",
"min(min(s(min(min(?x:Nat,?x:Nat),0)),s(?y:Nat)),min(min(s(min(0,?x:Nat)),?y:Nat),min(?x:Nat,0))) = min(min(s(min(0,?y:Nat)),?y:Nat),?z:Nat)",
"min(min(s(min(?x:Nat,0)),0),s(?x:Nat)) = min(s(min(min(?y:Nat,s(?x:Nat)),s(min(?y:Nat,?z:Nat)))),min(?x:Nat,min(s(min(?z:Nat,?y:Nat)),s(?x:Nat))))",
"min(min(min(0,?x:Nat),s(s(s(?x:Nat)))),s(min(?y:Nat,s(?z:Nat)))) = min(?x:Nat,min(min(s(?x:Nat),min(?y:Nat,?y:Nat)),min(min(s(?y:Nat),?y:Nat),?x:Nat)))",
"s(min(min(min(s(?x:Nat),min(?y:Nat,?x:Nat)),0),min(?x:Nat,min(min(?y:Nat,?z:Nat),min(?y:Nat,?x:Nat))))) = min(s(min(min(min(?y:Nat,?x:Nat),s(?x:Nat)),?z:Nat)),min(min(0,?y:Nat),min(?x:Nat,?y:Nat)))"]

val problemsMaxMinPlusYes =
IOMstrs.rdEqs [
"s(?x:Nat) = plus(?x:Nat,max(0,s(0)))",
"min(0,?z:Nat) = min(plus(?x:Nat,0),0)",
"s(max(0,0)) = plus(0,s(0))",
"plus(0,?y:Nat) = min(s(max(?y:Nat,?x:Nat)),?y:Nat)",
"s(0) = min(s(0),s(plus(?z:Nat,s(?z:Nat))))",
"max(?z:Nat,?x:Nat) = max(plus(?z:Nat,0),plus(0,?x:Nat))",
"min(?x:Nat,0) = min(max(?z:Nat,plus(0,?y:Nat)),0)",
"max(?x:Nat,s(plus(?x:Nat,?x:Nat))) = plus(s(?x:Nat),?x:Nat)",
"max(?x:Nat,min(plus(?y:Nat,?x:Nat),?x:Nat)) = min(?x:Nat,?x:Nat)",
"s(plus(0,min(?y:Nat,min(0,?z:Nat)))) = s(0)",
"s(?x:Nat) = s(min(s(max(plus(?x:Nat,?x:Nat),?x:Nat)),?x:Nat))",
"plus(0,?x:Nat) = plus(min(s(min(s(?y:Nat),?y:Nat)),0),?x:Nat)",
"s(max(?z:Nat,plus(min(0,?x:Nat),0))) = s(max(?z:Nat,?z:Nat))",
"min(0,plus(s(plus(s(?x:Nat),0)),?z:Nat)) = min(0,?x:Nat)",
"s(?x:Nat) = max(s(?x:Nat),min(0,s(plus(?z:Nat,plus(0,?z:Nat)))))",
"s(?z:Nat) = plus(plus(?z:Nat,min(min(?x:Nat,s(?y:Nat)),0)),s(0))",
"s(s(?x:Nat)) = s(min(s(?x:Nat),plus(0,s(max(0,?x:Nat)))))",
"min(?x:Nat,min(s(?x:Nat),plus(plus(?x:Nat,?x:Nat),0))) = plus(?x:Nat,0)",
"s(?x:Nat) = s(min(plus(min(?x:Nat,s(?y:Nat)),s(?x:Nat)),max(?x:Nat,?x:Nat)))",
"s(?y:Nat) = s(min(plus(plus(min(?x:Nat,0),max(?y:Nat,?z:Nat)),?y:Nat),?y:Nat))",
"plus(0,0) = min(0,s(s(plus(plus(?z:Nat,?z:Nat),max(?y:Nat,?x:Nat)))))",
"min(?x:Nat,s(?x:Nat)) = min(?x:Nat,plus(plus(max(s(?x:Nat),0),?z:Nat),?z:Nat))",
"min(?x:Nat,plus(plus(?y:Nat,?x:Nat),max(?x:Nat,?x:Nat))) = max(?x:Nat,max(0,?x:Nat))",
"s(max(max(?x:Nat,plus(min(?z:Nat,?x:Nat),0)),?x:Nat)) = max(s(?x:Nat),?x:Nat)",
"s(min(?x:Nat,min(plus(s(?x:Nat),s(0)),?x:Nat))) = max(?x:Nat,s(?x:Nat))",
"plus(min(?z:Nat,0),min(0,min(plus(?x:Nat,0),?y:Nat))) = min(?z:Nat,0)",
"s(?x:Nat) = min(s(?x:Nat),plus(plus(s(?x:Nat),?x:Nat),min(s(s(?z:Nat)),?x:Nat)))",
"s(?y:Nat) = s(max(min(0,plus(0,plus(?y:Nat,?y:Nat))),min(?y:Nat,s(?y:Nat))))",
"s(s(?x:Nat)) = s(s(min(plus(min(?x:Nat,?x:Nat),0),max(?x:Nat,s(?y:Nat)))))",
"min(s(?x:Nat),?x:Nat) = min(plus(s(plus(plus(?x:Nat,?z:Nat),s(?z:Nat))),?x:Nat),?x:Nat)",
"s(min(s(?x:Nat),?x:Nat)) = min(s(?x:Nat),s(plus(?z:Nat,s(min(?x:Nat,?x:Nat)))))",
"min(max(?x:Nat,0),?x:Nat) = min(plus(s(min(?x:Nat,plus(?x:Nat,?y:Nat))),?x:Nat),?x:Nat)",
"min(plus(max(plus(s(?x:Nat),?x:Nat),?y:Nat),0),0) = min(0,min(?z:Nat,?z:Nat))",
"min(?x:Nat,s(plus(min(plus(?y:Nat,?x:Nat),min(?x:Nat,?x:Nat)),0))) = plus(?x:Nat,0)",
"plus(?y:Nat,?x:Nat) = plus(?y:Nat,max(min(?x:Nat,min(max(?y:Nat,?x:Nat),max(?x:Nat,0))),0))",
"plus(0,?x:Nat) = min(max(s(?z:Nat),plus(?x:Nat,min(max(?y:Nat,0),s(?x:Nat)))),?x:Nat)",
"min(0,plus(max(0,?y:Nat),plus(s(min(?x:Nat,?x:Nat)),?z:Nat))) = min(s(?y:Nat),0)",
"max(?y:Nat,?y:Nat) = plus(plus(?y:Nat,min(plus(s(?z:Nat),min(?x:Nat,?x:Nat)),max(0,0))),0)",
"min(max(0,s(?x:Nat)),?x:Nat) = min(?x:Nat,plus(max(plus(?x:Nat,min(0,?z:Nat)),0),?x:Nat))",
"min(0,max(?x:Nat,min(?z:Nat,min(?x:Nat,max(?x:Nat,?z:Nat))))) = plus(0,min(s(?z:Nat),0))",
"plus(min(?x:Nat,s(?z:Nat)),max(?x:Nat,s(max(min(?z:Nat,?x:Nat),?z:Nat)))) = plus(?x:Nat,s(?z:Nat))",
"min(s(max(0,?x:Nat)),plus(s(?x:Nat),min(plus(plus(?x:Nat,?y:Nat),?x:Nat),?z:Nat))) = s(?x:Nat)",
"min(?x:Nat,?x:Nat) = min(plus(s(min(min(0,?x:Nat),?y:Nat)),plus(s(max(?x:Nat,?y:Nat)),?x:Nat)),?x:Nat)",
"plus(?x:Nat,0) = min(?x:Nat,s(plus(plus(plus(?y:Nat,?z:Nat),min(0,?x:Nat)),s(plus(0,?x:Nat)))))",
"plus(s(min(0,0)),?x:Nat) = s(max(min(s(?x:Nat),?x:Nat),min(s(max(?x:Nat,?z:Nat)),0)))",
"max(s(?x:Nat),s(min(s(?x:Nat),0))) = max(0,max(0,max(plus(s(?x:Nat),0),?x:Nat)))",
"plus(?y:Nat,max(min(min(0,?z:Nat),0),plus(0,plus(?x:Nat,?x:Nat)))) = plus(plus(?y:Nat,?x:Nat),?x:Nat)",
"min(?x:Nat,plus(?x:Nat,?y:Nat)) = max(min(min(plus(?x:Nat,s(?z:Nat)),plus(?x:Nat,plus(?z:Nat,?y:Nat))),?x:Nat),?x:Nat)",
"s(min(min(max(?x:Nat,?x:Nat),s(plus(?x:Nat,0))),?x:Nat)) = min(s(max(0,?x:Nat)),s(s(?x:Nat)))",
"min(min(max(plus(plus(0,?y:Nat),plus(?x:Nat,?x:Nat)),?x:Nat),?y:Nat),0) = min(s(max(?z:Nat,?z:Nat)),0)",
"min(0,s(0)) = min(s(min(max(max(?x:Nat,?x:Nat),0),min(plus(?z:Nat,?x:Nat),max(?z:Nat,?x:Nat)))),0)",
"plus(?z:Nat,min(max(?z:Nat,?x:Nat),0)) = min(s(max(s(?y:Nat),plus(plus(?y:Nat,0),min(?z:Nat,?z:Nat)))),?z:Nat)",
"s(plus(min(plus(max(?z:Nat,?x:Nat),max(?y:Nat,0)),?y:Nat),?x:Nat)) = plus(plus(?x:Nat,0),s(min(?y:Nat,?y:Nat)))",
"plus(?y:Nat,min(min(0,0),plus(?x:Nat,max(s(?x:Nat),s(?z:Nat))))) = max(max(?y:Nat,max(?y:Nat,?y:Nat)),?y:Nat)",
"min(?x:Nat,plus(plus(max(plus(?z:Nat,?z:Nat),max(0,?z:Nat)),plus(?z:Nat,s(?y:Nat))),s(?x:Nat))) = max(?x:Nat,0)",
"s(min(?x:Nat,max(min(?x:Nat,0),?x:Nat))) = min(s(?x:Nat),max(s(plus(?y:Nat,0)),s(plus(?x:Nat,s(?z:Nat)))))",
"max(min(max(max(?z:Nat,?x:Nat),s(?y:Nat)),?z:Nat),?x:Nat) = max(max(max(min(?x:Nat,plus(?y:Nat,?x:Nat)),?x:Nat),?z:Nat),?z:Nat)",
"min(plus(plus(?y:Nat,s(min(?x:Nat,?z:Nat))),plus(?z:Nat,?y:Nat)),s(min(s(?z:Nat),?z:Nat))) = min(s(?z:Nat),s(?z:Nat))",
"plus(s(0),?z:Nat) = plus(plus(s(0),?z:Nat),min(min(s(s(?y:Nat)),min(plus(?z:Nat,?z:Nat),plus(0,0))),?z:Nat))",
"min(?x:Nat,max(?x:Nat,?z:Nat)) = min(min(plus(min(plus(?z:Nat,?z:Nat),max(?y:Nat,?y:Nat)),?x:Nat),plus(s(s(?x:Nat)),?x:Nat)),?x:Nat)",
"min(s(plus(?x:Nat,max(?x:Nat,plus(0,?z:Nat)))),0) = min(0,min(0,plus(plus(plus(?x:Nat,0),?x:Nat),s(?z:Nat))))",
"min(0,min(max(max(s(?z:Nat),?x:Nat),s(min(0,?y:Nat))),min(?z:Nat,plus(?x:Nat,min(0,0))))) = max(0,0)",
"s(?x:Nat) = min(plus(?z:Nat,plus(s(s(0)),?x:Nat)),min(s(?x:Nat),plus(plus(max(?z:Nat,?x:Nat),plus(?z:Nat,?x:Nat)),s(?x:Nat))))",
"s(plus(?x:Nat,0)) = min(s(max(plus(plus(?y:Nat,?x:Nat),?x:Nat),plus(min(?y:Nat,?x:Nat),min(?x:Nat,0)))),plus(?x:Nat,s(0)))",
"max(s(?x:Nat),min(?x:Nat,min(plus(max(?z:Nat,?y:Nat),max(?x:Nat,?x:Nat)),?x:Nat))) = max(s(?x:Nat),min(s(?x:Nat),plus(?z:Nat,0)))",
"s(0) = max(min(max(?z:Nat,s(s(0))),min(min(0,plus(?y:Nat,?x:Nat)),0)),s(min(0,max(0,plus(?y:Nat,?x:Nat)))))",
"s(plus(0,?x:Nat)) = max(min(min(s(min(?y:Nat,0)),?x:Nat),plus(s(min(?x:Nat,0)),min(min(?z:Nat,?z:Nat),?y:Nat))),s(?x:Nat))",
"min(?x:Nat,max(?x:Nat,max(s(min(?x:Nat,?z:Nat)),0))) = min(?x:Nat,plus(max(?x:Nat,?z:Nat),min(plus(0,?y:Nat),plus(s(?y:Nat),?z:Nat))))",
"plus(?x:Nat,min(?y:Nat,min(?z:Nat,min(s(?x:Nat),0)))) = min(plus(?x:Nat,plus(?y:Nat,plus(plus(?z:Nat,?y:Nat),s(0)))),max(0,?x:Nat))",
"max(?x:Nat,plus(min(0,plus(?y:Nat,plus(?x:Nat,0))),s(?x:Nat))) = s(min(?x:Nat,max(max(?z:Nat,?x:Nat),plus(0,min(?x:Nat,?z:Nat)))))",
"min(max(max(min(s(?y:Nat),?z:Nat),max(plus(?z:Nat,?x:Nat),s(?z:Nat))),max(?y:Nat,min(plus(?z:Nat,?y:Nat),0))),0) = min(0,?y:Nat)",
"plus(min(max(s(s(?x:Nat)),?x:Nat),?x:Nat),0) = min(plus(?x:Nat,max(plus(s(?x:Nat),?z:Nat),plus(max(?y:Nat,?y:Nat),min(?y:Nat,?y:Nat)))),?x:Nat)",
"s(?x:Nat) = s(max(min(plus(plus(?x:Nat,0),max(0,?x:Nat)),max(?x:Nat,?x:Nat)),max(min(0,min(?y:Nat,?x:Nat)),min(?x:Nat,plus(?x:Nat,?z:Nat)))))",
"min(?x:Nat,max(max(?z:Nat,?x:Nat),?y:Nat)) = min(max(plus(plus(?x:Nat,?y:Nat),?y:Nat),max(max(?y:Nat,?x:Nat),plus(max(0,0),min(?z:Nat,?x:Nat)))),?x:Nat)",
"min(s(s(min(?x:Nat,plus(0,?x:Nat)))),?x:Nat) = min(plus(min(?x:Nat,?y:Nat),plus(s(max(0,?z:Nat)),min(plus(?x:Nat,?z:Nat),s(?x:Nat)))),?x:Nat)",
"max(s(max(0,?x:Nat)),max(min(max(0,min(?x:Nat,?x:Nat)),s(0)),?x:Nat)) = plus(s(?x:Nat),min(max(max(s(0),0),?y:Nat),0))",
"min(max(min(s(?x:Nat),plus(plus(?z:Nat,?z:Nat),?x:Nat)),s(?x:Nat)),plus(0,?x:Nat)) = min(?x:Nat,plus(s(?z:Nat),s(plus(plus(?x:Nat,?x:Nat),?z:Nat))))",
"max(min(min(plus(0,max(?y:Nat,0)),0),?x:Nat),max(min(min(?x:Nat,plus(?z:Nat,?z:Nat)),s(min(0,?y:Nat))),?x:Nat)) = min(?x:Nat,s(?x:Nat))",
"min(plus(min(s(s(0)),min(?x:Nat,s(?x:Nat))),s(?x:Nat)),max(s(?x:Nat),min(plus(min(?x:Nat,?x:Nat),?y:Nat),?x:Nat))) = s(min(?x:Nat,?x:Nat))",
"s(max(?y:Nat,min(s(s(?z:Nat)),?y:Nat))) = s(max(min(?x:Nat,0),min(max(s(?y:Nat),plus(?x:Nat,?x:Nat)),max(plus(0,?y:Nat),min(?y:Nat,0)))))",
"min(min(?x:Nat,plus(s(plus(0,?x:Nat)),?y:Nat)),?x:Nat) = min(max(min(max(max(?y:Nat,0),?x:Nat),min(plus(?x:Nat,?y:Nat),plus(?x:Nat,?x:Nat))),?x:Nat),?x:Nat)",
"min(?z:Nat,plus(s(?x:Nat),plus(plus(plus(?x:Nat,?x:Nat),min(?z:Nat,?z:Nat)),?z:Nat))) = min(?z:Nat,plus(max(min(plus(0,0),plus(0,?x:Nat)),?z:Nat),?z:Nat))",
"plus(min(min(max(max(?x:Nat,?y:Nat),min(?x:Nat,?z:Nat)),s(min(0,?x:Nat))),max(0,max(min(?z:Nat,0),0))),s(?y:Nat)) = s(plus(0,?y:Nat))",
"s(?x:Nat) = min(min(s(?x:Nat),max(s(plus(?x:Nat,?y:Nat)),0)),max(s(max(plus(?x:Nat,0),min(0,?y:Nat))),max(?x:Nat,plus(max(?x:Nat,0),0))))",
"s(min(?x:Nat,s(plus(plus(?y:Nat,?x:Nat),max(0,?x:Nat))))) = s(max(?x:Nat,min(min(s(?z:Nat),min(?z:Nat,0)),min(min(?y:Nat,?z:Nat),max(?y:Nat,?z:Nat)))))",
"plus(min(plus(max(plus(?x:Nat,?y:Nat),s(?x:Nat)),plus(max(?x:Nat,?x:Nat),plus(?x:Nat,?y:Nat))),?y:Nat),0) = min(plus(max(s(plus(?x:Nat,?x:Nat)),?y:Nat),?z:Nat),?y:Nat)",
"plus(?z:Nat,min(min(plus(min(?y:Nat,0),plus(?x:Nat,0)),plus(min(?y:Nat,?y:Nat),plus(?z:Nat,?x:Nat))),max(max(max(?x:Nat,?z:Nat),s(0)),0))) = plus(?x:Nat,?z:Nat)",
"min(?x:Nat,plus(min(?y:Nat,s(min(?x:Nat,0))),s(?x:Nat))) = min(?x:Nat,plus(min(plus(max(?z:Nat,?x:Nat),?x:Nat),s(max(?x:Nat,?x:Nat))),max(?y:Nat,plus(plus(?x:Nat,?y:Nat),?x:Nat))))",
"max(s(s(?x:Nat)),s(min(max(plus(?x:Nat,?z:Nat),?x:Nat),?x:Nat))) = s(min(max(max(min(?x:Nat,?x:Nat),s(?x:Nat)),s(min(?x:Nat,?y:Nat))),plus(max(?z:Nat,s(0)),s(?x:Nat))))",
"min(s(s(max(min(?x:Nat,?y:Nat),max(?y:Nat,?x:Nat)))),max(0,?x:Nat)) = max(min(plus(max(plus(?z:Nat,0),plus(0,?x:Nat)),max(max(?y:Nat,0),max(?z:Nat,?z:Nat))),0),?x:Nat)",
"min(min(?x:Nat,max(max(max(?x:Nat,?x:Nat),s(?y:Nat)),min(plus(?x:Nat,?x:Nat),?x:Nat))),?x:Nat) = min(min(plus(?x:Nat,?x:Nat),min(s(max(?z:Nat,?x:Nat)),?x:Nat)),max(plus(?x:Nat,?x:Nat),0))",
"s(min(max(plus(?y:Nat,?z:Nat),plus(?x:Nat,s(0))),?x:Nat)) = min(s(plus(plus(min(?x:Nat,?z:Nat),s(?z:Nat)),plus(s(?x:Nat),s(?x:Nat)))),max(0,s(min(?x:Nat,plus(?x:Nat,?y:Nat)))))",
"min(?x:Nat,max(?x:Nat,max(max(?x:Nat,plus(?x:Nat,?y:Nat)),min(?x:Nat,0)))) = min(plus(s(min(0,min(?x:Nat,?x:Nat))),plus(plus(plus(?x:Nat,?y:Nat),?x:Nat),max(?x:Nat,min(?y:Nat,0)))),?x:Nat)",
"min(min(?z:Nat,min(0,?x:Nat)),min(plus(s(min(?z:Nat,?y:Nat)),?x:Nat),plus(plus(max(?x:Nat,?x:Nat),s(?y:Nat)),min(0,plus(?x:Nat,?x:Nat))))) = min(s(s(plus(?y:Nat,s(?x:Nat)))),0)",
"min(max(s(max(s(?x:Nat),?y:Nat)),plus(?z:Nat,plus(min(?x:Nat,0),?z:Nat))),0) = min(0,plus(s(plus(max(?z:Nat,?x:Nat),?x:Nat)),max(s(min(?x:Nat,?z:Nat)),plus(max(?z:Nat,0),?x:Nat))))",
"min(?z:Nat,min(0,plus(?z:Nat,max(s(?x:Nat),s(0))))) = min(min(s(max(0,s(?x:Nat))),0),min(plus(max(plus(?x:Nat,0),plus(0,?x:Nat)),min(max(0,?x:Nat),plus(?y:Nat,?z:Nat))),?x:Nat))",
"plus(plus(plus(s(plus(?y:Nat,?x:Nat)),0),min(s(plus(?y:Nat,?y:Nat)),s(0))),0) = min(plus(0,s(plus(s(?x:Nat),s(?y:Nat)))),max(s(max(?x:Nat,s(0))),s(s(plus(?y:Nat,?x:Nat)))))",
"min(max(max(?x:Nat,?x:Nat),s(min(max(?x:Nat,?x:Nat),?x:Nat))),plus(0,?y:Nat)) = min(min(plus(plus(0,max(0,?z:Nat)),?y:Nat),plus(?x:Nat,s(plus(?x:Nat,0)))),plus(0,min(min(?y:Nat,?y:Nat),s(?x:Nat))))",
"min(max(0,max(plus(max(?z:Nat,?z:Nat),plus(0,?x:Nat)),?y:Nat)),?y:Nat) = min(?y:Nat,max(plus(max(min(?y:Nat,?z:Nat),?y:Nat),max(s(?x:Nat),min(?y:Nat,?x:Nat))),plus(min(max(?x:Nat,0),max(?y:Nat,?x:Nat)),plus(?x:Nat,plus(?x:Nat,0)))))",
"max(min(plus(min(s(0),min(0,0)),?x:Nat),max(min(?x:Nat,max(0,0)),max(plus(?x:Nat,?z:Nat),?y:Nat))),max(?x:Nat,?x:Nat)) = min(?x:Nat,max(plus(max(?y:Nat,s(?x:Nat)),?y:Nat),max(s(plus(?y:Nat,?y:Nat)),plus(plus(0,?x:Nat),plus(?x:Nat,?x:Nat)))))",
"min(min(min(0,plus(s(0),?z:Nat)),plus(plus(max(?y:Nat,?z:Nat),s(?y:Nat)),s(plus(?z:Nat,?x:Nat)))),plus(s(?y:Nat),min(min(s(0),?z:Nat),?x:Nat))) = min(min(?z:Nat,min(plus(max(?x:Nat,?x:Nat),min(?y:Nat,?x:Nat)),plus(0,s(?x:Nat)))),min(?y:Nat,0))",
"min(min(0,plus(min(?x:Nat,s(?x:Nat)),?x:Nat)),plus(min(s(min(?z:Nat,?x:Nat)),min(max(?x:Nat,?y:Nat),s(0))),min(max(max(?y:Nat,?x:Nat),?x:Nat),max(?x:Nat,?x:Nat)))) = min(0,plus(plus(plus(min(?z:Nat,?x:Nat),?x:Nat),s(?x:Nat)),max(s(s(?x:Nat)),s(min(?y:Nat,?x:Nat)))))",
"max(max(max(?x:Nat,min(min(?z:Nat,?y:Nat),plus(0,?x:Nat))),plus(0,min(max(0,?x:Nat),0))),min(max(min(max(?x:Nat,0),min(?y:Nat,?x:Nat)),s(plus(?z:Nat,?x:Nat))),min(?x:Nat,?x:Nat))) = min(?x:Nat,max(s(max(plus(?y:Nat,?x:Nat),min(0,?x:Nat))),max(?x:Nat,s(?z:Nat))))",
"min(min(?x:Nat,plus(?x:Nat,plus(?x:Nat,max(?z:Nat,?x:Nat)))),s(plus(min(plus(?x:Nat,?x:Nat),min(?y:Nat,?y:Nat)),plus(?x:Nat,max(?x:Nat,?x:Nat))))) = min(plus(min(0,plus(min(0,0),?y:Nat)),max(max(max(?x:Nat,?x:Nat),?x:Nat),?x:Nat)),plus(max(min(max(?x:Nat,?x:Nat),?z:Nat),s(plus(?x:Nat,?x:Nat))),max(plus(max(0,?z:Nat),min(?y:Nat,?x:Nat)),min(s(0),0))))"
]

val problemsMaxMinPlusNo =
IOMstrs.rdEqs [
"max(0,?x:Nat) = plus(?x:Nat,?x:Nat)",
"min(?x:Nat,0) = plus(?y:Nat,0)",
"min(0,?x:Nat) = plus(0,?y:Nat)",
"plus(?x:Nat,?x:Nat) = max(?z:Nat,?x:Nat)",
"plus(?x:Nat,0) = max(?z:Nat,0)",
"plus(?z:Nat,?z:Nat) = max(?x:Nat,?x:Nat)",
"plus(0,?x:Nat) = max(?y:Nat,0)",
"plus(0,?x:Nat) = min(?x:Nat,?z:Nat)",
"s(?x:Nat) = plus(?x:Nat,max(?x:Nat,?z:Nat))",
"s(?x:Nat) = plus(max(?z:Nat,?y:Nat),?y:Nat)",
"s(?z:Nat) = max(plus(?x:Nat,?y:Nat),?x:Nat)",
"s(?z:Nat) = min(0,plus(?y:Nat,?z:Nat))",
"s(?z:Nat) = min(plus(?z:Nat,?x:Nat),?x:Nat)",
"max(?z:Nat,?z:Nat) = s(plus(?x:Nat,?x:Nat))",
"min(?x:Nat,?z:Nat) = plus(s(0),?z:Nat)",
"max(s(?y:Nat),?x:Nat) = plus(?y:Nat,?y:Nat)",
"min(plus(?x:Nat,?x:Nat),?y:Nat) = s(?z:Nat)",
"plus(?x:Nat,min(?x:Nat,?y:Nat)) = s(?x:Nat)",
"plus(?x:Nat,min(?y:Nat,?y:Nat)) = s(0)",
"plus(?z:Nat,max(?x:Nat,?x:Nat)) = s(?x:Nat)",
"s(?x:Nat) = plus(max(s(?y:Nat),?y:Nat),0)",
"s(s(?x:Nat)) = plus(?y:Nat,min(?x:Nat,?x:Nat))",
"max(?x:Nat,?y:Nat) = plus(s(?x:Nat),s(?z:Nat))",
"plus(?x:Nat,0) = s(min(?z:Nat,s(?x:Nat)))",
"plus(?y:Nat,s(?x:Nat)) = max(?x:Nat,s(?y:Nat))",
"s(plus(?z:Nat,s(?y:Nat))) = max(0,0)",
"max(0,plus(?x:Nat,?z:Nat)) = plus(0,?x:Nat)",
"plus(max(?x:Nat,?z:Nat),0) = plus(?z:Nat,?x:Nat)",
"max(?x:Nat,plus(s(?x:Nat),?y:Nat)) = s(?y:Nat)",
"s(?x:Nat) = max(max(?x:Nat,?z:Nat),plus(0,?y:Nat))",
"max(?x:Nat,?x:Nat) = plus(s(?y:Nat),min(?x:Nat,?x:Nat))",
"max(?x:Nat,0) = plus(max(?y:Nat,?x:Nat),s(0))",
"plus(?z:Nat,?y:Nat) = plus(0,s(min(?y:Nat,0)))",
"max(s(?z:Nat),?y:Nat) = plus(0,s(s(?x:Nat)))",
"max(s(0),?x:Nat) = plus(s(s(?z:Nat)),?z:Nat)",
"plus(s(?x:Nat),?z:Nat) = plus(?x:Nat,min(?x:Nat,0))",
"max(?x:Nat,plus(?x:Nat,?x:Nat)) = s(max(?x:Nat,?x:Nat))",
"min(s(?y:Nat),s(0)) = plus(0,s(?y:Nat))",
"s(plus(?x:Nat,max(?x:Nat,?z:Nat))) = min(?x:Nat,?z:Nat)",
"max(s(plus(0,0)),?x:Nat) = plus(?x:Nat,?x:Nat)",
"min(plus(?z:Nat,s(?x:Nat)),?z:Nat) = max(?x:Nat,?y:Nat)",
"s(s(min(plus(?x:Nat,?z:Nat),?y:Nat))) = s(?z:Nat)",
"max(plus(?x:Nat,?x:Nat),max(?z:Nat,?z:Nat)) = s(0)",
"min(plus(0,0),plus(?z:Nat,0)) = s(?y:Nat)",
"plus(max(?x:Nat,plus(?y:Nat,?x:Nat)),0) = s(?y:Nat)",
"plus(max(?z:Nat,s(?x:Nat)),s(?x:Nat)) = s(0)",
"plus(max(s(?x:Nat),?z:Nat),s(?z:Nat)) = s(0)",
"s(?z:Nat) = plus(?z:Nat,plus(max(s(?z:Nat),?y:Nat),?z:Nat))",
"s(?z:Nat) = plus(plus(?x:Nat,max(0,s(?x:Nat))),?x:Nat)",
"s(s(?z:Nat)) = plus(plus(min(?x:Nat,0),0),?z:Nat)",
"min(?x:Nat,?x:Nat) = plus(?x:Nat,plus(max(0,?y:Nat),?y:Nat))",
"plus(s(?y:Nat),?x:Nat) = min(s(0),plus(?z:Nat,0))",
"plus(s(?y:Nat),?z:Nat) = plus(?x:Nat,min(?y:Nat,s(?x:Nat)))",
"s(s(plus(0,?x:Nat))) = max(?z:Nat,min(?x:Nat,?y:Nat))",
"s(s(max(s(?x:Nat),?z:Nat))) = s(plus(0,?x:Nat))",
"max(?x:Nat,min(?z:Nat,s(?x:Nat))) = plus(?z:Nat,s(?z:Nat))",
"min(s(?x:Nat),s(s(0))) = plus(?z:Nat,s(?y:Nat))",
"min(plus(?z:Nat,?x:Nat),s(?x:Nat)) = max(s(?x:Nat),?y:Nat)",
"s(s(plus(max(?x:Nat,?x:Nat),?z:Nat))) = plus(?x:Nat,?x:Nat)",
"s(max(0,s(plus(?y:Nat,0)))) = plus(?x:Nat,0)",
"s(plus(s(s(s(0))),?x:Nat)) = min(?x:Nat,?x:Nat)",
"max(plus(?x:Nat,min(?x:Nat,?z:Nat)),?x:Nat) = min(?y:Nat,?x:Nat)",
"min(?z:Nat,s(plus(s(?x:Nat),?x:Nat))) = max(0,0)",
"min(s(?y:Nat),s(plus(s(?x:Nat),?x:Nat))) = s(0)",
"s(?x:Nat) = max(s(plus(0,?x:Nat)),s(s(s(0))))",
"s(?x:Nat) = plus(min(?y:Nat,plus(min(?x:Nat,?z:Nat),?z:Nat)),?x:Nat)",
"s(?y:Nat) = s(s(plus(?y:Nat,min(plus(?z:Nat,?x:Nat),?x:Nat))))",
"s(?y:Nat) = s(plus(?z:Nat,s(plus(min(?x:Nat,?x:Nat),?z:Nat))))",
"s(?y:Nat) = min(?z:Nat,min(plus(?x:Nat,0),max(0,?x:Nat)))",
"s(0) = s(s(plus(max(s(?z:Nat),s(?y:Nat)),?x:Nat)))",
"s(0) = plus(plus(s(s(min(?x:Nat,?y:Nat))),?x:Nat),?x:Nat)",
"s(s(?x:Nat)) = plus(?y:Nat,s(s(s(max(?z:Nat,0)))))",
"max(?x:Nat,0) = plus(max(s(max(0,?y:Nat)),0),?y:Nat)",
"max(?y:Nat,?y:Nat) = plus(plus(0,0),max(s(?z:Nat),?y:Nat))",
"max(?y:Nat,0) = plus(min(s(?y:Nat),?x:Nat),min(?x:Nat,?z:Nat))",
"max(?z:Nat,?x:Nat) = s(s(s(min(?x:Nat,plus(?y:Nat,?x:Nat)))))",
"min(?y:Nat,0) = min(s(s(s(plus(?z:Nat,0)))),?x:Nat)",
"min(0,0) = plus(min(max(?z:Nat,?z:Nat),?x:Nat),s(?y:Nat))",
"plus(?x:Nat,?x:Nat) = min(plus(?x:Nat,s(plus(?y:Nat,?x:Nat))),?y:Nat)",
"plus(?x:Nat,?z:Nat) = plus(0,plus(min(s(?z:Nat),?x:Nat),?x:Nat))",
"plus(?z:Nat,?x:Nat) = s(min(s(min(?y:Nat,s(?z:Nat))),0))",
"plus(?z:Nat,?z:Nat) = max(?y:Nat,plus(s(0),s(s(?y:Nat))))",
"plus(0,?x:Nat) = plus(?z:Nat,max(?y:Nat,s(plus(0,?x:Nat))))",
"s(plus(?z:Nat,?x:Nat)) = s(max(?x:Nat,max(?y:Nat,s(?y:Nat))))",
"s(plus(?z:Nat,?z:Nat)) = plus(min(?z:Nat,plus(?y:Nat,?z:Nat)),?y:Nat)",
"max(s(0),?y:Nat) = plus(s(s(s(?z:Nat))),s(?x:Nat))",
"min(?x:Nat,s(?x:Nat)) = s(plus(?x:Nat,max(s(?z:Nat),?x:Nat)))",
"min(s(?x:Nat),?z:Nat) = s(s(s(s(plus(?x:Nat,?y:Nat)))))",
"plus(0,plus(?z:Nat,?x:Nat)) = s(s(max(?x:Nat,s(?x:Nat))))",
"s(plus(max(?x:Nat,?x:Nat),?y:Nat)) = min(?x:Nat,max(?x:Nat,0))",
"max(plus(?z:Nat,?z:Nat),s(0)) = min(max(0,?x:Nat),?y:Nat)",
"min(plus(?z:Nat,?y:Nat),s(0)) = plus(s(s(?x:Nat)),0)",
"max(?z:Nat,max(max(0,0),?x:Nat)) = s(plus(?x:Nat,?x:Nat))",
"plus(min(?y:Nat,min(?z:Nat,?x:Nat)),?y:Nat) = min(?y:Nat,s(0))",
"s(s(s(min(?z:Nat,plus(?x:Nat,0))))) = min(?x:Nat,?y:Nat)",
"s(s(min(plus(s(?y:Nat),?z:Nat),?x:Nat))) = s(s(?x:Nat))",
"s(min(?x:Nat,s(max(s(?x:Nat),?x:Nat)))) = plus(?x:Nat,?x:Nat)",
"plus(?x:Nat,s(s(max(s(0),?x:Nat)))) = plus(?x:Nat,?x:Nat)",
"plus(s(min(?y:Nat,max(?x:Nat,0))),0) = s(s(?y:Nat))",
"plus(max(plus(s(?z:Nat),?x:Nat),?z:Nat),0) = max(0,?x:Nat)",
"s(max(plus(?y:Nat,s(plus(?x:Nat,?z:Nat))),0)) = s(?z:Nat)",
"plus(?z:Nat,max(?z:Nat,plus(s(s(?z:Nat)),?y:Nat))) = s(?y:Nat)",
"plus(s(s(plus(0,min(?x:Nat,0)))),?x:Nat) = s(?x:Nat)",
"plus(s(min(0,0)),s(max(0,?x:Nat))) = s(?x:Nat)",
"s(?x:Nat) = s(max(max(?z:Nat,s(plus(?z:Nat,0))),s(?z:Nat)))",
"s(?x:Nat) = min(s(max(plus(?x:Nat,plus(?z:Nat,0)),?z:Nat)),?y:Nat)",
"s(?y:Nat) = max(?x:Nat,plus(max(max(s(?y:Nat),?y:Nat),0),?x:Nat))",
"s(?y:Nat) = plus(?x:Nat,s(plus(max(?x:Nat,max(?x:Nat,?x:Nat)),0)))",
"s(?z:Nat) = max(?y:Nat,min(plus(min(s(0),?x:Nat),?x:Nat),?y:Nat))",
"s(?z:Nat) = plus(s(?x:Nat),s(plus(max(s(?x:Nat),?x:Nat),?x:Nat)))",
"s(0) = s(max(min(s(plus(?x:Nat,?x:Nat)),?x:Nat),s(?x:Nat)))",
"s(0) = min(?y:Nat,max(0,max(plus(?z:Nat,s(?x:Nat)),0)))",
"s(0) = min(min(plus(min(s(?y:Nat),?x:Nat),?x:Nat),?y:Nat),?z:Nat)",
"s(s(?x:Nat)) = plus(?z:Nat,s(max(?x:Nat,s(plus(?y:Nat,?y:Nat)))))",
"s(s(?x:Nat)) = plus(0,plus(min(?y:Nat,s(?x:Nat)),s(?y:Nat)))",
"max(?x:Nat,?x:Nat) = s(s(plus(min(?x:Nat,?x:Nat),plus(?z:Nat,?x:Nat))))",
"max(?x:Nat,0) = plus(?y:Nat,min(?z:Nat,s(min(s(0),0))))",
"max(?z:Nat,?x:Nat) = min(s(s(plus(?z:Nat,?z:Nat))),max(?x:Nat,?y:Nat))",
"min(?x:Nat,0) = plus(?z:Nat,max(s(s(plus(?x:Nat,?x:Nat))),0))",
"min(0,?x:Nat) = s(plus(?y:Nat,max(?y:Nat,plus(s(?z:Nat),?x:Nat))))",
"plus(?x:Nat,0) = s(min(s(max(plus(?z:Nat,?y:Nat),0)),?x:Nat))",
"plus(?y:Nat,?x:Nat) = s(min(s(?z:Nat),plus(?z:Nat,s(s(0)))))",
"plus(?y:Nat,?x:Nat) = s(plus(?z:Nat,s(min(min(?z:Nat,0),?z:Nat))))",
"plus(?y:Nat,?x:Nat) = s(plus(s(?z:Nat),min(min(?z:Nat,?y:Nat),?x:Nat)))",
"plus(?z:Nat,?x:Nat) = s(max(plus(s(min(?x:Nat,?z:Nat)),?x:Nat),0))",
"plus(0,?x:Nat) = plus(min(?x:Nat,s(min(s(0),?x:Nat))),?x:Nat)",
"s(s(plus(?x:Nat,?x:Nat))) = max(?y:Nat,plus(?y:Nat,plus(?y:Nat,0)))",
"s(plus(s(0),s(?x:Nat))) = s(s(max(s(?z:Nat),?y:Nat)))",
"plus(plus(max(?y:Nat,?x:Nat),?x:Nat),?z:Nat) = s(plus(s(?y:Nat),0))",
"s(s(s(plus(?x:Nat,min(?z:Nat,?x:Nat))))) = s(min(?x:Nat,?x:Nat))",
"s(max(s(s(max(?z:Nat,?z:Nat))),?x:Nat)) = plus(s(?z:Nat),?z:Nat)",
"max(s(s(plus(s(?y:Nat),?x:Nat))),0) = max(s(?z:Nat),?z:Nat)",
"min(s(s(0)),s(max(?x:Nat,?x:Nat))) = s(plus(?x:Nat,?z:Nat))",
"min(s(max(plus(?y:Nat,?x:Nat),?x:Nat)),?x:Nat) = plus(0,s(?x:Nat))",
"plus(s(?z:Nat),plus(s(s(?z:Nat)),0)) = max(?x:Nat,s(?y:Nat))",
"s(min(s(plus(plus(?x:Nat,0),?y:Nat)),?x:Nat)) = plus(?z:Nat,?x:Nat)",
"min(plus(s(s(plus(?y:Nat,0))),?z:Nat),?x:Nat) = max(?x:Nat,?z:Nat)",
"plus(s(max(s(plus(?z:Nat,?z:Nat)),?z:Nat)),?x:Nat) = s(s(?z:Nat))",
"plus(min(s(s(?x:Nat)),0),min(?z:Nat,0)) = max(?x:Nat,?y:Nat)",
"min(?z:Nat,s(plus(min(0,max(?y:Nat,?y:Nat)),?y:Nat))) = s(0)",
"min(s(s(s(plus(?x:Nat,?y:Nat)))),min(0,0)) = s(?y:Nat)",
"plus(plus(min(?z:Nat,?z:Nat),plus(?y:Nat,?x:Nat)),s(?z:Nat)) = s(?x:Nat)",
"s(?x:Nat) = s(plus(s(?z:Nat),max(?z:Nat,plus(max(?x:Nat,0),0))))",
"s(?x:Nat) = plus(?y:Nat,plus(?x:Nat,max(0,plus(?x:Nat,min(0,?y:Nat)))))",
"s(?y:Nat) = min(s(plus(?x:Nat,max(s(?y:Nat),plus(?x:Nat,0)))),0)",
"s(?y:Nat) = plus(plus(?z:Nat,0),s(s(min(max(?x:Nat,0),?x:Nat))))",
"s(?z:Nat) = s(min(plus(s(?x:Nat),min(0,plus(?z:Nat,?x:Nat))),?x:Nat))",
"s(?z:Nat) = plus(0,min(?z:Nat,s(min(s(?x:Nat),plus(?x:Nat,?x:Nat)))))",
"s(0) = s(min(s(plus(max(?z:Nat,?x:Nat),plus(?z:Nat,?x:Nat))),?x:Nat))",
"s(0) = min(plus(?z:Nat,s(max(s(0),min(?z:Nat,?x:Nat)))),?z:Nat)",
"s(s(?x:Nat)) = max(plus(plus(0,0),?z:Nat),max(?z:Nat,s(?x:Nat)))",
"s(s(?x:Nat)) = plus(?x:Nat,s(plus(max(?x:Nat,plus(?x:Nat,?y:Nat)),?x:Nat)))",
"s(s(?x:Nat)) = plus(plus(?y:Nat,?z:Nat),plus(?x:Nat,min(?x:Nat,s(?z:Nat))))",
"s(s(?y:Nat)) = s(min(?x:Nat,max(plus(plus(?z:Nat,?x:Nat),0),?x:Nat)))",
"max(?x:Nat,?x:Nat) = s(plus(?z:Nat,plus(?x:Nat,min(s(?x:Nat),s(?y:Nat)))))",
"min(?x:Nat,?y:Nat) = plus(s(?x:Nat),s(plus(s(plus(?x:Nat,?y:Nat)),?x:Nat)))",
"min(?x:Nat,0) = plus(?y:Nat,plus(?y:Nat,s(plus(s(?x:Nat),s(?z:Nat)))))",
"min(?y:Nat,0) = s(s(s(plus(min(0,?y:Nat),max(?x:Nat,?x:Nat)))))",
"plus(?x:Nat,?x:Nat) = s(max(plus(?z:Nat,plus(0,max(0,?z:Nat))),?x:Nat))",
"plus(?x:Nat,0) = max(plus(?y:Nat,min(s(plus(?z:Nat,?x:Nat)),0)),?y:Nat)",
"plus(?y:Nat,?y:Nat) = plus(s(plus(min(max(0,0),0),?z:Nat)),0)",
"plus(?z:Nat,?x:Nat) = min(max(s(0),plus(?x:Nat,max(?y:Nat,?x:Nat))),?x:Nat)",
"plus(?z:Nat,0) = s(plus(?x:Nat,plus(min(plus(?z:Nat,?z:Nat),?z:Nat),?y:Nat)))",
"plus(0,?x:Nat) = plus(s(min(0,plus(?y:Nat,max(?x:Nat,?x:Nat)))),?x:Nat)",
"max(s(?x:Nat),?y:Nat) = s(plus(?y:Nat,max(?x:Nat,plus(?y:Nat,s(?z:Nat)))))",
"min(s(?x:Nat),?x:Nat) = plus(max(plus(?y:Nat,s(?z:Nat)),s(0)),0)",
"s(min(?x:Nat,s(?x:Nat))) = max(?x:Nat,s(plus(0,plus(?y:Nat,0))))",
"plus(s(?y:Nat),s(?x:Nat)) = s(min(?x:Nat,s(s(max(?x:Nat,?x:Nat)))))",
"s(max(?y:Nat,plus(?y:Nat,?y:Nat))) = plus(max(s(0),?x:Nat),s(0))",
"max(max(?z:Nat,s(?z:Nat)),?x:Nat) = plus(s(s(s(0))),s(0))",
"min(s(?x:Nat),s(s(0))) = s(plus(?y:Nat,s(max(?y:Nat,?z:Nat))))",
"min(?z:Nat,min(max(?y:Nat,?y:Nat),?y:Nat)) = plus(s(max(?x:Nat,?y:Nat)),?z:Nat)",
"plus(s(?z:Nat),min(?z:Nat,s(?z:Nat))) = s(max(0,s(s(0))))",
"s(s(max(s(plus(0,?x:Nat)),?z:Nat))) = min(s(s(?z:Nat)),?y:Nat)",
"s(max(?x:Nat,s(s(plus(?z:Nat,?y:Nat))))) = plus(?z:Nat,plus(?z:Nat,?x:Nat))",
"plus(?y:Nat,max(plus(s(0),?y:Nat),?y:Nat)) = s(s(min(?y:Nat,?y:Nat)))",
"s(max(plus(s(max(?x:Nat,?z:Nat)),?z:Nat),?x:Nat)) = plus(?y:Nat,s(?z:Nat))",
"max(?y:Nat,plus(?y:Nat,s(s(plus(?z:Nat,?y:Nat))))) = s(max(?y:Nat,?z:Nat))",
"plus(?x:Nat,s(max(max(?x:Nat,s(0)),?x:Nat))) = s(plus(?x:Nat,?x:Nat))",
"s(plus(plus(?y:Nat,max(plus(?z:Nat,?x:Nat),0)),0)) = s(s(?z:Nat))",
"s(plus(plus(plus(plus(?z:Nat,?x:Nat),?x:Nat),?y:Nat),0)) = min(?y:Nat,?y:Nat)",
"max(0,plus(?x:Nat,s(max(max(?x:Nat,?x:Nat),?x:Nat)))) = plus(?y:Nat,?x:Nat)",
"plus(s(s(max(s(?y:Nat),max(?z:Nat,?z:Nat)))),?x:Nat) = s(s(0))",
"plus(min(plus(?y:Nat,plus(s(?z:Nat),?y:Nat)),0),?y:Nat) = min(?z:Nat,?y:Nat)",
"s(s(plus(min(max(?z:Nat,?y:Nat),max(?z:Nat,?x:Nat)),?y:Nat))) = s(0)",
"max(max(s(plus(plus(?y:Nat,?x:Nat),?x:Nat)),s(?z:Nat)),0) = s(?z:Nat)",
"max(min(plus(min(plus(?z:Nat,?y:Nat),0),?x:Nat),0),?y:Nat) = s(?y:Nat)",
"max(plus(s(max(plus(?x:Nat,?y:Nat),s(0))),?x:Nat),?z:Nat) = s(0)",
"min(0,min(s(s(?x:Nat)),plus(s(?y:Nat),s(?x:Nat)))) = s(?x:Nat)",
"plus(?z:Nat,min(0,plus(plus(0,max(?x:Nat,?x:Nat)),?y:Nat))) = s(0)",
"plus(?z:Nat,plus(max(0,s(s(?x:Nat))),min(?x:Nat,0))) = s(?z:Nat)",
"plus(plus(max(?y:Nat,plus(plus(?x:Nat,?y:Nat),0)),0),?z:Nat) = s(?y:Nat)",
"s(?x:Nat) = s(plus(0,min(min(max(?x:Nat,?x:Nat),0),plus(?z:Nat,?y:Nat))))",
"s(?x:Nat) = plus(?z:Nat,plus(s(?x:Nat),plus(s(?z:Nat),s(min(?y:Nat,0)))))",
"s(s(?z:Nat)) = plus(?y:Nat,min(?y:Nat,plus(plus(?x:Nat,0),plus(?x:Nat,0))))",
"min(?x:Nat,0) = s(plus(s(?x:Nat),s(plus(s(?y:Nat),plus(0,?x:Nat)))))",
"min(?y:Nat,0) = max(plus(max(min(?x:Nat,?z:Nat),s(0)),?x:Nat),s(?z:Nat))",
"min(?z:Nat,?x:Nat) = s(plus(s(min(max(?x:Nat,?y:Nat),plus(?x:Nat,0))),?y:Nat))",
"plus(?x:Nat,?y:Nat) = min(s(plus(?x:Nat,?z:Nat)),s(s(s(max(?z:Nat,?x:Nat)))))",
"plus(?y:Nat,0) = min(plus(0,max(max(s(0),?x:Nat),s(?y:Nat))),?x:Nat)",
"s(s(s(?y:Nat))) = plus(min(min(?z:Nat,?x:Nat),?x:Nat),min(0,s(?x:Nat)))",
"s(min(?x:Nat,?z:Nat)) = s(max(s(s(plus(?z:Nat,?x:Nat))),plus(?x:Nat,?z:Nat)))",
"s(plus(?x:Nat,?x:Nat)) = s(max(?x:Nat,min(max(min(?x:Nat,?x:Nat),0),0)))",
"s(plus(?x:Nat,?x:Nat)) = plus(min(min(?y:Nat,min(?y:Nat,s(?y:Nat))),?x:Nat),?y:Nat)",
"max(s(?y:Nat),?z:Nat) = plus(0,max(min(s(min(?y:Nat,?x:Nat)),0),?x:Nat))",
"plus(s(?x:Nat),?x:Nat) = s(s(s(min(max(?z:Nat,?x:Nat),plus(0,?x:Nat)))))",
"plus(s(?y:Nat),?x:Nat) = max(?y:Nat,s(min(0,max(plus(0,?x:Nat),?y:Nat))))",
"max(plus(?z:Nat,?x:Nat),?z:Nat) = max(s(plus(s(plus(?y:Nat,?z:Nat)),?x:Nat)),?x:Nat)",
"plus(s(?y:Nat),s(?x:Nat)) = plus(s(?z:Nat),max(s(max(?z:Nat,?z:Nat)),?x:Nat))",
"plus(plus(0,?x:Nat),0) = plus(plus(s(0),?x:Nat),min(0,s(?x:Nat)))",
"plus(s(?z:Nat),s(s(?y:Nat))) = plus(s(max(s(s(?x:Nat)),?x:Nat)),?y:Nat)",
"s(plus(plus(?x:Nat,?x:Nat),s(0))) = plus(plus(min(?z:Nat,?y:Nat),?x:Nat),0)",
"min(plus(?z:Nat,?x:Nat),plus(?x:Nat,?y:Nat)) = s(plus(?y:Nat,plus(?x:Nat,s(?x:Nat))))",
"min(plus(0,?x:Nat),min(?x:Nat,?y:Nat)) = s(s(s(plus(s(0),?z:Nat))))",
"s(plus(?x:Nat,plus(?z:Nat,max(0,?z:Nat)))) = plus(?y:Nat,min(s(0),?x:Nat))",
"max(?y:Nat,plus(max(0,s(?y:Nat)),?z:Nat)) = plus(?x:Nat,max(s(?y:Nat),?x:Nat))",
"max(max(plus(?x:Nat,s(?x:Nat)),0),?y:Nat) = max(?x:Nat,s(s(s(?y:Nat))))",
"s(plus(s(?x:Nat),max(?y:Nat,s(s(?x:Nat))))) = max(0,min(?x:Nat,?y:Nat))",
"max(?z:Nat,plus(s(0),plus(s(?x:Nat),?x:Nat))) = s(s(plus(0,?z:Nat)))",
"max(0,min(s(max(?z:Nat,0)),s(0))) = plus(plus(?z:Nat,?x:Nat),0)",
"max(plus(s(?x:Nat),plus(?x:Nat,s(?y:Nat))),0) = max(max(?z:Nat,?x:Nat),?z:Nat)",
"plus(?x:Nat,s(plus(?x:Nat,s(min(?x:Nat,?x:Nat))))) = max(min(?y:Nat,0),?y:Nat)",
"plus(?z:Nat,min(0,min(?y:Nat,s(plus(?x:Nat,?x:Nat))))) = s(s(s(?x:Nat)))",
"plus(max(?x:Nat,plus(min(?x:Nat,s(?y:Nat)),?x:Nat)),?z:Nat) = s(max(?z:Nat,?x:Nat))",
"s(s(plus(max(plus(0,?y:Nat),max(?x:Nat,0)),?x:Nat))) = min(0,?y:Nat)",
"s(max(s(?x:Nat),s(plus(max(?x:Nat,0),s(?x:Nat))))) = min(0,?x:Nat)",
"s(min(s(?y:Nat),max(s(plus(?y:Nat,?x:Nat)),s(?x:Nat)))) = plus(?x:Nat,?z:Nat)",
"s(min(s(max(min(?x:Nat,?y:Nat),min(?x:Nat,?x:Nat))),?y:Nat)) = plus(?y:Nat,?x:Nat)",
"s(min(s(plus(min(?x:Nat,?x:Nat),plus(0,?x:Nat))),?x:Nat)) = max(?y:Nat,?y:Nat)",
"s(plus(s(plus(plus(?x:Nat,0),min(?x:Nat,?x:Nat))),?x:Nat)) = plus(?z:Nat,?x:Nat)",
"max(min(0,?z:Nat),s(min(?y:Nat,s(min(?x:Nat,?x:Nat))))) = plus(?y:Nat,?x:Nat)",
"max(min(plus(?z:Nat,plus(0,?x:Nat)),min(0,?z:Nat)),0) = s(s(?x:Nat))",
"min(?x:Nat,plus(?z:Nat,max(s(s(?x:Nat)),plus(?x:Nat,?x:Nat)))) = max(?x:Nat,?z:Nat)",
"min(plus(0,?z:Nat),s(min(s(max(?y:Nat,?z:Nat)),0))) = min(?z:Nat,?x:Nat)",
"plus(max(plus(?y:Nat,?z:Nat),min(s(s(?x:Nat)),0)),?z:Nat) = plus(?x:Nat,?z:Nat)",
"plus(plus(?y:Nat,s(plus(max(0,0),?z:Nat))),s(?z:Nat)) = plus(?y:Nat,?x:Nat)",
"s(s(min(s(min(0,?y:Nat)),min(0,plus(?x:Nat,?x:Nat))))) = s(?y:Nat)",
"s(s(plus(s(s(?z:Nat)),plus(max(?x:Nat,?y:Nat),s(?y:Nat))))) = s(?x:Nat)",
"s(plus(s(plus(0,s(?x:Nat))),s(min(?x:Nat,s(0))))) = s(?x:Nat)",
"s(plus(max(min(max(?z:Nat,?z:Nat),max(?x:Nat,0)),?z:Nat),?x:Nat)) = s(?y:Nat)",
"max(s(min(s(plus(?y:Nat,0)),?x:Nat)),max(0,s(?x:Nat))) = s(?z:Nat)",
"min(max(plus(?x:Nat,?x:Nat),plus(0,min(0,s(?x:Nat)))),?y:Nat) = s(?y:Nat)",
"plus(?x:Nat,max(?x:Nat,max(plus(s(0),plus(?x:Nat,?x:Nat)),?z:Nat))) = s(?x:Nat)",
"plus(s(?x:Nat),min(max(?z:Nat,plus(?y:Nat,max(0,0))),?x:Nat)) = s(?z:Nat)",
"s(?x:Nat) = max(s(max(plus(max(?z:Nat,0),plus(?x:Nat,?z:Nat)),s(?z:Nat))),?x:Nat)",
"s(?x:Nat) = plus(?y:Nat,plus(max(max(?y:Nat,max(0,?x:Nat)),plus(?x:Nat,?z:Nat)),?x:Nat))",
"s(?y:Nat) = max(min(?x:Nat,min(?z:Nat,max(plus(0,?z:Nat),plus(0,?z:Nat)))),?x:Nat)",
"s(?y:Nat) = min(?x:Nat,max(?x:Nat,max(0,min(plus(?x:Nat,?z:Nat),plus(?x:Nat,?y:Nat)))))",
"max(?x:Nat,?x:Nat) = plus(max(?x:Nat,min(plus(min(?x:Nat,?z:Nat),s(?x:Nat)),?y:Nat)),?x:Nat)",
"plus(?x:Nat,?y:Nat) = plus(plus(min(?x:Nat,?z:Nat),max(max(?y:Nat,s(0)),?x:Nat)),?x:Nat)",
"plus(?z:Nat,?y:Nat) = plus(0,max(s(s(s(?x:Nat))),s(plus(s(0),?z:Nat))))",
"plus(?z:Nat,?z:Nat) = min(plus(s(min(min(?y:Nat,?y:Nat),min(?x:Nat,0))),?x:Nat),?y:Nat)",
"s(max(0,0)) = min(min(s(plus(s(0),min(?x:Nat,?x:Nat))),?x:Nat),?x:Nat)",
"s(plus(?x:Nat,0)) = max(0,plus(min(?z:Nat,?x:Nat),max(s(?x:Nat),s(?y:Nat))))",
"max(?x:Nat,s(0)) = plus(min(max(max(plus(?x:Nat,?x:Nat),?x:Nat),?y:Nat),?x:Nat),?x:Nat)",
"max(s(?x:Nat),?z:Nat) = max(min(?x:Nat,0),s(s(min(0,plus(?z:Nat,?y:Nat)))))",
"s(s(s(s(?y:Nat)))) = s(min(s(?x:Nat),min(s(plus(?x:Nat,?x:Nat)),?z:Nat)))",
"plus(s(s(?x:Nat)),?z:Nat) = plus(plus(s(min(max(?x:Nat,?y:Nat),0)),0),?x:Nat)",
"plus(max(?z:Nat,?y:Nat),0) = s(plus(plus(plus(plus(?x:Nat,?x:Nat),?y:Nat),?y:Nat),?y:Nat))",
"plus(plus(?z:Nat,?x:Nat),0) = plus(s(min(?z:Nat,s(plus(?y:Nat,?z:Nat)))),s(?x:Nat))",
"min(max(0,s(?y:Nat)),?z:Nat) = plus(s(plus(?x:Nat,?y:Nat)),plus(?x:Nat,s(?z:Nat)))",
"s(s(plus(plus(?x:Nat,s(?z:Nat)),0))) = min(s(s(?x:Nat)),min(?z:Nat,0))",
"plus(s(0),max(s(s(?x:Nat)),0)) = plus(plus(?z:Nat,plus(?x:Nat,?z:Nat)),?x:Nat)",
"s(max(s(plus(?y:Nat,max(?x:Nat,?x:Nat))),?y:Nat)) = s(plus(?x:Nat,plus(0,?x:Nat)))",
"s(plus(plus(min(s(?x:Nat),?y:Nat),?x:Nat),0)) = max(?y:Nat,s(plus(?x:Nat,?x:Nat)))",
"max(0,plus(plus(?z:Nat,?x:Nat),s(s(?x:Nat)))) = plus(min(?x:Nat,s(?x:Nat)),0)",
"max(s(plus(s(min(?z:Nat,?x:Nat)),?x:Nat)),s(?x:Nat)) = plus(s(s(0)),?x:Nat)",
"min(?y:Nat,plus(s(s(plus(?x:Nat,?x:Nat))),s(?x:Nat))) = min(plus(?x:Nat,0),?y:Nat)",
"s(s(max(min(plus(?z:Nat,?y:Nat),min(?z:Nat,?z:Nat)),?x:Nat))) = plus(?z:Nat,s(?z:Nat))",
"max(s(plus(?x:Nat,?x:Nat)),s(s(plus(?x:Nat,s(?z:Nat))))) = plus(?z:Nat,s(?x:Nat))",
"max(s(plus(min(plus(?x:Nat,?x:Nat),s(?y:Nat)),?y:Nat)),?y:Nat) = s(s(s(0)))",
"min(?x:Nat,plus(plus(plus(plus(0,?x:Nat),?y:Nat),0),?x:Nat)) = plus(?z:Nat,s(?z:Nat))",
"plus(?y:Nat,plus(?x:Nat,s(max(s(?z:Nat),max(?z:Nat,?z:Nat))))) = max(0,s(?z:Nat))",
"plus(s(max(s(s(?z:Nat)),?x:Nat)),max(?z:Nat,s(?z:Nat))) = plus(s(?y:Nat),?z:Nat)",
"plus(plus(s(min(plus(?x:Nat,?z:Nat),?y:Nat)),?x:Nat),s(0)) = min(?y:Nat,s(0))",
"s(s(min(s(s(?y:Nat)),plus(plus(?x:Nat,0),s(?z:Nat))))) = s(s(?x:Nat))",
"s(min(s(0),plus(0,plus(plus(?y:Nat,?z:Nat),s(?z:Nat))))) = max(?x:Nat,0)",
"s(plus(?x:Nat,plus(plus(?x:Nat,?x:Nat),max(?y:Nat,max(?y:Nat,?y:Nat))))) = plus(?x:Nat,0)",
"s(plus(plus(0,plus(min(?z:Nat,0),plus(?y:Nat,?x:Nat))),0)) = max(0,?z:Nat)",
"plus(s(s(max(plus(0,?x:Nat),?x:Nat))),plus(?z:Nat,s(?x:Nat))) = plus(?y:Nat,?x:Nat)",
"plus(plus(s(min(max(?z:Nat,?x:Nat),?z:Nat)),min(?x:Nat,?x:Nat)),0) = min(?y:Nat,?x:Nat)",
"max(0,max(min(plus(plus(?x:Nat,?x:Nat),s(?x:Nat)),s(?x:Nat)),0)) = s(?z:Nat)",
"max(max(plus(?y:Nat,s(?y:Nat)),max(?x:Nat,s(plus(?z:Nat,?x:Nat)))),?z:Nat) = s(?y:Nat)",
"max(plus(?y:Nat,min(?x:Nat,max(max(?z:Nat,0),max(?x:Nat,?x:Nat)))),?x:Nat) = s(0)",
"min(min(0,plus(s(plus(?y:Nat,?z:Nat)),s(max(?z:Nat,?y:Nat)))),?x:Nat) = s(?z:Nat)",
"min(plus(?x:Nat,s(?x:Nat)),min(0,s(max(plus(0,?x:Nat),?x:Nat)))) = s(?y:Nat)",
"plus(?z:Nat,s(plus(max(s(?x:Nat),?x:Nat),plus(?x:Nat,plus(?x:Nat,?x:Nat))))) = s(?x:Nat)",
"plus(0,max(?y:Nat,plus(s(plus(?y:Nat,0)),s(min(?x:Nat,?z:Nat))))) = s(0)",
"plus(s(?x:Nat),plus(max(plus(plus(0,?z:Nat),0),s(?x:Nat)),?x:Nat)) = s(?z:Nat)",
"plus(max(?x:Nat,plus(s(max(?x:Nat,?y:Nat)),plus(s(?z:Nat),?x:Nat))),?x:Nat) = s(?z:Nat)",
"plus(plus(?z:Nat,?x:Nat),max(?y:Nat,s(min(plus(?z:Nat,?x:Nat),s(0))))) = s(?x:Nat)",
"s(?x:Nat) = min(min(?x:Nat,min(plus(s(?x:Nat),?x:Nat),?z:Nat)),plus(s(?z:Nat),s(?x:Nat)))",
"s(0) = s(s(min(plus(max(?x:Nat,?z:Nat),s(0)),plus(?z:Nat,max(?x:Nat,?x:Nat)))))",
"s(0) = max(s(min(0,?x:Nat)),s(plus(?z:Nat,max(min(?z:Nat,0),s(?y:Nat)))))",
"s(0) = plus(?x:Nat,plus(s(max(plus(?x:Nat,?x:Nat),0)),s(s(min(?z:Nat,0)))))",
"max(?x:Nat,?y:Nat) = max(s(min(plus(min(?x:Nat,0),?z:Nat),s(?x:Nat))),max(?y:Nat,?x:Nat))",
"max(?x:Nat,?z:Nat) = s(max(plus(0,?y:Nat),max(plus(s(?x:Nat),max(?y:Nat,?y:Nat)),0)))",
"max(?z:Nat,?z:Nat) = min(?x:Nat,s(max(max(0,plus(?x:Nat,0)),max(?x:Nat,s(0)))))",
"max(0,?z:Nat) = plus(s(s(?x:Nat)),max(?x:Nat,plus(max(?x:Nat,max(?x:Nat,0)),?x:Nat)))",
"plus(?x:Nat,?x:Nat) = plus(s(0),s(plus(s(plus(?y:Nat,0)),max(s(?y:Nat),?z:Nat))))",
"plus(?x:Nat,?y:Nat) = max(?y:Nat,max(min(?x:Nat,min(max(?y:Nat,?y:Nat),plus(?x:Nat,0))),0))",
"plus(?y:Nat,?z:Nat) = min(?x:Nat,max(s(s(?x:Nat)),s(plus(s(?z:Nat),plus(0,?x:Nat)))))",
"plus(?z:Nat,0) = s(s(min(plus(?x:Nat,plus(?z:Nat,?z:Nat)),max(?z:Nat,min(?x:Nat,?z:Nat)))))",
"s(max(0,?y:Nat)) = min(s(?x:Nat),min(s(plus(max(?z:Nat,?y:Nat),?y:Nat)),s(?z:Nat)))",
"plus(0,s(?x:Nat)) = s(max(plus(min(max(?z:Nat,?z:Nat),?y:Nat),?x:Nat),s(s(?x:Nat))))",
"max(plus(?y:Nat,?z:Nat),?x:Nat) = min(?x:Nat,s(s(min(min(?z:Nat,?x:Nat),max(?y:Nat,0)))))",
"plus(0,max(?y:Nat,0)) = max(s(?y:Nat),plus(?x:Nat,min(?x:Nat,min(?y:Nat,s(?x:Nat)))))",
"s(min(s(s(?x:Nat)),?x:Nat)) = plus(min(s(?x:Nat),plus(s(?y:Nat),?x:Nat)),s(?x:Nat))",
"plus(s(?x:Nat),plus(?z:Nat,?x:Nat)) = min(min(?y:Nat,plus(s(min(?x:Nat,0)),?y:Nat)),?x:Nat)",
"plus(s(min(?z:Nat,?z:Nat)),?y:Nat) = s(plus(s(?y:Nat),min(s(min(0,?x:Nat)),?x:Nat)))",
"s(s(plus(?y:Nat,s(s(?x:Nat))))) = s(s(s(plus(max(?x:Nat,?x:Nat),s(?x:Nat)))))",
"max(?y:Nat,min(?z:Nat,s(s(0)))) = s(plus(min(s(max(?y:Nat,0)),0),?x:Nat))",
"min(max(?x:Nat,s(s(?y:Nat))),0) = s(plus(max(s(plus(?x:Nat,?z:Nat)),0),?x:Nat))",
"plus(s(?x:Nat),plus(s(?z:Nat),?x:Nat)) = s(max(max(?x:Nat,?x:Nat),s(plus(?x:Nat,?x:Nat))))",
"s(min(s(plus(plus(?y:Nat,0),?x:Nat)),?x:Nat)) = s(max(s(plus(0,?x:Nat)),0))",
"s(min(min(?z:Nat,s(max(0,?x:Nat))),?x:Nat)) = s(plus(plus(s(0),?z:Nat),?x:Nat))",
"s(plus(?x:Nat,plus(s(plus(?z:Nat,?z:Nat)),?x:Nat))) = s(min(?z:Nat,plus(s(?y:Nat),?z:Nat)))",
"plus(s(min(plus(?z:Nat,s(?y:Nat)),0)),0) = s(plus(max(?x:Nat,s(0)),?y:Nat))",
"s(plus(plus(0,min(s(?y:Nat),s(?y:Nat))),?x:Nat)) = plus(max(0,s(?x:Nat)),?x:Nat)",
"plus(max(?x:Nat,?y:Nat),s(plus(plus(0,?y:Nat),?x:Nat))) = max(s(plus(?y:Nat,?x:Nat)),?x:Nat)",
"max(s(min(s(s(?y:Nat)),?z:Nat)),plus(?y:Nat,s(?z:Nat))) = min(max(?z:Nat,?x:Nat),?x:Nat)",
"max(min(?y:Nat,min(max(plus(?x:Nat,?z:Nat),?x:Nat),?y:Nat)),?y:Nat) = plus(?x:Nat,plus(?y:Nat,0))",
"max(min(plus(?x:Nat,max(0,plus(?x:Nat,0))),?y:Nat),?z:Nat) = max(min(?x:Nat,?x:Nat),?y:Nat)",
"plus(s(?z:Nat),s(min(s(?x:Nat),s(plus(?y:Nat,?x:Nat))))) = plus(plus(?x:Nat,?z:Nat),0)",
"s(max(?x:Nat,min(?x:Nat,plus(plus(0,?x:Nat),plus(?z:Nat,?z:Nat))))) = min(?y:Nat,s(?z:Nat))",
"s(plus(s(s(?x:Nat)),min(max(plus(0,?y:Nat),?x:Nat),?y:Nat))) = s(plus(?y:Nat,0))",
"min(?z:Nat,max(?z:Nat,plus(max(min(?x:Nat,?y:Nat),s(?x:Nat)),?y:Nat))) = s(plus(0,?x:Nat))",
"min(s(?y:Nat),s(s(max(plus(?y:Nat,?z:Nat),plus(?z:Nat,?y:Nat))))) = s(s(s(?z:Nat)))",
"plus(?y:Nat,min(s(max(?x:Nat,min(?z:Nat,?x:Nat))),s(s(?y:Nat)))) = s(max(0,?x:Nat))",
"plus(s(min(0,plus(min(?z:Nat,?x:Nat),min(0,?z:Nat)))),?y:Nat) = max(s(0),0)",
"s(s(min(min(?x:Nat,max(?x:Nat,0)),plus(plus(?x:Nat,?z:Nat),?y:Nat)))) = plus(0,?y:Nat)",
"max(plus(plus(?x:Nat,s(plus(?x:Nat,?y:Nat))),?y:Nat),min(?y:Nat,s(?y:Nat))) = max(?z:Nat,?z:Nat)",
"min(?x:Nat,plus(max(0,plus(plus(?x:Nat,?x:Nat),min(?y:Nat,?x:Nat))),?x:Nat)) = plus(?y:Nat,?x:Nat)",
"min(max(?y:Nat,plus(?x:Nat,plus(?z:Nat,s(?x:Nat)))),plus(s(?z:Nat),0)) = max(?x:Nat,?y:Nat)",
"min(min(0,?x:Nat),plus(s(plus(plus(?y:Nat,0),s(?x:Nat))),0)) = plus(0,?y:Nat)",
"min(plus(?x:Nat,max(?y:Nat,min(min(?x:Nat,0),plus(?y:Nat,?x:Nat)))),?z:Nat) = plus(?z:Nat,?x:Nat)",
"plus(s(s(?y:Nat)),s(min(min(plus(?z:Nat,?y:Nat),s(0)),0))) = max(?x:Nat,?x:Nat)",
"s(s(plus(min(plus(?x:Nat,0),max(?x:Nat,?x:Nat)),s(min(?y:Nat,0))))) = s(?x:Nat)",
"s(max(?x:Nat,plus(min(0,s(?z:Nat)),min(plus(?z:Nat,?x:Nat),s(0))))) = s(?z:Nat)",
"s(max(plus(?y:Nat,plus(?z:Nat,plus(?x:Nat,0))),s(s(min(0,?y:Nat))))) = s(0)",
"s(min(plus(s(s(?x:Nat)),max(plus(?y:Nat,0),max(?y:Nat,?x:Nat))),0)) = s(?x:Nat)",
"max(s(plus(min(?z:Nat,plus(?y:Nat,?y:Nat)),plus(s(0),s(?z:Nat)))),?y:Nat) = s(?z:Nat)",
"plus(?x:Nat,s(plus(plus(?y:Nat,plus(?y:Nat,?y:Nat)),max(?x:Nat,min(?x:Nat,?x:Nat))))) = s(0)",
"plus(min(s(0),plus(min(plus(?x:Nat,?x:Nat),s(0)),?x:Nat)),s(?x:Nat)) = s(?x:Nat)",
"plus(plus(?x:Nat,s(?y:Nat)),min(s(?x:Nat),min(s(max(?y:Nat,0)),?z:Nat))) = s(0)",
"plus(plus(min(s(plus(0,0)),?y:Nat),plus(s(s(?x:Nat)),0)),0) = s(?z:Nat)",
"s(?x:Nat) = plus(?z:Nat,max(?z:Nat,plus(min(?x:Nat,?x:Nat),max(plus(0,?y:Nat),plus(?y:Nat,0)))))",
"s(?x:Nat) = plus(min(?x:Nat,min(?x:Nat,s(min(0,?z:Nat)))),plus(?x:Nat,s(plus(?z:Nat,0))))",
"s(?z:Nat) = max(max(s(max(?x:Nat,?z:Nat)),max(plus(plus(?z:Nat,?z:Nat),s(?y:Nat)),0)),?z:Nat)",
"s(s(?y:Nat)) = s(s(min(max(s(?y:Nat),s(?x:Nat)),min(s(0),plus(0,?x:Nat)))))",
"min(0,?x:Nat) = min(?x:Nat,s(plus(max(max(?x:Nat,?x:Nat),max(?z:Nat,?y:Nat)),max(?x:Nat,?x:Nat))))",
"plus(?y:Nat,?x:Nat) = max(plus(0,plus(s(?x:Nat),min(?x:Nat,plus(0,0)))),plus(?z:Nat,?x:Nat))",
"plus(?y:Nat,?x:Nat) = plus(?z:Nat,min(?y:Nat,min(plus(?x:Nat,s(0)),min(plus(?x:Nat,?x:Nat),?x:Nat))))",
"plus(?y:Nat,?z:Nat) = s(max(s(s(max(0,?y:Nat))),plus(?x:Nat,max(min(0,?y:Nat),?z:Nat))))",
"plus(?z:Nat,?z:Nat) = min(?x:Nat,s(min(max(min(?x:Nat,?x:Nat),0),plus(min(?x:Nat,?y:Nat),?y:Nat))))",
"plus(?z:Nat,0) = plus(?x:Nat,min(?y:Nat,max(s(plus(0,?x:Nat)),plus(?y:Nat,min(?x:Nat,?x:Nat)))))",
"s(max(?z:Nat,0)) = plus(?y:Nat,max(min(?x:Nat,max(plus(0,?z:Nat),max(?z:Nat,?z:Nat))),?x:Nat))",
"max(?x:Nat,s(s(0))) = s(min(max(?x:Nat,?x:Nat),min(?x:Nat,max(plus(?y:Nat,?x:Nat),0))))",
"max(?x:Nat,max(0,?x:Nat)) = plus(max(?z:Nat,plus(min(min(?x:Nat,?z:Nat),0),?y:Nat)),s(0))",
"max(?x:Nat,plus(?y:Nat,?x:Nat)) = plus(?y:Nat,max(0,s(plus(max(?x:Nat,?x:Nat),plus(?z:Nat,0)))))",
"max(plus(?x:Nat,?z:Nat),?x:Nat) = plus(?y:Nat,s(max(?x:Nat,plus(max(?x:Nat,?y:Nat),plus(0,?z:Nat)))))",
"max(plus(?y:Nat,?y:Nat),?y:Nat) = plus(?y:Nat,plus(0,plus(s(plus(?y:Nat,?y:Nat)),plus(?y:Nat,?y:Nat))))",
"min(?y:Nat,min(?y:Nat,?z:Nat)) = plus(s(plus(max(?z:Nat,max(?z:Nat,?y:Nat)),0)),min(?z:Nat,0))",
"min(?z:Nat,plus(?x:Nat,?z:Nat)) = min(0,s(min(plus(min(?x:Nat,?x:Nat),plus(0,?x:Nat)),?z:Nat)))",
"plus(s(s(?x:Nat)),0) = s(min(?z:Nat,plus(s(?x:Nat),min(s(?y:Nat),plus(?x:Nat,?x:Nat)))))",
"plus(plus(?x:Nat,0),0) = min(?x:Nat,max(s(min(plus(0,?x:Nat),plus(?z:Nat,?y:Nat))),?y:Nat))",
"plus(plus(?z:Nat,?z:Nat),?x:Nat) = plus(s(plus(plus(max(?y:Nat,?z:Nat),s(?x:Nat)),s(?x:Nat))),0)",
"s(max(?y:Nat,min(0,?z:Nat))) = max(s(min(s(?x:Nat),s(plus(0,?x:Nat)))),s(?z:Nat))",
"plus(max(0,?z:Nat),s(?x:Nat)) = s(plus(plus(plus(?x:Nat,max(?z:Nat,?x:Nat)),s(?x:Nat)),?x:Nat))",
"s(max(plus(0,s(?z:Nat)),0)) = plus(?y:Nat,s(plus(0,plus(max(0,?z:Nat),?x:Nat))))",
"max(max(min(?x:Nat,?x:Nat),0),?x:Nat) = plus(max(?z:Nat,0),s(min(?x:Nat,max(?x:Nat,?y:Nat))))",
"max(min(?y:Nat,?y:Nat),plus(?x:Nat,?y:Nat)) = plus(max(?z:Nat,s(max(?z:Nat,min(?y:Nat,?x:Nat)))),?x:Nat)",
"plus(?x:Nat,max(plus(?x:Nat,?x:Nat),0)) = min(s(s(plus(s(?y:Nat),min(?x:Nat,?y:Nat)))),?x:Nat)",
"s(s(s(max(s(0),s(?x:Nat))))) = s(plus(?x:Nat,max(s(max(?x:Nat,?y:Nat)),?y:Nat)))",
"s(min(plus(plus(0,?y:Nat),?x:Nat),0)) = s(plus(max(s(plus(?y:Nat,?z:Nat)),?z:Nat),?y:Nat))",
"s(min(plus(s(0),s(s(?x:Nat))),0)) = s(plus(0,s(max(s(?z:Nat),?x:Nat))))",
"max(s(plus(s(min(0,?y:Nat)),?z:Nat)),0) = min(max(?y:Nat,plus(s(?z:Nat),?x:Nat)),?z:Nat)",
"max(0,min(s(s(min(?y:Nat,?x:Nat))),s(?x:Nat))) = plus(plus(s(?y:Nat),s(?x:Nat)),0)",
"plus(0,s(min(plus(min(0,?y:Nat),?y:Nat),?y:Nat))) = plus(s(max(?y:Nat,?z:Nat)),s(?x:Nat))",
"plus(?z:Nat,plus(s(?z:Nat),max(?x:Nat,min(s(?x:Nat),?z:Nat)))) = min(min(?x:Nat,s(?x:Nat)),?x:Nat)",
"plus(0,plus(s(?y:Nat),min(?x:Nat,s(plus(?x:Nat,?z:Nat))))) = min(s(?x:Nat),plus(?z:Nat,?x:Nat))",
"s(s(plus(max(s(?y:Nat),0),plus(?y:Nat,plus(0,?x:Nat))))) = min(max(?x:Nat,?x:Nat),?z:Nat)",
"s(min(min(plus(plus(?x:Nat,?z:Nat),plus(?x:Nat,?x:Nat)),?x:Nat),?y:Nat)) = min(s(?x:Nat),s(?x:Nat))",
"s(plus(min(min(0,?y:Nat),min(plus(?y:Nat,?z:Nat),?y:Nat)),?y:Nat)) = s(max(?z:Nat,s(0)))",
"plus(plus(?y:Nat,max(?x:Nat,max(plus(?x:Nat,0),s(0)))),?z:Nat) = min(?y:Nat,s(s(?y:Nat)))",
"plus(s(max(?y:Nat,?z:Nat)),s(plus(s(?z:Nat),s(plus(?z:Nat,?y:Nat))))) = plus(s(?y:Nat),0)",
"plus(plus(min(s(0),min(s(?x:Nat),plus(?y:Nat,0))),?x:Nat),?z:Nat) = plus(?y:Nat,s(0))",
"s(s(min(min(s(?z:Nat),max(?x:Nat,?x:Nat)),max(s(?z:Nat),s(0))))) = plus(?z:Nat,?z:Nat)",
"s(min(plus(max(max(?z:Nat,?y:Nat),max(?y:Nat,?x:Nat)),s(0)),s(?y:Nat))) = plus(?x:Nat,?x:Nat)",
"s(plus(plus(max(max(?x:Nat,?y:Nat),?x:Nat),plus(plus(?x:Nat,?z:Nat),?x:Nat)),?x:Nat)) = max(?x:Nat,?x:Nat)",
"max(max(?x:Nat,s(s(plus(?x:Nat,?z:Nat)))),plus(plus(?z:Nat,s(?z:Nat)),?x:Nat)) = plus(?z:Nat,?x:Nat)",
"min(?x:Nat,plus(plus(0,plus(s(?x:Nat),?y:Nat)),plus(?x:Nat,s(s(?x:Nat))))) = max(?x:Nat,?z:Nat)",
"plus(?x:Nat,max(s(?y:Nat),min(min(max(?x:Nat,0),min(?x:Nat,?x:Nat)),?x:Nat))) = max(?x:Nat,?z:Nat)",
"s(plus(max(plus(min(?z:Nat,?y:Nat),min(?y:Nat,?z:Nat)),plus(?z:Nat,0)),s(0))) = s(?y:Nat)",
"min(max(?z:Nat,max(s(0),?x:Nat)),min(s(0),max(plus(?x:Nat,?x:Nat),?z:Nat))) = s(0)",
"plus(min(0,?z:Nat),max(?x:Nat,min(plus(max(?y:Nat,?x:Nat),plus(?z:Nat,?x:Nat)),0))) = s(?x:Nat)",
"plus(plus(s(s(s(?x:Nat))),plus(0,min(?y:Nat,s(?x:Nat)))),s(s(?y:Nat))) = s(?z:Nat)",
"s(?x:Nat) = s(max(max(plus(s(?x:Nat),0),plus(min(?z:Nat,?y:Nat),s(?z:Nat))),min(?x:Nat,?x:Nat)))",
"s(?x:Nat) = s(min(plus(max(0,s(0)),plus(plus(?y:Nat,0),min(?x:Nat,0))),s(0)))",
"s(?x:Nat) = max(?x:Nat,plus(plus(s(0),plus(?z:Nat,0)),max(max(?z:Nat,min(?x:Nat,?x:Nat)),?x:Nat)))",
"s(?x:Nat) = max(plus(plus(max(plus(?z:Nat,0),max(?z:Nat,?x:Nat)),s(0)),s(?x:Nat)),s(?z:Nat))",
"s(?x:Nat) = min(s(?x:Nat),max(plus(max(?z:Nat,s(0)),s(0)),max(?y:Nat,max(?x:Nat,?x:Nat))))",
"s(?x:Nat) = plus(max(max(max(0,?x:Nat),min(plus(?y:Nat,0),max(?y:Nat,0))),s(0)),?x:Nat)",
"s(?y:Nat) = max(?x:Nat,plus(0,min(plus(min(?x:Nat,?x:Nat),s(0)),plus(?x:Nat,plus(?x:Nat,?y:Nat)))))",
"s(?z:Nat) = s(plus(s(min(?x:Nat,s(?y:Nat))),plus(min(?x:Nat,?x:Nat),plus(s(?z:Nat),s(0)))))",
"s(?z:Nat) = min(plus(s(min(?x:Nat,plus(?x:Nat,0))),max(plus(s(?z:Nat),s(0)),?x:Nat)),?x:Nat)",
"s(0) = max(max(?z:Nat,?y:Nat),min(0,plus(plus(?x:Nat,min(0,?x:Nat)),min(?x:Nat,s(?y:Nat)))))",
"s(s(?x:Nat)) = s(plus(s(max(0,s(?y:Nat))),plus(?x:Nat,max(plus(?x:Nat,?x:Nat),s(?x:Nat)))))",
"max(?x:Nat,?x:Nat) = max(min(?x:Nat,plus(plus(s(?x:Nat),plus(?y:Nat,?z:Nat)),plus(s(0),?z:Nat))),?z:Nat)",
"max(?z:Nat,0) = max(plus(plus(0,min(?x:Nat,s(?x:Nat))),s(min(?x:Nat,plus(?y:Nat,?z:Nat)))),?y:Nat)",
"min(?x:Nat,?x:Nat) = max(plus(?z:Nat,plus(?x:Nat,?x:Nat)),min(min(?y:Nat,min(s(?x:Nat),?x:Nat)),s(0)))",
"plus(?x:Nat,?z:Nat) = s(s(max(plus(max(?x:Nat,?x:Nat),min(0,?x:Nat)),plus(?x:Nat,max(?x:Nat,?x:Nat)))))",
"plus(?x:Nat,?z:Nat) = max(min(s(plus(max(?z:Nat,?z:Nat),plus(?z:Nat,?y:Nat))),plus(0,?x:Nat)),s(0))",
"plus(?y:Nat,?y:Nat) = plus(plus(?y:Nat,max(s(min(?y:Nat,0)),min(?x:Nat,max(?x:Nat,0)))),s(?y:Nat))",
"s(s(s(0))) = s(plus(plus(max(max(0,?x:Nat),?z:Nat),s(?x:Nat)),min(s(0),0)))",
"max(s(?z:Nat),?y:Nat) = max(?x:Nat,max(min(plus(s(?z:Nat),?z:Nat),plus(min(?z:Nat,?x:Nat),?y:Nat)),?y:Nat))",
"plus(?x:Nat,s(?y:Nat)) = s(min(0,plus(min(max(?x:Nat,?x:Nat),s(?z:Nat)),s(max(?x:Nat,?x:Nat)))))",
"min(s(?y:Nat),s(0)) = min(?x:Nat,min(min(s(plus(?z:Nat,?z:Nat)),max(?x:Nat,?y:Nat)),s(0)))",
"plus(?x:Nat,min(?x:Nat,?y:Nat)) = plus(0,min(?z:Nat,min(max(plus(?x:Nat,0),min(?y:Nat,?z:Nat)),?z:Nat)))",
"plus(plus(?x:Nat,0),s(?y:Nat)) = plus(s(max(?x:Nat,plus(max(?x:Nat,?x:Nat),max(?z:Nat,?x:Nat)))),?x:Nat)",
"s(s(s(plus(0,s(?x:Nat))))) = plus(?z:Nat,min(min(min(plus(?y:Nat,?x:Nat),?z:Nat),?y:Nat),?x:Nat))",
"plus(s(s(min(?x:Nat,0))),?z:Nat) = plus(s(max(?x:Nat,plus(?x:Nat,min(?z:Nat,?x:Nat)))),s(?x:Nat))",
"plus(plus(s(0),?y:Nat),s(?x:Nat)) = plus(0,plus(min(0,plus(?z:Nat,plus(0,?x:Nat))),?x:Nat))",
"plus(plus(min(?x:Nat,?z:Nat),?x:Nat),?y:Nat) = plus(plus(max(min(plus(?y:Nat,0),?y:Nat),0),0),?y:Nat)",
"s(max(min(max(?y:Nat,0),?z:Nat),?y:Nat)) = plus(?z:Nat,s(min(s(s(?x:Nat)),min(?x:Nat,?x:Nat))))",
"min(?z:Nat,plus(0,s(max(s(?z:Nat),0)))) = s(s(plus(s(s(?x:Nat)),plus(?x:Nat,?x:Nat))))",
"plus(?x:Nat,min(?z:Nat,plus(s(s(0)),?x:Nat))) = s(max(s(plus(?z:Nat,plus(0,?x:Nat))),?x:Nat))",
"plus(?z:Nat,s(max(max(0,s(?y:Nat)),0))) = max(plus(s(?x:Nat),0),plus(?y:Nat,s(?x:Nat)))",
"s(plus(s(0),plus(min(s(?x:Nat),0),?z:Nat))) = max(0,max(max(0,?y:Nat),s(?x:Nat)))",
"plus(max(0,s(plus(max(?y:Nat,?z:Nat),s(?x:Nat)))),?y:Nat) = s(max(?y:Nat,s(s(s(0)))))",
"s(plus(plus(plus(min(?x:Nat,0),?x:Nat),s(0)),min(0,0))) = max(s(?z:Nat),s(?x:Nat))",
"max(?y:Nat,max(max(min(s(?x:Nat),plus(?z:Nat,?z:Nat)),s(?x:Nat)),?y:Nat)) = min(0,s(s(?x:Nat)))",
"plus(0,max(?x:Nat,plus(s(s(?x:Nat)),min(plus(?z:Nat,?y:Nat),?x:Nat)))) = min(plus(?x:Nat,?z:Nat),?x:Nat)",
"s(s(max(max(min(?x:Nat,0),?x:Nat),plus(s(0),plus(?x:Nat,0))))) = plus(0,s(?z:Nat))",
"s(plus(max(s(min(?x:Nat,?z:Nat)),min(s(?y:Nat),min(?x:Nat,?y:Nat))),?x:Nat)) = plus(s(?y:Nat),?z:Nat)",
"min(min(s(plus(min(?y:Nat,?z:Nat),min(0,?y:Nat))),0),min(?y:Nat,?x:Nat)) = max(s(?x:Nat),?x:Nat)",
"plus(0,max(plus(?z:Nat,?x:Nat),min(?z:Nat,min(plus(?x:Nat,?x:Nat),s(?y:Nat))))) = min(0,s(?x:Nat))",
"plus(plus(max(s(s(?z:Nat)),?x:Nat),?z:Nat),min(s(s(s(?x:Nat))),?x:Nat)) = max(0,s(?y:Nat))",
"s(s(plus(plus(?z:Nat,plus(?x:Nat,?x:Nat)),plus(plus(0,0),min(0,0))))) = max(?y:Nat,?x:Nat)",
"s(max(0,plus(max(plus(?x:Nat,?x:Nat),min(?y:Nat,?z:Nat)),s(min(?y:Nat,0))))) = plus(?x:Nat,?x:Nat)",
"s(plus(plus(plus(s(?x:Nat),s(0)),?x:Nat),s(min(plus(?x:Nat,?z:Nat),?x:Nat)))) = plus(?x:Nat,?x:Nat)",
"max(min(plus(plus(?z:Nat,s(0)),plus(max(?z:Nat,?z:Nat),s(0))),?x:Nat),?y:Nat) = max(0,?x:Nat)",
"min(min(?x:Nat,min(?x:Nat,s(s(?x:Nat)))),s(s(plus(plus(?y:Nat,0),?y:Nat)))) = max(?y:Nat,?z:Nat)",
"min(plus(max(0,0),plus(plus(plus(0,?y:Nat),plus(?y:Nat,?x:Nat)),?x:Nat)),?x:Nat) = min(0,?z:Nat)",
"plus(?x:Nat,max(min(min(max(?y:Nat,?x:Nat),?x:Nat),plus(?x:Nat,max(?x:Nat,?y:Nat))),?y:Nat)) = s(s(?y:Nat))",
"plus(s(?x:Nat),s(plus(s(min(?x:Nat,?y:Nat)),max(plus(?y:Nat,?x:Nat),s(?x:Nat))))) = min(?x:Nat,0)",
"plus(s(plus(max(plus(?x:Nat,?x:Nat),min(0,?x:Nat)),plus(?x:Nat,?z:Nat))),s(?x:Nat)) = plus(?y:Nat,?x:Nat)",
"plus(plus(min(max(max(?x:Nat,?z:Nat),min(?y:Nat,?x:Nat)),s(?x:Nat)),?z:Nat),s(0)) = plus(?y:Nat,?x:Nat)",
"plus(plus(plus(?z:Nat,?x:Nat),plus(min(plus(0,0),?y:Nat),min(0,?x:Nat))),0) = min(?x:Nat,?z:Nat)",
"s(min(plus(plus(?x:Nat,s(?x:Nat)),s(0)),s(plus(max(0,?x:Nat),s(?z:Nat))))) = s(?y:Nat)",
"s(plus(plus(?x:Nat,s(s(0))),min(max(plus(0,0),max(?x:Nat,?z:Nat)),?y:Nat))) = s(0)",
"max(?z:Nat,s(plus(plus(plus(?z:Nat,?z:Nat),?y:Nat),min(min(?x:Nat,?x:Nat),max(?z:Nat,?y:Nat))))) = s(0)",
"min(?x:Nat,s(plus(max(min(0,?x:Nat),max(?x:Nat,?y:Nat)),max(s(?z:Nat),s(?x:Nat))))) = s(?z:Nat)",
"min(?x:Nat,plus(s(min(?z:Nat,max(?x:Nat,?x:Nat))),s(plus(s(?x:Nat),max(?y:Nat,?x:Nat))))) = s(?y:Nat)",
"min(s(?y:Nat),min(plus(?y:Nat,s(plus(?x:Nat,?z:Nat))),max(?z:Nat,s(plus(?x:Nat,0))))) = s(?z:Nat)",
"min(min(?y:Nat,?z:Nat),plus(s(s(?z:Nat)),s(max(max(0,?z:Nat),plus(?x:Nat,?x:Nat))))) = s(?x:Nat)",
"plus(?z:Nat,plus(min(s(?y:Nat),s(max(0,?x:Nat))),s(max(?y:Nat,plus(0,?x:Nat))))) = s(?x:Nat)",
"plus(min(?y:Nat,s(s(s(?z:Nat)))),min(s(0),max(?x:Nat,s(min(?x:Nat,?x:Nat))))) = s(0)",
"plus(plus(?z:Nat,max(plus(?y:Nat,s(?x:Nat)),min(plus(?x:Nat,?y:Nat),max(?z:Nat,?y:Nat)))),?x:Nat) = s(?x:Nat)",
"s(?y:Nat) = plus(?x:Nat,min(s(min(max(?x:Nat,?x:Nat),min(?z:Nat,?x:Nat))),s(plus(max(0,?z:Nat),0))))",
"s(?z:Nat) = plus(?y:Nat,plus(?x:Nat,max(max(plus(?y:Nat,?x:Nat),max(0,0)),plus(?x:Nat,min(?x:Nat,?x:Nat)))))",
"s(?z:Nat) = plus(min(0,min(plus(?x:Nat,s(?y:Nat)),s(max(?x:Nat,?z:Nat)))),s(s(s(s(?y:Nat)))))",
"s(0) = s(plus(plus(min(plus(?x:Nat,?x:Nat),max(?x:Nat,?x:Nat)),min(s(?z:Nat),0)),plus(0,?z:Nat)))",
"s(s(?x:Nat)) = max(s(s(max(?x:Nat,?x:Nat))),min(plus(min(min(?x:Nat,?x:Nat),?x:Nat),?x:Nat),s(?y:Nat)))",
"s(s(?y:Nat)) = max(?x:Nat,min(plus(plus(s(?z:Nat),0),plus(max(?x:Nat,0),max(0,?x:Nat))),?z:Nat))",
"max(?y:Nat,?x:Nat) = plus(?z:Nat,max(s(?y:Nat),min(min(?z:Nat,min(0,?z:Nat)),max(?x:Nat,plus(0,?y:Nat)))))",
"plus(?y:Nat,s(?x:Nat)) = max(s(0),min(min(min(plus(?y:Nat,?y:Nat),s(0)),plus(?y:Nat,0)),?y:Nat))",
"plus(?y:Nat,s(?x:Nat)) = max(max(s(plus(min(0,0),min(?z:Nat,?z:Nat))),min(?y:Nat,?x:Nat)),s(?x:Nat))",
"max(0,max(?x:Nat,?z:Nat)) = plus(min(s(s(max(?z:Nat,?x:Nat))),plus(s(min(?z:Nat,?x:Nat)),?z:Nat)),?y:Nat)",
"min(plus(?z:Nat,?z:Nat),?z:Nat) = max(?x:Nat,s(max(min(plus(0,?z:Nat),s(?x:Nat)),s(min(?y:Nat,?z:Nat)))))",
"plus(?y:Nat,s(s(?z:Nat))) = min(?z:Nat,plus(max(s(max(?x:Nat,?x:Nat)),max(?x:Nat,max(?z:Nat,?y:Nat))),0))",
"plus(0,max(?x:Nat,?x:Nat)) = plus(plus(min(min(s(?y:Nat),plus(?x:Nat,?x:Nat)),?x:Nat),?z:Nat),min(?z:Nat,0))",
"s(min(?x:Nat,min(?z:Nat,?x:Nat))) = s(plus(plus(s(plus(?y:Nat,?z:Nat)),?x:Nat),s(s(min(?x:Nat,?x:Nat)))))",
"plus(s(?x:Nat),s(s(0))) = max(min(?x:Nat,max(?x:Nat,s(min(?y:Nat,?x:Nat)))),plus(s(0),?y:Nat))",
"s(s(s(s(min(?x:Nat,?z:Nat))))) = plus(?z:Nat,min(plus(s(s(?y:Nat)),max(?y:Nat,s(0))),0))",
"s(plus(?y:Nat,plus(s(?y:Nat),?x:Nat))) = min(min(s(min(plus(?x:Nat,?y:Nat),max(?x:Nat,?x:Nat))),?y:Nat),?y:Nat)",
"plus(0,max(max(?y:Nat,?x:Nat),0)) = s(min(min(max(plus(?x:Nat,?x:Nat),max(?x:Nat,0)),?y:Nat),?y:Nat))",
"plus(s(s(s(s(?z:Nat)))),?y:Nat) = min(?y:Nat,min(max(plus(plus(0,?x:Nat),s(0)),?z:Nat),0))",
"max(s(plus(s(s(?z:Nat)),0)),?z:Nat) = plus(s(s(plus(plus(?x:Nat,0),max(0,0)))),?y:Nat)",
"plus(plus(min(0,?x:Nat),s(0)),?x:Nat) = s(min(max(0,?x:Nat),max(s(min(0,?x:Nat)),?y:Nat)))",
"s(max(max(s(?x:Nat),0),max(?x:Nat,0))) = s(plus(min(s(plus(?y:Nat,?z:Nat)),?z:Nat),s(?x:Nat)))",
"s(min(s(?x:Nat),s(s(plus(?x:Nat,?x:Nat))))) = s(min(?x:Nat,plus(?x:Nat,max(max(?z:Nat,?y:Nat),?x:Nat))))",
"max(plus(plus(s(?x:Nat),?z:Nat),s(?x:Nat)),?x:Nat) = plus(plus(0,s(plus(s(?z:Nat),s(0)))),0)",
"min(0,s(s(plus(plus(0,?y:Nat),?z:Nat)))) = max(max(s(plus(plus(?y:Nat,0),?x:Nat)),?x:Nat),0)",
"s(s(min(plus(s(?x:Nat),max(?z:Nat,?z:Nat)),?x:Nat))) = s(max(?x:Nat,max(?y:Nat,s(plus(?z:Nat,?z:Nat)))))",
"min(plus(s(0),s(plus(s(?z:Nat),?x:Nat))),?z:Nat) = s(max(?y:Nat,s(plus(0,plus(?y:Nat,?x:Nat)))))",
"s(min(?y:Nat,plus(min(s(?z:Nat),min(?x:Nat,?x:Nat)),0))) = min(?x:Nat,max(plus(s(0),?x:Nat),0))",
"s(plus(min(plus(?z:Nat,0),s(min(?x:Nat,?z:Nat))),?x:Nat)) = plus(s(s(?z:Nat)),min(s(0),?y:Nat))",
"min(plus(?z:Nat,min(?x:Nat,plus(?x:Nat,min(?x:Nat,?x:Nat)))),?y:Nat) = max(s(s(s(plus(0,?x:Nat)))),0)",
"plus(s(max(s(s(?z:Nat)),s(plus(?z:Nat,?y:Nat)))),?x:Nat) = plus(?y:Nat,plus(plus(s(0),?y:Nat),?x:Nat))",
"max(?x:Nat,plus(0,s(min(max(?z:Nat,?y:Nat),min(0,?y:Nat))))) = plus(?x:Nat,max(?x:Nat,min(?y:Nat,?y:Nat)))",
"max(max(plus(?z:Nat,min(s(?x:Nat),s(?x:Nat))),s(?x:Nat)),?y:Nat) = s(max(plus(?x:Nat,s(?y:Nat)),?x:Nat))",
"plus(plus(plus(?x:Nat,max(plus(?z:Nat,?x:Nat),?z:Nat)),s(?z:Nat)),?z:Nat) = plus(plus(?y:Nat,plus(?y:Nat,?z:Nat)),?z:Nat)",
"plus(min(?y:Nat,min(s(min(?z:Nat,?x:Nat)),s(plus(?x:Nat,0)))),?x:Nat) = max(s(max(?z:Nat,0)),?y:Nat)",
"s(min(min(0,max(?x:Nat,max(0,0))),s(s(plus(0,?x:Nat))))) = s(max(0,s(?x:Nat)))",
"max(?z:Nat,s(plus(max(plus(?y:Nat,?y:Nat),?x:Nat),plus(max(?y:Nat,?x:Nat),?x:Nat)))) = plus(0,plus(?x:Nat,0))",
"plus(?x:Nat,min(plus(s(0),?z:Nat),min(min(0,plus(?x:Nat,?y:Nat)),?x:Nat))) = plus(plus(?x:Nat,?y:Nat),0)",
"s(s(plus(plus(plus(0,?x:Nat),min(?z:Nat,?y:Nat)),min(s(0),s(?z:Nat))))) = min(s(?y:Nat),?z:Nat)",
"max(s(s(max(min(?x:Nat,?x:Nat),s(?y:Nat)))),plus(s(max(?y:Nat,?z:Nat)),?x:Nat)) = max(?y:Nat,s(?y:Nat))",
"max(plus(min(?x:Nat,s(s(?x:Nat))),min(s(0),s(s(0)))),s(?x:Nat)) = s(s(s(0)))",
"min(plus(min(min(?y:Nat,min(?y:Nat,?z:Nat)),min(?x:Nat,min(?z:Nat,?x:Nat))),?x:Nat),?x:Nat) = s(plus(?y:Nat,?y:Nat))",
"plus(0,plus(plus(min(?x:Nat,s(?x:Nat)),min(0,plus(?y:Nat,0))),s(0))) = plus(?x:Nat,s(?x:Nat))",
"plus(max(plus(0,?x:Nat),max(min(max(?x:Nat,?y:Nat),min(?x:Nat,?x:Nat)),?x:Nat)),?x:Nat) = max(s(?y:Nat),0)",
"plus(max(plus(0,?x:Nat),plus(s(?y:Nat),max(?x:Nat,min(?x:Nat,?y:Nat)))),s(?x:Nat)) = plus(s(?z:Nat),?x:Nat)",
"s(plus(min(min(s(0),s(?z:Nat)),?x:Nat),plus(?x:Nat,min(0,plus(?z:Nat,?x:Nat))))) = max(?z:Nat,?y:Nat)",
"max(?z:Nat,plus(min(?x:Nat,plus(min(?x:Nat,?x:Nat),max(?y:Nat,?x:Nat))),max(s(?z:Nat),?x:Nat))) = max(?y:Nat,?x:Nat)",
"max(max(?z:Nat,s(?x:Nat)),s(max(plus(s(?y:Nat),?x:Nat),min(plus(?x:Nat,?x:Nat),?z:Nat)))) = plus(?x:Nat,?x:Nat)",
"min(?z:Nat,min(s(s(plus(?z:Nat,?x:Nat))),max(s(min(?x:Nat,0)),s(s(?x:Nat))))) = plus(?y:Nat,0)",
"min(0,max(plus(?x:Nat,plus(s(?x:Nat),max(?x:Nat,?z:Nat))),s(s(plus(?x:Nat,?x:Nat))))) = max(?z:Nat,0)",
"min(s(plus(plus(max(0,?z:Nat),plus(?x:Nat,0)),min(?y:Nat,plus(?x:Nat,?x:Nat)))),?x:Nat) = max(?x:Nat,?y:Nat)",
"min(plus(?x:Nat,plus(?z:Nat,plus(?y:Nat,plus(?y:Nat,?x:Nat)))),s(s(plus(s(?x:Nat),?x:Nat)))) = max(?x:Nat,?z:Nat)",
"plus(?x:Nat,max(s(max(plus(?z:Nat,?x:Nat),s(?x:Nat))),min(?x:Nat,s(max(?z:Nat,?z:Nat))))) = max(?z:Nat,?x:Nat)",
"plus(max(?x:Nat,max(?x:Nat,?x:Nat)),min(s(0),min(max(?y:Nat,plus(0,?y:Nat)),?z:Nat))) = plus(?y:Nat,?x:Nat)",
"plus(plus(?y:Nat,plus(plus(s(?z:Nat),plus(?x:Nat,?y:Nat)),0)),s(min(?z:Nat,s(?x:Nat)))) = max(0,?z:Nat)",
"s(min(s(s(plus(?z:Nat,0))),min(plus(s(0),plus(0,?x:Nat)),plus(?x:Nat,0)))) = s(0)",
"max(s(s(plus(?x:Nat,s(0)))),s(max(s(?x:Nat),plus(max(?x:Nat,0),s(?z:Nat))))) = s(?z:Nat)",
"max(max(?x:Nat,s(?x:Nat)),min(plus(s(0),max(max(?x:Nat,?y:Nat),min(?z:Nat,?x:Nat))),?y:Nat)) = s(?x:Nat)",
"min(s(?x:Nat),plus(?x:Nat,max(plus(plus(?x:Nat,?x:Nat),?x:Nat),min(max(?z:Nat,?x:Nat),s(?y:Nat))))) = s(?x:Nat)",
"min(plus(?x:Nat,0),plus(plus(s(?y:Nat),s(max(?x:Nat,?z:Nat))),s(max(s(?z:Nat),?z:Nat)))) = s(?z:Nat)",
"plus(s(min(max(min(?z:Nat,?z:Nat),?x:Nat),plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat)))),s(?x:Nat)) = s(?x:Nat)",
"plus(max(?y:Nat,?x:Nat),s(max(min(min(?x:Nat,?x:Nat),s(?x:Nat)),plus(min(?y:Nat,?y:Nat),?y:Nat)))) = s(?z:Nat)",
"plus(min(s(plus(?z:Nat,plus(?x:Nat,0))),max(max(min(?x:Nat,?z:Nat),?x:Nat),s(?y:Nat))),?x:Nat) = s(?z:Nat)",
"plus(plus(s(plus(?z:Nat,plus(?z:Nat,?x:Nat))),s(min(plus(0,0),plus(?z:Nat,0)))),?x:Nat) = s(?z:Nat)",
"s(?x:Nat) = min(s(plus(plus(min(?x:Nat,?x:Nat),max(?y:Nat,?x:Nat)),plus(s(?x:Nat),max(?z:Nat,?z:Nat)))),s(0))",
"s(?x:Nat) = min(min(0,s(max(?y:Nat,plus(0,?x:Nat)))),max(max(?x:Nat,max(plus(?z:Nat,?x:Nat),?z:Nat)),0))",
"s(?x:Nat) = plus(min(s(?x:Nat),?x:Nat),plus(max(?z:Nat,max(?x:Nat,?x:Nat)),max(min(?x:Nat,0),plus(?x:Nat,?z:Nat))))",
"s(?y:Nat) = s(max(plus(plus(?x:Nat,s(0)),plus(min(?y:Nat,0),?x:Nat)),s(min(0,max(?x:Nat,?x:Nat)))))",
"s(?y:Nat) = max(plus(?y:Nat,0),min(max(plus(s(?y:Nat),?x:Nat),s(?z:Nat)),plus(s(s(?x:Nat)),s(?z:Nat))))",
"s(?y:Nat) = min(max(?x:Nat,max(min(max(?x:Nat,?z:Nat),s(?z:Nat)),max(plus(?x:Nat,?z:Nat),plus(?z:Nat,?x:Nat)))),?x:Nat)",
"s(?y:Nat) = plus(0,plus(max(s(max(0,?x:Nat)),s(?x:Nat)),plus(s(?x:Nat),plus(?y:Nat,min(?z:Nat,?x:Nat)))))",
"s(0) = plus(plus(s(?x:Nat),max(max(max(?y:Nat,0),min(?z:Nat,?y:Nat)),min(?x:Nat,min(?x:Nat,?x:Nat)))),?z:Nat)",
"min(0,0) = min(plus(max(0,plus(s(?z:Nat),?y:Nat)),s(min(max(?x:Nat,?x:Nat),max(?y:Nat,?x:Nat)))),?x:Nat)",
"plus(?x:Nat,?z:Nat) = min(max(plus(?x:Nat,plus(s(?y:Nat),min(?y:Nat,?z:Nat))),s(max(s(?y:Nat),s(?y:Nat)))),?x:Nat)",
"plus(?y:Nat,?x:Nat) = min(plus(?x:Nat,s(plus(s(?z:Nat),plus(?x:Nat,?x:Nat)))),plus(?z:Nat,s(s(plus(0,?y:Nat)))))",
"s(max(?x:Nat,?x:Nat)) = max(plus(0,s(plus(?z:Nat,min(?y:Nat,?x:Nat)))),s(min(min(0,s(?x:Nat)),?x:Nat)))",
"plus(?y:Nat,s(?z:Nat)) = plus(?x:Nat,min(min(plus(?x:Nat,min(0,?y:Nat)),s(plus(?x:Nat,?x:Nat))),s(s(?x:Nat))))",
"s(s(plus(?z:Nat,s(s(?x:Nat))))) = min(plus(min(?z:Nat,plus(max(?z:Nat,?x:Nat),plus(?z:Nat,?z:Nat))),?y:Nat),?z:Nat)",
"s(s(plus(s(s(?x:Nat)),?y:Nat))) = min(plus(min(s(max(?x:Nat,?x:Nat)),s(min(?z:Nat,0))),?x:Nat),?z:Nat)",
"max(plus(max(?x:Nat,?z:Nat),?y:Nat),?x:Nat) = max(?y:Nat,s(max(min(?y:Nat,max(?y:Nat,?z:Nat)),max(s(?x:Nat),?x:Nat))))",
"plus(?y:Nat,min(?x:Nat,plus(0,?x:Nat))) = min(plus(?y:Nat,?y:Nat),plus(s(min(s(?z:Nat),min(?x:Nat,?y:Nat))),?z:Nat))",
"plus(?z:Nat,plus(plus(?x:Nat,?x:Nat),?x:Nat)) = plus(?z:Nat,s(max(s(?y:Nat),max(min(?z:Nat,?x:Nat),plus(?z:Nat,?x:Nat)))))",
"plus(max(s(plus(0,0)),?z:Nat),?y:Nat) = max(s(plus(max(min(?z:Nat,?z:Nat),min(?x:Nat,?z:Nat)),?x:Nat)),?x:Nat)",
"s(s(plus(?x:Nat,plus(plus(?y:Nat,?x:Nat),?y:Nat)))) = plus(plus(s(s(min(?x:Nat,?y:Nat))),?z:Nat),min(?y:Nat,?x:Nat))",
"max(s(min(0,s(min(?x:Nat,?x:Nat)))),?z:Nat) = s(max(s(plus(plus(?y:Nat,?z:Nat),max(?z:Nat,?x:Nat))),0))",
"max(plus(?y:Nat,s(?x:Nat)),min(s(?x:Nat),?x:Nat)) = s(max(s(plus(max(?x:Nat,?x:Nat),plus(?x:Nat,0))),?x:Nat))",
"min(max(?x:Nat,plus(s(s(?x:Nat)),?y:Nat)),?x:Nat) = s(max(?x:Nat,max(s(min(?x:Nat,?y:Nat)),plus(0,?y:Nat))))",
"min(min(0,s(s(?y:Nat))),plus(?x:Nat,?y:Nat)) = plus(?x:Nat,max(s(?z:Nat),s(min(0,max(0,0)))))",
"plus(?x:Nat,s(plus(?x:Nat,plus(?z:Nat,s(0))))) = s(max(max(s(?x:Nat),plus(?x:Nat,plus(0,?y:Nat))),?x:Nat))",
"plus(?z:Nat,s(plus(?x:Nat,min(0,s(?z:Nat))))) = min(0,s(min(s(?x:Nat),plus(?x:Nat,plus(0,?x:Nat)))))",
"s(max(max(plus(0,plus(?x:Nat,?x:Nat)),?x:Nat),?y:Nat)) = max(plus(s(s(?y:Nat)),min(s(?x:Nat),0)),?z:Nat)",
"min(0,s(max(max(?z:Nat,min(?x:Nat,?x:Nat)),?x:Nat))) = s(min(?y:Nat,max(?x:Nat,max(plus(?x:Nat,0),?z:Nat))))",
"s(s(max(s(s(?z:Nat)),plus(plus(?z:Nat,?x:Nat),?z:Nat)))) = s(plus(?x:Nat,min(min(s(?x:Nat),?x:Nat),?x:Nat)))",
"s(s(plus(s(?x:Nat),plus(s(0),min(?y:Nat,?y:Nat))))) = max(?z:Nat,s(s(min(?z:Nat,min(0,?y:Nat)))))",
"s(plus(0,s(min(max(?x:Nat,?x:Nat),min(?z:Nat,?y:Nat))))) = min(plus(s(s(plus(?x:Nat,0))),?x:Nat),?x:Nat)",
"min(plus(max(?y:Nat,min(0,max(0,?x:Nat))),?y:Nat),?x:Nat) = max(plus(0,plus(s(s(?z:Nat)),?x:Nat)),?x:Nat)",
"plus(s(max(?x:Nat,plus(max(?z:Nat,?x:Nat),?z:Nat))),s(0)) = min(max(?x:Nat,s(s(min(?z:Nat,?y:Nat)))),?x:Nat)",
"plus(plus(?z:Nat,?z:Nat),s(s(min(max(?y:Nat,?x:Nat),?y:Nat)))) = s(s(plus(plus(max(?z:Nat,?x:Nat),?x:Nat),0)))",
"plus(plus(0,plus(min(?x:Nat,s(0)),?y:Nat)),s(0)) = s(plus(?x:Nat,s(max(?x:Nat,min(?x:Nat,?y:Nat)))))",
"s(max(?x:Nat,max(plus(min(?y:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),?y:Nat))) = min(plus(max(?x:Nat,?z:Nat),s(?x:Nat)),?x:Nat)",
"max(min(0,s(plus(plus(?z:Nat,?x:Nat),plus(?y:Nat,?y:Nat)))),?y:Nat) = s(s(s(plus(0,min(0,?y:Nat)))))",
"plus(?x:Nat,min(max(?z:Nat,min(plus(?y:Nat,?x:Nat),s(?x:Nat))),?y:Nat)) = s(min(?x:Nat,s(plus(s(?y:Nat),?x:Nat))))",
"plus(min(0,s(min(plus(?x:Nat,0),min(?y:Nat,?x:Nat)))),s(?y:Nat)) = max(max(min(?x:Nat,?x:Nat),?z:Nat),?x:Nat)",
"plus(min(s(plus(plus(?x:Nat,?x:Nat),min(?z:Nat,?y:Nat))),?z:Nat),s(?z:Nat)) = s(s(s(min(0,s(?x:Nat)))))",
"s(min(s(s(max(?z:Nat,?x:Nat))),s(plus(min(?x:Nat,?y:Nat),s(?x:Nat))))) = plus(?y:Nat,plus(?x:Nat,s(0)))",
"max(s(min(plus(min(0,?x:Nat),plus(?x:Nat,?x:Nat)),s(?x:Nat))),s(?x:Nat)) = max(s(?z:Nat),min(?y:Nat,?z:Nat))",
"max(min(s(0),min(?y:Nat,max(?x:Nat,max(?x:Nat,?y:Nat)))),s(s(?x:Nat))) = s(s(plus(s(?z:Nat),?y:Nat)))",
"min(plus(0,min(plus(s(?z:Nat),0),s(?y:Nat))),min(min(?x:Nat,?x:Nat),?y:Nat)) = min(plus(?x:Nat,?y:Nat),?z:Nat)",
"plus(s(s(min(max(?y:Nat,?x:Nat),plus(?z:Nat,?x:Nat)))),s(s(min(0,?z:Nat)))) = min(?y:Nat,min(?x:Nat,?x:Nat))",
"min(?x:Nat,max(plus(plus(?z:Nat,s(?y:Nat)),plus(max(?z:Nat,?z:Nat),max(0,0))),?x:Nat)) = plus(s(?y:Nat),?x:Nat)",
"min(min(s(s(s(0))),plus(0,min(min(?x:Nat,?y:Nat),min(?x:Nat,?x:Nat)))),0) = min(?z:Nat,s(?x:Nat))",
"plus(?z:Nat,plus(max(min(plus(?y:Nat,?x:Nat),plus(?z:Nat,?x:Nat)),s(max(?z:Nat,0))),?x:Nat)) = s(plus(?x:Nat,?x:Nat))",
"plus(s(s(s(?z:Nat))),min(?z:Nat,min(plus(?z:Nat,?z:Nat),min(?z:Nat,min(?x:Nat,?y:Nat))))) = plus(s(?z:Nat),?x:Nat)",
"plus(max(?x:Nat,0),max(plus(?y:Nat,min(min(?x:Nat,?x:Nat),?z:Nat)),plus(0,s(0)))) = max(0,s(?z:Nat))",
"s(min(plus(min(?y:Nat,s(?z:Nat)),s(s(?z:Nat))),s(max(s(?x:Nat),plus(?z:Nat,?z:Nat))))) = s(s(?x:Nat))",
"s(plus(s(s(plus(?x:Nat,0))),plus(s(min(?z:Nat,?y:Nat)),min(max(?z:Nat,0),?x:Nat)))) = plus(?x:Nat,0)",
"max(?y:Nat,min(plus(max(?y:Nat,max(?x:Nat,?y:Nat)),?x:Nat),plus(max(?z:Nat,max(?y:Nat,0)),?y:Nat))) = s(s(?x:Nat))",
"min(?x:Nat,min(s(plus(?y:Nat,min(0,?y:Nat))),max(s(?x:Nat),min(max(0,?x:Nat),?x:Nat)))) = s(s(?x:Nat))",
"min(s(plus(plus(s(?x:Nat),s(0)),0)),max(?z:Nat,s(max(?x:Nat,max(?x:Nat,?z:Nat))))) = min(?x:Nat,?x:Nat)",
"min(min(min(?x:Nat,s(s(?x:Nat))),s(s(plus(0,?x:Nat)))),s(min(s(?y:Nat),?x:Nat))) = plus(?y:Nat,?y:Nat)",
"min(min(plus(0,min(s(?z:Nat),plus(?x:Nat,?x:Nat))),max(?x:Nat,s(s(?x:Nat)))),s(?z:Nat)) = max(?z:Nat,?x:Nat)",
"min(max(min(plus(plus(?x:Nat,?x:Nat),min(?z:Nat,0)),plus(max(?x:Nat,0),s(?x:Nat))),0),0) = s(?y:Nat)",
"plus(?z:Nat,plus(?x:Nat,plus(plus(s(?x:Nat),max(?y:Nat,?x:Nat)),min(max(0,?x:Nat),max(?x:Nat,?z:Nat))))) = s(?x:Nat)",
"plus(s(plus(s(?z:Nat),0)),plus(min(s(min(?x:Nat,?x:Nat)),min(plus(?y:Nat,?x:Nat),?y:Nat)),?z:Nat)) = s(?z:Nat)",
"plus(s(plus(min(max(?x:Nat,?y:Nat),plus(0,?x:Nat)),plus(max(0,?y:Nat),min(?y:Nat,0)))),?z:Nat) = s(?y:Nat)",
"s(?x:Nat) = min(?x:Nat,plus(plus(s(?y:Nat),min(s(?x:Nat),max(?y:Nat,?x:Nat))),s(min(s(?z:Nat),min(?y:Nat,?x:Nat)))))",
"s(?x:Nat) = plus(min(?x:Nat,min(?x:Nat,max(s(?x:Nat),s(?z:Nat)))),plus(plus(?y:Nat,plus(min(?y:Nat,?y:Nat),?x:Nat)),?x:Nat))",
"min(0,?z:Nat) = min(plus(?x:Nat,plus(plus(min(0,?z:Nat),min(?x:Nat,0)),s(?z:Nat))),plus(s(0),s(?z:Nat)))",
"plus(?x:Nat,?z:Nat) = plus(s(s(s(?x:Nat))),max(plus(plus(plus(?z:Nat,?z:Nat),min(?x:Nat,?y:Nat)),s(s(?z:Nat))),?y:Nat))",
"plus(?y:Nat,0) = min(min(0,s(max(plus(?x:Nat,?x:Nat),?x:Nat))),plus(0,max(plus(s(0),s(?x:Nat)),?x:Nat)))",
"plus(?z:Nat,?x:Nat) = max(plus(plus(plus(plus(?x:Nat,?x:Nat),min(0,?x:Nat)),?x:Nat),s(plus(s(0),s(?x:Nat)))),?x:Nat)",
"max(?x:Nat,plus(?y:Nat,?x:Nat)) = s(min(plus(plus(?z:Nat,0),min(plus(?x:Nat,?x:Nat),max(?z:Nat,?z:Nat))),plus(?y:Nat,?z:Nat)))",
"max(0,plus(?x:Nat,0)) = min(min(min(min(s(0),?z:Nat),s(s(?z:Nat))),plus(min(?z:Nat,0),0)),?y:Nat)",
"max(min(?z:Nat,0),?x:Nat) = plus(s(s(?x:Nat)),min(?x:Nat,max(plus(max(?y:Nat,?x:Nat),0),max(s(?z:Nat),?y:Nat))))",
"plus(?z:Nat,s(s(?y:Nat))) = plus(max(min(max(max(?x:Nat,?y:Nat),s(0)),?y:Nat),s(s(max(?x:Nat,?y:Nat)))),0)",
"plus(s(?x:Nat),plus(?y:Nat,?x:Nat)) = s(min(max(?y:Nat,max(0,max(?z:Nat,0))),plus(s(?y:Nat),plus(?z:Nat,?z:Nat))))",
"s(plus(?z:Nat,min(?z:Nat,s(?x:Nat)))) = plus(plus(?y:Nat,?x:Nat),max(?y:Nat,min(plus(plus(?x:Nat,?x:Nat),s(?x:Nat)),0)))",
"plus(?x:Nat,s(s(plus(?z:Nat,?x:Nat)))) = s(max(?x:Nat,max(s(plus(?y:Nat,0)),plus(s(?x:Nat),plus(?z:Nat,?y:Nat)))))",
"s(plus(s(s(plus(?x:Nat,?x:Nat))),0)) = max(min(?y:Nat,plus(max(max(?x:Nat,0),s(?x:Nat)),s(?x:Nat))),?y:Nat)",
"max(?y:Nat,max(?x:Nat,max(s(0),?x:Nat))) = plus(?z:Nat,plus(?y:Nat,plus(?y:Nat,plus(min(?x:Nat,?y:Nat),max(?z:Nat,?z:Nat)))))",
"plus(plus(min(0,s(?z:Nat)),?y:Nat),?y:Nat) = plus(plus(s(s(0)),?x:Nat),s(s(min(?z:Nat,max(?z:Nat,?z:Nat)))))",
"s(plus(0,s(max(?z:Nat,plus(?y:Nat,?y:Nat))))) = s(plus(s(s(min(0,?y:Nat))),min(s(s(?x:Nat)),?z:Nat)))",
"max(plus(?y:Nat,max(min(?y:Nat,?z:Nat),?y:Nat)),0) = plus(max(s(?x:Nat),s(plus(s(?x:Nat),s(?x:Nat)))),s(?y:Nat))",
"min(?x:Nat,s(s(plus(0,plus(?x:Nat,0))))) = plus(plus(?y:Nat,s(max(?y:Nat,?y:Nat))),s(min(s(?x:Nat),?x:Nat)))",
"plus(?x:Nat,s(s(min(?y:Nat,plus(0,?x:Nat))))) = s(min(max(?x:Nat,s(0)),min(?x:Nat,s(min(0,?x:Nat)))))",
"plus(?x:Nat,plus(?x:Nat,s(s(plus(?x:Nat,?x:Nat))))) = plus(plus(min(s(?x:Nat),min(?y:Nat,min(0,?y:Nat))),?x:Nat),0)",
"plus(?z:Nat,s(s(plus(?z:Nat,plus(?y:Nat,?x:Nat))))) = s(plus(min(?x:Nat,min(min(?y:Nat,0),plus(?z:Nat,0))),?x:Nat))",
"max(0,max(max(?y:Nat,s(0)),s(s(?x:Nat)))) = min(s(plus(min(s(?x:Nat),max(0,?z:Nat)),?z:Nat)),?x:Nat)",
"max(min(plus(s(?z:Nat),?x:Nat),min(?y:Nat,?x:Nat)),?x:Nat) = plus(?y:Nat,max(?x:Nat,plus(0,max(min(?x:Nat,?z:Nat),0))))",
"max(plus(?y:Nat,s(plus(?x:Nat,plus(0,0)))),?z:Nat) = plus(?x:Nat,max(min(s(?x:Nat),max(?y:Nat,s(?x:Nat))),?z:Nat))",
"s(max(0,s(plus(min(?x:Nat,?z:Nat),min(?z:Nat,0))))) = min(s(s(min(plus(?x:Nat,?x:Nat),s(?x:Nat)))),?x:Nat)",
"s(min(?z:Nat,plus(s(max(?x:Nat,?z:Nat)),s(s(0))))) = plus(plus(min(plus(s(?y:Nat),?x:Nat),?z:Nat),?x:Nat),?x:Nat)",
"plus(s(max(?x:Nat,s(plus(?y:Nat,?x:Nat)))),min(0,?y:Nat)) = s(plus(plus(?y:Nat,min(0,min(?x:Nat,0))),0))",
"plus(s(max(min(min(?z:Nat,?x:Nat),s(?x:Nat)),?x:Nat)),?y:Nat) = plus(min(min(?x:Nat,plus(s(?x:Nat),0)),?x:Nat),?z:Nat)",
"plus(max(max(?y:Nat,plus(?x:Nat,plus(?x:Nat,?z:Nat))),s(?x:Nat)),?z:Nat) = min(0,plus(min(plus(?x:Nat,?z:Nat),0),?z:Nat))",
"plus(plus(?z:Nat,s(min(max(?x:Nat,?z:Nat),min(?y:Nat,?x:Nat)))),?x:Nat) = s(plus(?x:Nat,s(max(s(?z:Nat),s(0)))))",
"plus(plus(s(plus(max(0,?x:Nat),min(?z:Nat,?z:Nat))),?y:Nat),?x:Nat) = min(s(max(s(plus(?z:Nat,?x:Nat)),?y:Nat)),?z:Nat)",
"s(min(min(min(plus(?x:Nat,?x:Nat),min(?x:Nat,0)),s(?x:Nat)),0)) = s(max(0,s(min(s(0),0))))",
"min(0,max(plus(s(?x:Nat),plus(?y:Nat,plus(?x:Nat,0))),s(?z:Nat))) = min(?y:Nat,s(s(s(min(0,0)))))",
"s(max(plus(plus(max(?x:Nat,?z:Nat),?x:Nat),?z:Nat),s(max(?y:Nat,min(?y:Nat,?x:Nat))))) = min(max(0,s(0)),?x:Nat)",
"s(plus(max(plus(s(?z:Nat),s(?z:Nat)),?z:Nat),s(plus(min(?z:Nat,?x:Nat),?x:Nat)))) = max(min(0,?x:Nat),s(?x:Nat))",
"max(max(s(s(max(?y:Nat,?y:Nat))),s(min(s(0),max(?y:Nat,?x:Nat)))),?x:Nat) = plus(s(?x:Nat),plus(?z:Nat,?x:Nat))",
"s(s(min(max(min(?y:Nat,?x:Nat),max(?z:Nat,?x:Nat)),max(max(?x:Nat,?y:Nat),s(?y:Nat))))) = plus(?x:Nat,plus(?x:Nat,0))",
"s(max(plus(plus(plus(?x:Nat,?y:Nat),0),plus(?x:Nat,min(?x:Nat,?z:Nat))),max(?x:Nat,?x:Nat))) = s(plus(?x:Nat,s(?z:Nat)))",
"plus(min(min(plus(?x:Nat,plus(?x:Nat,?y:Nat)),?x:Nat),s(plus(?x:Nat,min(?x:Nat,?x:Nat)))),?y:Nat) = s(s(max(?z:Nat,?x:Nat)))",
"s(max(min(?x:Nat,min(plus(?x:Nat,?x:Nat),min(0,?x:Nat))),max(s(0),s(max(?x:Nat,?y:Nat))))) = min(?x:Nat,0)",
"max(?x:Nat,max(s(max(min(?x:Nat,?z:Nat),s(?z:Nat))),plus(min(plus(?y:Nat,?x:Nat),0),s(?y:Nat)))) = min(?x:Nat,0)",
"max(?x:Nat,plus(max(plus(min(?x:Nat,?y:Nat),s(?x:Nat)),max(plus(0,?x:Nat),plus(?y:Nat,0))),?y:Nat)) = max(0,0)",
"max(min(plus(min(plus(?x:Nat,?z:Nat),?y:Nat),s(s(0))),s(max(plus(?y:Nat,?x:Nat),?x:Nat))),?y:Nat) = s(s(?y:Nat))",
"min(plus(s(?x:Nat),?x:Nat),min(min(max(s(0),s(?y:Nat)),min(plus(?x:Nat,0),?x:Nat)),?x:Nat)) = min(?z:Nat,?y:Nat)",
"plus(min(s(max(min(0,?x:Nat),?z:Nat)),plus(s(s(?y:Nat)),plus(plus(?z:Nat,0),?y:Nat))),?z:Nat) = s(s(0))",
"max(min(min(?z:Nat,s(plus(0,?y:Nat))),s(s(s(0)))),plus(?y:Nat,plus(s(s(0)),0))) = s(?x:Nat)",
"s(?x:Nat) = plus(?x:Nat,plus(plus(plus(s(?x:Nat),s(?y:Nat)),max(?y:Nat,?z:Nat)),s(min(min(?x:Nat,?y:Nat),min(?x:Nat,?z:Nat)))))",
"s(?y:Nat) = max(plus(plus(?z:Nat,plus(?x:Nat,s(?x:Nat))),s(plus(min(?x:Nat,?x:Nat),max(0,?x:Nat)))),plus(0,s(?z:Nat)))",
"s(0) = max(max(min(?x:Nat,?x:Nat),?y:Nat),min(max(?x:Nat,s(max(?z:Nat,0))),min(?x:Nat,plus(0,plus(?z:Nat,?y:Nat)))))",
"s(s(?x:Nat)) = plus(plus(plus(max(s(?x:Nat),s(?z:Nat)),min(min(?x:Nat,?x:Nat),min(?y:Nat,?y:Nat))),0),plus(?x:Nat,?z:Nat))",
"s(s(?z:Nat)) = max(plus(max(?z:Nat,max(max(?x:Nat,?z:Nat),plus(0,?x:Nat))),?x:Nat),max(?x:Nat,min(0,plus(0,?z:Nat))))",
"max(?x:Nat,?x:Nat) = plus(?y:Nat,plus(plus(min(?y:Nat,?x:Nat),min(plus(?x:Nat,0),s(?z:Nat))),s(min(max(0,?z:Nat),?x:Nat))))",
"min(?x:Nat,?x:Nat) = min(?z:Nat,plus(max(min(?z:Nat,?y:Nat),0),plus(plus(plus(0,?z:Nat),min(?x:Nat,?z:Nat)),max(?y:Nat,0))))",
"min(?z:Nat,0) = plus(min(min(0,plus(0,min(?x:Nat,?z:Nat))),plus(s(plus(?y:Nat,?y:Nat)),s(max(?x:Nat,?y:Nat)))),?y:Nat)",
"plus(?x:Nat,?x:Nat) = min(min(s(plus(min(0,0),s(?z:Nat))),plus(plus(?x:Nat,plus(?z:Nat,?x:Nat)),0)),plus(?x:Nat,?z:Nat))",
"plus(?x:Nat,?z:Nat) = plus(min(max(s(?x:Nat),?z:Nat),0),s(plus(plus(?x:Nat,?z:Nat),plus(max(0,?x:Nat),plus(?y:Nat,?x:Nat)))))",
"plus(0,?z:Nat) = plus(min(plus(?x:Nat,plus(plus(?x:Nat,0),min(?x:Nat,0))),?x:Nat),max(s(s(plus(?y:Nat,?x:Nat))),?x:Nat))",
"s(min(?y:Nat,?x:Nat)) = plus(s(max(s(?x:Nat),max(?z:Nat,max(?x:Nat,?x:Nat)))),plus(min(s(plus(?z:Nat,?z:Nat)),?x:Nat),?x:Nat))",
"min(s(?x:Nat),?x:Nat) = max(plus(plus(0,min(min(0,?x:Nat),s(?x:Nat))),max(s(plus(?x:Nat,?y:Nat)),?x:Nat)),s(?y:Nat))",
"min(s(?y:Nat),?z:Nat) = s(plus(max(s(plus(0,0)),plus(s(0),plus(?x:Nat,?z:Nat))),s(s(min(?x:Nat,?y:Nat)))))",
"plus(?z:Nat,s(0)) = max(?x:Nat,max(min(0,max(s(?z:Nat),?y:Nat)),max(?x:Nat,max(max(?z:Nat,0),plus(?y:Nat,?z:Nat)))))",
"max(?x:Nat,plus(?y:Nat,s(?y:Nat))) = plus(plus(max(?x:Nat,s(min(?z:Nat,?x:Nat))),?x:Nat),plus(plus(0,?z:Nat),s(s(?z:Nat))))",
"s(max(max(?x:Nat,?y:Nat),s(0))) = s(max(?x:Nat,plus(min(0,max(?x:Nat,?z:Nat)),min(s(0),max(?y:Nat,?y:Nat)))))",
"s(s(max(?y:Nat,s(max(?x:Nat,?y:Nat))))) = plus(max(plus(min(?x:Nat,min(?x:Nat,?x:Nat)),min(?z:Nat,s(?y:Nat))),0),?z:Nat)",
"s(plus(max(?y:Nat,s(?x:Nat)),s(?x:Nat))) = plus(max(max(min(max(?y:Nat,?z:Nat),max(?x:Nat,0)),s(?x:Nat)),?z:Nat),?y:Nat)",
"max(?y:Nat,max(?x:Nat,plus(?y:Nat,s(?x:Nat)))) = plus(s(?z:Nat),min(plus(plus(plus(0,0),?y:Nat),max(?z:Nat,0)),0))",
"min(max(?x:Nat,s(plus(0,?z:Nat))),?z:Nat) = plus(max(plus(s(plus(?x:Nat,?y:Nat)),max(?x:Nat,max(?y:Nat,?x:Nat))),?z:Nat),?y:Nat)",
"plus(?x:Nat,s(plus(plus(?z:Nat,0),0))) = min(?z:Nat,plus(s(max(max(0,?z:Nat),?x:Nat)),max(plus(?y:Nat,0),0)))",
"plus(s(?x:Nat),max(plus(?z:Nat,?z:Nat),?y:Nat)) = min(max(?y:Nat,s(?x:Nat)),plus(?x:Nat,s(plus(s(0),plus(?x:Nat,?x:Nat)))))",
"s(max(s(max(?y:Nat,?x:Nat)),max(?x:Nat,?x:Nat))) = s(s(min(plus(min(?y:Nat,?x:Nat),min(?x:Nat,?y:Nat)),plus(?y:Nat,?y:Nat))))",
"plus(?y:Nat,min(?x:Nat,min(?x:Nat,s(s(?x:Nat))))) = max(s(min(min(0,min(?x:Nat,?x:Nat)),s(?x:Nat))),plus(?x:Nat,?x:Nat))",
"plus(0,s(s(plus(min(?x:Nat,0),?x:Nat)))) = plus(?y:Nat,plus(s(?x:Nat),max(min(?z:Nat,min(0,?x:Nat)),s(?z:Nat))))",
"s(min(s(?y:Nat),max(s(plus(?x:Nat,?x:Nat)),?x:Nat))) = min(s(s(max(min(?y:Nat,?z:Nat),max(?x:Nat,?y:Nat)))),s(?x:Nat))",
"s(plus(0,min(plus(s(?x:Nat),s(?x:Nat)),?z:Nat))) = s(s(plus(plus(?z:Nat,max(?z:Nat,?y:Nat)),s(plus(?y:Nat,0)))))",
"max(?z:Nat,plus(min(?x:Nat,plus(?x:Nat,?y:Nat)),s(0))) = s(s(max(max(s(0),?y:Nat),min(min(?x:Nat,?x:Nat),0))))",
"min(?y:Nat,s(max(plus(?y:Nat,plus(?x:Nat,?x:Nat)),?y:Nat))) = s(max(max(s(min(?x:Nat,0)),s(plus(0,?z:Nat))),?x:Nat))",
"plus(?y:Nat,max(max(s(min(?x:Nat,?y:Nat)),?x:Nat),?y:Nat)) = s(s(plus(s(?z:Nat),min(min(0,?x:Nat),plus(?x:Nat,0)))))",
"plus(plus(?y:Nat,max(?x:Nat,max(?y:Nat,s(?y:Nat)))),0) = max(s(plus(min(?x:Nat,s(?y:Nat)),s(plus(?x:Nat,0)))),?z:Nat)",
"min(?x:Nat,max(?z:Nat,s(plus(min(?y:Nat,0),s(?x:Nat))))) = max(?x:Nat,plus(?z:Nat,max(?x:Nat,min(s(?x:Nat),s(?x:Nat)))))",
"min(?z:Nat,s(plus(plus(?x:Nat,?x:Nat),s(plus(?x:Nat,?x:Nat))))) = min(s(max(plus(?x:Nat,?x:Nat),s(?y:Nat))),max(?x:Nat,?z:Nat))",
"plus(s(?y:Nat),max(s(0),plus(s(?y:Nat),s(?x:Nat)))) = plus(plus(?y:Nat,plus(min(?y:Nat,min(0,?y:Nat)),?z:Nat)),?z:Nat)",
"plus(plus(?x:Nat,s(min(?x:Nat,?y:Nat))),plus(s(?z:Nat),?z:Nat)) = s(s(max(?x:Nat,min(plus(0,?z:Nat),plus(?x:Nat,?x:Nat)))))",
"plus(plus(s(0),s(plus(s(?x:Nat),s(?y:Nat)))),?x:Nat) = s(plus(plus(plus(max(?x:Nat,?y:Nat),s(?x:Nat)),?z:Nat),?z:Nat))",
"min(?x:Nat,plus(max(s(s(?z:Nat)),s(min(?x:Nat,?y:Nat))),0)) = s(s(max(plus(0,max(?y:Nat,?x:Nat)),s(0))))",
"plus(0,max(max(?z:Nat,?x:Nat),s(max(plus(?x:Nat,?z:Nat),?y:Nat)))) = min(plus(plus(?z:Nat,?y:Nat),0),s(min(?z:Nat,?x:Nat)))",
"plus(min(min(min(?x:Nat,0),?x:Nat),s(?z:Nat)),s(s(0))) = s(max(s(min(max(?x:Nat,?y:Nat),s(?z:Nat))),?y:Nat))",
"s(plus(max(s(max(?x:Nat,0)),plus(plus(0,?z:Nat),?z:Nat)),?y:Nat)) = plus(?y:Nat,s(min(?x:Nat,max(s(?y:Nat),?z:Nat))))",
"max(plus(plus(max(plus(?z:Nat,?y:Nat),min(?z:Nat,0)),?y:Nat),?x:Nat),?z:Nat) = plus(?y:Nat,max(max(?z:Nat,?z:Nat),plus(?x:Nat,?z:Nat)))",
"plus(s(s(?z:Nat)),s(min(s(max(0,?y:Nat)),s(s(?x:Nat))))) = plus(?x:Nat,max(plus(?y:Nat,?x:Nat),plus(?x:Nat,?x:Nat)))",
"plus(s(plus(0,max(max(?x:Nat,?x:Nat),plus(?y:Nat,?x:Nat)))),s(?y:Nat)) = plus(plus(?z:Nat,0),min(plus(0,?y:Nat),?z:Nat))",
"plus(max(max(s(?z:Nat),?y:Nat),s(min(max(0,?x:Nat),?z:Nat))),?x:Nat) = s(plus(?x:Nat,s(max(?z:Nat,max(?x:Nat,?y:Nat)))))",
"s(min(plus(min(s(?z:Nat),plus(0,0)),s(s(0))),s(?z:Nat))) = s(max(plus(plus(?x:Nat,0),?x:Nat),0))",
"max(s(plus(min(min(?z:Nat,?y:Nat),plus(?x:Nat,?x:Nat)),?x:Nat)),s(s(0))) = s(s(plus(s(min(?z:Nat,?x:Nat)),?y:Nat)))",
"min(max(max(?y:Nat,min(min(?y:Nat,?x:Nat),min(?x:Nat,?y:Nat))),s(?x:Nat)),0) = s(max(?x:Nat,plus(?y:Nat,max(?x:Nat,?z:Nat))))",
"plus(min(?y:Nat,max(plus(?x:Nat,s(0)),?z:Nat)),max(plus(0,?x:Nat),0)) = max(s(max(?x:Nat,?x:Nat)),max(?x:Nat,?z:Nat))",
"max(?y:Nat,s(plus(min(0,s(0)),min(plus(?x:Nat,?z:Nat),max(?x:Nat,?z:Nat))))) = plus(plus(min(?x:Nat,?y:Nat),?z:Nat),?y:Nat)",
"plus(s(s(max(?z:Nat,plus(?x:Nat,0)))),s(s(max(?y:Nat,plus(?z:Nat,?y:Nat))))) = plus(min(0,0),max(?y:Nat,?y:Nat))",
"plus(min(?z:Nat,min(?x:Nat,?x:Nat)),min(?z:Nat,s(plus(min(?x:Nat,0),s(0))))) = plus(?x:Nat,min(?y:Nat,min(?x:Nat,?x:Nat)))",
"s(plus(max(?x:Nat,plus(max(?x:Nat,?x:Nat),?y:Nat)),s(plus(min(?x:Nat,?z:Nat),plus(?z:Nat,?z:Nat))))) = max(?y:Nat,plus(0,?x:Nat))",
"max(min(0,?x:Nat),plus(?x:Nat,max(min(?z:Nat,?x:Nat),max(min(?x:Nat,?x:Nat),plus(?y:Nat,0))))) = max(0,s(s(0)))",
"plus(?y:Nat,min(s(plus(plus(?x:Nat,?x:Nat),?z:Nat)),max(?x:Nat,min(s(?x:Nat),max(?x:Nat,?x:Nat))))) = min(min(0,?x:Nat),?z:Nat)",
"max(min(plus(plus(plus(?x:Nat,?z:Nat),s(?x:Nat)),min(min(?x:Nat,?x:Nat),min(?z:Nat,?x:Nat))),?y:Nat),?y:Nat) = s(s(s(0)))",
"min(s(min(plus(plus(0,?y:Nat),plus(?x:Nat,?y:Nat)),plus(max(?x:Nat,0),min(0,0)))),?y:Nat) = max(0,s(0))",
"plus(plus(min(plus(plus(?y:Nat,?x:Nat),?z:Nat),min(0,?z:Nat)),?y:Nat),s(plus(s(s(?y:Nat)),?z:Nat))) = plus(?y:Nat,s(?z:Nat))",
"s(min(plus(min(plus(?x:Nat,?y:Nat),?x:Nat),max(s(?z:Nat),0)),s(min(s(?x:Nat),min(?y:Nat,?x:Nat))))) = plus(?x:Nat,?x:Nat)",
"max(?x:Nat,min(max(min(s(0),plus(?x:Nat,?z:Nat)),?x:Nat),min(?x:Nat,max(max(?x:Nat,0),s(?x:Nat))))) = plus(?x:Nat,?z:Nat)",
"max(?x:Nat,plus(s(plus(plus(?z:Nat,?y:Nat),?y:Nat)),max(s(min(?z:Nat,?y:Nat)),plus(plus(0,?x:Nat),?z:Nat)))) = plus(?z:Nat,?z:Nat)",
"max(min(min(min(s(?x:Nat),?z:Nat),?x:Nat),plus(?x:Nat,max(plus(?y:Nat,?x:Nat),plus(?z:Nat,?z:Nat)))),s(?z:Nat)) = min(0,?y:Nat)",
"min(s(?x:Nat),min(min(?x:Nat,min(s(?x:Nat),plus(?z:Nat,?y:Nat))),plus(min(?x:Nat,?x:Nat),s(s(0))))) = min(?x:Nat,?z:Nat)",
"min(plus(max(plus(plus(?x:Nat,?x:Nat),min(?z:Nat,?y:Nat)),?x:Nat),plus(plus(s(?x:Nat),s(0)),?y:Nat)),?x:Nat) = min(?x:Nat,?y:Nat)",
"plus(s(min(min(min(?x:Nat,?y:Nat),?y:Nat),s(max(?y:Nat,?y:Nat)))),s(min(s(?x:Nat),s(s(?x:Nat))))) = min(?y:Nat,?x:Nat)",
"min(s(plus(max(plus(?y:Nat,0),plus(?y:Nat,?y:Nat)),?y:Nat)),s(s(max(max(?z:Nat,?x:Nat),max(?x:Nat,?z:Nat))))) = s(?x:Nat)",
"plus(max(plus(plus(?x:Nat,plus(?z:Nat,?x:Nat)),s(s(?x:Nat))),plus(plus(?y:Nat,max(0,?z:Nat)),0)),s(?x:Nat)) = s(?z:Nat)",
"s(?y:Nat) = max(max(0,s(?z:Nat)),plus(min(0,plus(min(?z:Nat,?y:Nat),plus(?y:Nat,?x:Nat))),s(plus(max(?y:Nat,?x:Nat),?z:Nat))))",
"s(?z:Nat) = max(min(min(s(?x:Nat),max(?z:Nat,plus(?y:Nat,?x:Nat))),?y:Nat),plus(?x:Nat,min(s(min(?x:Nat,?y:Nat)),max(?z:Nat,0))))",
"s(?z:Nat) = plus(plus(plus(min(s(?x:Nat),?z:Nat),max(?y:Nat,plus(?x:Nat,?x:Nat))),s(plus(max(0,?z:Nat),max(?y:Nat,?x:Nat)))),0)",
"max(?x:Nat,0) = min(s(plus(s(plus(?x:Nat,0)),?y:Nat)),s(max(s(plus(?z:Nat,0)),max(max(?z:Nat,?z:Nat),s(?z:Nat)))))",
"max(?y:Nat,?y:Nat) = min(s(plus(?x:Nat,plus(s(0),0))),s(plus(plus(plus(?y:Nat,?x:Nat),?x:Nat),max(?y:Nat,max(?y:Nat,?z:Nat)))))",
"min(?x:Nat,?x:Nat) = max(max(min(plus(0,s(?y:Nat)),max(?x:Nat,?y:Nat)),min(min(plus(?x:Nat,?y:Nat),min(?x:Nat,?z:Nat)),?x:Nat)),?x:Nat)",
"plus(?x:Nat,?x:Nat) = min(min(plus(plus(s(0),plus(?x:Nat,?x:Nat)),0),?x:Nat),max(s(s(s(?x:Nat))),max(s(?x:Nat),?y:Nat)))",
"plus(?x:Nat,?z:Nat) = max(plus(plus(min(s(?x:Nat),plus(?x:Nat,0)),s(?x:Nat)),plus(?z:Nat,s(s(?x:Nat)))),plus(0,s(0)))",
"plus(?z:Nat,?x:Nat) = plus(max(min(min(?x:Nat,max(?x:Nat,?y:Nat)),min(?x:Nat,min(?x:Nat,?y:Nat))),max(0,plus(?x:Nat,?x:Nat))),s(?z:Nat))",
"max(?z:Nat,s(?x:Nat)) = plus(min(min(s(plus(?x:Nat,?z:Nat)),max(s(?y:Nat),min(?y:Nat,?y:Nat))),s(plus(?x:Nat,s(?x:Nat)))),?x:Nat)",
"plus(0,s(?x:Nat)) = max(plus(s(plus(?y:Nat,s(?x:Nat))),s(0)),plus(plus(?x:Nat,?x:Nat),min(?z:Nat,s(plus(?z:Nat,?z:Nat)))))",
"plus(s(?y:Nat),?x:Nat) = plus(plus(plus(0,0),?z:Nat),min(plus(s(min(0,?z:Nat)),?x:Nat),plus(?z:Nat,plus(?x:Nat,s(?x:Nat)))))",
"plus(?y:Nat,min(?x:Nat,?y:Nat)) = s(max(plus(plus(?x:Nat,min(?x:Nat,?x:Nat)),plus(?x:Nat,s(?y:Nat))),max(?x:Nat,plus(s(0),?z:Nat))))",
"plus(?y:Nat,min(?x:Nat,?z:Nat)) = max(0,plus(max(s(?y:Nat),max(min(?x:Nat,?z:Nat),s(0))),max(?z:Nat,s(plus(?z:Nat,?x:Nat)))))",
"plus(min(0,?x:Nat),0) = plus(min(s(plus(s(?z:Nat),?x:Nat)),min(0,max(plus(?z:Nat,0),min(?x:Nat,?y:Nat)))),s(0))",
"s(min(?x:Nat,max(?z:Nat,?x:Nat))) = max(?z:Nat,plus(s(max(min(?x:Nat,?y:Nat),min(?x:Nat,?y:Nat))),s(plus(min(?z:Nat,?z:Nat),?x:Nat))))",
"plus(?z:Nat,plus(?y:Nat,s(?x:Nat))) = plus(s(max(s(s(?z:Nat)),?x:Nat)),min(plus(s(plus(?z:Nat,?z:Nat)),s(s(0))),?z:Nat))",
"min(s(?x:Nat),s(plus(0,0))) = min(plus(?x:Nat,?x:Nat),plus(s(plus(?z:Nat,?x:Nat)),plus(s(?y:Nat),plus(?y:Nat,s(?y:Nat)))))",
"min(plus(plus(?z:Nat,0),0),0) = s(min(s(max(?y:Nat,plus(?y:Nat,?z:Nat))),plus(s(?y:Nat),min(max(?z:Nat,?x:Nat),?x:Nat))))",
"plus(?x:Nat,max(?z:Nat,max(?x:Nat,0))) = min(s(s(?x:Nat)),plus(plus(max(?z:Nat,min(?z:Nat,0)),max(s(?x:Nat),0)),?y:Nat))",
"plus(s(?x:Nat),plus(s(?x:Nat),?z:Nat)) = plus(s(min(min(plus(?z:Nat,?y:Nat),min(?z:Nat,?x:Nat)),plus(plus(?z:Nat,?y:Nat),0))),?z:Nat)",
"plus(s(max(?z:Nat,?x:Nat)),s(?y:Nat)) = min(s(?x:Nat),max(max(s(min(?x:Nat,0)),max(max(?x:Nat,0),0)),s(0)))",
"s(max(?x:Nat,max(s(s(?x:Nat)),?y:Nat))) = s(plus(0,min(min(plus(0,?z:Nat),max(?z:Nat,?y:Nat)),plus(?z:Nat,s(?y:Nat)))))",
"plus(min(max(?x:Nat,s(?x:Nat)),?z:Nat),?x:Nat) = max(?y:Nat,plus(s(min(s(?y:Nat),0)),max(?z:Nat,plus(min(?x:Nat,?x:Nat),?y:Nat))))",
"s(plus(?z:Nat,plus(s(max(?y:Nat,?z:Nat)),?z:Nat))) = s(min(s(?y:Nat),plus(plus(s(?z:Nat),max(?z:Nat,0)),min(0,?x:Nat))))",
"min(max(s(plus(?y:Nat,s(?x:Nat))),?z:Nat),?x:Nat) = plus(plus(min(min(plus(?x:Nat,?x:Nat),s(0)),s(?z:Nat)),?x:Nat),s(0))",
"plus(0,max(?z:Nat,plus(?x:Nat,s(s(?y:Nat))))) = min(max(max(?z:Nat,min(?x:Nat,?x:Nat)),s(plus(0,min(?y:Nat,?z:Nat)))),?x:Nat)",
"min(?y:Nat,max(?x:Nat,s(plus(?x:Nat,max(?x:Nat,?y:Nat))))) = plus(min(s(plus(min(0,?x:Nat),s(?x:Nat))),?z:Nat),max(?x:Nat,?y:Nat))",
"plus(?x:Nat,max(s(max(?y:Nat,plus(?x:Nat,0))),?x:Nat)) = plus(plus(plus(min(?x:Nat,?y:Nat),0),s(?x:Nat)),plus(s(?z:Nat),?x:Nat))",
"plus(?z:Nat,s(max(?y:Nat,plus(plus(?x:Nat,0),?x:Nat)))) = max(?x:Nat,max(?x:Nat,max(max(min(?y:Nat,?y:Nat),min(?x:Nat,?x:Nat)),?x:Nat)))",
"plus(min(?y:Nat,?x:Nat),s(s(s(max(?z:Nat,?z:Nat))))) = max(?x:Nat,s(max(min(s(?x:Nat),?y:Nat),max(?x:Nat,plus(?z:Nat,?x:Nat)))))",
"s(s(max(s(s(0)),plus(?z:Nat,max(?x:Nat,?z:Nat))))) = max(?x:Nat,plus(s(max(max(0,?x:Nat),max(0,?y:Nat))),0))",
"min(min(?y:Nat,s(max(plus(0,0),?y:Nat))),s(?x:Nat)) = max(s(plus(?y:Nat,plus(min(0,?y:Nat),0))),plus(?x:Nat,?y:Nat))",
"plus(s(?y:Nat),s(plus(s(max(?x:Nat,?x:Nat)),s(?z:Nat)))) = plus(?x:Nat,max(plus(s(?x:Nat),plus(?x:Nat,plus(?x:Nat,?z:Nat))),?z:Nat))",
"plus(s(s(plus(plus(?x:Nat,?x:Nat),max(0,?x:Nat)))),?y:Nat) = max(plus(?z:Nat,s(?x:Nat)),s(min(?x:Nat,s(plus(0,0)))))",
"s(max(min(plus(s(?x:Nat),max(0,0)),s(?y:Nat)),?x:Nat)) = s(max(?x:Nat,s(max(plus(?z:Nat,0),min(?z:Nat,?z:Nat)))))",
"s(min(plus(?x:Nat,plus(max(?z:Nat,?x:Nat),min(?x:Nat,?x:Nat))),?z:Nat)) = plus(min(?x:Nat,min(s(?x:Nat),s(max(?y:Nat,0)))),?z:Nat)",
"max(max(max(plus(?y:Nat,s(0)),?y:Nat),?y:Nat),s(s(?x:Nat))) = min(s(min(plus(min(?y:Nat,?y:Nat),s(?x:Nat)),?y:Nat)),?x:Nat)",
"min(s(?x:Nat),plus(?x:Nat,plus(plus(0,plus(?z:Nat,?x:Nat)),?x:Nat))) = max(s(min(?y:Nat,min(max(0,?z:Nat),s(?y:Nat)))),?y:Nat)",
"plus(plus(plus(?z:Nat,s(s(?z:Nat))),?z:Nat),max(s(?x:Nat),s(0))) = plus(?x:Nat,plus(?x:Nat,plus(0,s(plus(?z:Nat,?y:Nat)))))",
"s(min(?z:Nat,min(max(plus(?z:Nat,?x:Nat),?y:Nat),plus(plus(0,?z:Nat),?y:Nat)))) = s(s(max(?x:Nat,min(?y:Nat,min(0,?x:Nat)))))",
"min(?y:Nat,max(0,max(plus(max(0,0),0),s(min(?z:Nat,?y:Nat))))) = max(min(?y:Nat,s(plus(s(0),?x:Nat))),?z:Nat)",
"s(min(max(?x:Nat,plus(s(?y:Nat),?x:Nat)),max(?y:Nat,plus(plus(0,?y:Nat),?y:Nat)))) = plus(min(s(s(s(0))),?x:Nat),?x:Nat)",
"s(plus(s(plus(s(0),?x:Nat)),plus(0,plus(plus(?x:Nat,?x:Nat),s(?x:Nat))))) = s(min(s(?y:Nat),s(s(s(?y:Nat)))))",
"max(min(max(0,plus(min(0,?z:Nat),plus(0,0))),max(?x:Nat,?x:Nat)),?y:Nat) = s(s(s(plus(?x:Nat,plus(?z:Nat,?z:Nat)))))",
"min(?x:Nat,plus(max(plus(?y:Nat,s(?z:Nat)),min(s(?z:Nat),plus(?z:Nat,0))),?y:Nat)) = plus(?x:Nat,plus(s(s(0)),s(?x:Nat)))",
"s(max(s(0),max(max(max(?x:Nat,0),min(?z:Nat,?x:Nat)),min(?y:Nat,s(?y:Nat))))) = plus(max(min(?z:Nat,0),?z:Nat),?x:Nat)",
"plus(s(plus(min(max(?x:Nat,0),min(?x:Nat,?y:Nat)),?x:Nat)),plus(plus(?x:Nat,?z:Nat),?z:Nat)) = plus(plus(0,s(?y:Nat)),s(?x:Nat))",
"plus(plus(min(min(?x:Nat,min(?z:Nat,?z:Nat)),s(?y:Nat)),s(s(max(?x:Nat,?z:Nat)))),?x:Nat) = plus(?x:Nat,s(max(s(?z:Nat),?y:Nat)))",
"max(?y:Nat,min(plus(?z:Nat,plus(?x:Nat,plus(?x:Nat,?y:Nat))),max(plus(min(?x:Nat,?x:Nat),?x:Nat),?x:Nat))) = s(max(max(?y:Nat,0),?x:Nat))",
"plus(min(0,min(min(max(0,?x:Nat),max(?z:Nat,0)),max(max(0,0),?y:Nat))),?z:Nat) = plus(s(plus(?z:Nat,?y:Nat)),0)",
"plus(0,min(s(0),min(max(min(?z:Nat,?y:Nat),?z:Nat),max(min(?x:Nat,?z:Nat),max(?z:Nat,?x:Nat))))) = s(s(plus(?x:Nat,?x:Nat)))",
"plus(s(?z:Nat),max(max(s(0),plus(0,min(?x:Nat,?y:Nat))),s(max(min(?y:Nat,?y:Nat),?x:Nat)))) = plus(?x:Nat,plus(?x:Nat,0))",
"plus(plus(min(plus(?x:Nat,min(?x:Nat,?x:Nat)),?z:Nat),plus(0,min(plus(?z:Nat,?x:Nat),s(?x:Nat)))),?y:Nat) = max(0,plus(?x:Nat,?y:Nat))",
"s(plus(plus(plus(plus(0,?y:Nat),min(?z:Nat,?z:Nat)),?y:Nat),s(plus(plus(?y:Nat,?x:Nat),min(?x:Nat,0))))) = max(s(?x:Nat),0)",
"s(plus(max(?y:Nat,min(plus(0,0),min(?x:Nat,?x:Nat))),max(plus(0,s(?x:Nat)),s(plus(?x:Nat,?z:Nat))))) = plus(?x:Nat,?x:Nat)",
"max(min(?y:Nat,0),min(min(s(s(?x:Nat)),s(max(?x:Nat,?z:Nat))),plus(s(?z:Nat),s(plus(0,?z:Nat))))) = min(0,?z:Nat)",
"max(plus(plus(s(?z:Nat),max(s(?z:Nat),plus(0,0))),max(plus(plus(?y:Nat,?x:Nat),?x:Nat),s(?x:Nat))),?x:Nat) = s(s(?x:Nat))",
"min(s(?y:Nat),plus(min(s(plus(?x:Nat,?x:Nat)),?x:Nat),plus(plus(plus(0,?y:Nat),plus(?z:Nat,?x:Nat)),s(?z:Nat)))) = min(?z:Nat,0)",
"plus(s(s(max(plus(?z:Nat,?x:Nat),min(?x:Nat,0)))),max(max(?z:Nat,min(min(?y:Nat,?y:Nat),?y:Nat)),s(?z:Nat))) = plus(0,?x:Nat)",
"min(0,min(max(s(max(?y:Nat,0)),plus(max(0,?z:Nat),s(?z:Nat))),max(plus(?z:Nat,s(0)),s(?y:Nat)))) = s(?x:Nat)",
"min(s(min(?z:Nat,0)),max(plus(plus(s(?x:Nat),0),s(s(?z:Nat))),plus(min(?y:Nat,max(0,?y:Nat)),?x:Nat))) = s(?z:Nat)",
"min(min(max(?z:Nat,0),max(max(plus(?z:Nat,?y:Nat),s(?x:Nat)),plus(?y:Nat,plus(?y:Nat,?x:Nat)))),s(min(?y:Nat,0))) = s(?x:Nat)",
"plus(s(s(min(s(?x:Nat),s(?x:Nat)))),min(s(?y:Nat),plus(max(?x:Nat,plus(?z:Nat,?x:Nat)),s(min(?x:Nat,0))))) = s(?z:Nat)",
"s(?x:Nat) = plus(min(max(plus(plus(0,?x:Nat),s(0)),s(?z:Nat)),?z:Nat),s(plus(?x:Nat,min(plus(0,?y:Nat),plus(?z:Nat,0)))))",
"s(?y:Nat) = max(plus(max(s(max(?x:Nat,?z:Nat)),?z:Nat),s(?y:Nat)),s(min(max(s(0),min(0,0)),s(max(?x:Nat,?y:Nat)))))",
"s(?z:Nat) = min(max(max(plus(plus(?x:Nat,?x:Nat),plus(?z:Nat,?z:Nat)),plus(max(?z:Nat,?z:Nat),s(0))),?x:Nat),plus(?y:Nat,min(0,?z:Nat)))",
"s(0) = max(min(plus(0,?x:Nat),max(?z:Nat,max(min(?z:Nat,0),?z:Nat))),s(plus(?y:Nat,min(plus(?y:Nat,0),max(?y:Nat,?z:Nat)))))",
"max(?x:Nat,?x:Nat) = plus(plus(s(s(min(?x:Nat,?x:Nat))),s(s(?y:Nat))),s(plus(min(max(?x:Nat,?x:Nat),?x:Nat),s(max(?x:Nat,0)))))",
"max(s(?z:Nat),?y:Nat) = min(plus(s(max(?x:Nat,min(?y:Nat,?x:Nat))),min(min(s(?z:Nat),plus(?x:Nat,?z:Nat)),s(max(?z:Nat,?y:Nat)))),?x:Nat)",
"max(s(0),?y:Nat) = min(max(max(plus(0,?z:Nat),?z:Nat),s(plus(?y:Nat,plus(?y:Nat,0)))),min(s(?x:Nat),s(s(s(?x:Nat)))))",
"min(?z:Nat,s(?x:Nat)) = max(plus(min(?z:Nat,max(?y:Nat,plus(0,?z:Nat))),plus(s(?x:Nat),max(max(?x:Nat,0),min(?y:Nat,0)))),?y:Nat)",
"plus(s(?z:Nat),?y:Nat) = min(plus(max(s(?x:Nat),plus(?x:Nat,s(?x:Nat))),max(s(min(?y:Nat,?x:Nat)),max(min(?z:Nat,?x:Nat),0))),?x:Nat)",
"s(plus(?y:Nat,s(?x:Nat))) = max(min(0,plus(?x:Nat,plus(s(?z:Nat),s(?x:Nat)))),max(min(?x:Nat,plus(?x:Nat,?x:Nat)),plus(0,?x:Nat)))",
"max(?x:Nat,min(0,?z:Nat)) = s(min(max(0,s(min(0,?x:Nat))),plus(plus(max(0,?y:Nat),plus(0,?x:Nat)),s(s(?x:Nat)))))",
"plus(s(s(s(0))),0) = min(plus(plus(s(min(0,?y:Nat)),min(?x:Nat,?x:Nat)),s(max(min(?z:Nat,0),s(?x:Nat)))),?z:Nat)",
"s(s(max(?x:Nat,max(0,?x:Nat)))) = plus(max(plus(s(s(?x:Nat)),?x:Nat),?x:Nat),plus(plus(?y:Nat,max(?z:Nat,plus(0,?z:Nat))),?x:Nat))",
"max(s(s(?x:Nat)),plus(0,s(?x:Nat))) = max(min(s(s(min(0,?z:Nat))),?z:Nat),s(min(min(s(0),s(?x:Nat)),?z:Nat)))",
"max(s(s(min(?x:Nat,?y:Nat))),s(0)) = max(?x:Nat,plus(max(s(s(?z:Nat)),min(s(?y:Nat),?z:Nat)),s(s(plus(?y:Nat,?y:Nat)))))",
"min(?z:Nat,s(min(s(s(?x:Nat)),?y:Nat))) = min(max(0,?x:Nat),s(plus(max(s(?z:Nat),plus(?z:Nat,?z:Nat)),plus(0,s(?z:Nat)))))",
"min(s(s(s(max(0,?z:Nat)))),?x:Nat) = min(max(max(plus(0,plus(0,?x:Nat)),?z:Nat),min(?y:Nat,?x:Nat)),max(0,s(?z:Nat)))",
"plus(plus(s(s(?y:Nat)),s(?x:Nat)),?y:Nat) = s(s(min(plus(s(?x:Nat),min(?x:Nat,?x:Nat)),plus(min(?x:Nat,?x:Nat),max(?x:Nat,0)))))",
"s(plus(?z:Nat,s(min(s(?x:Nat),s(0))))) = max(max(max(min(?x:Nat,s(0)),plus(?y:Nat,max(0,?z:Nat))),0),s(0))",
"s(plus(min(?z:Nat,s(min(?x:Nat,0))),0)) = s(plus(min(max(min(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),s(max(0,?x:Nat))),?y:Nat))",
"max(?x:Nat,plus(0,plus(s(s(?z:Nat)),?x:Nat))) = s(min(min(min(?x:Nat,max(?x:Nat,?x:Nat)),0),max(s(max(?x:Nat,?x:Nat)),?x:Nat)))",
"max(0,plus(plus(?y:Nat,s(s(?x:Nat))),?z:Nat)) = min(s(?z:Nat),min(?x:Nat,min(min(max(?y:Nat,0),0),plus(s(?z:Nat),?y:Nat))))",
"plus(?x:Nat,s(plus(?x:Nat,s(max(?x:Nat,?x:Nat))))) = min(s(s(0)),plus(0,max(plus(max(?x:Nat,?y:Nat),max(?z:Nat,?y:Nat)),?z:Nat)))",
"s(s(s(plus(plus(?x:Nat,0),plus(?x:Nat,?x:Nat))))) = plus(max(0,plus(s(min(?x:Nat,?x:Nat)),plus(plus(?z:Nat,0),?x:Nat))),?x:Nat)",
"s(plus(?y:Nat,plus(s(?y:Nat),s(min(?y:Nat,0))))) = min(?x:Nat,min(min(s(max(?x:Nat,?x:Nat)),plus(0,plus(?z:Nat,?z:Nat))),?y:Nat))",
"max(?x:Nat,plus(s(s(s(?y:Nat))),max(?y:Nat,0))) = s(plus(min(?y:Nat,max(max(0,?x:Nat),plus(?x:Nat,?z:Nat))),min(?x:Nat,?x:Nat)))",
"plus(?y:Nat,s(s(plus(s(?x:Nat),plus(?x:Nat,?x:Nat))))) = s(max(?y:Nat,plus(plus(s(?x:Nat),max(0,?z:Nat)),plus(s(?x:Nat),0))))",
"s(max(s(?x:Nat),max(?y:Nat,plus(min(?x:Nat,?z:Nat),?x:Nat)))) = s(plus(0,plus(plus(min(?z:Nat,?x:Nat),0),plus(?y:Nat,s(?x:Nat)))))",
"max(?x:Nat,min(min(?x:Nat,max(plus(?y:Nat,?z:Nat),0)),0)) = s(s(plus(max(s(?x:Nat),s(0)),min(max(0,?z:Nat),0))))",
"max(plus(s(0),min(s(plus(?y:Nat,?x:Nat)),?x:Nat)),?x:Nat) = plus(min(s(min(plus(?x:Nat,?z:Nat),?y:Nat)),?x:Nat),min(s(?y:Nat),?x:Nat))",
"plus(min(?y:Nat,max(plus(?x:Nat,?z:Nat),min(?x:Nat,?y:Nat))),?y:Nat) = max(plus(plus(?z:Nat,?z:Nat),?x:Nat),max(0,s(s(plus(0,?x:Nat)))))",
"min(?x:Nat,plus(plus(?z:Nat,s(min(?x:Nat,0))),min(?x:Nat,?z:Nat))) = s(plus(0,plus(min(max(?y:Nat,?x:Nat),min(0,?x:Nat)),?x:Nat)))",
"min(s(?x:Nat),plus(min(?x:Nat,plus(plus(?x:Nat,0),?y:Nat)),0)) = max(s(s(s(?z:Nat))),s(max(min(s(?x:Nat),?z:Nat),?x:Nat)))",
"min(s(?z:Nat),s(max(max(?z:Nat,0),max(s(0),?y:Nat)))) = plus(plus(?z:Nat,s(max(plus(?z:Nat,?x:Nat),min(0,?y:Nat)))),0)",
"plus(?z:Nat,plus(0,plus(plus(0,min(0,?x:Nat)),s(?x:Nat)))) = s(max(min(?z:Nat,min(plus(?x:Nat,?x:Nat),plus(0,?x:Nat))),?x:Nat))",
"plus(0,s(min(plus(s(?y:Nat),?x:Nat),s(min(?x:Nat,0))))) = min(?x:Nat,min(s(max(?x:Nat,0)),s(s(min(?x:Nat,?x:Nat)))))",
"plus(0,plus(plus(max(max(?y:Nat,?x:Nat),?x:Nat),s(?z:Nat)),?z:Nat)) = min(0,min(?z:Nat,s(min(plus(?x:Nat,?z:Nat),max(0,0)))))",
"plus(plus(s(plus(min(?z:Nat,?z:Nat),max(?x:Nat,?x:Nat))),?y:Nat),?x:Nat) = s(plus(plus(?y:Nat,plus(min(?x:Nat,?y:Nat),plus(?y:Nat,0))),0))",
"s(max(max(s(?z:Nat),max(max(?x:Nat,?x:Nat),min(?x:Nat,?x:Nat))),0)) = min(0,s(min(?z:Nat,max(s(?x:Nat),plus(?x:Nat,?x:Nat)))))",
"max(plus(?z:Nat,0),min(plus(max(?x:Nat,plus(0,0)),?x:Nat),?x:Nat)) = s(plus(s(?z:Nat),plus(?z:Nat,plus(?x:Nat,max(0,?x:Nat)))))",
"min(?x:Nat,plus(plus(max(plus(?x:Nat,?y:Nat),min(?z:Nat,0)),0),?y:Nat)) = max(min(?z:Nat,s(plus(?x:Nat,plus(?z:Nat,?x:Nat)))),s(?x:Nat))",
"plus(plus(min(?x:Nat,s(plus(?z:Nat,?x:Nat))),s(min(?y:Nat,?y:Nat))),?z:Nat) = min(?x:Nat,s(plus(?x:Nat,min(plus(?x:Nat,?x:Nat),s(0)))))",
"plus(plus(plus(s(plus(?z:Nat,?x:Nat)),?x:Nat),min(0,s(?x:Nat))),?y:Nat) = s(min(s(min(plus(?z:Nat,?x:Nat),max(?x:Nat,?x:Nat))),?y:Nat))",
"max(s(max(plus(0,0),max(0,plus(?z:Nat,?z:Nat)))),min(0,?z:Nat)) = plus(?z:Nat,plus(s(s(s(?z:Nat))),min(?x:Nat,?z:Nat)))",
"min(plus(?y:Nat,min(plus(max(?x:Nat,?x:Nat),0),s(max(?z:Nat,?y:Nat)))),?x:Nat) = s(plus(0,plus(max(min(0,?x:Nat),0),?x:Nat)))",
"plus(s(s(max(max(0,0),min(?y:Nat,0)))),s(max(?x:Nat,?z:Nat))) = plus(s(?x:Nat),s(max(s(max(?x:Nat,?x:Nat)),0)))",
"plus(plus(?x:Nat,max(plus(min(?x:Nat,0),min(?x:Nat,?x:Nat)),s(0))),?x:Nat) = max(s(min(0,plus(s(?x:Nat),s(?x:Nat)))),0)",
"plus(plus(min(?z:Nat,max(?z:Nat,0)),?x:Nat),s(max(plus(?x:Nat,?z:Nat),?y:Nat))) = plus(max(?x:Nat,plus(?x:Nat,plus(?z:Nat,s(?z:Nat)))),?y:Nat)",
"plus(plus(plus(plus(plus(?x:Nat,0),plus(?x:Nat,0)),?x:Nat),?y:Nat),s(?x:Nat)) = max(min(plus(s(plus(?y:Nat,?x:Nat)),?y:Nat),?x:Nat),?x:Nat)",
"plus(plus(?y:Nat,?z:Nat),min(s(?x:Nat),min(?z:Nat,plus(s(0),max(?z:Nat,?x:Nat))))) = max(0,plus(?z:Nat,max(?z:Nat,s(s(?x:Nat)))))",
"plus(0,plus(plus(?x:Nat,s(max(?x:Nat,0))),min(plus(s(?x:Nat),?x:Nat),s(0)))) = max(s(?z:Nat),plus(min(?z:Nat,0),?z:Nat))",
"plus(plus(?y:Nat,0),plus(min(plus(?x:Nat,s(?z:Nat)),plus(?y:Nat,min(0,?x:Nat))),?x:Nat)) = plus(?y:Nat,min(?z:Nat,s(s(s(?x:Nat)))))",
"max(s(plus(plus(s(?x:Nat),max(?x:Nat,0)),min(min(?x:Nat,?x:Nat),max(?y:Nat,0)))),?z:Nat) = max(s(s(?x:Nat)),max(?y:Nat,0))",
"min(0,plus(min(?z:Nat,plus(0,0)),plus(?x:Nat,plus(min(0,?z:Nat),plus(?y:Nat,?y:Nat))))) = max(min(?z:Nat,?z:Nat),plus(?x:Nat,0))",
"plus(min(s(plus(?z:Nat,s(?y:Nat))),?x:Nat),s(max(?x:Nat,min(plus(?y:Nat,0),min(0,?y:Nat))))) = min(plus(s(0),0),?x:Nat)",
"plus(max(min(max(s(?z:Nat),s(?z:Nat)),0),max(max(plus(?z:Nat,0),max(?x:Nat,0)),?x:Nat)),?x:Nat) = plus(?x:Nat,s(s(?x:Nat)))",
"max(max(s(plus(max(?y:Nat,?x:Nat),plus(0,?x:Nat))),max(min(?x:Nat,max(?x:Nat,?z:Nat)),min(?z:Nat,?x:Nat))),?x:Nat) = min(?x:Nat,s(?x:Nat))",
"min(max(max(plus(plus(0,?x:Nat),plus(?y:Nat,?x:Nat)),s(plus(0,?x:Nat))),s(s(min(?x:Nat,?x:Nat)))),?x:Nat) = min(s(?z:Nat),0)",
"s(plus(plus(plus(?y:Nat,?x:Nat),min(max(?x:Nat,?x:Nat),?x:Nat)),max(s(max(0,0)),max(?x:Nat,min(?x:Nat,?z:Nat))))) = plus(?y:Nat,?x:Nat)",
"max(max(plus(?y:Nat,s(plus(?y:Nat,0))),s(min(s(?y:Nat),plus(?x:Nat,?x:Nat)))),plus(?x:Nat,min(s(?y:Nat),?x:Nat))) = s(s(0))",
"max(plus(?y:Nat,?y:Nat),max(plus(min(s(?x:Nat),plus(?x:Nat,0)),s(?x:Nat)),min(plus(?x:Nat,plus(?z:Nat,?x:Nat)),?z:Nat))) = plus(?x:Nat,?x:Nat)",
"min(plus(plus(0,s(?z:Nat)),min(?x:Nat,plus(min(?y:Nat,0),min(?x:Nat,0)))),max(min(?x:Nat,s(?x:Nat)),?x:Nat)) = min(?x:Nat,?x:Nat)",
"plus(?x:Nat,plus(min(s(?x:Nat),max(max(0,?x:Nat),?y:Nat)),min(min(plus(?x:Nat,?x:Nat),max(?z:Nat,?x:Nat)),s(?x:Nat)))) = plus(?z:Nat,0)",
"s(max(min(max(max(?x:Nat,?x:Nat),?x:Nat),?x:Nat),max(plus(plus(?x:Nat,?z:Nat),plus(?y:Nat,0)),max(plus(?x:Nat,?y:Nat),?y:Nat)))) = s(?y:Nat)",
"min(s(plus(?x:Nat,?x:Nat)),min(plus(?x:Nat,s(min(0,0))),max(plus(min(?y:Nat,?x:Nat),?z:Nat),max(s(?z:Nat),?y:Nat)))) = s(?y:Nat)",
"min(min(s(plus(s(?y:Nat),max(?x:Nat,?x:Nat))),plus(min(s(?y:Nat),?y:Nat),plus(s(0),min(?z:Nat,?z:Nat)))),s(?z:Nat)) = s(?x:Nat)",
"min(plus(s(plus(?x:Nat,?x:Nat)),s(min(max(?z:Nat,?z:Nat),max(?y:Nat,?x:Nat)))),s(plus(max(s(?z:Nat),s(?z:Nat)),?x:Nat))) = s(?x:Nat)",
"s(?x:Nat) = plus(max(max(0,s(max(0,?x:Nat))),?y:Nat),plus(plus(max(s(?x:Nat),plus(?y:Nat,?x:Nat)),plus(?z:Nat,?x:Nat)),s(s(?y:Nat))))",
"s(?y:Nat) = min(min(max(plus(min(?x:Nat,?y:Nat),min(?x:Nat,?x:Nat)),?y:Nat),?z:Nat),plus(s(s(?z:Nat)),min(?z:Nat,plus(min(0,?x:Nat),?y:Nat))))",
"s(?z:Nat) = min(max(max(min(?z:Nat,?x:Nat),plus(plus(?x:Nat,0),max(?x:Nat,0))),plus(?z:Nat,plus(s(?y:Nat),min(?z:Nat,?y:Nat)))),s(0))",
"max(?x:Nat,0) = plus(?x:Nat,plus(max(max(s(?y:Nat),0),plus(?y:Nat,plus(0,?x:Nat))),plus(s(s(0)),plus(plus(?y:Nat,0),?z:Nat))))",
"min(?z:Nat,?z:Nat) = s(min(plus(plus(plus(?y:Nat,?z:Nat),min(0,?y:Nat)),s(plus(0,?x:Nat))),min(plus(s(?y:Nat),plus(?x:Nat,?z:Nat)),?x:Nat)))",
"plus(?x:Nat,?z:Nat) = min(plus(plus(max(plus(?y:Nat,?x:Nat),?x:Nat),?y:Nat),?x:Nat),max(min(max(s(?z:Nat),?x:Nat),min(s(0),s(?z:Nat))),?x:Nat))",
"plus(?x:Nat,?z:Nat) = plus(min(?x:Nat,plus(plus(min(?y:Nat,?z:Nat),s(0)),?x:Nat)),max(?x:Nat,plus(s(plus(?y:Nat,?x:Nat)),min(0,s(?x:Nat)))))",
"plus(?x:Nat,0) = plus(?y:Nat,plus(max(min(plus(0,?x:Nat),plus(?z:Nat,?y:Nat)),plus(?z:Nat,min(?z:Nat,?x:Nat))),plus(?x:Nat,s(plus(?x:Nat,0)))))",
"plus(0,?y:Nat) = min(?z:Nat,min(plus(max(plus(?x:Nat,?y:Nat),plus(?z:Nat,?x:Nat)),plus(0,?x:Nat)),plus(?y:Nat,max(s(0),plus(?x:Nat,0)))))",
"s(plus(?x:Nat,?x:Nat)) = max(plus(max(?y:Nat,?x:Nat),plus(max(min(?x:Nat,?x:Nat),max(0,?y:Nat)),plus(plus(?y:Nat,?x:Nat),max(?x:Nat,0)))),?y:Nat)",
"s(plus(?x:Nat,?x:Nat)) = plus(min(min(s(?x:Nat),?z:Nat),plus(plus(plus(?x:Nat,?y:Nat),s(?y:Nat)),plus(min(0,0),plus(?x:Nat,0)))),?x:Nat)",
"plus(s(?x:Nat),?y:Nat) = max(min(plus(plus(?x:Nat,?y:Nat),min(max(?x:Nat,0),s(?y:Nat))),?z:Nat),max(max(max(?x:Nat,s(0)),0),?y:Nat))",
"plus(s(0),?x:Nat) = plus(min(?x:Nat,plus(s(?z:Nat),plus(s(0),plus(?y:Nat,?x:Nat)))),plus(plus(?z:Nat,plus(min(?y:Nat,?x:Nat),?x:Nat)),0))",
"max(plus(?z:Nat,?x:Nat),?x:Nat) = min(max(s(plus(s(?z:Nat),min(?x:Nat,?y:Nat))),?z:Nat),min(plus(min(s(0),max(?y:Nat,?x:Nat)),0),?x:Nat))",
"plus(?x:Nat,plus(?x:Nat,?x:Nat)) = max(min(min(max(?z:Nat,plus(0,?y:Nat)),plus(s(0),plus(?x:Nat,?x:Nat))),?x:Nat),max(?y:Nat,min(?x:Nat,0)))",
"plus(min(?z:Nat,?z:Nat),0) = max(?y:Nat,min(s(plus(?x:Nat,s(?x:Nat))),min(min(plus(?y:Nat,0),?y:Nat),plus(plus(?z:Nat,?x:Nat),s(0)))))",
"s(s(min(?x:Nat,s(?x:Nat)))) = plus(plus(?z:Nat,s(max(min(?z:Nat,?x:Nat),max(?z:Nat,?z:Nat)))),plus(?z:Nat,s(min(s(?y:Nat),s(?x:Nat)))))",
"max(?x:Nat,s(max(?y:Nat,?z:Nat))) = s(min(plus(s(0),plus(plus(?z:Nat,?x:Nat),min(?z:Nat,?x:Nat))),max(?z:Nat,plus(s(0),s(?x:Nat)))))",
"max(min(0,0),s(?x:Nat)) = plus(0,plus(max(max(?x:Nat,?y:Nat),min(?x:Nat,?x:Nat)),max(plus(?y:Nat,s(0)),s(plus(?z:Nat,?x:Nat)))))",
"plus(s(s(?y:Nat)),max(?y:Nat,?x:Nat)) = plus(?z:Nat,plus(max(?x:Nat,?z:Nat),max(min(s(?z:Nat),plus(?z:Nat,?y:Nat)),plus(?z:Nat,plus(0,?x:Nat)))))",
"s(min(s(s(?y:Nat)),plus(0,?x:Nat))) = s(plus(min(?y:Nat,max(?y:Nat,min(?z:Nat,?z:Nat))),s(plus(max(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat)))))",
"max(s(plus(s(?x:Nat),s(?y:Nat))),?z:Nat) = max(?y:Nat,max(min(max(?y:Nat,min(?z:Nat,?x:Nat)),0),s(plus(plus(?x:Nat,?y:Nat),s(?x:Nat)))))",
"min(max(?x:Nat,min(?x:Nat,s(?x:Nat))),?y:Nat) = min(s(s(?x:Nat)),min(?x:Nat,max(min(?y:Nat,?z:Nat),plus(min(?z:Nat,?y:Nat),max(?x:Nat,?x:Nat)))))",
"plus(s(s(s(min(0,0)))),?x:Nat) = s(plus(min(max(0,s(?y:Nat)),min(max(?y:Nat,?x:Nat),?x:Nat)),s(s(min(?x:Nat,0)))))",
"plus(s(min(?x:Nat,?x:Nat)),plus(?y:Nat,?x:Nat)) = s(max(min(plus(?z:Nat,s(?y:Nat)),plus(?x:Nat,0)),min(?z:Nat,min(s(?x:Nat),s(?z:Nat)))))",
"min(?x:Nat,s(min(s(plus(?z:Nat,?y:Nat)),?x:Nat))) = s(plus(max(plus(?x:Nat,max(?y:Nat,?x:Nat)),s(?z:Nat)),plus(plus(?z:Nat,s(0)),0)))",
"min(min(s(s(?y:Nat)),s(?z:Nat)),s(?z:Nat)) = max(?z:Nat,plus(plus(s(0),s(s(?x:Nat))),max(?y:Nat,max(0,min(?z:Nat,0)))))",
"s(max(s(0),s(max(?z:Nat,plus(?x:Nat,?x:Nat))))) = min(s(s(s(max(?x:Nat,?x:Nat)))),min(s(max(min(?y:Nat,0),?x:Nat)),?z:Nat))",
"s(max(plus(?y:Nat,plus(min(?x:Nat,?x:Nat),?x:Nat)),?x:Nat)) = plus(min(s(max(?y:Nat,0)),max(min(s(?z:Nat),s(?x:Nat)),?x:Nat)),s(?x:Nat))",
"max(0,s(max(?x:Nat,max(?x:Nat,min(?x:Nat,?x:Nat))))) = plus(max(min(max(?x:Nat,max(?x:Nat,?x:Nat)),max(s(?x:Nat),s(?y:Nat))),?x:Nat),?x:Nat)",
"max(plus(plus(?x:Nat,?x:Nat),?y:Nat),max(s(?x:Nat),?x:Nat)) = max(?y:Nat,max(s(plus(plus(?z:Nat,?z:Nat),s(?x:Nat))),plus(?z:Nat,s(s(0)))))",
"plus(0,max(s(?y:Nat),s(s(plus(?y:Nat,?x:Nat))))) = min(plus(min(min(min(?y:Nat,?x:Nat),s(?z:Nat)),max(s(?x:Nat),?z:Nat)),?y:Nat),?y:Nat)",
"plus(0,min(s(?x:Nat),plus(s(?x:Nat),s(?z:Nat)))) = s(s(min(plus(?z:Nat,plus(?x:Nat,?z:Nat)),plus(max(?y:Nat,?x:Nat),min(?z:Nat,?y:Nat)))))",
"plus(s(?x:Nat),s(min(?y:Nat,s(min(?x:Nat,?x:Nat))))) = s(s(min(min(max(?z:Nat,?z:Nat),plus(?x:Nat,?x:Nat)),max(max(?y:Nat,0),0))))",
"plus(s(0),s(plus(plus(s(?z:Nat),?z:Nat),?x:Nat))) = max(plus(s(plus(0,s(?x:Nat))),max(plus(?z:Nat,0),s(s(?x:Nat)))),?x:Nat)",
"s(max(max(?x:Nat,0),s(min(?x:Nat,max(0,0))))) = max(max(?x:Nat,plus(max(plus(?y:Nat,0),?x:Nat),s(?x:Nat))),max(?z:Nat,?z:Nat))",
"plus(plus(plus(max(s(0),?z:Nat),?x:Nat),?x:Nat),s(?y:Nat)) = min(plus(?x:Nat,s(?x:Nat)),plus(?z:Nat,s(max(s(?x:Nat),plus(?x:Nat,?x:Nat)))))",
"s(plus(plus(max(plus(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),?x:Nat),?y:Nat)) = min(plus(s(s(?z:Nat)),?z:Nat),s(s(plus(s(0),s(?x:Nat)))))",
"max(max(?x:Nat,s(0)),min(0,s(s(max(?x:Nat,?x:Nat))))) = min(?x:Nat,max(?y:Nat,plus(s(?x:Nat),plus(s(?x:Nat),max(?z:Nat,0)))))",
"plus(plus(max(max(0,max(?z:Nat,?x:Nat)),?y:Nat),s(?x:Nat)),?x:Nat) = min(?z:Nat,min(plus(max(?x:Nat,?y:Nat),?x:Nat),s(s(max(?z:Nat,?x:Nat)))))",
"s(plus(?x:Nat,max(max(plus(?x:Nat,0),plus(0,?x:Nat)),s(?z:Nat)))) = s(s(max(max(0,s(0)),min(?x:Nat,plus(0,?x:Nat)))))",
"max(0,plus(min(min(plus(?x:Nat,?x:Nat),plus(?x:Nat,?z:Nat)),?z:Nat),0)) = max(?z:Nat,s(plus(min(?y:Nat,?x:Nat),plus(?y:Nat,min(?z:Nat,?x:Nat)))))",
"max(s(plus(?x:Nat,?y:Nat)),s(plus(min(plus(?y:Nat,?x:Nat),0),?z:Nat))) = max(s(plus(s(?x:Nat),s(min(?x:Nat,?y:Nat)))),plus(?x:Nat,?x:Nat))",
"min(plus(?x:Nat,s(0)),s(max(s(max(?x:Nat,?y:Nat)),s(0)))) = plus(?z:Nat,s(plus(max(min(0,?x:Nat),plus(?x:Nat,0)),?z:Nat)))",
"max(s(?z:Nat),plus(0,min(s(?y:Nat),min(s(?y:Nat),min(?x:Nat,?x:Nat))))) = min(0,s(max(plus(?x:Nat,?x:Nat),min(s(?x:Nat),0))))",
"plus(s(?z:Nat),min(plus(?z:Nat,min(min(0,?x:Nat),min(?z:Nat,?z:Nat))),?x:Nat)) = plus(s(?z:Nat),s(max(max(s(?y:Nat),?x:Nat),s(?z:Nat))))",
"min(?x:Nat,max(min(s(plus(?z:Nat,?x:Nat)),s(s(?x:Nat))),plus(0,s(?x:Nat)))) = s(max(min(min(max(?x:Nat,0),?z:Nat),?z:Nat),?y:Nat))",
"min(s(0),max(min(?x:Nat,?x:Nat),plus(max(s(?y:Nat),min(?y:Nat,?y:Nat)),?x:Nat))) = max(plus(?z:Nat,s(min(min(?x:Nat,?x:Nat),?x:Nat))),?z:Nat)",
"plus(s(s(plus(plus(?z:Nat,?y:Nat),?x:Nat))),s(max(?z:Nat,s(min(?y:Nat,?x:Nat))))) = max(max(max(0,?x:Nat),?z:Nat),plus(0,s(?z:Nat)))",
"plus(s(max(s(?x:Nat),plus(max(0,0),plus(0,?x:Nat)))),plus(?z:Nat,?y:Nat)) = max(plus(min(s(min(?z:Nat,?z:Nat)),?y:Nat),?y:Nat),?x:Nat)",
"max(s(min(min(s(0),plus(?z:Nat,0)),max(s(?x:Nat),max(0,?z:Nat)))),?x:Nat) = min(s(min(s(?y:Nat),0)),s(s(?x:Nat)))",
"max(max(s(?x:Nat),s(?z:Nat)),s(plus(max(plus(?y:Nat,?x:Nat),plus(?z:Nat,?z:Nat)),0))) = plus(?y:Nat,plus(0,s(plus(s(?z:Nat),?x:Nat))))",
"min(?y:Nat,max(max(?z:Nat,min(?x:Nat,plus(?x:Nat,?x:Nat))),s(max(max(?y:Nat,?z:Nat),0)))) = s(plus(s(s(s(?z:Nat))),s(s(?z:Nat))))",
"plus(max(0,plus(?y:Nat,?z:Nat)),min(?z:Nat,min(min(min(?y:Nat,?x:Nat),plus(?z:Nat,?z:Nat)),?y:Nat))) = min(s(?x:Nat),s(s(min(?x:Nat,?y:Nat))))",
"plus(max(plus(s(max(0,?y:Nat)),?x:Nat),s(max(plus(?x:Nat,?x:Nat),?y:Nat))),plus(0,?x:Nat)) = plus(plus(s(max(?x:Nat,?y:Nat)),0),?y:Nat)",
"s(plus(plus(max(min(?x:Nat,?x:Nat),plus(0,?x:Nat)),max(?y:Nat,min(?x:Nat,?y:Nat))),max(?z:Nat,?x:Nat))) = plus(s(plus(s(?x:Nat),?x:Nat)),?z:Nat)",
"plus(?z:Nat,max(s(s(plus(0,?x:Nat))),min(max(s(?x:Nat),?z:Nat),max(?z:Nat,max(?y:Nat,?y:Nat))))) = max(plus(?y:Nat,?x:Nat),plus(?x:Nat,?y:Nat))",
"plus(s(max(plus(?z:Nat,max(?x:Nat,?x:Nat)),min(s(?x:Nat),?x:Nat))),max(s(min(?x:Nat,?z:Nat)),?y:Nat)) = s(min(max(?y:Nat,?x:Nat),s(?z:Nat)))",
"max(plus(max(s(plus(?z:Nat,?z:Nat)),plus(min(?x:Nat,?x:Nat),max(?x:Nat,?z:Nat))),s(min(?y:Nat,?z:Nat))),0) = plus(s(min(0,?y:Nat)),0)",
"plus(s(min(0,plus(s(?x:Nat),plus(?x:Nat,0)))),s(plus(?y:Nat,plus(min(?z:Nat,?y:Nat),plus(?z:Nat,?y:Nat))))) = plus(?x:Nat,plus(?z:Nat,?y:Nat))",
"plus(?y:Nat,max(min(min(plus(?x:Nat,?z:Nat),0),s(min(?x:Nat,?y:Nat))),plus(s(max(0,?y:Nat)),s(s(?x:Nat))))) = s(s(s(?x:Nat)))",
"min(?x:Nat,min(min(s(min(?x:Nat,?x:Nat)),s(min(?z:Nat,0))),max(s(?y:Nat),max(plus(?z:Nat,0),plus(0,?z:Nat))))) = max(?z:Nat,0)",
"max(s(max(max(max(?y:Nat,0),s(0)),s(?x:Nat))),s(max(min(plus(0,?x:Nat),?z:Nat),min(?x:Nat,max(?z:Nat,0))))) = s(?x:Nat)",
"max(min(max(max(plus(?x:Nat,0),s(?z:Nat)),s(plus(?z:Nat,?y:Nat))),max(0,plus(min(?y:Nat,?z:Nat),plus(0,?y:Nat)))),?z:Nat) = s(?x:Nat)",
"min(max(plus(s(s(?x:Nat)),?y:Nat),min(max(plus(?z:Nat,?x:Nat),?x:Nat),?z:Nat)),s(s(plus(max(?x:Nat,?x:Nat),min(?z:Nat,?y:Nat))))) = s(?x:Nat)",
"plus(?x:Nat,max(min(plus(max(?z:Nat,?z:Nat),min(0,?x:Nat)),?y:Nat),plus(s(max(?x:Nat,?x:Nat)),plus(max(?x:Nat,0),s(?z:Nat))))) = s(?x:Nat)",
"plus(s(plus(max(min(?y:Nat,?x:Nat),0),min(s(?x:Nat),min(0,?y:Nat)))),plus(plus(max(?z:Nat,plus(?y:Nat,?z:Nat)),0),?x:Nat)) = s(?z:Nat)",
"s(?z:Nat) = plus(s(max(plus(plus(?x:Nat,?z:Nat),min(?z:Nat,?z:Nat)),s(s(?x:Nat)))),plus(max(?y:Nat,?x:Nat),s(max(s(0),plus(?z:Nat,?x:Nat)))))",
"s(0) = max(plus(max(?y:Nat,plus(s(?x:Nat),0)),min(?y:Nat,min(min(?y:Nat,0),s(?x:Nat)))),s(plus(?x:Nat,max(?z:Nat,max(0,?x:Nat)))))",
"plus(?x:Nat,?x:Nat) = min(?z:Nat,min(s(plus(min(?y:Nat,?x:Nat),plus(?x:Nat,0))),min(plus(plus(?z:Nat,?y:Nat),?y:Nat),min(max(0,?z:Nat),s(?x:Nat)))))",
"plus(?z:Nat,?y:Nat) = plus(s(plus(0,min(max(?y:Nat,?x:Nat),?x:Nat))),max(min(plus(s(?x:Nat),?x:Nat),min(plus(?z:Nat,?x:Nat),min(?x:Nat,?y:Nat))),?y:Nat))",
"min(0,s(0)) = max(?y:Nat,plus(max(plus(s(?x:Nat),max(0,?y:Nat)),?y:Nat),max(s(plus(?x:Nat,?y:Nat)),plus(s(?y:Nat),max(?z:Nat,?x:Nat)))))",
"s(max(s(?y:Nat),0)) = plus(plus(s(0),min(min(s(?x:Nat),s(?z:Nat)),s(max(?x:Nat,0)))),max(?x:Nat,s(max(?x:Nat,s(?x:Nat)))))",
"s(plus(s(?x:Nat),?z:Nat)) = plus(min(?x:Nat,s(?y:Nat)),plus(plus(s(max(?x:Nat,?x:Nat)),s(max(?x:Nat,?x:Nat))),s(max(?z:Nat,min(0,?x:Nat)))))",
"plus(s(s(?x:Nat)),?x:Nat) = max(plus(plus(min(plus(?x:Nat,0),?y:Nat),0),s(?x:Nat)),max(s(s(plus(?z:Nat,?x:Nat))),max(?z:Nat,s(?y:Nat))))",
"max(max(?z:Nat,s(?z:Nat)),?y:Nat) = plus(plus(s(?x:Nat),min(?y:Nat,plus(s(?x:Nat),min(?y:Nat,?z:Nat)))),min(max(s(min(?z:Nat,?x:Nat)),?y:Nat),?x:Nat))",
"plus(s(s(s(?z:Nat))),?y:Nat) = s(plus(plus(min(?z:Nat,?x:Nat),plus(plus(?z:Nat,?x:Nat),s(?y:Nat))),s(plus(min(?x:Nat,?y:Nat),plus(?z:Nat,?z:Nat)))))",
"max(max(plus(?x:Nat,?z:Nat),?x:Nat),?x:Nat) = plus(plus(s(?z:Nat),?z:Nat),s(min(plus(max(?x:Nat,?x:Nat),plus(?y:Nat,?y:Nat)),max(s(0),s(?y:Nat)))))",
"plus(s(min(s(?x:Nat),?x:Nat)),?x:Nat) = max(min(plus(0,max(?x:Nat,s(?z:Nat))),plus(?y:Nat,max(s(?y:Nat),plus(0,?x:Nat)))),min(?x:Nat,?x:Nat))",
"s(s(min(?y:Nat,s(min(?y:Nat,?x:Nat))))) = min(0,plus(s(min(s(?z:Nat),?x:Nat)),max(plus(plus(?z:Nat,?x:Nat),s(0)),s(s(?x:Nat)))))",
"plus(?x:Nat,plus(s(min(s(?y:Nat),?y:Nat)),?z:Nat)) = plus(max(plus(?z:Nat,s(?y:Nat)),max(min(min(?x:Nat,0),?y:Nat),min(?x:Nat,s(?x:Nat)))),?y:Nat)",
"s(s(s(plus(plus(?y:Nat,?x:Nat),max(?y:Nat,?x:Nat))))) = s(max(?z:Nat,max(min(plus(?x:Nat,?z:Nat),min(?x:Nat,?z:Nat)),min(?z:Nat,plus(0,?x:Nat)))))",
"max(?x:Nat,max(s(s(plus(?z:Nat,0))),s(?z:Nat))) = max(0,plus(s(plus(min(?z:Nat,?x:Nat),max(0,?x:Nat))),s(s(max(?y:Nat,?z:Nat)))))",
"plus(?x:Nat,plus(0,s(min(?y:Nat,max(?x:Nat,?y:Nat))))) = plus(?x:Nat,max(max(s(plus(?y:Nat,?z:Nat)),s(0)),max(?z:Nat,s(max(0,?z:Nat)))))",
"plus(plus(s(plus(0,max(?y:Nat,0))),0),?y:Nat) = min(0,max(0,plus(max(s(?x:Nat),min(?z:Nat,?y:Nat)),max(?z:Nat,plus(?x:Nat,?y:Nat)))))",
"plus(plus(plus(?x:Nat,s(max(?x:Nat,?x:Nat))),?x:Nat),?x:Nat) = s(min(max(min(0,min(?y:Nat,0)),?x:Nat),min(plus(min(0,?x:Nat),?x:Nat),?y:Nat)))",
"s(s(min(plus(min(0,?x:Nat),plus(?z:Nat,?x:Nat)),?x:Nat))) = s(plus(plus(min(?z:Nat,0),?x:Nat),s(min(min(?z:Nat,?x:Nat),plus(?x:Nat,?x:Nat)))))",
"max(s(plus(0,min(s(?y:Nat),max(?z:Nat,?z:Nat)))),?z:Nat) = min(plus(?z:Nat,max(s(?x:Nat),?x:Nat)),s(s(max(s(?y:Nat),plus(?y:Nat,?x:Nat)))))",
"min(?x:Nat,max(s(plus(?y:Nat,min(?y:Nat,0))),s(?z:Nat))) = max(?z:Nat,max(max(?y:Nat,?z:Nat),plus(s(min(?x:Nat,?x:Nat)),min(s(?x:Nat),?z:Nat))))",
"min(?z:Nat,min(max(0,max(0,min(?y:Nat,?y:Nat))),?x:Nat)) = max(s(s(max(?x:Nat,s(?x:Nat)))),max(s(plus(?x:Nat,min(?x:Nat,?x:Nat))),?z:Nat))",
"plus(0,max(s(min(?x:Nat,s(?x:Nat))),min(?z:Nat,?z:Nat))) = max(min(plus(min(s(?y:Nat),s(?y:Nat)),?x:Nat),min(max(?y:Nat,?z:Nat),?x:Nat)),0)",
"plus(plus(?y:Nat,0),s(s(min(max(?x:Nat,?y:Nat),?x:Nat)))) = max(max(plus(plus(?x:Nat,?y:Nat),?x:Nat),s(s(min(?x:Nat,?z:Nat)))),plus(?x:Nat,?z:Nat))",
"max(min(?z:Nat,s(min(max(?x:Nat,?y:Nat),?x:Nat))),s(s(?x:Nat))) = s(min(?x:Nat,plus(min(?z:Nat,?x:Nat),max(plus(?z:Nat,?x:Nat),plus(?x:Nat,?z:Nat)))))",
"min(?x:Nat,s(min(min(min(?z:Nat,?z:Nat),?y:Nat),plus(?z:Nat,?z:Nat)))) = max(min(?z:Nat,max(s(?y:Nat),max(max(?y:Nat,?x:Nat),min(?z:Nat,?x:Nat)))),0)",
"plus(?z:Nat,min(min(plus(?y:Nat,?z:Nat),plus(s(?z:Nat),?y:Nat)),?x:Nat)) = min(?y:Nat,min(s(s(plus(?x:Nat,?x:Nat))),plus(min(?x:Nat,s(?z:Nat)),?y:Nat)))",
"plus(plus(max(min(plus(?x:Nat,?z:Nat),s(0)),?x:Nat),?z:Nat),?z:Nat) = plus(plus(plus(?x:Nat,?x:Nat),s(plus(max(?y:Nat,0),min(?x:Nat,?z:Nat)))),0)",
"s(max(plus(0,?x:Nat),min(s(plus(?x:Nat,?x:Nat)),s(s(?y:Nat))))) = s(s(plus(s(s(?x:Nat)),max(min(0,?x:Nat),plus(0,?x:Nat)))))",
"max(plus(?z:Nat,min(plus(?x:Nat,min(?z:Nat,?y:Nat)),s(s(?x:Nat)))),?x:Nat) = plus(plus(min(?x:Nat,s(min(?y:Nat,?z:Nat))),s(?y:Nat)),plus(?y:Nat,0))",
"max(plus(s(plus(?x:Nat,plus(?y:Nat,0))),s(plus(?x:Nat,?x:Nat))),?y:Nat) = plus(plus(min(?z:Nat,plus(s(?y:Nat),plus(?x:Nat,?z:Nat))),s(?y:Nat)),?z:Nat)",
"min(max(?z:Nat,plus(min(?x:Nat,s(?x:Nat)),s(s(?y:Nat)))),s(0)) = min(?y:Nat,max(s(?z:Nat),s(plus(min(?x:Nat,?z:Nat),min(?y:Nat,?x:Nat)))))",
"plus(?x:Nat,s(max(min(min(0,?x:Nat),max(0,?z:Nat)),plus(0,?x:Nat)))) = s(plus(max(?x:Nat,min(plus(?x:Nat,?y:Nat),min(?x:Nat,?x:Nat))),?z:Nat))",
"plus(plus(s(max(0,plus(?z:Nat,?x:Nat))),s(?x:Nat)),min(?x:Nat,s(?x:Nat))) = max(?y:Nat,plus(plus(plus(min(?x:Nat,?x:Nat),s(?y:Nat)),0),?z:Nat))",
"plus(plus(plus(max(?y:Nat,plus(?y:Nat,?x:Nat)),s(min(?z:Nat,?z:Nat))),0),?x:Nat) = s(min(plus(s(0),plus(s(?x:Nat),min(?z:Nat,?x:Nat))),0))",
"s(plus(s(plus(?x:Nat,min(0,?y:Nat))),max(?z:Nat,plus(min(0,0),?x:Nat)))) = max(s(min(?y:Nat,min(s(?z:Nat),min(?y:Nat,?y:Nat)))),?z:Nat)",
"max(?z:Nat,min(s(plus(plus(?z:Nat,0),s(?y:Nat))),plus(?x:Nat,s(s(?z:Nat))))) = plus(0,max(s(max(plus(0,?x:Nat),s(?z:Nat))),?y:Nat))",
"max(0,plus(s(s(s(0))),min(?y:Nat,min(s(?z:Nat),max(?z:Nat,?x:Nat))))) = plus(plus(?x:Nat,s(min(s(0),plus(0,?x:Nat)))),?z:Nat)",
"max(max(s(plus(min(0,0),max(0,0))),min(?x:Nat,?z:Nat)),s(?y:Nat)) = max(min(max(?z:Nat,max(s(?x:Nat),?x:Nat)),0),s(0))",
"min(max(?z:Nat,0),min(max(min(s(0),max(?z:Nat,?z:Nat)),s(?y:Nat)),?x:Nat)) = plus(max(?y:Nat,s(max(s(?x:Nat),min(?y:Nat,0)))),?x:Nat)",
"plus(?x:Nat,plus(min(min(?y:Nat,?z:Nat),s(plus(?x:Nat,0))),s(plus(?x:Nat,?x:Nat)))) = plus(0,max(max(s(s(?x:Nat)),s(s(?x:Nat))),?x:Nat))",
"plus(?y:Nat,max(s(s(?y:Nat)),min(s(max(?x:Nat,?x:Nat)),s(max(?x:Nat,?z:Nat))))) = min(plus(0,?x:Nat),min(s(plus(?x:Nat,s(?x:Nat))),?x:Nat))",
"max(plus(min(s(?y:Nat),?z:Nat),plus(?x:Nat,max(plus(?x:Nat,0),min(?y:Nat,?x:Nat)))),?y:Nat) = min(s(plus(plus(?y:Nat,max(?x:Nat,?x:Nat)),?x:Nat)),?x:Nat)",
"min(max(min(?x:Nat,s(plus(?x:Nat,?x:Nat))),plus(?x:Nat,?x:Nat)),max(?y:Nat,min(?x:Nat,?y:Nat))) = max(?x:Nat,max(max(s(max(?y:Nat,?y:Nat)),?x:Nat),?z:Nat))",
"min(?z:Nat,max(plus(?x:Nat,plus(plus(?x:Nat,0),?y:Nat)),max(max(s(?z:Nat),s(?z:Nat)),?x:Nat))) = plus(?z:Nat,plus(s(s(?x:Nat)),max(?x:Nat,?x:Nat)))",
"max(0,plus(plus(plus(plus(?z:Nat,?y:Nat),min(?z:Nat,?y:Nat)),plus(s(?x:Nat),plus(?x:Nat,0))),0)) = s(plus(max(?y:Nat,s(s(?z:Nat))),?x:Nat))",
"min(plus(min(s(s(?z:Nat)),min(max(?x:Nat,0),0)),max(s(?y:Nat),s(plus(?y:Nat,?x:Nat)))),?x:Nat) = s(plus(plus(?z:Nat,?z:Nat),s(?y:Nat)))",
"plus(?x:Nat,max(s(min(?x:Nat,0)),plus(min(max(?z:Nat,0),max(0,?x:Nat)),s(max(?z:Nat,?z:Nat))))) = plus(?x:Nat,min(plus(?y:Nat,?y:Nat),?x:Nat))",
"max(plus(max(?z:Nat,plus(s(0),0)),?x:Nat),max(max(plus(plus(?y:Nat,?z:Nat),max(?x:Nat,?y:Nat)),?y:Nat),?y:Nat)) = s(s(s(min(?y:Nat,0))))",
"max(plus(max(s(plus(0,?z:Nat)),min(0,s(?z:Nat))),s(?x:Nat)),plus(plus(?x:Nat,max(?x:Nat,?x:Nat)),0)) = min(?z:Nat,plus(?x:Nat,s(?y:Nat)))",
"min(max(?x:Nat,s(plus(max(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat)))),max(plus(min(?z:Nat,max(?x:Nat,?x:Nat)),s(?x:Nat)),?y:Nat)) = max(?x:Nat,min(?y:Nat,0))",
"min(max(plus(max(min(?y:Nat,?x:Nat),s(?x:Nat)),?x:Nat),s(?x:Nat)),max(0,max(max(plus(?z:Nat,?y:Nat),?x:Nat),?z:Nat))) = s(min(?z:Nat,s(?z:Nat)))",
"plus(s(s(?x:Nat)),max(s(min(plus(?z:Nat,?y:Nat),s(0))),plus(?x:Nat,max(min(?z:Nat,?x:Nat),plus(?x:Nat,?z:Nat))))) = s(s(max(?x:Nat,0)))",
"plus(s(min(?y:Nat,min(s(?x:Nat),plus(?z:Nat,0)))),plus(?y:Nat,plus(s(max(?x:Nat,?x:Nat)),s(max(?x:Nat,?x:Nat))))) = min(?y:Nat,max(?y:Nat,?y:Nat))",
"plus(s(plus(?x:Nat,min(s(?z:Nat),max(?y:Nat,?y:Nat)))),plus(plus(?x:Nat,max(plus(?y:Nat,?y:Nat),max(?z:Nat,?y:Nat))),?z:Nat)) = max(plus(?z:Nat,?y:Nat),?x:Nat)",
"max(plus(plus(s(s(?x:Nat)),min(0,?x:Nat)),max(max(plus(?z:Nat,0),min(?z:Nat,?x:Nat)),min(s(0),?x:Nat))),0) = s(plus(?z:Nat,?z:Nat))",
"plus(max(0,s(?x:Nat)),plus(min(min(s(?y:Nat),min(0,?z:Nat)),s(?z:Nat)),plus(0,min(s(?x:Nat),s(0))))) = plus(s(?x:Nat),0)",
"plus(plus(?y:Nat,max(max(?y:Nat,?x:Nat),s(plus(?z:Nat,?x:Nat)))),s(min(max(max(?x:Nat,0),?z:Nat),plus(?y:Nat,s(0))))) = s(min(?y:Nat,?z:Nat))",
"max(max(plus(?x:Nat,?x:Nat),plus(max(max(?y:Nat,?z:Nat),max(?z:Nat,0)),max(max(?y:Nat,?z:Nat),plus(?x:Nat,0)))),plus(0,?z:Nat)) = max(?x:Nat,?x:Nat)",
"min(s(s(?y:Nat)),min(min(plus(s(?y:Nat),0),plus(min(0,?x:Nat),s(?x:Nat))),s(plus(s(?y:Nat),plus(?x:Nat,?z:Nat))))) = s(s(?x:Nat))",
"plus(?x:Nat,max(plus(plus(?x:Nat,plus(?x:Nat,?y:Nat)),?x:Nat),plus(min(plus(0,0),plus(?z:Nat,?x:Nat)),max(max(?x:Nat,?y:Nat),?x:Nat)))) = max(?x:Nat,0)",
"s(plus(min(plus(s(0),plus(?z:Nat,?y:Nat)),max(min(?z:Nat,?x:Nat),plus(?x:Nat,0))),plus(min(0,0),min(?x:Nat,s(?x:Nat))))) = s(?y:Nat)",
"plus(s(?x:Nat),plus(max(plus(?x:Nat,max(?y:Nat,?x:Nat)),min(plus(?x:Nat,?z:Nat),s(?x:Nat))),plus(?x:Nat,max(s(0),min(?x:Nat,?x:Nat))))) = s(?x:Nat)",
"plus(plus(s(?x:Nat),plus(plus(max(0,?x:Nat),?x:Nat),plus(plus(?z:Nat,?z:Nat),?y:Nat))),min(0,min(s(s(?z:Nat)),s(s(?x:Nat))))) = s(0)",
"s(?x:Nat) = plus(s(plus(max(?z:Nat,s(?z:Nat)),?y:Nat)),plus(min(0,plus(max(?x:Nat,?x:Nat),s(?z:Nat))),min(s(?z:Nat),plus(max(0,?y:Nat),?x:Nat))))",
"s(?x:Nat) = plus(plus(min(0,?x:Nat),plus(s(max(?z:Nat,0)),min(max(?z:Nat,0),?x:Nat))),plus(plus(0,?x:Nat),s(max(max(?y:Nat,?z:Nat),?y:Nat))))",
"s(?z:Nat) = max(plus(plus(s(s(0)),s(?z:Nat)),s(?y:Nat)),min(max(0,s(min(0,?y:Nat))),max(?y:Nat,max(s(?y:Nat),max(?z:Nat,?z:Nat)))))",
"s(s(?y:Nat)) = plus(plus(plus(s(max(0,?y:Nat)),max(plus(?y:Nat,?y:Nat),max(?z:Nat,?x:Nat))),min(s(?y:Nat),plus(?y:Nat,min(?x:Nat,?y:Nat)))),s(?z:Nat))",
"max(?x:Nat,?x:Nat) = max(plus(s(?x:Nat),plus(plus(0,0),max(s(?z:Nat),s(?x:Nat)))),plus(s(?y:Nat),plus(min(plus(?y:Nat,?y:Nat),s(0)),?x:Nat)))",
"min(?y:Nat,0) = max(min(?z:Nat,plus(min(min(?x:Nat,0),?x:Nat),?x:Nat)),max(plus(min(max(?y:Nat,?x:Nat),plus(?y:Nat,?x:Nat)),min(s(?x:Nat),?x:Nat)),?z:Nat))",
"s(s(s(?z:Nat))) = plus(0,plus(plus(s(min(?x:Nat,?z:Nat)),plus(s(0),min(?x:Nat,?y:Nat))),plus(plus(?y:Nat,?x:Nat),plus(s(?z:Nat),s(?z:Nat)))))",
"max(s(?z:Nat),?x:Nat) = min(?y:Nat,plus(plus(plus(plus(?y:Nat,?y:Nat),plus(0,0)),plus(max(?y:Nat,?x:Nat),min(?z:Nat,?x:Nat))),s(min(s(0),?z:Nat))))",
"plus(?x:Nat,max(?y:Nat,?x:Nat)) = plus(?x:Nat,max(plus(max(s(?z:Nat),s(?x:Nat)),?z:Nat),plus(max(plus(?x:Nat,?z:Nat),max(?y:Nat,?z:Nat)),max(?x:Nat,s(?x:Nat)))))",
"plus(s(0),s(?x:Nat)) = plus(min(0,s(min(s(0),?x:Nat))),min(max(?y:Nat,?z:Nat),max(s(plus(0,?x:Nat)),plus(?x:Nat,min(0,0)))))",
"s(s(plus(s(0),s(plus(?y:Nat,?y:Nat))))) = plus(s(?x:Nat),max(min(?y:Nat,?y:Nat),plus(plus(?z:Nat,min(?x:Nat,?z:Nat)),max(?x:Nat,plus(?y:Nat,?x:Nat)))))",
"s(min(s(s(?x:Nat)),max(0,s(?y:Nat)))) = max(s(?y:Nat),plus(plus(plus(0,0),?z:Nat),max(plus(min(?x:Nat,?x:Nat),max(?z:Nat,0)),0)))",
"s(min(max(?y:Nat,plus(s(0),?y:Nat)),?x:Nat)) = min(0,plus(max(s(?y:Nat),?z:Nat),min(min(max(?x:Nat,?x:Nat),?x:Nat),plus(?x:Nat,min(?x:Nat,?x:Nat)))))",
"min(min(0,min(?x:Nat,0)),max(0,?y:Nat)) = min(s(max(plus(plus(?x:Nat,0),s(?z:Nat)),max(?y:Nat,s(?z:Nat)))),s(plus(?z:Nat,s(?x:Nat))))",
"min(min(s(s(plus(0,?z:Nat))),?x:Nat),?y:Nat) = plus(plus(max(min(plus(0,0),?x:Nat),?x:Nat),?z:Nat),plus(?x:Nat,s(max(s(?z:Nat),s(?z:Nat)))))",
"s(s(s(plus(plus(?x:Nat,0),min(?y:Nat,?z:Nat))))) = plus(max(plus(s(plus(?x:Nat,?x:Nat)),0),plus(s(s(?z:Nat)),?x:Nat)),min(?x:Nat,s(?y:Nat)))",
"s(min(plus(plus(s(?x:Nat),s(?z:Nat)),0),0)) = s(min(min(?x:Nat,min(s(?z:Nat),s(0))),s(plus(max(?x:Nat,?z:Nat),plus(?x:Nat,?x:Nat)))))",
"max(plus(plus(s(s(0)),0),?x:Nat),s(?z:Nat)) = min(plus(s(?y:Nat),max(?x:Nat,plus(s(0),plus(?y:Nat,?z:Nat)))),s(s(s(s(0)))))",
"s(plus(?z:Nat,plus(min(?z:Nat,s(?x:Nat)),s(s(?z:Nat))))) = plus(?x:Nat,plus(s(plus(min(0,?z:Nat),s(?y:Nat))),s(max(?x:Nat,max(?z:Nat,?y:Nat)))))",
"max(?y:Nat,min(0,max(?z:Nat,plus(max(0,?y:Nat),0)))) = s(plus(max(?y:Nat,plus(plus(?z:Nat,0),?x:Nat)),max(min(?x:Nat,min(?x:Nat,0)),?x:Nat)))",
"min(max(plus(max(?z:Nat,0),?z:Nat),0),plus(?x:Nat,?x:Nat)) = max(min(?z:Nat,min(s(max(?z:Nat,0)),max(min(?x:Nat,?x:Nat),min(?x:Nat,?y:Nat)))),?z:Nat)",
"plus(s(s(s(?x:Nat))),plus(?x:Nat,max(s(?x:Nat),?z:Nat))) = max(min(s(plus(?y:Nat,s(?z:Nat))),max(s(min(0,0)),?z:Nat)),min(?x:Nat,?x:Nat))",
"plus(max(?x:Nat,s(min(s(?x:Nat),max(?x:Nat,?z:Nat)))),?y:Nat) = s(plus(s(s(plus(?x:Nat,?y:Nat))),min(0,plus(plus(?x:Nat,?y:Nat),max(?y:Nat,?x:Nat)))))",
"s(s(plus(s(?x:Nat),max(min(?x:Nat,?x:Nat),min(?y:Nat,?x:Nat))))) = s(min(plus(s(max(?y:Nat,?x:Nat)),max(s(?y:Nat),?x:Nat)),max(?x:Nat,s(0))))",
"max(?y:Nat,s(plus(min(min(?x:Nat,0),plus(?y:Nat,?y:Nat)),?x:Nat))) = plus(min(plus(?x:Nat,min(max(?y:Nat,?z:Nat),max(?x:Nat,0))),plus(?x:Nat,0)),?x:Nat)",
"max(s(min(plus(max(?x:Nat,?x:Nat),min(?z:Nat,?y:Nat)),?x:Nat)),0) = s(max(s(s(max(?x:Nat,?x:Nat))),s(max(plus(?z:Nat,?x:Nat),min(0,?y:Nat)))))",
"plus(s(plus(s(s(?y:Nat)),plus(?x:Nat,max(?z:Nat,?x:Nat)))),?y:Nat) = min(plus(plus(?z:Nat,min(min(?z:Nat,?x:Nat),?x:Nat)),s(min(s(0),?z:Nat))),?x:Nat)",
"s(min(plus(?y:Nat,s(s(?x:Nat))),min(s(plus(?y:Nat,?z:Nat)),?x:Nat))) = s(plus(max(max(max(?z:Nat,?z:Nat),?x:Nat),min(?y:Nat,max(?x:Nat,?x:Nat))),?z:Nat))",
"max(?x:Nat,plus(min(?y:Nat,max(max(?z:Nat,?x:Nat),plus(?x:Nat,?y:Nat))),0)) = s(plus(0,min(plus(?y:Nat,plus(0,0)),max(plus(?x:Nat,?y:Nat),?x:Nat))))",
"min(max(max(plus(?x:Nat,?z:Nat),plus(s(?y:Nat),s(?z:Nat))),?x:Nat),?z:Nat) = max(s(s(s(0))),s(min(?x:Nat,max(max(?x:Nat,?x:Nat),s(0)))))",
"s(plus(s(s(max(?x:Nat,0))),max(?z:Nat,plus(min(?x:Nat,?x:Nat),?x:Nat)))) = max(?y:Nat,s(min(max(min(?x:Nat,?x:Nat),?y:Nat),s(min(?x:Nat,?x:Nat)))))",
"s(min(min(min(min(0,?z:Nat),max(?y:Nat,0)),min(s(?x:Nat),?x:Nat)),?x:Nat)) = min(s(min(min(plus(?y:Nat,?y:Nat),min(0,?x:Nat)),?x:Nat)),?z:Nat)",
"max(0,plus(plus(max(min(?y:Nat,?x:Nat),plus(?z:Nat,0)),?x:Nat),plus(?z:Nat,?x:Nat))) = min(?y:Nat,plus(s(0),s(max(plus(?x:Nat,?z:Nat),s(?x:Nat)))))",
"plus(?x:Nat,min(max(s(max(?x:Nat,?x:Nat)),min(max(0,?y:Nat),s(?z:Nat))),?y:Nat)) = max(?y:Nat,plus(s(max(s(?y:Nat),plus(?x:Nat,?x:Nat))),s(?y:Nat)))",
"s(plus(s(0),plus(plus(max(?x:Nat,?x:Nat),min(?z:Nat,?x:Nat)),s(plus(?z:Nat,0))))) = max(s(max(s(min(0,0)),?x:Nat)),plus(?x:Nat,?z:Nat))",
"s(plus(plus(?x:Nat,?z:Nat),max(max(plus(?x:Nat,0),s(?z:Nat)),max(s(?x:Nat),0)))) = plus(?z:Nat,s(min(s(max(0,?x:Nat)),plus(?z:Nat,?y:Nat))))",
"max(max(0,max(min(max(?x:Nat,?z:Nat),s(?z:Nat)),s(s(?x:Nat)))),plus(?x:Nat,?z:Nat)) = s(max(?x:Nat,min(plus(s(?x:Nat),0),max(?x:Nat,?x:Nat))))",
"plus(max(?y:Nat,0),max(max(s(?x:Nat),min(plus(?z:Nat,?z:Nat),plus(?x:Nat,?x:Nat))),?y:Nat)) = plus(max(?x:Nat,s(plus(min(?x:Nat,?x:Nat),?z:Nat))),s(?y:Nat))",
"plus(plus(?x:Nat,plus(plus(?z:Nat,min(0,?x:Nat)),plus(s(0),?x:Nat))),s(s(0))) = plus(plus(plus(0,?x:Nat),s(?x:Nat)),s(s(s(?y:Nat))))",
"s(min(s(min(plus(0,?y:Nat),max(?z:Nat,?y:Nat))),max(max(min(?y:Nat,?x:Nat),?y:Nat),0))) = s(s(min(s(max(0,?y:Nat)),s(s(?y:Nat)))))",
"max(plus(s(?x:Nat),max(plus(s(?x:Nat),plus(?x:Nat,?x:Nat)),max(?x:Nat,plus(?z:Nat,0)))),0) = s(max(max(0,min(plus(?x:Nat,?z:Nat),?x:Nat)),?x:Nat))",
"min(?y:Nat,max(min(plus(s(?x:Nat),max(?x:Nat,?x:Nat)),plus(?y:Nat,?z:Nat)),s(s(s(0))))) = plus(s(min(plus(plus(?z:Nat,?x:Nat),?x:Nat),?z:Nat)),?y:Nat)",
"plus(?z:Nat,plus(plus(?x:Nat,min(plus(?x:Nat,?x:Nat),?z:Nat)),min(0,plus(?z:Nat,max(0,?y:Nat))))) = max(?x:Nat,plus(?x:Nat,plus(?z:Nat,plus(s(?x:Nat),?z:Nat))))",
"plus(min(plus(plus(?x:Nat,?y:Nat),?x:Nat),?x:Nat),min(s(plus(s(?z:Nat),s(?x:Nat))),s(?x:Nat))) = max(plus(?z:Nat,max(?y:Nat,s(max(?z:Nat,?y:Nat)))),?x:Nat)",
"plus(plus(min(max(max(0,?y:Nat),s(?z:Nat)),max(?z:Nat,plus(?x:Nat,?x:Nat))),?x:Nat),max(?x:Nat,0)) = plus(s(min(min(?x:Nat,s(?y:Nat)),?y:Nat)),?z:Nat)",
"plus(?x:Nat,max(min(max(max(?x:Nat,?y:Nat),plus(?x:Nat,0)),max(max(?y:Nat,?x:Nat),?z:Nat)),max(?y:Nat,?y:Nat))) = s(plus(0,min(plus(?x:Nat,?x:Nat),0)))",
"plus(max(s(?z:Nat),plus(?z:Nat,max(0,s(?x:Nat)))),s(plus(plus(plus(?x:Nat,0),s(?z:Nat)),0))) = max(0,min(max(0,?z:Nat),s(0)))",
"plus(s(plus(s(plus(?z:Nat,?x:Nat)),plus(0,max(?z:Nat,?z:Nat)))),max(min(?x:Nat,s(plus(?x:Nat,?z:Nat))),?z:Nat)) = plus(?z:Nat,max(max(?x:Nat,?x:Nat),?x:Nat))",
"plus(min(s(?z:Nat),max(plus(max(0,?x:Nat),?x:Nat),0)),plus(max(max(max(?x:Nat,?x:Nat),?y:Nat),?x:Nat),?y:Nat)) = plus(max(s(s(?x:Nat)),?y:Nat),?x:Nat)",
"min(s(s(plus(plus(?x:Nat,?x:Nat),min(0,0)))),min(max(0,s(plus(0,?x:Nat))),plus(s(?x:Nat),?x:Nat))) = min(min(s(?x:Nat),?z:Nat),?x:Nat)",
"min(min(plus(plus(?y:Nat,min(?z:Nat,?x:Nat)),min(s(?y:Nat),?z:Nat)),plus(plus(min(?z:Nat,?z:Nat),?z:Nat),s(?x:Nat))),?y:Nat) = s(plus(?z:Nat,s(s(?z:Nat))))",
"s(plus(plus(max(s(0),?y:Nat),min(min(0,?x:Nat),plus(?z:Nat,?x:Nat))),s(plus(min(?x:Nat,?x:Nat),plus(?x:Nat,?y:Nat))))) = min(plus(?x:Nat,0),?x:Nat)",
"max(0,plus(min(s(max(0,?x:Nat)),max(plus(?x:Nat,?x:Nat),0)),plus(?x:Nat,plus(min(?y:Nat,?x:Nat),min(?x:Nat,?x:Nat))))) = min(min(?x:Nat,0),0)",
"max(plus(min(?z:Nat,s(?x:Nat)),?z:Nat),plus(s(0),plus(plus(min(?y:Nat,?x:Nat),plus(?z:Nat,?x:Nat)),s(max(0,?x:Nat))))) = min(?x:Nat,max(?y:Nat,0))",
"plus(s(min(s(0),min(plus(?x:Nat,?x:Nat),?x:Nat))),s(plus(min(s(0),?x:Nat),plus(s(0),plus(?z:Nat,?y:Nat))))) = min(?x:Nat,plus(?x:Nat,?y:Nat))",
"plus(min(min(plus(s(?x:Nat),s(0)),plus(min(0,?x:Nat),?z:Nat)),min(plus(?x:Nat,?x:Nat),?y:Nat)),plus(?x:Nat,s(?x:Nat))) = plus(?z:Nat,min(0,?x:Nat))",
"min(s(s(s(plus(0,?y:Nat)))),s(min(max(max(?x:Nat,?z:Nat),min(?x:Nat,?x:Nat)),min(plus(?x:Nat,?y:Nat),max(0,?y:Nat))))) = min(?z:Nat,s(?y:Nat))",
"s(plus(plus(max(min(?x:Nat,?y:Nat),?y:Nat),?y:Nat),plus(plus(max(0,?y:Nat),plus(?z:Nat,?z:Nat)),min(plus(?z:Nat,0),plus(0,?x:Nat))))) = min(?x:Nat,0)",
"s(plus(plus(min(min(0,?z:Nat),?x:Nat),min(s(?y:Nat),plus(?x:Nat,?x:Nat))),plus(max(s(?y:Nat),plus(?y:Nat,?x:Nat)),s(s(?z:Nat))))) = plus(0,0)",
"min(s(plus(min(?y:Nat,max(?y:Nat,?x:Nat)),min(plus(?x:Nat,?x:Nat),min(?x:Nat,?x:Nat)))),plus(s(?y:Nat),max(s(plus(0,?z:Nat)),?y:Nat))) = min(?y:Nat,?x:Nat)",
"min(min(?z:Nat,min(0,min(max(?x:Nat,?x:Nat),max(?x:Nat,?x:Nat)))),plus(s(min(plus(?z:Nat,0),max(?x:Nat,?x:Nat))),max(?z:Nat,?y:Nat))) = s(s(?z:Nat))",
"max(plus(min(plus(min(?y:Nat,?x:Nat),?x:Nat),plus(s(?z:Nat),?x:Nat)),plus(0,plus(plus(?x:Nat,0),max(?z:Nat,0)))),min(0,s(?y:Nat))) = s(?x:Nat)",
"plus(?x:Nat,plus(min(plus(?x:Nat,plus(?x:Nat,?x:Nat)),min(min(?x:Nat,?x:Nat),max(?y:Nat,?x:Nat))),max(plus(min(?y:Nat,?z:Nat),?x:Nat),plus(?z:Nat,?y:Nat)))) = s(?y:Nat)",
"plus(max(s(plus(plus(0,?y:Nat),plus(?x:Nat,?y:Nat))),max(min(plus(?x:Nat,?y:Nat),min(?z:Nat,?z:Nat)),max(s(?y:Nat),0))),plus(?y:Nat,?x:Nat)) = s(0)",
"s(?y:Nat) = min(s(plus(max(min(?x:Nat,?x:Nat),plus(?z:Nat,?z:Nat)),min(min(?z:Nat,0),?x:Nat))),plus(plus(s(?x:Nat),max(min(?z:Nat,?x:Nat),s(?x:Nat))),?x:Nat))",
"min(0,?x:Nat) = min(min(max(plus(?z:Nat,max(0,?x:Nat)),plus(s(?x:Nat),?x:Nat)),max(plus(max(0,?z:Nat),s(?x:Nat)),max(?z:Nat,plus(?x:Nat,?x:Nat)))),?z:Nat)",
"plus(?y:Nat,0) = min(plus(max(?z:Nat,plus(plus(?z:Nat,?z:Nat),plus(?x:Nat,?x:Nat))),plus(plus(min(?y:Nat,?z:Nat),?x:Nat),0)),s(max(s(plus(?x:Nat,?x:Nat)),?z:Nat)))",
"plus(0,?x:Nat) = min(plus(max(?z:Nat,?x:Nat),s(plus(s(?x:Nat),min(?z:Nat,?y:Nat)))),s(max(max(plus(0,?x:Nat),plus(?x:Nat,?x:Nat)),min(s(0),0))))",
"plus(?x:Nat,s(?z:Nat)) = s(plus(min(plus(min(?x:Nat,0),plus(?y:Nat,?z:Nat)),?x:Nat),max(max(?x:Nat,plus(?y:Nat,?x:Nat)),max(max(?x:Nat,?z:Nat),plus(?z:Nat,?x:Nat)))))",
"plus(min(?x:Nat,?y:Nat),?z:Nat) = s(max(min(max(min(?x:Nat,?y:Nat),s(?z:Nat)),?y:Nat),min(max(plus(?x:Nat,?x:Nat),s(0)),max(s(?x:Nat),max(?x:Nat,0)))))",
"min(0,max(?x:Nat,s(?x:Nat))) = plus(plus(max(0,s(?x:Nat)),s(s(plus(?x:Nat,?z:Nat)))),max(?x:Nat,min(?x:Nat,plus(plus(?x:Nat,?y:Nat),max(?z:Nat,?x:Nat)))))",
"min(max(plus(?x:Nat,?y:Nat),?x:Nat),?x:Nat) = min(s(s(max(s(?x:Nat),plus(0,?x:Nat)))),s(min(max(s(?z:Nat),plus(?x:Nat,?x:Nat)),plus(?z:Nat,s(0)))))",
"s(max(0,plus(s(?y:Nat),s(?x:Nat)))) = plus(plus(plus(plus(s(?x:Nat),min(?x:Nat,?x:Nat)),?x:Nat),?x:Nat),max(?y:Nat,plus(0,plus(max(?x:Nat,0),?z:Nat))))",
"s(max(min(?y:Nat,0),max(0,?z:Nat))) = min(s(s(plus(0,min(?x:Nat,?x:Nat)))),max(plus(plus(?x:Nat,max(?x:Nat,?z:Nat)),s(s(?y:Nat))),s(?x:Nat)))",
"s(plus(plus(?y:Nat,?x:Nat),max(0,?x:Nat))) = max(s(?z:Nat),max(plus(s(plus(?z:Nat,?x:Nat)),plus(?x:Nat,min(?y:Nat,?y:Nat))),max(0,s(min(0,?y:Nat)))))",
"max(?z:Nat,plus(s(s(s(?y:Nat))),?y:Nat)) = max(max(s(s(min(?y:Nat,0))),s(min(s(?x:Nat),s(0)))),max(s(plus(?y:Nat,?y:Nat)),s(?z:Nat)))",
"max(s(plus(?x:Nat,s(s(?x:Nat)))),?y:Nat) = min(min(plus(?z:Nat,?y:Nat),min(plus(?z:Nat,min(?x:Nat,0)),max(s(0),min(?y:Nat,?x:Nat)))),plus(0,?x:Nat))",
"min(s(s(plus(0,s(0)))),?z:Nat) = max(min(min(?z:Nat,plus(plus(?z:Nat,?y:Nat),?y:Nat)),?y:Nat),min(plus(?x:Nat,?y:Nat),plus(min(0,s(?x:Nat)),?x:Nat)))",
"plus(?x:Nat,max(s(s(0)),s(0))) = plus(plus(max(plus(max(?y:Nat,?x:Nat),?x:Nat),s(plus(?x:Nat,?x:Nat))),s(?x:Nat)),min(min(s(0),?x:Nat),?x:Nat))",
"max(min(s(plus(?x:Nat,s(?x:Nat))),?x:Nat),?z:Nat) = min(min(s(?x:Nat),min(plus(?z:Nat,plus(0,?z:Nat)),?z:Nat)),max(?z:Nat,s(plus(?x:Nat,max(?y:Nat,?z:Nat)))))",
"min(?y:Nat,min(0,s(plus(s(?y:Nat),?y:Nat)))) = s(plus(s(s(max(?x:Nat,?y:Nat))),min(max(max(?z:Nat,?y:Nat),max(?x:Nat,0)),s(max(?y:Nat,?z:Nat)))))",
"s(min(s(?z:Nat),plus(min(?y:Nat,?x:Nat),s(s(?y:Nat))))) = min(max(s(s(0)),s(s(max(?y:Nat,?x:Nat)))),s(s(min(?x:Nat,plus(?x:Nat,0)))))",
"s(plus(?y:Nat,plus(min(0,s(?x:Nat)),s(s(?z:Nat))))) = max(?y:Nat,max(s(max(s(?x:Nat),?z:Nat)),plus(?y:Nat,max(max(?z:Nat,?x:Nat),min(?z:Nat,?y:Nat)))))",
"s(plus(s(max(plus(?x:Nat,?x:Nat),max(?x:Nat,?z:Nat))),?x:Nat)) = s(min(max(?z:Nat,max(s(?z:Nat),min(?x:Nat,?y:Nat))),plus(?z:Nat,plus(plus(0,?x:Nat),?y:Nat))))",
"plus(?z:Nat,s(min(max(plus(?z:Nat,?y:Nat),0),s(0)))) = max(?x:Nat,min(s(max(plus(?y:Nat,?x:Nat),?y:Nat)),plus(0,max(s(0),max(?x:Nat,?z:Nat)))))",
"plus(plus(?x:Nat,max(plus(?x:Nat,max(?x:Nat,?y:Nat)),?z:Nat)),?x:Nat) = s(plus(min(?x:Nat,plus(plus(?z:Nat,?z:Nat),s(?x:Nat))),plus(?x:Nat,plus(?y:Nat,max(?x:Nat,?x:Nat)))))",
"s(plus(0,max(?x:Nat,max(max(0,?x:Nat),max(?x:Nat,0))))) = min(s(0),max(plus(0,min(?x:Nat,min(?y:Nat,?x:Nat))),min(?y:Nat,plus(?x:Nat,?x:Nat))))",
"max(max(s(s(plus(?z:Nat,?x:Nat))),s(s(?x:Nat))),s(0)) = plus(plus(?x:Nat,?z:Nat),min(max(min(?x:Nat,?z:Nat),min(s(?x:Nat),?z:Nat)),plus(?y:Nat,?y:Nat)))",
"max(min(s(s(?x:Nat)),?y:Nat),plus(plus(?z:Nat,?x:Nat),s(?y:Nat))) = plus(?y:Nat,max(plus(max(plus(?x:Nat,?z:Nat),0),?z:Nat),plus(s(plus(?z:Nat,?x:Nat)),?x:Nat)))",
"min(plus(min(min(plus(?z:Nat,?x:Nat),0),0),s(?z:Nat)),?z:Nat) = plus(min(max(plus(max(?x:Nat,?y:Nat),?x:Nat),?z:Nat),min(s(plus(0,?x:Nat)),?z:Nat)),?y:Nat)",
"max(max(?y:Nat,plus(?z:Nat,min(min(?x:Nat,0),plus(?y:Nat,?x:Nat)))),0) = plus(?x:Nat,plus(s(max(plus(?z:Nat,?x:Nat),s(?z:Nat))),s(s(max(?x:Nat,?x:Nat)))))",
"min(s(?z:Nat),max(max(max(?z:Nat,?x:Nat),plus(s(?x:Nat),?x:Nat)),0)) = s(min(min(s(plus(?x:Nat,?x:Nat)),plus(plus(?y:Nat,0),plus(?x:Nat,?z:Nat))),0))",
"min(max(0,?y:Nat),plus(s(0),s(max(?z:Nat,plus(?x:Nat,?y:Nat))))) = plus(?z:Nat,s(plus(max(plus(?x:Nat,?z:Nat),min(?x:Nat,0)),s(max(?x:Nat,?z:Nat)))))",
"min(plus(plus(?y:Nat,plus(s(?z:Nat),s(?x:Nat))),s(plus(?z:Nat,s(?x:Nat)))),?x:Nat) = s(min(max(min(min(?y:Nat,0),plus(?y:Nat,0)),s(?x:Nat)),?x:Nat))",
"min(max(?y:Nat,min(s(min(?y:Nat,0)),?x:Nat)),s(min(?x:Nat,min(s(?x:Nat),0)))) = s(min(?x:Nat,max(s(plus(?z:Nat,?x:Nat)),plus(?x:Nat,s(?y:Nat)))))",
"plus(s(?x:Nat),max(max(s(max(?x:Nat,?x:Nat)),?x:Nat),s(min(?y:Nat,min(?y:Nat,?z:Nat))))) = plus(?z:Nat,plus(s(s(s(?x:Nat))),plus(s(?z:Nat),s(0))))",
"s(min(plus(max(?z:Nat,plus(?z:Nat,0)),?z:Nat),plus(plus(plus(?y:Nat,0),s(?z:Nat)),0))) = plus(min(s(0),s(plus(s(?x:Nat),s(?x:Nat)))),0)",
"s(plus(?z:Nat,min(max(min(?z:Nat,?x:Nat),max(?y:Nat,?y:Nat)),min(min(?z:Nat,?y:Nat),s(?y:Nat))))) = max(s(min(?z:Nat,plus(s(0),min(?z:Nat,?x:Nat)))),?x:Nat)",
"max(plus(min(s(s(?x:Nat)),?z:Nat),min(s(max(?z:Nat,?y:Nat)),s(max(?x:Nat,?x:Nat)))),?y:Nat) = plus(min(plus(s(0),s(plus(?x:Nat,?z:Nat))),?x:Nat),?z:Nat)",
"s(max(max(plus(s(?y:Nat),min(?x:Nat,?x:Nat)),plus(max(?z:Nat,?x:Nat),min(?x:Nat,?y:Nat))),s(0))) = max(s(?x:Nat),plus(?y:Nat,s(s(max(?x:Nat,?x:Nat)))))",
"max(?z:Nat,max(max(0,max(s(?z:Nat),plus(?x:Nat,?x:Nat))),min(plus(min(?x:Nat,?y:Nat),0),?y:Nat))) = plus(?z:Nat,min(plus(?y:Nat,s(max(?z:Nat,0))),?y:Nat))",
"plus(min(?x:Nat,?x:Nat),min(plus(min(?z:Nat,max(0,0)),s(min(?x:Nat,?y:Nat))),min(0,?x:Nat))) = max(s(?x:Nat),s(max(min(?z:Nat,s(0)),?z:Nat)))",
"plus(min(?z:Nat,plus(s(?x:Nat),?z:Nat)),plus(?x:Nat,plus(min(?x:Nat,s(0)),plus(plus(?y:Nat,?x:Nat),0)))) = s(s(s(max(s(?x:Nat),plus(0,?x:Nat)))))",
"plus(plus(s(plus(?y:Nat,0)),max(?x:Nat,plus(0,max(?x:Nat,?y:Nat)))),s(s(s(max(?y:Nat,?y:Nat))))) = s(max(?y:Nat,max(?x:Nat,plus(?y:Nat,s(?x:Nat)))))",
"min(?y:Nat,plus(min(s(plus(?x:Nat,?z:Nat)),?y:Nat),min(max(max(?x:Nat,?z:Nat),s(?x:Nat)),min(plus(?y:Nat,?y:Nat),?y:Nat)))) = plus(s(min(0,?y:Nat)),s(?z:Nat))",
"plus(?y:Nat,max(plus(0,min(max(?x:Nat,0),s(?x:Nat))),min(s(?x:Nat),max(max(?x:Nat,?y:Nat),max(?x:Nat,?z:Nat))))) = plus(s(s(s(s(?x:Nat)))),0)",
"plus(?z:Nat,max(min(plus(min(?x:Nat,0),?z:Nat),min(?x:Nat,s(?x:Nat))),plus(plus(?x:Nat,s(?z:Nat)),s(min(0,?x:Nat))))) = max(s(?x:Nat),plus(?x:Nat,?z:Nat))",
"max(min(s(?x:Nat),0),plus(max(min(?x:Nat,0),min(max(0,?y:Nat),s(?y:Nat))),max(?x:Nat,plus(?x:Nat,min(?x:Nat,?x:Nat))))) = plus(0,s(s(?x:Nat)))",
"max(s(min(min(?x:Nat,s(0)),plus(min(?x:Nat,?x:Nat),?x:Nat))),max(?z:Nat,max(min(s(?y:Nat),?z:Nat),plus(?x:Nat,plus(?y:Nat,0))))) = s(max(?y:Nat,0))",
"min(?y:Nat,max(min(plus(plus(?z:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),?x:Nat),plus(plus(min(?x:Nat,?x:Nat),min(0,?y:Nat)),max(s(0),?x:Nat)))) = s(plus(?z:Nat,?y:Nat))",
"min(max(min(s(0),plus(s(?x:Nat),max(?x:Nat,?z:Nat))),?x:Nat),s(plus(s(max(?x:Nat,?x:Nat)),min(plus(0,?y:Nat),s(?x:Nat))))) = s(plus(?z:Nat,?y:Nat))",
"plus(plus(max(max(?y:Nat,plus(?x:Nat,0)),?x:Nat),s(?x:Nat)),plus(min(0,?x:Nat),min(plus(max(0,?x:Nat),min(?z:Nat,?x:Nat)),?y:Nat))) = s(s(s(?z:Nat)))",
"min(min(0,min(s(plus(?x:Nat,0)),s(s(?x:Nat)))),s(max(plus(plus(?x:Nat,?x:Nat),plus(?y:Nat,?z:Nat)),plus(plus(?y:Nat,?z:Nat),?z:Nat)))) = s(s(?x:Nat))",
"plus(max(max(plus(max(?x:Nat,?z:Nat),min(?x:Nat,?y:Nat)),plus(s(?y:Nat),s(?y:Nat))),min(plus(min(0,?x:Nat),max(?y:Nat,?x:Nat)),?y:Nat)),?z:Nat) = plus(0,?y:Nat)",
"max(max(min(s(s(?x:Nat)),max(?z:Nat,?x:Nat)),max(plus(max(?x:Nat,?y:Nat),?z:Nat),?x:Nat)),plus(s(s(s(?x:Nat))),max(s(?x:Nat),s(?y:Nat)))) = s(?y:Nat)",
"min(min(0,s(?z:Nat)),max(plus(s(?y:Nat),plus(?y:Nat,s(?z:Nat))),plus(min(min(?x:Nat,?x:Nat),?x:Nat),max(plus(?y:Nat,0),plus(?x:Nat,?z:Nat))))) = s(?y:Nat)",
"plus(min(plus(?x:Nat,?z:Nat),max(s(0),s(?x:Nat))),min(plus(plus(plus(0,0),s(0)),plus(?x:Nat,min(?y:Nat,?y:Nat))),min(?x:Nat,?x:Nat))) = s(0)",
"max(?y:Nat,?x:Nat) = s(max(plus(max(min(?y:Nat,?z:Nat),s(?z:Nat)),s(plus(0,?z:Nat))),plus(min(plus(?z:Nat,?z:Nat),min(0,?x:Nat)),max(plus(?x:Nat,?y:Nat),?x:Nat))))",
"max(0,?z:Nat) = max(plus(max(?x:Nat,min(s(0),plus(?x:Nat,?x:Nat))),s(min(min(0,0),plus(?y:Nat,?y:Nat)))),s(max(?y:Nat,plus(min(?y:Nat,?x:Nat),?x:Nat))))",
"min(?x:Nat,?z:Nat) = min(s(plus(s(?x:Nat),0)),plus(min(max(s(?x:Nat),0),max(min(?x:Nat,?y:Nat),plus(0,0))),s(plus(min(0,?z:Nat),s(?z:Nat)))))",
"s(min(?z:Nat,?z:Nat)) = max(max(?x:Nat,?y:Nat),plus(plus(max(s(0),0),min(0,?y:Nat)),plus(min(s(?x:Nat),plus(0,0)),min(0,max(?z:Nat,0)))))",
"s(min(0,?x:Nat)) = min(max(s(plus(?x:Nat,plus(?x:Nat,?x:Nat))),max(s(min(?y:Nat,?x:Nat)),min(s(0),0))),s(s(min(max(?x:Nat,0),s(0)))))",
"s(plus(plus(?x:Nat,?x:Nat),0)) = max(s(max(s(plus(?y:Nat,0)),plus(min(?x:Nat,?y:Nat),s(?x:Nat)))),plus(plus(max(plus(?x:Nat,?x:Nat),s(?x:Nat)),?x:Nat),?z:Nat))",
"plus(s(s(plus(?x:Nat,?x:Nat))),?z:Nat) = max(0,plus(min(min(s(0),plus(?x:Nat,?z:Nat)),?y:Nat),min(min(0,min(?z:Nat,?x:Nat)),max(min(?y:Nat,?x:Nat),?x:Nat))))",
"s(s(max(?y:Nat,s(plus(?x:Nat,?z:Nat))))) = max(max(s(max(?x:Nat,s(0))),plus(plus(plus(?z:Nat,?y:Nat),?z:Nat),?z:Nat)),max(s(s(0)),plus(?z:Nat,?x:Nat)))",
"s(plus(0,min(0,min(?x:Nat,?x:Nat)))) = min(min(?x:Nat,plus(s(min(?z:Nat,?x:Nat)),?y:Nat)),min(?x:Nat,plus(plus(s(0),?z:Nat),plus(min(0,0),?z:Nat))))",
"max(?x:Nat,s(plus(?x:Nat,max(0,s(?y:Nat))))) = s(plus(min(plus(min(0,?x:Nat),min(?x:Nat,?x:Nat)),s(plus(?y:Nat,?y:Nat))),plus(plus(s(0),?y:Nat),?z:Nat)))",
"max(s(s(max(s(0),s(?x:Nat)))),?x:Nat) = s(max(plus(s(?x:Nat),plus(max(?x:Nat,0),0)),min(s(?x:Nat),plus(plus(0,0),max(?x:Nat,?x:Nat)))))",
"min(s(max(s(max(?x:Nat,?x:Nat)),?x:Nat)),?x:Nat) = s(max(plus(?z:Nat,plus(?z:Nat,plus(?z:Nat,0))),plus(min(?x:Nat,?x:Nat),plus(max(0,?z:Nat),min(?y:Nat,?x:Nat)))))",
"s(s(max(s(?x:Nat),plus(min(?z:Nat,?x:Nat),?x:Nat)))) = plus(0,plus(?y:Nat,max(max(plus(?y:Nat,?z:Nat),plus(0,?z:Nat)),plus(min(?x:Nat,?y:Nat),max(?y:Nat,?y:Nat)))))",
"s(plus(s(s(?y:Nat)),s(min(s(?y:Nat),?y:Nat)))) = max(max(min(s(plus(?y:Nat,0)),?x:Nat),max(s(?z:Nat),plus(min(?x:Nat,?x:Nat),plus(?y:Nat,?x:Nat)))),?x:Nat)",
"plus(plus(s(min(?x:Nat,0)),max(?x:Nat,?x:Nat)),?x:Nat) = plus(plus(?z:Nat,plus(plus(min(0,?x:Nat),plus(?z:Nat,0)),?x:Nat)),s(s(min(plus(0,?x:Nat),0))))",
"s(s(plus(s(plus(?z:Nat,?x:Nat)),plus(0,s(0))))) = plus(plus(plus(min(min(?y:Nat,?x:Nat),s(?x:Nat)),?z:Nat),?x:Nat),s(max(?y:Nat,min(?z:Nat,s(?x:Nat)))))",
"s(s(plus(plus(max(?x:Nat,?z:Nat),max(?y:Nat,?x:Nat)),?x:Nat))) = s(min(max(plus(0,?x:Nat),?z:Nat),min(plus(0,plus(?x:Nat,?x:Nat)),plus(plus(?y:Nat,?x:Nat),?x:Nat))))",
"min(?z:Nat,min(s(?x:Nat),s(min(0,min(?x:Nat,?x:Nat))))) = s(max(plus(min(s(?z:Nat),?x:Nat),plus(?x:Nat,0)),max(s(plus(?x:Nat,?y:Nat)),plus(0,?z:Nat))))",
"min(0,max(s(plus(s(?x:Nat),max(0,0))),?x:Nat)) = s(plus(max(?z:Nat,max(max(?x:Nat,?x:Nat),plus(?z:Nat,?y:Nat))),plus(min(s(?x:Nat),s(?z:Nat)),?x:Nat)))",
"plus(min(?z:Nat,plus(0,plus(plus(?y:Nat,0),0))),?z:Nat) = max(?y:Nat,max(max(plus(min(0,?x:Nat),plus(?x:Nat,?z:Nat)),max(min(?y:Nat,?x:Nat),s(0))),?z:Nat))",
"plus(min(min(min(?y:Nat,s(0)),0),s(?y:Nat)),?y:Nat) = max(plus(plus(s(plus(?y:Nat,?x:Nat)),plus(max(?x:Nat,?y:Nat),?x:Nat)),0),s(s(plus(?z:Nat,?x:Nat))))",
"s(s(plus(s(max(?y:Nat,?x:Nat)),min(?y:Nat,max(?x:Nat,?y:Nat))))) = min(min(?z:Nat,min(?x:Nat,?z:Nat)),plus(?x:Nat,min(?x:Nat,max(max(?z:Nat,?x:Nat),max(?x:Nat,?x:Nat)))))",
"max(s(max(plus(?x:Nat,s(?y:Nat)),max(min(?x:Nat,?x:Nat),0))),?x:Nat) = max(max(s(plus(?x:Nat,max(?y:Nat,?y:Nat))),max(?y:Nat,min(min(?y:Nat,?x:Nat),?x:Nat))),?x:Nat)",
"plus(plus(?z:Nat,s(s(min(0,?x:Nat)))),plus(min(0,?x:Nat),0)) = max(max(s(?x:Nat),min(s(s(?z:Nat)),max(s(?y:Nat),?x:Nat))),s(max(?y:Nat,?x:Nat)))",
"max(?x:Nat,plus(min(s(max(0,?z:Nat)),plus(?z:Nat,?x:Nat)),plus(?x:Nat,?z:Nat))) = min(s(?z:Nat),plus(max(0,?x:Nat),min(max(s(?y:Nat),max(?x:Nat,?y:Nat)),0)))",
"min(max(?x:Nat,?x:Nat),s(plus(s(plus(?x:Nat,0)),plus(?x:Nat,s(?x:Nat))))) = plus(s(s(plus(min(0,?x:Nat),s(?y:Nat)))),plus(min(?y:Nat,s(0)),?x:Nat))",
"max(?x:Nat,plus(0,min(max(max(?y:Nat,?y:Nat),max(?x:Nat,?x:Nat)),s(s(?y:Nat))))) = s(plus(plus(max(?x:Nat,s(?z:Nat)),max(?z:Nat,max(0,?y:Nat))),s(?z:Nat)))",
"max(?x:Nat,s(min(max(min(?x:Nat,?z:Nat),max(?x:Nat,?x:Nat)),plus(min(?z:Nat,?z:Nat),?z:Nat)))) = plus(0,plus(max(max(s(?x:Nat),?x:Nat),plus(s(0),?x:Nat)),?x:Nat))",
"max(0,min(plus(plus(plus(0,?y:Nat),max(?x:Nat,?z:Nat)),s(plus(?x:Nat,?x:Nat))),?z:Nat)) = max(max(plus(s(plus(?x:Nat,?x:Nat)),?x:Nat),?x:Nat),plus(?y:Nat,s(?x:Nat)))",
"min(?y:Nat,max(max(s(s(?y:Nat)),?y:Nat),max(min(max(?x:Nat,?x:Nat),s(?y:Nat)),?y:Nat))) = s(plus(plus(min(min(?x:Nat,?y:Nat),?y:Nat),s(0)),s(s(?z:Nat))))",
"plus(s(s(min(0,s(?z:Nat)))),max(?y:Nat,plus(min(s(?x:Nat),?y:Nat),s(?x:Nat)))) = s(max(plus(s(plus(?z:Nat,0)),s(min(?z:Nat,0))),s(?x:Nat)))",
"s(max(plus(plus(?x:Nat,s(?x:Nat)),plus(plus(?x:Nat,?z:Nat),?x:Nat)),plus(0,s(s(?x:Nat))))) = plus(max(plus(?x:Nat,max(?y:Nat,plus(?z:Nat,0))),s(?x:Nat)),?x:Nat)",
"s(min(plus(?z:Nat,max(?y:Nat,plus(?x:Nat,?x:Nat))),min(plus(s(?x:Nat),plus(?x:Nat,?x:Nat)),?x:Nat))) = plus(s(s(?y:Nat)),min(plus(?y:Nat,s(min(?y:Nat,?y:Nat))),?x:Nat))",
"s(min(plus(max(plus(?x:Nat,?z:Nat),s(?y:Nat)),max(plus(?x:Nat,?x:Nat),?x:Nat)),s(s(0)))) = s(s(min(s(max(?x:Nat,?x:Nat)),plus(plus(?z:Nat,?y:Nat),?x:Nat))))",
"s(plus(?x:Nat,max(plus(plus(?x:Nat,?z:Nat),max(?x:Nat,?x:Nat)),plus(s(0),plus(?y:Nat,?z:Nat))))) = plus(max(plus(s(0),s(?y:Nat)),?y:Nat),s(s(s(?x:Nat))))",
"min(s(min(s(?y:Nat),s(?x:Nat))),s(min(max(min(?y:Nat,?x:Nat),max(?x:Nat,0)),?y:Nat))) = s(min(s(0),min(plus(s(0),max(?y:Nat,?y:Nat)),?x:Nat)))",
"min(max(plus(max(plus(?y:Nat,?z:Nat),0),s(s(?x:Nat))),max(min(?x:Nat,0),?y:Nat)),?z:Nat) = s(plus(s(s(plus(?z:Nat,?y:Nat))),plus(0,min(?x:Nat,?y:Nat))))",
"min(max(plus(plus(?y:Nat,s(0)),min(?y:Nat,0)),?y:Nat),plus(?y:Nat,plus(s(?z:Nat),?y:Nat))) = s(max(0,min(max(s(0),min(?z:Nat,?y:Nat)),s(?x:Nat))))",
"plus(?x:Nat,min(plus(min(?x:Nat,s(?x:Nat)),?z:Nat),min(min(?z:Nat,?z:Nat),max(s(?x:Nat),?x:Nat)))) = max(min(max(min(?y:Nat,s(?z:Nat)),?z:Nat),?x:Nat),min(0,?z:Nat))",
"plus(?y:Nat,plus(plus(?x:Nat,max(0,min(?z:Nat,?y:Nat))),plus(?y:Nat,plus(s(?x:Nat),s(?x:Nat))))) = s(plus(plus(?y:Nat,?x:Nat),min(max(?x:Nat,max(0,?y:Nat)),?y:Nat)))",
"plus(min(?y:Nat,?y:Nat),plus(?x:Nat,max(max(s(?x:Nat),max(?z:Nat,?y:Nat)),max(?x:Nat,s(?y:Nat))))) = plus(max(?y:Nat,plus(plus(plus(?x:Nat,?x:Nat),s(?y:Nat)),?x:Nat)),?x:Nat)",
"plus(min(plus(0,s(min(?x:Nat,?x:Nat))),0),min(?z:Nat,max(?x:Nat,s(plus(?x:Nat,0))))) = s(max(plus(min(0,s(?z:Nat)),s(min(?z:Nat,?z:Nat))),?x:Nat))",
"min(?y:Nat,plus(s(plus(plus(?y:Nat,?x:Nat),max(0,?x:Nat))),s(max(min(?z:Nat,0),s(?x:Nat))))) = min(?x:Nat,plus(plus(?z:Nat,min(?z:Nat,max(?x:Nat,?x:Nat))),?x:Nat))",
"plus(?x:Nat,plus(min(plus(?z:Nat,?x:Nat),?x:Nat),max(min(s(0),plus(?y:Nat,?x:Nat)),min(?x:Nat,?x:Nat)))) = plus(plus(?z:Nat,?y:Nat),s(max(?x:Nat,s(min(?z:Nat,?y:Nat)))))",
"plus(s(max(max(min(?x:Nat,?y:Nat),min(?y:Nat,?y:Nat)),max(max(?x:Nat,?x:Nat),plus(?y:Nat,?z:Nat)))),?z:Nat) = plus(s(plus(?y:Nat,s(plus(?x:Nat,0)))),min(?z:Nat,0))",
"plus(max(plus(plus(?z:Nat,max(?z:Nat,?x:Nat)),max(max(?x:Nat,?x:Nat),min(?x:Nat,0))),s(?x:Nat)),s(?y:Nat)) = plus(plus(?y:Nat,s(s(s(?y:Nat)))),min(?x:Nat,?x:Nat))",
"min(s(s(?x:Nat)),min(max(s(min(?x:Nat,?x:Nat)),0),plus(s(min(?z:Nat,?x:Nat)),min(s(?y:Nat),0)))) = s(max(0,s(min(?x:Nat,plus(?z:Nat,?x:Nat)))))",
"plus(max(max(0,plus(?x:Nat,?z:Nat)),min(plus(plus(?z:Nat,?y:Nat),s(?y:Nat)),max(s(?y:Nat),?y:Nat))),s(0)) = s(max(s(?z:Nat),min(plus(?y:Nat,?x:Nat),?y:Nat)))",
"plus(plus(min(plus(?x:Nat,0),s(s(0))),max(min(max(?x:Nat,?x:Nat),s(?z:Nat)),min(0,?x:Nat))),?y:Nat) = max(plus(min(0,s(?z:Nat)),?y:Nat),s(?z:Nat))",
"plus(?x:Nat,plus(plus(plus(s(?x:Nat),max(?z:Nat,?x:Nat)),min(plus(?x:Nat,?x:Nat),max(?y:Nat,0))),s(s(s(?y:Nat))))) = s(s(s(plus(0,min(?x:Nat,?x:Nat)))))",
"plus(max(max(min(0,min(?y:Nat,?z:Nat)),?y:Nat),?x:Nat),s(max(min(plus(0,?x:Nat),plus(?x:Nat,?x:Nat)),s(?x:Nat)))) = s(max(max(0,s(s(0))),0))",
"min(min(max(?z:Nat,s(plus(?x:Nat,?z:Nat))),0),min(max(?x:Nat,max(s(0),0)),plus(?x:Nat,min(?x:Nat,s(?y:Nat))))) = s(plus(?x:Nat,s(min(?y:Nat,?x:Nat))))",
"plus(s(s(plus(plus(?x:Nat,?y:Nat),min(?z:Nat,?x:Nat)))),plus(?z:Nat,plus(max(plus(?y:Nat,?x:Nat),?z:Nat),s(plus(?x:Nat,?x:Nat))))) = plus(plus(?x:Nat,s(?y:Nat)),s(?x:Nat))",
"max(max(?z:Nat,max(plus(?y:Nat,s(?x:Nat)),min(plus(?y:Nat,?x:Nat),min(?y:Nat,?x:Nat)))),plus(max(s(max(?x:Nat,?x:Nat)),s(?x:Nat)),s(?y:Nat))) = plus(0,s(?x:Nat))",
"max(max(max(?x:Nat,min(?x:Nat,0)),max(min(s(?x:Nat),plus(?y:Nat,0)),min(max(0,?x:Nat),max(?z:Nat,?z:Nat)))),min(?y:Nat,s(?x:Nat))) = s(plus(?y:Nat,0))",
"max(min(s(min(s(?y:Nat),plus(?x:Nat,?x:Nat))),min(s(min(?x:Nat,0)),s(?y:Nat))),plus(max(max(plus(?y:Nat,?x:Nat),?y:Nat),?z:Nat),?x:Nat)) = plus(s(0),?x:Nat)",
"plus(max(?x:Nat,max(max(?x:Nat,min(?z:Nat,?x:Nat)),?x:Nat)),min(max(plus(?z:Nat,plus(?x:Nat,0)),plus(?y:Nat,max(?x:Nat,?y:Nat))),max(?x:Nat,?y:Nat))) = min(?x:Nat,s(?z:Nat))",
"plus(max(s(plus(plus(?x:Nat,?x:Nat),s(0))),?y:Nat),min(max(?y:Nat,s(min(?x:Nat,0))),plus(s(min(?y:Nat,?y:Nat)),s(s(0))))) = min(s(?x:Nat),?y:Nat)",
"max(min(s(min(plus(0,0),min(?y:Nat,?z:Nat))),s(?x:Nat)),max(?z:Nat,plus(plus(s(?z:Nat),s(?x:Nat)),min(s(?y:Nat),min(?y:Nat,0))))) = min(0,?y:Nat)",
"plus(s(plus(plus(max(?z:Nat,?x:Nat),min(0,?x:Nat)),min(s(?y:Nat),max(0,?x:Nat)))),min(?x:Nat,s(plus(max(?x:Nat,?x:Nat),max(?x:Nat,?z:Nat))))) = min(?x:Nat,?x:Nat)",
"plus(plus(max(0,s(?x:Nat)),max(?z:Nat,s(s(?y:Nat)))),plus(max(plus(min(?y:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),max(0,max(0,?y:Nat))),?x:Nat)) = s(s(0))",
"s(plus(plus(min(min(?x:Nat,0),?x:Nat),plus(?y:Nat,plus(?z:Nat,?x:Nat))),min(min(min(?x:Nat,?z:Nat),s(0)),min(min(?y:Nat,?x:Nat),plus(?x:Nat,?z:Nat))))) = s(?y:Nat)",
"max(max(min(min(s(?x:Nat),?y:Nat),s(plus(?x:Nat,0))),min(min(?x:Nat,0),?x:Nat)),min(plus(0,s(?x:Nat)),plus(min(s(0),?y:Nat),0))) = s(?z:Nat)",
"max(max(plus(s(min(?x:Nat,?y:Nat)),max(s(?x:Nat),min(?x:Nat,?x:Nat))),s(plus(plus(?x:Nat,?x:Nat),?x:Nat))),max(s(max(0,max(?y:Nat,?y:Nat))),0)) = s(?x:Nat)",
"min(plus(max(min(?y:Nat,0),s(?y:Nat)),min(max(?z:Nat,?z:Nat),?z:Nat)),max(plus(plus(?x:Nat,min(?y:Nat,0)),?y:Nat),plus(?z:Nat,max(?y:Nat,s(?z:Nat))))) = s(?z:Nat)",
"plus(s(?y:Nat),plus(min(max(max(?x:Nat,?x:Nat),plus(0,?z:Nat)),plus(s(?z:Nat),?y:Nat)),plus(max(?x:Nat,max(?z:Nat,0)),min(min(0,?x:Nat),?z:Nat)))) = s(?x:Nat)",
"s(?y:Nat) = min(plus(max(plus(s(?x:Nat),max(?x:Nat,?x:Nat)),plus(s(?x:Nat),s(?z:Nat))),max(s(?x:Nat),?x:Nat)),max(?x:Nat,s(min(plus(?z:Nat,?y:Nat),plus(?x:Nat,?y:Nat)))))",
"s(0) = max(min(max(s(s(?x:Nat)),s(s(0))),0),max(plus(plus(?y:Nat,s(?x:Nat)),min(s(?z:Nat),0)),s(plus(min(?x:Nat,?z:Nat),max(?x:Nat,?y:Nat)))))",
"s(s(?z:Nat)) = plus(s(min(max(max(?z:Nat,?y:Nat),max(?x:Nat,0)),s(s(?y:Nat)))),s(plus(max(max(?x:Nat,?y:Nat),min(?x:Nat,?x:Nat)),max(?x:Nat,plus(0,?x:Nat)))))",
"plus(?x:Nat,s(s(?z:Nat))) = max(plus(?z:Nat,plus(min(min(0,?z:Nat),?x:Nat),s(?x:Nat))),s(min(max(max(0,?z:Nat),plus(?x:Nat,?y:Nat)),plus(s(?y:Nat),s(?x:Nat)))))",
"min(?x:Nat,plus(?x:Nat,min(0,?y:Nat))) = plus(0,max(max(min(max(?x:Nat,0),?x:Nat),s(max(?x:Nat,0))),min(s(plus(?x:Nat,?z:Nat)),max(min(0,0),0))))",
"s(s(s(max(?x:Nat,plus(?x:Nat,?x:Nat))))) = max(plus(min(min(?y:Nat,plus(?z:Nat,?x:Nat)),max(max(0,?z:Nat),min(0,?x:Nat))),max(0,plus(0,s(?x:Nat)))),?z:Nat)",
"max(0,plus(min(?x:Nat,?x:Nat),s(s(?x:Nat)))) = plus(plus(?x:Nat,max(min(s(?y:Nat),max(?z:Nat,0)),s(max(?z:Nat,?x:Nat)))),s(plus(s(min(?x:Nat,?y:Nat)),?z:Nat)))",
"max(s(s(min(plus(?x:Nat,?y:Nat),?z:Nat))),?x:Nat) = min(s(plus(plus(0,s(?x:Nat)),plus(s(?x:Nat),0))),min(s(min(?y:Nat,plus(0,0))),min(?x:Nat,0)))",
"min(0,s(s(plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,?z:Nat))))) = max(min(s(plus(?z:Nat,plus(?z:Nat,?z:Nat))),s(plus(plus(?x:Nat,?z:Nat),?z:Nat))),max(?y:Nat,min(?x:Nat,?z:Nat)))",
"max(?x:Nat,max(s(?z:Nat),max(?z:Nat,min(?x:Nat,min(?z:Nat,?y:Nat))))) = min(max(?z:Nat,min(max(?x:Nat,s(?y:Nat)),plus(0,min(?y:Nat,?x:Nat)))),plus(s(s(?x:Nat)),0))",
"max(s(plus(plus(plus(0,?x:Nat),min(?z:Nat,0)),?x:Nat)),?x:Nat) = max(?z:Nat,min(?z:Nat,min(plus(plus(?x:Nat,?x:Nat),s(?z:Nat)),min(plus(0,?z:Nat),plus(?z:Nat,?z:Nat)))))",
"min(?x:Nat,s(max(max(?x:Nat,?y:Nat),max(?x:Nat,plus(?x:Nat,?x:Nat))))) = min(max(s(min(?z:Nat,min(?x:Nat,?x:Nat))),?z:Nat),plus(?y:Nat,plus(?x:Nat,plus(?z:Nat,plus(0,0)))))",
"s(min(max(plus(plus(?y:Nat,?x:Nat),plus(?z:Nat,?y:Nat)),s(?z:Nat)),?y:Nat)) = max(plus(plus(?y:Nat,s(0)),s(?x:Nat)),max(s(s(0)),s(s(max(?x:Nat,?x:Nat)))))",
"s(plus(plus(?y:Nat,s(0)),max(s(min(?x:Nat,?x:Nat)),s(?z:Nat)))) = plus(plus(?x:Nat,s(max(max(?z:Nat,?x:Nat),?y:Nat))),plus(?x:Nat,s(max(?y:Nat,max(?z:Nat,?x:Nat)))))",
"max(0,plus(plus(s(plus(0,?y:Nat)),s(plus(?x:Nat,?x:Nat))),?z:Nat)) = min(max(max(min(?x:Nat,?x:Nat),?x:Nat),?x:Nat),plus(s(?x:Nat),max(s(plus(?x:Nat,0)),0)))",
"max(plus(min(s(max(?z:Nat,?x:Nat)),s(min(?y:Nat,?z:Nat))),?z:Nat),?y:Nat) = max(plus(s(?x:Nat),min(max(max(?z:Nat,?z:Nat),?y:Nat),max(plus(?z:Nat,?z:Nat),s(?x:Nat)))),0)",
"min(?x:Nat,s(plus(plus(?x:Nat,s(?y:Nat)),min(s(?z:Nat),s(?x:Nat))))) = plus(s(?x:Nat),max(s(min(?y:Nat,min(?y:Nat,?z:Nat))),s(plus(plus(?z:Nat,?x:Nat),s(?z:Nat)))))",
"min(plus(s(?y:Nat),min(0,max(plus(?x:Nat,?y:Nat),s(?y:Nat)))),0) = min(s(0),plus(plus(0,s(?y:Nat)),plus(s(s(0)),max(plus(?x:Nat,?x:Nat),?y:Nat))))",
"s(plus(max(s(s(?z:Nat)),max(?x:Nat,max(?z:Nat,?x:Nat))),max(0,0))) = plus(?z:Nat,max(max(?z:Nat,min(s(0),max(?x:Nat,0))),s(min(s(?x:Nat),?z:Nat))))",
"max(s(s(s(s(0)))),s(plus(min(min(?x:Nat,?z:Nat),?y:Nat),?y:Nat))) = min(min(0,plus(plus(max(?x:Nat,?z:Nat),min(?x:Nat,?x:Nat)),max(s(?z:Nat),?y:Nat))),?x:Nat)",
"min(s(?y:Nat),min(plus(?x:Nat,max(plus(?x:Nat,?y:Nat),min(0,?z:Nat))),?z:Nat)) = min(s(?x:Nat),max(min(plus(0,s(?z:Nat)),min(plus(?x:Nat,?y:Nat),s(?y:Nat))),?x:Nat))",
"plus(?x:Nat,min(plus(?z:Nat,s(plus(0,?x:Nat))),plus(s(s(0)),?x:Nat))) = min(max(?y:Nat,s(max(0,0))),min(max(min(max(?y:Nat,?x:Nat),0),?z:Nat),?x:Nat))",
"min(s(min(?x:Nat,max(s(?x:Nat),?y:Nat))),min(s(max(?z:Nat,s(?x:Nat))),?x:Nat)) = plus(max(s(s(?z:Nat)),min(?y:Nat,plus(max(0,?y:Nat),min(?x:Nat,0)))),?x:Nat)",
"min(plus(?x:Nat,plus(plus(max(?x:Nat,?x:Nat),s(?y:Nat)),s(max(?x:Nat,?y:Nat)))),0) = max(?x:Nat,plus(max(?y:Nat,s(plus(0,?x:Nat))),max(s(max(0,?x:Nat)),?z:Nat)))",
"plus(max(s(?z:Nat),?x:Nat),s(min(plus(min(?z:Nat,0),max(?y:Nat,0)),?x:Nat))) = plus(s(s(?y:Nat)),max(s(plus(s(?y:Nat),s(?x:Nat))),max(0,s(?z:Nat))))",
"plus(plus(max(?z:Nat,s(plus(?y:Nat,?x:Nat))),plus(s(plus(?x:Nat,?y:Nat)),?y:Nat)),?z:Nat) = s(plus(s(s(max(?x:Nat,?y:Nat))),plus(?x:Nat,plus(s(?z:Nat),plus(0,?x:Nat)))))",
"s(s(min(max(min(?x:Nat,0),max(?z:Nat,?x:Nat)),plus(max(?z:Nat,?x:Nat),s(0))))) = max(s(?z:Nat),plus(?z:Nat,max(plus(s(?z:Nat),0),s(plus(?x:Nat,?y:Nat)))))",
"s(min(max(s(max(0,?y:Nat)),plus(?x:Nat,?x:Nat)),plus(?x:Nat,s(min(0,0))))) = s(max(max(s(s(?z:Nat)),plus(min(?x:Nat,0),plus(0,0))),?x:Nat))",
"plus(min(?x:Nat,?z:Nat),plus(min(s(plus(?x:Nat,?y:Nat)),min(max(?z:Nat,?z:Nat),?z:Nat)),?x:Nat)) = min(plus(0,s(max(plus(?z:Nat,?x:Nat),?y:Nat))),s(s(s(s(?x:Nat)))))",
"s(min(min(s(s(?y:Nat)),plus(plus(?y:Nat,?y:Nat),s(?x:Nat))),s(min(s(?y:Nat),0)))) = s(plus(max(min(0,?z:Nat),0),s(max(s(?x:Nat),s(0)))))",
"s(min(plus(max(min(0,?z:Nat),plus(?x:Nat,?x:Nat)),0),plus(min(?x:Nat,0),s(?z:Nat)))) = plus(0,plus(min(?x:Nat,0),s(max(plus(?x:Nat,?z:Nat),s(?x:Nat)))))",
"min(?x:Nat,min(s(s(s(?x:Nat))),plus(plus(0,s(?y:Nat)),min(plus(?x:Nat,?x:Nat),0)))) = max(max(min(0,plus(max(?x:Nat,0),?z:Nat)),s(0)),s(?z:Nat))",
"max(s(s(?z:Nat)),plus(min(min(s(?x:Nat),0),plus(plus(?z:Nat,?y:Nat),max(?z:Nat,?y:Nat))),?x:Nat)) = min(plus(plus(?z:Nat,?x:Nat),?x:Nat),max(?x:Nat,plus(s(?x:Nat),0)))",
"min(s(max(0,?y:Nat)),plus(?x:Nat,plus(s(plus(?x:Nat,?y:Nat)),plus(min(?z:Nat,?x:Nat),s(0))))) = max(?x:Nat,plus(?x:Nat,s(plus(max(?x:Nat,0),max(?z:Nat,?z:Nat)))))",
"plus(max(s(plus(s(?z:Nat),?x:Nat)),?z:Nat),s(plus(?x:Nat,plus(plus(?x:Nat,?z:Nat),plus(?x:Nat,?z:Nat))))) = max(0,plus(?x:Nat,max(s(plus(?x:Nat,?z:Nat)),s(s(0)))))",
"plus(min(s(plus(0,plus(?y:Nat,?z:Nat))),min(min(?z:Nat,?z:Nat),?y:Nat)),min(?z:Nat,plus(?z:Nat,?z:Nat))) = min(s(min(0,max(min(?y:Nat,?z:Nat),plus(?x:Nat,0)))),?x:Nat)",
"max(s(s(min(max(?x:Nat,?y:Nat),?y:Nat))),s(min(plus(?y:Nat,s(?x:Nat)),min(?x:Nat,plus(0,?y:Nat))))) = min(s(?z:Nat),s(max(plus(plus(?x:Nat,?x:Nat),?z:Nat),0)))",
"max(min(s(min(?x:Nat,?z:Nat)),plus(max(plus(0,?x:Nat),?x:Nat),plus(max(0,?x:Nat),?y:Nat))),s(?x:Nat)) = plus(plus(0,plus(min(max(?x:Nat,?x:Nat),?x:Nat),?x:Nat)),?x:Nat)",
"max(min(min(min(max(?x:Nat,?y:Nat),s(0)),max(max(?z:Nat,?x:Nat),max(?z:Nat,?x:Nat))),s(0)),0) = max(min(plus(?z:Nat,max(?x:Nat,min(0,0))),?z:Nat),?y:Nat)",
"min(s(?x:Nat),min(min(max(s(?x:Nat),plus(0,0)),s(0)),max(s(plus(?z:Nat,0)),0))) = min(plus(s(plus(s(0),?x:Nat)),plus(?x:Nat,?y:Nat)),?z:Nat)",
"min(max(min(s(?x:Nat),min(?z:Nat,plus(?y:Nat,0))),plus(?z:Nat,min(0,?x:Nat))),max(s(?x:Nat),?z:Nat)) = plus(?z:Nat,min(?z:Nat,s(plus(s(?x:Nat),plus(?x:Nat,?z:Nat)))))",
"plus(s(s(s(s(?z:Nat)))),s(plus(min(s(?z:Nat),plus(?x:Nat,?x:Nat)),plus(s(?x:Nat),s(0))))) = s(s(max(plus(max(?y:Nat,0),max(?x:Nat,0)),?y:Nat)))",
"plus(plus(max(?x:Nat,?z:Nat),0),min(s(min(s(?x:Nat),?x:Nat)),s(min(plus(?y:Nat,?y:Nat),s(?y:Nat))))) = s(s(min(min(plus(?x:Nat,?y:Nat),min(0,?y:Nat)),?y:Nat)))",
"min(max(max(plus(?x:Nat,plus(?x:Nat,?y:Nat)),min(plus(0,?x:Nat),s(?x:Nat))),max(s(s(?x:Nat)),?x:Nat)),?z:Nat) = s(s(max(plus(?z:Nat,min(?x:Nat,?y:Nat)),s(?x:Nat))))",
"plus(plus(max(s(?y:Nat),plus(plus(0,?y:Nat),plus(?z:Nat,?x:Nat))),plus(?x:Nat,s(?y:Nat))),plus(s(?z:Nat),?x:Nat)) = min(?x:Nat,s(plus(?x:Nat,plus(s(?x:Nat),s(?x:Nat)))))",
"max(?z:Nat,max(max(0,min(plus(0,?x:Nat),0)),min(plus(plus(?x:Nat,?y:Nat),min(?z:Nat,?y:Nat)),plus(s(0),?z:Nat)))) = min(max(0,s(s(?x:Nat))),s(?x:Nat))",
"plus(?x:Nat,min(max(max(max(?z:Nat,?z:Nat),plus(?x:Nat,?x:Nat)),0),min(plus(?z:Nat,plus(?y:Nat,0)),max(s(0),s(?x:Nat))))) = s(min(s(min(?z:Nat,?x:Nat)),?x:Nat))",
"plus(max(?y:Nat,s(plus(s(?x:Nat),max(?x:Nat,?x:Nat)))),plus(min(s(?x:Nat),?x:Nat),max(min(?x:Nat,min(?z:Nat,?y:Nat)),s(?x:Nat)))) = s(min(?x:Nat,min(?x:Nat,s(?y:Nat))))",
"s(min(max(min(min(?z:Nat,?x:Nat),max(?x:Nat,?x:Nat)),s(min(0,0))),min(s(s(?y:Nat)),plus(plus(?x:Nat,?y:Nat),min(?y:Nat,0))))) = plus(?x:Nat,min(?z:Nat,?x:Nat))",
"plus(max(s(0),s(s(plus(0,?x:Nat)))),plus(plus(s(min(0,0)),?y:Nat),plus(plus(plus(?z:Nat,0),?x:Nat),max(?x:Nat,?y:Nat)))) = plus(0,max(?y:Nat,?y:Nat))",
"max(plus(plus(min(plus(?x:Nat,0),min(?z:Nat,?y:Nat)),?y:Nat),min(s(s(?z:Nat)),?x:Nat)),plus(min(max(?y:Nat,max(?x:Nat,?z:Nat)),?z:Nat),s(?x:Nat))) = plus(?y:Nat,s(0))",
"min(max(plus(plus(s(?x:Nat),s(?z:Nat)),min(s(?x:Nat),s(?x:Nat))),s(s(max(?z:Nat,0)))),max(?z:Nat,s(min(max(?y:Nat,?x:Nat),0)))) = s(min(0,0))",
"min(plus(min(max(plus(?z:Nat,?y:Nat),plus(?z:Nat,?x:Nat)),plus(s(?z:Nat),max(?y:Nat,?z:Nat))),max(min(plus(?x:Nat,?x:Nat),?z:Nat),?z:Nat)),plus(?x:Nat,?x:Nat)) = s(s(s(?y:Nat)))",
"max(s(plus(min(min(0,?x:Nat),min(?z:Nat,?x:Nat)),plus(?x:Nat,min(?x:Nat,?z:Nat)))),s(plus(s(plus(?x:Nat,?z:Nat)),plus(plus(?x:Nat,?y:Nat),max(0,?x:Nat))))) = s(?x:Nat)",
"max(min(0,plus(max(?z:Nat,plus(?z:Nat,?x:Nat)),?x:Nat)),min(plus(s(plus(?x:Nat,?x:Nat)),s(plus(?x:Nat,?z:Nat))),min(max(?y:Nat,min(?z:Nat,0)),s(?y:Nat)))) = s(0)",
"plus(plus(s(max(max(?x:Nat,0),plus(0,0))),min(?x:Nat,s(max(?x:Nat,?x:Nat)))),max(?x:Nat,plus(min(plus(?z:Nat,?z:Nat),s(?y:Nat)),plus(?x:Nat,?x:Nat)))) = s(0)",
"s(?y:Nat) = plus(plus(max(plus(0,plus(?x:Nat,?x:Nat)),?y:Nat),max(0,s(plus(?y:Nat,?x:Nat)))),max(?z:Nat,plus(s(plus(0,?y:Nat)),plus(max(?x:Nat,?z:Nat),min(?z:Nat,?x:Nat)))))",
"max(?x:Nat,?y:Nat) = s(plus(min(plus(plus(0,?x:Nat),plus(?y:Nat,0)),plus(s(0),max(?x:Nat,?x:Nat))),max(max(min(?x:Nat,?z:Nat),?x:Nat),plus(s(?z:Nat),max(?y:Nat,?x:Nat)))))",
"max(?z:Nat,?x:Nat) = min(min(min(?z:Nat,min(max(?z:Nat,0),plus(?x:Nat,?x:Nat))),plus(s(?x:Nat),plus(?x:Nat,s(?y:Nat)))),max(?z:Nat,max(s(s(0)),max(?y:Nat,s(0)))))",
"min(?y:Nat,?x:Nat) = max(max(plus(plus(?x:Nat,s(?x:Nat)),max(plus(?y:Nat,?x:Nat),?z:Nat)),s(?x:Nat)),max(plus(plus(s(?y:Nat),plus(?z:Nat,0)),min(s(?x:Nat),s(?y:Nat))),0))",
"max(s(?z:Nat),?y:Nat) = min(plus(?x:Nat,?y:Nat),max(max(plus(max(?x:Nat,?z:Nat),?z:Nat),plus(min(0,?z:Nat),min(?x:Nat,?y:Nat))),min(min(max(?y:Nat,0),s(0)),s(0))))",
"s(min(max(s(?x:Nat),?x:Nat),?x:Nat)) = plus(s(min(plus(?y:Nat,s(?y:Nat)),0)),max(min(?z:Nat,s(min(0,?y:Nat))),min(0,max(max(?z:Nat,?x:Nat),plus(?y:Nat,?z:Nat)))))",
"max(s(min(?z:Nat,s(0))),0) = plus(plus(s(max(plus(?x:Nat,?x:Nat),s(0))),max(s(plus(?x:Nat,?y:Nat)),?y:Nat)),max(min(?x:Nat,min(?x:Nat,plus(0,?x:Nat))),?y:Nat))",
"plus(s(plus(s(?x:Nat),?x:Nat)),?z:Nat) = max(s(plus(plus(0,max(?x:Nat,?x:Nat)),min(s(?x:Nat),max(0,0)))),plus(min(max(min(?x:Nat,?x:Nat),0),s(?x:Nat)),?y:Nat))",
"s(plus(plus(?z:Nat,?x:Nat),s(min(?y:Nat,?x:Nat)))) = min(?z:Nat,plus(plus(s(min(?z:Nat,?x:Nat)),min(?z:Nat,?y:Nat)),max(max(?y:Nat,min(?x:Nat,?x:Nat)),max(?y:Nat,min(?z:Nat,0)))))",
"min(max(min(?x:Nat,?y:Nat),?z:Nat),min(?x:Nat,?x:Nat)) = min(plus(?x:Nat,plus(?x:Nat,min(?x:Nat,?y:Nat))),max(max(plus(?z:Nat,max(0,?y:Nat)),?x:Nat),min(s(plus(?x:Nat,?x:Nat)),0)))",
"min(plus(?x:Nat,s(s(?z:Nat))),min(?z:Nat,?x:Nat)) = min(s(plus(max(s(?y:Nat),max(?x:Nat,?y:Nat)),?y:Nat)),min(plus(max(?x:Nat,?z:Nat),min(s(?y:Nat),max(?z:Nat,?x:Nat))),0))",
"plus(?x:Nat,plus(?y:Nat,max(?x:Nat,plus(0,?y:Nat)))) = min(plus(min(s(min(?x:Nat,?x:Nat)),plus(?y:Nat,max(?x:Nat,?x:Nat))),plus(s(s(?z:Nat)),min(s(?z:Nat),s(0)))),0)",
"plus(s(plus(s(min(?y:Nat,0)),?x:Nat)),0) = max(plus(0,max(?y:Nat,max(s(?x:Nat),?x:Nat))),plus(0,plus(plus(?z:Nat,?x:Nat),plus(plus(?x:Nat,?y:Nat),min(?x:Nat,0)))))",
"plus(plus(min(s(?y:Nat),0),?x:Nat),s(?x:Nat)) = max(plus(plus(?y:Nat,max(?x:Nat,?y:Nat)),s(?x:Nat)),plus(min(s(min(?x:Nat,?y:Nat)),min(s(0),s(?y:Nat))),s(0)))",
"min(s(max(?z:Nat,?x:Nat)),min(plus(?z:Nat,?x:Nat),?y:Nat)) = max(plus(s(?y:Nat),0),plus(0,min(plus(min(?x:Nat,?y:Nat),s(?z:Nat)),max(max(?x:Nat,?y:Nat),max(?y:Nat,?x:Nat)))))",
"plus(max(s(s(max(?x:Nat,?x:Nat))),s(?x:Nat)),?x:Nat) = min(s(max(s(plus(?x:Nat,0)),s(max(?x:Nat,?z:Nat)))),max(s(?x:Nat),max(min(?z:Nat,min(0,0)),?x:Nat)))",
"min(s(s(plus(plus(?z:Nat,?y:Nat),plus(?x:Nat,?x:Nat)))),?y:Nat) = s(max(plus(plus(?z:Nat,max(?x:Nat,?y:Nat)),plus(?x:Nat,min(0,?x:Nat))),plus(0,min(s(?z:Nat),s(?x:Nat)))))",
"max(?x:Nat,min(plus(s(min(?y:Nat,?y:Nat)),0),plus(?x:Nat,?y:Nat))) = plus(max(plus(max(min(0,?y:Nat),max(?x:Nat,?z:Nat)),0),?x:Nat),s(max(s(max(?x:Nat,?x:Nat)),?x:Nat)))",
"max(s(?x:Nat),plus(?z:Nat,s(max(s(0),max(0,?z:Nat))))) = min(min(?y:Nat,0),max(plus(plus(max(?x:Nat,?z:Nat),max(?z:Nat,?x:Nat)),?y:Nat),min(s(s(?x:Nat)),?x:Nat)))",
"min(min(min(s(plus(?x:Nat,0)),s(s(?x:Nat))),?x:Nat),?z:Nat) = s(plus(min(max(?z:Nat,plus(?x:Nat,?z:Nat)),plus(s(0),plus(?x:Nat,?z:Nat))),plus(?x:Nat,max(?y:Nat,?z:Nat))))",
"plus(?y:Nat,max(?z:Nat,plus(max(plus(0,?x:Nat),?x:Nat),s(?z:Nat)))) = max(?z:Nat,max(max(max(max(?z:Nat,?x:Nat),max(?x:Nat,?z:Nat)),?x:Nat),plus(min(plus(?z:Nat,?x:Nat),?z:Nat),?x:Nat)))",
"s(s(plus(max(?z:Nat,?x:Nat),plus(min(?z:Nat,?x:Nat),plus(?x:Nat,?x:Nat))))) = min(s(plus(?x:Nat,plus(max(?x:Nat,?z:Nat),0))),plus(0,s(min(s(?z:Nat),min(0,?x:Nat)))))",
"min(?y:Nat,plus(plus(?x:Nat,min(plus(0,0),plus(?z:Nat,?x:Nat))),?y:Nat)) = plus(s(min(?x:Nat,s(0))),s(max(min(s(?x:Nat),s(?x:Nat)),min(s(?z:Nat),s(?x:Nat)))))",
"min(plus(?y:Nat,max(s(plus(?z:Nat,0)),s(plus(0,?x:Nat)))),?x:Nat) = max(s(min(plus(plus(?x:Nat,?x:Nat),min(?y:Nat,?z:Nat)),plus(plus(?y:Nat,?x:Nat),plus(?z:Nat,?x:Nat)))),?x:Nat)",
"plus(min(max(0,s(?x:Nat)),min(s(min(?y:Nat,?y:Nat)),?x:Nat)),?x:Nat) = plus(max(s(plus(plus(?x:Nat,?x:Nat),s(?z:Nat))),max(min(?y:Nat,max(?x:Nat,?x:Nat)),?y:Nat)),s(?z:Nat))",
"max(s(s(s(plus(?y:Nat,?y:Nat)))),plus(0,s(min(max(?x:Nat,?z:Nat),?z:Nat)))) = min(s(?x:Nat),plus(min(plus(plus(?x:Nat,?z:Nat),?x:Nat),max(plus(?x:Nat,?y:Nat),0)),?x:Nat))",
"max(s(plus(0,min(?z:Nat,0))),plus(?z:Nat,max(s(plus(?x:Nat,?z:Nat)),?x:Nat))) = plus(plus(plus(min(?x:Nat,s(?x:Nat)),min(max(?x:Nat,?y:Nat),min(?x:Nat,0))),0),?x:Nat)",
"min(plus(?x:Nat,s(s(plus(?y:Nat,?x:Nat)))),max(s(min(?z:Nat,s(?y:Nat))),?z:Nat)) = plus(plus(s(min(max(?x:Nat,0),?y:Nat)),min(?x:Nat,min(max(0,?y:Nat),?y:Nat))),?x:Nat)",
"max(plus(min(plus(?y:Nat,?x:Nat),plus(plus(?z:Nat,?z:Nat),s(?y:Nat))),?x:Nat),plus(?x:Nat,?x:Nat)) = min(max(?z:Nat,max(s(s(0)),max(s(?y:Nat),max(?z:Nat,?x:Nat)))),s(?y:Nat))",
"min(plus(s(s(plus(0,?x:Nat))),plus(s(plus(?y:Nat,?x:Nat)),?z:Nat)),max(?x:Nat,?x:Nat)) = max(max(?x:Nat,min(plus(?x:Nat,?x:Nat),plus(?x:Nat,plus(?x:Nat,?y:Nat)))),max(0,?x:Nat))",
"min(?x:Nat,max(plus(min(?z:Nat,?x:Nat),max(max(?y:Nat,?y:Nat),plus(0,?x:Nat))),plus(?x:Nat,?x:Nat))) = min(min(min(s(?y:Nat),max(plus(?y:Nat,?z:Nat),plus(0,?x:Nat))),?z:Nat),?x:Nat)",
"s(min(?z:Nat,plus(min(plus(?y:Nat,?y:Nat),plus(?z:Nat,?y:Nat)),plus(max(?y:Nat,0),plus(?z:Nat,?z:Nat))))) = s(plus(min(plus(plus(0,?x:Nat),?z:Nat),s(max(?y:Nat,?z:Nat))),0))",
"max(s(plus(s(s(?y:Nat)),max(s(0),?z:Nat))),plus(max(plus(?x:Nat,s(?x:Nat)),0),?z:Nat)) = min(s(s(plus(max(?z:Nat,?x:Nat),plus(?x:Nat,?y:Nat)))),min(?z:Nat,?x:Nat))",
"plus(max(min(plus(0,plus(?z:Nat,?x:Nat)),plus(s(?x:Nat),plus(0,?x:Nat))),?x:Nat),min(?y:Nat,?x:Nat)) = max(?x:Nat,min(max(s(max(?z:Nat,0)),?x:Nat),max(s(?z:Nat),?x:Nat)))",
"s(min(s(plus(max(?y:Nat,?x:Nat),max(?z:Nat,?x:Nat))),plus(min(min(?z:Nat,?x:Nat),plus(0,?x:Nat)),0))) = min(0,plus(max(?z:Nat,plus(s(?z:Nat),min(?z:Nat,?z:Nat))),?x:Nat))",
"max(plus(s(plus(s(?x:Nat),0)),plus(?x:Nat,s(?x:Nat))),plus(s(min(0,plus(?y:Nat,?x:Nat))),?x:Nat)) = s(min(max(?y:Nat,s(s(?y:Nat))),s(s(max(?y:Nat,0)))))",
"plus(min(plus(max(plus(?z:Nat,0),max(0,0)),min(s(?x:Nat),s(?y:Nat))),s(?y:Nat)),s(0)) = s(max(0,max(plus(plus(?x:Nat,0),s(0)),s(?y:Nat))))",
"max(s(max(?x:Nat,min(s(?z:Nat),s(?x:Nat)))),plus(max(max(0,?x:Nat),min(plus(?x:Nat,0),?z:Nat)),?z:Nat)) = min(?y:Nat,plus(min(?x:Nat,plus(0,plus(0,?x:Nat))),?z:Nat))",
"min(?y:Nat,plus(plus(max(s(?z:Nat),s(?x:Nat)),max(?x:Nat,?y:Nat)),max(min(?y:Nat,?x:Nat),s(max(?z:Nat,?y:Nat))))) = plus(s(min(plus(plus(?z:Nat,?y:Nat),0),s(0))),?z:Nat)",
"plus(0,min(plus(max(?x:Nat,plus(?z:Nat,?z:Nat)),?x:Nat),max(max(?x:Nat,max(?z:Nat,?y:Nat)),s(plus(0,?y:Nat))))) = plus(min(s(?x:Nat),plus(?x:Nat,min(?y:Nat,s(?x:Nat)))),?x:Nat)",
"plus(s(min(?z:Nat,s(?z:Nat))),min(min(0,s(s(0))),max(plus(plus(?z:Nat,?x:Nat),s(?z:Nat)),?z:Nat))) = min(s(?x:Nat),min(plus(s(?z:Nat),s(s(0))),0))",
"plus(s(min(s(?y:Nat),min(0,min(?x:Nat,?z:Nat)))),plus(plus(?z:Nat,plus(s(?z:Nat),?y:Nat)),plus(?y:Nat,?z:Nat))) = plus(?x:Nat,s(s(plus(plus(?y:Nat,?x:Nat),max(?z:Nat,?y:Nat)))))",
"s(min(plus(plus(?x:Nat,plus(?z:Nat,?y:Nat)),plus(?x:Nat,?x:Nat)),min(min(s(?z:Nat),plus(?y:Nat,?x:Nat)),plus(?x:Nat,?z:Nat)))) = min(?z:Nat,plus(?z:Nat,max(max(?x:Nat,s(0)),?z:Nat)))",
"max(plus(max(s(s(?y:Nat)),max(?z:Nat,plus(?y:Nat,?z:Nat))),0),max(s(min(max(?y:Nat,?z:Nat),?x:Nat)),s(?y:Nat))) = plus(plus(?x:Nat,s(min(s(?x:Nat),s(?x:Nat)))),?z:Nat)",
"min(?x:Nat,max(plus(?x:Nat,min(plus(?z:Nat,0),?y:Nat)),max(max(?x:Nat,max(?x:Nat,?x:Nat)),min(max(?y:Nat,0),?y:Nat)))) = s(min(plus(plus(s(?y:Nat),s(0)),?x:Nat),?z:Nat))",
"plus(min(plus(s(s(?x:Nat)),max(s(?x:Nat),max(0,?x:Nat))),s(plus(max(0,?y:Nat),min(0,?y:Nat)))),?y:Nat) = max(plus(s(max(plus(?x:Nat,?z:Nat),?y:Nat)),?x:Nat),?z:Nat)",
"plus(plus(min(?x:Nat,min(s(?y:Nat),max(?y:Nat,?x:Nat))),min(plus(s(0),max(?x:Nat,?z:Nat)),plus(?y:Nat,?y:Nat))),?x:Nat) = max(0,s(plus(?x:Nat,plus(0,min(?x:Nat,?x:Nat)))))",
"plus(max(?z:Nat,max(max(?x:Nat,max(?x:Nat,?y:Nat)),?x:Nat)),plus(?x:Nat,min(s(plus(0,?y:Nat)),min(0,min(0,?y:Nat))))) = plus(s(?y:Nat),s(min(plus(?x:Nat,?x:Nat),?y:Nat)))",
"max(plus(s(plus(max(?y:Nat,?x:Nat),plus(0,0))),s(?x:Nat)),min(plus(s(min(?y:Nat,?y:Nat)),?y:Nat),s(min(?z:Nat,?x:Nat)))) = plus(?x:Nat,min(s(?x:Nat),s(s(0))))",
"plus(plus(min(min(plus(?z:Nat,?z:Nat),s(?y:Nat)),plus(s(?z:Nat),0)),plus(max(max(?y:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),?z:Nat)),s(0)) = min(?x:Nat,max(max(?y:Nat,?x:Nat),?x:Nat))",
"max(min(plus(s(s(?x:Nat)),?x:Nat),?x:Nat),min(max(max(?x:Nat,min(0,?z:Nat)),?x:Nat),plus(s(s(?x:Nat)),plus(0,plus(?z:Nat,?x:Nat))))) = plus(?x:Nat,s(min(?z:Nat,?y:Nat)))",
"plus(s(min(s(max(?x:Nat,?x:Nat)),plus(s(?x:Nat),?x:Nat))),plus(0,plus(max(max(?x:Nat,?x:Nat),?z:Nat),min(plus(?x:Nat,?y:Nat),s(?z:Nat))))) = s(max(plus(?y:Nat,?z:Nat),?z:Nat))",
"plus(min(s(max(s(0),s(?z:Nat))),s(min(?x:Nat,s(?y:Nat)))),s(plus(min(s(?x:Nat),s(?y:Nat)),plus(max(0,?x:Nat),plus(?x:Nat,?z:Nat))))) = plus(0,s(?z:Nat))",
"plus(plus(?x:Nat,min(?z:Nat,max(plus(?y:Nat,?z:Nat),plus(?y:Nat,?z:Nat)))),min(max(plus(max(0,?z:Nat),s(0)),s(plus(?z:Nat,?z:Nat))),s(s(?x:Nat)))) = plus(s(?y:Nat),?x:Nat)",
"plus(plus(min(s(s(?z:Nat)),?x:Nat),max(s(min(?y:Nat,?z:Nat)),s(?y:Nat))),min(min(max(0,max(?z:Nat,0)),?y:Nat),s(plus(s(?x:Nat),?y:Nat)))) = max(?x:Nat,s(?y:Nat))",
"max(max(0,s(plus(?x:Nat,s(?z:Nat)))),max(max(min(plus(?y:Nat,?x:Nat),plus(0,0)),?y:Nat),plus(plus(s(?y:Nat),?x:Nat),max(max(?x:Nat,0),s(?x:Nat))))) = s(?x:Nat)",
"max(?x:Nat,0) = min(s(max(max(min(?x:Nat,?x:Nat),0),s(max(?x:Nat,?x:Nat)))),plus(max(max(plus(?z:Nat,?y:Nat),0),?x:Nat),plus(min(plus(?x:Nat,?x:Nat),s(?y:Nat)),s(?z:Nat))))",
"plus(?z:Nat,?y:Nat) = max(s(max(plus(s(?x:Nat),min(?y:Nat,0)),max(?y:Nat,?x:Nat))),max(max(s(0),max(plus(?x:Nat,?x:Nat),?x:Nat)),min(?z:Nat,max(plus(0,?x:Nat),s(?x:Nat)))))",
"max(s(?x:Nat),?x:Nat) = max(max(min(?z:Nat,plus(plus(?x:Nat,?x:Nat),max(0,?z:Nat))),plus(s(s(?z:Nat)),?z:Nat)),s(max(plus(?z:Nat,plus(?y:Nat,?x:Nat)),min(s(?x:Nat),s(?z:Nat)))))",
"max(s(?y:Nat),?x:Nat) = max(s(s(max(plus(?x:Nat,0),plus(?x:Nat,?y:Nat)))),plus(s(plus(s(0),max(?x:Nat,?y:Nat))),plus(plus(s(0),min(?x:Nat,0)),plus(?x:Nat,?x:Nat))))",
"plus(?y:Nat,s(?y:Nat)) = plus(max(max(s(s(?x:Nat)),plus(?x:Nat,plus(?z:Nat,?z:Nat))),s(min(plus(?z:Nat,?y:Nat),?x:Nat))),s(min(s(min(?x:Nat,?y:Nat)),max(max(?y:Nat,?z:Nat),?x:Nat))))",
"max(?y:Nat,plus(?z:Nat,?z:Nat)) = plus(min(s(?x:Nat),s(plus(min(0,?z:Nat),plus(?x:Nat,?x:Nat)))),plus(?x:Nat,max(max(min(0,0),plus(?x:Nat,?z:Nat)),max(0,max(0,?z:Nat)))))",
"max(min(?z:Nat,?x:Nat),?z:Nat) = max(max(s(plus(min(?z:Nat,0),max(?z:Nat,?x:Nat))),max(plus(?y:Nat,?x:Nat),plus(max(?x:Nat,0),s(?z:Nat)))),plus(?y:Nat,s(s(plus(?x:Nat,?x:Nat)))))",
"min(?x:Nat,plus(0,?x:Nat)) = plus(max(min(max(max(0,0),min(?y:Nat,?x:Nat)),max(min(?x:Nat,?z:Nat),plus(?y:Nat,?y:Nat))),plus(max(plus(?y:Nat,?y:Nat),s(?z:Nat)),s(?x:Nat))),?x:Nat)",
"s(s(min(?y:Nat,s(?x:Nat)))) = min(s(plus(plus(max(?z:Nat,?x:Nat),max(?x:Nat,?x:Nat)),max(max(?x:Nat,0),s(?x:Nat)))),plus(plus(plus(s(0),max(0,?x:Nat)),?x:Nat),?x:Nat))",
"min(plus(s(?x:Nat),?x:Nat),?x:Nat) = plus(plus(s(min(s(0),max(?x:Nat,?y:Nat))),s(plus(s(?z:Nat),s(?z:Nat)))),max(?z:Nat,min(?x:Nat,max(plus(?x:Nat,?y:Nat),min(?x:Nat,0)))))",
"plus(?x:Nat,max(s(?y:Nat),0)) = min(plus(min(s(0),min(s(?y:Nat),plus(?x:Nat,0))),s(plus(plus(?x:Nat,0),?x:Nat))),min(?x:Nat,min(s(s(?y:Nat)),max(0,0))))",
"max(max(?z:Nat,?x:Nat),max(?x:Nat,0)) = plus(min(max(?x:Nat,plus(min(0,0),min(?y:Nat,?x:Nat))),min(max(plus(0,?z:Nat),s(?x:Nat)),plus(s(?z:Nat),s(?z:Nat)))),s(?x:Nat))",
"plus(s(s(?z:Nat)),max(0,?x:Nat)) = min(?x:Nat,plus(plus(max(s(?y:Nat),s(0)),s(max(?z:Nat,?y:Nat))),min(min(plus(?x:Nat,?z:Nat),s(?x:Nat)),plus(min(?x:Nat,?x:Nat),?y:Nat))))",
"plus(s(max(?z:Nat,?z:Nat)),s(?z:Nat)) = min(plus(min(?y:Nat,max(?z:Nat,max(?z:Nat,?x:Nat))),max(s(s(?x:Nat)),?x:Nat)),max(s(0),min(plus(min(?x:Nat,?z:Nat),s(?x:Nat)),?x:Nat)))",
"plus(s(s(s(max(?x:Nat,?x:Nat)))),?z:Nat) = plus(min(?z:Nat,s(max(s(?y:Nat),s(?x:Nat)))),min(max(max(plus(?y:Nat,0),plus(?z:Nat,0)),min(0,plus(0,?x:Nat))),0))",
"min(?x:Nat,plus(s(s(plus(0,?y:Nat))),0)) = plus(plus(plus(min(?x:Nat,min(?z:Nat,0)),?x:Nat),plus(plus(s(?x:Nat),s(?y:Nat)),plus(plus(?y:Nat,?x:Nat),max(?x:Nat,?y:Nat)))),0)",
"plus(s(max(?x:Nat,?z:Nat)),max(s(?y:Nat),0)) = plus(s(max(s(?x:Nat),plus(?y:Nat,plus(?y:Nat,?z:Nat)))),plus(max(plus(plus(?x:Nat,?x:Nat),0),0),s(min(s(?y:Nat),?x:Nat))))",
"max(?y:Nat,plus(s(s(?x:Nat)),min(?y:Nat,s(?x:Nat)))) = min(?y:Nat,max(s(plus(plus(?z:Nat,?y:Nat),?x:Nat)),min(min(min(?x:Nat,?x:Nat),plus(?z:Nat,?x:Nat)),max(0,max(0,0)))))",
"min(s(min(?x:Nat,plus(0,max(?x:Nat,?y:Nat)))),?z:Nat) = plus(max(max(plus(plus(?x:Nat,0),plus(0,?x:Nat)),s(min(?x:Nat,?x:Nat))),plus(s(plus(0,?x:Nat)),?x:Nat)),s(?x:Nat))",
"min(plus(?y:Nat,s(min(min(?z:Nat,0),0))),?z:Nat) = max(s(s(s(plus(?z:Nat,?z:Nat)))),min(plus(s(?x:Nat),plus(?x:Nat,plus(?y:Nat,?z:Nat))),s(plus(?z:Nat,plus(?x:Nat,?x:Nat)))))",
"min(?z:Nat,min(?y:Nat,s(max(s(0),max(?y:Nat,0))))) = plus(s(max(plus(s(?x:Nat),plus(?x:Nat,?x:Nat)),max(max(?z:Nat,?y:Nat),?z:Nat))),plus(?x:Nat,max(s(s(?y:Nat)),?y:Nat)))",
"plus(plus(plus(?y:Nat,plus(plus(?x:Nat,?z:Nat),?z:Nat)),0),?z:Nat) = plus(?z:Nat,plus(min(plus(plus(?x:Nat,?y:Nat),plus(0,?y:Nat)),plus(s(?z:Nat),min(?z:Nat,?z:Nat))),s(s(s(?x:Nat)))))",
"plus(s(max(s(plus(0,0)),s(plus(?y:Nat,?x:Nat)))),?x:Nat) = min(?x:Nat,max(s(s(max(?x:Nat,?z:Nat))),max(s(min(?x:Nat,?x:Nat)),plus(min(?x:Nat,?y:Nat),plus(0,?x:Nat)))))",
"s(max(?z:Nat,plus(s(plus(?x:Nat,?x:Nat)),min(max(?z:Nat,0),?z:Nat)))) = s(max(max(plus(s(0),plus(?z:Nat,?x:Nat)),?z:Nat),plus(max(s(?x:Nat),?y:Nat),max(?y:Nat,s(?x:Nat)))))",
"min(plus(?y:Nat,s(?z:Nat)),max(s(0),plus(plus(?y:Nat,0),?y:Nat))) = max(plus(max(plus(max(?y:Nat,0),max(?x:Nat,?x:Nat)),min(?z:Nat,min(?x:Nat,?z:Nat))),?x:Nat),s(s(?y:Nat)))",
"s(min(s(max(s(?z:Nat),min(0,0))),max(s(s(?x:Nat)),?z:Nat))) = max(max(?z:Nat,?x:Nat),max(max(plus(plus(?y:Nat,0),max(?x:Nat,?z:Nat)),0),s(max(?y:Nat,0))))",
"max(s(?y:Nat),plus(plus(?y:Nat,min(s(?y:Nat),max(?y:Nat,?x:Nat))),s(?x:Nat))) = s(min(plus(plus(0,plus(0,?x:Nat)),max(?x:Nat,max(?y:Nat,?y:Nat))),plus(s(s(?x:Nat)),?x:Nat)))",
"plus(plus(?x:Nat,max(?x:Nat,max(min(?z:Nat,?x:Nat),s(?y:Nat)))),plus(?z:Nat,?x:Nat)) = max(min(plus(plus(s(0),plus(?x:Nat,?y:Nat)),min(plus(?y:Nat,?y:Nat),plus(?y:Nat,?z:Nat))),?x:Nat),?z:Nat)",
"max(max(max(s(plus(?y:Nat,0)),min(0,?z:Nat)),s(s(max(?z:Nat,?z:Nat)))),?z:Nat) = plus(plus(s(?y:Nat),min(0,s(?x:Nat))),max(plus(?x:Nat,plus(s(?x:Nat),?z:Nat)),?x:Nat))",
"min(s(s(?y:Nat)),min(?x:Nat,min(s(max(0,?x:Nat)),plus(?x:Nat,plus(?z:Nat,?y:Nat))))) = plus(?y:Nat,plus(plus(max(max(?z:Nat,?y:Nat),s(?z:Nat)),s(?z:Nat)),plus(?x:Nat,s(?z:Nat))))",
"s(plus(max(plus(plus(?x:Nat,?x:Nat),max(?z:Nat,?z:Nat)),s(s(0))),min(s(?x:Nat),?z:Nat))) = plus(plus(0,?y:Nat),plus(max(max(min(0,?y:Nat),?y:Nat),?x:Nat),min(?x:Nat,?z:Nat)))",
"min(plus(plus(?y:Nat,s(?z:Nat)),min(s(plus(0,?y:Nat)),min(0,0))),plus(?y:Nat,0)) = plus(min(s(plus(s(?z:Nat),?x:Nat)),s(min(0,s(0)))),max(0,?x:Nat))",
"plus(min(plus(plus(plus(?z:Nat,?x:Nat),s(?z:Nat)),plus(?x:Nat,s(?x:Nat))),s(?x:Nat)),s(?x:Nat)) = plus(plus(s(plus(s(?y:Nat),?y:Nat)),max(?y:Nat,min(s(?x:Nat),s(?x:Nat)))),0)",
"max(max(0,?x:Nat),min(?x:Nat,plus(max(s(?x:Nat),?z:Nat),plus(plus(?y:Nat,?y:Nat),plus(?y:Nat,?y:Nat))))) = min(min(s(?x:Nat),s(max(plus(?x:Nat,?x:Nat),s(0)))),max(?y:Nat,0))",
"plus(max(s(s(?x:Nat)),plus(s(max(?x:Nat,0)),min(s(?x:Nat),s(?z:Nat)))),max(?x:Nat,?y:Nat)) = plus(s(plus(s(s(?y:Nat)),max(plus(0,?x:Nat),max(?z:Nat,0)))),?y:Nat)",
"s(max(plus(max(max(0,?x:Nat),plus(?x:Nat,?x:Nat)),min(min(?x:Nat,?z:Nat),plus(?x:Nat,?z:Nat))),s(?x:Nat))) = plus(?z:Nat,plus(max(s(plus(0,?y:Nat)),s(min(0,?x:Nat))),?z:Nat))",
"s(min(max(max(min(?y:Nat,?y:Nat),?y:Nat),max(plus(?x:Nat,?x:Nat),s(?y:Nat))),s(s(max(?y:Nat,?x:Nat))))) = s(max(s(0),max(0,plus(min(?y:Nat,?x:Nat),max(?y:Nat,?x:Nat)))))",
"max(s(s(?y:Nat)),max(plus(max(max(?y:Nat,?z:Nat),?z:Nat),max(max(?x:Nat,?z:Nat),s(?x:Nat))),s(?y:Nat))) = min(min(?x:Nat,plus(s(min(0,?x:Nat)),min(s(0),?x:Nat))),?x:Nat)",
"s(plus(s(s(max(0,?z:Nat))),min(plus(min(?z:Nat,?y:Nat),max(?x:Nat,?z:Nat)),plus(?x:Nat,plus(?x:Nat,?x:Nat))))) = plus(min(0,plus(?z:Nat,max(s(?x:Nat),max(?x:Nat,?x:Nat)))),?x:Nat)",
"max(min(min(min(?x:Nat,plus(?x:Nat,?y:Nat)),0),min(min(plus(0,0),plus(?z:Nat,?y:Nat)),s(?x:Nat))),?x:Nat) = plus(min(?y:Nat,?x:Nat),max(s(plus(min(0,?y:Nat),?y:Nat)),?y:Nat))",
"min(?x:Nat,plus(max(?z:Nat,s(plus(?x:Nat,?x:Nat))),min(plus(max(?x:Nat,?x:Nat),min(?x:Nat,?z:Nat)),s(s(?x:Nat))))) = plus(s(plus(plus(s(?x:Nat),?x:Nat),?x:Nat)),plus(?x:Nat,s(?x:Nat)))",
"min(s(s(?x:Nat)),plus(plus(min(min(?y:Nat,0),s(0)),?x:Nat),s(max(max(?y:Nat,?z:Nat),s(?x:Nat))))) = max(max(s(?x:Nat),plus(s(?z:Nat),plus(s(?y:Nat),?y:Nat))),?x:Nat)",
"plus(s(min(plus(plus(?x:Nat,0),min(?y:Nat,0)),?z:Nat)),min(?x:Nat,min(s(max(?y:Nat,?x:Nat)),s(?x:Nat)))) = plus(s(min(?x:Nat,min(plus(?x:Nat,0),min(?y:Nat,?z:Nat)))),?x:Nat)",
"plus(plus(min(max(plus(?z:Nat,0),s(?y:Nat)),s(?y:Nat)),s(plus(max(?x:Nat,0),max(0,0)))),?z:Nat) = plus(0,plus(?y:Nat,max(s(s(?x:Nat)),max(s(?z:Nat),?x:Nat))))",
"max(max(?y:Nat,min(plus(plus(0,0),0),?x:Nat)),plus(s(s(plus(?x:Nat,?z:Nat))),s(s(max(0,?x:Nat))))) = plus(min(s(s(s(?x:Nat))),?x:Nat),plus(?x:Nat,s(?x:Nat)))",
"max(?x:Nat,max(plus(0,min(?x:Nat,0)),plus(max(plus(0,?x:Nat),min(?z:Nat,?z:Nat)),min(s(0),min(?y:Nat,?x:Nat))))) = s(s(min(s(?z:Nat),min(max(0,?y:Nat),?x:Nat))))",
"max(?x:Nat,min(plus(max(?x:Nat,?x:Nat),min(max(?x:Nat,0),min(?x:Nat,?x:Nat))),s(min(plus(?y:Nat,?x:Nat),plus(?z:Nat,?z:Nat))))) = s(max(s(0),max(s(?x:Nat),s(s(0)))))",
"plus(?y:Nat,min(plus(min(s(?y:Nat),plus(0,0)),?x:Nat),min(plus(max(?z:Nat,0),?y:Nat),max(max(?z:Nat,?x:Nat),?x:Nat)))) = s(s(plus(?x:Nat,max(s(?z:Nat),plus(?z:Nat,?y:Nat)))))",
"plus(plus(plus(plus(max(?y:Nat,?x:Nat),s(?x:Nat)),min(plus(?z:Nat,?x:Nat),plus(?x:Nat,?z:Nat))),plus(?z:Nat,plus(?x:Nat,?x:Nat))),?y:Nat) = s(plus(?y:Nat,plus(?x:Nat,max(s(?z:Nat),s(?x:Nat)))))",
"min(min(min(s(0),plus(s(0),min(?z:Nat,?x:Nat))),min(?x:Nat,?x:Nat)),min(plus(plus(0,max(?x:Nat,?x:Nat)),?x:Nat),?x:Nat)) = min(?x:Nat,min(s(?x:Nat),min(s(?y:Nat),0)))",
"plus(s(plus(?x:Nat,plus(plus(?z:Nat,?y:Nat),?x:Nat))),min(s(s(min(?x:Nat,?x:Nat))),min(plus(?y:Nat,plus(?x:Nat,?z:Nat)),s(0)))) = s(min(?x:Nat,plus(?y:Nat,s(plus(?z:Nat,?x:Nat)))))",
"max(?y:Nat,min(s(min(plus(?x:Nat,?z:Nat),min(0,0))),plus(min(plus(0,?z:Nat),?x:Nat),plus(max(?y:Nat,?y:Nat),min(?z:Nat,?x:Nat))))) = s(max(plus(min(?z:Nat,?x:Nat),?y:Nat),?x:Nat))",
"min(plus(plus(s(s(?z:Nat)),?y:Nat),s(s(0))),max(plus(?y:Nat,plus(s(?x:Nat),?x:Nat)),max(0,max(?z:Nat,plus(0,?x:Nat))))) = min(min(?z:Nat,s(s(s(?z:Nat)))),?x:Nat)",
"plus(max(max(s(s(0)),s(max(0,0))),max(0,?x:Nat)),plus(plus(plus(min(?x:Nat,?z:Nat),min(?x:Nat,?z:Nat)),?x:Nat),?x:Nat)) = max(min(?x:Nat,s(?z:Nat)),plus(0,?x:Nat))",
"plus(plus(plus(plus(?x:Nat,min(0,?y:Nat)),s(max(?z:Nat,?x:Nat))),min(plus(s(0),s(?x:Nat)),s(plus(0,?y:Nat)))),s(?y:Nat)) = plus(?y:Nat,plus(max(0,s(0)),0))",
"min(max(min(plus(?y:Nat,plus(?y:Nat,0)),max(s(0),plus(?x:Nat,?x:Nat))),s(s(max(?z:Nat,?z:Nat)))),max(?x:Nat,plus(?x:Nat,s(max(?x:Nat,?x:Nat))))) = plus(max(?y:Nat,?y:Nat),?x:Nat)",
"plus(min(s(max(max(0,?x:Nat),?y:Nat)),plus(plus(?z:Nat,plus(0,?y:Nat)),s(0))),s(max(max(0,plus(?x:Nat,?y:Nat)),min(?y:Nat,max(?x:Nat,?x:Nat))))) = s(max(?y:Nat,?x:Nat))",
"max(min(s(max(plus(?z:Nat,?z:Nat),max(0,?y:Nat))),s(s(s(?x:Nat)))),max(?z:Nat,plus(plus(min(?x:Nat,?x:Nat),?y:Nat),min(plus(?y:Nat,0),max(?z:Nat,?z:Nat))))) = max(?x:Nat,?z:Nat)",
"min(plus(s(max(max(?x:Nat,?y:Nat),plus(0,?z:Nat))),min(0,min(?y:Nat,min(0,?z:Nat)))),s(min(plus(plus(?y:Nat,0),s(?y:Nat)),s(max(?x:Nat,?z:Nat))))) = min(?z:Nat,0)",
"min(plus(s(min(min(?y:Nat,?x:Nat),0)),s(?x:Nat)),min(plus(s(s(?y:Nat)),s(plus(?z:Nat,?z:Nat))),min(s(?x:Nat),max(max(?x:Nat,?x:Nat),plus(0,0))))) = s(s(?z:Nat))",
"s(?x:Nat) = max(plus(min(plus(s(?x:Nat),plus(?z:Nat,?z:Nat)),?x:Nat),plus(?x:Nat,min(?y:Nat,plus(?y:Nat,0)))),plus(max(max(min(0,0),?z:Nat),?z:Nat),s(plus(max(?x:Nat,?z:Nat),?z:Nat))))",
"plus(?z:Nat,0) = min(plus(max(max(min(?z:Nat,?z:Nat),?x:Nat),max(max(?x:Nat,0),?y:Nat)),s(min(?x:Nat,?x:Nat))),s(plus(plus(min(?x:Nat,?y:Nat),max(0,?y:Nat)),s(plus(?x:Nat,0)))))",
"min(?z:Nat,s(?y:Nat)) = plus(s(s(s(s(?x:Nat)))),plus(min(s(min(?x:Nat,?z:Nat)),plus(plus(0,?x:Nat),plus(?x:Nat,?x:Nat))),min(max(min(?x:Nat,?x:Nat),?z:Nat),s(plus(?z:Nat,?y:Nat)))))",
"plus(min(?y:Nat,0),?y:Nat) = min(plus(max(min(0,?y:Nat),plus(min(?x:Nat,?x:Nat),?y:Nat)),?x:Nat),max(min(0,?y:Nat),plus(max(max(?x:Nat,?y:Nat),?z:Nat),plus(max(?x:Nat,?x:Nat),s(?x:Nat)))))",
"s(plus(?x:Nat,plus(?z:Nat,?x:Nat))) = max(plus(s(s(min(?x:Nat,?x:Nat))),s(min(s(?y:Nat),?x:Nat))),plus(s(plus(max(?x:Nat,?x:Nat),s(0))),min(s(plus(?z:Nat,?z:Nat)),s(?y:Nat))))",
"min(?x:Nat,min(?x:Nat,s(s(?z:Nat)))) = max(max(max(min(?x:Nat,plus(?x:Nat,?y:Nat)),max(plus(?x:Nat,?z:Nat),s(?z:Nat))),min(?y:Nat,min(s(?z:Nat),?y:Nat))),max(min(0,0),s(?x:Nat)))",
"min(?x:Nat,min(?x:Nat,max(?x:Nat,?x:Nat))) = plus(plus(s(max(min(?x:Nat,0),max(?x:Nat,?x:Nat))),?x:Nat),max(max(?y:Nat,?z:Nat),min(s(min(?x:Nat,?y:Nat)),plus(plus(?z:Nat,?x:Nat),s(?y:Nat)))))",
"s(s(min(plus(0,plus(?x:Nat,?z:Nat)),?x:Nat))) = max(min(max(?x:Nat,?x:Nat),plus(min(plus(?x:Nat,?z:Nat),?x:Nat),s(plus(?z:Nat,?y:Nat)))),max(s(s(plus(?z:Nat,0))),min(?x:Nat,?y:Nat)))",
"min(s(s(plus(max(?z:Nat,?y:Nat),0))),?x:Nat) = plus(min(?y:Nat,max(s(plus(?x:Nat,?z:Nat)),plus(plus(0,?z:Nat),?y:Nat))),min(s(?x:Nat),s(plus(plus(?x:Nat,?x:Nat),max(?z:Nat,0)))))",
"plus(?x:Nat,min(s(?y:Nat),min(?x:Nat,s(s(0))))) = min(s(plus(?x:Nat,s(s(?z:Nat)))),min(s(min(max(?z:Nat,?y:Nat),?z:Nat)),min(min(0,s(?x:Nat)),s(plus(0,?z:Nat)))))",
"s(min(s(?x:Nat),plus(?x:Nat,plus(?x:Nat,max(0,?y:Nat))))) = plus(s(plus(?y:Nat,?y:Nat)),plus(plus(?x:Nat,plus(min(?x:Nat,0),max(?x:Nat,?y:Nat))),max(plus(max(?x:Nat,?x:Nat),?x:Nat),?y:Nat)))",
"s(plus(max(s(max(?z:Nat,?x:Nat)),?y:Nat),min(?x:Nat,?z:Nat))) = plus(max(?x:Nat,plus(plus(?x:Nat,plus(0,?y:Nat)),?x:Nat)),min(min(s(min(?x:Nat,?x:Nat)),0),min(?y:Nat,s(s(0)))))",
"max(?z:Nat,s(plus(plus(plus(?x:Nat,?x:Nat),?x:Nat),s(0)))) = plus(s(0),max(plus(max(s(?y:Nat),max(?y:Nat,?x:Nat)),s(s(?y:Nat))),s(plus(max(0,?x:Nat),plus(?x:Nat,?z:Nat)))))",
"min(?x:Nat,max(s(?y:Nat),s(min(?x:Nat,max(?y:Nat,?x:Nat))))) = min(?z:Nat,max(plus(max(max(?y:Nat,?y:Nat),min(?y:Nat,?y:Nat)),max(s(?x:Nat),plus(?z:Nat,0))),s(min(s(?x:Nat),?x:Nat))))",
"min(min(0,plus(s(min(?z:Nat,?y:Nat)),0)),s(?x:Nat)) = min(min(plus(s(max(?x:Nat,0)),s(?y:Nat)),plus(s(s(?x:Nat)),min(max(?x:Nat,?x:Nat),plus(?x:Nat,?z:Nat)))),s(0))",
"max(plus(plus(plus(plus(?x:Nat,?y:Nat),0),?x:Nat),s(?y:Nat)),?z:Nat) = plus(min(min(plus(max(0,?z:Nat),max(0,?y:Nat)),?x:Nat),?x:Nat),min(max(min(plus(?x:Nat,?x:Nat),?z:Nat),0),?x:Nat))",
"min(s(0),plus(s(?x:Nat),s(min(?y:Nat,max(?x:Nat,?x:Nat))))) = min(s(?y:Nat),max(min(s(min(?z:Nat,?x:Nat)),plus(s(0),max(?z:Nat,?y:Nat))),min(?z:Nat,plus(?x:Nat,s(?x:Nat)))))",
"plus(?y:Nat,plus(plus(max(s(?x:Nat),plus(?z:Nat,0)),?x:Nat),?x:Nat)) = plus(plus(?y:Nat,min(?y:Nat,s(max(?x:Nat,?x:Nat)))),plus(0,plus(plus(s(?z:Nat),max(?x:Nat,?x:Nat)),min(?y:Nat,?x:Nat))))",
"plus(max(?z:Nat,?x:Nat),max(plus(?y:Nat,s(min(?y:Nat,?x:Nat))),?x:Nat)) = min(plus(?z:Nat,?x:Nat),max(min(s(max(0,?x:Nat)),min(plus(?x:Nat,0),?x:Nat)),s(min(?y:Nat,max(?z:Nat,?x:Nat)))))",
"s(s(max(min(min(?x:Nat,0),0),min(min(?x:Nat,?x:Nat),?x:Nat)))) = min(s(min(s(plus(?x:Nat,?x:Nat)),?y:Nat)),plus(s(s(s(?x:Nat))),s(max(plus(?x:Nat,0),s(?z:Nat)))))",
"plus(s(plus(plus(min(?z:Nat,?x:Nat),?x:Nat),s(max(0,?x:Nat)))),0) = plus(plus(?z:Nat,min(min(?x:Nat,plus(0,?x:Nat)),plus(min(?z:Nat,?x:Nat),s(0)))),plus(plus(?z:Nat,0),?x:Nat))",
"max(max(max(plus(?x:Nat,s(?z:Nat)),min(?x:Nat,max(?z:Nat,?x:Nat))),0),0) = plus(plus(?z:Nat,plus(?x:Nat,max(plus(0,0),?z:Nat))),s(plus(s(min(0,?z:Nat)),plus(?x:Nat,?y:Nat))))",
"min(s(0),s(max(plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),s(?z:Nat)))) = s(plus(s(min(s(?z:Nat),plus(?x:Nat,?x:Nat))),max(s(plus(0,?y:Nat)),plus(min(?z:Nat,?y:Nat),?z:Nat))))",
"s(max(plus(max(plus(?x:Nat,?x:Nat),min(0,?x:Nat)),s(max(?y:Nat,?x:Nat))),?z:Nat)) = s(min(plus(min(?x:Nat,?x:Nat),plus(min(0,?x:Nat),min(0,?y:Nat))),s(s(plus(?z:Nat,?x:Nat)))))",
"min(?z:Nat,s(min(s(plus(?z:Nat,?x:Nat)),min(min(0,?z:Nat),max(?x:Nat,?z:Nat))))) = max(?z:Nat,plus(max(max(plus(?y:Nat,?z:Nat),s(?x:Nat)),plus(plus(?x:Nat,?x:Nat),min(?x:Nat,?y:Nat))),?x:Nat))",
"plus(max(s(?z:Nat),plus(s(?x:Nat),0)),s(s(plus(plus(?y:Nat,?x:Nat),?y:Nat)))) = min(?x:Nat,max(s(min(0,?z:Nat)),plus(max(?x:Nat,min(?y:Nat,?x:Nat)),max(min(?x:Nat,0),0))))",
"s(plus(max(max(0,?z:Nat),0),plus(plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),?z:Nat))) = max(max(plus(max(plus(0,0),?z:Nat),min(0,s(?z:Nat))),s(max(?x:Nat,?x:Nat))),0)",
"plus(max(plus(plus(plus(?x:Nat,0),?x:Nat),?x:Nat),plus(plus(?x:Nat,?y:Nat),?z:Nat)),s(0)) = plus(?y:Nat,min(plus(?y:Nat,plus(0,?x:Nat)),plus(min(min(0,?y:Nat),?y:Nat),s(s(?y:Nat)))))",
"plus(plus(?x:Nat,s(plus(max(0,?y:Nat),0))),min(0,s(s(min(?y:Nat,0))))) = max(plus(plus(?z:Nat,?y:Nat),?y:Nat),plus(plus(s(?y:Nat),min(?x:Nat,plus(?y:Nat,?z:Nat))),s(0)))",
"s(max(plus(min(?x:Nat,min(?x:Nat,?y:Nat)),s(max(0,0))),max(s(?y:Nat),s(?x:Nat)))) = plus(min(max(plus(0,plus(?y:Nat,?z:Nat)),plus(?z:Nat,?y:Nat)),plus(?x:Nat,?x:Nat)),s(?x:Nat))",
"s(plus(plus(max(?z:Nat,?y:Nat),max(plus(0,?x:Nat),min(?y:Nat,?x:Nat))),max(s(0),0))) = max(max(max(s(?x:Nat),max(plus(0,?x:Nat),plus(0,?y:Nat))),?y:Nat),min(0,?x:Nat))",
"plus(?z:Nat,min(plus(plus(max(?y:Nat,?x:Nat),0),min(plus(?z:Nat,?z:Nat),max(0,?x:Nat))),?z:Nat)) = min(?x:Nat,plus(?x:Nat,plus(max(plus(?y:Nat,?z:Nat),s(?y:Nat)),min(s(?z:Nat),s(?z:Nat)))))",
"plus(s(?z:Nat),min(s(s(plus(?y:Nat,?z:Nat))),min(plus(?z:Nat,min(?x:Nat,?x:Nat)),s(?x:Nat)))) = plus(min(plus(s(?z:Nat),min(?z:Nat,?z:Nat)),max(0,max(?y:Nat,max(?x:Nat,?y:Nat)))),?x:Nat)",
"plus(plus(plus(s(?x:Nat),?z:Nat),?x:Nat),plus(?x:Nat,s(min(max(?z:Nat,?z:Nat),max(?z:Nat,?y:Nat))))) = max(plus(?y:Nat,s(?y:Nat)),plus(0,min(max(max(?x:Nat,?x:Nat),plus(0,?x:Nat)),?y:Nat)))",
"s(max(max(min(min(?z:Nat,?y:Nat),min(0,?x:Nat)),0),plus(?x:Nat,plus(s(?z:Nat),s(?y:Nat))))) = min(max(?z:Nat,plus(s(min(?y:Nat,?y:Nat)),plus(0,plus(?x:Nat,?y:Nat)))),s(?x:Nat))",
"s(min(plus(s(max(?x:Nat,?x:Nat)),plus(plus(0,?x:Nat),plus(?x:Nat,0))),min(s(0),?z:Nat))) = plus(plus(s(min(?z:Nat,?y:Nat)),s(plus(max(?z:Nat,0),min(?z:Nat,?x:Nat)))),0)",
"plus(min(s(?y:Nat),min(plus(0,plus(?z:Nat,?y:Nat)),plus(plus(?x:Nat,0),0))),max(?y:Nat,?x:Nat)) = min(min(min(s(min(?x:Nat,?x:Nat)),plus(s(0),s(?x:Nat))),s(?x:Nat)),?x:Nat)",
"plus(plus(plus(s(max(?z:Nat,?x:Nat)),min(?z:Nat,max(?z:Nat,?y:Nat))),s(s(s(?z:Nat)))),s(0)) = s(plus(min(plus(min(?z:Nat,?x:Nat),?y:Nat),?x:Nat),max(s(max(?z:Nat,?x:Nat)),?z:Nat)))",
"s(min(max(s(plus(?x:Nat,0)),max(?x:Nat,plus(?x:Nat,?z:Nat))),plus(?x:Nat,min(min(?y:Nat,?y:Nat),?x:Nat)))) = s(s(max(plus(?y:Nat,min(?y:Nat,?y:Nat)),plus(max(?x:Nat,?x:Nat),s(?x:Nat)))))",
"min(plus(s(min(?y:Nat,0)),s(min(0,s(?y:Nat)))),plus(?x:Nat,min(0,s(max(0,?y:Nat))))) = min(plus(s(s(?z:Nat)),max(plus(s(?z:Nat),?z:Nat),?x:Nat)),plus(?x:Nat,?x:Nat))",
"max(min(plus(?x:Nat,?z:Nat),plus(?x:Nat,max(min(?y:Nat,0),s(?z:Nat)))),s(plus(min(?x:Nat,s(?y:Nat)),0))) = min(s(s(?x:Nat)),max(min(min(min(?x:Nat,?x:Nat),?z:Nat),?y:Nat),?z:Nat))",
"plus(min(s(min(max(?x:Nat,?x:Nat),plus(0,?x:Nat))),s(0)),s(max(plus(plus(?x:Nat,?x:Nat),?x:Nat),?z:Nat))) = plus(min(max(plus(plus(?x:Nat,0),min(?y:Nat,?z:Nat)),?x:Nat),0),?x:Nat)",
"min(min(?x:Nat,s(max(s(0),?x:Nat))),min(s(0),min(plus(min(?y:Nat,?y:Nat),max(?x:Nat,?y:Nat)),s(?z:Nat)))) = plus(?x:Nat,max(max(s(min(?x:Nat,?x:Nat)),plus(?y:Nat,?x:Nat)),?x:Nat))",
"plus(?z:Nat,min(plus(plus(?x:Nat,?y:Nat),?z:Nat),min(max(plus(?z:Nat,0),min(?x:Nat,?x:Nat)),plus(s(?x:Nat),s(?z:Nat))))) = plus(max(plus(s(s(?x:Nat)),0),plus(0,s(?y:Nat))),?x:Nat)",
"plus(0,min(min(min(s(?y:Nat),0),max(s(0),min(?x:Nat,?z:Nat))),s(min(s(?x:Nat),plus(?x:Nat,0))))) = max(?z:Nat,s(plus(max(plus(?x:Nat,?x:Nat),max(?y:Nat,?x:Nat)),?x:Nat)))",
"plus(s(s(min(min(0,?x:Nat),0))),s(max(plus(max(?x:Nat,?y:Nat),s(?y:Nat)),plus(?z:Nat,plus(?z:Nat,?y:Nat))))) = plus(0,min(?z:Nat,max(s(?x:Nat),plus(0,plus(?y:Nat,0)))))",
"plus(max(s(?x:Nat),s(min(plus(0,0),plus(?z:Nat,?x:Nat)))),max(min(?x:Nat,s(plus(?x:Nat,?z:Nat))),s(?x:Nat))) = max(s(?z:Nat),plus(?y:Nat,s(max(plus(0,?x:Nat),s(?x:Nat)))))",
"s(max(min(plus(?x:Nat,max(0,?y:Nat)),plus(plus(?y:Nat,0),plus(?z:Nat,?y:Nat))),plus(plus(min(?x:Nat,0),s(?z:Nat)),?x:Nat))) = max(max(s(s(plus(?x:Nat,?x:Nat))),?x:Nat),s(?z:Nat))",
"plus(s(s(?y:Nat)),max(min(max(min(?x:Nat,?z:Nat),min(0,?y:Nat)),0),plus(plus(min(?x:Nat,?z:Nat),max(?x:Nat,?y:Nat)),?x:Nat))) = plus(min(?x:Nat,?y:Nat),max(min(0,s(?z:Nat)),0))",
"plus(plus(plus(max(?x:Nat,s(?y:Nat)),min(max(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat))),max(?x:Nat,max(min(?x:Nat,?y:Nat),plus(?x:Nat,?z:Nat)))),?y:Nat) = plus(s(0),max(s(?x:Nat),s(s(?z:Nat))))",
"plus(max(?z:Nat,max(plus(plus(?z:Nat,?y:Nat),min(0,0)),s(s(?z:Nat)))),plus(plus(max(?x:Nat,plus(?x:Nat,?z:Nat)),plus(?x:Nat,?y:Nat)),?x:Nat)) = s(plus(?x:Nat,s(plus(s(?z:Nat),?y:Nat))))",
"plus(min(s(?z:Nat),max(s(s(?y:Nat)),s(plus(?x:Nat,?y:Nat)))),max(?x:Nat,plus(min(?x:Nat,max(0,?x:Nat)),plus(?x:Nat,max(?x:Nat,?z:Nat))))) = plus(plus(?x:Nat,max(s(?y:Nat),?z:Nat)),0)",
"max(plus(plus(s(s(?z:Nat)),plus(s(?x:Nat),min(?x:Nat,?z:Nat))),?y:Nat),plus(min(?z:Nat,s(0)),plus(min(plus(?x:Nat,0),plus(?z:Nat,?y:Nat)),?x:Nat))) = plus(s(s(s(?y:Nat))),?x:Nat)",
"max(plus(plus(plus(0,?y:Nat),min(?x:Nat,?x:Nat)),0),max(max(s(min(?x:Nat,0)),min(s(?y:Nat),0)),max(plus(max(?y:Nat,?y:Nat),max(?y:Nat,?z:Nat)),0))) = s(plus(0,?z:Nat))",
"min(s(plus(max(s(0),?x:Nat),min(max(?x:Nat,?x:Nat),?y:Nat))),max(max(min(s(0),?x:Nat),min(plus(?y:Nat,?x:Nat),max(?y:Nat,0))),min(s(?x:Nat),?z:Nat))) = max(?y:Nat,s(?x:Nat))",
"s(plus(s(s(?x:Nat)),max(?x:Nat,?x:Nat))) = min(plus(plus(max(plus(?y:Nat,?y:Nat),plus(?x:Nat,?z:Nat)),plus(s(?x:Nat),?z:Nat)),plus(?y:Nat,s(min(0,0)))),s(min(s(s(?y:Nat)),?y:Nat)))",
"max(s(plus(s(max(0,?x:Nat)),?x:Nat)),s(?x:Nat)) = min(s(min(s(min(0,?x:Nat)),?y:Nat)),min(min(s(s(?x:Nat)),s(min(0,?x:Nat))),min(plus(min(?x:Nat,?x:Nat),?x:Nat),?x:Nat)))",
"plus(s(?z:Nat),plus(s(s(plus(?z:Nat,?z:Nat))),0)) = min(min(s(min(?z:Nat,plus(?x:Nat,?x:Nat))),min(?z:Nat,min(s(?x:Nat),?z:Nat))),plus(?z:Nat,s(plus(min(?z:Nat,?z:Nat),plus(0,?z:Nat)))))",
"s(max(max(s(s(0)),plus(min(?y:Nat,?y:Nat),?x:Nat)),0)) = max(?x:Nat,plus(min(max(plus(?y:Nat,?z:Nat),s(?x:Nat)),min(?y:Nat,max(?z:Nat,0))),min(?x:Nat,plus(min(?x:Nat,?x:Nat),?y:Nat))))",
"max(max(min(plus(?y:Nat,s(?z:Nat)),?y:Nat),?y:Nat),max(0,?y:Nat)) = min(s(0),min(s(max(?y:Nat,s(0))),plus(min(plus(?x:Nat,?x:Nat),plus(?z:Nat,?y:Nat)),plus(max(?z:Nat,?z:Nat),?x:Nat))))",
"min(?x:Nat,max(s(?z:Nat),plus(min(?y:Nat,?y:Nat),plus(?z:Nat,?y:Nat)))) = min(min(s(s(max(0,?x:Nat))),min(?x:Nat,plus(?x:Nat,min(0,?z:Nat)))),plus(plus(?x:Nat,s(plus(?x:Nat,?y:Nat))),0))",
"max(max(0,plus(plus(plus(?x:Nat,?z:Nat),min(0,0)),0)),0) = plus(plus(?z:Nat,?x:Nat),plus(min(s(?z:Nat),plus(?z:Nat,max(?x:Nat,?y:Nat))),s(min(plus(?x:Nat,?x:Nat),min(?z:Nat,?z:Nat)))))",
"plus(max(0,max(s(min(0,?x:Nat)),plus(?z:Nat,s(?y:Nat)))),?x:Nat) = max(plus(max(?x:Nat,?z:Nat),plus(plus(min(?x:Nat,?x:Nat),min(?z:Nat,?y:Nat)),?x:Nat)),plus(?z:Nat,max(s(s(?y:Nat)),0)))",
"s(s(min(plus(max(?x:Nat,?x:Nat),min(?z:Nat,?z:Nat)),max(?x:Nat,plus(?x:Nat,?x:Nat))))) = plus(plus(plus(?z:Nat,0),plus(plus(plus(?x:Nat,0),max(?z:Nat,?z:Nat)),max(?z:Nat,min(0,?x:Nat)))),?x:Nat)",
"s(plus(s(?x:Nat),plus(plus(plus(?x:Nat,0),min(0,?y:Nat)),plus(?y:Nat,?z:Nat)))) = max(0,min(min(plus(0,min(?x:Nat,?z:Nat)),s(s(0))),plus(plus(plus(?x:Nat,?x:Nat),?z:Nat),?x:Nat)))",
"max(plus(max(?x:Nat,plus(s(?y:Nat),plus(?z:Nat,?z:Nat))),plus(?x:Nat,s(?x:Nat))),?x:Nat) = max(max(min(s(0),plus(s(?x:Nat),s(?x:Nat))),max(?z:Nat,?x:Nat)),s(s(s(max(0,?z:Nat)))))",
"plus(s(s(plus(?z:Nat,s(0)))),s(min(plus(?y:Nat,max(?x:Nat,?z:Nat)),0))) = plus(plus(plus(?x:Nat,s(min(0,?y:Nat))),s(?y:Nat)),s(min(min(?z:Nat,s(0)),min(?y:Nat,?y:Nat))))",
"s(plus(plus(min(plus(?z:Nat,?y:Nat),plus(?z:Nat,0)),plus(?y:Nat,?x:Nat)),s(s(?y:Nat)))) = plus(max(min(0,?y:Nat),s(s(plus(0,?z:Nat)))),s(plus(plus(?x:Nat,s(?z:Nat)),s(?x:Nat))))",
"max(max(plus(?y:Nat,0),max(s(plus(?y:Nat,0)),plus(max(0,0),?z:Nat))),0) = max(s(plus(plus(?y:Nat,?x:Nat),max(plus(?y:Nat,?x:Nat),?x:Nat))),s(plus(s(min(?x:Nat,?y:Nat)),?y:Nat)))",
"plus(max(min(max(?x:Nat,?z:Nat),0),plus(plus(?y:Nat,?y:Nat),plus(max(?x:Nat,?y:Nat),?y:Nat))),?y:Nat) = min(?x:Nat,min(max(min(0,?x:Nat),max(?x:Nat,plus(?z:Nat,?z:Nat))),s(s(max(?y:Nat,?x:Nat)))))",
"s(min(plus(s(min(?x:Nat,0)),s(min(?x:Nat,?y:Nat))),s(plus(s(?x:Nat),plus(?y:Nat,?x:Nat))))) = s(min(0,min(max(max(?x:Nat,0),?x:Nat),plus(min(?z:Nat,?y:Nat),max(0,?z:Nat)))))",
"min(min(plus(plus(plus(0,?y:Nat),0),s(min(?y:Nat,?x:Nat))),s(max(?x:Nat,?z:Nat))),s(?x:Nat)) = min(plus(plus(?z:Nat,?y:Nat),?x:Nat),min(plus(?x:Nat,min(?z:Nat,plus(0,?x:Nat))),s(0)))",
"plus(s(plus(min(max(?z:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),max(?x:Nat,plus(?z:Nat,0)))),s(s(0))) = plus(min(0,plus(s(?z:Nat),plus(min(?z:Nat,?y:Nat),plus(0,?x:Nat)))),max(?x:Nat,0))",
"plus(min(?x:Nat,plus(s(max(0,?z:Nat)),plus(s(?x:Nat),?x:Nat))),s(s(min(0,s(0))))) = s(min(min(s(plus(?x:Nat,?z:Nat)),plus(?y:Nat,plus(0,?y:Nat))),s(s(s(?z:Nat)))))",
"plus(plus(s(s(min(?y:Nat,?x:Nat))),max(plus(?z:Nat,?y:Nat),min(s(?z:Nat),plus(?x:Nat,?y:Nat)))),?x:Nat) = plus(min(plus(?x:Nat,s(?y:Nat)),plus(min(min(0,?x:Nat),max(?y:Nat,?z:Nat)),?y:Nat)),?y:Nat)",
"min(max(?z:Nat,plus(plus(max(?x:Nat,?x:Nat),s(?z:Nat)),s(0))),max(0,min(?z:Nat,s(s(?x:Nat))))) = min(s(?z:Nat),plus(max(s(?x:Nat),max(min(0,?x:Nat),min(?x:Nat,?z:Nat))),?x:Nat))",
"max(s(?y:Nat),plus(min(min(plus(?x:Nat,?z:Nat),max(?x:Nat,?x:Nat)),?z:Nat),s(plus(max(?z:Nat,?z:Nat),s(?y:Nat))))) = s(max(min(max(?z:Nat,?y:Nat),max(plus(?z:Nat,?x:Nat),min(?y:Nat,?y:Nat))),?y:Nat))",
"plus(?z:Nat,max(plus(s(?x:Nat),plus(plus(?z:Nat,0),?x:Nat)),plus(0,plus(max(?x:Nat,?z:Nat),plus(?x:Nat,?x:Nat))))) = plus(0,plus(s(?z:Nat),min(plus(max(?x:Nat,0),plus(?z:Nat,?x:Nat)),0)))",
"plus(plus(?x:Nat,?x:Nat),plus(s(max(?x:Nat,0)),min(max(?x:Nat,s(?x:Nat)),max(plus(?x:Nat,?y:Nat),max(?x:Nat,?x:Nat))))) = max(s(s(?z:Nat)),s(s(plus(max(?x:Nat,0),max(?y:Nat,?x:Nat)))))",
"plus(min(min(?x:Nat,?y:Nat),s(plus(min(?y:Nat,?y:Nat),?x:Nat))),s(plus(plus(?y:Nat,plus(?x:Nat,?z:Nat)),plus(plus(?y:Nat,?x:Nat),?z:Nat)))) = plus(?x:Nat,s(plus(max(plus(?z:Nat,?x:Nat),s(?z:Nat)),?x:Nat)))",
"max(plus(min(max(plus(?x:Nat,?z:Nat),min(?y:Nat,0)),min(plus(?x:Nat,?z:Nat),max(?x:Nat,?z:Nat))),s(s(s(0)))),plus(?x:Nat,?z:Nat)) = max(?z:Nat,min(plus(s(?z:Nat),?x:Nat),plus(?x:Nat,?x:Nat)))",
"plus(s(plus(plus(?y:Nat,0),min(min(?x:Nat,?x:Nat),s(?x:Nat)))),s(plus(min(?z:Nat,s(?y:Nat)),plus(max(?x:Nat,?x:Nat),s(0))))) = min(min(0,s(plus(s(0),s(0)))),?x:Nat)",
"plus(min(plus(?x:Nat,?z:Nat),plus(s(?x:Nat),min(?z:Nat,min(0,?z:Nat)))),max(?x:Nat,min(s(min(?x:Nat,?y:Nat)),max(min(?x:Nat,?x:Nat),?z:Nat)))) = s(max(0,plus(min(?z:Nat,s(?x:Nat)),?x:Nat)))",
"min(plus(0,plus(min(?x:Nat,s(?z:Nat)),s(min(?z:Nat,?x:Nat)))),max(min(max(?z:Nat,min(?z:Nat,?x:Nat)),?z:Nat),plus(?z:Nat,s(plus(?y:Nat,?x:Nat))))) = max(plus(min(?z:Nat,s(?y:Nat)),?x:Nat),?x:Nat)",
"plus(max(plus(?x:Nat,max(plus(?x:Nat,?x:Nat),s(?x:Nat))),min(max(min(?x:Nat,0),?x:Nat),0)),s(min(min(?z:Nat,max(?x:Nat,?z:Nat)),s(?x:Nat)))) = plus(plus(min(s(?y:Nat),?x:Nat),?y:Nat),?x:Nat)",
"min(plus(plus(min(?x:Nat,plus(?x:Nat,?x:Nat)),?z:Nat),min(max(0,s(?y:Nat)),min(min(?x:Nat,?x:Nat),min(?x:Nat,?x:Nat)))),max(?z:Nat,s(s(plus(?y:Nat,?x:Nat))))) = min(0,plus(?x:Nat,s(0)))",
"max(max(min(plus(max(?y:Nat,?z:Nat),?y:Nat),plus(max(?x:Nat,0),min(?z:Nat,0))),max(plus(min(?y:Nat,?y:Nat),plus(?x:Nat,?x:Nat)),plus(max(?x:Nat,?z:Nat),?y:Nat))),?x:Nat) = plus(plus(?y:Nat,0),0)",
"plus(min(max(plus(?y:Nat,s(?x:Nat)),min(min(?z:Nat,0),plus(0,?x:Nat))),s(s(s(0)))),plus(min(s(s(?y:Nat)),plus(plus(0,?z:Nat),s(0))),s(?x:Nat))) = plus(?z:Nat,0)",
"plus(max(min(s(s(?y:Nat)),min(0,max(?x:Nat,?x:Nat))),max(s(?z:Nat),?x:Nat)),plus(max(max(s(?x:Nat),?x:Nat),min(s(?y:Nat),min(?x:Nat,?z:Nat))),min(max(?x:Nat,?x:Nat),?y:Nat))) = s(?x:Nat)",
"plus(?x:Nat,?x:Nat) = plus(plus(max(plus(plus(?z:Nat,?y:Nat),plus(?z:Nat,?z:Nat)),plus(s(0),?z:Nat)),max(?x:Nat,max(s(?z:Nat),?y:Nat))),s(plus(plus(s(0),plus(?y:Nat,?y:Nat)),min(s(?x:Nat),?x:Nat))))",
"s(s(s(s(s(0))))) = min(max(?x:Nat,plus(s(plus(?y:Nat,?x:Nat)),min(?x:Nat,?x:Nat))),min(plus(s(?y:Nat),plus(plus(?x:Nat,0),?x:Nat)),max(max(?x:Nat,?x:Nat),min(0,plus(?x:Nat,?x:Nat)))))",
"plus(?z:Nat,max(?x:Nat,plus(0,?x:Nat))) = plus(plus(?x:Nat,min(max(s(?x:Nat),?x:Nat),min(min(?y:Nat,?y:Nat),s(?z:Nat)))),min(plus(max(?y:Nat,plus(?y:Nat,?x:Nat)),max(s(?y:Nat),plus(?x:Nat,?z:Nat))),?x:Nat))",
"s(max(?x:Nat,s(max(s(?y:Nat),0)))) = min(plus(plus(min(s(?x:Nat),s(?y:Nat)),plus(max(?z:Nat,0),min(0,?x:Nat))),max(?z:Nat,s(?z:Nat))),min(min(?z:Nat,plus(?y:Nat,s(0))),0))",
"plus(plus(?y:Nat,s(s(plus(?x:Nat,?x:Nat)))),?x:Nat) = plus(max(plus(plus(?x:Nat,max(?x:Nat,0)),?x:Nat),plus(s(s(?z:Nat)),min(min(?x:Nat,?z:Nat),s(0)))),s(max(?z:Nat,s(min(?x:Nat,0)))))",
"s(max(max(min(s(?x:Nat),s(?x:Nat)),s(0)),s(?y:Nat))) = min(s(?z:Nat),plus(min(plus(s(?y:Nat),s(0)),s(plus(?x:Nat,?y:Nat))),max(s(s(?x:Nat)),plus(plus(0,?y:Nat),?y:Nat))))",
"s(plus(min(0,plus(plus(?z:Nat,?x:Nat),max(?z:Nat,0))),?z:Nat)) = min(max(s(max(?x:Nat,s(?y:Nat))),plus(s(?y:Nat),0)),plus(s(plus(?x:Nat,s(?z:Nat))),plus(?z:Nat,min(?x:Nat,s(?x:Nat)))))",
"max(min(?y:Nat,?x:Nat),plus(?x:Nat,plus(plus(?y:Nat,plus(?y:Nat,?y:Nat)),?y:Nat))) = min(?z:Nat,min(max(plus(min(?x:Nat,?z:Nat),min(?z:Nat,?y:Nat)),max(?y:Nat,min(?y:Nat,?y:Nat))),max(min(?y:Nat,s(?z:Nat)),0)))",
"s(plus(max(?y:Nat,s(s(?x:Nat))),max(s(plus(?x:Nat,?y:Nat)),s(plus(0,?y:Nat))))) = plus(?x:Nat,plus(min(max(0,min(?x:Nat,?z:Nat)),?y:Nat),max(0,plus(plus(?y:Nat,?y:Nat),min(?z:Nat,?y:Nat)))))",
"min(s(s(?x:Nat)),max(?y:Nat,min(min(plus(?z:Nat,?x:Nat),s(?x:Nat)),min(s(?x:Nat),?y:Nat)))) = s(min(plus(?x:Nat,s(plus(?z:Nat,?x:Nat))),min(min(plus(?x:Nat,?y:Nat),min(?x:Nat,?y:Nat)),s(?y:Nat))))",
"max(plus(plus(plus(s(?x:Nat),plus(?y:Nat,?x:Nat)),?y:Nat),s(plus(?y:Nat,s(?z:Nat)))),s(plus(?z:Nat,?y:Nat))) = max(?x:Nat,s(min(plus(?z:Nat,plus(?z:Nat,?x:Nat)),plus(max(?y:Nat,?y:Nat),min(?z:Nat,?x:Nat)))))",
"plus(min(?x:Nat,plus(?y:Nat,min(s(?z:Nat),s(?y:Nat)))),max(s(max(plus(0,?z:Nat),s(?x:Nat))),?x:Nat)) = plus(plus(plus(0,?x:Nat),?y:Nat),s(plus(s(min(?x:Nat,?z:Nat)),s(plus(?z:Nat,?z:Nat)))))",
"max(0,plus(plus(max(?x:Nat,min(?x:Nat,?y:Nat)),?x:Nat),min(s(plus(?x:Nat,?x:Nat)),min(?y:Nat,max(?y:Nat,?z:Nat))))) = min(?x:Nat,max(plus(s(?x:Nat),s(?y:Nat)),max(max(0,?z:Nat),min(?x:Nat,?z:Nat))))",
"plus(min(s(s(s(?x:Nat))),s(plus(s(?z:Nat),s(?y:Nat)))),min(?x:Nat,plus(?x:Nat,s(min(0,?x:Nat))))) = s(max(?x:Nat,max(min(s(?x:Nat),max(0,?x:Nat)),max(plus(0,?y:Nat),?z:Nat))))",
"plus(plus(?x:Nat,?x:Nat),plus(plus(max(plus(?y:Nat,?z:Nat),s(?x:Nat)),plus(?z:Nat,?x:Nat)),s(s(max(?x:Nat,?z:Nat))))) = s(max(min(?y:Nat,s(plus(0,0))),max(max(max(0,?z:Nat),?y:Nat),?y:Nat)))",
"max(max(?z:Nat,max(s(0),max(?x:Nat,plus(?z:Nat,?y:Nat)))),s(max(max(?y:Nat,plus(?x:Nat,0)),s(plus(?x:Nat,?y:Nat))))) = max(?x:Nat,max(?y:Nat,plus(?x:Nat,plus(plus(?x:Nat,?x:Nat),max(0,?z:Nat)))))",
"max(plus(max(min(0,max(0,?x:Nat)),plus(plus(?x:Nat,?x:Nat),plus(0,0))),plus(?x:Nat,s(plus(?z:Nat,0)))),0) = max(0,s(plus(max(?x:Nat,plus(0,?z:Nat)),max(?x:Nat,s(?x:Nat)))))",
"min(s(?x:Nat),min(max(?x:Nat,max(min(0,?y:Nat),min(?x:Nat,0))),min(min(max(0,?x:Nat),max(?x:Nat,?x:Nat)),?x:Nat))) = min(s(plus(0,?z:Nat)),s(plus(?z:Nat,min(?x:Nat,plus(?x:Nat,0)))))",
"plus(s(s(max(?z:Nat,?x:Nat))),plus(min(max(?y:Nat,?x:Nat),max(plus(?y:Nat,?y:Nat),plus(?x:Nat,?y:Nat))),plus(max(?y:Nat,?z:Nat),?z:Nat))) = plus(s(?z:Nat),plus(max(min(plus(?y:Nat,?z:Nat),?x:Nat),?x:Nat),?x:Nat))",
"plus(min(s(s(max(?x:Nat,0))),min(min(max(0,?y:Nat),?x:Nat),?y:Nat)),plus(max(plus(s(?y:Nat),plus(0,?x:Nat)),0),?x:Nat)) = min(plus(?z:Nat,?y:Nat),s(max(?x:Nat,s(max(?y:Nat,?y:Nat)))))",
"plus(min(?z:Nat,?z:Nat),plus(plus(plus(plus(?x:Nat,?z:Nat),min(?x:Nat,0)),min(s(?x:Nat),s(?x:Nat))),max(plus(plus(?x:Nat,?x:Nat),?z:Nat),0))) = min(?x:Nat,max(s(plus(0,plus(?x:Nat,?x:Nat))),0))",
"plus(plus(max(s(min(0,0)),plus(min(?y:Nat,?y:Nat),plus(?x:Nat,0))),s(?x:Nat)),plus(plus(?x:Nat,max(plus(?z:Nat,?x:Nat),?x:Nat)),?x:Nat)) = max(?x:Nat,min(?x:Nat,s(plus(s(?x:Nat),s(?y:Nat)))))",
"max(max(min(?z:Nat,plus(s(?x:Nat),min(0,?x:Nat))),plus(plus(?z:Nat,?z:Nat),plus(s(?z:Nat),?x:Nat))),plus(min(?x:Nat,max(max(?y:Nat,?x:Nat),?z:Nat)),0)) = plus(s(plus(?x:Nat,?x:Nat)),max(?z:Nat,?x:Nat))",
"plus(max(plus(min(plus(?z:Nat,0),s(?x:Nat)),s(s(0))),s(s(s(0)))),max(?x:Nat,max(plus(min(?z:Nat,0),max(?x:Nat,?y:Nat)),0))) = plus(?y:Nat,s(s(s(plus(0,?x:Nat)))))",
"plus(plus(max(max(s(0),max(?x:Nat,?z:Nat)),min(plus(?y:Nat,?y:Nat),max(?x:Nat,?y:Nat))),plus(min(min(?x:Nat,?x:Nat),s(?x:Nat)),plus(min(?x:Nat,0),min(?x:Nat,?x:Nat)))),?x:Nat) = s(min(?z:Nat,?z:Nat))",
"s(max(s(?x:Nat),plus(?y:Nat,?z:Nat))) = plus(plus(min(plus(?x:Nat,?y:Nat),s(plus(?z:Nat,?z:Nat))),max(plus(s(?z:Nat),?x:Nat),s(s(?x:Nat)))),max(?x:Nat,min(min(plus(?x:Nat,0),plus(?x:Nat,0)),?z:Nat)))",
"s(min(plus(s(?y:Nat),?z:Nat),?x:Nat)) = plus(s(plus(min(max(?x:Nat,?z:Nat),s(?y:Nat)),min(?x:Nat,s(?x:Nat)))),min(plus(0,plus(max(?x:Nat,0),min(?x:Nat,?x:Nat))),s(plus(max(?x:Nat,0),0))))",
"s(max(min(0,plus(min(?x:Nat,?x:Nat),s(?x:Nat))),?y:Nat)) = plus(s(?z:Nat),plus(max(plus(max(?y:Nat,?z:Nat),plus(?x:Nat,?y:Nat)),plus(plus(?x:Nat,0),s(?z:Nat))),min(?x:Nat,plus(max(?y:Nat,?x:Nat),0))))",
"max(max(s(max(plus(?x:Nat,?y:Nat),?z:Nat)),s(?x:Nat)),?z:Nat) = max(max(?z:Nat,?x:Nat),max(max(min(max(?y:Nat,0),?y:Nat),s(max(?z:Nat,?x:Nat))),min(max(?x:Nat,max(?x:Nat,?z:Nat)),s(plus(0,?x:Nat)))))",
"plus(?x:Nat,plus(0,s(max(min(?x:Nat,0),max(?x:Nat,?x:Nat))))) = min(s(plus(plus(s(?z:Nat),?x:Nat),plus(s(?x:Nat),plus(?x:Nat,?x:Nat)))),max(?x:Nat,max(s(max(?z:Nat,?z:Nat)),plus(s(0),?y:Nat))))",
"plus(max(plus(s(?x:Nat),min(0,s(?x:Nat))),s(?y:Nat)),?x:Nat) = plus(s(max(max(s(?x:Nat),max(?z:Nat,?y:Nat)),max(s(?z:Nat),plus(?y:Nat,?y:Nat)))),s(s(min(plus(0,0),plus(?x:Nat,0)))))",
"s(s(min(min(0,0),min(max(?y:Nat,?x:Nat),plus(?x:Nat,?z:Nat))))) = min(min(s(min(plus(0,?x:Nat),?y:Nat)),max(?x:Nat,s(max(?y:Nat,0)))),min(?x:Nat,plus(plus(?x:Nat,min(?x:Nat,?y:Nat)),?z:Nat)))",
"max(min(?z:Nat,s(0)),min(plus(plus(?z:Nat,s(?y:Nat)),?x:Nat),?z:Nat)) = plus(min(max(plus(plus(?y:Nat,?x:Nat),?x:Nat),max(?z:Nat,min(?z:Nat,0))),plus(plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,?y:Nat)),?y:Nat)),?y:Nat)",
"plus(plus(s(min(min(?z:Nat,?x:Nat),s(?y:Nat))),s(s(0))),s(?z:Nat)) = plus(plus(max(?x:Nat,?z:Nat),s(min(min(?x:Nat,?x:Nat),plus(?x:Nat,?y:Nat)))),plus(?x:Nat,max(s(?x:Nat),plus(?x:Nat,s(?x:Nat)))))",
"min(?x:Nat,s(max(plus(s(?y:Nat),?x:Nat),plus(plus(?z:Nat,?x:Nat),max(?x:Nat,0))))) = plus(min(min(?x:Nat,min(plus(0,?z:Nat),s(0))),max(max(?y:Nat,plus(0,0)),s(plus(?z:Nat,?z:Nat)))),?x:Nat)",
"min(max(plus(plus(s(?x:Nat),?x:Nat),?y:Nat),s(plus(?x:Nat,max(?x:Nat,0)))),?z:Nat) = min(max(min(plus(max(0,?x:Nat),plus(0,?x:Nat)),s(plus(?y:Nat,?y:Nat))),max(0,s(min(?z:Nat,0)))),?x:Nat)",
"plus(s(max(min(plus(?y:Nat,?z:Nat),min(?x:Nat,?y:Nat)),?x:Nat)),plus(?z:Nat,s(?y:Nat))) = max(max(?y:Nat,s(?x:Nat)),s(plus(min(max(?x:Nat,?z:Nat),max(?x:Nat,?y:Nat)),max(min(0,?x:Nat),plus(?y:Nat,?z:Nat)))))",
"max(s(?x:Nat),max(plus(max(?x:Nat,plus(?x:Nat,?x:Nat)),?z:Nat),s(min(s(?z:Nat),?y:Nat)))) = min(s(plus(plus(min(?x:Nat,?z:Nat),plus(?x:Nat,?y:Nat)),?y:Nat)),s(s(max(plus(0,0),max(?z:Nat,?y:Nat)))))",
"min(min(min(max(min(?x:Nat,0),max(?z:Nat,0)),?x:Nat),?z:Nat),plus(?z:Nat,s(?y:Nat))) = plus(max(?y:Nat,min(plus(s(?x:Nat),s(?x:Nat)),?z:Nat)),max(?x:Nat,s(plus(max(?y:Nat,?x:Nat),max(0,?x:Nat)))))",
"plus(?z:Nat,plus(s(?x:Nat),min(s(min(0,0)),min(plus(?x:Nat,0),s(0))))) = plus(max(max(plus(?z:Nat,s(?x:Nat)),min(0,max(?x:Nat,?z:Nat))),0),plus(?x:Nat,max(?x:Nat,s(s(?x:Nat)))))",
"s(max(min(min(s(0),?x:Nat),plus(?y:Nat,max(0,?y:Nat))),s(s(min(?y:Nat,?z:Nat))))) = s(max(max(?y:Nat,plus(plus(?y:Nat,?x:Nat),s(0))),max(?y:Nat,max(min(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat)))))",
"s(min(max(min(s(?z:Nat),plus(?x:Nat,?x:Nat)),min(0,0)),s(s(min(0,?x:Nat))))) = plus(max(min(0,max(s(?x:Nat),s(?x:Nat))),min(?x:Nat,plus(max(?x:Nat,?x:Nat),max(0,?z:Nat)))),?x:Nat)",
"max(min(max(s(?y:Nat),min(?z:Nat,min(?y:Nat,?x:Nat))),min(plus(?y:Nat,s(?y:Nat)),?z:Nat)),?z:Nat) = s(min(plus(plus(0,?x:Nat),plus(plus(0,?x:Nat),s(0))),min(min(?y:Nat,?x:Nat),s(s(0)))))",
"plus(min(min(max(min(?z:Nat,?x:Nat),0),min(min(?x:Nat,0),s(?x:Nat))),?y:Nat),s(?z:Nat)) = max(min(min(s(plus(?z:Nat,?x:Nat)),s(max(?x:Nat,?y:Nat))),max(?x:Nat,min(?y:Nat,plus(?z:Nat,?y:Nat)))),0)",
"max(?x:Nat,plus(plus(min(s(?y:Nat),min(?y:Nat,?x:Nat)),?y:Nat),min(max(plus(?z:Nat,?x:Nat),?x:Nat),?y:Nat))) = min(min(max(?y:Nat,?x:Nat),s(s(min(?y:Nat,0)))),s(s(plus(min(0,?x:Nat),s(?z:Nat)))))",
"min(max(min(max(max(?x:Nat,?z:Nat),?x:Nat),s(max(?z:Nat,?x:Nat))),?x:Nat),s(max(s(?x:Nat),?x:Nat))) = min(min(?z:Nat,?z:Nat),plus(min(min(s(?z:Nat),plus(0,0)),plus(?y:Nat,s(?z:Nat))),s(?x:Nat)))",
"plus(s(max(?y:Nat,s(0))),max(plus(s(s(?x:Nat)),s(plus(?z:Nat,?z:Nat))),s(s(s(?y:Nat))))) = plus(plus(max(min(plus(?x:Nat,?x:Nat),?z:Nat),min(?z:Nat,?x:Nat)),s(plus(?x:Nat,s(?y:Nat)))),?z:Nat)",
"plus(min(plus(plus(0,plus(0,?x:Nat)),plus(s(?x:Nat),s(?x:Nat))),s(s(?x:Nat))),plus(?x:Nat,?y:Nat)) = max(plus(plus(plus(s(0),?y:Nat),s(?z:Nat)),max(plus(?z:Nat,s(?z:Nat)),?z:Nat)),s(0))",
"s(max(plus(plus(s(?x:Nat),plus(?y:Nat,?y:Nat)),plus(min(?y:Nat,?x:Nat),?y:Nat)),s(max(plus(?z:Nat,?x:Nat),?y:Nat)))) = plus(?x:Nat,plus(min(max(plus(0,?x:Nat),max(0,0)),s(plus(0,?x:Nat))),?z:Nat))",
"max(plus(s(min(?z:Nat,plus(?z:Nat,?y:Nat))),?y:Nat),min(plus(plus(?z:Nat,max(?x:Nat,0)),s(max(?x:Nat,?y:Nat))),0)) = plus(min(?x:Nat,s(max(0,s(?z:Nat)))),s(s(plus(?z:Nat,plus(?x:Nat,?x:Nat)))))",
"min(max(plus(max(?y:Nat,?z:Nat),s(?z:Nat)),?x:Nat),plus(plus(?y:Nat,?y:Nat),s(plus(min(?y:Nat,?y:Nat),plus(?x:Nat,?x:Nat))))) = min(plus(max(plus(max(?x:Nat,?z:Nat),s(0)),?x:Nat),s(max(?x:Nat,?y:Nat))),?y:Nat)",
"min(plus(min(s(?y:Nat),s(?x:Nat)),plus(?z:Nat,plus(plus(0,?x:Nat),plus(?x:Nat,0)))),plus(max(?y:Nat,?y:Nat),0)) = min(?x:Nat,plus(plus(plus(max(?y:Nat,?x:Nat),s(?x:Nat)),s(min(?x:Nat,0))),0))",
"plus(min(min(?x:Nat,max(min(0,0),s(?z:Nat))),?x:Nat),max(s(max(?y:Nat,?x:Nat)),plus(?x:Nat,s(s(?x:Nat))))) = plus(?z:Nat,plus(min(plus(max(?y:Nat,?y:Nat),0),plus(s(0),s(?x:Nat))),?y:Nat))",
"plus(plus(?z:Nat,s(min(plus(?x:Nat,0),?x:Nat))),max(plus(?y:Nat,0),min(max(plus(?z:Nat,?x:Nat),s(?x:Nat)),?x:Nat))) = s(plus(?x:Nat,max(plus(min(?z:Nat,?x:Nat),s(?z:Nat)),plus(?x:Nat,min(?z:Nat,?x:Nat)))))",
"plus(plus(s(?z:Nat),min(s(plus(?y:Nat,?y:Nat)),?z:Nat)),s(plus(plus(min(?x:Nat,?z:Nat),plus(?x:Nat,?x:Nat)),min(?z:Nat,s(0))))) = plus(max(?x:Nat,?x:Nat),plus(s(0),s(plus(0,plus(?x:Nat,?y:Nat)))))",
"min(min(plus(s(plus(?x:Nat,?y:Nat)),plus(max(?x:Nat,0),min(?z:Nat,?y:Nat))),min(s(min(?x:Nat,0)),plus(plus(0,0),?x:Nat))),0) = plus(?x:Nat,plus(?x:Nat,s(max(plus(?x:Nat,?y:Nat),s(0)))))",
"plus(max(s(s(max(?x:Nat,?y:Nat))),s(max(?z:Nat,0))),max(plus(?y:Nat,?z:Nat),plus(s(min(?x:Nat,?x:Nat)),plus(?z:Nat,plus(?y:Nat,0))))) = max(?x:Nat,min(min(plus(0,?y:Nat),min(?y:Nat,?z:Nat)),?y:Nat))",
"plus(min(max(plus(s(?x:Nat),?z:Nat),max(?z:Nat,min(0,0))),plus(min(s(?y:Nat),s(?y:Nat)),s(?z:Nat))),plus(0,plus(max(?x:Nat,?y:Nat),plus(?x:Nat,max(?x:Nat,?x:Nat))))) = s(s(min(0,?x:Nat)))",
"min(s(min(?y:Nat,plus(plus(?z:Nat,?z:Nat),s(?y:Nat)))),min(plus(?x:Nat,plus(plus(?y:Nat,?x:Nat),plus(?z:Nat,?x:Nat))),min(min(plus(?x:Nat,?y:Nat),?z:Nat),max(plus(?x:Nat,?x:Nat),s(?y:Nat))))) = s(s(s(?x:Nat)))",
"max(max(s(max(?y:Nat,min(?z:Nat,?x:Nat))),max(max(?z:Nat,plus(?x:Nat,?z:Nat)),plus(plus(0,?z:Nat),s(?x:Nat)))),min(?z:Nat,min(max(s(0),max(?x:Nat,0)),s(plus(?z:Nat,?x:Nat))))) = plus(?y:Nat,?z:Nat)",
"max(max(s(max(max(?z:Nat,?x:Nat),?z:Nat)),min(plus(s(?y:Nat),max(?z:Nat,0)),max(s(?x:Nat),max(?z:Nat,?x:Nat)))),max(min(plus(plus(?x:Nat,?x:Nat),?x:Nat),plus(max(?x:Nat,?x:Nat),?y:Nat)),?z:Nat)) = s(?x:Nat)",
"min(plus(max(plus(max(?x:Nat,?z:Nat),?x:Nat),plus(plus(?x:Nat,?y:Nat),s(?y:Nat))),plus(plus(plus(?z:Nat,?x:Nat),plus(?x:Nat,?z:Nat)),min(?z:Nat,s(?y:Nat)))),plus(?x:Nat,s(min(plus(?z:Nat,0),?x:Nat)))) = s(?x:Nat)",
"max(s(0),0) = plus(min(plus(plus(s(0),s(?y:Nat)),s(plus(?x:Nat,?x:Nat))),min(?z:Nat,?z:Nat)),plus(s(max(plus(?x:Nat,0),?z:Nat)),max(max(max(?y:Nat,?x:Nat),min(?y:Nat,?x:Nat)),min(?z:Nat,?z:Nat))))",
"s(plus(?x:Nat,s(s(?x:Nat)))) = max(min(s(plus(?x:Nat,?y:Nat)),max(s(s(0)),plus(?x:Nat,s(?x:Nat)))),min(max(s(max(?x:Nat,?x:Nat)),max(s(?x:Nat),s(?x:Nat))),max(s(min(0,?z:Nat)),?x:Nat)))",
"min(s(max(0,min(?z:Nat,?y:Nat))),s(?x:Nat)) = min(plus(?z:Nat,?y:Nat),min(max(plus(max(0,?x:Nat),s(0)),min(plus(0,0),?x:Nat)),plus(plus(s(0),min(?z:Nat,0)),s(min(?z:Nat,?x:Nat)))))",
"s(max(?y:Nat,min(plus(0,plus(?z:Nat,?y:Nat)),?x:Nat))) = max(min(min(plus(s(?y:Nat),?y:Nat),0),plus(s(s(?z:Nat)),s(?y:Nat))),s(max(plus(s(0),s(?z:Nat)),min(s(?z:Nat),max(?x:Nat,?x:Nat)))))",
"plus(s(s(plus(s(?x:Nat),s(?x:Nat)))),s(?x:Nat)) = min(plus(max(s(min(?z:Nat,?x:Nat)),plus(max(?x:Nat,?x:Nat),min(?z:Nat,0))),max(s(0),?y:Nat)),max(?z:Nat,plus(?x:Nat,min(?y:Nat,min(0,0)))))",
"min(?z:Nat,plus(plus(s(s(?x:Nat)),s(?x:Nat)),s(?x:Nat))) = s(plus(max(plus(max(?x:Nat,?x:Nat),max(?z:Nat,0)),min(max(0,?y:Nat),s(0))),plus(s(max(?x:Nat,?z:Nat)),plus(?x:Nat,min(0,0)))))",
"plus(?z:Nat,plus(min(max(min(0,?y:Nat),?x:Nat),?y:Nat),?x:Nat)) = plus(min(0,max(plus(min(?x:Nat,?x:Nat),plus(?y:Nat,?z:Nat)),plus(plus(0,?y:Nat),min(?z:Nat,?x:Nat)))),s(plus(s(0),s(min(?x:Nat,?x:Nat)))))",
"s(s(min(s(min(0,?y:Nat)),min(s(?x:Nat),plus(?y:Nat,?x:Nat))))) = max(min(min(plus(s(?z:Nat),s(?y:Nat)),s(max(?x:Nat,?x:Nat))),max(plus(plus(?y:Nat,?x:Nat),s(?y:Nat)),s(0))),plus(?x:Nat,?z:Nat))",
"s(plus(max(max(plus(0,?x:Nat),?x:Nat),s(min(?x:Nat,?x:Nat))),?y:Nat)) = min(plus(?z:Nat,plus(s(s(?x:Nat)),s(0))),plus(max(0,max(s(?y:Nat),?x:Nat)),s(plus(plus(?x:Nat,0),plus(?x:Nat,?x:Nat)))))",
"max(?x:Nat,plus(plus(?z:Nat,plus(max(?x:Nat,?y:Nat),plus(?x:Nat,0))),min(0,s(?x:Nat)))) = max(min(s(plus(min(0,?x:Nat),min(?x:Nat,?x:Nat))),?y:Nat),s(plus(max(max(?z:Nat,?x:Nat),max(?x:Nat,0)),0)))",
"plus(max(s(max(?y:Nat,plus(?y:Nat,?z:Nat))),s(plus(s(?y:Nat),max(?x:Nat,?z:Nat)))),?x:Nat) = min(max(0,plus(?y:Nat,s(?y:Nat))),s(min(min(?z:Nat,min(?x:Nat,?x:Nat)),min(plus(?x:Nat,?x:Nat),min(?z:Nat,?y:Nat)))))",
"s(min(min(min(min(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),s(plus(0,?x:Nat))),max(?y:Nat,?x:Nat))) = plus(?y:Nat,max(s(max(plus(?z:Nat,?x:Nat),min(?z:Nat,?z:Nat))),plus(max(max(?x:Nat,?x:Nat),?x:Nat),plus(0,0))))",
"s(min(plus(?y:Nat,s(min(?x:Nat,0))),min(?x:Nat,plus(plus(?y:Nat,?x:Nat),max(0,?x:Nat))))) = min(plus(max(plus(plus(?x:Nat,0),?y:Nat),plus(?y:Nat,plus(?x:Nat,?z:Nat))),max(max(s(?y:Nat),?x:Nat),?x:Nat)),0)",
"min(plus(min(?y:Nat,s(plus(?x:Nat,?x:Nat))),s(?z:Nat)),plus(max(?x:Nat,s(0)),s(?z:Nat))) = min(s(?y:Nat),plus(plus(?y:Nat,plus(plus(?x:Nat,?y:Nat),min(?z:Nat,?y:Nat))),s(plus(s(?x:Nat),max(?z:Nat,?y:Nat)))))",
"plus(plus(s(plus(s(?z:Nat),plus(?y:Nat,?x:Nat))),plus(?z:Nat,min(?z:Nat,min(?x:Nat,?z:Nat)))),?y:Nat) = plus(min(s(min(plus(?y:Nat,?y:Nat),max(?z:Nat,?y:Nat))),s(?x:Nat)),max(?y:Nat,plus(s(plus(?x:Nat,?x:Nat)),?z:Nat)))",
"plus(max(plus(?x:Nat,?y:Nat),max(plus(0,plus(?z:Nat,?y:Nat)),?z:Nat)),min(s(0),max(?y:Nat,?x:Nat))) = max(s(min(min(s(?x:Nat),?x:Nat),?z:Nat)),max(?z:Nat,max(plus(plus(?x:Nat,?x:Nat),max(?y:Nat,?z:Nat)),?y:Nat)))",
"max(min(?x:Nat,s(0)),plus(?z:Nat,max(plus(min(?y:Nat,?x:Nat),?z:Nat),plus(min(?x:Nat,0),s(?x:Nat))))) = s(plus(min(min(s(?x:Nat),plus(?x:Nat,?y:Nat)),?x:Nat),min(plus(?x:Nat,max(?z:Nat,0)),s(?x:Nat))))",
"max(min(s(min(max(?x:Nat,?y:Nat),?y:Nat)),s(plus(min(?z:Nat,?x:Nat),min(?x:Nat,?z:Nat)))),plus(?y:Nat,?x:Nat)) = s(max(plus(s(min(?x:Nat,?y:Nat)),plus(plus(?y:Nat,?x:Nat),?y:Nat)),plus(?y:Nat,s(plus(?z:Nat,?x:Nat)))))",
"plus(?x:Nat,min(max(plus(?z:Nat,?z:Nat),plus(max(?x:Nat,0),min(?x:Nat,?y:Nat))),s(plus(plus(0,?y:Nat),plus(0,?x:Nat))))) = plus(min(plus(plus(s(?x:Nat),s(0)),s(s(?y:Nat))),plus(0,?x:Nat)),?z:Nat)",
"s(max(plus(max(?y:Nat,min(?x:Nat,0)),plus(s(?x:Nat),min(?y:Nat,?x:Nat))),plus(min(0,plus(0,?y:Nat)),s(s(?y:Nat))))) = s(max(plus(plus(?z:Nat,?y:Nat),s(?x:Nat)),min(?x:Nat,max(s(?z:Nat),?x:Nat))))",
"max(?z:Nat,plus(plus(s(s(?y:Nat)),s(max(?y:Nat,?y:Nat))),min(min(max(?x:Nat,0),min(?y:Nat,?y:Nat)),s(plus(?z:Nat,0))))) = min(max(max(min(plus(?x:Nat,?x:Nat),max(0,0)),s(?y:Nat)),0),?x:Nat)",
"min(plus(s(min(max(?x:Nat,?z:Nat),?z:Nat)),min(plus(min(0,0),s(?x:Nat)),plus(?y:Nat,plus(0,?x:Nat)))),plus(?y:Nat,?z:Nat)) = min(min(?y:Nat,?x:Nat),s(min(plus(max(?z:Nat,?x:Nat),?x:Nat),plus(0,?x:Nat))))",
"max(plus(min(plus(plus(?z:Nat,?x:Nat),?x:Nat),s(min(?y:Nat,0))),min(min(s(?y:Nat),max(?x:Nat,?z:Nat)),plus(s(?x:Nat),?y:Nat))),0) = s(plus(plus(?z:Nat,?z:Nat),s(max(min(?x:Nat,?x:Nat),max(0,?x:Nat)))))",
"min(s(max(max(?x:Nat,plus(?z:Nat,?y:Nat)),?y:Nat)),max(s(s(min(0,?z:Nat))),plus(plus(max(?x:Nat,?x:Nat),min(0,0)),?x:Nat))) = s(max(max(plus(s(?x:Nat),0),plus(?y:Nat,plus(?x:Nat,?x:Nat))),0))",
"plus(max(plus(?x:Nat,min(s(0),s(?z:Nat))),plus(?x:Nat,?y:Nat)),plus(min(plus(s(?x:Nat),?x:Nat),plus(min(?x:Nat,?x:Nat),?x:Nat)),?y:Nat)) = plus(s(0),plus(?z:Nat,max(min(max(?x:Nat,?y:Nat),?x:Nat),s(0))))",
"max(max(s(max(s(?x:Nat),s(?y:Nat))),plus(s(?z:Nat),s(plus(0,?y:Nat)))),s(plus(plus(plus(?x:Nat,?y:Nat),s(?x:Nat)),min(?x:Nat,s(?x:Nat))))) = s(max(s(s(?x:Nat)),plus(max(?y:Nat,0),?x:Nat)))",
"plus(min(?x:Nat,0),plus(plus(plus(?x:Nat,min(?x:Nat,0)),max(max(?x:Nat,?x:Nat),max(?z:Nat,0))),plus(0,max(min(?x:Nat,?x:Nat),plus(?z:Nat,?z:Nat))))) = min(min(max(?y:Nat,s(plus(?x:Nat,?x:Nat))),?y:Nat),0)",
"min(plus(min(plus(?x:Nat,s(?y:Nat)),max(?x:Nat,?z:Nat)),plus(max(?x:Nat,0),plus(s(?z:Nat),?x:Nat))),min(max(?x:Nat,plus(max(?x:Nat,?x:Nat),s(?z:Nat))),0)) = s(min(?y:Nat,s(max(min(0,?x:Nat),0))))",
"plus(max(plus(s(plus(?z:Nat,?x:Nat)),?x:Nat),min(min(plus(0,?x:Nat),max(?x:Nat,?y:Nat)),max(?x:Nat,s(?x:Nat)))),min(?x:Nat,min(0,s(max(?z:Nat,?x:Nat))))) = plus(s(s(min(plus(?x:Nat,0),0))),?y:Nat)",
"plus(plus(?x:Nat,plus(min(max(0,0),s(0)),max(plus(?y:Nat,0),s(?x:Nat)))),plus(?y:Nat,plus(min(min(?y:Nat,?y:Nat),min(?x:Nat,?x:Nat)),s(0)))) = s(plus(?x:Nat,max(?z:Nat,s(min(?z:Nat,?z:Nat)))))",
"plus(plus(0,min(plus(?x:Nat,s(?y:Nat)),max(s(?y:Nat),min(?z:Nat,?y:Nat)))),plus(plus(s(max(0,?x:Nat)),plus(?z:Nat,?x:Nat)),max(plus(?y:Nat,max(?x:Nat,?z:Nat)),plus(plus(?x:Nat,?x:Nat),?y:Nat)))) = plus(?z:Nat,?x:Nat)",
"s(s(max(plus(plus(?z:Nat,?z:Nat),?x:Nat),?x:Nat))) = plus(max(s(s(min(?z:Nat,0))),plus(max(min(?x:Nat,?z:Nat),min(?z:Nat,0)),s(s(?x:Nat)))),max(min(s(plus(?x:Nat,?x:Nat)),min(?z:Nat,s(?z:Nat))),?z:Nat))",
"s(s(max(s(plus(?y:Nat,?x:Nat)),max(0,0)))) = max(min(plus(?x:Nat,s(?y:Nat)),s(min(plus(?x:Nat,?z:Nat),?x:Nat))),max(plus(plus(?x:Nat,plus(?x:Nat,?z:Nat)),?y:Nat),s(min(min(?y:Nat,?z:Nat),min(?x:Nat,0)))))",
"max(plus(0,min(s(s(?y:Nat)),s(?y:Nat))),?x:Nat) = plus(max(?x:Nat,min(min(?z:Nat,plus(?z:Nat,?x:Nat)),plus(min(0,?x:Nat),min(?x:Nat,0)))),max(max(s(max(?z:Nat,0)),max(plus(?x:Nat,?x:Nat),?z:Nat)),?z:Nat))",
"max(max(min(0,plus(plus(?x:Nat,?z:Nat),0)),?x:Nat),s(?y:Nat)) = max(s(?y:Nat),min(max(plus(s(?z:Nat),0),plus(min(0,0),s(0))),plus(s(s(?y:Nat)),plus(min(?y:Nat,?y:Nat),plus(?x:Nat,?z:Nat)))))",
"plus(min(?x:Nat,?x:Nat),max(max(?z:Nat,?y:Nat),plus(max(?x:Nat,0),?y:Nat))) = min(max(s(plus(min(?z:Nat,?x:Nat),plus(?z:Nat,0))),s(?x:Nat)),min(?x:Nat,plus(min(max(?x:Nat,0),?x:Nat),min(plus(?x:Nat,?x:Nat),0))))",
"max(s(min(s(max(?x:Nat,0)),s(s(?x:Nat)))),s(plus(?x:Nat,?z:Nat))) = max(max(plus(plus(min(?x:Nat,?x:Nat),0),s(min(?y:Nat,?x:Nat))),?x:Nat),plus(plus(s(max(?z:Nat,?x:Nat)),s(min(?x:Nat,?x:Nat))),?y:Nat))",
"min(?z:Nat,max(?x:Nat,min(plus(plus(?x:Nat,?y:Nat),plus(?z:Nat,?y:Nat)),s(max(?z:Nat,?x:Nat))))) = max(min(plus(s(s(?y:Nat)),plus(?x:Nat,?z:Nat)),plus(?y:Nat,s(min(?x:Nat,?x:Nat)))),s(s(plus(max(?y:Nat,0),0))))",
"s(plus(max(s(plus(?x:Nat,?x:Nat)),plus(min(?y:Nat,?z:Nat),0)),plus(max(?x:Nat,?z:Nat),?x:Nat))) = s(plus(s(min(min(?x:Nat,0),min(?x:Nat,?x:Nat))),max(max(max(?x:Nat,?x:Nat),min(?x:Nat,?y:Nat)),s(s(?z:Nat)))))",
"max(min(?y:Nat,?x:Nat),s(min(s(plus(?y:Nat,?x:Nat)),min(min(?y:Nat,0),min(0,?x:Nat))))) = plus(s(s(s(plus(?y:Nat,?x:Nat)))),plus(?x:Nat,max(s(min(?y:Nat,?y:Nat)),plus(max(0,?y:Nat),plus(0,?x:Nat)))))",
"min(s(?y:Nat),plus(?x:Nat,min(plus(min(?z:Nat,?z:Nat),max(?x:Nat,?x:Nat)),min(?x:Nat,s(?y:Nat))))) = min(s(?x:Nat),plus(s(min(max(?x:Nat,0),s(?z:Nat))),plus(plus(s(?x:Nat),s(?z:Nat)),s(max(?x:Nat,0)))))",
"s(min(s(s(0)),plus(max(max(?x:Nat,?x:Nat),plus(?z:Nat,?x:Nat)),plus(s(?z:Nat),plus(?x:Nat,?x:Nat))))) = plus(plus(plus(?z:Nat,s(max(?z:Nat,0))),s(plus(plus(0,?x:Nat),max(?z:Nat,?z:Nat)))),min(?x:Nat,?x:Nat))",
"s(plus(plus(max(?z:Nat,min(?z:Nat,?x:Nat)),?z:Nat),plus(plus(plus(?x:Nat,?z:Nat),s(?x:Nat)),plus(s(?x:Nat),?z:Nat)))) = min(max(s(min(min(?x:Nat,?x:Nat),s(?y:Nat))),s(plus(plus(?x:Nat,?x:Nat),max(0,?x:Nat)))),?y:Nat)",
"max(?x:Nat,min(max(plus(min(?x:Nat,?x:Nat),s(?x:Nat)),0),plus(plus(max(0,?x:Nat),min(?z:Nat,?z:Nat)),?x:Nat))) = min(max(0,?x:Nat),plus(max(max(s(?y:Nat),?x:Nat),?x:Nat),plus(?x:Nat,plus(min(?z:Nat,?x:Nat),?x:Nat))))",
"s(max(plus(s(s(?z:Nat)),s(plus(?z:Nat,?y:Nat))),plus(plus(?x:Nat,plus(?z:Nat,?z:Nat)),plus(?x:Nat,plus(0,?x:Nat))))) = min(?y:Nat,max(min(plus(plus(?y:Nat,?x:Nat),plus(0,?y:Nat)),plus(?y:Nat,min(?x:Nat,?z:Nat))),?x:Nat))",
"max(?z:Nat,plus(min(max(plus(?y:Nat,?x:Nat),plus(0,0)),min(?y:Nat,?x:Nat)),min(max(0,s(?x:Nat)),s(?x:Nat)))) = plus(max(min(?x:Nat,min(s(?x:Nat),0)),plus(?z:Nat,?y:Nat)),plus(max(?x:Nat,s(?x:Nat)),0))",
"max(plus(s(0),plus(s(max(?x:Nat,?y:Nat)),?x:Nat)),min(min(plus(plus(?x:Nat,?z:Nat),plus(0,?x:Nat)),?z:Nat),0)) = plus(?z:Nat,min(s(min(max(?x:Nat,?y:Nat),s(?x:Nat))),s(plus(min(?y:Nat,?x:Nat),s(?x:Nat)))))",
"min(plus(s(min(s(?x:Nat),plus(0,0))),max(plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?z:Nat)),s(plus(?x:Nat,0)))),?y:Nat) = min(min(plus(max(s(?x:Nat),plus(?x:Nat,?y:Nat)),?x:Nat),max(s(s(?y:Nat)),?z:Nat)),?y:Nat)",
"plus(s(min(max(?z:Nat,plus(?x:Nat,?z:Nat)),plus(?x:Nat,0))),max(?x:Nat,plus(?y:Nat,min(plus(?z:Nat,?y:Nat),min(?z:Nat,?y:Nat))))) = max(plus(?x:Nat,min(plus(plus(?x:Nat,?y:Nat),s(?x:Nat)),max(s(0),s(?x:Nat)))),0)",
"min(min(plus(s(max(?y:Nat,?z:Nat)),s(plus(0,?z:Nat))),0),s(plus(plus(s(?x:Nat),0),min(s(0),s(?z:Nat))))) = min(s(?x:Nat),max(plus(plus(s(?y:Nat),0),plus(plus(?z:Nat,?x:Nat),?x:Nat)),?y:Nat))",
"min(0,min(plus(s(min(?x:Nat,?x:Nat)),s(plus(0,?y:Nat))),plus(plus(max(0,?y:Nat),?x:Nat),min(plus(?y:Nat,?z:Nat),min(?x:Nat,?x:Nat))))) = plus(plus(?y:Nat,?x:Nat),s(s(max(min(?y:Nat,0),min(?x:Nat,?y:Nat)))))",
"plus(min(s(s(?x:Nat)),plus(max(min(?z:Nat,0),?x:Nat),s(min(0,?x:Nat)))),s(max(s(max(0,?x:Nat)),min(?z:Nat,s(?x:Nat))))) = s(plus(s(s(s(?x:Nat))),min(s(?z:Nat),plus(?z:Nat,s(?z:Nat)))))",
"plus(plus(min(?y:Nat,?y:Nat),max(plus(?z:Nat,?x:Nat),max(plus(?y:Nat,?x:Nat),max(?y:Nat,?y:Nat)))),s(max(?y:Nat,plus(plus(0,0),max(?x:Nat,?y:Nat))))) = min(?z:Nat,plus(0,plus(min(plus(?z:Nat,?y:Nat),s(0)),?z:Nat)))",
"s(s(?z:Nat)) = plus(min(min(?x:Nat,min(?y:Nat,0)),max(?x:Nat,s(plus(?x:Nat,?x:Nat)))),plus(plus(max(plus(?x:Nat,0),?y:Nat),s(s(?z:Nat))),max(max(min(?x:Nat,?z:Nat),?x:Nat),min(plus(?y:Nat,?z:Nat),plus(?x:Nat,?z:Nat)))))",
"plus(?x:Nat,?y:Nat) = max(max(min(s(?y:Nat),min(?z:Nat,plus(?y:Nat,?x:Nat))),max(plus(min(?y:Nat,0),?x:Nat),?x:Nat)),max(plus(plus(?z:Nat,?x:Nat),plus(max(?y:Nat,0),?x:Nat)),plus(min(min(?y:Nat,0),plus(?x:Nat,?x:Nat)),0)))",
"min(?x:Nat,plus(s(?y:Nat),?x:Nat)) = plus(max(min(s(min(?x:Nat,?x:Nat)),min(min(?z:Nat,?x:Nat),?x:Nat)),plus(0,?x:Nat)),min(min(s(min(?y:Nat,?z:Nat)),plus(s(0),max(?x:Nat,?x:Nat))),plus(min(?x:Nat,s(?x:Nat)),?y:Nat)))",
"s(s(s(min(0,plus(?x:Nat,0))))) = max(plus(s(?x:Nat),min(s(?y:Nat),s(plus(?x:Nat,?x:Nat)))),min(s(plus(s(?z:Nat),plus(0,?x:Nat))),min(plus(0,s(?x:Nat)),max(max(?z:Nat,?x:Nat),max(?x:Nat,?x:Nat)))))",
"s(max(?y:Nat,min(plus(s(0),0),s(?x:Nat)))) = max(min(s(?z:Nat),min(?z:Nat,max(s(?x:Nat),plus(0,?x:Nat)))),plus(max(plus(plus(0,?y:Nat),min(0,?z:Nat)),min(max(?y:Nat,?x:Nat),min(?x:Nat,?x:Nat))),0))",
"s(max(plus(max(min(?z:Nat,?y:Nat),?y:Nat),?y:Nat),?z:Nat)) = max(min(plus(min(?z:Nat,s(?z:Nat)),?x:Nat),plus(plus(?x:Nat,?y:Nat),max(min(?z:Nat,?y:Nat),?x:Nat))),plus(max(s(0),max(min(?z:Nat,?x:Nat),min(?x:Nat,?x:Nat))),?y:Nat))",
"max(min(0,min(?x:Nat,?z:Nat)),min(s(0),s(?y:Nat))) = plus(?y:Nat,max(plus(s(min(?x:Nat,?y:Nat)),max(min(?x:Nat,?x:Nat),plus(?x:Nat,0))),plus(plus(plus(?z:Nat,?x:Nat),s(0)),max(min(0,?x:Nat),s(?x:Nat)))))",
"s(plus(?z:Nat,plus(plus(s(?z:Nat),max(?x:Nat,?z:Nat)),max(?x:Nat,?z:Nat)))) = max(max(s(max(?y:Nat,s(0))),s(?x:Nat)),plus(plus(plus(?z:Nat,plus(?x:Nat,?y:Nat)),s(plus(?x:Nat,?y:Nat))),min(?z:Nat,s(plus(?x:Nat,?y:Nat)))))",
"min(plus(plus(plus(min(?z:Nat,?z:Nat),plus(?x:Nat,?y:Nat)),0),0),?y:Nat) = min(min(max(plus(0,0),s(?z:Nat)),plus(s(s(?x:Nat)),plus(?y:Nat,min(0,0)))),plus(?z:Nat,max(min(?x:Nat,plus(?y:Nat,0)),?z:Nat)))",
"min(min(max(max(min(0,?z:Nat),?z:Nat),s(max(?x:Nat,?x:Nat))),?x:Nat),?z:Nat) = min(max(plus(?x:Nat,min(s(?y:Nat),plus(?y:Nat,?x:Nat))),plus(s(max(0,?x:Nat)),s(?x:Nat))),plus(?x:Nat,s(min(min(?x:Nat,?y:Nat),0))))",
"plus(s(?y:Nat),plus(plus(plus(max(?y:Nat,?x:Nat),plus(?x:Nat,?x:Nat)),?y:Nat),?x:Nat)) = plus(max(s(?x:Nat),s(min(plus(?z:Nat,?z:Nat),?x:Nat))),plus(max(?x:Nat,s(max(?y:Nat,0))),s(max(max(?y:Nat,?x:Nat),max(0,0)))))",
"max(min(s(0),s(?z:Nat)),plus(?x:Nat,max(max(0,min(0,?x:Nat)),?z:Nat))) = plus(min(?x:Nat,max(s(0),plus(?x:Nat,max(?x:Nat,?y:Nat)))),s(min(min(?z:Nat,min(?x:Nat,?x:Nat)),min(plus(?z:Nat,?y:Nat),s(?x:Nat)))))",
"max(min(max(?x:Nat,min(plus(?z:Nat,?x:Nat),s(?z:Nat))),0),max(min(?x:Nat,?x:Nat),?z:Nat)) = s(plus(min(plus(0,0),plus(?y:Nat,min(?y:Nat,?x:Nat))),max(plus(plus(?x:Nat,?x:Nat),plus(0,0)),max(s(?z:Nat),0))))",
"max(?x:Nat,plus(s(?x:Nat),plus(min(max(?z:Nat,0),s(?y:Nat)),min(plus(0,?y:Nat),s(?x:Nat))))) = plus(s(plus(s(plus(?y:Nat,?x:Nat)),min(0,s(?z:Nat)))),min(s(max(plus(?x:Nat,?x:Nat),min(?z:Nat,?z:Nat))),?z:Nat))",
"plus(plus(min(plus(s(?z:Nat),min(?y:Nat,?x:Nat)),?y:Nat),min(min(plus(?x:Nat,?x:Nat),0),?x:Nat)),?z:Nat) = max(plus(max(min(max(?x:Nat,?y:Nat),max(?x:Nat,?z:Nat)),s(?z:Nat)),min(?x:Nat,max(?z:Nat,s(?x:Nat)))),plus(?x:Nat,?y:Nat))",
"s(plus(min(?y:Nat,plus(?x:Nat,min(?y:Nat,?x:Nat))),min(min(?x:Nat,?y:Nat),min(min(?x:Nat,?y:Nat),s(?x:Nat))))) = max(max(plus(?z:Nat,min(max(?x:Nat,?x:Nat),min(0,0))),0),max(s(plus(min(0,?y:Nat),?x:Nat)),?y:Nat))",
"min(?y:Nat,plus(plus(max(s(0),s(?x:Nat)),max(min(?x:Nat,?x:Nat),?y:Nat)),s(s(plus(?x:Nat,0))))) = plus(plus(s(?z:Nat),s(?y:Nat)),plus(min(?z:Nat,plus(plus(?y:Nat,0),s(?x:Nat))),s(s(plus(0,?x:Nat)))))",
"plus(max(plus(plus(plus(?y:Nat,?y:Nat),max(?x:Nat,?x:Nat)),min(?y:Nat,?x:Nat)),max(?x:Nat,s(?x:Nat))),s(?z:Nat)) = s(plus(max(max(s(?x:Nat),0),plus(s(0),?y:Nat)),plus(min(s(?x:Nat),?z:Nat),s(plus(0,?x:Nat)))))",
"max(max(plus(s(max(0,?x:Nat)),min(s(0),min(0,0))),plus(s(s(0)),min(plus(?y:Nat,?x:Nat),?x:Nat))),?x:Nat) = plus(s(min(min(s(?x:Nat),s(?y:Nat)),?z:Nat)),s(s(max(min(?z:Nat,0),?y:Nat))))",
"plus(max(0,s(plus(plus(0,?x:Nat),min(?y:Nat,?x:Nat)))),plus(?x:Nat,plus(s(?y:Nat),max(plus(?x:Nat,?x:Nat),min(0,?x:Nat))))) = min(s(plus(s(max(?z:Nat,?x:Nat)),plus(min(?x:Nat,?z:Nat),plus(?z:Nat,?x:Nat)))),s(?x:Nat))",
"plus(min(max(max(?z:Nat,?x:Nat),max(plus(?y:Nat,?x:Nat),plus(?x:Nat,?x:Nat))),s(plus(min(?x:Nat,?x:Nat),plus(?y:Nat,0)))),plus(0,?x:Nat)) = plus(plus(min(?y:Nat,min(max(?x:Nat,?x:Nat),plus(?y:Nat,?x:Nat))),?z:Nat),max(?x:Nat,?x:Nat))",
"plus(min(plus(min(max(?x:Nat,?y:Nat),0),plus(?x:Nat,min(?y:Nat,?y:Nat))),plus(?y:Nat,min(plus(?y:Nat,?x:Nat),min(0,?x:Nat)))),s(?x:Nat)) = s(min(?y:Nat,max(s(max(?x:Nat,?x:Nat)),max(min(?z:Nat,?x:Nat),plus(?x:Nat,?x:Nat)))))",
"plus(max(?x:Nat,min(?x:Nat,min(?x:Nat,plus(?y:Nat,?z:Nat)))),max(plus(?x:Nat,plus(s(0),plus(?z:Nat,?x:Nat))),plus(?y:Nat,s(min(0,?y:Nat))))) = plus(plus(0,plus(?y:Nat,max(?x:Nat,s(?x:Nat)))),min(plus(?x:Nat,?x:Nat),?z:Nat))",
"max(min(max(s(?y:Nat),plus(min(?z:Nat,?z:Nat),?y:Nat)),max(?y:Nat,plus(s(0),plus(?y:Nat,0)))),max(?x:Nat,min(max(s(?z:Nat),0),?y:Nat))) = s(min(min(plus(?x:Nat,s(?y:Nat)),plus(max(?z:Nat,?x:Nat),?x:Nat)),?z:Nat))",
"plus(min(plus(min(s(?x:Nat),s(?x:Nat)),max(?x:Nat,plus(0,0))),min(?x:Nat,?x:Nat)),plus(0,plus(plus(min(?x:Nat,?x:Nat),plus(0,?y:Nat)),?y:Nat))) = s(plus(max(?z:Nat,?x:Nat),min(s(?x:Nat),s(plus(?x:Nat,?x:Nat)))))",
"plus(s(plus(max(s(?z:Nat),s(?y:Nat)),min(s(?z:Nat),min(?x:Nat,?y:Nat)))),max(?x:Nat,max(plus(min(?x:Nat,?x:Nat),s(?x:Nat)),max(max(?z:Nat,?z:Nat),s(?x:Nat))))) = min(?y:Nat,plus(0,s(min(plus(?z:Nat,?x:Nat),?y:Nat))))",
"plus(max(?x:Nat,max(plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?y:Nat)),?x:Nat)),plus(min(s(plus(?y:Nat,?x:Nat)),plus(?y:Nat,plus(?z:Nat,?x:Nat))),s(max(plus(?x:Nat,?z:Nat),0)))) = s(plus(?y:Nat,s(min(s(?y:Nat),min(?z:Nat,?z:Nat)))))",
"max(min(plus(?y:Nat,max(?y:Nat,s(0))),min(plus(?x:Nat,?z:Nat),min(min(0,?x:Nat),s(?z:Nat)))),plus(plus(s(plus(0,?y:Nat)),min(?z:Nat,max(?y:Nat,?x:Nat))),max(?z:Nat,max(0,plus(?x:Nat,?x:Nat))))) = s(s(?x:Nat))",
"s(plus(s(?x:Nat),plus(?z:Nat,max(?y:Nat,0)))) = min(plus(max(min(?x:Nat,max(0,?x:Nat)),max(0,0)),s(plus(max(?x:Nat,?y:Nat),min(?x:Nat,?y:Nat)))),max(plus(?x:Nat,plus(min(?x:Nat,?x:Nat),?y:Nat)),s(s(s(?x:Nat)))))",
"max(s(s(max(?z:Nat,s(?z:Nat)))),plus(s(plus(?y:Nat,?y:Nat)),0)) = max(plus(s(max(max(?x:Nat,?x:Nat),?x:Nat)),s(s(min(0,?x:Nat)))),plus(min(max(plus(?x:Nat,?x:Nat),min(?x:Nat,?z:Nat)),s(plus(?z:Nat,?z:Nat))),?x:Nat))",
"max(s(min(?y:Nat,plus(min(?y:Nat,?z:Nat),min(?x:Nat,?x:Nat)))),s(max(0,0))) = min(s(?x:Nat),plus(plus(plus(s(?x:Nat),min(?x:Nat,?x:Nat)),s(s(?x:Nat))),plus(plus(max(?z:Nat,?x:Nat),s(?x:Nat)),s(min(?y:Nat,0)))))",
"min(max(?y:Nat,plus(?x:Nat,min(min(0,0),plus(?x:Nat,0)))),max(0,?y:Nat)) = max(plus(plus(?y:Nat,min(s(?z:Nat),max(?x:Nat,?y:Nat))),min(plus(max(?z:Nat,?y:Nat),s(?x:Nat)),plus(?x:Nat,?z:Nat))),plus(plus(?y:Nat,?y:Nat),?x:Nat))",
"min(min(plus(?x:Nat,min(s(?y:Nat),s(?x:Nat))),s(max(?z:Nat,max(?x:Nat,?z:Nat)))),0) = min(min(min(min(min(?x:Nat,?x:Nat),s(?x:Nat)),min(max(?z:Nat,0),max(?z:Nat,?z:Nat))),plus(s(plus(?z:Nat,?x:Nat)),s(?x:Nat))),?z:Nat)",
"plus(max(plus(plus(s(?z:Nat),plus(0,?x:Nat)),?z:Nat),s(s(min(?x:Nat,0)))),0) = max(?y:Nat,plus(max(plus(max(?y:Nat,?z:Nat),max(?y:Nat,?y:Nat)),s(?x:Nat)),plus(max(plus(?y:Nat,?x:Nat),s(?x:Nat)),s(min(?x:Nat,?x:Nat)))))",
"min(?z:Nat,s(min(plus(max(?x:Nat,?x:Nat),plus(?y:Nat,?y:Nat)),min(min(?x:Nat,?y:Nat),max(?z:Nat,?x:Nat))))) = s(min(plus(plus(min(?x:Nat,?z:Nat),plus(?x:Nat,0)),0),min(plus(max(?y:Nat,?z:Nat),s(?z:Nat)),min(s(?x:Nat),?x:Nat))))",
"plus(max(min(max(?y:Nat,?z:Nat),0),max(min(s(?y:Nat),min(?y:Nat,?y:Nat)),s(s(?x:Nat)))),?x:Nat) = max(min(min(min(plus(?z:Nat,?x:Nat),min(?x:Nat,?x:Nat)),min(?x:Nat,plus(?x:Nat,0))),plus(?z:Nat,s(max(?x:Nat,?x:Nat)))),?x:Nat)",
"max(max(?x:Nat,min(?z:Nat,0)),min(plus(max(plus(?x:Nat,?z:Nat),min(?x:Nat,?x:Nat)),s(?x:Nat)),max(?y:Nat,?z:Nat))) = min(plus(plus(min(max(?x:Nat,?x:Nat),s(?x:Nat)),s(max(?y:Nat,?x:Nat))),?z:Nat),max(0,max(s(?y:Nat),?x:Nat)))",
"plus(0,plus(min(max(s(?y:Nat),?x:Nat),0),min(min(plus(?x:Nat,?x:Nat),0),plus(s(?z:Nat),plus(?x:Nat,?x:Nat))))) = min(s(plus(?x:Nat,max(plus(?x:Nat,?x:Nat),max(?y:Nat,?z:Nat)))),plus(min(?y:Nat,s(min(?x:Nat,0))),0))",
"max(s(s(plus(?z:Nat,plus(?x:Nat,?y:Nat)))),min(s(?x:Nat),plus(min(max(?y:Nat,?y:Nat),?x:Nat),min(s(?y:Nat),min(?y:Nat,?z:Nat))))) = plus(s(s(min(?x:Nat,s(?y:Nat)))),plus(?x:Nat,min(s(max(?y:Nat,?x:Nat)),s(s(?y:Nat)))))",
"plus(min(s(max(s(?z:Nat),plus(?x:Nat,?x:Nat))),?z:Nat),max(s(?z:Nat),max(plus(s(?x:Nat),min(?y:Nat,?x:Nat)),max(plus(?z:Nat,?x:Nat),?z:Nat)))) = max(plus(plus(plus(s(?x:Nat),s(?y:Nat)),s(?x:Nat)),plus(s(?z:Nat),?x:Nat)),?y:Nat)",
"min(plus(plus(s(?z:Nat),s(?x:Nat)),?z:Nat),max(s(plus(max(?x:Nat,?y:Nat),?x:Nat)),min(min(plus(?x:Nat,0),plus(?x:Nat,?y:Nat)),max(?y:Nat,0)))) = max(s(s(?x:Nat)),s(max(max(plus(?x:Nat,?x:Nat),max(?y:Nat,?z:Nat)),?y:Nat)))",
"max(plus(min(max(s(?z:Nat),plus(?x:Nat,?z:Nat)),plus(?x:Nat,s(?x:Nat))),max(max(plus(?x:Nat,0),?y:Nat),?x:Nat)),s(s(plus(s(?z:Nat),s(0))))) = max(?x:Nat,min(max(s(?y:Nat),?y:Nat),s(plus(0,plus(0,0)))))",
"max(min(s(plus(max(?z:Nat,0),plus(?z:Nat,?z:Nat))),s(plus(plus(0,?z:Nat),plus(?y:Nat,0)))),s(min(max(s(?x:Nat),plus(?y:Nat,?x:Nat)),s(plus(?x:Nat,?z:Nat))))) = s(plus(min(?x:Nat,0),max(s(s(?x:Nat)),?x:Nat)))",
"max(plus(s(plus(max(?x:Nat,?x:Nat),?y:Nat)),plus(?x:Nat,s(plus(0,?y:Nat)))),plus(min(?x:Nat,min(?x:Nat,?x:Nat)),min(s(plus(0,?x:Nat)),plus(?z:Nat,max(0,?y:Nat))))) = s(min(min(0,min(plus(?z:Nat,?x:Nat),0)),0))",
"min(max(?y:Nat,s(?y:Nat)),plus(max(plus(plus(?y:Nat,?y:Nat),max(?y:Nat,?z:Nat)),s(s(?x:Nat))),max(plus(min(?y:Nat,?y:Nat),max(?x:Nat,?y:Nat)),min(max(0,?x:Nat),0)))) = plus(max(plus(s(plus(?y:Nat,?x:Nat)),?z:Nat),?z:Nat),?y:Nat)",
"min(plus(s(max(max(?x:Nat,?y:Nat),plus(0,?y:Nat))),max(max(plus(?x:Nat,?x:Nat),plus(?z:Nat,?x:Nat)),plus(s(0),plus(0,?x:Nat)))),plus(plus(?z:Nat,?y:Nat),min(s(plus(?z:Nat,?x:Nat)),?z:Nat))) = max(max(?y:Nat,0),s(?x:Nat))",
"min(plus(plus(s(0),min(plus(?x:Nat,?x:Nat),plus(?z:Nat,0))),min(?x:Nat,min(0,0))),plus(max(plus(plus(?x:Nat,?x:Nat),0),max(?y:Nat,s(?z:Nat))),min(0,min(max(?y:Nat,?x:Nat),s(?z:Nat))))) = s(plus(?z:Nat,?x:Nat))",
"plus(plus(plus(plus(min(?y:Nat,0),?z:Nat),max(min(?x:Nat,?y:Nat),min(0,0))),plus(max(s(?z:Nat),?x:Nat),plus(?z:Nat,s(?y:Nat)))),plus(plus(min(s(?x:Nat),s(?z:Nat)),s(max(?x:Nat,?y:Nat))),min(?x:Nat,0))) = s(0)",
"s(?y:Nat) = min(plus(max(min(min(?x:Nat,?z:Nat),?x:Nat),plus(?y:Nat,plus(0,?x:Nat))),max(plus(max(0,?x:Nat),min(?x:Nat,?x:Nat)),max(plus(?y:Nat,0),s(?x:Nat)))),min(max(min(min(?x:Nat,?y:Nat),s(0)),?x:Nat),s(s(?z:Nat))))",
"plus(?x:Nat,s(?x:Nat)) = min(plus(max(s(0),plus(plus(?x:Nat,?x:Nat),min(?x:Nat,?z:Nat))),plus(s(0),?y:Nat)),max(plus(min(?z:Nat,s(?x:Nat)),?x:Nat),plus(max(min(0,?x:Nat),min(?x:Nat,?z:Nat)),plus(plus(?x:Nat,?x:Nat),s(?x:Nat)))))",
"s(min(s(s(max(0,?x:Nat))),?z:Nat)) = plus(s(min(max(?y:Nat,plus(0,?y:Nat)),min(plus(?z:Nat,?x:Nat),max(?z:Nat,?z:Nat)))),min(min(max(?z:Nat,plus(?z:Nat,?y:Nat)),s(min(?y:Nat,0))),plus(min(?x:Nat,max(?y:Nat,?y:Nat)),?z:Nat)))",
"max(0,min(s(min(s(?y:Nat),?x:Nat)),?x:Nat)) = min(min(plus(plus(plus(?y:Nat,?y:Nat),plus(?x:Nat,?x:Nat)),min(plus(0,0),?x:Nat)),s(s(plus(?x:Nat,?z:Nat)))),s(min(s(s(?x:Nat)),min(max(?z:Nat,0),plus(?x:Nat,?z:Nat)))))",
"max(s(?x:Nat),s(min(s(s(?x:Nat)),?x:Nat))) = plus(plus(?z:Nat,plus(max(s(?x:Nat),max(?x:Nat,?y:Nat)),s(?z:Nat))),plus(plus(plus(min(?z:Nat,?x:Nat),plus(?z:Nat,0)),?y:Nat),min(s(plus(?y:Nat,?x:Nat)),plus(max(?x:Nat,0),0))))",
"min(0,s(min(s(plus(?y:Nat,0)),?x:Nat))) = min(plus(0,min(min(plus(0,?z:Nat),max(?x:Nat,?y:Nat)),max(s(?x:Nat),plus(?z:Nat,?x:Nat)))),s(min(max(plus(?z:Nat,?z:Nat),s(?y:Nat)),max(plus(?x:Nat,0),plus(?x:Nat,0)))))",
"min(plus(s(s(plus(0,?x:Nat))),?x:Nat),s(?y:Nat)) = min(plus(max(?z:Nat,min(min(?y:Nat,?x:Nat),0)),plus(0,plus(max(?z:Nat,?z:Nat),?z:Nat))),s(plus(max(min(?x:Nat,?x:Nat),s(?z:Nat)),max(plus(?x:Nat,0),plus(?y:Nat,?y:Nat)))))",
"s(s(plus(min(min(?x:Nat,?y:Nat),s(?z:Nat)),s(?z:Nat)))) = max(s(plus(?z:Nat,?x:Nat)),min(plus(plus(?z:Nat,?x:Nat),plus(s(?x:Nat),max(0,?y:Nat))),min(plus(min(?z:Nat,?y:Nat),max(?x:Nat,?x:Nat)),max(s(?x:Nat),max(?y:Nat,?x:Nat)))))",
"plus(?x:Nat,plus(?x:Nat,min(?z:Nat,plus(s(?z:Nat),plus(?x:Nat,0))))) = plus(plus(?z:Nat,min(s(max(?x:Nat,0)),min(plus(0,?x:Nat),min(?x:Nat,?z:Nat)))),min(0,min(max(?x:Nat,min(?y:Nat,?y:Nat)),min(max(?z:Nat,?x:Nat),s(?y:Nat)))))",
"plus(?y:Nat,plus(s(max(s(?y:Nat),plus(?z:Nat,0))),max(min(s(0),s(?y:Nat)),0))) = plus(plus(s(min(0,plus(?x:Nat,0))),?x:Nat),max(s(0),plus(s(plus(0,?y:Nat)),max(min(?y:Nat,?z:Nat),plus(?y:Nat,?y:Nat)))))",
"min(plus(plus(min(?x:Nat,?z:Nat),?x:Nat),plus(s(max(0,?x:Nat)),plus(plus(?x:Nat,?z:Nat),?y:Nat))),s(?z:Nat)) = min(s(s(plus(s(?z:Nat),max(?x:Nat,?x:Nat)))),plus(plus(?x:Nat,?x:Nat),plus(s(plus(?x:Nat,?y:Nat)),min(?x:Nat,s(?x:Nat)))))",
"plus(s(plus(min(max(?y:Nat,?z:Nat),?y:Nat),min(plus(?x:Nat,?y:Nat),0))),s(s(max(s(?y:Nat),?y:Nat)))) = min(min(?y:Nat,plus(s(max(?x:Nat,?z:Nat)),min(?y:Nat,max(?x:Nat,?x:Nat)))),plus(s(s(max(?x:Nat,0))),min(?z:Nat,?x:Nat)))",
"plus(s(plus(?x:Nat,?z:Nat)),plus(min(max(0,s(?y:Nat)),?x:Nat),plus(max(s(?y:Nat),plus(?z:Nat,0)),?y:Nat))) = max(plus(max(s(plus(?x:Nat,0)),s(max(0,?y:Nat))),?z:Nat),min(?z:Nat,plus(plus(?x:Nat,plus(?x:Nat,?x:Nat)),?y:Nat)))",
"s(max(min(?x:Nat,s(max(?x:Nat,0))),max(plus(max(?x:Nat,0),?x:Nat),plus(max(?y:Nat,0),max(?x:Nat,?x:Nat))))) = s(plus(plus(min(max(?x:Nat,?x:Nat),min(?z:Nat,?x:Nat)),0),min(plus(plus(0,?x:Nat),?y:Nat),min(?x:Nat,?x:Nat))))",
"max(plus(?z:Nat,plus(max(s(0),plus(?x:Nat,?y:Nat)),max(s(?y:Nat),plus(?x:Nat,?z:Nat)))),min(?y:Nat,min(0,0))) = max(min(plus(s(?x:Nat),plus(plus(?y:Nat,?x:Nat),?x:Nat)),plus(plus(?z:Nat,max(?y:Nat,?z:Nat)),plus(?y:Nat,?z:Nat))),?x:Nat)",
"min(?x:Nat,min(s(min(s(0),plus(?y:Nat,?y:Nat))),min(min(max(?y:Nat,?y:Nat),s(0)),max(max(?x:Nat,0),plus(0,?y:Nat))))) = max(?x:Nat,plus(s(max(min(?x:Nat,?y:Nat),plus(0,?x:Nat))),s(max(plus(?z:Nat,?x:Nat),?x:Nat))))",
"min(min(plus(s(s(?z:Nat)),max(?x:Nat,max(?x:Nat,?z:Nat))),?x:Nat),max(max(s(plus(0,?x:Nat)),?x:Nat),plus(?y:Nat,max(0,max(?x:Nat,?x:Nat))))) = max(0,min(max(?y:Nat,0),plus(min(max(?z:Nat,?x:Nat),plus(?x:Nat,?y:Nat)),?y:Nat)))",
"plus(plus(plus(s(s(0)),min(s(?y:Nat),plus(?x:Nat,0))),?x:Nat),plus(s(s(s(0))),min(s(s(?y:Nat)),plus(min(?x:Nat,0),?x:Nat)))) = max(min(max(?y:Nat,min(min(?z:Nat,?y:Nat),s(?x:Nat))),s(s(?y:Nat))),0)",
"plus(max(0,min(plus(plus(0,?z:Nat),?z:Nat),s(max(?x:Nat,?y:Nat)))),max(plus(?x:Nat,plus(min(?z:Nat,?x:Nat),max(?x:Nat,0))),s(plus(0,s(?x:Nat))))) = max(?z:Nat,s(min(min(?y:Nat,plus(?z:Nat,?z:Nat)),min(s(0),?x:Nat))))",
"max(plus(s(?y:Nat),min(s(plus(0,?z:Nat)),plus(0,s(?z:Nat)))),plus(plus(s(s(?x:Nat)),plus(?x:Nat,s(?y:Nat))),max(plus(plus(?x:Nat,?x:Nat),s(?x:Nat)),plus(s(?x:Nat),max(?z:Nat,0))))) = s(s(s(min(?x:Nat,?z:Nat))))",
"min(?z:Nat,plus(plus(s(min(?x:Nat,?x:Nat)),?x:Nat),plus(max(s(0),0),?x:Nat))) = max(max(max(max(plus(?x:Nat,?z:Nat),max(?z:Nat,?x:Nat)),?x:Nat),s(plus(plus(0,?x:Nat),s(?y:Nat)))),max(?x:Nat,s(plus(s(?x:Nat),max(?z:Nat,?x:Nat)))))",
"plus(?y:Nat,plus(plus(?x:Nat,min(s(?x:Nat),max(?x:Nat,?x:Nat))),max(?y:Nat,plus(?x:Nat,0)))) = max(plus(s(s(?x:Nat)),plus(max(s(?x:Nat),max(?y:Nat,?z:Nat)),plus(?x:Nat,s(?x:Nat)))),plus(s(plus(max(?x:Nat,?x:Nat),s(?x:Nat))),s(?x:Nat)))",
"plus(min(plus(s(max(?z:Nat,?y:Nat)),min(plus(?x:Nat,?y:Nat),max(?z:Nat,?y:Nat))),0),?x:Nat) = s(plus(min(plus(plus(?z:Nat,?z:Nat),plus(?x:Nat,?x:Nat)),min(s(?x:Nat),max(?z:Nat,?y:Nat))),max(plus(min(0,?x:Nat),plus(?y:Nat,?x:Nat)),s(?z:Nat))))",
"plus(plus(max(s(min(?x:Nat,?x:Nat)),?z:Nat),max(?x:Nat,plus(?y:Nat,max(?x:Nat,?x:Nat)))),0) = s(max(max(s(s(?y:Nat)),min(min(?z:Nat,?y:Nat),min(?z:Nat,?x:Nat))),plus(min(plus(0,?y:Nat),plus(?z:Nat,?y:Nat)),max(plus(?x:Nat,?x:Nat),?z:Nat))))",
"s(max(plus(min(max(?x:Nat,?z:Nat),?y:Nat),?x:Nat),max(max(s(?x:Nat),max(?y:Nat,?y:Nat)),?x:Nat))) = max(s(?x:Nat),plus(min(plus(max(?z:Nat,?x:Nat),max(0,0)),max(min(?y:Nat,?y:Nat),?x:Nat)),max(plus(s(?z:Nat),?x:Nat),min(?x:Nat,?x:Nat))))",
"plus(min(s(s(plus(?y:Nat,?z:Nat))),plus(s(max(?x:Nat,0)),s(max(?y:Nat,?x:Nat)))),0) = min(plus(plus(min(plus(?x:Nat,?z:Nat),?x:Nat),plus(?x:Nat,max(?x:Nat,?x:Nat))),plus(min(?z:Nat,min(0,0)),?z:Nat)),s(s(plus(?x:Nat,0))))",
"min(min(?y:Nat,max(s(0),s(plus(?x:Nat,?z:Nat)))),min(plus(s(min(?y:Nat,?x:Nat)),0),?z:Nat)) = s(plus(plus(?z:Nat,plus(plus(?y:Nat,0),min(?y:Nat,?x:Nat))),min(min(plus(?y:Nat,?y:Nat),plus(?x:Nat,?x:Nat)),min(s(?y:Nat),s(?y:Nat)))))",
"s(min(min(0,s(?y:Nat)),min(plus(s(?x:Nat),s(?x:Nat)),min(min(?x:Nat,?x:Nat),plus(?y:Nat,0))))) = min(min(plus(s(0),plus(min(?x:Nat,?x:Nat),min(?x:Nat,?x:Nat))),?x:Nat),max(?x:Nat,min(s(?x:Nat),max(min(?y:Nat,?z:Nat),?y:Nat))))",
"plus(min(min(min(?x:Nat,plus(0,0)),?x:Nat),0),plus(?y:Nat,plus(plus(0,plus(?x:Nat,?x:Nat)),?x:Nat))) = plus(plus(s(min(?z:Nat,0)),max(max(?z:Nat,min(?z:Nat,?x:Nat)),max(plus(0,0),?x:Nat))),max(s(s(s(?y:Nat))),?y:Nat))",
"max(s(s(max(min(?y:Nat,0),min(?x:Nat,?z:Nat)))),max(max(s(max(?y:Nat,0)),?x:Nat),min(?y:Nat,?z:Nat))) = s(min(s(plus(?x:Nat,s(?z:Nat))),max(plus(max(?z:Nat,?x:Nat),plus(?x:Nat,?z:Nat)),plus(min(?x:Nat,?z:Nat),min(0,0)))))",
"min(?z:Nat,max(plus(min(max(?x:Nat,0),?x:Nat),plus(max(?x:Nat,?y:Nat),0)),s(max(min(?z:Nat,?x:Nat),?x:Nat)))) = min(plus(max(max(s(?y:Nat),s(?x:Nat)),?x:Nat),max(0,plus(?x:Nat,?y:Nat))),plus(max(s(plus(?x:Nat,0)),?x:Nat),?x:Nat))",
"min(plus(s(?y:Nat),max(s(max(?y:Nat,?x:Nat)),max(?x:Nat,max(?z:Nat,?x:Nat)))),min(0,plus(0,plus(?z:Nat,plus(?x:Nat,?x:Nat))))) = plus(?x:Nat,plus(plus(?x:Nat,min(plus(?z:Nat,?x:Nat),?x:Nat)),max(s(plus(?z:Nat,?z:Nat)),plus(?x:Nat,s(?y:Nat)))))",
"min(min(plus(0,min(min(0,?z:Nat),plus(?x:Nat,?z:Nat))),0),min(?z:Nat,min(s(min(?x:Nat,?x:Nat)),max(plus(?z:Nat,?x:Nat),plus(?x:Nat,0))))) = plus(s(max(0,plus(plus(?z:Nat,0),plus(0,?x:Nat)))),max(s(?x:Nat),s(?x:Nat)))",
"max(plus(plus(plus(s(?y:Nat),?x:Nat),plus(plus(?x:Nat,0),max(?x:Nat,?x:Nat))),max(max(s(?y:Nat),?x:Nat),max(?z:Nat,min(?x:Nat,?x:Nat)))),s(plus(?x:Nat,?x:Nat))) = min(min(?z:Nat,?x:Nat),plus(plus(s(plus(?x:Nat,?x:Nat)),plus(0,?z:Nat)),?x:Nat))",
"min(max(min(plus(?y:Nat,plus(?y:Nat,0)),max(?y:Nat,?z:Nat)),s(s(?x:Nat))),min(s(max(min(?y:Nat,?y:Nat),?z:Nat)),min(max(?x:Nat,?x:Nat),plus(?x:Nat,plus(?x:Nat,?x:Nat))))) = min(s(min(?y:Nat,max(min(?z:Nat,0),min(?x:Nat,?x:Nat)))),?y:Nat)",
"plus(max(?x:Nat,min(?z:Nat,max(s(?y:Nat),s(?x:Nat)))),min(min(plus(plus(?z:Nat,?z:Nat),plus(?x:Nat,0)),max(max(0,?z:Nat),s(?y:Nat))),plus(?y:Nat,plus(?z:Nat,plus(?x:Nat,?y:Nat))))) = s(s(plus(plus(plus(?y:Nat,?y:Nat),s(0)),0)))",
"plus(min(min(s(min(0,?x:Nat)),min(0,min(0,?z:Nat))),max(min(plus(?z:Nat,?x:Nat),s(0)),min(max(?x:Nat,0),plus(?x:Nat,?y:Nat)))),plus(min(?z:Nat,max(plus(?x:Nat,?x:Nat),0)),s(s(max(?z:Nat,?y:Nat))))) = s(s(?x:Nat))",
"plus(s(plus(max(?z:Nat,0),0)),?x:Nat) = max(max(?x:Nat,min(min(plus(?x:Nat,?y:Nat),?x:Nat),?y:Nat)),plus(max(max(?y:Nat,min(?x:Nat,?x:Nat)),plus(plus(?x:Nat,?x:Nat),min(?x:Nat,?x:Nat))),plus(plus(plus(?x:Nat,?x:Nat),plus(?y:Nat,?y:Nat)),max(0,0))))",
"min(?y:Nat,min(?x:Nat,plus(max(s(?y:Nat),plus(?z:Nat,0)),?y:Nat))) = min(s(0),min(max(plus(?z:Nat,s(0)),plus(max(?x:Nat,?x:Nat),min(?z:Nat,?x:Nat))),plus(max(plus(?x:Nat,0),min(?x:Nat,0)),min(plus(0,?x:Nat),max(?x:Nat,0)))))",
"max(min(s(s(plus(?x:Nat,0))),max(?x:Nat,s(?y:Nat))),max(?x:Nat,0)) = min(min(plus(?y:Nat,s(?x:Nat)),min(s(s(?x:Nat)),plus(plus(?z:Nat,?x:Nat),0))),plus(s(?x:Nat),plus(max(0,min(?x:Nat,?x:Nat)),plus(min(0,?x:Nat),?x:Nat))))",
"s(plus(min(?x:Nat,max(s(0),plus(?y:Nat,?y:Nat))),s(s(plus(0,?x:Nat))))) = plus(plus(min(?x:Nat,min(0,?y:Nat)),max(plus(plus(?y:Nat,?x:Nat),s(?x:Nat)),max(plus(?x:Nat,?y:Nat),max(?x:Nat,?z:Nat)))),s(plus(?x:Nat,s(max(?y:Nat,?y:Nat)))))",
"min(?x:Nat,max(plus(max(plus(0,?x:Nat),s(?y:Nat)),?x:Nat),min(0,s(min(?x:Nat,?z:Nat))))) = min(s(s(max(plus(0,?y:Nat),s(?z:Nat)))),min(s(max(plus(?z:Nat,?y:Nat),min(?y:Nat,?z:Nat))),min(plus(plus(?x:Nat,?y:Nat),?x:Nat),s(?z:Nat))))",
"min(s(s(plus(?z:Nat,min(?z:Nat,0)))),s(max(plus(plus(?y:Nat,?z:Nat),plus(?x:Nat,?x:Nat)),s(s(?x:Nat))))) = plus(s(plus(max(?z:Nat,min(?z:Nat,?z:Nat)),min(max(?z:Nat,?x:Nat),0))),min(0,max(s(plus(?x:Nat,?x:Nat)),plus(0,?x:Nat))))",
"max(?x:Nat,min(plus(?y:Nat,plus(min(?x:Nat,?x:Nat),?z:Nat)),plus(max(?y:Nat,plus(?z:Nat,?z:Nat)),max(max(0,0),s(0))))) = s(min(plus(plus(0,0),0),min(plus(plus(0,?x:Nat),max(?z:Nat,?z:Nat)),plus(max(0,?x:Nat),s(?y:Nat)))))",
"max(min(s(plus(min(?x:Nat,?x:Nat),?z:Nat)),?z:Nat),max(plus(?y:Nat,max(max(?x:Nat,?x:Nat),?x:Nat)),min(?x:Nat,s(s(?x:Nat))))) = plus(plus(plus(s(s(?z:Nat)),s(max(?z:Nat,?y:Nat))),max(min(?x:Nat,s(?y:Nat)),max(max(?x:Nat,?x:Nat),0))),?y:Nat)",
"plus(max(max(plus(?x:Nat,0),max(plus(?z:Nat,0),?x:Nat)),plus(max(plus(?z:Nat,0),?x:Nat),plus(s(0),?z:Nat))),?x:Nat) = plus(plus(0,plus(plus(?x:Nat,0),max(min(?x:Nat,?x:Nat),s(?z:Nat)))),s(plus(?x:Nat,plus(?y:Nat,max(0,?x:Nat)))))",
"max(max(max(?y:Nat,min(plus(?x:Nat,?x:Nat),?z:Nat)),max(plus(plus(?y:Nat,?x:Nat),?y:Nat),plus(s(?x:Nat),?z:Nat))),s(max(?x:Nat,?y:Nat))) = plus(max(?x:Nat,min(0,s(max(?y:Nat,?x:Nat)))),s(min(s(min(?x:Nat,?y:Nat)),min(?y:Nat,plus(0,?x:Nat)))))",
"max(min(min(s(?z:Nat),0),?y:Nat),max(plus(min(?x:Nat,max(?z:Nat,0)),max(s(?x:Nat),plus(?x:Nat,?z:Nat))),max(?x:Nat,?x:Nat))) = min(plus(plus(?x:Nat,s(min(?x:Nat,?x:Nat))),min(plus(0,plus(?z:Nat,?x:Nat)),min(max(?x:Nat,0),?x:Nat))),?x:Nat)",
"s(plus(min(plus(?y:Nat,plus(?z:Nat,?x:Nat)),plus(0,?y:Nat)),plus(min(?z:Nat,min(0,?y:Nat)),max(plus(?z:Nat,?y:Nat),plus(?y:Nat,?x:Nat))))) = plus(max(s(s(?x:Nat)),s(max(plus(0,?x:Nat),plus(0,?y:Nat)))),max(0,s(s(s(?y:Nat)))))",
"max(plus(?x:Nat,max(?x:Nat,plus(s(0),min(?y:Nat,?x:Nat)))),min(max(max(max(?z:Nat,?y:Nat),?z:Nat),?z:Nat),min(max(?y:Nat,?y:Nat),0))) = max(min(plus(s(?y:Nat),plus(min(?z:Nat,0),?z:Nat)),?x:Nat),s(max(s(s(?z:Nat)),s(s(?x:Nat)))))",
"plus(?z:Nat,plus(min(s(min(?x:Nat,?y:Nat)),plus(s(?z:Nat),?z:Nat)),max(plus(min(?x:Nat,?y:Nat),?x:Nat),plus(plus(?z:Nat,0),min(0,?z:Nat))))) = min(plus(0,s(?y:Nat)),min(min(plus(s(0),s(?z:Nat)),plus(min(?y:Nat,?z:Nat),?z:Nat)),0))",
"plus(s(?z:Nat),plus(plus(max(plus(?z:Nat,0),max(?y:Nat,?z:Nat)),plus(plus(?x:Nat,?x:Nat),min(?x:Nat,?x:Nat))),s(max(?x:Nat,min(?x:Nat,?x:Nat))))) = s(min(plus(s(s(?x:Nat)),plus(s(?x:Nat),s(?z:Nat))),max(s(s(?x:Nat)),max(?x:Nat,?x:Nat))))",
"plus(min(s(max(?z:Nat,plus(?x:Nat,?x:Nat))),plus(min(s(?z:Nat),max(0,?z:Nat)),?z:Nat)),plus(0,min(plus(min(?z:Nat,0),?y:Nat),?x:Nat))) = s(min(min(plus(max(?z:Nat,?z:Nat),min(?x:Nat,?z:Nat)),min(min(?z:Nat,?y:Nat),max(?z:Nat,0))),?x:Nat))",
"s(plus(plus(max(plus(?z:Nat,?x:Nat),min(?x:Nat,0)),max(?x:Nat,0)),plus(plus(?y:Nat,plus(?x:Nat,?x:Nat)),min(plus(?x:Nat,?x:Nat),plus(0,?y:Nat))))) = min(?y:Nat,plus(min(?z:Nat,?y:Nat),max(max(s(?y:Nat),?z:Nat),plus(s(?x:Nat),plus(?x:Nat,?x:Nat)))))",
"min(plus(?x:Nat,max(min(?x:Nat,?z:Nat),s(0))),min(max(plus(min(?z:Nat,?x:Nat),plus(?y:Nat,?y:Nat)),?z:Nat),min(plus(s(?y:Nat),s(?z:Nat)),plus(?y:Nat,min(?z:Nat,?y:Nat))))) = s(s(plus(plus(?x:Nat,plus(?z:Nat,?x:Nat)),plus(?y:Nat,plus(?y:Nat,0)))))",
"plus(plus(min(plus(s(0),s(?x:Nat)),min(max(?y:Nat,?y:Nat),?z:Nat)),plus(s(s(?x:Nat)),plus(plus(?y:Nat,0),plus(?y:Nat,?y:Nat)))),s(plus(s(max(?x:Nat,?x:Nat)),?y:Nat))) = min(s(?z:Nat),plus(plus(s(?x:Nat),s(plus(0,?x:Nat))),?x:Nat))",
"plus(plus(?x:Nat,plus(?y:Nat,s(max(?x:Nat,?x:Nat)))),?z:Nat) = plus(max(max(min(s(?z:Nat),plus(?z:Nat,?z:Nat)),s(max(?y:Nat,?x:Nat))),max(plus(?x:Nat,?z:Nat),?y:Nat)),max(plus(max(?x:Nat,s(?z:Nat)),plus(plus(?z:Nat,?y:Nat),s(0))),s(s(s(?y:Nat)))))",
"s(plus(0,s(max(max(0,0),plus(?y:Nat,?y:Nat))))) = min(max(s(plus(plus(?x:Nat,?z:Nat),max(?x:Nat,?x:Nat))),plus(s(?x:Nat),?x:Nat)),max(plus(plus(plus(0,?z:Nat),plus(?y:Nat,?x:Nat)),?x:Nat),plus(max(s(0),0),s(max(?x:Nat,?x:Nat)))))",
"s(max(min(min(min(0,?y:Nat),?x:Nat),?z:Nat),max(0,?y:Nat))) = max(s(plus(max(?y:Nat,plus(?z:Nat,0)),s(?z:Nat))),max(s(max(plus(0,0),min(?y:Nat,0))),plus(min(plus(?z:Nat,0),s(?x:Nat)),plus(plus(?y:Nat,?x:Nat),s(?y:Nat)))))",
"max(min(plus(plus(plus(?y:Nat,?z:Nat),?z:Nat),0),s(min(?y:Nat,?z:Nat))),?x:Nat) = min(plus(min(plus(s(?x:Nat),s(?z:Nat)),0),?x:Nat),min(s(max(plus(0,?x:Nat),?x:Nat)),max(plus(plus(?y:Nat,0),min(?y:Nat,0)),plus(0,min(?y:Nat,?z:Nat)))))",
"max(max(?z:Nat,plus(?z:Nat,?x:Nat)),s(max(min(plus(?x:Nat,?z:Nat),?x:Nat),s(?x:Nat)))) = max(plus(min(?z:Nat,plus(min(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat))),plus(min(plus(?z:Nat,?y:Nat),s(0)),?z:Nat)),plus(min(plus(plus(?z:Nat,?x:Nat),s(?y:Nat)),?x:Nat),?y:Nat))",
"plus(max(plus(?x:Nat,?y:Nat),plus(s(max(?z:Nat,0)),s(?x:Nat))),s(plus(plus(max(?x:Nat,?x:Nat),min(?x:Nat,?x:Nat)),s(plus(?x:Nat,0))))) = max(plus(min(s(plus(0,?x:Nat)),0),min(max(plus(?y:Nat,?y:Nat),s(?z:Nat)),s(s(?x:Nat)))),?x:Nat)",
"plus(plus(s(min(?z:Nat,min(?x:Nat,?y:Nat))),max(max(min(?x:Nat,?x:Nat),s(?x:Nat)),plus(?z:Nat,?y:Nat))),s(plus(?x:Nat,s(plus(?x:Nat,?x:Nat))))) = plus(s(?y:Nat),s(plus(min(max(0,?x:Nat),plus(0,?z:Nat)),max(plus(0,?x:Nat),max(?z:Nat,0)))))",
"max(plus(min(?y:Nat,min(s(?z:Nat),s(?y:Nat))),min(s(max(?y:Nat,?x:Nat)),plus(plus(?x:Nat,?y:Nat),plus(?y:Nat,?x:Nat)))),max(min(s(?y:Nat),?y:Nat),max(s(max(0,?y:Nat)),?y:Nat))) = s(plus(min(?z:Nat,min(?x:Nat,max(?x:Nat,?y:Nat))),min(?z:Nat,?x:Nat)))",
"plus(min(plus(s(?y:Nat),s(min(?y:Nat,?x:Nat))),?y:Nat),plus(min(min(max(?x:Nat,?z:Nat),plus(?x:Nat,?y:Nat)),plus(plus(?y:Nat,?y:Nat),min(?z:Nat,?y:Nat))),min(min(s(?y:Nat),plus(?x:Nat,0)),s(0)))) = s(s(s(plus(s(0),plus(?x:Nat,?z:Nat)))))",
"plus(s(max(min(s(0),0),plus(max(?z:Nat,?x:Nat),?y:Nat))),min(min(min(s(0),s(?z:Nat)),s(max(?x:Nat,0))),max(max(s(?z:Nat),min(?x:Nat,?y:Nat)),plus(s(?y:Nat),plus(?x:Nat,?y:Nat))))) = max(?z:Nat,min(max(s(0),?x:Nat),?x:Nat))",
"max(?x:Nat,s(min(max(?z:Nat,min(?z:Nat,0)),?x:Nat))) = plus(plus(max(s(?x:Nat),min(plus(?x:Nat,?x:Nat),plus(?y:Nat,?x:Nat))),s(plus(s(?z:Nat),?y:Nat))),min(min(s(plus(0,?x:Nat)),plus(s(?y:Nat),?x:Nat)),plus(plus(s(0),?y:Nat),min(?y:Nat,?x:Nat))))",
"s(plus(0,s(plus(plus(?x:Nat,?y:Nat),plus(?x:Nat,?z:Nat))))) = plus(plus(plus(plus(plus(0,?y:Nat),?x:Nat),plus(plus(?x:Nat,?x:Nat),plus(?x:Nat,?y:Nat))),min(min(max(?x:Nat,?x:Nat),?y:Nat),?x:Nat)),s(min(max(?x:Nat,0),plus(max(?x:Nat,?x:Nat),max(?y:Nat,?x:Nat)))))",
"plus(max(?y:Nat,min(min(?x:Nat,max(0,?z:Nat)),plus(?y:Nat,min(0,0)))),?x:Nat) = max(s(s(s(?y:Nat))),plus(plus(min(min(?z:Nat,0),min(0,?x:Nat)),s(min(?z:Nat,?y:Nat))),plus(max(0,min(?x:Nat,?x:Nat)),max(s(?y:Nat),min(?x:Nat,?x:Nat)))))",
"s(min(max(plus(?x:Nat,?x:Nat),?x:Nat),plus(?z:Nat,max(plus(0,0),plus(?z:Nat,?x:Nat))))) = max(max(0,0),max(plus(plus(max(?z:Nat,0),plus(?z:Nat,?y:Nat)),min(?z:Nat,s(?x:Nat))),plus(min(0,plus(?z:Nat,?x:Nat)),max(s(?y:Nat),min(?z:Nat,?y:Nat)))))",
"max(s(0),plus(min(min(s(?x:Nat),?y:Nat),max(max(?x:Nat,?x:Nat),min(0,?z:Nat))),?x:Nat)) = plus(max(min(0,?y:Nat),min(s(s(?x:Nat)),s(max(?y:Nat,?x:Nat)))),min(?y:Nat,plus(max(?z:Nat,plus(0,?y:Nat)),plus(plus(?y:Nat,?y:Nat),min(?x:Nat,?z:Nat)))))",
"plus(?y:Nat,s(max(min(max(?x:Nat,?y:Nat),plus(?z:Nat,?y:Nat)),min(max(0,?x:Nat),s(0))))) = plus(max(?x:Nat,min(0,plus(s(?x:Nat),max(?x:Nat,?y:Nat)))),max(min(max(?x:Nat,s(?z:Nat)),s(plus(?z:Nat,?y:Nat))),plus(?x:Nat,plus(plus(?z:Nat,?x:Nat),?x:Nat))))",
"max(max(s(?y:Nat),s(?x:Nat)),min(?y:Nat,plus(plus(max(?y:Nat,?x:Nat),s(?x:Nat)),s(s(0))))) = plus(s(min(min(?x:Nat,max(0,?x:Nat)),s(plus(0,?z:Nat)))),max(plus(?x:Nat,plus(?x:Nat,s(?z:Nat))),s(plus(plus(?x:Nat,?y:Nat),max(?x:Nat,?y:Nat)))))",
"min(max(min(0,0),s(plus(min(?x:Nat,?y:Nat),s(?y:Nat)))),max(s(max(?z:Nat,plus(?y:Nat,?x:Nat))),?z:Nat)) = min(s(plus(?z:Nat,plus(0,plus(0,?x:Nat)))),min(min(max(max(?x:Nat,0),0),?x:Nat),plus(?x:Nat,max(min(?x:Nat,?x:Nat),s(?x:Nat)))))",
"plus(max(plus(min(min(?z:Nat,0),min(?x:Nat,?x:Nat)),plus(s(?x:Nat),s(?y:Nat))),min(max(plus(?x:Nat,?y:Nat),?x:Nat),?x:Nat)),?x:Nat) = plus(plus(min(?z:Nat,0),plus(s(max(?x:Nat,?x:Nat)),?x:Nat)),plus(?x:Nat,plus(?x:Nat,min(plus(0,?x:Nat),max(0,?x:Nat)))))",
"plus(s(plus(?y:Nat,plus(plus(?y:Nat,?y:Nat),s(?z:Nat)))),min(max(plus(max(?z:Nat,0),s(0)),max(plus(0,?x:Nat),min(0,0))),plus(max(?y:Nat,plus(?y:Nat,?z:Nat)),?y:Nat))) = plus(?x:Nat,max(plus(?y:Nat,0),max(0,plus(?x:Nat,plus(0,?x:Nat)))))",
"plus(max(max(?y:Nat,max(?y:Nat,max(0,?y:Nat))),max(?x:Nat,s(s(?z:Nat)))),plus(plus(plus(?y:Nat,plus(?y:Nat,?z:Nat)),min(max(?z:Nat,?x:Nat),?y:Nat)),s(plus(max(0,?z:Nat),?y:Nat)))) = max(max(plus(?x:Nat,max(max(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat))),?z:Nat),?z:Nat)",
"plus(s(max(s(?x:Nat),plus(plus(?y:Nat,?x:Nat),max(?x:Nat,?y:Nat)))),plus(min(s(max(?x:Nat,?z:Nat)),plus(plus(?z:Nat,?y:Nat),?z:Nat)),min(min(max(?y:Nat,?y:Nat),plus(0,?x:Nat)),min(?x:Nat,plus(0,?z:Nat))))) = plus(?z:Nat,plus(s(?z:Nat),plus(0,s(?x:Nat))))",
"s(plus(plus(plus(min(0,?x:Nat),?x:Nat),s(?y:Nat)),s(s(min(0,?x:Nat))))) = plus(min(max(max(s(0),?x:Nat),plus(0,s(?x:Nat))),plus(min(min(?y:Nat,?x:Nat),?z:Nat),max(s(0),s(?z:Nat)))),s(min(min(s(?x:Nat),plus(?z:Nat,0)),0)))",
"plus(plus(min(s(?z:Nat),s(?x:Nat)),0),s(s(max(min(?y:Nat,?z:Nat),?x:Nat)))) = plus(s(max(max(?x:Nat,0),s(?y:Nat))),max(max(s(s(0)),max(min(?x:Nat,?z:Nat),?y:Nat)),plus(plus(min(0,?x:Nat),min(?x:Nat,?x:Nat)),plus(max(?z:Nat,0),?x:Nat))))",
"s(min(plus(?y:Nat,?z:Nat),max(plus(plus(0,?y:Nat),min(?z:Nat,?x:Nat)),plus(?x:Nat,s(0))))) = min(s(max(0,max(?x:Nat,0))),min(max(plus(plus(?x:Nat,?x:Nat),s(?x:Nat)),max(plus(?z:Nat,?x:Nat),s(?x:Nat))),plus(0,plus(s(?y:Nat),plus(?y:Nat,?x:Nat)))))",
"max(s(0),max(plus(plus(s(?x:Nat),min(?x:Nat,?x:Nat)),plus(s(?z:Nat),min(?x:Nat,?z:Nat))),?z:Nat)) = plus(max(min(0,s(max(0,?z:Nat))),min(min(?x:Nat,?y:Nat),plus(min(?z:Nat,?z:Nat),s(?z:Nat)))),min(max(plus(min(?x:Nat,?z:Nat),0),s(?x:Nat)),?y:Nat))",
"plus(0,min(plus(?x:Nat,min(s(0),plus(?z:Nat,?x:Nat))),s(min(max(0,?x:Nat),plus(?x:Nat,?y:Nat))))) = min(min(max(?x:Nat,0),s(max(s(?y:Nat),min(?x:Nat,?x:Nat)))),min(s(plus(plus(0,?x:Nat),max(?x:Nat,?x:Nat))),min(s(max(?x:Nat,?z:Nat)),0)))",
"s(plus(plus(s(s(?x:Nat)),max(plus(0,?x:Nat),max(0,?y:Nat))),max(plus(s(0),?x:Nat),s(max(?y:Nat,?x:Nat))))) = max(0,min(max(max(plus(?x:Nat,?x:Nat),min(?z:Nat,?x:Nat)),min(max(?y:Nat,?x:Nat),?y:Nat)),s(plus(min(?x:Nat,?x:Nat),max(?y:Nat,?x:Nat)))))",
"plus(plus(?z:Nat,max(min(0,?x:Nat),max(min(?x:Nat,0),s(?x:Nat)))),s(plus(plus(?x:Nat,s(?x:Nat)),min(s(?z:Nat),?z:Nat)))) = min(s(min(min(s(0),max(0,0)),plus(?x:Nat,plus(0,?x:Nat)))),s(min(max(max(?x:Nat,?z:Nat),?z:Nat),s(0))))",
"min(plus(min(0,plus(?x:Nat,plus(?y:Nat,?x:Nat))),min(min(s(?y:Nat),?x:Nat),?x:Nat)),s(s(min(plus(?x:Nat,?z:Nat),plus(0,0))))) = s(plus(max(?x:Nat,min(max(0,0),s(?z:Nat))),plus(s(max(0,?x:Nat)),min(min(0,?z:Nat),plus(?x:Nat,?y:Nat)))))",
"max(plus(s(s(s(?x:Nat))),max(max(min(?x:Nat,?z:Nat),?z:Nat),?y:Nat)),min(?y:Nat,max(max(plus(?x:Nat,?x:Nat),?z:Nat),plus(s(?x:Nat),s(?z:Nat))))) = min(max(plus(?z:Nat,max(min(?x:Nat,?z:Nat),plus(0,0))),?x:Nat),min(s(min(min(?y:Nat,?z:Nat),?y:Nat)),?x:Nat))",
"plus(min(min(s(max(?y:Nat,?x:Nat)),s(min(?x:Nat,?x:Nat))),?z:Nat),plus(max(s(?x:Nat),?x:Nat),max(min(min(?x:Nat,0),s(?x:Nat)),s(plus(0,?x:Nat))))) = max(plus(plus(?z:Nat,plus(plus(?x:Nat,?z:Nat),?y:Nat)),s(max(s(?x:Nat),plus(?y:Nat,?z:Nat)))),s(0))",
"max(plus(s(plus(plus(?x:Nat,?x:Nat),min(0,?z:Nat))),?y:Nat),min(s(plus(min(?x:Nat,?x:Nat),s(?x:Nat))),min(max(s(?y:Nat),plus(?y:Nat,?y:Nat)),s(s(?x:Nat))))) = min(s(max(s(?x:Nat),0)),plus(?x:Nat,max(?x:Nat,plus(min(?x:Nat,?z:Nat),max(0,0)))))",
"max(plus(plus(max(?y:Nat,?z:Nat),s(?z:Nat)),plus(?x:Nat,0)),plus(plus(s(s(?z:Nat)),min(min(?x:Nat,?z:Nat),?y:Nat)),max(min(0,plus(?x:Nat,?y:Nat)),max(s(?x:Nat),max(?y:Nat,0))))) = min(max(max(min(min(?y:Nat,0),max(?x:Nat,?x:Nat)),0),?y:Nat),?y:Nat)",
"min(plus(max(s(plus(?x:Nat,?x:Nat)),plus(?y:Nat,plus(0,?y:Nat))),max(plus(s(?x:Nat),max(?x:Nat,?x:Nat)),s(?x:Nat))),min(plus(plus(plus(0,?x:Nat),min(?x:Nat,?x:Nat)),0),plus(?x:Nat,s(plus(0,?x:Nat))))) = plus(0,s(plus(0,s(min(?x:Nat,?x:Nat)))))",
"min(s(min(max(min(?x:Nat,?x:Nat),s(?x:Nat)),s(plus(0,?y:Nat)))),max(0,?y:Nat)) = max(max(max(s(plus(0,?z:Nat)),?z:Nat),min(max(plus(?y:Nat,?y:Nat),plus(?z:Nat,?y:Nat)),0)),max(max(max(?y:Nat,s(0)),?x:Nat),s(min(min(?z:Nat,?x:Nat),s(?z:Nat)))))",
"min(min(min(0,min(max(?x:Nat,?x:Nat),s(?z:Nat))),?x:Nat),plus(max(min(plus(?y:Nat,?z:Nat),plus(?x:Nat,?z:Nat)),?z:Nat),?z:Nat)) = min(?x:Nat,max(plus(s(plus(0,?x:Nat)),max(0,?x:Nat)),plus(min(plus(?y:Nat,?z:Nat),plus(0,?z:Nat)),plus(s(?z:Nat),max(?z:Nat,?y:Nat)))))",
"max(min(plus(?x:Nat,s(s(?z:Nat))),plus(?x:Nat,0)),min(min(s(?z:Nat),s(plus(0,0))),plus(min(?x:Nat,0),?y:Nat))) = plus(max(plus(plus(plus(?z:Nat,?x:Nat),?x:Nat),?x:Nat),?y:Nat),min(max(?y:Nat,s(max(?y:Nat,?y:Nat))),plus(?y:Nat,min(max(?z:Nat,?y:Nat),?y:Nat))))",
"plus(?x:Nat,max(min(plus(?y:Nat,max(?x:Nat,0)),s(max(?z:Nat,?z:Nat))),plus(s(?z:Nat),min(max(?y:Nat,?z:Nat),plus(?y:Nat,?x:Nat))))) = plus(max(min(?x:Nat,plus(s(?z:Nat),s(0))),plus(plus(s(?x:Nat),s(?y:Nat)),s(plus(?x:Nat,?x:Nat)))),plus(?x:Nat,plus(?z:Nat,?x:Nat)))",
"plus(plus(?x:Nat,plus(?y:Nat,min(max(?y:Nat,?x:Nat),s(0)))),min(plus(s(?z:Nat),?x:Nat),min(?x:Nat,max(plus(?z:Nat,?y:Nat),s(?x:Nat))))) = min(plus(min(max(min(?y:Nat,0),max(?x:Nat,0)),s(min(?x:Nat,?x:Nat))),min(?x:Nat,max(s(0),max(?x:Nat,?x:Nat)))),0)",
"max(plus(?z:Nat,plus(max(plus(?x:Nat,?x:Nat),plus(0,?y:Nat)),plus(?z:Nat,min(0,0)))),min(plus(s(s(?x:Nat)),plus(0,plus(?x:Nat,?x:Nat))),s(?y:Nat))) = min(max(plus(?x:Nat,s(min(?x:Nat,?z:Nat))),s(max(?y:Nat,s(?y:Nat)))),s(s(s(plus(?x:Nat,?x:Nat)))))",
"min(max(s(?x:Nat),plus(?z:Nat,plus(?x:Nat,max(?x:Nat,?x:Nat)))),plus(max(plus(s(?z:Nat),max(0,0)),0),plus(plus(?y:Nat,s(?x:Nat)),plus(?y:Nat,s(0))))) = plus(?y:Nat,plus(plus(max(max(0,?x:Nat),?x:Nat),plus(max(0,0),s(?x:Nat))),s(s(?x:Nat))))",
"plus(s(min(min(min(?z:Nat,0),0),s(s(?x:Nat)))),min(s(plus(min(0,0),s(?z:Nat))),plus(s(?x:Nat),min(min(0,?x:Nat),plus(?x:Nat,?y:Nat))))) = plus(plus(0,max(?y:Nat,min(s(?y:Nat),plus(?z:Nat,0)))),plus(0,s(s(plus(?z:Nat,?x:Nat)))))",
"plus(s(plus(plus(min(?z:Nat,0),?x:Nat),?z:Nat)),plus(plus(?y:Nat,plus(plus(?x:Nat,0),s(?y:Nat))),plus(plus(0,s(?x:Nat)),plus(plus(0,?y:Nat),min(0,?z:Nat))))) = plus(min(?x:Nat,s(s(max(?x:Nat,?x:Nat)))),max(s(min(plus(?y:Nat,?x:Nat),?x:Nat)),s(?y:Nat)))",
"plus(max(max(max(plus(?y:Nat,?x:Nat),min(?z:Nat,?z:Nat)),plus(s(?x:Nat),plus(?z:Nat,?y:Nat))),s(plus(plus(0,?x:Nat),?z:Nat))),s(plus(s(min(?x:Nat,?y:Nat)),plus(min(?x:Nat,0),s(?x:Nat))))) = plus(min(max(?x:Nat,min(s(0),s(?x:Nat))),?x:Nat),plus(?z:Nat,?z:Nat))",
"plus(min(plus(s(plus(0,0)),?y:Nat),plus(s(max(?y:Nat,?x:Nat)),plus(min(0,?y:Nat),?x:Nat))),plus(min(s(?z:Nat),plus(plus(?x:Nat,0),?y:Nat)),plus(plus(min(?z:Nat,0),?z:Nat),?z:Nat))) = plus(plus(?x:Nat,min(plus(max(?x:Nat,?x:Nat),max(?x:Nat,?z:Nat)),0)),?y:Nat)",
"s(s(plus(plus(s(?x:Nat),?y:Nat),?y:Nat))) = max(min(plus(s(?y:Nat),s(?x:Nat)),plus(min(max(?x:Nat,?y:Nat),s(?x:Nat)),plus(min(?z:Nat,?x:Nat),min(0,?z:Nat)))),max(plus(max(max(?z:Nat,?x:Nat),min(?z:Nat,0)),?x:Nat),max(max(?z:Nat,plus(?x:Nat,?z:Nat)),s(s(0)))))",
"min(max(plus(plus(max(0,?z:Nat),max(?x:Nat,?z:Nat)),s(plus(?y:Nat,0))),min(min(?x:Nat,plus(?x:Nat,?x:Nat)),0)),s(?x:Nat)) = plus(plus(?y:Nat,max(?z:Nat,max(max(?x:Nat,0),0))),max(min(plus(?x:Nat,min(?y:Nat,?y:Nat)),s(min(?y:Nat,?x:Nat))),max(?y:Nat,s(?x:Nat))))",
"plus(min(min(max(s(0),plus(0,?x:Nat)),max(?z:Nat,max(0,0))),?x:Nat),s(plus(?x:Nat,max(max(?x:Nat,?x:Nat),?x:Nat)))) = plus(min(min(plus(s(?x:Nat),max(0,?y:Nat)),max(min(?x:Nat,?y:Nat),plus(?x:Nat,?y:Nat))),?x:Nat),min(max(s(plus(0,?y:Nat)),?z:Nat),?x:Nat))",
"plus(s(s(?y:Nat)),min(plus(plus(0,min(?x:Nat,0)),min(max(0,?y:Nat),min(0,0))),max(s(plus(?x:Nat,?y:Nat)),?x:Nat))) = min(?x:Nat,plus(plus(max(min(?x:Nat,0),plus(0,?y:Nat)),?y:Nat),plus(max(plus(?z:Nat,?x:Nat),0),plus(max(?x:Nat,?z:Nat),s(0)))))",
"plus(min(?y:Nat,plus(plus(?x:Nat,max(?x:Nat,?y:Nat)),max(s(?x:Nat),plus(?x:Nat,?z:Nat)))),plus(?x:Nat,plus(plus(0,max(?z:Nat,?z:Nat)),0))) = max(s(min(plus(s(?x:Nat),plus(?z:Nat,?x:Nat)),max(plus(?z:Nat,0),min(0,?x:Nat)))),s(plus(s(0),s(plus(?x:Nat,?y:Nat)))))",
"plus(?x:Nat,min(plus(min(plus(?y:Nat,?x:Nat),max(0,0)),s(?x:Nat)),min(s(min(?x:Nat,?x:Nat)),min(min(?y:Nat,?z:Nat),plus(?x:Nat,?x:Nat))))) = max(min(plus(0,s(min(0,?y:Nat))),max(0,?x:Nat)),s(plus(min(?x:Nat,min(0,?y:Nat)),min(min(?x:Nat,?x:Nat),?y:Nat))))",
"plus(max(max(?x:Nat,plus(0,s(?y:Nat))),plus(max(plus(?x:Nat,0),0),s(plus(0,?y:Nat)))),plus(s(plus(s(?z:Nat),?x:Nat)),?y:Nat)) = s(plus(plus(s(max(?x:Nat,0)),s(plus(?x:Nat,0))),max(plus(max(?z:Nat,?x:Nat),min(?y:Nat,?z:Nat)),plus(?x:Nat,s(?y:Nat)))))",
"min(plus(plus(?x:Nat,s(plus(?x:Nat,?x:Nat))),s(min(plus(?z:Nat,?y:Nat),max(?z:Nat,?x:Nat)))),plus(plus(max(s(0),?y:Nat),?y:Nat),s(s(0)))) = plus(min(?x:Nat,?x:Nat),max(plus(min(?y:Nat,max(?x:Nat,?x:Nat)),0),min(plus(?z:Nat,plus(?y:Nat,?z:Nat)),min(s(?y:Nat),?z:Nat))))",
"plus(plus(plus(plus(?x:Nat,s(?x:Nat)),min(?z:Nat,plus(0,?x:Nat))),?y:Nat),plus(s(min(s(?x:Nat),0)),plus(plus(?x:Nat,?y:Nat),min(max(?x:Nat,?x:Nat),0)))) = max(?x:Nat,plus(min(plus(s(0),max(0,?z:Nat)),0),min(s(min(?x:Nat,?x:Nat)),max(s(?y:Nat),?y:Nat))))",
"s(min(max(min(?x:Nat,max(?z:Nat,?x:Nat)),?x:Nat),plus(?y:Nat,max(?z:Nat,s(?y:Nat))))) = min(min(plus(max(min(?z:Nat,?x:Nat),?x:Nat),max(s(?z:Nat),?z:Nat)),s(max(plus(?y:Nat,?y:Nat),max(?z:Nat,?y:Nat)))),plus(s(min(?y:Nat,plus(?x:Nat,?z:Nat))),s(max(min(?z:Nat,?y:Nat),s(?x:Nat)))))",
"plus(s(plus(s(?z:Nat),max(?y:Nat,s(?x:Nat)))),s(min(s(s(?y:Nat)),s(plus(0,0))))) = plus(plus(s(min(plus(?z:Nat,?x:Nat),?x:Nat)),?y:Nat),min(max(plus(plus(?x:Nat,?x:Nat),s(?y:Nat)),max(plus(0,?x:Nat),plus(?x:Nat,?y:Nat))),plus(0,plus(s(?y:Nat),s(?x:Nat)))))",
"s(plus(plus(max(?y:Nat,max(0,0)),min(?x:Nat,plus(0,?x:Nat))),plus(s(min(0,?x:Nat)),min(s(?y:Nat),s(?z:Nat))))) = plus(s(plus(s(min(?x:Nat,?y:Nat)),?y:Nat)),plus(min(plus(?x:Nat,?x:Nat),max(min(?y:Nat,?y:Nat),plus(?y:Nat,?y:Nat))),plus(0,max(?y:Nat,s(?y:Nat)))))",
"min(s(max(plus(max(?z:Nat,?y:Nat),min(?x:Nat,?y:Nat)),plus(s(?x:Nat),plus(?x:Nat,?x:Nat)))),plus(max(max(min(?y:Nat,?z:Nat),0),?x:Nat),?y:Nat)) = plus(plus(plus(?y:Nat,?x:Nat),min(?y:Nat,max(?x:Nat,plus(?z:Nat,?y:Nat)))),s(max(plus(?x:Nat,plus(?x:Nat,?x:Nat)),min(min(?x:Nat,?x:Nat),0))))",
"plus(min(?x:Nat,max(?x:Nat,plus(plus(?z:Nat,?y:Nat),min(?z:Nat,?z:Nat)))),min(plus(plus(s(?y:Nat),plus(0,?x:Nat)),s(0)),plus(0,s(s(?x:Nat))))) = min(min(s(?y:Nat),plus(min(plus(?z:Nat,0),?x:Nat),?x:Nat)),min(max(plus(plus(?z:Nat,?x:Nat),max(?x:Nat,?x:Nat)),?y:Nat),?x:Nat))",
"plus(min(s(min(?x:Nat,plus(0,?z:Nat))),plus(s(?x:Nat),s(min(?x:Nat,?x:Nat)))),plus(min(?y:Nat,min(s(?y:Nat),?x:Nat)),max(min(s(?z:Nat),s(?x:Nat)),s(?x:Nat)))) = s(min(plus(?y:Nat,min(plus(?x:Nat,?z:Nat),max(?z:Nat,?x:Nat))),plus(s(?z:Nat),min(?x:Nat,min(?z:Nat,0)))))",
"max(max(s(?z:Nat),max(plus(plus(?z:Nat,?y:Nat),s(?y:Nat)),max(?z:Nat,max(?z:Nat,?x:Nat)))),max(plus(?x:Nat,plus(s(?x:Nat),0)),plus(?x:Nat,plus(s(?x:Nat),max(?y:Nat,?x:Nat))))) = max(s(s(plus(?z:Nat,?z:Nat))),min(min(s(plus(?x:Nat,?x:Nat)),min(?y:Nat,min(?x:Nat,?z:Nat))),0))",
"max(plus(plus(s(?x:Nat),?y:Nat),min(max(plus(?y:Nat,?z:Nat),plus(?y:Nat,0)),min(min(?x:Nat,?z:Nat),0))),plus(min(plus(?z:Nat,s(?x:Nat)),max(?z:Nat,plus(?z:Nat,?y:Nat))),s(?x:Nat))) = max(max(s(min(plus(0,?x:Nat),plus(0,0))),s(max(plus(?z:Nat,?x:Nat),?x:Nat))),?x:Nat)",
"min(plus(plus(plus(s(?z:Nat),?x:Nat),min(plus(?x:Nat,?x:Nat),?y:Nat)),0),min(plus(?x:Nat,min(s(0),max(0,?x:Nat))),max(s(plus(?x:Nat,?y:Nat)),max(min(0,?x:Nat),min(?z:Nat,?x:Nat))))) = s(max(0,min(min(s(?x:Nat),max(?z:Nat,?y:Nat)),max(?y:Nat,plus(?x:Nat,?x:Nat)))))",
"max(?x:Nat,min(s(?x:Nat),max(?x:Nat,max(s(?x:Nat),s(?z:Nat))))) = plus(plus(min(s(min(?x:Nat,0)),?x:Nat),plus(max(min(0,?x:Nat),s(?x:Nat)),max(plus(0,?x:Nat),min(?z:Nat,?x:Nat)))),max(min(plus(plus(?x:Nat,?y:Nat),?z:Nat),s(max(?y:Nat,0))),s(plus(?x:Nat,s(?z:Nat)))))",
"max(max(0,s(?x:Nat)),min(min(plus(s(?x:Nat),?x:Nat),?x:Nat),0)) = plus(plus(min(max(max(?x:Nat,?y:Nat),s(?z:Nat)),plus(min(?x:Nat,?z:Nat),max(?y:Nat,?y:Nat))),max(min(?z:Nat,max(?z:Nat,?x:Nat)),plus(0,max(0,0)))),plus(min(?x:Nat,max(s(0),max(?x:Nat,?y:Nat))),?x:Nat))",
"max(s(max(plus(plus(0,0),?z:Nat),?y:Nat)),s(max(s(s(?z:Nat)),?z:Nat))) = max(max(s(min(s(?z:Nat),?x:Nat)),plus(min(0,0),plus(max(?y:Nat,?y:Nat),max(?x:Nat,?z:Nat)))),max(s(min(?x:Nat,s(?y:Nat))),max(plus(s(0),plus(?x:Nat,?x:Nat)),s(max(?x:Nat,?x:Nat)))))",
"min(min(min(?y:Nat,s(plus(?z:Nat,?x:Nat))),?x:Nat),max(0,max(plus(min(?x:Nat,0),s(?z:Nat)),plus(s(?y:Nat),plus(?x:Nat,?x:Nat))))) = max(min(?x:Nat,max(s(s(0)),plus(s(?y:Nat),min(?x:Nat,?x:Nat)))),min(max(min(plus(?y:Nat,?x:Nat),plus(?z:Nat,?x:Nat)),min(?z:Nat,0)),?x:Nat))",
"min(0,plus(min(plus(max(?y:Nat,0),?x:Nat),s(min(?y:Nat,?y:Nat))),plus(plus(min(?x:Nat,0),max(?z:Nat,?y:Nat)),plus(?x:Nat,s(?y:Nat))))) = plus(plus(max(?x:Nat,s(min(?y:Nat,?x:Nat))),min(s(0),plus(?x:Nat,plus(?z:Nat,?x:Nat)))),plus(s(0),s(plus(0,min(?x:Nat,0)))))",
"plus(max(max(min(?z:Nat,?x:Nat),s(max(?y:Nat,?y:Nat))),min(plus(min(?x:Nat,?x:Nat),min(0,?y:Nat)),min(s(?x:Nat),plus(0,?x:Nat)))),?y:Nat) = max(plus(max(plus(0,max(?y:Nat,?y:Nat)),plus(?x:Nat,?x:Nat)),plus(max(?x:Nat,min(?x:Nat,?x:Nat)),min(plus(?z:Nat,?y:Nat),plus(?x:Nat,?x:Nat)))),?y:Nat)",
"min(min(0,max(plus(?z:Nat,plus(0,0)),max(?x:Nat,max(0,0)))),plus(s(?z:Nat),max(max(plus(?x:Nat,?z:Nat),max(?x:Nat,?x:Nat)),max(s(?x:Nat),?y:Nat)))) = max(plus(max(plus(?x:Nat,s(?x:Nat)),min(s(?x:Nat),max(?z:Nat,?z:Nat))),plus(max(?z:Nat,max(?x:Nat,?y:Nat)),?x:Nat)),0)",
"min(plus(plus(plus(0,?x:Nat),max(s(?x:Nat),plus(0,?y:Nat))),max(max(max(0,?z:Nat),max(?z:Nat,?x:Nat)),s(s(?y:Nat)))),plus(plus(min(max(?y:Nat,?y:Nat),plus(?x:Nat,?y:Nat)),?x:Nat),?z:Nat)) = s(min(min(max(s(?x:Nat),0),max(min(0,?z:Nat),min(0,?z:Nat))),s(?x:Nat)))",
"min(s(?z:Nat),plus(max(?x:Nat,?z:Nat),plus(?x:Nat,s(max(?z:Nat,?x:Nat))))) = plus(min(plus(max(max(?y:Nat,?x:Nat),?x:Nat),plus(?y:Nat,plus(?z:Nat,?z:Nat))),min(s(s(0)),max(?y:Nat,s(?x:Nat)))),max(s(min(min(?x:Nat,?z:Nat),s(?x:Nat))),plus(s(max(0,?x:Nat)),plus(0,s(?x:Nat)))))",
"max(min(s(?z:Nat),max(min(min(?z:Nat,?x:Nat),plus(?x:Nat,0)),?z:Nat)),plus(?y:Nat,?x:Nat)) = max(plus(plus(plus(s(?x:Nat),min(0,0)),?y:Nat),max(?x:Nat,plus(?x:Nat,min(?x:Nat,?x:Nat)))),max(s(plus(s(?y:Nat),plus(0,0))),plus(min(plus(0,?z:Nat),?y:Nat),s(min(?y:Nat,?x:Nat)))))",
"plus(plus(max(plus(?y:Nat,max(0,?z:Nat)),0),min(s(plus(?x:Nat,?z:Nat)),plus(?x:Nat,plus(0,0)))),?z:Nat) = plus(max(max(max(s(?x:Nat),min(?x:Nat,?x:Nat)),min(s(?x:Nat),?z:Nat)),s(?x:Nat)),max(plus(?x:Nat,s(s(?z:Nat))),plus(plus(0,s(?x:Nat)),min(?x:Nat,max(0,?y:Nat)))))",
"min(plus(plus(s(s(?z:Nat)),?y:Nat),min(?z:Nat,?x:Nat)),min(min(min(min(?y:Nat,0),plus(?x:Nat,0)),0),plus(?x:Nat,max(?y:Nat,s(?x:Nat))))) = min(min(min(s(plus(?z:Nat,?x:Nat)),min(max(?x:Nat,?z:Nat),max(?y:Nat,?x:Nat))),plus(s(s(0)),plus(?x:Nat,min(?x:Nat,?x:Nat)))),s(?x:Nat))",
"plus(max(s(min(?x:Nat,plus(?x:Nat,?z:Nat))),min(s(max(?x:Nat,?x:Nat)),min(?x:Nat,min(?x:Nat,0)))),plus(?z:Nat,s(max(plus(?z:Nat,?y:Nat),max(?y:Nat,?x:Nat))))) = plus(plus(max(plus(max(?x:Nat,?x:Nat),?x:Nat),s(min(?z:Nat,?z:Nat))),plus(plus(max(?z:Nat,?x:Nat),s(?z:Nat)),?x:Nat)),min(?y:Nat,?x:Nat))",
"min(min(plus(plus(s(?x:Nat),?z:Nat),s(min(?y:Nat,0))),plus(min(?x:Nat,?y:Nat),s(?z:Nat))),plus(max(max(?y:Nat,plus(?x:Nat,?x:Nat)),plus(s(?x:Nat),?y:Nat)),0)) = min(min(min(s(min(?z:Nat,?x:Nat)),max(max(?z:Nat,?z:Nat),max(?z:Nat,?z:Nat))),s(plus(plus(?x:Nat,?z:Nat),s(?z:Nat)))),?x:Nat)",
"plus(plus(plus(s(s(?x:Nat)),max(plus(?y:Nat,?y:Nat),plus(?x:Nat,?z:Nat))),?z:Nat),min(min(max(plus(?x:Nat,?z:Nat),max(?y:Nat,?z:Nat)),plus(s(?x:Nat),min(?x:Nat,?z:Nat))),s(s(plus(0,?x:Nat))))) = max(?x:Nat,plus(?z:Nat,max(plus(min(?x:Nat,?x:Nat),min(?x:Nat,?z:Nat)),plus(max(?x:Nat,?x:Nat),?y:Nat))))",
"max(0,plus(min(plus(0,min(?y:Nat,?x:Nat)),plus(s(0),max(?y:Nat,0))),min(min(min(?y:Nat,?x:Nat),?x:Nat),?x:Nat))) = max(plus(s(?z:Nat),plus(plus(min(0,0),plus(?z:Nat,?y:Nat)),s(min(?y:Nat,0)))),min(0,max(min(max(?z:Nat,0),plus(?x:Nat,?z:Nat)),min(?y:Nat,s(?y:Nat)))))",
"plus(max(?x:Nat,plus(s(?x:Nat),min(plus(?x:Nat,?y:Nat),min(?y:Nat,?z:Nat)))),min(s(plus(?y:Nat,0)),s(min(?z:Nat,plus(?x:Nat,?x:Nat))))) = min(min(plus(min(max(?y:Nat,?z:Nat),max(?y:Nat,?z:Nat)),s(0)),min(0,s(0))),min(s(s(?y:Nat)),max(max(0,plus(?x:Nat,0)),s(?x:Nat))))",
"plus(min(plus(min(min(?x:Nat,?z:Nat),max(?y:Nat,?y:Nat)),plus(min(?x:Nat,?y:Nat),min(?y:Nat,?x:Nat))),min(?x:Nat,s(max(?y:Nat,0)))),?z:Nat) = max(s(max(max(max(?z:Nat,0),?z:Nat),plus(min(?x:Nat,?z:Nat),?y:Nat))),min(plus(plus(plus(?x:Nat,?x:Nat),?x:Nat),max(s(?x:Nat),max(?x:Nat,?x:Nat))),s(0)))",
"plus(plus(s(max(?x:Nat,0)),min(?x:Nat,max(?y:Nat,0))),plus(max(?y:Nat,s(plus(?x:Nat,?x:Nat))),min(max(?x:Nat,plus(?x:Nat,?y:Nat)),s(0)))) = min(max(plus(plus(min(0,0),plus(0,?x:Nat)),min(plus(?y:Nat,0),plus(?y:Nat,?x:Nat))),?x:Nat),s(max(min(?x:Nat,min(0,0)),?z:Nat)))",
"plus(s(plus(?y:Nat,?x:Nat)),min(plus(min(?z:Nat,min(?y:Nat,0)),plus(plus(0,0),max(?x:Nat,?y:Nat))),min(max(plus(?x:Nat,?y:Nat),plus(?y:Nat,?x:Nat)),plus(plus(?x:Nat,?z:Nat),min(?x:Nat,?x:Nat))))) = s(plus(max(plus(max(0,?x:Nat),plus(?y:Nat,?z:Nat)),?z:Nat),plus(?x:Nat,max(max(0,?x:Nat),?y:Nat))))",
"plus(plus(max(plus(plus(?x:Nat,?z:Nat),plus(?x:Nat,?z:Nat)),min(s(0),?x:Nat)),max(plus(min(?x:Nat,0),0),plus(plus(?x:Nat,?z:Nat),?y:Nat))),plus(max(plus(plus(?z:Nat,?x:Nat),?y:Nat),max(s(0),s(?z:Nat))),min(s(min(0,0)),s(?y:Nat)))) = s(min(?y:Nat,s(s(plus(0,?y:Nat)))))",
"s(s(plus(min(max(?z:Nat,?x:Nat),?x:Nat),max(?x:Nat,?x:Nat)))) = min(min(plus(min(plus(?y:Nat,?y:Nat),min(0,?x:Nat)),min(s(?x:Nat),s(?x:Nat))),max(?y:Nat,plus(?z:Nat,?z:Nat))),max(min(min(s(?x:Nat),plus(?y:Nat,?z:Nat)),s(?x:Nat)),plus(min(s(?z:Nat),?z:Nat),min(min(0,?z:Nat),min(?x:Nat,0)))))",
"max(max(min(min(s(?x:Nat),?x:Nat),?z:Nat),max(?y:Nat,min(plus(0,?y:Nat),min(?x:Nat,?y:Nat)))),s(plus(0,?x:Nat))) = min(plus(min(min(plus(?z:Nat,?x:Nat),max(0,?y:Nat)),plus(s(?x:Nat),max(0,?z:Nat))),?x:Nat),max(plus(plus(?x:Nat,s(?z:Nat)),s(plus(?z:Nat,?x:Nat))),plus(s(s(?y:Nat)),?y:Nat)))",
"max(plus(s(max(min(?x:Nat,?z:Nat),max(?y:Nat,?x:Nat))),?z:Nat),plus(max(?x:Nat,min(s(?y:Nat),min(?x:Nat,?x:Nat))),0)) = plus(min(plus(?x:Nat,max(0,min(?z:Nat,?y:Nat))),s(s(s(0)))),min(plus(s(s(0)),s(?x:Nat)),min(plus(max(0,?z:Nat),s(?x:Nat)),min(plus(?z:Nat,?x:Nat),0))))",
"plus(0,max(min(min(s(0),plus(?x:Nat,0)),plus(plus(0,?x:Nat),plus(?z:Nat,0))),min(s(min(0,?x:Nat)),s(max(?y:Nat,?x:Nat))))) = min(s(min(?z:Nat,plus(s(?y:Nat),0))),plus(plus(s(plus(0,0)),min(plus(?x:Nat,?x:Nat),s(?x:Nat))),plus(?z:Nat,max(?y:Nat,plus(?x:Nat,?x:Nat)))))",
"min(plus(max(plus(max(?z:Nat,?x:Nat),?x:Nat),max(max(?x:Nat,0),max(?z:Nat,?x:Nat))),plus(plus(?z:Nat,plus(?x:Nat,?x:Nat)),s(min(0,?y:Nat)))),s(max(s(?y:Nat),0))) = plus(s(plus(?z:Nat,s(?y:Nat))),max(max(s(?y:Nat),plus(?x:Nat,min(?x:Nat,?y:Nat))),min(max(0,plus(0,?x:Nat)),s(0))))",
"plus(plus(min(s(?x:Nat),plus(plus(?y:Nat,?x:Nat),plus(?x:Nat,?x:Nat))),min(s(min(?x:Nat,?y:Nat)),plus(max(?z:Nat,?y:Nat),min(?y:Nat,?y:Nat)))),?x:Nat) = min(min(s(max(s(?x:Nat),s(?x:Nat))),min(?y:Nat,min(plus(?y:Nat,?z:Nat),s(?y:Nat)))),max(s(min(?x:Nat,?y:Nat)),min(0,min(max(?y:Nat,?x:Nat),s(?y:Nat)))))",
"max(max(s(min(?z:Nat,s(0))),max(plus(?z:Nat,?x:Nat),?x:Nat)),plus(min(s(max(?x:Nat,?x:Nat)),?y:Nat),s(plus(s(?z:Nat),plus(?y:Nat,?y:Nat))))) = plus(s(max(plus(plus(?x:Nat,?y:Nat),s(0)),min(s(?z:Nat),min(?z:Nat,?x:Nat)))),plus(?x:Nat,max(s(s(?z:Nat)),min(plus(0,0),plus(?z:Nat,?z:Nat)))))",
"plus(max(plus(?x:Nat,plus(plus(?z:Nat,?x:Nat),s(?x:Nat))),max(plus(?x:Nat,?x:Nat),plus(max(?z:Nat,0),s(?y:Nat)))),max(?x:Nat,s(min(max(?z:Nat,?z:Nat),s(?z:Nat))))) = plus(plus(?z:Nat,plus(max(plus(?y:Nat,?x:Nat),?x:Nat),?y:Nat)),max(s(?x:Nat),min(s(plus(?y:Nat,?x:Nat)),plus(max(0,?y:Nat),plus(?y:Nat,?x:Nat)))))",
"plus(plus(s(max(plus(?y:Nat,?x:Nat),plus(?x:Nat,?y:Nat))),max(min(?y:Nat,?x:Nat),max(max(?y:Nat,0),min(?y:Nat,?z:Nat)))),plus(s(?y:Nat),s(plus(plus(?x:Nat,0),?x:Nat)))) = max(max(?x:Nat,min(plus(max(?x:Nat,?x:Nat),min(?x:Nat,?z:Nat)),?z:Nat)),max(s(?x:Nat),min(s(s(?y:Nat)),min(?x:Nat,max(?x:Nat,?x:Nat)))))",
"max(min(max(?z:Nat,max(s(?z:Nat),max(?y:Nat,?x:Nat))),s(s(max(?x:Nat,?y:Nat)))),plus(min(?y:Nat,max(plus(?z:Nat,?x:Nat),s(0))),plus(?z:Nat,min(?x:Nat,max(?x:Nat,0))))) = min(plus(plus(plus(s(?y:Nat),max(?x:Nat,?z:Nat)),?z:Nat),plus(max(min(?z:Nat,0),s(?y:Nat)),max(plus(0,?x:Nat),?y:Nat))),?y:Nat)",
"min(max(max(min(?z:Nat,max(?x:Nat,0)),?x:Nat),min(min(max(?z:Nat,?x:Nat),s(?y:Nat)),min(min(?x:Nat,0),?x:Nat))),plus(plus(?y:Nat,plus(?y:Nat,plus(?z:Nat,?z:Nat))),max(s(plus(?x:Nat,?y:Nat)),max(0,plus(?x:Nat,?y:Nat))))) = min(?y:Nat,plus(plus(0,0),plus(?z:Nat,max(plus(0,?x:Nat),plus(?y:Nat,?y:Nat)))))",
"max(min(0,plus(s(s(0)),min(s(?z:Nat),0))),min(max(?x:Nat,max(s(?x:Nat),min(?x:Nat,?x:Nat))),0)) = plus(plus(max(min(max(?x:Nat,?y:Nat),min(?x:Nat,?x:Nat)),min(s(?x:Nat),?x:Nat)),?y:Nat),max(min(min(0,max(?z:Nat,?z:Nat)),plus(min(?z:Nat,?z:Nat),?x:Nat)),min(?x:Nat,min(plus(?y:Nat,0),?z:Nat))))",
"plus(?z:Nat,plus(min(plus(plus(?y:Nat,?z:Nat),min(0,?z:Nat)),max(?x:Nat,?x:Nat)),max(?x:Nat,min(plus(?x:Nat,?z:Nat),max(?y:Nat,0))))) = max(min(plus(min(plus(?y:Nat,?x:Nat),max(?z:Nat,?x:Nat)),plus(?z:Nat,?x:Nat)),plus(?x:Nat,s(min(?y:Nat,?x:Nat)))),max(min(0,plus(?y:Nat,?z:Nat)),max(max(s(0),s(?x:Nat)),?x:Nat)))",
"max(?x:Nat,max(plus(min(?x:Nat,?z:Nat),plus(plus(?x:Nat,?y:Nat),s(?x:Nat))),plus(plus(min(?x:Nat,?y:Nat),max(?z:Nat,?y:Nat)),min(0,0)))) = plus(min(?y:Nat,plus(plus(?y:Nat,?z:Nat),min(plus(?z:Nat,?x:Nat),plus(?z:Nat,?z:Nat)))),plus(plus(?z:Nat,plus(max(?x:Nat,?x:Nat),min(?x:Nat,?y:Nat))),max(plus(?z:Nat,plus(0,?z:Nat)),?z:Nat)))",
"min(plus(plus(plus(s(?y:Nat),plus(?y:Nat,?x:Nat)),min(?y:Nat,max(?y:Nat,?x:Nat))),plus(min(s(?y:Nat),min(?x:Nat,?x:Nat)),?x:Nat)),plus(min(s(s(0)),max(max(?x:Nat,?x:Nat),s(?y:Nat))),s(?z:Nat))) = max(max(?z:Nat,max(plus(?z:Nat,plus(?x:Nat,?z:Nat)),plus(min(?z:Nat,0),?z:Nat))),s(s(s(max(?z:Nat,?y:Nat)))))",
"max(s(min(?x:Nat,plus(min(?z:Nat,?y:Nat),plus(?x:Nat,0)))),s(plus(s(plus(?x:Nat,?z:Nat)),max(plus(?x:Nat,?z:Nat),s(?x:Nat))))) = min(min(max(s(?x:Nat),plus(plus(?x:Nat,?x:Nat),min(0,?x:Nat))),s(s(min(?y:Nat,?x:Nat)))),plus(0,plus(plus(min(?x:Nat,?x:Nat),plus(?x:Nat,?y:Nat)),min(min(0,?x:Nat),s(?z:Nat)))))",
"max(min(plus(?x:Nat,min(min(?x:Nat,?z:Nat),0)),s(s(s(0)))),plus(plus(?z:Nat,plus(min(?z:Nat,?x:Nat),0)),plus(plus(min(?z:Nat,?x:Nat),?z:Nat),?x:Nat))) = plus(s(plus(?z:Nat,min(max(?x:Nat,?x:Nat),s(0)))),max(max(min(0,max(0,?y:Nat)),plus(max(?x:Nat,?x:Nat),s(?z:Nat))),min(max(?x:Nat,s(?x:Nat)),?z:Nat)))",
"max(plus(plus(max(0,min(0,?x:Nat)),plus(plus(0,?y:Nat),plus(?y:Nat,?z:Nat))),?x:Nat),plus(s(plus(?z:Nat,plus(?x:Nat,?x:Nat))),plus(min(s(?z:Nat),?z:Nat),plus(?x:Nat,?x:Nat)))) = plus(max(min(min(s(?z:Nat),max(?x:Nat,0)),max(min(?x:Nat,?z:Nat),min(?z:Nat,?y:Nat))),s(?x:Nat)),min(s(?x:Nat),s(s(min(?x:Nat,?z:Nat)))))",
"plus(plus(plus(s(max(?z:Nat,0)),s(s(?x:Nat))),min(min(s(?x:Nat),s(?y:Nat)),plus(min(?x:Nat,0),max(?y:Nat,0)))),plus(plus(plus(s(?x:Nat),max(?x:Nat,0)),s(?z:Nat)),s(s(max(0,?x:Nat))))) = max(?x:Nat,plus(min(s(plus(?y:Nat,?z:Nat)),plus(plus(?z:Nat,?z:Nat),plus(?x:Nat,?x:Nat))),plus(?x:Nat,s(s(0)))))",
"plus(s(plus(max(s(0),s(?x:Nat)),plus(max(?x:Nat,?x:Nat),?y:Nat))),plus(min(s(plus(?x:Nat,?z:Nat)),?x:Nat),min(s(max(?x:Nat,?x:Nat)),0))) = plus(plus(min(max(?x:Nat,plus(?x:Nat,?x:Nat)),s(s(0))),s(max(s(?y:Nat),0))),max(max(plus(min(?x:Nat,?x:Nat),0),min(0,max(?x:Nat,?x:Nat))),max(?x:Nat,s(s(?z:Nat)))))",
"max(max(plus(?x:Nat,s(min(?x:Nat,?y:Nat))),s(plus(max(?z:Nat,?x:Nat),s(?z:Nat)))),max(plus(max(?z:Nat,plus(?z:Nat,?x:Nat)),plus(plus(0,?x:Nat),?x:Nat)),s(s(max(?y:Nat,?y:Nat))))) = plus(plus(plus(min(s(?x:Nat),0),plus(s(?x:Nat),max(?x:Nat,0))),plus(plus(?y:Nat,max(?z:Nat,?y:Nat)),?z:Nat)),s(s(min(plus(0,0),?z:Nat))))",
"max(plus(min(plus(?x:Nat,max(0,?z:Nat)),min(min(?y:Nat,?z:Nat),plus(?y:Nat,?x:Nat))),min(0,max(?x:Nat,?y:Nat))),plus(max(s(?y:Nat),plus(0,?x:Nat)),max(plus(?x:Nat,plus(?x:Nat,?y:Nat)),min(max(?x:Nat,?z:Nat),plus(?x:Nat,?x:Nat))))) = s(plus(s(max(s(0),min(?z:Nat,0))),max(max(min(?x:Nat,?x:Nat),min(?z:Nat,0)),s(?z:Nat))))",
"plus(plus(plus(s(plus(?x:Nat,0)),s(0)),min(max(min(?x:Nat,?x:Nat),?z:Nat),min(?y:Nat,?z:Nat))),plus(plus(plus(plus(0,?x:Nat),s(?y:Nat)),max(s(?z:Nat),plus(?y:Nat,?z:Nat))),min(max(s(?x:Nat),max(?y:Nat,0)),min(min(?y:Nat,?z:Nat),plus(?z:Nat,?z:Nat))))) = max(max(?z:Nat,0),s(s(min(plus(?x:Nat,?x:Nat),plus(?x:Nat,?x:Nat)))))",
"min(plus(min(plus(plus(?x:Nat,?y:Nat),s(?y:Nat)),plus(0,s(?z:Nat))),min(min(plus(?z:Nat,0),?y:Nat),?y:Nat)),?z:Nat) = max(plus(max(min(s(?z:Nat),?z:Nat),plus(plus(?x:Nat,?y:Nat),plus(?x:Nat,?y:Nat))),plus(max(plus(0,0),0),s(?y:Nat))),min(plus(s(max(?x:Nat,0)),?z:Nat),plus(s(plus(?y:Nat,?x:Nat)),min(?x:Nat,min(0,?y:Nat)))))",
"max(max(plus(0,s(plus(?z:Nat,?y:Nat))),s(plus(s(0),max(?z:Nat,?x:Nat)))),min(min(min(0,?x:Nat),s(plus(?x:Nat,?y:Nat))),max(s(?y:Nat),max(max(0,?x:Nat),0)))) = max(s(max(?x:Nat,min(min(?z:Nat,0),plus(0,0)))),plus(s(?y:Nat),plus(min(plus(?z:Nat,?y:Nat),min(?z:Nat,?y:Nat)),min(s(0),plus(?y:Nat,?z:Nat)))))",
"s(plus(min(0,min(plus(?x:Nat,0),max(?x:Nat,?y:Nat))),max(plus(min(?z:Nat,?y:Nat),min(0,?x:Nat)),plus(plus(0,?x:Nat),max(?z:Nat,0))))) = min(plus(plus(min(max(0,?y:Nat),plus(?x:Nat,?x:Nat)),max(?x:Nat,plus(?x:Nat,?x:Nat))),min(plus(s(?x:Nat),s(0)),0)),max(s(0),min(min(?x:Nat,max(?x:Nat,?y:Nat)),min(?z:Nat,max(?y:Nat,?x:Nat)))))",
"min(plus(s(min(max(?x:Nat,0),?z:Nat)),plus(max(?x:Nat,max(0,?x:Nat)),?x:Nat)),plus(min(plus(min(?x:Nat,?y:Nat),plus(?x:Nat,0)),s(?y:Nat)),plus(s(plus(?y:Nat,0)),s(?y:Nat)))) = max(plus(?z:Nat,min(s(s(?z:Nat)),?x:Nat)),max(max(plus(?x:Nat,?z:Nat),min(min(?x:Nat,?z:Nat),plus(0,?y:Nat))),max(plus(plus(?x:Nat,?y:Nat),?x:Nat),plus(?z:Nat,s(?z:Nat)))))",
"max(min(plus(s(min(?y:Nat,?y:Nat)),plus(min(?x:Nat,0),0)),s(plus(s(?x:Nat),plus(?y:Nat,?x:Nat)))),plus(min(?z:Nat,?y:Nat),plus(max(plus(?x:Nat,?x:Nat),?z:Nat),plus(?y:Nat,max(?x:Nat,?z:Nat))))) = plus(max(min(min(min(?x:Nat,?z:Nat),max(0,0)),plus(plus(0,?x:Nat),?x:Nat)),s(?x:Nat)),plus(max(?x:Nat,plus(?z:Nat,s(?y:Nat))),plus(?z:Nat,min(plus(0,?x:Nat),?z:Nat))))",
"plus(max(min(0,min(plus(0,0),s(?x:Nat))),min(plus(?x:Nat,?x:Nat),s(?x:Nat))),min(plus(max(plus(?x:Nat,0),s(?x:Nat)),s(s(0))),max(?x:Nat,min(plus(?y:Nat,?z:Nat),plus(0,0))))) = min(s(plus(?y:Nat,max(plus(?y:Nat,?x:Nat),?y:Nat))),max(max(plus(?y:Nat,plus(0,?y:Nat)),?x:Nat),max(min(max(0,?y:Nat),s(?z:Nat)),plus(max(?z:Nat,?z:Nat),max(?x:Nat,?z:Nat)))))",
"max(s(min(max(s(?x:Nat),s(?y:Nat)),max(min(?x:Nat,?y:Nat),plus(?x:Nat,?z:Nat)))),plus(max(plus(min(?y:Nat,?y:Nat),s(?z:Nat)),min(?y:Nat,min(?z:Nat,?y:Nat))),plus(plus(plus(?x:Nat,?y:Nat),s(?z:Nat)),min(plus(?x:Nat,?z:Nat),min(?x:Nat,?x:Nat))))) = plus(plus(?x:Nat,plus(?z:Nat,s(s(?y:Nat)))),min(min(max(min(?y:Nat,0),plus(?x:Nat,?z:Nat)),min(max(?y:Nat,?y:Nat),s(?x:Nat))),s(0)))",
"plus(0,min(max(max(?y:Nat,plus(?x:Nat,?y:Nat)),plus(s(0),s(?x:Nat))),plus(max(?x:Nat,?x:Nat),max(?x:Nat,plus(?x:Nat,?x:Nat))))) = min(min(plus(max(s(0),max(?x:Nat,?z:Nat)),?z:Nat),plus(plus(s(?y:Nat),max(?y:Nat,0)),plus(?x:Nat,?x:Nat))),max(min(plus(min(?x:Nat,0),plus(0,?y:Nat)),plus(plus(?y:Nat,?x:Nat),?z:Nat)),plus(s(plus(0,?z:Nat)),plus(min(?x:Nat,?y:Nat),max(?x:Nat,?x:Nat)))))",
"min(min(max(?x:Nat,plus(s(?x:Nat),min(?z:Nat,?x:Nat))),max(max(0,s(?x:Nat)),s(?x:Nat))),min(max(plus(?x:Nat,plus(?y:Nat,?x:Nat)),plus(max(?y:Nat,?x:Nat),plus(?y:Nat,?x:Nat))),max(s(max(0,?x:Nat)),min(max(0,?x:Nat),?y:Nat)))) = plus(plus(?x:Nat,plus(plus(min(?z:Nat,?x:Nat),min(?x:Nat,?z:Nat)),plus(s(?z:Nat),max(?x:Nat,?z:Nat)))),min(plus(plus(s(?z:Nat),?y:Nat),0),plus(?x:Nat,plus(min(?z:Nat,0),0))))"]


val data = [
    (2, true, problemsPlusWithoutZeroYes, "Plus-yes"),
    (2, false, problemsPlusWithoutZeroNo, "Plus-no"),
    (12, true, problemsPlusWithZeroYes, "PlusZero-yes"),
    (12, false, problemsPlusWithZeroNo, "PlusZero-no"),
    (1,true, problemsTimesWithoutZeroYes, "Times-yes"),
    (1,false, problemsTimesWithoutZeroNo, "Times-no"),
    (11,true, problemsTimesWithZeroYes, "TimesZero-yes"),
    (11,false, problemsTimesWithZeroNo, "TimesZero-no"),
    (3,true, problemsListLenYes, "List-yes"),
    (3,false, problemsListLenNo, "List-no"),
    (7,true, problemsMaxMinPlusYes, "MaxMinPlus-yes"),
    (7,false, problemsMaxMinPlusNo, "MaxMinPlus-no"),
    (4,true, problemsMaxMinYes, "MaxMin-yes"),
    (4,false, problemsMaxMinNo, "MaxMin-no"),
    (5,true, problemsMaxYes, "Max-yes"),
    (5,false, problemsMaxNo, "Max-no"),
    (6,true, problemsMinYes, "Min-yes"),
    (6,false, problemsMinNo, "Min-no")
]

fun mainSub num (u,v) = 
    let val n = num mod 10
	val s = case Term.attachSortToTerm (checkDecls n) u of 
		    NONE => (print ("not well-sorted term: " ^ Term.toString u); raise DecError "check")
		  | SOME u' => u'
	val t = case Term.attachSortToTerm (checkDecls n) v of 
		    NONE => (print ("not well-sorted term: " ^ Term.toString v); raise DecError "check")
		  | SOME v' => v'
	val _ = print ((Int.toString (!count)) ^ ". ");
	val _ = print ("Conjecture:" ^ (Term.toString s) ^ " = " ^ (Term.toString t) ^ "\n")
	(* val result = runInd n (s,t)  *)
	val result = runInd 7 (s,t)
	val _ = print ("answer: " ^ (if result then "yes\n" else "no\n"))
	val _ = count:= (!count) + 1;
    in result end


fun makeFiles trans (_,_,eqs,name) =
    let	val inTmpFile = OS.Path.joinDirFile {dir=".",file= "spike" ^ name ^ ".txt"}
	val outs = TextIO.openOut inTmpFile
	val _ = count := 1;
	val _ =  L.app (fn (l,r) => 
			   (TextIO.output (outs, "\n./run.sh " ^ (Int.toString (!count))
					     ^ " \""  ^ (trans l) ^ " = " ^ (trans r) ^ "\"");
			    count := (!count) +1))  eqs
	val _ = TextIO.output (outs, "\n")
	val _ = TextIO.flushOut outs
	val _ = TextIO.closeOut outs
    in ()
    end

(* to make a shell script file for runnig other tools *)
(* val _ = L.app (makeFiles toStringSpikeOrg) data  *)

val data2 = [
    (2, true, problemsPlusWithoutZeroYes, "Plus-yes"),
    (2, false, problemsPlusWithoutZeroNo, "Plus-no"),
    (12, true, problemsPlusWithZeroYes, "PlusZero-yes"),
    (12, false, problemsPlusWithZeroNo, "PlusZero-no"),
    (4,true, problemsMaxMinYes, "MaxMin-yes"),
    (4,false, problemsMaxMinNo, "MaxMin-no"),
    (5,true, problemsMaxYes, "Max-yes"),
    (5,false, problemsMaxNo, "Max-no"),
    (6,true, problemsMinYes, "Min-yes"),
    (6,false, problemsMinNo, "Min-no")
]

(* val _ = L.app (fn (num,yesno,conjs,name)=> *)
(* 		  let val _ = print ("File: " ^ name ^ "\n") *)
(* 		      val timer = Timer.startRealTimer () *)
(* 		      val start = Timer.checkRealTimer timer *)
(* 		      val _ = count:= 1 *)
(* 		      val result = L.map (mainSub num) conjs *)
(* 		      val stop = Timer.checkRealTimer timer *)
(* 		      fun reportTime (t1,t2) = Int.toString (LargeInt.toInt  *)
(* 								 ((Time.toMilliseconds t2) -  *)
(* 								  (Time.toMilliseconds t1))) *)
(* 		      val _ = print (" ("  ^ (reportTime (start,stop)) ^ " msec.)\n") *)
(* 		  in () end) *)
(* 	 (*     data *) *)
(* 	      data2 *)

(* val _ = (count := 1; *)
(* 	 L.app (fn (l,r) => (print ("\n./run.sh " ^ (Int.toString (!count)) *)
(* 				   ^ " \""  ^ (toStringSpike l) ^ " = " ^ (toStringSpike r) ^ "\""); *)
(* 			     count := (!count) +1)) *)
(* 	       problemsNatMaxMinPlusNo) *)


end (* of local *)
end; (* of structure *)

